(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2006
              The University of Iowa

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(** infers sorts based on a clause set

    Implements sort inference, see the paper
    New Techniques that Improve MACE-style Finite Model Finding
    by
    Koen Classen and Niklas Soerensson.
    
    For an unsorted first-order clause set,
    the sorts of each function and predicate symbol (arguments and results)
    are first assumed to be different.
    Then sorts are collapsed,
    if their occurrences in the clauses require two sorts to be identical.
    For example,
    - g(a) and g(b) requires a and b to have the same sort,
    - f(x, x) and f(a, b) requires a and b to have the same sort,
    - (a = b) requires a and b to have the same sort.
    
    Sorts can be used for static symmetry reduction ({!Finite_domain}).
*)

(** {6 Types} *)


type symbol = Symbol.symbol
type clause = Term.clause

(** the sorts of a clause set. *)
type sorts


(** {6 Functions} *)

(** infers the sorts for the clause set. *)
val infer: print:bool -> clause list -> sorts

(** [add_constant sorts constant existing]
    registers the new [constant] symbol.
    Its sort is of the result sort of the symbol [existing],
    which must already be registerd.
*)
val add_constant: sorts -> symbol -> symbol -> unit

(** returns the constants partitioned by sorts. *)
val constants_partition: sorts -> symbol list list

(** returns the biggest partition in {!Sort_inference.constants_partition}. *)
val max_constant_partition_size: sorts -> int

(** prints the sorts. *)
val print: sorts -> unit

