(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


(****************************************************************************)
(** Indexargs: Using only certain arguments of a term for indexing.
This is maintained by the Reader and Writer intermediate access
layers. *)
(****************************************************************************)

let tuplesym = Term.gensym ()

(** Extracts the portion used for indexing from the given term.
    Assumes that term is not a variable. Does not need a context
    argument, since it rearranges only the toplevel structure of the 
    input term. However the result must be used with the proper context.
*)
let key_term indexargs term =
  match indexargs with
    Opt.WholeTerm -> term
  | Opt.Argument i ->
      begin
	match term with
	  Term.Compound k ->
	    (Term.compound_args k).(i)
	| _ ->
	    raise (Err.Error "Bad indexargs (1) for Opt.Argument.")
      end
  | Opt.Arguments is ->
      begin
	match term with
	  Term.Compound k ->
	    let a = Term.compound_args k
	    and r = Array.make (Array.length is) Term.Unbound in
	    for i = 0 to (Array.length is) - 1 do
	      r.(i) <- a.(is.(i))
	    done;
	    Term.Compound (Term.make_compound tuplesym r)
	| _ -> 
	    print_string "IA: ";
	    print_int (Array.length is);print_string "   -> ";PrologPrint.print term;
	    raise (Err.Error "Bad indexargs (2) for Opt.Arguments.")
      end


(****************************************************************************)
(** Store - lowest access layer: direct term index access. We want
to use different index implementations together, e.g. in a list, so
this explicit runtime dispatch seems necessary. *)
(****************************************************************************)
      
module Store =
  struct

(*
    module HashDtIndex =
      DtIndex.Make(Fact)(DtIndex.HashTableImplementation)
                      
    module AlistDtIndex =
      DtIndex.Make(Fact)(DtIndex.AlistTableImplementation)
*)

    module HashDtIndex =
      LayeredIndex.Make(DtIndex.Make(Fact)(DtIndex.HashTableImplementation))
                      
    module AlistDtIndex =
      LayeredIndex.Make(DtIndex.Make(Fact)(DtIndex.AlistTableImplementation))

    module NoIndex1 =
      LayeredIndex.Make(NoIndex)
      
    type store =
	B1 of HashDtIndex.index
      | B2 of AlistDtIndex.index
      | B3 of NoIndex1.index

    type state =
	S1 of HashDtIndex.state
      | S2 of AlistDtIndex.state
      | S3 of NoIndex1.state
	    
    let make_store indextype =
      match indextype with
	Opt.HashDt -> B1 (HashDtIndex.make_index ())
      | Opt.AlistDt -> B2 (AlistDtIndex.make_index ())
      | Opt.NoIndex -> B3 (NoIndex1.make_index ())

    let insert store =
      match store with
	B1 store -> HashDtIndex.insert store
      | B2 store -> AlistDtIndex.insert store
      | B3 store -> NoIndex1.insert store
	    
    let remove_variant store =
      match store with
	B1 store -> HashDtIndex.remove_variant store
      | B2 store -> AlistDtIndex.remove_variant store
      | B3 store -> NoIndex1.remove_variant store

    let iter_all f store =
      match store with
	B1 store -> HashDtIndex.iter_all f store
      | B2 store -> AlistDtIndex.iter_all f store
      | B3 store -> NoIndex1.iter_all f store

    let iter_variant_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_variant_candidates f store
      | B2 store -> AlistDtIndex.iter_variant_candidates f store
      | B3 store -> NoIndex1.iter_variant_candidates f store      
      
    let iter_instance_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_instance_candidates f store
      | B2 store -> AlistDtIndex.iter_instance_candidates f store
      | B3 store -> NoIndex1.iter_instance_candidates f store
	    
    let iter_generalization_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_generalization_candidates f store
      | B2 store -> AlistDtIndex.iter_generalization_candidates f store
      | B3 store -> NoIndex1.iter_generalization_candidates f store

    let iter_unification_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_unification_candidates f store
      | B2 store -> AlistDtIndex.iter_unification_candidates f store
      | B3 store -> NoIndex1.iter_unification_candidates f store

      
    let iter_variant_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_variant_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_variant_candidates_ml f store
      | B3 store -> NoIndex1.iter_variant_candidates_ml f store      
      
    let iter_instance_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_instance_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_instance_candidates_ml f store
      | B3 store -> NoIndex1.iter_instance_candidates_ml f store
	    
    let iter_generalization_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_generalization_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_generalization_candidates_ml f store
      | B3 store -> NoIndex1.iter_generalization_candidates_ml f store

    let iter_unification_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_unification_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_unification_candidates_ml f store
      | B3 store -> NoIndex1.iter_unification_candidates_ml f store      
      
      
      
      
    let clear store =
      match store with
	B1 store -> HashDtIndex.clear store
      | B2 store -> AlistDtIndex.clear store
      | B3 store -> NoIndex1.clear store

    let size store =
      match store with
	B1 store -> HashDtIndex.size store
      | B2 store -> AlistDtIndex.size store
      | B3 store -> NoIndex1.size store

    let new_layer store =
      match store with
	B1 store -> S1 (HashDtIndex.new_layer store)
      | B2 store -> S2 (AlistDtIndex.new_layer store)
      | B3 store -> S3 (NoIndex1.new_layer store)

    let add_layer state =
      match state with
	S1 state -> S1 (HashDtIndex.add_layer state)
      | S2 state -> S2 (AlistDtIndex.add_layer state)
      | S3 state -> S3 (NoIndex1.add_layer state)

    let obtain_state store =
      match store with
	B1 store -> S1 (HashDtIndex.obtain_state store)
      | B2 store -> S2 (AlistDtIndex.obtain_state store)
      | B3 store -> S3 (NoIndex1.obtain_state store)

    let install_state state =
      match state with
	S1 state -> HashDtIndex.install_state state
      | S2 state -> AlistDtIndex.install_state state
      | S3 state -> NoIndex1.install_state state

  end

(****************************************************************************)
(* Heuristics *)
(****************************************************************************)

(* Term indices are in principle applicable for all binding patterns, however
   this works more or less good.

   An applicability pattern for an index is an array of float with the
   same length as the predicate arity. The sum of the float values is
   1.0. Each position that contributes to indexed access if bound has
   a value > 0.0. An "applicability hint" for the index and a binding
   pattern is computed by summing up the values in bound positions.

   A sum of 0.0 means the index is "not efficiently applicable", i.e.
   access works as if unindexed. 1.0 means it is "fully applicable".
   Intermediate values express that it is more or less effective for
   the given binding.
*)

type applicability_pattern = Opt.applicability_value array

let applicability_relative_value_fun indexlength position =
  let n = float (indexlength - position) in
  n ** 2.0 (* ### Just some formula such that early args are preferred. *)

let applicability_value indexlength position =
  let sum = ref 0.0 in
  for i = 0 to indexlength - 1 do
    sum := !sum +. (applicability_relative_value_fun indexlength i)
  done;
  let factor = 1.0 /. !sum in
  (* ### Must we ensure that the sum is really 1.0 ? *)
  (applicability_relative_value_fun indexlength position) *. factor

let mk_applicability_pattern arity indexargs =
   let ap = Array.make arity 0.0
   and indexlength = (Array.length indexargs) in
   for i = 0 to indexlength - 1 do
     ap.(indexargs.(i)) <- applicability_value indexlength i
   done;
   ap

let mk_array_up_to n =
  let a = Array.make n 0 in
  for i = 0 to n - 1 do
    a.(i) <- i
  done;
  a
   
let indexargs_applicability_pattern store arity indexargs =
  match store with
    Store.B1 _ | Store.B2 _ ->
      let args =
	match indexargs with
	  Opt.WholeTerm -> mk_array_up_to arity
	| Opt.Argument n -> [|n|]
	| Opt.Arguments ns -> ns in
      mk_applicability_pattern arity args
  | Store.B3 _ ->
      mk_applicability_pattern arity (mk_array_up_to arity)


(****************************************************************************)
(** Writer - write access layer above Store. *)
(****************************************************************************)

module Writer =
  struct

    type writer = {
	indexargs: Opt.indexargs;
	store: Store.store;
	predicate_arity: int
      }

    let get_store writer =
      writer.store

    let make_writer indextype indexargs predicate_arity =

      let store = Store.make_store indextype in
      {indexargs=indexargs;
       store=store;
       predicate_arity=predicate_arity}

    let indexargs writer =
      writer.indexargs

    let store writer =
      writer.store

    let predicate_arity writer =
      writer.predicate_arity

    let insert writer term entry =
      Store.insert 
	writer.store 
	(key_term writer.indexargs term)
	entry

    let remove_variant writer term =
      Store.remove_variant 
	writer.store 
	(key_term writer.indexargs term)
	
    let clear writer =
      Store.clear writer.store

    let size writer =
      Store.size writer.store
  end


(****************************************************************************)
(** Reader - read access layer above Store. The same store can be
associated with several readers, e.g. for different binding patterns.
*)
(****************************************************************************)

module Reader =
  struct

    type reader = {
	applicability_pattern: applicability_pattern;
	indexargs: Opt.indexargs;
	store: Store.store;
	predicate_arity: int
      }

     (* DEBUG *)
    let print_info reader =
      print_string "READER: ";
      match reader.indexargs with
	Opt.WholeTerm ->
	  print_string "WholeTerm"
      | _ ->
	  print_string "Selected"

    let get_indexargs reader =
      reader.indexargs

    let make_readers writer =
      let store = Writer.store writer
      and indexargs = Writer.indexargs writer
      and arity = Writer.predicate_arity writer in
      let ap =
	indexargs_applicability_pattern store arity indexargs in
      [{applicability_pattern=ap;
	indexargs=indexargs;
	store=store;
	predicate_arity=arity}]

    let null_reader_obj = {
      applicability_pattern=[||];
      indexargs=Opt.Arguments [||];
      store=Store.make_store Opt.AlistDt;
      predicate_arity=0
    }

    let null_reader () =
      null_reader_obj

    let applicability_pattern reader =
      reader.applicability_pattern

    let applicability reader binding_pattern =
      if reader.predicate_arity = 0 then
	1.0
      else
	let sum = ref 0.0 in
	for i = 0 to reader.predicate_arity - 1 do
	  match binding_pattern.(i) with
	    Opt.Bound ->
	      sum := !sum +. reader.applicability_pattern.(i)
	  | Opt.Free ->
	      ()
	done;
	!sum

    let iter_unification_candidates f reader term context =
      Store.iter_unification_candidates
	f 
	reader.store
	(key_term reader.indexargs term)
	context
	
    let iter_instance_candidates f reader term context =
      Store.iter_instance_candidates
	f
	reader.store
	(key_term reader.indexargs term)
	context
	
    let iter_variant_candidates f reader term context =
      Store.iter_variant_candidates
	f
	reader.store
	(key_term reader.indexargs term)
	context	

    let iter_generalization_candidates f reader term context =
      Store.iter_generalization_candidates
	f
	reader.store
	(key_term reader.indexargs term)
	context

	
	let iter_unification_candidates_ml f reader term context =
      Store.iter_unification_candidates_ml
	f 
	reader.store
	(key_term reader.indexargs term)
	context
	
    let iter_instance_candidates_ml f reader term context =
      Store.iter_instance_candidates_ml
	f
	reader.store
	(key_term reader.indexargs term)
	context
	
    let iter_variant_candidates_ml f reader term context =
      Store.iter_variant_candidates_ml
	f
	reader.store
	(key_term reader.indexargs term)
	context	

    let iter_generalization_candidates_ml f reader term context =
      Store.iter_generalization_candidates_ml
	f
	reader.store
	(key_term reader.indexargs term)
	context
	
	
	
    let iter_all f reader =
      Store.iter_all
	f
	reader.store

	
	

  end















