(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(****************************************************************************)
(* Deterministic Builtins - they may fail or succeed once. *)
(****************************************************************************)

open Term

let dont_gc_these_symbols = ref []

let register_builtin symbol_table symbol_name arity f =
  let symbol = find_symbol symbol_table symbol_name arity in
  match symbol_predicate symbol with
    PUnbound ->
      set_symbol_predicate symbol f;
      dont_gc_these_symbols := symbol :: !dont_gc_these_symbols
  | _ -> raise (Failure ("Trying to reassign predicate symbol: "^
			 symbol_name^"/"^(string_of_int arity)^"."))

let clear_builtins symbol_table =
  symbol_table_iter
    (fun symbol ->
      set_symbol_predicate symbol PUnbound)
    symbol_table;
  dont_gc_these_symbols := []


(****************************************************************************)
(* ISO-Prolog Arithmetic Builtins *)
(****************************************************************************)

let builtin_is term context foreign_context trail =
  (* foreign_context is not actually used here,
     perhaps we could use match_term. *)
  match term with
    Compound k ->
      let args = compound_args k in
      unify
	args.(0)
	context
	(Evaluable.eval args.(1) context)
	context
	trail
  | _ ->
      raise (Failure "Bad call to builtin is/2.")

let arith_equal_terms x y =
  match x, y with
    Integer x, Integer y -> x = y
  | Integer x, Float y -> float(x) = y
  | Float x, Integer y -> x = float(y)
  | Float x, Float y -> x = y
  | _ -> raise (Failure "Bad call to arithmetic builtin.")
	
let arith_not_equal_terms x y =
  match x, y with
    Integer x, Integer y -> x <> y
  | Integer x, Float y -> float(x) <> y
  | Float x, Integer y -> x <> float(y)
  | Float x, Float y -> x <> y
  | _ -> raise (Failure "Bad call to arithmetic builtin.")
	
let arith_lessq_terms x y =
  match x, y with
    Integer x, Integer y -> x <= y
  | Integer x, Float y -> float(x) <= y
  | Float x, Integer y -> x <= float(y)
  | Float x, Float y -> x <= y
  | _ -> raise (Failure "Bad call to arithmetic builtin.")

let arith_less_terms x y =
  match x, y with
    Integer x, Integer y -> x < y
  | Integer x, Float y -> float(x) < y
  | Float x, Integer y -> x < float(y)
  | Float x, Float y -> x < y
  | _ -> raise (Failure "Bad call to arithmetic builtin.")

let builtin_arith_equal term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_equal_terms
	  (Evaluable.eval args.(0) context)
	  (Evaluable.eval args.(1) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin =:=/2.")

let builtin_arith_not_equal term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_not_equal_terms
	  (Evaluable.eval args.(0) context)
	  (Evaluable.eval args.(1) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin =\\=/2.")
	
let builtin_arith_lessq term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_lessq_terms
	  (Evaluable.eval args.(0) context)
	  (Evaluable.eval args.(1) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin =</2.")

let builtin_arith_less term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_less_terms
	  (Evaluable.eval args.(0) context)
	  (Evaluable.eval args.(1) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin </2.")

let builtin_arith_greaterq term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_lessq_terms
	  (Evaluable.eval args.(1) context)
	  (Evaluable.eval args.(0) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin >=/2.")

let builtin_arith_greater term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      if arith_less_terms
	  (Evaluable.eval args.(1) context)
	  (Evaluable.eval args.(0) context)
      then
	trail
      else
	fail
  | _ ->
      raise (Failure "Bad call to builtin >/2.")


(****************************************************************************)
(* Ordset Builtins. *)
(****************************************************************************)

module StdTermOrd = Ordset.MakeAux(
  struct
    type t = Term.term
    type aux = Term.context
    let compare = 
      (fun context t1 t2 -> StdOrder.term_compare t1 context t2 context)
  end)

let term_list_to_list term context =
  Convert.term_list_to_list (apply term context)
      
(*
* The idea here is to dereference just on the list level (not
* into the list elements) but this is bugged, an example that does
* not work:
*
* p(s(A), [A]).
* sum(Z) :-
* 	p(s(b), X),     % also buggy with: p(s(a), X)
* 	ord_union(X, [b], Z).
* 
* let rec term_list_to_list term context =
*    let (term, context1) = dereference term context in
*    match term with
*      Compound k ->
*        let a = compound_args k in
*        if Array.length a <> 2 then
*  	raise (Err.Error "List expected.")
*        else
*  	a.(0) :: term_list_to_list a.(1) context1
*    | _ -> []
*)

let make_builtin_sort symbol_table =
  let dot = (find_symbol symbol_table "." 2)
  and nil = (find_symbol symbol_table "[]" 0) in
  ( fun term context foreign_context trail ->
    match term with
      Compound k ->
	let args = compound_args k in
	let input_list = term_list_to_list args.(0) context in
	let output_list =
	  Convert.list_of_term_to_term_1
	    dot nil
	    (StdTermOrd.list_to_ord_set context input_list) in
	unify
	  args.(1)
	  context
	  output_list
	  context
	  trail 
    | _ ->
	raise (Failure "Bad call to builtin sort/2."))

let make_builtin_ord_insert symbol_table =
  let dot = (find_symbol symbol_table "." 2)
  and nil = (find_symbol symbol_table "[]" 0) in
  ( fun term context foreign_context trail ->
    match term with
      Compound k ->
	let args = compound_args k in
	let input_list = term_list_to_list args.(0) context
	and input_elem = args.(1) in
	let output_list =
	  Convert.list_of_term_to_term_1
	    dot nil
	    (StdTermOrd.ord_insert context input_list input_elem) in
	unify
	  args.(2)
	  context
	  output_list
	  context
	  trail 
    | _ ->
	raise (Failure "Bad call to builtin ord_insert/3."))
	  
let make_builtin_binary_ord_op setop symbol_table =
  let dot = (find_symbol symbol_table "." 2)
  and nil = (find_symbol symbol_table "[]" 0) in
  ( fun term context foreign_context trail ->

    match term with
      Compound k ->
	let args = compound_args k in
	let input_list_1 = term_list_to_list args.(0) context
	and input_list_2 = term_list_to_list args.(1) context in
	let output_list =
	  Convert.list_of_term_to_term_1
	    dot nil
	    (setop context input_list_1 input_list_2) in
	unify
	  args.(2)
	  context
	  output_list
	  context
	  trail 
    | _ ->
	raise (Failure "Bad call to binary set operation builtin."))

let make_builtin_binary_ord_test test =
  ( fun term context foreign_context trail ->
    match term with
      Compound k ->
	let args = compound_args k in
	let input_list_1 = term_list_to_list args.(0) context
	and input_list_2 = term_list_to_list args.(1) context in
	if (test context input_list_1 input_list_2) then
	  trail
	else
	  fail
    | _ ->
	raise (Failure "Bad call to binary set test builtin."))


(****************************************************************************)
(* Gensym Builtin *)

let builtin_gensym term context foreign_context trail =
  match term with
    Compound k ->
      let args = compound_args k in
      unify
	args.(0)
	context
	(Term.Constant (Term.gensym ()))
	context
	trail
  | _ ->
      raise (Failure "Bad call to builtin gensym/1.")

(****************************************************************************)
(* List Stuff *)
(****************************************************************************)

(****************************************************************************)
(* Append *)

(* The 2nd argument to append actually is unconstrained - it 
 * may be e.g. be a variable. The 1st argument must be a ground list.
 * (It must be fully ground, i.e. may not contain any variables 
 * within list elements).
 *)

let term_list_to_list term context =
  Convert.term_list_to_list (apply term context)

let rec term_list_append dot terms tlist2  =
  match terms with
    [] ->
      tlist2
  | term :: terms1 ->
      Compound 
	(make_compound
	   dot
	   [|term; term_list_append dot terms1 tlist2|])

let make_builtin_append symbol_table =
  let dot = (find_symbol symbol_table "." 2)
(*  and nil = (find_symbol symbol_table "[]" 0)*) in
  ( fun term context foreign_context trail ->
    match term with
      Compound k ->
	let args = compound_args k in
	let input_list_1 = term_list_to_list args.(0) context in
	let output_list = term_list_append dot input_list_1 args.(1) in
	unify
	  args.(2)
	  context
	  output_list
	  context
	  trail 
    | _ ->
	raise (Failure "Bad call to builtin append."))


(****************************************************************************)

let init_builtins symbol_table =
  register_builtin symbol_table "is" 2
    (PDeterministic
       (builtin_is,
	([|Opt.OutputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "=:=" 2 
    (PDeterministic
       (builtin_arith_equal,
	([|Opt.InputGround; Opt.InputGround|],
         [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "=\\=" 2 
    (PDeterministic 
       (builtin_arith_not_equal,
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "=<" 2
    (PDeterministic 
       (builtin_arith_lessq,
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "<" 2
    (PDeterministic
       (builtin_arith_less,
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table ">=" 2
    (PDeterministic
       (builtin_arith_greaterq,
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table ">" 2
    (PDeterministic
       (builtin_arith_greater,
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "sort" 2
    (PDeterministic
       ((make_builtin_sort symbol_table),
	([|Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_union" 3
    (PDeterministic
       ((make_builtin_binary_ord_op StdTermOrd.ord_union symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_intersection" 3
    (PDeterministic
       ((make_builtin_binary_ord_op StdTermOrd.ord_intersection symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_subtract" 3
    (PDeterministic
       ((make_builtin_binary_ord_op StdTermOrd.ord_subtract symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_symdiff" 3
    (PDeterministic
       ((make_builtin_binary_ord_op StdTermOrd.ord_symdiff symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_subset" 2
    (PDeterministic
       ((make_builtin_binary_ord_test StdTermOrd.ord_subset),
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_disjoint" 2
    (PDeterministic
       ((make_builtin_binary_ord_test StdTermOrd.ord_disjoint),
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_intersect" 2
    (PDeterministic
       ((make_builtin_binary_ord_test StdTermOrd.ord_intersect),
	([|Opt.InputGround; Opt.InputGround|],
	 [|1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "ord_insert" 3
    (PDeterministic
       ((make_builtin_ord_insert symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)));
  register_builtin symbol_table "append" 3
    (PDeterministic
       ((make_builtin_append symbol_table),
	([|Opt.InputGround; Opt.InputGround; Opt.OutputGround|],
	 [|1.0;1.0;1.0|],
	 Opt.BuiltinProof)))


(* genym(-) binds the argument to a fresh uninterned symbol.
 * - The implementation should work but commented out since it may only be useful with additional
 *   stuff such as special subsumption
 *
 * register_builtin symbol_table "gensym" 1
 *   (PDeterministic
 *      (builtin_gensym,
 *	([|Opt.OutputGround|],
 *	 [|1.0|],
 *	 Opt.BuiltinProof)))
 *)

