(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

open Term

exception Convert_error of string

let term_to_string term =
  match term with
    Constant symbol ->
      symbol_name symbol
  | _ -> raise (Convert_error "Atom expected.")

let rec map_term f term =
  match term with
    Constant _ ->
      []
  | Compound l ->
      if (symbol_arity (compound_functor l)) <> 2 then
	raise (Convert_error "List expected.")
      else
	(f (compound_args l).(0)) :: map_term f (compound_args l).(1)
  | _ -> raise (Convert_error "List expected.")
      
let term_to_alist term =
  map_term
    (fun e ->
      match e with
	Compound kv ->
	  if (symbol_arity (compound_functor kv)) <> 2 then
	    raise (Convert_error "Key=Value pair expected.")
	  else
	    (term_to_string (compound_args kv).(0), (compound_args kv).(1))
      | _ -> raise (Convert_error "Key=Value pair expected."))
    term

let term_to_int term =
  match term with
    Integer n ->
      n
  | _ -> raise (Convert_error "Integer expected.")

let term_to_float term =
  match term with
    Float n ->
      n
  | _ -> raise (Convert_error "Float expected.")

let term_to_array_of_int term =
  Array.of_list (map_term term_to_int term)

let term_to_array_of_float term =
  Array.of_list (map_term term_to_float term)

let term_nth_arg term n =
  match term with
    Compound k ->
      if (symbol_arity (compound_functor k)) <= n then
	raise (Convert_error "Compound with larger arity expected.")
      else
	(compound_args k).(n)
  | _ -> raise (Convert_error "Compound expected.")

let term_pair term =
  match term with
    Compound k ->
      if (symbol_arity (compound_functor k)) < 2 then
	raise (Convert_error "Compound with arity >= 2 expected.")
      else
	((compound_args k).(0), (compound_args k).(1))
  | _ -> raise (Convert_error "Compound with arity >= 2 expected.")
	
let symbol_to_string symbol =
  Printf.sprintf "%s/%d"
    (PrologPrint.print_to_string (Constant symbol))
    (symbol_arity symbol)

let term_functor_symbol term = 
  match term with
    Compound k ->
      compound_functor k
  | Constant s ->
      s
  | _ -> (raise (Failure "Bad argument to term_functor_symbol."))

let term_functor_pair term =
  let sym = term_functor_symbol term in
  (symbol_name sym, symbol_arity sym)
    
let predspec_pair term =
  let (name, arity) = term_pair term in
  (term_to_string name, term_to_int arity)
   
let rec list_of_int_to_term st ns =
  match ns with
    [] -> 
      Constant (find_symbol st "[]" 0)
  | n :: ns1 ->
      Compound 
	(make_compound
	   (find_symbol st "." 2)
	   [|Integer n; list_of_int_to_term st ns1|])

let array_of_int_to_term st ns =
  list_of_int_to_term st (Array.to_list ns)

let rec list_of_float_to_term st ns =
  match ns with
    [] -> 
      Constant (find_symbol st "[]" 0)
  | n :: ns1 ->
      Compound 
	(make_compound
	   (find_symbol st "." 2)
	   [|Float n; list_of_float_to_term st ns1|])

let array_of_float_to_term st ns =
  list_of_float_to_term st (Array.to_list ns)

let rec alist_to_term st pairs =
  match pairs with
    [] -> 
      Constant (find_symbol st "[]" 0)
  | (k,v) :: pairs1 ->
      Compound 
	(make_compound
	   (find_symbol st "." 2)
	   [|Compound
	       (make_compound
		  (find_symbol st "=" 2)
		  [|Constant (find_symbol st k 0); v|]);
	     alist_to_term st pairs1|])

let symbol_to_pred_spec_term st symbol =
  Compound
    (make_compound
       (find_symbol st "/" 2)
       [|Constant (find_symbol st (symbol_name symbol) 0);
	 Integer (symbol_arity symbol)|])

let rec term_sequence_to_list symbol term =
  match term with
    Compound k ->
      if (compound_functor k) == symbol then
	(compound_args k).(0) ::
	term_sequence_to_list symbol (compound_args k).(1)
      else
	[term]
  | _ -> [term]

let array_to_compound symbol_table functor_name args =
  Compound 
    (make_compound
       (find_symbol symbol_table functor_name (Array.length args))
       (Array.copy args))

let int_to_term n =
  Integer n

let float_to_term n =
  Float n

let string_to_term symbol_table string =
  Constant (find_symbol symbol_table string 0)

let rec array_of_conj_term_to_term st terms =
	array_of_term_to_term_1 
   		(find_symbol st "," 2)
   		(find_symbol st "[]" 0) 
   		terms 
   		0
		
and array_of_term_to_term_1 dot nil terms i =
	if i = Array.length terms then
		(Constant nil)
	else
		(if i = ((Array.length terms)-1) then
			terms.(i)
		else
			(Compound
				(make_compound dot [|terms.(i); array_of_term_to_term_1 dot nil terms (i+1)|])) 
		)
		
let array_of_disj_term_to_term st terms =
	array_of_term_to_term_1 
   		(find_symbol st ";" 2)
   		(find_symbol st "[]" 0) 
   		terms 
   		0
						  
let rec list_of_term_to_term st terms =
  list_of_term_to_term_1
    (find_symbol st "." 2)
    (find_symbol st "[]" 0)
    terms

and list_of_term_to_term_1 dot nil terms =
  match terms with
    [] -> 
      Constant nil
  | term :: terms1 ->
      Compound 
	(make_compound
	   dot
	   [|term; list_of_term_to_term_1 dot nil terms1|])

let rec term_list_to_list term =
  match term with
    Compound k ->
      let a = compound_args k in
      if Array.length a <> 2 then
	raise (Convert_error "List expected.")
      else
	a.(0) :: term_list_to_list a.(1)
  | Constant _ ->
      []
  | _ ->
      raise (Convert_error "List expected.")
	
let term_is_list term =
  match term with
    Compound k ->
      let s = compound_functor k in
      symbol_name s = "." && symbol_arity s = 2
  | Constant s ->
      symbol_name s = "[]"
  | _ -> false


