(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Convenience functions for accessing terms with OCaml. *)

(** Raised if the input term is not suited for application of a
conversion function. *)
exception Convert_error of string

(** {6 Converting from Terms to OCaml} *)

(** Input must be an atom term. Output is its symbol name*)
val term_to_string : Term.term -> string

(** Input must be a list term. The function is applied to all members and
a list of result is returned. *)
val map_term : (Term.term -> 'a) -> Term.term -> 'a list


(** Input must be a list term of Key=Value pairs. Output is an association
list, where the keys are strings obtained by [term_to_string] from
the original keys, and the values are the original terms. *)
val term_to_alist : Term.term -> (string * Term.term) list

(** Returns true if the input term has ./2 or []/0 as main functor. *)
val term_is_list : Term.term -> bool

(** Input must be a proper list, i.e. each tail must be a list.
   The result is a list of the element terms. *)
val term_list_to_list : Term.term -> Term.term list

(** Input must be a sequence constructed from a binary functor, such
    that the first argument is the element and the second argument the
    rest sequence or the last element (which must have a different main
    functor). [symbol] specifies the binary functor. The output is a list
    of the elements.
*) 
val term_sequence_to_list : Term.symbol -> Term.term -> Term.term list

(** Input must be an integer term. Output is an integer. *)
val term_to_int : Term.term -> int

(** Input must be a float term. Output is a float. *)
val term_to_float : Term.term -> float

(** Input must be a list of integer terms. Output is an array of integers. *)
val term_to_array_of_int : Term.term -> int array

(** Input must be a list of float terms. Output is an array of floats. *)
val term_to_array_of_float : Term.term -> float array

(** {6 Functions for Accessing Terms in Special Ways} *)

(** Return the nth argument of a compound term. *)
val term_nth_arg : Term.term -> int -> Term.term

(** Return a pair of the first and second argument of a compound term. *)
val term_pair : Term.term -> (Term.term * Term.term)

(** Returns a string represenation "symbol-name/arity" of the given symbol. *)
val symbol_to_string : Term.symbol -> string

(** Input must be a constant or compound. Result is its functor symbol. *)
val term_functor_symbol : Term.term -> Term.symbol

(** Input must be a constant or compound. Result is its functor as
[(name, arity)] pair.
*)
val term_functor_pair : Term.term -> (string * int)

(** Input must be a term of the form [functor/arity]. Result is the
corresponding [(name, arity)] pair. *)
val predspec_pair : Term.term ->
  (string * int)

(** {6 Converting from OCaml to Terms} *)

val list_of_term_to_term : Term.symbol_table -> Term.term list -> Term.term

(* Like list_of_term_to_term but has the list constructors './2' and '[]/0'
    as arguments. *)
val list_of_term_to_term_1 : 
    Term.symbol -> Term.symbol -> Term.term list -> Term.term
    
val array_of_conj_term_to_term : Term.symbol_table -> Term.term array -> Term.term
val array_of_disj_term_to_term : Term.symbol_table -> Term.term array -> Term.term


val array_of_int_to_term : Term.symbol_table -> int array -> Term.term

val list_of_int_to_term : Term.symbol_table -> int list -> Term.term

val array_of_float_to_term : Term.symbol_table -> float array -> Term.term

val list_of_float_to_term : Term.symbol_table -> float list -> Term.term

val alist_to_term : Term.symbol_table -> (string * Term.term) list -> Term.term

val symbol_to_pred_spec_term : Term.symbol_table -> Term.symbol -> Term.term

val array_to_compound : Term.symbol_table -> string -> Term.term array ->
  Term.term

val string_to_term : Term.symbol_table -> string -> Term.term

val int_to_term : int -> Term.term

val float_to_term : float -> Term.term


