(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(****************************************************************************)
(****************************************************************************)
(** Discrimination Tree Indexing. *)
(****************************************************************************)
(****************************************************************************)

(* See e.g.: "Experiments with Discrimination-Tree Indexing and Path
   Indexing for Term Retrieval" by William McCune, 1990.
*)


(****************************************************************************)
(* Term Functor (i.e. Index Key) Representation *) 
(****************************************************************************)

(* - Pervasives.compare could be taken as comparison function, provided
     no special ordering is required.
 *)

type hkey =
    Variable |
    Symbol of Term.symbol |
    Integer of int |
    Float of float

let print_hkey k =
	match k with
		Variable -> print_string "X"
		| Symbol s -> print_string (Term.symbol_name s);print_string "/";print_int (Term.symbol_arity s)
		| Integer i -> print_int i
		| Float f -> print_float f  
		
let print_fterm fterm =
	List.iter
		(fun k -> print_hkey k;print_string " ")
		fterm   		  
    
let hkey_equal k1 k2 =
  match k1, k2 with
    Variable, Variable ->
      true
  | Symbol s1, Symbol s2 ->
      s1 == s2
  | Integer n1, Integer n2 ->
      n1 = n2
  | Float n1, Float n2 ->
      n1 = n2
  | _, _ ->
      false

let term_hkey term =
  match term with
    Term.Compound k ->
      Symbol (Term.compound_functor k)
  | Term.Constant s ->
      Symbol s
  | Term.Variable _ ->
      Variable
  | Term.Integer n ->
      Integer n
  | Term.Float n ->
      Float n
  | Term.Unbound ->
      raise (Failure "Bad term used for indexing.")
	
let hkey_arity hkey =
  match hkey with
    Symbol s ->
      Term.symbol_arity s
  | _ -> 0


(****************************************************************************)
(* TableType *)
(****************************************************************************)

module type TableType =
  sig
    type key = hkey
    type 'a t
    val make_table: unit -> 'a t
    val add: 'a t -> key -> 'a -> unit
    val remove: 'a t -> key -> unit
    val find: 'a t -> key -> 'a
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    val iter_and_throw_away: (key -> 'a -> unit) -> 'a t -> unit
  end

module Make (Entry: Index.EntryType) (Table: TableType) =
  struct

(****************************************************************************)
(* Index Data Structure *)
(****************************************************************************)

    type entry = Entry.entry

    type entries = {mutable list: entry list}

    type node =
	DtInnerNode of table
      | DtLeafNode of entries
	    
    and table =
	node Table.t
	  
    type index =
	{mutable root:node;
	 mutable size:int}
	  
    let index_root index = index.root
	
    let make_index () =
      {root=DtInnerNode (Table.make_table ()); size=0}

    let clear index =
      index.root <- DtInnerNode (Table.make_table ());
      index.size <- 0

    let size index =
      index.size

(****************************************************************************)
(* Insertion *)
(****************************************************************************)
	
    let rec insert index term entry = 
      match index_root index with
	DtInnerNode table ->
	  ignore (insert_1 entry term table 1);
	  index.size <- index.size + 1
      | _ ->
	  raise (Failure "Bad index toplevel type.")
	    
    and insert_1 entry term table n = 
      (* n is a counter of the terms to be processed. It is used to
	 find out when we are at the last subterm in the preorder. *)
      let hkey = term_hkey term in
      let hkey_arity = (hkey_arity hkey) in
      let n1 = n + hkey_arity - 1 in
      let node =
	try 
	  Table.find table hkey
	with Not_found ->
	  let node =
	    if n1 = 0 then (* At the last subterm in the preorder. *)
	      DtLeafNode {list=[]}
	    else
	      DtInnerNode (Table.make_table ())
	  in
	  Table.add table hkey node; (* We could also use replace here. *)
	  node
      in
      match n1, node with
	0, DtLeafNode entries ->
	  entries.list <- entry :: entries.list;
	  table, n1 (* Return the table just to stay well typed. *)
      | _, DtInnerNode table1 ->
	  ( match term with
	    Term.Compound k ->
	      insert_1_args
		entry
		(Term.compound_args k)
		hkey_arity
		0
		table1
		n1
	  | _ -> table1, n1 )
      | _ , _ ->
	  raise (Failure "Bad index structure (insert).")
	    
    and insert_1_args entry args arity i table n =
      if i >= arity then
	table, n
      else
	let table1, n1 = insert_1 entry args.(i) table n in
	insert_1_args entry args arity (i+1) table1 n1
	  
	  
(****************************************************************************)
(* Retrieve Whole Content *)
(****************************************************************************)
	  
    let rec iter_all f index =
      if index.size > 0 then
	iter_all_1 f (index_root index)
	
    and iter_all_1 f node =
      match node with
	DtInnerNode table ->
	  Table.iter (fun k v -> iter_all_1 f v) table
      | DtLeafNode entries ->
	  List.iter f entries.list
	    

(****************************************************************************)
(* Retrieve Whole Content And Throw Away While Iterating *)
(****************************************************************************)
	    
    let rec iter_all_and_throw_away f index =
      if index.size > 0 then
	iter_all_and_throw_away_1 f (index_root index)
	
    and iter_all_and_throw_away_1 f node =
      match node with
	DtInnerNode table ->
	  Table.iter_and_throw_away 
	    (fun k v -> iter_all_and_throw_away_1 f v) table
      | DtLeafNode entries ->
	  entries_iter_and_throw_away f entries
	    
    and entries_iter_and_throw_away f entries =
      match entries.list with
	[] -> ()
      | a::l ->
	  entries.list <- l;
	  f a;
	  entries_iter_and_throw_away f entries


(****************************************************************************)
(* Auxiliary Function: Flatten Term *)
(****************************************************************************)
	    
(* - A flattened term is just a list of hkeys in preorder. Skipping is
   done by using hkey_arity.
   
   - This implementation of flatten_term is not tail recursive. It could
   be if pointers were used.
   
   - A flattened term is created for a query term before lookup
   in the access functions. This simplifies the lookup functions.
   
   - The lookup functions work without backtracking and without allocation
   of extra data.
 *)

  
    let rec flatten_term term context =
      flatten_term_1 context term []
	
    and flatten_term_1 context term rest =
      let (term, context) = Term.dereference term context in
      let hkey = term_hkey term in
      match term with
	Term.Compound k ->
	  hkey :: flatten_term_1_args
		   context
		   rest
		   (Term.compound_args k)
		   (hkey_arity hkey)
		   0
      | _ -> hkey :: rest
		      
    and flatten_term_1_args context rest args arity i =
      if i >= arity then
	rest
      else
	flatten_term_1
	  context
	  args.(i)
	  (flatten_term_1_args
	     context
	     rest
	     args
	     arity
	     (i + 1))
	  
(* Skip some terms in fterm. count is the number of terms to be skipped. *)

    let rec skip_terms_fterm fterm count =
      if count = 0 then
	fterm
      else
	match fterm with 
	  hkey::fterm1 ->
	    skip_terms_fterm fterm1 (count + (hkey_arity hkey) - 1)
	| _ ->
	    raise (Failure "Bad flattened term structure.")
	      
	      
(****************************************************************************)
(* Retrieve Instance Candidates *)
(****************************************************************************)
	      
    let rec iter_instance_candidates f index term context =
      if index.size > 0 then
	iter_inst f (flatten_term term context) (index_root index)
	
    and iter_inst f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    match hkey with
	      Variable ->
		iter_inst_skip f fterm1 node 1
	    | _ ->
		try
		  let node1 = Table.find table hkey in
		  iter_inst f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (instance).")
	    
    and iter_inst_skip f fterm node count =
      if count = 0 then
	iter_inst f fterm node
      else
	match node with
	  DtInnerNode table ->
	    Table.iter
	      (fun hkey node1 ->
		iter_inst_skip f fterm node1 (count + (hkey_arity hkey) - 1))
	      table
	| _ -> raise (Failure "Bad index structure (skip/inst).")
	
   
	      
    let rec iter_variant_candidates f index term context =
      if index.size > 0 then
	iter_variant f (flatten_term term context) (index_root index)
	
    and iter_variant f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
		try
		  let node1 = Table.find table hkey in
		  iter_variant f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (variant).")	
	
(****************************************************************************)
(* Retrieve Generalization Candidates *)
(****************************************************************************)
	      
    let rec iter_generalization_candidates f index term context =
      if index.size > 0 then
	iter_gen f (flatten_term term context) (index_root index)
	
    and iter_gen f fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    begin
	      try
		let node1 = Table.find table Variable in
		iter_gen 
		  f
		  (skip_terms_fterm fterm 1)
		  node1
	      with Not_found ->
		()
	    end;
	    match hkey with
	      Variable ->
		()
	    | _ ->
		try 
		  let node1 = Table.find table hkey in
		  iter_gen f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (generalization).")
	    
	    
(****************************************************************************)
(* Retrieve Unification Candidates *)
(****************************************************************************)
	    
    let rec iter_unification_candidates f index term context =
      if index.size > 0 then
	iter_unif f (flatten_term term context) (index_root index)
	
    and iter_unif f fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    match hkey with
	      Variable ->
		iter_unif_skip f fterm1 node 1
	    | _ ->
		begin
		  try 
		    let node1 = Table.find table Variable in
		    iter_unif
		      f
		      (skip_terms_fterm fterm 1)
		      node1
		  with Not_found ->
		    ()
		end;
		begin
		  try 
		    let node1 = Table.find table hkey in
		    iter_unif f fterm1 node1
		  with Not_found ->
		    ()
		end
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (unif).")	
	    
    and iter_unif_skip f fterm node count =
      if count = 0 then
	iter_unif f fterm node
      else
	match node with
	  DtInnerNode table ->
	    Table.iter
	      (fun hkey node1 ->
		iter_unif_skip f fterm node1 (count + (hkey_arity hkey) - 1))
	      table
	| _ -> raise (Failure "Bad index structure (skip/unif).")
	      
	      
(****************************************************************************)
(* Remove Variant *)
(****************************************************************************)
	      
(* - remove_instances seems more useful, but difficult to implement
   with tables, that do not have simultaneous iteration and removing
   operations.
 *)
	      
    exception Found
	
    let is_empty_table table =
      try
	Table.iter 
	  (fun k v -> raise Found) table;
	true
      with Found ->
	false
	  
    let rec remove_variant index term =
      ignore
	(rm_variant term 
	   (flatten_term term Term.null_context)
	   (index_root index));
      index.size <- index.size - 1
	
    and rm_variant term fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      let node1 = Table.find table hkey in
	      if rm_variant term fterm1 node1 then
		begin
		  Table.remove table hkey;
		  is_empty_table table
		end
	      else
		false
	    with Not_found ->
	      false
	  end
      | [], DtLeafNode entries ->
	  entries.list <- list_remove_variant term entries.list;
	  entries.list = []	  
      | _, _ ->
	  raise (Failure "Bad index structure (generalization).")
	    
    and list_remove_variant term entries =
      match entries with
	[] ->
	  []
      | entry :: entries1 ->
	  if Term.variant term (Entry.get_term entry) then
	    entries1
	  else
	    entry :: (list_remove_variant term entries1)
		      
		      
(****************************************************************************)
(* Stuff for Debugging *)
(****************************************************************************)
		      
    let dbg_print_hkey hkey =
      match hkey with
	Symbol sym ->
	  print_string (Term.symbol_name sym);
	  print_string "/";
	  print_int (Term.symbol_arity sym)
      | Variable ->
	  print_string "*/0"
      | _ ->
	  print_string "?/0"
	    
    let rec indent d =
      if d <= 0 then
	()
      else
	begin
	  print_string "  ";
	  indent (d - 1)
	end
	  
    let rec dbg_print_index index =
      let node = index_root index in
      dbg_print_index_1 node 0
	
    and dbg_print_index_1 node d =
      match node with
	DtInnerNode table ->
	  Table.iter
	    (fun k v ->
	      indent d;
	      dbg_print_hkey k;
	      print_newline ();
	      dbg_print_index_1 v (d + 1))
	    table
      | DtLeafNode entries ->
	  indent d;
	  print_string "[";
	  List.iter 
	    (fun e ->
	      PrologPrint.print (Entry.get_term e);
	      print_string " ")
	    entries.list;
	  print_string "]\n"

end (* Make *)

(****************************************************************************)
(* TermEntry *)
(****************************************************************************)

(* Implementation of EntryType, that uses terms directly as entries. *)

module TermEntry =
  struct
    type entry = Term.term
    let get_term entry = entry
  end

(****************************************************************************)
(* HKeyHashtbl *)
(****************************************************************************)

(* Implementation of TableType as hash table. *)

module HKeyHashtbl =
  Hashtbl.Make(
  struct
    type t = hkey
    let equal = hkey_equal
    let hash = Hashtbl.hash
  end)

(* Hmm, Hashtbl.hash looks for "meaningful" components of its
   argument. It may be more efficient to use the object identity directly
   for symbols. Is there a function associating integers with object
   identities? *)    

module HashTableImplementation =
  struct
    type key = hkey
    type 'a t = 'a HKeyHashtbl.t
    let make_table () = HKeyHashtbl.create 8
    let add = HKeyHashtbl.add
    let remove = HKeyHashtbl.remove
    let find = HKeyHashtbl.find
    let iter = HKeyHashtbl.iter
    let iter_and_throw_away f t = 
      (* Hashtables from the standard library do not to allow removal
	 of entries while iterating. *)
      HKeyHashtbl.iter f t;
      HKeyHashtbl.clear t
  end

(****************************************************************************)
(* Alist Table *)
(****************************************************************************)

(* Implementation of TableType as association list. *)

module AlistTableImplementation =
  struct

    type key = hkey

    type 'a t =  {mutable alist: (hkey * 'a) list}

    let make_table () = {alist=[]}

    let add table k v = table.alist <- (k,v) :: table.alist

    let rec remove table k =
      table.alist <- remove_assoc k table.alist
    and remove_assoc x = function
	[] -> []
      | (a, b as pair) :: l -> 
	  if hkey_equal a x then l else pair :: remove_assoc x l

    let rec find table k =
      assoc k table.alist
    and assoc x = function
	[] -> raise Not_found
      | (a, b)::l -> if hkey_equal a x then b else assoc x l

    let rec iter f table =
      iter_1 f table.alist
    and iter_1 f = function
	[] -> ()
      | (a,b)::l -> f a b; iter_1 f l

    let rec iter_and_throw_away f table =
      match table.alist with
	[] -> ()
      | (a,b)::l ->
	  table.alist <- l;
	  iter_and_throw_away f table
  end

(****************************************************************************)
(* Exported Combinations *)
(****************************************************************************)

module TermHashIndex =
  Make(TermEntry)(HashTableImplementation)

module TermAlistIndex =
  Make(TermEntry)(AlistTableImplementation)

