(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Discrimination tree indexing. *)

(** {6 Predefined Index Implementations} *)

(** A predefined discrimination tree implementation that uses terms as
   entries and hashtables for the internal tree representation.
 *)
module TermHashIndex : (Index.Index with type entry=Term.term)

(** A predefined discrimination tree implementation that uses terms as
   entries and association lists for the internal tree representation.
 *)
module TermAlistIndex : (Index.Index with type entry=Term.term)


(** {6 Constructing Index Implementations} *)

(** Functor interface to construct index implementations. *)

(** Used by [TableType]. *)
type hkey

(** An input signature of the Make functor. The type of tables, that are used
to represent the discrimination tree. This signature is a subset of the
signature of [Hashtbl] in the OCaml standard library.
*)
module type TableType =
  sig
    type key=hkey
    type 'a t
    val make_table: unit -> 'a t
    val add: 'a t -> key -> 'a -> unit
    val remove: 'a t -> key -> unit
    val find: 'a t -> key -> 'a
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    (** Like [iter], but the implementation may remove entries
       from the table as they are iterated over. The table may no longer
       be used afterwards. *)
    val iter_and_throw_away: (key -> 'a -> unit) -> 'a t -> unit
  end

(** Functor for building an index implementation. *)
module Make (Entry : Index.EntryType) (Table : TableType) : 
    (Index.Index with type entry = Entry.entry)

(** Implementation of [TableType] using hash tables. *)
module HashTableImplementation : TableType

(** Implementation of [TableType] using association lists. *)
module AlistTableImplementation : TableType
