(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(*
let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "
*)




type env =
    { symbol_table: Term.symbol_table;
      predicate_table: Predicate.predicate_table;
      negative_units: NegativeUnits.store;
      subterm_table: Ic.subterm_table;
      mutable rules: Term.term list;
      mutable clauses: Ic.clause list;
      redundant_store: Ic.redundant_store;
      cfg: Cfg.cfg;
      clause_aux: Ic.Aux.aux;
      mutable shared_input_table: SharedInput.table;
      mutable current_result:Tableau.result
    }

let make_env cfg =
  let st = Term.make_symbol_table () 
  and pt = Predicate.make_predicate_table ()
  and nu = NegativeUnits.make_store ()
  and sut = Ic.make_subterm_table ()
  and rs = Ic.make_redundant_store ()
   in
  let ca = Ic.Aux.make_aux cfg pt st nu 
   in
  Builtin.init_builtins st;
  Evaluable.init_evaluables st;
  ignore(Ic.find_subterm sut (Term.get_equality_symbol st));	(* install the equality symbol as subterm
  right away - it is the only predicate symbol to be used as a subterm as well *)
  let krh_dom_pred = Predicate.find_predicate pt (Term.get_krh_dom_symbol st) 
  in
  Predicate.ensure_ready_to_assert krh_dom_pred;
  {symbol_table=st;
   predicate_table=pt;
   negative_units=nu;
   subterm_table=sut;
   rules=[];
   clauses=[];
   redundant_store=rs;
   cfg=cfg;
   clause_aux=ca;
   shared_input_table=SharedInput.null_table;
   current_result=Tableau.initial_result cfg pt st nu sut ca
 }

let get_cfg env =
  env.cfg

let get_clause_aux env =
  env.clause_aux  
  
let get_symbol_table env =
  env.symbol_table

let get_predicate_table env =
  env.predicate_table

let get_rules env =
  env.rules

let get_clauses env =
	env.clauses
	
let get_negative_units env =
  env.negative_units
  
let get_subterm_table env =
  env.subterm_table
  
let get_redundant_store env =
	env.redundant_store

type statement_type =
    Meta
  | Unit
  | Rule
  | Disjunction
  | False

let classify_statement_functor env sym =
  let st = env.symbol_table in
  if sym == Term.get_meta_functor_symbol st then
    Meta
  else if sym == Term.get_rule_symbol st then
    Rule
  else if sym == Term.get_disjunction_symbol st then
    Disjunction
  else if sym == Term.get_false_symbol st then
    False
  else
    Unit

let ensure_ready_to_assert env symbol =
  let pred = Predicate.find_predicate env.predicate_table symbol in
  Predicate.ensure_ready_to_assert pred;
  pred

let ensure_ready_to_run env symbol =
  let pred = Predicate.find_predicate env.predicate_table symbol in
  Predicate.ensure_ready_to_run pred;
  pred

let lookup_predicate env name arity =
  Predicate.PredicateTable.find 
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

let abolish_predicate env name arity =
  Predicate.abolish
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity);
  NegativeUnits.clear env.negative_units
    
  
let find_predicate env name arity =
  Predicate.find_predicate
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

(*    
let add_rule env clause =
  env.rules <- clause :: env.rules
*)

	
let make_and_add_clause env term =
	let clause = Ic.make_clause env.clause_aux term
	in	
		Ic.set_proof clause Fact.InputClauseProof;
		if (Ic.is_indexed env.subterm_table clause) then ((*ppp (Ic.get_source clause);pe "!!!"*))
		else (
			try (
				Ic.np_subsumption_processing_clause 
					env.subterm_table
					env.redundant_store
					clause;
				Ic.index_clause_subterms
					env.subterm_table
					clause;
				env.clauses <- clause :: env.clauses)
			with Ic.Redundant -> ())

let add_rule env term =
	make_and_add_clause env term;
	env.rules <- term :: env.rules

   
let add_disjunction env term =
	let disj_term = 
		(Term.Compound
       		(Term.make_compound
	  			(Term.get_rule_symbol env.symbol_table)
	  			[|term; (Term.Constant (Term.get_true_symbol env.symbol_table))|]))
	in
		add_rule env disj_term

let add_false env =
	let false_term =
		(Term.Compound
       		(Term.make_compound
	  			(Term.get_rule_symbol env.symbol_table)
	  				[|(Term.Constant (Term.get_false_symbol env.symbol_table));
	    			(Term.Constant (Term.get_true_symbol env.symbol_table))|]))
	in
		add_rule env false_term


let clear_rules env =
  env.rules <- [];
  env.clauses <- [];
  Ic.SubtermTable.iter 
  	(fun symbol subterm ->
  		Ic.clear subterm)
  	env.subterm_table;
  NegativeUnits.clear env.negative_units;
  Cfg.set_ehyper env.cfg false

let reset_clauses env =
	env.clauses <- []; 
  	Ic.SubtermTable.iter 
  		(fun symbol subterm ->
  			Ic.clear subterm)
  		env.subterm_table;
  	NegativeUnits.clear env.negative_units;
  	List.iter 
  		(fun rule ->
  			make_and_add_clause env rule)
  		env.rules;
  Cfg.set_ehyper env.cfg false
  
let begin_shared_input env =
  if env.shared_input_table == SharedInput.null_table then
    env.shared_input_table <- SharedInput.make_table ()
  else
    ()

let end_shared_input env =
  env.shared_input_table <- SharedInput.null_table

let unit_for_env env term =
  if env.shared_input_table == SharedInput.null_table then
    term
  else
    SharedInput.share_subterms env.shared_input_table term

let get_current_result env =
    env.current_result

let set_current_result env model =
  env.current_result <- model
