(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(** Entry point to "global" settings and datastructures. 

   Processing happens in the context of an [env] object. Usually a
   single [env] is used in a invocation of the system as standalone
   program. But of course also several [env]s could be used at the
   same time. 

   Associted with an [env] are

   - A [Cfg.cfg] representing configuration information.

   - A [Term.symbol_table].

   - A [Predicate.predicate_table]. Via the predicate table,
     meta information about predicates, such as index declarations,
     as well as the extent of predicates (derived or input unit clauses) 
     can be accessed. 

   - A list of nonunit clauses, represented as terms.
   
   [Env] provides some functions from [Predicate], with a "symbolic"
   interface, i.e. instead of taking symbol objects as input, name
   and arity are used.
*)

type env

val make_env : Cfg.cfg -> env

val get_cfg : env -> Cfg.cfg

val get_clause_aux : env -> Ic.Aux.aux

val get_symbol_table : env -> Term.symbol_table

val get_predicate_table : env -> Predicate.predicate_table

val get_negative_units : env -> NegativeUnits.store

val get_subterm_table : env -> Ic.subterm_table

val get_redundant_store : env -> Ic.redundant_store

val get_clauses : env -> Ic.clause list

val get_rules : env -> Term.term list

val get_current_result : env -> Tableau.result

val set_current_result : env -> Tableau.result -> unit

val ensure_ready_to_assert : env -> Term.symbol -> Predicate.predicate

val ensure_ready_to_run : env -> Term.symbol -> Predicate.predicate

type statement_type =
    Meta
  | Unit
  | Rule
  | Disjunction
  | False

val classify_statement_functor : env -> Term.symbol -> statement_type
    
(** Raises [Not_found] if the predicate does not exist. *)
val lookup_predicate : env -> string -> int -> Predicate.predicate

(** Ensures that the predicate exists. *)
val find_predicate : env -> string -> int -> Predicate.predicate

val add_rule : env -> Term.term -> unit

val add_disjunction : env -> Term.term -> unit

val add_false : env -> unit

(** Abolishes the specified predicate, if it exists. This means that 
all its unit clauses (input or derived) and declarations are removed.
It has no effect on any nonunit input clauses. 
*)
val abolish_predicate : env -> string -> int -> unit

(** Effect that all input nonunit clauses are removed. *)
val clear_rules : env -> unit

(** Enable sharing of subsequent input terms. Input terms are only
   shared within dynamic begin_shared_input/end_shared_input contexts.
   The behavior for nested begin_shared_input/end_shared_input
   contexts is undefined. 
*) 
val begin_shared_input : env -> unit

(** Disable sharing of subsequent input terms and free resources
   used for sharing.
 *)
val end_shared_input : env -> unit

(** Returns a possibly optimized variant of the input term (a unit clause),
   e.g. by using subterms that are shared with other clauses.
 *)
val unit_for_env : env -> Term.term -> Term.term


