(*
* Copyright (C) 2006 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(* miscellaneous equality related operations *)


(* debug *)
(*
let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i
*)


exception Failed

let multi_ident l1 r1 l2 r2 =
	(Termutils.is_ground l1)			
	&& (Termutils.is_ground r1)
	&& (((Term.term_ident l1 l2)&&(Term.term_ident r1 r2))
		||
		((Term.term_ident l1 r2)&&(Term.term_ident r1 l2)))

	
type equation =
	{	mutable term : Term.term;
		mutable left : Term.term;
		mutable right: Term.term;
		mutable is_oriented: bool;
		mutable left_is_ground: bool;
		mutable right_is_ground: bool }
			
let get_is_oriented eq =
	eq.is_oriented

let get_term eq =
	eq.term	
	
let get_left eq =
	eq.left

let get_right eq =
	eq.right
	
let get_is_ground eq =
	eq.left_is_ground
	&& eq.right_is_ground

let make_unchecked_equation term left right =
	{	term = term;
		left = left;
		right = right;
		is_oriented = false;	
		left_is_ground = Termutils.is_ground left;
		right_is_ground = Termutils.is_ground right }
		
let make_checked_equation term left right =
	if (LrpOrder.lrpo_greater_shared_vars left right) then
		{	term = Term.Compound (Term.make_compound (Convert.term_functor_symbol term) [|left; right|]);
			left = left;
			right = right;
			is_oriented = true;
			left_is_ground = Termutils.is_ground left;
			right_is_ground = Termutils.is_ground right }
	else
		if (LrpOrder.lrpo_greater_shared_vars right left) then
			{	term = Term.Compound (Term.make_compound (Convert.term_functor_symbol term) [|right; left|]);
				left = right;
				right = left;
				is_oriented = true;
				left_is_ground = Termutils.is_ground right;
				right_is_ground = Termutils.is_ground left }
		else
			if (LrpOrder.ground_like_lrpo_greater right left) then 	
			{	term = Term.Compound (Term.make_compound (Convert.term_functor_symbol term) [|right; left|]);
				left = right;
				right = left;
				is_oriented = false;
				left_is_ground = Termutils.is_ground right;
				right_is_ground = Termutils.is_ground left }
			else	
				{	term = Term.Compound (Term.make_compound (Convert.term_functor_symbol term) [|left; right|]);
					left = left;
					right = right;
					is_oriented = false;
					left_is_ground = Termutils.is_ground left;
					right_is_ground = Termutils.is_ground right }

let make_oriented_equation term left right =
	{	term = Term.Compound (Term.make_compound (Convert.term_functor_symbol term) [|right; left|]);
		left = left;
		right = right;
		is_oriented = true;
		left_is_ground = Termutils.is_ground left;
		right_is_ground = Termutils.is_ground right }
		
let make_tmp_equation term =
	{	term = term;
		left = term;
		right = Term.Unbound;
		is_oriented = true;
		left_is_ground = Termutils.is_ground term;
		right_is_ground = true }
		
let set_is_oriented eq is_oriented =
	eq.is_oriented <- is_oriented
	
let set_left eq left =
	eq.left <- left
	
let set_right eq right =
	eq.right <- right
	
let set_is_ground eq =
	eq.left_is_ground <- Termutils.is_ground eq.left;	
	eq.right_is_ground <- Termutils.is_ground eq.right
	
let copy_equation eq =
	{	term = Term.copy_term eq.term;	
		left = Term.copy_term eq.left;
		right = Term.copy_term eq.right;
		is_oriented = eq.is_oriented;
		left_is_ground = eq.left_is_ground;
		right_is_ground = eq.right_is_ground }
		
let term_to_equation term =
	let l, r = Convert.term_pair term
	in
	make_checked_equation term l r

(**************************************************************		
* Place (pointer-like datatype)		
**************************************************************)		
		
type place =
	ArrayPlace of (Term.term array * int)
	| EqLeftPlace of equation
	| EqRightPlace of equation
	
let set_place place term =
	match place with
		ArrayPlace (args, index) ->
			args.(index) <- term
		| EqLeftPlace eq ->
			set_left eq term
		| EqRightPlace eq ->
			set_right eq term
			
let get_place place =
	match place with
		ArrayPlace (args, index) ->
			args.(index)
		| EqLeftPlace eq ->
			get_left eq
		| EqRightPlace eq ->
			get_right eq

let make_array_place term =
	ArrayPlace ((Array.make 1 term), 0)
	
let null_place =
	make_array_place Term.Unbound
	
let get_i place i =
	match place with
		ArrayPlace (args, _) ->
			args.(i)
		|_ -> raise Failed
		
(**************************************************************		
* subterm iteration 		
**************************************************************)							
let rec iter_nonvar_subterms_place_no_top f term place =
	match term with
		Term.Variable v -> ()
		|Term.Compound k ->
			let args = Term.compound_args k
			in 
				for i=0 to ((Array.length args)-1) do
					iter_nonvar_subterms_place f args.(i) (ArrayPlace (args, i))
				done
		|_ -> f term place		

and iter_nonvar_subterms_place f term place =
	match term with
		Term.Variable v -> ()
		|Term.Compound k ->
			f term place;
			let args = Term.compound_args k
			in 
				for i=0 to ((Array.length args)-1) do
					iter_nonvar_subterms_place f args.(i) (ArrayPlace (args, i))
				done
		|_ -> f term place
				

let iter_nonvar_subterms f term =
	let place = make_array_place term in
		iter_nonvar_subterms_place f term place					
		
		


exception Substitution_successful		
	

let sup_right_substitute l r s t context_lr context_st trail term place =
	let saved_trail = Term.save_trail trail
	and trail = Term.unify term context_st l context_lr trail
	in 
		if Term.failed trail then
			()
		else (
			let l = Term.apply l context_lr
			and r = Term.apply r context_lr
			in 
				if (Term.term_ident l r) || (LrpOrder.lrpo_greater_shared_vars r l) then 
					(ignore (Term.undo_bindings trail saved_trail))
				else (
					let s = Term.apply s context_st
					and t = Term.apply t context_st
					in
						if (multi_ident l r s t) || (LrpOrder.eq_lrpo_greater l r s t) then 
							(ignore (Term.undo_bindings trail saved_trail))
						else (
							if (Term.term_ident s t) || (LrpOrder.lrpo_greater_shared_vars t s) then 
								(ignore (Term.undo_bindings trail saved_trail))
							else (
								set_place place r;
								raise Substitution_successful)
						)
				)
			)
	
let sup_left_substitute l r s t context_lr context_st trail term place =
	let saved_trail = Term.save_trail trail
	and trail = Term.unify term context_st l context_lr trail
	in  
		if Term.failed trail then
			()
		else (
			let l = Term.apply l context_lr
			and r = Term.apply r context_lr
			in 
				if (Term.term_ident l r) || (LrpOrder.lrpo_greater_shared_vars r l) then 
					(ignore (Term.undo_bindings trail saved_trail))
				else (
					let s = Term.apply s context_st
					and t = Term.apply t context_st
					in
						if (Term.term_ident s t) || (LrpOrder.lrpo_greater_shared_vars t s) then 
							(ignore (Term.undo_bindings trail saved_trail))
						else 
							set_place place r;
							raise Substitution_successful)
							
			)
		
			
			
			
			



let substitute_sup_right l r st place context_lr context_st trail left_is_indexed sup =
		if Term.failed (Term.unify (get_place place) context_st l context_lr trail) then
			(raise Failed)
		else (
			let s_index, t_index =
				if left_is_indexed then
					(0, 1)
				else
					(1, 0)
			in
			let s2 = Term.apply (Term.get_compound_args st).(s_index) context_st
			and t2 = Term.apply (Term.get_compound_args st).(t_index) context_st
			in 	
				if sup && ((Term.term_ident s2 t2) || (LrpOrder.lrpo_greater_shared_vars t2 s2)) then
					(raise Failed)
				else (
					let l2 = Term.apply l context_lr
					and r2 = Term.apply r context_lr
					in
						if (Term.term_ident l2 r2) 
						|| (LrpOrder.lrpo_greater_shared_vars r2 l2)
						|| (multi_ident l2 r2 s2 t2) 
						|| (LrpOrder.eq_lrpo_greater l2 r2 s2 t2) then
							(raise Failed)
						else (
							let saved_place = get_place place
							in
							set_place place r2;
							let rewritten_s = Term.copy_term (Term.get_compound_args st).(s_index)
							in	
								set_place place saved_place;
								let rewritten_s = Term.apply rewritten_s context_st
								in 
									(rewritten_s, t2, context_st)
						)
				)
		)
		


let substitute_sup_left l r st place context_lr context_st trail equals left_is_indexed sup =
		if Term.failed (Term.unify (get_place place) context_st l context_lr trail) then
			(raise Failed)
		else (
			let s_index, t_index =
				if left_is_indexed then
					(0, 1)
				else
					(1, 0)
			in
			let s2 = Term.apply (Term.get_compound_args st).(s_index) context_st
			and t2 = Term.apply (Term.get_compound_args st).(t_index) context_st
			in
				if sup && ((Term.term_ident s2 t2) || (LrpOrder.lrpo_greater_shared_vars t2 s2)) then
					(raise Failed)
				else (
					let l2 = Term.apply l context_lr
					and r2 = Term.apply r context_lr
					in
						if (Term.term_ident l2 r2) || (LrpOrder.lrpo_greater_shared_vars r2 l2) then
							(raise Failed)
						else (		
							let saved_place = get_place place
							in
							set_place place r2;	
							let rewritten_s = Term.copy_term (Term.get_compound_args st).(s_index)
							in
								set_place place saved_place;
								let rewritten_s = Term.apply rewritten_s context_st
								in 
									(rewritten_s, t2, context_st)
						)
				)
		)	
		
		
	
		
let sup_right_place l r st place left_is_indexed sup =
	let l2 = Term.copy_term l
	and r2 = Term.copy_term r
	and renum_table = Array.make Term.max_vars (-1)
	and context_lr = Term.make_context 0
	and context_st = Term.make_context 0
	and trail = Term.make_trail ()
	in
		let renum_compound = Term.make_anonymous_binary_compound l2 r2
		in  	
			Term.init_renum_table renum_table st;
			Term.renumber_variables_term_specific renum_table renum_compound;
			let  s2, t2, context_st2 = substitute_sup_right l2 r2 st place context_lr context_st trail left_is_indexed sup
			in 
				(s2, t2, context_st2)

let sup_left_place l r st place equals st_term left_is_indexed sup =
	let l2 = Term.copy_term l
	and r2 = Term.copy_term r
	and renum_table = Array.make Term.max_vars (-1)
	and context_lr = Term.make_context 0
	and context_st = Term.make_context 0
	and trail = Term.make_trail ()
	in
		let renum_compound = Term.make_anonymous_binary_compound l2 r2
		in  	
			Term.init_renum_table renum_table st_term;
			Term.renumber_variables_term_specific renum_table renum_compound;
			let  s2, t2, context_st2 = substitute_sup_left l2 r2 st place context_lr context_st trail equals left_is_indexed sup
			in
				(s2, t2, context_st2)				
				
				
				
			
						
let sup_right l r s t =
	let l2 = Term.copy_term l
	and r2 = Term.copy_term r
	and s2 = Term.copy_term s
	and t2 = Term.copy_term t
	and renum_table = Array.make Term.max_vars (-1)
	in
		let renum_compound = Term.make_anonymous_binary_compound l2 r2
		in  	
		Term.init_renum_table renum_table s2;
		Term.init_renum_table renum_table t2;
		Term.renumber_variables_term_specific renum_table renum_compound;
		let place = make_array_place s2
		and context_lr = Term.make_context 0
		and context_st = Term.make_context 0
		and trail = Term.make_trail ()
		and result_s = ref s
		and result_t = ref t
		and iter_nonvar_subterms_place_gen =
			if t2==Term.Unbound then
				(iter_nonvar_subterms_place_no_top)
			else
				(iter_nonvar_subterms_place)
		in 	
			(try
				iter_nonvar_subterms_place_gen
					(sup_right_substitute
							l2
							r2
							s2
							t2
							context_lr
							context_st
							trail)
					s2
					place

			with Substitution_successful -> let s3 = Term.apply s context_st
											and t3 = Term.apply t context_st
											in 
												if not((Term.term_ident s3 t3)||(LrpOrder.lrpo_greater_shared_vars t3 s3)) then begin
													result_s := Term.apply (get_place place) context_st;
													result_t := Term.apply t2 context_st
												end
												else ());
			(!result_s, !result_t, context_st)

			

			
let sup_left l r s t st_term sup =
	let l2 = Term.copy_term l
	and r2 = Term.copy_term r
	and s2 = Term.copy_term s
	and t2 = Term.copy_term t
	and renum_table = Array.make Term.max_vars (-1)
	in 
		let renum_compound = Term.make_anonymous_binary_compound l2 r2
		in  	
		Term.init_renum_table renum_table st_term;
		Term.renumber_variables_term_specific renum_table renum_compound;
		let place = make_array_place s2
		and context_lr = Term.make_context 0
		and context_st = Term.make_context 0
		and trail = Term.make_trail ()
		and result_s = ref s
		and result_t = ref t
		and iter_nonvar_subterms_place_gen =
			if t2==Term.Unbound then
				(iter_nonvar_subterms_place_no_top)
			else
				(iter_nonvar_subterms_place)
		in 	
			(try
				iter_nonvar_subterms_place_gen
					(sup_left_substitute
							l2
							r2
							s2
							t2
							context_lr
							context_st
							trail)
					s2
					place

			with Substitution_successful -> let s3 = Term.apply s context_st	
											and t3 = Term.apply t context_st
											in 
												if sup then (
													if not((Term.term_ident s3 t3)||(LrpOrder.lrpo_greater_shared_vars t3 s3)) then (
														result_s := Term.apply (get_place place) context_st;
														result_t := Term.apply t2 context_st)
													else ())
												else (
													result_s := Term.apply (get_place place) context_st;
													result_t := Term.apply t2 context_st));
			(!result_s, !result_t, context_st)			

				
				
			
			
let unit_sup_right_lr l r term equals =
	let results = ref []
	in
		if equals == (Convert.term_functor_symbol term) then (
			let s, t = Convert.term_pair term
			in
				(let s_new, t_new, context = sup_right l r s t
				in
					if not (Term.term_ident s s_new) then
						results := (Term.Compound (Term.make_compound equals [|s_new; t_new|]), context) :: !results);
				(let t_new, s_new, context  = sup_right l r t s
				in
					if not (Term.term_ident t t_new) then
						results := (Term.Compound (Term.make_compound equals [|t_new; s_new|]), context) :: !results))
		else (
			let term_new, t_new, context = sup_right l r term Term.Unbound
			in
				if not (Term.term_ident term term_new) then
					results := (term_new, context) :: !results);
		!results

let unit_sup_right_lr_into_eq l r s t equals =
	let results = ref []
	in	
		let s_new, t_new, context = sup_right l r s t
		in 
			if not (Term.term_ident s s_new) then
				results := (Term.Compound (Term.make_compound equals [|s_new; t_new|]), context) :: !results
			else ();
			!results		
		
		
let unit_sup_right_lr_into_eq_place l r eq place equals left_is_indexed sup =
	let s_new, t_new, context = sup_right_place l r eq place left_is_indexed sup
	in 
		(Term.Compound (Term.make_compound equals [|s_new; t_new|]), context)

let unit_sup_right_lr_into_term_place l r term place equals sup =
	let compatibility_term =
		Term.Compound (Term.make_compound equals [|term; Term.Unbound|])
	in 
	let term_new, _, context = sup_right_place l r compatibility_term place true sup
	in 
		(term_new, context)

		
				
let unit_sup_right_lr_into_term l r term =
	let results = ref []
	in
		let term_new, t_new, context = sup_right l r term Term.Unbound
		in
			if not (Term.term_ident term term_new) then
				results := (term_new, context) :: !results
			else ();
			!results
			
	
			
		
let sup_left_lr_into_eq l r s t equals st_term sup =
	let results = ref []
	in
		let s_new, t_new, context = sup_left l r s t st_term sup
		in 
			if not (Term.term_ident s s_new) then
				results := (Term.Compound (Term.make_compound equals [|s_new; t_new|]), context) :: !results
			else ();
			!results		
				
let sup_left_lr_into_term l r term st_term sup =
	let results = ref []
	in
		let term_new, t_new, context = sup_left l r term Term.Unbound st_term sup
		in
			if not (Term.term_ident term term_new) then
				results := (term_new, context) :: !results
			else ();
			!results
			
			
let sup_left_lr_into_eq_place l r st place equals st_term left_is_indexed sup =
	let s_new, t_new, context = sup_left_place l r st place equals st_term left_is_indexed sup
	in 
		(Term.Compound (Term.make_compound equals [|s_new; t_new|]), context)
	
			
let sup_left_lr_into_term_place l r term place equals st_term sup =
	let compatibility_term =
		Term.Compound (Term.make_compound equals [|term; Term.Unbound|])
	in
	let term_new, _, context = sup_left_place l r compatibility_term place equals st_term true sup
	in
		(term_new, context)
			