(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Arithmetic evaluation *)

open Term

exception Evaluable_error of string

let rec eval term context =
  let (term, context) = dereference term context in
  match term with
    Variable _ ->
      raise (Evaluable_error "Unbound variable in arithmetic evaluation.")
  | Constant s ->
      raise 
	(Evaluable_error ("No evaluable constant: "^(symbol_name s)^"."))
  | Integer n ->
      term
  | Float n ->
      term
  | Unbound ->
      raise (Failure "Bad term used for evaluation.")
  | Compound k ->
      let sym = (compound_functor k) in
      let sf = symbol_function sym and
	  arity = symbol_arity sym and
	  args = compound_args k in
      match sf, arity with
	FUnbound, _ ->
	  raise (Evaluable_error 
		   ("No evaluable function: "^(symbol_name sym)^"."))
      | FUnary f, 1 ->
	  f (eval args.(0) context)
      | FBinary f, 2 ->
	  f (eval args.(0) context) (eval args.(1) context)
      | _ -> raise (Failure 
		 ("Bad function used for evaluation: "^
		  (symbol_name sym)^"."))

let dont_gc_these_symbols = ref []
 
let register_evaluable symbol_table symbol_name arity f =
  let symbol = find_symbol symbol_table symbol_name arity in
  match symbol_function symbol with
    FUnbound ->
      set_symbol_function symbol f;
      dont_gc_these_symbols := symbol :: !dont_gc_these_symbols
  | _ -> raise (Failure ("Try to reassign function symbol: "^
			 symbol_name^"/"^(string_of_int arity)^"."))

let clear_evaluables symbol_table =
  symbol_table_iter
    (fun symbol ->
      set_symbol_function symbol FUnbound)
    symbol_table;
  dont_gc_these_symbols := []

	
(****************************************************************************)
(* Arithmetic Functions *)
(* Should correspond to ISO-Prolog, for now just an excerpt. *)
(****************************************************************************)

let eval_plus x y =
  match x, y with
    Integer n, Integer m -> Integer (n + m)
  | Float n, Integer m -> Float (n +. float(m))
  | Integer n, Float m -> Float (float(n) +. m)
  | Float n, Float m -> Float (n +. m)
  | _ -> raise (Evaluable_error "Wrong argument type in +/2.")

let eval_minus x y =
  match x, y with
    Integer n, Integer m -> Integer (n - m)
  | Float n, Integer m -> Float (n -. float(m))
  | Integer n, Float m -> Float (float(n) -. m)
  | Float n, Float m -> Float (n -. m)
  | _ -> raise (Evaluable_error "Wrong argument type in -/2.")

let eval_times x y =
  match x, y with
    Integer n, Integer m -> Integer (n * m)
  | Float n, Integer m -> Float (n *. float(m))
  | Integer n, Float m -> Float (float(n) *. m)
  | Float n, Float m -> Float (n *. m)
  | _ -> raise (Evaluable_error "Wrong argument type in */2.")

let eval_div x y =
  match x, y with
    Integer n, Integer m -> Float (float(n) /. float(m))
  | Float n, Integer m -> Float (n /. float(m))
  | Integer n, Float m -> Float (float(n) /. m)
  | Float n, Float m -> Float (n /. m)
  | _ -> raise (Evaluable_error "Wrong argument type in //2.")

let eval_minus_unary x =
  match x with
    Integer n -> Integer (- n)
  | Float n -> Float (-.  n)
  | _ -> raise (Evaluable_error "Wrong argument type in -/1.")

let eval_abs x =
  match x with
    Integer n -> Integer (abs n)
  | Float n -> Float (abs_float  n)
  | _ -> raise (Evaluable_error "Wrong argument type in abs/1.")

let eval_float x =
  match x with
    Integer n -> Float (float n)
  | Float _ -> x
  | _ -> raise (Evaluable_error "Wrong argument type in float/1.")

let eval_truncate x =
  match x with
  | Float n -> Integer (truncate n)
  | _ -> raise (Evaluable_error "Wrong argument type in truncate/1.")

(* ### ... *)

(****************************************************************************)

let init_evaluables symbol_table =
  register_evaluable symbol_table "+" 2 (FBinary eval_plus);
  register_evaluable symbol_table "-" 2 (FBinary eval_minus);
  register_evaluable symbol_table "*" 2 (FBinary eval_times);
  register_evaluable symbol_table "/" 2 (FBinary eval_div);
  register_evaluable symbol_table "-" 1 (FUnary eval_minus_unary);
  register_evaluable symbol_table "abs" 1 (FUnary eval_abs);
  register_evaluable symbol_table "float" 1 (FUnary eval_float);
  register_evaluable symbol_table "truncate" 1 (FUnary eval_truncate)

(****************************************************************************)
