(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(** Representation of derived facts and associated meta information. *)

type fact =
    InputFact of Term.term * bool * fact ref
  | DerivedFact of Term.term * bool * fact ref	
  | DerivedFactWithProof of Term.term * proof * bool * fact ref
  | InputEquationFact of Equality.equation * bool * bool * fact ref
  | DerivedEquationFact of Equality.equation * bool * bool * fact ref
  | DerivedEquationFactWithProof of Equality.equation * bool * proof * bool * fact ref  
  | FactWithUnknownProof
  | NullFact


and proof =
    BuiltinProof
  | InputFactProof
  | InputClauseProof
  | Proof of Term.term * fact array
  | DisjProof of int * Term.term * fact array * fact list
  | NegProof of Term.term * fact array * fact
  | UnknownProof (* Similar to [BuiltinProof], however no assumptions
		    about the proof may be made, e.g. [UnknownProof]
		    may not be used to justify a level cut by missing
		    presence of a literal. *)
  | UnitSupRightProof of Term.term * Term.term * fact * fact
  | NegUnitSupRightProof of Term.term * Term.term * fact * fact * fact  
  | SupLeftProof of Term.term * Term.term * fact * fact	 
  | EqRefProof of Term.term * fact 	
  | HyperPureProof of Term.term * fact array	    

(* val print_proof : proof -> unit  *)
  
val is_redundant : fact -> bool

val is_ground : fact -> bool

val declare_redundant : fact -> unit

val declare_redundant_by_fact : fact -> fact -> unit

val undeclare_redundant : fact -> unit 

val get_reducer : fact -> fact

val null_fact : fact

val fact_with_unknown_proof : fact

val make_input_fact : Term.term -> fact

val make_input_equation_fact : Equality.equation -> bool -> fact

val make_derived_fact : Term.term -> fact

val make_derived_equation_fact : Equality.equation -> bool -> fact

val make_derived_fact_with_proof : Term.term -> proof -> fact

val make_derived_equation_fact_with_proof : Equality.equation -> bool -> proof -> fact

val is_equation : fact -> bool

val get_equation : fact -> Equality.equation

(** Result is like the given fact, but with the given term. *)
val make_fact_like : fact -> Term.term -> fact

(** {6 Entry Interface} 
   This interface is required by functions in the [Access] module.
 *)

type entry = fact

val get_term : fact -> Term.term

val get_proof : fact -> proof

(** {6 Proof Terms} *)

(** A proof term is structured as follows:

      node(Id, Goal, Rule, Subproofs, NegativeSubproofs) |
      split(Id, Rule, Subproofs, NegativeSubproofs) |
      branch(Id, Goal, Subproof) |
      ref(Id)

{ul
   {- Id     is an atom identifying the node.}
   {- Goal   is the subgoal proven by the node.}
   {- Rule   is an atom or compound representing the applied deduction rule.
           It is one of
   {ul
           {- builtin       - not further specified}
           {- input_fact    - an input fact}
           {- hyper(Clause) - hypertableau/resolution inference
                              step with the specified
                              input clause.
                              Clause is given as term.}
   }}
   {- Subproofs is a list of proof terms. It represents the
      subproofs used to prove Goal with Rule.}
   {- NegativeSubproofs are subproofs that resolve head literals.
      They could result from negative input literals and complement
      splitting.}
   {- A proof term with node/5 structure represents a proof of Goal
      by applying Rule to Subproofs and NegativeSubproofs}
   {- A proof term with split/4 structure represents a disjunction, that
      has been derived by applying Rule to Subproofs and NegativeSubproofs.}
   {- A proof term with branch/3 structure represents a disjunct contained
      in the disjunction of the split/4 structure Subproof.}
   {- A proof term with ref/1 structure is a reference to
      the node with the given identifier.}
}
*)

val proof_term_for_fact : Term.symbol_table -> fact -> Term.term

type l_proof = proof list

val proof_term_for_l_proof : Term.symbol_table -> l_proof -> Term.term

val make_subproofs : int -> fact array

type proof_term_aux

(** A proof term aux can be shared in separate calls to proof_term_r_proof,
   to ensure sharing of subproofs... *)
val make_proof_term_aux :  unit -> proof_term_aux

val proof_term_for_r_proof : Term.symbol_table -> proof_term_aux -> proof ->
   Term.term

