(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type 'a s_graph = ('a * ('a list)) list

let rec warshall g =
   (* A straightforward port of Richard O'Keefe's Prolog implementation
      of the Warshall algorithm.    
    *)
  warshall_3 g g
    
and warshall_3 g1 g2 =
  match g1 with
    [] ->
      g2
  | (v, _) :: g3 ->
      warshall_3 g3 (warshall_4 g2 v (List.assoc v g2))
	
and warshall_4 g1 v y =
  match g1 with
    [] ->
      []
  | (x, neibs) :: g2 ->
      if List.mem v neibs then
	(x, Ordset.ord_union neibs y) :: warshall_4 g2 v y
      else
	(x, neibs) :: warshall_4 g2 v y


type 'a p_graph = ('a * 'a) list


let rec p_to_s_graph pg =
  let edge_set = Ordset.list_to_ord_set pg in
  p_to_s_group 
    (Ordset.list_to_ord_set (p_to_s_vertices edge_set))
    edge_set
    
and p_to_s_vertices es =
  match es with
    [] ->
      []
  | (a,z) :: es1 ->
      a :: z :: p_to_s_vertices es1

and p_to_s_group vs es =
  match vs with
    [] ->
      []
  | v :: vs1 ->
      let (neibs, es1) = p_to_s_group_1 es v in
      (v, neibs) :: p_to_s_group vs1 es1

and p_to_s_group_1 es v =
   (* ### Probably not as tail recursive as
      possible. *)
    match es with
      [] ->
	([], es)
    | (v1, x) :: es1 ->
	if (v1 = v) then
	  let (neibs, es2) = p_to_s_group_1 es1 v in
	  (x::neibs, es2)
	else
	  ([], es)

let rec s_to_p_graph sg =
  match sg with
    [] ->
      []
  | (_, []) :: sg1 ->
      s_to_p_graph sg1
  | (v, (x :: xs)) :: sg1 ->
      (v, x) :: s_to_p_graph ((v, xs) :: sg1)

let p_transpose pg =
  List.map (fun (x, y) -> (y, x)) pg

let rec top_sort g =
  (* A quick port of the Prolog implementation. *)
  let vcs = count_edges g (List.map (fun (v,_) -> (v,0)) g) in
  top_sort_1
    (List.map fst (List.filter (fun (v, m) -> m = 0) vcs))
    g
    vcs

and count_edges g vcs =
  match g with
    [] ->
      vcs
  | (_, ns) :: g ->
      count_edges g (incr_list ns vcs)

and incr_list ns vcs =
  match ns, vcs with
    [], _ ->
      vcs
  | n :: ns0, (v, m) :: vcs0 ->
      if n = v then
	(v, (m+1)) :: incr_list ns0 vcs0
      else
	(v, m) :: incr_list ns vcs0
  | _ ->
      raise (Failure "This should not happen: incr_list.")

and top_sort_1 zvs g vcs =
  match zvs with
    [] ->
      if List.exists (fun (_, c) -> c <> 0) vcs then
	raise Not_found
      else
	[]
  | v :: zvs0 ->
      let (vcs2, new_zeros) = 
	decr_list (List.assoc v g) vcs zvs0 in
      v :: top_sort_1 new_zeros g vcs2

and decr_list ns vcs zvs =
  let new_zvs = ref zvs in
  let rec decr_list_1 ns vcs =
    match ns, vcs with
      [], _ ->
	vcs
    | n :: ns0, (v,c) :: vcs0 ->
	if n = v then
	  if c = 1 then
	    begin
	      new_zvs := v :: !new_zvs;
	      (v,0) :: decr_list_1 ns0 vcs0
	    end
	  else
	    (v,(c-1)) :: decr_list_1 ns0 vcs0
	else
	  (v,c) :: decr_list_1 ns vcs0
    | _ ->
	raise (Failure "This should not happen: decr_list.")
  in 
  let vcs1 = decr_list_1 ns vcs in
  (vcs1, !new_zvs)


