(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(** Compiled input clauses. *)

(** Notes:

   The compiled clauses intern references to predicate objects (i.e.
   objects of type Predicate.predicate) and information obtained by
   them. Abolishing a predicate can make a compiled clause unusable.

   See also notes for module [IcSet].
*)

exception Redundant

type atom =
	Term of Term.term
	| Eq of Equality.equation

type sign =
    Pos (** Positive. *)
  | StratifiedNegAsFailure (** Stratified negation as failure. 

                 [\+(A)], where [A] must be a non-builtin atom.
		 Succeeds iff [A] is not unifiable with a fact,
		 at time of the call. [A] is handled like a builtin
		 that has to be ground in subgoal sorting. 

		 This kind of negation should work well with
		 stratification and ground facts. More complex
		 expressions within the scope of [\+] (conjunctions,
		 existential variables) can be translated into this
		 form. *)
  | StratifiedNegAsFailureNotSubsumed 
    (** See explanation for 'not_subsumed' in user documentation. *)
  | Findall
  | Neg2

module BodyLiteral :
  sig

    type literal

    val get_sign : literal -> sign

    val get_predicate : literal -> Predicate.predicate

    val get_atom : literal -> atom
    
    val get_atom_term : literal -> Term.term

    val get_reader : literal -> Access.Reader.reader

    val get_foreign_context : literal -> Term.context

    val get_is_delta : literal -> bool

    val get_builtin : literal -> Term.p

    val make_literal : Predicate.predicate_table ->
      Term.symbol_table ->
      bool ->
      Term.term -> literal

    val print_optim_info : literal -> unit

  end


module HeadLiteral :
    sig

      type literal
	    
      val get_atom : literal -> atom

      val get_atom_term : literal -> Term.term
      
      val get_predicate : literal -> Predicate.predicate

      val make_literal : Predicate.predicate_table -> atom -> literal

    end

module Aux :
    sig

      (** Used to hold auxiliary datastructures. *)
      type aux

      (** Creates an [aux] for the given [Cfg.cfg].*)
      val make_aux : Cfg.cfg -> 
	Predicate.predicate_table -> Term.symbol_table ->
	  NegativeUnits.store -> aux

    end

module DisjunctiveHead :
    sig
      type head
      val get_literals : head -> HeadLiteral.literal list
    end

type body =
    Body of BodyLiteral.literal list
	
type head =
    False
  | Unit of HeadLiteral.literal
  | Conjunction of HeadLiteral.literal list
  | Disjunction of DisjunctiveHead.head

type clause

val get_ground_head_eqs : clause -> Term.term array

val get_non_ground_head_eqs : clause -> Term.term array

val get_body : clause -> body

val get_body_predicates : clause list -> Predicate.predicate list

val get_head : clause -> head

val get_news : clause -> News.news

val get_id : clause -> string

val get_source : clause -> Term.term

val get_aux : clause -> Aux.aux

val get_sigterm : clause -> Term.term

val get_fuzzy_sigterm : clause -> Term.term

val is_eq_unit : clause -> bool

val is_oriented : clause -> bool

val is_redundant : clause -> bool

val is_ground : clause -> bool

val set_proof : clause -> Fact.proof -> unit

val is_in_current_stratum : clause -> bool

val set_is_in_current_stratum : bool -> clause -> unit

val make_clause : Aux.aux -> Term.term -> clause

(** Returns a list of clauses, that contains for each body literal
   with a predicate that appears in delta_predicates, a copy of the
   [clause], with the body literal replaced by ist delta version. 
*)
val clause_delta_clauses : Predicate.predicate list -> clause -> clause list

(** A generalized version of [clause_delta_clauses]; uses no delta predicates,
creates all delta versions of clause. *)
val clause_delta_clauses_unchecked : clause -> clause list

(** Prepare a clause for [eval_clause]. *)
val initialize_clause : clause -> News.news -> unit

type redundant_store

val make_redundant_store : unit -> redundant_store

val declare_fact_redundant : redundant_store -> Fact.fact -> unit

val declare_clause_redundant : redundant_store -> clause -> unit

val undeclare_redundancies : redundant_store -> unit


(** Raised if [False] head has been derived (a hypertableau branch has been
 closed. 
*)
exception BranchClosed of Fact.proof


module SubtermEntry :
	sig
		type entry =
			Clause of (clause * int * bool * Equality.place)
			|Fact of (Fact.fact * Equality.place)
		val get_term : entry -> Term.term
		val get_place_term : entry -> Term.term
		val is_redundant : entry -> bool
		val is_ground : entry -> bool
		val declare_redundant : redundant_store -> entry -> unit
	end
	
(** {6 Support of Modifiable States } *)

module SubtermStore :
    sig
      type store
      type state
      val obtain_state : store -> state
      val install_state : state -> unit
      val add_layer : state -> state
      val new_layer : store -> state
    end

(** {6 Access Objects } *)

module Writer :
  sig
    type writer
    val make_writer : Opt.indextype -> Opt.indexargs -> int -> writer
    val insert : writer -> Term.term -> SubtermEntry.entry -> unit
    val remove_variant : writer -> Term.term -> unit
    val clear : writer -> unit
    val get_store : writer -> SubtermStore.store
    val size : writer -> int
  end

module Reader :
  sig
    type reader
    val make_readers : Writer.writer -> reader list
    val null_reader : unit -> reader
    val applicability : reader -> Opt.binding_pattern ->
      Opt.applicability_value
    val iter_unification_candidates :
      (SubtermEntry.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_instance_candidates :
      (SubtermEntry.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_generalization_candidates :
      (SubtermEntry.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_variant_candidates :
    	(SubtermEntry.entry -> unit) ->
      reader -> Term.term -> Term.context -> unit
    val iter_all : (SubtermEntry.entry -> unit) -> reader -> unit
    val get_indexargs : reader -> Opt.indexargs
(** DEBUG *)
    val print_info : reader -> unit
  end

type subterm

type subterm_table

module SubtermTable :
    sig
      val find : subterm_table -> Term.symbol -> subterm
      val iter : (Term.symbol -> subterm -> unit) -> subterm_table -> unit
    end

val make_subterm_table : unit -> subterm_table

val find_subterm : subterm_table -> Term.symbol -> subterm

val get_subterm_reader : subterm -> Reader.reader

val get_subterm_writer : subterm -> Writer.writer

val get_delta_subterm : subterm -> subterm

(** [index_subterms subterm_table clause] inserts clause into
subterm_table. *)
val index_clause_subterms : subterm_table -> clause -> unit
val index_literal_term : subterm_table -> Term.term -> clause -> int -> bool -> Equality.place -> unit
val index_literal_eq : subterm_table -> Term.term -> clause -> int -> unit
val index_unit_subterms : subterm_table -> Term.term -> Fact.entry -> Equality.place -> unit

(** Same as [index_unit_subterms], but does not use the top functor (i.e. predicate)
for indexing. *)
val index_unit_subterms_no_top : subterm_table -> Term.term -> Fact.entry -> unit

(** [is_indexed subterm_table clause] checks whether clause or a variant of
clause has been indexed in subterm_table, returns boolean true or false. *)
val is_indexed : subterm_table -> clause -> bool

val has_delta_subterm : subterm -> bool

val ensure_delta_ready_to_run : subterm -> unit

val clear : subterm -> unit

val clear_delta_subterm : subterm -> unit	

val indexed_in_extent : Predicate.predicate -> Term.term -> bool

val nonred_indexed_in_extent : Predicate.predicate -> Term.term -> bool

val red_indexed_in_extent : Predicate.predicate -> Term.term -> bool

module NewClauseCandidates :
	sig

		type new_clause_candidates
		
		val make_new_clause_candidates : Cfg.cfg -> new_clause_candidates
		
		val get_current_stratum_candidates : new_clause_candidates -> clause list
			
		val get_other_stratum_candidates : new_clause_candidates -> clause list
		
		val get_delayed_candidates : new_clause_candidates -> clause list
		
		val insert_current_stratum_candidate : new_clause_candidates -> clause -> unit
				
		val insert_other_stratum_candidate : new_clause_candidates -> clause -> unit
				
		val clear_new_clause_candidates : new_clause_candidates -> unit
		
		val clear_delayed_clause_candidates : new_clause_candidates -> unit
		
		val set_max_literal_weight : new_clause_candidates -> int -> unit
		
		val get_max_literal_weight : new_clause_candidates -> int

		val reset_weight_limit_exceeded : new_clause_candidates -> unit

		val weight_limit_exceeded : new_clause_candidates -> int		
		
		val get_ncc_id : new_clause_candidates -> string

	end

(** Raises [BranchClosed]. *)
val eval_clause : clause -> bool -> subterm_table 
	-> clause list -> NewClauseCandidates.new_clause_candidates -> redundant_store -> unit

val reset_gen_id : unit -> unit

val remove_clause_duplicates : clause list -> clause list

val remove_subsumed_clauses : clause list -> clause list






val get_subterm_reader : subterm -> Reader.reader
	

(** [np_subsumption_processing_clause subterm_table redundant_store clause] compares [clause] with
the current clause set regarding non-proper subsumption. All clauses non-properly subsumed
by [clause] get blocked. If another clause non-properly subsumes [clause], then a 
[Redundant]-exception is raised. *)
val np_subsumption_processing_clause : subterm_table -> redundant_store -> clause -> unit

(** Test of unit subsumption, analogous to [np_subsumption_processing_clause]. *)
val np_subsumption_processing_unit : bool -> redundant_store -> Predicate.predicate -> Fact.fact -> unit

(** [clause_np_subsumed_by_clauses subterm_table clause] returns true if [clause] is non-properly
subsumed by the current clause extent and false otherwise.*)
val clause_np_subsumed_by_clauses : subterm_table -> clause -> bool


(** [minimize_term subterm_table equals term] reduces [term] as far as possible
using consecutive reductive superposition inferences. 
[term] may not be a literal, since all symbols in [term] are open for rewriting, whereas
predicates may not be overwritten.
[term] should be ground. The output of the function for a non-ground [term] is undefined.
A minimized term is returned; this will be [term] itself if no reduction was possible. *)
val minimize_term : subterm_table -> Term.symbol -> Term.term -> Term.term


(** [minimize_literal_term subterm_table equals term]: same as
[minimize_term], but preserves the top symbol of [term] and is thus suitable for atoms. *)
val minimize_literal_term : Term.symbol -> subterm_table -> Term.term -> Term.term

(** [minimize_clause clause subterm_table] reduces [clause] as far as possible while ensuring
that the result makes [clause] redundant. *) 
val minimize_clause : clause -> subterm_table -> clause