(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


module DependencyGraph =
  struct

    (* ### Check this stuff for
       - disjunctive heads
         
       - negation:
         - is it handled as builtin?
     *)

    let pred_node predicate =
      let sym = Predicate.get_symbol predicate in
      ((Term.symbol_name sym),
       (Term.symbol_arity sym))

    let head_lit_node lit =
      pred_node (Ic.HeadLiteral.get_predicate lit)

    let body_lit_node lit =
      pred_node (Ic.BodyLiteral.get_predicate lit)

    let nonbuiltin_body_lits clause =
      List.filter 
	(fun lit -> 
	  match Ic.BodyLiteral.get_builtin lit with
	    Term.PUnbound ->
	      true
	  | Term.PDeterministic _ ->
	      false)
	(match Ic.get_body clause with
	  Ic.Body body_lits ->
	    body_lits)

    let iter_head_nodes f clause =
      match Ic.get_head clause with
	Ic.Unit lit ->
	  f (head_lit_node lit)
      | Ic.False ->
	  ()
      | Ic.Conjunction lits ->
	  List.iter
	    (fun lit -> f (head_lit_node lit))
	    lits
      | Ic.Disjunction head ->
	  List.iter
	    (fun lit -> f (head_lit_node lit))
	    (Ic.DisjunctiveHead.get_literals head)
    let iter_nonbuiltin_body_nodes f clause =
      List.iter
      	(fun lit -> f (body_lit_node lit))
	(nonbuiltin_body_lits clause)

    let make_dependency_pgraph clauses =
      (* pgraph, arrows -> is read "requires" *)
      let g = ref [] in
      List.iter
	(fun clause ->
	  iter_head_nodes
	    (fun hsym ->
	      iter_nonbuiltin_body_nodes
		(fun bsym ->
		  Listutils.push (hsym, bsym) g)
		clause)
	    clause)
	clauses;
      Listutils.sort_without_duplicates !g

    (** Select the clauses with [Fail] head, in which all body literals
       have predicates in [nodes]. *)
    let select_emptyhead_clauses_for_nodes nodes clauses =
      let result = ref [] in
      List.iter
	(fun clause ->
	  match Ic.get_head clause with
	    Ic.False ->
	      if List.for_all
		  (fun lit ->
		    List.mem (body_lit_node lit) nodes)
		  (nonbuiltin_body_lits clause) then
		Listutils.push clause result
	      else
		()

	  | _ ->
	      ())
	clauses;
      !result

    (** Select the clauses with a head literal with a predicate in
        [nodes]. *)
    let select_clauses_for_nodes nodes clauses =
      let result = ref [] in
      List.iter 
	(fun clause ->
	  match Ic.get_head clause with
	    Ic.False ->
	      ()
	  | Ic.Unit lit ->
	      if List.mem (head_lit_node lit) nodes then
		Listutils.push clause result
	      else
		()
	  | Ic.Conjunction lits ->
	      if List.exists
		  (fun lit -> List.mem (head_lit_node lit) nodes)
		  lits
	      then
		Listutils.push clause result
	  | Ic.Disjunction head ->
	      if List.exists
		  (fun lit -> List.mem (head_lit_node lit) nodes)
		  (Ic.DisjunctiveHead.get_literals head)
	      then
		Listutils.push clause result)
	clauses;
      !result

		
    (** Pair of list of node. Fst are the nonrecursive, Snd the
       recursive predicates. *)
    let dependency_sgraph_nonrecs_recs sgraph =
      let nonrecs = ref []
      and recs = ref []
      and tc = Graph.warshall sgraph in
      List.iter
	(fun (v, vs) ->
	  if List.mem v vs then
	    Listutils.push v recs
	  else
	    Listutils.push v nonrecs)
	tc;
      (!nonrecs, !recs)
      
    (** Return [nodes], which must be nonrecursive, in a toplogic
       ordering, such that each predicate depends only on predicates
       that appeared before in the result list. [pgraph] is the dependency
       graph as produced by [make_dependency_pgraph].
     *)
    let topsort_nonrecursives nodes pgraph =
      let pgraph1 = ref [] in
      List.iter
	(fun (v1, v2) ->
	  if List.mem v1 nodes then
	    Listutils.push (v2, v1) pgraph1
	  else
	    ())
	pgraph;
      Graph.top_sort (Graph.p_to_s_graph !pgraph1)
	
    let relevant_nodes node sgraph =
      let tc = Graph.warshall sgraph in
      try
	let (v, vs) = List.find (fun (v, vs) -> v = node) tc in
	Ordset.ord_insert vs v
      with Not_found ->
	[]
      
    let node_to_predicate symbol_table predicate_table (name, arity) =
      Predicate.find_predicate 
	predicate_table
	(Term.find_symbol symbol_table name arity)
  end    

type clauses = Ic.clause list

let rec negative_units_to_store table clauses =
  match clauses with
    [] -> []
  | clause :: clauses1 ->
      if 
	match Ic.get_head clause with
	  Ic.False ->
	    begin
	      match Ic.get_body clause with 
		Ic.Body [lit] ->
		  begin
		    match Ic.BodyLiteral.get_sign lit with Ic.Pos ->
		      let atom = Ic.BodyLiteral.get_atom_term lit in
		      NegativeUnits.insert table (Fact.make_input_fact atom);
		      true
		    | _ -> false
		  end
	      | _ -> false
	    end
	| _ -> false
      then
	(* Clause could be removed, if handling of positive input
	   units also considers the negatives. *)
        (* negative_units_to_store table clauses1 *)
	clause :: negative_units_to_store table clauses1
      else
	clause :: negative_units_to_store table clauses1

let make_eval_plain_clauses cfg predicate_table symbol_table neg_units terms = 
  let aux = Ic.Aux.make_aux cfg predicate_table symbol_table neg_units in
  List.map 
    (fun term -> Ic.make_clause aux term) 
    terms
        
let make_eval_inc_clauses plain_clauses delta_predicates = (*List.iter (fun x ->PrologPrint.print (Ic.get_source x);print_string "; ") plain_clauses;print_string " |||| ";*)
  List.fold_right 
    List.append
    (List.map 
       (fun clauses -> 
	 (Ic.clause_delta_clauses delta_predicates clauses))
       plain_clauses)
    []

let idb_predicates clauses =
  let preds = ref [] in
  List.iter
    (fun clause ->
      match Ic.get_head clause with
	Ic.False ->
	  ()
      | Ic.Unit lit ->
	  (* ### we may handle empty bodies specially here, but for now
	     a clause "p :- true" produces p in the 1st round. *)
	  Listutils.pushnewq (Ic.HeadLiteral.get_predicate lit) preds
      | Ic.Conjunction lits ->
          (* ### we may handle empty bodies specially here, but for now
	     a clause "p :- true" produces p in the 1st round. *)
          List.iter
	    (fun lit ->
	      Listutils.pushnewq (Ic.HeadLiteral.get_predicate lit)
		preds)
	    lits
      | Ic.Disjunction head ->
	  List.iter
	    (fun lit ->
	      Listutils.pushnewq (Ic.HeadLiteral.get_predicate lit) preds)
	    (Ic.DisjunctiveHead.get_literals head))  
    clauses;
  !preds
    
let select_relevant_clauses clauses goal_pred =
  let pdg = DependencyGraph.make_dependency_pgraph clauses in
  let dg = Graph.p_to_s_graph pdg in
  let nodes =
    DependencyGraph.relevant_nodes 
      (DependencyGraph.pred_node goal_pred)
      dg in
  List.append 
    (DependencyGraph.select_emptyhead_clauses_for_nodes nodes clauses)
    (DependencyGraph.select_clauses_for_nodes nodes clauses)

let split_clauses clauses =
  let pdg = DependencyGraph.make_dependency_pgraph clauses in
  let dg = Graph.p_to_s_graph pdg in
  let (nonrecs1, _) = DependencyGraph.dependency_sgraph_nonrecs_recs dg in
  let nonrecs2 = DependencyGraph.topsort_nonrecursives nonrecs1 pdg in
  let nonrec_clause_sets =
    List.map (fun node -> (node, ref [])) nonrecs2
  and rec_clauses = ref []
  and emptyhead_clauses = ref [] in
  List.iter
    (fun clause ->
      match Ic.get_head clause with
	Ic.Unit lit ->
	  begin
	    try 
	      Listutils.push 
		clause
		(List.assoc
		   (DependencyGraph.head_lit_node lit)
		   nonrec_clause_sets)
	    with Not_found ->
	      Listutils.push clause rec_clauses
	  end
      | Ic.False ->
	  Listutils.push clause emptyhead_clauses
      | Ic.Disjunction _ | Ic.Conjunction _ ->
	  raise (Err.Error 
		   "Split clauses for nonunit heads is not implemented."))
    clauses;
  (List.map (fun (_, clauses) -> !clauses) nonrec_clause_sets,
   !rec_clauses,
   !emptyhead_clauses)


(****************************************************************************)
(* Stratify *)	      
(****************************************************************************)

let iter_head_lits f clause =
  match Ic.get_head clause with
    Ic.False ->
      ()
  | Ic.Unit lit ->
      f lit
  | Ic.Conjunction lits ->
      List.iter	(fun lit -> f lit) lits
  | Ic.Disjunction head ->
      List.iter	(fun lit -> f lit) (Ic.DisjunctiveHead.get_literals head)

let iter_body_lits f clause =
  match Ic.get_body clause with
    Ic.Body lits ->
      List.iter f lits

let needs_stratification lit =
  match Ic.BodyLiteral.get_sign lit with
    Ic.StratifiedNegAsFailure -> true
(*  | Ic.StratifiedNegAsFailureNotSubsumed -> true *)
  | Ic.StratifiedNegAsFailureNotSubsumed -> false
  | Ic.Findall -> true
  | Ic.Neg2 -> false
  | Ic.Pos -> false

let head_lit_psym lit =
  let sym = Predicate.get_symbol (Ic.HeadLiteral.get_predicate lit) in
  ((Term.symbol_name sym),
   (Term.symbol_arity sym))

let body_lit_psym lit =
  let sym = Predicate.get_symbol (Ic.BodyLiteral.get_predicate lit) in
  ((Term.symbol_name sym),
   (Term.symbol_arity sym))
      
exception Not_stratifiable of string

let falsehead_clause_stratum table clause =
  let stratum = ref 1 in
  iter_body_lits
    (fun lit ->
      let v =
	try Hashtbl.find table (body_lit_psym lit)
	with Not_found -> -1 in
      stratum :=
	max (if needs_stratification lit then (v+1) else v) !stratum)
    clause;
  !stratum
	  
let clause_stratum table clause =
  match Ic.get_head clause with
    Ic.False ->
      falsehead_clause_stratum table clause
  | _ -> 
      let stratum = ref (-1) in
      iter_head_lits
	(fun lit ->
	  let v = Hashtbl.find table (head_lit_psym lit) in
	  if !stratum = -1 then
	    stratum := v
	  else
	    if !stratum <> v then
	      raise (Not_stratifiable 
		       ("Head literals with different strata in clause "
			^PrologPrint.print_to_string 
			    (Ic.get_source clause)^"."))
	    else
	      ())
	clause;
      !stratum
	      
let stratify clauses =
  let table = Hashtbl.create 32 in

  List.iter
    (fun clause ->
      iter_head_lits 
	(fun lit -> Hashtbl.replace table (head_lit_psym lit) 1)
	clause)
    clauses;	
  let num_of_preds = ref 0 in
  Hashtbl.iter (fun k v -> num_of_preds := !num_of_preds + 1) table;	
  let max_stratum = ref 1 in
  let changed = ref true in
  while !changed do		
    changed := false;
    List.iter
      (fun clause ->	
	iter_head_lits			
	  (fun plit ->
	    let p = (head_lit_psym plit) in
	    let vp = Hashtbl.find table p in	
	    iter_body_lits						
	      (fun qlit ->
		let q = (body_lit_psym qlit) in
		let vq = 
		  try Hashtbl.find table q			
		  with Not_found -> -1 in
		if vq = -1 then
		  ()								
		else
		  if needs_stratification qlit then
		    if vp <= vq then
		      let vq1 = (vq + 1) in
		      if vq1 > !num_of_preds then
			raise 
			  (Not_stratifiable 
			     ("Stratification failed for some predicate on \
			       which "
			       ^fst q^"/"^string_of_int (snd q)^" depends."))
		      else
			begin
			  Hashtbl.replace table p vq1;
			  max_stratum := max !max_stratum vq1;
			  changed := true
			end
		    else
		      ()
		  else
		    if vp < vq then
		      begin
			Hashtbl.replace table p vq;
			changed := true
		      end
		    else
		      ())
	      clause)
	  clause)
      clauses
  done;
  let strata = Array.make (!max_stratum + 1) [] in
  List.iter
    (fun clause -> (*PrologPrint.print (Ic.get_source clause);print_endline "!";*)
      let stratum = clause_stratum table clause in (*print_int stratum;print_endline "!";*)
      strata.(stratum-1) <- clause::strata.(stratum-1))
    clauses;
  let strata_list = ref [] in
  Array.iter
    (fun stratum ->
      if stratum = [] then
	()
      else
	Listutils.push stratum strata_list)
    strata;
  List.rev !strata_list


(****************************************************************************)
(* Sort Clauses *)	      
(****************************************************************************)

let ic_head_size ic =
  match Ic.get_head ic with
    Ic.False -> 0
  | Ic.Unit _ -> 1
  | Ic.Conjunction _ -> 1
  | Ic.Disjunction head -> 
      List.length (Ic.DisjunctiveHead.get_literals head)

let ic_body_size ic =
  match Ic.get_body ic with
    Ic.Body lits -> List.length lits

let sort_clauses ics =
  List.stable_sort
    (fun ic1 ic2 ->
      match compare (ic_head_size ic1) (ic_head_size ic2) with
	0 -> compare (ic_body_size ic1) (ic_body_size ic2)
      | c -> c)
    ics
    


