(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Index interface. *)

(** An index associates values of arbitrary type ("entries") 
with terms. *)
   
(** {6 Index Signature} *)

(** Signature of the index entry types.
*)
module type EntryType =
    sig

      (** The type of the entries. *)
      type entry

      val get_term : entry -> Term.term

    end

(** Signature of an index. *)
module type Index =
  sig
    
    (** The type of an index. *)
    type index

    (* ### How to express the relship of entry to EntryType in OCaml ? *)
    (** The type of the index entries. *)
    type entry
	  
    (** [make_index ()] returns a newly created index. *)
    val make_index : unit -> index

    (** [clear index] removes all elements from [index]. *)
    val clear : index -> unit
	
    (** [insert index term entry] inserts a new value, [entry], into
       [index], associated with key [term]. *)
    val insert : index -> Term.term -> entry -> unit

    (** [remove_variant index term context] deletes a single entry
       which has a variant of [term] as key from the
       index. It has no effect, if the index does not contain such an
       entry. 
     *) 
    val remove_variant : index -> Term.term -> unit

    (** [iter_all f index] applies [f] to all entries in the index. *)
    val iter_all : (entry -> unit) -> index -> unit

    (** [iter_all_and_throw_away f index] is similar to [iter_all f
       index], except, that the implementation is allowed to remove
       entries as they are iterated over from the index. The index
       must be reinitialized with [clear] if it is to be used
       afterwards.

       (Note: A kind of iteration, that allows to remove the
       current entry would be more desirable. See Java hashtables.)
       *) 
    val iter_all_and_throw_away : (entry -> unit) -> index -> unit

    (** [iter_instance_candidates f index term context] applies [f] to
       all those entries in the index, whose keys are variants or proper 
       instances of [term] in [context] , and maybe to more entries
       in the index.

       [context] may be [Term.null_context].
     *) 
    val iter_instance_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
	
    val iter_variant_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit	

    (** [iter_generalization_candidates f index term context] applies [f]
       to all those entries in the index, whose keys are variants or
       more general than [term] in [context], and maybe to more
       entries in the index.

       [context] may be [Term.null_context].
     *) 
    val iter_generalization_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit

    (** [iter_unification_candidates f index term] applies [f] to all those
       entries in the index, whose keys are unifiable with of [term] in
       [context], and maybe to more entries in the index.

       [context] may be [Term.null_context].
     *) 
    val iter_unification_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit

    val size : index -> int

  end
