(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let compare_symbols s1 s2 =
  if (s1 == s2) then
    0
  else 
    let c1 = compare (Term.symbol_name s1) (Term.symbol_name s2) in
    if c1 = 0 then
      compare (Term.symbol_arity s1) (Term.symbol_arity s2)
    else
      c1

let print_predicates_info predicate_table =
  let total = ref 0 in
  let preds = ref [] in
  Predicate.PredicateTable.iter
    (fun _ predicate ->
      preds := predicate :: !preds)
    predicate_table;
  List.iter 
    (fun predicate ->
      Printf.printf "%% %-30s [extent-size: %8d%s]\n"
	(Convert.symbol_to_string (Predicate.get_symbol predicate))
	(Predicate.extent_size predicate)
	(if Predicate.has_delta_predicate predicate 
	then " has-delta" 
	else "");
      total := !total + (Predicate.extent_size predicate))
    (List.sort 
       (fun p1 p2 -> compare_symbols 
	   (Predicate.get_symbol p1)
	   (Predicate.get_symbol p2))
       !preds);
  print_string "% Number of predicates: ";
  print_int (List.length !preds);
  print_endline "";
  print_string "% Total extent size: ";
  print_int !total;
  print_endline ""

let print_symbols_info symbol_table =
  let syms = ref [] in
  Term.symbol_table_iter
    (fun symbol ->
      syms := symbol :: !syms)
    symbol_table;
  List.iter
    (fun symbol ->
      Printf.printf "%% %-50s"
	(Convert.symbol_to_string symbol);
      begin
	match Term.symbol_predicate symbol with
	  Term.PUnbound ->
	    ()
	| _ -> print_string " [builtin predicate]"
      end;
      begin
	match Term.symbol_function symbol with
	  Term.FUnbound ->
	    ()
	| _ -> print_string " [evaluable function]"
      end;
      print_endline "")
    (List.sort compare_symbols !syms);
  print_string "% Number of symbols: ";
  print_int (List.length !syms);
  print_endline ""

(****************************************************************************)

let print_body_literal_info lit =
  begin
    match Ic.BodyLiteral.get_sign lit with
      Ic.Pos -> ()
    | _ -> print_string "SIGN "
  end;
  PrologPrint.print (Ic.BodyLiteral.get_atom_term lit);
  if (Ic.BodyLiteral.get_is_delta lit) then
    print_string " delta"
  else
    ();
  begin
    match Ic.BodyLiteral.get_builtin lit with
      Term.PUnbound -> ()
    | _ ->  print_string " builtin"
  end;
  print_string "\n";
  Ic.BodyLiteral.print_optim_info lit

let print_head_literal_info lit =
  PrologPrint.print (Ic.HeadLiteral.get_atom_term lit);
  print_endline ""
        
let print_clause_info clause =
  print_string "==================== CLAUSE ";
  print_string (Ic.get_id clause);
  print_endline " HEAD:";
  begin
    match (Ic.get_head clause) with
      Ic.False ->
	print_string "[false]\n"
    | Ic.Unit lit ->
	print_head_literal_info lit
    | Ic.Disjunction head ->
	List.iter
	  (fun lit ->
	    print_endline "--------------- DISJ";
	    print_head_literal_info lit)
	  (Ic.DisjunctiveHead.get_literals head)
    | Ic.Conjunction lits ->
	List.iter
	  (fun lit ->
	    print_endline "--------------- CONJ";
	    print_head_literal_info lit)
	  lits
  end;
  print_endline "= = = = = = = = = =  BODY:";	  
  match Ic.get_body clause with
    Ic.Body lits ->
      List.iter
	(fun lit ->
	  print_endline "--------------------";
	  print_body_literal_info lit)
	lits
	
    
let print_clauses_info clauses =
  List.iter print_clause_info clauses

(***************************************************************************)

let print_body_literal_as_term lit =
  begin
    match (Ic.BodyLiteral.get_sign lit) with
      Ic.Pos ->
	PrologPrint.print (Ic.BodyLiteral.get_atom_term lit)
    | Ic.StratifiedNegAsFailure ->
	print_string("\\+ ");
	PrologPrint.print (Ic.BodyLiteral.get_atom_term lit);
    | _ ->
	print_string("'$unspecified_sign'(");
	PrologPrint.print (Ic.BodyLiteral.get_atom_term lit);
	print_string(")")
  end
	
let print_clause_as_term clause =
  begin
    match (Ic.get_head clause) with
      Ic.False ->
	print_string "false"
    | Ic.Unit lit ->
	PrologPrint.print (Ic.HeadLiteral.get_atom_term lit)
    | Ic.Disjunction head ->
	begin
	  match (Ic.DisjunctiveHead.get_literals head) with
	    [] -> print_string "false"
	  | lit :: lits ->
	      PrologPrint.print (Ic.HeadLiteral.get_atom_term lit);
	      List.iter
		(fun lit ->
		  print_string " ; ";
		  PrologPrint.print (Ic.HeadLiteral.get_atom_term lit))
		lits
	end
    | Ic.Conjunction lits ->
	match lits with
	  [] -> print_string "false"
	| lit :: lits ->
	    PrologPrint.print (Ic.HeadLiteral.get_atom_term lit);
	    List.iter
	      (fun lit ->
		print_string " , ";
		PrologPrint.print (Ic.HeadLiteral.get_atom_term lit))
	      lits
  end;
  begin
    match Ic.get_body clause with
      Ic.Body([]) -> ()
    | Ic.Body(lit :: lits) ->
	print_endline " :-";
	print_string "        ";
	print_body_literal_as_term lit;
	List.iter
	  (fun lit ->
	    print_endline ", ";
	    print_string "        ";
	    print_body_literal_as_term lit)
	  lits
  end;
  print_endline ".\n"

let print_clauses_as_terms clauses =
  List.iter print_clause_as_term clauses

