(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

module type LayeredIndex =
  sig
    type entry
    type index
    val make_index : unit -> index
    val clear : index -> unit
    val insert : index -> Term.term -> entry -> unit
    val remove_variant : index -> Term.term -> unit
    val iter_all : (entry -> unit) -> index -> unit
    val iter_all_and_throw_away : (entry -> unit) -> index -> unit
    val iter_instance_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_variant_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_instance_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_variant_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit	
    val size : index -> int
	
    type state
    val obtain_state : index -> state
    val install_state : state -> unit
    val add_layer : state -> state
    val new_layer : index -> state

  end

module Make (LayerType : Index.Index) =
  struct

    type entry = LayerType.entry
    type layer = LayerType.index
    type index1 = (layer list) ref
    type index = (layer list) ref
    type state = index * (layer list)

    let mutable_layer index =
      match !index with
	layer :: _ -> layer
      | _ ->
	  raise (Failure "Attempt to access layered index without layer.")

    let only_layer index =
      match !index with
	[layer] -> layer
      | _ ->
	  raise (Failure "Layered index should have exactly one layer.")
	
    let make_index () = 
      ref [LayerType.make_index ()]
	    
    let clear index =
      LayerType.clear (only_layer index)

    let remove_variant index = 
      LayerType.remove_variant (only_layer index)

    let iter_all_and_throw_away f index =
      LayerType.iter_all_and_throw_away f (only_layer index)

    let insert index t e  =
      (*
      print_string "III ";
      print_int (List.length !index);
      print_endline "";
      *)	 
      LayerType.insert (mutable_layer index) t e

    let iter_all f index =
      List.iter (fun layer -> LayerType.iter_all f layer) !index

    let iter_instance_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_instance_candidates f layer t c)
	!index

	let iter_instance_candidates_ml f index t c =
		LayerType.iter_instance_candidates f (mutable_layer index) t c	
	
    let iter_variant_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_variant_candidates f layer t c)
	!index	
	
	let iter_variant_candidates_ml f index t c =
		LayerType.iter_variant_candidates f (mutable_layer index) t c
	
		
    let iter_generalization_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_generalization_candidates f layer t c)
	!index

	
	let iter_generalization_candidates_ml f index t c =
		LayerType.iter_generalization_candidates f (mutable_layer index) t c	
	
    let iter_unification_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_unification_candidates f layer t c)
	!index

	let iter_unification_candidates_ml f index t c =
		LayerType.iter_unification_candidates f (mutable_layer index) t c	
	
    let size index =
      List.fold_left (fun x y -> x + (LayerType.size y)) 0 !index

    let obtain_state index =
      (index, !index)
	
    let install_state (index, content) =
      index := content

    let add_layer (index, content) = 
      (index, (LayerType.make_index ()) :: content)
	
    let new_layer index =
      (index, [LayerType.make_index ()])

  end
