(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Indexes with undoable modifications. *)

(** Extends Index with support for undoable modifications. The
   interface for undoable modifications can be explained with the help
   of considering a possible implementation:

   An index is a store that is destructively modified. A layered index
   is a reference to a list of indexes. The list is not modified
   destructively. Read access on a layered index is performed on all
   the list elements. Write access on the first element. A state is
   represented by a list/reference pair. obtain_state returns such a
   state for a given layered index. install_state sets the reference value to
   the list. add_layer returns a new pair with the same reference and
   a list that is the cons of a fresh index with the input list. 

   [Note that a program may perform modifications on a layered index
   that shares a part of its list with a state that has been extracted
   previously with obtain_state. When this state is installed later,
   the modifications in the shared indexes remain in effect. This property
   is no longer used, see tableau.ml.]
*) 

module type LayeredIndex = 
  sig type entry type index
      
    (** The index is created with a single initial layer. *)
    val make_index : unit -> index
	
    (** Undefined if the index contains more than the single initial layer. *)
    val clear : index -> unit
    val insert : index -> Term.term -> entry -> unit

    (** Undefined if the index contains more than the single initial layer. *)
    val remove_variant : index -> Term.term -> unit

    val iter_all : (entry -> unit) -> index -> unit
    val iter_all_and_throw_away : (entry -> unit) -> index -> unit
    val iter_instance_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_variant_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_instance_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_variant_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
	
	
    val size : index -> int

    type state
    val obtain_state : index -> state
    val install_state : state -> unit
    val add_layer : state -> state
    val new_layer : index -> state
	
end

(** Functor for building a layered index implementation based on an
implementation of [Index.Index]. *)
module Make (LayerType : Index.Index) : 
    (LayeredIndex with type entry = LayerType.entry)








