(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(** Comparison of terms according to the lexicographic recursive path
ordering of terms (LRPO). *)

(** [is_subterm term1 term2] tests whether term1 is a subterm of term2 and returns
* true or false accordingly. *)	
val is_subterm :
	Term.term -> Term.term -> bool
	
(** [functor_symbol_precedence_compare term1 term2] compares term1 and term2
	according to the symbol precedence order of the leading symbols (functors in case
	of compounds, names in case of constants and the terms themselves otherwise). It returns:
    -  -1 if the functor of term1 is less than the functor of term2.

    -  0 if the functor of term1 and the functor of term2 are equal.

    -  1 if the functor of term1 is greater than the functor of term2.
	Symbols are ordered by arity, then by name (for example: f/3 > a/2 > f/2 > g/2 > g/1).
*)		
val functor_symbol_precedence_compare :
	Term.term -> Term.term -> int

(** [lrpo_greater term1 term2] tests whether term1 is greater than term2 in the LRPO
and returns true or false. *)	
val lrpo_greater :
	Term.term -> Term.term -> bool

(** [lrpo_greater_unshared_vars term1 term2] works like [lrpo_greater], except that
the terms are treated as havin different variable numbering domains; i.e. a variable
from [term1] will not count as occurring in [term2] and vice versa, even if both terms
have a variable with the same number. *)	
val lrpo_greater_unshared_vars : Term.term -> Term.term -> bool	

(** [lrpo_greater_shared_vars term1 term2] works like [lrpo_greater], except that
before the actual LRPO-test a filtering is applied: if the variables of [term2] are not
a subset of the variables of [term1], then [lrpo_greater_shared_vars] returns false. *)	
val lrpo_greater_shared_vars : Term.term -> Term.term -> bool

(** [ground_compare_lrpo_greater term1 term2] compares term1 and term2 according to the LRPO
and returns 1 iff term1>term2, 0 iff term1=term2 and -1 iff term1<term2. The input terms
should be ground - the function will raise Err.Error if the terms cannot be compared. *)		
val ground_compare_lrpo_greater : Term.term -> Term.term -> int	
	


	
	
(** [ground_like_functor_symbol_precedence_compare] works the same as [functor_symbol_precedence_compare],
except for treating variables as lowest class constants, and comparing variables by their variable
numbers. *)	
val ground_like_functor_symbol_precedence_compare :
		Term.term -> Term.term -> int	

(** [ground_like_lrpo_greater] works the same as [lrpo_greater], except for treating variables as lowest 
class constants, and comparing variables by their variable numbers. *)
val ground_like_lrpo_greater :
	Term.term -> Term.term -> bool

	
(** [ground_like_compare_lrpo_greater] works the same as [ground_like_compare_lrpo_greater], except for
treating variables as lowest class constants, and comparing variables by their variable numbers. *)	
val ground_like_compare_lrpo_greater : Term.term -> Term.term -> int		
		
(** [unit_lrpo_greater term1 term2 symbol_table] works analogous to lrpo_greater but is intended
to handle unit clauses instead of terms. Uses the multiset-extension of the LRPO. Needs the equality
symbol as input in order to distinguish between equations and non-equations. *)	
val unit_lrpo_greater :
	Term.term -> Term.term -> Term.symbol -> bool

(** [eq_lrpo_greater l1 r1 l2 r2] works analogous to lrpo_greater but is intended
to handle equations instead of any terms. Uses the multiset-extension of the LRPO. Compares
the multisets {l1, r1} and {l2, r2}. *)		
val eq_lrpo_greater :
	Term.term -> Term.term -> Term.term -> Term.term -> bool	

				
(** [lrpo_greater_c term1 context1 term2 context2] works the same as [lrpo_greater] except for
taking non-dereferenced terms. *)	
val lrpo_greater_c :
	Term.term -> Term.context -> Term.term -> Term.context -> bool
	
