(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "

(****************************************************************************)
(* Notes *)
(****************************************************************************)
(* 
==============================================================================
There are different possiblilities to arrange subsumption:
==============================================================================

   1.
   At deriving:
     subs_fwd(extent)
     subs_fwd(news)
     subs_bwd(news)
   At inserting:
     subs_bwd(extent)


   2.
   At deriving:
     subs_fwd(news)  (or do nothing)
   At inserting
     subs_fwd(extent)
     subs_bwd(extent)
     subs_bwd(delta)

==============================================================================
Disjunctions...
==============================================================================

------------------------------------------------------------------------------
At insert_unit
------------------------------------------------------------------------------

     - may (must?) also consider also subsumption of a unit by
       - a delayed unit
       - a disjunction
     - delayed unit subsumer must be lighter or equal under a term
       weight that counts vars lighter or equal to other symbols.
     - only a factor of a disjunction can subsume a units
     - if disjunctions are ground, such a factor would have converted the
       disjunction to a unit before
     - probably not "must" - see hypertableau properties with regularity


==============================================================================
Proofs
==============================================================================


Redirection of proofs in case of back subsumption is not so
simple: consider p(X) :- p(a), which may lead to a loop.

*)


module NewsEntry =
  struct
    type entry = (Predicate.predicate * Fact.fact)
    let get_term (_ , fact) = Fact.get_term fact
    let get_predicate = fst
  end

module Disjunct =
  struct
    type disjunct = 
	Disjunct of Predicate.predicate * Fact.fact * Fact.fact list * Term.term * Term.term array * Term.term array
      | WpSemanticsTrue

    let make_disjunct p f c i s u = Disjunct (p, f, c, i, s, u) 	  

    let is_wp_semantics_true disjunct =
      match disjunct with
	WpSemanticsTrue ->
	  true
      | Disjunct _ -> false

    let get_term disjunct =
      match disjunct with 
	Disjunct (_ , fact, _, _, _, _) -> 
	  Fact.get_term fact
      | _ -> raise (Failure "Invalid disjunct access.")

    let get_fact disjunct =
      match disjunct with 
	Disjunct (_ , fact, _, _, _, _) -> fact
      | _ -> raise (Failure "Invalid disjunct access.")

    let get_predicate disjunct =
      match disjunct with
	Disjunct (pred , _, _, _, _, _) -> pred
      | _ -> raise (Failure "Invalid disjunct access.")

    let get_complements disjunct =
      match disjunct with
	Disjunct (_, _, comp, _, _, _) -> comp 
      | _ -> raise (Failure "Invalid disjunct access.")
      
    let get_impure_term disjunct =
      match disjunct with
	Disjunct (_, _, _, it, _, _) -> it 
      | _ -> raise (Failure "Invalid disjunct access.")
    
    let get_impure_bindings disjunct =
      match disjunct with
	Disjunct (_, _, _, _, sb, _) -> sb 
      | _ -> raise (Failure "Invalid disjunct access.")
      
    let get_pure_bindings disjunct =
      match disjunct with
	Disjunct (_, _, _, _, _, ub) -> ub 
      | _ -> raise (Failure "Invalid disjunct access.")
  end

type delayed_clause =
    DelayedDisjunction of Disjunct.disjunct list * int

type delayed_clauses = delayed_clause list

let get_weight delayed_clause =
  match delayed_clause with
    DelayedDisjunction (_, w) -> w
	
let get_disjuncts delayed_clause =
  match delayed_clause with
    DelayedDisjunction (ds, _) -> ds

module NewsTable =
  DtIndex.Make(NewsEntry)(DtIndex.HashTableImplementation)

  
let id = ref 0

let gen_id () =
  id := !id + 1;
  "news_"^string_of_int(!id)

let reset_gen_id () =
  id := 0  
  
type news = {unit_table:NewsTable.index;
		delayed_unit_table:NewsTable.index;
		news_id:string; 
	     mutable delayed_clauses:delayed_clauses;
	     mutable impure_delayed_clauses:delayed_clauses;
	     cfg_back_subsumption:bool;
	     cfg_back_subsumption_within_news:bool;
	     mutable max_literal_weight:int;
	     cfg_wp_semantics:bool;
	     cfg_show_derived_literals:bool;
	     cfg_full_branch_redundancy:bool;
	     mutable unit_table_size:int;
	     mutable delayed_unit_table_size:int;
	     term_weighing_method:Weight.term_weighing_method;
	     mutable weight_limit_exceeded:int;
	     cfg_ehyper:bool;
	     krh_dom_symbol:Term.symbol
	   }
	
let set_delayed_clauses news delayed_clauses =
	news.delayed_clauses <- delayed_clauses
	
let get_delayed_clauses news =
	news.delayed_clauses	 
	
(*let set_delayed_facts news delayed_facts =
	news.delayed_facts <- delayed_facts
	
let get_delayed_facts news =
	news.delayed_facts		
*)	
let set_impure_delayed_clauses news impure_delayed_clauses =
	news.impure_delayed_clauses <- impure_delayed_clauses
	
let get_impure_delayed_clauses news =
	news.impure_delayed_clauses	  
	   
let set_max_literal_weight news weight =
  news.max_literal_weight <- weight

let reset_weight_limit_exceeded news =
  news.weight_limit_exceeded <- 0

let unit_table_size news =
  news.unit_table_size

let delayed_unit_table_size news =
  news.delayed_unit_table_size  
  
let delayed_clauses_size news =
  (List.length news.delayed_clauses)

let weight_limit_exceeded news =
  news.weight_limit_exceeded

let make_news cfg krh_dom_symbol = (*print_endline "Making news.";*)
  {unit_table=NewsTable.make_index ();
  delayed_unit_table=NewsTable.make_index ();
  news_id=gen_id ();
   delayed_clauses=[];
   impure_delayed_clauses=[];
   cfg_back_subsumption = Cfg.get_back_subsumption_flag cfg;
   cfg_back_subsumption_within_news =
   Cfg.get_back_subsumption_within_news_flag cfg;
   max_literal_weight=Cfg.get_max_weight_initial_param cfg;
   cfg_wp_semantics=Cfg.get_wp_semantics_flag cfg;
   cfg_show_derived_literals=Cfg.get_show_derived_literals_flag cfg;
   cfg_full_branch_redundancy=Cfg.get_full_branch_redundancy_flag cfg;
   unit_table_size=0;
   delayed_unit_table_size=0;
   term_weighing_method=
   (match Cfg.get_weighing_method_param cfg with
     0 -> Weight.SymbolCount
   | 1 -> Weight.Depth
   | _ -> raise
	 (Err.Error "Bad value for term_weighing_method parameter."));
   weight_limit_exceeded=0;
   cfg_ehyper=Cfg.get_ehyper cfg;
   krh_dom_symbol=krh_dom_symbol
 }
    
let null_news = 
  {unit_table=NewsTable.make_index ();
   delayed_unit_table=NewsTable.make_index ();
   delayed_clauses=[];
   impure_delayed_clauses=[];
   news_id=gen_id ();
   cfg_back_subsumption=false;
   cfg_back_subsumption_within_news=false;
   max_literal_weight=0;
   cfg_wp_semantics=false;
   cfg_show_derived_literals=false;
   cfg_full_branch_redundancy=false;
   unit_table_size=0;
   delayed_unit_table_size=0;
   term_weighing_method=Weight.SymbolCount;
   weight_limit_exceeded=0;
   cfg_ehyper=false;
   krh_dom_symbol=Term.gensym ()
 }

let get_news_id news =
	news.news_id
	
let get_cfg_ehyper news =
	news.cfg_ehyper
	
let get_cfg_show_derived_literals news =
	news.cfg_show_derived_literals
	
let get_cfg_full_branch_redundancy news =
	news.cfg_full_branch_redundancy
	
(****************************************************************************)
(** Handling of Undoable Modifications *)
(****************************************************************************)

(* Notes:
   - delayed_clauses is a list that is not destructively modified.
     Each version in the "layers" contains all of the clauses. (In
     contrast to index representations that are composed by several
     indexes in each of the layers.)
   - unit_table is emptied after each round, so its modifications
     need not to be undoable.
*)

type state = news * delayed_clauses 

let obtain_state news =
  (news, news.delayed_clauses)

let install_state (news, delayed_clauses) =
  news.delayed_clauses <- delayed_clauses


(****************************************************************************)
(** Support for Units *)
(****************************************************************************)

exception Found

let remove_subsumed_from_extent predicate reader new_fact =
  let new_term = (Fact.get_term new_fact)
  and to_remove = ref [] in
  Access.Reader.iter_instance_candidates
    (fun old_fact ->
      if Term.subsumes_chk new_term (Fact.get_term old_fact) then
	Listutils.push old_fact to_remove
      else
	())
    reader
    new_term
    Term.null_context;
  List.iter
    (fun old_fact ->
      Predicate.remove_variant predicate (Fact.get_term old_fact))
    !to_remove

let remove_subsumed_from_news news new_fact =
  let new_term = (Fact.get_term new_fact)
  and to_remove = ref [] in
  NewsTable.iter_instance_candidates
    (fun old_fact ->
      if Term.subsumes_chk new_term (NewsEntry.get_term old_fact) then
	Listutils.push old_fact to_remove
      else
	())
    news.unit_table
    new_term
    Term.null_context;
  List.iter
    (fun old_fact ->
      NewsTable.remove_variant news.unit_table (NewsEntry.get_term old_fact);
      news.unit_table_size <- news.unit_table_size - 1)
    !to_remove
    
let remove_subsumed_from_delayed_news news new_fact =
  let new_term = (Fact.get_term new_fact)
  and to_remove = ref [] in
  NewsTable.iter_instance_candidates
    (fun old_fact ->
      if Term.subsumes_chk new_term (NewsEntry.get_term old_fact) then
	Listutils.push old_fact to_remove
      else
	())
    news.delayed_unit_table
    new_term
    Term.null_context;
  List.iter
    (fun old_fact ->
      NewsTable.remove_variant news.delayed_unit_table (NewsEntry.get_term old_fact);
      news.delayed_unit_table_size <- news.delayed_unit_table_size - 1)
    !to_remove    


let unit_is_subsumed_by_extent predicate term =
  try
    Access.Reader.iter_generalization_candidates 
      (fun fact -> 
	if Term.subsumes_chk (Fact.get_term fact) term then
	  raise Found
	else
	  ())
(*      (Predicate.get_all_bound_reader predicate)*)
	(Predicate.default_reader predicate)
      term
      Term.null_context;
    false
  with Found ->
    true

let unit_is_subsumed_by_new_units news term =
  try
    NewsTable.iter_generalization_candidates 
      (fun fact -> 
	if Term.subsumes_chk (NewsEntry.get_term fact) term then
	  raise Found
	else
	  ())
      news.unit_table
      term
      Term.null_context;
    false
  with Found ->
    true

    
let unit_is_subsumed_by_delayed_units news term =
  try
    NewsTable.iter_generalization_candidates 
      (fun fact -> 
	if Term.subsumes_chk (NewsEntry.get_term fact) term then
	  raise Found
	else
	  ())
      news.delayed_unit_table
      term
      Term.null_context;
    false
  with Found ->
    true    

let insert_unit news predicate fact = 
  (* ??? ### also check subsumption by delayed disjunction units ? *)
  let term = (Fact.get_term fact) 
  in	(*ppp term;pe ".";*)
	if news.max_literal_weight != -1 &&
	    (let w =
	    	if (Convert.term_functor_symbol term)==news.krh_dom_symbol then
	    		(2*(Weight.term_weight
		       news.term_weighing_method
		       term) - news.max_literal_weight) 
	    	else
	    		((Weight.term_weight
		       news.term_weighing_method
		       term) - news.max_literal_weight)
		 in
		    if w > 0 then (	
				news.weight_limit_exceeded <- 
			  		if news.weight_limit_exceeded > 0 then
			    		min news.weight_limit_exceeded w
			  		else w;
					true
			)			
		    else false) 
		then 
			(if (news.cfg_ehyper) then 
				(if ((unit_is_subsumed_by_extent predicate term)
	  				||(unit_is_subsumed_by_new_units news term) 
	  				||(unit_is_subsumed_by_delayed_units news term))
	  				then ()
	  			else
	  				(if news.cfg_back_subsumption_within_news then
						(remove_subsumed_from_delayed_news news fact)
	  				else ();
	  				NewsTable.insert news.delayed_unit_table term (predicate, fact);
	  				news.delayed_unit_table_size <- news.delayed_unit_table_size + 1))
	  		else ())
  	else (	
  		if (unit_is_subsumed_by_extent predicate term)
  			||(unit_is_subsumed_by_new_units news term) 
  			then ()
  		else ( 
			if news.cfg_back_subsumption_within_news then
				remove_subsumed_from_news news fact
  			else ();		
			NewsTable.insert news.unit_table term (predicate, fact);
  			news.unit_table_size <- news.unit_table_size + 1
   		)
   	)
   	
(*   	
let insert_rewritten_unit news predicate fact = 
  let term = (Fact.get_term fact) 
  in 						(*PrologPrint.print term;print_string " in ";print_endline news.news_id;*)
	if (unit_is_subsumed_by_extent predicate term)
		||(unit_is_subsumed_by_new_units news term) 
		then ()
	else ( 		(*PrologPrint.print (Fact.get_term fact);print_endline ".";*)
	if news.cfg_back_subsumption_within_news then
		remove_subsumed_from_news news fact
		else ();		(*print_int news.unit_table_size;*)
	NewsTable.insert news.unit_table term (predicate, fact);
	news.unit_table_size <- news.unit_table_size + 1	(*;print_int news.unit_table_size*)
	)
*)  	

(****************************************************************************)
(* Delayed Stuff and Disjunctions *)
(****************************************************************************)

let rec ord_insert_to_delayed_clauses set el =
  match set with
    [] -> 
      [el]
  | h :: t ->
      let v = compare (get_weight h) (get_weight el) in
      if v < 0 then
	h :: ord_insert_to_delayed_clauses t el
      else
	el :: set

let disjunction_is_subsumed_by_extent news disjunction =
  if news.cfg_wp_semantics then
    List.for_all
      (fun disjunct ->
			unit_is_subsumed_by_extent 
			  (Disjunct.get_predicate disjunct)
			  (Disjunct.get_term disjunct))
      (get_disjuncts disjunction)
  else
    List.exists
      (fun disjunct ->
			unit_is_subsumed_by_extent 
			  (Disjunct.get_predicate disjunct)
			  (Disjunct.get_term disjunct))
      (get_disjuncts disjunction)
      
let disjunction_is_subsumed_by_new_units news disjunction =
  if news.cfg_wp_semantics then
    List.for_all
      (fun disjunct ->
			unit_is_subsumed_by_new_units news (Disjunct.get_term disjunct))
      (get_disjuncts disjunction)
  else
    List.exists
      (fun disjunct ->
			unit_is_subsumed_by_new_units news (Disjunct.get_term disjunct))
      (get_disjuncts disjunction)

let insert_disjunction news ptwisucs = 
	let rejected_eqs = ref []
	in
  if news.max_literal_weight != -1 &&
    (let w = (List.fold_left (fun w1 (_,_,w2,_,_,_,_) -> max w1 w2) 0 ptwisucs) -
	news.max_literal_weight in
    if w > 0 then
      ( 
		news.weight_limit_exceeded <- 
		  if news.weight_limit_exceeded > 0 then
		    min news.weight_limit_exceeded w
		  else
		    w;
		true
      )
    else
      (false)) 
  then 
  	(if (news.cfg_ehyper) then 
  		(List.iter
  			(fun (predicate, fact, _,_,_,_,_) ->
  				let term = Fact.get_term fact
  				in
	  				if (Fact.is_equation fact) then
	  					(rejected_eqs := term :: !rejected_eqs)
	  				else
						(if ((unit_is_subsumed_by_extent predicate term)
								||(unit_is_subsumed_by_new_units news term) 
								||(unit_is_subsumed_by_delayed_units news term))
								then ()
						else
							(NewsTable.insert news.delayed_unit_table term (predicate, fact);
							news.delayed_unit_table_size <- news.delayed_unit_table_size + 1)))
			ptwisucs)
	 else ())
  else
    (let weight =  List.fold_left (fun w1 (_,_,w2,_,_,_,_) -> w1 + w2) 0 ptwisucs
    in
    let disjunction =
      DelayedDisjunction 
	((List.map (fun (p,t,w,i,s,u,c) -> Disjunct.make_disjunct p t c i s u) ptwisucs), weight) 
	in
    (* "Heuristic" early subsumption. Actual subsumption can only happen
     when attaching to the tableau. *)
    if disjunction_is_subsumed_by_extent news disjunction then
      ()
    else if disjunction_is_subsumed_by_new_units news disjunction then
      ()
    else
      (news.delayed_clauses <-
	ord_insert_to_delayed_clauses news.delayed_clauses disjunction));
 !rejected_eqs
	  
(* replaces last "else"	
	else (
		if news.cfg_ehyper && (List.length shared_varnums)>0 then
			(news.impure_delayed_clauses <-
				ord_insert_to_delayed_clauses news.impure_delayed_clauses disjunction)
		else
			(news.delayed_clauses <-
				ord_insert_to_delayed_clauses news.delayed_clauses disjunction))
  *)
(****************************************************************************)
(* Process News *)
(****************************************************************************)

let insert_to_extent_and_delta news predicate fact delta_count new_facts = (*print_string " Inserting term!";*)
  if news.cfg_back_subsumption then
    remove_subsumed_from_extent
      predicate 
      (Predicate.get_all_bound_reader predicate)
      fact
  else
    ();
  let term = (Fact.get_term fact) in

  Predicate.insert predicate term fact;

  if news.cfg_show_derived_literals then
    begin
      print_string "derived_fact(";
      PrologPrint.print term;
      print_endline ")."
    end;

  if Predicate.has_delta_predicate predicate then (
    let delta_predicate = Predicate.get_delta_predicate predicate in
    Predicate.insert delta_predicate term fact
	)
  else
    ();
    delta_count := !delta_count + 1;
    new_facts := fact :: !new_facts

		
					
let process_news news =	(*print_int news.unit_table_size;print_endline ".";print_endline "Size.";*)
  let new_facts = ref []
  in (*print_int news.unit_table_size;print_endline "."; *)
  NewsTable.iter_all_and_throw_away
    (fun (predicate, fact) ->	(*PrologPrint.print (Fact.get_term fact);*)
		if news.cfg_back_subsumption then
			(remove_subsumed_from_extent
				predicate 
				(Predicate.get_all_bound_reader predicate)
				fact)
  		else ();
  		new_facts := (predicate, fact) :: !new_facts
   		(*insert_to_extent_and_delta news predicate fact delta_count new_facts*)
    )
    news.unit_table;
  NewsTable.clear news.unit_table;
  news.unit_table_size <- 0;		(*print_int !delta_count;print_endline ".";*)
  !new_facts

let process_delayed_news news =
  let new_facts = ref []
  in (*print_int news.unit_table_size;print_endline "."; *)
  NewsTable.iter_all_and_throw_away
    (fun (predicate, fact) ->	(*PrologPrint.print (Fact.get_term fact);*)
		if news.cfg_back_subsumption then
			(remove_subsumed_from_extent
				predicate 
				(Predicate.get_all_bound_reader predicate)
				fact)
  		else ();
  		new_facts := (predicate, fact) :: !new_facts
   		(*insert_to_extent_and_delta news predicate fact delta_count new_facts*)
    )
    news.delayed_unit_table;
  NewsTable.clear news.delayed_unit_table;
  news.delayed_unit_table_size <- 0;		(*pe "ALMOST";*)
  !new_facts
  
let clear_news news =
  NewsTable.clear news.unit_table;
  news.unit_table_size <- 0

let clear_delayed_news news =
  NewsTable.clear news.delayed_unit_table;
  news.delayed_unit_table_size <- 0
  
(****************************************************************************)
(* Select Delayed Disjunction *)
(****************************************************************************)

(********************)

let rec pick_disjunction_std news =
  match news.delayed_clauses with
    [] -> raise Not_found
  | clause :: clauses ->
      news.delayed_clauses <- clauses;
      if disjunction_is_subsumed_by_extent news clause then
	pick_disjunction_std news
      else
	clause

(********************)

let rec reduce_disjuncts_for_wp_semantics disjuncts =
  let sst_occured = ref false in
  let disjuncts1 = 
    reduce_disjuncts_for_wp_semantics_1 disjuncts sst_occured in
  if !sst_occured then
    Disjunct.WpSemanticsTrue :: disjuncts1
  else
    disjuncts1
  
and reduce_disjuncts_for_wp_semantics_1 disjuncts sst_occured =
  match disjuncts with
    [] ->
      []
  | disjunct :: disjuncts1 ->
      if unit_is_subsumed_by_extent
	  (Disjunct.get_predicate disjunct)
	  (Disjunct.get_term disjunct) then
	begin
	  sst_occured := true;
	  reduce_disjuncts_for_wp_semantics_1 disjuncts1 sst_occured
	end
      else
	disjunct :: 
	reduce_disjuncts_for_wp_semantics_1 disjuncts1 sst_occured
  
let rec pick_disjunction_wp_semantics news =
  match news.delayed_clauses with
    [] -> raise Not_found
  | clause :: clauses ->
      news.delayed_clauses <- clauses;
      match clause with
	DelayedDisjunction (disjuncts, x) ->
	  let reduced_clause =
	    reduce_disjuncts_for_wp_semantics disjuncts in
	  match reduced_clause with
	    [] | [Disjunct.WpSemanticsTrue] ->
	      pick_disjunction_wp_semantics news
	  | _ -> 
	      DelayedDisjunction (reduced_clause, x)

let pick_disjunction news =
  if news.cfg_wp_semantics then
    pick_disjunction_wp_semantics news
  else
    pick_disjunction_std news

(****************************************************************************)

let sel_fd_clause_equal news c1 c2 =
  (* An unnecessary fail of sel_fd_clause_equal has no effect on
     correctness/completeness. *)
  let disjuncts1 = get_disjuncts c1
  and disjuncts2 = get_disjuncts c2 in
  let disjuncts3 =
    if news.cfg_wp_semantics then
      (* ### With WP there could be a "partial subsumption", i.e.
	 a;b;c;d will become a;b;<true> if c and d are present.
	 This is not considered here, i.e. equal fails for them.
       *)
      reduce_disjuncts_for_wp_semantics disjuncts2
    else
      disjuncts2 in
  List.length disjuncts1 = List.length disjuncts3 &&
  List.for_all2 
    (fun e1 e2 -> 
      (* Or perhaps use term_variant to handle disjuncts with variables,
	 which had different history (e.g. were produced with different
	 literal ordering before renumber_variables). *)
      Term.term_ident (Disjunct.get_term e1) (Disjunct.get_term e2))
    disjuncts1 disjuncts3

(* ### TODO just check up to weight... *)



