(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Storing and processing of newly derived facts (similar to a SOS). *)

(* # TODO ... bulk/single implementations, fact types, disjunction *)

(** The type of [news] objects. *)
type news

module Disjunct :
  sig
    type disjunct = 
    	Disjunct of Predicate.predicate * Fact.fact * Fact.fact list * Term.term * Term.term array * Term.term array
      | WpSemanticsTrue
    val is_wp_semantics_true : disjunct -> bool
    val get_fact : disjunct -> Fact.fact
    val get_predicate : disjunct -> Predicate.predicate
    val get_complements : disjunct -> Fact.fact list
    val get_term : disjunct -> Term.term
    val get_impure_term : disjunct -> Term.term
    val get_impure_bindings : disjunct -> Term.term array
    val get_pure_bindings : disjunct -> Term.term array
    val make_disjunct : Predicate.predicate -> Fact.fact -> 
    					(Fact.fact list) -> Term.term -> 
    					Term.term array -> Term.term array -> disjunct
  end

type delayed_clause =
	DelayedDisjunction of (Disjunct.disjunct list) * int

val get_news_id : news -> string

val get_cfg_ehyper : news -> bool

val get_cfg_show_derived_literals : news -> bool

val get_cfg_full_branch_redundancy : news -> bool

val set_delayed_clauses : news -> delayed_clause list -> unit
	
val get_delayed_clauses : news -> delayed_clause list

val set_impure_delayed_clauses : news -> delayed_clause list -> unit
	
val get_impure_delayed_clauses : news -> delayed_clause list

(** Create a [news] object. *)
val make_news : Cfg.cfg -> Term.symbol -> news

(** An unusable [news] object to initialize typed fields. *)
val null_news : news

val set_max_literal_weight : news -> int -> unit

val reset_weight_limit_exceeded : news -> unit

val unit_table_size : news -> int

val delayed_unit_table_size : news -> int

val delayed_clauses_size : news -> int 

(** Is > 0 iff an inferred clause has been discarded since it exceeded
  the weight limit. The returned amount is the difference from the weight of
  the lightest such discarded clause to the limit.
*)
val weight_limit_exceeded : news -> int

(** Insert a fact into a [news] object. *)
val insert_unit : news -> Predicate.predicate -> Fact.fact -> unit

(** Insert a rewritten fact into a [news] object. Does not use a term weight check. *)
(*val insert_rewritten_unit : news -> Predicate.predicate -> Fact.fact -> unit*)

(** Insert a disjunction into a [news] object. The disjunction is
    represented as a list of length > 1 of quadruples. A quadruple
    contains: the predicate, the atom, the weight of the atom, and
    a list of terms used for complement splitting.
    A possibly empty list is returned. This list contains the equational
    terms of the disjunction iff the disjunction exceeds the weight limit.
    As equational terms must be turned into Ic.clause type records, the news object cannot
    handle them.
  *)       
val insert_disjunction : news -> 
  (Predicate.predicate * Fact.fact * int * Term.term * Term.term array
   * Term.term array * (Fact.fact list)) list -> Term.term list

(** [process_news news] processes facts in [news]. Processed facts
   are removed from [news]. 

   The number of actually new facts is returned, i.e. processed facts
   that are not subsumed by previously derived facts.

   The precise meaning of [process_news]  depends on the particular strategy.

   - For seminaive delta iteration, all facts in news are removed and, if
     they are not redundant by subsumption, are asserted as facts and
     as delta facts.
*)
val process_news : news -> (Predicate.predicate * Fact.fact) list

val process_delayed_news : news -> (Predicate.predicate * Fact.fact) list

(** Remove all unit facts from news. To be called in place of
   process_news when the facts should be thrown away (e.g. after
   branch closing) , but the news object should be used again (e.g. for
   alternative branches). *) 
val clear_news : news -> unit

val clear_delayed_news : news -> unit




val get_disjuncts : delayed_clause -> Disjunct.disjunct list
val get_weight : delayed_clause -> int

(** Raises [Not_found] if no disjunction that is not subsumed by a unit in
   the extent is found. 
 *)
val pick_disjunction : news -> delayed_clause

type state

val obtain_state : news -> state

val install_state : state -> unit

val unit_is_subsumed_by_extent : Predicate.predicate -> Term.term -> bool