(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Data structures used for optimization. *)

type bound =
    Free
  | Bound

      
type binding_pattern =
    bound array


(** True iff the first binding pattern is stronger or equal to the second,
in each position, where Ground is stronger than Bound and
Bound is stronger than Free.

Both arguments are assumed to have the same length. *)
val binding_pattern_covers : binding_pattern -> binding_pattern -> bool

type mode =
    InputGround (** A parameter that must be ground at input time. *)
  | OutputGround (** A parameter that will be bound to a
		    ground term at output time. *)
  | OutputFree (** A parameter that may be nonground at output time. *)

(*
   OutputBound (* A parameter that is bound at output time. *)
   InputBound (* A parameter that must not be a variable at input time. *)
*)

type mode_specifier =
    mode array

(** Used as hint for optimizations. A float between 0.0 and 1.0.
   It estimates for an argument position:

   NumberOfDifferentArgumentValues/PredicateSize

   (### Maybe use a different formula, that is better suited for
   combination of values - ask a statistican.)

   So, an argument corresonding to a key attribute has a discrimination
   value of 1.0. An argument which has the same value for the whole relation
   has a discrimination value of 0.0.
*)
type discrimination_value =
   float

(** Array of discrimination values corresponding to the arguments
   of a predicate. *)
type discrimination_values =
    discrimination_value array

val compute_discrimination_hint : binding_pattern ->
  discrimination_values -> discrimination_value

val make_default_discrimination_hints : int -> discrimination_values

(** Between 0.0 and 1.0. *)
type applicability_value =
    float

(** Specifies the dependency of builtin predicates. *)
type proof =
    BuiltinProof (* All information used for evaluation is contained
		    in the literal, as e.g. for arithmetics. *)
  | UnknownProof (* Evaluation of the builtin involves information about
		    other derived facts. See also [Fact.UnknownProof]. *)

(** Specifiers for different kinds of predicates. *)
type predicate_kind =
    Edb (** Consists only of facts (i.e. unit clauses). *)
  | Idb (** Consists of facts and/or rules (i.e. may have non-unit clauses). *)
  | BuiltIn (** Built-in predicate. *)

(** Specifiers for arguments to be indexed. *)
type indexargs =
    WholeTerm
  | Argument of int (** Argument position starting at 0 from left. *)
  | Arguments of int array (** Argument positions starting at 0
			     from left. *) 

(** Specifiers for index types. *)
type indextype =
    HashDt
  | AlistDt
  | NoIndex

type declaration =
    DiscriminationHintsDecl of discrimination_values
  | IndexDecl of indextype * indexargs
  | ModeDecl of mode_specifier
