(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* Algorithms are ported from O'Keefe's Prolog library. *)

module type AuxOrderedType =
  sig
    type t
    type aux
    val compare: aux -> t -> t -> int
  end

module type AuxS =
  sig
    type aux
    type elem
    val list_to_ord_set : aux -> elem list -> elem list
    val ord_insert : aux -> elem list -> elem -> elem list
    val ord_union : aux -> elem list -> elem list -> elem list
    val ord_intersection : aux -> elem list -> elem list -> elem list
    val ord_subtract : aux -> elem list -> elem list -> elem list
    val ord_symdiff : aux -> elem list -> elem list -> elem list
    val ord_subset : aux -> elem list -> elem list -> bool
    val ord_disjoint : aux -> elem list -> elem list -> bool
    val ord_intersect : aux -> elem list -> elem list -> bool
  end


module MakeAux(Ord: AuxOrderedType) = 
  struct
    
    type elem = Ord.t
    type aux = Ord.aux

    let rec ord_intersect aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    true
	  else if v < 0  then
	    ord_intersect aux t1 set2
	  else
	    ord_intersect aux set1 t2
      | _ -> false
	    
    let rec ord_disjoint aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    false
	  else if v < 0 then
	    ord_disjoint aux t1 set2
	  else
	    ord_disjoint aux set1 t2
      | _ -> true
	    
    let rec ord_subset aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    ord_subset aux t1 t2
	  else if v < 0 then
	    false
	  else
	    ord_subset aux set1 t2
      | [], _ -> true
      | _ -> false
	    
    let rec ord_symdiff aux set1 set2 =
      match set1, set2 with
	h1::t1, h2::t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    ord_symdiff aux t1 t2
	  else if v < 0 then
	    h1 :: ord_symdiff aux t1 set2
	  else
	    h2 :: ord_symdiff aux set1 t2
      | set1, [] -> set1
      | [], set2 -> set2
	    
    let rec ord_subtract aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    ord_subtract aux t1 t2
	  else if v < 0 then
	    h1 :: ord_subtract aux t1 set2
	  else
	    ord_subtract aux set1 t2
      | _, [] -> set1
      | [], _ -> []
	    
    let rec ord_intersection aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    h1 :: ord_intersection aux t1 t2
	  else if v < 0 then
	    ord_intersection aux t1 set2
	  else
	    ord_intersection aux set1 t2
      | _ -> []
	    

    let rec ord_union aux set1 set2 =
      match set1, set2 with
	h1 :: t1, h2 :: t2 ->
	  let v = Ord.compare aux h1 h2 in
	  if v = 0 then
	    h1 :: ord_union aux t1 t2
	  else if v < 0 then
	    h1 :: ord_union aux t1 set2
	  else
	    h2 :: ord_union aux set1 t2
      | [], set1 -> set1
      | set2, [] -> set2
	    
    let rec ord_insert aux set el =
      match set with
	[] -> 
	  [el]
      | h :: t ->
	  let v = Ord.compare aux h el in
	  if v = 0 then
	    set
	  else if v < 0 then
	    h :: ord_insert aux t el
	  else
	    el :: set
		   
    let rec rm_duplicates aux list =
      match list with
	e1 :: e2 :: list1 ->
	  if Ord.compare aux e1 e2 = 0 then
	    rm_duplicates aux (e2 :: list1)
	  else
	    e1 :: rm_duplicates aux (e2 :: list1)
      | _ -> list
	    
    let list_to_ord_set aux list =
      rm_duplicates aux	(List.fast_sort (Ord.compare aux) list)
	
  end


(** ### Hmm this is a copy of the other ordset stuff, with the aux
   argument removed. It seems not easy to express such code sharing
   of polymorphic functions with the Functor stuff. *)

let rec ord_intersect set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	true
      else if v < 0  then
	ord_intersect t1 set2
      else
	ord_intersect set1 t2
  | _ -> false
	
let rec ord_disjoint set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	false
      else if v < 0 then
	ord_disjoint t1 set2
      else
	ord_disjoint set1 t2
  | _ -> true
	
let rec ord_subset set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	ord_subset t1 t2
      else if v < 0 then
	false
      else
	ord_subset set1 t2
  | [], _ -> true
  | _ -> false
	
let rec ord_symdiff set1 set2 =
  match set1, set2 with
    h1::t1, h2::t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	ord_symdiff t1 t2
      else if v < 0 then
	h1 :: ord_symdiff t1 set2
      else
	h2 :: ord_symdiff set1 t2
  | set1, [] -> set1
  | [], set2 -> set2
	
let rec ord_subtract set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	ord_subtract t1 t2
      else if v < 0 then
	h1 :: ord_subtract t1 set2
      else
	ord_subtract set1 t2
  | _, [] -> set1
  | [], _ -> []
	
let rec ord_intersection set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	h1 :: ord_intersection t1 t2
      else if v < 0 then
	ord_intersection t1 set2
      else
	ord_intersection set1 t2
  | _ -> []
	
	
let rec ord_union set1 set2 =
  match set1, set2 with
    h1 :: t1, h2 :: t2 ->
      let v = Pervasives.compare h1 h2 in
      if v = 0 then
	h1 :: ord_union t1 t2
      else if v < 0 then
	h1 :: ord_union t1 set2
      else
	h2 :: ord_union set1 t2
  | [], set1 -> set1
  | set2, [] -> set2
	
let rec ord_insert set el =
  match set with
    [] -> 
      [el]
  | h :: t ->
      let v = Pervasives.compare h el in
      if v = 0 then
	set
      else if v < 0 then
	h :: ord_insert t el
      else
	el :: set
	       
let rec rm_duplicates list =
  match list with
    e1 :: e2 :: list1 ->
      if Pervasives.compare e1 e2 = 0 then
	rm_duplicates (e2 :: list1)
      else
	e1 :: rm_duplicates (e2 :: list1)
  | _ -> list
	
let list_to_ord_set list =
  rm_duplicates (List.fast_sort Pervasives.compare list)
    
