(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Representation of sets as ordered list.  This module is similar to the
   [ordsets] library module common in Prolog systems.
*)

(** {6 Comparison using Pervasives.compare} *)


(** [list_to_ord_set list] returns an ordered set with the
  elements of [list], i.e. a copy of [list] with duplicate 'aents
  removed. 
  *) 
val list_to_ord_set : 'a list -> 'a list
    
(** [ord_insert set element] returns a [set] with [element] inserted
   into it, preserving the order. [set] is an ordered set.
 *) 
val ord_insert : 'a list -> 'a-> 'a list
	  
    (** [ord_union set1 set2] returns the ordered union of [set1] and
       [set2]. If both sets contain an element, it is retained only
       once in the result. [set1] and [set1] are ordered sets.
     *)
val ord_union : 'a list -> 'a list -> 'a list

    (** [ord_intersection set1 set2] returns the ordered intersection of
       [set1] and [set2]. [set1] and [set1] are ordered sets.
     *)
val ord_intersection : 'a list -> 'a list -> 'a list

    (** [ord_subtract set1 set2] returns the elements in [set1] which are not
       also in [set2] as ordered set.  [set1] and [set1] are ordered sets.
     *)
val ord_subtract : 'a list -> 'a list -> 'a list

    (** [ord_symdiff set1 set2] returns the symmetric difference
       of [set1] and [set2] as ordered set.  [set1] and [set1] are
       ordered sets.
     *)
val ord_symdiff : 'a list -> 'a list -> 'a list

    (** [ord_subset set1 set2] is true iff every element of the ordered
       set [set1] appears in the ordered set [set2]. 
       [set1] and [set1] are ordered sets.
     *)
val ord_subset : 'a list -> 'a list -> bool

    (** [ord_disjoint set1 set2] is true iff [set1] and [set2] have no
       elements in common. [set1] and [set1] are ordered sets.
     *)
val ord_disjoint : 'a list -> 'a list  -> bool

    (** [ord_intersect set1 set2] is true iff [set1] and [set2] have
       at least one element in common. [set1] and [set1] are ordered sets.
     *)
val ord_intersect : 'a list -> 'a list -> bool

(** {6 Functorial Interface} *)

(** Input signature of the [MakeAux] functor. 
*)
module type AuxOrderedType = 
  sig
    type t

    type aux

    (** A total ordering function over the keys.
       This is a two-argument function [f] such that
       [f aux e1 e2] is zero if the keys [e1] and [e2] are equal,
       [f aux e1 e2] is strictly negative if [e1] is smaller than [e2],
       and [f aux e1 e2] is strictly positive if [e1] is greater than [e2].

       [aux] is an auxiliary argument.
     *)
    val compare : aux -> t -> t -> int
	
  end


(** Output signature of the MakeAux functor. This is similar to the
   functions directly in [Ordset], except that each function takes 
   an additional first argument
   which it passed to [compare]. *)
module type AuxS =
  sig
    type elem
    type aux
    val list_to_ord_set : aux -> elem list -> elem list
    val ord_insert : aux -> elem list -> elem -> elem list
    val ord_union : aux -> elem list -> elem list -> elem list
    val ord_intersection : aux -> elem list -> elem list -> elem list
    val ord_subtract : aux -> elem list -> elem list -> elem list
    val ord_symdiff : aux -> elem list -> elem list -> elem list
    val ord_subset : aux -> elem list -> elem list -> bool
    val ord_disjoint : aux -> elem list -> elem list -> bool
    val ord_intersect : aux -> elem list -> elem list -> bool
  end

(** Functor building an implementation of the [Ordset.AuxS] type
   given a totally ordered type. *)
module MakeAux (Ord : AuxOrderedType) : 
    AuxS with type elem = Ord.t and type aux = Ord.aux



