(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


(****************************************************************************)
(* Data Types. *)
(****************************************************************************)

module PredicateTable = Hashtbl.Make(
  struct
    type t = Term.symbol
    let equal = (==)
    let hash = Hashtbl.hash
  end)

type processing_status =
    New
  | ReadyToAssert
  | ReadyToRun

type predicate =
  { symbol: Term.symbol;
    mutable is_builtin: bool;
    mutable is_delta: bool;
    mutable declarations: Opt.declaration list;
    mutable status: processing_status;
    mutable readers: Access.Reader.reader list;
    mutable writers: Access.Writer.writer list;
    mutable discrimination_hints: Opt.discrimination_values;
    mutable delta_predicate: predicate;
      (* Used as pointer to the main predicate in delta predicates. *)
    mutable all_bound_reader: Access.Reader.reader
  }
  
let get_predicate_symbols pt =
	let symbols = ref []
	in
		PredicateTable.iter
			(fun sym _ ->  
				symbols := sym :: !symbols)
			pt;
		!symbols

(* debug *)
let pretty_pred pred =
  Convert.symbol_to_string pred.symbol^
  (if pred.is_delta then ":delta" else "")

(* debug *)
let print_pred p =
  print_endline (pretty_pred p)

let get_symbol pred =
  pred.symbol

let get_declarations pred =
  pred.declarations

let get_is_builtin pred =
  pred.is_builtin

let get_delta_predicate pred =
  pred.delta_predicate

let get_main_predicate pred =
  pred.delta_predicate

let get_discrimination_hints pred =
  pred.discrimination_hints

let get_readers pred =
  pred.readers

let get_writers pred =
  pred.writers

let get_all_bound_reader pred =
  pred.all_bound_reader

let has_delta_predicate pred =
  pred.delta_predicate != pred

let extent_size pred =
  match pred.writers with
    [] -> 0
  | w :: _ -> Access.Writer.size w
  
type predicate_table = predicate PredicateTable.t

let make_predicate_table () =
  PredicateTable.create 16

let make_predicate symbol =
  let rec pred =
    {status=New;
     symbol=symbol;
     declarations=[];
     is_builtin=false;
     is_delta=false;
     readers=[];
     writers=[];
     delta_predicate=pred;
     discrimination_hints=Opt.make_default_discrimination_hints
       (Term.symbol_arity symbol);
     all_bound_reader=Access.Reader.null_reader ()}
  in
  pred

let make_delta_predicate main_pred =
  {status=New;
   symbol=main_pred.symbol;
   declarations=main_pred.declarations;
   is_builtin=false;
   is_delta=true;
   readers=[];
   writers=[];
   delta_predicate=main_pred; 
   discrimination_hints=Opt.make_default_discrimination_hints
     (Term.symbol_arity main_pred.symbol);
   all_bound_reader=Access.Reader.null_reader ()}


(****************************************************************************)
(* First Step: Make A Predicate Ready To Assert. *)
(****************************************************************************)

let default_indextype pred = 
  let arity = (Term.symbol_arity pred.symbol) in 
  if arity = 0 then
    Opt.NoIndex
  else 
    Opt.AlistDt
  
let add_index pred indextype indexargs =
  let writer =
    Access.Writer.make_writer
      indextype indexargs (Term.symbol_arity pred.symbol) in
  let readers =
    Access.Reader.make_readers writer in
  pred.writers <- writer :: pred.writers;
  pred.readers <- List.append readers pred.readers

let process_declarations pred = 
  (* Declarations in pred.declarations are in reverse order as read, i.e.
     those read at last are preferred. *)
  (* 1. Look for a discrimination hints declaration. *)
  begin
    try
      let disc_hints =
	List.find 
	  (fun decl -> 
	    match decl with
	      Opt.DiscriminationHintsDecl _ -> true | _ -> false)
	  pred.declarations in
      match disc_hints with
	Opt.DiscriminationHintsDecl disc_hints_1 ->
	  pred.discrimination_hints <- disc_hints_1
      | _ -> raise (Failure "This should not happen - disc_hints.")
    with Not_found ->
      ()
  end;
  (* 2. Process all index declarations. *)
  List.iter
    (fun decl ->
      match decl with
	Opt.IndexDecl (indextype, indexargs) ->
	  add_index pred indextype indexargs
      | _ ->
	  ())
    pred.declarations

  
let make_ready_to_assert pred =
  begin
    match Term.symbol_predicate pred.symbol with
      Term.PUnbound ->
	pred.is_builtin <- false;
	process_declarations pred;
	if pred.readers = [] then
	  begin
	    (* Add a default index as extent store. *)
	    Msg.msg 2 
	      ("Creating default index for "^(pretty_pred pred)^".");
	    add_index pred (default_indextype pred) Opt.WholeTerm
	  end
	else
	  ();
	pred.status <- ReadyToAssert
    | Term.PDeterministic (f, (modespec, discvalues, _)) ->
	pred.is_builtin <- true;
	pred.discrimination_hints <- discvalues;
	pred.status <- ReadyToRun
  end


let find_predicate predicate_table symbol =
  try 
    PredicateTable.find predicate_table symbol 
  with Not_found ->
    let pred = make_predicate symbol in
    PredicateTable.add predicate_table symbol pred;
    pred
      
let ensure_ready_to_assert pred =
  match pred.status with
    New ->
      make_ready_to_assert pred
  | ReadyToAssert -> ()
  | ReadyToRun -> ()


(****************************************************************************)
(* Second Step: Make a Predicate Ready to Run. *)
(****************************************************************************)

let applicability pred binding_pattern =
  List.fold_left
    (fun n1 r2 ->
      let n2 = Access.Reader.applicability r2 binding_pattern
      in 
      if n1 < n2 then n2 else n1)
    0.0
    pred.readers

let choose_best_reader pred binding_pattern =
  match pred.readers with
    [] -> 
      raise (Failure "No reader.")
  | [r] -> r
  | r :: rs ->
      List.fold_left
	(fun r1 r2 ->
	  if (Access.Reader.applicability r1 binding_pattern) <
	    (Access.Reader.applicability r2 binding_pattern) 
	  then
	    r2
	  else
	    r1)
	r
	rs

let make_ready_to_run pred = 

  (* Sort readers such that those on more args come first. 
     This sorting is assumed by choose_best_reader. *)
  let arity = (Term.symbol_arity pred.symbol) in
  pred.readers <-
    List.sort 
      (fun r1 r2 ->
	let v1 = 
	  match Access.Reader.get_indexargs r1 with
	    Opt.WholeTerm -> arity
	  | Opt.Argument _ -> 1
	  | Opt.Arguments a -> Array.length a
	and v2 =
	  match Access.Reader.get_indexargs r2 with
	    Opt.WholeTerm -> arity
	  | Opt.Argument _ -> 1
	  | Opt.Arguments a -> Array.length a in
	compare v2 v1)
      pred.readers;
	  	 
  let all_bound_bp =
    Array.make (Term.symbol_arity pred.symbol) Opt.Bound in
  pred.all_bound_reader <- choose_best_reader pred all_bound_bp;
  pred.status <- ReadyToRun
      
let ensure_ready_to_run pred =
  ensure_ready_to_assert pred;
  match pred.status with
    ReadyToRun ->
      ()
  | ReadyToAssert ->
      make_ready_to_run pred
  | _ ->
      raise (Failure "Something is wrong with the ensure_ready stuff.")

(* ### If backward subsumption is off and delta is always the 1st predicate
   in a body, no index to it is needed. For now delta is created with
   the same indices as the plain predicate. *)

let ensure_delta_ready_to_run pred =
  ensure_ready_to_run pred;
  if has_delta_predicate pred then
    ()
  else
    let delta_pred = make_delta_predicate pred in
    ensure_ready_to_run delta_pred;
    pred.delta_predicate <- delta_pred


(****************************************************************************)
(* Access. *)
(****************************************************************************)

let insert predicate term entry =
  List.iter 
    (fun writer -> Access.Writer.insert writer term entry)
    predicate.writers

let remove_variant predicate term =
  List.iter
    (fun writer -> Access.Writer.remove_variant writer term)
    predicate.writers

let clear predicate =
  List.iter Access.Writer.clear predicate.writers
      
let clear_delta_predicate predicate =
  if has_delta_predicate predicate then
    clear predicate.delta_predicate
  else
    ()
      
let assert_unit_clause predicate clause fact =
  (* Assumed status is at least ReadyToAssert. *)
  insert predicate clause fact

let check_declaration predicate declaration =
  match declaration with
    Opt.IndexDecl (indextype, indexargs) ->
      begin
	match indexargs with
	  Opt.Argument n ->
	    if n >= (Term.symbol_arity predicate.symbol) then
	      raise 
		(Err.Error "Indexarg larger than predicate arity.")
	    else
	      ()
	| Opt.Arguments ns ->
	    let arity = (Term.symbol_arity predicate.symbol) in
	    Array.iter
	      (fun n ->
		if n >= arity then
		  raise 
		    (Err.Error "An indexarg is larger than predicate arity.")
		else
		  ())
	      ns
	| _ ->
	    ()
      end
  | Opt.DiscriminationHintsDecl vals ->
      if Array.length vals <> (Term.symbol_arity predicate.symbol) then
	raise 
	  (Err.Error "Length of discrimination hints does not match arity.")
      else
	();
      for i = 0 to (Array.length vals) - 1 do
	if vals.(i) < 0.0 || vals.(i) > 1.0 then
	  raise (Err.Error "Discrimination hint value out of range.")
	else
	  ()
      done
  | Opt.ModeDecl mode ->
      if Array.length mode <> (Term.symbol_arity predicate.symbol) then
	raise 
	  (Err.Error "Length of mode declaration does not match arity.")


let add_declaration predicate decl =
  match predicate.status with
    New ->
      check_declaration predicate decl;
      predicate.declarations <-
	decl :: predicate.declarations
  | _ ->
      raise 
	(Err.Error "Attempt to add declaration to predicate already in use.")

let default_reader predicate =
  match predicate.readers with
    [] ->
      raise Not_found
  | reader :: _ ->
      reader

let abolish predicate_table symbol =
  PredicateTable.remove predicate_table symbol

  

