(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Information associated with predicates. This includes:

   - Declarations that have been submitted as user input.
   - Extents, i.e. unit clauses that have been submitted as input or
     derived.
   - Special objects used to access the extent. See module [access].
   - Information about the processing status of a predicate, e.g.
     if it is ready to accept declarations or new facts.
*)

(** {6 Predicate Objects} *)

(** The type of predicate objects. *)
type predicate

(** Get the symbol associated with the given predicate. *)
val get_symbol : predicate -> Term.symbol

(** Get the declarations associated with the given predicate. *)
val get_declarations : predicate -> Opt.declaration list

val get_discrimination_hints : predicate -> Opt.discrimination_values

val get_delta_predicate : predicate -> predicate

val get_main_predicate : predicate -> predicate

val get_is_builtin : predicate -> bool

val get_all_bound_reader : predicate -> Access.Reader.reader

val get_writers : predicate -> Access.Writer.writer list

(** Note: [predicate] is assumed to be not a delta predicate. *)
val has_delta_predicate : predicate -> bool

(** Add a unit clause to the given predicate's extent. The predicate's
   status must at least be ready_to_assert. *)
val assert_unit_clause : predicate -> Term.term -> Fact.fact -> unit

(** Associate the given declaration object with the given predicate. *)
val add_declaration : predicate -> Opt.declaration -> unit

(** Return the size of the given predicate's extent in the current model. *)
val extent_size : predicate -> int

(** Return some reader for the predicate, e.g. for unindexed iteration. 
    Raises [Not_found] if there is no reader associated with the predicate. *)
val default_reader : predicate -> Access.Reader.reader

val applicability : predicate -> Opt.binding_pattern -> 
  Opt.applicability_value

val choose_best_reader : predicate ->
  Opt.binding_pattern -> Access.Reader.reader

(** {6 Processing Status} *)

val ensure_ready_to_assert : predicate -> unit

val ensure_ready_to_run : predicate -> unit

val ensure_delta_ready_to_run : predicate -> unit


(** {6 Predicate Table} 

A predicate table maintains an association between predicate symbols
and predicate objects.
*)

type predicate_table

module PredicateTable :
    sig
      val find : predicate_table -> Term.symbol -> predicate
      val iter : (Term.symbol -> predicate -> unit) -> predicate_table -> unit
    end

val make_predicate_table : unit -> predicate_table

val find_predicate : predicate_table -> Term.symbol -> predicate

val abolish : predicate_table -> Term.symbol -> unit

val get_predicate_symbols : predicate_table -> Term.symbol list

(** {6 Modifying Access} *)

val insert : predicate -> Term.term -> Fact.entry -> unit

val remove_variant : predicate -> Term.term -> unit

val clear_delta_predicate : predicate -> unit

val print_pred : predicate -> unit
