(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i

(****************************************************************************)
(* Handing of Declarations - Converting from and to Terms *)
(****************************************************************************)

let indextype_of_spec term =
  match Convert.term_to_string term with
    "alist_dt" ->
      Opt.AlistDt
  | "hash_dt" ->
      Opt.HashDt
  | "none" ->
      Opt.NoIndex
  | _ ->
      raise (Err.Error "Bad index type specifier.")

let indexargs_of_spec term =
  match term with
    Term.Constant sym ->
      begin
	match Term.symbol_name sym with
	  "all" ->
	    Opt.WholeTerm
	| _ ->
	    raise (Err.Error "Bad index args specifier.")
      end
  | Term.Integer int ->
      Opt.Argument int
  | _ ->
      let args = Convert.term_to_array_of_int term in
      let len = Array.length args in
      if len > 1 then
	Opt.Arguments args
      else if len = 1 then
	Opt.Argument args.(0)
      else
	Opt.WholeTerm

let process_index_decl env term =
  let (name, arity) = 
    Convert.predspec_pair (Convert.term_nth_arg term 0) in
  let pred = (Env.find_predicate env name arity) in
  let args = Convert.term_to_alist (Convert.term_nth_arg term 1) in
  let indextype_spec =
    try List.assoc "type" args
    with Not_found ->
      raise (Err.Error "Missing 'type' specifier in index declaration.")
  and indexargs_spec =
    try List.assoc "args" args
    with Not_found ->
      raise (Err.Error "Missing 'args' specifier in index declaration.")
  in
  Predicate.add_declaration
    pred
    (Opt.IndexDecl
       ((indextype_of_spec indextype_spec),
	(indexargs_of_spec indexargs_spec)))

let process_mode_decl env term =
  let template = (Convert.term_nth_arg term 0) in
  match template with
    Term.Compound compound ->
      let (name, arity) =
	Convert.term_functor_pair template in
      let pred = (Env.find_predicate env name arity)
      and args = Term.compound_args compound
      and mode = Array.make arity Opt.OutputGround in
      for i = 0 to arity - 1 do
	match Convert.term_to_string args.(i) with
	  "b" ->
	    mode.(i) <- Opt.OutputGround
	| "f" ->
	    mode.(i) <- Opt.OutputFree	    
	| _ ->
	    raise (Err.Error "Bad argument in 'mode' declaration template.")
      done;
      Predicate.add_declaration
	pred
	(Opt.ModeDecl mode)
  | _ ->
      ()

let process_disc_hints_decl env term =
  let (name, arity) = 
    Convert.predspec_pair (Convert.term_nth_arg term 0) in
  let pred = (Env.find_predicate env name arity) in
  Predicate.add_declaration
    pred
    (Opt.DiscriminationHintsDecl
       (Convert.term_to_array_of_float (Convert.term_nth_arg term 1)))

let decl_to_term env predicate decl =
  let st = Env.get_symbol_table env in
  match decl with
    Opt.IndexDecl (indextype, indexargs) ->
      begin
	let itype =
	  Term.Constant
	    (Term.find_symbol
	       st
	       ( match indextype with
		 Opt.HashDt -> "hash_dt"
	       | Opt.AlistDt -> "alist_dt"
	       | Opt.NoIndex -> "none")
	       0) in
	let iargs =
	  match indexargs with
	    Opt.WholeTerm -> Term.Constant (Term.find_symbol st "all" 0)
	  | Opt.Argument n -> Term.Integer n
	  | Opt.Arguments ns -> Convert.array_of_int_to_term st ns in
	Term.Compound
	  (Term.make_compound
	     (Term.find_symbol st "index" 2)
	     [|Convert.symbol_to_pred_spec_term
		 st
		 (Predicate.get_symbol predicate);
	       (Convert.alist_to_term
		  st
		  [("type", itype); ("args", iargs)])|])
      end
  | Opt.DiscriminationHintsDecl vals ->
      Term.Compound
	(Term.make_compound
	   (Term.find_symbol st "discrimination_hints" 2)
	   [|Convert.symbol_to_pred_spec_term
	       st
	       (Predicate.get_symbol predicate);
	     Convert.array_of_float_to_term st vals|])
  | Opt.ModeDecl mode ->
      let arity =  (Array.length mode) in
      let args = Array.make arity Term.Unbound in
      for i = 0  to arity - 1 do
	args.(i) <-
	  Term.Constant 
	    (Term.find_symbol
	       st 
	       ( match mode.(i) with
		 Opt.OutputGround -> "b"
	       | Opt.OutputFree -> "f"
	       | Opt.InputGround -> "i" )
	       0)
      done;
      Term.Compound
	(Term.make_compound
	   (Term.find_symbol st "mode" 1)
	   [|(Term.Compound
		(Term.make_compound
		   (Predicate.get_symbol predicate)
		   args))|])

let print_declarations env pred =
  List.iter
    (fun decl -> 
      PrologPrint.print
	(decl_to_term env pred decl);
      print_endline ".")
    (List.rev (Predicate.get_declarations pred))


(****************************************************************************)
(* Commands *)
(****************************************************************************)

let model_count = ref 0

let scan_format string len i =
  if !i >= len then
    raise (Err.Error "No format action specified.");
  begin
    match String.get string !i with
      'm' ->
	print_int !model_count
    | '~' ->
	print_char '~'
    | 'z' ->
	flush stdout
    | _ ->
	raise (Err.Error "Unknown format action.")
  end;
  i := !i + 1

let write_formatted string =
  let len = String.length string
  and i = ref 0 in
  while !i < len
  do
    match String.get string !i with
      '~' ->
	i := !i+1;
	scan_format string len i
    | c -> 
	i := !i+1;
	print_char c;
  done


let ensure_branch_reset env =
  let result = Env.get_current_result env in
  if Tableau.is_open result then
    Msg.msg 1 "Abandoning current model.";
  Env.set_current_result env (Tableau.abandon_result result);
  model_count := 0

let abandon_cmd =
  (fun env term ->
    ensure_branch_reset env)
(*
let make_clauses env = 
	let new_made_clauses, old_made_clauses =
		(List.partition
			(fun clause -> 
				if (Ic.is_indexed (Env.get_rule_subterm_table env) clause) then 
					false
				else
					(Ic.index_clause_subterms 
						(Env.get_rule_subterm_table env) 
						clause;
					true))
			(Ic.remove_clause_duplicates
	  			(IcSet.make_eval_plain_clauses
				     (Env.get_cfg env)
				     (Env.get_predicate_table env)
				     (Env.get_symbol_table env)
				     (Env.get_negative_units env)
				     (Env.get_rules env))))
	in
		new_made_clauses
*)		
(*		
let make_clauses env = 
	(Ic.remove_clause_duplicates
		(IcSet.make_eval_plain_clauses
		     (Env.get_cfg env)
		     (Env.get_predicate_table env)
		     (Env.get_symbol_table env)
		     (Env.get_negative_units env)
		     (Env.get_rules env)))
*)		     
let make_clauses env = 
		(IcSet.make_eval_plain_clauses
		     (Env.get_cfg env)
		     (Env.get_predicate_table env)
		     (Env.get_symbol_table env)
		     (Env.get_negative_units env)
		     (Env.get_rules env))

(*		     
let make_clauses env = 
	(Ic.remove_subsumed_clauses
		(IcSet.make_eval_plain_clauses
		     (Env.get_cfg env)
		     (Env.get_predicate_table env)
		     (Env.get_symbol_table env)
		     (Env.get_negative_units env)
		     (Env.get_rules env)))
*)		     
		     		

let szs_output env sat =
(*	let szs_problem_name = Cfg.get_szs_problem_name (Env.get_cfg env)
	in	*)
	if (Cfg.get_szs_output_flag (Env.get_cfg env)) then
			(print_endline "";
			if sat then
				(print_endline "SZS status Satisfiable")
			else
				(print_endline "SZS status Unsatisfiable"))


		
let next_cmd =
  (fun env term ->
    let t0 = Sys.time () 
    and sat = ref true
    in
    let result =
      Tableau.next_result (Env.get_current_result env) in
    if Tableau.is_open result then
      (
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0);
	   )
    else
      begin
    if !model_count=0 then
    	(sat := false);   
	model_count := 0;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
    Env.set_current_result env result;
    Tableau.info_msg result;
    szs_output env !sat)

let run_cmd =
  (fun env term ->
    ensure_branch_reset env;
    let t0 = Sys.time () 
    and sat = ref true
    in
    let result =
      Tableau.first_result
        (Env.get_cfg env)
        (Env.get_predicate_table env)
        (Env.get_symbol_table env)
		(Env.get_negative_units env)
		(Env.get_subterm_table env)
		(Env.get_clause_aux env)
        (Env.get_clauses env) in
    if Tableau.is_open result then
      begin
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0)
      end
    else
      begin
	model_count := 0;
	sat := false;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
    Env.set_current_result env result;
    Tableau.info_msg result;
    szs_output env !sat)
	
let run_relevant_cmd =
  (fun env term ->
    let (name, arity) =
      Convert.predspec_pair (Convert.term_nth_arg term 0) in
    ensure_branch_reset env;
    let t0 = Sys.time ()
    and sat = ref true
    in
    let pred = Env.find_predicate env name arity in
    let result =
      Tableau.first_result
        (Env.get_cfg env)
        (Env.get_predicate_table env)
        (Env.get_symbol_table env)
		(Env.get_negative_units env)
		(Env.get_subterm_table env)
		(Env.get_clause_aux env)
	(IcSet.select_relevant_clauses (Env.get_clauses env) pred) in
    if Tableau.is_open result then
      begin
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0)
      end
    else
      begin
	model_count := 0;
	sat := false;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
    Env.set_current_result env result;
    Tableau.info_msg result;
    szs_output env !sat)


(****************************************************************************)
(* Process Statement *)
(****************************************************************************)

exception Interrupted

let set_signals () =
  Sys.set_signal
    Sys.sigint 
    (Sys.Signal_handle
       (fun s ->
	 Msg.msg 1 "Interrupted";
	 raise Interrupted))

exception Found
	 
let unit_is_indexed predicate symbol_table new_term = (*	PrologPrint.print new_term;print_string "   ";	*)
	let equals = Term.get_equality_symbol symbol_table
	and sym = Convert.term_functor_symbol new_term
	in
		if equals == sym then (
			let flip_term = Term.binary_flip new_term
			in
				try
		    		Access.Reader.iter_unification_candidates 
		      			(fun fact ->
		      				let term = Fact.get_term fact
		      				in 
								if (Term.variant_chk term new_term) then
				  					raise Found
								else
				  					())
		      			(Predicate.default_reader predicate)
		      			new_term
		      			Term.null_context;
		      		Access.Reader.iter_unification_candidates 
		      			(fun fact ->
		      				let term = Fact.get_term fact
		      				in 
								if (Term.variant_chk term flip_term) then
				  					raise Found
								else
				  					())
		      			(Predicate.default_reader predicate)
		      			flip_term
		      			Term.null_context;
		    		false
		  		with Found ->
		    		true
		  )
		  else(			(*	PrologPrint.print new_term;print_string "   ";		*)
		  		try
		    		Access.Reader.iter_unification_candidates 
		      			(fun fact -> 
							if (Term.variant_chk (Fact.get_term fact) new_term) then
			  					raise Found
							else
			  					())
			      		(Predicate.default_reader predicate)
			      		new_term
			      		Term.null_context;
		    		false
		  		with Found ->
		    		true)	 
			 
	 
	 
let rec process_statement env term =
  if Cfg.get_echo_flag (Env.get_cfg env) then
    begin
      PrologPrint.print term;
      print_string ".";
      print_newline ();
    end
  else
    ();
  let sym = Convert.term_functor_symbol term in
  match Env.classify_statement_functor env sym with
    Env.Meta ->
      (try
	process_meta_statement env (Convert.term_nth_arg term 0)
      with ex ->
	ensure_branch_reset env;
	raise ex)
  | Env.Unit ->
      ensure_branch_reset env;
(* Equation units are treated as as rules. *)
      let equals = (Term.get_equality_symbol (Env.get_symbol_table env)) 
      in
	      	if sym == equals then (
				Env.add_rule env term
			)
			else (
				if (Cfg.get_hyperextension_flag (Env.get_cfg env)) then
					(try 
						(let new_term = (Env.unit_for_env env term)
						in	
						let fact = (Fact.make_input_fact new_term)
						in	
							Ic.np_subsumption_processing_unit
								(Cfg.get_full_branch_redundancy_flag (Env.get_cfg env))
								(Env.get_redundant_store env)
								(Env.ensure_ready_to_assert env sym)
								fact;
							Predicate.assert_unit_clause 
								(Env.ensure_ready_to_assert env sym) 
								new_term
								fact;
							Ic.index_unit_subterms_no_top
								(Env.get_subterm_table env)
								new_term
								fact
						) 
					with Ic.Redundant -> ())
				else
					(let eq_term =
						Term.Compound
							(Term.make_compound
								equals
								[|term; Term.Unbound|])
					in
						Env.add_rule env eq_term));
      true
  | Env.Rule ->
      ensure_branch_reset env;
      Env.add_rule env term;
      true
  | Env.Disjunction ->
      ensure_branch_reset env;
      Env.add_disjunction env term;
      true
  | Env.False ->
      ensure_branch_reset env;
      Env.add_false env;
      true

      
and process_meta_statement env term =
  match Convert.term_functor_pair term with
    ("halt", 0) ->
      false
  | _ ->
      try
	Sys.set_signal
	  Sys.sigint 
	  (Sys.Signal_handle
	     (fun s -> raise Interrupted));
	process_meta_statement_1 env term;
	Sys.set_signal Sys.sigint Sys.Signal_default;
	true
      with Interrupted ->
	Msg.msg 1 "Interrupted.";
	true

and cmd_table =
  [
   (("help", 0),
    ("'help' prints documentation.",
     (fun env term -> help env)));
   (("run", 0),
    ("'run' computes the first model or a refutation. If a model is found,\n\
      it becomes the current model.",
     run_cmd));
   (("run", 1),
    ("'run(p/n)' is like run/0, but only the subset of input rules\n\
     that is relevant for the given predicate is used.", 
     run_relevant_cmd));
   (("next", 0),
    ("'next' computes the next model. If a model is found, it becomes the\n\
      current model.",
     next_cmd));
   (("abandon", 0),
    ("'abandon' abandons the current model.",
     abandon_cmd));
   (("szs_problem_name", 1),
    ("'szs_problem_name/1' enters the problem name required for SZS-conforming output.",
     (fun env term ->
     	Cfg.set_szs_problem_name
     		(Env.get_cfg env)
     		(Convert.term_to_string
     			(Convert.term_nth_arg term 0)))));     
     
(* debug *)     
  (*   (("block_dummy",0),("bla",
     (fun env term ->
     	(let dummy = Fact.InputFact (Term.Constant (Term.gensym ()))
     		in
     		Ic.block_fact (Env.get_blocked_table env) dummy))));*)
	(("print_current_subterm_table_keys", 0),
		("Debugging function, prints the keys of the current subterm_table.",
		(fun env term ->
		     Ic.SubtermTable.iter
		  		(fun symbol subterm -> ps (Convert.symbol_to_string symbol);pe ".")
		  		(Env.get_subterm_table env))));
   (("print_current_compiled_rules", 0),
    ("'print_current_compiled_rules' prints the set of compiled rules in the current\n\
       model.",
     (fun env term ->
       (let clauses = ref []
       in
       Ic.SubtermTable.iter
		  (fun symbol subterm -> (*ps (Convert.symbol_to_string symbol);pe "  !";*)
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (
				  				(*ps (Convert.symbol_to_string symbol);ps " : ";ppp (Ic.get_source clause);pe ".";*)
				  			
				  			clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ())
	  (Env.get_subterm_table env);
(*	  List.iter 
	  	(fun c -> 
	  		ps (Ic.get_id c);
	  		if (Ic.is_eq_unit c) then 
	  			(if (Ic.is_oriented c) then ps " reductive"; ps " (EQ-UNIT)");
	  		ps " : ";ppp (Ic.get_source c);
	  		pe ".") 
	  	!clauses	*)
	  Info.print_clauses_as_terms !clauses	))));
   
(("print_current_equation_units", 0),
    ("'print_current_equation_units' prints the set of equation units in the current\n\
       model.",
     (fun env term ->
       (let clauses = ref []
       in
       Ic.SubtermTable.iter
		  (fun symbol subterm ->
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (not (Ic.is_eq_unit clause))||(Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ())
	  (Env.get_subterm_table env);
	  Info.print_clauses_as_terms !clauses))));
 	  
   (("print_extent", 0),
    ("'print_extent' prints the extents of all predicates in the current\n\
       model.",
     (fun env term ->
       (Predicate.PredicateTable.iter
	  (fun symbol predicate ->
	    try 
	      Access.Reader.iter_all
		(fun entry -> 
		  	if (Fact.is_redundant entry) then ((*ps "RED: ";PrologPrint.print (Fact.get_term entry);
		  		print_endline "."*))
		  	else (
				PrologPrint.print (Fact.get_term entry);
		  		print_endline "."))
		(Predicate.default_reader predicate)
	    with Not_found ->
              (* May be thrown by Predicate.default_reader predicate *)
	      ())
	  (Env.get_predicate_table env))
     )));
     
(("print_model", 0),
    ("'print_model' prints the current model, which means\n\
     all positive unit clauses in the current branch.",
     (fun env term ->
     	let krh_dom_symbol = Term.get_krh_dom_symbol (Env.get_symbol_table env)
     	and clauses = ref []
     	in
       (Predicate.PredicateTable.iter
	  (fun symbol predicate ->
	  	if symbol == krh_dom_symbol then ()
	  	else
		    (try 
		      Access.Reader.iter_all
			(fun entry -> 
			  	if (Fact.is_redundant entry) then ((*ps "RED: ";PrologPrint.print (Fact.get_term entry);
			  		print_endline "."*))
			  	else (
					PrologPrint.print (Fact.get_term entry);
			  		print_endline "."))
			(Predicate.default_reader predicate)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ()))
	  (Env.get_predicate_table env));
	  (Ic.SubtermTable.iter
		  (fun symbol subterm ->
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (not (Ic.is_eq_unit clause))||(Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
		      ())
	  (Env.get_subterm_table env);
	  Info.print_clauses_as_terms !clauses)
     )));     
     
     
   (("print_extent", 1),
    ("'print_extent(p/n)' prints the extent of the predicate p/n in the\n\
      current model.",
     (fun env term ->
      begin
	let (name, arity) = 
	  Convert.predspec_pair
	    (Convert.term_nth_arg term 0) in
	try 
	  Access.Reader.iter_all
	    (fun entry -> 
	      PrologPrint.print (Fact.get_term entry);
	      print_endline ".")
	    (Predicate.default_reader (Env.lookup_predicate env name arity))
	with Not_found ->
	  print_endline "% Predicate not found."
      end)));
   (("print_extent_with_proofs", 1),
    ("'print_extent_with_proofs(p/n)' prints a Fact-Proof pair for each\n\
       fact in the extent of the predicate p/n in the current model.\n\
      [Works not yet with disjunctions].",
	   (fun env term ->
	     begin
	       let (name, arity) = 
		 Convert.predspec_pair
		   (Convert.term_nth_arg term 0) in
	       try 
		 Access.Reader.iter_all
		   (fun entry -> 
		     PrologPrint.print (Fact.get_term entry);
		     print_endline " - ";
		     PrologPrint.print
		       (Fact.proof_term_for_fact
			  (Env.get_symbol_table env)
			  entry);
		     print_endline ".")
		   (Predicate.default_reader
		      (Env.lookup_predicate env name arity))
	       with Not_found ->
		 print_endline "% Predicate not found."
	     end)));
   (("print_proof", 0),
    ("'print_proof' prints the proof if the current result is closed.\n\
       Output format is the same as of 'print_extent_with_proofs'.\n\
      [Output for non-Horn problems and open results is still experimental].",
     (fun env term ->
       let r = Tableau.get_refutation (Env.get_current_result env) in
       PrologPrint.print 
	 (Tableau.refutation_to_term
	    (Env.get_symbol_table env)
	    r);
       print_endline("."))));
   (("print_szs_proof", 0),
    ("'print_szs_proof' prints the proof if the current result is closed.\n\
       Output format is similar to that of 'print_proof', except that\n\
       extra control clauses are added to make the output suitable for the\n\
       'post_szs.pl' postprocessor.\n\
      [Output for non-Horn problems and open results is still experimental].",
     (fun env term ->
     if !model_count=0 then (
       let r = Tableau.get_refutation (Env.get_current_result env) in
       
       	match r with
       		Tableau.NullRefutation ->
       			(print_endline "szs_unfinished.")
       		|_ ->
       (print_endline "szs_refutation_start.";
       PrologPrint.print 
	 (Tableau.refutation_to_term
	    (Env.get_symbol_table env)
	    r);
       print_endline(".");
       print_endline "szs_refutation_stop."))
     else
     	(print_endline "szs_satisfiable.");)));       
   (("print_proof_1", 0),
    ("'print_proof_1' prints the proof if the current result is closed.\n\
       Output format is the same as of 'print_extent_with_proofs'.\n\
      [Output for non-Horn problems and open results are still experimental].\
      \n [Deprecated - old version of print_proof]",
      (fun env term ->
       let proof = Tableau.get_l_proof (Env.get_current_result env) in
       PrologPrint.print 
	 (Fact.proof_term_for_l_proof
	    (Env.get_symbol_table env)
	    proof);
       print_endline("."))));
   (("print_negative_units", 0),
    ("'print_negative_units' prints the negative clauses of the current\n\
       model which are maintained for heuristic purposes, such as complement\n\
      splitting.",
      (fun env term ->
       let rule_sym =
	 (Term.get_rule_symbol (Env.get_symbol_table env))
       and false_term =
	 Term.Constant 
	   (Term.get_false_symbol (Env.get_symbol_table env))
       in
       NegativeUnits.iter_all
	 (fun fact ->
	   PrologPrint.print 
	     (Term.Compound 
		(Term.make_compound rule_sym [|false_term; 
					       (NegativeUnits.FactEntry.get_term fact)|]));
	   print_endline ".")
	 (Env.get_negative_units env))));
   (("print_rules", 0),
    ("'print_rules' prints all rules.",
     (fun env term ->
       (List.iter
	  (fun rule ->
	    PrologPrint.print rule;
	    print_endline ".")
	  (Env.get_rules env)))));
   (("print_declarations", 1),
    ("'print_declarations(p/n)' prints the currently effective declarations\n\
      for the predicate p/n.",
     (fun env term ->
      begin
	let (name, arity) = 
	  Convert.predspec_pair
	    (Convert.term_nth_arg term 0) in
	try 
	  let pred = Env.lookup_predicate env name arity in
	  print_declarations env pred
	with Not_found ->
	  print_endline "% Predicate not found."
      end)));
   (("print_declarations", 0),
    ("'print_declarations' prints the currently effective declarations.",
     (fun env term ->
       (Predicate.PredicateTable.iter
	  (fun symbol predicate -> print_declarations env predicate)
	  (Env.get_predicate_table env)))));
   (("print_settings", 0),
    ("'print_settings' prints the currently effective flag and parameter\n\
       settings.",
     (fun env term ->
      Cfg.print_settings (Env.get_cfg env))));
   (("print_compiled_rules", 0),
    ("'print_compiled_rules' prints all rules after preprocessing,\n\
       especially body literals are sorted heuristically and obeying\n\
       builtins. This command may be useful to generate input for other\n\
       preprocessing operations such as a magic sets transformation.\n\
       [Currently does not properly print literals with findall/3 and\n\
        negation operators other than \\+/1.]",
     (fun env term ->
       ensure_branch_reset env;   
       Tableau.print_compiled_rules
         (Env.get_cfg env)
         (Env.get_predicate_table env)
         (Env.get_symbol_table env)
	 	 (Env.get_negative_units env)
	 	 (Env.get_subterm_table env)
	 	 (Env.get_clause_aux env)
         (Env.get_clauses env))));
   (("show_symbols", 0),
    ("'show_symbols' prints information about symbols in the symbol table.",
     (fun env term ->
      Info.print_symbols_info (Env.get_symbol_table env))));
   (("show_predicates", 0),
    ("'show_predicates' prints information about predicates with respect\n\
      to the current model.",
     (fun env term ->
      Info.print_predicates_info (Env.get_predicate_table env))));
   (("show_extent_size", 1),
    ("'show_extent_size(p/n)' prints information about the extent size of\n\
       predicate p/n in the current model.",
     (fun env term ->
      let (name, arity) = 
	Convert.predspec_pair
	  (Convert.term_nth_arg term 0) in
      print_string "% ";
      print_int
	(Predicate.extent_size (Env.lookup_predicate env name arity));
      print_endline "")));
   (("show_stratification", 0),
    ("show_stratification prints information about the stratification\n\
      of input rules performed by the system.",
    (fun env term ->
      let strata = IcSet.stratify (Env.get_clauses env)
      and n = ref 0 in
      print_string "% ========== Number of strata: ";
      print_int (List.length strata);
      print_endline " ==========";
      List.iter
	(fun stratum ->
	  print_string "% ========== Stratum: ";
	  print_int !n;
	  let stratum_size = (List.length stratum) in
	  print_string " (";
	  print_int stratum_size;
	  if (stratum_size = 1) then print_string " clause)" else print_string " clauses)";
	  print_endline " ==========";
	  List.iter 
	    (fun clause ->
	      PrologPrint.print (Ic.get_source clause);
	      print_endline ".")
	    stratum;
	  n := !n + 1)
      	strata)));
   (("set_flag", 2),
    ("'set_flag(Flag, Boolean)' assigns the given truth value to the flag.\n\
       The boolean can be specified as 'true'/'false' or 'on'/'off'.",
     (fun env term ->
       Cfg.set_flag 
	 (Env.get_cfg env)
	 (Convert.term_to_string (Convert.term_nth_arg term 0))
	 (match (Convert.term_to_string (Convert.term_nth_arg term 1)) with 
	   "true" | "t" | "on" -> true
	 | "false" | "f" | "off" -> false
	 | _ -> raise
	       (Err.Error "Value of set_flag does not specify a boolean.")))));
   (("set_parameter", 2),
    ("'set_parameter(Parameter, Integer)' assigns the given integer value to\n\
       the parameter.",
       (fun env term ->
	 Cfg.set_parameter
	   (Env.get_cfg env)
	   (Convert.term_to_string (Convert.term_nth_arg term 0))
	   (Convert.term_to_int (Convert.term_nth_arg term 1)))));
   (("index", 2),
    ("'index(p/n, [type=Type, args=Args])' declares an index for the\n\
       predicate p/n. Type is one of 'none', 'alist_dt' or 'hash_dt'.\n\
       Args is either 'all' (index on the whole term), an integer n (index\n\
       on the nth argument) or a list of integers (index on a subterm\n\
       composed of the respective arguments in the given ordering).",
	   (fun env term ->
	     ensure_branch_reset env;
	     process_index_decl env term)));
   (("discrimination_hints", 2),
    ("'discrimination_hints(p/n, Hints)' declares Hints as discrimination\n\
      hints for the predicate p/n. Hints is a list of floats between\n\
      0.0 and 1.0 and has length n.",
     (fun env term ->
       ensure_branch_reset env;
       process_disc_hints_decl env term)));
   (("mode", 1),
    ("'mode(Template)' declares the ``mode'' of a predicate. It is used\n\
       for body literal ordering. 'Template' is a term with\n\
       functor and arity of the predicate. The arguments are 'f', which\n\
       means that the respective argument can have nonground values in\n\
       facts with that predicate, or 'b', which means that it has always\n\
       ground values. By default a mode that assigns 'b' to all argument\n\
       positions is assumed. Note that, unless builtins are used, mode\n\
       declarations play only a heuristic role. The effect of different\n\
       mode declarations for a predicate is undefined.",
    (fun env term ->
      ensure_branch_reset env;
      process_mode_decl env term)));
   (("begin_shared_input", 0),
    ("'begin_shared_input' enables sharing of subsequent input terms.\n\
       Input terms are only shared within dynamic begin_shared_input/\n\
       end_shared_input contexts. The behavior for nested\n\
       begin_shared_input/end_shared_input contexts is undefined.",
	   (fun env term ->
	     Env.begin_shared_input env)));
   (("end_shared_input", 0),
    ("'end_shared_input' disables sharing of subsequent input terms\n\
       and frees resources used for sharing.",
	   (fun env term ->
	     Env.end_shared_input env)));
   (("abolish", 1),
    ("'abolish(p/n)' abolishes predicate p/n. This means its extent and\n\
      its declarations are removed.",
     (fun env term ->
       ensure_branch_reset env;
       let (name, arity) = 
	 Convert.predspec_pair
	   (Convert.term_nth_arg term 0) in
       Env.abolish_predicate env name arity)));
   (("clear_rules", 0),
    ("'clear_rules' removes all rules.",
     (fun env term ->
       ensure_branch_reset env;
       Env.clear_rules env)));
   (("load", 1),
    ("'load(Filename)' processes the statements in the specified file.\n\
      Filename can be a relative pathname. If the load command is called\n\
      from an input file, the pathname is resolved using the directory of\n\
      that file. If load is called from standard input, the pathname is\n\
      resolved relative to the current working directory.",	   
     (fun env term ->
       let filename = Convert.term_to_string (Convert.term_nth_arg term 0) in
       process_file env filename)));
   (("load_shell_output", 1),
    ("'load_shell_output(ShellCommand)' runs the given shell command\n\
     in parallel with the krh program. The standard output of the shell\n\
     command is processed as input by krh. The shell command is interpreted\n\
     by /bin/sh.\n\
     Example: #(load_shell_command('canonicalize.pl myfile.tme'))",
     (fun env term ->
       let shell_cmd = Convert.term_to_string (Convert.term_nth_arg term 0) in
       let input = Unix.open_process_in shell_cmd in
       process_in_channel env input)));
   (("onto_file", 2),
    ("'onto_file(StatementList, Filename)' subsequently processes each\n\
       Statement in StatementList with redirecting standard output to the\n\
       specified file.",
     (fun env term ->
       let input_stmts = (Convert.term_nth_arg term 0) in
       let stmts =
	 if Convert.term_is_list input_stmts then
	   Convert.term_list_to_list input_stmts
	 else
	   begin
	     Msg.msg 1 ("Warning: Deprecated use of onto_file/2.");
	     [input_stmts]
	   end
       and filename = Convert.term_to_string (Convert.term_nth_arg term 1) in
       Msg.msg 1 ("Writing file: "^filename^".");
       Fromonto.onto_file
	 (fun () -> 
	   List.iter (fun stmt -> ignore (process_statement env stmt)) stmts)
	 filename
     )));
   (("progn", 1),
    ("'progn(StatementList)' subsequently processes each statement in\n\
       the list.",
     (fun env term ->
       List.iter
	 (fun stmt ->
	   ignore (process_statement env stmt))
	 (Convert.term_list_to_list
	    (Convert.term_nth_arg term 0)))));
   (("do_models", 1),
    ("'do_models(StatementList)' processes the statements in StatementList\n\
       like progn but for each model that is found. During processing of\n\
       the statements, the current model is set to the respective model.",
     do_models_cmd));
   (("write", 1),
    ("'write(String)' writes the content of the given string to the\n\
       standard output (similar to write/1 in Prolog). String is a\n\
       Prolog-syntax atom. Special sequences, starting with the tilde (~),\n\
       are interpreted as formatting actions:\n\
       \n    ~~    Output the tilde itself.\
       \n    ~m    Output the number of the current model.\
       \n    ~z    Flush the output stream.",
     (fun env term ->
       (write_formatted
	  (Convert.term_to_string (Convert.term_nth_arg term 0))))));
   (("assert", 1),
    ("'assert(Statement)' processes Statement.",
     (fun env term ->
      ignore (process_statement env (Convert.term_nth_arg term 0)))));
   (("time", 1),
    ("'time(StatementList)' subsequently processes each Statement in\n\
       StatementList and prints information about the execution time.",
    (fun env term ->
      let input_stmts = (Convert.term_nth_arg term 0) in
      let stmts =
	if Convert.term_is_list input_stmts then
	  Convert.term_list_to_list input_stmts
	else
	  begin
	    Msg.msg 1 ("Warning: Deprecated use of time/1.");
	    [input_stmts]
	  end
      in
      let t0 = Sys.time () in
      List.iter (fun stmt -> ignore (process_statement env stmt)) stmts;
      let t1 = Sys.time () in
      print_string "% CPU Time: ";
      print_float (t1 -. t0);
      print_endline " seconds.")));
   (("set_verbosity", 1),
    ("'set_verbosity(Integer)' assigns a verbosity value, that determines\n\
      which messages are printed. It should range from -1 (no messages)\n\
      to 3 (maximum messages).",
      (fun env term ->
	Cfg.set_verbosity
	  (Convert.term_to_int (Convert.term_nth_arg term 0)))));
   (("set_message_output", 1),
    ("'set_message_output(OutputStream)' determines the output stream on\n\
      which messages are printed. By default messages are printed to the\n\
      standard error stream. Since standard error and standard output use\n\
      different buffers internally, they are not synchronized. OutputStream\n\
      can be 'stdout' or 'stderr'.",
    (fun env term ->
      Cfg.set_msg_output
	(Convert.term_to_string (Convert.term_nth_arg term 0)))));
   (("clear_builtins", 0),
    ("'clear_builtins' disables builtin predicates, arithmetic evaluation\n\
       and logic operators with exception of: :-/2, ','/2, ';'/2, 'false'/0.\n\
       It must be called before any facts or rules are read in.",
      (fun env _ ->
	 (let st = (Env.get_symbol_table env) in
	 Evaluable.clear_evaluables st;
	 Builtin.clear_builtins st;
	 Term.set_true_symbol st (Term.gensym ());
	 Term.set_stratified_neg_as_failure_symbol st (Term.gensym ());
	 Term.set_stratified_neg_as_failure_alternate_symbol
	   st (Term.gensym ());
	 Term.set_stratified_neg_as_failure_not_subsumed_symbol
	   st (Term.gensym ());
	 Term.set_findall_symbol st (Term.gensym ());
	 Term.set_neg2_symbol st (Term.gensym ())))));
   (("sys_gc", 0),
    ("'sys_gv' forces an OCaml garbage collection.",
     (fun env term ->
      Gc.full_major ())));
   (("sys_print_stat", 0),
    ("'sys_print_stat' prints some OCaml system statistics. See\n\
      documentation of Gc.print_stat in the OCaml standard library for\n\
      more information.",
     (fun env term -> Gc.print_stat stdout)));
   (("man", 1),
    ("'man(Section)' prints the specified section of the documentation.\n\
      Currently the sections 'general', 'commands', 'flags' and 'parameters'\n\
      are supported.",
    (fun env term ->
      let page = Convert.term_to_string (Convert.term_nth_arg term 0) in
      match page with
	"general" -> man_general env
      | "commands" -> man_commands env
      | "flags" -> man_flags env
      | "parameters" -> man_parameters env
      | _ -> print_endline "<No such documentation page available>")));
   (("print_version", 0),
    ("'print_version' prints a fact version(Version) with Version an atom\n\
       indicating the version of krh.",
	   (fun env term -> 
	     print_string ("version('" ^ (Release.release ()) ^ "').");
	     print_newline ())));
   (("exit", 0),
    ("'exit' terminates krh with status code 0.",
     (fun env term -> 
       Msg.msg 1 "Bye.";
       exit 0 )))]

and process_meta_statement_1 env term =
  let (_, cmd) =
    try
      List.assoc
	(Convert.term_functor_pair term)
	cmd_table
    with Not_found ->
      raise (Err.Error "Unknown command.") in
  cmd env term

and bigheader x =
  print_string "========================================";
  print_string "======================================\n";
  print_string x;
  print_string "\n========================================";
  print_string "======================================\n\n"

and smallheader x =
  print_string "----------------------------------------";
  print_string "--------------------------------------\n";
  print_string x;
  print_string "\n----------------------------------------";
  print_string "--------------------------------------\n"

and man_general env =
  bigheader "Invocation";
  Printf.printf "    krh [-tty] [FILE]...\n\n";
  Printf.printf 
    "Process the statements in the given files.\n\
     \n\
     With no FILE, or when FILE is -, process standard input.\n\
     \n\
     If standard input is processed, and it is a terminal, the system\n\
     prints a prompt whenever it is ready to read a statement.\n\
     \n\
     Signal SIGINT (CTRL-C) can be used to abort processing of a proving/\n\
     model generation task. In a terminal CTRL-D (end of file) can be\n\
     used to quit the program.\n\
     \n\
     Options:\n\
     \n\
     -tty\
     \n        Forces to handle the standard input in terminal mode,\
     \n        also if it is not actually connected to a terminal.\n\n";
  bigheader "Statements";
  Printf.printf "A statement is either of the form\n\n  #(Command).\n\n\
    which effects that the given command will be executed. Or it is a\n\
    clause that will be asserted.\n\n\
    Statements must be written in canonical ISO-Prolog syntax. Lists\n\
    may also be written in Prolog list notation, using square brackets and\n\
    the vertical bar. Strings in double-quotes are not supported.\n\n";
  flush stdout

and man_commands env =
  bigheader "Commands";
  Printf.printf "Note: Commands starting with 'print_' produce Prolog \
    readable output.\n\n";
  List.iter
    (fun ((name, arity), (doc, _)) ->
      smallheader (Printf.sprintf "Command: %s/%d" name arity);
      Printf.printf "%s\n\n" doc)
    cmd_table;
  flush stdout

and man_flags env =
  bigheader "Flags";
  Cfg.print_flags_doc (Env.get_cfg env);
  flush stdout

and man_parameters env =
  bigheader "Parameters";
  Cfg.print_parameters_doc (Env.get_cfg env);
  flush stdout
    
and help env =
  man_general env;
  man_commands env;
  man_flags env;
  man_parameters env

and process_file env filename =
  let cfg = Env.get_cfg env in 
  let old_cd = Cfg.get_current_directory cfg in
  let absfilename =
    if Filename.is_relative filename then
      Filename.concat old_cd filename
    else
      filename in
  Msg.msg 1 ("Processing file: "^absfilename^".");
  let input =  open_in_bin absfilename in
  try 
    Cfg.set_current_directory cfg (Filename.dirname absfilename);
    process_in_channel env input
  with e ->
    Cfg.set_current_directory cfg old_cd;
    close_in input;
    raise e
    
and process_in_channel env input =
  let p =
    PrologRead.prolog_parser
      (Env.get_symbol_table env)
      (Stream.of_channel input) in 
      try
		  (Stream.iter 
		    (fun term -> if process_statement env term then () else exit 0)
		    p)
		with Stream.Failure -> ();
  close_in input
      
and do_models_cmd =
  (fun env term ->
    let stmts = 
      Convert.term_list_to_list
	(Convert.term_nth_arg term 0) in
    ensure_branch_reset env;
    let t0 = Sys.time () 
    and sat = ref false
    in
    let result =
      ref (Tableau.first_result
             (Env.get_cfg env)
             (Env.get_predicate_table env)
             (Env.get_symbol_table env)
	     	 (Env.get_negative_units env)
	     	 (Env.get_subterm_table env)
	     	 (Env.get_clause_aux env)
             (Env.get_clauses env)) in
    while (Tableau.is_open !result) do
      sat := true;
      model_count := !model_count + 1;
      Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0);
      szs_output env !sat;
      Env.set_current_result env !result;
      Tableau.info_msg !result;
      List.iter
	(fun stmt ->
	  ignore (process_statement env stmt))
	stmts;
      result := Tableau.next_result !result;
    done;
    model_count := 0;
    Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0);
    Env.set_current_result env !result;
    Tableau.info_msg !result;
    szs_output env !sat)


(****************************************************************************)
(* Interactive Processing. *)
(****************************************************************************)

let rec connected_to_terminal () =
  isatty (Unix.descr_of_in_channel stdin)
   
and isatty fd =
  match Sys.os_type with
    "Unix" ->
      (try 
	ignore (Unix.tcgetattr fd);
	true
      with Unix.Unix_error (Unix.ENOTTY,_,_) 
      |  Unix.Unix_error (Unix.EINVAL,_,_) -> 
	  false
      | Unix.Unix_error (e, _, _) ->
	  raise (Failure ("Isatty hack failure: "^(Unix.error_message e))))
  | _ -> false

let prompt () =
  print_string "KRH> ";
  flush stdout
    
let process_stdin env force_tty =
  let tty = force_tty || connected_to_terminal () in
  if tty then
    begin
      Msg.msg 1 ("Welcome to KRH - " ^ (Release.release ()));
      Msg.msg 1 "Enter #(help). for help."
    end;
  let p =
    ref (PrologRead.prolog_parser
	   (Env.get_symbol_table env)
	   (Stream.of_channel stdin)) in
  let loop = ref true in
  while !loop
  do
    if tty then
      prompt ();
    try
      let term =
	try
	  Stream.next !p
	with Stream.Failure ->
	  (* ### how to find out if stdin is at EOF ? Stream.is_empty
	     is not suited, since it seems to parse the next term. *)
	  loop := false;
	  raise Stream.Failure
	| e ->
	  p := (PrologRead.prolog_parser
		  (Env.get_symbol_table env)
		  (Stream.of_channel stdin));
	  raise (Failure ("Syntax Error: "^(Printexc.to_string e)))
      in
      loop := process_statement env term
    with e -> (* ### Error handling *)
      if !loop then
	Msg.msg 0 ("Error: "^(Printexc.to_string e))
      else
	()
  done;
  if tty then
    Msg.msg 1 "Bye."



(****************************************************************************)
(* Invocation. *)
(****************************************************************************)

let is_option string =
  String.length string > 1 && String.get string 0 = '-'

let main () =
  let force_tty = ref false 
  and env = Env.make_env (Cfg.make_cfg ())
  and n = (Array.length Sys.argv) - 1
  and i = ref 1 in
  while !i <= n && is_option Sys.argv.(!i)  do
    begin
      let option = Sys.argv.(!i) in
      match option with
	"-tty" ->
	  force_tty := true;
      | _ ->
	  prerr_string "krh: invalid option '";
	  prerr_string option;
	  prerr_endline "'.";
	  exit 1
    end;
    i := !i + 1
  done;
  if !i > n then
    process_stdin env !force_tty
  else
    for j = !i to n do
      let arg = Sys.argv.(j) in
      if arg = "-tty" then
	force_tty := true
      else
	if arg = "-" then
	  process_stdin env !force_tty
	else
	  process_file env arg
    done
  
let interact () =
  let env = Env.make_env (Cfg.make_cfg ()) in
  process_stdin env true

