(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* This is based on the code of the OCaml Genlex module. See also module 
   prologPrint. *)

type prolog_token =
    PunctuationPrologToken of string
  | VarPrologToken of string
  | ConstPrologToken of string
  | IntPrologToken of int
  | FloatPrologToken of float

(* The string buffering machinery *)

let initial_buffer = String.create 32

let buffer = ref initial_buffer

let bufpos = ref 0

let reset_buffer () = buffer := initial_buffer; bufpos := 0

let store c =
  if !bufpos >= String.length !buffer then
    begin
      let newbuffer = String.create (2 * !bufpos) in
      String.blit !buffer 0 newbuffer 0 !bufpos; buffer := newbuffer
    end;
  String.set !buffer !bufpos c;
  incr bufpos

let get_string () =
  let s = String.sub !buffer 0 !bufpos in buffer := initial_buffer; s

(* The lexer *)

let prolog_lexer =
  let kwd_table = Hashtbl.create 17 in
  List.iter (fun s -> Hashtbl.add kwd_table s (PunctuationPrologToken s))
    (* Punctuation Chars, except '%' correspond to keywords *)
    ["("; ")"; ","; "["; "]"; "{"; "|"; "}"];
  let const_or_keyword id =
    try Hashtbl.find kwd_table id with
      Not_found -> ConstPrologToken id
  and keyword_or_error c =
    let s = String.make 1 c in
    try Hashtbl.find kwd_table s with
      Not_found -> raise (Stream.Error ("Illegal character " ^ s))
  in
  let rec next_token (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      (* Layout Char *)
      Some ('\000'..'\032' | '\127'..'\159' as c) ->
        Stream.junk strm__; next_token strm__
    | (* CapitalLetter or Underline *)
      Some
	('\065'..'\090' | '\192'..'\214' | '\216'..'\222' | '\095' as c) ->
	  Stream.junk strm__;
          let s = strm__ in reset_buffer (); store c; var s
    | (* SmallLetter *)
      Some
	('\097'..'\122' | '\223'..'\246' | '\248'..'\255' as c) ->
          Stream.junk strm__;
          let s = strm__ in reset_buffer (); store c; const s
    | (* SymbolChar, except '-' (\045) and '/' (\047) *)
      Some
	('\035' | '\036' | '\038' | '\042' | '\043' | '\046' |
	'\058' | '\060' | '\061' | '\062' | '\063' | '\064' |
	'\092' | '\094' | '\096' | '\126' | '\215' | '\247' |
	'\160'..'\191' as c) ->
          Stream.junk strm__;
          let s = strm__ in reset_buffer (); store c; const2 s
    | (* SoloChar, i.e.  '!' (\033) and ';' (\059) *)
      Some
	('\033' | '\059' as c ) ->
          Stream.junk strm__;
          let s = strm__ in reset_buffer ();
	  Some (ConstPrologToken (String.make 1 c)) 
    | Some '\'' ->
        Stream.junk strm__;
        let s = strm__ in reset_buffer (); Some (ConstPrologToken (quoted s))
    | Some ('0'..'9' as c) ->
        Stream.junk strm__;
        let s = strm__ in reset_buffer (); store c; number s
    | Some '-' -> Stream.junk strm__; neg_number strm__
    | Some '%' -> Stream.junk strm__; skip_to_eol strm__
    | Some '/' -> Stream.junk strm__; maybe_comment strm__
    | Some c -> Stream.junk strm__; Some (keyword_or_error c)
    | _ -> None
  and var (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
    (* CaptitalLetter, Underline, SmallLetter, or Number *)
      Some
	('\065'..'\090' | '\192'..'\214' | '\216'..'\222' | '\095' |
	'\097'..'\122' | '\223'..'\246' | '\248'..'\255' |
	'0'..'9' as c) ->
	  Stream.junk strm__; let s = strm__ in store c; var s
    | _ ->  Some (VarPrologToken (get_string ()))
  and const (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
    (* CaptitalLetter, Underline, SmallLetter, or Number *)
      Some
	('\065'..'\090' | '\192'..'\214' | '\216'..'\222' | '\095' |
	'\097'..'\122' | '\223'..'\246' | '\248'..'\255' |
	'0'..'9' as c) ->
	  Stream.junk strm__; let s = strm__ in store c; const s
    | _ -> Some (ConstPrologToken (get_string ()))
  and const2 (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      (* SymbolChar, except '.' (\046) *)
      Some
	('\035' | '\036' | '\038' | '\042' | '\043' |
	'\058' | '\060' | '\061' | '\062' | '\063' | '\064' |
	'\092' | '\094' | '\096' | '\126' | '\215' | '\247' |
	'\160'..'\191' |
	'-' | '/' as c) ->
          Stream.junk strm__; let s = strm__ in store c; const2 s
    | _ -> Some (const_or_keyword (get_string ()))
  and neg_number (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some ('0'..'9' as c) ->
        Stream.junk strm__;
        let s = strm__ in reset_buffer (); store '-'; store c; number s
    | _ -> let s = strm__ in reset_buffer (); store '-'; const2 s
  and number (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some ('0'..'9' as c) ->
        Stream.junk strm__; let s = strm__ in store c; number s
    | Some '.' ->
	let s = strm__ in  maybe_decimal_part s
    | Some ('e' | 'E') ->
        Stream.junk strm__; let s = strm__ in store 'E'; exponent_part s
    | _ -> 
	Some (IntPrologToken (int_of_string (get_string ())))
  and maybe_decimal_part (strm__ : _ Stream.t) =
    try 
      let second = List.nth (Stream.npeek 2 strm__) 1 in
	match second with
	  ('0'..'9' as c) ->
            Stream.junk strm__;
	    Stream.junk strm__; 
	    let s = strm__ in
	    store '.';
	    store c; 
	    decimal_part s
	| _ -> Some (IntPrologToken (int_of_string (get_string ())))
    with (Failure _) ->
      Some (IntPrologToken (int_of_string (get_string ())))
  and decimal_part (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some ('0'..'9' as c) ->
        Stream.junk strm__; let s = strm__ in store c; decimal_part s
    | Some ('e' | 'E') ->
        Stream.junk strm__; let s = strm__ in store 'E'; exponent_part s
    | _ -> Some (FloatPrologToken (float_of_string (get_string ())))
  and exponent_part (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some ('+' | '-' as c) ->
        Stream.junk strm__; let s = strm__ in store c; end_exponent_part s
    | _ -> end_exponent_part strm__
  and end_exponent_part (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some ('0'..'9' as c) ->
        Stream.junk strm__; let s = strm__ in store c; end_exponent_part s
    | _ -> Some (FloatPrologToken (float_of_string (get_string ())))
  and quoted (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some '\'' -> Stream.junk strm__; get_string ()
    | Some '\\' ->
        Stream.junk strm__;
        let c =
          try escape strm__ with
            Stream.Failure -> raise (Stream.Error "")
        in
        let s = strm__ in store c; quoted s
    | Some c -> Stream.junk strm__; let s = strm__ in store c; quoted s
    | _ -> raise Stream.Failure
  and escape (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some 'n' -> Stream.junk strm__; '\n'
    | Some 'r' -> Stream.junk strm__; '\r'
    | Some 't' -> Stream.junk strm__; '\t'
    | Some ('0'..'9' as c1) ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some ('0'..'9' as c2) ->
            Stream.junk strm__;
            begin match Stream.peek strm__ with
	      Some ('0'..'9' as c3) ->
                Stream.junk strm__;
                Char.chr
                  ((Char.code c1 - 48) * 100 + (Char.code c2 - 48) * 10 +
                     (Char.code c3 - 48))
            | _ -> raise (Stream.Error "")
            end
        | _ -> raise (Stream.Error "")
        end
    | Some c -> Stream.junk strm__; c
    | _ -> raise Stream.Failure
  and maybe_comment (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some '*' ->
        Stream.junk strm__; let s = strm__ in comment s; next_token s
    | _ -> let s = strm__ in reset_buffer (); store '/'; const2 s
  and comment (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some '*' -> Stream.junk strm__; maybe_end_comment strm__
    | Some c -> Stream.junk strm__; comment strm__
    | _ -> raise Stream.Failure
  and maybe_end_comment (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some '/' -> Stream.junk strm__; ()
    | Some '*' -> Stream.junk strm__; maybe_end_comment strm__
    | Some c -> Stream.junk strm__; comment strm__
    | _ -> raise Stream.Failure
  and skip_to_eol (strm__ : _ Stream.t) =
    match Stream.peek strm__ with
      Some '\n' ->
	Stream.junk strm__; next_token strm__
    | Some c -> Stream.junk strm__; skip_to_eol strm__
    | _ -> raise Stream.Failure	  
  in
  fun input -> Stream.from (fun count -> next_token input)

