(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


(* This is a bit complicated since OCAML has no string streams, and
  also %a and %t seem work with sprintf not in the same way as with
  printf. *)

open Term

module type OutputType =
  sig
    type t
    val	output_char: t -> char -> unit
    val output_string: t -> string -> unit
  end

(* These character classes are taken from the Sicstus Prolog manual.
   But that seems not fully correct, e.g [] and {} are not considered.
   See also module GenlexProlog. *)

type char_class =
    CharClassLayout
  | CharClassSmallLetter
  | CharClassCapitalLetter
  | CharClassDigit
  | CharClassSymbolChar
  | CharClassSoloChar
  | CharClassPunctuationChar
  | CharClassQuoteChar
  | CharClassUnderline
  | CharClassUnspecified

let classify_char c =
  match c with
    ('\000'..'\032' | '\127'..'\159') ->
      CharClassLayout
  | ('\097'..'\122' |'\223'..'\246' | '\248'..'\255') ->
      CharClassSmallLetter
  | ('\065'..'\090' |'\192'..'\214' | '\216'..'\222') ->
      CharClassCapitalLetter
  | ('\048'..'\057') ->
      CharClassDigit
  | ('\035' | '\036' | '\038' | '\042' | '\043' | '\045' | '\046' |
    '\047' | '\058' | '\060' | '\061' | '\062' | '\063' | '\064' |
    '\092' | '\094' | '\096' | '\126' | '\215' | '\247' |
    '\160'..'\191') ->
      CharClassSymbolChar
  | ('\033' | '\059') ->
      CharClassSoloChar
  | ('\037' | '\040' | '\041' | '\044' | '\091' | '\093' | '\123' |
    '\124' | '\125') ->
      CharClassPunctuationChar
  | ('\034' | '\039') ->
      CharClassQuoteChar
  | '\095' ->
    CharClassUnderline

let rec alphanum_string s i =
  if i >= String.length s then
    true
  else 
    match classify_char (String.get s i) with
      CharClassCapitalLetter 
    | CharClassSmallLetter 
    | CharClassDigit 
    | CharClassUnderline ->
	alphanum_string s (i+1)
    | _ -> false
    
let rec symbol_string s i =
  if i >= String.length s then
    true
  else 
    match classify_char (String.get s i) with
      CharClassSymbolChar ->
	symbol_string s (i+1)
    | _ -> false


module Make (Output: OutputType) =
  struct
    let output_name oc s =
      (* Just considers \ and ' as escape and quote characters, anything
	 else is printed literally. *)
      if String.length s == 0 then
	Output.output_string oc "''"
      else if 
	(( match classify_char (String.get s 0) with
	  CharClassSmallLetter ->
	    alphanum_string s 1
	| _ -> 
	    false )
       || (s <> "." && symbol_string s 0 && 
	   not (String.get s 0 == '/' 
		  && (String.length s) > 1
		  && String.get s 1 == '*'))
       || s = "!"
       || s = ";"
       || s = "[]" 
       || s = "{}" ) then
	Output.output_string oc s
      else
	begin
	  Output.output_char oc '\'';
	  for i = 0 to (String.length s)-1 do
	    let c = String.get s i in
	    if c == '\'' then
	      Output.output_string oc "\\'"
	    else if c == '\\' then
	      Output.output_string oc "\\\\"
	    else
	      Output.output_char oc c
	  done;
	  Output.output_char oc '\'';
	end
	  
    let rec output oc t =
      match t with
	Variable v ->
	  Output.output_string oc "X";
	  Output.output_string oc (string_of_int (variable_number v))
      | Constant s ->
	  output_name oc (symbol_name s)
      | Compound k ->
	  let name = (symbol_name (compound_functor k))
	  and arity = (symbol_arity (compound_functor k)) in
	  if name = "." && arity = 2 then
	    begin
	      Output.output_string oc "[";
	      output oc (compound_args k).(0);
	      output_cdr oc (compound_args k).(1);
	      Output.output_string oc "]"
	    end
	  else
	    begin
	      output_name oc name;
	      Output.output_string oc "(";
	      output oc (compound_args k).(0);
	      for i = 1 to (symbol_arity (compound_functor k))-1 do
		Output.output_string oc ", ";
		output oc (compound_args k).(i)
	      done;
	      Output.output_string oc ")"
	    end
      | Integer n ->
	  Output.output_string oc (string_of_int n)
      | Float n ->
	  let s1 = (string_of_float n) in
	  let s2 = 
	    if String.get s1 ((String.length s1) - 1) = '.' then
	      s1^"0"
	    else
	      s1 in
	  Output.output_string oc s2
      | Unbound ->
	  Output.output_string oc "'<unbound>'"
	    
    and output_cdr oc t =
      match t with
	Compound k ->
	  let name = (symbol_name (compound_functor k))
	  and arity = (symbol_arity (compound_functor k)) in
	  if name = "." && arity = 2 then
	    begin
	      Output.output_string oc ", ";
	      output oc (compound_args k).(0);
	      output_cdr oc (compound_args k).(1)
	    end
	  else
	    begin
	      Output.output_string oc " | ";
	      output oc t
	    end
      | Constant s ->
	  if (symbol_name s) = "[]" then
	    ()
	  else
	    begin
	      Output.output_string oc " | ";
	      output oc t
	    end
      | _ ->
	  Output.output_string oc " | ";
	  output oc t
  end	

module OutChannelPrint =
  Make(
  struct
    type t = Pervasives.out_channel
    let output_char = Pervasives.output_char
    let output_string = Pervasives.output_string
  end)

let print term =
  OutChannelPrint.output stdout term
	
let output out term =
  OutChannelPrint.output out term

let sbuff = Buffer.create 64

module StringPrint =
  Make(
  struct
    type t = unit
    let output_char oc char =
      Buffer.add_char sbuff char
    let output_string oc string =
      Buffer.add_string sbuff string
  end)

let print_to_string term =
  Buffer.reset sbuff;
  StringPrint.output () term;
  Buffer.contents sbuff


 
