(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

open PrologGenlex
open Term

exception At_end

(* Hashtable might be a bit heavy here. *)

let make_vartable () = Hashtbl.create 8

let vartable_find vt = Hashtbl.find vt

let vartable_add vt k v = Hashtbl.add vt k v


let prolog_parser symtable =

  let rec next_clause (s : prolog_token Stream.t) =
    try
      let t = next_term true (make_vartable ()) (ref (-1)) s in
      match Stream.peek s with
	Some ConstPrologToken "." ->
	  Stream.junk s;
	  Some t
      | _ -> raise (Stream.Error ("Missing full stop."))
    with At_end ->
      None

  and next_term toplevelp vt vc (s : prolog_token Stream.t) =
    match Stream.peek s with
	Some VarPrologToken varname ->
	  if varname = "_" then
	    begin
	      vc := !vc + 1;
	      if !vc >= Term.max_vars then
		raise Term.Max_vars_exceeded;
	      Stream.junk s;
	      Variable (make_variable !vc)
	    end
	  else
	    begin
	      try 
		let vnum = vartable_find vt varname in
		Stream.junk s; 
		Variable (Term.make_variable vnum)
	      with Not_found ->
		vc := !vc + 1;
		if !vc >= Term.max_vars then
		  raise Term.Max_vars_exceeded;
		vartable_add vt varname !vc;
		Stream.junk s;
		Variable (make_variable !vc)
	    end
      | Some IntPrologToken n -> Stream.junk s; Integer n
      | Some FloatPrologToken n -> Stream.junk s; Float n
      | Some ConstPrologToken z ->
	  Stream.junk s;
	  parse_const vt vc z s
      | Some PunctuationPrologToken "[" ->
	  Stream.junk s;
	  begin
	    match Stream.peek s with
	      Some PunctuationPrologToken "]" ->
		Stream.junk s;
		Constant (find_symbol symtable "[]" 0)
	    | _ -> 
		let hd = next_term false vt vc s in
		construct_compound "." [hd;parse_list vt vc s]
	  end
      | Some _ ->
	  raise (Stream.Error "")
      | None ->
	  if toplevelp then
	    raise At_end
	  else raise (Stream.Error "")

  and parse_const vt vc z s =
    match Stream.peek s with
      Some PunctuationPrologToken "(" -> 
	Stream.junk s;
	let hd = next_term false vt vc s in
	construct_compound z (hd :: parse_args vt vc s)
(*    | _ -> Constant (find_symbol symtable z 0)*)
	  | _ ->
	  	let sym = (find_symbol symtable z 0)
	  	in
	  		if sym == (get_t_constant_symbol symtable) then
	  			(Unbound)
	  		else
	  			(Constant (sym))

  and parse_args vt vc s =
    match Stream.peek s with
      Some PunctuationPrologToken "," ->
	Stream.junk s;
	let hd = next_term false vt vc s in
	hd :: parse_args vt vc s
    | Some PunctuationPrologToken ")" ->
	Stream.junk s;
	[]
    | _ ->
	raise (Stream.Error "Bad compound arguments.")

  and parse_list vt vc s =
    match Stream.peek s with
      Some PunctuationPrologToken "," ->
	Stream.junk s;
	let hd = next_term false vt vc s in
	construct_compound "." [hd;parse_list vt vc s]
    | Some PunctuationPrologToken "|" ->
	Stream.junk s;
	let result = next_term false vt vc s in
	begin
	  match Stream.peek s with
	    Some PunctuationPrologToken "]" ->
	      Stream.junk s;
	      result
	  | _ -> raise (Stream.Error ("Bad cons."))
	end
    | Some PunctuationPrologToken "]" ->
	Stream.junk s;
	Constant (find_symbol symtable "[]" 0)
    | _ ->
	raise (Stream.Error "Bad list.")


  and construct_compound z args =
    Compound (make_compound
		(find_symbol symtable z (List.length args))
		(Array.of_list args))

  in
  fun input -> 
    Stream.from (fun count -> next_clause (prolog_lexer input))

