(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(* let print_debug x = print_string "DEBUG: "; print_endline x *)

let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "
let pi i = print_int i

let print_current_compiled_rules subterm_table =
       (let clauses = ref []
       in
       Ic.SubtermTable.iter
		  (fun symbol subterm -> 
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ())
	  subterm_table;
	  List.iter 
	  	(fun c -> 
	  		ps (Ic.get_id c);
	  		if (Ic.is_eq_unit c) then 
	  			(if (Ic.is_oriented c) then ps " reductive"; ps " (EQ-UNIT)");
	  		ps " : ";ppp (Ic.get_source c);
	  		pe ".") 
	  	!clauses
	  (*Info.print_clauses_as_terms !clauses*)	)

let print_debug x = ()
 



exception Found
exception Failed

(****************************************************************************)
(* State *)
(****************************************************************************)

module State =
  struct

    type state =
	{ extent_state:Access.Store.state list;
	  negative_units_state:NegativeUnits.state;
	  subterm_state:Ic.SubtermStore.state list
	}

    let obtain_extent_state predicate_table =
      let states = ref [] in
      Predicate.PredicateTable.iter
	(fun psymbol predicate ->
	  List.iter 
	    (fun writer -> 
	      states := 
		Access.Store.obtain_state (Access.Writer.get_store writer) ::
		!states)
	    (Predicate.get_writers predicate))
	predicate_table;
      !states

    let obtain_subterm_state subterm_table =
      let states = ref [] in
	      Ic.SubtermTable.iter
			(fun subterm_symbol subterm_entry ->
		    	states := 
					Ic.SubtermStore.obtain_state (Ic.Writer.get_store (Ic.get_subterm_writer subterm_entry)) ::
					!states)
			subterm_table;	
	      !states

	      
	          let obtain_subterm_state2 subterm_table =
      let states = ref [] in
	      Ic.SubtermTable.iter
			(fun subterm_symbol subterm_entry ->
		    	states := 
					Ic.SubtermStore.obtain_state (Ic.Writer.get_store (Ic.get_subterm_writer subterm_entry)) ::
					!states)
			subterm_table;
	      !states
	      

    
    let obtain_state predicate_table negative_units subterm_table =
      { extent_state = obtain_extent_state predicate_table;
	negative_units_state = NegativeUnits.obtain_state negative_units;
	subterm_state = obtain_subterm_state subterm_table}


(*	
	let adda s = pe "ADD TO EXTENT"; Access.Store.add_layer s
	let adds s = pe "ADD TO SUBTERMS"; Ic.SubtermStore.add_layer s
	
	let add_layer state =
      { extent_state = List.map adda state.extent_state;
		negative_units_state =	NegativeUnits.add_layer state.negative_units_state;
		subterm_state = List.map adds state.subterm_state}
*)		
	
    let add_layer state =
      { extent_state = List.map Access.Store.add_layer state.extent_state;
		negative_units_state =	NegativeUnits.add_layer state.negative_units_state;
		subterm_state = List.map Ic.SubtermStore.add_layer state.subterm_state}
	
    let install_state state =
      List.iter Access.Store.install_state state.extent_state;
      NegativeUnits.install_state state.negative_units_state;
      List.iter Ic.SubtermStore.install_state state.subterm_state

  end


(****************************************************************************)
(* Refutation *)
(****************************************************************************)

type refutation =
    NullRefutation
  | PrimitiveRefutation of Fact.proof
  | ComposedRefutation of (refutation list)


(****************************************************************************)
(* Aux *)
(****************************************************************************)

module Aux =
  struct
    type aux = {
	predicate_table:Predicate.predicate_table;
	symbol_table:Term.symbol_table;
	symbol_list:Term.symbol list;
	cfg:Cfg.cfg;
	clause_aux:Ic.Aux.aux;
	news:News.news;
	negative_units:NegativeUnits.store;
	subterm_table:Ic.subterm_table;
	input_state:State.state;
	use_proof_terms:bool;
	use_negative_units:bool;
	use_level_cut:bool;
	mutable proofs:Fact.proof list; (* deprecated *)
	mutable refutation:refutation;
	mutable stat_disjunction_nodes:int;
	mutable stat_stretchings:int;
	mutable stat_closed_branches:int;
	mutable stat_models:int;
	mutable stat_eval_rounds:int;
	mutable t_start:float;
	mutable t_model_start:float;
	mutable t_round_start:float;
	mutable n_strata:int;
	mutable stat_max_weight:int;
	mutable stat_depth:int;
	mutable stat_level_cuts:int;
	new_clause_candidates: Ic.NewClauseCandidates.new_clause_candidates
      }

    let make_aux cfg predicate_table symbol_table negunits subterm_table clause_aux = (*pe "Making aux.";*)
      let news = News.make_news cfg (Term.get_krh_dom_symbol symbol_table) in
      let input_state =
	State.obtain_state predicate_table negunits subterm_table
	  and symbol_list =
	  	(let pf_symbols = Term.get_symbol_list symbol_table
	  	and p_symbols = Predicate.get_predicate_symbols predicate_table
	  	in
	  		List.filter
	  			(fun sym ->
	  				not (List.memq sym p_symbols))
	  			pf_symbols)
	  in
      { predicate_table = predicate_table;
	symbol_table = symbol_table;
	symbol_list = symbol_list;
	cfg = cfg;
	clause_aux = clause_aux;
	news = news;
	negative_units = negunits;
	subterm_table = subterm_table;
	input_state = input_state;
	use_proof_terms = Cfg.get_proof_terms_flag cfg;
	use_negative_units = Cfg.get_negative_units_flag cfg;
	use_level_cut =
	Cfg.get_proof_terms_flag cfg &&	Cfg.get_level_cut_flag cfg;
	proofs = [];
	refutation = NullRefutation;
	stat_disjunction_nodes = 0;
	stat_stretchings = 0;
	stat_closed_branches = 0;
	stat_models = 0;
	stat_eval_rounds = 0;
	t_start=0.0;
	t_round_start=0.0;
	t_model_start=0.0;
	n_strata=0;
	stat_max_weight=(-1);
	stat_depth = 0;
	stat_level_cuts = 0;
	new_clause_candidates = Ic.NewClauseCandidates.make_new_clause_candidates cfg
      }


 		
    let inc_stat_disjunction_nodes aux =
      aux.stat_disjunction_nodes <- aux.stat_disjunction_nodes + 1
      (*
      Msg.msg1 3 "                Disjunction Node (#%d)."
	aux.stat_disjunction_nodes
       *)

    let inc_stat_stretchings aux =
      aux.stat_stretchings <- aux.stat_stretchings + 1;
      Msg.msg2 3 "                Resetting max weight to %d (#%d)."
	aux.stat_max_weight
	aux.stat_stretchings

    let inc_stat_closed_branches aux =
      aux.stat_closed_branches <- aux.stat_closed_branches + 1
      (*	  
      Msg.msg1 3 "                Branch closed (#%d)."
	aux.stat_closed_branches *)

    let inc_stat_models aux =
      aux.stat_models <- aux.stat_models + 1

    let inc_stat_eval_rounds aux =
      aux.stat_eval_rounds <- aux.stat_eval_rounds + 1

    let inc_stat_level_cuts aux =
      aux.stat_level_cuts <- aux.stat_level_cuts + 1;
      Msg.msg1 3 "                Level cut (#%d)."
	aux.stat_level_cuts

	
    let info_msg aux =
      let t = Sys.time () in
      Msg.msg1 2 "Model number:       %9d." aux.stat_models;
      Msg.msg1 2 "Time:               %8.3fs." (t -. aux.t_start);
      Msg.msg1 2 "Time in last model: %8.3fs." (t -. aux.t_model_start);
      Msg.msg1 2 "Disjunction nodes:  %9d." aux.stat_disjunction_nodes;
      Msg.msg1 2 "Level cuts:         %9d." aux.stat_level_cuts;
      Msg.msg1 2 "Tableau depth:      %9d." aux.stat_depth;
      Msg.msg1 2 "Closed branches:    %9d." aux.stat_closed_branches;
      Msg.msg1 2 "Evaluation rounds:  %9d." aux.stat_eval_rounds;
      Msg.msg1 2 "Max weight resets:  %9d." aux.stat_stretchings;
      Msg.msg1 2 "Latest max weight:  %9d." aux.stat_max_weight;
  end

  

(****************************************************************************)
(* Stratum *)
(****************************************************************************)

module Stratum =
  struct

    type stratum = {
	plain_clauses:IcSet.clauses;
	inc_clauses:IcSet.clauses;
	delta_predicates:Predicate.predicate list;
	mutable needs_compilation:bool;
	aux:Aux.aux
      }

    let make_stratum aux plain_clauses =
      let plain_clauses =
	if aux.Aux.use_negative_units then
	  IcSet.negative_units_to_store aux.Aux.negative_units plain_clauses
	else
	  plain_clauses in
      let delta_predicates = (IcSet.idb_predicates plain_clauses) in
      { plain_clauses=IcSet.sort_clauses plain_clauses;
	inc_clauses=
	IcSet.sort_clauses
	  (IcSet.make_eval_inc_clauses plain_clauses delta_predicates);
	delta_predicates=delta_predicates;
	needs_compilation=true;
	aux=aux
      }

    let print_fat_line () =
      print_string "============================";
      print_endline "=================================================="
    
    let print_task_debug_info
	predicate_table
	delta_predicates
	plain_clauses
	inc_clauses =
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: BEGIN";
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: PREDICATES";
      print_fat_line ();
      Info.print_predicates_info predicate_table;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: PLAIN CLAUSES";
      print_fat_line ();
      Info.print_clauses_info plain_clauses;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: INC CLAUSES";
      print_fat_line ();
      Info.print_clauses_info inc_clauses;
      print_fat_line ();
      print_endline "COMPILATION INFORMATION: END";
      print_fat_line ()
	
   (* This can be called at any time to re-compile a clause set before
      evaluation. Actual extent sizes then are considered for 
      join ordering. It must be called at least once before evaluation. *)
    let ensure_compiled stratum = (*print_endline "ensure_compiled:";*)
      if stratum.needs_compilation then
	let aux = stratum.aux in
	let news = aux.Aux.news in	(*print_string "news_id: "; print_endline (News.get_news_id news);*)
	Msg.msg1 2 "Compiling stratum (%d rules)."
	  (List.length stratum.plain_clauses);
	List.iter (fun clause -> Ic.initialize_clause clause news)
	  stratum.plain_clauses;
	List.iter (fun clause -> Ic.initialize_clause clause news)
	  stratum.inc_clauses;
	stratum.needs_compilation <- false;
	Msg.msg 2 "Compilation of stratum finished.";
	if Cfg.get_debug_flag aux.Aux.cfg then
	  print_task_debug_info 
	    aux.Aux.predicate_table
	    stratum.delta_predicates
	    stratum.plain_clauses 
	    stratum.inc_clauses	    
	    
    let print_compiled_clauses stratum =
      ensure_compiled stratum;
      Info.print_clauses_as_terms stratum.plain_clauses

      
      
    			      
      
      
      
      
  end


(****************************************************************************)
(* Node *)
(****************************************************************************)

type literal = 
    Literal of News.Disjunct.disjunct
  | UnboundLiteral

type node =
    { mutable strata:Stratum.stratum list;
      mutable post_eval_strata:Stratum.stratum list;
      mutable delayed_units:(Predicate.predicate * Fact.fact) list;
      mutable delayed_rules:Ic.clause list;
      aux:Aux.aux;
      mutable max_weight:int;
      parent:node;
      mutable state:State.state;
      mutable initial_news_state:News.state;
      literal: literal;
      right_brother_literals:News.Disjunct.disjunct list;
      mutable stretch_ancestor:stretch_ancestor;
      mutable left_brother_refutations: refutation list;
      mutable refutation: refutation;
      depth:int;
      redundant_store: Ic.redundant_store
    }


(****************************************************************************)
(* Stretch Ancestor *)

and stretch_ancestor =
    StretchAncestor of node * int (* node * weight_increment *)
  | UnboundStretchAncestor

let make_stretch_ancestor node weight_increment =
  StretchAncestor (node, weight_increment)

(****************************************************************************)
(* Weight Management Parameters *)

let wm_initial_max_weight cfg =
  Cfg.get_max_weight_initial_param cfg

let wm_initial_max_weight_next_stratum node =
    node.max_weight

let wm_child_max_weight n = n

let wm_increase_max_weight_of_stretch_root aux node required_increment = 
  let max_weight =
    node.max_weight + 
      (max 
	 (Cfg.get_max_weight_increment_param aux.Aux.cfg)
	 required_increment) in
  node.max_weight <- max_weight;
  aux.Aux.stat_max_weight <- max_weight


(****************************************************************************)
(* Node Construction and Access *)

let make_initial_node clauses aux = (*print_int (List.length clauses);*)
  let cfg = aux.Aux.cfg 
  and news = aux.Aux.news in
  let strata =
    List.map 
      (fun clauses1 -> Stratum.make_stratum aux clauses1) 
      (IcSet.stratify clauses) in
  aux.Aux.n_strata <- (List.length strata);
  

  Msg.msg1 2 "Number of strata: %5d" (List.length strata);
  let rec node =
    { strata = strata;
      post_eval_strata = [];
      delayed_units = [];
      delayed_rules = [];
      aux = aux;
      max_weight = wm_initial_max_weight cfg;
      parent = node; (* self for "unbound" *)
      state = 
      if Cfg.get_merge_inf_input_flag cfg then
	aux.Aux.input_state
      else
	State.add_layer aux.Aux.input_state;
      initial_news_state = News.obtain_state news;
      literal = UnboundLiteral;
      right_brother_literals = [];
      stretch_ancestor = UnboundStretchAncestor;
      left_brother_refutations = [];
      refutation = NullRefutation;
      depth = 0;
      redundant_store = Ic.make_redundant_store ()} in
  node

let has_next_stratum node =
  match node.strata with
    _ :: _ :: _ -> true
  | [_] | [] -> false

let make_next_stratum_node node news_state =
  let strata =
    match node.post_eval_strata with
      _ :: strata1 -> strata1 
    | _ -> raise (Failure "No next stratum.");
  in
  { strata = strata;
    post_eval_strata = [];
    delayed_units = node.delayed_units;
    delayed_rules = node.delayed_rules;
    aux = node.aux;
    max_weight = wm_initial_max_weight_next_stratum node;
    parent = node; (* for backtracking (find_next_node) *)
    state = node.state; (* continue with state passed from lower stratum *)
    initial_news_state = news_state;
    literal = UnboundLiteral;
    right_brother_literals = [];
    stretch_ancestor = UnboundStretchAncestor;
    left_brother_refutations = [];
    refutation = NullRefutation;
    depth = node.depth + 1;
    redundant_store = Ic.make_redundant_store () }

let make_child_node node lit lits news_state =
  let aux = node.aux in
  aux.Aux.stat_depth <- max aux.Aux.stat_depth (node.depth + 1);
  { strata = node.post_eval_strata;
    post_eval_strata = [];
    delayed_units = node.delayed_units;
    delayed_rules = node.delayed_rules;
    aux = aux;
    max_weight = wm_child_max_weight node.max_weight;
    parent = node;
    state = State.add_layer node.state;
    initial_news_state = news_state;
    literal = Literal lit;
    right_brother_literals = lits;
    stretch_ancestor = node.stretch_ancestor;
    left_brother_refutations = [];
    refutation = NullRefutation;
    depth = node.depth + 1;
    redundant_store = Ic.make_redundant_store ()
  }

let make_right_brother_node node lit lits =
  { strata = node.strata;
    post_eval_strata = [];
    delayed_units = node.parent.delayed_units;
    delayed_rules = node.parent.delayed_rules;
    aux = node.aux;
    max_weight = node.max_weight;
    parent = node.parent;
    state = State.add_layer node.parent.state;
    initial_news_state = node.initial_news_state;
    literal = Literal lit;
    right_brother_literals = lits;
    stretch_ancestor = node.stretch_ancestor;
    left_brother_refutations =
    node.refutation :: node.left_brother_refutations;
    refutation = NullRefutation;
    depth = node.depth;
    redundant_store = Ic.make_redundant_store ()}

(****************************************************************************)
(* For Debugging *)
(*
let debug_print_node node =
  match node.literal with
    Literal disjunct ->
      PrologPrint.print
	(Fact.get_term (News.Disjunct.get_fact disjunct))
  | UnboundLiteral ->
      print_string "[UnboundLiteral]"
*)
(****************************************************************************)


let gensym_count = ref 0

let gensym string =
  gensym_count := !gensym_count+1;
  string^string_of_int !gensym_count


let rec rtt st aux r =
  match r with
    NullRefutation ->
      Fact.proof_term_for_r_proof st aux Fact.BuiltinProof
  | PrimitiveRefutation proof ->
      Fact.proof_term_for_r_proof st aux proof
  | ComposedRefutation rs ->
      Convert.array_to_compound
	st
	"node"
	[|Convert.string_to_term st (gensym "composed");
	  Convert.string_to_term st "---";
	  Convert.string_to_term st "composed";
	  (Convert.list_of_term_to_term
	     st
	     (List.map (fun r1 -> rtt st aux r1) rs));
	  Convert.list_of_term_to_term st []|]

let refutation_to_term symbol_table r =
  let aux = Fact.make_proof_term_aux () in
  rtt symbol_table aux r



(* 'beheading' of rules: if a unit subsumes a head literal of a rule, then there
is no point in keeping that rule, as anything it derives will be redundant - the rule
will therefore be declared as redundant itself. *)
let remove_subsumed_rules term reader redundant_store =
	Ic.Reader.iter_instance_candidates
		(fun entry ->
			if (Ic.SubtermEntry.is_redundant entry) then ()
			else
				(if (Term.np_subsumes_chk 
						term 
						(Ic.SubtermEntry.get_place_term 
							entry))
					then
						(Ic.SubtermEntry.declare_redundant redundant_store entry)))
		reader
		term
		Term.null_context	


let insert_disjunct node disjunct =
	let fresh_clauses = ref []
	in
  if Cfg.get_wp_semantics_flag node.aux.Aux.cfg && 
    News.Disjunct.is_wp_semantics_true disjunct then
    ()
  else (
    let fact = News.Disjunct.get_fact disjunct in
    let term = Fact.get_term fact
    and predicate = News.Disjunct.get_predicate disjunct in			(*ppp term;pe ".";*)
(*PrologPrint.print term;print_string ";  ";*)
    if Cfg.get_show_derived_literals_flag node.aux.Aux.cfg then
      begin
	print_string "selected_disjunct(";
	PrologPrint.print term;
	print_endline ", _)."
      end;
(*ppp term;pe "<DISJ";*)
    match fact with
    	Fact.DerivedEquationFact _ ->
			(let clause_aux = node.aux.Aux.clause_aux
			in	
			let clause = Ic.make_clause clause_aux term
			in
				if (Ic.is_indexed node.aux.Aux.subterm_table clause) then ()
				else (
					Ic.index_clause_subterms
						node.aux.Aux.subterm_table
						clause;
					Ic.initialize_clause clause node.aux.Aux.news;
					fresh_clauses := clause :: !fresh_clauses;

					let meta_symbol = Term.get_meta_symbol node.aux.Aux.symbol_table 1
					and eq_array =
						if (Ic.is_ground clause) then
							(Ic.get_ground_head_eqs clause)
						else 
							(Ic.get_non_ground_head_eqs clause)					
					in
					let key_term =
						Term.Compound
							(Term.make_compound
								meta_symbol
								eq_array)
					and meta_reader =
						Ic.get_subterm_reader 
							(Ic.find_subterm node.aux.Aux.subterm_table meta_symbol)
					in
						remove_subsumed_rules key_term meta_reader node.redundant_store;
						if (not (Ic.is_oriented clause)) then
							(let flipped_key_term =
								Term.Compound
									(Term.make_compound
										meta_symbol
										[|(Term.binary_flip term)|])
							in
								remove_subsumed_rules flipped_key_term meta_reader node.redundant_store)
					
					))
		|Fact.DerivedEquationFactWithProof _ ->
			(let clause_aux = node.aux.Aux.clause_aux
			and proof = Fact.get_proof fact	
			in	
			let clause = Ic.make_clause clause_aux term
			in
				Ic.set_proof clause proof;
				if (Ic.is_indexed node.aux.Aux.subterm_table clause) then ()
				else (
					Ic.index_clause_subterms
						node.aux.Aux.subterm_table
						clause;
					Ic.initialize_clause clause node.aux.Aux.news;
					fresh_clauses := clause :: !fresh_clauses;

					let meta_symbol = Term.get_meta_symbol node.aux.Aux.symbol_table 1
					and eq_array =
						if (Ic.is_ground clause) then
							(Ic.get_ground_head_eqs clause)
						else 
							(Ic.get_non_ground_head_eqs clause)
					in
					let key_term =
						Term.Compound
							(Term.make_compound
								meta_symbol
								eq_array)
					and meta_reader =
						Ic.get_subterm_reader 
							(Ic.find_subterm node.aux.Aux.subterm_table meta_symbol)
					in
						remove_subsumed_rules key_term meta_reader node.redundant_store;
						if (not (Ic.is_oriented clause)) then
							(let flipped_key_term =
								Term.Compound
									(Term.make_compound
										meta_symbol
										[|(Term.binary_flip term)|])
							in
								remove_subsumed_rules flipped_key_term meta_reader node.redundant_store)
					
					
					))					
		|_ -> 
			(try 
				(if (Cfg.get_ehyper node.aux.Aux.cfg) then
					(Ic.np_subsumption_processing_unit
						(Cfg.get_full_branch_redundancy_flag node.aux.Aux.cfg)
						node.redundant_store
						predicate
						fact;
					Ic.index_unit_subterms_no_top 
						node.aux.Aux.subterm_table 
						term 
						fact);
				Predicate.insert predicate term fact;
				if Predicate.has_delta_predicate predicate then (
		      		Predicate.insert
						(Predicate.get_delta_predicate predicate) 
						term 
						fact);

(*       PrologPrint.print 
	 (refutation_to_term
	    (node.aux.Aux.symbol_table)
	    (PrimitiveRefutation (Fact.get_proof fact)));
pe "";pe "";*)
				
				let meta_symbol = Term.get_meta_symbol node.aux.Aux.symbol_table 1
				in
				let key_term =
					Term.Compound
						(Term.make_compound
							meta_symbol
							[|term|])
				and meta_reader = Ic.get_subterm_reader
									(Ic.find_subterm node.aux.Aux.subterm_table meta_symbol)
				in
					remove_subsumed_rules key_term meta_reader node.redundant_store;
				if node.aux.Aux.use_negative_units then
		      		(let negunits = node.aux.Aux.negative_units in
		      			List.iter 
							(fun complement -> 
			  					NegativeUnits.insert negunits complement)
							(News.Disjunct.get_complements disjunct))) 
			with Ic.Redundant -> ()));		

	!fresh_clauses	
	

	
	
(****************************************************************************)
(* Eval *)


type eval_result =
    Open of int * int * Ic.clause list * Ic.clause list * Ic.clause list * Ic.clause list (* of weight_limit_exceeded * delta_count * new_clauses * new_inc_clauses * new_other_stratum_clauses *)
  | Closed of Fact.proof

let process_delayed_rule_candidates news aux =
	let new_rules = ref []
	in
	List.iter
		(fun clause ->
			if (Ic.is_indexed aux.Aux.subterm_table clause) then ()
			else (
				Ic.initialize_clause clause news;
				if (Ic.clause_np_subsumed_by_clauses aux.Aux.subterm_table clause) then ()
				else 
					(new_rules := clause :: !new_rules)))
		(Ic.NewClauseCandidates.get_delayed_candidates aux.Aux.new_clause_candidates);
	Ic.NewClauseCandidates.clear_delayed_clause_candidates aux.Aux.new_clause_candidates;
	!new_rules
  
let process_new_rule_candidates news aux redundant_store = 
	let delta_count = ref 0
	and new_rules = ref []
	and new_other_stratum_rules = ref []
	and meta_symbol = Term.get_meta_symbol aux.Aux.symbol_table 1
	in	(*ps (Ic.NewClauseCandidates.get_ncc_id aux.Aux.new_clause_candidates);pe " <<<<<<< ";*)
	(*pe "START1";*)
	let meta_reader = 
		Ic.get_subterm_reader
			(Ic.find_subterm aux.Aux.subterm_table meta_symbol)
	in
	List.iter
		(fun clause -> 		(*ppp (Ic.get_source clause);pe ". <<<<<<<<<<<<<<<<<<<<";*)
			if (Ic.is_indexed aux.Aux.subterm_table clause) then ((*pe "INDEXED"*))
			else ( (*pe "GO";*)
				Ic.initialize_clause clause news;
				try 
					(Ic.np_subsumption_processing_clause aux.Aux.subterm_table redundant_store clause;
					delta_count := !delta_count + 1;	(*ppp (Ic.get_source clause);pe ". <";*)
					Ic.index_clause_subterms
						aux.Aux.subterm_table
						clause;		
					new_rules := clause :: !new_rules;
					if (Ic.is_eq_unit clause) then
						(let eq_array =
							if (Ic.is_ground clause) then
								(Ic.get_ground_head_eqs clause)
							else 
								(Ic.get_non_ground_head_eqs clause)
						in 
						let key_term =
							Term.Compound
								(Term.make_compound
									meta_symbol
									eq_array)
						in
							remove_subsumed_rules key_term meta_reader redundant_store;
							if (not (Ic.is_oriented clause)) then
								(let flipped_key_term =
									Term.Compound
										(Term.make_compound
											meta_symbol
											[|eq_array.(0)|])
								in
									remove_subsumed_rules flipped_key_term meta_reader redundant_store)))
				with Ic.Redundant -> ()))
		(Ic.NewClauseCandidates.get_current_stratum_candidates aux.Aux.new_clause_candidates);

	List.iter
		(fun clause -> (*ps (Ic.get_id clause);ppp (Ic.get_source clause);pe ".";*)
			if (Ic.is_indexed aux.Aux.subterm_table clause) then ()
			else (
				Ic.initialize_clause clause news;
				try
					(Ic.np_subsumption_processing_clause aux.Aux.subterm_table redundant_store clause;
					delta_count := !delta_count + 1;
					Ic.index_clause_subterms
						aux.Aux.subterm_table
						clause;		(*ppp (Ic.get_source clause);pe "<<<<<<<<<<!";*)
					new_other_stratum_rules := clause :: !new_other_stratum_rules;
					if (Ic.is_eq_unit clause) then
						(let eq_array =
							if (Ic.is_ground clause) then
								(Ic.get_ground_head_eqs clause)
							else 
								(Ic.get_non_ground_head_eqs clause)
						in
						let key_term =
							Term.Compound
								(Term.make_compound
									meta_symbol
									(Ic.get_ground_head_eqs clause))
						in
							remove_subsumed_rules key_term meta_reader redundant_store;
							if (not (Ic.is_oriented clause)) then
								(let flipped_key_term =
									Term.Compound
										(Term.make_compound
											meta_symbol
											[|eq_array.(0)|])
								in
									remove_subsumed_rules flipped_key_term meta_reader redundant_store)))
				with Ic.Redundant -> ()))
		(Ic.NewClauseCandidates.get_other_stratum_candidates aux.Aux.new_clause_candidates);
	Ic.NewClauseCandidates.clear_new_clause_candidates aux.Aux.new_clause_candidates;	
	(*pe "STOP1";*)
	(!delta_count, !new_rules, !new_other_stratum_rules)

  
	
let make_post_eval_strata 
		node 
		post_eval_clauses 
		post_eval_last_fresh_clauses 
		post_eval_inc_clauses 
		post_eval_other_stratum_clauses =
	(*pe " MAKE_POST_EVAL_STRATA ";*)
	if (List.length post_eval_other_stratum_clauses)>0 then (
		let other_strata_clauses =
			match node.strata with
				_ :: other_strata ->
					let other_strata_clauses = ref []
					in
						List.iter
							(fun stratum ->
								other_strata_clauses :=
									List.append 
										!other_strata_clauses
										(List.append 
											stratum.Stratum.plain_clauses
											stratum.Stratum.inc_clauses))
							other_strata;
						!other_strata_clauses
				|_ -> []
		in
		let clauses = 
				List.append
					(List.append 
						post_eval_clauses 
						post_eval_other_stratum_clauses)
					other_strata_clauses
		in 	
	    let strata =
    		List.map 
      			(fun clauses1 -> Stratum.make_stratum node.aux clauses1) 
      			(IcSet.stratify clauses) 
      	in
      		strata)
	else (
		if (List.length post_eval_clauses)>0 then (
			let new_inc_clauses = 
				IcSet.sort_clauses
					(List.concat 
						(List.map 
							(fun fresh_clause -> 
								Ic.clause_delta_clauses_unchecked fresh_clause) 
							post_eval_last_fresh_clauses))								
			in				
			match node.strata with
				stratum :: strata_tail ->
					{Stratum.plain_clauses = IcSet.sort_clauses 
												(List.concat 
													[stratum.Stratum.plain_clauses; (*experimental*)
													post_eval_clauses	(*;
													post_eval_last_fresh_clauses*)]);
					Stratum.inc_clauses = IcSet.sort_clauses 
											(List.concat 
												[	(*stratum.Stratum.inc_clauses; *)	(*experimental*)
												post_eval_inc_clauses;
												new_inc_clauses]);
					Stratum.delta_predicates = stratum.Stratum.delta_predicates;
					Stratum.needs_compilation = true;
					Stratum.aux = stratum.Stratum.aux}
					:: strata_tail
				|[] -> (raise (Err.Error ("Empty strata in Tableau.make_post_eval_strata.")))
		)
		else (	(*pe "SAME!";*)node.strata)
	)  	
	
	

	
let process_news subterm_table symbol_table redundant_store news =
	let new_unit_delta_count = ref 0 
	and new_pred_facts = News.process_news news
	and meta_symbol = Term.get_meta_symbol symbol_table 1
	in
	let meta_reader = Ic.get_subterm_reader (Ic.find_subterm subterm_table meta_symbol)
	in
		if (News.get_cfg_ehyper news) then (
			List.iter
				(fun (predicate, fact) ->
					try
						(let term = (Fact.get_term fact) in
							(Ic.np_subsumption_processing_unit 
								(News.get_cfg_full_branch_redundancy news) 
								redundant_store 
								predicate 
								fact);
							Predicate.insert predicate term fact;
							
							let key_term =
								Term.Compound
									(Term.make_compound
										meta_symbol
										[|term|])
							in
								remove_subsumed_rules key_term meta_reader redundant_store;

							if (News.get_cfg_show_derived_literals news) then
								begin
									print_string "derived_fact(";
									PrologPrint.print term;
									print_endline ")."
								end;		(*ppp term;pe "!!!";*)
							if Predicate.has_delta_predicate predicate then (
								let delta_predicate = Predicate.get_delta_predicate predicate in
									Predicate.insert delta_predicate term fact)
							else ();
							Ic.index_unit_subterms_no_top subterm_table term fact;
							new_unit_delta_count := 1 + !new_unit_delta_count)
					with Ic.Redundant -> ())
				new_pred_facts;
				!new_unit_delta_count
			)
		else 
			(List.iter
				(fun (predicate, fact) -> 
					let term = (Fact.get_term fact) 
					in
						Predicate.insert predicate term fact;
						if (News.get_cfg_show_derived_literals news) then
							begin
								print_string "derived_fact(";
								PrologPrint.print term;
								print_endline ")."
							end;
							
						let key_term =
							Term.Compound
								(Term.make_compound
									meta_symbol
									[|term|])
						in
							remove_subsumed_rules key_term meta_reader redundant_store;								
			
						if Predicate.has_delta_predicate predicate then (
							let delta_predicate = Predicate.get_delta_predicate predicate in
								Predicate.insert delta_predicate term fact)
						else ();
						new_unit_delta_count := 1 + !new_unit_delta_count)
				new_pred_facts;
			!new_unit_delta_count)
			
let minmax x y =
	if x>y then (y, x)
	else (x, y)		

let rec eval_1 
    		aux 
    		news 
    		delta_predicates
    		weight_limit_exceeded 
    		delta_clauses 					(* empty in first round - "inc_clauses" *)
    		fresh_clauses 				(* "plain_clauses" in first round *)
    		post_eval_clauses 
    		post_eval_other_stratum_clauses 
    		stratum_clauses 
    		redundant_store =	
(* delta_clauses : eq-units and deltified clauses that are applied in each round, fresh_clauses will be added to these after their first undeltified run *)
(* fresh_clauses : clauses generated last round, never applied so far *)
(* post_eval_clauses : clauses generated so far in this node (possibly over multiple rounds), including fresh_clauses *)
(* post_eval_other_stratum_clauses : other stratum clauses generated so far in this node (possibly over multiple rounds) *)    	
(* stratum_clauses : all non-delta-clauses of the current stratum, used to check if a new clause is element of the
current stratum *) 
  Aux.inc_stat_eval_rounds aux;	
(*List.iter (fun x -> PrologPrint.print (Ic.get_source x);print_endline ".") post_eval_clauses;*)
  aux.Aux.t_round_start <- (Sys.time ());	(*print_string " --> ";print_int(List.length fresh_clauses);print_string "; "; *)
  	let next_stratum_clauses = (List.append stratum_clauses fresh_clauses) (* "next" as in "for next round"... *)
	and ehyper = Cfg.get_ehyper aux.Aux.cfg
	in	
(*	List.iter (fun cl -> PrologPrint.print (Ic.get_source cl);print_endline " <STRATUM";) stratum_clauses;
	List.iter (fun cl -> PrologPrint.print (Ic.get_source cl);print_endline " <FRESH";) fresh_clauses;*)
  let round_result = (	
  	try										(*pe "START";*)
  	(*print_int (Ic.NewClauseCandidates.get_max_literal_weight aux.Aux.new_clause_candidates);pe "...";*)
(*    		List.iter (fun cl -> PrologPrint.print (Ic.get_source cl);print_endline ".";) fresh_clauses;*)
(*List.iter (fun c -> PrologPrint.print (Ic.get_source c);print_endline ".") clauses1;*)
(*List.iter (fun c -> PrologPrint.print (Ic.get_source c);print_endline ".") post_eval_clauses;*)
(*pe "";pi (List.length post_eval_clauses);pe "";*)
      List.iter
		(fun c ->
			Ic.set_is_in_current_stratum true c)
		next_stratum_clauses;
(*      List.iter
		(fun c ->
			Ic.set_is_in_current_stratum true c)
		delta_clauses;
	  List.iter
		(fun c ->
			Ic.set_is_in_current_stratum true c)
		fresh_clauses;*)
      List.iter 
      	(fun clause -> 		
      	      Ic.eval_clause 
	      		clause 
	      		true	(* delta *)
	      		aux.Aux.subterm_table
	      		next_stratum_clauses
	      		aux.Aux.new_clause_candidates
	      		redundant_store) 
      	delta_clauses;
      List.iter 
      	(fun clause -> 		
      		Ic.eval_clause 
	      		clause
	      		false 	(* no delta *)
	      		aux.Aux.subterm_table
	      		next_stratum_clauses
	      		aux.Aux.new_clause_candidates
	      		redundant_store) 
      	fresh_clauses;								(*pe "STOP";*)
	  List.iter
		(fun c ->
			Ic.set_is_in_current_stratum false c)
		next_stratum_clauses;
      
     
      let weight_limit_exceeded_1 =
			if weight_limit_exceeded = 0 then
				(let minwle, maxwle = minmax
	  				(News.weight_limit_exceeded news) 
	  				(Ic.NewClauseCandidates.weight_limit_exceeded aux.Aux.new_clause_candidates)
	  			in
	  				if minwle = 0 then maxwle
	  				else (minwle))
(*	  			(min
	  				(News.weight_limit_exceeded news) 
	  				(Ic.NewClauseCandidates.weight_limit_exceeded aux.Aux.new_clause_candidates))*)
			else
	  			weight_limit_exceeded 
	  in
      	if ehyper then
	      	(Predicate.PredicateTable.iter
	      		(fun symbol pred ->
	      			Predicate.clear_delta_predicate pred)
	      		aux.Aux.predicate_table;

	      	Ic.SubtermTable.iter
	      		(fun symbol subterm ->
	      			Ic.clear_delta_subterm subterm)
	      		aux.Aux.subterm_table)
	    else
	    	(List.iter Predicate.clear_delta_predicate delta_predicates);

	    	
	    	
      	let unit_delta_count = process_news aux.Aux.subterm_table aux.Aux.symbol_table redundant_store news
      	and clause_delta_count, round_new_clauses, round_new_other_stratum_clauses = process_new_rule_candidates news aux redundant_store
      	in 

      	let delta_count = unit_delta_count + clause_delta_count
      	in	
       		Open (weight_limit_exceeded_1, delta_count, round_new_clauses, fresh_clauses, delta_clauses, round_new_other_stratum_clauses)
    with Ic.BranchClosed proof ->
      List.iter
		(fun c ->
			Ic.set_is_in_current_stratum false c)
		delta_clauses;
	  List.iter
		(fun c ->
			Ic.set_is_in_current_stratum false c)
		fresh_clauses;    
      (			
      if ehyper then
	  	(Predicate.PredicateTable.iter
      		(fun symbol pred ->
      			Predicate.clear_delta_predicate pred)
      		aux.Aux.predicate_table;
	      
	      Ic.SubtermTable.iter
	      		(fun symbol subterm ->
	      			Ic.clear_delta_subterm subterm)
	      		aux.Aux.subterm_table)
	  else
	  	(List.iter Predicate.clear_delta_predicate delta_predicates);
      		
      News.clear_news news;
      Ic.NewClauseCandidates.clear_new_clause_candidates aux.Aux.new_clause_candidates;
      Closed proof) 
	)      
      in
      
	match round_result with
		Open (wle, delta_count, next_fresh_clauses, _, delta_clauses, round_new_other_stratum_clauses) ->
		
			Msg.msg4 2 "Round: %9d, Delta: %9d, Delayed: %9d, Time: %7.2fs"
				aux.Aux.stat_eval_rounds
				delta_count
				(News.delayed_clauses_size news)
				(Sys.time () -. aux.Aux.t_round_start);
				
      		if delta_count > 0 then (	
				let new_inc_clauses = 
					IcSet.sort_clauses
						(List.concat 
							(List.map 
								(fun fresh_clause -> 
									Ic.clause_delta_clauses_unchecked fresh_clause) 
								fresh_clauses))
				in 		
				List.iter
					(fun c ->
						Ic.initialize_clause c news)
					new_inc_clauses;
				let next_delta_clauses = List.append 
											delta_clauses 
											new_inc_clauses
				and next_post_eval_clauses = List.append 
												next_fresh_clauses 
												post_eval_clauses
				and next_post_eval_other_stratum_clauses = List.append 
																round_new_other_stratum_clauses 
																post_eval_other_stratum_clauses
				in 
					eval_1 
				  		aux 
				  		news 
				  		delta_predicates 
				  		wle 
				  		next_delta_clauses 
				  		next_fresh_clauses 
				  		next_post_eval_clauses 
				  		next_post_eval_other_stratum_clauses
				  		next_stratum_clauses
				  		redundant_store)
      else (	
      Open (wle, delta_count, post_eval_clauses, fresh_clauses, delta_clauses, post_eval_other_stratum_clauses))
  | Closed proof ->
      Closed proof


      
      
let neg_proof fact neg_fact =
  match Fact.get_proof fact with
    Fact.BuiltinProof ->
      Fact.BuiltinProof
  | Fact.DisjProof _ -> 
      Fact.NegProof
	(Term.Unbound,
	 [|fact|],
	 neg_fact)
  | _ ->
      raise (Failure "Unexpected proof type for tableau node literal.")
	
      
let print_extent node =		
	print_endline "================================================";
		(Predicate.PredicateTable.iter
	  (fun symbol predicate ->
	    try 
	      Access.Reader.iter_all
		(fun entry -> 
		  	if (Fact.is_redundant entry) then ()
		  	else (
				PrologPrint.print (Fact.get_term entry);
		  		print_endline "."))
		(Predicate.default_reader predicate)
	    with Not_found ->
	      ())
	  (node.aux.Aux.predicate_table));
	  print_endline "================================================"
      
      
      
      
      
let eval node =	
  print_debug "EVAL";
  let aux = node.aux in
  let news = aux.Aux.news in
  State.install_state node.state;
  News.install_state node.initial_news_state;
  News.set_max_literal_weight news node.max_weight;	
  News.reset_weight_limit_exceeded news;
  Ic.NewClauseCandidates.set_max_literal_weight aux.Aux.new_clause_candidates node.max_weight;
  Ic.NewClauseCandidates.reset_weight_limit_exceeded aux.Aux.new_clause_candidates;
  
  
  try
    if aux.Aux.use_negative_units then 
      (match node.literal with
	Literal lit ->
	  let fact = News.Disjunct.get_fact lit in 
	  let neg_fact = 
	    NegativeUnits.lookup node.aux.Aux.negative_units fact in 
	  Closed
	    (if aux.Aux.use_proof_terms then
	      begin
		print_debug "EVAL-CLOSED-BY-COMPLEMENT";
		neg_proof fact neg_fact
	      end
	    else
	      Fact.BuiltinProof)
      | UnboundLiteral ->
	  (raise Not_found))
    else
      (raise Not_found)
  with Not_found -> 
    (
    match node.strata with
      stratum :: _ ->
	(
	
	  match node.literal with
	    Literal literal ->
	      (* DISJUNCTION NODE *)
	      let fresh_clauses = insert_disjunct node literal
	      in 	
	      (match node.strata with
	      	stratum :: _ ->
			  Stratum.ensure_compiled stratum; 
		      eval_1
				aux
				news
				stratum.Stratum.delta_predicates
				0
				stratum.Stratum.inc_clauses
				fresh_clauses
				fresh_clauses	(* was [] *)
				[]								
				stratum.Stratum.plain_clauses
				node.redundant_store
			|_ -> 
				(raise (Err.Error ("Stratum lost in Tableau.eval."))))
	  | UnboundLiteral -> 
	      (* FIRST NODE OR FIRST NODE OF A HIGHER STRATUM *)
	      Stratum.ensure_compiled stratum;
	      eval_1
			aux
			news
			stratum.Stratum.delta_predicates
			0
		(*	stratum.Stratum.inc_clauses*)[] 
			stratum.Stratum.plain_clauses
			[]
			[]
			stratum.Stratum.plain_clauses
			node.redundant_store
	)
    | [] ->
	Msg.msg 2 "No stratum.";
	Open (0, 0, [], [], [], []))
	  
(****************************************************************************)
(* Find Model Loop *)
(****************************************************************************)

type result =
    Model of node
  | NoModel of Aux.aux

type disjunction =
    Disjunction of News.delayed_clause
  | UnboundDisjunction

type next_node =
    NextNode of node
  | NoNextNode of node (* returns the top node *)

let info_find_model node =
  let aux = node.aux in
  let t = Sys.time () in
  Msg.msg4 2
    "Depth: %d, Weight: %d, Stratum: %d, Model:%7.2fs" 
    (* (t -. aux.Aux.t_start) *)
    node.depth
    node.max_weight
    (aux.Aux.n_strata - (List.length node.strata))
    (t -. aux.Aux.t_model_start)

let install_top_state aux =
  State.install_state aux.Aux.input_state

(* ### Use level (depth) of Closed as parameter to level_cut ! *)
(* but a refutation propagated upwards may use a different depth... *)

(** Raises Not_found if proof is Fact.UnknownProof. *)
let rec proof_iter_facts f proof =
  match proof with
    Fact.BuiltinProof | Fact.InputFactProof | Fact.InputClauseProof ->
      ()
  | Fact.Proof (_, facts) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts
  | Fact.DisjProof (_, _, facts, neg_facts) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts;
      List.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	neg_facts
  | Fact.NegProof (_, facts, neg_fact) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts;
      f neg_fact;
      proof_iter_facts f (Fact.get_proof neg_fact)
  | Fact.UnknownProof ->
      (raise Not_found)
  | Fact.UnitSupRightProof (_, _, into_fact, eq_fact) ->
  		f into_fact;
  		f eq_fact;
  		proof_iter_facts f (Fact.get_proof into_fact);    
  		proof_iter_facts f (Fact.get_proof eq_fact)
  | Fact.NegUnitSupRightProof (_, _, into_fact, eq_fact, neg_fact) ->
  		f neg_fact;
  		f into_fact;
  		f eq_fact;
  		proof_iter_facts f (Fact.get_proof neg_fact);
  		proof_iter_facts f (Fact.get_proof into_fact);    
  		proof_iter_facts f (Fact.get_proof eq_fact)  		
  | Fact.SupLeftProof (_, _, into_fact, eq_fact) ->
  		f into_fact;
  		f eq_fact;
  		proof_iter_facts f (Fact.get_proof into_fact);    
  		proof_iter_facts f (Fact.get_proof eq_fact)
  | Fact.EqRefProof (_, fact) ->
  		f fact;
  		proof_iter_facts f (Fact.get_proof fact)  
  | Fact.HyperPureProof (_, facts) ->
      Array.iter 
	(fun fact -> f fact; proof_iter_facts f (Fact.get_proof fact))
	facts		
  		
let rec refutation_iter_facts f refutation =
  match refutation with
    NullRefutation ->
      ()
  | PrimitiveRefutation proof ->
      (proof_iter_facts f proof)
  | ComposedRefutation refutations ->
      List.iter (fun r -> refutation_iter_facts f r) refutations

(* must be FALSE for level cut execution *)
let used_in_refutation facts refutation depth =
  (* Depth could be used to restrict the set of facts to iterate over. *)
  try				(*List.iter (fun f -> ppp (Fact.get_term f);pe "!") facts;*)
    refutation_iter_facts
      (fun fact -> (*ppp (Fact.get_term fact);pe " <ARG";*)
	if List.mem fact facts then raise Found) 
        (* ### Fact-identity: = (and not ==) *)
      refutation;
    false

    with Found -> (true)
    | Not_found -> (true)

let level_cut node depth =
  if  depth >= 0 && node.aux.Aux.use_level_cut then
    match node.literal with
      UnboundLiteral ->
	(* Top node or top node of stratum. *)
	false
    | Literal lit ->
	(* Disjunction node. *)	(*ppp (Fact.get_term (News.Disjunct.get_fact lit));pe " <DDD";*)
	let facts = 
	  (News.Disjunct.get_fact lit) ::
	  (News.Disjunct.get_complements lit) in
	not (used_in_refutation facts node.refutation depth) 
  else
    false

let rec find_next_node node depth = 
(*	Ic.undeclare_redundancies node.redundant_store;*) (*This is bad when dealing with disjunctions,
as "node" is then the parent of the node we want to undo, resulting in the wrong redundancies being
undeclared. *)
  if level_cut node depth then
    begin
      Aux.inc_stat_level_cuts node.aux;
      node.parent.refutation <- node.refutation;
      Ic.undeclare_redundancies node.redundant_store;
      find_next_node
	node.parent 
	(if depth >= 0 then (depth + 1) else depth)
    end
  else (
    match node.right_brother_literals with
      [] -> 
	begin
	  if node.parent == node then
	    (* Top node. *)
	    begin	
		  Ic.undeclare_redundancies node.redundant_store;    
	      install_top_state node.aux;
	      NoNextNode node
	    end
	  else
	    match node.literal with
		    Literal _ -> 
				(* Disjunction branch node. *)
				print_debug "NN_PARENT";
				node.parent.refutation <-
				  ComposedRefutation 
				    (node.refutation :: node.left_brother_refutations);
				Ic.undeclare_redundancies node.redundant_store;    
				find_next_node
				  node.parent 
				  (if depth >= 0 then (depth + 1) else depth)
		    |UnboundLiteral ->
				(* Top node of stratum, the parent is on a lower stratum *)
				node.parent.refutation <-
				  ComposedRefutation 
				    (node.refutation :: node.left_brother_refutations);
		(*		Ic.undeclare_redundancies node.redundant_store;    *)
				find_next_node node.parent (-1)
	end
    | lit :: lits ->
	print_debug "NN_BROTHER";
	Ic.undeclare_redundancies node.redundant_store;
	NextNode (make_right_brother_node node lit lits))
	

let print_disjuncts disjuncts =
	List.iter
		(fun d ->
			PrologPrint.print(Fact.get_term (News.Disjunct.get_fact d));print_string "; ")
		(News.get_disjuncts disjuncts);print_endline "."

		
		
(*********************************************************************)
(* picking a disjunction 	*)
(*********************************************************************)		

		
let minimize_disjunct aux equals var_table weighing_method disjunct =
	let pure_term = News.Disjunct.get_term disjunct
	and impure_term = News.Disjunct.get_impure_term disjunct
	and impure_bindings = News.Disjunct.get_impure_bindings disjunct
	and pure_bindings = News.Disjunct.get_pure_bindings disjunct
	in
		let minimized_impure_bindings =
				(Array.map
					(fun term ->
						Ic.minimize_term
							aux.Aux.subterm_table
							equals
							term)
					impure_bindings)
		in
		let minimized_context =
			(let minimized_bindings = Array.make Term.max_vars Term.Unbound
			in
				Array.iteri
					(fun i term ->
						if (minimized_bindings.(i) == Term.Unbound) then
							(minimized_bindings.(i) <- term))
					pure_bindings;
				Array.iteri
					(fun i term ->
						if (minimized_bindings.(i) == Term.Unbound) then
							(minimized_bindings.(i) <- term))
					minimized_impure_bindings;
				Term.make_bound_context minimized_bindings 0)
		in
		let purified_term = Term.apply impure_term minimized_context
		in
			Term.renumber_variables_term var_table purified_term;
			let minimized_weight = Weight.term_weight weighing_method pure_term
			and minimized_fact =
				(if (Convert.term_functor_symbol purified_term) == equals then
					(let l, r = Convert.term_pair purified_term
					in
					let pure_eq = Equality.make_checked_equation purified_term l r
					in
						Fact.make_derived_equation_fact pure_eq true)
				else
					(Fact.make_fact_like (News.Disjunct.get_fact disjunct) purified_term))
			in
				((News.Disjunct.make_disjunct
					(News.Disjunct.get_predicate disjunct)
					minimized_fact
					(News.Disjunct.get_complements disjunct)
					impure_term
					impure_bindings
					pure_bindings),
				minimized_weight)
		
				
let rec factorize_identical_minimized_literals lits =
	match lits with
		l1 :: l2 :: lits1 ->
			(match l1, l2 with
	  			((News.Disjunct.Disjunct (_,f1,_,_,_,_)), _), ((News.Disjunct.Disjunct (_,f2,_,_,_,_)), _) ->
	  				if Term.term_ident (Fact.get_term f1) (Fact.get_term f2) then
	  					(factorize_identical_minimized_literals (l2 :: lits1))
	  				else
						(l1 :: factorize_identical_minimized_literals (l2 :: lits1))
				|_ -> (l1 :: factorize_identical_minimized_literals (l2 :: lits1)))
		|_ -> lits				
				
let minimize_disjunction aux equals redundant_store disjunction =
	let var_table = Array.make Term.max_vars (-1)
	and weighing_method =
        (match Cfg.get_weighing_method_param aux.Aux.cfg with
		      0 -> Weight.SymbolCount
		    | 1 -> Weight.Depth
		    | _ -> raise
			  (Err.Error "Bad value for term_weighing_method parameter."))
	in
		let lits1 =
			List.map
				(fun disjunct ->
					minimize_disjunct aux equals var_table weighing_method disjunct)
				(News.get_disjuncts disjunction)
		in
			let lits2 =
				(List.stable_sort
					(fun (d1,_) (d2,_) ->
						LrpOrder.ground_like_compare_lrpo_greater 
							(News.Disjunct.get_term d1)
							(News.Disjunct.get_term d2))
					lits1)
			in
			let lits3 =
				factorize_identical_minimized_literals lits2		
			in		
			let lits4 =
					(List.map
						(fun (disjunct, _) -> disjunct)
						lits3)
			and minimized_weight =
				List.fold_left (fun w1 (_,w2) -> w1 + w2) 0 lits3
			in
			
			(*List.iter (fun d -> ppp (News.Disjunct.get_term d);pe ".") lits4;*)
			
				News.DelayedDisjunction 
						(lits4, minimized_weight)

			
let rec pick_disjunction news aux redundant_store =
	
	if not (News.get_cfg_ehyper news) then
		(News.pick_disjunction news)
	else 	
		(pick_pure_disjunction news aux redundant_store)

      
and pick_pure_disjunction news aux redundant_store =
	match (News.get_delayed_clauses news) with
		[] -> (raise Not_found)
		|pure_disjunctions ->
			let equals = Term.get_equality_symbol aux.Aux.symbol_table
			in
				try
					(pick_pure_disjunction_1 news aux equals redundant_store pure_disjunctions)
				with Not_found -> (raise Not_found)
			

	
and pick_pure_disjunction_1 news aux equals redundant_store disjunctions =
	match disjunctions with
		[] -> (raise Not_found)
		|disjunction :: rest_disjunctions -> 
 			if ((disjunction_is_reducible aux disjunction)
 				||(disjunction_contains_indexed_disjunct disjunction aux equals)) then
				(pick_pure_disjunction_1 news aux equals redundant_store rest_disjunctions)
			else ((*pe "GO";*)
				News.set_delayed_clauses news rest_disjunctions;
				disjunction)

and disjunction_is_reducible aux disjunction =
	let krh_dom_sym =
		Term.get_krh_dom_symbol aux.Aux.symbol_table
	in
	let krh_dom_pred = Predicate.find_predicate 
							aux.Aux.predicate_table
							krh_dom_sym
	and purifiers =
		match (News.get_disjuncts disjunction) with
			[] -> (raise (Err.Error ("Bad disjunction in Tableau.disjunction_is_reducible.")))
			| disjunct :: _ ->
				(News.Disjunct.get_impure_bindings disjunct)					
	in
		try
			(Array.iter
				(fun t ->
					if (t==Term.Unbound) then ()
					else
						(let domain_term =
							Term.Compound
								(Term.make_compound
									krh_dom_sym
									[|t|])
						in
							if (Ic.red_indexed_in_extent krh_dom_pred domain_term) then
								(raise Found)))
				purifiers;
			false)
		with Found -> true
				
and disjunction_contains_indexed_disjunct disjunction aux equals =
	List.exists
		(fun disjunct ->
			is_indexed_disjunct disjunct aux equals)
		(News.get_disjuncts disjunction)					

		
and disjunction_is_fully_indexed disjunction aux equals =
	List.for_all
		(fun disjunct ->
			is_indexed_disjunct disjunct aux equals)
		(News.get_disjuncts disjunction)			
						
and is_indexed_disjunct disjunct aux equals =
	let term = Fact.get_term (News.Disjunct.get_fact disjunct)
	in	(*ppp term;pe ".";*)
		try
			(if (Convert.term_functor_symbol term) == equals then
				(let eq_subterm = Ic.find_subterm aux.Aux.subterm_table equals
				and flip_term = Term.binary_flip term
				in
					Ic.Reader.iter_instance_candidates 
						(fun entry -> 
							match entry with
								Ic.SubtermEntry.Clause (candidate, _, _, _) ->
									if (Term.term_ident 
											(Ic.get_source candidate) 
											term)
									|| (Term.term_ident 
											(Ic.get_source candidate) 
											flip_term) then
										(raise Found)
									else ()
								|_ -> ())
						(Ic.get_subterm_reader eq_subterm)
						term
						Term.null_context;
					false)
			else (
				let pred = News.Disjunct.get_predicate disjunct
				in
					Access.Reader.iter_instance_candidates
						(fun fact ->
							if (Term.term_ident (Fact.get_term fact) term) then
								(raise Found)
							else ())
						(Predicate.get_all_bound_reader pred)
						term
						Term.null_context;
					false))
		with Found -> true	
	
let rec undeclare_stretched_redundancies node snode =
	Ic.undeclare_redundancies node.redundant_store;
	if (node == snode) then ()
	else
		(undeclare_stretched_redundancies node.parent snode)	

(*	This one looks good, but keep in mind that disjunctions are declared redundant
when first collected into news. Not undeclaring them in the node of their collection
will lead to them not being used at all. 				
let rec undeclare_stretched_redundancies node snode =
	if (node == snode) then ()
	else
		(Ic.undeclare_redundancies node.redundant_store;
		undeclare_stretched_redundancies node.parent snode)					
*)
	
let redundancy_check_delayed_rule clause subterm_table =
	let minimized_clause =
		(Ic.minimize_clause clause subterm_table)
	in
		(if (Ic.clause_np_subsumed_by_clauses subterm_table minimized_clause) then 
			(true)
		else
			(false))

		
					
let redundancy_check_delayed_unit predicate fact subterm_table equals =
	let term = Fact.get_term fact
	in	(*ppp term;pe "!";*)
	let minimized_term =
		(if (Termutils.is_ground term) then
			(Ic.minimize_literal_term equals subterm_table term)		
		else
			(term))
	in	
		if (News.unit_is_subsumed_by_extent predicate minimized_term) then 
			(true)
		else
			(false)
		
			
(*********************************************************************)
						
let rec find_model node = 
  info_find_model node;
(*  print_extent node; *)
  
  let aux = node.aux in
  let news = aux.Aux.news in 
  match eval node with
    Open (weight_limit_exceeded, _, post_eval_clauses, post_eval_last_fresh_clauses, post_eval_inc_clauses, post_eval_other_stratum_clauses) ->
    			
  (*  pe "POST_EVAL_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		ppp (Ic.get_source c);pe ".")
			  	post_eval_clauses;
			  pe "POST_EVAL_INC_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		ppp (Ic.get_source c);pe ".")
			  	post_eval_inc_clauses;
			  pe "POST_EVAL_OTHER_STRATUM_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		ppp (Ic.get_source c);pe ".")
			  	post_eval_other_stratum_clauses;
			  pe "POST_EVAL_LAST_FRESH_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		ppp (Ic.get_source c);pe ".")
			  	post_eval_last_fresh_clauses;			  	
			  pe "THATS ALL!";*)
(*pe "POST_EVAL_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		if ((Ic.get_id c)="clause_109") then (pe "Here.";))
			  	post_eval_clauses;
			  pe "POST_EVAL_INC_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		if ((Ic.get_id c)="clause_109") then (pe "Here.";))
			  	post_eval_inc_clauses;
			  pe "POST_EVAL_OTHER_STRATUM_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		if ((Ic.get_id c)="clause_109") then (pe "Here.";))
			  	post_eval_other_stratum_clauses;
			  pe "POST_EVAL_LAST_FRESH_CLAUSES:";
			  List.iter
			  	(fun c ->
			  		if ((Ic.get_id c)="clause_109") then (pe "Here.";))
			  	post_eval_last_fresh_clauses;			  	
			  pe "THATS ALL!";
	*)			  
			  
      node.post_eval_strata <- make_post_eval_strata 
      								node 
      								post_eval_clauses 
      								post_eval_last_fresh_clauses
      								post_eval_inc_clauses 
      								post_eval_other_stratum_clauses;

				
  	List.iter
  		(fun du ->
  			node.delayed_units <- du :: node.delayed_units)
  			(News.process_delayed_news news);
  	List.iter
  		(fun dr ->
  			node.delayed_rules <- dr :: node.delayed_rules)
  			(process_delayed_rule_candidates news aux);
      
    	(* TABLEAU BELOW NODE IS OPEN *)
      (match node.stretch_ancestor with
			UnboundStretchAncestor ->
				  if weight_limit_exceeded > 0 then
				    if (Cfg.get_max_weight_increment_param aux.Aux.cfg) >= 0 then
				      (* SET STRETCH ANCESTOR *)
				      begin
					Msg.msg 3 "Weight limit exceeded.";
					print_debug "SET STRETCH ANCESTOR";
					node.stretch_ancestor <-
					  StretchAncestor (node, weight_limit_exceeded)
				      end
				    else
				      Msg.msg 3 "Weight limit exceeded: model may be incomplete.";
(*		      | StretchAncestor _ ->
			  ());*)
			  | StretchAncestor (snode, prev_weight_limit_exceeded) ->
				  if (weight_limit_exceeded > 0) && (weight_limit_exceeded < prev_weight_limit_exceeded) then
				    if (Cfg.get_max_weight_increment_param aux.Aux.cfg) >= 0 then
				      (* LOWER STRETCH ANCESTOR'S WEIGHT *)
				      begin
					Msg.msg 3 "Weight limit exceeded with lower difference.";
					print_debug "MODIFY STRETCH ANCESTOR";
					node.stretch_ancestor <-
					  StretchAncestor (snode, weight_limit_exceeded)
				      end);			  
      (* TRY TO EXTEND HYPERTABLEAU *)
      (match 
			(try
			  Disjunction (pick_disjunction news aux node.redundant_store)
			with Not_found ->
			  (UnboundDisjunction)) 
	  	with
	Disjunction disjunction ->	(*print_disjuncts disjunction;*)
	  print_debug("EXTEND");
	  Aux.inc_stat_disjunction_nodes aux;
	  print_debug "CHILD";
	  (* PICKED A DISJUNCTION FROM SOS *)
	  (match (News.get_disjuncts disjunction) with
	    lit :: lits ->	
	      let child =
		make_child_node node lit lits (News.obtain_state news)
	      in	
	      if Cfg.get_show_derived_literals_flag aux.Aux.cfg then
		print_endline "attached_disjunction(_).";
	      find_model child
	  | [] ->	
	      raise (Failure "Empty disjunction."))
      | UnboundDisjunction ->
	  (* SOS EMPTY *)
	  match node.stretch_ancestor with
	    UnboundStretchAncestor ->
			      print_debug("UNBOUND SA");
		
			      (* MODEL FOR THIS STRATUM FOUND *)
			      if has_next_stratum node then
				(* NEXT STRATUM *)
				begin
				  print_debug "NEXT STRATUM";
				  find_model 
				    (make_next_stratum_node 
				       node 
				       (News.obtain_state news)) (* should be empty here *)
				end
			      else
				(* MODEL FOR ALL STRATA FOUND *)
				begin
				  Aux.inc_stat_models aux;
				  Model node
				end
	  | StretchAncestor (snode, weight_increment) ->	 
	  		(* if redundant forall news.delayed_units then model,
	  			else continue down there *)
	  			
 				let equals = Term.get_equality_symbol aux.Aux.symbol_table
 				in
	  			try
	  				(	(*pi ((List.length node.delayed_units)+(List.length node.delayed_rules));pe "!";*)
	  				if (not (Cfg.get_ehyper node.aux.Aux.cfg)) then raise Failed;
	  				List.iter
	  					(fun (pred, fact) -> (*ppp (Fact.get_term fact);pe "!";*)
	  						if (redundancy_check_delayed_unit 
	  								pred 
	  								fact 
	  								aux.Aux.subterm_table 
	  								equals)
	  							then ()
	  						else (raise Failed))
	  					node.delayed_units;
	  				List.iter
	  					(fun clause -> (*ppp (Ic.get_source clause);pe "!";*)
	  						if (redundancy_check_delayed_rule 
	  								clause
	  								aux.Aux.subterm_table)
	  							then ()
	  						else (raise Failed))
	  					node.delayed_rules;
	  				if has_next_stratum node then
						(print_debug "NEXT STRATUM";
						 find_model 
						    (make_next_stratum_node 
						       node 
						       (News.obtain_state news)) (* should be empty here *)
						)
			  	    else
						(
						 Aux.inc_stat_models aux;
				 		 Model node))
	  			with Failed -> 
	      (* REDO FROM ANCESTOR WITH INCREASED MAX-WEIGHT *)
	    (*  Ic.undeclare_redundancies node.redundant_store; *)(*YYYY*)
		(*  Ic.undeclare_redundancies snode.redundant_store;*)
		  (undeclare_stretched_redundancies node snode;
	      print_debug("STRETCH");
	      wm_increase_max_weight_of_stretch_root
			aux
			snode
			weight_increment;
	      Aux.inc_stat_stretchings aux;
	      snode.stretch_ancestor <- UnboundStretchAncestor;
	      snode.delayed_units <- [];
		  snode.delayed_rules <- [];
	      match snode.literal with
			UnboundLiteral ->	(*pe "HERE";*)
			  (* FIRST NODE OR FIRST NODE OF A HIGHER STRATUM *)
			  (* RE-USE THE NODE'S STATE AS LEMMAS *)
			  print_debug("STRETCH FIRST");
			  snode.strata <- snode.post_eval_strata;
	(*	 	    pe "1:";print_current_compiled_rules snode.aux.Aux.subterm_table;
		      install_top_state snode.aux;
		      pe "2:";print_current_compiled_rules snode.aux.Aux.subterm_table;*)
			  find_model snode
		    |Literal lit -> 
			  print_debug("STRETCH DISJ");
			(*  Ic.undeclare_redundancies node.redundant_store;*)
			  snode.state <- State.add_layer snode.parent.state;
			  find_model snode))
  | Closed proof ->
      Aux.inc_stat_closed_branches aux;
      if aux.Aux.use_proof_terms then
	begin
	  aux.Aux.proofs <- proof :: aux.Aux.proofs;
	  node.refutation <- PrimitiveRefutation proof
	end;
	  Ic.undeclare_redundancies node.redundant_store;
      match find_next_node node 0 with
	NextNode next_node ->
	  find_model next_node
      | NoNextNode top_node ->
	  aux.Aux.refutation <- top_node.refutation;
	  NoModel aux


(****************************************************************************)
(* Invocation *)
(****************************************************************************)

let result_aux result =
  match result with
    Model node -> node.aux 
  | NoModel aux -> aux

let is_open result =
  match result with
    Model _ -> true
  | NoModel _ -> false

let info_msg result =
  Aux.info_msg (result_aux result)

let initial_result cfg predicate_table symbol_table negunits subterm_table clause_aux =
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits subterm_table clause_aux in
  NoModel aux


  
let first_result cfg predicate_table symbol_table negunits subterm_table clause_aux clauses =
  NegativeUnits.clear negunits;
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits subterm_table clause_aux in
  let t = Sys.time () in
  aux.Aux.t_start <- t;
  aux.Aux.t_model_start <- t;
  let node = (make_initial_node clauses aux) in
  find_model node
    
let abandon_result result =
  if is_open result then
    let aux = (result_aux result) in
    install_top_state aux;
    NoModel aux
  else
    let aux = (result_aux result) in
    install_top_state aux;
    result

let next_result result =
  match result with
    Model node ->
      Ic.undeclare_redundancies node.redundant_store;
      node.aux.Aux.t_model_start <-  Sys.time ();
      (match find_next_node node (-1) with
	NextNode next_node ->
	  find_model next_node
      | NoNextNode top_node ->
	  node.aux.Aux.refutation <- top_node.refutation;
	  NoModel node.aux)
  | NoModel _ ->
      result

let get_l_proof result =
  (result_aux result).Aux.proofs

let get_refutation result =
  (result_aux result).Aux.refutation

(****************************************************************************)

(*xxxx*)

(****************************************************************************)


let print_compiled_rules cfg predicate_table symbol_table negunits subterm_table clause_aux clauses =
  (* Mimic some steps performed at first_result *)
  NegativeUnits.clear negunits;
  let aux = Aux.make_aux cfg predicate_table symbol_table negunits subterm_table clause_aux in
  let t = Sys.time () in
  aux.Aux.t_start <- t;
  aux.Aux.t_model_start <- t;

  let strata =
    List.map 
      (fun clauses1 -> Stratum.make_stratum aux clauses1) 
      (IcSet.stratify clauses)
  in
  aux.Aux.n_strata <- (List.length strata);
  let n = ref 0 in
  print_string "% ========== Number of strata: ";
  print_int (List.length strata);
  print_endline " ==========";
  List.iter
    (fun stratum ->
      print_string "% ========== Stratum: ";
      print_int !n;
      print_endline " ==========";
      Stratum.print_compiled_clauses stratum;
      n := !n + 1)
    strata



