(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Evaluation of clause sets. *)

(** Operations are performed on hypertableau like structures. The fact base is
    modified by side effects: at a point in time, there is a single 
    "current result" (i.e. "current model") (with respect to a 
    [Predicate.predicate_table] and [NegativeUnits.store]).
*)

(** Represents an open hypertableau fragment or a closed hypertableau. *)
type result

(** Create a [result] object. The [result] object is not open (see
  [is_open]). It is used to initialize an environment structure 
  (see module [Env]). It may be seen as representing input facts,
  without inference rules applied.
*) 
val initial_result : Cfg.cfg ->
  Predicate.predicate_table -> Term.symbol_table ->
  NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> result

(** Derive the first model (or a closed tableau). *)
val first_result : Cfg.cfg -> Predicate.predicate_table -> Term.symbol_table ->
  NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> IcSet.clauses -> result

(** Derive the next model (or a closed tableau) with respect to the
   given model. If the given [result] is not open, it is returned directly.
   *)
val next_result : result -> result

(** Reset the fact base to its original state. The effect of calling
   [next_result] on the input [result] after calling this function is
   undefined. The [result] object returned is not open (see [is_open]).
   Notice that [first_result] and [next_result] automatically reset the
   fact base if they return a closed result.
 *)
val abandon_result : result -> result

(** True iff [result] is open (i.e. is a model). If [result] is not
 open, and has been returned by [first_result] or [next_result], it is closed
 (i.e. represents a refutation). *)
val is_open : result -> bool

(** Print information messages, such as statistics, about [result] to
   the message output. Reported values may only be accurate if this
   function is called immediately after the result has been returned
   by [first_result] or [next_result].
 *)
val info_msg : result -> unit

type refutation =
	NullRefutation
  | PrimitiveRefutation of Fact.proof
  | ComposedRefutation of (refutation list)

val get_refutation : result -> refutation

val refutation_to_term : Term.symbol_table -> refutation -> Term.term




(** Deprecated *)
val get_l_proof : result -> Fact.l_proof

val print_compiled_rules : 
   Cfg.cfg -> Predicate.predicate_table -> Term.symbol_table ->
   NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> IcSet.clauses -> unit
