(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)

(****************************************************************************)
(****************************************************************************)
(** Term Representation and Operations *)
(****************************************************************************)
(****************************************************************************)

(* Term representation with trail, contexts and multipliers operations
   roughly follows the Otter implementation. *)

(****************************************************************************)
(* Terms  *)
(****************************************************************************)

type variable = {
    mutable variable_number: int;
  }

and compound = {
    compound_functor: symbol;
    compound_args: term array; 
  }

and term = 
    Variable of variable
  | Constant of symbol
  | Compound of compound
  | Integer of int
  | Float of float
  | Unbound

and f =
    FUnbound
  | FUnary of (term -> term)
  | FBinary of (term -> term -> term)

and p =
    PUnbound
  | PDeterministic of 
      deterministic_builtin *
	(Opt.mode_specifier * Opt.discrimination_values * Opt.proof)

and deterministic_builtin = term -> context -> context -> trail -> trail

and symbol = 
    { mutable symbol_name: string; 
(* Mutable just to ensure object identity. *)
      symbol_arity: int;
      mutable symbol_function: f;
      mutable symbol_predicate: p
    }

and context = {
    context_terms: term array;
    context_contexts: context array;
    context_multiplier: int
  }

and trail = {mutable trail_varnum: int; 
	     (* Mutable just to ensure object identity. *)
	     trail_context: context;
	     trail_previous: trail}


(****************************************************************************)
(* Symbols *)
(****************************************************************************)

(* A symbol corresponds to a constant or functor symbol along with its
   arity. Object identity (i.e. '==') is used as symbol identifier.
   The symbol table is maintained by a weak hash set to effect symbol
   garbage collection. *)

let gensym_counter = ref (-1)

let gensym () =
  gensym_counter := !gensym_counter + 1;
  {symbol_name="$G_"^(string_of_int !gensym_counter);
   symbol_arity=0;
   symbol_function=FUnbound;
   symbol_predicate=PUnbound}
   



	    
(****************************************************************************)
(* Hashtbl as symbol table *)

type symbol_table = 
    {table:((string * int) , symbol) Hashtbl.t;
     mutable meta_functor_symbol:symbol;
     mutable rule_symbol:symbol;
     mutable disjunction_symbol:symbol;
     mutable conjunction_symbol:symbol;
     mutable true_symbol:symbol;
     mutable false_symbol:symbol;
     mutable stratified_neg_as_failure_symbol:symbol;
     mutable stratified_neg_as_failure_alternate_symbol:symbol;
     mutable stratified_neg_as_failure_not_subsumed_symbol:symbol;
     mutable findall_symbol:symbol;
     mutable neg2_symbol:symbol;
     mutable equality_symbol:symbol;
     mutable meta_0_symbol:symbol;
     mutable meta_1_symbol:symbol;
     mutable meta_2_symbol:symbol;
     mutable meta_3_symbol:symbol;
     mutable meta_4_symbol:symbol;
     mutable meta_5_symbol:symbol;
     mutable meta_6_symbol:symbol;
     mutable meta_7_symbol:symbol;
     mutable meta_8_symbol:symbol;
     mutable krh_dom_symbol:symbol;
     mutable krh_constant_symbol:symbol;
     mutable t_constant_symbol:symbol
     }

    
     
let find_symbol symbol_table name arity =
  try
    Hashtbl.find symbol_table.table (name, arity)
  with Not_found ->
    let sym = {symbol_name = name;
	       symbol_arity = arity; 
	       symbol_function=FUnbound;
	       symbol_predicate=PUnbound} in
    Hashtbl.add symbol_table.table (name, arity) sym;
    sym

       
    
let make_symbol_table () =
  let table = Hashtbl.create 64
  and null_sym = gensym () in
  let st =
    { table=table;
      meta_functor_symbol=null_sym;
      rule_symbol=null_sym;
      disjunction_symbol=null_sym;
      conjunction_symbol=null_sym;
      true_symbol=null_sym;
      false_symbol=null_sym;
      stratified_neg_as_failure_symbol=null_sym;
      stratified_neg_as_failure_alternate_symbol=null_sym;
      stratified_neg_as_failure_not_subsumed_symbol=null_sym;
      findall_symbol=null_sym;
      neg2_symbol=null_sym;
      equality_symbol=null_sym;
      meta_0_symbol=null_sym;
      meta_1_symbol=null_sym;
      meta_2_symbol=null_sym;
      meta_3_symbol=null_sym;
      meta_4_symbol=null_sym;
      meta_5_symbol=null_sym;
      meta_6_symbol=null_sym;
      meta_7_symbol=null_sym;
      meta_8_symbol=null_sym;
      krh_dom_symbol=null_sym;
      krh_constant_symbol=null_sym;
      t_constant_symbol=null_sym
    }
  in
  st.meta_functor_symbol<-find_symbol st "#" 1;
  st.rule_symbol<-find_symbol st ":-" 2;
  st.disjunction_symbol<-find_symbol st ";" 2;
  st.conjunction_symbol<-find_symbol st "," 2;
  st.true_symbol<-find_symbol st "true" 0;
  st.false_symbol<-find_symbol st "false" 0;
  st.stratified_neg_as_failure_symbol<-find_symbol st "\\+" 1;
  st.stratified_neg_as_failure_alternate_symbol<-find_symbol st "not" 1;
  st.stratified_neg_as_failure_not_subsumed_symbol<-find_symbol st 
      "not_subsumed" 1;
  st.findall_symbol<-find_symbol st "findall" 3;
  st.neg2_symbol<-find_symbol st "~" 1;
  st.equality_symbol<-find_symbol st "=" 2;
  st.meta_0_symbol<-find_symbol st "#0" 0;
  st.meta_1_symbol<-find_symbol st "#1" 1;
  st.meta_2_symbol<-find_symbol st "#2" 1;
  st.meta_3_symbol<-find_symbol st "#3" 1;
  st.meta_4_symbol<-find_symbol st "#4" 1;
  st.meta_5_symbol<-find_symbol st "#5" 1;
  st.meta_6_symbol<-find_symbol st "#6" 1;
  st.meta_7_symbol<-find_symbol st "#7" 1;
  st.meta_8_symbol<-find_symbol st "#8" 1;
  st.krh_dom_symbol<-find_symbol st "$krh_dom" 1;
  st.krh_constant_symbol<-find_symbol st "$krh_c" 0;
  st.t_constant_symbol<-find_symbol st "$t" 0;
  st
      
let symbol_table_iter f symbol_table =
  Hashtbl.iter (fun _ sym -> f sym) symbol_table.table

let get_meta_functor_symbol st = st.meta_functor_symbol
let get_rule_symbol st = st.rule_symbol
let get_disjunction_symbol st = st.disjunction_symbol
let get_conjunction_symbol st = st.conjunction_symbol
let get_true_symbol st = st.true_symbol
let get_false_symbol st = st.false_symbol
let get_stratified_neg_as_failure_symbol st =
  st.stratified_neg_as_failure_symbol
let get_stratified_neg_as_failure_alternate_symbol st =
  st.stratified_neg_as_failure_alternate_symbol
let get_stratified_neg_as_failure_not_subsumed_symbol st =
  st.stratified_neg_as_failure_not_subsumed_symbol
let get_findall_symbol st = st.findall_symbol
let get_neg2_symbol st = st.neg2_symbol
let get_equality_symbol st = st.equality_symbol
let get_meta_symbol st i =
	match i with
		0 -> st.meta_0_symbol
		|1 -> st.meta_1_symbol
		|2 -> st.meta_2_symbol
		|3 -> st.meta_3_symbol
		|4 -> st.meta_4_symbol
		|5 -> st.meta_5_symbol
		|6 -> st.meta_6_symbol
		|7 -> st.meta_7_symbol
		|8 -> st.meta_8_symbol
		|_ -> raise (Err.Error ("Bad argument in Term.set_meta_symbol."))
let get_krh_dom_symbol st = st.krh_dom_symbol
let get_krh_constant_symbol st = st.krh_constant_symbol
let get_t_constant_symbol st = st.t_constant_symbol


let set_meta_functor_symbol st symbol = st.meta_functor_symbol <- symbol
let set_rule_symbol st symbol = st.rule_symbol <- symbol
let set_disjunction_symbol st symbol = st.disjunction_symbol <- symbol
let set_conjunction_symbol st symbol = st.disjunction_symbol <- symbol
let set_true_symbol st symbol = st.true_symbol <- symbol
let set_false_symbol st symbol = st.false_symbol <- symbol
let set_stratified_neg_as_failure_symbol st symbol =
  st.stratified_neg_as_failure_symbol <- symbol
let set_stratified_neg_as_failure_alternate_symbol st symbol =
  st.stratified_neg_as_failure_alternate_symbol <- symbol
let set_stratified_neg_as_failure_not_subsumed_symbol st symbol =
  st.stratified_neg_as_failure_not_subsumed_symbol <- symbol

let set_findall_symbol st symbol = st.findall_symbol <- symbol
let set_neg2_symbol st symbol = st.neg2_symbol <- symbol
let set_equality_symbol st symbol = st.equality_symbol <- symbol 
let set_meta_symbol st i symbol =
	match i with
		0 -> st.meta_0_symbol <- symbol
		|1 -> st.meta_1_symbol <- symbol
		|2 -> st.meta_2_symbol <- symbol
		|3 -> st.meta_3_symbol <- symbol
		|4 -> st.meta_4_symbol <- symbol
		|5 -> st.meta_5_symbol <- symbol
		|6 -> st.meta_6_symbol <- symbol
		|7 -> st.meta_7_symbol <- symbol
		|8 -> st.meta_8_symbol <- symbol
		|_ -> raise (Err.Error ("Bad argument in Term.set_meta_symbol."))
let set_krh_dom_symbol st symbol = st.krh_dom_symbol <- symbol
let set_krh_constant_symbol st symbol = st.krh_constant_symbol <- symbol
let set_t_constant_symbol st symbol = st.t_constant_symbol <- symbol		

let sym_compare s1 s2 =
	if (s1.symbol_arity < s2.symbol_arity) then (-1)
	else 
		(if (s1.symbol_arity > s2.symbol_arity) then (1)
		else
			(Pervasives.compare s1.symbol_name s2.symbol_name))

let get_symbol_list st =
	let metasymbols =
		[(get_meta_functor_symbol st);
	     (get_rule_symbol st);
	     (get_disjunction_symbol st);
	     (get_conjunction_symbol st);
	     (get_true_symbol st);
	     (get_false_symbol st);
	     (get_stratified_neg_as_failure_symbol st);
	     (get_stratified_neg_as_failure_alternate_symbol st);
	     (get_stratified_neg_as_failure_not_subsumed_symbol st);
	     (get_findall_symbol st);
	     (get_neg2_symbol st);
	     (get_equality_symbol st);
	     (get_meta_symbol st 0);
	     (get_meta_symbol st 1);
	     (get_meta_symbol st 2);
	     (get_meta_symbol st 3);
	     (get_meta_symbol st 4);
	     (get_meta_symbol st 5);
	     (get_meta_symbol st 6);
	     (get_meta_symbol st 7);
	     (get_meta_symbol st 8);
	     (get_krh_dom_symbol st);
	     (get_krh_constant_symbol st);	     
	     (get_t_constant_symbol st)]
	and symbols = ref []
	in
		symbol_table_iter
			(fun s ->
				if (List.memq s metasymbols) then ()
				else (symbols := s :: !symbols))
			st;
		symbols := List.sort sym_compare !symbols;
		!symbols
		
let get_first_symbol st symbols =
	match symbols with
		[] -> (get_krh_constant_symbol st)
		|s :: _ -> s
				
exception Found		
		
let get_greater_symbol st symbols s =
	match symbols with
		[] -> (get_krh_constant_symbol st)
		|_ -> 
			(let cnt = ref 0
			in
				try
					(List.iter
						(fun s2 ->
							if not (s == s2) then
								(cnt := !cnt + 1)
							else (raise Found))
						symbols;
					raise (Err.Error ("Unlisted symbol in Term.get_greater_symbol.")))
				with Found ->
					(cnt := !cnt+1;
					if (!cnt > (List.length symbols)) then (s)
					else (List.nth symbols !cnt)))
			
		 

(****************************************************************************)
(* Alternative: Weak as symbol table (should effect symbol GC) *)
(*
module SymbolTable =
  Weak.Make(
  struct
    type t = symbol
    let equal s1 s2 = 
      s1.symbol_name = s2.symbol_name &&
      s1.symbol_arity = s2.symbol_arity
    let hash = function s -> Hashtbl.hash s.symbol_name
  end)
  
type symbol_table = SymbolTable.t

let make_symbol_table () = SymbolTable.create 64

let find_symbol symbol_table name arity =
  (* ### Hmm - costly entry creation just for lookup *)
  let sym = {symbol_name = name;
	     symbol_arity = arity; 
	     symbol_function=FUnbound;
	     symbol_predicate=PUnbound} in
  try
    let sym1 = 
      SymbolTable.find symbol_table sym
    in sym1
  with Not_found ->
    SymbolTable.add symbol_table sym;
    sym
      
let symbol_table_iter = SymbolTable.iter
*)


(****************************************************************************)
(* Abstract Export Interface *)
(****************************************************************************)

let make_variable number = {variable_number = number}

let variable_number variable = variable.variable_number

let symbol_name symbol = symbol.symbol_name

let symbol_arity symbol = symbol.symbol_arity

let symbol_function symbol = symbol.symbol_function

let symbol_predicate symbol = symbol.symbol_predicate

let set_symbol_function symbol f =
  symbol.symbol_function <- f

let set_symbol_predicate symbol p =
  symbol.symbol_predicate <- p

let make_compound symbol array = 
  {compound_functor = symbol; compound_args = array}

let compound_functor compound = compound.compound_functor

let compound_args compound = compound.compound_args

let get_compound_args term =
	match term with
		Compound k -> k.compound_args
		|_ -> raise (Err.Error "Non-compound in Term.get_compound_args")

let context_multiplier context =
  context.context_multiplier

let context_terms context =
  context.context_terms
  
let context_contexts context =
  context.context_contexts  
  

(****************************************************************************)
(* Context *)
(****************************************************************************)

(** LIMIT: Maximum number of distinct variables in a clause. *)   
let max_vars = 64
(* let max_vars = 128 *)

let null_context = {context_terms = [||];
		    context_contexts = [||];
		    context_multiplier = 0}

let make_context multiplier = 
  {context_terms = Array.make max_vars Unbound;
   context_contexts = Array.make max_vars null_context;
   context_multiplier = multiplier}

let make_bound_context variable_bindings multiplier = 
  {context_terms = variable_bindings;
   context_contexts = Array.make max_vars null_context;
   context_multiplier = multiplier}
   

   
let rec dereference t c =
  if c == null_context then
    (* This case may result e.g. from match_term. *)
    (t, c)
  else
    match t with
      Variable v ->
	let vn = v.variable_number in
	( match c.context_terms.(vn) with
	  Unbound ->
	    (t, c)
	| _ ->
	    dereference c.context_terms.(vn) c.context_contexts.(vn)
	 )
    | _ ->
	(t, c)


			

(****************************************************************************)
(* Trail *)
(****************************************************************************)

let rec fail = {trail_varnum=(-1);
		trail_context=null_context;
		trail_previous=fail}

let base_trail = {trail_varnum=(-1);
		  trail_context=null_context;
		  trail_previous=fail} 

let make_trail () = base_trail

let failed trail = trail == fail

type saved_trail = trail

let save_trail trail = trail

let bind vn c1 t2 c2 trail = 
  c1.context_terms.(vn) <- t2;
  c1.context_contexts.(vn) <- c2;
  {trail_varnum=vn;
   trail_context=c1;
   trail_previous=trail}

let rec undo_bindings trail saved_trail = 
  if trail == saved_trail then
    fail
  else
    begin
      trail.trail_context.context_terms.(trail.trail_varnum) <- Unbound;
      undo_bindings trail.trail_previous saved_trail
    end


(****************************************************************************)
(* Unify Terms *)
(****************************************************************************)

let rec occur_check vn vc t c = 
  match t with
    Compound k ->
      occur_check_args
	(k.compound_functor.symbol_arity-1)
	vn
	vc
	k.compound_args
	c
  | Variable v ->
      let tvn = v.variable_number in
      if tvn = vn && c == vc then
	false
      else (
	match c.context_terms.(tvn) with
	  Unbound ->
	    true
	| _ -> occur_check vn vc c.context_terms.(tvn) c.context_contexts.(tvn)
       )  
  | _ ->
      true

and occur_check_args i vn vc a c = 
  if i < 0 then
    true
  else
    occur_check vn vc a.(i) c && occur_check_args (i-1) vn vc a c

let rec unify t1 c1 t2 c2 trail = 
  let (t1, c1) = dereference t1 c1
  and (t2, c2) = dereference t2 c2 in
  match t1, t2 with
    Variable v1, Variable v2 ->
      let vn1 = v1.variable_number in
      if vn1 = v2.variable_number && c1 == c2 then
	trail
      else
	bind vn1 c1 t2 c2 trail
  | Variable v1, _ ->
      let vn1 = v1.variable_number in
      if occur_check vn1 c1 t2 c2 then
	bind vn1 c1 t2 c2 trail
      else
	fail
  | _, Variable v2 ->
      let vn2 = v2.variable_number in
	if occur_check vn2 c2 t1 c1 then
	  bind vn2 c2 t1 c1 trail
	else
	  fail
  | Constant s1, Constant s2 ->
      if s1 == s2 then
	trail
      else
	fail
  | Compound k1, Compound k2 ->
      if k1.compound_functor == k2.compound_functor then
	unify_args
	  k1.compound_args
	  c1
	  k2.compound_args
	  c2
	  (save_trail trail)
	  trail
	  ( k1.compound_functor.symbol_arity - 1)
      else 
	fail
  | _, _ ->
      if t1 = t2 then trail else fail

and unify_args a1 c1 a2 c2 saved_trail trail i = 
  if i < 0 then
    trail
  else
    let new_trail = unify a1.(i) c1 a2.(i) c2 trail in
    if failed new_trail then
      undo_bindings trail saved_trail
    else
      unify_args a1 c1 a2 c2 saved_trail new_trail (i-1)


(****************************************************************************)
(* Match Terms *)
(****************************************************************************)
	
let rec term_ident t1 t2 =
  match t1, t2 with
    Compound k1, Compound k2 ->
      if k1.compound_functor == k2.compound_functor then
	term_ident_args 
	  k1.compound_args
	  k2.compound_args
	  ( k1.compound_functor.symbol_arity - 1)
      else false
  | Variable v1, Variable v2 ->
      v1.variable_number = v2.variable_number
  | Constant c1, Constant c2 ->
      c1 == c2
  | _, _ ->
      t1 = t2

and term_ident_args a1 a2 i =
  if i < 0 then
    true
  else
    ( term_ident a1.(i) a2.(i) && term_ident_args a1 a2 (i - 1) )

let rec term_non_ident t1 c1 t2 c2 =
	match t1, t2 with
		|Constant c1, Constant c2 ->
			not (c1 == c2)
		|Constant c1, _ -> true
		| _, Constant c2 -> true
		|Compound k1, Compound k2 ->
			if not (k1.compound_functor == k2.compound_functor) then true
			else let len = k1.compound_functor.symbol_arity
				in
					if not (len = k2.compound_functor.symbol_arity) then true
					else let index = ref 0
						in
							while !index<len && not (term_non_ident k1.compound_args.(!index) c1 k2.compound_args.(!index) c2) do
								index := !index+1
							done;
							if !index=len then false
							else true
		|Compound k1, _ -> true
		| _, Compound k2 -> true
		|Variable v1, _ ->
			let t1, c1 = dereference t1 c1 in
				term_non_ident t1 c1 t2 c2
		| _, Variable v2 ->
			let t2, c2 = dereference t2 c2 in
			term_non_ident t1 c1 t2 c2
		| _, _ ->
			not (t1 = t2)    
    
    
let rec match_terms t1 c1 t2 trail =
  match t1, t2 with
    Variable v1, _ ->
      let vn = v1.variable_number in
      ( match c1.context_terms.(vn) with
	Unbound ->
	  bind vn c1 t2 null_context trail
      | _ -> if term_ident c1.context_terms.(vn) t2 then trail else fail)
  | _, Variable _ ->
      fail
  | Compound k1, Compound k2 ->
      if k1.compound_functor == k2.compound_functor then
	match_terms_args 
	  k1.compound_args
	  c1 
	  k2.compound_args 
	  (save_trail trail)
	  trail
	  ( k1.compound_functor.symbol_arity - 1)
      else
	fail
  | Constant s1, Constant s2 ->
      if s1 == s2 then
	trail
      else
	fail
  | _, _ ->
      if t1 = t2 then trail else fail

and match_terms_args a1 c1 a2 saved_trail trail i =
  if i < 0 then
    trail
  else
      let new_trail = match_terms a1.(i) c1 a2.(i) trail in
      if failed new_trail then
	undo_bindings trail saved_trail
      else
	match_terms_args a1 c1 a2 saved_trail new_trail (i-1)


(****************************************************************************)
(* SubsumesChk *)
(****************************************************************************)

(* Similar to match_terms, but more efficient if the binding is not needed. *)

(*
let subsumes_chk t1 t2 =
  not (failed (match_terms t1 (make_context 0) t2 base_trail))
*)

let rec subsumes_chk t1 t2 =
  subsumes_chk_1 t1 (Array.make max_vars Unbound) t2

and subsumes_chk_1 t1 c1 t2 =
  match t1, t2 with
    Variable v1, _ ->
      begin
	let vn = v1.variable_number in
	match c1.(vn) with
	  Unbound ->
	    c1.(vn) <- t2;
	    true
	| _ -> (term_ident c1.(vn) t2)
      end
  | _, Variable _ ->
      false
  | Compound k1, Compound k2 ->
      if k1.compound_functor == k2.compound_functor then
	subsumes_chk_args 
	  k1.compound_args
	  c1 
	  k2.compound_args 
	  ( k1.compound_functor.symbol_arity - 1)
      else
	false
  | Constant s1, Constant s2 ->
      if s1 == s2 then
	true
      else
	false
  | _, _ ->
      if t1 = t2 then true else false

and subsumes_chk_args a1 c1 a2 i =
  if i < 0 then
    true
  else
    subsumes_chk_1 a1.(i) c1 a2.(i) && subsumes_chk_args a1 c1 a2 (i-1)

let variant term1 term2 =
  (* for now *)
  subsumes_chk term1 term2 && subsumes_chk term2 term1


(****************************************************************************)
(* Apply Bindings *)
(****************************************************************************)


let rec copy_apply t c = 
  let (t, c) = dereference t c in
  match t with
    Variable v ->
      Variable 
	{variable_number = 
	 if c == null_context then
	   v.variable_number
	 else
	   c.context_multiplier * max_vars + v.variable_number}
  | Compound k ->
      Compound
	{compound_functor = k.compound_functor;
	 compound_args = Array.map (fun t1 -> copy_apply t1 c) k.compound_args}
  | _ ->
      t


(****************************************************************************)
(* Apply Bindings - Experimental Version *)
(****************************************************************************)

(* Returns the input term if it contains no variable (bound or unbound) to
   save memory. *)

let rec apply_shared t c = 
  let (t1, c) = dereference t c in
  match t1 with
    Variable v ->
      (true,
       Variable 
	 {variable_number = 
	  if c == null_context then
	    v.variable_number
	  else
	    c.context_multiplier * max_vars + v.variable_number})
  | Compound k ->
      let copy_needed = ref false in
      let args_copy = 
	Array.map
	  (fun t2 ->
	    let (copy_needed_1, t3) = apply_shared t2 c in
	    copy_needed := (!copy_needed || copy_needed_1);
	    t3)
	  k.compound_args in
      if !copy_needed then
	(true, 
	 Compound {compound_functor = k.compound_functor;
		   compound_args = args_copy})
      else
	(t != t1,
	 t1)
  | _ ->
      (t != t1,
       t1)

let apply t c =
  snd (apply_shared t c)


(****************************************************************************)
(* Copy Term *)
(****************************************************************************)

let copy_term t = 
  copy_apply t null_context


(****************************************************************************)
(* Renumber Variables *)
(****************************************************************************)

exception Max_vars_exceeded
 
let rec init_renum_table varnums t =
	match t with
		Compound k ->
			Array.iter 
				(fun t1 ->
					init_renum_table varnums t1)
				k.compound_args
	| Variable v ->
		let varnum = v.variable_number
		in
			if varnum >= max_vars then
				raise Max_vars_exceeded
			else 
				varnums.(v.variable_number) <- v.variable_number 
	| _ -> ()

let rec renumber_variables_term varnums t = 
  match t with
    Compound k ->
      Array.iter (fun t1 -> renumber_variables_term varnums t1) k.compound_args
  | Variable v ->
      let i = ref 0 in
      while !i < max_vars
	  && varnums.(!i) != -1
	  && varnums.(!i) != v.variable_number do
	i := !i+1
      done;
      if !i = max_vars then
	raise Max_vars_exceeded
      else if varnums.(!i) = -1 then
	begin
	  varnums.(!i) <- v.variable_number;
	  v.variable_number <- !i
	end
      else
      	v.variable_number <- !i
  | _ -> ()

let rec renumber_variables_term_specific varnums t =
	let offset = ref 0 in
		while !offset < max_vars
			&& varnums.(!offset) != -1 do 
				offset := !offset + 1
		done;
		renumber_variables_term_specific_from_offset varnums t !offset

and renumber_variables_term_specific_from_offset varnums t offset =
  match t with
    Compound k ->
      Array.iter (fun t1 -> renumber_variables_term_specific_from_offset varnums t1 offset) k.compound_args
  | Variable v ->
      let i = ref offset in
      while !i < max_vars
	  && varnums.(!i) != -1
	  && varnums.(!i) != v.variable_number do
	i := !i+1
      done;
      if !i = max_vars then
	raise Max_vars_exceeded
      else if varnums.(!i) = -1 then
	begin
	  varnums.(!i) <- v.variable_number;
	  v.variable_number <- !i
	end
      else
      	v.variable_number <- !i
  | _ -> ()  
  	

	
(** [rule_subsumes_chk term_1 term_2 symbol_table] is true iff [term_2] is an instance of [term_1]. If the arguments are
rule terms, then the commutativity of head and body literals is accounted for. *)
(*
rule_subsumes_chk t1 t2 symbol_table =
	let rule_symbol = get_rule_symbol symbol_table
	in
		if (Convert.term_functor_symbol t1) == rule_symbol then (
			if (Convert.term_functor_symbol t2) == rule_symbol then (
				)
			else false)
		else (
			if (Convert.term_functor_symbol t2) == rule_symbol then false
			else subsumes_chk t1 t2)
			
			
*)


exception Failed

let rec variant_chk t1 t2 =
	variant_chk_1 t1 (Array.make max_vars (-1)) t2 (Array.make max_vars (-1))
		
and variant_chk_1 t1 b1 t2 b2 = 
	match t1, t2 with
		Compound k1, Compound k2 ->
			if k1.compound_functor == k2.compound_functor then
				try
					(Array.iteri 
						(fun i arg -> 
							if (variant_chk_1 
									arg 
									b1 
									k2.compound_args.(i)
									b2) then ()
							else raise Failed) 
						k1.compound_args;
						true)
				with Failed -> false
			else false
		| Variable v1, Variable v2 ->
			if b1.(v1.variable_number)<0 && b2.(v2.variable_number)<0 then
				(b1.(v1.variable_number) <- v2.variable_number;
				b2.(v2.variable_number) <- v1.variable_number;
				true)
			else
				(if b1.(v1.variable_number) == v2.variable_number then true
				else false)
		| Compound _, _ -> false
		| _, Compound _ -> false
		| Variable _, _ -> false
		| _, Variable _ -> false	
		| Constant c1, Constant c2 ->
      		c1 == c2
  		| _, _ ->
      		t1 = t2
      		
let rec variant_chk_bindings t1 t2 index tob1 tob2 byb1 byb2 = 
	if (variant_chk_bindings_1 t1 t2 index tob1 tob2 byb1 byb2) then true
	else
		(variant_chk_bindings_reset_bindings index tob1 tob2 byb1 byb2;
		false)

and variant_chk_bindings_reset_bindings index tob1 tob2 byb1 byb2 =
	Array.iteri
		(fun i pos ->
			if (pos = index) then
				(tob1.(i) <- (-1);
				byb1.(i) <- (-1)))
		byb1;
	Array.iteri
		(fun i pos ->
			if (pos = index) then
				(tob2.(i) <- (-1);
				byb2.(i) <- (-1)))
		byb2      	
		
and variant_chk_bindings_1 t1 t2 index tob1 tob2 byb1 byb2 =		
	match t1, t2 with
		Compound k1, Compound k2 ->
			if k1.compound_functor == k2.compound_functor then ( 
				try
					(Array.iteri 
						(fun i arg -> 
							if (variant_chk_bindings_1 
									arg 
									k2.compound_args.(i)
									index
									tob1
									tob2
									byb1
									byb2) then ()
							else raise Failed) 
						k1.compound_args;
						true)
				with Failed -> false)
			else false
		| Variable v1, Variable v2 ->
			if tob1.(v1.variable_number)<0 && tob2.(v2.variable_number)<0 then
				(tob1.(v1.variable_number) <- v2.variable_number;
				tob2.(v2.variable_number) <- v1.variable_number;
				byb1.(v1.variable_number) <- index;
				byb2.(v2.variable_number) <- index;
				true)
			else
				(if tob1.(v1.variable_number) == v2.variable_number then true
				else false)
		| Compound _, _ -> false
		| _, Compound _ -> false
		| Variable _, _ -> false
		| _, Variable _ -> false	
		| Constant c1, Constant c2 ->
      		c1 == c2
  		| _, _ ->
      		t1 = t2
      		
let binary_flip term =
	match term with
		Compound k ->
			if (Array.length k.compound_args)=2 then
				Compound (make_compound k.compound_functor [|(k.compound_args.(1)); (k.compound_args.(0))|])
			else raise Failed
		| _ -> raise Failed
		
(*XXXXXXXXXXXXXXXXX*)

module type OutputType =
  sig
    type t
    val	output_char: t -> char -> unit
    val output_string: t -> string -> unit
  end

(* These character classes are taken from the Sicstus Prolog manual.
   But that seems not fully correct, e.g [] and {} are not considered.
   See also module GenlexProlog. *)

type char_class =
    CharClassLayout
  | CharClassSmallLetter
  | CharClassCapitalLetter
  | CharClassDigit
  | CharClassSymbolChar
  | CharClassSoloChar
  | CharClassPunctuationChar
  | CharClassQuoteChar
  | CharClassUnderline
  | CharClassUnspecified

let classify_char c =
  match c with
    ('\000'..'\032' | '\127'..'\159') ->
      CharClassLayout
  | ('\097'..'\122' |'\223'..'\246' | '\248'..'\255') ->
      CharClassSmallLetter
  | ('\065'..'\090' |'\192'..'\214' | '\216'..'\222') ->
      CharClassCapitalLetter
  | ('\048'..'\057') ->
      CharClassDigit
  | ('\035' | '\036' | '\038' | '\042' | '\043' | '\045' | '\046' |
    '\047' | '\058' | '\060' | '\061' | '\062' | '\063' | '\064' |
    '\092' | '\094' | '\096' | '\126' | '\215' | '\247' |
    '\160'..'\191') ->
      CharClassSymbolChar
  | ('\033' | '\059') ->
      CharClassSoloChar
  | ('\037' | '\040' | '\041' | '\044' | '\091' | '\093' | '\123' |
    '\124' | '\125') ->
      CharClassPunctuationChar
  | ('\034' | '\039') ->
      CharClassQuoteChar
  | '\095' ->
    CharClassUnderline

let rec alphanum_string s i =
  if i >= String.length s then
    true
  else 
    match classify_char (String.get s i) with
      CharClassCapitalLetter 
    | CharClassSmallLetter 
    | CharClassDigit 
    | CharClassUnderline ->
	alphanum_string s (i+1)
    | _ -> false
    
let rec symbol_string s i =
  if i >= String.length s then
    true
  else 
    match classify_char (String.get s i) with
      CharClassSymbolChar ->
	symbol_string s (i+1)
    | _ -> false


module Make (Output: OutputType) =
  struct
    let output_name oc s =
      (* Just considers \ and ' as escape and quote characters, anything
	 else is printed literally. *)
      if String.length s == 0 then
	Output.output_string oc "''"
      else if 
	(( match classify_char (String.get s 0) with
	  CharClassSmallLetter ->
	    alphanum_string s 1
	| _ -> 
	    false )
       || (s <> "." && symbol_string s 0 && 
	   not (String.get s 0 == '/' 
		  && (String.length s) > 1
		  && String.get s 1 == '*'))
       || s = "!"
       || s = ";"
       || s = "[]" 
       || s = "{}" ) then
	Output.output_string oc s
      else
	begin
	  Output.output_char oc '\'';
	  for i = 0 to (String.length s)-1 do
	    let c = String.get s i in
	    if c == '\'' then
	      Output.output_string oc "\\'"
	    else if c == '\\' then
	      Output.output_string oc "\\\\"
	    else
	      Output.output_char oc c
	  done;
	  Output.output_char oc '\'';
	end
	  
    let rec output oc t =
      match t with
	Variable v ->
	  Output.output_string oc "X";
	  Output.output_string oc (string_of_int (variable_number v))
      | Constant s ->
	  output_name oc (symbol_name s)
      | Compound k ->
	  let name = (symbol_name (compound_functor k))
	  and arity = (symbol_arity (compound_functor k)) in
	  if name = "." && arity = 2 then
	    begin
	      Output.output_string oc "[";
	      output oc (compound_args k).(0);
	      output_cdr oc (compound_args k).(1);
	      Output.output_string oc "]"
	    end
	  else
	    begin
	      output_name oc name;
	      Output.output_string oc "(";
	      output oc (compound_args k).(0);
	      for i = 1 to (symbol_arity (compound_functor k))-1 do
		Output.output_string oc ", ";
		output oc (compound_args k).(i)
	      done;
	      Output.output_string oc ")"
	    end
      | Integer n ->
	  Output.output_string oc (string_of_int n)
      | Float n ->
	  let s1 = (string_of_float n) in
	  let s2 = 
	    if String.get s1 ((String.length s1) - 1) = '.' then
	      s1^"0"
	    else
	      s1 in
	  Output.output_string oc s2
      | Unbound ->
	  Output.output_string oc "'<unbound>'"
	    
    and output_cdr oc t =
      match t with
	Compound k ->
	  let name = (symbol_name (compound_functor k))
	  and arity = (symbol_arity (compound_functor k)) in
	  if name = "." && arity = 2 then
	    begin
	      Output.output_string oc ", ";
	      output oc (compound_args k).(0);
	      output_cdr oc (compound_args k).(1)
	    end
	  else
	    begin
	      Output.output_string oc " | ";
	      output oc t
	    end
      | Constant s ->
	  if (symbol_name s) = "[]" then
	    ()
	  else
	    begin
	      Output.output_string oc " | ";
	      output oc t
	    end
      | _ ->
	  Output.output_string oc " | ";
	  output oc t
  end	

module OutChannelPrint =
  Make(
  struct
    type t = Pervasives.out_channel
    let output_char = Pervasives.output_char
    let output_string = Pervasives.output_string
  end)

let print term =
  OutChannelPrint.output stdout term
	
let output out term =
  OutChannelPrint.output out term

let sbuff = Buffer.create 64

module StringPrint =
  Make(
  struct
    type t = unit
    let output_char oc char =
      Buffer.add_char sbuff char
    let output_string oc string =
      Buffer.add_string sbuff string
  end)

let print_to_string term =
  Buffer.reset sbuff;
  StringPrint.output () term;
  Buffer.contents sbuff

  let ppp t = print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "
let pa a = Array.iter (fun t -> ppp t;ps " ") a
	

let rec np_subsumes_chk_bindings t1 t2 index tob1 byb1 byb2 =
	if (np_subsumes_chk_bindings_1 t1 t2 index tob1 byb1 byb2) then true
	else 
		(np_subsumes_chk_bindings_reset_bindings index tob1 byb1 byb2;
		false)			

and np_subsumes_chk_bindings_1 t1 t2 index tob1 byb1 byb2 = 
	match t1, t2 with
		Compound k1, Compound k2 ->
			if k1.compound_functor == k2.compound_functor then ( 
				try
					(Array.iteri 
						(fun i arg -> 
							if (np_subsumes_chk_bindings_1 
									arg 
									k2.compound_args.(i)
									index
									tob1
									byb1
									byb2) then ()
							else raise Failed) 
						k1.compound_args;
						true)
				with Failed -> false)
			else false
		| Variable v1, Variable v2 ->
			if tob1.(v1.variable_number)==Unbound then
				(tob1.(v1.variable_number) <- t2;
				byb1.(v1.variable_number) <- index;
				byb2.(v2.variable_number) <- index;
				true)
			else
				(if (term_ident tob1.(v1.variable_number) t2) then (true)
				else (false))
		| Variable v1, _ ->
			if tob1.(v1.variable_number)==Unbound then
				(tob1.(v1.variable_number) <- t2;
				byb1.(v1.variable_number) <- index;
				true)
			else
				(if (term_ident tob1.(v1.variable_number) t2) then true
				else false)				
		| Compound _, _ -> false
		| _, Compound _ -> false
		| _, Variable _ -> false	
		| Constant c1, Constant c2 ->
      		c1 == c2
  		| _, _ ->
      		t1 = t2		
			
and np_subsumes_chk_bindings_reset_bindings index tob1 byb1 byb2 =
	Array.iteri
		(fun i pos ->
			if (pos = index) then
				(tob1.(i) <- Unbound;
				byb1.(i) <- (-1)))
		byb1;
	Array.iteri
		(fun i pos ->
			if (pos = index) then
				(byb2.(i) <- (-1)))
		byb2      		
      		
let rec np_subsumes_chk t1 t2 =
	np_subsumes_chk_bindings t1 t2 0 (Array.make max_vars Unbound) (Array.make max_vars (-1)) (Array.make max_vars (-1))		
				
let make_anonymous_binary_compound t1 t2 =
	Compound(
		make_compound 
			({symbol_name="";
			symbol_arity=2;
			symbol_function=FUnbound;
			symbol_predicate=PUnbound})
			[|t1; t2|])
			
(*			
let split_context context special_varnums =
	let normal_contextbindings = Array.make Term.max_vars Term.Unbound
	and normal_context_contexts = Array.make max_vars null_context
	and special_bindings = Array.make Term.max_vars Term.Unbound
	and special_context_contexts = Array.make max_vars null_context
	in
		Array.iteri
			context.context_terms
*)

let extract_flat_bindings context =
	let bindings = Array.make max_vars Unbound
	in
		for i = 0 to (max_vars - 1) do
			(bindings.(i) <- apply (Variable (make_variable i)) context)
		done;
		bindings   
		
		
let rec iter_term f t =
	f t;
	match t with
		Compound k ->
			Array.iter
				(fun arg ->
					iter_term f arg)
				k.compound_args
		|_ -> () 