(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(** Term representation and operations on terms. *)

(** Term representation and operations roughly follow the implementation
   in the OTTER theorem prover. *)

(** {6 Term Structure} *)

type variable
(** Different occurences of the same variable are not shared as
   objects. They are identified by their variable_number. *)
   
and symbol
(** A symbol corresponds to a constant or functor symbol along with its
   arity. Object identity (i.e. '==') is used as symbol identifier.
 *)

and compound

and term =
    Variable of variable
  | Constant of symbol
  | Compound of compound
  | Integer of int
  | Float of float
  | Unbound

val make_variable : int -> variable

val variable_number : variable -> int

(** [gensym ()] returns a newly generated symbol that is not
   maintained by a symbol table. *) 
val gensym : unit -> symbol

val symbol_name : symbol -> string

val symbol_arity : symbol -> int

val make_compound : symbol -> term array -> compound

val compound_functor : compound -> symbol

val compound_args : compound -> term array

val get_compound_args : term -> term array

(** {6 Symbol Table} *)

(** Symbols are maintained by a symbol table.
*)
type symbol_table

(** Create a symbol table. *)
val make_symbol_table : unit -> symbol_table

(** [find_symbol name arity] returns the symbol corresponding to [name] and
   [arity] in the symbol table. If no such symbol is in the table, one is
   created and put into the table.
   *)
val find_symbol : symbol_table -> string -> int -> symbol


(** Iterate over the symbols in the symbol table. *)
val symbol_table_iter : (symbol -> unit) -> symbol_table -> unit

(** {6 Symbols for Logic Operators} *)

val get_meta_functor_symbol : symbol_table -> symbol
val get_rule_symbol : symbol_table -> symbol
val get_disjunction_symbol : symbol_table -> symbol
val get_conjunction_symbol : symbol_table -> symbol
val get_true_symbol : symbol_table -> symbol
val get_false_symbol : symbol_table -> symbol
val get_stratified_neg_as_failure_symbol : symbol_table -> symbol
val get_stratified_neg_as_failure_alternate_symbol : symbol_table -> symbol
val get_stratified_neg_as_failure_not_subsumed_symbol : symbol_table -> symbol
val get_findall_symbol : symbol_table -> symbol
val get_neg2_symbol : symbol_table -> symbol
val get_equality_symbol : symbol_table -> symbol
val get_meta_symbol : symbol_table -> int -> symbol
val get_krh_dom_symbol : symbol_table -> symbol
val get_krh_constant_symbol : symbol_table -> symbol
val get_t_constant_symbol : symbol_table -> symbol

val set_meta_functor_symbol : symbol_table -> symbol -> unit
val set_rule_symbol : symbol_table -> symbol -> unit
val set_disjunction_symbol : symbol_table -> symbol -> unit
val set_conjunction_symbol : symbol_table -> symbol -> unit
val set_true_symbol : symbol_table -> symbol -> unit
val set_false_symbol : symbol_table -> symbol -> unit
val set_stratified_neg_as_failure_symbol : symbol_table -> symbol -> unit
val set_stratified_neg_as_failure_alternate_symbol :
    symbol_table -> symbol -> unit
val set_stratified_neg_as_failure_not_subsumed_symbol : 
    symbol_table -> symbol -> unit
val set_findall_symbol : symbol_table -> symbol -> unit
val set_neg2_symbol : symbol_table -> symbol -> unit
val set_equality_symbol : symbol_table -> symbol -> unit
val set_meta_symbol : symbol_table -> int -> symbol -> unit
val set_krh_dom_symbol : symbol_table -> symbol -> unit
val set_krh_constant_symbol : symbol_table -> symbol -> unit
val set_t_constant_symbol : symbol_table -> symbol -> unit

val get_symbol_list : symbol_table -> symbol list

(** {6 Context} *)

(** A context is used to represent undoable variable bindings within a term.

   With a context, a so called multiplier is associated. It is used to
   distinguish variables from different terms when applying a substitution.
   (See also the OTTER implementation.)
*)
type context

val context_terms : context -> term array
val context_contexts : context -> context array

(** LIMIT: The maximum number of distinct variables in a clause. *)
val max_vars : int

(** Thrown if the maximum number of distinct variables in a clause is exceeded. *)
exception Max_vars_exceeded

(** Create a context. The int argument is the multiplier, i.e. an
   integer >= 0, of the context. *)
val make_context : int -> context

(** Create a context with predefined variable bindings. The term array argument
must be of length [max_vars]. *) 
val make_bound_context : term array -> int -> context

(** Creates an array of variable bindings from [context]. *)
val extract_flat_bindings : context -> term array

(** A context that represents the empty binding. It is allowed as argument
   only where explicitely specified. *)
val null_context : context

(** Returns the multiplier of the given context.
*)
val context_multiplier : context -> int

(** If [term] is a variable, [dereference_term term context] returns
its binding in [context] along with the context too look up bindings in
subterms, otherwise the input is returned. This function is not called
recursively, i.e. a returned term may contain bound variables.

   [context] may be [null_context].
*)
val dereference : term -> context -> (term * context)

(** {6 Trail} *)

(** A trail is used to record variable bindings, that may be undone.
   Also, a special instance of type trail is used to represent unification
   failure.
 *)
type trail

(** [make_trail] returns an empty trail that is different from [fail]. *)
val make_trail : unit -> trail

(** [fail] represents unification failure. This object is identified with the
   [failed] function. *)
val fail : trail

(** [failed trail] tests if the given trail is [fail]. *)
val failed : trail -> bool

(** A representation of a trail state, that can be used to undo bindings. *)
type saved_trail

(** [save_trail trail] returns a representation of the given trail's state,
   that can be used to undo bindings. *)
val save_trail : trail -> saved_trail

(** [undo_bindings trail saved_trail] undoes bindings in [trail] until
   the state represented by [saved_trail] is reached. The given trail
   may be destructively modified by the implementation.

   Returns [fail].
*)
val undo_bindings : trail -> saved_trail -> trail


(** {6 Term Operations} *)

(** [unify term1 context1 term2 context2] attempts to unify
   term1 in context1 with term2 in context2. If unification
   succeeds, the contexts may be modified and the 
   resulting trail is returned. If the unification fails, fail
   is returned and the contexts have the same state as before.
*)
val unify : term -> context -> term -> context -> trail -> trail

(** [apply term context] returns a copy of [term] with the substitution
[context] applied; subterms are shared between original and copy. *)
val apply : term -> context -> term

(** [copy_apply term context] returns a copy of [term] with the substitution
[context] applied; no subterms are shared between original and copy. *)
val copy_apply : term -> context -> term

(** [copy_term] returns a copy of [term]. *)
val copy_term : term -> term

(** [match_terms term1 context1 term2 trail] computes the substitution
   [context1] under which [term2] is an instance of [term1].

   Both terms may not share variables.

   Variables in [term2] must be uninstantiated.

   [context1] must be fresh or modified by calls to [match_terms] only
   (and no call to [unify] (???)).

   Similar as in [unify], if the matching succeeds, the resulting trail
   is returned. If the matching fails, the context has the same state as
   before and fail is returned.
*)
val match_terms : term -> context -> term -> trail -> trail

(** [subsumes_chk term1 term2] is true iff [term2] is an instance of [term1].

   Both terms may not share variables.
*)
val subsumes_chk : term -> term -> bool

(** [subsumes_chk_1 term1 variable_bindings term2] is true iff [term2] is an instance of [term1].
Supporting function for [subsumes_chk], but can be useful on its own.

   Both terms may not share variables.
*)
val subsumes_chk_1 : term -> term array -> term -> bool


(** [variant term1 term2] is true iff both terms are variants.

   Both terms may not share variables.
*)
val variant : term -> term -> bool

(** [term_ident term1 term2] returns true iff both terms are identical, i.e.
   have the same structure and identical variables. *)
val term_ident : term -> term -> bool


(** [term_non_ident term1 context1 term2 context2] returns true iff both terms are not identical
under the given contexts. Experimental. *)
val term_non_ident : term -> context -> term -> context -> bool

(** [renumber_variables_term table term] reassigns variable numbers
   starting from 0 to the variables in [term]. [table] is an int array
   of length [max_vars], that maps old var numbers to the new ones. It
   must be initialized by -1. [table] is modified and can be passed to the
   next call of [renumber_variables] to consistently renumber
   variables in a set of terms. 

   Note: variables in [term] may not be shared as objects.
*) 
val renumber_variables_term : int array -> term -> unit

(** [renumber_variables_term_specific table term] does the same as [renumber_variables_term],
but assigns a new namespace for term, so that multiple calls will not lead to multiple terms
sharing variable numbers.
*)
val renumber_variables_term_specific : int array -> term -> unit

(** [renumber_variables_from_offset term offset] renumbers the variables in [term], with
[offset] being the minimum new variable number. *)
val renumber_variables_term_specific_from_offset : int array -> term -> int -> unit


(** [init_renum_table varnums term] works in a way similar to [renumber_variables_term],
but does not actually renumber the variables. Instead, it merely initializes the table
[varnums] with the variables from [term]. *)
val init_renum_table : int array -> term -> unit

val occur_check : int -> context -> term -> context -> bool

(** {6 Built-Ins and Evaluable Arithmetic Functions} *)

(** Function type for a built-in that succeeds at most once. It is performed by
   calling that function, say [f], in the following way: [f term context1
   context2 trail]. [term] in [context1] is the input term. [context2]
   is an auxiliary fresh context with proper multiplier, that may
   be used, and has to be restored by this function when it returns
   [fail]. The result is a trail corresponding to the input trail
   extended by variable bindings performed as side effects in the
   contexts or [fail]. See example implementations in module
   [builtins]. *)
type deterministic_builtin =
   term -> context -> context -> trail -> trail

(** The type of functions which are associated with symbols in predicate role.
   They are used for implementing built-in predicates.
 *)
type p =
    PUnbound
  | PDeterministic of 
   deterministic_builtin *
	(Opt.mode_specifier * Opt.discrimination_values * Opt.proof)

val symbol_predicate : symbol -> p

val set_symbol_predicate : symbol -> p -> unit

(** The type of functions which are associated with symbols in
   functional role. They are used for implementing Prolog-style arithmetic
   evaluation.
*) 
type f = 
    FUnbound 
  | FUnary of (term -> term)
  | FBinary of (term -> term -> term)


val symbol_function : symbol -> f

val set_symbol_function : symbol -> f -> unit

val variant_chk : term -> term -> bool

val variant_chk_1 : term -> int array -> term -> int array -> bool

val variant_chk_bindings : term -> term -> int -> int array -> int array -> int array -> int array -> bool

val binary_flip : term -> term

val np_subsumes_chk_bindings : term -> term -> int -> term array -> int array -> int array -> bool

val np_subsumes_chk_bindings_reset_bindings : int -> term array -> int array -> int array -> unit


val np_subsumes_chk : term -> term -> bool

val make_anonymous_binary_compound : term -> term -> term 

val iter_term : (term -> unit) -> term -> unit