/* ---------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#include "Solver.h"
#include "MiniSatInstantiateClause.h"

#include <stdio.h>
#include <stdlib.h>

#if (__GNUC__ > 2)
#include <ext/hash_map>
using namespace __gnu_cxx;
#else
#include <hash_map>
#endif

typedef unsigned int uint;
typedef void         ExtSolver;

struct SolverWrapper {
  Solver   s;
  vec<Lit> prefix;
  FOClause clause;
  Loc*     loc;
  uint     sig;
  vec<int> args;
  vec<int> dummy;
};

extern "C" {
  SolverWrapper* solver_new(void) { 
    SolverWrapper* s = (SolverWrapper*)malloc(sizeof(SolverWrapper));
    new (s) SolverWrapper();
    return s;
  }

  void     solver_free(SolverWrapper* w)                    { w->~SolverWrapper(); free(w); }

  Loc*     loc_new    (uint a) { Loc* l = (Loc*)malloc(sizeof(Loc)); new (l) Loc(a); return l; }
  void     loc_free   (Loc* l) { free(l); }
  uint     loc_arity  (Loc* l) { return l->getArity(); }

  // build a clause
  void solver_clause_begin       (SolverWrapper* w)                 { w->clause.clear();}
  void solver_clause_add_lit     (SolverWrapper* w, Loc* l, uint s) { w->clause.addLit(l,(bool)s); }
  void solver_clause_add_lit_var (SolverWrapper* w, int v)          {
    w->clause.lastLit().addArg(varArg(v-1));}
  void solver_clause_add_lit_con (SolverWrapper* w, int c)          { 
    w->clause.lastLit().addArg(conArg(c-1));}
  void solver_clause_add_size    (SolverWrapper* w, int s)          { w->clause.addSize(s); }
  uint solver_clause_commit      (SolverWrapper* w, int fresh)      { 
    return w->clause.instantiate(w->s, fresh) ? 1 : 0; }

  // setting a literal to be read or assumed during solving
  void solver_lit_begin   (SolverWrapper* w, Loc* l, uint sign)  { 
    w->args.clear(); w->loc = l; w->sig = sign; }
  void solver_lit_add_con (SolverWrapper* w, int c)              { w->args.push(conArg(c-1)); }

  // reading the value of a predicate (location)
  uint solver_read        (SolverWrapper* w) { 
      // creates variables if they don't exists!! fix this !! 
      int ind = w->loc->get(w->s, w->args, w->dummy);
      //return w->sig ^ w->s.model[ind] ? 0 : 1; }
    return w->sig == (w->s.model[ind] == l_True); }

  uint solver_solve(SolverWrapper* w) {
    vec<Lit>  prefix;
    if(w->loc != NULL) {
      Var v = w->loc->get(w->s, w->args, w->dummy); 
      prefix.push(Lit(v,!w->sig));
    }
    //return w->s.solve(prefix) ? 1 : 0; 
    return w->s.solve(prefix) ? 1 : 0; 
  }

  void solver_dimacs(SolverWrapper* w, char* name) { 
//    ofstream out;
//    out.open(name, ofstream::out);
//    if(!out.is_open()) { cout << "Could not open file " << name << "!\n"; exit(-1); }
//    vector<Lit>  prefix;
//    if(w->loc != NULL) {
//      Var v = w->loc->get(w->s, w->args, w->dummy); 
//      prefix.push_back(lit(v,w->sig));
//    }
//    w->s.export_dimacs(out,w->prefix);
//    out.close();
  }

 uint solver_simplify(SolverWrapper* w) { w->s.simplifyDB(); return w->s.okay(); }


  // statistics
  uint     solver_get_nof_clauses   (SolverWrapper* w)       { return w->s.nClauses(); }
  uint     solver_get_nof_vars      (SolverWrapper* w)       { return w->s.nVars(); }
  uint     solver_get_nof_conflicts (SolverWrapper* w)       { return w->s.stats.conflicts; }

  void     solver_set_verbose       (SolverWrapper* w, uint b) { w->s.verbosity =  b ? 1 : 0; }

  char    *identify (void) { return "MiniSat"; }
}
