------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module ParseProblem where

import System
  ( exitWith
  , ExitCode(..)
  )

import Char
  ( isSpace
  , isAlpha
  , isAlphaNum
  , isDigit
  , isUpper
  )

import List
  ( intersperse
  , (\\)
  , tails
  )

import IO
  ( hFlush
  , stdout
  , try
  )

import System.IO.Error
  ( ioError
  , userError
  )

import Monad
  ( guard
  )

import Data
import Set
import Parsek as P
import Clausify

import Array

-------------------------------------------------------------------------
-- reading

readProblemWithRoots :: [FilePath] -> FilePath -> IO [(String,CType,Clause)]
readProblemWithRoots roots name =
  do putStr ("Reading file '" ++ name ++ "' ... ")
     hFlush stdout
     mes <- findFile roots name
     case mes of
       Nothing ->
         do putStrLn "FILE NOT FOUND"
            exitWith (ExitFailure (-1))

       Just s ->
         case parseP s of
           Left err ->
             do putStr $ unlines $ [ "PARSE ERROR:" ] ++ map ("*** " ++) err
                exitWith (ExitFailure (-1))

           Right (includes,clauses) ->
             do putStrLn "OK"
                hFlush stdout
                sets <- sequence [ readProblemWithRoots roots incl | incl <- includes ]
                return (concat sets ++ clauses)
 where
  findFile []     name =
    do return Nothing
  
  findFile (r:rs) name =
    do ees <- IO.try (readFile (r++name))
       case ees of
         Left _  -> findFile rs name
         Right s -> return (Just s)
  

readProblem :: FilePath -> IO [(String,CType,Clause)]
readProblem name = readProblemWithRoots [""] name

-------------------------------------------------------------------------
-- parsing

type P = Parser Char

-- white space

white :: P ()
white =
  do munch isSpace
     option () $
       do char '%' <?> ""
          many (satisfy (/= '\n')) 
          char '\n'
          white

token :: String -> P String
token s =
  do white
     string s
 <?> show s

pname :: P Symbol
pname =
  do white
     c <- satisfy isAlphaNum
     s <- munch isIdfChar
     return (if (c:s) == "equal" then Equal else name (c:s) ::- Unknown)
 <?> "name"

isIdfChar :: Char -> Bool
isIdfChar c = isAlphaNum c || c == '_'

parens :: P a -> P a
parens = between (token "(") (token ")")

bracks :: P a -> P a
bracks = between (token "[") (token "]")

-- terms

term :: P Term
term =
  do s <- pname
     if isVarSymbol s
       then do return (Var s)
       else do xs <- args
               return (s `app` xs)
 <|>
  do parens term
 <?> "term"

args :: P [Term]
args =
  do return []
 <|>
  do parens (term `sepBy` token ",")
 <?> "arguments"

-- atoms

atom' :: P Atom
atom' =
  do s  <- pname
     xs <- args
     let at = s `prd` xs
     (at == at) `seq` return at
 <?> "atom"

atom :: P Atom
atom =
  do atom'
 <|>
  do parens atom
 <?> "atom"

-- literals

lit :: P Lit
lit =
  do sgn <- (do token "++"; return Pos) <|> (do token "--"; return Neg)
     at  <- atom
     return (sgn at)

-- forms

form :: P Form
form =
  do foper ops
 <?> "formula"    
 where
  ops = [ ("<=>", (<->.))
        , ("<~>", \x y -> nt (x <->. y))
        , ("=>",  (->.))
        , ("<=",  flip (->.))
        , ("|",   (|.))
        , ("~|",  \x y -> nt (x |. y))
        , ("&",   (&.))
        , ("~&",  \x y -> nt (x &. y))
        ]

foper :: [(String, Form->Form->Form)] -> P Form
foper []                   = funit
foper ops@((sym,fun):ops') =
  do a <- foper ops'
     option a $
       do token sym
          b <- foper ops
          return (a `fun` b)

funit :: P Form
funit =
  do parens form
 <|>
  do a <- atom'
     return (Atom a)
 <|>
  do token "~"
     f <- funit
     return (nt f)
 <|>
  do q <- (do token "!"; return All) <|> (do token "?"; return Exi)
     vs <- bracks (pname `sepBy` token ",")
     if any (not . isVarSymbol) vs
       then fail "quantification over non-variable"
       else return ()
     token ":"
     f <- funit
     return (foldr q f vs)
 <?> "formula unit"    

-- clauses

clause :: P (String,CType,Clause)
clause =
  do token "input_clause"
     x <- parens $
       do white
          s <- munch isIdfChar
          token ","
          white
          t <- typ `fmap` munch isIdfChar
          token ","
          ls <- bracks (lit `sepBy` token ",")
          return (s,t,set ls)
     token "."
     return x
 where
  typ "axiom"      = Assumption
  typ "hypothesis" = Assumption
  typ _            = Conjecture

-- clauses

formula :: P (String,CType,Form)
formula =
  do token "input_formula"
     x <- parens $
       do white
          s <- munch isIdfChar
          token ","
          white
          t <- typ `fmap` munch isIdfChar
          token ","
          f <- form
          if size (free f) == 0
            then return (s,t,if t == Conjecture then nt f else f)
            else fail "unbound variable"
     token "."
     return x
 where
  typ "axiom"      = Assumption
  typ "hypothesis" = Assumption
  typ _            = Conjecture

-- includes

include :: P FilePath
include =
  do token "include"
     s <- parens (white >> filePath)
     token "."
     return s

filePath :: P FilePath
filePath =
  do q <- char '\'' <|> char '\"'
     s <- munch (\c -> c /= q && c /= '\n')
     char q
     return s
 <?> "file path"

prob :: P ([FilePath],[(String,CType,Clause)])
prob =
  do incls <- many include
     clses <- do many clause
             <|>
              do fs <- many1 formula
                 let cs = clausify (And [ f | (_,_,f) <- fs ])
                 return [ ("fof",Conjecture,c) | c <- cs ]
     white
     return (incls,clses)
 <|>
  do otter
 <|>
  do qbf

parseP :: String -> Either [String] ([FilePath],[(String,CType,Clause)])
parseP s =
  case parse prob completeResultsWithLine s of
    Left (n, exp, unexp) ->
      Left $
        [ "On line:    " ++ show n ] ++
        [ "Unexpected: " ++ commas "and" unexp | not (null unexp) ] ++
        [ "Expected:   " ++ commas "or" exp    | not (null exp) ]

    Right [x] ->
      Right x

    Right _ ->
      Left $
        [ "Internal error: Ambiguous parse!"
        , "Please report this as a bug in the parser."
        ]
 where
  commas op = concat . intersperse (", " ++ op ++ " ") 

-------------------------------------------------------------------------
-- otter/mace/gandalf syntax

-- this is dirty but does the job

osettings :: P Bool
osettings =
  do bs <- many osetting
     return (or bs)

osetting :: P Bool
osetting =
  do b <- choice [ do token "set" <|> token "clear"
                      parens $
                        do white
                           s <- munch isIdfChar
                           return (s == "prolog_style_variables")
                 , do token "assign"
                      parens $
                        do white
                           munch isIdfChar
                           token ","
                           white
                           munch isIdfChar
                           return False
                 ]
     token "."
     return b

olit :: P Lit
olit =
  do sgn1      <- (do return True) <|> (do token "-"; return False)
     (sgn2,at) <- (do a <- atom; return (True, a)) <|> eqatom
     return (if sgn1 == sgn2 then Pos at else Neg at)
 where
  eqatom =
    do x <- term
       sgn <- (do token "="; return True) <|> (do token "!="; return False)
       y <- term
       return (sgn, Equal `prd` [x,y])
   <|>
    parens eqatom

oclause :: P Clause
oclause =
  do ls <- olit `sepBy` token "|"
     case ls of
       [Pos (Pred _ (MkName _ (Extern "end_of_list") ::- _) _)] -> fail ""
       _                                                        -> return ()
     token "."
     return (set ls)

oblock :: P [(String,CType,Clause)]
oblock =
  do token "list"
     s <- parens (token "sos" <|> token "usable")
     token "."
     cs <- many oclause
     token "end_of_list"
     token "."
     return [ ("", typ s, c) | c <- cs ]
 where
  typ "sos" = Conjecture
  typ _     = Assumption

otter :: P ([FilePath],[(String,CType,Clause)])
otter =
  do b <- osettings
     if not b then fail "no 'set(prolog_style_variables).'" else return ()
     blocks <- many oblock
     white
     return ([],concat blocks)

-------------------------------------------------------------------------
-- qbf

number :: P Int
number =
  do white
     c <- satisfy (\c -> '1' <= c && c <= '9')
     s <- munch isDigit 
     return (read (c:s))
 <?> show "number"

qbf :: P ([FilePath],[(String,CType,Clause)])
qbf =
  do -- header
     many (do token "c"
              many (satisfy (/= '\n')) 
              char '\n')

     token "p"
     token "cnf"
     n <- number
     number
     
     -- quantifiers
     qs <- many (do [q] <- token "a" <|> do token "e"
                    xs  <- many number
                    token "0"
                    return (q, xs))

     let quant xs [] = []
         
         quant xs (('a',ys):qs) =
           (ys `zip` [ p (Var v) | v <- vs ]) ++ quant (xs ++ vs) qs
          where
           vs = [ name "A" % i ::- Unknown | i <- ys ]
         
         quant xs (('e',ys):qs) =
           (ys `zip` ts) ++ quant xs qs
          where
           ts = [ (name "q" % i ::- Unknown) `prd` [ Var x | x <- xs ]
                | i <- ys
                ]
         
         err   = error "variable not defined"
         table = accumArray (\_ x -> x) err (1,n) (quant [] qs)

     -- clauses
     cs <- many (do sxs <- many (do s <- do token "-"; return Neg
                                     <|> do return Pos
                                    x <- number
                                    return (s,x))
                    token "0"
                    return [ s (table ! x)
                           | (s,x) <- sxs
                           ])
     
     -- clauses
     let cs' =
           [ set [ Var x =. false, Var x =. true ]
           , set [ Neg (p false) ]
           ] ++
           [ set c
           | c <- cs
           ]
     white
     return ([],[ ("qbf",Conjecture,c) | c <- cs' ])
 where
  true   = (name' "true"  ::- Unknown) `app` []
  false  = (name' "false" ::- Unknown) `app` []
  x =. y = Pos (Equal `prd` [x,y])
  x      = name' "X" ::- Unknown
  p x    = Equal `prd` [x,true]

{-
qbf' :: P ([FilePath],[(String,CType,Clause)])
qbf' =
  do -- header
     many (do token "c"
              many (satisfy (/= '\n')) 
              char '\n')

     token "p"
     token "cnf"
     number
     number
     
     -- quantifiers
     qs <- many (do q <- do token "a"; return All
                     <|> do token "e"; return Exi
                    xs <- many number
                    token "0"
                    return (foldr (.) id [ q (name x ::- Unknown) | x <- xs]))

     -- clauses
     cs <- many (do sxs <- many (do s <- do token "-"; return Not
                                     <|> do return id
                                    x <- number
                                    return (s,x))
                    token "0"
                    return [ s (Atom (p (Var (name x ::- Unknown))))
                           | (s,x) <- sxs
                           ])
     
     -- formula
     let f  = And [ All x (Or [ Var x =. false, Var x =. true])
                  , Not (Atom (p false))
                  , Atom (p true)
                  , foldr (.) id qs (And [ Or c | c <- cs ])
                  ]

     white
     return ([],[ ("qbf",Conjecture,c) | c <- clausify f ])
 where
  true   = (name' "true"  ::- Unknown) `app` []
  false  = (name' "false" ::- Unknown) `app` []
  x =. y = Atom (Equal `prd` [x,y])
  x      = name' "x" ::- Unknown
  p x    = (name' "p" ::- Unknown) `prd` [x]
-}
 
-------------------------------------------------------------------------
-- the end.

