/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef ALLOCATION_FAILURE_HANDLING_HPP
#define ALLOCATION_FAILURE_HANDLING_HPP
#include "jargon.hpp"
#include "Exit.hpp"
#include "RuntimeError.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "Debug.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOCATION_FAILURE_HANDLING
#  define DEBUG_NAMESPACE "AllocationFailureHandling<InstanceId>"
#endif 
#include "debugMacros.hpp"
namespace BK 
{
  template <class InstanceId> 
  class AllocationFailureHandling
  {
  public: 
    class Activator
    {
    public:
      Activator(AllocationFailureHandling* handler)
      {
	CALL("Activator::constructor Activator(AllocationFailureHandling* handler)");
	_handler = handler;
	_handler->activate();
      };
      ~Activator() 
      {
	CALL("Activator::destructor ~Activator()"); 
	if (DestructionMode::isThorough())
	  _handler->deactivate();
	BK_CORRUPT(*this);
      };
   private: 
    Activator();
    AllocationFailureHandling* _handler;
  };
 public:
  AllocationFailureHandling() :
    _memoryReleaseRequest(0),
    _massiveMemoryReleaseRequest(0),
    _inheritUnrecoverableAllocationFailureHandler(false),
    _unrecoverableAllocationFailureHandler(0),
    _previousHandler(0)
  {
  };
  ~AllocationFailureHandling()
  {
    BK_CORRUPT(*this);
  };
  void activate() 
  {
   CALL("activate()");
   if (_previousHandler)
    {
     RuntimeError::report("activating allocation failure handler which is active");
    }
   else
    {
     _previousHandler = currentHandler();
     currentHandler() = this;
    };
  };
  void deactivate()
  {
   CALL("deactivate()");
   currentHandler() = _previousHandler;
   _previousHandler = 0;
  };
  void setMemoryReleaseRequest(bool (*memoryReleaseRequest)(ulong size)) 
  {
   _memoryReleaseRequest = memoryReleaseRequest;
  };
  void setMassiveMemoryReleaseRequest(bool (*massiveMemoryReleaseRequest)()) 
  {
   _massiveMemoryReleaseRequest = massiveMemoryReleaseRequest;
  };
  bool& inheritUnrecoverableAllocationFailureHandler() 
  { 
   return _inheritUnrecoverableAllocationFailureHandler;
  };
  bool inheritUnrecoverableAllocationFailureHandler() const
  { 
   return _inheritUnrecoverableAllocationFailureHandler;
  };
  void setUnrecoverableAllocationFailureHandler(void (*unrecoverableAllocationFailureHandler)(ulong size)) 
  {
   _unrecoverableAllocationFailureHandler = unrecoverableAllocationFailureHandler;
  };
  static bool requestMemoryRelease(ulong size) 
  {
   CALL("requestMemoryRelease(ulong size)");
   return currentHandler()->callMemoryRelease(size);
  };
  static bool requestMassiveMemoryRelease() 
  {
   CALL("requestMassiveMemoryRelease()");
   return currentHandler()->callMassiveMemoryRelease();
  };
  static void reportUnrecoverableAllocationFailure(ulong size) 
  {
   CALL("reportUnrecoverableAllocationFailure(ulong size)");
   currentHandler()->reportUltimateFailure(size);
  };
 private: 
  static AllocationFailureHandling*& currentHandler()
  {
    static AllocationFailureHandling* _currentHandler = 0;
    return _currentHandler;
  };
  bool callMemoryRelease(ulong size)
  {
    return this && _memoryReleaseRequest && _memoryReleaseRequest(size);         
  };
  bool callMassiveMemoryRelease()
  {
   return 
     this && _massiveMemoryReleaseRequest && _massiveMemoryReleaseRequest();
  };
  void reportUltimateFailure(ulong size)
  {
   if (this)
    {
     if (_inheritUnrecoverableAllocationFailureHandler)
      {
       _previousHandler->reportUltimateFailure(size);
      }
     else
      if (_unrecoverableAllocationFailureHandler)
       _unrecoverableAllocationFailureHandler(size); 
    };
   RuntimeError::report("unhandled unrecoverable allocation failure");
  };
 private:
  bool (*_memoryReleaseRequest)(ulong size);
  bool (*_massiveMemoryReleaseRequest)();
  bool _inheritUnrecoverableAllocationFailureHandler;
  void (*_unrecoverableAllocationFailureHandler)(ulong size);
  AllocationFailureHandling* _previousHandler;
}; 
}; 
#endif
