/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef NO_DEBUG
#ifndef DEBUG_HPP
#define DEBUG_HPP
#include <iostream>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
using namespace std;
namespace BK 
{
class Debug
{
 public:
  class InitStatic
    {
    public: 
      InitStatic();
      ~InitStatic();
    private:
      static long _count;
    }; 
  class Call;
  class OpenDataBracket;
  class CloseDataBracket;
  class DummyOpenDataBracket;
  class DummyCloseDataBracket;
  template <unsigned long DefaultLimit>
  class LimitedCounter
  {
  public:
    LimitedCounter() : _limit(DefaultLimit), _accumulated(0UL) {};
    LimitedCounter(unsigned long lim) : _limit(lim), _accumulated(0UL) {};
    void reset() { _accumulated = 0UL; };
    void reset(unsigned long lim) { _limit = lim; _accumulated = 0UL; };
    bool overflow() const { return _accumulated > _limit; };
    void inc() { ++_accumulated; };
  private:
    unsigned long _limit;
    unsigned long _accumulated;
  };
 public:  
  static void assignJob(const char* jobId) { job = jobId; }; 
  static void setTracing(ostream* str,bool traceCOPs,bool traceAfterErr) 
    { 
      trace = str; 
      traceCP = traceCOPs; 
      traceAfterError = traceAfterErr;
    };
  static void tracingOff() { trace = 0; };
  static void setNextEventHook(void (*fun)(ulong))
    {
      nextEventHook = fun;
    };
  static void setErrorHook(void (*fun)())
    {
      errorHook = fun;
    };
  static void abort() { errorFlag = true; };
  static void cp(const char* cop);
  static void error(std::ostream& str);
  static void error(std::ostream& str,const char* msg);
  static void debugFeatureInternalFault(std::ostream& str,const char* msg);
  static void reportCP(std::ostream& str);
  static void history(std::ostream& str);
  static ulong event() { return eventId; };         
 private:
  Debug() {};       
  ~Debug() {}; 
  static void nextEvent() 
    { 
      eventId++; 
      if (nextEventHook) nextEventHook(eventId); 
    }; 
  static void initSigHandlers();
  static void restoreSigHandlers();
  static void reportCP(std::ostream& str,
		       const char* cn,
		       const char* cp,
		       const char* on); 
  static void handleSIGFPE(int);
  static void handleSIGILL(int);
  static void handleSIGSEGV(int);
  static void handleSIGBUS(int);
  static void handleSIGABRT(int);
  static void handleSIGTRAP(int);
 private:
  static const char* job;  
  static ulong eventId;
  static Call* currentCall;
  static Call* bottomCall;
  static long callDepth;
  static const char* errorDescription;
  static ostream* trace;
  static bool traceCP;
  static bool traceAfterError;
  static void (*nextEventHook)(ulong eventNum);
  static void (*errorHook)();
  static bool errorFlag;
  friend class Call;
  friend class InitStatic;
  friend class OpenDataBracket;
  friend class CloseDataBracket;
}; 
class Debug::Call
{
public:
  Call(const char* callNm);
  Call(const char* callNm,const char* objName);
  ~Call();
private:
  Call() {};
private:
  const char* _callName;
  const char* _objName;
  const char* _controlPoint;
  OpenDataBracket* _dataBrackets;
  Call* _outerCall;
  friend class Debug;
  friend class Debug::InitStatic;
  friend class Debug::OpenDataBracket;
  friend class Debug::CloseDataBracket;
}; 
class Debug::OpenDataBracket
{
public:
  enum Status
  {
    InitOpen,
    InitClose,
    DestrOpen,
    DestrClose
  };
public:
  OpenDataBracket(const char* name);
  ~OpenDataBracket();
private:
  const char* _name;
  OpenDataBracket* _outerBracket;
  Status _status;
  friend class Debug;
  friend class Debug::Call;
  friend class Debug::CloseDataBracket;
}; 
class Debug::CloseDataBracket
{
public:
  CloseDataBracket();
  ~CloseDataBracket();
private:
  OpenDataBracket* _match;
  friend class Debug;
  friend class Debug::Call;
}; 
class Debug::DummyOpenDataBracket
{
public:
  DummyOpenDataBracket() {};
  ~DummyOpenDataBracket() {};
}; 
class Debug::DummyCloseDataBracket
{
public:
  DummyCloseDataBracket() {};
  ~DummyCloseDataBracket() {};
}; 
inline void Debug::cp(const char* cop) 
{    
  if (!errorFlag)
    {
      currentCall->_controlPoint = cop;
      nextEvent();
      if (trace && traceCP) 
	*trace << eventId << " COP " << cop << '\n' << flush; 
    };
};
inline
Debug::Call::Call(const char* callNm) : 
  _callName(callNm),
  _objName("?"),
  _controlPoint("start"),
  _dataBrackets(0),
  _outerCall(Debug::currentCall)
{
  if (!Debug::errorFlag)
    {
      callDepth++;
      Debug::currentCall = this;
      nextEvent();
      if (trace)
	{
	  *trace << eventId << " -> [" << callDepth << "] " << callNm 
		 << " [?]\n" << flush;
	};
    };
}; 
inline
Debug::Call::Call(const char* callNm,const char* objName) : 
  _callName(callNm),
  _objName(objName),
  _controlPoint("start"),
  _dataBrackets(0),
  _outerCall(Debug::currentCall)
{
  if (!Debug::errorFlag)
    {
      callDepth++;
      Debug::currentCall = this;
      nextEvent();
      if (trace)
	{
	  *trace << eventId << " -> [" << callDepth << "] " << callNm 
		 << " [" << objName << "]\n" << flush;
	};
    };
}; 
inline
Debug::Call::~Call()
{
  if (!Debug::errorFlag)
    {
      if (_dataBrackets)
	{
	  switch (_dataBrackets->_status)
	    {
	    case OpenDataBracket::InitOpen:
	      Debug::error((trace)? (*trace) : nullStream,
			   "Data bracket mismatch: missing closing bracket");
	    case OpenDataBracket::InitClose:
	      Debug::debugFeatureInternalFault((trace)? (*trace) : nullStream,
					       "Return from call with a persistent data bracket with InitClose status");
	    case OpenDataBracket::DestrOpen:
	      Debug::error((trace)? (*trace) : nullStream,
			   "Data bracket mismatch: missing opening bracket");
	    case OpenDataBracket::DestrClose:
	      Debug::debugFeatureInternalFault((trace)? (*trace) : nullStream,
					       "Return from call with a persistent data bracket with DestrClose status");
	    };	  
	};
      nextEvent();
      if (trace)
	{
	  *trace << eventId  << " <- [" << callDepth << "] " << _callName  << '\n' << flush; 
	};
      Debug::currentCall = _outerCall;
      callDepth--;
    };
}; 
inline
Debug::OpenDataBracket::OpenDataBracket(const char* name) :
  _name(name),
  _outerBracket(Debug::currentCall->_dataBrackets),
  _status(InitOpen)
{
  if (!Debug::errorFlag)
    {
      callDepth++;
      Debug::currentCall->_dataBrackets = this;
      nextEvent();  
      if (trace)
	{
	  *trace << eventId << " -> [" << callDepth << "] " 
		 << "INIT BRACKET " << name << '\n' << flush;
	};
    };
}; 
inline
Debug::OpenDataBracket::~OpenDataBracket()
{   
  if (!Debug::errorFlag)
    {
      _status = DestrClose;
      nextEvent();      
      if (trace)
	{
	  *trace << eventId  << " <- [" << callDepth << "] " 
		 << " DESTR BRACKET " << _name << '\n' << flush;
	};
      Debug::currentCall->_dataBrackets = _outerBracket;
      callDepth--;
    };
}; 
inline
Debug::CloseDataBracket::CloseDataBracket()
{
  if (!Debug::errorFlag)
    {
      _match = Debug::currentCall->_dataBrackets;
      if ((!_match) || 
	  _match->_status != OpenDataBracket::InitOpen)
	{
	  Debug::error((trace)? (*trace) : nullStream,
		       "Data bracket mismatch: missing opening bracket");
	}
      else
	{
	  _match->_status = OpenDataBracket::InitClose;
	  nextEvent();
	  if (trace)
	    {
	      *trace << eventId  << " <- [" << callDepth << "] " 
		     << " INIT BRACKET " << _match->_name << '\n' << flush; 
	    };
	  Debug::currentCall->_dataBrackets = _match->_outerBracket;
	  callDepth--; 
	};
    };
}; 
inline
Debug::CloseDataBracket::~CloseDataBracket()
{    
  if (!Debug::errorFlag)
    {
      callDepth++;
      _match->_outerBracket = Debug::currentCall->_dataBrackets;
      Debug::currentCall->_dataBrackets = _match;
      _match->_status = OpenDataBracket::DestrOpen;
      nextEvent(); 
      if (trace)
	{
	  *trace << eventId << " -> [" << callDepth << "] " 
		 << "DESTR BRACKET " << _match->_name 
		 << '\n' << flush;
	};
    };
}; 
static Debug::InitStatic debugInitStatic;
}; 
#endif
#endif
