/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef EXPANDING_STACK_HPP
#define EXPANDING_STACK_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "Array.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_EXPANDING_STACK
# define DEBUG_NAMESPACE "ExpandingStack<Alloc,ElType,ElRelocation,MinIncrementSize>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class Alloc,class ElType,class ElRelocation,ulong MinIncrementSize>
class ExpandingStack
{
 public:
  ExpandingStack() 
   : _memory(DOP("ExpandingStack<..>::_memory"))
  {
   CALL("constructor ExpandingStack()"); 
   _cursor = _memory.memory();
   #ifdef DEBUG_NAMESPACE
    _name = "?";
   #endif
  };
  ExpandingStack(const char* name) 
   : _memory(DOP("ExpandingStack<..>::_memory"))
  {
   CALLM(name,"constructor ExpandingStack(const char* name)");
   _cursor = _memory.memory();
   #ifdef DEBUG_NAMESPACE
    _name = name;
   #endif
  };
  ExpandingStack(const ExpandingStack& s)
   : _memory(s._memory)
     #ifdef DEBUG_NAMESPACE
      , _name(s._name)
     #endif
  {
   CALLM(_name,"constructor ExpandingStack(const ExpandingStack& s)");
   _cursor = _memory.memory() + (s._cursor - s._memory.memory());
  };  
  ~ExpandingStack() 
  {
    CALLM(_name,"destructor ~ExpandingStack()");
    BK_CORRUPT(_cursor);
  };
  void relocateTo(void* target)
  {
    CALLM(_name,"relocateTo(void* target)");
    long cusorOffset = _cursor - _memory.memory();     
    _memory.relocateTo(static_cast<void*>(&(static_cast<ExpandingStack*>(target)->_memory)));
    static_cast<ExpandingStack*>(target)->_cursor = 
      static_cast<ExpandingStack*>(target)->_memory.memory() + cusorOffset;    
#ifdef DEBUG_NAMESPACE
    static_cast<ExpandingStack*>(target)->_name = _name;
#endif
    BK_CORRUPT(*this);
  }; 
  void reset()
  {
   CALLM(_name,"reset()");
   _cursor = _memory.memory();
  }; 
  operator bool() const
  {
   CALLM(_name,"operator bool() const");
   return _cursor != _memory.memory();   
  };
  bool empty() const
  {
   CALLM(_name,"bool empty() const");
   return _cursor == _memory.memory();  
  };
  bool nonempty() const
  {
   CALLM(_name,"nonempty() const"); 
   return _cursor != _memory.memory(); 
  };
  ulong size() const
  {
   CALLM(_name,"size() const");
   return _cursor - _memory.memory();
  };
  ElType* begin() { return _memory.memory(); };
  ElType* end() { return _cursor; };
  ElType* last() 
  { 
    CALL("last()");
    ASSERT(_memory.memory());
    return _cursor - 1; 
  };
  const ElType* begin() const { return _memory.memory(); }; 
  const ElType* end() const { return _cursor; }; 
  const ElType* last() const { return _cursor - 1; };
  ElType& top()
  {
   CALLM(_name,"ElType& top()");
   return *(_cursor - 1);
  }; 
  const ElType& top() const
  {
   CALLM(_name,"const ElType& top() const");
   return *(_cursor - 1);
  };
  void push(const ElType& el)  
  {
   CALLM(_name,"push(const ElType& el)");
   ASSERT(_cursor < _memory.endOfMemory());
   *_cursor = el;
   _cursor++;
  };
  void pushSafe(const ElType& el)
  {
   CALLM(_name,"pushSafe(const ElType& el)");
   if (_cursor >= _memory.endOfMemory()) 
    {
     ulong sz = _cursor - _memory.memory(); 
     _memory.expand();
     _cursor = _memory.memory() + sz;
    };
   ASSERT(_cursor < _memory.endOfMemory());
   *_cursor = el;
   _cursor++;
  };
  const ElType& pop()
  {
   CALLM(_name,"pop()");
   ASSERT(nonempty());
   _cursor--;
   return *_cursor;
  };
  void skipTop()
  {
   CALLM(_name,"skipTop()");
   ASSERT(nonempty());
   _cursor--;
  };
  const ElType& nth(ulong n) const 
  {
   CALLM(_name,"nth(ulong n) const");
   ASSERT(n < _memory.size());
   return _memory[n];
  };
  ElType& nth(ulong n)  
  {
   CALLM(_name,"nth(ulong n)");
   ASSERT(n < _memory.size());
   return _memory[n];
  };
  ElType& nthSafe(ulong n)
  {
   CALLM(_name,"nthSafe(ulong n)");
   if (n >= _memory.size()) 
    {
     ulong sz = _cursor - _memory.memory(); 
     _memory.expand(n);
     _cursor = _memory.memory() + sz;
    };
   return _memory[n];  
  };
  void expand(ulong n)
  {
   CALLM(_name,"expand(ulong n)");
   ulong sz = _cursor - _memory.memory(); 
   _memory.expand(n);
   _cursor = _memory.memory() + sz;
  };
  bool tryToExpand(ulong n)
  {
   CALLM(_name,"tryToExpand(ulong n)");
   ulong sz = _cursor - _memory.memory(); 
   if (_memory.tryToExpand(n))
     {
       _cursor = _memory.memory() + sz;
       return true;
     };
   return false;
  };
  void expand()
  {
   CALLM(_name,"expand()");
   ulong sz = _cursor - _memory.memory(); 
   _memory.expand();
   _cursor = _memory.memory() + sz;
  };
  #ifndef NO_DEBUG
   void freeze() { _memory.freeze(); };
   void unfreeze() { _memory.unfreeze(); };
  #endif  
 private: 
  typedef Array<Alloc,ElType,ElRelocation,MinIncrementSize> Array;
 private:
  Array _memory;
  ElType* _cursor;
  #ifdef DEBUG_NAMESPACE
   const char* _name;
  #endif
}; 
}; 
#endif
