/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef OBJECT_CACHE_HPP
#define OBJECT_CACHE_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_CACHE
#  define DEBUG_NAMESPACE "ObjectCache<Alloc,PageSize,C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
  template <class Alloc,ulong PageSize,class C>
  class ObjectCache 
  {
  public:
    ObjectCache();
    ~ObjectCache();
    void reset() { reclaimAllObjects(); };
    C* reserveObject();
    C* tryToReserveObject();
    void reclaimAllObjects();
  private:
    class Page
    {
    public:
      Page() : _next(0) {};
      ~Page() {};
      Page*& next() { return _next; };
      C* objects() { return _objects; };
      C* endOfObjects() { return _objects + PageSize; };
      void* operator new(size_t) 
      {
	CALL("Page::operator new(size_t)");
	return Alloc::allocate(sizeof(Page));
      };
      void* operator new(size_t,bool) 
      {
	CALL("Page::operator new(size_t,bool)");
	return Alloc::tryToAllocate(sizeof(Page));
      };
      void operator delete(void* obj) 
      {
	CALL("Page::operator delete(void* obj)");
	Alloc::deallocate(obj,sizeof(Page));
      };
    private:
      Page* _next;
      C _objects[PageSize];
    }; 
  private:
    Page _mainPage;
    Page* _currentPage;
    C* _nextAvailableObject;
    C* _endOfCurrentPage;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_CACHE
#  define DEBUG_NAMESPACE "ObjectCache<Alloc,PageSize,C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
  template <class Alloc,ulong PageSize,class C>
  inline
  ObjectCache<Alloc,PageSize,C>::ObjectCache() :
    _currentPage(&_mainPage),
    _nextAvailableObject(_mainPage.objects()),
    _endOfCurrentPage(_mainPage.endOfObjects())
  {
    CALL("constructor ObjectCache()");
  };
  template <class Alloc,ulong PageSize,class C>
  inline
  ObjectCache<Alloc,PageSize,C>::~ObjectCache()
  {
    CALL("destructor ~ObjectCache()");
    if (DestructionMode::isThorough())
      {
	Page* p1 = _mainPage.next();
	while (p1)
	  {
	    Page* p2 = p1->next();
	    delete p1;
	    p1 = p2;
	  };
      };
  };
  template <class Alloc,ulong PageSize,class C>
  inline
  C* ObjectCache<Alloc,PageSize,C>::reserveObject()
  {
    CALL("reserveObject()");
    if (_nextAvailableObject == _endOfCurrentPage)
      {
	if (!_currentPage->next())
	  _currentPage->next() = new Page();
	_currentPage = _currentPage->next();
	_nextAvailableObject = _currentPage->objects();
	_endOfCurrentPage = _currentPage->endOfObjects();
      };
    C* res = _nextAvailableObject;
    ++_nextAvailableObject;
    return res;
  };
  template <class Alloc,ulong PageSize,class C>
  inline
  C* ObjectCache<Alloc,PageSize,C>::tryToReserveObject()
  {
    CALL("tryToReserveObject()");
    if (_nextAvailableObject == _endOfCurrentPage)
      {
	if (!_currentPage->next())
	  {
	    Page* newPage = new(true) Page();
	    if (!newPage) return static_cast<C*>(0);
	    _currentPage->next() = newPage;
	  };
	_currentPage = _currentPage->next();
	_nextAvailableObject = _currentPage->objects();
	_endOfCurrentPage = _currentPage->endOfObjects();
      };
    C* res = _nextAvailableObject;
    ++_nextAvailableObject;
    return res;
  };
  template <class Alloc,ulong PageSize,class C>
  inline
  void ObjectCache<Alloc,PageSize,C>::reclaimAllObjects()
  {
    CALL("reclaimAllObjects()");
    _currentPage = &_mainPage;
    _nextAvailableObject = _mainPage.objects();
    _endOfCurrentPage = _mainPage.endOfObjects();
  };
}; 
#endif
