/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "InferenceInfo.hpp"
#include "Inference.hpp"
#include "XML.hpp"
#include "Tracer.hpp"
#include "DestructionMode.hpp"
#include "VampireException.hpp"
using namespace Shell;
InferenceInfo::InferenceInfo (InferenceRule rule, 
			      const Position& position, 
			      InferenceInfo* next)
  : _rule(rule),
    _position(position),
    _next(next)
{
} 
InferenceInfo::InferenceInfo (InferenceRule rule, 
			      InferenceInfo* next)
  : _rule(rule),
    _next(next)
{
} 
void InferenceInfo::destroy ()
{
  TRACER("InferenceInfo::destroy");
  if (DestructionMode::isFast()) {
    return;
  }
  switch (rule()) {
  case IR_SKOLEMIZE:
  case IR_CHOICE_AXIOM:
  case IR_FORALL_ELIMINATION:
  case IR_FORALL_OR:
  case IR_EXISTS_AND:
  case IR_DUMMY_QUANTIFIER_REMOVAL:
    delete static_cast<InferenceInfoTermList*>(this);
    return;
  case IR_KERNEL:
    delete static_cast<KernelInferenceInfo*>(this);
    return;
  default: 
    delete this;
    return;
  }
} 
void InferenceInfo::destroyList ()
{
  TRACER("InferenceInfo::destroyList");
  if (DestructionMode::isFast()) {
    return;
  }
  InferenceInfo* inf = this;
  while (inf) {
    InferenceInfo* nxt = inf->next();
    inf->destroy();
    inf = nxt;
  }
} 
InferenceInfoTermList::InferenceInfoTermList (InferenceRule rule,
					      const Position& position, 
					      InferenceInfo* next,
					      const TermList& terms)
  : InferenceInfo(rule,position,next),
    _terms(terms)
{
  ASS(rule == IR_SKOLEMIZE ||
      rule == IR_CHOICE_AXIOM ||
      rule == IR_FORALL_ELIMINATION ||
      rule == IR_FORALL_OR ||
      rule == IR_EXISTS_AND || 
      rule == IR_DUMMY_QUANTIFIER_REMOVAL);
} 
XML::Element* InferenceInfo::toXML () const
{
  TRACER("InferenceInfo::toXML");
  XML::DeepElement* xml = new XML::DeepElement("inference_info");
  xml->addAttribute(new XML::TextAttribute("rule",
					   Inference::toString(rule())));
  xml->addChild(new XML::TextElement("position",position().toString()));
  switch (rule()) {
  case IR_SKOLEMIZE:
  case IR_CHOICE_AXIOM:
  case IR_FORALL_ELIMINATION:
  case IR_FORALL_OR:
  case IR_EXISTS_AND:
  case IR_DUMMY_QUANTIFIER_REMOVAL:
    {
      const InferenceInfoTermList* inf = static_cast<const InferenceInfoTermList*>(this);
      XML::DeepElement* ts = new XML::DeepElement("terms");
      xml->addChild(ts);
      Iterator<Term> tts(inf->terms());
      while (tts.hasNext()) {
	ts->addChild(new XML::TextElement("term",tts.next().toString()));
      }
    }
    break;
  default: 
    break;
  }
  return xml;
} 
string InferenceInfo::toString () const
{
  TRACER("InferenceInfo::toString");
  string result = (string)"inference_info (rule=" + 
                  Inference::toString(rule()) +
                  ", position=" +
                  position().toString();
  switch (rule()) {
  case IR_SKOLEMIZE:
  case IR_CHOICE_AXIOM:
  case IR_FORALL_ELIMINATION:
  case IR_FORALL_OR:
  case IR_EXISTS_AND:
  case IR_DUMMY_QUANTIFIER_REMOVAL:
    {
      const InferenceInfoTermList* inf = static_cast<const InferenceInfoTermList*>(this);
      result += (string)", terms=" + inf->terms().toString(",");
    }
    break;
  default: 
    break;
  }
  return result + ")";
} 
KernelInferenceInfo::KernelInferenceInfo (VampireKernel::Clause::BackgroundFlagIterator& flags)
  : InferenceInfo(IR_KERNEL,0),
    _inferenceCodes(0)
{
  VampireKernel::Clause::BackgroundFlag flag;
  while (flags.getCurrentFlag(flag)) {
    switch (flag) {
    case VampireKernel::Clause::IsInputClause:
      _inferenceCodes |= KR_IsInput;
      break;
    case VampireKernel::Clause::IsObtainedDuringPreprocessing:
      _inferenceCodes |= KR_Preprocessing;
      break;
    case VampireKernel::Clause::BinaryResolutionUsed:
      _inferenceCodes |= KR_BinaryResolution;
      break;
    case VampireKernel::Clause::ForwardTransitiveChainingUsed:
      _inferenceCodes |= KR_ForwardTransitiveChaining;
      break;
    case VampireKernel::Clause::BackwardTransitiveChainingUsed:
      _inferenceCodes |= KR_BackwardTransitiveChaining;
      break;
    case VampireKernel::Clause::ForwardSuperpositionUsed:
      _inferenceCodes |= KR_ForwardSuperposition;
      break;
    case VampireKernel::Clause::BackwardSuperpositionUsed:
      _inferenceCodes |= KR_BackwardSuperposition;
      break;
    case VampireKernel::Clause::EqualityResolutionUsed:
      _inferenceCodes |= KR_EqualityResolution;
      break;
    case VampireKernel::Clause::EqualityFactoringUsed:
      _inferenceCodes |= KR_EqualityFactoring;
      break;
    case VampireKernel::Clause::ForwardDemodulationUsed:
      _inferenceCodes |= KR_ForwardDemodulation;
      break;
    case VampireKernel::Clause::BackwardDemodulationUsed:
      _inferenceCodes |= KR_BackwardDemodulation;
      break;
    case VampireKernel::Clause::SimplificationByEqualityResolutionUsed:
      _inferenceCodes |= KR_SimplificationByEqualityResolution;
      break;
    case VampireKernel::Clause::ForwardSubsumptionResolutionUsed:
      _inferenceCodes |= KR_ForwardSubsumptionResolution;
      break;
    case VampireKernel::Clause::BackwardSubsumptionResolutionUsed:
      _inferenceCodes |= KR_BackwardSubsumptionResolution;
      break;
    case VampireKernel::Clause::IsReanimatedPassive:
      _inferenceCodes |= KR_IsReanimatedPassive;
      break;
    case VampireKernel::Clause::ExternalEvaluationUsed:
      _inferenceCodes |= KR_Evaluation;
      break;
    case VampireKernel::Clause::IsNameIntro:
      _inferenceCodes |= KR_IsNameIntro;
      break;
    case VampireKernel::Clause::FoldingUsed:
      _inferenceCodes |= KR_Folding;
      break;
    case VampireKernel::Clause::ClausificationUsed:
      _inferenceCodes |= KR_Clausification;
      break;
#if DEBUG_SHELL
    default:
      ASS(false);
#endif
    }
    flags.goToNextFlag();
  }
} 
string KernelInferenceInfo::rulesAsString () const
{
  string result = "";
  unsigned infCodes = _inferenceCodes;
  if ((infCodes & KR_Preprocessing) && infCodes != KR_Preprocessing) {
    infCodes ^= KR_Preprocessing;
  }
  for (int code = 1; code <= KR_Clausification; code *= 2) {
    if (infCodes & code) {
      if (result != "") {
	result += ", ";
      }
      result += toString((KernelRule)code);
    }
  }
  return result;
} 
string KernelInferenceInfo::toString (KernelRule rule)
{
  switch (rule) {
  case KR_IsInput:
    return "input";
  case KR_Preprocessing:
    return "literal permutation";
  case KR_BinaryResolution:
    return "resolution";
  case KR_ForwardTransitiveChaining:
    return "forward transitivity chaining";
  case KR_BackwardTransitiveChaining:
    return "backward transitivity chaining";
  case KR_ForwardSuperposition:
    return "forward superposition";
  case KR_BackwardSuperposition:
    return "backward superposition";
  case KR_EqualityResolution:
  case KR_SimplificationByEqualityResolution:
    return "equality resolution";
  case KR_EqualityFactoring:
    return "equality factoring";
  case KR_ForwardDemodulation:
    return "forward demodulation";
  case KR_BackwardDemodulation:
    return "backward demodulation";
  case KR_ForwardSubsumptionResolution:
    return "forward subsumption resolution";
  case KR_BackwardSubsumptionResolution:
    return "backward subsumption resolution";
  case KR_IsReanimatedPassive:
    return "passive clause reanimation";
  case KR_Evaluation:
    return "evaluation";
  case KR_IsNameIntro:
    return "definition introduction";
  case KR_Folding:
    return "definition folding";
  case KR_Clausification:
    return "cnf transformation";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
