/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cmath>
#include <cstdlib>
#include "Int.hpp"
#include "InterpretedConstant.hpp"
#include "Random.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "VampireKernelEnvironment.hpp"
using namespace VampireKernelEnvironment;
using namespace std;
using namespace Shell;
InterpretedConstantValue* 
InterpretedConstantValue::_args[MAX_ARITY_FOR_INTERPRETED_SYMBOLS];
int InterpretedConstantValue::_nextArgumentIndex = 0;
InterpretedConstant::~InterpretedConstant () {}
InterpretedConstant::InterpretedConstant (void* implementation) 
  : _implementation(implementation)
{
  ASS(implementation);
} 
void InterpretedConstant::operator = (const InterpretedConstant& t)
{
  _implementation = t._implementation;
} 
string InterpretedConstant::toString () const
{
  ASS(_implementation);
  return (static_cast<InterpretedConstantValue*>(_implementation))->toString();
} 
int InterpretedConstant::compare (const InterpretedConstant& c) const
{
  TRACER("InterpretedConstant::compare");
  ASS(_implementation && c._implementation);
  InterpretedConstantValue* v1 = static_cast<InterpretedConstantValue*>(_implementation);
  InterpretedConstantValue* v2 = static_cast<InterpretedConstantValue*>(c._implementation);
  return v1->compare(v2);
} 
int InterpretedConstantValue::compare (const InterpretedConstantValue* v) const
{
  TRACER("InterpretedConstantValue::compare");
  ICSort sort = tag();
  int cmp = Int::compare(sort,v->tag());
  if (cmp != 0) {
    return cmp;
  }
  switch (sort) { 
  case IC_INTEGER:
    return Int::compare(static_cast<const IntegerConstantValue*>(this)->value(),
			static_cast<const IntegerConstantValue*>(v)->value());
  case IC_REAL:
    return Int::compare(static_cast<const RealConstantValue*>(this)->value(),
			static_cast<const RealConstantValue*>(v)->value());
  case IC_STRING:
    return static_cast<const StringConstantValue*>(this)->value().compare(
	   static_cast<const StringConstantValue*>(v)->value());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string StringConstantValue::toString () const
{
  TRACER("StringConstantValue::toString");
  return (string)"\"" + _value + '\"';
} 
InterpretedConstant::InterpretedConstant (const InterpretedConstant& t)
  :
  _implementation (t._implementation)
{
} 
void InterpretedConstant::useAsNextArgument () const
{
  TRACER("InterpretedConstant::useAsNextArgument");
  ASS (InterpretedConstantValue::_nextArgumentIndex < 
       MAX_ARITY_FOR_INTERPRETED_SYMBOLS);
  InterpretedConstantValue::_args[InterpretedConstantValue::_nextArgumentIndex++] = 
    static_cast<InterpretedConstantValue*>(_implementation);
} 
bool InterpretedConstant::evaluateFunction (int kernelFunctionNumber,
 					    InterpretedConstant& result)
{
  TRACER("InterpretedConstant::evaluateFunction");
  InterpretedConstantValue* val = 
    InterpretedConstantValue::evaluate((EvaluableFunction)Signature::sig.getFunctionByItsKernelNumber(kernelFunctionNumber).evaluatorNumber());
  if (val) {
    result._implementation = val;
    return true;
  }
  return false;
} 
bool InterpretedConstant::evaluatePredicate (int kernelPredicateNumber,
					     bool& result)
{
  TRACER("InterpretedConstant::evaluatePredicate");
  return InterpretedConstantValue::evaluate((EvaluablePredicate)Signature::sig.getPredicateByItsKernelNumber(kernelPredicateNumber).evaluatorNumber(),
					    result);
} 
InterpretedConstantValue* 
InterpretedConstantValue::evaluate (EvaluableFunction fun)
{
  TRACER("InterpretedConstantValue::evaluate/1");
  int next = _nextArgumentIndex;
  _nextArgumentIndex = 0;
  switch (fun) {
  case GENERAL_TO_STRING:
    break;
  case INTEGER_ABS:
  case INTEGER_COMPARE:
  case INTEGER_DIV:
  case INTEGER_MAX:
  case INTEGER_MIN:
  case INTEGER_MINUS:
  case INTEGER_MOD:
  case INTEGER_MULTIPLY:
  case INTEGER_NEGATE:
  case INTEGER_PLUS:
  case INTEGER_RANDOM:
  case INTEGER_SIGN:
  case INTEGER_TO_STRING:
    for (int i = 0;i < next;i++) {
      if (_args[i]->tag() != IC_INTEGER) {
	return 0;
      }
    }
    break;
  case REAL_ABS:
  case REAL_ACOS:
  case REAL_ASIN:
  case REAL_ATAN:
  case REAL_ATAN2:
  case REAL_CEIL:
  case REAL_COMPARE:
  case REAL_COS:
  case REAL_COSH:
  case REAL_DIV:
  case REAL_EXP:
  case REAL_FLOOR:
  case REAL_LN:
  case REAL_LOG10:
  case REAL_MAX:
  case REAL_MIN:
  case REAL_MINUS:
  case REAL_MULTIPLY:
  case REAL_NEGATE:
  case REAL_PLUS:
  case REAL_POWER:
  case REAL_RANDOM:
  case REAL_ROUND:
  case REAL_SIGN:
  case REAL_SIN:
  case REAL_SINH:
  case REAL_SQRT:
  case REAL_TAN:
  case REAL_TANH:
  case REAL_TO_STRING:
  case REAL_TRUNC:
    for (int i = 0;i < next;i++) {
      ICSort srt = _args[i]->tag();
      if (srt != IC_REAL && srt != IC_INTEGER) {
	return 0;
      }
    }
    break;
  case STRING_COMPARE:
  case STRING_CONCAT:
  case STRING_SIZE:
  case INTEGER_FROM_STRING:
  case REAL_FROM_STRING:
    for (int i = 0;i < next;i++) {
      if (_args[i]->tag() != IC_STRING) {
	return 0;
      }
    }
    break;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
  InterpretedConstantValue* v1 = _args[0];
  InterpretedConstantValue* v2 = _args[1];
  switch (fun) {
  case GENERAL_TO_STRING:
    if (v1->tag() == IC_STRING) {
      return v1;
    }
    return new StringConstantValue(v1->toString());
  case INTEGER_ABS:
    return new IntegerConstantValue(abs(static_cast<IntegerConstantValue*>(v1)->value()));
  case INTEGER_COMPARE:
    return new IntegerConstantValue(Int::compare(static_cast<IntegerConstantValue*>(v1)->value(),
						 static_cast<IntegerConstantValue*>(v2)->value()));
  case INTEGER_DIV:
    return new IntegerConstantValue(static_cast<IntegerConstantValue*>(v1)->value() /
				    static_cast<IntegerConstantValue*>(v2)->value());
  case INTEGER_FROM_STRING: 
    {
      long res;
      if (Int::stringToLong(static_cast<StringConstantValue*>(v1)->value().c_str(),res)) {
	return new IntegerConstantValue(res);
      }
      return 0;
    }
  case INTEGER_MAX:
    return new IntegerConstantValue(max(static_cast<IntegerConstantValue*>(v1)->value(),
					static_cast<IntegerConstantValue*>(v2)->value()));
  case INTEGER_MIN:
    return new IntegerConstantValue(min(static_cast<IntegerConstantValue*>(v1)->value(),
					static_cast<IntegerConstantValue*>(v2)->value()));
  case INTEGER_MINUS:
    return new IntegerConstantValue(static_cast<IntegerConstantValue*>(v1)->value() -
				    static_cast<IntegerConstantValue*>(v2)->value());
  case INTEGER_MOD:
    return new IntegerConstantValue(static_cast<IntegerConstantValue*>(v1)->value() %
				    static_cast<IntegerConstantValue*>(v2)->value());
  case INTEGER_MULTIPLY:
    return new IntegerConstantValue(static_cast<IntegerConstantValue*>(v1)->value() *
				    static_cast<IntegerConstantValue*>(v2)->value());
  case INTEGER_NEGATE:
    return new IntegerConstantValue(-static_cast<IntegerConstantValue*>(v1)->value());
  case INTEGER_PLUS:
    return new IntegerConstantValue(static_cast<IntegerConstantValue*>(v1)->value() +
				    static_cast<IntegerConstantValue*>(v2)->value());
  case INTEGER_RANDOM:
    return new IntegerConstantValue(Random::getInteger(static_cast<IntegerConstantValue*>(v1)->value()));
  case INTEGER_SIGN:
    return new IntegerConstantValue(Int::sign(static_cast<IntegerConstantValue*>(v1)->value()));
  case INTEGER_TO_STRING:
    return new StringConstantValue(v1->toString());
  case REAL_ABS:
    return new RealConstantValue(abs(v1->realValue()));
  case REAL_ACOS:
    return new RealConstantValue(acos(v1->realValue()));
  case REAL_ASIN:
    return new RealConstantValue(asin(v1->realValue()));
  case REAL_ATAN:
    return new RealConstantValue(atan(v1->realValue()));
  case REAL_ATAN2:
    return new RealConstantValue(atan2(v1->realValue(),v2->realValue()));
  case REAL_CEIL:
    return new IntegerConstantValue((long)ceil(v1->realValue()));
  case REAL_COMPARE:
    return new RealConstantValue(Int::compare(v1->realValue(),v2->realValue()));
  case REAL_COS:
    return new RealConstantValue(cos(v1->realValue()));
  case REAL_COSH:
    return new RealConstantValue(cosh(v1->realValue()));
  case REAL_DIV:
    return new RealConstantValue(v1->realValue() / v2->realValue());
  case REAL_EXP:
    return new RealConstantValue(exp(v1->realValue()));
  case REAL_FLOOR:
    return new IntegerConstantValue((long)floor(v1->realValue()));
  case REAL_FROM_STRING: 
    {
      double res;
      if (Int::stringToDouble(static_cast<StringConstantValue*>(v1)->value().c_str(),res)) {
	return new RealConstantValue(res);
      }
      return 0;
    }
  case REAL_LN:
    return new RealConstantValue(log(v1->realValue()));
  case REAL_LOG10:
    return new RealConstantValue(log10(v1->realValue()));
  case REAL_MAX:
    return new RealConstantValue(max(v1->realValue(),v2->realValue()));
  case REAL_MIN:
    return new RealConstantValue(min(v1->realValue(),v2->realValue()));
  case REAL_MINUS:
    return new RealConstantValue(v1->realValue() - v2->realValue());
  case REAL_MULTIPLY:
    return new RealConstantValue(v1->realValue() * v2->realValue());
  case REAL_NEGATE:
    return new RealConstantValue(-v1->realValue());
  case REAL_PLUS:
    return new RealConstantValue(v1->realValue() + v2->realValue());
  case REAL_POWER:
    return new RealConstantValue(pow(v1->realValue(),v2->realValue()));
  case REAL_RANDOM:
    return new RealConstantValue(drand48() * v1->realValue());
  case REAL_ROUND:
    return new RealConstantValue(round(v1->realValue()));
  case REAL_SIGN:
    return new IntegerConstantValue(Int::sign(v1->realValue()));
  case REAL_SIN:
    return new RealConstantValue(sin(v1->realValue()));
  case REAL_SINH:
    return new RealConstantValue(sinh(v1->realValue()));
  case REAL_SQRT:
    return new RealConstantValue(sqrt(v1->realValue()));
  case REAL_TAN:
    return new RealConstantValue(tan(v1->realValue()));
  case REAL_TANH:
    return new RealConstantValue(tanh(v1->realValue()));
  case REAL_TO_STRING:
    return new StringConstantValue(v1->toString());
  case REAL_TRUNC:
    return new RealConstantValue(trunc(v1->realValue()));
  case STRING_COMPARE:
    return new IntegerConstantValue(static_cast<StringConstantValue*>(v1)->value().compare(static_cast<StringConstantValue*>(v2)->value()));
  case STRING_CONCAT:
    return new StringConstantValue(static_cast<StringConstantValue*>(v1)->value() +
				   static_cast<StringConstantValue*>(v2)->value());
  case STRING_SIZE:
    return new IntegerConstantValue(static_cast<StringConstantValue*>(v1)->value().length());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string IntegerConstantValue::toString () const 
{ 
  return Int::toString(_value); 
} 
string RealConstantValue::toString () const 
{ 
  return Int::toString(_value); 
} 
double StringConstantValue::realValue() const
{
#if DEBUG_SHELL
  ASSERTION_VIOLATION;
#endif
  return 0.0;
}
bool InterpretedConstantValue::evaluate (EvaluablePredicate pred,bool& result)
{
  TRACER("InterpretedConstantValue::evaluate/2");
  int next = _nextArgumentIndex;
  _nextArgumentIndex = 0;
  switch (pred) {
  case GENERAL_IS_INTEGER:
  case GENERAL_IS_REAL:
  case GENERAL_IS_STRING:
    break;
  case INTEGER_GEQ:
  case INTEGER_GREATER:
  case INTEGER_LEQ:
  case INTEGER_LESS:
    for (int i = 0;i < next;i++) {
      if (_args[i]->tag() != IC_INTEGER) {
	return 0;
      }
    }
    break;
  case REAL_GEQ:
  case REAL_GREATER:
  case REAL_LEQ:
  case REAL_LESS:
    for (int i = 0;i < next;i++) {
      ICSort srt = _args[i]->tag();
      if (srt != IC_REAL && srt != IC_INTEGER) {
	return 0;
      }
    }
    break;
  case STRING_GEQ:
  case STRING_GREATER:
  case STRING_LEQ:
  case STRING_LESS:
    for (int i = 0;i < next;i++) {
      if (_args[i]->tag() != IC_STRING) {
	return 0;
      }
    }
    break;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
  InterpretedConstantValue* v1 = _args[0];
  InterpretedConstantValue* v2 = _args[1];
  switch (pred) {
  case GENERAL_IS_INTEGER:
    result = (v1->tag() == IC_INTEGER);
    return true;
  case GENERAL_IS_REAL:
    result = (v1->tag() == IC_REAL);
    return true;
  case GENERAL_IS_STRING:
    result = (v1->tag() == IC_STRING);
    return true;
  case INTEGER_GEQ:
    result = (static_cast<IntegerConstantValue*>(v1)->value() >=
	      static_cast<IntegerConstantValue*>(v2)->value());
    return true;
  case INTEGER_GREATER:
    result = (static_cast<IntegerConstantValue*>(v1)->value() >
	      static_cast<IntegerConstantValue*>(v2)->value());
    return true;
  case INTEGER_LEQ:
    result = (static_cast<IntegerConstantValue*>(v1)->value() <=
	      static_cast<IntegerConstantValue*>(v2)->value());
    return true;
  case INTEGER_LESS:
    result = (static_cast<IntegerConstantValue*>(v1)->value() <
	      static_cast<IntegerConstantValue*>(v2)->value());
    return true;
  case REAL_GEQ:
    result = (v1->realValue() >= v2->realValue());
    return true;
  case REAL_GREATER:
    result = (v1->realValue() > v2->realValue());
    return true;
  case REAL_LEQ:
    result = (v1->realValue() <= v2->realValue());
    return true;
  case REAL_LESS:
    result = (v1->realValue() < v2->realValue());
    return true;
  case STRING_GEQ:
    result = 
      (static_cast<StringConstantValue*>(v1)->value().compare(static_cast<StringConstantValue*>(v2)->value()) != -1);
    return true;
  case STRING_GREATER:
    result = 
      (static_cast<StringConstantValue*>(v1)->value().compare(static_cast<StringConstantValue*>(v2)->value()) == 1);
    return true;
  case STRING_LEQ:
    result = 
      (static_cast<StringConstantValue*>(v1)->value().compare(static_cast<StringConstantValue*>(v2)->value()) != 1);
    return true;
  case STRING_LESS:
    result = 
      (static_cast<StringConstantValue*>(v1)->value().compare(static_cast<StringConstantValue*>(v2)->value()) == -1);
    return true;
    break;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
