/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <climits>
#include "FormulaVarIterator.hpp"
#include "GlobAlloc.hpp"
#include "Environment.hpp"
#include "Int.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Problem.hpp"
#include "RenamingToNew.hpp"
#include "Signature.hpp"
#include "TPTP.hpp" 
#include "Timer.hpp"
#include "Tracer.hpp"
#include "VarCounter.hpp"
namespace Shell {
class Kernel::Register {
public:
  inline explicit Register(Kernel* kernel)
    {Environment::kernel = kernel;}
  inline ~Register()
    {Environment::kernel = 0;}
};
int Kernel::passClauses (const Problem& problem)
{
  TRACER("Kernel::passClauses");
  _kernel.beginInputClauses();
  int lastClauseNumber = 0;
  UnitChain::Iterator units (problem.units());
  while (units.hasNext()) {
    Unit u (units.next());
    ASS(u.unitType() == CLAUSE);
    RenamingToNew renaming;
    int number = u.number();
    if (number > lastClauseNumber) {
      lastClauseNumber = number;
    }
    bool sos = (_options.sos() && u.inputType() == AXIOM);
    _kernel.beginInputClause(number, 
			     u.giveAway(),
			     u.inputType() != AXIOM,
			     sos);
    Iterator<Literal> lits (u.clause().literals());
    while (lits.hasNext()) {
      Literal& lit = lits.next();
      const Atom& atom = lit.atom();
      const PredicateSymbol& pred = atom.functor();
      _kernel.beginLiteral(lit.isPositive(),pred.kernelNumber());
      if (pred.isEvaluable()) {
	_kernel.markUnselected();
      }
      else if (pred.isEqualityProxy() && 
	       lit.isPositive() && 
	       _options.equalityProxy() == Options::EP_EXP1) {
	_kernel.markSelected();
      }
      Iterator<Term> args (atom.args());
      while (args.hasNext()) {
	passTerm(args.next(), renaming);
      }
      _kernel.endLiteral();		  
    }
    _kernel.endInputClause();
  }
  _kernel.endInputClauses();
  return lastClauseNumber;
} 
void Kernel::passTerm (const Term& t, 
		       RenamingToNew& renaming)
{
  TRACER("Kernel::passTerm");
  switch (t.tag()) 
    {
    case VAR: 
      _kernel.receiveVariable(renaming.findOrBindToNew(t.var()).toInt());
      return;
    case COMPOUND: 
      {
	_kernel.receiveFunction (t.functor().kernelNumber());
	Iterator<Term> ts (t.args());
	while (ts.hasNext()) {
	  passTerm(ts.next(), renaming);
	}
      }
      return;
    case INTERPRETED:
      _kernel.receiveInterpretedConstant(t.value());
      return;
    }
} 
void Kernel::passOptions ()
{
  TRACER("Kernel::passOptions");
  _kernelOptions.age_weight_ratio_numerator = _options.ageRatio();
  _kernelOptions.age_weight_ratio_denominator = _options.weightRatio();
  _kernelOptions.allocation_buffer_price = _options.allocationBufferPrice ();
  _kernelOptions.allocation_buffer_size = _options.allocationBufferSize ();
  switch (_options.backwardDemodulation()) 
    {
    case Options::DEMODULATION_OFF:
      _kernelOptions.backward_demodulation = 0;
      _kernelOptions.bd_on_splitting_branches = false;
      break;
    case Options::DEMODULATION_PREORDERED:
      _kernelOptions.backward_demodulation = 1;
      _kernelOptions.bd_on_splitting_branches = true;
      break;
    case Options::DEMODULATION_ALL:
      _kernelOptions.backward_demodulation = 4;
      _kernelOptions.bd_on_splitting_branches = true;
      break;
    }
  if (_options.backwardSubsumption()) {
    _kernelOptions.backward_subsumption = true;
  }
  else {
    _kernelOptions.backward_subsumption = false;
  }
  if (_options.basicStrategy()) {
    _kernelOptions.prohibit_non_basic_superposition = true;
    _kernelOptions.prohibit_non_basic_fd = true;
    _kernelOptions.prohibit_non_basic_bd = true;
    _kernelOptions.reduce_non_basic = true;
    _kernelOptions.input_non_basicness = true;
  }
  else {
    _kernelOptions.prohibit_non_basic_superposition = false;
    _kernelOptions.prohibit_non_basic_fd = false;
    _kernelOptions.prohibit_non_basic_bd = false;
    _kernelOptions.reduce_non_basic = false;
    _kernelOptions.input_non_basicness = false;
  }
  _kernelOptions.simplify_by_eq_res = _options.equalityResolutionWithDeletion();
  _kernelOptions.normalize_by_commutativity = _options.fastDemodulationByCommutativity();
  _kernelOptions.fd_on_splitting_branches = true;
  switch (_options.forwardDemodulation()) 
    {
    case Options::DEMODULATION_OFF:
      _kernelOptions.forward_demodulation = 0;
      break;
    case Options::DEMODULATION_PREORDERED:
      _kernelOptions.forward_demodulation = 1;
      break;
    case Options::DEMODULATION_ALL:
      _kernelOptions.forward_demodulation = 4;
      break;
    }
  if (_options.forwardSubsumption()) {
    _kernelOptions.forward_subsumption = true;
  }
  else {
    _kernelOptions.forward_subsumption = false;
  }
  _kernelOptions.simplify_by_fs = _options.forwardSubsumptionResolution ();
  _kernelOptions.static_neg_eq_splitting = _options.inequalitySplitting ();
  _kernelOptions.inherited_negative_selection = _options.inheritedNegativeSelection ();
  _kernelOptions.first_mem_check = INT_MAX;
  _kernelOptions.first_time_check = _options.lrsFirstTimeCheck();
  _kernelOptions.simulated_time_limit = _options.simulatedTimeLimit();
  switch (_options.mainAlg()) 
    {
    case Options::LRS:
      _kernelOptions.main_alg = 0;
      _kernelOptions.lrs = 1;
      break;
    case Options::OTTER:
      _kernelOptions.main_alg = 0;
      _kernelOptions.lrs = 0;
      break;
    case Options::DISCOUNT:
      _kernelOptions.main_alg = 3;
      _kernelOptions.lrs = 0;
      break;
    }
  _kernelOptions.main_selected_lit_demodulation = _options.mainPremiseDemodulation();
  _kernelOptions.main_premise_subsumption = _options.mainPremiseSubsumption();
  int max = _options.maxActive();
  _kernelOptions.max_active =  max ? max : LONG_MAX;
  max = _options.maxInferenceDepth();
  _kernelOptions.max_inference_depth = max ? max : LONG_MAX;
  max = _options.maxPassive();
  _kernelOptions.max_passive = max ? max : LONG_MAX;
  _kernelOptions.max_skolem = _options.maxSkolem();
  max = _options.maxWeight();
  _kernelOptions.max_weight = max ? max : LONG_MAX;
  _kernelOptions.memory_limit = _options.memoryLimit() * 1000 - 
                        GlobalAllocator::getUsedMemoryInKilobytes();
  _kernelOptions.no_inferences_between_rules= _options.noInferencesBetweenRules();
  _kernelOptions.nongoal_penalty_coefficient = _options.nongoalWeightCoefficient();
  _kernelOptions.orphan_murder = _options.orphanDeletion();
  _kernelOptions.paramodulation = true;
  _kernelOptions.paramodulate_from_variables = _options.superpositionFromVariables ();
  _kernelOptions.selected_pos_eq_penalty_coefficient = _options.selectedEqualityWeightCoefficient();    
  switch (_options.selection()) {
  case -11:
    _kernelOptions.selection = -3;
    break;
  case -16:
    _kernelOptions.selection = -6;
    break;
  case -1012:
    _kernelOptions.selection = -1003;
    break;
  case 0:
    _kernelOptions.selection = 0;
    break;
  case 8:
    _kernelOptions.selection = 1;
    break;
  case 9:
    _kernelOptions.selection = 2;
    break;
  case 10:
    _kernelOptions.selection = 4;
    break;
  case 11:
    _kernelOptions.selection = 3;
    break;
  case 13:
    _kernelOptions.selection = 5;
    break;
  case 16:
    _kernelOptions.selection = 6;
    break;
  case 17:
    _kernelOptions.selection = 7;
    break;
  case 1010:
    _kernelOptions.selection = 1001;
    break;
  case 1011:
    _kernelOptions.selection = 1002;
    break;
  case 1012:
    _kernelOptions.selection = 1003;
    break;
  case 1013:
    _kernelOptions.selection = 1004;
    break;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
  _kernelOptions.show_active = _options.showActive();
  _kernelOptions.show_gen = _options.showGenerated();
  _kernelOptions.show_input = _options.showInput();
  _kernelOptions.show_opt = _options.showOptions();
  _kernelOptions.show_passive = _options.showPassive();
  _kernelOptions.show_profile = _options.showProfile();
  _kernelOptions.show_weight_limit_changes = _options.showWeightLimitChanges();
  _kernelOptions.side_selected_lit_demodulation = _options.sidePremiseDemodulation();
  _kernelOptions.side_premise_subsumption = _options.sidePremiseSubsumption();
  _kernelOptions.simplification_ordering = VampireKernel::Options::StandardKBOrdering;
  _kernelOptions.sim_back_sup = _options.simultaneousBackwardSuperposition();
  switch (_options.splitting ()) 
    {
    case Options::SPLIT_OFF:
      _kernelOptions.static_splitting = false;
      _kernelOptions.dynamic_splitting = false;
      break;
    case Options::SPLIT_INPUT_ONLY:
      _kernelOptions.static_splitting = true;
      _kernelOptions.dynamic_splitting = false;
      break;
    case Options::SPLIT_ON:
      _kernelOptions.static_splitting = true;
      _kernelOptions.dynamic_splitting = true;
      break;
    }
  _kernelOptions.term_weighting_scheme = VampireKernel::Options::NonuniformConstantTermWeighting;
  string jobId = Int::toString(_options.testNumber());
  _kernelOptions.job_id = jobId.c_str();
  switch(_options.transitivityChaining()) 
    {
    case Options::TC_NONE:
      _kernelOptions.chaining_for_transitive = false;
      break;
    case Options::TC_ONLY_NONVARIABLES:
      _kernelOptions.chaining_for_transitive = true;
      _kernelOptions.chaining_through_variables = false;
      break;
    case Options::TC_FULL:
      _kernelOptions.chaining_for_transitive = true;
      _kernelOptions.chaining_through_variables = true;
      break;
    }
  _kernelOptions.std_output = 0;
} 
void Kernel::passSignature ()
{
  TRACER ("Kernel::passSignature");
  _kernel.beginInputSignature();
  Signature::PredIterator preds(_signature);
  while (preds.hasNext()) {
    PredicateSymbol& pred = static_cast<PredicateSymbol&>(preds.next());
    ulong symbolId; 
    if (pred.isEquality()) {
      _kernel.registerEquality(symbolId,
			       pred.level(),
			       pred.negativeLevel());
    }
    else {
      _kernel.registerInputPredicate(pred.name(),
				     pred.arity(),
				     pred.isAnswer(),
				     symbolId,
				     pred.weight(),
				     pred.precedence(),
				     pred.level(),
				     pred.negativeLevel());
    }
    _signature.registerKernelNumber(pred,symbolId);
    if (pred.isEvaluable()) {
      _kernel.registerInterpretedPredicate(symbolId);
    }
  }
  Signature::FunIterator funs(_signature);
  while (funs.hasNext()) {
    ulong symbolId; 
    Symbol& fun = funs.next();
    _kernel.registerInputFunction(fun.name(),
				  fun.arity(),
				  symbolId,
				  fun.weight(),
				  fun.precedence());
    _signature.registerKernelNumber(fun,symbolId);
    if (fun.isEvaluable()) {
      _kernel.registerInterpretedFunction(symbolId);
    }
  }
  _kernel.endInputSignature();
} 
void Kernel::getClause (const VampireKernel::Clause* clause, 
			Clause& result) const
{
  TRACER("Kernel::getClause");
  LiteralList literals;
  getLiteralList(clause->literals(), literals);
  result = Clause(literals);
} 
void Kernel::getDefinition (const VampireKernel::Clause* clause, 
			    Formula& result) const
{
  TRACER("Kernel::getDefinition");
  const VampireKernel::Literal* definedLiteral;
  const VampireKernel::Literal* bodyLiterals;
#if DEBUG_SHELL
  ASS(clause->isLiteralDefinition(definedLiteral,bodyLiterals));
#else
  clause->isLiteralDefinition(definedLiteral,bodyLiterals);
#endif
  Literal l;
  getLiteral(definedLiteral,l);
  LiteralList ls;
  getLiteralList(bodyLiterals,ls);
  Formula lhs(l.atom());
  Formula rhs;
  Formula::makeFromLiteralList(ls,rhs);
  FormulaVarIterator fvs(lhs);
  VarCounter vc;
  while (fvs.hasNext()) {
    Var v(fvs.next());
    vc.incCount(v);
  }
  VarList ws; 
  Stack<Var> stack(ws); 
  FormulaVarIterator fws(rhs);
  while (fws.hasNext()) {
    Var v(fws.next());
    if (vc.getCount(v) == 0) {
      stack.push(v);
      vc.incCount(v);
    }
  }
  if (! ws.isEmpty()) {
    rhs = Formula(FORALL,ws,rhs);
  }
  if (l.isNegative()) {
    rhs = Formula(NOT,rhs);
  }
  result = Formula(IFF,lhs,rhs);
} 
void Kernel::getLiteral (const VampireKernel::Literal* literal, 
			 Literal& result) const
{
  TRACER("Kernel::getLiteral");
  TermList ts;
  getTermList(literal->firstArgument(), ts);
  ulong sym = literal->predicate();
  const PredicateSymbol& p = 
    _signature.getPredicateByItsKernelNumber(sym,
					     _kernel.symbolArity(sym));
  Atom a (p,ts);
  bool sign = literal->isPositive();
  if (p.isName()) {
    sign = !sign;
  }
  result = Literal(sign,a);
} 
void Kernel::getLiteralList (const VampireKernel::Literal* literal,
			     LiteralList& result) const
{
  TRACER("Kernel::getLiteralList");
  if (! literal) {
    return;
  }
  Literal l;
  getLiteral(literal,l);
  LiteralList ls;
  getLiteralList(literal->next(),ls);
  result = LiteralList(l,ls);
} 
void Kernel::getTerm (const VampireKernel::Subterm* term, 
		      Term& result) const
{
  TRACER("Kernel::getTerm");
  if (term->isVariable()) {
    result = Term(Var(term->variableNumber()+1));
    return;
  }
  ulong symbol = term->topFunction();
  if (_kernel.isInterpretedConstantId(symbol)) {
    InterpretedConstant ic;
    _kernel.interpretedConstantValue(symbol,ic);
    result = Term(ic);
    return;
  }
  TermList ts;
  getTermList(term->firstArgument(),ts);
  result = Term(_signature.getFunctionByItsKernelNumber(symbol,
							_kernel.symbolArity(symbol)),
		ts);
} 
void Kernel::getTermList (const VampireKernel::Subterm* terms, 
			  TermList& result) const
{
  TRACER("Kernel::getTermList");
  if (! terms) {
    return;
  }
  Term t;
  getTerm(terms,t);
  TermList ts;
  getTermList(terms->next(),ts);
  result = TermList(t,ts);
} 
RunResult Kernel::rerun (long timeLimit)
{
  TRACER("rerun");
  Register reg(this); 
  try 
    {
      _kernel.lookForAnotherAnswer(timeLimit);
    }
  catch (const VampireKernel::Emergency& e)
    {
      switch (e.type())
	{
	case VampireKernel::Emergency::AllocationFailure:
	  return UNKNOWN;
	case VampireKernel::Emergency::TimeLimit:
	  return UNKNOWN;
	};
    };
  RunResult result = UNKNOWN;
  switch (_kernel.terminationReason())
    {
    case VampireKernel::UnknownTerminationReason:
      ICP("ICP0"); 
      BK::RuntimeError::report("Kernel session terminated for unknown reason");
    case VampireKernel::RefutationFound:
      result = REFUTATION;
      break;
    case VampireKernel::Satisfiable:
      result = SATISFIABLE;
      break;
    case VampireKernel::EmptyPassive:
      break;
    case VampireKernel::MaxPassive:
      break;
    case VampireKernel::MaxActive:
      break;
    #if DEBUG_SHELL
    default:
      ASS (false);
    #endif
    };
  return result;
} 
RunResult Kernel::run (long timeLimitInDeciseconds,
		       Problem& problem,
		       Timer& timer,
		       XML::DeepElement* log)
{
  TRACER ("Kernel::run");
  Register reg(this); 
  passOptions();
  bool weightIncrement = _options.weightIncrement();
  int weightLimit = 10;
  long initialTime = timer.elapsedDeciseconds();
  for (;;) {
    bool emergency = false;
    try 
      {
	BK::GlobAlloc::reset ();
	if (weightIncrement) {
	  _kernelOptions.max_weight = weightLimit;
	}
	_kernel.openSession(_kernelOptions);
	passSignature();
	int lastClauseNumber = passClauses(problem);
	int kernelTimeLimit = _options.timeLimitInDeciseconds() ?
	  timeLimitInDeciseconds + initialTime - timer.elapsedDeciseconds() :
	  INT_MAX;
 	_kernel.saturate(lastClauseNumber+1,kernelTimeLimit);
      }
    catch (const VampireKernel::Emergency& e)
      {
	switch (e.type())
	  {
	  case VampireKernel::Emergency::AllocationFailure:
	    emergency = true;
	    break;
	  case VampireKernel::Emergency::TimeLimit:
	    return UNKNOWN;
	  };
      };
    if (! emergency) {
      switch (_kernel.terminationReason()) {
      case VampireKernel::RefutationFound:
	return REFUTATION;
      case VampireKernel::Satisfiable:
	return SATISFIABLE;
      default:
	break;
      }
    }
    if (weightIncrement) {
      BK::DestructionMode::makeFast();
      _kernel.closeSession();
      BK::DestructionMode::makeThorough();
      weightLimit = (weightLimit * 6) / 5;
    }
    else {
      return UNKNOWN;
    }
  }
} 
Kernel::Kernel (Options& options,Signature& signature)
  : _options(options),
    _signature(signature)
{
}
}
