/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Map__
#define __Map__
#include "Tracer.hpp"
#include "VampireException.hpp"
namespace Shell {
template <typename Key, typename Val>
class Map 
{
  friend class Debugger;
 public:
  Map ();
  ~Map ();
  bool find (Key key);
  bool find (Key key, Val& value);
  void insert (Key key, Val value);
  void replace (Key key, Val value);
#if DEBUG_SHELL
#else
  void* operator new (size_t);
#endif
 protected:
  class Entry
  {
  public:
    inline 
    Entry ()
      : _occupied (false)
    {
    } 
    inline
    void fill (Key key, Val value)
    {
      ASS(! _occupied);
      _occupied = true;
      _key = key;
      _value = value;
    } 
    inline
    bool isOccupied () const
    {
      return _occupied;
    } 
    inline void setValue (Val value) { _value = value; }
    inline Key key () const { return _key; }
    inline Val value () const { return _value; }
    void* operator new (size_t);
  private:
    bool _occupied;
    Key _key;
    Val _value;
  }; 
  int* _capacityPointer;
  int _capacity;
  int _noOfEntries;
  Entry* _entries;
  Entry* _afterLast; 
  int _maxEntries;
  void expand ();
  Entry* findEntry (Key); 
}; 
class MapConstants {
public:
  static int capacities[];
};
template <class Key, class Val>
inline
Map<Key,Val>::Map ()
  : _capacityPointer (MapConstants::capacities),
    _noOfEntries (0),
    _entries (0)
{
  expand ();
} 
template <class Key, class Val>
void Map<Key, Val>::expand ()
{
  if (_capacityPointer[1] == 0) {
    ALLOCATION_FAILURE;
  }
  _capacityPointer ++;
  _capacity = *_capacityPointer;
  Entry* oldEntries = _entries;
  _entries = new Entry [_capacity];
  if (! _entries) {
    ALLOCATION_FAILURE;
  }
  _afterLast = _entries + _capacity;
  _maxEntries = (int)(_capacity * 0.7);
  Entry* current = oldEntries;
  int remaining = _noOfEntries;
  _noOfEntries = 0;
  while (remaining != 0) {
    while (! current->isOccupied()) {
      current ++;
    }
    insert (current->key(), current->value());
    current ++;
    remaining --;
  }
  delete [] oldEntries;
} 
template <class Key, class Val>
Map<Key,Val>::~Map ()
{
  delete [] _entries;
} 
template <class Key, class Val>
bool Map<Key,Val>::find (Key key, Val& found)
{
  Entry* entry = findEntry (key);
  if (entry->isOccupied()) {
    found = entry->value();
    return true;
  }
  return false;
} 
template <class Key, class Val>
inline
bool Map<Key,Val>::find (Key key)
{
  Entry* entry = findEntry (key);
  return entry->isOccupied();
} 
template <class Key, class Val>
void Map<Key,Val>::insert (Key key, Val value)
{
  if (_noOfEntries >= _maxEntries) { 
    expand ();
  }
  _noOfEntries ++;
  Entry* entry = findEntry(key);
  ASS (! entry->isOccupied());
  entry->fill (key, value);
} 
template <class Key, class Val>
inline
void Map<Key,Val>::replace (Key key, Val value)
{
  Entry* entry = findEntry(key);
  ASS (entry->isOccupied());
  entry->setValue(value);
} 
template <class Key, class Val>
typename Map<Key, Val>::Entry* Map<Key, Val>::findEntry (Key key)
{
  Entry* entry = _entries + ((unsigned int)key % _capacity);
  while ( entry->isOccupied() ) {
    if (entry->key() == key) {
      return entry;
    }
    entry ++;
    if (entry ==_afterLast) {
      entry =_entries;
    }
  }
  return entry;
} 
}
#endif 
