/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Tracer.hpp"
#include "NNF.hpp"
#include "Unit.hpp"
#include "InferenceInfo.hpp"
#include "Inference.hpp"
#include "Position.hpp"
using namespace Shell;
void NNF::ennf (Unit& u)
{
  TRACER("NNF::ennf (Unit& u)");
  ASS(u.unitType() == FORMULA);
  Formula f(u.formula());
  InferenceInfo* info = 0;
  ennf(f,true,info,Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_ENNF,u);
    inf->setInfo(info);
    u = Unit(f,inf);
  }
} 
void NNF::nnf (Unit& u)
{
  TRACER("NNF::nnf (Unit& u)");
  ASS(u.unitType() == FORMULA);
  Formula f(u.formula());
  InferenceInfo* info = 0;
  nnf(f,true,info,Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_NNF,u);
    inf->setInfo(info);
    u = Unit(f,inf);
  }
} 
void NNF::ennf (Formula& f, bool polarity, 
		InferenceInfo*& info, const Position& pos)
{
  Connective c = f.connective();
  switch (c) {
  case ATOM:
    if (! polarity) {
      f = Formula (NOT, f);
    }
    return;
  case AND: 
  case OR: 
    {
      FormulaList fs (f.args());
      ennf(fs,polarity,0,info,pos);
      if (fs == f.args()) { 
	return;
      }
      f = Formula ((polarity == (c == AND)) ? AND : OR, fs);
      if (! polarity) {
	info = 
	  new InferenceInfo(c == AND ? IR_NOT_AND : IR_NOT_OR,
			    pos, info);
      }
    }
    return;
  case IMP: 
    {
      Formula l (f.left());
      Formula r (f.right());
      ennf(l, ! polarity, info, Position(0,pos));
      ennf(r, polarity, info, Position(1,pos));
      if (polarity) {
	f = Formula(OR, FormulaList(l,FormulaList(r)));
	info = new InferenceInfo(IR_IMP_TO_OR, pos, info);
      }
      else {
	f = Formula(AND, FormulaList(l,FormulaList(r)));
	info = new InferenceInfo(IR_NOT_IMP, pos, info);
      }
    }
    return;
  case IFF: 
  case XOR: 
    {
      Formula l(f.left());
      ennf(l,true, info, Position(0,pos));
      Formula r(f.right());
      ennf(r,true, info, Position(1,pos));
      if (polarity && l == f.left() && r == f.right()) { 
	return;
      }
      f = Formula(polarity == (c == IFF) ? IFF : XOR, l, r);
      if (! polarity) {
	info = 
	  new InferenceInfo(c == IFF ? IR_NOT_IFF : IR_NOT_XOR,
			    pos, info);
      }
    }
    return;
    case NOT:
      if (polarity && f.uarg().connective() == ATOM) {
        return;
      }
      f = f.uarg();
      ennf(f,! polarity,info,pos);
      if (! polarity) {
	info = new InferenceInfo(IR_NOT_NOT, pos, info);
      }
      return;
  case FORALL:
  case EXISTS: 
    {
      Formula arg(f.qarg());
      ennf(arg,polarity,info,Position(0,pos));
      if (arg == f.qarg()) {
	return;
      }
      f = Formula(polarity == (c == FORALL) ? FORALL : EXISTS,
		  f.vars(),
		  arg);
      if (! polarity) {
	info = new InferenceInfo(c == FORALL ? IR_NOT_FORALL : IR_NOT_EXISTS,
				 pos, info);
      }
    }
    return;
  case FALSE:
  case TRUE:
    return;
#if DEBUG_SHELL
  default:
    ASS( false );
#endif
  }
} 
void NNF::ennf (FormulaList& fs, bool polarity, int n,
		InferenceInfo*& info, const Position& pos)
{
  if (fs.isEmpty()) {
    return;
  }
  Formula hd (fs.head());
  ennf(hd,polarity,info,Position(n,pos));
  FormulaList tl (fs.tail());
  ennf(tl,polarity,n+1,info,pos);
  if (hd == fs.head() && tl == fs.tail()) {
    return;
  }
  fs = FormulaList (hd,tl);
} 
void NNF::nnf (Formula& f, bool polarity, 
	       InferenceInfo*& info, const Position& pos)
{
  Connective c = f.connective();
  switch (c) {
  case ATOM:
    if (! polarity) {
      f = Formula (NOT, f);
    }
    return;
  case AND: 
  case OR: 
    {
      FormulaList fs (f.args());
      nnf(fs,polarity,0,info,pos);
      if (fs == f.args()) { 
	return;
      }
      f = Formula ((polarity == (c == AND)) ? AND : OR, fs);
      if (! polarity) {
	info = 
	  new InferenceInfo(c == AND ? IR_NOT_AND : IR_NOT_OR,
			    pos, info);
      }
    }
    return;
  case IMP: 
    {
      Formula l (f.left());
      Formula r (f.right());
      nnf(l, ! polarity, info, Position(0,pos));
      nnf(r, polarity, info, Position(1,pos));
      if (polarity) {
	f = Formula(OR, FormulaList(l,FormulaList(r)));
	info = new InferenceInfo(IR_IMP_TO_OR, pos, info);
      }
      else {
	f = Formula(AND, FormulaList(l,FormulaList(r)));
	info = new InferenceInfo(IR_NOT_IMP, pos, info);
      }
    }
    return;
  case IFF: 
  case XOR: 
    {
      Formula l(f.left());
      Formula r(f.right());
      if (polarity == (c == IFF)) {
	f = Formula(AND,
		    FormulaList(Formula(IMP,l,r),
				FormulaList(Formula(IMP,r,l))));
      }
      else {
	f = Formula(AND,
		    FormulaList(Formula(OR,FormulaList(l,
                                           FormulaList(r))),
		    FormulaList(Formula(OR,FormulaList(Formula(NOT,l),
                                           FormulaList(Formula(NOT,r)))))));
      }
      nnf(f,true,info,pos);
      if (polarity) {
	if (c == IFF) {
	  info = new InferenceInfo(IR_IFF_TO_AND, pos, info);
	}
	else {
	  info = new InferenceInfo(IR_XOR_TO_AND, pos, info);
	}
      }
      else {
	if (c == IFF) {
	  info = new InferenceInfo(IR_XOR_TO_AND, pos, info);
	  info = new InferenceInfo(IR_NOT_IFF, pos, info);
	}
	else {
	  info = new InferenceInfo(IR_IFF_TO_AND, pos, info);
	  info = new InferenceInfo(IR_NOT_XOR, pos, info);
	}
      }
    }
    return;
    case NOT:
      if (polarity && f.uarg().connective() == ATOM) {
        return;
      }
      f = f.uarg();
      nnf(f,! polarity,info,pos);
      if (! polarity) {
	info = new InferenceInfo(IR_NOT_NOT, pos, info);
      }
      return;
  case FORALL:
  case EXISTS: 
    {
      Formula arg(f.qarg());
      nnf(arg,polarity,info,Position(0,pos));
      if (arg == f.qarg()) {
	return;
      }
      f = Formula(polarity == (c == FORALL) ? FORALL : EXISTS,
		  f.vars(),
		  arg);
      if (! polarity) {
	info = new InferenceInfo(c == FORALL ? IR_NOT_FORALL : IR_NOT_EXISTS,
				 pos, info);
      }
    }
    return;
  case FALSE:
  case TRUE:
    return;
#if DEBUG_SHELL
  default:
    ASS( false );
#endif
  }
} 
void NNF::nnf (FormulaList& fs, bool polarity, int n,
	       InferenceInfo*& info, const Position& pos)
{
  if (fs.isEmpty()) {
    return;
  }
  Formula hd (fs.head());
  nnf(hd,polarity,info,Position(n,pos));
  FormulaList tl (fs.tail());
  nnf(tl,polarity,n+1,info,pos);
  if (hd == fs.head() && tl == fs.tail()) {
    return;
  }
  fs = FormulaList (hd,tl);
} 
