/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Inference.hpp"
#include "Int.hpp"
#include "Naming.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
#include "VampireException.hpp"
using namespace Shell;
struct Naming::ClsRep {
  int conCls;
  int disCls;
  int conRep;
  int disRep;
};
Naming::Naming (int threshold, Signature& signature)
  : _threshold(threshold+1),
    _signature(signature),
    _calc(threshold+1)
{
  ASS(threshold >= 1);
  ASS(threshold < 32768);
} 
void Naming::apply (Unit& u)
{
  TRACER("Naming::apply(Unit&)");
  Formula f(u.formula());
  switch (f.connective()) {
  case TRUE:
  case FALSE:
    return;
  default:
    break;
  }
  ClsRep info;
  apply(f,ON_TOP,info);
  if (f != u.formula()) {
    u = Unit(f,new Inference(IR_DEFINITION_FOLDING,_definitions));
  }
} 
void Naming::apply (Formula& f,Where where,ClsRep& info)
{
  TRACER("Naming::apply(Formula& ...)");
  switch (f.connective()) {
  case ATOM:
  case NOT:
    info.conCls = 1;
    info.disCls = 1;
    info.conRep = 0;
    info.disRep = 0;
    return;
  case AND: {
    FormulaList fs(f.args());
    int length = fs.length();
    Formula* forms = listToArray(fs,length);
    ClsRep* infos = new ClsRep[length];
    bool changed = apply(fs,where,forms,infos);
    if (changed) {
      arrayToList(fs,length,forms);
      f = Formula(AND,fs);
    }
    if (where == ON_TOP) {
      delete [] infos;
      delete [] forms;
      return;
    }
    changed = false;
    int conCls = 0;
    int conRep = 1;
    for (int i = length-1;i >= 0;i--) {
      ClsRep& inf = infos[i];
      if (_calc.infinite(inf.conRep)) {
	Formula name;
	introduceDefinition(forms[i],name,where == UNDER_IFF);
	forms[i] = name;
	inf.conRep = 0;
	inf.disRep = 0;
	inf.conCls = 1;
	inf.disCls = 1;
	changed = true;
      }
      conCls += inf.conCls;
      conRep = Int::max(conRep,inf.conRep);
    }
    info.conCls = conCls;
    info.conRep = Int::max(1,conRep);
    if (changed) {
      arrayToList(fs,length,forms);
      f = Formula(AND,fs);
    }
    changed = false;
    int prod = 1;
    Calculator big(_threshold * _threshold);
    for (int i = length-1;i >= 0;i--) {
      int p = big.mult(prod,infos[i].disCls);
      if (big.infinite(p)) {
	for (int j = i;j >= 0;j--) {
	  if (infos[j].disRep) {
	    changed = true;
	    Formula name;
	    introduceDefinition(forms[j],name,where == UNDER_IFF);
	    forms[j] = name;
	    infos[j].disRep = 0;
	    infos[j].disCls = 1;
	  }
	}
	break;
      }
      prod = p;
    }
    for (int i = length-1; i >= 0;i--) {
      if (_calc.infinite((prod/infos[i].disCls)*infos[i].disRep)) {
	changed = true;
	Formula name;
	introduceDefinition(forms[i],name,where == UNDER_IFF);
	forms[i] = name;
	infos[i].disRep = 0;
	infos[i].disCls = 1;
	prod /= infos[i].disCls;
      }
    }
    info.disCls = _calc.plus(prod,0);
    int disRep = 0;
    for (int i = length-1; i >= 0;i--) {
      int rep = (prod/infos[i].disCls)*infos[i].disRep;
      if (rep > disRep) {
	disRep = rep;
      }
    }
    info.disRep = disRep;
    if (changed) {
      arrayToList(fs,length,forms);
      f = Formula(AND,fs);
    }
    delete [] infos;
    delete [] forms;
    return;
  }
  case OR: { 
    FormulaList fs(f.args());
    int length = fs.length();
    Formula* forms = listToArray(fs,length);
    ClsRep* infos = new ClsRep[length];
    bool changed = apply(fs,
			 where == UNDER_IFF ? UNDER_IFF : OTHER,
			 forms,
			 infos);
    if (changed) {
      f = Formula(OR,fs);
    }
    changed = false;
    int disCls = 0;
    int disRep = 1;
    for (int i = length-1;i >= 0;i--) {
      ClsRep& inf = infos[i];
      if (_calc.infinite(inf.disRep)) {
	Formula name;
	introduceDefinition(forms[i],name,where == UNDER_IFF);
	forms[i] = name;
	inf.disRep = 0;
	inf.conRep = 0;
	inf.disCls = 1;
	inf.conCls = 1;
	changed = true;
      }
      disCls += inf.disCls;
      disRep = Int::max(disRep,inf.disRep);
    }
    info.disCls = disCls;
    info.disRep = Int::max(1,disRep);
    if (changed) {
      arrayToList(fs,length,forms);
      f = Formula(OR,fs);
    }
    changed = false;
    int prod = 1;
    Calculator big(_threshold * _threshold);
    for (int i = length-1;i >= 0;i--) {
      int p = big.mult(prod,infos[i].conCls);
      if (big.infinite(p)) {
	for (int j = i;j >= 0;j--) {
	  if (infos[j].conRep) {
	    changed = true;
	    Formula name;
	    introduceDefinition(forms[j],name,where == UNDER_IFF);
	    forms[j] = name;
	    infos[j].conRep = 0;
	    infos[j].conCls = 1;
	  }
	}
	break;
      }
      prod = p;
    }
    for (int i = length-1; i >= 0;i--) {
      if (_calc.infinite((prod/infos[i].conCls)*infos[i].conRep)) {
	changed = true;
	Formula name;
	introduceDefinition(forms[i],name,where == UNDER_IFF);
	forms[i] = name;
	infos[i].conRep = 0;
	infos[i].conCls = 1;
	prod /= infos[i].conCls;
      }
    }
    info.conCls = _calc.plus(prod,0);
    int conRep = 0;
    for (int i = length-1; i >= 0;i--) {
      int rep = (prod/infos[i].conCls)*infos[i].conRep;
      if (rep > conRep) {
	conRep = rep;
      }
    }
    info.conRep = conRep;
    if (changed) {
      arrayToList(fs,length,forms);
      f = Formula(OR,fs);
    }
    delete [] infos;
    delete [] forms;
    return;
  } 
  case IFF: {
    Formula lf(f.left());
    ClsRep linfo;
    apply(lf,UNDER_IFF,linfo);
    Formula rf(f.right());
    ClsRep rinfo;
    apply(rf,UNDER_IFF,rinfo);
    bool name = false;
    bool start = true;
    for (;;) {
      if (! name) {
	if (! start) { 
	  if (lf != f.left() || rf != f.right()) {
	    f = Formula(IFF,lf,rf);
	  }
	  info.conCls = iffConCls(linfo,rinfo);
	  info.conRep = iffConRep(linfo,rinfo);
	  info.disCls = iffDisCls(linfo,rinfo);
	  info.disRep = iffDisRep(linfo,rinfo);
	  return;
	}
	start = false;
      }
      else if (linfo.conRep + linfo.disRep >= rinfo.conRep + rinfo.disRep) {
	Formula name;
	introduceDefinition(lf,name,true);
	lf = name;
	linfo.disRep = 0;
	linfo.disCls = 1;
	linfo.conRep = 0;
	linfo.conCls = 1;
      }
      else {
	Formula name;
	introduceDefinition(rf,name,true);
	rf = name;
	rinfo.disRep = 0;
	rinfo.disCls = 1;
	rinfo.conRep = 0;
	rinfo.conCls = 1;
      }
      name = false;
      if (_calc.infinite(iffConRep(linfo,rinfo)) ||
	  (where != ON_TOP &&
	   _calc.infinite(iffDisRep(linfo,rinfo)))) {
	name = true;
      }
    }
  }
  case XOR: {
    Formula lf(f.left());
    ClsRep linfo;
    apply(lf,UNDER_IFF,linfo);
    Formula rf(f.right());
    ClsRep rinfo;
    apply(rf,UNDER_IFF,rinfo);
    bool name = false;
    bool start = true;
    for (;;) {
      if (! name) {
	if (! start) { 
	  if (lf != f.left() || rf != f.right()) {
	    f = Formula(XOR,lf,rf);
	  }
	  info.conCls = iffDisCls(linfo,rinfo);
	  info.conRep = iffDisRep(linfo,rinfo);
	  info.disCls = iffConCls(linfo,rinfo);
	  info.disRep = iffConRep(linfo,rinfo);
	  return;
	}
	start = false;
      }
      else if (linfo.conRep + linfo.disRep >= rinfo.conRep + rinfo.disRep) {
	Formula name;
	introduceDefinition(lf,name,true);
	lf = name;
	linfo.disRep = 0;
	linfo.disCls = 1;
	linfo.conRep = 0;
	linfo.conCls = 1;
      }
      else {
	Formula name;
	introduceDefinition(rf,name,true);
	rf = name;
	rinfo.disRep = 0;
	rinfo.disCls = 1;
	rinfo.conRep = 0;
	rinfo.conCls = 1;
      }
      name = false;
      if (_calc.infinite(iffDisRep(linfo,rinfo)) ||
	  _calc.infinite(iffConRep(linfo,rinfo))) {
	name = true;
      }
    }
  }
  case FORALL:
  case EXISTS: {
    Formula g(f.qarg());
    apply(g,where,info);
    if (g != f.qarg()) {
      f = Formula(f.connective(),f.vars(),g);
    }
    return;
  }
#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
void Naming::introduceDefinition (const Formula& f,Formula& name,bool iff)
{
  TRACER("Naming::introduceDefinition");
  ASS(f.connective() != ATOM);
  ASS(f.connective() != NOT);
  VarList vs;
  f.freeVariables(vs);
  PredicateSymbol& nm = _signature.newName(vs.length());
  Atom atom(nm,TermList(vs));
  name = Formula(atom);
  Formula def;
  if (iff) {
    def = Formula(IFF,name,f);
  }
  else {
    FormulaList fs(f.connective() == OR ?
		   f.args() :
		   FormulaList(f));
    fs.push(Formula(NOT,name));
    def = Formula(OR,fs);
  }
  if (vs.isNonEmpty()) {
    def = Formula(FORALL,vs,def);
  }
  Unit definition(def,new Inference(IR_PREDICATE_DEFINITION));
  _definitions.push(definition);
} 
bool Naming::apply (FormulaList& fs,Where where,Formula* result,ClsRep* infos)
{
  TRACER("Naming::apply (FormulaList&...)");
  bool changed = false;
  Iterator<Formula> fit(fs);
  while (fit.hasNext()) {
    Formula g(fit.next());
    Formula h(g);
    apply(g,where,*infos++);
    if (g != h) {
      changed = true;
    }
    *result++ = g;
  }
  return result;
} 
Formula* Naming::listToArray (FormulaList& fs,int length)
{
  TRACER("Naming::listToArray");
  Formula* forms = new Formula[length];
  Formula* cursor = forms;
  Iterator<Formula> fit(fs);
  while (fit.hasNext()) {
    *cursor++ = fit.next();
  }
  return forms;
} 
int Naming::iffConCls (const ClsRep& lhs,const ClsRep& rhs)
{
  TRACER("Naming::iffConCls");
  return _calc.plus(_calc.mult(lhs.disCls,rhs.conCls),
		    _calc.mult(lhs.conCls,rhs.disCls));
} 
int Naming::iffConRep (const ClsRep& lhs,const ClsRep& rhs)
{
  TRACER("Naming::iffConRep");
  return Int::max(1,
		  _calc.plus(Int::max(_calc.mult(lhs.disRep,rhs.conCls),
				      _calc.mult(rhs.disRep,lhs.conCls)),
			     Int::max(_calc.mult(lhs.conRep,rhs.disCls),
				      _calc.mult(rhs.conRep,lhs.disCls))));
} 
int Naming::iffDisCls (const ClsRep& lhs,const ClsRep& rhs)
{
  TRACER("Naming::iffDisCls");
  return _calc.plus(_calc.mult(lhs.conCls,rhs.conCls),
		    _calc.mult(lhs.disCls,rhs.disCls));
} 
int Naming::iffDisRep (const ClsRep& lhs,const ClsRep& rhs)
{
  TRACER("Naming::iffDisRep");
  return Int::max(1,
		  Int::max(_calc.mult(lhs.conRep,rhs.conCls),
			   _calc.mult(rhs.conRep,lhs.conCls)));
} 
void Naming::arrayToList (FormulaList& writeTo,int length,Formula* array)
{
  TRACER("Naming::arrayToList");
  FormulaList ls;
  for (int i = length-1;i >= 0;i--) {
    ls.push(array[i]);
  }
  writeTo = ls;
} 
