/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Literal.hpp"
#include "Permutation.hpp"
#include "Renaming.hpp"
#include "ShapeComparator.hpp"
#include "Sort.hpp"
#include "Tracer.hpp"
using namespace Shell;
bool Permutation::match (const TermList& ts1,
			 const TermList& ts2,
			 Renaming& ren)
{
  TRACER("Permutation::match (const TermList& ts1...)");
  if (ts1.isEmpty()) {
    return ts2.isEmpty();
  }
  return ts2.isNonEmpty() &&
    match(ts1.head(),ts2.head(),ren) &&
    match(ts1.tail(),ts2.tail(),ren);
} 
bool Permutation::match (const Atom& a1,
			 const Atom& a2,
			 bool swap,
			 Renaming& ren)
{
  TRACER("Permutation::match (const Atom&...)");
  ASS(a1.functor() == a2.functor());
  const TermList& ts1 (a1.args());
  const TermList& ts2 (a2.args());
  if (! swap) {
    return match(ts1,ts2,ren);
  }
  if (! a1.isEquality()) {
    return false;
  }
  return match(ts1.head(),ts2.second(),ren) &&
         match(ts1.second(),ts2.head(),ren);
} 
bool Permutation::match (const Term& t1,
			 const Term& t2,
			 Renaming& ren)
{
  TRACER("Permutation::match (const Term&...)");
  if (t1.tag() != t2.tag()) {
    return false;
  }
  switch (t1.tag()) {
  case VAR:
    return ren.bind(t1.var().toInt(),t2.var().toInt());
  case COMPOUND:
    return t1.functor() == t2.functor() &&
      match(t1.args(),t2.args(),ren);
  case INTERPRETED:
    return t1.value().equals(t2.value());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
bool Permutation::isPermutation (const LiteralList& ls1,
				 const LiteralList& ls2)
{
  TRACER("Permutation::isPermutation");
  int length = ls1.length();
  if (length != ls2.length()) {
    return false;
  }
  if (ls1.isEmpty()) {
    return true;
  }
  Literal* lits1 = sort(ls1,length);
  Literal* lits2 = sort(ls2,length);
  Renaming ren(true);
  return permutation(lits1,lits2,length-1,length-1,false,ren);
} 
Literal* Permutation::sort (const LiteralList& ls,int length)
{
  TRACER("Permutation::sort");
  Permutation perm;
  Sort<Literal,Permutation> sorter(length,perm);
  Iterator<Literal> lits(ls);
  while (lits.hasNext()) {
    sorter.add(lits.next());
  }
  sorter.sort();
  Literal* result = new Literal[length];
  if (! result) {
    ALLOCATION_FAILURE;
  }
  for (int i = length-1;i >= 0; i--) {
    result[i] = sorter[i];
  }
  return result;
} 
bool Permutation::lessThan (const Literal& l1,const Literal& l2)
{
  TRACER("Permutation::lessThan");
  return ShapeComparator::compare(l1,l2) == LESS;
} 
bool Permutation::permutation (Literal* ls1,
			       Literal* ls2,
			       int last,
			       int ind,
			       bool swap,
			       Renaming& ren)
{
  TRACER("Permutation::permutation (Literal* ls1...)");
  ASS(last >= -1);
  ASS(ind >= -1);
  ASS(ind <= last);
  if (last == -1) {
    return true;
  }
  if (ind == -1) {
    return false;
  }
  if (ShapeComparator::compare(ls1[last],ls2[ind]) != EQUAL) {
    return false;
  }
  int bp = ren.backtrackPoint();
  if (match(ls1[last].atom(),ls2[ind].atom(),swap,ren)) {
    Literal last2(ls2[ind]);
    ls2[ind] = ls2[last];
    if (permutation(ls1,ls2,last-1,last-1,false,ren)) {
      return true;
    }
    ls2[ind] = last2;
    if (! swap) {
      if (permutation(ls1,ls2,last,ind,true,ren)) {
	return true;
      }
    }
  }
  ren.backtrack(bp);
  return permutation(ls1,ls2,last,ind-1,false,ren);
} 
