/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "Position.hpp"
#include "Formula.hpp"
#include "VampireException.hpp"
using namespace Shell;
const Formula& Position::subformula (const Formula& f) const
{
  if (isEmpty()) {
    return f;
  }
  const Formula& g = tail().subformula(f);
  int n = head();
  switch (g.connective()) {
  case ATOM:
  case FALSE:
  case TRUE:
    throw NonExistingPositionException();
  case AND:
  case OR: 
    {
      FormulaList fs (g.args());
      while (fs.isNonEmpty()) {
	if (n == 0) {
	  return fs.head();
	}
	fs.pop();
	n--;
      }
      throw NonExistingPositionException();
    }
  case IMP:
  case IFF:
  case XOR:
    switch (n) {
    case 0:
      return g.left();
    case 1:
      return g.right();
    default:
      throw NonExistingPositionException();
    }
  case NOT:
    if (n == 0) {
      return g.uarg();
    }
    throw NonExistingPositionException();
  case FORALL:
  case EXISTS:
    if (n == 0) {
      return g.qarg();
    }
    throw NonExistingPositionException();
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
}
void Position::replace (Formula& f, const Formula& g) const
{
  Position rev;
  reverse(rev);
  rev.replaceReverse(f,g);
} 
void Position::replaceReverse (Formula& f, const Formula& g) const
{
  if (isEmpty()) {
    f = g;
    return;
  }
  switch (f.connective()) {
  case ATOM:
  case FALSE:
  case TRUE:
    throw NonExistingPositionException();
  case AND:
  case OR:
    {
      FormulaList fs (f.args());
      tail().replaceReverse(fs,g,head());
      f = Formula(f.connective(),fs);
    }
    return;
  case IMP:
  case IFF:
  case XOR:
    {
      Formula l(f.left());
      Formula r(f.right());
      switch (head()) {
      case 0:
	tail().replaceReverse(l,g);
	break;
      case 1:
	tail().replaceReverse(r,g);
	break;
      default:
	throw NonExistingPositionException();
      }
      f = Formula(f.connective(),l,r);
    }
    return;
  case NOT:
    {
      if (head() != 0) {
	throw NonExistingPositionException();
      }
      Formula arg(f.uarg());
      tail().replaceReverse(arg,g);
      f = Formula(NOT,arg);
    }
    return;
  case FORALL:
  case EXISTS:
    {
      if (head() != 0) {
	throw NonExistingPositionException();
      }
      Formula arg(f.qarg());
      tail().replaceReverse(arg,g);
      f = Formula(f.connective(),f.vars(),arg);
    }
    return;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
void Position::replaceReverse (FormulaList& fs, const Formula& g, int k) const
{
  ASS (k >= 0);
  if (fs.isEmpty()) {
    throw NonExistingPositionException();
  }
  if (k == 0) {
    Formula f(fs.head());
    replaceReverse(f,g);
    fs = FormulaList(f,fs.tail());
    return;
  }
  FormulaList gs(fs.tail());
  replaceReverse(gs,g,k-1);
  fs = FormulaList(fs.head(),gs);
} 
int Position::polarity (const Formula& f) const
{
  Position rev;
  reverse(rev);
  return rev.polarityReverse(f);
} 
int Position::polarityReverse (const Formula& f) const
{
  if (isEmpty()) {
    return 1;
  }
  switch (f.connective()) {
  case ATOM:
  case FALSE:
  case TRUE:
    throw NonExistingPositionException();
  case AND:
  case OR:
    return tail().polarityReverse(head(),f.args());
  case IMP:
    switch (head()) {
    case 0:
      return -tail().polarityReverse(f.left());
    case 1:
      return tail().polarityReverse(f.right());
    default:
      throw NonExistingPositionException();
    }
  case IFF:
  case XOR:
    switch (head()) {
    case 0:
    case 1:
      return 0;
    default:
      throw NonExistingPositionException();
    }
  case NOT:
    if (head() != 0) {
      throw NonExistingPositionException();
    }
    return -tail().polarityReverse(f.uarg());
  case FORALL:
  case EXISTS:
    if (head() != 0) {
      throw NonExistingPositionException();
    }
    return tail().polarityReverse(f.qarg());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
int Position::polarityReverse (int k, const FormulaList& fs) const
{
  if (fs.isEmpty()) {
    throw NonExistingPositionException();
  }
  if (k == 0) {
    return polarityReverse(fs.head());
  }
  return polarityReverse(k-1,fs.tail());
} 
