/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <ctime>
#include <signal.h>
#include "CASCMode.hpp"
#include "Environment.hpp"
#include "Options.hpp"
#include "Run.hpp"
#include "Statistics.hpp"
#include "System.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
Run::Run (Options& options,const VarList& answerVars)
  : _saved(false),
    _options(options),
    _answerVariables(answerVars)
{
  time_t jobTime = time(static_cast<time_t*>(0));
  tm* gmt = gmtime(&jobTime); 
  int day = gmt->tm_mday;
  int month = gmt->tm_mon + 1;
  int year = 1900 + gmt->tm_year;
  cout <<
    "********************* WARNING! ****************************\n"
    "  This version of Vampire has been prepared and compiled\n"
    "  only for running at CASC-J3. You are not allowed to use\n"
    "  this version for anything else. Please consult the Vampire\n"
    "  home page on how to obtain a licence for running Vampire.\n"
    "***********************************************************\n";
  if (year > 2006 || month > 8) {
    exit(22);
  }
  _timer.reset();
  _timer.start();
  _log = 0;
  if (! (Environment::outputSyntax == OS_KIF ||
	 ! options.outputMessages() ||
	 options.outputSuppressed())) {
    _silent = false;
  }
  else {
    _silent = true;
  }
} 
Run::~Run()
{
  saveLog();
  if (_current == this) {
    _current = 0;
  }
} 
string Run::toString (RunResult krr)
{
  TRACER("Run::toString (RunResult krr)");
  switch (krr) {
  case REFUTATION:
    return "refutation";
  case SATISFIABLE:
    return "satisfiable";
  case UNKNOWN:
    return "unknown";
#if DEBUG_SHELL
  default:
    ASS(false);
    return "";
#endif
  }
} 
void Run::saveLog () 
{
  TRACER("Run::saveLog");
  if (_log && ! _saved) {
    _saved = true;
    _log->append(_options.logFile());
  }
} 
void Run::terminateBySignal (string signalDescription,int exitCode)
{
  TRACER("Run::terminateBySignal (string signalDescription)");
  cancelSignalHandling();
  if (Environment::kernel) {
    Statistics statistics(*Environment::kernel);
    XML::Element* xmlStat = statistics.toXML();
    if (xmlStat && _log) {
      _log->addChild(xmlStat);
    }
    if (_log) {
      XML::Element* err = new XML::TextElement("aborted","");
      err->addAttribute(new XML::TextAttribute("signal",signalDescription));
      _log->addChild(err);
    }
    string resultString = (string)"aborted by signal " + signalDescription;
    statistics.setResult(resultString);
    if (! _silent) {
      statistics.output(cout,_options);
    }
  }
  saveLog();
  exit(exitCode);
} 
void Run::cancelSignalHandling ()
{
  TRACER("cancelSignalHandling");
  signal(SIGTERM,SIG_DFL);
#ifndef _MSC_VER
  signal(SIGQUIT,SIG_DFL);
  signal(SIGHUP,SIG_DFL);
  signal(SIGXCPU,SIG_DFL);
#endif
  signal(SIGINT,SIG_DFL);
#ifndef DEBUG_ERROR_SIGNALS
  signal(SIGILL,SIG_DFL);
  signal(SIGFPE,SIG_DFL);
  signal(SIGSEGV,SIG_DFL);
#ifndef _MSC_VER
  signal(SIGBUS,SIG_DFL);
  signal(SIGTRAP,SIG_DFL);
#endif
  signal(SIGABRT,SIG_DFL);
#endif
} 
void handleSignal (int sigNum)
{
  static bool handled = false;
  string signalDescription = Run::signalToString(sigNum);
  switch (sigNum)
    {
    case SIGTERM:
# ifndef _MSC_VER
    case SIGQUIT:
    case SIGHUP:
    case SIGXCPU:
      if (handled) {
	exit(0);
      }
      handled = true;
      cout << "% Aborted by signal " << signalDescription << "\n";
      if (VampireKernel::currentSession()) {
	cout << "% Trying to shut down current kernel session...\n";
      }
      cout << "% Bailing out now: aborted by interrupt signal " 
	   << signalDescription << "\n";
      Run::currentRun()->terminateBySignal(signalDescription,EXIT_FAILURE);
      return;
# endif
    case SIGINT:
      {
	if (VampireKernel::currentSession())
	  {
	    VampireKernel::currentSession()->suspend();
	  }
	else
	  {
	    cout << "% No kernel session running at the moment.\n";
	  };
	cout << "Continue? (y/n)\n";
	char a;
	cin >> a;
	if ((a == 'n') || (a == 'N'))
	  {
	    cout << "% Terminated by the user.\n";
	    Run::currentRun()->terminateBySignal(signalDescription,
						 EXIT_SUCCESS);
	    return;
	  }
	else
	  {
            if (VampireKernel::currentSession())
	      VampireKernel::currentSession()->resume();
	    cout << "\nContinuing...\n";
	  }
      }
      return;      
    case SIGILL:
    case SIGFPE:
    case SIGSEGV:
# ifndef _MSC_VER
    case SIGBUS:
    case SIGTRAP:
# endif
    case SIGABRT: 
      {
	if (handled) {
	  exit(0);
	}
	handled = true;
	cout << "% Error signal " << signalDescription << " caught\n";
	VampireKernel* kernel = VampireKernel::currentSession ();
	if (kernel) {
	  cout << "% Trying to shut down current kernel session...\n";
	  kernel->reportErrorSignal(sigNum);
	}
#if DEBUG_SHELL
	MyTracer::printStack(cout);
#endif
	cout << "% Bailing out now: aborted by error signal " << signalDescription << "\n";
	Run::currentRun()->terminateBySignal(signalDescription,EXIT_FAILURE);
	return;
      }
    default:
      break;
    };
} 
void Run::setSignalHandlers()
{
  signal(SIGTERM,handleSignal);
#ifndef _MSC_VER
  signal(SIGQUIT,handleSignal);
  signal(SIGHUP,handleSignal);
  signal(SIGXCPU,handleSignal);
#endif
  signal(SIGINT,handleSignal);
#ifndef DEBUG_ERROR_SIGNALS
  signal(SIGILL,handleSignal);
  signal(SIGFPE,handleSignal);
  signal(SIGSEGV,handleSignal);
#ifndef _MSC_VER
  signal(SIGBUS,handleSignal);
  signal(SIGTRAP,handleSignal);
#endif
  signal(SIGABRT,handleSignal);
#endif
} 
string Run::signalToString (int sigNum)
{
  switch (sigNum)
    {
    case SIGTERM: return "SIGTERM";
# ifndef _MSC_VER
    case SIGQUIT: return "SIGQUIT";
    case SIGHUP: return "SIGHUP";
    case SIGXCPU: return "SIGXCPU";
    case SIGBUS: return "SIGBUS";
    case SIGTRAP: return "SIGTRAP";
# endif
    case SIGINT: return "SIGINT";
    case SIGILL: return "SIGILL";
    case SIGFPE: return "SIGFPE";
    case SIGSEGV: return "SIGSEGV";
    case SIGABRT: return "SIGABRT";
    default: return "UNKNOWN SIGNAL";
    };
} 
void Run::logResult (const string terminationReason,RunResult result)
{
} 
void Run::addToLog (XML::Element* element)
{
  if (_log) {
    _log->addChild(element);
  }
  else {
    element->destroy();
  }
} 
