/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Position.hpp"
#include "SimplifyFalseTrue.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
using namespace Shell;
bool SimplifyFalseTrue::simplify (Unit& u)
{
  TRACER("SimplifyFalseTrue::simplify/1");
  ASS(u.unitType() == FORMULA);
  CONTROL("2");
  Formula f(u.formula());
  SimplifyFalseTrue simp;
  simp.simplify(f,Position());
  if (simp._info) { 
    ASS(! f.equals(u.formula()));
    Inference* inf = new Inference(IR_REDUCE_FALSE_TRUE,UnitList(u));
    inf->setInfo(simp._info);
    u = Unit(f,inf);
    return true;
  }
  ASS (f.equals(u.formula()));
  return false;
} 
void SimplifyFalseTrue::simplify (Formula& f, const Position& pos)
{
  TRACER("SimplifyFalseTrue::simplify/2");
  Connective con = f.connective();
  switch (con) {
  case ATOM:
  case TRUE:
  case FALSE:
    return;
  case AND:
  case OR: 
    {
      FormulaList simplifiedArgs (f.args());
      if (simplify(simplifiedArgs,f.connective(),pos,0)) {
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
      }
      else if (simplifiedArgs == f.args()) { 
	return;
      }
      if (simplifiedArgs.hasLength(1)) {
	f = simplifiedArgs.head();
	return;
      }
      f = Formula (f.connective(), simplifiedArgs);
    }
    return;
  case IMP:
  case IFF:
  case XOR: 
    {
      Formula simplifiedLeft(f.left());
      simplify(simplifiedLeft,Position(0,pos));
      Formula simplifiedRight(f.right());
      simplify(simplifiedRight,Position(1,pos));
      Connective lc = simplifiedLeft.connective();
      Connective rc = simplifiedRight.connective();
      if (lc == TRUE ||
	  lc == FALSE ||
	  rc == TRUE ||
	  rc == FALSE) {
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
      }
      switch (lc) {
      case FALSE:
	switch (rc) {
	case FALSE: 
	  f = Formula(con == XOR ? false : true);
	  return;
	case TRUE: 
	  f = Formula(con == IFF ? false : true);
	  return;
	default: 
	  switch (con) {
	  case IMP:
	    f = Formula(true);
	    return;
	  case IFF:
	    f = Formula(NOT,simplifiedRight);
	    return;
	  case XOR: 
	    f = simplifiedRight;
	    return;
#if DEBUG_SHELL
	  default:
	    ASS(false);
#endif
	  }
	}
      case TRUE:
	switch (rc) {
	case FALSE: 
	  f = Formula(con == XOR ? true : false);
	  return;
	case TRUE: 
	  f = Formula(con == XOR ? false : true);
	  return;
	default: 
	  switch (con) {
	  case IMP:
	  case IFF:
	    f = simplifiedRight;
	    return;
	  case XOR: 
	    f = Formula(NOT,simplifiedRight);
	    return;
#if DEBUG_SHELL
	  default:
	    ASS(false);
#endif
	  }
	}
      default:
	switch (rc) {
	case FALSE: 
	  switch (con) {
	  case IMP:
	  case IFF:
	    f = Formula(NOT,simplifiedLeft);
	    return;
	  case XOR: 
	    f = simplifiedLeft;
	    return;
#if DEBUG_SHELL
	  default:
	    ASS(false);
#endif
	  }
	case TRUE:  
	  switch (con) {
	  case IMP:
	    f = Formula(true);
	    return;
	  case IFF:
	    f = simplifiedLeft;
	    return;
	  case XOR: 
	    f = Formula(NOT,simplifiedLeft);
	    return;
#if DEBUG_SHELL
	  default:
	    ASS(false);
#endif
	  }
	default: 
	  if (simplifiedLeft == f.left() && simplifiedRight == f.right()) {
	    return;
	  }
	  f = Formula (f.connective(), simplifiedLeft, simplifiedRight);
	  return;
	}
      }
    }
  case NOT: 
    {
      Formula simplifiedArg (f.uarg());
      simplify(simplifiedArg,Position(0,pos));
      switch (simplifiedArg.connective()) {
      case FALSE:
	f = Formula(true);
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
	return;
      case TRUE:
	f = Formula(false);
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
	return;
      default:
	if (simplifiedArg == f.uarg()) {
	  return;
	}
	f = Formula (NOT, simplifiedArg);
	return;
      }
    }
  case FORALL:
  case EXISTS: 
    {
      Formula simplifiedArg(f.qarg());
      simplify(simplifiedArg,Position(0,pos));
      switch (simplifiedArg.connective()) {
      case FALSE:
	f = Formula(false);
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
	return;
      case TRUE:
	f = Formula(true);
	_info = new InferenceInfo(IR_REDUCE_FALSE_TRUE,pos,_info);
	return;
      default:
	if (simplifiedArg == f.qarg()) {
	  return;
	}
	f = Formula (f.connective(), f.vars(), simplifiedArg);
	return;
      }
    }
  default:
    ASS( false );
    return;
  }
} 
bool SimplifyFalseTrue::simplify (FormulaList& fs, 
				  Connective c, 
				  const Position& pos,
				  int k)
{
  TRACER("SimplifyFalseTrue::simplify/4");
  if (fs.isEmpty()) {
    return false;
  }
  Formula simplifiedHead(fs.head()); 
  simplify(simplifiedHead,Position(k,pos));
  if (fs.tail().isEmpty()) {
    if (simplifiedHead != fs.head()) {
      fs = FormulaList(simplifiedHead);
    }
    return false;
  }
  FormulaList simplifiedTail(fs.tail());
  bool simplified = simplify(simplifiedTail,c,pos,k+1);
  Connective hc = simplifiedHead.connective();
  Connective tc = simplifiedTail.head().connective();
  switch (tc) {
  case FALSE:
    if (c == AND) { 
      fs = simplifiedTail;
      return true;
    }
    fs = FormulaList(simplifiedHead);
    return true;
  case TRUE:
    if (c == OR) { 
      fs = simplifiedTail;
      return true;
    }
    fs = FormulaList(simplifiedHead);
    return true;
  default: 
    switch (hc) {
    case FALSE:
      if (c == AND) { 
	fs = FormulaList(simplifiedHead);
	return true;
      }
      fs = simplifiedTail;
      return true;
    case TRUE:
      if (c == OR) { 
	fs = FormulaList(simplifiedHead);
	return true;
      }
      fs = simplifiedTail;
      return true;
    default:
      if (simplifiedHead == fs.head() && simplifiedTail == fs.tail()) {
	return false;
      }
      fs = FormulaList (simplifiedHead, simplifiedTail);
      return simplified;
    }
  }
} 
