/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Tracer.hpp"
#include "SymCounter.hpp"
#include "Problem.hpp"
#include "Unit.hpp"
#include "Signature.hpp"
#include "VampireException.hpp"
using namespace Shell;
SymCounter::SymCounter (Signature& sig)
  :
  _noOfPreds (sig.noOfPreds()),
  _noOfFuns (sig.noOfFuns()),
  _preds ( new Pred[_noOfPreds] ),
  _funs ( new Fun[_noOfFuns] )
{
  TRACER("SymCounter::SymCounter");
  if (! _preds || ! _funs) {
    ALLOCATION_FAILURE;
  }
  Signature::Iterator ps (sig._pbuckets);
  while (ps.hasNext()) {
    PredicateSymbol& p  = static_cast<PredicateSymbol&>(ps.next());
    _preds[p.number()].pred(p);
  }
  Signature::Iterator fs (sig._fbuckets);
  while ( fs.hasNext() ) {
    FunctionSymbol& f = static_cast<FunctionSymbol&>(fs.next());
    _funs[f.number()].fun(f);
  }
} 
SymCounter::~SymCounter ()
{
  TRACER("SymCounter::~SymCounter");
  delete [] _preds;
  delete [] _funs;
} 
void SymCounter::count (const Problem& p,int c)
{
  UnitChain::Iterator us ( p.units() );
  while ( us.hasNext() ) {
    count (us.next(),c);
  }
} 
void SymCounter::count (const Unit& u,int c)
{
  switch (u.unitType()) {
    case FORMULA:
      count (u.formula(), 1, c);
      return;
    case CLAUSE:
      count (u.clause(), c);
      return;
  }
} 
void SymCounter::count (const Clause& u,int add)
{
  Iterator<Literal> ls (u.literals());
  while ( ls.hasNext() ) {
    count (ls.next(), add);
  }
} 
void SymCounter::count (const Formula& f,int polarity,int add)
{
  switch (f.connective()) {
    case ATOM:
      count (f.atom(), polarity, add);
      return;
    case AND:
    case OR: {
      Iterator<Formula> fs (f.args());
      while (fs.hasNext()) {
        count (fs.next(), polarity, add);
      }
      return;
    }
    case IMP:
      count (f.left(), -polarity, add);
      count (f.right(), polarity, add);
      return;
    case NOT:
      count (f.uarg(), -polarity, add);
      return;
    case IFF:
    case XOR:
      count (f.left(), 0, add);
      count (f.right(), 0, add);
      return;
    case FORALL:
    case EXISTS:
      count (f.qarg(), polarity, add);
      return;
  case TRUE:
  case FALSE:
    return;
    default:
      ASS(false);
      return;
  }
} 
void SymCounter::count (const Literal& l,int add)
{
  count (l.atom(), l.isPositive() ? 1 : -1, add);
} 
void SymCounter::count (const Atom& a,int polarity,int add)
{
  get(a.functor())->add (polarity,add);
  Iterator<Term> ts (a.args());
  while ( ts.hasNext() ) {
    count (ts.next(), add);
  }
} 
void SymCounter::count (const Term& t,int add)
{
  switch (t.tag())
    {
    case VAR:
    case INTERPRETED:
      return;
    case COMPOUND:
      {
	get(t.functor())->add (add);
	Iterator<Term> ts (t.args());
	while ( ts.hasNext() ) {
	  count (ts.next(), add);
	}
      }
    }
} 
void SymCounter::Pred::add (int polarity, int add)
{
  switch (polarity) {
    case -1:
      _nocc += add;
      return;
    case 0:
      _docc += add;
      return;
    case 1:
      _pocc += add;
      return;
    default:
      ASS(false);
      return;
  }
} 
