/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TPTP.hpp"
#include "Unit.hpp"
#include "Int.hpp"
#include "Problem.hpp"
#include "Refutation.hpp"
#include "VampireException.hpp"
using namespace Shell;
string TPTP::toString (const Atom& a)
{
  return a.functor().name() + a.args().toString(",");
} 
string TPTP::toString (const Clause& c)
{
  if (c.isEmpty()) {
    return "[]";
  }
  string result = string("[") + toString(c.literals().head());
  Iterator<Literal> lits (c.literals().tail());
  while (lits.hasNext()) {
    result += string(",\n  ") + toString(lits.next());
  }
  return result + "]";;
} 
string TPTP::toString (const Formula& f)
{
  static string names [] =
    { "", " & ", " | ", " => ", " <=> ", " <~> ", 
      "~", "!", "?", "false", "true"};
  Connective c = f.connective();
  string con = names[(int)c];
  switch (c) {
  case ATOM:
    return toString(f.atom());
  case AND:
  case OR: 
    {
      string result = toString(f.args().head(),c);
      Iterator<Formula> arg (f.args().tail());
      while (arg.hasNext()) {
	result += con + toString(arg.next(),c);
      }
      return result;
    }
  case IMP:
  case IFF:
  case XOR:
    return toString(f.left(),c) + con + toString(f.right(),c);
  case NOT:
    return con + toString(f.uarg(),c);
  case FORALL:
  case EXISTS:
    {
      string result = con + "[" + f.vars().head().toString();
      Iterator<Var> vs(f.vars().tail());
      while (vs.hasNext()) {
	result += string(",") + vs.next().toString();
      }
      return result + "] : " + toString(f.qarg(),c);
    }
  case FALSE:
  case TRUE:
    return con;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toString (const Literal& l)
{
  return string(l.isPositive() ? "++" : "--") + toString(l.atom());
} 
string TPTP::toString (const Unit& u)
{
  static string inputTypes[] = {"axiom","conjecture","hypothesis","derived"};
  string nm = u.name() != "" ? 
              u.name() : 
              string("u_") + Int::toString(u.number());
  switch (u.unitType()) {
  case CLAUSE:
    return string("input_clause(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.clause()) + ").";
  case FORMULA:
    return string("input_formula(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.formula()) + ").";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toStringAsInference (const Unit& u)
{
  return toString(u);
} 
string TPTP::toString (const Formula& f, Connective outer)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toString(f) + ")" :
         toString(f);
} 
string TPTP::toString (const Refutation& ref)
{
  string result = "%=========== Refutation ==========\n";
  Iterator<Unit> units(ref.refutation());
  while (units.hasNext()) {
    result += toStringAsInference(units.next());
  }
  return result + "%======= End of refutation =======\n";
} 
string TPTP::toString (const Problem& problem)
{
  string result("");
  UnitChain::Iterator us (problem.units());
  while (us.hasNext()) {
    result += toString(us.next()) + "\n\n";
  }
  return result;
} 
