/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "TPTPLexer.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
using namespace Shell;
TPTPLexer::TPTPLexer (istream& in)
  : Lexer (in)
{
} 
void TPTPLexer::skipWhiteSpacesAndComments ()
{
  TRACER("TPTPLexer::skipWhiteSpacesAndComments");
  bool comment = false;
  for (;;) {
    if (_eof) {
      return;
    }
    switch (_lastCharacter) {
    case ' ':
    case '\t':
    case '\r':
    case '\f':
      readNextChar();
      break;
    case '\n': 
      comment = false;
      readNextChar();
      break;
    case '%': 
      comment = true;
      readNextChar();
      break;
    default:
      if (comment) {
	readNextChar();
	break;
      }
      return;
    }
  }
} 
void TPTPLexer::readToken (Token& token)
{
  TRACER("TPTPLexer::readToken");
  skipWhiteSpacesAndComments();
  _charCursor = 0;
  if (_eof) {
    token.tag = TT_EOF;
    token.text = "";
    return;
  }
  token.line = _lineNumber;
  switch (currentCharacterType()) {
  case UPPER_CASE_LETTER:
    readName(token);
    token.tag = TT_VAR;
    return;
  case LOWER_CASE_LETTER:
    readName(token);
    token.tag = detectNameTokenTypeOfLastToken();
    return;
  case DIGIT:
    readNumber(token);
    return;
  case SINGLE:
    readSingle(token);
    return;
  case SYMBOLIC:
    readSymbolic(token);
    token.tag = detectSymbolicTokenTypeOfLastToken();
    return;
  case QUOTE:
    readQuotedString(token);
    return;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
TPTPLexer::CharType TPTPLexer::currentCharacterType () const
{
  TRACER("TPTPLexer::currentCharacterType");
  switch (_lastCharacter) {
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
  case '_':
    return UPPER_CASE_LETTER;
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
  case 'g':
  case 'h':
  case 'i':
  case 'j':
  case 'k':
  case 'l':
  case 'm':
  case 'n':
  case 'o':
  case 'p':
  case 'q':
  case 'r':
  case 's':
  case 't':
  case 'u':
  case 'v':
  case 'w':
  case 'x':
  case 'y':
  case 'z':
  case '$':
    return LOWER_CASE_LETTER;
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    return DIGIT;
  case '(':
  case ')':
  case '[':
  case ']':
  case ',':
    return SINGLE;
  case '.':
  case '?':
  case '!':
  case '<':
  case '>':
  case '+':
  case '-':
  case '=':
  case ':':
  case '&':
  case '~':
  case '|':
    return SYMBOLIC;
  case '\'':
    return QUOTE;
  case ' ':
  case '\t':
  case '\r':
  case '\f':
  case '\n':
    return WHITE;
  case '%':
    return OTHER;
  default:
    throw LexerException((string)"illegal character " + (char)_lastCharacter,
			 *this);
  }
} 
void TPTPLexer::readSymbolic (Token& token)
{
  TRACER("TPTPLexer::readSymbolic");
  saveLastChar();
  while (readNextChar()) {
    switch (currentCharacterType()) {
    case SYMBOLIC:
      saveLastChar();
      break;
    default:
      saveTokenText(token);
      return;
    }
  }
  saveTokenText(token);
} 
void TPTPLexer::readSingle (Token& token)
{
  TRACER("TPTPLexer::readSingle");
  switch (_lastCharacter) {
  case '(':
    token.tag = TT_LPAR;
    break;
  case ')':
    token.tag = TT_RPAR;
    break;
  case '[':
    token.tag = TT_LBRA;
    break;
  case ']':
    token.tag = TT_RBRA;
    break;
  case ',':
    token.tag = TT_COMMA;
    break;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
  saveLastChar();
  readNextChar();
#if DEBUG_SHELL
  saveTokenText(token);
#endif
} 
void TPTPLexer::readName (Token& token)
{
  TRACER("TPTPLexer::readName");
  saveLastChar();
  while (readNextChar()) {
    switch (currentCharacterType()) {
    case UPPER_CASE_LETTER:
    case LOWER_CASE_LETTER:
    case DIGIT:
      saveLastChar();
      break;
    default:
      saveTokenText(token);
      return;
    }
  }
  saveTokenText(token);
} 
TokenType TPTPLexer::detectNameTokenTypeOfLastToken ()
{
  TRACER("TPTPLexer::detectNameTokenTypeOfLastToken");
  switch (_charBuffer[0]) {
  case 'c':
    if (! strcmp(_charBuffer,"cnf")) {
      return TT_CNF;
    }
    return TT_NAME;    
  case 'f':
    if (! strcmp(_charBuffer,"false")) {
      return TT_FALSE;
    }
    if (! strcmp(_charBuffer,"fof")) {
      return TT_INPUT_FORMULA;
    }
    return TT_NAME;
  case 'i':
    if (! strcmp(_charBuffer,"input_formula")) {
      return TT_INPUT_FORMULA;
    }
    if (! strcmp(_charBuffer,"input_clause")) {
      return TT_INPUT_CLAUSE;
    }
    if (! strcmp(_charBuffer,"include")) {
      return TT_INCLUDE;
    }
    return TT_NAME;
  case 't':
    if (! strcmp(_charBuffer,"true")) {
      return TT_TRUE;
    }
    return TT_NAME;
  default:
    return TT_NAME;
  }
} 
TokenType TPTPLexer::detectSymbolicTokenTypeOfLastToken ()
{
  TRACER("TPTPLexer::detectSymbolicTokenTypeOfLastToken");
  switch (_charBuffer[0]) {
  case ':':
    if (_charBuffer[1]) {
      break;
    }
    return TT_COLON;
  case '.':
    if (_charBuffer[1]) {
      break;
    }
    return TT_DOT;
  case '&':
    if (_charBuffer[1]) {
      break;
    }
    return TT_AND;
  case '~':
    if (_charBuffer[1]) {
      break;
    }
    return TT_NOT;
  case '|':
    if (_charBuffer[1]) {
      break;
    }
    return TT_OR;
  case '!':
    if (! strcmp(_charBuffer+1,"=")) {
      return TT_NEQ;
    }
    if (_charBuffer[1]) {
      break;
    }
    return TT_FORALL;
  case '?':
    if (_charBuffer[1]) {
      break;
    }
    return TT_EXISTS;
  case '<':
    if (! strcmp(_charBuffer+1,"=>")) {
      return TT_IFF;
    }
    if (! strcmp(_charBuffer+1,"=")) {
      return TT_RIMP;
    }
    if (! strcmp(_charBuffer+1,"~>")) {
      return TT_XOR;
    }
    break;
  case '=':
    if (! _charBuffer[1]) {
      return TT_EQUAL;
    }
    if (! strcmp(_charBuffer+1,">")) {
      return TT_IMP;
    }
    break;
  case '+':
    if (! strcmp(_charBuffer+1,"+")) {
      return TT_PP;
    }
    break;
  case '-':
    if (! strcmp(_charBuffer+1,"-")) {
      return TT_MM;
    }
    break;
  default:
    break;
  }
  throw LexerException((string)"illegal symbolic character " + _charBuffer,
		       *this);
} 
void TPTPLexer::readQuotedString (Token& token)
{
  TRACER("TPTPLexer::readQuotedString");
  while (readNextChar()) {
    if (_lastCharacter == '\'') {
      saveTokenText(token);
      readNextChar();
      token.tag = TT_QUOTED_STRING;
      return;
    }
    else {
      saveLastChar();
    }
  }
  throw LexerException((string)"file ended while reading quoted string ",
		       *this);
} 
