/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Term__
#define __Term__
#include <string>
#include "Var.hpp"
#include "VampireKernelEnvironment.hpp"
#include "XML.hpp"
using namespace std;
using namespace VampireKernelEnvironment;
namespace Shell {
class TermList;
class FunctionSymbol;
class Substitution;
enum TermType {
  VAR = 0u,        
  INTERPRETED = 1u,
  COMPOUND = 2u
};
class Term 
{
 private:
  class Data;
  class VarData;
  class CompoundData;
  class InterpretedData;
 public:
  Term () : _data(0) {}
  Term (const FunctionSymbol& f, const TermList& args); 
  Term (const FunctionSymbol& f);                       
  Term (const Term& t);
  explicit Term (int value);
  explicit Term (double value);
  explicit Term (const string& value);
  ~Term ();
  explicit Term (Var v);
  explicit Term (const InterpretedConstant&);
  void operator= (const Term& rhs);
  TermType tag () const;
  const Var& var () const;
  const InterpretedConstant& value () const;
  const FunctionSymbol& functor () const;
  const TermList& args () const;
  bool operator == (const Term& rhs) const;
  bool operator != (const Term& rhs) const;
  void* operator new (size_t);
  XML::Element* toXML () const;
  bool occurs (Var v) const;
  bool equals (const Term& t) const;
  bool hasVarsNotIn (const VarListList&) const;
  string toString () const;
 private:
  Data* _data;
}; 
class TermList 
  : public Lst<Term>
{
 public:
  TermList () : Lst<Term> () {}
  TermList (const TermList& ts) : Lst<Term> (ts) {}
  explicit TermList (const Term& t);
  TermList (const Term& head, const TermList& tail);
  explicit TermList (LstData<Term>* d) : Lst<Term> (d) {}
  explicit TermList (const VarList&);
  const TermList& tail () const
    { return static_cast<const TermList&>(Lst<Term>::tail()); }
  bool occurs (Var v) const;
  bool equals (const TermList& ts) const;
  bool hasVarsNotIn (const VarListList& ) const;
  bool isLinearSequenceOfVariables () const;
  string toString (string separator) const;
}; 
class Term::Data 
{
private:
  Data ();
public:
  explicit Data (TermType tag);
  ~Data () {}
  TermType tag () const { return _tag; }
  void ref () { _counter++; }
  void deref ();
  void destroy ();
private:
  void* operator new (size_t);
 protected:
  unsigned int _counter : 30;
  TermType _tag : 2;
}; 
class Term::VarData 
: public Term::Data
{
 public:
  explicit VarData (Var v);
  const Var& var () const {ASS(_tag == VAR); return _var;}
  CLASS_NAME("Term::VarData");
  USE_ALLOCATOR(VarData);
 private:
  Var _var;
}; 
class Term::CompoundData 
: public Term::Data
{
 public:
  CompoundData (const FunctionSymbol& f, const TermList& args);
  explicit CompoundData (const FunctionSymbol& f);
  const FunctionSymbol& functor () const {return _functor;}
  const TermList& args () const {return _args;}
  CLASS_NAME("Term::CompoundData");
  USE_ALLOCATOR(CompoundData);
 private:
  const FunctionSymbol& _functor;
  TermList _args;
}; 
class Term::InterpretedData 
: public Term::Data
{
 public:
  explicit InterpretedData (const InterpretedConstant&);
   const InterpretedConstant& value () const { return _value; }
  CLASS_NAME("Term::InterpretedData");
  USE_ALLOCATOR(InterpretedData);
 private:
  InterpretedConstant _value;
}; 
inline
Term::Term (const Term& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 
inline
Term::Term (const FunctionSymbol& f, const TermList& args)
  :
  _data (new CompoundData (f,args))
{
} 
inline
Term::Term (const FunctionSymbol& f)
  :
  _data (new CompoundData (f))
{
} 
inline
TermType Term::tag () const 
{ 
  return _data->tag(); 
} 
inline
bool Term::operator == (const Term& rhs) const
{ 
  ASS (_data && rhs._data);
  return _data == rhs._data; 
} 
inline
bool Term::operator != (const Term& rhs) const
{ 
  ASS (_data && rhs._data);
  return _data != rhs._data; 
} 
inline
const Var& Term::var () const
{ 
  ASS (this && tag() == VAR);
  return (static_cast<Term::VarData*>(_data))->var();
} 
inline
const FunctionSymbol& Term::functor () const
{ 
  ASS (this && tag() == COMPOUND);
  return (static_cast<Term::CompoundData*>(_data))->functor();
} 
inline
const TermList& Term::args () const
{ 
  ASS (this && tag() == COMPOUND);
  return (static_cast<Term::CompoundData*>(_data))->args();
} 
inline
const InterpretedConstant& 
Term::value () const
{ 
  ASS (this && tag() == INTERPRETED);
  return (static_cast<Term::InterpretedData*>(_data))->value();
} 
inline
Term::Data::Data (TermType tag)
  : 
  _counter (1),
  _tag (tag)
{
} 
}
#endif 
