/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string.h>
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Int.hpp"
#include "ShellReader.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
long Unit::_lastNumber = 0;
UnitList::UnitList () 
  : Lst<Unit> ()
{
} 
UnitList::UnitList (const Unit &hd)
  : Lst<Unit> (hd)
{
} 
UnitList::UnitList (const Unit &hd, const UnitList& tl)
  :
  Lst<Unit> (hd,tl)
{
} 
UnitList::UnitList (LstData<Unit>* d)
  :
  Lst<Unit> (d)
{
} 
UnitList::UnitList (const UnitList& ts)
  :
  Lst<Unit> (ts)
{
} 
void Unit::operator = (const Unit& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
void Unit::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;
  if (_counter == 0) {
    switch (_untype) {
    case CLAUSE:
      delete static_cast<ClauseData*>(this);
      return;
    case FORMULA:
      delete static_cast<FormulaData*>(this);
      return;
#if DEBUG_SHELL
    default:
      ASSERTION_VIOLATION;
#endif
    }
  } 
} 
string Unit::toString(InputType t)
{
  TRACER("Unit::toString(InputType t)");
  switch (t) {
  case AXIOM:
    return "axiom";
  case CONJECTURE:
    return "conjecture";
  case ASSUMPTION:
    return "assumption";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
    return 0;
#endif
  }
} 
string Unit::toString(UnitType t)
{
  TRACER("Unit::toString(UnitType t)");
  switch (t) {
  case CLAUSE:
    return "clause";
  case FORMULA:
    return "formula";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
    return 0;
#endif
  }
} 
string Unit::toString(DefType dt)
{
  TRACER("Unit::toString(DefType dt)");
  switch (dt) {
  case DT_NONE:
    return "none";
  case DT_PRED:
    return "pred";
  case DT_FUN:
    return "fun";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
    return 0;
#endif
  }
} 
ostream& operator << (ostream& str, InputType t)
{
  TRACER("operator << (ostream& str, InputType t)");
  str << Unit::toString(t);
  return str;
} 
ostream& operator << (ostream& str, UnitType t)
{
  TRACER("operator << (ostream& str, UnitType t)");
  switch ( t ) {
    case CLAUSE:
      str << "input_clause";
      break;
    case FORMULA:
      str << "input_formula";
      break;
  }
  return str;
} 
Unit::Data::Data (UnitType tp,const string& name, InputType intype)
  :
  _counter (1),
  _untype (tp),
  _intype (intype),
  _defType (DT_NONE),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _name (name),
  _inference (new Inference(IR_INPUT))
{
  TRACER( "Unit::Data::Data-1" );
} 
Unit::ClauseData::ClauseData (const string& name,
			      InputType intype,
			      const Clause& c)
  : Data(CLAUSE,name,intype),
    _clause (c)
{
  TRACER( "Unit::ClauseData::ClauseData-1" );
} 
Unit::FormulaData::FormulaData (const string& name,
				InputType intype,
				const Formula& f)
  : Data(FORMULA,name,intype),
    _formula (f)
{
  TRACER( "Unit::FormulaData::FormulaData-1" );
} 
Unit::Data::Data (UnitType tp,Inference* inf,long number)
  : _counter(1),
    _untype(tp),
    _intype(inf->premises().maxInputType()),
    _defType(DT_NONE),
    _number(number),
    _numberInRefutation(0),
    _inference(inf)
{
  TRACER( "Unit::Data::Data-2" );
}
Unit::FormulaData::FormulaData (const Formula& f, Inference* inf)
  : Data(FORMULA,inf,++_lastNumber),
    _formula(f)
{
  TRACER( "Unit::FormulaData::FormulaData-2" );
}
Unit::ClauseData::ClauseData (const Clause& c,Inference* inf)
  : Data(CLAUSE,inf,++_lastNumber),
    _clause(c)
{
  TRACER("Unit::ClauseData::ClauseData-2");
}
Unit::Data::Data (UnitType tp,InferenceRule ir, const Unit& parent)
  :
  _counter (1),
  _untype (tp),
  _intype (parent.inputType()),
  _defType (DT_NONE),
  _number (++ _lastNumber),
  _numberInRefutation(0),
  _inference (new Inference(ir,parent))
{
  TRACER( "Unit::Data::Data-3" );
} 
Unit::FormulaData::FormulaData (InferenceRule ir, 
				const Formula& f, 
				const Unit& parent)
  : Data(FORMULA,ir,parent),
    _formula (f)
{
  TRACER( "Unit::FormulData::FormulaData-3" );
} 
Unit::ClauseData::ClauseData (InferenceRule ir,
			      const Clause& c, 
			      const Unit& parent)
  : Data(CLAUSE,ir,parent),
    _clause(c)
{
  TRACER( "Unit::ClauseData::ClauseData-3" );
} 
Unit::ClauseData::ClauseData (const Clause& c,
			      const UnitList& parents,
			      long number,
			      VampireKernel::Clause::BackgroundFlagIterator& flag)
  : Data(CLAUSE,new Inference(parents,flag),number),
    _clause(c)
{
  TRACER( "Unit::ClauseData::ClauseData-4" );
} 
Unit::FormulaData::FormulaData (const Formula& f,long number)
  : Data(FORMULA,new Inference(IR_PREDICATE_DEFINITION),number),
    _formula(f)
{
  TRACER( "Unit::FormulaData::FormulaData-4" );
} 
Unit::FormulaData::FormulaData (const Formula& f, const UnitList& parents)
  : Data(FORMULA,new Inference(IR_KERNEL,parents),++_lastNumber),
    _formula(f)
{
  TRACER( "Unit::FormulaData::FormulaData-5" );
} 
Unit::Data::~Data ()
{
  TRACER( "Unit::Data::~Data" );
  ASS (_counter == 0);
  delete _inference;
} 
XML::Element* Unit::toXML() const
{
  TRACER("Unit::toXML()");
  XML::DeepElement* unit = new XML::DeepElement("unit");
  unit->addAttribute(new XML::TextAttribute("input_type",
					    toString(inputType())));
  unit->addAttribute(new XML::LongAttribute("number",number()));
  if (numberInRefutation()) {
    unit->addAttribute(new XML::IntegerAttribute("refutation_number",
						 numberInRefutation()));
  }
  unit->addChild(new XML::TextElement("is_definition",
					   toString(defType())));
  if (name() != "") {
    unit->addChild(new XML::TextElement("name",name()));
  }
  unit->addChild(inference()->toXML());
  if (unitType() == CLAUSE) {
    unit->addChild(new XML::TextElement("clause",clause().toString()));
  }
  else { 
    unit->addChild(new XML::TextElement("formula",formula().toString()));
  }
  return unit;
} 
string Unit::toString () const
{
  TRACER("Unit::toString");
  string prefix = Int::toString(number()) + ". ";
  string postfix = string(" (") + toString(inputType());
  if (name() != "") {
    postfix += string(",name=") + name();
  }
  postfix += ")";
  switch (unitType()) {
  case CLAUSE:
    return prefix + clause().toString() + postfix;
  case FORMULA:
    return prefix + formula().toString() + postfix;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string Unit::toStringAsInference () const
{
  TRACER("Unit::toStringAsInference");
  static string stars = "***********";
  UnitList parents (inference()->premises());
  string result = stars + " [";
  Iterator<Unit> ps1 (parents);
  while (ps1.hasNext()) {
    result += Int::toString(ps1.next().number());
    if (ps1.hasNext()) {
      result += ",";
    }
  }
  if (parents.isNonEmpty()) {
    result += "->";
  }
  Inference* inf = inference();
  string infString = inf->rule() == IR_KERNEL ?
    static_cast<const KernelInferenceInfo*>(inf->info())->rulesAsString() :
    Inference::toString(inf->rule());
  result += Int::toString(number()) + ", " + 
            infString + "] " + 
            stars + "\n";
  Iterator<Unit> ps (parents);
  while (ps.hasNext()) {
    result += "  ";
    Unit nxt (ps.next());
    switch (nxt.unitType()) {
    case CLAUSE:
      result += nxt.clause().toString();
      break;
    case FORMULA:
      result += nxt.formula().toString();
      break;
    }
    result += "\n";
  }
  if (parents.isNonEmpty()) {
    result += "-----------------------------\n  ";
  }
  switch (unitType()) {
  case CLAUSE:
    result += clause().toString();
    break;
  case FORMULA:
    result += formula().toString();
    break;
  }
  return result + "\n";
} 
InputType Unit::max (InputType type1, InputType type2)
{
  return type1 > type2 ? type1 : type2;
}
InputType UnitList::maxInputType () const
{
  TRACER("UnitList::maxInputType");
  InputType result = AXIOM;
  Iterator<Unit> us(*this);
  while (us.hasNext()) {
    switch (us.next().inputType()) {
    case AXIOM:
      break;
    case ASSUMPTION:
      result = ASSUMPTION;
      break;
    case CONJECTURE:
      return CONJECTURE;
    }
  }
  return result;
} 
