/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef BS_SUBSTITUTION_CACHE_HPP
#define BS_SUBSTITUTION_CACHE_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#ifndef NO_DEBUG
#  include "LinearCombination.hpp"
#endif
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_SUBSTITUTION_CACHE
#define DEBUG_NAMESPACE "BSSubstitutionCache"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSSubstitutionCache
  {
  public: 
    class Instance;
  public:
    BSSubstitutionCache() {};
    ~BSSubstitutionCache() {};
    void reset() 
    {
      DOP(_debugVariables.reset());
    };
    void associate(ulong var,const Term& term)
    {
      CALL("associate(ulong var,const Term& term)");
      ASSERT(!_debugVariables.coefficient(var));
      _subst[var] = term;
      DOP(_debugVariables.add1(var));
    };
    const Term& map(ulong var) const 
    {
      CALL("map(ulong var) const");
      ASSERT(_debugVariables.coefficient(var));
      return _subst[var]; 
    };
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
      for (const ulong* v = _debugVariables.begin(); v < _debugVariables.end(); ++v)
	str << "X" << *v << " ---> " << map(*v) << "\n";
      return str;
    };
#endif
  private:
    Term _subst[VampireKernelConst::MaxNumOfVariables];
#if (defined DEBUG_NAMESPACE) || (!(defined NO_DEBUG_VIS))
    BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _debugVariables;
#endif
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,const VK::BSSubstitutionCache* subst)
  {
    return subst->output(str);
  };
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_SUBSTITUTION_CACHE
#define DEBUG_NAMESPACE "BSSubstitutionCache::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class BSSubstitutionCache::Instance
  {
  public:
    Instance() : 
      _dummyTerm0((Term::Arity)0,(Term::Functor)0,Signature::PositivePolarity)  
    {
      _dummyTerm1[0].MakeComplex((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity);
    };
    ~Instance() {};
    void reset(const Term& t,const BSSubstitutionCache* subst);
    void reset(const Term* t,const BSSubstitutionCache* subst);
    const Term& symbol() const;
    bool next();
    static 
    bool 
    equal(const Term& t1,const Term& t2,const BSSubstitutionCache* subst);
    static 
    bool 
    equal(const Term* t1,const Term* t2,const BSSubstitutionCache* subst);
  private:
    const BSSubstitutionCache* _subst;
    Term::Iterator _topLevel;
    Term::Iterator _substIter;
    bool _onTop;
    Term _dummyTerm0;
    Term _dummyTerm1[2];
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BS_SUBSTITUTION_CACHE
#define DEBUG_NAMESPACE "BSSubstitutionCache::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void 
  BSSubstitutionCache::Instance::reset(const Term& t,
				       const BSSubstitutionCache* subst)
  {
    CALL("reset(const Term& t,const BSSubstitutionCache* subst)");
    if (t.isVariable())
      {
	_subst = subst;
	if (subst->map(t.var()).isVariable())
	  {
	    _dummyTerm1[1] = subst->map(t.var());
	    _substIter.Reset(_dummyTerm1);
	    ALWAYS(_substIter.Next());
	  }
	else
	  {
	    ASSERT(subst->map(t.var()).IsReference());
	    _substIter.Reset(subst->map(t.var()).deref());
	  };
	_topLevel.Reset(&_dummyTerm0);
	NEVER(_topLevel.Next());
	_onTop = false;
      }
    else
      {
	ASSERT(t.IsReference());
	reset(t.deref(),subst);
      };
  }; 
  inline
  void 
  BSSubstitutionCache::Instance::reset(const Term* t,
				       const BSSubstitutionCache* subst)
  {
    CALL("reset(const Term* t,const BSSubstitutionCache* subst)");
    ASSERT(t->isComplex());
    _subst = subst;
    _topLevel.Reset(t);
    _onTop = true;
  }; 
  inline const Term& BSSubstitutionCache::Instance::symbol() const
  {
    CALL("symbol() const");
    if (_onTop) return _topLevel.symbolRef();
    return _substIter.symbolRef();
  }; 
  inline bool BSSubstitutionCache::Instance::next()
  {
    CALL("next()");
    if (_onTop)
      {
      next_on_top:
	if (!_topLevel.Next()) 
	  {
	    return false;
	  };
	if (_topLevel.CurrentSymIsVar())
	  {
	    _onTop = false;
	    if (_subst->map(_topLevel.symbolRef().var()).isVariable())
	      {
		_dummyTerm1[1] = _subst->map(_topLevel.symbolRef().var());
		_substIter.Reset(_dummyTerm1);
		ALWAYS(_substIter.Next());
	      }
	    else
	      {
		ASSERT(_subst->map(_topLevel.symbolRef().var()).IsReference());
		_substIter.Reset(_subst->map(_topLevel.symbolRef().var()).deref());
	      };
	  };
	return true;
      }
    else 
      {
	if (_substIter.Next())
	  {
	    return true;
	  };
	_onTop = true;
	goto next_on_top;
      };
  }; 
  inline
  bool 
  BSSubstitutionCache::Instance::equal(const Term& t1,
				       const Term& t2,
				       const BSSubstitutionCache* subst)
  {
    CALL("equal(const Term& t1,..)");
    Instance iter1;
    Instance iter2;
    iter1.reset(t1,subst);
    iter2.reset(t2,subst);
    do
      {
	if (iter1.symbol() != iter2.symbol()) return false;
	iter2.next();
      }
    while (iter1.next());
    return true;
  }; 
  inline
  bool 
  BSSubstitutionCache::Instance::equal(const Term* t1,
				       const Term* t2,
				       const BSSubstitutionCache* subst)
  {
    CALL("equal(const Term* t1,..)");
    ASSERT(t1->isComplex() && t2->isComplex()); 
    Instance iter1;
    Instance iter2;
    iter1.reset(t1,subst);
    iter2.reset(t2,subst);
    do
      {
	if (iter1.symbol() != iter2.symbol()) return false;
	iter2.next();
      }
    while (iter1.next());
    return true;
  }; 
}; 
#endif
