/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BackDemodulation.hpp"
#include "Clause.inl"
#include "DestructionMode.hpp"
#include "TemporaryClauseCollector.hpp"
using namespace BK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
#define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.hpp"        
VK::BackDemodulation::BackDemodulation() : 
  _bDOnSplittingBranchesFlag(false),
  _clauseCollector(0),
  _splittingBranches(DOP("BackDemodulation::_splittingBranches")),
  _numOfBDOnSplittingBranches(0UL)
{
  CALL("constructor BackDemodulation()");
  DOP(_splittingBranches.freeze());
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForInstanceRetrieval();
  unlock();
}; 
VK::BackDemodulation::~BackDemodulation() 
{ 
  CALL("destructor ~BackDemodulation()");  
  if (DestructionMode::isThorough()) 
    {
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForInstanceRetrieval(_compiledConstraint);
    };
}; 
void VK::BackDemodulation::reset()
{
  CALL("reset()");
  BK::destroy(*this);
  BK::init(*this);
}; 
void VK::BackDemodulation::integrate(Clause* cl)
{
  CALL("integrate(Clause* cl)");   
  ASSERT(!isLocked());
  lock();
  cl->protect();
  Term* lit;
  ulong ln = 0;
  for(LiteralList::Iterator litIter(cl->LitList());
      litIter.notEnd();
      litIter.next())
    {
      lit = litIter.currentLiteral();
      if (lit->arity())
	{
	  integrateLiteral(lit,ln,cl);
	};         
      ln++;
    };
  cl->unprotect();
  unlock();
}; 
inline
void VK::BackDemodulation::integrateLiteral(Term* lit,ulong litNum,Clause* cl)
{
  CALL("integrateLiteral(Term* lit,ulong litNum,Clause* cl)"); 
  ASSERT(isLocked());
  Term* arg = lit->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = lit->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      integrateComplexTerm(arg->deref(),litNum,cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable())
	    {
	      integrateComplexTerm(arg->deref(),litNum,cl);
	    };
	  arg = arg->Next();
	};
    };
}; 
void VK::BackDemodulation::integrateComplexTerm(Term* term,ulong litNum,Clause* cl)
{
  CALL("integrateComplexTerm(Term* term,ulong litNum,Clause* cl)");
  ASSERT(isLocked());
  ASSERT(term->isComplex());
  term->RegisterOccurence(cl,litNum);
  _index.integrate(term);
  Term* arg = term->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = term->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      integrateComplexTerm(arg->deref(),litNum,cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable())
	    {
	      integrateComplexTerm(arg->deref(),litNum,cl);
	    };
	  arg = arg->Next();
	};
    };
}; 
void VK::BackDemodulation::remove(Clause* cl) 
{
  CALL("remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  Term* lit;
  for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
    {
      lit = litIter.currentLiteral();
      if (lit->arity())
	{
	  removeLiteral(lit,cl);
	};   
    };  
  cl->unprotect();
  unlock();
}; 
inline
void VK::BackDemodulation::removeLiteral(Term* lit,Clause* cl)
{
  CALL("removeLiteral(Term* lit,Clause* cl)");
  ASSERT(isLocked());
  Term* arg = lit->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = lit->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      removeComplexTerm(arg->deref(),cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable())
	    {
	      removeComplexTerm(arg->deref(),cl);
	    };
	  arg = arg->Next();
	};
    };
}; 
void VK::BackDemodulation::removeComplexTerm(Term* term,Clause* cl)
{
  CALL("removeComplexTerm(Term* term,Clause* cl)");
  ASSERT(isLocked());
  ASSERT(term->isComplex());
  term->RemoveOccurences(cl);
  if (!term->Occurences()) 
    _index.remove(term);
  Term* arg = term->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = term->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      removeComplexTerm(arg->deref(),cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->isNotArg())
	{
	  if (!arg->isVariable())
	    {
	      removeComplexTerm(arg->deref(),cl);
	    };
	  arg = arg->Next();
	};
    };
}; 
bool VK::BackDemodulation::eligibleAsQuery(const Clause* cl) const
{
  CALL("eligibleAsQuery(const Clause* cl) const");
  if (cl->Unit())
    {
      const LiteralList::Element* elem = cl->LitList()->firstElem();
      const Term* lit = elem->literal();
      if (lit->Negative() || !lit->IsEquality()) return false;
      if (!elem->isOrientableEq()) return false;
      if (!elem->isOrientedEq() && _degree <= 1) return false; 
    }
  else 
    {
      if (!_bDOnSplittingBranchesFlag) return false;
      if (cl->LitList()->numOfSelectedLits() != 1) return false;
      const LiteralList::Element* elem = cl->LitList()->firstElem(); 
      const Term* lit = elem->literal();
      if (lit->Negative() || !lit->IsEquality()) return false;
      if (!elem->isOrientableEq()) return false;
      if (!elem->isOrientedEq() && _degree <= 1) return false;     
      LiteralList::Iterator iter(cl->LitList());
      iter.next(); 
      while (iter.notEnd())
	{
	  if (!iter.currentLiteral()->IsSplitting()) return false;
	  iter.next();
	};
    };
  return true;
}; 
bool VK::BackDemodulation::query(Clause* cl) 
{ 
  CALL("query(Clause* cl)");  
  ASSERT(!isLocked());
  lock();
  cl->protect();
  if (cl->flag(Clause::FlagContainsAnswerLiterals)) goto return_false;
  if (cl->isSupportClause() && (!_demodulateFromSupport)) goto return_false;
  if (!cl->isPersistent()) goto return_false;;
  if (cl->Unit() || 
      (_bDOnSplittingBranchesFlag && 
       (cl->LitList()->numOfSelectedLits() == 1)))
    {   
      const LiteralList::Element* elem = cl->LitList()->firstElem(); 
      if (!elem->isOrientableEq()) goto return_false;
      const Term* lit = elem->literal();
      if (lit->Positive() && lit->IsEquality())
	{
	  if (_bDOnSplittingBranchesFlag)
	    {
	      _splittingBranches.reset();
	      LiteralList::Iterator iter(cl->LitList());
	      iter.next();
	      DOP(_splittingBranches.unfreeze());
	      while (iter.notEnd())
		{
		  if (iter.currentLiteral()->IsSplitting()) 
		    {
		      ASSERT(!iter.currentLiteral()->arity());
		      _splittingBranches.addSafe(iter.currentLiteral()->HeaderNum()); 
		    }
		  else goto return_false;;
		  iter.next();
		};
	      DOP(_splittingBranches.freeze());
	    }
	  else 
	    if (!cl->Unit()) goto return_false;;
	  _rule = cl;
	  _candidates = 0;
	  if (elem->isOrientedEq())
	    {
	      _swapQuerySides = false;
	      if (elem->orientationFollowsOrdering())
		{
		  ASSERT(!elem->equalityLHS()->isVariable());
		  if (_checkSubstSafetyWRTInputNonBasicness &&
		      !safeRuleWRTInputNonBasicness(*(elem->equalityLHS()),
						    *(elem->equalityRHS())))
		    goto return_false;
		  _ruleLHS = elem->equalityLHS()->deref();
		  _ruleRHS = *(elem->equalityRHS());
		use_without_constraint:
		  if (!_index.newQuery(_ruleLHS))
		    {
		      goto return_false;
		    };
		  _currentConstraint = 0;
		  ASSERT(isLocked());
		  ASSERT(cl->isProtected());
		  return true;
		}
	      else 
		{
		  if (elem->equalityLHS()->isVariable() ||
		      elem->equalityRHS()->isVariable()) 
		    goto return_false;
		  if (_checkSubstSafetyWRTInputNonBasicness &&
		      !safeRuleWRTInputNonBasicness(*(elem->equalityLHS()),
						    *(elem->equalityRHS())))
		    goto return_false;
		  _ruleLHS = elem->equalityLHS()->deref();
		  _ruleRHS = *(elem->equalityRHS());
		  _compiledConstraint->loadNotEqual(_ruleLHS,_ruleRHS.deref());
		  if (!_compiledConstraint->lhsVarInstMayOrder()) 
		    goto return_false;
		  ASSERT(_compiledConstraint->category() != 
			 SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Contradiction);
		  if (_compiledConstraint->category() == 
		      SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology)
		    {
		      goto use_without_constraint;
		    };
		  if (_index.newQuery(_ruleLHS))
		    {
		      _compiledConstraint->compile(_index.substitution());
		      _currentConstraint = _compiledConstraint;
		      ASSERT(isLocked());
		      ASSERT(cl->isProtected());
		      return true;
		    }
		  else
		    {
		      goto return_false;
		    };
		}; 
	    }
	  else 
	    {
	      if (_degree <= 1) goto return_false;
	      _swapQuerySides = true;
	      if (query(lit->Arg1(),lit->Arg2())) 
		{
		  ASSERT(isLocked());
		  ASSERT(cl->isProtected());
		  return true;
		};
	      _swapQuerySides = false;
	      if (!query(lit->Arg2(),lit->Arg1()))
		goto return_false;
	      ASSERT(isLocked());
	      ASSERT(cl->isProtected());
	      return true;
	    };
	};
    };
 return_false:
  unlock();
  cl->unprotect();
  return false; 
}; 
void VK::BackDemodulation::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked());
  unlock();
  _rule->unprotect();
}; 
bool VK::BackDemodulation::query(const Term* lhsref,const Term* rhsref) 
{
  CALL("query(const Term* lhsref,const Term* rhsref)");
  ASSERT(isLocked());
  if (lhsref->isVariable() || rhsref->isVariable()) return false;
  if (_checkSubstSafetyWRTInputNonBasicness &&
      !safeRuleWRTInputNonBasicness(*lhsref,*rhsref))
    return false;
  ASSERT(lhsref->IsReference());
  ASSERT(rhsref->IsReference());
  _ruleLHS = lhsref->deref();
  ASSERT(_ruleLHS->isComplex());
  _ruleRHS = *rhsref;
  _compiledConstraint->loadGreater(_ruleLHS,_ruleRHS.deref());
  if ((!_compiledConstraint->lhsVarInstMayOrder()) ||
      (!appropriateDegree(_compiledConstraint->category())))
    {
      return false;
    };
  if (_index.newQuery(_ruleLHS))
    {
      _compiledConstraint->compile(_index.substitution());
      _currentConstraint = _compiledConstraint;
      return true;
    }
  else
    {
      return false;
    };
}; 
bool VK::BackDemodulation::safeRuleWRTInputNonBasicness(const Term& lhs,const Term& rhs)
{
  CALL("safeRuleWRTInputNonBasicness(const Term& lhs,const Term& rhs)");
  _inputNonBasicVarsInRHS.reset();
  collectInputNonBasicVars(rhs,false);
  if (_inputNonBasicVarsInRHS.length() == 0UL) return true;
  cancelInputNonBasicVars(lhs,false);
  return _inputNonBasicVarsInRHS.length() == 0UL;
}; 
void VK::BackDemodulation::collectInputNonBasicVars(const Term& term,bool alreadyInputNonBasic)
{
  CALL("collectInputNonBasicVars(const Term& term,bool alreadyInputNonBasic)");
  if (term.isVariable())
    {
      if (alreadyInputNonBasic) _inputNonBasicVarsInRHS.add1(term.var());
    }
  else
    {
      const Term* complexTerm = term.deref();
      const bool* nonBasicArgs;
      if (alreadyInputNonBasic ||
	  (!((nonBasicArgs = complexTerm->nonBasicArguments()))))
	{
	  for (const Term* arg = complexTerm->Args(); 
	       !arg->isNotArg();
	       arg = arg->Next())
	    collectInputNonBasicVars(*arg,alreadyInputNonBasic);
	}
      else
	{
	  ASSERT(!alreadyInputNonBasic && nonBasicArgs);
	  ulong argNum = 0UL;
	  for (const Term* arg = complexTerm->Args(); 
	       !arg->isNotArg();
	       arg = arg->Next())
	    {
	      collectInputNonBasicVars(*arg,nonBasicArgs[argNum]);
	      ++argNum;
	    };
	};
    };
}; 
void VK::BackDemodulation::cancelInputNonBasicVars(const Term& term,bool alreadyInputNonBasic)
{
  CALL("cancelInputNonBasicVars(const Term& term,bool alreadyInputNonBasic)");
  if (term.isVariable())
    {
      if (alreadyInputNonBasic) _inputNonBasicVarsInRHS.remove(term.var());
    }
  else
    {
      const Term* complexTerm = term.deref();
      const bool* nonBasicArgs;
      if (alreadyInputNonBasic ||
	  (!((nonBasicArgs = complexTerm->nonBasicArguments()))))
	{
	  for (const Term* arg = complexTerm->Args(); 
	       !arg->isNotArg();
	       arg = arg->Next())
	    cancelInputNonBasicVars(*arg,alreadyInputNonBasic);
	}
      else
	{
	  ASSERT(!alreadyInputNonBasic && nonBasicArgs);
	  ulong argNum = 0UL;
	  for (const Term* arg = complexTerm->Args(); 
	       !arg->isNotArg();
	       arg = arg->Next())
	    {
	      cancelInputNonBasicVars(*arg,nonBasicArgs[argNum]);
	      ++argNum;
	    };
	};
    };
}; 
bool VK::BackDemodulation::nextSimplified(bool& onBranch) 
{ 
  CALL("nextSimplified(bool& onBranch)");
  ASSERT(isLocked());
  if (!_rule->isPersistent()) return false;
  onBranch = false;
 check_candidates:
  if (_candidates) 
    {   
      _recentlySimplified = _candidates->key();
      if (_recentlySimplified == _rule) 
	{ 
	  _candidates = _candidates->next();
	  goto check_candidates; 
	}; 
      if (!_recentlySimplified)
	{
	  _candidates = _candidates->next();
	  goto check_candidates; 
	};
      ASSERT(_recentlySimplified);
      if (!_recentlySimplified->isPersistent())
	{ 
	  _candidates = _candidates->next();
	  goto check_candidates; 
	}; 
      if (_bDOnSplittingBranchesFlag)
	{
	  ulong numOfBranchesToCover = _splittingBranches.numOfDiffElements();
	  if (!numOfBranchesToCover) return true;
	  if (_recentlySimplified->numOfAllLiterals() <= numOfBranchesToCover)
	    {
	      _candidates = _candidates->next();
	      goto check_candidates; 
	    };     
	  LiteralList::Iterator iter(_recentlySimplified->LitList());
	  do
	    {
	      if ((iter.currentLiteral()->IsSplitting()) 
		  && (_splittingBranches.safeScoreOf(iter.currentLiteral()->HeaderNum())))
		{
		  ASSERT(!iter.currentLiteral()->arity());          
		  ASSERT(numOfBranchesToCover);
		  numOfBranchesToCover--;
		  if (!numOfBranchesToCover) 
		    { 
		      _numOfBDOnSplittingBranches++;
		      onBranch = true;
		      return true; 
		    };
		};          
	      iter.next();
	    }
	  while (iter.notEnd());
	  ASSERT(numOfBranchesToCover);
	  _candidates = _candidates->next();
	  goto check_candidates; 
	};
      return true;
    };
 try_next_match:
  if (_index.nextMatch(_redex))
    {
      if (_currentConstraint && 
	  (!_currentConstraint->holds()))
	{
	  goto try_next_match;
	};
    }
  else
    {
      if (_swapQuerySides)
	{
	  _swapQuerySides = false;
	  const Term* lit = _rule->LitList()->firstLit();
	  if (!query(lit->Arg2(),lit->Arg1())) return false;
	  goto try_next_match;  
	}
      else 
        return false;
    };
  ASSERT(_redex && _redex->isComplex());
  const Term::OccList* occ = _redex->Occurences();   
  if (occ)
    {
      _candidates = occ->first();
      goto check_candidates;
    }
  else goto try_next_match; 
}; 
bool VK::BackDemodulation::collectInference() 
{
  CALL("collectInference()");
  ASSERT(isLocked());
  ASSERT(_candidates);
  if (rewrite(_candidates)) 
    { 
      _candidates = _candidates->next(); 
      return true; 
    };
  _candidates = _candidates->next();   
  return false; 
}; 
bool VK::BackDemodulation::rewrite(const Term::OccList::Node* occ)
{
  CALL("rewrite(const Term::OccList::Node* occ)");
  _clauseCollector->openNewClause(ClauseBackground::BackDemod);  
  _clauseCollector->addSimplified((Clause*)_recentlySimplified);
  _clauseCollector->addSimplifier((Clause*)_rule);
  _clauseCollector->makeVIP();
  _clauseCollector->announceNewClause();
  Term::LitNumMask* litMask = occ->value();
  ASSERT(litMask);
  ulong litNum = 0;
  for (LiteralList::Iterator iter(_recentlySimplified->LitList());
       iter.notEnd();
       iter.next())
    {
      if (litMask->contains(litNum)) 
	{
	  if (!rewriteLit(iter.currentElement()))
	    goto rejected;
	}
      else 
	if (!_clauseCollector->pushLiteral(iter.currentLiteral(),
					   iter.currentElement()->properties())) 
	  goto rejected;  
      litNum++;
    };
  if (!_clauseCollector->finaliseClause())
    {
    rejected:
      return _clauseCollector->rejectedAsRedundant();
    };
  return true; 
}; 
bool VK::BackDemodulation::rewriteLit(const LiteralList::Element* lit)             
{
  CALL("rewriteLit(const LiteralList::Element* lit)");
  ASSERT(_redex->isComplex());
  const Term* litBody = lit->literal();
  LiteralProperties prop = lit->properties();
  Term header = *litBody;
  if (header.IsEquality())
    {
     if (header.IsOrderedEq()) header.MakeUnordered(); 
     if (lit->isPreorientedEq())
       {
	 const Term* lhs = lit->equalityLHS();
	 if (!lhs->isVariable())
	   {
	     _rewrittenLitIter.Reset(lhs->deref());
	   check_lhs_subterm:
	     if (_rewrittenLitIter.CurrPos() == _redex)
	       {
		 prop.setDefaultEqualityOrientationMode();
	       }
	     else
	       {
		 if (_rewrittenLitIter.Next()) goto check_lhs_subterm;
	       };
	   };
       };
    };
  if (!_clauseCollector->openNewNonpropLiteral(header,prop)) 
    return false; 
  _rewrittenLitIter.Reset(litBody);
  ALWAYS(_rewrittenLitIter.Next());
 check_position:    
  if (_rewrittenLitIter.CurrPos() == _redex)
    {
      if (!collectRHSInstance()) return false;      
      if (_rewrittenLitIter.SkipTerm()) goto check_position;
    }
  else
    {
      if (!_clauseCollector->pushSymbol(_rewrittenLitIter.symbolRef(),false)) 
	return false;
      if (_rewrittenLitIter.Next()) goto check_position;
    };
  return _clauseCollector->closeNonpropLiteral();
}; 
inline bool VK::BackDemodulation::collectRHSInstance()
{
  CALL("collectRHSInstance()");
  _rhsInstance.reset(_ruleRHS,_index.substitution());
  do
    {
      if (!_clauseCollector->pushSymbol(_rhsInstance.symbol(),false)) 
	return false;
    }
  while (_rhsInstance.next());
  return true;
}; 
