/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BackwardFolding.hpp"
#include "TemporaryClauseCollector.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_FOLDING
#define DEBUG_NAMESPACE "BackwardFolding"
#endif
#include "debugMacros.hpp"
namespace VK {
  void BackwardFolding::receive(Clause* const & cl)
  {
    CALL("receive(Clause* const & cl)");
    if (_foldForSplitting && cl->isProducedByPropSplitting())
      {
	_definitionQueue.push(cl);
      };
  }; 
  bool BackwardFolding::nextDefinition() {
    CALL("nextDefinition()");
    ASSERT(!isLocked());
    lock();
  check_reuse:
    _inversePolarity = false;
    if (_useCurrentDefForNegations)
      {
	_currentDefinition->protect();
	_useCurrentDefForNegations = false;
	if (preproCurrentDefForNegations()) 
	  {
	    _inversePolarity = true;
	    return true;
	  };
	_currentDefinition->unprotect();
      };
    ASSERT(!_useCurrentDefForNegations);
    ASSERT(!_inversePolarity);
    if (_definitionQueue.empty()) 
      {
	unlock();
	return false;
      };
    _currentDefinition = _definitionQueue.front();
    _currentDefinition->protect();
    _definitionQueue.pop();
    bool preproCurrent = preproCurrentDef();
    _useCurrentDefForNegations = 
      _reusePropSplNamesForNegations &&
      _currentDefinition->numOfAllLiterals() == 2UL &&
      _boundVariables.isEmpty();
    if (preproCurrent) return true;
    _currentDefinition->unprotect();
    goto check_reuse;
  }; 
  bool BackwardFolding::preproCurrentDef() {
    CALL("preproCurrentDef()");
    ASSERT(isLocked() && _index->isLocked());
    if (!_machine.resetQuery(_currentDefinition->numOfAllLiterals() - 1,
			     _index->index())) 
      return false; 
    LiteralList::Iterator iter(_currentDefinition->LitList());
    ASSERT(iter.notEnd());
    _definedLit = iter.currentLiteral();
    _freeVariables.reset();
    Term::Iterator litIter(_definedLit); 
    while (litIter.NextVar())
      {
	_freeVariables.add(litIter.symbolRef().var());
      };
    iter.next();
    _boundVariables.reset();
    _groundQueryLiterals.reset();
    for (; iter.notEnd(); iter.next())
      {
	Term* lit = iter.currentLiteral();
	if (!preproQueryLit(lit,false)) 
	  {
            return false;
	  };
	if (lit->isGround()) 
	  {
	    _groundQueryLiterals.pushSafe(lit);
	  };
      };
    return _machine.endOfQuery();
  }; 
  bool BackwardFolding::preproCurrentDefForNegations() {
    CALL("preproCurrentDefForNegations()");
    ASSERT(isLocked() && _index->isLocked());
    ASSERT(_currentDefinition->numOfAllLiterals() == 2UL);
    ASSERT(_boundVariables.isEmpty());
    if (!_machine.resetQuery(_currentDefinition->numOfAllLiterals() - 1,
			     _index->index())) 
      return false;
    Term* lit = _currentDefinition->LitList()->nth(1UL)->literal();
    _groundQueryLiterals.reset(); 
    if (lit->isGround()) 
      {
	_groundQueryLiterals.pushSafe(lit);
      };
    if (!preproQueryLit(lit,true)) return false;
    return _machine.endOfQuery();
  }; 
  bool BackwardFolding::preproQueryLit(const Term* lit,bool inversePolarity) {
    CALL("preproQueryLit(const Term* lit)");
    if (lit->arity())
      { 
	 Term::Iterator litIter(lit); 
	 while (litIter.NextVar())
	   {
	     if (!_freeVariables.scoreOf(litIter.symbolRef().var()))
	       {
		 _boundVariables.add(litIter.symbolRef().var());
	       };
	   };
	if (lit->IsEquality()) 
	  { 
	    return 
	      preproQuerySymLit(lit,inversePolarity) && 
	      _machine.queryEndOfSymLit(); 
	  }
	else 
	  {
	    Term header = *lit;
	    if (inversePolarity) header.flipPolarity();
	    if (!_retrieval.litHeader(header)) return false;
	    _machine.queryNextLit();     
	    return  
	      preproQueryComplexTermArguments(lit) && _machine.queryEndOfLit();  
	  };
      }
    else 
      {
	const CSkipList* plcands;
	Term header = *lit;
	if (inversePolarity) header.flipPolarity();
	return 
	  _retrieval.propLit(header,plcands) && _machine.queryPropLit(plcands);
      };
  }; 
  inline 
  bool BackwardFolding::preproQueryComplexTermArguments(const Term* t) 
  {
    CALL("preproQueryComplexTermArguments(const Term* t)");
    ASSERT(t->isComplex());
    ASSERT(t->arity());
    const CLTSkipList* vcands;
    bool allVars = true;
    const Term* argRef;
    for (argRef = t->Args(); ((!argRef->isNotArg()) && allVars); argRef = argRef->Next())
      allVars = argRef->isVariable();
    argRef = t->Args();
    if (argRef->isVariable())
      {
	if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,allVars)))
	  return false;  
      }
    else 
      if (!preproQueryComplexTerm(argRef->deref())) return false;
    argRef = argRef->Next();
    while (!argRef->isNotArg())
      {
	if (argRef->isVariable())
	  {
	    if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,false)))
	      return false;  
	  }
	else 
	  if (!preproQueryComplexTerm(argRef->deref())) return false; 
	argRef = argRef->Next();    
      };   
    return true;
  }; 
  inline bool BackwardFolding::preproQueryComplexTerm(const Term* t)
  {
    CALL("preproQueryComplexTerm(const Term* t)");
    ASSERT(t->isComplex());
    const CLSkipList* ccands;
    if (t->arity())
      {
	return _retrieval.nonconstFunc(*t) && preproQueryComplexTermArguments(t);
      }
    else 
      return _retrieval.constant(*t,ccands) && _machine.queryConstant(*t,ccands);
  }; 
  inline 
  bool 
  BackwardFolding::preproQuerySymLit(const Term* lit,bool inversePolarity)
  {
    CALL("preproQuerySymLit(const Term* lit,bool inversePolarity)");
    ASSERT(lit->IsEquality());
    Term header = *lit;
    if (inversePolarity) header.flipPolarity();
    if (!_retrieval.symLitHeader(header)) return false;
    _machine.queryNextSymLit();
    const Term* argRef1 = lit->Args();
    const Term* argRef2 = argRef1->Next(); 
    const CLATSkipList* slvcands;
    if (argRef1->isVariable()) 
      {   
	if ((!_retrieval.symLitVariable(slvcands)) 
	    || (!_machine.querySymLitVariable(*argRef1,slvcands,argRef2->isVariable())))
	  return false;      
      }
    else 
      if (!preproQuerySymLitComplexTerm(argRef1->deref())) return false;
    if (!_retrieval.symLitSecondArg()) return false;
    _machine.querySymLitSecondArg();
    if (argRef2->isVariable()) 
      {   
	return (_retrieval.symLitVariable(slvcands)) 
	  && (_machine.querySymLitVariable(*argRef2,slvcands,false)); 
      }
    else return preproQuerySymLitComplexTerm(argRef2->deref()); 
  }; 
  inline bool BackwardFolding::preproQuerySymLitComplexTerm(const Term* t)
  {
    CALL("preproQuerySymLitComplexTerm(const Term* t)");
    ASSERT(t->isComplex());
    const CLASkipList* slccands;
    if (t->arity())
      {
	return _retrieval.symLitNonconstFunc(*t) && preproQuerySymLitComplexTermArguments(t);
      }
    else 
      return _retrieval.symLitConstant(*t,slccands) && _machine.querySymLitConstant(*t,slccands);
  }; 
  inline bool BackwardFolding::preproQuerySymLitComplexTermArguments(const Term* t) 
  {
    CALL("preproQuerySymLitComplexTermArguments(const Term* t)");
    ASSERT(t->isComplex());
    ASSERT(t->arity());
    const CLATSkipList* slvcands;
    bool allVars = true;
    const Term* argRef;
    for (argRef = t->Args(); ((!argRef->isNotArg()) && allVars); argRef = argRef->Next())
      allVars = argRef->isVariable();
    argRef = t->Args();          
    if (argRef->isVariable())
      {
	if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,allVars)))
	  return false;  
      }
    else 
      if (!preproQuerySymLitComplexTerm(argRef->deref())) return false;
    argRef = argRef->Next();
    while (!argRef->isNotArg())
      {
	if (argRef->isVariable())
	  {
	    if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,false)))
	      return false;  
	  }
	else 
	  if (!preproQuerySymLitComplexTerm(argRef->deref())) return false; 
	argRef = argRef->Next();    
      };   
    return true;
  }; 
  void BackwardFolding::abandonCurrentDefinition() {
    CALL("abandonCurrentDefinition()");
    if (!_useCurrentDefForNegations)
      {
	_inversePolarity = false;
	_currentDefinition->unprotect();
      };
    unlock();
  };
  bool BackwardFolding::foldNext() {
    CALL("foldNext()");
    ASSERT(isLocked() && _index->isLocked());
    ulong subsumedClNum;
    while (_machine.subsumeNextMultisetMode(subsumedClNum)) 
      {
	_currentFolded = _clauseNumIndex->find(subsumedClNum);
	ASSERT(_currentFolded);    
	if (_currentFolded->isPersistent() 
	    && (!_foldForSplitting || !_currentFolded->isPartOfNameIntroForPropSplitting())
	    && preliminarySubstitutionCheck()
	    && separateRemainderLiterals())
	  {
	    if (_foldForSplitting &&
		_currentDefinition->isProducedByPropSplitting() &&
		!_foldSingleComponentsForPropSpl)
	      {
		const LiteralList::Element* const * l = _remainderLiterals.begin();
		for (;
		     l != _remainderLiterals.end() && (*l)->literal()->IsSplitting(); 
                     ++l);
		if (l == _remainderLiterals.end())
		  return false;
	      };
	    _currentDefinition->MakeVIP();
	    return true;
	  };
      };
    return false;     
  }; 
  bool BackwardFolding::preliminarySubstitutionCheck()
  { 
    CALL("preliminarySubstitutionCheck()");
    _subst.reset();
    _machine.resetSubstitutionCollection();
    ulong var;
    Term term;
    while (_machine.substNextPair(var,term)) 
      {
	_subst.associate(var,term);
      };
    for (ulong v = 0UL; v < _boundVariables.numOfDiffElements(); ++v)
      {
	Term boundVarSubst = _subst.map(_boundVariables.nth(v));
	if (!boundVarSubst.isVariable()) return false;
	for (ulong u = 0UL; u < _boundVariables.numOfDiffElements(); ++u)
	  if (u != v && boundVarSubst == _subst.map(_boundVariables.nth(u)))
	    return false;
	for (ulong u = 0UL; u < _freeVariables.numOfDiffElements(); ++u)
	  {
	    Term substTerm = _subst.map(_boundVariables.nth(u));
	    if (substTerm == boundVarSubst) return false;
	    if (substTerm.IsReference() && 
		substTerm.deref()->containsSymbol(boundVarSubst))
	      return false;
	  };
      };    
    return true;
  }; 
  bool BackwardFolding::separateRemainderLiterals() {
    CALL("separateRemainderLiterals()");
    ulong litNum = 1UL; 
    _remainderLiterals.reset();
    for(LiteralList::Iterator iter(_currentFolded->LitList());
	iter.notEnd();
	iter.next())
      {
	const LiteralList::Element* el = iter.currentElement();
	const Term* lit = el->literal();
	if (!_machine.litIsSubsumedByNonGround(litNum) &&
	    !isSubsumedByGround(lit))
	  {
	    for (ulong u = 0UL; u < _boundVariables.numOfDiffElements(); ++u)
	      if (lit->containsSymbol(_subst.map(_boundVariables.nth(u))))
		return false;
	    _remainderLiterals.pushSafe(el);
	  };
	++litNum;
      };
    return true;
  }; 
  bool BackwardFolding::isSubsumedByGround(const Term* lit) const {
    CALL("isSubsumedByGround(const Term* lit)");
   for (Term* const * l = _groundQueryLiterals.begin();
	 l != _groundQueryLiterals.end();
	 ++l)
      {
	if (_inversePolarity)
	  {
	    (*l)->flipPolarity();
	    bool complementary = Term::EqualComplexTerms(*l,lit);
	    (*l)->flipPolarity();
	    if (complementary) return true;
	  }
	else
	  if (*l == lit ||
	      Term::EqualComplexTerms(*l,lit))
	    return true;
      };
    return false;
  }; 
  bool BackwardFolding::collectResult() {
    CALL("collectResult()");
    LiteralProperties defaultProp;
    Term header;
    if (_inversePolarity &&
	!_currentDefinition->isNameIntroWithAllPartsDeployed())
      {
	_clauseCollector->openNewClause(ClauseBackground::Clausification,
					_currentDefinition);
	_clauseCollector->addRule(ClauseBackground::PartOfNameIntroForPropSplitting);
	_clauseCollector->announceNewClause();
	if (!_clauseCollector->pushLiteral(_definedLit,defaultProp))
	  return false;
	ASSERT(_currentDefinition->numOfAllLiterals() == 2UL);
	Term* lit = _currentDefinition->LitList()->nth(1UL)->literal();
	lit->flipPolarity();
	if (!_clauseCollector->pushLiteral(lit,defaultProp))
	  {
	    lit->flipPolarity();
	    return false;
	  };
	lit->flipPolarity();
	if (!_clauseCollector->finaliseClause() &&
	    !_clauseCollector->rejectedAsRedundant())
	  return false;
	_currentDefinition->markAsNameIntroWithAllPartsDeployed();
      };
    _clauseCollector->openNewClause(ClauseBackground::Folding,
				    _currentFolded);
    _clauseCollector->addSimplifier(_currentDefinition);
    _clauseCollector->makeVIP();
    _clauseCollector->announceNewClause();
    const LiteralList::Element* el;
    while (!_remainderLiterals.empty())
      {
	el = _remainderLiterals.pop();
	if (!_clauseCollector->pushLiteral(el->literal(),
					   el->properties())) 
	  {
	    return false;
	  };
      };
    header = *_definedLit;
    if (_inversePolarity) header.flipPolarity();
    if (header.arity())
      {
	if (!_clauseCollector->openNewNonpropLiteral(header,
						     defaultProp)) 
	  return false;
	_definedLitInstance.reset(_definedLit,&_subst);
	while (_definedLitInstance.next())
	  {
	    if (!_clauseCollector->pushSymbol(_definedLitInstance.symbol(),
					      false)) 
	      return false;
	  };
	if (!_clauseCollector->closeNonpropLiteral())
	  return false;
      }
    else
      {
	if (!_clauseCollector->pushPropLiteral(header,defaultProp))
	  return false;
      };
    if (!_clauseCollector->finaliseClause())
      {
	return _clauseCollector->rejectedAsRedundant();
      };
    return true;
  }; 
}; 
