/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "LinearCombination.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
using namespace std;
using namespace BK;
using namespace VK;
Clause::Context* 
Clause::Context::_current; 
void Clause::Destroy()         
{
  CALL("Destroy()");
  delete litList;
  delete this;
};  
void Clause::DestroyUnshared()         
{
  CALL("DestroyUnshared()");
  for (LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
    iter.currentLiteral()->destroyUnsharedLit();
  delete litList;
  delete this;
};  
bool Clause::isSymmetryLaw() const 
{
  if (numOfAllLiterals() != 2) return false;
  const Term* lit1 = LitList()->firstLit();
  const Term* lit2 = LitList()->nth(1)->literal();
  if (!Term::Symmetry(lit1,lit2)) return false;
  return lit1->polarity() != lit2->polarity();
}; 
bool Clause::isCommutativityLaw(ulong& func) const
{
  CALL("isCommutativityLaw(ulong& func) const");
  if (Unit())
    {
      const Term* lit = LitList()->firstLit();
      if  (lit->Positive() 
	   && lit->IsEquality()
	   && lit->Arg1()->IsReference()
	   && lit->Arg2()->IsReference()
	   && (Term::Symmetry(lit->Arg1()->deref(),lit->Arg2()->deref())))
	{
	  func = lit->Arg1()->deref()->functor();
	  return true;
	};
    }; 
  return false;     
}; 
bool Clause::isTransitivityLaw(ulong& pred) const
{
  CALL("isTransitivityLaw(ulong& pred) const");
  if ((numOfAllLiterals() != 3) ||
      (flag(FlagContainsAnswerLiterals))) 
    return false;
  const Term* lit1 = LitList()->firstLit();
  const Term* lit2 = LitList()->nth(1)->literal();
  const Term* lit3 = LitList()->nth(2)->literal();
  pred = lit1->functor();
  if ((lit2->functor() != pred) ||
      (lit3->functor() != pred) ||
      (lit1->arity() != 2UL))
    return false;
  if ((!lit1->Arg1()->isVariable()) ||
      (!lit1->Arg2()->isVariable()) ||
      (*(lit1->Arg1()) == *(lit1->Arg2())))
    return false;
  if ((!lit2->Arg1()->isVariable()) ||
      (!lit2->Arg2()->isVariable()) ||
      (*(lit2->Arg1()) == *(lit2->Arg2())))
    return false;
  if ((!lit2->Arg1()->isVariable()) ||
      (!lit2->Arg2()->isVariable()) ||
      (*(lit2->Arg1()) == *(lit2->Arg2())))
    return false;
  const Term* negLit1 = 0;
  const Term* negLit2 = 0;
  const Term* posLit = 0;
  if (lit1->Negative()) 
    {
      negLit1 = lit1;
      if (lit2->Negative())
	{
	  negLit2 = lit2;
	  if (lit3->Negative()) return false;
	  posLit = lit3;
	}
      else
	{
	  posLit = lit2;
	  if (lit3->Positive()) return false;
	  negLit2 = lit3;
	};
    }
  else
    {
      posLit = lit1;
      if (lit2->Positive() || lit3->Positive()) return false;
      negLit1 = lit2;
      negLit2 = lit3;
    };
  if (*(posLit->Arg1()) == *(negLit1->Arg1()))
    {
      if ((*(negLit1->Arg2()) != *(negLit2->Arg1())) || 
	  (*(negLit2->Arg2()) != *(posLit->Arg2())))
	return false;
    }
  else
    if (*(posLit->Arg1()) == *(negLit2->Arg1()))
      {
	if ((*(negLit2->Arg2()) != *(negLit1->Arg1())) || 
	    (*(negLit1->Arg2()) != *(posLit->Arg2())))
	  return false;
      }
    else
      return false;
  return true;
}; 
ostream& Clause::outputBackground(ostream& str) const
{
  if (IsVIP()) str << "vip ";
  str << '['; 
  BitWord rules = _background.properties();
  ClauseBackground::outputRules(str,rules);
  bool parent;
  if (!isInputClause())
    {
      for (const ClauseBackground::AncestorList* anc = _background.ancestors(); 
	   anc; 
	   anc = anc->tl())
	{
	  str << anc->hd().strip(parent)->Number();
	  if (anc->tl()) str << ',';
	};
    };
  return str << ']';
}; 
ostream& Clause::output(ostream& str) const  
{
  if (mainSet() == MainSetActive) str << "* ";
  if (isNameIntro()) str << "NAME INTRO ";
  str << Number() << ". ";
  if (isNameIntro())
    {	  
      LiteralList::Iterator iter(LitList());
      const Term* head = iter.currentLiteral();
      iter.next();
      ASSERT(iter.notEnd());
      Term::Iterator litIter(head);
      static LinearCombination<ulong,VampireKernelConst::MaxNumOfVariables> varRegister;
      varRegister.reset();
      while (litIter.NextVar())
	{
	  ulong varNum = litIter.symbolRef().var();
	  if (varRegister.coefficient(varNum) == 0UL)
	    {
	      if (varRegister.isZero())
		{
		  str << "! [";
		}
	      else
		str << ",";
	      str << litIter.symbolRef();		  
	      varRegister.add1(varNum);
	    };
	};
      if (!varRegister.isZero()) str << "] : ";
      str << head << " <=> (";
      bool innerVar = false;
      while (iter.notEnd())
	{
	  const Term* lit = iter.currentLiteral();
	  iter.next();
	  litIter.Reset(lit);
	  while (litIter.NextVar())
	    {
	      ulong varNum = litIter.symbolRef().var();
	      if (varRegister.coefficient(varNum) == 0UL)
		{
		  if (innerVar)
		    {
		      str << ","; 
		    }
		  else
		    {
		      str << "! [";
		      innerVar = true;
		    };
		  str << litIter.symbolRef();
		  varRegister.add1(varNum);
		};
	    };
	};
      if (innerVar) str << "] : (";
      iter.reset(LitList());
      iter.next();
      while (iter.notEnd())
	{
	  str << iter.currentLiteral();
	  if (!iter.last()) str << " | ";
	  iter.next();
	};
      if (innerVar) str << ')';
      str << ')';
    }
  else 
    {
      if (Empty()) { str << '#'; }
      else 
	LitList()->output(str,Context::getCurrent()->_showLiteralProperties);
    };
  str << " /" << weight() << '/' << size() << '/' << practicallyMinResidualSize() << "/"
      << inferenceDepth() << " ";
  return outputBackground(str);   
}; 
