/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef DISC_TREE_HPP
#define DISC_TREE_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_DISC_TREE
#  define DEBUG_NAMESPACE "DiscTree"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class DiscTree
  {
  public:
    class Node
    {
    public:
      class Tag
      {
      public:
	Tag() {};
	Tag(Term* p) { _symbol.makeReference(p); }; 
	inline Tag& operator= (Term* p) 
	{
	  _symbol.makeReference(p);
	  return *this; 
	}; 
	Tag(const Term& sym) 
	  : _symbol(sym) 
	{ 
	};
	~Tag() {};
	inline Tag& operator= (const Tag& tag) 
	{
	  _symbol = tag._symbol;
	  return *this;
	};
	inline bool IsLeaf() const { return _symbol.IsReference(); };
	inline bool equal(const Term& sym) const 
	{
	  return _symbol == sym;
	};
	inline bool Greater(const Term& sym) const
	{
	  if (_symbol.typeTag() == sym.typeTag())
	    return _symbol > sym;
	  return _symbol.typeTag() > sym.typeTag();
	};
	inline bool equal(const Tag& tag) const 
	{
	  return equal(tag._symbol);
	};
	inline bool Greater(const Tag& tag) const 
	{
	  return Greater(tag._symbol);
	}; 
	inline Term* term() const { return _symbol.deref(); };
	inline void SetTerm(Term* t) { _symbol.makeReference(t); }; 
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str) const { return str << _symbol; };
#endif 
      private:
	Term _symbol;
      }; 
    public:
      Node(Term* p) : _tag(p), _next(0) {}; 
      Node(const Term& sym)  
	: _tag(sym),
	  _less(0), 
	  _greater(0)
      {};
      ~Node() {};
      inline void Copy(const Node& node)
      {
	_tag = node._tag;
	_next = node._next;
	_less = node._less;
	_greater = node._greater;
      };  
      inline void* operator new(size_t)
      {
	return BK::GlobAlloc::allocate(sizeof(Node));
      };
      inline void operator delete(void* obj)
      {
	BK::GlobAlloc::deallocate(obj,sizeof(Node));
      };
      Tag& tag() { return _tag; };
      const Tag& tag() const { return _tag; }; 
      inline Term* term() const { return _tag.term(); }; 
      inline void SetTerm(Term* t) { _tag.SetTerm(t); };
      inline bool IsLeaf() const { return _tag.IsLeaf(); };
      inline Node*& Next() { return _next; }; 
      inline Node* const & Next() const { return _next; }; 
      inline long& entryMultiplicity() { return (long&)_next; }; 
      inline Node*& less() { return _less; }; 
      inline Node* const & less() const { return _less; }; 
      inline Node*& greater() { return _greater; };
      inline Node* const & greater() const { return _greater; };
      inline bool Fork() { return (less() || greater()); }; 
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const
      {
	str << '[' << (unsigned long)this << "] ";
	if (IsLeaf()) return str << "LEAF " << term();
	return tag().output(str) << " N[" << (unsigned long)Next()
				 << "] L[" << (unsigned long)less()
				 << "] G[" << (unsigned long)greater() << "] "; 
      };
#endif
    private:
      Node();
    private:
      Tag _tag;
      Node* _next;
      Node* _less;
      Node* _greater;
    }; 
    class Surfer;
  public:
    static inline Node** Track(const Term& sym,Node** node);
    static inline void InfoForDeletion(Term::Iterator& trm,
				       Node** tree,
				       Node**& branch_to_destroy,
				       Node*& leaf,
				       Node*& last_fork);
    static inline Node* CreateBranch(Term::IteratorWithDepth& trm,
				     Node*& leaf);
    static inline void DestroyBranchWithoutForks(Node* branch);
    static inline void ReduceFork(Node* fork);
    static inline void Destroy(Node* tree,void (*DestroyLeafContent)(Node* leaf));
  private:
    static inline Node* Minimal(Node* node);
    static inline Node* Merge(Node* less,Node* greater);
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_DISC_TREE
#  define DEBUG_NAMESPACE "DiscTree::Surfer"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class DiscTree::Surfer
  {
  public:
    Surfer() 
    {
    };
    ~Surfer() {};
    inline void Reset(DiscTree::Node** tree)
    {
      node = tree;
    };
    inline bool Read(const Term& sym)
    {
      CALL("Read(const Term& sym)");   
      node = DiscTree::Track(sym,node);
      if (*node)
	{
	  node = &((*node)->Next());
	  return true;
	}
      else return false;
    }; 
    inline void Write(const Term& sym)
    {
      CALL("Write(const Term& sym)");
      *node = new DiscTree::Node(sym);
      node = &((*node)->Next());
    }; 
    inline void CreateLeaf(Term* t)
    {
      *node = new DiscTree::Node(t);
    };
    inline DiscTree::Node* Leaf() { return *node; };
  private:
    DiscTree::Node** node;
  }; 
}; 
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_DISC_TREE
#  define DEBUG_NAMESPACE "DiscTree"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline DiscTree::Node* DiscTree::Minimal(Node* node)
  {
    CALL("Minimal(Node* node)");
    ASSERT(node->less());
  next_less:
    node = node->less();
    if (node->less()) { goto next_less; }
    else { return node; };
  }; 
  inline DiscTree::Node* DiscTree::Merge(Node* less,Node* greater)
  {
    CALL("Merge(Node* less,Node* greater)");
    ASSERT(less);
    ASSERT(greater);
    if (less->greater())
      {
	if (greater->less())
	  {
	    Minimal(greater)->less() = less;
	    return greater;
	  }
	else
	  {
	    greater->less() = less;
	    return greater;
	  };
      }
    else
      {
	less->greater() = greater;
	return less;
      };
  }; 
  inline void DiscTree::ReduceFork(Node* fork)
  {
    CALL("ReduceFork(Node* fork)");
    ASSERT(fork);
    ASSERT(fork->Fork());
    Node* node_to_delete;
    if (fork->less())
      {
	if (fork->greater())
	  {
	    node_to_delete = Merge(fork->less(),fork->greater());
	    fork->Copy(*node_to_delete);
	  }
	else
	  {
	    node_to_delete = fork->less();
	    fork->Copy(*(fork->less()));
	  };
      }
    else 
      {
	node_to_delete = fork->greater();
	fork->Copy(*(fork->greater()));
      };
    delete node_to_delete;
  }; 
  inline DiscTree::Node** DiscTree::Track(const Term& sym,Node** node)
  {
  check_node:
    if (*node)
      {
	if ((*node)->tag().equal(sym)) { return node; };
	if ((*node)->tag().Greater(sym))
	  {
	    node = &((*node)->less());
	  }
	else
	  {
	    node = &((*node)->greater());
	  }; 
	goto check_node; 
      }
    else { return node; };
  }; 
  inline DiscTree::Node* DiscTree::CreateBranch(Term::IteratorWithDepth& trm,
						Node*& leaf)
  {
    Node* branch = new Node(trm.symbolRef());
    Node* last_node = branch;
  next_sym:
    if (trm.Next())
      {
	last_node->Next() = new Node(trm.symbolRef()); 
	last_node = last_node->Next();
	goto next_sym;
      }
    else
      {
	leaf = new Node(0);
	last_node->Next() = leaf;
      };   
    return branch;
  }; 
  inline void DiscTree::DestroyBranchWithoutForks(Node* branch)
  {
    CALL("DestroyBranchWithoutForks(Node* branch)");
    ASSERT(branch);
    Node* node_to_delete;
  next_node:
    if (branch->IsLeaf()) { delete branch; }
    else
      {
	node_to_delete = branch;
	branch = branch->Next();
	delete node_to_delete;
	goto next_node;
      };   
  }; 
  inline void DiscTree::Destroy(Node* tree,void (*DestroyLeafContent)(Node* leaf))
  {
    CALL("Destroy(Node* tree,void (*DestroyLeafContent)(Node* leaf))");
    if (tree)
      {    
	if (tree->IsLeaf()) 
	  {
	    DestroyLeafContent(tree);
	    delete tree;
	  }
	else
	  {
	    if (tree->less()) Destroy(tree->less(),DestroyLeafContent);
	    if (tree->greater()) Destroy(tree->greater(),DestroyLeafContent);
	    if (tree->Next()) Destroy(tree->Next(),DestroyLeafContent);
	    delete tree;
	  };
      };
  }; 
  inline void DiscTree::InfoForDeletion(Term::Iterator& trm,
					Node** tree,
					Node**& branch_to_destroy,
					Node*& leaf,
					Node*& last_fork)
  {
    CALL("InfoForDeletion(Term::Iterator&,Node**,Node**&,Node*&,Node*&)");
    ASSERT(*tree);
    ASSERT(!((*tree)->IsLeaf()));
    last_fork = 0; 
    branch_to_destroy = tree;
    Node** current_branch;
    Node* current_node;
  track_sym:
    current_branch = Track(trm.symbolRef(),tree);
    current_node = *current_branch;
    ASSERT(current_node);
    ASSERT(!(current_node->IsLeaf()));
    ASSERT(current_node->tag().equal(trm.symbolRef()));
    if (current_node->Fork())  
      {
	branch_to_destroy = &(current_node->Next());
	last_fork = current_node;
      }
    else
      {
	if (current_branch != tree) 
	  { 
	    ASSERT(!((*tree)->tag().equal(trm.symbolRef())));
	    branch_to_destroy = current_branch;
	    last_fork = 0;
	  }
	else
	  {
	    ASSERT((*tree)->tag().equal(trm.symbolRef()));
	  };
      };
    tree = &(current_node->Next());
    ASSERT(*tree);
    if (trm.Next()) { goto track_sym; };
    leaf = *tree;
    ASSERT(leaf->IsLeaf());
    ASSERT(*branch_to_destroy);
    ASSERT((!last_fork) || (!(last_fork->IsLeaf())));
    ASSERT((!last_fork) || (last_fork->Fork()));
    ASSERT((!last_fork) || (branch_to_destroy == &(last_fork->Next()))); 
    ASSERT((*branch_to_destroy)->IsLeaf() ? (last_fork != 0): true);
    ASSERT((*branch_to_destroy)->IsLeaf() ? (&(last_fork->Next()) == branch_to_destroy) : true);
  }; 
}; 
#endif
