/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardDemodulation.hpp"
#include "Clause.inl"
#include "SimplificationOrdering.hpp"
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
#  define DEBUG_NAMESPACE "ForwardDemodulation"
#endif 
#include "debugMacros.hpp"
using namespace VK;
using namespace BK;
ForwardDemodulation::ForwardDemodulation() : 
  fDOnSplittingBranchesFlag(false)
{
  CALL("constructor ForwardDemodulation()");
  _constrCheck = 
    SimplificationOrdering::current()->reserveConstraintCheckOnForwardMatchingSubst();
  unlock();
};
ForwardDemodulation::~ForwardDemodulation() 
{
  CALL("destructor ~ForwardDemodulation()");
  if (DestructionMode::isThorough()) 
    {
      ASSERT(_constrCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnForwardMatchingSubst(_constrCheck);
    };
};
void ForwardDemodulation::reset()
{
  CALL("reset()");
  BK::destroy(*this);
  BK::init(*this);
}; 
void ForwardDemodulation::Integrate(Clause* cl) 
{ 
  CALL("Integrate(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  if (CanBeRewritingRule(cl)) index.integrate(cl);  
  cl->unprotect();
  unlock();
}; 
void ForwardDemodulation::Remove(Clause* cl) 
{ 
  CALL("Remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  if (CanBeRewritingRule(cl)) index.remove(cl);   
  cl->unprotect();  
  unlock();
};
bool 
ForwardDemodulation::Rewrite(const Flatterm* query,
					  const SetOfSplittingBranches& splittingBranches)
{
  CALL("Rewrite(const Flatterm* query,const SetOfSplittingBranches& splittingBranches)");
  ASSERT(query->symbolRef().isComplex());
  ASSERT(_degree > 0); 
  ASSERT(!isLocked());
  lock();
  branchesUsed = false;
  for (FMCodeTree<Index::LeafInfo>::Leaf* 
	 leaf = index.Match(query); 
       leaf; 
       leaf = index.MatchAgain())
    {
      ASSERT(leaf->isLeaf());  
      Index::ConstrRuleList* crl =  leaf->info().rules(); 
      ASSERT(crl);  
      do
	{
#ifdef DEBUG_ALLOC
	  ASSERT(crl->checkObj());
#endif
	  rule = crl->hd().clause;
	  if (rule->isPersistent() && (!rule->isBlockedForDemodulationFrom()))
	    {
	      SimplificationOrdering::StoredConstraint* constr = crl->hd().constr;
	      collectSubstCache(crl->hd().subst);
	      if (!constr) 
		{
		  ASSERT(crl->hd().lhs == rule->LitList()->firstElem()->equalityLHS()->deref());
		  if (CheckSplittingBranches(rule,splittingBranches))
		    {
		      _ruleRHSTerm = *(rule->LitList()->firstElem()->equalityRHS());
		      _ruleRHSInstance.reset(_ruleRHSTerm,&_substCache);
		      return true;
		    };
		}
	      else
		{
		  if (_constrCheck->holds(constr,&_substCache) && 
		      (CheckSplittingBranches(rule,splittingBranches)))
		    {
		      const Term* lhs = crl->hd().lhs;
		      ASSERT(lhs);
		      ASSERT(lhs->isComplex());
		      if (rule->LitList()->firstLit()->Arg1()->deref() == lhs)
			{
			  _ruleRHSTerm = *(rule->LitList()->firstLit()->Arg2());
			}          
		      else
			{
			  ASSERT(rule->LitList()->firstLit()->Arg2()->deref() == lhs);
			  _ruleRHSTerm = *(rule->LitList()->firstLit()->Arg1());
			  ASSERT(_ruleRHSTerm.deref()->isComplex());
			};
		      _ruleRHSInstance.reset(_ruleRHSTerm,&_substCache);  
		      return true;      
		    };              
		};
	    };
	  crl = crl->tl();
	}
      while (crl);
    }; 
  unlock();
  return false;      
}; 
void ForwardDemodulation::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked());
  unlock();
};
ForwardDemodulation::Index::ConstrRuleList* 
ForwardDemodulation::removeConstrRule(Clause* cl,
				      Term* t,
				      Index::ConstrRuleList* lst)
{
  CALL("removeConstrRule(Clause* cl,Term* t,Index::ConstrRuleList* lst)");
  if (!lst) { return 0; };
  if ((lst->hd().clause == cl) && (lst->hd().lhs == t))
    {
      Index::ConstrRuleList* res = lst->tl();
      delete lst;
      return res;
    };
  lst->tl() = removeConstrRule(cl,t,lst->tl());
  return lst;
}; 
ForwardDemodulation::Index::ConstrRuleList* 
ForwardDemodulation::insertConstrRule(Clause* cl,
				      SimplificationOrdering::StoredConstraint* constr,
				      Index::StoredSubstitution* subst,
				      Term* t,
				      Index::ConstrRuleList* lst)
{
  CALL("insertConstrRule(..)");
  Index::ConstrRuleList* res = new Index::ConstrRuleList(lst); 
  res->hd().replace(cl,constr,subst,t); 
  return res;
}; 
inline 
bool 
ForwardDemodulation::CanBeRewritingRule(const Clause* cl)
{
  CALL("CanBeRewritingRule(const Clause* cl)");
  if (cl->flag(Clause::FlagContainsAnswerLiterals)) return false;
  if (cl->isSupportClause() && (!_demodulateFromSupport)) return false;
  Term* lit; 
  if (cl->Unit())
    {
      lit = cl->LitList()->firstLit();
      return (lit->Positive() && lit->IsEquality());
    }
  else  
    {
      if (!fDOnSplittingBranchesFlag) return false;
      lit = cl->LitList()->firstLit();
      if (lit->Negative() || (!lit->IsEquality()) || (cl->LitList()->numOfSelectedLits() != 1)) return false;
      LiteralList::Iterator iter(cl->LitList());
      iter.next();
      ASSERT(iter.notEnd());
      do
	{
	  if (!iter.currentLiteral()->IsSplitting()) return false;
	  iter.next();
	}
      while (iter.notEnd());
      return true;
    };
}; 
inline 
bool 
ForwardDemodulation::CheckSplittingBranches(const Clause* cl,
                                            const SetOfSplittingBranches& splittingBranches)
{
  CALL("CheckSplittingBranches(const Clause* cl,const SetOfSplittingBranches& splittingBranches)");
  if ((!fDOnSplittingBranchesFlag) || cl->Unit())
    {
      ASSERT(cl->Unit());
      return true; 
    }; 
  LiteralList::Iterator iter(cl->LitList());
  iter.next(); 
  while (iter.notEnd()) 
    {
      if (!splittingBranches.safeScoreOf(iter.currentLiteral()->HeaderNum())) 
	return false;
      iter.next();
    };
  branchesUsed = true;
  return true;
}; 
inline
void 
ForwardDemodulation::collectSubstCache(const ForwardDemodulation::Index::StoredSubstitution* storedSubst)
{
  CALL("collectSubstCache(const ForwardDemodulation::Index::StoredSubstitution* storedSubst)");
  _substCache.reset(); 
  while (storedSubst) 
    {
#ifdef DEBUG_ALLOC
      ASSERT(storedSubst->checkObj());
#endif
      _substCache.associate(storedSubst->hd().first,*(storedSubst->hd().second));
      storedSubst = storedSubst->tl();
    }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_DEMODULATION
#define DEBUG_NAMESPACE "ForwardDemodulation::Index"
#endif 
#include "debugMacros.hpp"
void ForwardDemodulation::Index::integrate(Clause* cl)
{
  CALL("integrate(Clause* cl)");
#ifdef DEBUG_NAMESPACE
  if (cl->numOfNonanswerLiterals() > 1)
    {
      LiteralList::Iterator iter(cl->LitList());
      iter.next(); 
      do
	{
	  if ((!iter.currentLiteral()->IsSplitting()) &&
	      (!iter.currentLiteral()->isAnswerLiteral()))
	    {
	      ICP("E1");
	    };
	  iter.next();
	}
      while (iter.notEnd());
    };  
#endif          
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 
  LiteralList::Element* elem = cl->LitList()->firstElem();
  ASSERT(elem->positive());
  ASSERT(elem->isEquality());
  if (!elem->isOrientableEq()) return;
  if (elem->isOrientedEq())
    {
      if (elem->orientationFollowsOrdering())
	{
	integrate_without_constraint:
	  ASSERT(!elem->equalityLHS()->isVariable());
	  Term* lhs = elem->equalityLHS()->deref();
	  FMCodeTree<Index::LeafInfo>::Leaf* leaf = Integrate(lhs); 
	  Index::ConstrRuleList* crl = leaf->info().rules();   
#ifdef DEBUG_ALLOC
	  ASSERT(!crl || crl->checkObj());
#endif
	  crl = new Index::ConstrRuleList(crl); 
	  crl->hd().replace(cl,0,createStoredSubstitution(),lhs);
	  leaf->info().rules() = crl;
	}
      else 
	{	  
	  if (elem->equalityLHS()->isVariable()) return;
	  _specialisedArgComparison->load(*(elem->equalityLHS()),
					  *(elem->equalityRHS()));
	  if (_specialisedArgComparison->lhsVarInstMayOrder())
	    {
	      _specialisedArgComparison->specialiseNotEqual();
	      ASSERT(!_specialisedArgComparison->isAlwaysEqual());
	      if (_specialisedArgComparison->isAlwaysLess() || 
		  _specialisedArgComparison->isAlwaysGreater())
		goto integrate_without_constraint;
	      SimplificationOrdering::StoredConstraint* constraint = 
		SimplificationOrdering::current()->storedConstraintNotEqual(_specialisedArgComparison);
	      if (constraint)
		{
		  integrate(cl,elem->equalityLHS()->deref(),constraint);
		  SimplificationOrdering::current()->releaseConstraint(constraint);
		};
	    };
	}; 
    }      
  else 
    {
      Term* lit = elem->literal();
      Term* arg1ref = lit->Args();
      Term* arg2ref = arg1ref->Next();
      if ((arg1ref->isVariable()) || (arg2ref->isVariable())) return;        
      _specialisedArgComparison->load(*arg1ref,*arg2ref);
      if (_specialisedArgComparison->lhsVarInstMayOrder())
	{
	  _specialisedArgComparison->specialiseGreater();
	  if (appropriateDegreeForGreater(_specialisedArgComparison->category()))
	    {
	      SimplificationOrdering::StoredConstraint* constraint = 
		SimplificationOrdering::current()->storedConstraintGreater(_specialisedArgComparison);
	      if (constraint)
		{
		  integrate(cl,arg1ref->deref(),constraint);
		  SimplificationOrdering::current()->releaseConstraint(constraint);
		};
	    };
	};
      if (_specialisedArgComparison->rhsVarInstMayOrder())
	{
	  _specialisedArgComparison->specialiseLess();
	  if (appropriateDegreeForLess(_specialisedArgComparison->category()))
	    {
	      SimplificationOrdering::StoredConstraint* constraint = 
		SimplificationOrdering::current()->storedConstraintLess(_specialisedArgComparison);
	      if (constraint)
		{
		  integrate(cl,arg2ref->deref(),constraint);
		  SimplificationOrdering::current()->releaseConstraint(constraint);
		};
	    };
	};
    };
}; 
void ForwardDemodulation::Index::integrate(Clause* cl,
					   Term* t,
					   SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrate(Clause* cl,..)");
  FMCodeTree<Index::LeafInfo>::Leaf* 
    leaf = Integrate(t);
  Index::ConstrRuleList* crl = leaf->info().rules();  
#ifdef DEBUG_ALLOC
  ASSERT((!crl) || crl->checkObj());
#endif 
  crl = insertConstrRule(cl,constr,createStoredSubstitution(),t,crl);
  leaf->info().rules() = crl;
}; 
void ForwardDemodulation::Index::remove(Clause* cl) 
{ 
  CALL("remove(Clause* cl)"); 
#ifdef DEBUG_NAMESPACE
  if (cl->numOfNonanswerLiterals() > 1)
    {
      LiteralList::Iterator iter(cl->LitList());
      iter.next(); 
      do
	{
	  if ((!iter.currentLiteral()->IsSplitting()) &&
	      (!iter.currentLiteral()->isAnswerLiteral()))
	    {
	      ICP("E1");
	    };
	  iter.next();
	}
      while (iter.notEnd());
    };  
#endif          
  ASSERT(cl->LitList()->numOfSelectedLits() == 1); 
  LiteralList::Element* elem = cl->LitList()->firstElem();
  ASSERT(elem->positive());
  ASSERT(elem->isEquality());
  if (!elem->isOrientableEq()) return;
  if (elem->isOrientedEq())
    { 
      if (elem->equalityLHS()->isVariable()) return;
      remove(cl,elem->equalityLHS()->deref());
    }
  else
    {      
      Term* lit = cl->LitList()->firstLit(); 
      Term* arg1ref = lit->Arg1();
      Term* arg2ref = lit->Arg2();
      if (arg1ref->isVariable() || arg2ref->isVariable()) return;
      Term* arg1 = arg1ref->deref();
      Term* arg2 = arg2ref->deref();
      remove(cl,arg1);
      remove(cl,arg2);    
    };  
}; 
void ForwardDemodulation::Index::remove(Clause* cl,Term* lhs) 
{ 
  CALL("remove(Clause* cl,Term* lhs)");
  FMCodeTree<Index::LeafInfo>::Leaf* leaf;
  FMCodeTree<Index::LeafInfo>::Instruction** branch_to_destroy;
  if (Remove(lhs,leaf,branch_to_destroy))
    { 
      ASSERT(leaf->isLeaf());
      Index::ConstrRuleList* crl = leaf->info().rules();      
      ASSERT(crl);   
      crl = removeConstrRule(cl,lhs,crl);
      leaf->info().rules() = crl;
      if (!crl) 
	{
	  FMCodeTree<Index::LeafInfo>::DestroyBranch(branch_to_destroy);
	};
    };
}; 
inline 
bool 
ForwardDemodulation::Index::appropriateDegreeForGreater(const SimplificationOrdering::SpecialisedComparison::Category& cat) const 
{
  CALL("appropriateDegreeForGreater(const SimplificationOrdering::SpecialisedComparison::Category& cat) const");
  switch (cat)
    {
    case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
    case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
      return false;
    case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
      return true;
    case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
      return (_degree >= 4);
    case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
      return (_degree >= 2);
    case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:
      return (_degree >= 4);
    case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater:
      return (_degree >= 2);
    case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
      return (_degree >= 3);
    };
  ICP("ICP1");
  return false;
}; 
inline 
bool 
ForwardDemodulation::Index::appropriateDegreeForLess(const SimplificationOrdering::SpecialisedComparison::Category& cat) const 
{
  CALL("appropriateDegreeForLess(const SimplificationOrdering::SpecialisedComparison::Category& cat) const");
  switch (cat)
    {
    case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
      return true;
    case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
    case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
      return false;
    case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
      return (_degree >= 2);
    case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
      return (_degree >= 4);
    case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:
      return (_degree >= 2);
    case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater:
      return (_degree >= 4);
    case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
      return (_degree >= 3);
    };
  ICP("ICP1");
  return false;
}; 
inline 
FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf* 
ForwardDemodulation::Index::Integrate(Term* t)
{
  CALL("Integrate(Term* t)");
  ASSERT(t->isComplex()); 
  _matchingIndex.integrationReset(*t);
  Term::Iterator iter(t); 
  while (iter.Next()) 
    if (iter.CurrentSymIsVar()) 
      { _matchingIndex.integrationVar(iter.symbolRef()); }
    else _matchingIndex.integrationFunc(iter.symbolRef());
  bool newLeaf;
  FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf* res =  _matchingIndex.integrationGetLeaf(newLeaf);
  return res;
}; 
inline 
bool 
ForwardDemodulation::Index::Remove(Term* t,
				   FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf*& leaf,
				   FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Instruction**& branch_to_destroy)
{
  CALL("Remove(Term* t,FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf*& leaf,FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Instruction**& branch_to_destroy)");
  if (!_matchingIndex.removalReset(*t)) return false;
  Term::Iterator iter(t);
  while (iter.Next())  
    {    
      if (iter.CurrentSymIsVar())
	{
	  if (!_matchingIndex.removalVar(iter.symbolRef())) return false;
	}
      else if (!_matchingIndex.removalFunc(iter.symbolRef())) return false;
    };
  return _matchingIndex.removalFinish(leaf,branch_to_destroy);
}; 
inline 
FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf* 
ForwardDemodulation::Index::Match(const Flatterm* query) 
{ 
  CALL("Match(const Flatterm* query)");
  return _matchingIndex.Match(query);
}; 
inline 
FMCodeTree<ForwardDemodulation::Index::LeafInfo>::Leaf* 
ForwardDemodulation::Index::MatchAgain() 
{ 
  return _matchingIndex.MatchAgain();    
}; 
inline
ForwardDemodulation::Index::StoredSubstitution* 
ForwardDemodulation::Index::createStoredSubstitution()
{
  CALL("createStoredSubstitution()");
  StoredSubstitution* result = 0;
  for (ulong n = 0; 
       n < _matchingIndex.integrationGetVariables().numOfDiffElements(); 
       ++n)
    {
      ulong v = _matchingIndex.integrationGetVariables().nth(n);
      result = 
	new StoredSubstitution(VarInstanceLocationPair(v,
						       _matchingIndex.integrationMapVariable(v)),
			       result);
    };
  return result;
}; 
