/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardSubsumptionCode.hpp"
#include "Term.hpp"
#include "Clause.inl"
#include "Flatterm.hpp"
#include "DestructionMode.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCode"
#endif
#include "debugMacros.hpp"
ForwardSubsumptionCode::ForwardSubsumptionCode(const Flatterm** sbst) 
  : tree(0),
    subst(sbst),
    integrator(subst), 
    removal(subst),
    _maxlNumOfLits(0UL) 
{
  CALL("constructor ForwardSubsumptionCode(const Flatterm** sbst)");
};
ForwardSubsumptionCode::~ForwardSubsumptionCode() 
{ 
  CALL("destructor ~ForwardSubsumptionCode()"); 
  if (DestructionMode::isThorough())
    {
      if (tree) ForwardSubsumptionCommand::DestroyTree(tree);
    };
};
void ForwardSubsumptionCode::Remove(Clause* cl) 
{
  CALL("Remove(Clause* cl)"); 
  if (!tree) return;
  removal.Reset(&tree,cl);
  LiteralList::Iterator iter(cl->LitList());
  Term* lit;
 next_lit:
  if (iter.end()) 
    {
      if (removal.FindSuccess())
	{
	  removal.DestroyBranch();
	}
      else
	{ 
	  COP("COP1");
	  DMSG("Removed clause is not found in the FS index:\n"); 
	  DMSG(cl->Number());
	  DMSG("\n");
	  ICP(""); 
	};
      return;
    };
  lit = iter.currentLiteral();
  iter.next();
  if (removal.SkipLiteral(lit)) goto next_lit;
  COP("COP2");
  DMSG("Removed clause is not found in the FS index:\n"); 
  DMSG(cl->Number());
  DMSG("\n");
  ICP(""); 
}; 
void ForwardSubsumptionCode::Integrate(Clause* cl) 
{
  CALL("Integrate(Clause* cl)");
  if (cl->numOfAllLiterals() > _maxlNumOfLits) 
    _maxlNumOfLits = cl->numOfAllLiterals();
  integrator.Reset(&tree,cl);
  LiteralList::Iterator iter(cl->LitList());
  Term* lit;
  if (!tree) goto write_mode_next_lit;
 read_mode_next_lit:
  if (iter.end()) 
    {
      integrator.InsertSuccess(cl);
      return;
    }; 
  lit = iter.currentLiteral();
  iter.next();
  if (integrator.ReadLiteral(lit)) 
    {
      goto read_mode_next_lit;
    }
  else
    {
      ASSERT(!integrator.CurrInstr());
      goto write_mode_next_lit;
    }; 
 write_mode_next_lit:
  if (iter.end()) 
    {
      integrator.WriteSuccess(cl);
      return;
    }; 
  lit = iter.currentLiteral();
  iter.next();
  integrator.WriteLiteral(lit);
  goto write_mode_next_lit;
}; 
void ForwardSubsumptionCode::Integrate(const BasicTempLitList& component,void* indObj)
{
  CALL("Integrate(const BasicTempLitList& component,void* indObj)");
  if (component.length() > _maxlNumOfLits) 
    _maxlNumOfLits = component.length();
  integrator.Reset(&tree,component);
  const BasicTempLitList::Element* el = component.first();
  if (!tree) goto write_mode_next_lit;
 read_mode_next_lit:
  if (!el) 
    {
      integrator.InsertSuccess(indObj,component.length());
      return;
    }; 
  if (integrator.ReadLiteral(el->literal()->flatterm())) 
    {
      el = el->next();
      goto read_mode_next_lit;
    }
  else
    {
      ASSERT(!integrator.CurrInstr());
      el = el->next();
      goto write_mode_next_lit;
    }; 
 write_mode_next_lit:
  if (!el) 
    {
      integrator.WriteSuccess(indObj,component.length());
      return;
    }; 
  integrator.WriteLiteral(el->literal()->flatterm());  
  el = el->next();
  goto write_mode_next_lit;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCodeIntegrator"
#endif
#include "debugMacros.hpp"
ForwardSubsumptionCodeIntegrator::ForwardSubsumptionCodeIntegrator(const Flatterm** subst) 
  : constr_to_insert("constr_to_insert"),
    constr_to_shift("constr_to_shift"),
    constr_passed("constr_passed"),
    constr_passed_in_shift("constr_passed_in_shift"),
    constr_integrated("constr_integrated"),
    constr_old("constr_old"),
    fs_subst(subst)
{ 
  CALL("constructor ForwardSubsumptionCodeIntegrator(const Flatterm** subst)");
};
ForwardSubsumptionCodeIntegrator::~ForwardSubsumptionCodeIntegrator() 
{ 
  CALL("destructor ~ForwardSubsumptionCodeIntegrator()");  
  if (DestructionMode::isThorough())
    {
    };
};
void ForwardSubsumptionCodeIntegrator::WriteEqLiteral(Term* lit)
{
  CALL("WriteEqLiteral(Term* lit)");
  ASSERT(lit->IsEquality());
  Term* args = lit->Args();
  Term arg1 = *args;
  Term arg2 = *(args->Next());
  if (lit->IsOrderedEq())
    {
      WriteOrdEq(*lit); 
    }
  else
    {
      WriteUnordEq(*lit); 
    }; 
  WriteTerm(arg1);
  WriteSecondEqArg();
  WriteTerm(arg2);
}; 
void ForwardSubsumptionCodeIntegrator::WriteLiteral(Term* lit)
{ 
  if (lit->IsEquality()) { WriteEqLiteral(lit); return; };
  Term::Iterator iter;
  Term sym;
  if (lit->arity())
    {
      WriteHeader(*lit);
      iter.Reset(lit);
    write_mode_next_sym:
      if (!(iter.Next())) return;
      sym = iter.symbolRef();
      if (iter.CurrentSymIsVar())
	{
	  WriteVar(sym.var());
	}
      else 
	{
	  WriteFunc(sym);
	}; 
      goto write_mode_next_sym; 
    }
  else 
    {
      WritePropLit(*lit);
      return;
    }; 
}; 
void ForwardSubsumptionCodeIntegrator::WriteEqLiteral(const Flatterm* lit)
{
  CALL("WriteEqLiteral(Flatterm* lit)");
  ASSERT(lit->symbolRef().IsEquality());
  const Flatterm* arg1 = lit->arg1();
  const Flatterm* arg2 = lit->arg2();
  if (lit->symbolRef().IsOrderedEq())
    {
      WriteOrdEq(lit->symbolRef()); 
    }
  else
    {
      WriteUnordEq(lit->symbolRef()); 
    }; 
  WriteTerm(arg1);
  WriteSecondEqArg();
  WriteTerm(arg2);
}; 
void ForwardSubsumptionCodeIntegrator::WriteLiteral(const Flatterm* lit)
{ 
  if (lit->symbolRef().IsEquality()) { WriteEqLiteral(lit); return; };
  Flatterm::ConstIterator iter;
  if (lit->arity())
    {
      WriteHeader(lit->symbolRef());
      iter.Reset(lit);
    write_mode_next_sym:
      iter.NextSym();
      if (iter.EndOfTerm()) return;
      if (iter.CurrPos()->isVariable())
	{
	  WriteVar(iter.CurrPos()->Var());
	}
      else 
	{
	  WriteFunc(iter.CurrPos()->symbolRef());
	}; 
      goto write_mode_next_sym; 
    }
  else 
    {
      WritePropLit(lit->symbolRef());
      return;
    }; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadVar(const ulong& var)
{
  CALL("ReadVar(ulong var)");
  if (!(ReadVar())) 
    { 
      InsertVar(fs_var);
      if (vars.scoreOf(var))
	{
	  COP("was before");
	  EnterWriteMode(); 
	  WriteCompare(first_occ[var],fs_var);
	}
      else 
	{
	  COP("first occ");
	  vars.add(var);
	  first_occ[var] = fs_var;
	  EnterWriteMode(); 
	}; 
      fs_var++;
      return false; 
    };
  if (vars.scoreOf(var))
    {
      ReadCompare(first_occ[var],fs_var);
    } 
  else 
    {
      vars.add(var);
      first_occ[var] = fs_var;
    }; 
  fs_var++;
  return true;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadComplexTerm(Term* t)
{
  CALL("ReadComplexTerm(Term* t)");
  Term::Iterator iter;
  Term sym;
  iter.Reset(t);
  if (ReadFunc(*t)) 
    {
    read_mode_next_sym:
      if (!iter.Next()) return true; 
      sym = iter.symbolRef();
      if (iter.CurrentSymIsVar())
	{
	  if (!(ReadVar(sym.var()))) goto write_mode_next_sym; 
	}
      else 
	{
	  if (!(ReadFunc(sym))) 
	    { 
	      InsertFunc(sym);
	      COP("func inserted 1");
	      EnterWriteMode(); 
	      goto write_mode_next_sym; 
	    };
	}; 
      goto read_mode_next_sym;
    };
  InsertFunc(*t);
  COP("func inserted 2");
  EnterWriteMode();  
 write_mode_next_sym:
  if (!iter.Next()) return false;
  sym = iter.symbolRef();
  if (iter.CurrentSymIsVar())
    {
      WriteVar(sym.var());
    }
  else 
    {
      WriteFunc(sym);
    }; 
  goto write_mode_next_sym; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadComplexTerm(const Flatterm* t)
{
  CALL("ReadComplexTerm(const Flatterm* t)");
  Flatterm::ConstIterator iter;
  iter.Reset(t);
  if (ReadFunc(t->symbolRef())) 
    {
    read_mode_next_sym:
      iter.NextSym();
      if (iter.EndOfTerm()) return true; 
      if (iter.CurrPos()->isVariable())
	{
	  if (!(ReadVar(iter.CurrPos()->Var()))) goto write_mode_next_sym; 
	}
      else 
	{
	  if (!ReadFunc(iter.CurrPos()->symbolRef())) 
	    { 
	      InsertFunc(iter.CurrPos()->symbolRef());
	      COP("func inserted 1");
	      EnterWriteMode(); 
	      goto write_mode_next_sym; 
	    };
	}; 
      goto read_mode_next_sym;
    };
  InsertFunc(t->symbolRef());
  COP("func inserted 2");
  EnterWriteMode();  
 write_mode_next_sym:  
  iter.NextSym();
  if (iter.EndOfTerm()) return false;
  if (iter.CurrPos()->isVariable())
    {
      WriteVar(iter.CurrPos()->Var());
    }
  else 
    {
      WriteFunc(iter.CurrPos()->symbolRef());
    }; 
  goto write_mode_next_sym; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadEqLiteral(Term* lit)
{
  CALL("ReadEqLiteral(Term* lit)");
  ASSERT(lit->IsEquality());
  Term* args = lit->Args();
  Term arg1 = *args;
  Term arg2 = *(args->Next());
  if (lit->IsOrderedEq())
    {
      if (!ReadOrdEq(*lit)) { WriteOrdEq(*lit); goto write_mode_first_arg; };
    }
  else
    {
      if (!ReadUnordEq(*lit)) { WriteUnordEq(*lit); goto write_mode_first_arg; };
    }; 
  if (!ReadTerm(arg1)) goto write_mode_second_arg; 
  ReadSecondEqArg(); 
  return (ReadTerm(arg2)) && (ReadCollectedComps()); 
 write_mode_first_arg:
  WriteTerm(arg1);
 write_mode_second_arg:
  WriteSecondEqArg();
  WriteTerm(arg2);
  return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadLiteral(Term* lit)
{
  CALL("ReadLiteral(Term* lit)"); 
  ASSERT((!debugLitNum) || constr_integrated.Implies(constr_old));
  ASSERT((!debugLitNum) || constr_old.Implies(constr_passed));
  ASSERT((!debugLitNum) || constr_passed.Implies(constr_integrated));
  DOP(debugLitNum++);
  if (lit->IsEquality()) return ReadEqLiteral(lit);
  Term::Iterator iter;
  Term sym;
  if (lit->arity())
    {
      if (ReadHeader(*lit))
	{      
	  iter.Reset(lit);
	read_mode_next_sym:
	  if (!iter.Next()) 
	    {
	      return ReadCollectedComps();
	    };
	  sym = iter.symbolRef();
	  if (iter.CurrentSymIsVar())
	    {
	      if (!(ReadVar(sym.var()))) 
		{
		  goto write_mode_next_sym;
		}; 
	    }
	  else 
	    {
	      if (!(ReadFunc(sym))) 
		{ 
		  InsertFunc(sym);
		  EnterWriteMode(); 
		  ASSERT_IN(!CurrInstr(),"A30");
		  goto write_mode_next_sym; 
		};
	    }; 
	  goto read_mode_next_sym; 
	} 
      else 
	{
	  ASSERT_IN(!CurrInstr(),"A40");
	  WriteHeader(*lit);
	  goto write_mode_lit_args;
	};  
    }
  else 
    {
      if (ReadPropLit(*lit)) return true;
      ASSERT_IN(!CurrInstr(),"A50");
      WritePropLit(*lit);
      return false;
    }; 
 write_mode_lit_args: 
  iter.Reset(lit);
 write_mode_next_sym:       
  if (!(iter.Next())) return false;
  sym = iter.symbolRef();
  if (iter.CurrentSymIsVar())
    {
      WriteVar(sym.var());
    }
  else 
    {
      WriteFunc(sym);
    }; 
  goto write_mode_next_sym; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadEqLiteral(const Flatterm* lit)
{
  CALL("ReadEqLiteral(Flatterm* lit)");
  ASSERT(lit->symbolRef().IsEquality());
  const Flatterm* arg1 = lit->arg1();
  const Flatterm* arg2 = lit->arg2();
  if (lit->symbolRef().IsOrderedEq())
    {
      if (!ReadOrdEq(lit->symbolRef())) 
	{ 
	  WriteOrdEq(lit->symbolRef()); 
	  goto write_mode_first_arg; 
	};
    }
  else
    {
      if (!ReadUnordEq(lit->symbolRef())) 
	{ 
	  WriteUnordEq(lit->symbolRef()); 
	  goto write_mode_first_arg; 
	};
    }; 
  if (!ReadTerm(arg1)) goto write_mode_second_arg; 
  ReadSecondEqArg(); 
  return (ReadTerm(arg2)) && (ReadCollectedComps()); 
 write_mode_first_arg:
  WriteTerm(arg1);
 write_mode_second_arg:
  WriteSecondEqArg();
  WriteTerm(arg2);
  return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadLiteral(const Flatterm* lit)
{
  CALL("ReadLiteral(const Flatterm* lit)"); 
  ASSERT((!debugLitNum) || constr_integrated.Implies(constr_old));
  ASSERT((!debugLitNum) || constr_old.Implies(constr_passed));
  ASSERT((!debugLitNum) || constr_passed.Implies(constr_integrated));
  DOP(debugLitNum++);
  if (lit->symbolRef().IsEquality()) return ReadEqLiteral(lit);
  Flatterm::ConstIterator iter;
  if (lit->arity())
    {
      if (ReadHeader(lit->symbolRef()))
	{      
	  iter.Reset(lit);
	read_mode_next_sym:
          iter.NextSym();
	  if (iter.EndOfTerm()) return ReadCollectedComps();
	  if (iter.CurrPos()->isVariable())
	    {
	      if (!(ReadVar(iter.CurrPos()->Var())))
		goto write_mode_next_sym; 
	    }
	  else 
	    {
	      if (!ReadFunc(iter.CurrPos()->symbolRef())) 
		{ 
		  InsertFunc(iter.CurrPos()->symbolRef());
		  EnterWriteMode(); 
		  ASSERT_IN(!CurrInstr(),"A30");
		  goto write_mode_next_sym; 
		};
	    }; 
	  goto read_mode_next_sym; 
	} 
      else 
	{
	  ASSERT_IN(!CurrInstr(),"A40");
	  WriteHeader(lit->symbolRef());
	  goto write_mode_lit_args;
	};  
    }
  else 
    {
      if (ReadPropLit(lit->symbolRef())) return true;
      ASSERT_IN(!CurrInstr(),"A50");
      WritePropLit(lit->symbolRef());
      return false;
    }; 
 write_mode_lit_args: 
  iter.Reset(lit);
 write_mode_next_sym:  
  iter.NextSym();     
  if (iter.EndOfTerm()) return false;
  if (iter.CurrPos()->isVariable())
    {
      WriteVar(iter.CurrPos()->Var());
    }
  else 
    {
      WriteFunc(iter.CurrPos()->symbolRef());
    }; 
  goto write_mode_next_sym; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadHeader(const Term& header)
{
  CALL("ReadHeader(const Term& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  ASSERT(!header.IsEquality()); 
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::LitHeader :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  MoveDown(); 
	  return true; 
	};
    case ForwardSubsumptionCommand::PropLit :  
    case ForwardSubsumptionCommand::UnordEq :  
    case ForwardSubsumptionCommand::OrdEq :  
      if (curr_instr->Fork())
	{
	  MoveRight();
	  goto check_tag;
	};
      MoveRight();
      return false;
    case ForwardSubsumptionCommand::Success : 
      DWARN("Clause being integrated into FS index is subsumed (in set mode) by already integrated\n");
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif  
    };
	ICP("END");
	return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadPropLit(const Term& header)
{
  CALL("ReadPropLit(const Term& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::PropLit :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  MoveDown(); 
	  return true; 
	};
    case ForwardSubsumptionCommand::LitHeader : 
    case ForwardSubsumptionCommand::UnordEq : 
    case ForwardSubsumptionCommand::OrdEq :
      if (curr_instr->Fork())
	{
	  MoveRight();
	  goto check_tag;
	};
      MoveRight();
      return false;
    case ForwardSubsumptionCommand::Success : 
      DWARN("Clause being integrated into FS index is subsumed (in set mode) by already integrated\n");
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif  
    };
	ICP("END");
	return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadUnordEq(const Term& header)
{
  CALL("ReadUnordEq(const Term& header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::UnordEq :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  MoveDown(); 
	  return true; 
	};
    case ForwardSubsumptionCommand::LitHeader : 
    case ForwardSubsumptionCommand::PropLit : 
    case ForwardSubsumptionCommand::OrdEq :   
      if (curr_instr->Fork())
	{
	  MoveRight();
	  goto check_tag;
	};
      MoveRight();
      return false;
    case ForwardSubsumptionCommand::Success : 
      DWARN("Clause being integrated into FS index is subsumed (in set mode) by already integrated\n");
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif  
    };
	ICP("END");
	return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadOrdEq(Term header)
{
  CALL("ReadOrdEq(Term header)");
  ASSERT(*node_addr);
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Identity());
  ForwardSubsumptionCommand* curr_instr;
  ulong header_num = header.HeaderNum();
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::OrdEq :
      if (curr_instr->HeaderNum() == header_num) 
	{
	  MoveDown(); 
	  header.MakeUnordered();
	  return true; 
	};
    case ForwardSubsumptionCommand::LitHeader : 
    case ForwardSubsumptionCommand::PropLit : 
    case ForwardSubsumptionCommand::UnordEq :   
      if (curr_instr->Fork())
	{
	  MoveRight();
	  goto check_tag;
	};
      MoveRight();
      header.MakeUnordered();
      return false;
    case ForwardSubsumptionCommand::Success : 
      DWARN("Clause being integrated into FS index is subsumed (in set mode) by already integrated\n");
      MoveRight();
      if (*node_addr) { goto check_tag; };
      return false;
#ifdef DEBUG_NAMESPACE
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif  
    };
	ICP("END");
	return false;
}; 
void ForwardSubsumptionCodeIntegrator::ReadSecondEqArg()
{
  CALL("ReadSecondEqArg()");
  ForwardSubsumptionCommand* curr_instr;
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::SecondEqArg : MoveDown(); return;
    case ForwardSubsumptionCommand::CompareNF :
      PassCompNF(curr_instr);
      goto check_tag; 
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return;
#endif 
#endif 
    }; 
}; 
bool ForwardSubsumptionCodeIntegrator::ReadFunc(const Term& func)
{
  CALL("ReadFunc(const Term& func)");
  ForwardSubsumptionCommand* curr_instr;
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::Func :
      if (curr_instr->func() == func) 
	{
	  DOP(CheckConstrIntegrity());
	  if (curr_instr->Fork()) 
	    {
	      if (!constr_to_shift.Identity())  
		{       
		  constr_passed_in_shift = constr_passed; 
		  CopyCompsToShift(curr_instr->ForkAddr()); 
		};
	    };  
	  MoveDown(); 
	  return true; 
	};
      if (curr_instr->func() > func)
	{
	  if (!constr_to_shift.Identity())  
	    {
	      constr_passed_in_shift = constr_passed;
	      InsertCompsToShift(node_addr);
	    };
	  curr_instr = *node_addr;
	  return false;
	};  
    case ForwardSubsumptionCommand::Var :
      DOP(CheckConstrIntegrity());
      if (curr_instr->Fork())
	{
	  if (!constr_to_shift.Identity())
	    {
	      constr_passed_in_shift = constr_passed;
	      CopyCompsToShift(curr_instr->NextAddr());
	    };
	  MoveRight();
	  goto check_tag;
	};
      if (!constr_to_shift.Identity())
	{
	  constr_passed_in_shift = constr_passed;
	  InsertCompsToShift(curr_instr->NextAddr());
	};    
      MoveRight();
      return false;
    case ForwardSubsumptionCommand::CompareNF :
      PassCompNF(curr_instr);
      goto check_tag;
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif 
    };
	ICP("END");
	return false;
}; 
bool ForwardSubsumptionCodeIntegrator::ReadVar()
{
  CALL("ReadVar()");
  ForwardSubsumptionCommand* curr_instr;
 check_tag:
  curr_instr = *node_addr;
  switch (curr_instr->Tag())
    {
    case ForwardSubsumptionCommand::Var : 
      DOP(CheckConstrIntegrity());
      if (curr_instr->Fork()) 
	{ 
	  if (!constr_to_shift.Identity())
	    {
	      constr_passed_in_shift = constr_passed; 
	      CopyCompsToShift(curr_instr->ForkAddr()); 
	    };
	};
      MoveDown(); 
      return true;
    case ForwardSubsumptionCommand::Func :
      if (!constr_to_shift.Identity())
	{
	  constr_passed_in_shift = constr_passed;
	  InsertCompsToShift(node_addr);
	};
      curr_instr = *node_addr;
      return false;
    case ForwardSubsumptionCommand::CompareNF :
      PassCompNF(curr_instr);
      goto check_tag;
#ifdef DEBUG_NAMESPACE 
    default : ICP("X0"); return false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default : return false;
#endif 
#endif 
    };
	ICP("END");
	return false;
}; 
void ForwardSubsumptionCodeIntegrator::ReadCompare(const Flatterm** v1,const Flatterm** v2)
{
  CALL("ReadCompare(const Flatterm** v1,const Flatterm** v2)");
  ASSERT(v1 < v2);
  ulong var_num1 = VarNum(v1);
  ulong var_num2 = VarNum(v2);
  MakeIntegrated(var_num1,var_num2);
  constr_to_insert.MakeEquivalent(var_num1,var_num2); 
  DOP(CheckConstrIntegrity());
}; 
bool ForwardSubsumptionCodeIntegrator::ReadCollectedComps()
{
  CALL("ReadCollectedComps()");
  DOP(CheckConstrIntegrity());
  ForwardSubsumptionCommand* curr_instr;
 check_instr:
  curr_instr = *node_addr;
  if (curr_instr->IsCompare() || curr_instr->IsCompareNF())
    {
      if (PassComp(curr_instr) || (*node_addr)) goto check_instr;
      DOP(CheckConstrIntegrity());
      MoveCompsToInsert(node_addr);
      constr_to_shift.Reset();
      ASSERT(!(*node_addr));
      return false;
    };
  if (constr_to_shift.Identity())
    {      
      ForwardSubsumptionCommand** node_addr_backup = node_addr;
      *node_addr = 0;
      DOP(CheckConstrIntegrity());
      if (MoveCompsToInsert(node_addr))
	{
	  ASSERT(!(*node_addr));
	  ASSERT((*node_addr_backup)->IsCompareNF());
	  (*node_addr_backup)->SetTag(ForwardSubsumptionCommand::Compare);
	  (*node_addr_backup)->SetFork(curr_instr);
	  return false;
	}
      else 
	{
	  ASSERT(node_addr == node_addr_backup);
	  ASSERT(!(*node_addr));
	  *node_addr = curr_instr;          
	  return true; 
	};
    }
  else 
    {    
      constr_passed_in_shift = constr_passed;
      InsertCompsToShift(node_addr);
      ASSERT((*node_addr)->IsCompareNF());    
      (*node_addr)->SetTag(ForwardSubsumptionCommand::Compare);       
      ASSERT((!constr_integrated.Equivalent(VarNum((*node_addr)->Var1()),VarNum((*node_addr)->Var2()))));
      MoveRight();
      MoveCompsToInsert(node_addr);
      ASSERT(!(*node_addr));
      return false; 
    };
}; 
inline void ForwardSubsumptionCodeIntegrator::EnterWriteMode()
{
  CALL("EnterWriteMode()");
  ASSERT(constr_to_shift.Identity());
  ASSERT(constr_to_insert.Implies(constr_integrated));
  ASSERT(constr_passed.Implies(constr_integrated));
  ASSERT(constr_passed.Implies(constr_old));
  MoveCompsToInsert(node_addr);
  COP("END1");
  ASSERT(!(*node_addr));
}; 
inline void ForwardSubsumptionCodeIntegrator::MakeShifted(const ulong& v1,const ulong& v2) 
{
  CALL("MakeShifted(const ulong& v1,const ulong& v2)");
  ASSERT(v1 < v2);
  ASSERT(!constr_integrated.Equivalent(v1,v2));
  constr_to_shift.MakeEquivalent(v1,v2);
}; 
#ifdef DEBUG_NAMESPACE
inline void ForwardSubsumptionCodeIntegrator::CheckConstrIntegrity()
{ 
  CALL("CheckConstrIntegrity()");
  ASSERT(constr_to_insert.Implies(constr_integrated));      
  ASSERT(constr_passed.Implies(constr_integrated));   
  ASSERT(constr_to_shift.Implies(constr_old));        
  ASSERT(constr_passed.Implies(constr_old));
  static VarEquivalence debugER("debugER");  
  debugER.MakeComposition(constr_to_shift,constr_passed);
  ASSERT(constr_old.Implies(debugER));
  debugER.MakeComposition(constr_to_insert,constr_passed);
  ASSERT(constr_integrated.Implies(debugER));
}; 
#endif 
inline void ForwardSubsumptionCodeIntegrator::PassCompNF(ForwardSubsumptionCommand* instr)
{
  CALL("PassCompNF(ForwardSubsumptionCommand* instr)");
  DOP(CheckConstrIntegrity());
  ASSERT(instr->IsCompareNF());
  const Flatterm** var1 = instr->Var1();
  const Flatterm** var2 = instr->Var2();
  const ulong var_num1 = VarNum(var1);
  const ulong var_num2 = VarNum(var2);
  ASSERT(!(instr->Fork()));
  ASSERT(!constr_to_shift.Equivalent(var_num1,var_num2));
  ASSERT(!constr_old.Equivalent(var_num1,var_num2));
  ASSERT(!constr_passed.Equivalent(var_num1,var_num2));
  bool shared = constr_integrated.Equivalent(var_num1,var_num2);
  if (shared)
    {
      constr_passed.MakeEquivalent(var_num1,var_num2);
      node_addr = instr->NextAddr();
    }
  else 
    {
      *node_addr = instr->Next();
      delete instr;
    };
  instr = *node_addr;
  VarEquivalence::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
  ulong merged1;
  ulong merged2;
  while (mergedPairs.nextPair(merged1,merged2))
    {
      ASSERT(merged1 < merged2);     
      ASSERT(!constr_old.Equivalent(merged1,merged2));
      if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
	{
	  shared = true;
	  constr_passed.MakeEquivalent(merged1,merged2);
	  *node_addr = ForwardSubsumptionCommand::NewCompareNF(Var(merged1),Var(merged2));
	  (*node_addr)->SetNext(instr);
	  MoveDown();
	};
    };    
  constr_old.MakeEquivalent(var_num1,var_num2);
  if (!shared) MakeShifted(var_num1,var_num2);
  ASSERT((*node_addr) == instr);
  DOP(CheckConstrIntegrity());
}; 
inline bool ForwardSubsumptionCodeIntegrator::PassComp(ForwardSubsumptionCommand* instr)
{
  CALL("PassComp(ForwardSubsumptionCommand* instr)");
  DOP(CheckConstrIntegrity()); 
  ASSERT(instr->IsCompareNF() ||instr->IsCompare() );
  const Flatterm** var1 = instr->Var1();
  const Flatterm** var2 = instr->Var2();
  ulong var_num1 = VarNum(var1);
  ulong var_num2 = VarNum(var2);
  ASSERT(!constr_to_shift.Equivalent(var_num1,var_num2));
  ASSERT(!constr_old.Equivalent(var_num1,var_num2)); 
  ASSERT(!constr_passed.Equivalent(var_num1,var_num2));
  ASSERT((*node_addr) == instr);
  if (constr_integrated.Equivalent(var_num1,var_num2))
    {
      goto shift_and_insert_shared;      
    }
  else 
    {
      VarEquivalence::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
      ulong merged1;
      ulong merged2; 
      while (mergedPairs.nextPair(merged1,merged2))
	{
	  ASSERT(merged1 < merged2);     
	  ASSERT(!constr_old.Equivalent(merged1,merged2));
	  if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
	    {
	      var_num1 = merged1;
	      var_num2 = merged2;
	      instr->SetVar1(Var(var_num1));
	      instr->SetVar2(Var(var_num2));         
	      goto shift_and_insert_shared;
	    };
	};
      if (instr->IsCompareNF()) instr->SetTag(ForwardSubsumptionCommand::Compare); 
      if (!constr_to_shift.Identity())
	{
	  constr_passed_in_shift = constr_passed;
	  constr_passed_in_shift.MakeEquivalent(var_num1,var_num2);
	  CopyCompsToShift(instr->NextAddr());
	};
      MoveRight();  
      DOP(CheckConstrIntegrity());
      return false;
    };
 shift_and_insert_shared:
  COP("shift_and_insert_shared");
  if (instr->Fork()) 
    {
      if (!constr_to_shift.Identity())
        {
	  constr_passed_in_shift = constr_passed; 
	  CopyCompsToShift(instr->ForkAddr());  
        };      
    };
  constr_passed.MakeEquivalent(var_num1,var_num2);
  MoveDown();
  instr = *node_addr; 
  VarEquivalence::MergedPairs mergedPairs(constr_old,var_num1,var_num2);
  ulong merged1;
  ulong merged2; 
  while (mergedPairs.nextPair(merged1,merged2))
    {
      ASSERT(merged1 < merged2);     
      ASSERT(!constr_old.Equivalent(merged1,merged2));
      if (constr_integrated.Equivalent(merged1,merged2) && (!constr_passed.Equivalent(merged1,merged2)))
        {
	  constr_passed.MakeEquivalent(merged1,merged2);
	  *node_addr = ForwardSubsumptionCommand::NewCompareNF(Var(merged1),Var(merged2));
	  (*node_addr)->SetNext(instr);
	  MoveDown();
        };
    };    
  ASSERT((*node_addr) == instr);
  constr_old.MakeEquivalent(var_num1,var_num2); 
  COP("end");
  DOP(CheckConstrIntegrity());     
  return true;
}; 
inline void ForwardSubsumptionCodeIntegrator::CopyCompsToShift(ForwardSubsumptionCommand** addr)
{
  CALL("CopyCompsToShift(ForwardSubsumptionCommand** addr)");
  DOP(bool debugAtLeastOneShifted = constr_to_shift.Identity());  
  ASSERT(constr_passed.Implies(constr_passed_in_shift));
  ulong vn1;
  ulong vn2;
  VarEquivalence::Iterator iter(constr_to_shift);
  ForwardSubsumptionCommand* old_instr = *addr;
  while (iter.Next(vn2,vn1)) 
    {
      ASSERT(vn1 < vn2); 
      ASSERT(!constr_integrated.Equivalent(vn1,vn2));
      ASSERT(constr_old.Equivalent(vn1,vn2));
      if (!constr_passed_in_shift.Equivalent(vn1,vn2))
	{
	  DOP(debugAtLeastOneShifted = true);  
	  constr_passed_in_shift.MakeEquivalent(vn1,vn2);
	  ForwardSubsumptionCommand* new_instr = ForwardSubsumptionCommand::NewCompareNF(Var(vn1),Var(vn2));
	  *addr = new_instr;
	  addr = new_instr->NextAddr();
        }; 
    };
  *addr = old_instr;
  ASSERT(debugAtLeastOneShifted);
}; 
inline void ForwardSubsumptionCodeIntegrator::InsertCompsToShift(ForwardSubsumptionCommand** addr)
{
  CALL("InsertCompsToShift(ForwardSubsumptionCommand** addr)");
  DOP(bool debugAtLeastOneShifted = constr_to_shift.Identity());  
  DOP(CheckConstrIntegrity());     
  ASSERT(constr_passed.Implies(constr_passed_in_shift));
  ulong vn1;
  ulong vn2;
  ForwardSubsumptionCommand* old_instr = *addr;
  while (constr_to_shift.Split(vn2,vn1)) 
    { 
      ASSERT(vn1 < vn2);
      ASSERT(!constr_integrated.Equivalent(vn1,vn2));
      ASSERT(constr_old.Equivalent(vn1,vn2));
      if (!constr_passed_in_shift.Equivalent(vn1,vn2))
	{
	  DOP(debugAtLeastOneShifted = true);  
	  constr_passed_in_shift.MakeEquivalent(vn1,vn2);
	  ForwardSubsumptionCommand* new_instr = ForwardSubsumptionCommand::NewCompareNF(Var(vn1),Var(vn2));
	  *addr = new_instr;
	  addr = new_instr->NextAddr(); 
        };
    };
  ASSERT(constr_to_shift.Identity());
  *addr = old_instr;
  ASSERT(debugAtLeastOneShifted);
}; 
inline bool ForwardSubsumptionCodeIntegrator::MoveCompsToInsert(ForwardSubsumptionCommand**& addr)
{
  CALL("MoveCompsToInsert(ForwardSubsumptionCommand**& addr)");
  ulong vn1;
  ulong vn2;
  ForwardSubsumptionCommand* old_instr = *addr;
  bool res = false;
  while (constr_to_insert.Split(vn2,vn1)) 
    { 
      ASSERT(constr_integrated.Equivalent(vn1,vn2));
      ASSERT(vn1 < vn2);  
      if (!constr_passed.Equivalent(vn1,vn2))
	{
	  ASSERT(!constr_old.Equivalent(vn1,vn2));
	  constr_passed.MakeEquivalent(vn1,vn2);
	  ForwardSubsumptionCommand* new_instr = ForwardSubsumptionCommand::NewCompareNF(Var(vn1),Var(vn2));
	  *addr = new_instr;
	  addr = new_instr->NextAddr();
	  res = true;
        };
    };
  ASSERT(constr_to_insert.Identity());
  *addr = old_instr;
  COP("end");
  ASSERT(constr_to_insert.Identity());
  ASSERT(constr_passed.Implies(constr_integrated));
  ASSERT(constr_integrated.Implies(constr_passed));
  return res;
}; 
inline void ForwardSubsumptionCodeIntegrator::InsertVar(const Flatterm** v)
{
  ForwardSubsumptionCommand* tmp = *node_addr;
  *node_addr = ForwardSubsumptionCommand::NewVar(v);
  (*node_addr)->SetFork(tmp);
  MoveDown();
}; 
#ifndef NO_DEBUG_VIS
ostream& ForwardSubsumptionCodeIntegrator::outputComps(ostream& str) 
{
  constr_integrated.Normalize();
  constr_old.Normalize();
  constr_passed.Normalize();
  constr_to_shift.Normalize();
  constr_to_insert.Normalize();
  return str << constr_integrated << '\n'
	     << constr_old << '\n'
	     << constr_passed << '\n'
	     << constr_to_shift << '\n'
	     << constr_to_insert << '\n';      
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_CODE
#define DEBUG_NAMESPACE "ForwardSubsumptionCodeRemoval"
#endif
#include "debugMacros.hpp"
ForwardSubsumptionCodeRemoval::ForwardSubsumptionCodeRemoval(const Flatterm** subst)
  : constr(), 
  fs_subst(subst)
{
  CALL("constructor ForwardSubsumptionCodeRemoval(const Flatterm** subst)");
};
ForwardSubsumptionCodeRemoval::~ForwardSubsumptionCodeRemoval() { CALL("destructor ~FS_CODE_REMOVAL()"); };
bool ForwardSubsumptionCodeRemoval::SkipComplexTerm(Term* t)
{
  Term::Iterator iter;
  iter.Reset(t);
 skip_sym:
  if (iter.CurrentSymIsVar())
    {
      if (!SkipVar(iter.symbolRef().var())) return false;
    }
  else 
    {
      if (!(SkipFunc(iter.symbolRef()))) return false;
    };
  if (iter.Next()) goto skip_sym; 
  return true; 
}; 
bool ForwardSubsumptionCodeRemoval::SkipEqLiteral(Term* lit)
{
  ASSERT(lit->IsEquality());               
  if (lit->IsOrderedEq())
    {
      if (!(SkipOrdEq(*lit))) return false;
    }
  else
    {
      if (!(SkipUnordEq(*lit))) return false;
    }; 
  Term* args = lit->Args();
  Term arg1 = *args;
  Term arg2 = *(args->Next());
  return (SkipTerm(arg1)) && (SkipSecondEqArg()) && (SkipTerm(arg2));
}; 
bool ForwardSubsumptionCodeRemoval::SkipLiteral(Term* lit)
{
  if (lit->IsEquality()) return SkipEqLiteral(lit);
  Term::Iterator iter;
  Term sym;
  if (lit->arity())
    {
      if (SkipHeader(*lit))
	{
	  iter.Reset(lit);
	next_sym:
	  if (!(iter.Next())) return true;
	  sym = iter.symbolRef();
	  if (iter.CurrentSymIsVar())
	    {
	      if (!SkipVar(sym.var())) return false;
	    }
	  else 
	    {
	      if (!(SkipFunc(sym))) return false;
	    };
	  goto next_sym;   
	}
      else return false;
    }
  else 
    {
      return SkipPropLit(*lit);
    };
}; 
