/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ForwardSuperposition.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
#include "Clause.inl"
using namespace BK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "ForwardSuperposition"
#endif
#include "debugMacros.hpp"
VK::ForwardSuperposition::ForwardSuperposition(UnifIndex* unifIndex) : 
  OPEN_BR(1,"br 1") DCOMMA
  _statistics(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _noDuplicateRedexesFlag(false),
  _noInfBetweenRulesFlag(false),
  _useInputNonBasicness(false),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _mainPremiseVarIndex(0UL), 
  _mainSelectedLiterals(DOP("ForwardSuperposition::_mainSelectedLiterals")),
  _mainSelectedLitProperties(DOP("ForwardSuperposition::_mainSelectedLitProperties")),
  _eligibleSubtermPositions("_eligibleSubtermPositions"),
  _sidePremiseVarIndex(1UL),
  _pseudoHeader((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity),
  _dummyClauseCollector(DOP("ForwardSuperposition::_dummyClauseCollector"))
{
  CALL("constructor ForwardSuperposition(UnifIndex* unifIndex)");
  _mainPremiseResidualSizeFormula.SetIndex(_mainPremiseVarIndex);
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);  
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _storedConstraintCheck = 
    SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; 
VK::ForwardSuperposition::~ForwardSuperposition() 
{ 
  CALL("destructor ~ForwardSuperposition()"); 
  if (DestructionMode::isThorough())
    {
      ASSERT(_storedConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
    };
};
bool VK::ForwardSuperposition::Load(Clause* cl)
{
  CALL("Load(Clause* cl)");
  DOP(_mainSelectedLiterals.unfreeze());
  if (!_mainSelectedLiterals.tryToExpand(cl->numOfAllLiterals()))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return false;
    };
  DOP(_mainSelectedLiterals.freeze());   
  DOP(_mainSelectedLitProperties.unfreeze());
  if (!_mainSelectedLitProperties.tryToExpand(cl->numOfAllLiterals()))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return false;
    };
  DOP(_mainSelectedLitProperties.freeze());   
  _equalityResolutionApplicable = false;
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _numOfSelectedMainLits = _allMainPremiseLiterals->numOfSelectedLits();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_numOfSelectedMainLits);
  LiteralList::Iterator iter(_allMainPremiseLiterals);  
  _mainSelectedLiterals.reset();
  _freeMainSelectedLitMemory = _mainSelectedLiteralMemory;
  for (ulong ln = 0; ln < _numOfSelectedMainLits; ln++)
    {
      ASSERT(iter.notEnd());
      _mainSelectedLiterals.push(_freeMainSelectedLitMemory);
      UnifTerm::StandardTermToUnifTerm(_mainPremiseVarIndex,iter.currentLiteral(),_freeMainSelectedLitMemory);
      ASSERT(_freeMainSelectedLitMemory - _mainSelectedLiteralMemory <= (long)VampireKernelConst::MaxClauseSize);
      _mainSelectedLitProperties[ln] = iter.currentElement()->properties();
      if ((iter.currentLiteral()->Negative())
	  && (iter.currentLiteral()->IsEquality())
	  && (iter.currentLiteral()->IsUnorderedEq()))
	_equalityResolutionApplicable = true;
      iter.next();   };
  compileMainPremiseResidualSizeFormula();
  _nextMainSelectedLitNum = _numOfSelectedMainLits; 
  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();
  UnifTerm* lit;
  LiteralProperties prop;
 try_next_lit:
  lit = nextSelectedMainLit(prop);
  if (!lit) 
    {
      if (_equalityResolutionApplicable)
	{
	  _nextMainSelectedLitNum = _numOfSelectedMainLits; 
	  _continuation = ContEqRes;
	  return true;
	}
      else
	return false;
    };
  if (LoadLiteral(lit,prop)) { _continuation = ContFSup; return true; };
  goto try_next_lit; 
}; 
bool VK::ForwardSuperposition::NextInference()
{ 
  CALL("NextInference()");
  switch (_continuation)
    {
    case ContFSup:
      while (NextFSupInference())
	if (CollectResult()) return true;
      if (!_equalityResolutionApplicable) return false;
      _nextMainSelectedLitNum = _numOfSelectedMainLits; 
      _continuation = ContEqRes;
    case ContEqRes:
      while (NextEqRes())
	{
	  if (CollectEqResResult())
	    {
	      Unifier::current()->UnbindLastSection();         
	      return true;
	    };
	  Unifier::current()->UnbindLastSection(); 
	}; 
      return false;
    };
  ICP("ICP1");
  return false;
}; 
void VK::ForwardSuperposition::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();
  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());
  _nextMainSelectedLitNum = _numOfSelectedMainLits; 
  LiteralProperties prop;
  for (UnifTerm* m_lit = nextSelectedMainLit(prop); 
       m_lit; 
       m_lit = nextSelectedMainLit(prop))
    {
      Term header = m_lit->symbol().Func();
      if (header.IsEquality())
	{
	  if (prop.isPreorientedEq())
	    {
	      UnifTerm* rhs;
	      if (prop.isReverseOrientedEq())
		{
		  rhs = m_lit->Arg1();
		}
	      else
		rhs = m_lit->Arg2();
	      _mainPremiseResidualSizeFormula.Compile(rhs);
	    }
	  else
	    if (prop.isOrientableEq() && header.IsOrderedEq())
	      {
		_mainPremiseResidualSizeFormula.Compile(m_lit->Arg2());
	      };
	}; 
    };      
}; 
bool VK::ForwardSuperposition::NextFSupInference()
{
  if (Paramodulate()) return true;  
  UnifTerm* lit;
  LiteralProperties prop;
 try_next_lit:
  lit = nextSelectedMainLit(prop);
  if (!lit) return false;
  if (LoadLiteral(lit,prop) && (Paramodulate())) 
    return true;
  goto try_next_lit; 
}; 
bool VK::ForwardSuperposition::NextEqRes()
{
  CALL("NextEqRes()");
  UnifTerm* lit; 
  LiteralProperties dummyLitProp;
 try_next_lit:
  lit = nextSelectedMainLit(dummyLitProp);
  if (!lit) return false;
  Term header =  lit->symbol().Func();
  if (header.Negative() && header.IsEquality() && header.IsUnorderedEq())
    {
      UnifTerm* arg1 = lit+1;
      UnifTerm* arg2 = arg1->Bracket();
      if (Unification::Unify(arg1,arg2)) 
	{
	  return true;
	};
    };
  goto try_next_lit; 
}; 
bool VK::ForwardSuperposition::LoadLiteral(UnifTerm* lit,const LiteralProperties& prop)
{
  CALL("LoadLiteral(UnifTerm* lit,const LiteralProperties& prop)");
  if (!lit->symbol().Func().arity()) return false; 
  _currentMainLiteral = lit;
  _eligibleSubtermPositions.reset();
  if (lit->symbol().Func().IsEquality())
    {
      _eqArg1 = lit->Arg1();
      _eqArg2 = lit->Arg2();
      bool notToBeOriented = 
	lit->symbol().Func().Positive() && !prop.isOrientableEq();
      if (!notToBeOriented && 
	  (lit->symbol().Func().IsOrderedEq() || prop.isPreorientedEq()))
	{
	  _currentCompiledConstraint = 0;
	  if (prop.isReverseOrientedEq())
	    {
	      loadEligibleSubtermPositions(_eqArg2);
	    }
	  else
	    {
	      loadEligibleSubtermPositions(_eqArg1);
	    };
	  _swapArguments = false;
	}
      else 
	{ 
	  if (_eqArg1->symbol().isVariable() || _eqArg2->symbol().isVariable())
	    {   
	      _currentCompiledConstraint = 0;
	      if (_eqArg1->symbol().isVariable())
		{
		  loadEligibleSubtermPositions(_eqArg2);
		}
	      else
		loadEligibleSubtermPositions(_eqArg1);
	      _swapArguments = false;
	    }
	  else 
	    {
	      _currentCompiledConstraint = 0;
	      if (!notToBeOriented)
		{
		  _compiledConstraint->load(_eqArg2,_eqArg1);
		  if (_compiledConstraint->rhsVarInstMayOrder())
		    {
		      _compiledConstraint->compileGreaterOrEqual();
		      _currentCompiledConstraint = _compiledConstraint;
		    };
		};
	      loadEligibleSubtermPositions(_eqArg1);
	      _swapArguments = true;
	    };
	};
    }
  else 
    {
      _currentCompiledConstraint = 0;
      UnifTerm* endOfLit = lit->Bracket();
      UnifTerm* arg = lit + 1; 
      ulong argNum = 0UL;
      const bool* nonBasicArgs = lit->symbol().Func().nonBasicArguments();
      while (arg != endOfLit)
	{
	  if (!_useInputNonBasicness ||
	      !nonBasicArgs ||
	      !nonBasicArgs[argNum])
	    loadEligibleSubtermPositions(arg);
	  arg = arg->Bracket();
	  ++argNum;
	};
      _swapArguments = false;
    };
  while (!_eligibleSubtermPositions.empty())
    {
      if (LoadTermToRewrite(_eligibleSubtermPositions.pop()))
	{
	  _currentMainLitProperties = prop;
	  return true;
	}
      else
	++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
    };
  return false;
}; 
bool VK::ForwardSuperposition::LoadRestOfLit()
{
  CALL("LoadRestOfLit()");
  if (!_swapArguments) return false;
  _swapArguments = false;
  _currentCompiledConstraint = 0;
  _compiledConstraint->load(_eqArg1,_eqArg2);
  if (_compiledConstraint->rhsVarInstMayOrder())
    {
      _compiledConstraint->compileGreaterOrEqual();
      _currentCompiledConstraint = _compiledConstraint;
    };
  loadEligibleSubtermPositions(_eqArg2);
  return !_eligibleSubtermPositions.empty();
}; 
inline
void VK::ForwardSuperposition::loadEligibleSubtermPositions(UnifTerm* term)
{
  CALL("loadEligibleSubtermPositions(UnifTerm* term)");
  if (term->symbol().isVariable()) return;
  if (_noDuplicateRedexesFlag)
    {
      for (UnifTerm** prev =  _eligibleSubtermPositions.begin();
           prev != _eligibleSubtermPositions.end();
           ++prev)
	{
	  if (UnifTerm::identical(*prev,term)) return;
	};
    };
  _eligibleSubtermPositions.push(term);
  UnifTerm* endOfTerm = term->Bracket();
  UnifTerm* arg = term + 1; 
  ulong argNum = 0UL;
  const bool* nonBasicArgs = term->symbol().Func().nonBasicArguments();
  while (arg != endOfTerm)
    {
      if (!_useInputNonBasicness ||
	  !nonBasicArgs ||
	  !nonBasicArgs[argNum])
	loadEligibleSubtermPositions(arg);
      arg = arg->Bracket();
      ++argNum;
    };
}; 
bool VK::ForwardSuperposition::Paramodulate() 
{
  CALL("Paramodulate()");
 try_another_candidate:
  _currentCandidate = NextCandidate(_currentCompiledConstraint);
  if (_currentCandidate) 
    {
      _sidePremise = _currentCandidate->TheClause();
      if (_noRulesAsSidePremises && _sidePremise->IsRule())
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.discardedForOtherReasons);
	  goto try_another_candidate;
	};
      if (demodulateSelectedLits())
	goto try_another_candidate;
      if (subsumeSidePremise())
	goto try_another_candidate;
      return true; 
    }
  else 
    {
    try_another_position:
      if (_eligibleSubtermPositions.empty()) 
	{
	  if (!LoadRestOfLit()) return false;
	};
      ASSERT(!_eligibleSubtermPositions.empty());
      if (!LoadTermToRewrite(_eligibleSubtermPositions.pop()))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
	  goto try_another_position;
	};
      _currentCandidate = NextCandidate(_currentCompiledConstraint);
      if (_currentCandidate) 
	{
	  _sidePremise = _currentCandidate->TheClause();
	  if (_noRulesAsSidePremises && _sidePremise->IsRule())
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.discardedForOtherReasons);
	    goto try_another_candidate;
	    };
	  if (demodulateSelectedLits())
	    goto try_another_candidate;
	  if (subsumeSidePremise())
	    goto try_another_candidate;
	  return true;
	};
      goto try_another_position;
    };
}; 
VK::UnifIndex::IndexedObject* 
VK::ForwardSuperposition::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr) 
{
  CALL("NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr)");
  UnifIndex::IndexedObject* result;
  switch (_paramodulationMode)
    {
    case Init:
      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaTree(_currentParamodulationTargetTerm->symbol().Func()))))
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
	  _mainSelectedLitDemodulationAttempted = false;
	  _paramodulationMode = ParaFromComplex;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidate();
	  ASSERT_IN(result,"A10");
	  return result;
	};
      goto try_para_from_var; 
    case ParaFromComplex:
      result = _retrievalOfUnifiable.NextCandidate();
      if (result) return result;
    try_next_leaf:
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };  
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidate();
	  ASSERT_IN(result,"A20");
	  return result;
	};
      goto try_para_from_var;
    case ParaFromVar:
      result = _retrievalOfUnifiable.NextCandidate();
      if (result) return result;
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
	  _mainSelectedLitDemodulationAttempted = false;
	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidate();
	  ASSERT_IN(result,"A30"); 
	  return result;
	};
      return 0;
    };
 try_para_from_var:
  _paramodulationMode = ParaFromVar;
  if (!_retrievalOfUnifiable.CompilePseudoLiteral(_pseudoHeader,
						  _currentParamodulationTargetTerm))
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedForOtherReasons);
      return 0;
    };
  if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaFromVarTree())))
    {
      _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
      _mainSelectedLitDemodulationAttempted = false;
      _retrievalOfUnifiable.ResetCandidatesFromLeaf();
      result = _retrievalOfUnifiable.NextCandidate();
      ASSERT_IN(result,"A20");
      return result;
    };
  return 0; 
}; 
bool VK::ForwardSuperposition::CollectResult()
{
  CALL("CollectResult()");
  if (!_mainPremise->isPersistent()) return false;
  if (!_sidePremise->isPersistent()) return false;
  const SimplificationOrdering::StoredConstraint* constr = 
    _currentCandidate->orderingConstraint();
  if (constr)
    {
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantDiscarded);
	  return false;
	};
    };
  if (_mainPremiseResidualSize
      + _sidePremise->practicallyMinResidualSize() 
      > _clauseSizeLimit)
    {
      ++(_statistics->inferences.deduction.forwardSuperposition.discardedBySize);
      return false;
    }; 
  _clauseCollector->openNewClause(ClauseBackground::ForwSup,
				  _mainPremise,
				  _sidePremise);
  _clauseCollector->announceNewClause();
  LiteralProperties::SLModifier rewLitSLMod = 
    LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier mainPremSLMod = 
    LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier sidePremSLMod = 
    LiteralProperties::DefaultSLModifier;
  if (_mainSelectedLitProperties[_nextMainSelectedLitNum].deploymentMode() !=
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_mainSelectedLitProperties[_nextMainSelectedLitNum].deploymentMode() ==
	  LiteralProperties::RuleHead)
	{
	  mainPremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_mainSelectedLitProperties[_nextMainSelectedLitNum].deploymentMode() ==
		 LiteralProperties::Subquery);
	  mainPremSLMod = LiteralProperties::PostponeSelection;
	};
    };
  ulong eqLitNum = _currentCandidate->LitNum();
  LiteralProperties eqProp = 
    _sidePremise->LitList()->nth(eqLitNum)->properties();
  if (eqProp.deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (eqProp.deploymentMode() == LiteralProperties::RuleHead)
	{
	  sidePremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(eqProp.deploymentMode() == LiteralProperties::Subquery);
	  sidePremSLMod = LiteralProperties::PostponeSelection;
	};
    };
  return 
    collectMainPremiseLiterals(mainPremSLMod) && 
    collectRewrittenLiteral(rewLitSLMod) &&
    collectSidePremiseLiterals(sidePremSLMod) &&
    _clauseCollector->finaliseClause();
}; 
bool VK::ForwardSuperposition::CollectEqResResult()
{
  CALL("CollectEqResResult()");
  if (!_mainPremise->isPersistent()) return false;
  _clauseCollector->openNewClause(ClauseBackground::EqRes,_mainPremise);
  _clauseCollector->announceNewClause();
  return 
    collectMainPremiseLiterals(LiteralProperties::DefaultSLModifier) &&
    _clauseCollector->finaliseClause();
}; 
bool VK::ForwardSuperposition::collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod)");
  LiteralProperties prop;
  prop.inheritFrom(_currentMainLitProperties,slmod);
  if (_currentMainLitProperties.isPreorientedEq())
    {
      UnifTerm* lhs;
      if (_currentMainLitProperties.isReverseOrientedEq())
	{
	  lhs = _currentMainLiteral->Arg2();
	}
      else
	{
	  lhs = _currentMainLiteral->Arg1();
	};
      if (UnifTerm::containsSubterm(lhs,_currentParamodulationTargetTerm))
	prop.setDefaultEqualityOrientationMode();
    }
  else
    if (!_currentMainLitProperties.isOrientableEq())
      prop.setDefaultEqualityOrientationMode();
  return 
    _clauseCollector->pushLiteralWithReplacement(_currentMainLiteral,
						 _currentParamodulationTargetTerm,
						 _currentCandidate->ReplacingTerm(),
						 _sidePremiseVarIndex,
						 prop);
}; 
bool VK::ForwardSuperposition::collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod)");
  unsigned ln = 0;
  unsigned used_lit_num = _currentCandidate->LitNum();
  LiteralProperties prop;
  for (LiteralList::Iterator iter(_sidePremise->LitList());
       iter.notEnd();
       iter.next())
    {
      if (ln != used_lit_num) 
	{
	  prop.inheritFrom(iter.currentElement()->properties(),slmod);
	  if (!_clauseCollector->pushLiteralWithUnifier(_sidePremiseVarIndex,
							iter.currentLiteral(),
							prop)) 
	    return false;
	};
      ln++;
    }; 
  return true;
}; 
bool VK::ForwardSuperposition::collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod)
{
  CALL("collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod)");
  LiteralProperties prop;
  for (ulong ln = _numOfSelectedMainLits; ln; ln--)
    if ((ln - 1) != _nextMainSelectedLitNum)
      {
	prop.inheritFrom(_mainSelectedLitProperties[ln-1],slmod);
	if (!_clauseCollector->pushLiteralWithUnifier(_mainSelectedLiterals.nth(ln-1),
						      prop)) 
	  return false;     
      };
  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    {
      prop.inheritFrom(iter.currentElement()->properties(),slmod);
      if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						    iter.currentLiteral(),
						    prop)) 
	return false; 
    }; 
  return true; 
}; 
bool VK::ForwardSuperposition::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (collectMainPremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _mainPremise->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesSubsumed);
 discarded:
  _mainPremise->unblockForSubsumption();
  return true;
}; 
bool VK::ForwardSuperposition::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; 
bool VK::ForwardSuperposition::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  ASSERT(_sidePremise == _currentCandidate->TheClause());
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
  if (collectSidePremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.redundantDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _sidePremise->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.forwardSuperposition.sidePremisesSubsumed);
 discarded:
  _sidePremise->unblockForSubsumption();
  return true;
}; 
bool VK::ForwardSuperposition::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");  
  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;  
}; 
bool VK::ForwardSuperposition::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_mainPremise->paramodulationsIntoProtected() &&
      _sidePremise->Unit())
    return false; 
  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesDemodulated);
	      return true;
	    };
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
	  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() &&
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {	  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++(_statistics->inferences.deduction.forwardSuperposition.mainPremisesDemodulated);
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	    }
	  else 
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (tautology)
		++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	      return tautology;
	    };
	};
    };
  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
      if (collectSidePremiseSubtermsAvailableForParamodulationInto() &&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{	    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++(_statistics->inferences.deduction.forwardSuperposition.sidePremisesDemodulated);
	      return true;
	    }
	  else
	    return false;
	}
      else 
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  bool tautology = 
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (tautology)
	    ++(_statistics->inferences.deduction.forwardSuperposition.redundantBranchesDiscarded);
	  return tautology;
	};
    }
  else
    return false;
}; 
bool 
VK::ForwardSuperposition::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");  
  LiteralProperties prop;
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      Term* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_mainPremiseVarIndex,
								      Term::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; 
bool 
VK::ForwardSuperposition::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      Term* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_sidePremiseVarIndex,
								      Term::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; 
