/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "LPO.hpp"
#include "Unifier.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "DestructionMode.hpp"
#include "Unification.hpp"
#include "InstanceWithUnifier.hpp"
#define LPO_SMALL_ARITY
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO"
#endif
#include "debugMacros.hpp"
using namespace VK;
using namespace BK;
LPO::LPO() : 
  _specialisedComparisonPool(0),
  _constraintCheckOnUnifierPool(0),
  _constraintCheckOnForwardMatchingSubstPool(0),
  _compiledConstraintOnUnifiersPool(0),
  _compiledConstraintForInstanceRetrievalPool(0),
  _numberOfCompiledConstraintsForInstanceRetrieval(1UL),
  _numberOfConstraintChecksForInstanceRetrieval(1UL),
  _numberOfCompiledConstraintsOnUnifiers(1UL),
  _numberOfConstraintChecksOnUnifiers(1UL)
{    
  setType(SimplificationOrdering::LPO);
};
LPO::~LPO()
{
  CALL("destructor ~LPO()");
  if (DestructionMode::isThorough()) 
    {
      while (_specialisedComparisonPool)
	{
	  SpecialisedComparison* tmp = _specialisedComparisonPool;
	  _specialisedComparisonPool = _specialisedComparisonPool->getNextInPool();
	  delete tmp;
	};
      while (_constraintCheckOnUnifierPool)
	{
	  ConstraintCheckOnUnifier* tmp = _constraintCheckOnUnifierPool;
	  _constraintCheckOnUnifierPool = _constraintCheckOnUnifierPool->getNextInPool();
	  delete tmp;
	};
      while (_constraintCheckOnForwardMatchingSubstPool)
	{
	  ConstraintCheckOnForwardMatchingSubst* tmp = _constraintCheckOnForwardMatchingSubstPool;
	  _constraintCheckOnForwardMatchingSubstPool = _constraintCheckOnForwardMatchingSubstPool->getNextInPool();
	  delete tmp;
	};
      while (_compiledConstraintOnUnifiersPool)
	{
	  CompiledConstraintForUnifiers* tmp = _compiledConstraintOnUnifiersPool;
	  _compiledConstraintOnUnifiersPool = _compiledConstraintOnUnifiersPool->getNextInPool();      
	  delete tmp;
	};
      while (_compiledConstraintForInstanceRetrievalPool)
	{
	  CompiledConstraintForInstanceRetrieval* tmp = _compiledConstraintForInstanceRetrievalPool;
	  _compiledConstraintForInstanceRetrievalPool = _compiledConstraintForInstanceRetrievalPool->getNextInPool();  
	  delete tmp;
	};  
    };
}; 
void* LPO::operator new(size_t)
{
  return GlobAlloc::allocate(sizeof(LPO));
};
void LPO::operator delete(void* obj)
{
  GlobAlloc::deallocate(obj,sizeof(LPO));
}; 
Comparison LPO::compare(const Flatterm* term1,
			const WeightPolynomial& weight1,
			const Flatterm* term2,
			const WeightPolynomial& weight2)
{
  CALL("compare(..)#1");
  if (term1->isVariable())
    {
      if (term2->isVariable())
	{
	  if (term1->symbolRef() == term2->symbolRef())
	    return Equal;
	  return Incomparable;
	}
      else 
	{
	  if (weight2.containsVariable(term1->Var()))
	    return Less;
	  return Incomparable;
	};
    }
  else
    if (term2->isVariable())
      {
	if (weight1.containsVariable(term2->Var()))
	    return Greater;
	  return Incomparable;
      }
    else 
      {
	if (weight1.containsVariableWhichIsNotIn(weight2))
	  {
	    if (weight2.containsVariableWhichIsNotIn(weight1))
	      {
		return Incomparable;
	      }
	    else
	      {
		if (complexGreater(term1,term2))
		  {
		    ASSERT(!complexGreater(term2,term1));
		    ASSERT(complexGreaterThanProperSubterms(term1,term2));
		    return Greater;
		  };
		return Incomparable; 
	      };	      
	  }
	else
	  if (weight2.containsVariableWhichIsNotIn(weight1))
	    {
	      if (complexGreater(term2,term1))
		{
		  ASSERT(!complexGreater(term1,term2));
		  ASSERT(complexGreaterThanProperSubterms(term2,term1));
		  return Less;
		};
	      return Incomparable; 
	    }
	  else
	    {
	      Comparison res = compareComplex(term1,term2);
	      ASSERT(res == ::inverse(compareComplex(term2,term1)));
	      ASSERT(res != BK::Greater || 
		     complexGreaterThanProperSubterms(term1,term2));
	      ASSERT(res != BK::Less || 
		     complexGreaterThanProperSubterms(term2,term1));
	      return res;
	    }; 
      };
  ICP("END");
  return Incomparable;
}; 
Comparison LPO::compareAtoms(const Flatterm* lit1,
			     const WeightPolynomial& weight1,
			     const Flatterm* lit2,
			     const WeightPolynomial& weight2)
{
  CALL("compareAtoms(..)");
  if (weight1.containsVariableWhichIsNotIn(weight2))
    {
      if (weight2.containsVariableWhichIsNotIn(weight1))
	{
	  return Incomparable;
	}
      else
	{
	  if (complexGreater(lit1,lit2))
	    {
	      ASSERT(!complexGreater(lit2,lit1));
	      ASSERT(complexGreaterThanProperSubterms(lit1,lit2));
	      return Greater;
	    };
	  return Incomparable; 
	};	      
    }
  else
    if (weight2.containsVariableWhichIsNotIn(weight1))
      {
	if (complexGreater(lit2,lit1))
	  {
	    ASSERT(!complexGreater(lit1,lit2));
	    ASSERT(complexGreaterThanProperSubterms(lit2,lit1));
	    return Less;
	  };
	return Incomparable; 
      }
    else
      {
	Comparison res = compareComplex(lit1,lit2);
	ASSERT(res == ::inverse(compareComplex(lit2,lit1)));
	ASSERT(res != BK::Greater || 
	       complexGreaterThanProperSubterms(lit1,lit2));
	ASSERT(res != BK::Less || 
	       complexGreaterThanProperSubterms(lit2,lit1));
	return res;
      }; 
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::compare(const Flatterm* term1,
			const Flatterm* term2)
{
  CALL("compare(const Flatterm* term1,const Flatterm* term2)");
  if (term1->isVariable())
    {
      if (term2->isVariable())
	{
	  if (term1->symbolRef() == term2->symbolRef())
	    return Equal;
	  return Incomparable;
	}
      else 
	{
	  if (term2->containsSymbol(term1->symbolRef()))
	    return Less;
	  return Incomparable;
	};
    }
  else
    if (term2->isVariable())
      {
	if (term1->containsSymbol(term2->symbolRef())) 
	  return Greater;
	return Incomparable;
      }
    else 
      {
	Comparison res = compareComplex(term1,term2);
	ASSERT(res == ::inverse(compareComplex(term2,term1)));
	ASSERT(res != BK::Greater || 
	       complexGreaterThanProperSubterms(term1,term2));
	ASSERT(res != BK::Less || 
	       complexGreaterThanProperSubterms(term2,term1));
	return res;
      };
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::compareComplex(const Flatterm* complexTerm1,
			       const Flatterm* complexTerm2)
{
  CALL("compareComplex(const Flatterm* complexTerm1,..)");
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compare(complexTerm1->next(),
			   complexTerm2->next());
	  };
#endif
      return cLMA(complexTerm1,
		  complexTerm2,
		  complexTerm1->next(),
		  complexTerm1->after(),
		  complexTerm2->next(),
		  complexTerm2->after());
    }
  else 
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConst(complexTerm2,
					 complexTerm1->symbolRef().symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConst(complexTerm1,
			     complexTerm2->symbolRef().symbolPrecedence());
	};
#endif
      if (complexTerm1->symbolRef().hasGreaterPrecedenceThan(complexTerm2->symbolRef()))
	{
	  return 
	    cMA(complexTerm1,complexTerm2->next(),complexTerm2->after());
	}
      else 
	{
	  return 
	    BK::inverse(cMA(complexTerm2,
			    complexTerm1->next(),
			    complexTerm1->after()));
	};
    };
  ICP("END");
  return Incomparable;
}; 
inline
BK::Comparison LPO::cMA(const Flatterm* complexTerm1,
			const Flatterm* args2,
			const Flatterm* endOfArgs2)
{
  CALL("cMA(const Flatterm* complexTerm1,..)");
  ASSERT(args2 <= endOfArgs2);
  while (args2 != endOfArgs2)
    {
      Comparison cmp;
      if (args2->isVariable())
	{
	  if (complexTerm1->containsSymbol(args2->symbolRef()))
	    {
	      cmp = Greater;
	    }
	  else
	    cmp = Incomparable;
	}
      else
	cmp = compareComplex(complexTerm1,args2);
      switch (cmp) 
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->after(),endOfArgs2,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->after();
    };
  return Greater;
}; 
inline
BK::Comparison LPO::cLMA(const Flatterm* complexTerm1,
			 const Flatterm* complexTerm2,
			 const Flatterm* args1,
			 const Flatterm* endOfArgs1,
			 const Flatterm* args2,
			 const Flatterm* endOfArgs2)
{
  CALL("cLMA(const Flatterm* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  ASSERT(complexTerm1 < args1 && args1 <= endOfArgs1);
  ASSERT(complexTerm2 < args2 && args2 <= endOfArgs2);
  while (args1 != endOfArgs1)
    {
      switch (compare(args1,args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->after(),endOfArgs1));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->after(),endOfArgs2);
	case Incomparable:
	  if (alpha(args1->after(),endOfArgs1,complexTerm2))
	    return Greater;
	  if (alpha(args2->after(),endOfArgs2,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->after();
      args2 = args2->after();
    };
  ASSERT(args2 == endOfArgs2);
  return Equal;
}; 
inline
bool LPO::alpha(const Flatterm* args1,
		const Flatterm* endOfArgs1,
		const Flatterm* complexTerm2)
{
  CALL("alpha(const Flatterm* args1,..)");
  ASSERT(args1 <= endOfArgs1);
  while (args1 != endOfArgs1)
    {
      if (!args1->isVariable())
	{
	  if (complexGreaterOrEqual(args1,complexTerm2)) return true;
	};      
      args1 = args1->after();
    };
  return false;
}; 
inline
bool LPO::greater(const Flatterm* term1,const Flatterm* term2)
{
  CALL("greater(const Flatterm* term1,const Flatterm* term2)");
  if (term1->isVariable())
    {
      return false;
    }
  else
    if (term2->isVariable())
      {
	return
	  term1->containsSymbol(term2->symbolRef());
      }
    else
      return complexGreater(term1,term2);
}; 
inline
bool LPO::greaterOrEqual(const Flatterm* term1,const Flatterm* term2)
{
  CALL("greaterOrEqual(const Flatterm* term1,const Flatterm* term2)");
  if (term1->isVariable())
    {
      return term1->symbolRef() == term2->symbolRef();
    }
  else
    if (term2->isVariable())
      {
	return
	  term1->containsSymbol(term2->symbolRef());
      }
    else
      return complexGreaterOrEqual(term1,term2);
}; 
inline
bool LPO::complexGreater(const Flatterm* complexTerm1,
			 const Flatterm* complexTerm2)
{
  CALL("complexGreater(const Flatterm* complexTerm1,..)");
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greater(complexTerm1->next(),
			   complexTerm2->next());
	  };
#endif
      const Flatterm* args1 = complexTerm1->next();
      const Flatterm* endOfArgs1 = complexTerm1->after();
      const Flatterm* args2 = complexTerm2->next();
      const Flatterm* endOfArgs2 = complexTerm2->after();
      while (args1 != endOfArgs1)
	{
	  switch (compare(args1,args2))
	    {
	    case Less:
	      args1 = args1->after();
	      while (args1 != endOfArgs1)
		{
		  if (args1->isVariable())
		    {
		      if (!complexTerm2->containsSymbol(args1->symbolRef()))
			return 
			  alpha(args1->after(),endOfArgs1,complexTerm2);
		    }
		  else
		    {
		      switch (compareComplex(complexTerm2,args1)) 
			{
			case Less:
			case Equal:
			  return true;
			case Greater:
			  break; 
			case Incomparable:
			  return 
			    alpha(args1->after(),endOfArgs1,complexTerm2);
			};
		    };
		  args1 = args1->after();
		};
	      return false;
	    case Equal:
	      break;
	    case Greater:
	      args2 = args2->after();
	      while (args2 != endOfArgs2)
		{
		  if (args2->isVariable())
		    {
		      if (!complexTerm1->containsSymbol(args2->symbolRef()))
			return false;
		    }
		  else
		    {
		      if (!complexGreater(complexTerm1,args2))
			return false;
		    };
		  args2 = args2->after();
		};
	      return true;
	    case Incomparable:
	      return 
		alpha(args1->after(),endOfArgs1,complexTerm2);
	    };
	  args1 = args1->after();
	  args2 = args2->after();
	}; 
      return false; 
    }
  else 
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   complexTerm1->symbolRef().symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       complexTerm2->symbolRef().symbolPrecedence());
	};
#endif
      if (complexTerm1->symbolRef().hasGreaterPrecedenceThan(complexTerm2->symbolRef()))
	{
	  const Flatterm* args2 = complexTerm2->next();
	  const Flatterm* endOfArgs2 = complexTerm2->after();
	  while (args2 != endOfArgs2)
	    {
	      if (args2->isVariable())
		{
		  if (!complexTerm1->containsSymbol(args2->symbolRef())) 
		    return false;
		}
	      else
		{
		  if (!complexGreater(complexTerm1,args2))
		    return false;
		};
	      args2 = args2->after();
	    }; 
	  return true;
	}
      else 
	{
	  const Flatterm* args1 = complexTerm1->next();
	  const Flatterm* endOfArgs1 = complexTerm1->after();
	  while (args1 != endOfArgs1)
	    {
	      if (args1->isVariable())
		{
		  if (!complexTerm2->containsSymbol(args1->symbolRef()))
		    return alpha(args1->after(),endOfArgs1,complexTerm2);
		}
	      else
		{
		  switch (compareComplex(complexTerm2,args1)) 
		    {
		    case Less:
		    case Equal:
		      return true;
		    case Greater:
		      break; 
		    case Incomparable:
		      return alpha(args1->after(),endOfArgs1,complexTerm2);
		    };
		};
	      args1 = args1->after();
	    };
	  return false;
	}; 
    };
  ICP("END");
  return false;
}; 
bool LPO::complexGreaterOrEqual(const Flatterm* complexTerm1,
				const Flatterm* complexTerm2)
{
  CALL("complexGreaterOrEqual(const Flatterm* complexTerm1,..)");
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return true; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greaterOrEqual(complexTerm1->next(),
				  complexTerm2->next());
	  };
#endif
      const Flatterm* args1 = complexTerm1->next();
      const Flatterm* endOfArgs1 = complexTerm1->after();
      const Flatterm* args2 = complexTerm2->next();
      const Flatterm* endOfArgs2 = complexTerm2->after();
      while (args1 != endOfArgs1)
	{
	  switch (compare(args1,args2))
	    {
	    case Less:
	      args1 = args1->after();
	      while (args1 != endOfArgs1)
		{
		  if (args1->isVariable())
		    {
		      if (!complexTerm2->containsSymbol(args1->symbolRef()))
			return 
			  alpha(args1->after(),endOfArgs1,complexTerm2);
		    }
		  else
		    {
		      switch (compareComplex(complexTerm2,args1)) 
			{
			case Less:
			case Equal:
			  return true;
			case Greater:
			  break; 
			case Incomparable:
			  return 
			    alpha(args1->after(),endOfArgs1,complexTerm2);
			};
		    };
		  args1 = args1->after();
		};
	      return false;
	    case Equal:
	      break;
	    case Greater:
	      args2 = args2->after();
	      while (args2 != endOfArgs2)
		{
		  if (args2->isVariable())
		    {
		      if (!complexTerm1->containsSymbol(args2->symbolRef()))
			return false;
		    }
		  else
		    {
		      if (!complexGreater(complexTerm1,args2))
			return false;
		    };
		  args2 = args2->after();
		};
	      return true;
	    case Incomparable:
	      return 
		alpha(args1->after(),endOfArgs1,complexTerm2);
	    };
	  args1 = args1->after();
	  args2 = args2->after();
	}; 
      ASSERT(args2 == endOfArgs2);
      return true; 
    }
  else 
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   complexTerm1->symbolRef().symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       complexTerm2->symbolRef().symbolPrecedence());
	};
#endif
      if (complexTerm1->symbolRef().hasGreaterPrecedenceThan(complexTerm2->symbolRef()))
	{
	  const Flatterm* args2 = complexTerm2->next();
	  const Flatterm* endOfArgs2 = complexTerm2->after();
	  while (args2 != endOfArgs2)
	    {
	      if (args2->isVariable())
		{
		  if (!complexTerm1->containsSymbol(args2->symbolRef())) 
		    return false;
		}
	      else
		{
		  if (!complexGreater(complexTerm1,args2))
		    return false;
		};
	      args2 = args2->after();
	    }; 
	  return true;
	}
      else 
	{
	  const Flatterm* args1 = complexTerm1->next();
	  const Flatterm* endOfArgs1 = complexTerm1->after();
	  while (args1 != endOfArgs1)
	    {
	      if (args1->isVariable())
		{
		  if (!complexTerm2->containsSymbol(args1->symbolRef()))
		    return alpha(args1->after(),endOfArgs1,complexTerm2);
		}
	      else
		{
		  switch (compareComplex(complexTerm2,args1)) 
		    {
		    case Less:
		    case Equal:
		      return true;
		    case Greater:
		      break; 
		    case Incomparable:
		      return alpha(args1->after(),endOfArgs1,complexTerm2);
		    };
		};
	      args1 = args1->after();
	    };
	  return false;
	}; 
    };  
  ICP("END");
  return false;
}; 
bool LPO::complexContainsVar(const Term* complexTerm,
			     const ForwardMatchingSubstitutionCache* subst,
			     const Term& var)
{
  CALL("complexContainsVar(..)");
  ForwardMatchingSubstitutionCache::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (iter.symbol() == var)
	return true;
    };
  return false;
}; 
bool LPO::complexGreaterThanProperSubterms(const Flatterm* complexTerm1,
					   const Flatterm* complexTerm2)
{
  CALL("complexGreaterThanProperSubterms(..)");
  Flatterm::ConstIterator iter(complexTerm2);
  iter.NextSym();
  while (iter.NotEndOfTerm())
    {
      if (iter.CurrPos()->isVariable())
	{
	  if (!complexTerm1->containsSymbol(iter.CurrPos()->symbolRef()))
	    return false;
	}
      else
	if (!complexGreater(complexTerm1,iter.CurrPos())) 
	  return false;
      iter.NextSym();
    };
  return true;
}; 
inline
bool 
LPO::groundWithAllPrecBelow(const Flatterm* complexTerm,
			    long prec)
{
  CALL("groundWithAllPrecBelow(const Flatterm* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolRef().symbolPrecedence() != prec);
  if (complexTerm->symbolRef().symbolPrecedence() > prec)
    return false;
  Flatterm::ConstIterator iter(complexTerm);
  iter.NextSym();
  while (iter.NotEndOfTerm())
    {
      if (iter.CurrPos()->isVariable() ||
	  iter.CurrPos()->symbolRef().symbolPrecedence() >= prec)
	return false;
      iter.NextSym();      
    };
  return true;
}; 
inline
bool 
LPO::containsEqOrGrPrec(const Flatterm* complexTerm,
			long prec)
{
  CALL("containsEqOrGrPrec(const Flatterm* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolRef().symbolPrecedence() != prec);
  if (complexTerm->symbolRef().symbolPrecedence() > prec)
    return true;
  Flatterm::ConstIterator iter(complexTerm);
  iter.NextSym();
  while (iter.NotEndOfTerm())
    {
      if (!iter.CurrPos()->isVariable() &&
	  iter.CurrPos()->symbolRef().symbolPrecedence() >= prec)
	return true;
      iter.NextSym();      
    };
  return false;
}; 
inline
Comparison
LPO::compareWithConst(const Flatterm* complexTerm,
		      long constPrec)
{
  CALL("compareWithConst(const Flatterm* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolRef().symbolPrecedence() != constPrec);
  if (complexTerm->symbolRef().symbolPrecedence() > constPrec)
    return Greater;
  Flatterm::ConstIterator iter(complexTerm);
  iter.NextSym();
  bool ground = true;
  while (iter.NotEndOfTerm())
    {
      if (iter.CurrPos()->isVariable())
	{
	  ground = false;
	}
      else
	if (iter.CurrPos()->symbolRef().symbolPrecedence() >= constPrec)
	  return Greater;
      iter.NextSym();      
    };
  return (ground)? Less : Incomparable;
}; 
Comparison LPO::compareOnCurrentUnifier(UnifTerm* term1,
					UnifTerm* term2)
{
  CALL("compareOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)");
  UnifTerm::CorrectBracket(term1);
  UnifTerm::CorrectBracket(term2);
  bool freeVar1 = false;
  bool freeVar2 = false;
  void* instance1 = 0;
  void* instance2 = 0;
  if (term1->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term1->symbol().var())));
      instance1 = term1->symbol().var()->Unref(freeVar1);
    }
  else
    instance1 = term1;
  if (term2->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term2->symbol().var())));
      instance2 = term2->symbol().var()->Unref(freeVar2);
    }
  else
    instance2 = term2;
  if (freeVar1)
    {
      if (freeVar2)
	{
	  if (instance1 == instance2) return Equal;
	}
      else
	{
	  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance2));
	  if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance2),
					      static_cast<Unifier::Variable*>(instance1)))
	    return Less;
	};
      return Incomparable;
    }
  else
    if (freeVar2)
      {
	UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
	if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance1),
					    static_cast<Unifier::Variable*>(instance2)))
	    return Greater;
	return Incomparable;
      }
    else 
      {
	term1 = static_cast<UnifTerm*>(instance1);
	term2 = static_cast<UnifTerm*>(instance2);
	UnifTerm::CorrectBracket(term1);
	UnifTerm::CorrectBracket(term2);
	if (term1->symbol().Func().functor() == term2->symbol().Func().functor())
	  {
#ifdef LPO_SMALL_ARITY
	    if (term1->symbol().Func().arity() == 0UL)
	      {
		return Equal;
	      }
	    else
	      if (term1->symbol().Func().arity() == 1UL)
		{
		  Comparison res = compareOnCurrentUnifier(term1 + 1,term2 + 1);
		  ASSERT(res != Greater ||
			 complexGreaterThanProperSubtermsOnCurrentUnifier(term1,
									  term2));
		  ASSERT(res != Less ||
			 complexGreaterThanProperSubtermsOnCurrentUnifier(term2,
									  term1));
		  return res;
		};
#endif
	    Comparison res = cLMA(term1,
				  term2,
				  term1 + 1,
				  term1->Bracket(),
				  term2 + 1,
				  term2->Bracket());
	    ASSERT(res != Greater ||
		   complexGreaterThanProperSubtermsOnCurrentUnifier(term1,
								    term2));
	    ASSERT(res != Less ||
		   complexGreaterThanProperSubtermsOnCurrentUnifier(term2,
								    term1));
	    return res;
	  }
	else 
	  {
#ifdef LPO_SMALL_ARITY
	    if (term1->symbol().Func().arity() == 0UL)
	      {
		Comparison res = 
		  BK::inverse(compareWithConstOnCurrentUnifier(term2,
							       term1->symbol().Func().symbolPrecedence()));
		ASSERT(res != Greater ||
		       complexGreaterThanProperSubtermsOnCurrentUnifier(term1,
									term2));
		ASSERT(res != Less ||
		       complexGreaterThanProperSubtermsOnCurrentUnifier(term2,
									term1));
		return res;
	      };
	    if (term2->symbol().Func().arity() == 0UL)
	      {
		Comparison res = 
		  compareWithConstOnCurrentUnifier(term1,
						   term2->symbol().Func().symbolPrecedence());
		ASSERT(res != Greater ||
		       complexGreaterThanProperSubtermsOnCurrentUnifier(term1,
									term2));
		ASSERT(res != Less ||
		       complexGreaterThanProperSubtermsOnCurrentUnifier(term2,
									term1));
		return res;
	      };
#endif
	    Comparison res;
	    if (term1->symbol().Func().hasGreaterPrecedenceThan(term2->symbol().Func()))
	      {
		res =
		  cMA(term1,term2 + 1,term2->Bracket());
	      }
	    else
	      {
		res =
		  BK::inverse(cMA(term2,term1 + 1,term1->Bracket()));
	      };
	    ASSERT(res != Greater ||
		   complexGreaterThanProperSubtermsOnCurrentUnifier(term1,
								    term2));
	    ASSERT(res != Less ||
		   complexGreaterThanProperSubtermsOnCurrentUnifier(term2,
								    term1));
	    return res;
	  };
      };
  ICP("END");
  return Incomparable;
}; 
bool 
LPO::greaterOrEqualOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)
{
  CALL("greaterOrEqualOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)");
  UnifTerm::CorrectBracket(term1);
  UnifTerm::CorrectBracket(term2);
  bool freeVar1 = false;
  bool freeVar2 = false;
  void* instance1 = 0;
  void* instance2 = 0;
  if (term1->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term1->symbol().var())));
      instance1 = term1->symbol().var()->Unref(freeVar1);
    }
  else
    instance1 = term1;
  if (term2->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term2->symbol().var())));
      instance2 = term2->symbol().var()->Unref(freeVar2);
    }
  else
    instance2 = term2;
  if (freeVar1)
    {
      return instance1 == instance2;
    }
  else
    if (freeVar2)
      {
	UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
	return
	  termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance1),
					  static_cast<Unifier::Variable*>(instance2));
      }
    else 
      {
	return 
	  complexGreaterOrEqualOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
						static_cast<UnifTerm*>(instance2));
      };
  ICP("END");
  return false;
}; 
bool 
LPO::greaterOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)
{
  CALL("greaterOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)");
  bool freeVar1 = false;
  bool freeVar2 = false;
  void* instance1 = 0;
  void* instance2 = 0;
  if (term1->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term1->symbol().var())));
      instance1 = term1->symbol().var()->Unref(freeVar1);
      if (freeVar1) return false;
    }
  else
    instance1 = term1;
  if (term2->symbol().isVariable())
    {
      ASSERT(Unifier::current()->isVariable(static_cast<void*>(term2->symbol().var())));
      instance2 = term2->symbol().var()->Unref(freeVar2);
    }
  else
    instance2 = term2;
  ASSERT(!freeVar1);
  if (freeVar2)
    {
      UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
      return
	termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance1),
					static_cast<Unifier::Variable*>(instance2));
    }
  else 
    {
      return 
	complexGreaterOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
				       static_cast<UnifTerm*>(instance2));
    };
  ICP("END");
  return false;
}; 
bool 
LPO::complexGreaterThanProperSubtermsOnCurrentUnifier(UnifTerm* complexTerm1,
						      UnifTerm* complexTerm2)
{
  CALL("complexGreaterThanProperSubtermsOnCurrentUnifier(..)#1");
  UnifTerm::CorrectBracket(complexTerm1);
  UnifTerm::CorrectBracket(complexTerm2);
  UnifTerm::Iter iter(complexTerm2);
  iter.SkipSym(); 
  while (iter.NotEmpty())
    {
      if (!greaterOnCurrentUnifier(complexTerm1,
				   iter.CurrSubterm()))
	return false;
      iter.SkipSym();
    };
  return true;
}; 
inline
bool 
LPO::complexGreaterOrEqualOnCurrentUnifier(UnifTerm* complexTerm1,
					   UnifTerm* complexTerm2)
{
  CALL("complexGreaterOrEqualOnCurrentUnifier(UnifTerm* complexTerm1,..)");
  UnifTerm::CorrectBracket(complexTerm1);
  UnifTerm::CorrectBracket(complexTerm2);
  if (complexTerm1->symbol().Func().functor() == complexTerm2->symbol().Func().functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->symbol().Func().arity() == 0UL)
	{
	  return true; 
	}
      else
	if (complexTerm1->symbol().Func().arity() == 1UL)
	  {
	    return
	      greaterOrEqualOnCurrentUnifier(complexTerm1 + 1,
					     complexTerm2 + 1);
	  };
#endif
      return lexMAOrEqual(complexTerm1,
			  complexTerm2,
			  complexTerm1 + 1,
			  complexTerm1->Bracket(),
			  complexTerm2 + 1,
			  complexTerm2->Bracket());
    }
  else 
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->symbol().Func().arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelowOnCurrentUnifier(complexTerm2,
						   complexTerm1->symbol().Func().symbolPrecedence());
	};
      if (complexTerm2->symbol().Func().arity() == 0UL)
	{
	  return
	    containsEqOrGrPrecOnCurrentUnifier(complexTerm1,
					       complexTerm2->symbol().Func().symbolPrecedence());
	};
#endif
      if (complexTerm1->symbol().Func().hasGreaterPrecedenceThan(complexTerm2->symbol().Func()))
	{
	  return majo(complexTerm1,complexTerm2 + 1,complexTerm2->Bracket());
	}
      else
	{
	  return alpha(complexTerm1 + 1,complexTerm1->Bracket(),complexTerm2);
	};
    };
}; 
inline
bool 
LPO::complexGreaterOnCurrentUnifier(UnifTerm* complexTerm1,
				    UnifTerm* complexTerm2)
{
  CALL("complexGreaterOnCurrentUnifier(UnifTerm* complexTerm1,..)");
  UnifTerm::CorrectBracket(complexTerm1);
  UnifTerm::CorrectBracket(complexTerm2);
  if (complexTerm1->symbol().Func().functor() == complexTerm2->symbol().Func().functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->symbol().Func().arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->symbol().Func().arity() == 1UL)
	  {
	    return
	      greaterOnCurrentUnifier(complexTerm1 + 1,
				      complexTerm2 + 1);
	  };
#endif
      return lexMA(complexTerm1,
		   complexTerm2,
		   complexTerm1 + 1,
		   complexTerm1->Bracket(),
		   complexTerm2 + 1,
		   complexTerm2->Bracket());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->symbol().Func().arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelowOnCurrentUnifier(complexTerm2,
						   complexTerm1->symbol().Func().symbolPrecedence());
	};
      if (complexTerm2->symbol().Func().arity() == 0UL)
	{
	  return
	    containsEqOrGrPrecOnCurrentUnifier(complexTerm1,
					       complexTerm2->symbol().Func().symbolPrecedence());
	};
#endif
      if (complexTerm1->symbol().Func().hasGreaterPrecedenceThan(complexTerm2->symbol().Func()))
	{
	  return majo(complexTerm1,complexTerm2 + 1,complexTerm2->Bracket());
	}
      else
	{
	  return alpha(complexTerm1 + 1,complexTerm1->Bracket(),complexTerm2);
	};
    };
}; 
Comparison 
LPO::compareOnCurrentUnifier(const Term& term1,
			     const Term& term2,
			     ulong variableIndex)
{
  CALL("compareOnCurrentUnifier(const Term& term1,const Term& term2,ulong variableIndex)");
  if (term1 == term2) return Equal;  
  if (term1.isVariable())
    {
      Unifier::Variable* var1 = 
	Unifier::current()->variable(term1.var(),variableIndex);
      bool freeVar1;
      void* instance1 = var1->Unref(freeVar1);
      if (term2.isVariable())
	{
	  Unifier::Variable* var2 = 
	    Unifier::current()->variable(term2.var(),variableIndex);
	  bool freeVar2;
	  void* instance2 = var2->Unref(freeVar2);
	  if (freeVar1)
	    {
	      if (freeVar2)
		{
		  if (instance1 == instance2) return Equal;
		}
	      else
		{
		  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance2));
		  if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance2),
						      static_cast<Unifier::Variable*>(instance1)))
		    return Less;
		};
	      return Incomparable;
	    }
	  else
	    if (freeVar2)
	      {
		UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
		if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance1),
						    static_cast<Unifier::Variable*>(instance2)))
		  return Greater;
		return Incomparable;
	      }
	    else
	      {
		UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
		UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance2));
		return 
		  compareOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
					  static_cast<UnifTerm*>(instance2));
	      };
	}
      else 
	{
	  Unifier::Variable* var1 = 
	    Unifier::current()->variable(term1.var(),variableIndex);
	  bool freeVar1;
	  void* instance1 = var1->Unref(freeVar1);
	  if (freeVar1)
	    {
	      if (termWithUnifierContainsVariable(term2.deref(),
						  variableIndex,
						  static_cast<Unifier::Variable*>(instance1)))
		return Less;
	      return Incomparable;
	    }
	  else
	    {
	      UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
	      Comparison res =
		compareComplexOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
					       term2.deref(),
					       variableIndex);
	      ASSERT(res != Greater ||
		     complexGreaterThanProperSubtermsOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
								      term2.deref(),
								      variableIndex));
	      ASSERT(res != Less ||
		     complexGreaterThanProperSubtermsOnCurrentUnifier(term2.deref(),
								      variableIndex,
								      static_cast<UnifTerm*>(instance1)));
	      return res;
	    };
	};
    }
  else 
    if (term2.isVariable())
      {
	Unifier::Variable* var2 = 
	  Unifier::current()->variable(term2.var(),variableIndex);
	bool freeVar2;
	void* instance2 = var2->Unref(freeVar2);
	if (freeVar2)
	  {
	    if (termWithUnifierContainsVariable(term1.deref(),
						variableIndex,
						static_cast<Unifier::Variable*>(instance2)))
	      {
		return Greater;
	      };
	    return Incomparable;
	  }
	else
	  {
	    UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance2));
	    Comparison res =
	      BK::inverse(compareComplexOnCurrentUnifier(static_cast<UnifTerm*>(instance2),
							 term1.deref(),
							 variableIndex));
	    ASSERT(res != Greater ||
                   complexGreaterThanProperSubtermsOnCurrentUnifier(term1.deref(),
                                                                    variableIndex,
                                                                    static_cast<UnifTerm*>(instance2)));
	    ASSERT(res != Less ||
                   complexGreaterThanProperSubtermsOnCurrentUnifier(static_cast<UnifTerm*>(instance2),
                                                                    term1.deref(),
                                                                    variableIndex));
	    return res;
	  };	
      }
    else 
      {
	Comparison res = 
	  compareComplexOnCurrentUnifier(term1.deref(),
					 term2.deref(),
					 variableIndex);
	ASSERT(res == ::inverse(compareComplexOnCurrentUnifier(term2.deref(),
							       term1.deref(),
							       variableIndex)));
	ASSERT(res != Greater ||
	       complexGreaterThanProperSubtermsOnCurrentUnifier(term1.deref(),
								term2.deref(),
								variableIndex));
	ASSERT(res != Less ||
	       complexGreaterThanProperSubtermsOnCurrentUnifier(term2.deref(),
								term1.deref(),
								variableIndex));
	return res;
      };
  ICP("END");
  return Incomparable;
}; 
inline
BK::Comparison LPO::cMA(UnifTerm* complexTerm1,
			UnifTerm* args2,
			UnifTerm* endOfArgs2)
{
  CALL("cMA(UnifTerm* complexTerm1,..)");
  ASSERT(args2 <= endOfArgs2);
  while (args2 != endOfArgs2)
    { 
      switch (compareOnCurrentUnifier(complexTerm1,args2))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Bracket(),endOfArgs2,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Bracket();
    };
  return Greater;
}; 
inline
BK::Comparison LPO::cMA(UnifTerm* complexTerm1,
			const Term* args2,
			ulong variableIndex)
{
  CALL("cMA(UnifTerm* complexTerm1,..)");
  while (!args2->isNotArg())
    { 
      switch (compareOnCurrentUnifier(complexTerm1,*args2,variableIndex))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),complexTerm1,variableIndex))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline
Comparison LPO::cMA(const Term* complexTerm1,
		    const Term* args2,
		    ulong variableIndex)
{
  CALL("cMA(const Term* complexTerm1,..)#1");
  Term term1(const_cast<Term*>(complexTerm1));
  while (!args2->isNotArg())
    {
      switch (compareOnCurrentUnifier(term1,*args2,variableIndex))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),complexTerm1,variableIndex))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline
Comparison LPO::cMA(const Term* complexTerm1,
		    UnifTerm* args2,
		    UnifTerm* endOfArgs2,
		    ulong variableIndex)
{
  CALL("cMA(const Term* complexTerm1,..)#2"); 
  ASSERT(args2 <= endOfArgs2);
  while (args2 != endOfArgs2)
    { 
      switch (BK::inverse(compareOnCurrentUnifier(args2,complexTerm1,variableIndex)))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Bracket(),endOfArgs2,complexTerm1,variableIndex))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Bracket();
    };
  return Greater;
}; 
inline
Comparison LPO::cLMA(UnifTerm* complexTerm1,
		     UnifTerm* complexTerm2,
		     UnifTerm* args1,
		     UnifTerm* endOfArgs1,
		     UnifTerm* args2,
		     UnifTerm* endOfArgs2)
{
  CALL("cLMA(UnifTerm* complexTerm1,..)");
  ASSERT(complexTerm1->symbol().Func().functor() == complexTerm2->symbol().Func().functor());
  ASSERT(args1 <= endOfArgs1 && args2 <= endOfArgs2);
  while (args1 != endOfArgs1)
    {
      switch (compareOnCurrentUnifier(args1,args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->Bracket(),endOfArgs1));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->Bracket(),endOfArgs2);
	case Incomparable:
	  if (alpha(args1->Bracket(),endOfArgs1,complexTerm2))
	    return Greater;
	  if (alpha(args2->Bracket(),endOfArgs2,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Bracket();
      args2 = args2->Bracket();
    };
  ASSERT(args2 == endOfArgs2);
  return Equal;
}; 
inline
Comparison LPO::cLMA(const Term* complexTerm1,
		     const Term* complexTerm2,
		     const Term* args1,
		     const Term* args2,
		     ulong variableIndex)
{
  CALL("cLMA(const Term* complexTerm1,..)");
  while (!args1->isNotArg())
    {
      switch (compareOnCurrentUnifier(*args1,*args2,variableIndex))
	{
	case Less:
	  return 
	    BK::inverse(cMA(complexTerm2,args1->Next(),variableIndex));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->Next(),variableIndex);
	case Incomparable:
	  if (alpha(args1->Next(),complexTerm2,variableIndex))
	    return Greater;
	  if (alpha(args2->Next(),complexTerm1,variableIndex))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return Equal;
}; 
inline
Comparison LPO::cLMA(UnifTerm* complexTerm1,
		     const Term* complexTerm2,
		     UnifTerm* args1,
		     UnifTerm* endOfArgs1,
		     const Term* args2,
		     ulong variableIndex)
{
  CALL("cLMA(UnifTerm* complexTerm1,const Term* complexTerm2,..");
  ASSERT(complexTerm1->symbol().Func().functor() == complexTerm2->functor());
  ASSERT(args1 <= endOfArgs1);
  while (args1 != endOfArgs1)
    {
      ASSERT(!args2->isNotArg());
      switch (compareOnCurrentUnifier(args1,*args2,variableIndex))
	{
	case Less:
	  return 
	    BK::inverse(cMA(complexTerm2,args1->Bracket(),endOfArgs1,variableIndex));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->Next(),variableIndex);
	case Incomparable:
	  if (alpha(args1->Bracket(),endOfArgs1,complexTerm2,variableIndex))
	    return Greater;
	  if (alpha(args2->Next(),complexTerm1,variableIndex))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Bracket();
      args2 = args2->Next();
    };
  return Equal;
}; 
inline
bool 
LPO::lexMAOrEqual(UnifTerm* complexTerm1,
		  UnifTerm* complexTerm2,
		  UnifTerm* args1,
		  UnifTerm* endOfArgs1,
		  UnifTerm* args2,
		  UnifTerm* endOfArgs2)
{
  CALL("lexMAOrEqual(UnifTerm* complexTerm1,..)");
  ASSERT(complexTerm1->symbol().Func().functor() ==
	 complexTerm2->symbol().Func().functor());
  ASSERT(args1 <= endOfArgs1 && args2 <= endOfArgs2);
  while (args1 != endOfArgs1)
    {
      switch (compareOnCurrentUnifier(args1,args2))
	{
	case Less:
	  return alpha(args1->Bracket(),endOfArgs1,complexTerm2);
	case Equal:
	  break;
	case Greater:
	  return 
	    majo(complexTerm1,args2->Bracket(),endOfArgs2);
	case Incomparable:
	  return alpha(args1->Bracket(),endOfArgs1,complexTerm2);
	};
      args1 = args1->Bracket();
      args2 = args2->Bracket();
    };
  ASSERT(args2 == endOfArgs2);
  return true; 
}; 
inline
bool 
LPO::lexMA(UnifTerm* complexTerm1,
	   UnifTerm* complexTerm2,
	   UnifTerm* args1,
	   UnifTerm* endOfArgs1,
	   UnifTerm* args2,
	   UnifTerm* endOfArgs2)
{
  CALL("lexMA(UnifTerm* complexTerm1,..)");
  ASSERT(complexTerm1->symbol().Func().functor() ==
	 complexTerm2->symbol().Func().functor());
  ASSERT(args1 <= endOfArgs1 && args2 <= endOfArgs2);
  while (args1 != endOfArgs1)
    {
      switch (compareOnCurrentUnifier(args1,args2))
	{
	case Less:
	  return alpha(args1->Bracket(),endOfArgs1,complexTerm2);
	case Equal:
	  break;
	case Greater:
	  return 
	    majo(complexTerm1,args2->Bracket(),endOfArgs2);
	case Incomparable:
	  return alpha(args1->Bracket(),endOfArgs1,complexTerm2);
	};
      args1 = args1->Bracket();
      args2 = args2->Bracket();
    };
  ASSERT(args2 == endOfArgs2);
  return false; 
}; 
inline 
bool 
LPO::majo(UnifTerm* complexTerm1,
	  UnifTerm* args2,
	  UnifTerm* endOfArgs2)
{
  CALL("majo(UnifTerm* complexTerm1,..)");
  ASSERT(args2 <= endOfArgs2);
  while (args2 != endOfArgs2)
    {
      if (!greaterOnCurrentUnifier(complexTerm1,args2))
	return false;
      args2 = args2->Bracket();
    };
  return true;
}; 
inline
bool LPO::alpha(UnifTerm* args1,
		UnifTerm* endOfArgs1,
		UnifTerm* complexTerm2)
{
  CALL("alpha(UnifTerm* args1,..)");
  ASSERT(args1 <= endOfArgs1);
  while (args1 != endOfArgs1)
    {
      if (greaterOrEqualOnCurrentUnifier(args1,complexTerm2))
	return true;
      args1 = args1->Bracket();
    };
  return false;
}; 
inline
bool LPO::alpha(const Term* args1,
		const Term* complexTerm2,
		ulong variableIndex)
{
  CALL("alpha(const Term* args1,const Term* complexTerm2,ulong variableIndex)");
  Term term2(const_cast<Term*>(complexTerm2));
  while (!args1->isNotArg())
    {
      switch (compareOnCurrentUnifier(*args1,term2,variableIndex))
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false; 
}; 
inline
bool LPO::alpha(UnifTerm* args1,
		UnifTerm* endOfArgs1,
		const Term* complexTerm2,
		ulong variableIndex)
{
  CALL("alpha(UnifTerm* args1,..)");
  ASSERT(args1 <= endOfArgs1);
  Term term2(const_cast<Term*>(complexTerm2));
  while (args1 != endOfArgs1)
    {
      switch (compareOnCurrentUnifier(args1,term2,variableIndex))
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Bracket();
    };
  return false;
}; 
inline
bool LPO::alpha(const Term* args1,
		UnifTerm* complexTerm2,
		ulong variableIndex)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      switch (compareOnCurrentUnifier(complexTerm2,*args1,variableIndex))
	{	 
	case Less:
	case Equal:
	  return true;
	case Greater:
	  break;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false; 
}; 
Comparison LPO::compare(const Term* complexTerm1,
			const Term* complexTerm2,
			const InstRet::Substitution* subst)
{
  CALL("compare(const Term* complexTerm1,..)");
  if (complexTerm1 == complexTerm2) return Equal;
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compare(*(complexTerm1->Args()),
			   subst,
			   *(complexTerm2->Args()));
	  };
#endif
      Comparison res =
	cLMA(complexTerm1,
	     subst,
	     complexTerm2,
	     complexTerm1->Args(),
	     complexTerm2->Args());
      ASSERT(res != Greater || 
	     complexGreaterThanProperSubterms(complexTerm1,subst,complexTerm2));
      ASSERT(res != Less || 
	     complexGreaterThanProperSubterms(complexTerm2,subst,complexTerm1));
      return res;
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConst(complexTerm2,
					 subst,
					 complexTerm1->symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConst(complexTerm1,
			     subst,
			     complexTerm2->symbolPrecedence());
	};
#endif
      Comparison res; 
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  res =
	    cMA(complexTerm1,subst,complexTerm2->Args());
	}
      else
	{
	  res =
	    BK::inverse(cMA(complexTerm2,subst,complexTerm1->Args()));
	};
      ASSERT(res != Greater || 
	     complexGreaterThanProperSubterms(complexTerm1,subst,complexTerm2));
      ASSERT(res != Less || 
	     complexGreaterThanProperSubterms(complexTerm2,subst,complexTerm1));
      return res;
    };
  ICP("END");
  return Incomparable;
}; 
inline
bool LPO::complexGreater(const Term* complexTerm1,
			 const InstRet::Substitution* subst,
			 const Term* complexTerm2)
{
  CALL("complexGreater(const Term* complexTerm1,..)");
  if (complexTerm1 == complexTerm2) return false;
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greater(*(complexTerm1->Args()),
			   subst,
			   *(complexTerm2->Args()));
	  };
#endif
      return
	lexMA(complexTerm1,
	      subst,
	      complexTerm2,
	      complexTerm1->Args(),
	      complexTerm2->Args());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   subst,
				   complexTerm1->symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       subst,
			       complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return
	    majo(complexTerm1,subst,complexTerm2->Args());
	}
      else
	{
	  return
	    alpha(complexTerm1->Args(),subst,complexTerm2);
	};
    };
}; 
bool 
LPO::complexGreaterThanProperSubterms(const Term* complexTerm1,
				      const InstRet::Substitution* subst,
				      const Term* complexTerm2)
{
  CALL("complexGreaterThanProperSubterms(..)");
  Term::Iterator iter(complexTerm2);
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar())
	{
	  Term instance = subst->map(iter.CurrPos()->var());
	  if (instance.isVariable())
	    {
	      if (!complexContainsVar(complexTerm1,subst,instance))
		return false;
	    }
	  else
	    {
	      if (!complexGreater(subst,complexTerm1,instance.deref()))
		return false;
	    };
	}
      else
	{
	  if (!complexGreater(complexTerm1,subst,iter.CurrPos()))
	    return false;
	};
    };
  return true;
}; 
inline
bool LPO::complexGreater(const InstRet::Substitution* subst,
			 const Term* complexTerm1,
			 const Term* complexTerm2)
{
  CALL("complexGreater(const InstRet::Substitution* subst,..)");
  ASSERT(complexTerm1->Positive() && complexTerm2->Positive());
  if (*complexTerm1 == *complexTerm2)
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greater(subst,
			   *(complexTerm1->Args()),
			   *(complexTerm2->Args()));
	  };
#endif
      return
	lexMA(subst,
	      complexTerm1,
	      complexTerm2,
	      complexTerm1->Args(),
	      complexTerm2->Args());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   complexTerm1->symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       subst,
			       complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return
	    majo(subst,complexTerm1,complexTerm2->Args());
	}
      else
	{
	  return
	    alpha(subst,complexTerm1->Args(),complexTerm2);
	};
    };
}; 
inline
bool LPO::complexGreater(const Term* complexTerm1,
			 const Term* complexTerm2,
			 const InstRet::Substitution* subst)
{
  CALL("complexGreater(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->Positive() && complexTerm2->Positive());
  if (*complexTerm1 == *complexTerm2)
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greater(*(complexTerm1->Args()),
			   *(complexTerm2->Args()),
			   subst);
	  };
#endif
      return
	lexMA(complexTerm1,
	      complexTerm2,
	      subst,
	      complexTerm1->Args(),
	      complexTerm2->Args());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   subst,
				   complexTerm1->symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return
	    majo(complexTerm1,complexTerm2->Args(),subst);
	}
      else
	{
	  return
	    alpha(complexTerm1->Args(),complexTerm2,subst);
	};
    };
}; 
inline
bool LPO::greater(const Term* complexTerm1,
		  const InstRet::Substitution* subst,
		  const Term& term2)
{
  CALL("greater(const Term* complexTerm1,..)");
  if (term2.isVariable())
    {
      Term instance2 = subst->map(term2.var());
      if (instance2.isVariable())
	{
	  return
	    complexContainsVar(complexTerm1,subst,instance2);
	}
      else
	return complexGreater(subst,complexTerm1,instance2.deref());
    }
  else
    return complexGreater(complexTerm1,subst,term2.deref());
}; 
inline
bool LPO::greater(const Term& term1,
		  const InstRet::Substitution* subst,
		  const Term& term2)
{
  CALL("greater(const Term& term1,..)");
  if (term1.isVariable())
    {
      return greater(subst->map(term1.var()),term2,subst);
    }
  else
    return greater(term1.deref(),subst,term2);
}; 
inline
bool LPO::greater(const Term& term1,
		  const Term& term2,
		  const InstRet::Substitution* subst)
{
  CALL("greater(const Term& term1,..)");
  if (term1.isVariable())
    {
      return false;
    }
  else
    return greater(term1.deref(),term2,subst);
}; 
inline
bool LPO::greater(const Term* complexTerm1,
		  const Term& term2,
		  const InstRet::Substitution* subst)
{
  CALL("greater(const Term* complexTerm1,..)");
  if (term2.isVariable())
    {
      return greater(complexTerm1,subst->map(term2.var()));
    }
  else
    return 
      complexGreater(complexTerm1,
		     term2.deref(),
		     subst);
}; 
inline
bool LPO::greater(const InstRet::Substitution* subst,
		  const Term* complexTerm1,
		  const Term& term2)
{
  CALL("greater(const InstRet::Substitution* subst,..)");
  if (term2.isVariable())
    {
      return
	complexContainsVar(complexTerm1,subst,term2);
    }
  else
    return complexGreater(subst,complexTerm1,term2.deref());
}; 
inline
bool LPO::greater(const InstRet::Substitution* subst,
		  const Term& term1,
		  const Term& term2)
{
  CALL("greater(const InstRet::Substitution* subst,..)");
  if (term1.isVariable())
    {
      return greater(subst->map(term1.var()),term2);
    }
  else
    return greater(subst,term1.deref(),term2);
}; 
inline
Comparison LPO::cMA(const Term* complexTerm1,
		    const InstRet::Substitution* subst,
		    const Term* args2)
{
  CALL("cMA(const Term* complexTerm1,..)#3");
  while (!args2->isNotArg())
    {      
      Comparison cmp;
      if (args2->isVariable())
	{
	  cmp = compare(subst,complexTerm1,subst->map(args2->var()));
	}
      else
	cmp = compare(complexTerm1,args2->deref(),subst);
      switch (cmp) 
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),subst,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline    
Comparison LPO::cMA(const InstRet::Substitution* subst,
		    const Term* complexTerm1,
		    const Term* args2)
{
  CALL("cMA(const InstRet::Substitution* subst,..)");
  while (!args2->isNotArg())
    {
      switch (compare(subst,complexTerm1,*args2))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline     
Comparison LPO::cMA(const Term* complexTerm1,
		    const Term* args2,
		    const InstRet::Substitution* subst)
{
  CALL("cMA(const Term* complexTerm1,..)#4");
  while (!args2->isNotArg())
    {
      switch (BK::inverse(compare(subst,*args2,complexTerm1)))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(subst,args2->Next(),complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline     
Comparison LPO::cLMA(const Term* complexTerm1,
		     const InstRet::Substitution* subst,
		     const Term* complexTerm2,
		     const Term* args1,
		     const Term* args2)
{
  CALL("cLMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,subst,*args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,subst,args1->Next()));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,subst,args2->Next());
	case Incomparable:
	  if (alpha(args1->Next(),subst,complexTerm2))
	    return Greater;
	  if (alpha(args2->Next(),subst,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return Equal;
}; 
inline
Comparison LPO::cLMA(const InstRet::Substitution* subst,
		     const Term* complexTerm1,
		     const Term* complexTerm2,
		     const Term* args1,
		     const Term* args2)
{
  CALL("cLMA(const InstRet::Substitution* subst,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(subst,*args1,*args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->Next(),subst));
	case Equal:
	  break;
	case Greater:
	  return cMA(subst,complexTerm1,args2->Next());
	case Incomparable:
	  if (alpha(subst,args1->Next(),complexTerm2))
	    return Greater;
	  if (alpha(args2->Next(),complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return Equal;  
}; 
inline
bool LPO::lexMA(const Term* complexTerm1,
		const InstRet::Substitution* subst,
		const Term* complexTerm2,
		const Term* args1,
		const Term* args2)
{
  CALL("lexMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,subst,*args2))
	{
	case Less:
	  return alpha(args1->Next(),subst,complexTerm2);
	case Equal:
	  break;
	case Greater:
	  return majo(complexTerm1,subst,args2->Next());
	case Incomparable:
	  return alpha(args1->Next(),subst,complexTerm2);
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return false; 
}; 
inline
bool LPO::lexMA(const InstRet::Substitution* subst,
		const Term* complexTerm1,
		const Term* complexTerm2,
		const Term* args1,
		const Term* args2)
{
  CALL("lexMA(const InstRet::Substitution* subst,..)");
  ASSERT(complexTerm1->Positive() && complexTerm2->Positive());
  while (!args1->isNotArg())
    {
      switch (compare(subst,*args1,*args2))
	{
	case Less:
	  return alpha(subst,args1->Next(),complexTerm2);
	case Equal:
	  break;
	case Greater:
	  return majo(subst,complexTerm1,args2->Next());
	case Incomparable:
	  return alpha(subst,args1->Next(),complexTerm2);
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return false; 
}; 
inline
bool LPO::lexMA(const Term* complexTerm1,
		const Term* complexTerm2,
		const InstRet::Substitution* subst,
		const Term* args1,
		const Term* args2)
{
  CALL("lexMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->Positive() && complexTerm2->Positive());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,*args2,subst))
	{
	case Less:
	  return alpha(args1->Next(),complexTerm2,subst);
	case Equal:
	  break;
	case Greater:
	  return majo(complexTerm1,args2->Next(),subst);
	case Incomparable:
	  return alpha(args1->Next(),complexTerm2,subst);
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return false; 
}; 
inline
bool LPO::majo(const Term* complexTerm1,
	       const InstRet::Substitution* subst,
	       const Term* args2)
{
  CALL("majo(const Term* complexTerm1,..)");
  while (!args2->isNotArg())
    {
      if (!greater(complexTerm1,subst,*args2))
	return false;
      args2 = args2->Next();
    };
  return true;
}; 
inline
bool LPO::majo(const InstRet::Substitution* subst,
	       const Term* complexTerm1,
	       const Term* args2)
{
  CALL("majo(const InstRet::Substitution* subst,..)");
  while (!args2->isNotArg())
    {
      if (!greater(subst,complexTerm1,*args2))
	return false;
      args2 = args2->Next();
    };
  return true;
}; 
inline
bool LPO::majo(const Term* complexTerm1,
	       const Term* args2,
	       const InstRet::Substitution* subst)
{
  CALL("majo(const Term* complexTerm1,..)");
  while (!args2->isNotArg())
    {
      if (!greater(complexTerm1,*args2,subst))
	return false;
      args2 = args2->Next();
    };
  return true;
}; 
inline
bool LPO::alpha(const Term* args1,
		const InstRet::Substitution* subst,
		const Term* complexTerm2)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      Comparison cmp;
      if (args1->isVariable())
	{
	  cmp = 
	    BK::inverse(compare(subst,complexTerm2,subst->map(args1->var())));
	}
      else
	{
	  cmp = 
	    compare(args1->deref(),complexTerm2,subst);
	};
      switch (cmp)
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false;
}; 
inline
bool LPO::alpha(const Term* args1,
		const Term* complexTerm2,
		const InstRet::Substitution* subst)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      switch (BK::inverse(compare(subst,complexTerm2,*args1)))
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false;
}; 
bool LPO::alpha(const InstRet::Substitution* subst,
		const Term* args1,
		const Term* complexTerm2)
{
  CALL("alpha(const InstRet::Substitution* subst,..)");
  while (!args1->isNotArg())
    {
      switch (compare(subst,*args1,complexTerm2))
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false;
}; 
inline
bool LPO::complexContainsVar(const Term* complexTerm,
			     const InstRet::Substitution* subst,
			     const Term& var)
{
  CALL("complexContainsVar(const Term* complexTerm,..)");
  InstRet::Substitution::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (iter.symbol() == var) return true;
    };
  return false;
}; 
Comparison LPO::compare(const Term& term1,
			const Term& term2,
			const ForwardMatchingSubstitutionCache* subst)
{
  CALL("compare(const Term& term1,const Term& term2,const ForwardMatchingSubstitutionCache* subst)");
  if (term1 == term2) return Equal;
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  return compare(subst->map(term1.var()),
			 subst->map(term2.var()));
	}
      else 
	{
	  return BK::inverse(compare(term2,subst,subst->map(term1.var())));
	};
    }
  else
    if (term2.isVariable())
      {
	return compare(term1,subst,subst->map(term2.var()));
      }
    else 
      {
	const Term* complexTerm1 = term1.deref();
	const Term* complexTerm2 = term2.deref();
	if (*complexTerm1 == *complexTerm2)
	  {
#ifdef LPO_SMALL_ARITY
	    if (complexTerm1->arity() == 0UL)
	      {
		return Equal;
	      }
	    else
	      if (complexTerm1->arity() == 1UL)
		{
		  return compare(*(complexTerm1->Args()),
				 *(complexTerm2->Args()),
				 subst);
		};
#endif
	    Comparison res = 
	      cLMA(complexTerm1,
		   complexTerm2,
		   complexTerm1->Args(),
		   complexTerm2->Args(),
		   subst);
	    ASSERT(res != Greater ||
		   complexGreaterThanProperSubterms(complexTerm1,
						    complexTerm2,
						    subst));
	    ASSERT(res != Less ||
		   complexGreaterThanProperSubterms(complexTerm2,
						    complexTerm1,
						    subst));
	    return res;
	  }
	else
	  {
#ifdef LPO_SMALL_ARITY
	    if (complexTerm1->arity() == 0UL)
	      {
		return 
		  BK::inverse(compareWithConst(complexTerm2,
					       subst,
					       complexTerm1->symbolPrecedence()));
	      };
	    if (complexTerm2->arity() == 0UL)
	      {
		return 
		  compareWithConst(complexTerm1,
				   subst,
				   complexTerm2->symbolPrecedence());
	      };
#endif
	    Comparison res;
	    if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	      {
		res = 
		  cMA(complexTerm1,complexTerm2->Args(),subst);
	      }
	    else
	      {
		res = 
		  BK::inverse(cMA(complexTerm2,complexTerm1->Args(),subst));
	      };
	    ASSERT(res != Greater ||
		   complexGreaterThanProperSubterms(complexTerm1,
						    complexTerm2,
						    subst));
	    ASSERT(res != Less ||
		   complexGreaterThanProperSubterms(complexTerm2,
						    complexTerm1,
						    subst));
	    return res;
	  };
      };
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::cMA(const Term* complexTerm1,
		    const Term* args2,
		    const ForwardMatchingSubstitutionCache* subst)
{
  CALL("cMA(const Term* complexTerm1,..)#5");
  while (!args2->isNotArg())
    {      
      Comparison cmp;
      Term term1(const_cast<Term*>(complexTerm1));
      if (args2->isVariable())
	{
	  cmp = compare(term1,subst,subst->map(args2->var()));
	}
      else
	cmp = compare(term1,*args2,subst);
      switch (cmp) 
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline     
Comparison LPO::cLMA(const Term* complexTerm1,
		     const Term* complexTerm2,
		     const Term* args1,
		     const Term* args2,
		     const ForwardMatchingSubstitutionCache* subst)
{
  CALL("cLMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,*args2,subst))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->Next(),subst));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->Next(),subst);
	case Incomparable:
	  if (alpha(args1->Next(),complexTerm2,subst))
	    return Greater;
	  if (alpha(args2->Next(),complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return Equal;
}; 
inline
bool LPO::alpha(const Term* args1,
		const Term* complexTerm2,
		const ForwardMatchingSubstitutionCache* subst)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      Comparison cmp;
      Term term2(const_cast<Term*>(complexTerm2));
      if (args1->isVariable())
	{
	  cmp = 
	    BK::inverse(compare(term2,subst,subst->map(args1->var())));
	}
      else
	{
	  cmp = 
	    compare(*args1,term2,subst);
	};
      switch (cmp)
	{	 
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false;
}; 
inline
bool LPO::alpha(const Flatterm* args1,
		const Flatterm* endOfArgs1,
		const Term* complexTerm2,
		const ForwardMatchingSubstitutionCache* subst)
{
  CALL("alpha(const Flatterm* args1,..)");
  Term term2(const_cast<Term*>(complexTerm2));
  while (args1 != endOfArgs1)
    {
      switch (BK::inverse(compare(term2,subst,args1)))
	{
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->after();
    };
  return false;  
}; 
inline
bool LPO::alpha(const Term* args1,
		const ForwardMatchingSubstitutionCache* subst,
		const Flatterm* complexTerm2)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      switch (compare(*args1,subst,complexTerm2))
	{
	case Less:
	  break;
	case Equal:
	case Greater:
	  return true;
	case Incomparable:
	  break;
	};
      args1 = args1->Next();
    };
  return false;  
}; 
SimplificationOrdering::SpecialisedComparison* 
LPO::reserveSpecialisedComparison()
{
  CALL("reserveSpecialisedComparison()");
  SpecialisedComparison* result = _specialisedComparisonPool;
  if (result)
    {
      _specialisedComparisonPool = _specialisedComparisonPool->getNextInPool();
    }
  else
    {
      result = new SpecialisedComparison(this);
    };
  return 
    static_cast<SimplificationOrdering::SpecialisedComparison*>(result);
}; 
void LPO::releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc)");
  (static_cast<LPO::SpecialisedComparison*>(sc))->setNextInPool(_specialisedComparisonPool);
  _specialisedComparisonPool = static_cast<LPO::SpecialisedComparison*>(sc);
}; 
SimplificationOrdering::StoredConstraint*
LPO::storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintGreater();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
LPO::storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintGreaterOrEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
LPO::storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintLess();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
LPO::storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintLessOrEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
LPO::storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintNotEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return 
    static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint* LPO::copy(SimplificationOrdering::StoredConstraint* constr)
{
  CALL("copy(SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(constr);
  (static_cast<StoredConstraint*>(static_cast<void*>(constr)))->incReferenceCounter();
  return constr;
}; 
void LPO::releaseConstraint(SimplificationOrdering::StoredConstraint* constr)
{
  CALL("releaseConstraint(SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(constr);
  (static_cast<StoredConstraint*>(static_cast<void*>(constr)))->decReferenceCounter();
  if (!(static_cast<StoredConstraint*>(static_cast<void*>(constr)))->referenceCounter())
    delete (static_cast<StoredConstraint*>(static_cast<void*>(constr)));
}; 
SimplificationOrdering::ConstraintCheckOnUnifier* 
LPO::reserveConstraintCheckOnUnifier()
{
  CALL("reserveConstraintCheckOnUnifier()");  
  ConstraintCheckOnUnifier* result = _constraintCheckOnUnifierPool;
  if (result)
    {
      _constraintCheckOnUnifierPool = _constraintCheckOnUnifierPool->getNextInPool();
    }
  else
    {
      result = new ConstraintCheckOnUnifier(this);
    };
  return 
    static_cast<SimplificationOrdering::ConstraintCheckOnUnifier*>(result);
}; 
void 
LPO::releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu)
{
  CALL("releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu)");
  (static_cast<LPO::ConstraintCheckOnUnifier*>(ccu))->setNextInPool(_constraintCheckOnUnifierPool);
  _constraintCheckOnUnifierPool = static_cast<LPO::ConstraintCheckOnUnifier*>(ccu);
}; 
SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* 
LPO::reserveConstraintCheckOnForwardMatchingSubst()
{
  CALL("reserveConstraintCheckOnForwardMatchingSubst()");
  ConstraintCheckOnForwardMatchingSubst* result = _constraintCheckOnForwardMatchingSubstPool;
  if (result)
    {
      _constraintCheckOnForwardMatchingSubstPool = 
	_constraintCheckOnForwardMatchingSubstPool->getNextInPool();
    }
  else
    {
      result = new ConstraintCheckOnForwardMatchingSubst(this);
    };
  return 
    static_cast<SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst*>(result);
}; 
void 
LPO::releaseConstraintCheckOnForwardMatchingSubst(SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* ccfms)
{
  CALL("releaseConstraintCheckOnForwardMatchingSubst(..)");
  (static_cast<LPO::ConstraintCheckOnForwardMatchingSubst*>(ccfms))->
    setNextInPool(_constraintCheckOnForwardMatchingSubstPool);
  _constraintCheckOnForwardMatchingSubstPool = 
    static_cast<LPO::ConstraintCheckOnForwardMatchingSubst*>(ccfms);
}; 
SimplificationOrdering::CompiledConstraintForUnifiers* 
LPO::reserveCompiledConstraintForUnifiers()
{
  CALL("reserveCompiledConstraintForUnifiers()");
  LPO::CompiledConstraintForUnifiers* result = _compiledConstraintOnUnifiersPool;
  if (result)
    {
      _compiledConstraintOnUnifiersPool = result->getNextInPool();
    }
  else
    {
      result = new CompiledConstraintForUnifiers(this);
    };
  ASSERT(result);
  return 
    static_cast<SimplificationOrdering::CompiledConstraintForUnifiers*>(result);
}; 
void 
LPO::releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc)
{
  CALL("releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc)");
  (static_cast<LPO::CompiledConstraintForUnifiers*>(cc))->setNextInPool(_compiledConstraintOnUnifiersPool);
  _compiledConstraintOnUnifiersPool = static_cast<LPO::CompiledConstraintForUnifiers*>(cc);
}; 
SimplificationOrdering::CompiledConstraintForInstanceRetrieval* 
LPO::reserveCompiledConstraintForInstanceRetrieval()
{
  CALL("reserveCompiledConstraintForInstanceRetrieval()");
  LPO::CompiledConstraintForInstanceRetrieval* result = _compiledConstraintForInstanceRetrievalPool;
  if (result)
    {
      _compiledConstraintForInstanceRetrievalPool = result->getNextInPool();
    }
  else
    {
      result = new CompiledConstraintForInstanceRetrieval(this);
    };
  ASSERT(result);
  return
    static_cast<SimplificationOrdering::CompiledConstraintForInstanceRetrieval*>(result);
}; 
void 
LPO::releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc)
{
  CALL("releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc)");
  (static_cast<LPO::CompiledConstraintForInstanceRetrieval*>(cc))->setNextInPool(_compiledConstraintForInstanceRetrievalPool);
  _compiledConstraintForInstanceRetrievalPool = static_cast<LPO::CompiledConstraintForInstanceRetrieval*>(cc);
}; 
Comparison LPO::compareWithTransposition(const Flatterm* term)
{  
  CALL("compareWithTransposition(const Flatterm* term)");
 ASSERT(term->isComplex());
 ASSERT(term->arity() == 2);
 const Flatterm* arg1 = term->next();
 const Flatterm* arg2 = arg1->after();
 return compare(arg1,arg2);
}; 
inline
bool 
LPO::termWithUnifierContainsVariable(UnifTerm* complexTerm,
				     Unifier::Variable* var)
{ 
  CALL("termWithUnifierContainsVariable(UnifTerm* complexTerm,Unifier::Variable* var)");
  ObjectPool<UnifTerm::Iter>::EntryHandle 
    iter(static_cast<ObjectPool<UnifTerm::Iter>&>(_prefixIterPool));
  iter->Reset(complexTerm);
  iter->SkipSym();
  while (iter->NotEmpty())
    {
      if (iter->CurrVar() == var) return true;
      iter->SkipSym();
    };
  return false;
}; 
inline
bool 
LPO::termWithUnifierContainsVariable(const Term* complexTerm,
				     ulong variableIndex,
				     Unifier::Variable* var)
{
  CALL("termWithUnifierContainsVariable(const Term* complexTerm,ulong variableIndex,Unifier::Variable* var)");
  ObjectPool<InstanceWithUnifier>::EntryHandle iter(_stTermWithSubstPool);
  iter->Reset(variableIndex,complexTerm);
  iter->SkipSym();
  while (!iter->Empty())
    {
      if (iter->CurrSym().isVariable() &&
	  (iter->CurrSym().var() == var))
	return true;
      iter->SkipSym();
    };
  return false;
}; 
inline
bool 
LPO::termWithUnifierContainsVariable(const Term& term,
				     ulong variableIndex,
				     Unifier::Variable* var)
{
  CALL("termWithUnifierContainsVariable(const Term& term,ulong variableIndex,Unifier::Variable* var)");
  if (term.isVariable())
    {
      Unifier::Variable* v = 
	Unifier::current()->variable(term.var(),variableIndex);
      bool freeVar;
      void* instance = v->Unref(freeVar);
      if (freeVar)
	{
	  return (static_cast<Unifier::Variable*>(instance)) == var;
	}
      else
	{
	  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance));
	  return 
	    termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance),var);
	};
    } 
  else
    {
      return termWithUnifierContainsVariable(term.deref(),
					     variableIndex,
					     var);
    };
}; 
inline
Comparison LPO::compareComplexOnCurrentUnifier(const Term* complexTerm1,
					       const Term* complexTerm2,
					       ulong variableIndex)
{
  CALL("compareComplexOnCurrentUnifier(const Term* complexTerm1,..)");
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compareOnCurrentUnifier(*(complexTerm1->Args()),
					   *(complexTerm2->Args()),
					   variableIndex);
	  };
#endif
      return 
	cLMA(complexTerm1,
	     complexTerm2,
	     complexTerm1->Args(),
	     complexTerm2->Args(),
	     variableIndex);
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConstOnCurrentUnifier(complexTerm2,
							 variableIndex,
							 complexTerm1->symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConstOnCurrentUnifier(complexTerm1,
					     variableIndex,
					     complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{	
	  return
	    cMA(complexTerm1,complexTerm2->Args(),variableIndex);
	}
      else
	{
	  return 
	    BK::inverse(cMA(complexTerm2,complexTerm1->Args(),variableIndex));
	};
    };
}; 
bool 
LPO::complexGreaterThanProperSubtermsOnCurrentUnifier(const Term* complexTerm1,
						      const Term* complexTerm2,
						      ulong variableIndex)
{
  CALL("complexGreaterThanProperSubtermsOnCurrentUnifier(..)#2");
  Term::Iterator iter(complexTerm2);
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar())
	{
	  Unifier::Variable* var = 
	    Unifier::current()->variable(iter.CurrPos()->var(),variableIndex);
	  bool freeVar;
	  void* instance = var->Unref(freeVar);
	  if (freeVar)
	    {
	      if (!termWithUnifierContainsVariable(complexTerm1,
						   variableIndex,
						   static_cast<Unifier::Variable*>(instance)))
		return false;
	    }
	  else
	    {
	      UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance));
	      if (compareComplexOnCurrentUnifier(static_cast<UnifTerm*>(instance),
						 complexTerm1,
						 variableIndex) != 
		  Less)
		return false;
	    };
	}
      else
	{
	  if (compareComplexOnCurrentUnifier(complexTerm1,
					     iter.CurrPos(),
					     variableIndex) !=
	      Greater)
	    return false;
	};
    };
  return true;
} 
inline
Comparison 
LPO::compareComplexOnCurrentUnifier(UnifTerm* complexTerm1,
				    const Term* complexTerm2,
				    ulong variableIndex)
{
  CALL("compareComplexOnCurrentUnifier(UnifTerm* complexTerm1,const Term* complexTerm2,..)");
  ASSERT(complexTerm1->symbol().IsFunctor());
  UnifTerm::CorrectBracket(complexTerm1);
  if (complexTerm1->symbol().Func().functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm2->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm2->arity() == 1UL)
	  {
	    return compareOnCurrentUnifier(complexTerm1 + 1,
					   *(complexTerm2->Args()),
					   variableIndex);
	  };
#endif
      return
	cLMA(complexTerm1,
	     complexTerm2,
	     complexTerm1 + 1,
	     complexTerm1->Bracket(),
	     complexTerm2->Args(),
	     variableIndex);
    }
  else
    {      
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->symbol().Func().arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConstOnCurrentUnifier(complexTerm2,
							 variableIndex,
							 complexTerm1->symbol().Func().symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConstOnCurrentUnifier(complexTerm1,
					     complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->symbol().Func().hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return
	    cMA(complexTerm1,
		complexTerm2->Args(),
		variableIndex);
	}
      else
	{
	  return
	    BK::inverse(cMA(complexTerm2,
			    complexTerm1 + 1,
			    complexTerm1->Bracket(),
			    variableIndex));
	};
    };
}; 
bool 
LPO::complexGreaterThanProperSubtermsOnCurrentUnifier(const Term* complexTerm1,
						      ulong variableIndex,
						      UnifTerm* complexTerm2)
{
  CALL("complexGreaterThanProperSubtermsOnCurrentUnifier(..)#3");
  UnifTerm::CorrectBracket(complexTerm2);
  UnifTerm::Iter iter(complexTerm2);
  iter.SkipSym(); 
  while (iter.NotEmpty())
    {
      if (compareOnCurrentUnifier(iter.CurrSubterm(),
				  complexTerm1,
				  variableIndex) != 
	  Less)
	return false;
      iter.SkipSym();
    };
  return true;
}; 
bool 
LPO::complexGreaterThanProperSubtermsOnCurrentUnifier(UnifTerm* complexTerm1,
						      const Term* complexTerm2,
						      ulong variableIndex)
{
  CALL("complexGreaterThanProperSubtermsOnCurrentUnifier(..)#4");
  UnifTerm::CorrectBracket(complexTerm1);
  Term::Iterator iter(complexTerm2);
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar())
	{
	  Unifier::Variable* var = 
	    Unifier::current()->variable(iter.CurrPos()->var(),variableIndex);
	  bool freeVar;
	  void* instance = var->Unref(freeVar);
	  if (freeVar)
	    {
	      if (!termWithUnifierContainsVariable(complexTerm1,
						   static_cast<Unifier::Variable*>(instance)))
		return false;
	    }
	  else
	    {
	      UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance));
	      if (!complexGreaterOnCurrentUnifier(complexTerm1,
						  static_cast<UnifTerm*>(instance)))
		return false;
	    };
	}
      else
	{
	  if (compareComplexOnCurrentUnifier(complexTerm1,
					     iter.CurrPos(),
					     variableIndex) !=
	      Greater)
	    return false;
	};
    };
  return true;
}; 
inline
Comparison LPO::compareOnCurrentUnifier(UnifTerm* term1,
					const Term* complexTerm2,
					ulong variableIndex)
{
  CALL("compareOnCurrentUnifier(UnifTerm* term1,..)");
  if (term1->symbol().isVariable())
    {
      bool freeVar1;
      void* instance1 = term1->symbol().var()->Unref(freeVar1);
      if (freeVar1)
	{
	  if (termWithUnifierContainsVariable(complexTerm2,
					      variableIndex,
					      static_cast<Unifier::Variable*>(instance1)))
	    return Less;
	  return Incomparable;
	}
      else
	{
	  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
	  return 
	    compareComplexOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
					   complexTerm2,
					   variableIndex);
	};
    }
  else
    return compareComplexOnCurrentUnifier(term1,
					  complexTerm2,
					  variableIndex);
}; 
inline
Comparison LPO::compareOnCurrentUnifier(UnifTerm* term1,
					const Term& term2,
					ulong variableIndex)
{
  CALL("compareOnCurrentUnifier(UnifTerm* term1,..)");
  UnifTerm::CorrectBracket(term1);
  if (term2.isVariable())
    {
      Unifier::Variable* var2 = 
	Unifier::current()->variable(term2.var(),variableIndex);
      bool freeVar2;
      void* instance2 = var2->Unref(freeVar2);
      if (freeVar2)
	{
	  if (term1->symbol().isVariable())
	    {
	      Unifier::Variable* var1 = term1->symbol().var();
	      bool freeVar1;
	      void* instance1 = var1->Unref(freeVar1);
	      if (freeVar1)
		{
		  if (instance1 == instance2) return Equal;
		  return Incomparable; 
		}
	      else
		{
		  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance1));
		  if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance1),
						      static_cast<Unifier::Variable*>(instance2)))
		    return Greater;
		  return Incomparable;
		};
	    }
	  else
	    {
	      if (termWithUnifierContainsVariable(term1,
						  static_cast<Unifier::Variable*>(instance2)))
		return Greater;
	      return Incomparable;
	    };
	}
      else
	{
	  UnifTerm::CorrectBracket(static_cast<UnifTerm*>(instance2));
	  return
	    compareOnCurrentUnifier(term1,static_cast<UnifTerm*>(instance2));
	};
    }
  else
    return 
      compareOnCurrentUnifier(term1,term2.deref(),variableIndex);
  ICP("END");
  return Incomparable;
}; 
inline
bool 
LPO::groundWithAllPrecBelowOnCurrentUnifier(UnifTerm* complexTerm,
					    long prec)
{
  CALL("groundWithAllPrecBelowOnCurrentUnifier(UnifTerm* complexTerm,..)");
  ASSERT(complexTerm->symbol().IsFunctor());
  ASSERT(complexTerm->symbol().Func().symbolPrecedence() != prec);
  if (complexTerm->symbol().Func().symbolPrecedence() > prec)
    return false;
  UnifTerm::Iter iter(complexTerm);
  iter.SkipSym();
  while (iter.NotEmpty())
    {
      if (iter.CurrVar() ||
	  iter.CurrSubterm()->symbol().Func().symbolPrecedence() >= prec)
	return false;
      iter.SkipSym();
    };
  return true;
}; 
inline
bool
LPO::containsEqOrGrPrecOnCurrentUnifier(UnifTerm* complexTerm,
					long prec)
{
  CALL("containsEqOrGrPrecOnCurrentUnifier(UnifTerm* complexTerm,..)");
  ASSERT(complexTerm->symbol().IsFunctor());
  ASSERT(complexTerm->symbol().Func().symbolPrecedence() != prec);
  if (complexTerm->symbol().Func().symbolPrecedence() > prec)
    return true;
  UnifTerm::Iter iter(complexTerm);
  iter.SkipSym();
  while (iter.NotEmpty())
    {
      if (!iter.CurrVar() &&
	  iter.CurrSubterm()->symbol().Func().symbolPrecedence() >= prec)
	return true;
      iter.SkipSym();
    };
  return false;
}; 
inline
Comparison 
LPO::compareWithConstOnCurrentUnifier(UnifTerm* complexTerm,
				      long constPrec)
{
  CALL("compareWithConstOnCurrentUnifier(UnifTerm* complexTerm,..)");
  ASSERT(complexTerm->symbol().IsFunctor());
  ASSERT(complexTerm->symbol().Func().symbolPrecedence() != constPrec);
  if (complexTerm->symbol().Func().symbolPrecedence() > constPrec)
    return Greater;
  UnifTerm::Iter iter(complexTerm);
  iter.SkipSym();
  bool ground = true;
  while (iter.NotEmpty())
    {
      if (iter.CurrVar())
	{
	  ground = false;
	}
      else
	if (iter.CurrSubterm()->symbol().Func().symbolPrecedence() >= constPrec)
	  return Greater;
      iter.SkipSym();
    };
  return (ground)? Less : Incomparable;
}; 
bool 
LPO::groundWithAllPrecBelowOnCurrentUnifier(const Term* complexTerm,
					    ulong variableIndex,
					    long prec)
{
  CALL("groundWithAllPrecBelowOnCurrentUnifier(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return false;
  ObjectPool<InstanceWithUnifier>::EntryHandle iter(_stTermWithSubstPool);
  iter->Reset(variableIndex,complexTerm);
  iter->SkipSym();
  while (!iter->Empty())
    {
      if (iter->CurrSym().isVariable() ||
	  iter->CurrSym().Func().symbolPrecedence() >= prec) 
	return false;
      iter->SkipSym();
    };
  return true;
}; 
inline
bool 
LPO::containsEqOrGrPrecOnCurrentUnifier(const Term* complexTerm,
					ulong variableIndex,
					long prec)
{
  CALL("containsEqOrGrPrecOnCurrentUnifier(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return true;
  ObjectPool<InstanceWithUnifier>::EntryHandle iter(_stTermWithSubstPool);
  iter->Reset(variableIndex,complexTerm);
  iter->SkipSym();
  while (!iter->Empty())
    {
      if (!iter->CurrSym().isVariable() &&
	  iter->CurrSym().Func().symbolPrecedence() >= prec) 
	return true;
      iter->SkipSym();
    };
  return false;
}; 
inline
Comparison 
LPO::compareWithConstOnCurrentUnifier(const Term* complexTerm,
				      ulong variableIndex,
				      long constPrec)
{
  CALL("compareWithConstOnCurrentUnifier(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != constPrec);
  if (complexTerm->symbolPrecedence() > constPrec) return Greater;
  ObjectPool<InstanceWithUnifier>::EntryHandle iter(_stTermWithSubstPool);
  iter->Reset(variableIndex,complexTerm);
  iter->SkipSym();
  bool ground = true;
  while (!iter->Empty())
    {
      if (iter->CurrSym().isVariable())
	{
	  ground = false;
	}
      else
	if (iter->CurrSym().Func().symbolPrecedence() >= constPrec) 
	  return Greater;
      iter->SkipSym();
    };
  return (ground)? Less : Incomparable;
}; 
inline
Comparison 
LPO::compare(const InstRet::Substitution* subst,
	     const Term* complexTerm1,
	     const Term& term2)
{
  CALL("compare(const Term* complexTerm1,..)");
  if (term2.isVariable())
    {
      if (complexContainsVar(complexTerm1,subst,term2))
	return Greater;
      return Incomparable;
    }
  else
    {
      const Term* complexTerm2 = term2.deref();
      if (*complexTerm1 == *complexTerm2)
	{
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compare(subst,
			   *(complexTerm1->Args()),
			   *(complexTerm2->Args()));
	  };
#endif
	  return cLMA(subst,
		      complexTerm1,
		      complexTerm2,
		      complexTerm1->Args(),
		      complexTerm2->Args());
	}
      else
	{
#ifdef LPO_SMALL_ARITY
	  if (complexTerm1->arity() == 0UL)
	    {
	      return 
		BK::inverse(compareWithConst(complexTerm2,
					     complexTerm1->symbolPrecedence()));
	    };
	  if (complexTerm2->arity() == 0UL)
	    {
	      return 
		compareWithConst(complexTerm1,
				 subst,
				 complexTerm2->symbolPrecedence());
	    };
#endif
	  if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	    {
	      return 
		cMA(subst,
		    complexTerm1,
		    complexTerm2->Args());
	    }
	  else
	    {
	      return
		BK::inverse(cMA(complexTerm2,
				complexTerm1->Args(),
				subst));
	    };
	};
    };
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::compare(const InstRet::Substitution* subst,
			const Term* complexTerm1,
			const Term* complexTerm2)
{
  CALL("compare(const InstRet::Substitution* subst,..)");
  if (*complexTerm1 == *complexTerm2)
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compare(subst,
			   *(complexTerm1->Args()),
			   *(complexTerm2->Args()));
	  };
#endif
      return cLMA(subst,
		  complexTerm1,
		  complexTerm2,
		  complexTerm1->Args(),
		  complexTerm2->Args());
    }
  else 
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConst(complexTerm2,
					 complexTerm1->symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConst(complexTerm1,
			     subst,
			     complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return 
	    cMA(subst,
		complexTerm1,
		complexTerm2->Args());
	}
      else
	{
	  return
	    BK::inverse(cMA(complexTerm2,
			    complexTerm1->Args(),
			    subst));
	};
    };
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::compare(const InstRet::Substitution* subst,
			const Term& term1,
			const Term* complexTerm2)
{
  CALL("compare(const Term& term1,..)");
  if (term1.isVariable())
    {
      Term t1(subst->map(term1.var()));
      if (t1.isVariable())
	{
	  if (complexTerm2->containsSymbol(t1)) return Less;
	  return Incomparable;
	}
      else
	{
	  return compare(t1.deref(),complexTerm2);
	};
    }
  else
    {
      const Term* complexTerm1 = term1.deref();
      if (*complexTerm1 == *complexTerm2)
	{
#ifdef LPO_SMALL_ARITY
	  if (complexTerm1->arity() == 0UL)
	    {
	      return Equal;
	    }
	  else
	    if (complexTerm1->arity() == 1UL)
	      {
		return compare(subst,
			       *(complexTerm1->Args()),
			       *(complexTerm2->Args()));
	      };
#endif
	  return cLMA(subst,
		      complexTerm1,
		      complexTerm2,
		      complexTerm1->Args(),
		      complexTerm2->Args());
	}
      else 
	{
#ifdef LPO_SMALL_ARITY
	  if (complexTerm1->arity() == 0UL)
	    {
	      return 
		BK::inverse(compareWithConst(complexTerm2,
					     complexTerm1->symbolPrecedence()));
	    };
	  if (complexTerm2->arity() == 0UL)
	    {
	      return 
		compareWithConst(complexTerm1,
				 subst,
				 complexTerm2->symbolPrecedence());
	    };
#endif
	if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	  {
	    return 
	      cMA(subst,
		  complexTerm1,
		  complexTerm2->Args());
	  }
	else
	  {
	    return
	      BK::inverse(cMA(complexTerm2,
			      complexTerm1->Args(),
			      subst));
	  };
	};
    };
  ICP("END");
  return Incomparable;
}; 
inline
Comparison LPO::compare(const InstRet::Substitution* subst,
			const Term& term1,
			const Term& term2)
{
  CALL("compare(const InstRet::Substitution* subst,..)");
  if (term1.isVariable())
    {
      return compare(subst->map(term1.var()),term2);
    }
  else
    return compare(subst,term1.deref(),term2);
}; 
inline
Comparison LPO::compare(const Term& term1,
			const Term& term2,
			const InstRet::Substitution* subst)
{
  CALL("const Term& term1,..)");
  return 
    BK::inverse(compare(subst,term2,term1));
}; 
inline
Comparison LPO::compare(const Term& term1,
			const Term* complexTerm2,
			const InstRet::Substitution* subst)
{
  CALL("compare(const Term& term1,..)");
  if (term1.isVariable())
    {
      if (complexContainsVar(complexTerm2,subst,term1))
	return Less;
      return Incomparable;
    }
  else
    return BK::inverse(compare(subst,complexTerm2,term1.deref()));
}; 
inline
bool 
LPO::groundWithAllPrecBelow(const Term* complexTerm,
			    const InstRet::Substitution* subst,
			    long prec)
{
  CALL("groundWithAllPrecBelow(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return false;
  InstRet::Substitution::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (iter.symbol().isVariable() ||
	  iter.symbol().symbolPrecedence() >= prec) 
	return false;
    };
  return true;
}; 
inline
bool 
LPO::containsEqOrGrPrec(const Term* complexTerm,
			const InstRet::Substitution* subst,
			long prec)
{
  CALL("containsEqOrGrPrec(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return true;
  InstRet::Substitution::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (!iter.symbol().isVariable() &&
	  iter.symbol().symbolPrecedence() >= prec) 
	return true;
    };
  return false;
}; 
inline
Comparison
LPO::compareWithConst(const Term* complexTerm,
		      const InstRet::Substitution* subst,
		      long constPrec)
{
  CALL("compareWithConst(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != constPrec);
  if (complexTerm->symbolPrecedence() > constPrec) return Greater;
  InstRet::Substitution::Instance iter;
  iter.reset(complexTerm,subst);
  bool ground = true;
  while (iter.next())
    {
      if (iter.symbol().isVariable())
	{
	  ground = false;
	}
      else
	if (iter.symbol().symbolPrecedence() >= constPrec) 
	  return Greater;
    };
  return (ground)? Less : Incomparable;
}; 
inline 
Comparison 
LPO::compare(const Term* complexTerm1,const Term* complexTerm2)
{
  CALL("compare(const Term* complexTerm1,const Term* complexTerm2)");
  if (complexTerm1 == complexTerm2) return Equal;
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return Equal;
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return compare(*(complexTerm1->Args()),
			   *(complexTerm2->Args()));
	  };
#endif
      return 
	cLMA(complexTerm1,
	     complexTerm2,
	     complexTerm1->Args(),
	     complexTerm2->Args());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    BK::inverse(compareWithConst(complexTerm2,
					 complexTerm1->symbolPrecedence()));
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return 
	    compareWithConst(complexTerm1,
			     complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return cMA(complexTerm1,complexTerm2->Args());
	}
      else
	{
	  return BK::inverse(cMA(complexTerm2,complexTerm1->Args()));
	};
    };
}; 
Comparison 
LPO::compare(const Term& term1,const Term& term2)
{
  CALL("compare(const Term& term1,const Term& term2)");
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  if (term1 == term2) return Equal;
	}
      else
	{
	  if (term2.deref()->containsSymbol(term1)) return Less;
	};
      return Incomparable;
    }
  else
    if (term2.isVariable())
      {
	if (term1.deref()->containsSymbol(term2)) return Greater;
	return Incomparable;
      }
    else 
      return compare(term1.deref(),term2.deref());
}; 
inline
bool 
LPO::greater(const Term* complexTerm1,const Term* complexTerm2)
{
  CALL("greater(const Term* complexTerm1,const Term* complexTerm2)");
  if (complexTerm1 == complexTerm2) return false;
  if (complexTerm1->functor() == complexTerm2->functor())
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return false; 
	}
      else
	if (complexTerm1->arity() == 1UL)
	  {
	    return greater(*(complexTerm1->Args()),
			   *(complexTerm2->Args()));
	  };
#endif
      return lexMA(complexTerm1,
		   complexTerm2,
		   complexTerm1->Args(),
		   complexTerm2->Args());
    }
  else
    {
#ifdef LPO_SMALL_ARITY
      if (complexTerm1->arity() == 0UL)
	{
	  return 
	    groundWithAllPrecBelow(complexTerm2,
				   complexTerm1->symbolPrecedence());
	};
      if (complexTerm2->arity() == 0UL)
	{
	  return
	    containsEqOrGrPrec(complexTerm1,
			       complexTerm2->symbolPrecedence());
	};
#endif
      if (complexTerm1->hasGreaterPrecedenceThan(*complexTerm2))
	{
	  return 
	    majo(complexTerm1,complexTerm2->Args());
	}
      else
	{
	  return 
	    alpha(complexTerm1->Args(),complexTerm2);
	};
    };
}; 
inline
bool
LPO::greater(const Term* complexTerm1,const Term& term2)
{
  CALL("greater(const Term* complexTerm1,const Term& term2)");
  if (term2.isVariable())
    {
      return complexTerm1->containsSymbol(term2);
    }
  else
    return greater(complexTerm1,term2.deref());
}; 
inline
bool
LPO::greater(const Term& term1,const Term& term2)
{
  CALL("greater(const Term& term1,const Term& term2)");
  if (term1.isVariable())
    {
      return false;
    }
  else
    return greater(term1.deref(),term2);
}; 
inline 
Comparison LPO::cMA(const Term* complexTerm1,
		    const Term* args2)
{
  CALL("cMA(const Term* complexTerm1,..)#6");
  while (!args2->isNotArg())
    {
      Comparison cmp;
      if (args2->isVariable())
	{
	  if (complexTerm1->containsSymbol(*args2))
	    {
	      cmp = Greater;
	    }
	  else
	    cmp = Incomparable;
	}
      else
	cmp = compare(complexTerm1,args2->deref());
      switch (cmp) 
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
inline 
Comparison LPO::cLMA(const Term* complexTerm1,
		     const Term* complexTerm2,
		     const Term* args1,
		     const Term* args2)
{
  CALL("cLMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,*args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->Next()));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,args2->Next());
	case Incomparable:
	  if (alpha(args1->Next(),complexTerm2))
	    return Greater;
	  if (alpha(args2->Next(),complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return Equal;
}; 
inline 
bool LPO::alpha(const Term* args1,
		const Term* complexTerm2)
{
  CALL("alpha(const Term* args1,..)");
  while (!args1->isNotArg())
    {
      if (!args1->isVariable())
	{
	  Comparison cmp = compare(args1->deref(),complexTerm2);
	  if (cmp == Greater || cmp == Equal) return true;
	};      
      args1 = args1->Next();
    };
  return false;
}; 
inline
bool LPO::lexMA(const Term* complexTerm1,
		const Term* complexTerm2,
		const Term* args1,
		const Term* args2)
{
  CALL("lexMA(const Term* complexTerm1,..)");
  ASSERT(complexTerm1->functor() == complexTerm2->functor());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,*args2))
	{
	case Less:
	  return alpha(args1->Next(),complexTerm2);
	case Equal:
	  break;
	case Greater:
	  return majo(complexTerm1,args2->Next());
	case Incomparable:
	  return alpha(args1->Next(),complexTerm2);
	};
      args1 = args1->Next();
      args2 = args2->Next();
    };
  ASSERT(args2->isNotArg());
  return false;
}; 
inline
bool LPO::majo(const Term* complexTerm1,const Term* args2)
{
  CALL("majo(const Term* complexTerm1,const Term* args2)");
  while (!args2->isNotArg())
    {
      if (!greater(complexTerm1,*args2)) return false;
      args2 = args2->Next();
    };
  return true;
}; 
inline
bool 
LPO::groundWithAllPrecBelow(const Term* complexTerm,
			    long prec)
{
  CALL("groundWithAllPrecBelow(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return false;
  Term::Iterator iter(complexTerm);
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar() ||
	  iter.symbolRef().symbolPrecedence() >= prec) 
	return false;
    };
  return true;
}; 
inline
bool 
LPO::containsEqOrGrPrec(const Term* complexTerm,
			long prec)
{
  CALL("containsEqOrGrPrec(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return true;
  Term::Iterator iter(complexTerm);
  while (iter.Next())
    {
      if (!iter.CurrentSymIsVar() &&
	  iter.symbolRef().symbolPrecedence() >= prec) 
	return true;
    };
  return false;
}; 
inline
Comparison
LPO::compareWithConst(const Term* complexTerm,
		      long constPrec)
{
  CALL("compareWithConst(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != constPrec);
  if (complexTerm->symbolPrecedence() > constPrec) return Greater;
  Term::Iterator iter(complexTerm);
  bool ground = true;
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar())
	{
	  ground = false;
	}
      else
	if (iter.symbolRef().symbolPrecedence() >= constPrec) 
	  return Greater;
    };
  return (ground)? Less : Incomparable;
}; 
inline 
Comparison 
LPO::compare(const Term& term1,
	     const InstRet::Substitution* subst,
	     const Term& term2)
{
  CALL("compare(const Term& term1,..)");
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  return compare(subst->map(term1.var()),subst->map(term2.var()));
	}
      else
	{
	  Comparison cmp =
	    compare(subst,term2.deref(),subst->map(term1.var()));
	  return BK::inverse(cmp);
	};
    }
  else
    if (term2.isVariable())
      {
	return compare(subst,term1.deref(),subst->map(term2.var()));
      }
    else
      return compare(term1.deref(),term2.deref(),subst);
}; 
inline
Comparison 
LPO::compare(const Term& term1,
	     const ForwardMatchingSubstitutionCache* subst,
	     const Flatterm* term2)
{
  CALL("compare(const Term& term1,const ForwardMatchingSubstitutionCache* subst,const Flatterm* term2)");
  if (term1.isVariable())
    {
      return compare(subst->map(term1.var()),term2);
    }
  else
    {
      if (term2->isVariable())
	{
	  ForwardMatchingSubstitutionCache::Instance iter1;
	  iter1.reset(term1.deref(),subst);
	  while (iter1.next())
	    {
	      if (iter1.symbol() == term2->symbolRef())
		return Greater;
	    };
	  return Incomparable;
	}
      else 
	{
	  const Term* complexTerm1 = term1.deref();
	  if (*complexTerm1 == term2->symbolRef())
	    {
#ifdef LPO_SMALL_ARITY
	    if (complexTerm1->arity() == 0UL)
	      {
		return Equal;
	      }
	    else
	      if (complexTerm1->arity() == 1UL)
		{
		  return compare(*(complexTerm1->Args()),
				 subst,
				 term2->next());
		};
#endif
	      return cLMA(complexTerm1,
			  subst,
			  term2,
			  complexTerm1->Args(),
			  term2->next(),
			  term2->after());
	    }
	  else 
	    {
#ifdef LPO_SMALL_ARITY
	    if (complexTerm1->arity() == 0UL)
	      {
		return 
		  BK::inverse(compareWithConst(term2,
					       complexTerm1->symbolPrecedence()));
	      };
	    if (term2->arity() == 0UL)
	      {
		return 
		  compareWithConst(complexTerm1,
				   subst,
				   term2->symbolRef().symbolPrecedence());
	      };
#endif
	      if (complexTerm1->hasGreaterPrecedenceThan(term2->symbolRef()))
		{
		  return cMA(complexTerm1,subst,term2->next(),term2->after());	
		}
	      else
		{
		  return BK::inverse(cMA(term2,complexTerm1->Args(),subst));
		};
	    };
	};
    };
  ICP("END");
  return Incomparable;
}; 
bool
LPO::complexGreaterThanProperSubterms(const Term* complexTerm1,
				      const Term* complexTerm2,
				      const ForwardMatchingSubstitutionCache* subst)
{
  CALL("complexGreaterThanProperSubterms(..)");
  Term term1(const_cast<Term*>(complexTerm1));
  Term::Iterator iter(complexTerm2);
  while (iter.Next())
    {
      if (iter.CurrentSymIsVar())
	{
	  if (compare(term1,
		      subst,
		      subst->map(iter.CurrPos()->var())) != Greater)
	      return false;
	}
      else
	{
	  if (compare(term1,
		      Term(const_cast<Term*>(iter.CurrPos())),
		      subst) != Greater)
	    return false;
	};
    };
  return true;
}; 
Comparison LPO::cMA(const Term* complexTerm1,
		    const ForwardMatchingSubstitutionCache* subst,
		    const Flatterm* args2,
		    const Flatterm* endOfArgs2)
{
  CALL("cMA(const Term* complexTerm1,..)#7");
  while (args2 != endOfArgs2)
    {
      Comparison cmp;
      if (args2->isVariable())
	{
          if (complexContainsVar(complexTerm1,subst,args2->symbolRef()))
	    {
	      cmp = Greater;
	    }
	  else
	    cmp = Incomparable;
	}
      else
	{
	  Term term1(const_cast<Term*>(complexTerm1));
	  cmp = compare(term1,subst,args2);
	};
      switch (cmp) 
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->after(),endOfArgs2,complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->after();
    };
  return Greater;
}; 
Comparison LPO::cMA(const Flatterm* complexTerm1,
		    const Term* args2,
		    const ForwardMatchingSubstitutionCache* subst)
{
  CALL("cMA(const Flatterm* complexTerm1,..)");
  while (!args2->isNotArg())
    {
      switch (BK::inverse(compare(*args2,subst,complexTerm1)))
	{
	case Less:
	case Equal:
	  return Less;
	case Greater:
	  break; 
	case Incomparable:
	  if (alpha(args2->Next(),subst,complexTerm1))
	    return Less;
	  return Incomparable;
	};
      args2 = args2->Next();
    };
  return Greater;
}; 
Comparison LPO::cLMA(const Term* complexTerm1,
		     const ForwardMatchingSubstitutionCache* subst,
		     const Flatterm* complexTerm2,
		     const Term* args1,
		     const Flatterm* args2,
		     const Flatterm* endOfArgs2)
{
  CALL("cLMA(const Term* complexTerm1,..)");
  ASSERT(*complexTerm1 == complexTerm2->symbolRef());
  while (!args1->isNotArg())
    {
      switch (compare(*args1,subst,args2))
	{
	case Less:
	  return BK::inverse(cMA(complexTerm2,args1->Next(),subst));
	case Equal:
	  break;
	case Greater:
	  return cMA(complexTerm1,subst,args2->after(),endOfArgs2);
	case Incomparable:
	  if (alpha(args1->Next(),subst,complexTerm2))
	    return Greater;
	  if (alpha(args2->after(),endOfArgs2,complexTerm1,subst))
	    return Less;
	  return Incomparable;
	};
      args1 = args1->Next();
      args2 = args2->after();
    };
  ASSERT(args2 == endOfArgs2);
  return Equal;
}; 
inline
bool 
LPO::groundWithAllPrecBelow(const Term* complexTerm,
			    const ForwardMatchingSubstitutionCache* subst,
			    long prec)
{
  CALL("groundWithAllPrecBelow(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return false;
  ForwardMatchingSubstitutionCache::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (iter.symbol().isVariable() ||
	  iter.symbol().symbolPrecedence() >= prec) 
	return false;
    };
  return true;
}; 
inline
bool 
LPO::containsEqOrGrPrec(const Term* complexTerm,
			const ForwardMatchingSubstitutionCache* subst,
			long prec)
{
  CALL("containsEqOrGrPrec(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != prec);
  if (complexTerm->symbolPrecedence() > prec) return true;
  ForwardMatchingSubstitutionCache::Instance iter;
  iter.reset(complexTerm,subst);
  while (iter.next())
    {
      if (!iter.symbol().isVariable() &&
	  iter.symbol().symbolPrecedence() >= prec) 
	return true;
    };
  return false;
}; 
inline
Comparison
LPO::compareWithConst(const Term* complexTerm,
		      const ForwardMatchingSubstitutionCache* subst,
		      long constPrec)
{
  CALL("compareWithConst(const Term* complexTerm,..)");
  ASSERT(complexTerm->isComplex());
  ASSERT(complexTerm->symbolPrecedence() != constPrec);
  if (complexTerm->symbolPrecedence() > constPrec) return Greater;
  ForwardMatchingSubstitutionCache::Instance iter;
  iter.reset(complexTerm,subst);
  bool ground = true;
  while (iter.next())
    {
      if (iter.symbol().isVariable())
	{
	  ground = false;
	}
      else
	if (iter.symbol().symbolPrecedence() >= constPrec) 
	  return Greater;
    };
  return (ground)? Less : Incomparable;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
LPO::SpecialisedComparison::~SpecialisedComparison()
{ 
  CALL("destructor ~SpecialisedComparison()");
  if (DestructionMode::isThorough())
    {  
      if (_constraintGreater)
	{
	  _constraintGreater->decReferenceCounter();
	  if (!_constraintGreater->referenceCounter())
	    delete _constraintGreater;
	};
      if (_constraintGreaterOrEqual)
	{
	  _constraintGreaterOrEqual->decReferenceCounter();
	  if (!_constraintGreaterOrEqual->referenceCounter())
	    delete _constraintGreaterOrEqual;
	};
      if (_constraintLess)
	{
	  _constraintLess->decReferenceCounter();
	  if (!_constraintLess->referenceCounter())
	    delete _constraintLess;
	};  
      if (_constraintLessOrEqual)
	{
	  _constraintLessOrEqual->decReferenceCounter();
	  if (!_constraintLessOrEqual->referenceCounter())
	    delete _constraintLessOrEqual;
	};
      if (_constraintNotEqual)
	{
	  _constraintNotEqual->decReferenceCounter();
	  if (!_constraintNotEqual->referenceCounter())
	    delete _constraintNotEqual;
	};
    };
}; 
void LPO::SpecialisedComparison::load(const Term& lhs,const Term& rhs)
{
  CALL("load(const Term& lhs,const Term& rhs)");
  _lhs = lhs;
  _rhs = rhs;
  lhs.collectVariables(_lhsVars);
  rhs.collectVariables(_rhsVars);
  _specialised = false;
  if (_constraintGreater)
    {
      _constraintGreater->decReferenceCounter();
      if (!_constraintGreater->referenceCounter())
	delete _constraintGreater;
      _constraintGreater = 0;
    };
  if (_constraintGreaterOrEqual)
    {
      _constraintGreaterOrEqual->decReferenceCounter();
      if (!_constraintGreaterOrEqual->referenceCounter())
	delete _constraintGreaterOrEqual;
      _constraintGreaterOrEqual = 0;
    };
  if (_constraintLess)
    {
      _constraintLess->decReferenceCounter();
      if (!_constraintLess->referenceCounter())
	delete _constraintLess;
      _constraintLess = 0;
    };
  if (_constraintLessOrEqual)
    {
      _constraintLessOrEqual->decReferenceCounter();
      if (!_constraintLessOrEqual->referenceCounter())
	delete _constraintLessOrEqual;
      _constraintLessOrEqual = 0;
    };
  if (_constraintNotEqual)
    {
      _constraintNotEqual->decReferenceCounter();
      if (!_constraintNotEqual->referenceCounter())
	delete _constraintNotEqual;
      _constraintNotEqual = 0;
    };
}; 
bool LPO::SpecialisedComparison::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return _rhsVars.isSubsetOf(_lhsVars);
}; 
bool LPO::SpecialisedComparison::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return _lhsVars.isSubsetOf(_rhsVars);
}; 
void LPO::SpecialisedComparison::specialiseGreater()
{
  specialise();
};
void LPO::SpecialisedComparison::specialiseGreaterOrEqual()
{
  specialise();
};
void LPO::SpecialisedComparison::specialiseLess()
{
  specialise();
};
void LPO::SpecialisedComparison::specialiseLessOrEqual()
{
  specialise();
};
void LPO::SpecialisedComparison::specialiseNotEqual()
{
  specialise();
};
inline
LPO::StoredConstraint* LPO::SpecialisedComparison::storedConstraintGreater()
{
  CALL("storedConstraintGreater()");
  if (!_constraintGreater) 
    {
      specialise();
      _constraintGreater = 
	new StoredConstraint(StoredConstraint::Greater,
			     _lhs,
			     _rhs);
      _constraintGreater->incReferenceCounter();      
    };
  return _constraintGreater;
}; 
inline
LPO::StoredConstraint* LPO::SpecialisedComparison::storedConstraintGreaterOrEqual()
{
  CALL("storedConstraintGreaterOrEqual()");
  if (!_constraintGreaterOrEqual) 
    {
      specialise();
      _constraintGreaterOrEqual = 
	new StoredConstraint(StoredConstraint::GreaterOrEqual,
			     _lhs,
			     _rhs);
      _constraintGreaterOrEqual->incReferenceCounter();      
    };
  return _constraintGreaterOrEqual;
}; 
inline
LPO::StoredConstraint* 
LPO::SpecialisedComparison::storedConstraintLess()
{
  CALL("storedConstraintLess()");
  if (!_constraintLess) 
    {
      specialise();
      _constraintLess = 
	new StoredConstraint(StoredConstraint::Less,
			     _lhs,
			     _rhs);
      _constraintLess->incReferenceCounter();      
    };
  return _constraintLess;
}; 
inline
LPO::StoredConstraint* 
LPO::SpecialisedComparison::storedConstraintLessOrEqual()
{
  CALL("storedConstraintLessOrEqual()");  
  if (!_constraintLessOrEqual) 
    {
      specialise();
      _constraintLessOrEqual = 
	new StoredConstraint(StoredConstraint::LessOrEqual,
			     _lhs,
			     _rhs);
      _constraintLessOrEqual->incReferenceCounter();      
    };
  return _constraintLessOrEqual;
}; 
inline
LPO::StoredConstraint* 
LPO::SpecialisedComparison::storedConstraintNotEqual()
{
  CALL("storedConstraintNotEqual()");  
  if (!_constraintNotEqual) 
    {
      specialise();
      _constraintNotEqual = 
	new StoredConstraint(StoredConstraint::NotEqual,
			     _lhs,
			     _rhs);
      _constraintNotEqual->incReferenceCounter();      
    };
  return _constraintNotEqual;
}; 
#ifndef NO_DEBUG_VIS
ostream& LPO::SpecialisedComparison::output(ostream& str) const
{
  str << _lhs << " VS " << _rhs << "\n";
  if (_specialised)
    {
      str << "CATEGORY: " << category() << "\n";
    }
  else
    {
      str << " NOT SPECIALISED\n";
    };
  return str;
};
#endif
inline
void LPO::SpecialisedComparison::specialise()
{
  CALL("specialise()");
  if (_specialised) return;
  switch (_ordering->compare(_lhs,_rhs)) {
  case Less:
    _category = 
      SimplificationOrdering::SpecialisedComparison::AlwaysLess;
    break;
  case Equal:
    _category = 
      SimplificationOrdering::SpecialisedComparison::AlwaysEqual;
    break;
  case Greater:
    _category = 
      SimplificationOrdering::SpecialisedComparison::AlwaysGreater;
    break;
  case Incomparable:
    _category = 
      SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile;
    break;
  };
  _specialised = true;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#  define DEBUG_NAMESPACE "LPO::ConstraintCheckOnUnifier"
#endif
#include "debugMacros.hpp"
LPO::ConstraintCheckOnUnifier::~ConstraintCheckOnUnifier()
{
};
bool 
LPO::ConstraintCheckOnUnifier::holds(const SimplificationOrdering::StoredConstraint* constr,
					  ulong variableIndex)
{
  CALL("holds(..)");
  ASSERT(constr);
  const StoredConstraint* nativeConstr = 
    static_cast<const StoredConstraint*>(static_cast<const void*>(constr));
  _currentVariableIndex = variableIndex;
  bool result;
#if (defined DEBUG_NAMESPACE) || (defined _SUPPRESS_WARNINGS_)
  result = false;
#endif
  DOP(Comparison debugCmp = \
      _ordering->compareOnCurrentUnifier((nativeConstr)->lhs(),\
					 (nativeConstr)->rhs(),\
					 variableIndex));
  switch ((nativeConstr)->targetRelation())
    {
    case StoredConstraint::Greater:
      result = holdsGreater(nativeConstr);
      ASSERT(result == (debugCmp == Greater));
      break;
    case StoredConstraint::GreaterOrEqual:
      result = holdsGreaterOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Greater) || (debugCmp == Equal)));
      break;
    case StoredConstraint::Less:
      result = holdsLess(nativeConstr);
      ASSERT(result == (debugCmp == Less));
      break;
    case StoredConstraint::LessOrEqual:
      result = holdsLessOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Less) || (debugCmp == Equal)));
      break;
    case StoredConstraint::NotEqual:
      result = holdsNotEqual(nativeConstr);
      ASSERT(debugCmp != Equal || !result);
      break;
    };
  return result;
}; 
inline
bool LPO::ConstraintCheckOnUnifier::holdsGreater(const LPO::StoredConstraint* constr)
{
  CALL("holdsGreater(..)");
  return 
    _ordering->compareOnCurrentUnifier(constr->lhs(),
				       constr->rhs(),
				       _currentVariableIndex) == 
    Greater;
}; 
inline
bool LPO::ConstraintCheckOnUnifier::holdsGreaterOrEqual(const LPO::StoredConstraint* constr)
{
  CALL("holdsGreaterOrEqual(..)");
  Comparison cmp = 
    _ordering->compareOnCurrentUnifier(constr->lhs(),
				       constr->rhs(),
				       _currentVariableIndex);
  return cmp == Greater || cmp == Equal;
}; 
inline
bool LPO::ConstraintCheckOnUnifier::holdsLess(const LPO::StoredConstraint* constr)
{
  CALL("holdsLess(..)");
  Comparison cmp = 
    _ordering->compareOnCurrentUnifier(constr->lhs(),
				       constr->rhs(),
				       _currentVariableIndex);
  return cmp == Less;
}; 
inline
bool LPO::ConstraintCheckOnUnifier::holdsLessOrEqual(const LPO::StoredConstraint* constr)
{
  CALL("holdsLessOrEqual(..)");
  Comparison cmp = 
    _ordering->compareOnCurrentUnifier(constr->lhs(),
				       constr->rhs(),
				       _currentVariableIndex);
  return cmp == Less || cmp == Equal;
}; 
inline
bool LPO::ConstraintCheckOnUnifier::holdsNotEqual(const LPO::StoredConstraint* constr)
{
  CALL("holdsNotEqual(..)");
  return !InstanceWithUnifier::equal(_currentVariableIndex,
				  constr->lhs(),
				  _currentVariableIndex,
				  constr->rhs());
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#  define DEBUG_NAMESPACE "LPO::ConstraintCheckOnForwardMatchingSubst"
#endif
#include "debugMacros.hpp"
LPO::ConstraintCheckOnForwardMatchingSubst::~ConstraintCheckOnForwardMatchingSubst()
{
};
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holds(const SimplificationOrdering::StoredConstraint* constr,
								 const ForwardMatchingSubstitutionCache* subst)
{
  CALL("holds(..)");
  ASSERT(constr);
  const StoredConstraint* nativeConstr = 
    static_cast<const StoredConstraint*>(static_cast<const void*>(constr));
  bool result;
  DOP(Comparison debugCmp = \
      _ordering->compare(nativeConstr->lhs(),\
			 nativeConstr->rhs(),\
			 subst));
  _subst = subst;
  switch (nativeConstr->targetRelation())
    {
    case StoredConstraint::Greater:
      result = holdsGreater(nativeConstr);
      ASSERT(result == (debugCmp == Greater));
      return result;
    case StoredConstraint::GreaterOrEqual:
      result = holdsGreaterOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Greater) || (debugCmp == Equal)));
      return result;
    case StoredConstraint::Less:
      result = holdsLess(nativeConstr);
      ASSERT(result == (debugCmp == Less));
      return result;
    case StoredConstraint::LessOrEqual:
      result = holdsLessOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Less) || (debugCmp == Equal)));
      return result;
    case StoredConstraint::NotEqual:
      result = holdsNotEqual(nativeConstr);
      ASSERT(debugCmp != Equal || !result);
      return result;
    };
  ICP("ICP0");
  return false;
}; 
inline
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holdsGreater(const StoredConstraint* constr)
{
  CALL("holdsGreater(..)");
  Comparison cmp = 
    _ordering->compare(constr->lhs(),
		       constr->rhs(),
		       _subst);
  return cmp == Greater;
}; 
inline
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holdsGreaterOrEqual(const StoredConstraint* constr)
{
  CALL("holdsGreaterOrEqual(..)");
  Comparison cmp = 
    _ordering->compare(constr->lhs(),
		       constr->rhs(),
		       _subst);
  return cmp == Greater || cmp == Equal;
}; 
inline
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holdsLess(const StoredConstraint* constr)
{
  CALL("holdsLess(..)");
  Comparison cmp = 
    _ordering->compare(constr->lhs(),
		       constr->rhs(),
		       _subst);
  return cmp == Less;
}; 
inline
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holdsLessOrEqual(const StoredConstraint* constr)
{
  CALL("holdsLessOrEqual(..)");
  Comparison cmp = 
    _ordering->compare(constr->lhs(),
		       constr->rhs(),
		       _subst);
  return cmp == Less || cmp == Equal;
}; 
inline
bool 
LPO::ConstraintCheckOnForwardMatchingSubst::holdsNotEqual(const StoredConstraint* constr)
{
  CALL("holdsNotEqual(..)");
  return 
    !ForwardMatchingSubstitutionCache::Instance::equal(constr->lhs(),
						       constr->rhs(),
						       _subst);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::CompiledConstraintForUnifiers"
#endif
#include "debugMacros.hpp"
LPO::CompiledConstraintForUnifiers::~CompiledConstraintForUnifiers() 
{
};  
void LPO::CompiledConstraintForUnifiers::load(UnifTerm* lhs,UnifTerm* rhs)
{
  CALL("load(UnifTerm* lhs,UnifTerm* rhs)");
  _lhs = lhs;
  _rhs = rhs;
  _lhsVars.reset();
  if (_lhs->symbol().isVariable())
    {
      _lhsVars.add(_lhs->symbol().var()->relativeNum());
    }
  else
    {
      _lhsIter.Reset(_lhs);
      _lhsIter.SkipSym();
      while (_lhsIter.NotEmpty())
	{
	  const Unifier::Variable* var = _lhsIter.CurrVar();
	  if (var) _lhsVars.add(var->relativeNum());
	  _lhsIter.SkipSym();
	};
    };
  _rhsVars.reset();
  if (_rhs->symbol().isVariable())
    {
      _rhsVars.add(_rhs->symbol().var()->relativeNum());
    }
  else
    {
      _rhsIter.Reset(_rhs);
      _rhsIter.SkipSym();
      while (_rhsIter.NotEmpty())
	{
	  const Unifier::Variable* var = _rhsIter.CurrVar();
	  if (var) _rhsVars.add(var->relativeNum());
	  _rhsIter.SkipSym();
	};
    };
}; 
bool LPO::CompiledConstraintForUnifiers::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return _rhsVars.isSubsetOf(_lhsVars);
}; 
bool LPO::CompiledConstraintForUnifiers::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return _lhsVars.isSubsetOf(_rhsVars);
}; 
void LPO::CompiledConstraintForUnifiers::compileGreaterOrEqual()
{
  CALL("compileGreaterOrEqual()");
  ++(_ordering->_numberOfCompiledConstraintsOnUnifiers);
  _targetRelation = GreaterOrEqual;
}; 
void LPO::CompiledConstraintForUnifiers::compileGreater()
{
  CALL("compileGreater()");
  ++(_ordering->_numberOfCompiledConstraintsOnUnifiers);
  _targetRelation = Greater;
}; 
bool LPO::CompiledConstraintForUnifiers::holdsOnCurrentUnifier()
{
  CALL("holdsOnCurrentUnifier()");
  ++(_ordering->_numberOfConstraintChecksOnUnifiers);
  if (_targetRelation == GreaterOrEqual)
    {
      return _ordering->greaterOrEqualOnCurrentUnifier(_lhs,_rhs);
    }
  else 
    {
      ASSERT(_targetRelation == Greater);
      return _ordering->greaterOnCurrentUnifier(_lhs,_rhs);
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& LPO::CompiledConstraintForUnifiers::output(ostream& str) const
{
  UnifTerm::outputTerm(str << "CONSTRAINT: ",_lhs);
  switch (_targetRelation)
    {
    case GreaterOrEqual: str << " >= "; break;
    case Greater: str << " > "; break;
    default: str << "???";
    };
  UnifTerm::outputTerm(str,_rhs) << "\n";
  return str;
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LPO
#  define DEBUG_NAMESPACE "LPO::CompiledConstraintForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
LPO::CompiledConstraintForInstanceRetrieval::~CompiledConstraintForInstanceRetrieval()
{
};
void 
LPO::CompiledConstraintForInstanceRetrieval::loadGreater(const Term* lhs,
							 const Term* rhs)
{
  CALL("loadGreater(const Term* lhs,const Term* rhs)");
  ASSERT(lhs->isComplex());
  ASSERT(rhs->isComplex());
  _targetRelation = Greater;
  _lhs.makeReference(const_cast<Term*>(lhs));
  _rhs.makeReference(const_cast<Term*>(rhs));
  _lhs.collectVariables(_lhsVars);
  _rhs.collectVariables(_rhsVars);
  switch (_ordering->compare(lhs,rhs))
    {
    case BK::Less: 
    case BK::Equal:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Contradiction);
      break;
    case BK::Greater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology);
      break;
    case BK::Incomparable:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile);
      break;
    };
}; 
void 
LPO::CompiledConstraintForInstanceRetrieval::loadNotEqual(const Term* lhs,const Term* rhs)
{
  CALL("loadNotEqual(const Term* lhs,const Term* rhs)");
  ASSERT(lhs->isComplex());
  ASSERT(rhs->isComplex());
  _targetRelation = NotEqual;
  _lhs.makeReference(const_cast<Term*>(lhs));
  _rhs.makeReference(const_cast<Term*>(rhs));
  _lhs.collectVariables(_lhsVars);
  _rhs.collectVariables(_rhsVars);
  switch (_ordering->compare(lhs,rhs))
    {
    case BK::Less:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology);
      break;
    case BK::Equal:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Contradiction);
      break;
    case BK::Greater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology);
      break;
    case BK::Incomparable:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile);
      break;
    };
}; 
bool LPO::CompiledConstraintForInstanceRetrieval::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return _rhsVars.isSubsetOf(_lhsVars);
}; 
bool LPO::CompiledConstraintForInstanceRetrieval::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return _lhsVars.isSubsetOf(_rhsVars);
}; 
void LPO::CompiledConstraintForInstanceRetrieval::compile(const InstRet::Substitution* subst)
{
  CALL("compile(const InstRet::Substitution* subst)");
  ASSERT(_targetRelation == Greater || 
	 _targetRelation == NotEqual);
  ++(_ordering->_numberOfCompiledConstraintsForInstanceRetrieval);
  _subst = subst;
}; 
bool LPO::CompiledConstraintForInstanceRetrieval::holds()
{
  CALL("holds()");
  ++(_ordering->_numberOfConstraintChecksForInstanceRetrieval);
  ASSERT(_targetRelation == Greater || 
	 _targetRelation == NotEqual);
  if (_targetRelation == NotEqual)
    {
      return
	!InstRet::Substitution::Instance::equal(_lhs,_rhs,_subst);
    }
  else
    {
      ASSERT(_targetRelation == Greater);
      return 
	_ordering->complexGreater(_lhs.deref(),_subst,_rhs.deref());
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& LPO::CompiledConstraintForInstanceRetrieval::output(ostream& str) const
{
  str << "CONSTRAINT: " << _lhs;
  switch (_targetRelation)
    {
    case Greater: str << " > "; break;
    default: str << "???";
    };
  str << _rhs << "\n";
  return str;
};
#endif
