/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "LocalUnification.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "LocalUnification"
#endif
#include "debugMacros.hpp"
using namespace VK;
LocalUnification::LocalUnification() 
{
  for (ulong v = 0; v < VampireKernelConst::MaxNumOfVariables; ++v)
    _unifier[v] = 0;
};
LocalUnification::~LocalUnification() {};
void LocalUnification::reset()
{
  CALL("reset()");
  while (_instantiated)
    *(_instantiated.pop()) = 0;
}; 
bool LocalUnification::unify(const Term* complexTerm1,const Term* complexTerm2)
{
  CALL("unify(const Term* complexTerm1,const Term* complexTerm2)");
  const UnifierState* initialUnifierState = getUnifierState();
  if (unify1(complexTerm1,complexTerm2))
    return true;
  rollUnifierBackTo(initialUnifierState);
  return false;
}; 
bool LocalUnification::unify(const Term& term1,const Term& term2)
{
  CALL("unify(const Term& term1,const Term& term2)");
  const UnifierState* initialUnifierState = getUnifierState();
  if (unify1(term1,term2))
    return true;
  rollUnifierBackTo(initialUnifierState);
  return false;
}; 
void LocalUnification::rollUnifierBackTo(const UnifierState* notch)
{
  CALL("rollUnifierBackTo(const UnifierState* notch)");
  while (_instantiated.end() != static_cast<const void*>(notch))
    {
      *(_instantiated.pop()) = 0;
    };
}; 
inline
bool LocalUnification::unify1(const Term* complexTerm1,const Term* complexTerm2)
{
  CALL("unify1(const Term* complexTerm1,const Term* complexTerm2)");
  ASSERT(complexTerm1->isComplex());
  ASSERT(complexTerm2->isComplex());
  if (*complexTerm1 != *complexTerm2) 
    {
      return false;
    };
  const Term* arg1 = complexTerm1->Args();
  const Term* arg2 = complexTerm2->Args();
  while (!arg1->isNotArg())
    {
      if (!unify1(*arg1,*arg2)) 
	return false;
      arg1 = arg1->Next();
      arg2 = arg2->Next();
    };
  return true;
}; 
inline
bool LocalUnification::unify1(const Term& term1,const Term& term2)
{
  CALL("unify1(const Term& term1,const Term& term2)");
  if (term1 == term2) return true;  
  if (term1.isVariable())
    {
      const Term* instance = _unifier[term1.var()];
      if (instance)
	{
	  return unify1(*instance,term2);
	}
      else 
 	{
	  if (term2.isVariable())
	    {
	      instance = _unifier[term2.var()];
	      if (instance)
		{
		  return unify1(term1,*instance);
		}
	      else 
		{		  
		  instantiate(term1.var(),&term2);
		  return true;
		};
	    }
	  else
	    {
	      if (occursIn(term1,term2)) 
		{
		  return false;
		};
	      instantiate(term1.var(),&term2);
	      return true;
	    };
	};
    }
  else
    if (term2.isVariable())
      {
	const Term* instance = _unifier[term2.var()];
	if (instance)
	  {	   
	    return unify1(term1,*instance);
	  }
	else 
	  {
	    if (occursIn(term2,term1)) 
	      {
		return false;
	      };
	    instantiate(term2.var(),&term1);
	    return true;
	  };
      }
    else
      return unify1(term1.deref(),term2.deref()); 
}; 
inline
bool LocalUnification::occursIn(const Term& var,const Term& term) const
{
  CALL("occursIn(const Term& var,const Term& term) const");
  if (var == term) return true;
  if (term.isVariable())
    {
      const Term* instance = _unifier[term.var()];
      if (instance)
	{
	  return occursIn(var,*instance);
	}
      else
	return false;
    }
  else
    {
      ASSERT(term.IsReference());
      for (const Term* arg = term.deref()->Args(); !arg->isNotArg(); arg = arg->Next())
	if (occursIn(var,*arg)) return true;
      return false;
    };
}; 
inline
void LocalUnification::instantiate(ulong varNum,const Term* instance)
{
  CALL("instantiate(ulong varNum,const Term* instance)");
  ASSERT(!_unifier[varNum]);
  _unifier[varNum] = instance;
  _instantiated.push(_unifier + varNum);
}; 
#ifndef NO_DEBUG_VIS
ostream& LocalUnification::output(ostream& str) const
{
  CALL("output(ostream& str) const");
  str << "{";
  for (const Term** const * v = _instantiated.begin();
       v != _instantiated.end();
       ++v)
    {
      str << "X" << ((*v) - _unifier) << " --> "
	  << ***v;
      if (v + 1 != _instantiated.end())
	str << " | ";
    };
  return str << "}";
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "LocalUnification::TermTraversal"
#endif
#include "debugMacros.hpp"
LocalUnification::TermTraversal::TermTraversal() 
{
  _dummyTerm[0] = Term::dummyPositiveHeader1();  
};
LocalUnification::TermTraversal::TermTraversal(const Term& term,const LocalUnification* unifier) 
{
  CALL("constructor TermTraversal(const Term& term,const LocalUnification* unifier)");
  _dummyTerm[0] = Term::dummyPositiveHeader1();  
  _unifier = unifier->_unifier;
  _dummyTerm[1] = term;
  _term = _dummyTerm + 1;
  _after = _dummyTerm;
  canonizeVariableInstance();
};
LocalUnification::TermTraversal::~TermTraversal() {};
void LocalUnification::TermTraversal::reset(const Term& term,const LocalUnification* unifier)
{
  CALL("reset(const Term& term)");
  _unifier = unifier->_unifier;
  _dummyTerm[1] = term;
  _term = _dummyTerm + 1;
  _after = _dummyTerm;
  _backtrackPoints.reset();
  canonizeVariableInstance();
}; 
