/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PROBLEM_PROFILE_HPP
#define PROBLEM_PROFILE_HPP
#include <iostream>      
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "FIFO.hpp"
#include "Clause.hpp"
#include "Term.hpp"
#include "SymbolInfo.hpp"
#include "ClauseSet.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROBLEM_PROFILE
#  define DEBUG_NAMESPACE "ProblemProfile"
#endif
#include "debugMacros.hpp"
namespace VK
{
class ProblemProfile
{
 public:
  class PredInfo
    {
    public:
      PredInfo() 
	{
	  CALL("constructor PredInfo()");	
	  _registered = 0UL;
	  _symmetric = false;
	  _transitive = false;
	};
      ~PredInfo() {};
      void reset()
      {
	CALL("reset()");	  
	_registered = 0UL;
	_symmetric = false;
	_transitive = false;
      }; 
      ostream& output(ostream& str) const 
	{
	  str << _symbol << "/" << _symbol.arity();
	  if (_symmetric) str << " sym";
	  if (_transitive) str << " tra";
	  return str;
	};  
    private:
      void regPred(const Term& header) 
	{
	  _symbol = header;
	  _symbol.makePositive();
	  _registered++;
	};  
    private:
      Term _symbol;
      ulong _registered;
      bool _symmetric;  
      bool _transitive;
      friend class ProblemProfile;
    }; 
  class FuncInfo
    {
    public:
      FuncInfo() 
	{
	  CALL("constructor FuncInfo()");
	  _registered = 0UL;
	  _commutative = false;
	};
      ~FuncInfo() {};
      void reset()
      {
	CALL("reset()");
	_registered = 0UL;
	_commutative = false;
      };
      const Term& symbol() const { return _symbol; };
      ulong registered() const { return _registered; };
      bool commutative() const { return _commutative; };
      ostream& output(ostream& str) const 
	{
	  str << _symbol << "/" << _symbol.arity();
	  if (_commutative) str << " : COM";
	  return str;
	}; 
    private:
      Term _symbol;
      ulong _registered;
      bool _commutative;
      friend class ProblemProfile; 
    }; 
 public:
  ProblemProfile();
  ~ProblemProfile();
  void reset();
  void load(const Clause* cl);
  void loadBackground(const Clause* cl);
  void load(const BK::FIFO<BK::GlobAlloc,Clause*>& clauses)
    {
      CALL("load(const BK::FIFO<BK::GlobAlloc,Clause*>& clauses)");
      for (const BK::FIFO<BK::GlobAlloc,Clause*>::Entry* qe = clauses.begin(); 
	   qe; 
	   qe = qe->next())
	{
	  this->load(qe->value());
	};
    };                 
  void loadBackground(const BK::FIFO<BK::GlobAlloc,Clause*>& clauses)
    {
      CALL("load(const BK::FIFO<BK::GlobAlloc,Clause*>& clauses)");
      for (const BK::FIFO<BK::GlobAlloc,Clause*>::Entry* qe = clauses.begin(); qe; qe = qe->next())
	{
	  this->loadBackground(qe->value());
	};
    };                     
  void loadBackground(const ClauseSet& clauses) {
    CALL("load(const ClauseSet& clauses)");
    ClauseSet::ConstIter iter(clauses);
    while (!iter.isEnd())
      {
	loadBackground(iter.clause());
	iter.next();
      };
  };
  void summarise();                        
  const SymbolInfo* commutativityTable() const 
    {
      if (_symbolInfoTable.numOfCommutative()) 
	return &_symbolInfoTable;
      return 0;
    };
  const SymbolInfo* symmetryTable() const 
    {
      if (_symbolInfoTable.numOfSymmetric()) 
	return &_symbolInfoTable;
      return 0;   
    };
  const SymbolInfo* transitivityTable() const 
  {
    if (_symbolInfoTable.numOfTransitive()) 
      return &_symbolInfoTable;
    return 0;   
  };
  ulong numOfEqLiterals() const { return _numOfEqLiterals; }; 
  ulong numOfCommutative() const { return _symbolInfoTable.numOfCommutative(); };
  ulong numOfTransitive() const { return _symbolInfoTable.numOfTransitive(); };
  ostream& output(ostream& str) const;
 private: 
  PredInfo& predInfo(ulong pred) 
    {
      CALL("predInfo(ulong pred)");
      DOP(_predInfoTable.unfreeze());
      PredInfo* res = &(_predInfoTable.sub(pred));
      DOP(_predInfoTable.freeze());
      return *res;
    }; 
  FuncInfo& funcInfo(ulong func) 
    {
      CALL("funcInfo(ulong func)");
      DOP(_funcInfoTable.unfreeze());
      FuncInfo* res = &(_funcInfoTable.sub(func));
      DOP(_funcInfoTable.freeze());
      return *res;
    }; 
 private:
  BK::FIFO<BK::GlobAlloc,const Clause*> _clauses;
  ulong _numOfEqLiterals;
  ulong _numOfPred;
  ulong _numOfFuncs;
  BK::Array<BK::GlobAlloc,PredInfo,BK::FastRelocation<PredInfo>,128UL> _predInfoTable;
  BK::Array<BK::GlobAlloc,FuncInfo,BK::FastRelocation<FuncInfo>,128UL> _funcInfoTable;
  SymbolInfo _symbolInfoTable;
}; 
}; 
#endif
