/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Term.hpp"
#include "Signature.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
#define DEBUG_NAMESPACE "Term"
#endif
#include "debugMacros.hpp"
bool Term::isPXYZ() const 
{
  CALL("isPXYZ()");
  ASSERT(isComplex());
  for (const Term* ar1 = Args(); !ar1->isNotArg(); ar1 = ar1->Next())
    {
      if (!ar1->isVariable()) return false; 
      for (const Term* ar2 = Args(); ar2 != ar1; ar2 = ar2->Next())     
	if (*ar2 == *ar1) return false;
    };
  return true;
}; 
bool Term::Symmetry(const Term* t1,const Term* t2)
{
  CALL("Symmetry(const Term* t1,const Term* t2)");
  ASSERT(t1->isComplex() && t2->isComplex());
  if (t1->functor() != t2->functor()) return false;    
  if (t1->arity() != 2) return false;
  if ((!t1->isPXYZ()) || (!t2->isPXYZ())) return false;
  if ((*(t1->Arg1()) != *(t2->Arg2()))
      || (*(t1->Arg2()) != *(t2->Arg1()))) return false; 
  return true;
}; 
void Term::destroyUnsharedComplex()
{
  CALL("destroyUnsharedComplex()");
  ASSERT(isComplex());
  for (Term* argRef = Args(); !argRef->isNotArg(); argRef = argRef->Next())
    if (argRef->IsReference()) 
      argRef->deref()->destroyUnsharedComplex();
  delete this;
}; 
void Term::destroyUnsharedLit()
{
  CALL("destroyUnsharedLit()");
  ASSERT(isComplex());      
  for (Term* argRef = Args(); !argRef->isNotArg(); argRef = argRef->Next())
    {
      if (argRef->IsReference()) argRef->deref()->destroyUnsharedComplex();
    };
  delete this;
}; 
void Term::DestroyOccurences() const
{
  CALL("DestroyOccurences() const");               
  ASSERT(isComplex());
  OccList* occList = Occurences();
  if (occList) 
    {
      for (OccList::Node* n = occList->first();
	   n;
	   n = n->next())
	{
	  delete n->value();
	};
      delete occList;
      InitOccurences();
    };
}; 
void Term::RegisterOccurence(Clause* cl,ulong litNum) const
{
  CALL("RegisterOccurence(Clause* cl,ulong litNum) const");
  ASSERT(isComplex());
  OccList* occList = Occurences();
  if (!occList) 
    { 
      occList = new OccList(); 
      associatedInfo()->setOccurences(occList);
    };
  bool newNode;
  OccList::Node* node = occList->insert(cl,newNode);
  if (newNode) node->value() = new LitNumMask();
  if (!node->value()->write(litNum))
    {
      LitNumMask* newLitNumMask = node->value()->add(litNum);
      delete node->value();
      node->value() = newLitNumMask;
    };
}; 
void Term::RemoveOccurences(Clause* cl) const
{
  CALL("RemoveOccurences(Clause* cl) const");
  ASSERT(isComplex());
  OccList* occList = Occurences();
  if (!occList) return;
  LitNumMask* litNumMask;
  if (occList->remove(cl,litNumMask))
    {
      ASSERT(litNumMask);
      delete litNumMask;
    };
  ASSERT(!occList->contains(cl));
  if (!(*occList)) { delete occList; InitOccurences(); };
}; 
void Term::markBasicArgumentsInSignature() const
{
  CALL("markBasicArgumentsInSignature() const");
  ASSERT(isComplex());
  ulong argNum = 0UL;
  for (const Term* arg = Args(); !arg->isNotArg(); arg = arg->Next())
    {
      if (!arg->isVariable())
	{
	  Signature::current()->markBasicArgument(Positive(),functor(),argNum);
	  arg->deref()->markBasicArgumentsInSignature();
	};
      ++argNum;
    };
}; 
bool Term::containsComplexSubterm(const Term* t) const
{
  CALL("containsComplexSubterm(const Term* t) const");
  ASSERT(isComplex() && t->isComplex());
  if (this == t) return true;
  for (const Term* arg = Args(); !arg->isNotArg(); arg = arg->Next())
    if (!arg->isVariable())
      {
	if (arg->deref()->containsComplexSubterm(t)) return true;
      };
  return false;
}; 
bool Term::isGround() const
{
  CALL("isGround() const");
  ASSERT(isComplex());
  for (const Term* arg = Args(); !arg->isNotArg(); arg = arg->Next())
    if (arg->isVariable() || !arg->deref()->isGround()) return false;
  return true;
}; 
void Term::collectVariables(VariableSet& varSet) const {
  varSet.reset();
  if (isVariable()) 
    { 
      varSet.add(var());
    }
  else
    {
      Iterator iter(IsReference()? deref() : this); 
      while (iter.NextVar())
	{
	  varSet.add(iter.symbolRef().var());
	};
    };
}; 
ostream& Term::output(ostream& str,const Term& t) 
{ 
  CALL("output(ostream& str,const Term& t)");      
  if (t.isVariable()) { return str << 'X' << t.var(); }
  else
    if (t.isComplex())
      {
        bool isAnswerPred = Signature::current()->isAnswerPredicate(t.functor()); 
	if (t.IsSplitting()) str << "[";
        if (isAnswerPred) str << "<<";
	if (t.Negative()) str << ((t.IsEquality()) ? '!' : '~');
	if (t.IsEquality()) 
	  if (t.IsOrderedEq()) { return str << "=="; }    
	  else return str << '=';
	if (t.isBuiltInTruthValue())
	  {
	    return str << "**true**";
	  };
	str << Signature::current()->symbolName(t.functor());
        if (isAnswerPred) str << ">>";
	if (t.IsSplitting()) str << "]";
	return str;
      }
    else 
      { 
	return output(str,t.deref());
      };         
}; 
ostream& Term::output(ostream& str,const Term* t) 
{   
  CALL("output(ostream& str,const Term* t)");
  if (t->IsEquality())
    { 
      output(str,*(t->Args()));
      output(str,*t);
      return output(str,*(t->Args()->Next()));   
    };
  output(str,*t);
  if (!t->arity()) return str; 
  str << '(';
  for (const Term* arg = t->Args(); !arg->isNotArg(); arg = arg->Next())
    { 
      output(str,*arg);
      if (!arg->Next()->isNotArg()) str << ',';
    };
  return str << ')';
}; 
