/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "UnifIndex.hpp"
#include "Clause.inl"
#include "Unification.hpp"
#include "SimplificationOrdering.hpp"
#include "SymbolInfo.hpp"
#include "LiteralProperties.inl"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex"
#endif 
#include "debugMacros.hpp"
UnifIndex::UnifIndex() : 
  _literalIndex(static_cast<TreeNode*>(0),"UnifIndex::_literalIndex"),
  forward_para_from_var_tree(0),
  backward_para_tree(0),
  _pos1ChainingTree(static_cast<TreeNode*>(0),"UnifIndex::_pos1ChainingTree"),
  _pos2ChainingTree(static_cast<TreeNode*>(0),"UnifIndex::_pos2ChainingTree"),
  _neg1ChainingTree(static_cast<TreeNode*>(0),"UnifIndex::_neg1ChainingTree"),
  _neg2ChainingTree(static_cast<TreeNode*>(0),"UnifIndex::_neg2ChainingTree"),
  _chainingTable(0)
{
  CALL("constructor UnifIndex()");
  DOP(_literalIndex.freeze());
  pseudo_literal[0] = Term((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity);
  _specialisedArgComparison = 
    SimplificationOrdering::current()->reserveSpecialisedComparison(); 
}; 
UnifIndex::~UnifIndex() 
{ 
  CALL("destructor ~UnifIndex()"); 
  ulong i;
  if (DestructionMode::isThorough())
    {        
      for (i = 0; i < _literalIndex.size(); i++)
	{
	  TreeNode::Destroy(_literalIndex[i]); 
	};
      ASSERT(_paramodulateFromVariables || (!forward_para_from_var_tree));
      TreeNode::Destroy(forward_para_from_var_tree); 
      TreeNode::DestroyBPI(backward_para_tree);
      for (i = 0; i < _pos1ChainingTree.size(); i++)
	{
	  TreeNode::Destroy(_pos1ChainingTree[i]);
	};
      for (i = 0; i < _pos2ChainingTree.size(); i++)
	{
	  TreeNode::Destroy(_pos2ChainingTree[i]);
	};
      for (i = 0; i < _neg1ChainingTree.size(); i++)
	{
	  TreeNode::Destroy(_neg1ChainingTree[i]);
	};
      for (i = 0; i < _neg2ChainingTree.size(); i++)
	{
	  TreeNode::Destroy(_neg2ChainingTree[i]);
	};
      ASSERT(_specialisedArgComparison);
      SimplificationOrdering::current()->releaseSpecialisedComparison(_specialisedArgComparison);
    };
}; 
void UnifIndex::reset()
{
  CALL("reset()");
  ulong i;  
  for (i = 0; i < _literalIndex.size(); i++)
    {
      TreeNode::Destroy(_literalIndex[i]);  
      _literalIndex[i] = 0;
    };
  ASSERT(_paramodulateFromVariables || (!forward_para_from_var_tree));
  TreeNode::Destroy(forward_para_from_var_tree); 
  forward_para_from_var_tree = 0;
  TreeNode::DestroyBPI(backward_para_tree);
  backward_para_tree = 0;
  for (i = 0; i < _pos1ChainingTree.size(); i++)
    {
      TreeNode::Destroy(_pos1ChainingTree[i]);
      _pos1ChainingTree[i] = 0;
    };
  for (i = 0; i < _pos2ChainingTree.size(); i++)
    {
      TreeNode::Destroy(_pos2ChainingTree[i]);
      _pos2ChainingTree[i] = 0;
    };
  for (i = 0; i < _neg1ChainingTree.size(); i++)
    {
      TreeNode::Destroy(_neg1ChainingTree[i]);
      _neg1ChainingTree[i] = 0;
    };
  for (i = 0; i < _neg2ChainingTree.size(); i++)
    {
      TreeNode::Destroy(_neg2ChainingTree[i]);
      _neg2ChainingTree[i] = 0;
    };
}; 
void UnifIndex::Integrate(Clause* cl)
{
  CALL("Integrate(Clause* cl)");
  LiteralList* lits = cl->LitList();
  LiteralList::Element* firstNonselectedLit = lits->nth(lits->numOfSelectedLits());
  ulong litNum = 0;
  for(LiteralList::Iterator iter(lits);iter.currentElement() != firstNonselectedLit;iter.next()) 
    {
      Integrate(iter.currentElement(),cl,litNum);
      litNum++;
    };
}; 
void UnifIndex::Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)
{
  CALL("Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)");
  const Term* litBody = lit->literal();
  if (litBody->arity())
    {
      if (paramodulation) 
	integrateForBackwardSuperposition(lit,cl,lit_num);
      if (litBody->IsEquality())
	{      
	  if (lit->properties().equalityMode() == LiteralProperties::DefaultEqualityMode)
	    {
	      if (litBody->Positive()) integrateForForwardSuperposition(lit,cl,lit_num);
	    }
	  else 
	    {
	      ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolution ||
		     lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	      TreeNode* leaf = integrateEqForResolution(litBody);
	      ASSERT(leaf->indexedTerm().IsReference());
	      ASSERT_IN(!leaf->indexedTerm().deref() || leaf->indexedTerm().deref() == litBody,"COP10");
	      leaf->SetTerm(litBody);
	      leaf->AddPair(lit_num,cl); 
	      if (lit->properties().equalityMode() == LiteralProperties::UseEqForResolution)
		{
		  if (litBody->Positive()) integrateForForwardSuperposition(lit,cl,lit_num);
		}
	      else
		{
		  ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
		};
	    };
	}
      else 
	{
	  TreeNode* leaf = IntegrateNonpropLit(litBody);
	  ASSERT(leaf->indexedTerm().IsReference());
	  ASSERT_IN(!leaf->indexedTerm().deref() || leaf->indexedTerm().deref() == litBody,"COP20");
	  leaf->SetTerm(litBody);
	  leaf->AddPair(lit_num,cl); 
	};
      if (_chainingTable && _chainingTable->transitivityLaw(litBody->functor()))
	{
	  integrateForChaining(litBody,cl,lit_num);
	};
    }
  else
    {
      TreeNode* leaf = IntegratePropLit(*litBody);    
      ASSERT(leaf->indexedTerm().IsReference());
      ASSERT_IN(!leaf->indexedTerm().deref() || leaf->indexedTerm().deref() == litBody,"COP30");
      leaf->SetTerm(litBody);
      leaf->AddPair(lit_num,cl); 
    };  
}; 
void UnifIndex::integrateBackSupRedex(const Term* redex,
				      Clause* cl,
				      unsigned lit_num,
				      SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateBackSupRedex(const Term* redex,Clause* cl,unsigned lit_num,SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(redex->isComplex());
  TreeNode** tree = &backward_para_tree;
  TreeNode* leaf;
  pseudo_literal->Args()->makeReference(const_cast<Term*>(redex));
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  leaf = IntegrateArgs(tree);
  leaf->SetTerm(redex);
  leaf->addBackSupTriple(lit_num,cl,redex,constr);  
}; 
void 
UnifIndex::integrateForBackwardSuperposition(const LiteralList::Element* lit,
					    Clause* cl,
					    unsigned lit_num)
{
  CALL("integrateForBackwardSuperposition(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->arity());
  _previousBackSupRedexes.reset();
  if (lit->isEquality())
    {
      if (cl->isSupportClause())
	{
	  if ((lit->positive() && 
	       !_orientPosEqualitiesInSupport &&
	       !lit->isPreorientedEq()) ||
	      (lit->negative() && !_orientNegEqualitiesInSupport))
	    goto integrate_ordinary; 
	};
      if (lit->isOrientedEq())
	{
	  if (!lit->equalityLHS()->isVariable())
	    integrateComplexSubtermForBackwardSuperposition(lit->equalityLHS()->deref(),
							    cl,
							    lit_num,
							    0);
	}
      else 
	{
	  const Term* arg1ref = lit->arg1();
	  const Term* arg2ref = arg1ref->Next();
	  if (arg1ref->isVariable() || 
	      arg2ref->isVariable() || 
	      !lit->isOrientableEq()) 
	    goto integrate_ordinary;
	  SimplificationOrdering::StoredConstraint* constraint = 0;
	  _specialisedArgComparison->load(*arg1ref,*arg2ref);
	  if (_specialisedArgComparison->lhsVarInstMayOrder())
	    {
	      constraint = 
		SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
	    };
	  integrateComplexSubtermForBackwardSuperposition(arg1ref->deref(),
							  cl,
							  lit_num,
							  constraint);
	  if (constraint)
	    SimplificationOrdering::current()->releaseConstraint(constraint);
	  constraint = 0;
	  if (_specialisedArgComparison->rhsVarInstMayOrder())
	    {
	      constraint = 
		SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison);
	    };
	  integrateComplexSubtermForBackwardSuperposition(arg2ref->deref(),
							  cl,
							  lit_num,
							  constraint);
	  if (constraint)
	    SimplificationOrdering::current()->releaseConstraint(constraint);
	};
    }
  else 
    {
    integrate_ordinary:
      const bool* nonBasicArgs;
      if (_useInputNonBasicness &&
	  ((nonBasicArgs = lit->literal()->nonBasicArguments())))
	{
	  ASSERT(!lit->isEquality());
	  ulong argNum = 0UL;
	  for (const Term* arg = lit->args(); !arg->isNotArg(); arg = arg->Next())
	    {
	      if (!arg->isVariable() &&
		  !nonBasicArgs[argNum])
		integrateComplexSubtermForBackwardSuperposition(arg->deref(),
								cl,
								lit_num,
								0);
	      ++argNum;
	    };
	}
      else
	{
	  for (const Term* arg = lit->args(); !arg->isNotArg(); arg = arg->Next())
	    {
	      if (!arg->isVariable())
		integrateComplexSubtermForBackwardSuperposition(arg->deref(),
								cl,
								lit_num,
								0);
	    };
	};
    };
}; 
void 
UnifIndex::integrateComplexSubtermForBackwardSuperposition(const Term* subterm,
							  Clause* cl,
							  unsigned litNum,
							  SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateComplexSubtermForBackwardSuperposition(..)");
  ASSERT(subterm->isComplex());
  if (_backSupNoDuplicateRedexes)
    {
      for (const Term** prev = _previousBackSupRedexes.begin();
	   prev != _previousBackSupRedexes.end();
	   ++prev)
	{
	  if (*prev == subterm) return;
	};
    };
  _previousBackSupRedexes.push(subterm);
  integrateBackSupRedex(subterm,cl,litNum,constr);
  const bool* nonBasicArgs;
  if (_useInputNonBasicness &&
      ((nonBasicArgs = subterm->nonBasicArguments())))
    {      
      ulong argNum = 0UL;
      for (const Term* arg = subterm->Args();
	   !arg->isNotArg();
	   arg = arg->Next())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum]) 
	    integrateComplexSubtermForBackwardSuperposition(arg->deref(),
							    cl,
							    litNum,
							    constr);
	  ++argNum;
	};
    }
  else
    {       
      for (const Term* arg = subterm->Args();
	   !arg->isNotArg();
	   arg = arg->Next())
	{
	  if (!arg->isVariable())
	    integrateComplexSubtermForBackwardSuperposition(arg->deref(),
							    cl,
							    litNum,
							    constr);
	};
    };
}; 
void 
UnifIndex::integrateForForwardSuperposition(const LiteralList::Element* lit,
					   Clause* cl,
					   unsigned lit_num)
{
  CALL("integrateForForwardSuperposition(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->isEquality());
  ASSERT(lit->positive());
  TreeNode* leaf;
  const Term* arg;
  Term replacing_term;
  bool stop;
  const Term* arg1ref;
  const Term* arg2ref;
  if (!_orientPosEqualitiesInSupport && 
      cl->isSupportClause() &&
      !lit->isPreorientedEq())
    goto integrate_unconstrained;
  if (lit->isOrientedEq()) 
    { 
      integrateOrientedForForwardSuperposition(lit,cl,lit_num); 
      return; 
    };
  if (!lit->isOrientableEq()) goto integrate_unconstrained;
  arg1ref = lit->args();
  arg2ref = arg1ref->Next();
  if (arg1ref->IsReference() && arg2ref->IsReference())
    {
      SimplificationOrdering::StoredConstraint* constraint;
      _specialisedArgComparison->load(*arg1ref,*arg2ref);
      if (_specialisedArgComparison->lhsVarInstMayOrder())
	{
	  constraint = SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
	  integrateConstrainedForForwardSuperposition(*arg1ref,*arg2ref,cl,lit_num,constraint); 
	  if (constraint) SimplificationOrdering::current()->releaseConstraint(constraint);
	}
      else
	{
	  integrateConstrainedForForwardSuperposition(*arg1ref,*arg2ref,cl,lit_num,0);
	};
      if (_specialisedArgComparison->rhsVarInstMayOrder())
	{
	  constraint = SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison);   
	  integrateConstrainedForForwardSuperposition(*arg2ref,*arg1ref,cl,lit_num,constraint);
	  if (constraint) SimplificationOrdering::current()->releaseConstraint(constraint);
	}
      else
	{
	  integrateConstrainedForForwardSuperposition(*arg2ref,*arg1ref,cl,lit_num,0);
	};
      return;
    };
 integrate_unconstrained:
  arg = lit->args();
  replacing_term = *(arg->Next());
  stop = false;
 integrate_arg: 
  if (arg->isVariable())
    {
      if (_paramodulateFromVariables)
	{
	  pseudo_literal[1] = *arg;
	  l_iter.Reset(pseudo_literal,1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(&forward_para_from_var_tree);
	  leaf->SetTerm(*arg);
	  leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);
	}
      else
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedWithParamodulationsFromVariables);
	};
    }
  else 
    {
      if (arg->deref()->arity())
	{
	  l_iter.Reset(arg->deref(),1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(FParaTree(*(arg->deref())));
	}
      else
	{
	  leaf = *(FParaTree(*(arg->deref())));
	  if (!leaf) 
	    {
	      leaf = TreeNode::NewLeaf();
	      *(FParaTree(*(arg->deref()))) = leaf;
	    };
	}; 
      leaf->SetTerm(arg->deref());
      leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);
    };    
  if (stop) return; 
  replacing_term = *arg;
  arg = arg->Next();
  stop = true;
  goto integrate_arg;
}; 
inline
void UnifIndex::integrateConstrainedForForwardSuperposition(const Term& lhs,
							    const Term& rhs,
							    Clause* cl,
							    unsigned lit_num,
							    SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateConstrainedForForwardSuperposition(..)");
  ASSERT(lhs.IsReference());
  ASSERT(rhs.IsReference());  
  TreeNode* leaf;
  if (lhs.deref()->arity())
    {
      l_iter.Reset(lhs.deref(),1UL);
      l_iter.Next();
      leaf = IntegrateArgs(FParaTree(*(lhs.deref())));
    }
  else
    {
      leaf = *(FParaTree(*(lhs.deref())));
      if (!leaf) 
	{
	  leaf = TreeNode::NewLeaf();
	  *(FParaTree(*(lhs.deref()))) = leaf;
	};
    }; 
  leaf->SetTerm(lhs.deref());
  leaf->addForwardSupTriple(lit_num,cl,rhs,constr); 
}; 
void 
UnifIndex::integrateOrientedForForwardSuperposition(const LiteralList::Element* lit,
						     Clause* cl,
						     unsigned lit_num)
{
  CALL("integrateOrientedForForwardSuperposition(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->isOrientedEq());
  TreeNode* leaf;
  const Term* arg = lit->equalityLHS();
  Term replacing_term = *(lit->equalityRHS());   
  if (arg->isVariable())
    {
      ASSERT(lit->isPreorientedEq());
      if (_paramodulateFromVariables)
	{
	  pseudo_literal[1] = *arg;
	  l_iter.Reset(pseudo_literal,1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(&forward_para_from_var_tree);
	  leaf->SetTerm(*arg);
	  leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);
	}
      else
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedWithParamodulationsFromVariables);
	};
    }
  else 
    {
      if (arg->deref()->arity())
	{
	  l_iter.Reset(arg->deref(),1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(FParaTree(*(arg->deref())));
	}
      else
	{
	  leaf = *(FParaTree(*(arg->deref())));
	  if (!leaf) 
	    {
	      leaf = TreeNode::NewLeaf();
	      *(FParaTree(*(arg->deref()))) = leaf;
	    };
	}; 
      leaf->SetTerm(arg->deref());
      leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);  
    };
}; 
void UnifIndex::integrateForChaining(const Term* lit,Clause* cl,unsigned litNum)
{
  CALL("integrateForChaining(const Term* lit,Clause* cl,unsigned litNum)");
  ASSERT(_chainingTable && _chainingTable->transitivityLaw(lit->functor()));
  if (lit->Positive())
    {
      integratePosForChaining(lit,cl,litNum);
    }
  else
    {
      integrateNegForChaining(lit,cl,litNum);
    };
}; 
void UnifIndex::integratePosForChaining(const Term* lit,Clause* cl,unsigned litNum)
{
  CALL("integratePosForChaining(const Term* lit,Clause* cl,unsigned litNum)");
  const Term* arg1ref = lit->Args();
  const Term* arg2ref = arg1ref->Next();
  bool integrateArg1 = true;
  bool integrateArg2 = true;
  bool constrainArg1 = true;
  bool constrainArg2 = true;
  SimplificationOrdering::StoredConstraint* constraint1 = 0;
  SimplificationOrdering::StoredConstraint* constraint2 = 0;
  _specialisedArgComparison->load(*arg1ref,*arg2ref);
  if (!_specialisedArgComparison->lhsVarInstMayOrder())
    constrainArg1 = false;
  if (!_specialisedArgComparison->rhsVarInstMayOrder())
    constrainArg2 = false;
  if (arg1ref->isVariable())
    {
      if (!_chainingThroughVariables)
	{
	  integrateArg1 = false;
	  constrainArg1 = false;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	};
    };
  if (integrateArg1 && constrainArg1)
    {
      _specialisedArgComparison->specialiseLessOrEqual();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess: 
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess: 
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break; 
	};
    };
  if (arg2ref->isVariable())
    {
      if (!_chainingThroughVariables)
	{
	  integrateArg2 = false;
	  constrainArg2 = false;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	};
    };
  if (integrateArg2 && constrainArg2)
    {
      _specialisedArgComparison->specialiseGreaterOrEqual();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
	  constrainArg2 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual: 
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:  
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater:  
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile: 
	  break; 
	};
    };
  ASSERT(!constrainArg1 || integrateArg1);
  ASSERT(!constrainArg2 || integrateArg2);
  if (constrainArg1)
    {
      constraint1 = 
	SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
    };
  if (constrainArg2)
    {
      constraint2 = 
	SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison);
    };
  if (integrateArg1)
    integratePos1ForChaining(lit->functor(),*arg1ref,*arg2ref,cl,litNum,constraint1);
  if (integrateArg2)
    integratePos2ForChaining(lit->functor(),*arg2ref,*arg1ref,cl,litNum,constraint2);
  if (constraint1) 
    SimplificationOrdering::current()->releaseConstraint(constraint1);
  if (constraint2) 
    SimplificationOrdering::current()->releaseConstraint(constraint2);
}; 
void 
UnifIndex::integratePos1ForChaining(ulong predNum,
				   const Term& head,
				   const Term& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integratePos1ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  TreeNode* leaf = 
    IntegrateArgs(&(_pos1ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; 
void 
UnifIndex::integratePos2ForChaining(ulong predNum,
				   const Term& head,
				   const Term& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integratePos2ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  TreeNode* leaf = 
    IntegrateArgs(&(_pos2ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; 
void UnifIndex::integrateNegForChaining(const Term* lit,Clause* cl,unsigned litNum)
{
  CALL("integrateNegForChaining(const Term* lit,Clause* cl,unsigned litNum)");
  const Term* arg1ref = lit->Args();
  const Term* arg2ref = arg1ref->Next();
  bool integrateArg1 = true;
  bool integrateArg2 = true;
  bool constrainArg1 = true;
  bool constrainArg2 = true;
  SimplificationOrdering::StoredConstraint* constraint1 = 0;
  SimplificationOrdering::StoredConstraint* constraint2 = 0;
  _specialisedArgComparison->load(*arg1ref,*arg2ref);
  if (!_specialisedArgComparison->lhsVarInstMayOrder())
    constrainArg1 = false;
  if (!_specialisedArgComparison->rhsVarInstMayOrder())
    constrainArg2 = false;
  if (arg1ref->isVariable())
    {
      if (cl->singletonVar(arg1ref->var()))
	{	  
	  integrateArg1 = false;
	  constrainArg1 = false;
	}
      else
	if (!_chainingThroughVariables)
	  {
	    integrateArg1 = false;
	    constrainArg1 = false;
	    ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	    ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  };
    };
  if (integrateArg1 && constrainArg1)
    {
      _specialisedArgComparison->specialiseLessOrEqual();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess: 
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess: 
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break; 
	};
    };
  if (arg2ref->isVariable())
    {
      if (cl->singletonVar(arg2ref->var()))
	{	  
	  integrateArg2 = false;
	  constrainArg2 = false;
	}
      else
	if (!_chainingThroughVariables)
	  {
	    integrateArg2 = false;
	    constrainArg2 = false;
	    ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	    ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  };
    };
  if (integrateArg2 && constrainArg2)
    {
      _specialisedArgComparison->specialiseGreater();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  constrainArg2 = false;
	  break; 
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break; 
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  constrainArg2 = false;
	  break; 
	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break;
	};
    };
  ASSERT(!constrainArg1 || integrateArg1);
  ASSERT(!constrainArg2 || integrateArg2);
  if (constrainArg1)
    {
      constraint1 = 
	SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
    };
  if (constrainArg2)
    {
      constraint2 = 
	SimplificationOrdering::current()->storedConstraintGreater(_specialisedArgComparison);
    };
  if (integrateArg1)
    integrateNeg1ForChaining(lit->functor(),*arg1ref,*arg2ref,cl,litNum,constraint1);
  if (integrateArg2)
    integrateNeg2ForChaining(lit->functor(),*arg2ref,*arg1ref,cl,litNum,constraint2);
  if (constraint1) 
    SimplificationOrdering::current()->releaseConstraint(constraint1);
  if (constraint2) 
    SimplificationOrdering::current()->releaseConstraint(constraint2);
}; 
void 
UnifIndex::integrateNeg1ForChaining(ulong predNum,
				   const Term& head,
				   const Term& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integrateNeg1ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  TreeNode* leaf = 
    IntegrateArgs(&(_neg1ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; 
void 
UnifIndex::integrateNeg2ForChaining(ulong predNum,
				   const Term& head,
				   const Term& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integrateNeg2ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  TreeNode* leaf = 
    IntegrateArgs(&(_neg2ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; 
UnifIndex::TreeNode* UnifIndex::IntegratePropLit(Term lit)
{
  CALL("IntegratePropLit(Term lit)");
  TreeNode** tree = Tree(lit);
  if (!(*tree))
    {
      *tree = TreeNode::NewLeaf();
    };
  return *tree;
}; 
UnifIndex::TreeNode* UnifIndex::IntegrateNonpropLit(const Term* lit)
{
  CALL("IntegrateNonpropLit(const Term* lit)");
  TreeNode** tree = Tree(*lit);
  l_iter.Reset(lit,1UL);
  l_iter.Next();
  return IntegrateArgs(tree);
}; 
UnifIndex::TreeNode* UnifIndex::IntegrateArgs(TreeNode** tree)
{
  CALL("IntegrateArgs(TreeNode** tree)");
  var_counter.reset();
  if (*tree)
    {
      TreeNode* node = *tree;
      UnifTerm::Symbol sym;
    take_next_sym:
      if (l_iter.CurrentSymIsVar())
	{
	  ulong var_num = l_iter.symbolRef().var();
	  sym.MkVar(Unifier::current()->variableBase(var_num));
	  var_counter.add(var_num);
	}
      else
	{
	  sym.MkFunc(l_iter.symbolRef());
	};
    check_node:
      if (node->Symbol() == sym)
	{
	  node = node->Below();
	  if (l_iter.Next())
	    {
	      goto take_next_sym;
	    }
	  else 
	    {
	      ASSERT(node->IsLeaf());
	      return node;
	    };
	}
      else 
	{
	  if (node->Alternative())
	    {
	      node = node->Alternative();
	      goto check_node;
	    }
	  else 
	    {
	      TreeNode* leaf;
	      node->SetAlternative(CreateBranch(leaf));
	      return leaf;
	    };
	};      
    }
  else 
    {
      TreeNode* leaf;
      *tree = CreateBranch(leaf);
      return leaf;
    };
}; 
UnifIndex::TreeNode* 
UnifIndex::integrateEqForResolution(const Term* lit)
{
  CALL("integrateEqForResolution(const Term* lit)");
  Term header = *lit;
  ASSERT(header.IsEquality());
  header.MakeUnordered();
  TreeNode** tree = Tree(header);
  l_iter.Reset(lit,1UL);
  l_iter.Next();
  return IntegrateArgs(tree);
}; 
void UnifIndex::Remove(Clause* cl)
{   
  LiteralList* allLiterals = cl->LitList();
  LiteralList::Element* firstNonselectedLit = allLiterals->nth(allLiterals->numOfSelectedLits());
  for(LiteralList::Iterator iter(allLiterals); 
      iter.currentElement() != firstNonselectedLit;
      iter.next())
    Remove(iter.currentElement(),cl); 
}; 
void UnifIndex::RemovePropLit(const Term* lit,Clause* cl)
{
  CALL("RemovePropLit(const Term* lit,Clause* cl)");
  ASSERT(!lit->arity());
  TreeNode** tree = Tree(*lit);
  ASSERT(*tree);
  ASSERT((*tree)->IsLeaf());
  (*tree)->Remove(cl);
  if (!(*tree)->indObjects()) 
    {
      (*tree)->DeleteLeaf();
      *tree = 0;
    };
}; 
UnifIndex::TreeNode* 
UnifIndex::LeafAndBranch(const Term* t,TreeNode** tree,TreeNode**& branch)
{
  CALL("LeafAndBranch(const Term* t,TreeNode** tree,TreeNode**& branch)");
  ASSERT(tree);
  ASSERT(t->arity());
  branch = tree;
  TreeNode* node = *tree;
  if (!node) return 0; 
  TreeNode** node_addr = tree; 
  Term::Iterator iter;  
  iter.Reset(t);
 next_sym:
  if (!iter.Next())
    {
      ASSERT(node->IsLeaf());
      return node;
    };
 check_node:
  if (node->SameSymbol(iter.symbolRef()))
    {
      if (node->Alternative()) branch = node_addr;
      node_addr = node->BelowAddr();
      node = node->Below();
      ASSERT(node);
      goto next_sym;     
    }; 
  if (node->Alternative())
    {
      node_addr = node->AlternativeAddr();
      branch = node_addr;
      node = node->Alternative();
      goto check_node;
    };
  return 0;
}; 
void UnifIndex::Remove(const Term* t,Clause* cl,TreeNode** tree)
{  
  CALL("Remove(const Term* t,Clause* cl,TreeNode** tree)");
  ASSERT(t->arity());
  TreeNode** branch;
  TreeNode* branch_to_delete;
  TreeNode* node_to_delete;
  TreeNode* leaf = LeafAndBranch(t,tree,branch);
  if (!leaf) return;
  leaf->Remove(cl);
  if (!(leaf->indObjects())) 
    {
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
	{
	  node_to_delete = branch_to_delete;
	  branch_to_delete = branch_to_delete->Below();
	  node_to_delete->DeleteNode();
	};
      branch_to_delete->DeleteLeaf(); 
    };  
}; 
void 
UnifIndex::removeEqFromResolutionIndex(const Term* eqLit,Clause* cl)
{
  CALL("removeEqFromResolutionIndex(const Term* eqLit,Clause* cl)");
  Term header = *eqLit;
  ASSERT(header.IsEquality());
  header.MakeUnordered();
  Remove(eqLit,cl,Tree(header));
}; 
void UnifIndex::RemoveBPCandidates(const Term* t,Clause* cl,TreeNode** tree)
{  
  CALL("RemoveBPCandidates(const Term* t,Clause* cl,TreeNode** tree)");
  ASSERT(t->arity());
  TreeNode** branch;
  TreeNode* branch_to_delete;
  TreeNode* node_to_delete;
  TreeNode* leaf = LeafAndBranch(t,tree,branch);
  if (!leaf) 
    {
      return;
    };
  leaf->RemoveBPCandidates(cl);
  if (!(leaf->BPCandidates())) 
    {
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
	{
	  node_to_delete = branch_to_delete;
	  branch_to_delete = branch_to_delete->Below();
	  node_to_delete->DeleteNode();
	};
      branch_to_delete->DeleteLeaf(); 
    };
}; 
void UnifIndex::RemoveFromBPI(const LiteralList::Element* lit,Clause* cl)
{
  CALL("RemoveFromBPI(const LiteralList::Element* lit,Clause* cl)");
  _previousBackSupRedexes.reset();
  if (lit->isEquality()) 
    {
      if (cl->isSupportClause())
	{
	  if ((lit->positive() && 
	       !_orientPosEqualitiesInSupport &&
	       !lit->isPreorientedEq()) ||
	      (lit->negative() && !_orientNegEqualitiesInSupport))
	    goto remove_ordinary;
	};
      if (lit->isOrientedEq())
	{
	  if (!lit->equalityLHS()->isVariable())
	    removeComplexSubtermFromBPI(lit->equalityLHS()->deref(),cl);
	  return;
	};
    };
 remove_ordinary:
  const bool* nonBasicArgs;
  if (_useInputNonBasicness &&
      ((nonBasicArgs = lit->literal()->nonBasicArguments())))
    {
      ASSERT(!lit->isEquality());
      ulong argNum = 0UL;
      for (const Term* arg = lit->args(); !arg->isNotArg(); arg = arg->Next())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    removeComplexSubtermFromBPI(arg->deref(),cl);
	  ++argNum;
	};
    }
  else
    {
      for (const Term* arg = lit->args(); !arg->isNotArg(); arg = arg->Next())
	{
	  if (!arg->isVariable())
	    removeComplexSubtermFromBPI(arg->deref(),cl);
	};
    };
}; 
void UnifIndex::removeComplexSubtermFromBPI(const Term* subterm,Clause* cl)
{
  CALL("removeComplexSubtermFromBPI(const Term* subterm,Clause* cl)");
  ASSERT(subterm->isComplex());
  if (_backSupNoDuplicateRedexes)
    {
      for (const Term** prev = _previousBackSupRedexes.begin();
	   prev != _previousBackSupRedexes.end();
	   ++prev)
	{
	  if (*prev == subterm) return;
	};
    };
  _previousBackSupRedexes.push(subterm);
  RemoveBPRedex(subterm,cl);
  const bool* nonBasicArgs;
  if (_useInputNonBasicness &&
      ((nonBasicArgs = subterm->nonBasicArguments())))
    {      
      ulong argNum = 0UL;
      for (const Term* arg = subterm->Args();
	   !arg->isNotArg();
	   arg = arg->Next())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    removeComplexSubtermFromBPI(arg->deref(),cl);
	  ++argNum;
	};
    }
  else
    {
      for (const Term* arg = subterm->Args();
	   !arg->isNotArg();
	   arg = arg->Next())
	{
	  if (!arg->isVariable())
	    removeComplexSubtermFromBPI(arg->deref(),cl);
	};
    };
}; 
void UnifIndex::RemoveBPRedex(const Term* t,Clause* cl)
{
  CALL("RemoveBPRedex(const Term* t,Clause* cl)");
  pseudo_literal->Args()->makeReference(const_cast<Term*>(t));
  RemoveBPCandidates(pseudo_literal,cl,&backward_para_tree);
}; 
void UnifIndex::RemoveFromFPI(const LiteralList::Element* lit,Clause* cl)
{
  CALL("RemoveFromFPI(const LiteralList::Element* lit,Clause* cl)");
  bool treatAsUnoriented = 
    cl->isSupportClause() &&
    !_orientPosEqualitiesInSupport &&
    !lit->isPreorientedEq();
  bool stop = false;
  const Term* arg = lit->equalityLHS();
 remove_arg:
  if (arg->isVariable())
    {
      if (_paramodulateFromVariables)
	{
	  pseudo_literal[1] = *arg;
	  Remove(pseudo_literal,cl,&forward_para_from_var_tree);
	};
    }
  else 
    if (arg->deref()->arity())
      {
	Remove(arg->deref(),cl,FParaTree(*(arg->deref()))); 
      }
    else 
      {
	TreeNode** tree = FParaTree(*(arg->deref()));
	if (*tree)
	  {
	    (*tree)->Remove(cl);
	    if (!(*tree)->indObjects()) 
	      {
		(*tree)->DeleteLeaf();
		*tree = 0;
	      };
	  };
      };
  if ((!treatAsUnoriented && lit->isOrientedEq()) || stop) 
    { 
      return;
    }; 
  stop = true;
  arg = lit->equalityRHS();
  goto remove_arg;          
}; 
void UnifIndex::removeFromChainingIndex(const Term* lit,Clause* cl)
{
  CALL("removeFromChainingIndex(const Term* lit,Clause* cl)");
  ASSERT(_chainingTable && _chainingTable->transitivityLaw(lit->functor()));
  if (lit->Positive())
    {
      removePosFromChainingIndex(lit,cl);
    }
  else
    {
      removeNegFromChainingIndex(lit,cl);
    };
}; 
void UnifIndex::removePosFromChainingIndex(const Term* lit,Clause* cl)
{
  CALL("removePosFromChainingIndex(const Term* lit,Clause* cl)");
  const Term* arg1ref = lit->Args();
  const Term* arg2ref = arg1ref->Next();
  pseudo_literal[1] = *arg1ref;
  Remove(pseudo_literal,cl,&(_pos1ChainingTree.sub(lit->functor())));
  pseudo_literal[1] = *arg2ref;
  Remove(pseudo_literal,cl,&(_pos2ChainingTree.sub(lit->functor())));  
}; 
void UnifIndex::removeNegFromChainingIndex(const Term* lit,Clause* cl)
{
  CALL("removeNegFromChainingIndex(const Term* lit,Clause* cl)");
  const Term* arg1ref = lit->Args();
  const Term* arg2ref = arg1ref->Next();
  pseudo_literal[1] = *arg1ref;
  Remove(pseudo_literal,cl,&(_neg1ChainingTree.sub(lit->functor())));
  pseudo_literal[1] = *arg2ref;
  Remove(pseudo_literal,cl,&(_neg2ChainingTree.sub(lit->functor())));  
}; 
void UnifIndex::Remove(const LiteralList::Element* lit,Clause* cl)
{
  CALL("Remove(const LiteralList::Element* lit,Clause* cl)");
  const Term* litBody = lit->literal();
  if (!litBody->arity()) { RemovePropLit(litBody,cl); return; };
  if (paramodulation) RemoveFromBPI(lit,cl);
  if (litBody->IsEquality())
    {     
      if (lit->properties().equalityMode() == LiteralProperties::DefaultEqualityMode)
	{
	  if (litBody->Positive()) RemoveFromFPI(lit,cl);
	}
      else
	{
	  ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolution ||
		 lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	  removeEqFromResolutionIndex(litBody,cl);
	  if (lit->properties().equalityMode() == LiteralProperties::UseEqForResolution)
	    {
	      if (litBody->Positive()) RemoveFromFPI(lit,cl);
	    }
	  else
	    {
	      ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	    };
	};
    }
  else 
    {
      Remove(litBody,cl,Tree(*litBody));
    };
  if (_chainingTable && _chainingTable->transitivityLaw(litBody->functor()))
    {
      removeFromChainingIndex(litBody,cl);
    };
}; 
UnifIndex::TreeNode* UnifIndex::CreateBranch(TreeNode*& leaf)
{
  UnifTerm::Symbol sym;
  TreeNode::Tag tag;
  if (l_iter.CurrentSymIsVar())
    {
      if (var_counter.scoreOf(l_iter.symbolRef().var()))
	{
	  tag = TreeNode::Var;
	}
      else 
	{
	  var_counter.add(l_iter.symbolRef().var());
	  tag = TreeNode::VarFirstOcc;
	};
      sym.MkVar(Unifier::current()->variableBase(l_iter.symbolRef().var()));
    }
  else
    {
      if (l_iter.symbolRef().arity())
	{
	  tag = TreeNode::Func;
	}
      else
	{
	  tag = TreeNode::Const;
	};
      sym.MkFunc(l_iter.symbolRef());
    };
  TreeNode* result = TreeNode::NewNode(tag,sym,l_iter.Depth());
  TreeNode* node = result;
  while (l_iter.Next())
    {
      if (l_iter.CurrentSymIsVar())
	{
	  if (var_counter.scoreOf(l_iter.symbolRef().var()))
	    {
	      tag = TreeNode::Var;
	    }
	  else 
	    {
	      var_counter.add(l_iter.symbolRef().var());
	      tag = TreeNode::VarFirstOcc;
	    };
	  sym.MkVar(Unifier::current()->variableBase(l_iter.symbolRef().var()));
	}
      else
	{
	  if (l_iter.symbolRef().arity())
	    {
	      tag = TreeNode::Func;
	    }
	  else
	    {
	      tag = TreeNode::Const;
	    };
	  sym.MkFunc(l_iter.symbolRef());
	};
      node->SetBelow(TreeNode::NewNode(tag,sym,l_iter.Depth()));
      node = node->Below();
    };
  leaf = TreeNode::NewLeaf();
  node->SetBelow(leaf);
  return result;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::TreeNode"
#endif 
#include "debugMacros.hpp"
void UnifIndex::TreeNode::Destroy(TreeNode* tree)
{
  CALL("Destroy(TreeNode* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
    {
      tree->DeleteLeaf();
    }
  else
    {
      TreeNode* tree_below = tree->Below();
      TreeNode* tree_alt = tree->Alternative(); 
      tree->DeleteNode();
      Destroy(tree_below);
      if (tree_alt)
	{
	  Destroy(tree_alt);
	};   
    };
}; 
void UnifIndex::TreeNode::DestroyBPI(TreeNode* tree)
{
  CALL("DestroyBPI(TreeNode* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
    {
      tree->DeleteBPILeaf();
    }
  else
    {
      TreeNode* tree_below = tree->Below();
      TreeNode* tree_alt = tree->Alternative(); 
      tree->DeleteNode();
      DestroyBPI(tree_below);
      if (tree_alt)
	{
	  DestroyBPI(tree_alt);
	};   
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& UnifIndex::TreeNode::output(ostream& str,unsigned var_offset) const
{
  str << '[' << (ulong)this << "] " << tag() << ' ';
  switch (tag())
    {
    case Leaf: str << ' ' << indexedTerm(); break;
    case Func: 
    case Const:
    case Var: 
    case VarFirstOcc:    
      Symbol().output(str,var_offset) << " [" << (ulong)Below() << "] ["
                                      << (ulong)Alternative() << "] "
                                      << Depth();
      break;
    };
  return str;
}; 
#endif
namespace std
{
ostream& operator<<(ostream& str,const VK::UnifIndex::TreeNode::Tag& tag)
{
  switch (tag)
    {
    case UnifIndex::TreeNode::Leaf: return str << "Leaf";
    case UnifIndex::TreeNode::Func: return str << "Func";
    case UnifIndex::TreeNode::Const: return str << "Const"; 
    case UnifIndex::TreeNode::Var: return str << "Var";
    case UnifIndex::TreeNode::VarFirstOcc: return str << "VarFirstOcc"; 
    default: return str << "UNKNOWN Tag"; 
    };
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_INDEX
#define DEBUG_NAMESPACE "UnifIndex::Retrieval"
#endif 
#include "debugMacros.hpp"
inline bool UnifIndex::Retrieval::CompleteExtraction()
{
  CALL("CompleteExtraction()");
 check_iter:
  if (curr_iter->Empty()) 
    {
      return true; 
    };
 check_sym:
  Unifier::Variable* var = curr_iter->CurrVar();
  if (var) 
    {
      ASSERT(!(var->Binding()));
      bool same_var;
      if (BindWithOccCheck(var,same_var))
	{
	  ASSERT(!(same_var && var->Binding()));
	  if (same_var) { *curr_sit = VarSameVar; }
	  else { *curr_sit = VarAnything; };
	  curr_sit++;
	  curr_iter->SkipSym();
	  goto check_iter;
	}
      else
	{
	  if (BacktrackExtraction()) { goto check_sym; }
	  else { return false; };
	};
    }
  else 
    {
    check_node:
      switch (CurrNode()->tag())
        {
	case UnifIndex::TreeNode::Func : 
	case UnifIndex::TreeNode::Const : 
          if (CurrNode()->Symbol() == curr_iter->CurrSubterm()->symbol())
	    { 
	      curr_iter->SkipSym();
	      Below();
	      *curr_sit = FuncFunc;
	      curr_sit++;
	      goto check_iter;
	    }
          else 
	    {
	      Alternative();
	      if (CurrNode())
		{
		  goto check_node;
		}
	      else 
		{
		  if (BacktrackExtraction()) { goto check_sym; }
		  else { return false; };
		};
	    };
	case UnifIndex::TreeNode::Var :
          {
	    bool free_var;
	    void* bind = Shift(CurrNode()->Symbol().var())->Unref(free_var);
	    UnifTerm::CorrectBracket(curr_iter->CurrSubterm());
	    ASSERT(UnifTerm::CorrectComplexTerm(curr_iter->CurrSubterm()));
	    if (free_var)
	      {
		if (UnifTerm::Occurs(static_cast<Unifier::Variable*>(bind),curr_iter->CurrSubterm()))
		  {
		    Alternative();
		    if (CurrNode())
		      {
			goto check_node;
		      }
		    else 
		      {
			if (BacktrackExtraction()) { goto check_sym; }
			else { return false; };
		      };
		  }
		else 
		  {
		    Unifier::current()->Bind(static_cast<Unifier::Variable*>(bind),curr_iter->CurrSubterm());
		    curr_iter->SkipTerm();
		    Below();
		    *curr_sit = FuncFreeVar;
		    curr_sit++;
		    goto check_iter;
		  };
	      }
	    else 
	      {
		if (Unification::UnifyComplex(static_cast<UnifTerm*>(bind),curr_iter->CurrSubterm()))
		  {
		    curr_iter->SkipTerm();
		    Below();
		    *curr_sit = FuncBoundVar;
		    curr_sit++;
		    goto check_iter;
		  }
		else 
		  {
		    Alternative();
		    if (CurrNode())
		      {
			goto check_node;
		      }
		    else 
		      {
			if (BacktrackExtraction()) { goto check_sym; }
			else { return false; };
		      };
		  };
	      };
          };
	case UnifIndex::TreeNode::VarFirstOcc :
          UnifTerm::CorrectBracket(curr_iter->CurrSubterm());
          ASSERT(UnifTerm::CorrectComplexTerm(curr_iter->CurrSubterm()));
          Unifier::current()->Bind(Shift(CurrNode()->Symbol().var()),
				curr_iter->CurrSubterm());
          curr_iter->SkipTerm();
          Below();
          *curr_sit = FuncFreeVar;
          curr_sit++;
          goto check_iter;         
#ifdef DEBUG_NAMESPACE
	default : ICP("A2"); return false;
#elif defined _SUPPRESS_WARNINGS_
	default : return false;
#endif 
        }; 
      ICP("A3");
      return false;
    };  
}; 
inline bool UnifIndex::Retrieval::BacktrackExtraction()
{
  CALL("BacktrackExtraction()");
 rollback:
  ASSERT((!(CurrNode())) || CurrNode()->IsLeaf());
  curr_sit--;
  switch (*curr_sit)
    {
    case StartOfExtraction : 
      ForgetIter(); 
      return false;
    case VarSameVar :
      {
	bool same_var;
	curr_iter->Backtrack();
	ASSERT(curr_iter->CurrVar());
	ASSERT(!(curr_iter->CurrVar()->Binding()));
	if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
	  {
	    if (!(same_var)) { *curr_sit = VarAnything; };
	    curr_sit++;
	    curr_iter->SkipSym();
	    return true; 
	  }
	else 
	  {
	    goto rollback; 
	  };
      };
    case VarAnything : 
      {
	bool same_var;
	curr_iter->Backtrack();
	if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
	  {
	    if (same_var) 
	      {
		*curr_sit = VarSameVar; 
	      };
	    curr_sit++; 
	    curr_iter->SkipSym();
	    return true; 
	  }
	else 
	  {
	    goto rollback; 
	  };
      };
    case FuncFunc : 
      curr_iter->Backtrack();
      ASSERT(curr_iter->CurrSubterm()->symbol().IsFunctor());
      RestoreCurrent();
      Alternative();
      if (CurrNode()) { return true; }
      else 
	{
	  goto rollback; 
	};
    case FuncFreeVar :
      Unifier::current()->UnbindLast();
      curr_iter->Backtrack();
      ASSERT(curr_iter->CurrSubterm()->symbol().IsFunctor());
      RestoreCurrent();
      Alternative();
      if (CurrNode()) { return true; }
      else 
	{
	  goto rollback; 
	};
    case FuncBoundVar :
      Unifier::current()->UnbindLastSection();
      curr_iter->Backtrack();
      ASSERT(curr_iter->CurrSubterm()->symbol().IsFunctor());
      RestoreCurrent();
      Alternative();
      if (CurrNode()) 
	{
	  return true; 
	}
      else 
	{
	  goto rollback; 
	};
#ifdef DEBUG_NAMESPACE
    default : ICP("ICP1"); return false;
#endif
    };  
  ICP("ICP2"); 
  return false; 
}; 
bool UnifIndex::Retrieval::ExtractUnifiable(UnifTerm* word) 
{
  NextIter();
  curr_iter->Reset(word);
  *curr_sit = StartOfExtraction;
  curr_sit++;
  return CompleteExtraction();
}; 
bool UnifIndex::Retrieval::AlterUnifiable()
{
  if (BacktrackExtraction()) 
    {
      return CompleteExtraction(); 
    }
  else 
    {
      return false; 
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& UnifIndex::Retrieval::outputPath(ostream& str) const
{
  for (const UnifIndex::TreeNode* const * p = path; p<next; p++)    
    (*p)->output(str,var_offset) << '\n'; 
  str << "CURRENT NODE = ";
  if (curr_node) { curr_node->output(str,var_offset); }
  else str << "[0]";  
  return str << '\n'; 
}; 
#endif
#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const VK::UnifIndex::Retrieval::Situation& sit) 
{
  switch (sit)
    {
    case UnifIndex::Retrieval::StartOfExtraction : str << "StartOfExtraction"; break;
    case UnifIndex::Retrieval::VarSameVar : str << "VarSameVar"; break;
    case UnifIndex::Retrieval::VarAnything : str << "VarAnything"; break;
    case UnifIndex::Retrieval::FuncFunc : str << "FuncFunc"; break;
    case UnifIndex::Retrieval::FuncFreeVar : str << "FuncFreeVar"; break;
    case UnifIndex::Retrieval::FuncBoundVar : str << "FuncBoundVar"; break;
    };
  return str;
}; 
};
#endif
