/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <climits>
#include <cstdio>
#include <cmath>
#include "GlobAlloc.hpp"
#include "GlobalClock.hpp"
#include "RuntimeError.hpp"
#include "GlobalStopFlag.hpp"
#include "FIFO.hpp"
#include "Receiver.hpp"
#include "VampireKernel.hpp"
#include "Signature.hpp"
#include "Clause.hpp"
#include "ProblemProfile.hpp"
#include "Master.hpp"
#include "ClauseNumIndex.hpp"
#include "TermWeighting.hpp"
#include "UniformTermWeighting.hpp"
#include "NonuniformConstantTermWeighting.hpp"
#include "StandardKBOrdering.hpp"
#include "LPO.hpp"
#include "WeightPolynomial.hpp"
#include "ClauseSet.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
#include "TermSharing.hpp"
#include "LiteralSharing.hpp"
#include "ClauseAssembler.hpp"
#include "LiteralProperties.inl"
#include "ComponentName.hpp"
using namespace BK;
using namespace VK;
using namespace VampireKernelEnvironment;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#  define DEBUG_NAMESPACE "VampireKernel::_Implementation"
#endif    
#include "debugMacros.hpp"
class VampireKernel::_Implementation
{
public:
  void init(const VampireKernel::Options& opt);
  void destroy();
  void openSession()
  {
    _currentSession = this;
  };
  static void closeSession()
  {
    _currentSession = 0;
  };
  void* operator new(size_t)
  {
    return GlobAlloc::allocate(sizeof(_Implementation));
  };
  void operator delete(void* obj)
  {
    GlobAlloc::deallocate(obj,sizeof(_Implementation));
  };
  static _Implementation* allocate()
  { 
    return 
      static_cast<_Implementation*>(GlobAlloc::allocate(sizeof(_Implementation)));
  };
  static void deallocate(_Implementation* obj)
  {
    GlobAlloc::deallocate(static_cast<void*>(obj),sizeof(_Implementation));
  };
  void beginInputSignature();
  void endInputSignature();
  void registerInputFunction(const char* name,
			     ulong arity,
			     ulong& id,
			     const long& weight,
			     long precedence);
  void registerInputPredicate(const char* name,
			      ulong arity,
			      bool isAnswer,
			      ulong& id,
			      const long& weight,
			      long precedence,
			      long positiveHeaderRank,
			      long negativeHeaderRank);
  void registerEquality(ulong& id,
			long positiveHeaderRank,
			long negativeHeaderRank); 
  void setSymbolWeightConstPart(ulong symId,
				const long& weight);
  void setSymbolPrecedence(ulong symId,long prec);
  void setHeaderRank(bool positive,ulong symId,long prec); 
  void registerInterpreted(bool predicate,ulong symId);
  void beginInputClauses();
  void endInputClauses();
  void beginInputClause(ulong clauseNumber,void* origin,const ClauseAnnotation& annotation);
  void endInputClause();
  bool beginLiteral(bool positive,ulong predicateId);
  void markSelected();
  void markUnselected();
  void markRuleHead();
  void markSubquery();
  void markEqWithResolution();
  void markEqWithResolutionOnly();
  void markGivenPosEqOrientation();
  void markReversePosEqOrientation();
  void endLiteral();
  void receiveFunction(ulong symId);
  void receiveInterpretedConstant(const InterpretedConstant& value);
  void receiveVariable(ulong varNum);
  void saturate(ulong nextVacantClauseId,long timeLimit);
  void inputToMasterAndRunMainLoop();
  void lookForAnotherAnswer(int timeLimitInDeciseconds);
  void suspend();
  void resume(); 
  void reportInterrupt(int signalNumber);
  void reportErrorSignal(int signalNumber);
  void reportError();
  VampireKernel::TerminationReason terminationReason() const;
  const VampireKernel::Clause* refutation() const;
  VampireKernel::ClauseSetIterator* activeClauses();
  bool isInterpretedConstantId(ulong symbolId) const;
  bool isInputSymbol(ulong symbolId) const; 
  bool isAnswerPredicateId(ulong symbolId) const;
  ulong symbolArity(ulong symbolId) const;
  const char* symbolPrintName(ulong symbolId) const; 
  const InterpretedConstant& interpretedConstantValue(ulong symbolId) const;
private:
  class ClauseSetIterator : public VampireKernel::ClauseSetIterator
  {
  public:
    ClauseSetIterator(const ClauseSet& set) :
      _iter(set)
    {
    };
    ~ClauseSetIterator();
    void* operator new(size_t)
    {
      return GlobAlloc::allocate(sizeof(ClauseSetIterator));
    };
    void operator delete(void* obj)
    {
      GlobAlloc::deallocate(obj,sizeof(ClauseSetIterator));
    };
    bool isEnd() const;
    void next();
    const VampireKernel::Clause* clause() const;
  private:
    ClauseSet::ConstIter _iter;
  }; 
private:
  static void handleAlarm(int);
  static void globalStopFlagHandler();
  void loadInputQueryClausesIntoMaster();
  void loadInputSupportClausesIntoMaster();
  void prepareMasterForInput(long timeLimit, 
			     float memoryLimit); 
  void prepareMasterForSaturation(long timeLimit, 
				  float memoryLimit); 
  void reportTermination(); 
  void reportTimeLimitExpired();
  static void setHardMemoryLimit(float lim); 
  static void setAllocationBufferSize(long percentage);
  static bool releaseMemory(ulong size);
  static bool releaseALotOfMemory();
  static void handleAllocationFailure(ulong size); 
  void reportAllocationFailure(ulong size); 
  TempLitListExtras::ComparisonMaker* 
  translate(const Options::LitListOrdering* ord);
  static void outputClause(const VK::Clause* cl);
  static void outputWeightChange(long oldWL,long newWL);
private:
  enum Continuation
  {
    Continuation_Init,
    Continuation_LoadInputQuery,
    Continuation_LoadInputSupport,
    Continuation_MainLoop
  };
  class TermWeightingWrapper
  {
  public:
    TermWeightingWrapper(const VampireKernel::Options::TermWeightingScheme& sc)
    {
      CALL("TermWeightingWrapper(const VampireKernel::Options::TermWeightingScheme& sc)");      
      switch (sc)
	{
	case VampireKernel::Options::UniformTermWeighting:
	  _uniformTermWeighting = new UniformTermWeighting();
	  TermWeighting::setCurrent(static_cast<TermWeighting*>(_uniformTermWeighting));
	  _nonuniformConstantTermWeighting = 0;
	  break;
	case VampireKernel::Options::NonuniformConstantTermWeighting:
	  _nonuniformConstantTermWeighting = new NonuniformConstantTermWeighting();
	  TermWeighting::setCurrent(static_cast<TermWeighting*>(_nonuniformConstantTermWeighting));
	  _uniformTermWeighting = 0;
	  break;
	};
    };
    ~TermWeightingWrapper()
    {
      if (DestructionMode::isThorough())
	{
	  if (_uniformTermWeighting) delete _uniformTermWeighting;
	  if (_nonuniformConstantTermWeighting) delete _nonuniformConstantTermWeighting;
	};
    };
  private:
    UniformTermWeighting* _uniformTermWeighting;
    NonuniformConstantTermWeighting* _nonuniformConstantTermWeighting;
  }; 
  class SimplificationOrderingWrapper
  {
  public:
    SimplificationOrderingWrapper(const VampireKernel::Options::SimplificationOrdering& so)
    {
      switch (so)
	{ 
	case VampireKernel::Options::StandardKBOrdering:
	  _standardKBOrdering = new StandardKBOrdering();
	  SimplificationOrdering::setCurrent(static_cast<SimplificationOrdering*>(_standardKBOrdering));
	  _lexPathOrdering = 0;
	  break;	
	case VampireKernel::Options::LexPathOrdering:
	  _lexPathOrdering = new LPO();
	  SimplificationOrdering::setCurrent(static_cast<SimplificationOrdering*>(_lexPathOrdering));
	  _standardKBOrdering = 0;
	  break;	
	};
    };
    ~SimplificationOrderingWrapper() 
    { 
      if (DestructionMode::isThorough())
	{
	  if (_standardKBOrdering) delete _standardKBOrdering;
	  if (_lexPathOrdering) delete _lexPathOrdering;
	};
    };
  private:
    StandardKBOrdering* _standardKBOrdering;
    LPO* _lexPathOrdering;
  }; 
  class NameIntroReceiver : public BK::Receiver<VK::Clause*>
  {
  public:
    NameIntroReceiver(_Implementation* host) : _hostKernelSession(host) {};
    void receive(VK::Clause* const & cl);
  private:
    _Implementation* _hostKernelSession;
  }; 
private:
  void outputInputSupportClauses();
  void outputInputQueryClauses();
  const VampireKernel::Statistics& statistics();
  void updateExportedStatistics();
  void outputStatistics(ostream* str,bool full);
  void autoStatOutput(ostream* str);
  const char* flagValue(bool flag);
  void switchBuiltinEqualityOff();
  void switchChainingForTransitiveOff();
  void analyseInput();
  void analyseOptions();
  static bool isRelevantBackgroundFlag(const VK::ClauseBackground::Rule& fl);
  static 
  VampireKernel::Clause::BackgroundFlag
  convertBackgroundFlag(const VK::ClauseBackground::Rule& fl);
  static 
  VK::ClauseBackground::Rule
  convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl);
private:  
  static VampireKernel::_Implementation* _currentSession;
  WeightPolynomial::Sharing _weightPolynomialSharing;
  TermWeightingWrapper _termWeightingWrapper;
  SimplificationOrderingWrapper _simplificationOrderingWrapper;
  VampireKernel* _wrapper;
  VampireKernel::Options _options;
  TempLitListExtras::ComparisonMaker* _compMakerForBasicComp;
  GlobalClock::Timer _timer;
  long _saturationTimeLimit;
  bool _saturationRunning;
  bool _mainLoopRunning;
  Signature _signature;
  ulong _currentInputClauseNumber;
  void* _currentInputClauseOrigin;
  BK::FIFO<BK::GlobAlloc,VK::Clause*> _inputSupportClauses;
  BK::FIFO<BK::GlobAlloc,VK::Clause*> _inputQueryClauses;
  ulong _numberOfSubmittedInputClauses;
  bool _inputClauseRejected;  
  ulong _numberOfRejectedInputClauses;
  bool _incompleteProcedure;
  ProblemProfile _problemProfile;
  VK::Clause::NewNumber _clauseNumberGenerator;
  VK::Clause::Context _clauseContext;
  VK::Clause::Context::AutoActivator _clauseContextActivator;
  TemporaryLiteral::Context _temporaryLiteralContext;
  TemporaryLiteral::Context::AutoActivator _temporaryLiteralContextActivator;
  TempLitListExtras::Context _temporaryLitListContext;
  TempLitListExtras::Context::AutoActivator _temporaryLitListContextActivator;
  TemporaryClause::Context _temporaryClauseContext;
  TemporaryClause::Context::AutoActivator _temporaryClauseContextActivator;
  BasicTempLitList::Context _basicTempLitListContext;
  BasicTempLitList::Context::AutoActivator _basicTempLitListContextActivator;
  ComponentName _componentNamingService;
  TermSharing _staticTermSharingService;
  LiteralSharing _staticLiteralSharing;
  ClauseAssembler _staticClauseAssembler;
  Master _master;
  VK::TemporaryClauseCollector _inputClauseCollector;
  BK::FIFO<BK::GlobAlloc,VK::Clause*> _inputHistory;
  ClauseSet _nameIntroClauses;
  NameIntroReceiver _nameIntroReceiver;
  ClauseNumIndex _numberToClauseMapping;
  VampireKernel::TerminationReason _terminationReason;
  bool _symbolWeightConstantPartInputIgnored;
  bool _unaryFunctionWithZeroWeightExists;
  ulong _unaryFunctionWithZeroWeight;
  Continuation _continuation;
  VK::Statistics _statistics;
  VK::Statistics _inputStatistics;
  VampireKernel::Statistics _exportedStatistics;
  bool _lookingForMoreAnswers;
  long _numberOfSupportClauses;
  bool _equalityInSupport;
  bool _clientEnforcedLiteralSelection;
  bool _clientRestrictedEquality;
  bool _clientSpecifiedPositiveEqualityOrientation;
  VK::Clause* _currentInputQueryClause;
  VK::Clause* _currentInputSupportClause;
  friend class VampireKernel;
  friend class VampireKernel::Clause;
  friend class VampireKernel::Clause::BackgroundFlagIterator;
  friend class NameIntroReceiver;
}; 
VampireKernel::_Implementation* 
VampireKernel::_Implementation::_currentSession;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation"
#endif    
#include "debugMacros.hpp"
void VampireKernel::_Implementation::init (const VampireKernel::Options& opt) 
{
  CALL("init(const VampireKernel::Options& opt)");
  BK::init(_weightPolynomialSharing);
  BK::init(_termWeightingWrapper,opt.term_weighting_scheme);
  BK::init(_simplificationOrderingWrapper,opt.simplification_ordering);
  _options = opt;
  _compMakerForBasicComp = 0;
  BK::init(_timer);
  _saturationRunning = false;
  _mainLoopRunning = false;
  BK::init(_signature);
  BK::init(_inputSupportClauses);
  BK::init(_inputQueryClauses);
  _incompleteProcedure = false;
  BK::init(_problemProfile);
  BK::init(_clauseNumberGenerator);
  BK::init(_clauseContext);
  BK::init(_clauseContextActivator,&_clauseContext);
  BK::init(_temporaryLiteralContext);
  BK::init(_temporaryLiteralContextActivator,&_temporaryLiteralContext); 
  BK::init(_temporaryLitListContext);
  BK::init(_temporaryLitListContextActivator,&_temporaryLitListContext); 
  BK::init(_temporaryClauseContext);
  BK::init(_temporaryClauseContextActivator,&_temporaryClauseContext); 
  BK::init(_basicTempLitListContext);
  BK::init(_basicTempLitListContextActivator,&_basicTempLitListContext);
  BK::init(_componentNamingService,&_staticClauseAssembler,&_nameIntroReceiver);
  BK::init(_staticTermSharingService);
  BK::init(_staticLiteralSharing,&_staticTermSharingService);
  BK::init(_staticClauseAssembler,&_staticLiteralSharing,static_cast<LiteralSelection*>(0));
  _master.init(&_componentNamingService);
  BK::init(_inputClauseCollector);
  BK::init(_inputHistory);
  BK::init(_nameIntroClauses,VK::Clause::MainSetNameIntro,"_nameIntroClauses");
  BK::init(_nameIntroReceiver,this);
  BK::init(_numberToClauseMapping);
  _terminationReason = VampireKernel::UnknownTerminationReason;
  _continuation = Continuation_Init;
  BK::init(_statistics);
  BK::init(_inputStatistics);
  BK::init(_exportedStatistics);
  _lookingForMoreAnswers = false;
  _numberOfSupportClauses = 0L;
  _clientEnforcedLiteralSelection = false;
  _clientRestrictedEquality = false;
  _clientSpecifiedPositiveEqualityOrientation = false;
  Signature::setCurrent(&_signature);
  _signature.setMaxlNumOfSkolemSymbols(_options.max_skolem);
  _master.setClauseNumberGenerator(&_clauseNumberGenerator);
  WeightPolynomial::setSharing(&_weightPolynomialSharing);
  _inputClauseCollector.setShowAnnouncedClauses(0);
  _inputClauseCollector.setShowCompleteClauses(0);
  _inputClauseCollector.setPreprocessingInProgress(false);
  _inputClauseCollector.setStatistics(&_inputStatistics);
  _inputClauseCollector.setTautologyCheck(false);
  _inputClauseCollector.setLiteralFactoring(false);
  if (opt.basic_comp_in_spl_is_max_wrt)
    _compMakerForBasicComp = translate(opt.basic_comp_in_spl_is_max_wrt);
}; 
void  
VampireKernel::_Implementation::destroy() 
{ 
  CALL("destroy()");
  if (DestructionMode::isThorough())
    {
      VK::Clause* clause;
	while (_inputSupportClauses.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_inputQueryClauses.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_inputHistory.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_nameIntroClauses.Nonempty())
	  {
	    clause = _nameIntroClauses.Pop();
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
	  };
    };
  BK::destroy(_exportedStatistics);
  BK::destroy(_inputStatistics);
  BK::destroy(_statistics);
  BK::destroy(_numberToClauseMapping);
  BK::destroy(_inputHistory);
  BK::destroy(_nameIntroReceiver);
  BK::destroy(_nameIntroClauses);
  BK::destroy(_inputClauseCollector);
  _master.destroy();
  BK::destroy(_staticClauseAssembler);
  BK::destroy(_staticLiteralSharing);
  BK::destroy(_staticTermSharingService);
  BK::destroy(_componentNamingService);
  BK::destroy(_basicTempLitListContextActivator);
  BK::destroy(_basicTempLitListContext);
  BK::destroy(_temporaryClauseContextActivator); 
  BK::destroy(_temporaryClauseContext);
  BK::destroy(_temporaryLitListContextActivator); 
  BK::destroy(_temporaryLitListContext);
  BK::destroy(_temporaryLiteralContextActivator); 
  BK::destroy(_temporaryLiteralContext);
  BK::destroy(_clauseContextActivator);
  BK::destroy(_clauseContext);
  BK::destroy(_clauseNumberGenerator);
  BK::destroy(_problemProfile);
  BK::destroy(_inputQueryClauses);
  BK::destroy(_inputSupportClauses);
  BK::destroy(_signature);
  BK::destroy(_timer);	
  BK::destroy(_simplificationOrderingWrapper);
  BK::destroy(_termWeightingWrapper);
  BK::destroy(_weightPolynomialSharing);
}; 
void  
VampireKernel::_Implementation::beginInputSignature() 
{
  _symbolWeightConstantPartInputIgnored = false;
  _unaryFunctionWithZeroWeightExists = false;
};
void  
VampireKernel::_Implementation::endInputSignature() 
{
  if (_symbolWeightConstantPartInputIgnored)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: input of some symbol weight constant parts ignored\n"
			     << "%          because of --term_weighting_scheme " 
			     << _options.term_weighting_scheme << "\n";
    };
  _signature.endOfInput();
};
void  
VampireKernel::_Implementation::registerInputFunction(const char* name,
						      ulong arity,
						      ulong& id,
						      const long& weight,
						      long precedence)
{
  id = _signature.inputSymbolNumber(false,name,arity);
  setSymbolWeightConstPart(id,weight);
  setSymbolPrecedence(id,precedence);
};
inline
void  
VampireKernel::_Implementation::registerInputPredicate(const char* name,
						       ulong arity,
						       bool isAnswer,
						       ulong& id,
						       const long& weight,
						       long precedence,
						       long positiveHeaderRank,
						       long negativeHeaderRank)
{
  id = _signature.inputSymbolNumber(true,name,arity);
  if (isAnswer) _signature.makeAnswerPredicate(id);
  setSymbolWeightConstPart(id,weight);
  setHeaderRank(true,id,positiveHeaderRank);
  setHeaderRank(false,id,negativeHeaderRank);
  setSymbolPrecedence(id,precedence); 
};
inline
void
VampireKernel::_Implementation::registerEquality(ulong& id,
						 long positiveHeaderRank,
						 long negativeHeaderRank)
{
  id = Signature::equalityId();
  setHeaderRank(true,id,positiveHeaderRank);
  setHeaderRank(false,id,negativeHeaderRank);
};
inline
void  
VampireKernel::_Implementation::setSymbolWeightConstPart(ulong symId,
							 const long& weight)
{
  if (_options.term_weighting_scheme == 
      VampireKernel::Options::UniformTermWeighting) 
    {
      return;
    };
  long w = weight;
  if (w == 0)
    {
      ulong arity = _signature.arity(symId);
      if (arity == 0UL)
	{
	  throw 
	    VampireKernel::InputError("Zero weight is assigned to a constant or a propositional variable.");
	}
      else
	if ((arity == 1UL) &&
	    (!_signature.isPredicate(symId)))
	  {
	    if (_unaryFunctionWithZeroWeightExists)
	      {
		throw 
		  VampireKernel::InputError("Zero weight is assigned to more than one unary function.");
	      }
	    else
	      {
		_unaryFunctionWithZeroWeightExists = true;
		_unaryFunctionWithZeroWeight = symId;
	      };	      
	  };
    };
  _signature.setSymbolWeightConstantPart(symId,w);
};
inline
void  
VampireKernel::_Implementation::setSymbolPrecedence(ulong symId,long prec)
{
  CALL("setSymbolPrecedence(ulong symId,long prec)");
  _signature.setInputSymbolPrecedence(symId,prec);
};
inline
void  
VampireKernel::_Implementation::setHeaderRank(bool positive,ulong symId,long prec)
{
  CALL("setHeaderRank(bool positive,ulong symId,long prec)");
  Signature::Polarity polarity = 
    (positive)? Signature::PositivePolarity : Signature::NegativePolarity;
  _signature.setInputHeaderRank(polarity,symId,prec);
}; 
void 
VampireKernel::_Implementation::registerInterpreted(bool predicate,ulong symId)
{
  CALL("registerInterpreted(bool predicate,ulong symId)");
  ASSERT(_signature.isPredicate(symId) == predicate);
  _signature.makeEvaluable(symId);
}; 
void  
VampireKernel::_Implementation::beginInputClauses()
{
  CALL("beginInputClauses()");
  ASSERT(_inputSupportClauses.empty());
  ASSERT(_inputQueryClauses.empty());
  _numberOfSubmittedInputClauses = 0UL;
  _numberOfRejectedInputClauses = 0UL;
  _numberOfSupportClauses = 0L;
  _clientEnforcedLiteralSelection = false;
  _clientRestrictedEquality = false;
  _clientSpecifiedPositiveEqualityOrientation = false;
  _equalityInSupport = false;
};
void  
VampireKernel::_Implementation::endInputClauses()
{
  CALL("endInputClauses()");
  if (_clientEnforcedLiteralSelection && _options.std_output)
    {	
      *_options.std_output << "% WARNING: some input clauses have preselected literals.\n" 
			   << "%          Incompleteness is not assumed!\n";
    };
  if (_clientRestrictedEquality && _options.std_output)
    {	
      *_options.std_output << "% WARNING: some (possibly all) equality literals have restricted semantics.\n"
			   << "%          Incompleteness is not assumed!\n";
    };  
  if (_clientSpecifiedPositiveEqualityOrientation && _options.std_output)
    {	
      *_options.std_output << "% WARNING: some equality literals have non-default orientation mode.\n"
			   << "%          Incompleteness is not assumed!\n";
    };
};
void  
VampireKernel::_Implementation::beginInputClause(ulong clauseNumber,
						 void* origin,
						 const ClauseAnnotation& annotation)
{
  CALL("beginInputClause(ulong clauseNumber,void* origin,const ClauseAnnotation& annotation)");
  ++_numberOfSubmittedInputClauses;
  _inputClauseRejected = false;
  _inputClauseCollector.openNewClause(ClauseBackground::Input);
  _inputClauseCollector.announceNewClause();
  if (annotation.isGoal()) _inputClauseCollector.markAsSubgoal();
  if (annotation.isSupport()) 
    {
      _inputClauseCollector.markAsSupport();
      ++_numberOfSupportClauses;        
    };
  if (annotation.isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
  if (annotation.isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
  _currentInputClauseNumber = clauseNumber;
  _currentInputClauseOrigin = origin;
};
void  
VampireKernel::_Implementation::endInputClause()
{
  CALL("endInputClause()");
  ObjectPool<TemporaryClause>::EntryHandle collectedClause;
  VK::Clause* clause;
  _inputClauseRejected = 
    _inputClauseRejected ||
    (!_inputClauseCollector.finaliseClause(collectedClause)) ||
    (!((clause = _staticClauseAssembler.assembleSharedWithoutSelection(*collectedClause))));
  if (_inputClauseRejected)
    {
      ++_numberOfRejectedInputClauses;
    }
  else
    {
      clause->SetNumber(_currentInputClauseNumber);
      clause->setInputClauseOrigin(_currentInputClauseOrigin);
      if (clause->isSupportClause())
	{
	  ulong pred;
	  if (_options.chaining_for_transitive &&
	      clause->isTransitivityLaw(pred) &&
	      !VampireKernelConst::isEquality(pred))
	    {
	      clause->markAsQuery();
	      clause->setMainSet(VK::Clause::MainSetInputQuery);
	      _inputQueryClauses.enqueue(clause);
	    }
	  else
	    {
	      if (!_equalityInSupport && clause->containsEquality())
		_equalityInSupport = true;
	      clause->setMainSet(VK::Clause::MainSetInputSupport);
	      _inputSupportClauses.enqueue(clause);
	    };
	}
      else
	{
	  clause->setMainSet(VK::Clause::MainSetInputQuery);
	  _inputQueryClauses.enqueue(clause);
	};
      if (clause->Empty())
	{
	  if (_options.std_output)
	    *_options.std_output << "% WARNING: empty input clause.\n";
	};
    };
}; 
bool VampireKernel::_Implementation::beginLiteral(bool positive,
						  ulong predicateId)
{
  CALL("beginLiteral(..)");
  if (_inputClauseRejected) return false;
  VK::LiteralProperties prop;
  Term::Polarity polarity = 
    (positive)? Signature::PositivePolarity : Signature::NegativePolarity;
  Term header((Term::Arity)_signature.arity(predicateId),
	      (Term::Functor)predicateId,
	      polarity);
  if (!_inputClauseCollector.openNewLiteral(header,prop))
    {
      _inputClauseRejected = true;
      return false;
    };
  return true;
}; 
void VampireKernel::_Implementation::markSelected() {
  CALL("markSelected()");
  _inputClauseCollector.currLitProperties().setSelectabilityCategory(LiteralProperties::HighSelectability);
  _clientEnforcedLiteralSelection = true;
}; 
void VampireKernel::_Implementation::markUnselected() {
  CALL("markUnselected()");
  _inputClauseCollector.currLitProperties().setSelectabilityCategory(LiteralProperties::LowSelectability);
  _clientEnforcedLiteralSelection = true;
};
void VampireKernel::_Implementation::markRuleHead() {
  CALL("markRuleHead()");
  _inputClauseCollector.currLitProperties().makeRuleHead();
  _clientEnforcedLiteralSelection = true;
};
void VampireKernel::_Implementation::markSubquery() {
  CALL("markSubquery()");
  _inputClauseCollector.currLitProperties().makeSubquery();
  _clientEnforcedLiteralSelection = true;
};
void VampireKernel::_Implementation::markEqWithResolution() {
  CALL("markEqWithResolution()");
  _inputClauseCollector.currLitProperties().setEqualityMode(LiteralProperties::UseEqForResolution);
};
void VampireKernel::_Implementation::markEqWithResolutionOnly() {
  CALL("markEqWithResolutionOnly()");
  _inputClauseCollector.currLitProperties().setEqualityMode(LiteralProperties::UseEqForResolutionOnly);
  _clientRestrictedEquality = true;
};
void VampireKernel::_Implementation::markGivenPosEqOrientation() {
  CALL("markGivenPosEqOrientation()");
  _inputClauseCollector.currLitProperties().setEqualityOrientationMode(LiteralProperties::GivenEqualityOrientation);
  _clientSpecifiedPositiveEqualityOrientation = true;
};
void VampireKernel::_Implementation::markReversePosEqOrientation() {
  CALL("markReversePosEqOrientation()");
  _inputClauseCollector.currLitProperties().setEqualityOrientationMode(LiteralProperties::ReverseGivenEqualityOrientation);
  _clientSpecifiedPositiveEqualityOrientation = true;
};
void  
VampireKernel::_Implementation::endLiteral()
{
  if (_inputClauseRejected) return;
  if (!_inputClauseCollector.closeLiteral())
    _inputClauseRejected = true;
};
inline
void  
VampireKernel::_Implementation::receiveFunction(ulong symId)
{
  CALL("receiveFunction(ulong symId)");
  if (_inputClauseRejected) return;
  Term symbol((Term::Arity)_signature.arity(symId),
	      (Term::Functor)symId,
	      Signature::PositivePolarity);
  if (!_inputClauseCollector.pushFunc(symbol))
    _inputClauseRejected = true;
}; 
inline
void 
VampireKernel::_Implementation::receiveInterpretedConstant(const InterpretedConstant& value) 
{
  CALL("receiveInterpretedConstant(const InterpretedConstant& value)");
  ulong id = _signature.interpretedConstantNumber(value);
  receiveFunction(id);
};
inline
void  
VampireKernel::_Implementation::receiveVariable(ulong varNum)
{
  CALL("receiveVariable(ulong varNum)");
  if (_inputClauseRejected) return;
  if (!_inputClauseCollector.pushVar(varNum))
    _inputClauseRejected = true;
};
void  
VampireKernel::_Implementation::saturate(ulong nextVacantClauseId,long timeLimit)
{
  CALL("saturate(ulong nextVacantClauseId,long timeLimit)");
  GlobalStopFlag::reset();
  _saturationRunning = true;
  _saturationTimeLimit = timeLimit;
  if (_saturationTimeLimit < 0L) _saturationTimeLimit = 0L;
  _mainLoopRunning = false;
  _timer.reset();
  _timer.start();
  GlobalStopFlag::HandlerActivator 
    globalStopFlagHandlerActivator(globalStopFlagHandler);
  GlobalClock::AlarmHandler alarmHandler(handleAlarm);
  _clauseNumberGenerator.reset(nextVacantClauseId);
  if (_options.std_output)
    *_options.std_output << "% New Vampire Kernel session, time limit: " 
			 << _saturationTimeLimit/10 << "." 
			 << _saturationTimeLimit%10 << " sec.\n";
  if (_numberOfRejectedInputClauses)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: " << _numberOfRejectedInputClauses 
			     << " input clauses rejected.\n";
      _incompleteProcedure = true;
    };
  if (_options.show_opt) 
    {	
      if (_options.std_output)
	{
	  *_options.std_output << "%================== Options: =======================\n";
	  *_options.std_output<< "% time limit = " << ((float)_saturationTimeLimit)/10 << "\n";
	  _options.output(*_options.std_output);
	  *_options.std_output << "%========== End of options: ======================\n";
	};
    };
  _clauseContext.setSelectedPosEqPenaltyCoefficient(_options.selected_pos_eq_penalty_coefficient);
  _clauseContext.setNongoalPenaltyCoefficient(_options.nongoal_penalty_coefficient);
  _clauseContext.setShowLiteralProperties(_options.show_lit_prop);
  if (_options.show_input && _options.std_output)
    {
      *_options.std_output << "%================== Input clauses (support + query): ===================\n";
      outputInputSupportClauses();
      outputInputQueryClauses();
      *_options.std_output << "%============== End of input clauses: ================\n";
    };
  if (_options.show_profile && _options.std_output)
    {
      _problemProfile.loadBackground(_inputSupportClauses);
      _problemProfile.loadBackground(_nameIntroClauses);
      _problemProfile.load(_inputQueryClauses);
      _problemProfile.summarise();
      *_options.std_output << "%========= Profile of input (support + query):\n"; 
      _problemProfile.output(*_options.std_output);
      *_options.std_output << "%========= End of profile. ========\n";
      *_options.std_output << _signature;
    };
  GlobalStopFlag::check();
  inputToMasterAndRunMainLoop();
}; 
void  
VampireKernel::_Implementation::inputToMasterAndRunMainLoop()
{
  CALL("inputToMasterAndRunMainLoop()");    
  long remainingTime;
  float remainingMemory;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_LoadInputQuery:
      if (_options.std_output)
	*_options.std_output << "% Resuming loading input clauses...\n";
      goto load_input_query;
    case Continuation_LoadInputSupport:
      if (_options.std_output)
	*_options.std_output << "% Resuming loading input clauses...\n";
      goto load_input_support;
    case Continuation_MainLoop:
      if (_options.std_output)
	*_options.std_output << "% Resuming main loop...\n";
      goto main_loop;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::inputToMasterAndRunMainLoop()");
    };
  _continuation = Continuation_Init;
  _signature.setCommutativity(_problemProfile.commutativityTable(),
			      _problemProfile.symmetryTable());
  _signature.makeCommutative(VampireKernelConst::UnordEqNum);
  _signature.makeCommutative(VampireKernelConst::OrdEqNum);
  _problemProfile.loadBackground(_inputSupportClauses);
  _problemProfile.load(_inputQueryClauses);
  _problemProfile.summarise();
  if (_options.show_profile && _options.std_output)
    {
      *_options.std_output << "%========= Profile of query clause set:\n"; 
      _problemProfile.output(*_options.std_output);
      *_options.std_output << "%========= End of profile for query clause set. ========\n";
    }; 
  if (!_problemProfile.numOfEqLiterals() && !_equalityInSupport)
    switchBuiltinEqualityOff();
  if (!_problemProfile.numOfTransitive())
    {
      switchChainingForTransitiveOff();
    };
  GlobalStopFlag::check();
  remainingTime = _saturationTimeLimit - _timer.elapsedDeciseconds();
  remainingMemory = 
    (((float)_options.memory_limit)*1024) - GlobAlloc::occupiedByObjects();
  if (remainingMemory < 0.0) remainingMemory = 0.0;
  analyseInput();
  if (_options.input_non_basicness && 
      _options.show_profile && 
      _options.std_output)
    {
      if (_signature.nonBasicArgumentExists())
	{
	  *_options.std_output << "% Functions/headers with input-non-basic arguments:\n";
	  _signature.outputNonBasicArguments(*_options.std_output);
	}
      else
	{
	  *_options.std_output << "% No non-basic arguments in the input.\n";
	};
    };
  analyseOptions();
  remainingTime = _saturationTimeLimit - _timer.elapsedDeciseconds();
  remainingMemory = 
    (((float)_options.memory_limit)*1024) - GlobAlloc::occupiedByObjects();
  if (remainingMemory < 0.0) remainingMemory = 0.0;
  prepareMasterForInput(remainingTime,remainingMemory);
 load_input_support:
  loadInputSupportClausesIntoMaster();
  if (_master.terminated())
    {
      reportTermination();
      _timer.stop();
      _saturationRunning = false; 
      _continuation = Continuation_LoadInputSupport;
      return;
    };
  _continuation = Continuation_Init;
 load_input_query:
  loadInputQueryClausesIntoMaster();
  if (_master.terminated())
    {
      reportTermination();
      _timer.stop();
      _saturationRunning = false; 
      _continuation = Continuation_LoadInputQuery;
      return;
    };
  _continuation = Continuation_Init;
  _master.endOfInput();
  remainingTime = _saturationTimeLimit - _timer.elapsedDeciseconds();
  remainingMemory = 
    (((float)_options.memory_limit)*1024) - GlobAlloc::occupiedByObjects();
  if (remainingMemory < 0.0) remainingMemory = 0.0;
  prepareMasterForSaturation(remainingTime,remainingMemory);
  GlobalStopFlag::check();
 main_loop:
  _mainLoopRunning = true;
  _master.mainLoop();
  _mainLoopRunning = false;
  reportTermination();
  _timer.stop();
  _saturationRunning = false; 
  _continuation = Continuation_MainLoop;
}; 
void VampireKernel::_Implementation::lookForAnotherAnswer(int timeLimit)
{
  CALL("lookForAnotherAnswer()");
  ASSERT(refutation());
  ASSERT(!_saturationRunning);
  GlobalStopFlag::reset();
  _lookingForMoreAnswers = true;
  _timer.reset();
  _timer.start();
  _saturationTimeLimit = timeLimit;
  _saturationRunning = true;
  GlobalStopFlag::HandlerActivator 
    globalStopFlagHandlerActivator(globalStopFlagHandler);
  GlobalClock::AlarmHandler alarmHandler(handleAlarm);
  switch (_continuation)
    {
    case Continuation_LoadInputQuery:
    case Continuation_LoadInputSupport:
    case Continuation_MainLoop:
      if (_options.std_output)
	*_options.std_output << "% Resuming saturation in attempt to find another answer...\n";
      inputToMasterAndRunMainLoop();
      break;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::lookForAnotherAnswer()");
    };
}; 
void  
VampireKernel::_Implementation::suspend()
{
  if (_saturationRunning)
    {
      _master.suspend();
      _timer.stop();
    };
  if (_options.std_output)
    {
      *_options.std_output << "% Kernel session suspended.\n";   
      autoStatOutput(_options.std_output);
    };
};
void  
VampireKernel::_Implementation::resume()
{
  if (_options.std_output)
    *_options.std_output << "% Kernel session resumed.\n";
  if (_saturationRunning)
    {
      _timer.start();
      _master.resume();
    };
}; 
void  
VampireKernel::_Implementation::reportInterrupt(int signalNumber)
{
  CALL("reportInterrupt(int signalNumber)");
  _terminationReason = VampireKernel::AbortedOnExternalInterrupt;
  if (_options.std_output)
    {
      autoStatOutput(_options.std_output);
      *_options.std_output << "% Proof not found. Aborted on external signal " 
			   << signalNumber << ".\n";
      _options.std_output->flush();
    };
}; 
void  
VampireKernel::_Implementation::reportErrorSignal(int signalNumber)
{
  CALL("reportErrorSignal(int signalNumber)");
  _terminationReason = VampireKernel::AbortedOnErrorSignal;
  if (_options.std_output)
    *_options.std_output << "% Proof not found. Kernel session aborted on error signal " 
			 << signalNumber << ".\n";
}; 
void  
VampireKernel::_Implementation::reportError()
{
  CALL("reportError()");
  _terminationReason = VampireKernel::AbortedOnError;
  if (_options.std_output)
    *_options.std_output << "% Proof not found. Kernel session aborted due to an error.\n";
}; 
VampireKernel::TerminationReason  
VampireKernel::_Implementation::terminationReason() const
{
  return _terminationReason;
};
const VampireKernel::Clause*  
VampireKernel::_Implementation::refutation() const
{
  return static_cast<const VampireKernel::Clause*>(static_cast<const void*>(_master.proof()));
};
VampireKernel::ClauseSetIterator*  
VampireKernel::_Implementation::activeClauses()
{
  CALL("activeClauses()");
  return 
    static_cast<VampireKernel::ClauseSetIterator*>(new ClauseSetIterator(_master.activeClauses()));
};
bool  
VampireKernel::_Implementation::isInterpretedConstantId(ulong symbolId) const
{
  return _signature.isInterpretedConstant(symbolId);
};
bool  
VampireKernel::_Implementation::isInputSymbol(ulong symbolId) const
{
  return !_signature.isSkolem(symbolId);
}; 
bool  
VampireKernel::_Implementation::isAnswerPredicateId(ulong symbolId) const
{
  return _signature.isAnswerPredicate(symbolId);
};
ulong  
VampireKernel::_Implementation::symbolArity(ulong symbolId) const
{
  return _signature.arity(symbolId);
};
const char*  
VampireKernel::_Implementation::symbolPrintName(ulong symbolId) const
{
  CALL("symbolPrintName(ulong symbolId) const");
  return _signature.symbolName(symbolId);
}; 
const InterpretedConstant& 
VampireKernel::_Implementation::interpretedConstantValue(ulong symbolId) const
{
  CALL("interpretedConstantValue(ulong symbolId) const");
  ASSERT(isInterpretedConstantId(symbolId));
  return _signature.interpretedConstantValue(symbolId);
};
void  
VampireKernel::_Implementation::handleAlarm(int)
{
  if (_currentSession && _currentSession->_saturationRunning)
    {	
      if (_currentSession->_timer.elapsedDeciseconds() > _currentSession->_saturationTimeLimit)
	{
	  GlobalStopFlag::raise();
	}
      else
	{
	  if (_currentSession->_mainLoopRunning)
	    {
	      _currentSession->_master.externalTick();
	    };
	};
    };
};
void  
VampireKernel::_Implementation::globalStopFlagHandler()
{
  CALL("globalStopFlagHandler()");
  if (_currentSession) 
    { 
      _currentSession->reportTimeLimitExpired();
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::TimeLimit);
    };
}; 
void  
VampireKernel::_Implementation::loadInputQueryClausesIntoMaster()
{
  CALL("loadInputQueryClausesIntoMaster()");
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_LoadInputQuery:
      goto load_clause;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::loadInputQueryClausesIntoMaster()");
    };
 next_clause:
  if (_inputQueryClauses.dequeue(_currentInputQueryClause))
    {
      _currentInputQueryClause->setMainSet(VK::Clause::MainSetInputHistory);
      _inputHistory.enqueue(_currentInputQueryClause);
    load_clause:
      _master.inputClause(_currentInputQueryClause);
      if (_master.terminated()) 
	{
	  return;
	};
      _continuation = Continuation_Init;	
      goto next_clause;
    };
  ASSERT(_inputQueryClauses.empty());
}; 
void  
VampireKernel::_Implementation::loadInputSupportClausesIntoMaster()
{
  CALL("loadInputSupportClausesIntoMaster()");
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_LoadInputSupport:
      goto load_clause;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::loadInputSupportClausesIntoMaster()");
    };
 next_clause:
  if (_inputSupportClauses.dequeue(_currentInputSupportClause))
    {
      _currentInputSupportClause->setMainSet(VK::Clause::MainSetInputHistory);
      _inputHistory.enqueue(_currentInputSupportClause);
    load_clause:
      _master.inputClause(_currentInputSupportClause);
      if (_master.terminated()) 
	{
	  return;
	};
      _continuation = Continuation_Init;
      goto next_clause;
    };
  ASSERT(_inputSupportClauses.empty());
}; 
void  
VampireKernel::_Implementation::prepareMasterForSaturation(long timeLimit, 
							   float memoryLimit) 
{
  _master.options.incompleteAPriori = _incompleteProcedure;
  _master.options.main_alg = _options.main_alg;
  _master.options.binary_resolution = true;
  _master.options.no_inferences_between_rules = _options.no_inferences_between_rules;
  _master.options.splitting = _options.dynamic_splitting;  
  _master.options.small_basic_comp_in_hyp_splitting = _options.small_basic_comp_in_hyp_splitting;
  _master.options.fold_single_comp_in_splitting = _options.fold_single_comp_in_splitting;
  _master.options.single_ground_lit_comp_min_size = _options.single_ground_lit_comp_min_size;
  _master.options.reuse_names_for_negated_ground_lits_in_splitting = _options.reuse_names_for_negated_ground_lits_in_splitting;
  _master.options.basic_comp_in_spl_is_max_wrt = _compMakerForBasicComp;
  _master.options.back_fold_in_spl = _options.back_fold_in_spl;
  _master.options.neg_eq_splitting = _options.dynamic_neg_eq_splitting; 
  _master.options.neg_eq_spl_var_count_limit = _options.dyn_neg_eq_spl_var_count_limit; 
  _master.options.paramodulation = _options.paramodulation;
  _master.options.paramodulate_from_variables = _options.paramodulate_from_variables;
  _master.options.sim_back_sup = _options.sim_back_sup;
  _master.options.forw_sup_no_dup_red = _options.forw_sup_no_dup_red;
  _master.options.back_sup_no_dup_red = _options.back_sup_no_dup_red;
  _master.options.prohibit_non_basic_superposition = 
    _options.prohibit_non_basic_superposition;
  _master.options.prohibit_non_basic_fd = 
    _options.prohibit_non_basic_fd;
  _master.options.prohibit_non_basic_bd = 
    _options.prohibit_non_basic_bd;
  _master.options.reduce_non_basic = 
    _options.reduce_non_basic;
  _master.options.input_non_basicness = 
    _options.input_non_basicness;
  _master.options.run_time_non_basicness = 
    _options.run_time_non_basicness;
  _master.options.selection = _options.selection;
  _master.options.inherited_negative_selection = _options.inherited_negative_selection;
  _master.options.external_evaluation = 
    _options.external_evaluation &&
    (_signature.someSymbolsAreEvaluable() ||
     _signature.someConstantsAreInterpreted());
  _master.options.handle_equality = _options.paramodulation;
  _master.options.forward_subsumption = _options.forward_subsumption;
  _master.options.simplify_by_fs = _options.simplify_by_fs;
  _master.options.main_premise_subsumption = _options.main_premise_subsumption;
  _master.options.side_premise_subsumption = _options.side_premise_subsumption;
  _master.options.main_selected_lit_demodulation = _options.main_selected_lit_demodulation;
  _master.options.side_selected_lit_demodulation = _options.side_selected_lit_demodulation;
  _master.options.forward_demodulation = _options.forward_demodulation;
  _master.options.fd_on_splitting_branches = _options.fd_on_splitting_branches;
  if (_options.normalize_by_commutativity)
    {
      _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
    }
  else
    _master.options.normalize_by_commutativity = 0;
  if (_options.eq_taut_mod_C)
    {
      _master.options.eq_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.eq_taut_mod_C = 0;
  if (_options.prop_taut_mod_C)
    {
      _master.options.prop_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.prop_taut_mod_C = 0;
  if (_options.trivial_eq_res_mod_C)
    {
      _master.options.trivial_eq_res_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.trivial_eq_res_mod_C = 0;
  if (_options.identical_lit_factoring_mod_C)
    {
      _master.options.identical_lit_factoring_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.identical_lit_factoring_mod_C = 0;
  if (_options.chaining_for_transitive)
    {
      _master.options.chaining_for_transitive = _problemProfile.transitivityTable();
    }
  else
    _master.options.chaining_for_transitive = 0;
  _master.options.chaining_through_variables = _options.chaining_through_variables;
  _master.options.backward_demodulation = _options.backward_demodulation;
  _master.options.bd_on_splitting_branches = _options.bd_on_splitting_branches;
  _master.options.simplify_by_eq_res = _options.simplify_by_eq_res;
  _master.options.backward_subsumption = _options.backward_subsumption;
  _master.options.orphan_murder = _options.orphan_murder;
  _master.options.lrs = _options.lrs;
  _master.options.orient_pos_equalities_in_support = _options.orient_pos_equalities_in_support;
  _master.options.orient_neg_equalities_in_support = _options.orient_neg_equalities_in_support;
  _master.options.age_weight_ratio_numerator = _options.age_weight_ratio_numerator;
  _master.options.age_weight_ratio_denominator = _options.age_weight_ratio_denominator;
  _master.options.best_good_weight_ratio_numerator = _options.best_good_weight_ratio_numerator;
  _master.options.best_good_weight_ratio_denominator = _options.best_good_weight_ratio_denominator;
  _master.options.timeLimit = timeLimit;
  _master.options.memoryLimit = memoryLimit;
  _master.options.allocation_buffer_price = (float)_options.allocation_buffer_price;
  _master.options.maxActive = _options.max_active;
  _master.options.maxPassive = _options.max_passive;
  _master.options.maxWeight = _options.max_weight;
  _master.options.maxInferenceDepth = _options.max_inference_depth;
  _master.options.simulatedTimeLimit = _options.simulated_time_limit ?
                                       _options.simulated_time_limit :
                                       timeLimit;
  _master.options.first_time_check = ((float)_options.first_time_check)/100;
  _master.options.first_mem_check = ((float)_options.first_mem_check)/100;
  _master.options.outputStreamForGenerated = (_options.show_gen) ? _options.std_output : 0;
  _master.options.showPassive = (_options.show_passive) ? outputClause : 0;
  _master.options.showActive = (_options.show_active) ? outputClause : 0;
  _master.options.showLitProp = _options.show_lit_prop;
  _master.options.outputStreamForPassiveProfiles = _options.std_output;
  _master.options.showPassiveProfileInterval = _options.show_passive_profile_interval;
  _master.options.showWeightChange = (_options.show_weight_limit_changes) ? outputWeightChange : 0;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _master.options.step_resolution_literal_selection = _options.step_resolution_literal_selection;
#endif
  _master.setStrategy();
}; 
void  
VampireKernel::_Implementation::prepareMasterForInput(long timeLimit, 
						      float memoryLimit) 
{
  _master.reset();
  _master.options.incompleteAPriori = _incompleteProcedure;
  _master.options.main_alg = _options.main_alg;
  _master.options.binary_resolution = true;
  _master.options.no_inferences_between_rules = _options.no_inferences_between_rules;
  _master.options.splitting = _options.static_splitting;  
  _master.options.small_basic_comp_in_hyp_splitting = _options.small_basic_comp_in_hyp_splitting;
  _master.options.fold_single_comp_in_splitting = _options.fold_single_comp_in_splitting;
  _master.options.single_ground_lit_comp_min_size = _options.single_ground_lit_comp_min_size;
  _master.options.reuse_names_for_negated_ground_lits_in_splitting = _options.reuse_names_for_negated_ground_lits_in_splitting;
  _master.options.basic_comp_in_spl_is_max_wrt = _compMakerForBasicComp;
  _master.options.back_fold_in_spl = _options.back_fold_in_spl;
  _master.options.neg_eq_splitting = _options.static_neg_eq_splitting; 
  _master.options.neg_eq_spl_var_count_limit = _options.dyn_neg_eq_spl_var_count_limit; 
  _master.options.paramodulation = _options.paramodulation;
  _master.options.paramodulate_from_variables = _options.paramodulate_from_variables;
  _master.options.sim_back_sup = _options.sim_back_sup;
  _master.options.forw_sup_no_dup_red = _options.forw_sup_no_dup_red;
  _master.options.back_sup_no_dup_red = _options.back_sup_no_dup_red;
  _master.options.prohibit_non_basic_superposition = 
    _options.prohibit_non_basic_superposition;
  _master.options.prohibit_non_basic_fd = 
    _options.prohibit_non_basic_fd;
  _master.options.prohibit_non_basic_bd = 
    _options.prohibit_non_basic_bd;
  _master.options.reduce_non_basic = 
    _options.reduce_non_basic;
  _master.options.input_non_basicness = 
    _options.input_non_basicness;
  _master.options.run_time_non_basicness = 
    _options.run_time_non_basicness;
  _master.options.selection = _options.selection;
  _master.options.inherited_negative_selection = _options.inherited_negative_selection;
  _master.options.external_evaluation = 
    _options.external_evaluation &&
    (_signature.someSymbolsAreEvaluable() ||
     _signature.someConstantsAreInterpreted());
  _master.options.handle_equality = _options.paramodulation;
  _master.options.forward_subsumption = _options.forward_subsumption;
  _master.options.simplify_by_fs = _options.simplify_by_fs;
  _master.options.main_premise_subsumption = _options.main_premise_subsumption;
  _master.options.side_premise_subsumption = _options.side_premise_subsumption;
  _master.options.main_selected_lit_demodulation = _options.main_selected_lit_demodulation;
  _master.options.side_selected_lit_demodulation = _options.side_selected_lit_demodulation;
  _master.options.forward_demodulation = _options.forward_demodulation;
  _master.options.fd_on_splitting_branches = _options.fd_on_splitting_branches;
  if (_options.normalize_by_commutativity)
    {
      _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
    }
  else
    _master.options.normalize_by_commutativity = 0;
  if (_options.eq_taut_mod_C)
    {
      _master.options.eq_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.eq_taut_mod_C = 0;
  if (_options.prop_taut_mod_C)
    {
      _master.options.prop_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.prop_taut_mod_C = 0;
  if (_options.trivial_eq_res_mod_C)
    {
      _master.options.trivial_eq_res_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.trivial_eq_res_mod_C = 0;
  if (_options.identical_lit_factoring_mod_C)
    {
      _master.options.identical_lit_factoring_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.identical_lit_factoring_mod_C = 0;
  if (_options.chaining_for_transitive)
    {
      _master.options.chaining_for_transitive = _problemProfile.transitivityTable();
    }
  else
    _master.options.chaining_for_transitive = 0;
  _master.options.chaining_through_variables = _options.chaining_through_variables;
  _master.options.backward_demodulation = _options.backward_demodulation;
  _master.options.bd_on_splitting_branches = _options.bd_on_splitting_branches;
  _master.options.simplify_by_eq_res = _options.simplify_by_eq_res;
  _master.options.backward_subsumption = _options.backward_subsumption;
  _master.options.orphan_murder = _options.orphan_murder;
  _master.options.lrs = _options.lrs;
  _master.options.orient_pos_equalities_in_support = _options.orient_pos_equalities_in_support;
  _master.options.orient_neg_equalities_in_support = _options.orient_neg_equalities_in_support;
  _master.options.age_weight_ratio_numerator = _options.age_weight_ratio_numerator;
  _master.options.age_weight_ratio_denominator = _options.age_weight_ratio_denominator;
  _master.options.best_good_weight_ratio_numerator = _options.best_good_weight_ratio_numerator;
  _master.options.best_good_weight_ratio_denominator = _options.best_good_weight_ratio_denominator;
  _master.options.timeLimit = timeLimit;
  _master.options.memoryLimit = memoryLimit;
  _master.options.allocation_buffer_price = (float)_options.allocation_buffer_price;
  _master.options.maxActive = _options.max_active;
  _master.options.maxPassive = _options.max_passive;
  _master.options.maxWeight = _options.max_weight;
  _master.options.maxInferenceDepth = _options.max_inference_depth;
  _master.options.simulatedTimeLimit = _options.simulated_time_limit ?
                                       _options.simulated_time_limit :
                                       timeLimit;
  _master.options.first_time_check = ((float)_options.first_time_check)/100;
  _master.options.first_mem_check = ((float)_options.first_mem_check)/100;
  _master.options.outputStreamForGenerated = (_options.show_gen) ? _options.std_output : 0;
  _master.options.showPassive = (_options.show_passive) ? outputClause : 0;
  _master.options.showActive = (_options.show_active) ? outputClause : 0;
  _master.options.showLitProp = _options.show_lit_prop;
  _master.options.outputStreamForPassiveProfiles = _options.std_output;
  _master.options.showPassiveProfileInterval = _options.show_passive_profile_interval;
  _master.options.showWeightChange = (_options.show_weight_limit_changes) ? outputWeightChange : 0;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _master.options.step_resolution_literal_selection = _options.step_resolution_literal_selection;
#endif
  _master.setStrategy();
}; 
void  
VampireKernel::_Implementation::reportTermination()
{
  CALL("reportTermination()");
  autoStatOutput(_options.std_output);
  switch (_master.terminationReason())
    {
    case Master::UnknownTR:
      ICP("ICP0");
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Internal error: unknown termination type.\n"; 
      _terminationReason = VampireKernel::UnknownTerminationReason;
      break;  
    case Master::Proof: 
      if (_options.std_output)
	*_options.std_output << "% Proof found. Thanks to Tanya!\n"; 
      _terminationReason = VampireKernel::RefutationFound;
      break; 
    case Master::EmptyPassive: 
      if (_master.completeStrategyUsed() && 
	  (!_lookingForMoreAnswers))
	{
	  if (_options.std_output)
	    *_options.std_output << "% Unprovable.\n";  
	  _terminationReason = VampireKernel::Satisfiable;
	}
      else 
	{ 
	  if (_options.std_output)
	    *_options.std_output << "% Proof not found. Empty passive.\n"; 
	  _terminationReason = VampireKernel::EmptyPassive;
	};
      break;
    case Master::MaxPassive: 
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Max. num. of passive reached. \n";
      _terminationReason = VampireKernel::MaxPassive;
      break; 
    case Master::MaxActive: 
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Max. num. of active reached. \n"; 
      _terminationReason = VampireKernel::MaxActive;
      break; 
    }; 
  if (_options.std_output) _options.std_output->flush();
}; 
void  
VampireKernel::_Implementation::reportTimeLimitExpired()
{
  CALL("reportTimeLimitExpired()");  
  _terminationReason = VampireKernel::TimeLimitExpired;
  if (_saturationRunning)
    {
      autoStatOutput(_options.std_output);
    };
  if (_options.std_output)
    {
      *_options.std_output << "% Proof not found. Time limit expired.\n";
      _options.std_output->flush();
    };
}; 
void  
VampireKernel::_Implementation::setHardMemoryLimit(float lim) 
{
  CALL("setHardMemoryLimit(float lim)");
  if (lim < 0.0) lim = 0.0;
  long truncated;
  if (lim > (LONG_MAX/2))
    {
      truncated = LONG_MAX/2;
    }
  else
    truncated = (long)lim;
  GlobAlloc::setHardMemoryLimit(truncated);
  GlobAlloc::basicAllocationFailureHandling().setMemoryReleaseRequest(releaseMemory);
  GlobAlloc::basicAllocationFailureHandling().setMassiveMemoryReleaseRequest(releaseALotOfMemory);
  GlobAlloc::basicAllocationFailureHandling().setUnrecoverableAllocationFailureHandler(handleAllocationFailure);
}; 
void  
VampireKernel::_Implementation::setAllocationBufferSize(long percentage)
{
  GlobAlloc::setAllocationBufferSize(percentage);
}; 
bool  
VampireKernel::_Implementation::releaseMemory(ulong size) 
{
  CALL("releaseMemory(ulong size)");
  if (_currentSession && (_currentSession->_mainLoopRunning))
    {
      return _currentSession->_master.releaseMemory(size);
    }
  else
    return false;
}; 
bool  
VampireKernel::_Implementation::releaseALotOfMemory()
{
  CALL("releaseALotOfMemory()");
  if (_currentSession && (_currentSession->_mainLoopRunning))
    {
      return _currentSession->_master.releaseALotOfMemory();
    }
  else
    return false;
}; 
void  
VampireKernel::_Implementation::handleAllocationFailure(ulong size) 
{
  CALL("handleAllocationFailure(ulong size)");
  if (_currentSession) 
    {
      _currentSession->reportAllocationFailure(size);
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::AllocationFailure);
    }
  else
    {
      cout << "% Allocation failure occured when no kernel session is running.\n";
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::AllocationFailure);
    };
}; 
void  
VampireKernel::_Implementation::reportAllocationFailure(ulong size)
{
  CALL("reportAllocationFailure(ulong size)");
  _terminationReason = VampireKernel::AbortedOnAllocationFailure;
  if (_options.std_output)
    {
      *_options.std_output << "% ------------- ALLOCATION FAILURE: " 
			   << size << " bytes -----------------\n";
      autoStatOutput(_options.std_output);
      *_options.std_output << "% Proof not found. Allocation failure.\n";
      _options.std_output->flush();
    };
}; 
TempLitListExtras::ComparisonMaker* 
VampireKernel::_Implementation::translate(const Options::LitListOrdering* ord)
{
  CALL("translate(const Options::LitListOrdering* ord)");
  ASSERT(ord);
  Options::ElementaryLitListOrdering elTag;
  const Options::LitListOrdering* ord1;
  const Options::LitListOrdering* ord2;
  if (ord->isElementary(elTag))
    {
      TempLitListExtras::ComparableCharacteristic* c1;
      TempLitListExtras::ComparableCharacteristic* c2;
      switch (elTag)
	{
	case Options::BySize:
	  c1 = new TempLitListExtras::Size();
	  c2 = new TempLitListExtras::Size();
	  break;
	case Options::ByNumberOfVarOccurences:
	  c1 = new TempLitListExtras::NumOfVarOccurences();
	  c2 = new TempLitListExtras::NumOfVarOccurences();
	  break;
	case Options::ByNumberOfLiterals:
	  c1 = new TempLitListExtras::NumOfLiterals();
	  c2 = new TempLitListExtras::NumOfLiterals();
	  break;
	};
      return 
	new TempLitListExtras::CharacteristicComparisonMaker(c1,c2);
    }
  else if (ord->isInverse(ord1))
    {
      return
	new TempLitListExtras::ComplexComparisonMaker(translate(ord1));
    }
  else
    {
      bool comp = ord->isComposition(ord1,ord2);
      ASSERT(comp);
      return
	new TempLitListExtras::ComplexComparisonMaker(translate(ord1),
						      translate(ord2));
    };
}; 
void  
VampireKernel::_Implementation::outputClause(const VK::Clause* cl)
{
  if (_currentSession && (_currentSession->_options.std_output))
    *(_currentSession->_options.std_output) << "% " << cl << "\n";
};
void  
VampireKernel::_Implementation::outputWeightChange(long oldWL,long newWL)
{
  if (_currentSession && (_currentSession->_options.std_output))
    *(_currentSession->_options.std_output) << "% WEIGHT LIMIT CHANGED: " << oldWL << " ------> " << newWL << "\n";
};
void  
VampireKernel::_Implementation::outputInputSupportClauses()
{
  if (!_options.std_output) return;
  for (const BK::FIFO<BK::GlobAlloc,VK::Clause*>::Entry* cl = _inputSupportClauses.begin();
       cl;
       cl = cl->next())
    {
      *_options.std_output << "% " << cl->value() << "\n";
    };
};
void  
VampireKernel::_Implementation::outputInputQueryClauses()
{
  if (!_options.std_output) return;
  for (const BK::FIFO<BK::GlobAlloc,VK::Clause*>::Entry* cl = _inputQueryClauses.begin();
       cl;
       cl = cl->next())
    {
      *_options.std_output << "% " << cl->value() << "\n";
    };
};
const VampireKernel::Statistics&
VampireKernel::_Implementation::statistics()
{
  updateExportedStatistics();
  return _exportedStatistics;
};
void VampireKernel::_Implementation::updateExportedStatistics()
{
  _statistics.reset();
  _statistics.absorb(_inputStatistics);
  _statistics.absorb(_master.statistics());
  _exportedStatistics.time = _timer.elapsedSeconds();
  _exportedStatistics.memory = GlobAlloc::occupiedByChunks();
  VampireKernel::Statistics::DeductionInferences& dedInf = 
    _exportedStatistics.deductionInferences;
  VampireKernel::Statistics::SimplifyingInferences& simInf = 
    _exportedStatistics.simplifyingInferences;
  VampireKernel::Statistics::PropositionalSplitting& propSpl =
    _exportedStatistics.propositionalSplittingInferences;
  VampireKernel::Statistics::NegativeEqualitySplitting& negEqSpl = 
    _exportedStatistics.negativeEqualitySplittingInferences;
  VampireKernel::Statistics::GeneratedClauses& genCl = 
    _exportedStatistics.generatedClauses;
  VampireKernel::Statistics::RetainedClauses& retCl = 
    _exportedStatistics.retainedClauses;
  dedInf.totalAccomplished = 
    _statistics.inferences.deduction.totalAccomplished();
  dedInf.resolutionAccomplished = 
    _statistics.inferences.deduction.resolution.accomplished;
  dedInf.forwardSuperposition = 
    _statistics.inferences.deduction.forwardSuperposition.accomplished;
  dedInf.backwardSuperposition = 
    _statistics.inferences.deduction.backwardSuperposition.accomplished;
  dedInf.equalityFactoring = 
    _statistics.inferences.deduction.equalityFactoring.accomplished;
  dedInf.equalityResolution = 
    _statistics.inferences.deduction.equalityResolution.accomplished;
  dedInf.forwardChaining = 
    _statistics.inferences.deduction.forwardChaining.accomplished;
  dedInf.backwardChaining = 
    _statistics.inferences.deduction.backwardChaining.accomplished;
  dedInf.totalRedundantDiscarded = 
    _statistics.inferences.deduction.totalRedundantDiscarded();
  dedInf.totalRedundantBranchesDiscarded = 
    _statistics.inferences.deduction.totalRedundantBranchesDiscarded();
  dedInf.totalNonredundantDiscarded = 
    _statistics.inferences.deduction.totalNonredundantDiscarded();
  dedInf.totalNonredundantBranchesDiscarded = 
    _statistics.inferences.deduction.totalNonredundantBranchesDiscarded();
  simInf.totalForward = 
    _statistics.inferences.simplifying.totalForward();
  simInf.propositionalTautologyDeletion = 
    _statistics.inferences.simplifying.propositionalTautologyDeletion;
  simInf.equationalTautologyDeletion = 
    _statistics.inferences.simplifying.equationalTautologyDeletion;
  simInf.forwardSubsumption = 
    _statistics.inferences.simplifying.forwardSubsumption;
  simInf.forwardSubsumptionResolution = 
    _statistics.inferences.simplifying.forwardSubsumptionResolution;
  simInf.forwardDemodulation = 
    _statistics.inferences.simplifying.forwardDemodulation;
  simInf.onBranchForwardDemodulation = 
    _statistics.inferences.simplifying.onBranchForwardDemodulation;
  simInf.destructiveEqualityResolution = 
    _statistics.inferences.simplifying.destructiveEqualityResolution;
  simInf.forwardFolding = 
    _statistics.inferences.simplifying.forwardFolding;
  simInf.externalEvaluation = 
    _statistics.inferences.simplifying.externalEvaluation;
  simInf.totalBackward = 
    _statistics.inferences.simplifying.totalBackward();
  simInf.backwardSubsumption = 
    _statistics.inferences.simplifying.backwardSubsumption;
  simInf.backwardDemodulation = 
    _statistics.inferences.simplifying.backwardDemodulation;
  simInf.onBranchBackwardDemodulation = 
    _statistics.inferences.simplifying.onBranchBackwardDemodulation;
  propSpl.accomplished = 
    _statistics.inferences.propositionalSplitting.accomplished;
  if (_statistics.inferences.propositionalSplitting.accomplished)
    {
      propSpl.averageNumOfComponents = 
	(static_cast<float>(_statistics.inferences.propositionalSplitting.components)) /
	_statistics.inferences.propositionalSplitting.accomplished;
    }
  else
    propSpl.averageNumOfComponents = 0.0;
  propSpl.introducedComponentNames = 
    _statistics.inferences.propositionalSplitting.introducedComponentNames;
  negEqSpl.accomplished = 
    _statistics.inferences.negativeEqualitySplitting.accomplished;
  if (_statistics.inferences.negativeEqualitySplitting.accomplished)
    {
      negEqSpl.averageNumOfArgumentsSplitOff = 
	_statistics.inferences.negativeEqualitySplitting.argumentsSplitOff /
	_statistics.inferences.negativeEqualitySplitting.accomplished;
    }
  else
    negEqSpl.averageNumOfArgumentsSplitOff = 0.0;
  negEqSpl.introducedNames = 
    _statistics.inferences.negativeEqualitySplitting.introducedNames;
  genCl.total = 
    _statistics.clauses.generated.total();
  genCl.byDeductionInferences = 
    _statistics.clauses.generated.totalGeneratedByDeduction();
  genCl.bySimplificationInferences = 
    _statistics.clauses.generated.totalGeneratedBySimplification();
  genCl.byPropositionalSplitting = 
    _statistics.clauses.generated.byPropositionalSplitting;
  genCl.byNegativeEqualitySplitting = 
    _statistics.clauses.generated.byNegativeEqualitySplitting;
  genCl.byOther = 
    _statistics.clauses.generated.byOther;
  genCl.totalRedundantDiscarded = 
    _statistics.clauses.generated.totalRedundant();
  genCl.discardedPropositionalTautologies = 
    _statistics.clauses.generated.discardedPropositionalTautologies;
  genCl.discardedEquationalTautologies = 
    _statistics.clauses.generated.discardedEquationalTautologies;
  genCl.discardedSubsumed = 
    _statistics.clauses.generated.subsumed;
  genCl.totalNonredundantDiscarded = 
    _statistics.clauses.generated.totalNonredundantDiscarded();
  genCl.totalSimplified = 
    _statistics.clauses.generated.totalSimplified();
  genCl.transformedByPropositionalSplitting = 
    _statistics.clauses.generated.transformedByPropositionalSplitting;
  genCl.transformedByNegativeEqualitySplitting = 
    _statistics.clauses.generated.transformedByNegativeEqualitySplitting;
  retCl.total = 
    _statistics.clauses.retained.total;
  retCl.selected = 
    _statistics.clauses.retained.selected;
  retCl.totalSubsumed = 
    _statistics.clauses.retained.passiveSubsumed +
    _statistics.clauses.retained.activeSubsumed;
  retCl.totalSimplified = 
    _statistics.clauses.retained.totalSimplified();
  retCl.totalSimplifiedByDemodulation = 
    _statistics.clauses.retained.passiveSimplifiedByDemodulation +
    _statistics.clauses.retained.activeSimplifiedByDemodulation;
  retCl.totalSimplifiedByOnBranchDemodulation = 
    _statistics.clauses.retained.passiveSimplifiedByOnBranchDemodulation +
    _statistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
  retCl.totalOrphansMurdered = 
    _statistics.clauses.retained.passiveOrphansMurdered +
    _statistics.clauses.retained.activeOrphansMurdered; 
  retCl.totalRecycled = 
    _statistics.clauses.retained.passiveRecycled +
    _statistics.clauses.retained.activeRecycled;
  retCl.currentlyPassive = _master.numOfCurrentlyPassive();
  retCl.currentlyActive = _master.numOfCurrentlyActive();
}; 
void 
VampireKernel::_Implementation::outputStatistics(ostream* str,bool full)
{
  if (!str) return;
  _statistics.reset();
  _statistics.absorb(_inputStatistics);
  _statistics.absorb(_master.statistics());
  *str << "%========================================================================\n";
  if (full)
    {
      *str << "%============== Full statistics: =============\n";
    }
  else
    {
      *str << "%============ Summary of statistics: =========\n";
    };
  *str << "% time " <<  _timer.elapsedSeconds()
       << " (" << _master.timer().elapsedSeconds() << " current loop)\n";
  ulong defragCalls;
  float defragAverageInputSize;
  ulong defragNumOfInputPieces;
  float defragAverageOutputSize;
  ulong defragNumOfOutputPieces;
  GlobAlloc::getDefragmentationStatistics(defragCalls,
					  defragAverageInputSize,
					  defragNumOfInputPieces,
					  defragAverageOutputSize,
					  defragNumOfOutputPieces);
  *str << "% memory " << (GlobAlloc::occupiedByObjects()/1024) 
       << "Kb (" << (GlobAlloc::occupiedByChunks()/1024) << "Kb in chunks)\n"
       << "%     buffer deficit " << (GlobAlloc::totalBufferDeficit()/1024) << "Kb ("
       << (GlobAlloc::activeBufferDeficit()/1024) << "Kb act. + "
       << (GlobAlloc::suspendedBufferDeficit()/1024) << "Kb susp.)\n"
       << "%     defragmentation: " << defragCalls << " calls";
  if (defragCalls)
    *str << ", " << defragNumOfInputPieces << "*" << floor(defragAverageInputSize)
	 << " bytes ---> " 
	 << defragNumOfOutputPieces << "*" << floor(defragAverageOutputSize) << " bytes";
  *str << "\n";
  if (full)
    {
      _statistics.outputDetailed(*str);
    }
  else
    {
      _statistics.outputSummary(*str);
    };
  *str << "%== Currently passive " << _master.numOfCurrentlyPassive()
       << " (" << _master.numOfReachablePassive() << " reachable)\n";
  *str << "%== Currently active " << _master.numOfCurrentlyActive() << "\n"; 
  if (full)
    {
      *str << "%================ End of full statistics. ====================\n";
    }
  else
    {
      *str << "%=============== End of statistics summary. ==================\n";
    };
}; 
void 
VampireKernel::_Implementation::autoStatOutput(ostream* str)
{
  if (!str) return;
  if (_options.auto_stat_full) outputStatistics(str,true);
  if (_options.auto_stat_summary) outputStatistics(str,false);
}; 
inline  
const char*  
VampireKernel::_Implementation::flagValue(bool flag)
{
  if (flag) return "on";
  return "off";
};
void  
VampireKernel::_Implementation::switchBuiltinEqualityOff()
{
  CALL("switchBuiltinEqualityOff()");
  if  (_options.paramodulation 
       || _options.run_time_non_basicness
       || _options.input_non_basicness
       || _options.run_time_non_basicness
       || _options.inherited_negative_selection
       || _options.forward_demodulation
       || _options.fd_on_splitting_branches
       || _options.backward_demodulation
       || _options.bd_on_splitting_branches
       || _options.simplify_by_eq_res
       || _options.main_selected_lit_demodulation
       || _options.side_selected_lit_demodulation)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: No equality. All equality related features are switched off!\n";   
      _options.paramodulation = false;
      _options.prohibit_non_basic_superposition = false;
      _options.prohibit_non_basic_fd = false;
      _options.prohibit_non_basic_bd = false;
      _options.reduce_non_basic = false;
      _options.input_non_basicness = false;
      _options.run_time_non_basicness = false;
      _options.inherited_negative_selection = false;
      _options.forward_demodulation = false;
      _options.fd_on_splitting_branches = false;
      _options.backward_demodulation = false;
      _options.bd_on_splitting_branches = false;
      _options.simplify_by_eq_res = false;
      _options.main_selected_lit_demodulation = 0;
      _options.side_selected_lit_demodulation = 0;
    };
}; 
void  
VampireKernel::_Implementation::switchChainingForTransitiveOff()
{
  CALL("switchChainingForTransitiveOff()");
  if (_options.chaining_for_transitive)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: No transitive predicates. Chaining is switched off!\n";  
      _options.chaining_for_transitive = false;
    };
}; 
void  
VampireKernel::_Implementation::analyseInput()
{
  CALL("analyseInput()");
  if (_options.input_non_basicness)
    {
      _signature.resetAllArgumentsAsNonBasic();
      const BK::FIFO<BK::GlobAlloc,VK::Clause*>::Entry* entry;
      for (entry = _inputSupportClauses.begin();
	   entry;
	   entry = entry->next())
	{
	  for (const LiteralList::Element* el = entry->value()->LitList()->firstElem();
	       el;
	       el = el->next())
	    {
	      el->literal()->markBasicArgumentsInSignature();
	    };
	};
      for (entry = _inputQueryClauses.begin();
	   entry;
	   entry = entry->next())
	{
	  for (const LiteralList::Element* el = entry->value()->LitList()->firstElem();
	       el;
	       el = el->next())
	    {
	      el->literal()->markBasicArgumentsInSignature();
	    };
	};
      _signature.finaliseArgumentBasicness();
    };
}; 
void  
VampireKernel::_Implementation::analyseOptions()
{
  CALL("analyseOptions()");
  switch (_options.selection)
    { 
    case -1004L:
      _options.inherited_negative_selection = false;
      if (_options.std_output)
	*_options.std_output << "% WARNING: bidirectional literal selection "
			     << _options.selection 
			     << " switches inherited_negative_selection off!\n";
    };
  if (_options.input_non_basicness)
    {
      if (!_signature.nonBasicArgumentExists())
	{
	  if (_options.std_output)
	    *_options.std_output << "% WARNING: no non-basic arguments in the input.\n"
				 << "% input_non_basicness is switched off.\n";
	  _options.input_non_basicness = false;
	};
    };
  if (!_options.input_non_basicness)
    {
      if (_options.prohibit_non_basic_superposition || 
	  _options.prohibit_non_basic_bd)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: input_non_basicness is off.\n";
	      if (_options.prohibit_non_basic_superposition)
		*_options.std_output << "% prohibit_non_basic_superposition is switched off.\n";
	      if (_options.prohibit_non_basic_bd)
		*_options.std_output << "% prohibit_non_basic_bd is switched off.\n";
	    };
	  _options.prohibit_non_basic_superposition = false;
	  _options.prohibit_non_basic_bd = false;
	};
    };
  if (!_options.input_non_basicness && !_options.run_time_non_basicness)
    {
      if (_options.prohibit_non_basic_fd ||
	  _options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: no non-basic subterms can be identified.\n";
	      if (_options.prohibit_non_basic_fd)
		*_options.std_output << "% prohibit_non_basic_fd is switched off.\n";
	      if (_options.reduce_non_basic)
		*_options.std_output << "% reduce_non_basic is switched off.\n";
	    };
	  _options.prohibit_non_basic_fd = false;
	  _options.reduce_non_basic = false;
	};
    };
  if (_options.input_non_basicness)
    {
      if (!_options.prohibit_non_basic_superposition &&
	  !_options.prohibit_non_basic_fd &&
	  !_options.prohibit_non_basic_bd &&
	  !_options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: input_non_basicness is switched off since it cannot be used.\n";
	    };
	  _options.input_non_basicness = false;
	};
    };
  if (_options.run_time_non_basicness)
    {
      if (!_options.prohibit_non_basic_fd &&
	  !_options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: run_time_non_basicness is switched off since it cannot be used.\n";
	    };
	  _options.run_time_non_basicness = false;
	};
    };
  if (_options.reduce_non_basic)
    {
      if (!_options.prohibit_non_basic_fd || 
	  !_options.prohibit_non_basic_bd)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: non-basic demodulation makes no sense.\n"
				   << "% since reduce_non_basic is on.\n";
	      if (!_options.prohibit_non_basic_fd)
		*_options.std_output << "% prohibit_non_basic_fd is switched on.\n";
	      if (!_options.prohibit_non_basic_bd)
		*_options.std_output << "% prohibit_non_basic_bd is switched on.\n";
	    };
	  _options.prohibit_non_basic_fd = true;
	  _options.prohibit_non_basic_bd = true;
	};
    };
  if (_options.prohibit_non_basic_superposition &&
      (!_options.prohibit_non_basic_fd ||
       !_options.prohibit_non_basic_bd))
    {	
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: non-basic demodulation without non-basic superposition\n"
			       << "% causes incompleteness.\n";
	};
      _incompleteProcedure = true;
    };
  if (_options.forw_sup_no_dup_red ||
      _options.back_sup_no_dup_red)
    {
      _incompleteProcedure = true;
    };
  if (_options.fold_single_comp_in_splitting)
    {
      _incompleteProcedure = true;
    };
}; 
bool  
VampireKernel::_Implementation::isRelevantBackgroundFlag(const VK::ClauseBackground::Rule& fl)
{
  switch (fl)
    {
    case VK::ClauseBackground::Input:
    case VK::ClauseBackground::Prepro:
    case VK::ClauseBackground::BinRes:
    case VK::ClauseBackground::ForwTransChain:
    case VK::ClauseBackground::BackTransChain:
    case VK::ClauseBackground::ForwSup: 
    case VK::ClauseBackground::BackSup:
    case VK::ClauseBackground::EqRes:
    case VK::ClauseBackground::EqFact:
    case VK::ClauseBackground::ForwDemod:
    case VK::ClauseBackground::BackDemod: 
    case VK::ClauseBackground::EqResSimp:
    case VK::ClauseBackground::ForwSubsRes:
    case VK::ClauseBackground::BackSubsRes:  
    case VK::ClauseBackground::Reanimation:
    case VK::ClauseBackground::ExternalEvaluation:
    case VK::ClauseBackground::NameIntro:
    case VK::ClauseBackground::Folding:
    case VK::ClauseBackground::Clausification:
      return true;
    default:
      return false;
    };
}; 
VampireKernel::Clause::BackgroundFlag
VampireKernel::_Implementation::convertBackgroundFlag(const VK::ClauseBackground::Rule& fl)
{
  switch (fl)
    {
    case VK::ClauseBackground::Input: 
      return VampireKernel::Clause::IsInputClause;
    case VK::ClauseBackground::Prepro: 
      return VampireKernel::Clause::IsObtainedDuringPreprocessing;
    case VK::ClauseBackground::BinRes: 
      return VampireKernel::Clause::BinaryResolutionUsed;
    case VK::ClauseBackground::ForwTransChain: 
      return VampireKernel::Clause::ForwardTransitiveChainingUsed;
    case VK::ClauseBackground::BackTransChain: 
      return VampireKernel::Clause::BackwardTransitiveChainingUsed;
    case VK::ClauseBackground::ForwSup: 
      return VampireKernel::Clause::ForwardSuperpositionUsed; 
    case VK::ClauseBackground::BackSup: 
      return VampireKernel::Clause::BackwardSuperpositionUsed;
    case VK::ClauseBackground::EqRes: 
      return VampireKernel::Clause::EqualityResolutionUsed;
    case VK::ClauseBackground::EqFact: 
      return VampireKernel::Clause::EqualityFactoringUsed;
    case VK::ClauseBackground::ForwDemod: 
      return VampireKernel::Clause::ForwardDemodulationUsed;
    case VK::ClauseBackground::BackDemod: 
      return VampireKernel::Clause::BackwardDemodulationUsed; 
    case VK::ClauseBackground::EqResSimp: 
      return VampireKernel::Clause::SimplificationByEqualityResolutionUsed;
    case VK::ClauseBackground::ForwSubsRes: 
      return VampireKernel::Clause::ForwardSubsumptionResolutionUsed;
    case VK::ClauseBackground::BackSubsRes: 
      return VampireKernel::Clause::BackwardSubsumptionResolutionUsed; 
    case VK::ClauseBackground::Reanimation: 
      return VampireKernel::Clause::IsReanimatedPassive;
    case VK::ClauseBackground::ExternalEvaluation: 
      return VampireKernel::Clause::ExternalEvaluationUsed;
     case VK::ClauseBackground::NameIntro:
       return VampireKernel::Clause::IsNameIntro;
     case VK::ClauseBackground::Folding:
       return VampireKernel::Clause::FoldingUsed;
     case VK::ClauseBackground::Clausification:
       return VampireKernel::Clause::ClausificationUsed;
    default:
      return VampireKernel::Clause::IsInputClause; 
    };
}; 
VK::ClauseBackground::Rule
VampireKernel::_Implementation::convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl)
{
  CALL("convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl)");
  switch (fl)
    {
    case VampireKernel::Clause::IsInputClause:
      return VK::ClauseBackground::Input;
    case VampireKernel::Clause::IsObtainedDuringPreprocessing:
      return VK::ClauseBackground::Prepro;
    case VampireKernel::Clause::BinaryResolutionUsed:
      return VK::ClauseBackground::BinRes;
    case VampireKernel::Clause::ForwardTransitiveChainingUsed:
      return VK::ClauseBackground::ForwTransChain;
    case VampireKernel::Clause::BackwardTransitiveChainingUsed:
      return VK::ClauseBackground::BackTransChain;
    case VampireKernel::Clause::ForwardSuperpositionUsed:
      return VK::ClauseBackground::ForwSup;
    case VampireKernel::Clause::BackwardSuperpositionUsed:
      return VK::ClauseBackground::BackSup;
    case VampireKernel::Clause::EqualityResolutionUsed:
      return VK::ClauseBackground::EqRes;
    case VampireKernel::Clause::EqualityFactoringUsed:
      return VK::ClauseBackground::EqFact;
    case VampireKernel::Clause::ForwardDemodulationUsed:
      return VK::ClauseBackground::ForwDemod;
    case VampireKernel::Clause::BackwardDemodulationUsed:
      return VK::ClauseBackground::BackDemod;
    case VampireKernel::Clause::SimplificationByEqualityResolutionUsed:
      return VK::ClauseBackground::EqResSimp;
    case VampireKernel::Clause::ForwardSubsumptionResolutionUsed:
      return VK::ClauseBackground::ForwSubsRes;
    case VampireKernel::Clause::BackwardSubsumptionResolutionUsed:
      return VK::ClauseBackground::BackSubsRes;
    case VampireKernel::Clause::IsReanimatedPassive:
      return VK::ClauseBackground::Reanimation;
    case VampireKernel::Clause::ExternalEvaluationUsed:
      return VK::ClauseBackground::ExternalEvaluation;
    case VampireKernel::Clause::IsNameIntro:
      return VK::ClauseBackground::NameIntro;
    case VampireKernel::Clause::FoldingUsed:
      return VK::ClauseBackground::Folding;
    case VampireKernel::Clause::ClausificationUsed:
      return VK::ClauseBackground::Clausification;
    default:
      return VK::ClauseBackground::Input; 
    };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation::ClauseSetIterator"
#endif    
#include "debugMacros.hpp"
VampireKernel::_Implementation::ClauseSetIterator::~ClauseSetIterator()
{
  CALL("destructor ~ClauseSetIterator()");
};
bool VampireKernel::_Implementation::ClauseSetIterator::isEnd() const
{
  return _iter.isEnd();
}; 
void VampireKernel::_Implementation::ClauseSetIterator::next()
{
  _iter.next();
};
const VampireKernel::Clause* VampireKernel::_Implementation::ClauseSetIterator::clause() const
{
  return 
    static_cast<const VampireKernel::Clause*>(static_cast<const void*>(_iter.clause()));
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation::NameIntroReceiver"
#endif    
#include "debugMacros.hpp"
void VampireKernel::_Implementation::NameIntroReceiver::receive(VK::Clause* const & cl)
{
  CALL("receive(VK::Clause* const & cl)");
  cl->SetNumber(_hostKernelSession->_clauseNumberGenerator.generate());
  _hostKernelSession->_nameIntroClauses.Insert(cl);
  _hostKernelSession->_master.nameIntroReceiver()->receive(cl);
  if (_hostKernelSession->_options.std_output &&
      _hostKernelSession->_options.show_name_intro)
    *(_hostKernelSession->_options.std_output) << "% " << cl << "\n";
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel"
#endif    
#include "debugMacros.hpp"
VampireKernel::VampireKernel() 
{
  CALL("constructor VampireKernel()");
  _implementation = 0; 
};
VampireKernel::~VampireKernel() 
{
  CALL("destructor ~VampireKernel()");
  if (DestructionMode::isThorough())
    {
      if (_implementation)
	{
	  _Implementation::closeSession();
	  _implementation->destroy();
	  _Implementation::deallocate(_implementation);	
	};
    };
};
void VampireKernel::openSession(const Options& opt) 
{
  CALL("openSession(const Options& opt)");
  ASSERT(!GlobAlloc::occupiedByObjects());
  GlobalStopFlag::reset();
  _Implementation::setHardMemoryLimit(((float)opt.memory_limit)*1024);
  _Implementation::setAllocationBufferSize(opt.allocation_buffer_size);
  _implementation = _Implementation::allocate();
  _implementation->init(opt);
  _implementation->openSession();
  _implementation->_wrapper = this;
}; 
void VampireKernel::closeSession()
{
  _Implementation::closeSession();
  _implementation = 0;
}; 
float VampireKernel::versionNumber()
{
  return 7.45f;
};
string VampireKernel::versionDescription() 
{
  return string("8.01 (for CASC-J3 only)");
};
void VampireKernel::beginInputSignature() 
{
  _implementation->beginInputSignature();
};  
void VampireKernel::endInputSignature() 
{
  _implementation->endInputSignature();
};
void VampireKernel::registerInputFunction(string name,
					  ulong arity,
					  ulong& id,
					  const long& weight,
					  long precedence)
{
  _implementation->registerInputFunction(name.c_str(),
					 arity,
					 id,
					 weight,
					 precedence);
}; 
void VampireKernel::registerInputPredicate(string name,
					   ulong arity,
					   bool isAnswer,
					   ulong& id,
					   const long& weight,
					   long precedence,
					   long positiveHeaderRank,
					   long negativeHeaderRank)
{
  _implementation->registerInputPredicate(name.c_str(),
					  arity,
					  isAnswer,
					  id,
					  weight,
					  precedence,
					  positiveHeaderRank,
					  negativeHeaderRank);
};  
void VampireKernel::registerEquality(ulong& id,
				     long positiveHeaderRank,
				     long negativeHeaderRank)
{
  _implementation->registerEquality(id,
				    positiveHeaderRank,
				    negativeHeaderRank);
};
void VampireKernel::registerInterpretedFunction(ulong id)
{
  _implementation->registerInterpreted(false,id);
};
void VampireKernel::registerInterpretedPredicate(ulong id)
{
  _implementation->registerInterpreted(true,id);
}; 
void VampireKernel::beginInputClauses() 
{
  _implementation->beginInputClauses();
};
void VampireKernel::endInputClauses() 
{
  _implementation->endInputClauses();
};
void VampireKernel::beginInputClause(ulong clauseNumber,
				     void* origin,
				     bool goal,
				     bool support) 
{
  ClauseAnnotation annotation;
  if (goal) annotation.markAsGoal();
  if (support) annotation.markAsSupport();
  _implementation->beginInputClause(clauseNumber,origin,annotation); 
};
void VampireKernel::beginInputClause(ulong clauseNumber,
				     void* origin,
				     const ClauseAnnotation& annotation)
{
  _implementation->beginInputClause(clauseNumber,origin,annotation); 
};
void VampireKernel::endInputClause() 
{
  _implementation->endInputClause();
};
bool VampireKernel::beginLiteral(bool positive,ulong predicateId) 
{
  return _implementation->beginLiteral(positive,predicateId);
};
void VampireKernel::markSelected() {
  CALL("markSelected()");
  _implementation->markSelected();
}; 
void VampireKernel::markUnselected() {
  CALL("markUnselected()");
  _implementation->markUnselected();
};
void VampireKernel::markRuleHead() {
  CALL("markRuleHead()");
  _implementation->markRuleHead();
};
void VampireKernel::markSubquery() {
  CALL("markSubquery()");
  _implementation->markSubquery();
};
void VampireKernel::markEqWithResolution() {
  CALL("markEqWithResolution()");
  _implementation->markEqWithResolution();
};
void VampireKernel::markEqWithResolutionOnly() {
  CALL("markEqWithResolutionOnly()");
  _implementation->markEqWithResolutionOnly();
};
void VampireKernel::markGivenPosEqOrientation() {
  CALL("markGivenPosEqOrientation()");
  _implementation->markGivenPosEqOrientation();
};
void VampireKernel::markReversePosEqOrientation() {
  CALL("markReversePosEqOrientation()");
  _implementation->markReversePosEqOrientation();
};
void VampireKernel::endLiteral() 
{
  _implementation->endLiteral();
};
void VampireKernel::receiveFunction(ulong symId) 
{
  _implementation->receiveFunction(symId);
};
void 
VampireKernel::receiveInterpretedConstant(const VampireKernelEnvironment::InterpretedConstant& value)
{
  _implementation->receiveInterpretedConstant(value);
};
void VampireKernel::receiveVariable(ulong varNum) 
{
  _implementation->receiveVariable(varNum);
};
void VampireKernel::saturate(ulong nextVacantClauseId,long timeLimit) 
{
  _implementation->saturate(nextVacantClauseId,timeLimit);
};
void VampireKernel::lookForAnotherAnswer(int timeLimit)
{
  _implementation->lookForAnotherAnswer(timeLimit);
}
VampireKernel* VampireKernel::currentSession()
{
  if (_Implementation::_currentSession)
    {
      return _Implementation::_currentSession->_wrapper;
    }
  else
    return 0;
};
void VampireKernel::suspend()
{
  _implementation->suspend();
};
void VampireKernel::resume()
{
  _implementation->resume();
}; 
void VampireKernel::reportInterrupt(int signalNumber)
{
  if (_implementation)
    {
      _implementation->reportInterrupt(signalNumber);
    };
};
void VampireKernel::reportErrorSignal(int signalNumber)
{
  if (_implementation)
    {
      _implementation->reportErrorSignal(signalNumber);
    };
};
void VampireKernel::reportError()
{
  if (_implementation)
    {
      _implementation->reportError();
    };
};
VampireKernel::TerminationReason VampireKernel::terminationReason() const
{
  return _implementation->terminationReason();
};
const VampireKernel::Clause* VampireKernel::refutation() const 
{
  return _implementation->refutation();
};
VampireKernel::ClauseSetIterator* VampireKernel::activeClauses()
{
  return _implementation->activeClauses();
};
void VampireKernel::disposeOf(VampireKernel::ClauseSetIterator* itr)
{
  delete static_cast<ClauseSetIterator*>(itr);
};
bool VampireKernel::isInterpretedConstantId(ulong symbolId) const
{
  return _implementation->isInterpretedConstantId(symbolId);
};
bool VampireKernel::isInputSymbol(ulong symbolId) const
{
  return _implementation->isInputSymbol(symbolId);
}; 
bool VampireKernel::isAnswerPredicateId(ulong symbolId) const
{
  return _implementation->isAnswerPredicateId(symbolId);
};
ulong VampireKernel::symbolArity(ulong symbolId) const
{
  return _implementation->symbolArity(symbolId);
};
string VampireKernel::symbolPrintName(ulong symbolId) const
{
  return string(_implementation->symbolPrintName(symbolId));
}; 
void 
VampireKernel::interpretedConstantValue(ulong symbolId,
					VampireKernelEnvironment::InterpretedConstant& value) const
{
  value = _implementation->interpretedConstantValue(symbolId);
}; 
const VampireKernel::Statistics& VampireKernel::statistics()
{
  return _implementation->statistics();
};
void VampireKernel::outputStatistics(ostream* str,bool full)
{
  if (!str) return;
  _implementation->outputStatistics(str,full);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Options::LitListOrdering"
#endif    
#include "debugMacros.hpp"
VampireKernel::Options::LitListOrdering::LitListOrdering(ElementaryLitListOrdering tag)
{
  _type = 0;
  _tag = tag;
};
VampireKernel::Options::LitListOrdering::LitListOrdering(const LitListOrdering* ord)
{
  _type = 1;
  _ord1 = ord;
};
VampireKernel::Options::LitListOrdering::LitListOrdering(const LitListOrdering* ord1,
							 const LitListOrdering* ord2)
{
  _type = 2;
  _ord1 = ord1;
  _ord2 = ord2;
};
bool
VampireKernel::Options::LitListOrdering::isElementary(ElementaryLitListOrdering& tag) const
{
  tag = _tag;
  return _type == 0;
};
bool
VampireKernel::Options::LitListOrdering::isInverse(const LitListOrdering*& ord) const
{
  ord = _ord1;
  return _type == 1;
};
bool
VampireKernel::Options::LitListOrdering::isComposition(const LitListOrdering*& ord1,
						       const LitListOrdering*& ord2) const
{  
  ord1 = _ord1;
  ord2 = _ord2;
  return _type == 2;
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Options"
#endif    
#include "debugMacros.hpp"
void VampireKernel::Options::setDefaultValues()
{
  job_id = "default_job_id";
  main_alg = 0L;
  no_inferences_between_rules = false;
  static_splitting = true;
  dynamic_splitting = false;
  small_basic_comp_in_hyp_splitting = false;
  fold_single_comp_in_splitting = false;
  single_ground_lit_comp_min_size = LONG_MAX;
  reuse_names_for_negated_ground_lits_in_splitting = false;
  basic_comp_in_spl_is_max_wrt = 0;
  back_fold_in_spl = false;
  static_neg_eq_splitting = 3L;
  stat_neg_eq_spl_var_count_limit = 0L;
  dynamic_neg_eq_splitting = 0L;
  dyn_neg_eq_spl_var_count_limit = 0L;
  paramodulation = true;
  paramodulate_from_variables = true;
  sim_back_sup = 0;
  forw_sup_no_dup_red = false;
  back_sup_no_dup_red = false;
  prohibit_non_basic_superposition = false;
  prohibit_non_basic_fd = true;
  prohibit_non_basic_bd = true;
  reduce_non_basic = false;
  input_non_basicness = false;
  run_time_non_basicness = false;
  selection = 4L;
  inherited_negative_selection = false;
  external_evaluation = true;
  forward_subsumption = true;
  simplify_by_fs = 1L;
  main_premise_subsumption = 0L;
  side_premise_subsumption = 0L;
  main_selected_lit_demodulation = 0L;
  side_selected_lit_demodulation = 0L;
  forward_demodulation = 4L;
  fd_on_splitting_branches = true;
  normalize_by_commutativity = true;
  eq_taut_mod_C = false;
  prop_taut_mod_C = false;
  trivial_eq_res_mod_C = false;
  identical_lit_factoring_mod_C = false;
  chaining_for_transitive = false;
  chaining_through_variables = true;
  backward_demodulation = 4L;
  bd_on_splitting_branches = true;
  simplify_by_eq_res = true;
  backward_subsumption = true;
  orphan_murder = false;
  lrs = 1L;
  simplification_ordering = StandardKBOrdering;
  term_weighting_scheme = UniformTermWeighting;
  orient_pos_equalities_in_support = true;
  orient_neg_equalities_in_support = true;
  memory_limit = LONG_MAX; 
  allocation_buffer_size = 10L;
  allocation_buffer_price = 2.0;
  max_skolem = 1024; 
  max_active = LONG_MAX;
  max_passive = LONG_MAX;
  max_weight = LONG_MAX;
  max_inference_depth = LONG_MAX;
  age_weight_ratio_numerator = 1L;
  age_weight_ratio_denominator = 5L;
  best_good_weight_ratio_numerator = 1;
  best_good_weight_ratio_denominator = 0;
  selected_pos_eq_penalty_coefficient = 1.0;  
  nongoal_penalty_coefficient = 1.0;
  simulated_time_limit = 0;
  first_time_check = 5L;   
  first_mem_check = 200L;   
  show_opt = false;
  show_input = false;
  show_gen = false;
  show_passive = false;
  show_active = false;
  show_name_intro = false;
  show_lit_prop = false; 
  show_passive_profile_interval = LONG_MAX;
  auto_stat_full = false;
  auto_stat_summary = true;
  show_profile = false;
  show_weight_limit_changes = false;  
  std_output = &cout;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  step_resolution_literal_selection = false;
#endif
}; 
ostream& VampireKernel::Options::output(ostream& str)
{
  str << "% job_id = " << job_id << '\n';
  str << "% main_alg = " << main_alg << '\n';
  str << "% no_inferences_between_rules = " << no_inferences_between_rules << '\n';
  str << "% static_splitting = " << static_splitting << '\n';
  str << "% dynamic_splitting = " << dynamic_splitting << '\n';
  str << "% small_basic_comp_in_hyp_splitting = " << small_basic_comp_in_hyp_splitting << '\n'; 
  str << "% fold_single_comp_in_splitting = " << fold_single_comp_in_splitting << '\n'; 
  str << "% single_ground_lit_comp_min_size = " <<  single_ground_lit_comp_min_size << '\n';
  str << "% reuse_names_for_negated_ground_lits_in_splitting = " << reuse_names_for_negated_ground_lits_in_splitting  << '\n'; 
  str << "% basic_comp_in_spl_is_max_wrt = "
      << ((basic_comp_in_spl_is_max_wrt) ? "nonempty" : "empty") << '\n'; 
  str << "% back_fold_in_spl = " << back_fold_in_spl << '\n';
  str << "% static_neg_eq_splitting = " << static_neg_eq_splitting << '\n';
  str << "% stat_neg_eq_spl_var_count_limit = " << stat_neg_eq_spl_var_count_limit << '\n';
  str << "% dynamic_neg_eq_splitting = " << dynamic_neg_eq_splitting << '\n';
  str << "% dyn_neg_eq_spl_var_count_limit = " << dyn_neg_eq_spl_var_count_limit << '\n';
  str << "% paramodulation = " << paramodulation << '\n'; 
  str << "% paramodulate_from_variables = " << paramodulate_from_variables << '\n'; 
  str << "% sim_back_sup = " << sim_back_sup << '\n';  
  str << "% forw_sup_no_dup_red = " << forw_sup_no_dup_red << '\n';   
  str << "% back_sup_no_dup_red = " << back_sup_no_dup_red << '\n';   
  str << "% prohibit_non_basic_superposition = " << prohibit_non_basic_superposition << '\n'; 
  str << "% prohibit_non_basic_fd = " << prohibit_non_basic_fd << '\n'; 
  str << "% prohibit_non_basic_bd = " << prohibit_non_basic_bd << '\n'; 
  str << "% reduce_non_basic = " << reduce_non_basic << '\n';  
  str << "% input_non_basicness = " << input_non_basicness << '\n'; 
  str << "% run_time_non_basicness = " << run_time_non_basicness << '\n';   
  str << "% selection = " << selection << '\n';
  str << "% inherited_negative_selection = " << inherited_negative_selection << '\n'; 
  str << "% external_evaluation = " << external_evaluation << '\n'; 
  str << "% forward_subsumption = " << forward_subsumption << '\n'; 
  str << "% simplify_by_fs = " << simplify_by_fs << '\n';
  str << "% main_premise_subsumption = " << main_premise_subsumption << '\n';
  str << "% side_premise_subsumption = " << side_premise_subsumption << '\n';
  str << "% main_selected_lit_demodulation = " << main_selected_lit_demodulation << '\n';
  str << "% side_selected_lit_demodulation = " << side_selected_lit_demodulation << '\n';
  str << "% forward_demodulation = " << forward_demodulation << '\n';
  str << "% fd_on_splitting_branches = " << fd_on_splitting_branches << '\n';
  str << "% normalize_by_commutativity = " << normalize_by_commutativity << '\n';
  str << "% eq_taut_mod_C = " << eq_taut_mod_C << '\n';
  str << "% prop_taut_mod_C = " << prop_taut_mod_C << '\n';
  str << "% trivial_eq_res_mod_C = " << trivial_eq_res_mod_C << '\n';
  str << "% identical_lit_factoring_mod_C = " << identical_lit_factoring_mod_C << '\n';
  str << "% chaining_for_transitive = " << chaining_for_transitive << '\n';
  str << "% chaining_through_variables = " << chaining_through_variables << '\n';
  str << "% backward_demodulation = " << backward_demodulation << '\n';
  str << "% bd_on_splitting_branches = " << bd_on_splitting_branches << '\n';
  str << "% simplify_by_eq_res = " << simplify_by_eq_res << '\n'; 
  str << "% backward_subsumption = " << backward_subsumption << '\n'; 
  str << "% orphan_murder = " << orphan_murder << '\n'; 
  str << "% lrs = " << lrs << '\n';
  str << "% simplification_ordering = " << simplification_ordering << '\n';
  str << "% term_weighting_scheme = " << term_weighting_scheme << '\n';
  str << "% orient_pos_equalities_in_support = " << orient_pos_equalities_in_support <<  '\n';
  str << "% orient_neg_equalities_in_support = " << orient_neg_equalities_in_support <<  '\n';
  str << "% memory_limit = " << memory_limit << '\n';
  str << "% allocation_buffer_size = " << allocation_buffer_size << '\n';
  str << "% allocation_buffer_price = " << allocation_buffer_price << '\n';
  str << "% max_skolem = " << max_skolem << '\n';
  str << "% max_active = " << max_active << '\n'; 
  str << "% max_passive = " << max_passive << '\n';  
  str << "% max_weight = " << max_weight << '\n';    
  str << "% max_inference_depth = " << max_inference_depth << '\n';  
  str << "% age_weight_ratio_numerator = " << age_weight_ratio_numerator << '\n';   
  str << "% age_weight_ratio_denominator = " << age_weight_ratio_denominator << '\n'; 
  str << "% best_good_weight_ratio_numerator = " << best_good_weight_ratio_numerator << '\n';
  str << "% best_good_weight_ratio_denominator = " << best_good_weight_ratio_denominator << '\n'; 
  str << "% selected_pos_eq_penalty_coefficient = " 
      << selected_pos_eq_penalty_coefficient << '\n';  
  str << "% nongoal_penalty_coefficient = "
      << nongoal_penalty_coefficient << '\n';
  str << "% first_time_check = " << first_time_check << '\n';   
  str << "% first_mem_check = " << first_mem_check << '\n';  
  str.flush();
  return str;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause"
#endif    
#include "debugMacros.hpp"
ulong VampireKernel::Clause::number() const
{
  CALL("number() const");
  return (static_cast<const VK::Clause*>(static_cast<const void*>(this)))->Number();
};
bool VampireKernel::Clause::isLiteralDefinition(const Literal*& definedLiteral,
					       const Literal*& bodyLiterals) const
{
  CALL("isLiteralDefinition(const Literal*& definedLiteral,const Literal*& bodyLiterals) const");
  if ((static_cast<const VK::Clause*>
       (static_cast<const void*>(this)))->isNameIntro())
    {
      definedLiteral = static_cast<const VampireKernel::Literal*>
	(static_cast<const void*>
	 ((static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->LitList()->firstElem()));
      bodyLiterals = definedLiteral->next();
      return true;
    }
  else
    return false;
};
const VampireKernel::Literal* VampireKernel::Clause::literals() const
{
  CALL("literals() const");
  ASSERT(!(static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->isNameIntro());
  return 
    static_cast<const VampireKernel::Literal*>
    (static_cast<const void*>
     ((static_cast<const VK::Clause*>
       (static_cast<const void*>(this)))->LitList()->firstElem()));
};
ulong VampireKernel::Clause::numberOfSelectedLiterals() const
{
  CALL("numberOfSelectedLiterals() const");
  ASSERT(!(static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->isNameIntro());
  return 
    (static_cast<const VK::Clause*>
     (static_cast<const void*>(this)))->LitList()->numOfSelectedLits();
};
bool VampireKernel::Clause::isInputClause() const
{
  CALL("isInputClause() const");  
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->isInputClause();
};
bool VampireKernel::Clause::isNameIntroduction() const
{
  CALL("isNameIntroduction() const");
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->isNameIntro();
};
void* VampireKernel::Clause::inputClauseOrigin() const
{
  CALL("inputClauseOrigin() const");
  ASSERT(isInputClause());
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->inputClauseOrigin();
};
const VampireKernel::Clause::Ancestor* 
VampireKernel::Clause::firstAncestor() const
{
  CALL("firstAncestor() const");  
  if (isInputClause()) return 0;
  return 
    static_cast<const VampireKernel::Clause::Ancestor*>
    (static_cast<const void*>
     ((((static_cast<const VK::Clause*>
	 (static_cast<const void*>(this)))->background().ancestors()))));
}; 
VampireKernel::Clause::BackgroundFlagIterator 
VampireKernel::Clause::backgroundFlagIterator() const
{
  CALL("backgroundFlagIterator() const");
  BackgroundFlagIterator res;
  res._base = static_cast<const void*>(this);
  ulong fl = 0UL;
  while ((fl < 32UL) && 
	 (!(VampireKernel::_Implementation::isRelevantBackgroundFlag((VK::ClauseBackground::Rule)fl) &&
	    (static_cast<const VK::Clause*>
	     (static_cast<const void*>(this)))->background().properties().bit(fl))))
    {
      ++fl;
    };
  res._offset = fl;
  return res;
}; 
bool VampireKernel::Clause::backgroundFlagIsSet(BackgroundFlag fl) const
{
  CALL("backgroundFlagIsSet(BackgroundFlag fl) const");
  ulong nativeFlagNum = 
    (ulong)VampireKernel::_Implementation::convertBackgroundFlag(fl);
  return 
    (static_cast<const VK::Clause*>
     (static_cast<const void*>(this)))->background().properties().bit((ulong)nativeFlagNum);
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause::Ancestor"
#endif    
#include "debugMacros.hpp"
const VampireKernel::Clause* VampireKernel::Clause::Ancestor::clause() const
{
  CALL("clause() const");
  return 
    static_cast<const VampireKernel::Clause*>
    (static_cast<const void*>
     ((static_cast<const VK::ClauseBackground::AncestorList*>
       (static_cast<const void*>(this)))->hd().strip()));
}; 
const VampireKernel::Clause::Ancestor* VampireKernel::Clause::Ancestor::next() const
{
  CALL("next() const");  
  return 
    static_cast<const VampireKernel::Clause::Ancestor*>
    (static_cast<const void*>
     ((static_cast<const VK::ClauseBackground::AncestorList*>
       (static_cast<const void*>(this)))->tl()));
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause::BackgroundFlagIterator"
#endif    
#include "debugMacros.hpp"
VampireKernel::Clause::BackgroundFlagIterator::BackgroundFlagIterator()
{
  _base = 0;
};
VampireKernel::Clause::BackgroundFlagIterator::~BackgroundFlagIterator()
{
};
bool VampireKernel::Clause::BackgroundFlagIterator::getCurrentFlag(BackgroundFlag& flag) const
{
  CALL("getCurrentFlag(BackgroundFlag& flag) const");
  if (_base && (_offset < 32UL)) 
    {
      flag = 
	VampireKernel::_Implementation::convertBackgroundFlag((const VK::ClauseBackground::Rule&)_offset);
      return true;
    }
  else
    return false;
};
void VampireKernel::Clause::BackgroundFlagIterator::goToNextFlag()
{
  CALL("goToNextFlag()");
  ++_offset;
  while ((_offset < 32UL) && 
	 (!(VampireKernel::_Implementation::isRelevantBackgroundFlag((VK::ClauseBackground::Rule)_offset) &&
	    (static_cast<const VK::Clause*>(_base))->background().properties().bit(_offset))))
    {
      ++_offset;
    };
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Literal"
#endif    
#include "debugMacros.hpp"
const VampireKernel::Literal* 
VampireKernel::Literal::next() const
{
  CALL("next() const");
  return
    static_cast<const VampireKernel::Literal*>
    (static_cast<const void*>    
     ((static_cast<const VK::LiteralList::Element*>
       (static_cast<const void*>(this)))->next()));
}; 
bool VampireKernel::Literal::isPositive() const 
{
  CALL("isPositive() const");
  return 
    (static_cast<const VK::LiteralList::Element*>(static_cast<const void*>(this)))->positive();
};
ulong VampireKernel::Literal::predicate() const
{
  CALL("predicate() const");
  ulong predId = 
    (static_cast<const VK::LiteralList::Element*>
     (static_cast<const void*>(this)))->literal()->functor();
  if (predId == VampireKernelConst::OrdEqNum) predId = VampireKernelConst::UnordEqNum;
  return predId;
};
const VampireKernel::Subterm* VampireKernel::Literal::firstArgument() const
{
  CALL("firstArgument() const");
  const Term* args = 
    ((static_cast<const VK::LiteralList::Element*>
      (static_cast<const void*>(this)))->literal()->Args());
  if (args->isNotArg()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(args));
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Subterm"
#endif    
#include "debugMacros.hpp"
const VampireKernel::Subterm* VampireKernel::Subterm::next() const
{
  CALL("next() const");
  const Term* nxt = (static_cast<const Term*>(static_cast<const void*>(this)))->Next();
  if (nxt->isNotArg()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(nxt));
};
bool VampireKernel::Subterm::isVariable() const
{
  CALL("isVariable() const");
  return (static_cast<const Term*>(static_cast<const void*>(this)))->isVariable();
};
ulong VampireKernel::Subterm::variableNumber() const
{
  CALL("variableNumber() const");
  ASSERT(isVariable());
  return (static_cast<const Term*>(static_cast<const void*>(this)))->var();
};
ulong VampireKernel::Subterm::topFunction() const
{
  CALL("topFunction() const");
  ASSERT(!isVariable());
  return (static_cast<const Term*>(static_cast<const void*>(this)))->deref()->functor();
};
const VampireKernel::Subterm* VampireKernel::Subterm::firstArgument() const
{
  CALL("firstArgument() const");
  ASSERT(!isVariable());
  const Term* args = (static_cast<const Term*>(static_cast<const void*>(this)))->deref()->Args();  
  if (args->isNotArg()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(args));
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::ClauseSetIterator"
#endif    
#include "debugMacros.hpp"
VampireKernel::ClauseSetIterator::~ClauseSetIterator()
{
  CALL("destructor ~ClauseSetIterator()");
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#  define DEBUG_NAMESPACE "VampireKernel::ClauseAnnotation"
#endif    
#include "debugMacros.hpp"
VampireKernel::ClauseAnnotation::ClauseAnnotation() 
{
  makeDefault();
};
void VampireKernel::ClauseAnnotation::makeDefault()
{
  _isGoal = false;
  _isSupport = false;
  _isNonSubsuming = false;
  _isNonSubsumable = false;
};
void VampireKernel::ClauseAnnotation::make(bool isGoal,bool isSupport,bool isNonSubsuming,bool isNonSubsumable)
{
  _isGoal = isGoal;
  _isSupport = isSupport;
  _isNonSubsuming = isNonSubsuming;
  _isNonSubsumable = isNonSubsumable;
}; 
bool VampireKernel::ClauseAnnotation::isGoal() const
{
  return _isGoal;
};
bool VampireKernel::ClauseAnnotation::isSupport() const
{
  return _isSupport;
};
bool VampireKernel::ClauseAnnotation::isNonSubsuming() const
{
  return _isNonSubsuming;
};
bool VampireKernel::ClauseAnnotation::isNonSubsumable() const
{
  return _isNonSubsumable;
};
void VampireKernel::ClauseAnnotation::markAsGoal()
{
  _isGoal = true;
};
void VampireKernel::ClauseAnnotation::markAsSupport()
{
  _isSupport = true;
};
void VampireKernel::ClauseAnnotation::markAsNonSubsuming()
{
  _isNonSubsuming = true;
};
void VampireKernel::ClauseAnnotation::markAsNonSubsumable()
{
  _isNonSubsumable = true;
};
