/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef VAMPIRE_KERNEL_HPP
#define VAMPIRE_KERNEL_HPP
#include <iostream>
#include <fstream>
#include <string>
#include "jargon.hpp"
#include "VampireKernelEnvironment.hpp"
class VampireKernel
{
 public:
  class Options;
  class InputError;
  class Emergency;
  enum TerminationReason
  {
    UnknownTerminationReason,
    RefutationFound,
    Satisfiable,
    EmptyPassive,
    MaxPassive,
    MaxActive,
    TimeLimitExpired,
    AbortedOnAllocationFailure,
    AbortedOnExternalInterrupt,
    AbortedOnErrorSignal,
    AbortedOnError
  };
  enum 
  {
    DefaultPositiveHeaderRank = 0L,
    DefaultNegativeHeaderRank = 0L
  };
  class Statistics;
  class Subterm;
  class Literal;
  class Clause;
  class ClauseSetIterator;
  class ClauseAnnotation;
 public:
  VampireKernel();
  ~VampireKernel();
  void openSession(const Options& opt);
  void closeSession();
  static float versionNumber();
  static string versionDescription();
  void beginInputSignature();  
  void endInputSignature();
  void registerInputFunction(string name,
			     ulong arity,
			     ulong& id,
			     const long& weight,
			     long precedence); 
  void registerInputPredicate(string name,
			      ulong arity,
			      bool isAnswer,
			      ulong& id,
			      const long& weight,
			      long precedence,
			      long positiveHeaderRank,
			      long negativeHeaderRank); 
  void registerEquality(ulong& id,
			long positiveHeaderRank,
			long negativeHeaderRank); 
  void registerInterpretedFunction(ulong id);
  void registerInterpretedPredicate(ulong id);
  void beginInputClauses();
  void endInputClauses();
  void beginInputClause(ulong clauseNumber,
			void* origin,
			bool goal,
			bool support);
  void beginInputClause(ulong clauseNumber,
			void* origin,
			const ClauseAnnotation& annotation);
  void endInputClause();
  bool beginLiteral(bool positive,ulong predicateId);
  void markSelected();
  void markUnselected();
  void markRuleHead();
  void markSubquery();
  void markEqWithResolution();
  void markEqWithResolutionOnly();
  void markGivenPosEqOrientation();
  void markReversePosEqOrientation();
  void endLiteral();  
  void receiveFunction(ulong symId);
  void 
  receiveInterpretedConstant(const VampireKernelEnvironment::InterpretedConstant& value);
  void receiveVariable(ulong varNum);
  void saturate(ulong nextVacantClauseId,
		long timeLimit); 
  void lookForAnotherAnswer(int timeLimit);
  static VampireKernel* currentSession();
  void suspend();
  void resume();
  void reportInterrupt(int signalNumber);
  void reportErrorSignal(int signalNumber);
  void reportError();
  TerminationReason terminationReason() const;
  const Clause* refutation() const;
  ClauseSetIterator* activeClauses(); 
  void disposeOf(ClauseSetIterator* iterator); 
  bool isInterpretedConstantId(ulong symbolId) const;
  bool isInputSymbol(ulong symbolId) const; 
  bool isAnswerPredicateId(ulong symbolId) const;
  ulong symbolArity(ulong symbolId) const;
  string symbolPrintName(ulong symbolId) const; 
  void interpretedConstantValue(ulong symbolId,
				VampireKernelEnvironment::InterpretedConstant& value) const; 
  const Statistics& statistics(); 
  void outputStatistics(ostream* str,bool full);
 private:
  class _Implementation; 
 private:
  _Implementation* _implementation;
}; 
class VampireKernel::Emergency
{
 public:
  enum Type
  {
    AllocationFailure,
    TimeLimit
  };
 public:
  Emergency(const Type& t) : type_(t) {}; 
  ~Emergency() {};
  const Type& type() const { return type_; };
 private:
  Emergency() {};
 private:
  Type type_;
}; 
class VampireKernel::InputError
{
 public:
  InputError(const std::string& description) : description_(description) {}; 
  ~InputError() {};
  const std::string& description() const { return description_; };
 private:
  InputError() {};
 private:
  std::string description_;
}; 
class VampireKernel::Statistics
{
public:
  class DeductionInferences
  {
  public:
    ulong totalAccomplished;
    ulong resolutionAccomplished;
    ulong forwardSuperposition;
    ulong backwardSuperposition;
    ulong equalityFactoring;
    ulong equalityResolution;
    ulong forwardChaining;
    ulong backwardChaining;
    ulong totalRedundantDiscarded;
    ulong totalRedundantBranchesDiscarded;
    ulong totalNonredundantDiscarded;
    ulong totalNonredundantBranchesDiscarded;
  }; 
  class SimplifyingInferences
  {
  public:
    ulong totalForward;
    ulong propositionalTautologyDeletion;
    ulong equationalTautologyDeletion;
    ulong forwardSubsumption;
    ulong forwardSubsumptionResolution;
    ulong forwardDemodulation;
    ulong onBranchForwardDemodulation;
    ulong destructiveEqualityResolution;
    ulong forwardFolding;
    ulong externalEvaluation;    
    ulong totalBackward;
    ulong backwardSubsumption;
    ulong backwardDemodulation;
    ulong onBranchBackwardDemodulation;
  }; 
  class PropositionalSplitting
  {
  public:
    ulong accomplished;
    float averageNumOfComponents;
    ulong introducedComponentNames;
  }; 
  class NegativeEqualitySplitting
  {
  public:
    ulong accomplished;
    float averageNumOfArgumentsSplitOff;
    ulong introducedNames;
  }; 
  class GeneratedClauses
  {
  public:
    ulong total;
    ulong byDeductionInferences;
    ulong bySimplificationInferences;
    ulong byPropositionalSplitting;
    ulong byNegativeEqualitySplitting;
    ulong byOther;    
    ulong totalRedundantDiscarded; 
    ulong discardedPropositionalTautologies; 
    ulong discardedEquationalTautologies; 
    ulong discardedSubsumed; 
    ulong totalNonredundantDiscarded;
    ulong totalSimplified;
    ulong transformedByPropositionalSplitting;
    ulong transformedByNegativeEqualitySplitting;
  }; 
  class RetainedClauses
  {
  public:
    ulong total;
    ulong selected;
    ulong totalSubsumed;
    ulong totalSimplified;
    ulong totalSimplifiedByDemodulation;
    ulong totalSimplifiedByOnBranchDemodulation;
    ulong totalOrphansMurdered; 
    ulong totalRecycled;
    ulong currentlyPassive;
    ulong currentlyActive;
  }; 
public:
  float time; 
  ulong memory; 
  DeductionInferences deductionInferences;
  SimplifyingInferences simplifyingInferences;
  PropositionalSplitting propositionalSplittingInferences;
  NegativeEqualitySplitting negativeEqualitySplittingInferences;
  GeneratedClauses generatedClauses;
  RetainedClauses retainedClauses;
}; 
class VampireKernel::Subterm
{
 public:
  const Subterm* next() const; 
  bool isVariable() const;
  ulong variableNumber() const; 
  ulong topFunction() const;
  const Subterm* firstArgument() const; 
 private:
  Subterm() {};
  ~Subterm() {};
  friend class VampireKernel;
}; 
class VampireKernel::Literal
{
 public:
  const Literal* next() const;
  bool isPositive() const;
  ulong predicate() const;
  const Subterm* firstArgument() const; 
 private:
  Literal() {};
  ~Literal() {};
  friend class VampireKernel;
}; 
class VampireKernel::Clause
{
 public:
  class Ancestor
    {
    public:
      const Clause* clause() const;
      const Ancestor* next() const;
    private:
      Ancestor() {};
      ~Ancestor() {};
      friend class Clause;
    };
  enum BackgroundFlag
    { 
      IsInputClause,
      IsObtainedDuringPreprocessing,
      BinaryResolutionUsed,
      ForwardTransitiveChainingUsed,
      BackwardTransitiveChainingUsed,
      ForwardSuperpositionUsed,
      BackwardSuperpositionUsed,
      EqualityResolutionUsed,
      EqualityFactoringUsed,
      ForwardDemodulationUsed,
      BackwardDemodulationUsed,
      SimplificationByEqualityResolutionUsed,
      ForwardSubsumptionResolutionUsed,
      BackwardSubsumptionResolutionUsed,
      IsReanimatedPassive,
      ExternalEvaluationUsed,
      IsNameIntro,
      FoldingUsed,
      ClausificationUsed
    };
  class BackgroundFlagIterator
  {
  public:
    BackgroundFlagIterator();
    ~BackgroundFlagIterator();
    bool getCurrentFlag(BackgroundFlag& flag) const;
    void goToNextFlag(); 
  private:
    const void* _base;
    unsigned long _offset;
    friend class Clause;
  };
 public:
  ulong number() const;
  bool isLiteralDefinition(const Literal*& definedLiteral,
			   const Literal*& bodyLiterals) const;
  const Literal* literals() const; 
  ulong numberOfSelectedLiterals() const; 
  bool isInputClause() const;
  void* inputClauseOrigin() const;
  bool isNameIntroduction() const; 
  const Ancestor* firstAncestor() const;
  BackgroundFlagIterator backgroundFlagIterator() const;
  bool backgroundFlagIsSet(BackgroundFlag fl) const; 
 private:
  Clause() {};
  ~Clause() {};
  friend class VampireKernel;
}; 
class VampireKernel::ClauseSetIterator
{
public:
  ClauseSetIterator() {};
  virtual ~ClauseSetIterator();
  operator bool() const { return !isEnd(); };
  virtual bool isEnd() const = 0;
  ClauseSetIterator& operator ++() { next(); return *this; };
  virtual void next() = 0;
  virtual const Clause* clause() const = 0;
}; 
class VampireKernel::ClauseAnnotation
{
public:
  ClauseAnnotation(); 
  void makeDefault();
  void make(bool isGoal,bool isSupport,bool isNonSubsuming,bool isNonSubsumable);
  bool isGoal() const;
  bool isSupport() const;
  bool isNonSubsuming() const;
  bool isNonSubsumable() const;  
  void markAsGoal();
  void markAsSupport();
  void markAsNonSubsuming();
  void markAsNonSubsumable();
private:
  bool _isGoal;
  bool _isSupport;
  bool _isNonSubsuming;
  bool _isNonSubsumable;
}; 
class VampireKernel::Options
{
 public:
  enum SimplificationOrdering
  {
    StandardKBOrdering,
    LexPathOrdering
  };
  enum TermWeightingScheme
  {
    UniformTermWeighting,
    NonuniformConstantTermWeighting
  };
  enum ElementaryLitListOrdering
  {
    BySize,
    ByNumberOfVarOccurences,
    ByNumberOfLiterals
  };
  class LitListOrdering 
  {
  public:
    explicit 
    LitListOrdering(ElementaryLitListOrdering tag); 
    explicit 
    LitListOrdering(const LitListOrdering* ord); 
    explicit
    LitListOrdering(const LitListOrdering* ord1,
		    const LitListOrdering* ord2); 
    bool isElementary(ElementaryLitListOrdering& tag) const;
    bool isInverse(const LitListOrdering*& ord) const;
    bool isComposition(const LitListOrdering*& ord1,
		       const LitListOrdering*& ord2) const;
  private:
    LitListOrdering();
  private:
    int _type;
    union 
    {
      ElementaryLitListOrdering _tag;
      const LitListOrdering* _ord1;
    };
    const LitListOrdering* _ord2;
  }; 
 public:
  Options() { setDefaultValues(); };
  ~Options() {};
  void setDefaultValues();
  std::ostream& output(std::ostream& str);
 public:
  const char* job_id;
  long main_alg;
  bool no_inferences_between_rules;
  bool static_splitting;
  bool dynamic_splitting;
  bool small_basic_comp_in_hyp_splitting;
  bool fold_single_comp_in_splitting;
  long single_ground_lit_comp_min_size;
  bool reuse_names_for_negated_ground_lits_in_splitting;
  const LitListOrdering* basic_comp_in_spl_is_max_wrt;
  bool back_fold_in_spl;
  long static_neg_eq_splitting;
  long stat_neg_eq_spl_var_count_limit;
  long dynamic_neg_eq_splitting;
  long dyn_neg_eq_spl_var_count_limit;
  bool paramodulation;
  bool paramodulate_from_variables;
  long sim_back_sup;
  bool forw_sup_no_dup_red;
  bool back_sup_no_dup_red;
  bool prohibit_non_basic_superposition;
  bool prohibit_non_basic_fd;
  bool prohibit_non_basic_bd;
  bool reduce_non_basic;
  bool input_non_basicness;
  bool run_time_non_basicness;
  long selection;
  bool inherited_negative_selection;
  bool external_evaluation;
  bool forward_subsumption;
  long simplify_by_fs;
  long main_premise_subsumption;
  long side_premise_subsumption;
  long main_selected_lit_demodulation;
  long side_selected_lit_demodulation;
  long forward_demodulation;
  bool fd_on_splitting_branches;
  bool normalize_by_commutativity;
  bool eq_taut_mod_C;
  bool prop_taut_mod_C;
  bool trivial_eq_res_mod_C;
  bool identical_lit_factoring_mod_C;
  bool chaining_for_transitive;
  bool chaining_through_variables;
  long backward_demodulation;
  bool bd_on_splitting_branches;
  bool simplify_by_eq_res;
  bool backward_subsumption;
  bool orphan_murder;
  long lrs;
  SimplificationOrdering simplification_ordering;
  TermWeightingScheme term_weighting_scheme;
  bool orient_pos_equalities_in_support;
  bool orient_neg_equalities_in_support;
  long memory_limit; 
  long allocation_buffer_size;
  double allocation_buffer_price;
  long max_skolem; 
  long max_active;
  long max_passive;
  long max_weight;
  long max_inference_depth;
  long age_weight_ratio_numerator;
  long age_weight_ratio_denominator;
  long best_good_weight_ratio_numerator;
  long best_good_weight_ratio_denominator;
  float selected_pos_eq_penalty_coefficient;    
  float nongoal_penalty_coefficient;
  long simulated_time_limit;
  long first_time_check;   
  long first_mem_check;   
  bool show_opt;
  bool show_input;
  bool show_gen;
  bool show_passive;
  bool show_active;
  bool show_name_intro;
  bool show_lit_prop; 
  long show_passive_profile_interval;
  bool auto_stat_full;
  bool auto_stat_summary;
  bool show_profile;
  bool show_weight_limit_changes;  
  std::ostream* std_output;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
  bool step_resolution_literal_selection;
#endif
}; 
namespace std
{
inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::SimplificationOrdering& so)
{
  switch (so)
    {
    case VampireKernel::Options::StandardKBOrdering: 
      str << "standard_KB";
      break;
    case VampireKernel::Options::LexPathOrdering: 
      str << "LPO";
      break;
    };
  return str;
};
inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::TermWeightingScheme& tws)
{
  switch (tws)
    {
    case VampireKernel::Options::UniformTermWeighting:
      str << "uniform";
      break;
    case VampireKernel::Options::NonuniformConstantTermWeighting:
      str << "nonuniform_constant";
      break;
    };
  return str;
};
};
#endif
