/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "WeightPolynomial.hpp"
#include "WeightPolynomialSpecialisedComparisonCommand.hpp"
#include "Math.hpp"
using namespace BK;
using namespace VK;
WeightPolynomial::Sharing* WeightPolynomial::_sharing; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
void 
WeightPolynomial::SpecialisedComparison::compare(const WeightPolynomial& weight1,
						 const WeightPolynomial& weight2)
{
  CALL("compare(..)");
  _positivePart.reset();
  _negativePart.reset();
  const ulong* endOfVars = weight1.monomials().end();
  const ulong* v;
  for (v = weight1.monomials().begin();
       v != endOfVars;
       ++v)
    {
      long coeffDiff = weight1.coefficient(*v) - weight2.coefficient(*v);
      if (coeffDiff > 0)
	{
	  _positivePart.add(coeffDiff,*v);
	}
      else
	if (coeffDiff < 0)
	  {
	    _negativePart.add(-coeffDiff,*v);
	  };
    };
  endOfVars = weight2.monomials().end();
  for (v = weight2.monomials().begin();
       v != endOfVars;
       ++v)
    {
      if (!weight1.containsVariable(*v))
	{
	  _negativePart.add(weight2.coefficient(*v),*v);
	};
    };
  if (weight1.freeMember() > weight2.freeMember())
    {
      _positivePart.add(weight1.freeMember() - weight2.freeMember());
    }
  else
    {
      _negativePart.add(weight2.freeMember() - weight1.freeMember());
    };
  if (_positivePart.isConstant())
    { 
      if (_negativePart.isConstant())
	{ 
	  if (_positivePart.freeMember() == _negativePart.freeMember())
	    {
	      ASSERT(!_positivePart.freeMember());
	      _category = AlwaysEqual;
	    }
	  else
	    if (_positivePart.freeMember())
	      {
		ASSERT(!_negativePart.freeMember());
		_category = AlwaysGreater;
	      }
	    else
	      {
		ASSERT(_negativePart.freeMember());
		_category = AlwaysLess;
	      };
	}
      else
	{
	  long negMinVal = _negativePart.minimalValue();
	  if (_positivePart.freeMember() < negMinVal)
	    {
	      _category = AlwaysLess;
	    }
	  else
	    if (_positivePart.freeMember() == negMinVal)
	      {
		_category = CanBeLessOrEqual;
	      }
	    else
	      _category = VolatileButNearlyAlwaysLess;
	};
    }
  else
    if (_negativePart.isConstant())
      {    
	long posMinVal = _positivePart.minimalValue();
	if (posMinVal > _negativePart.freeMember())
	  {
	    _category = AlwaysGreater;
	  }
	else
	  if (posMinVal == _negativePart.freeMember())
	    {
	      _category = CanBeGreaterOrEqual;
	    }
	  else
	    _category = VolatileButNearlyAlwaysGreater;
      }
    else
      _category = AbsolutelyVolatile;
  factoriseCoefficients();
}; 
void WeightPolynomial::SpecialisedComparison::factoriseCoefficients()
{
  CALL("factoriseCoefficients()");
  long gcdPos = _positivePart.greatestCommonDivisorOfCoefficients();
  long gcdNeg = _negativePart.greatestCommonDivisorOfCoefficients();
  long gcd = Math::greatestCommonDivisor(gcdPos,gcdNeg);
  if (gcd != 1)
    {
      _positivePart.divideBy(gcd);
      _negativePart.divideBy(gcd);
    };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
#  define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison::Extras"
#endif
#include "debugMacros.hpp"
bool 
WeightPolynomial::SpecialisedComparison::Extras::simplifyByEqualities(const SpecialisedComparison& comp, 
								      const SpecialisedComparison& relations,
								      Handle& simplified,
								      Handle& spareComparisonObject1,
								      Handle& spareComparisonObject2)
{
  CALL("simplifyByEqualities(..)");
  ASSERT(!simplified.isNull());
  ASSERT(!spareComparisonObject1.isNull());
  ASSERT(!spareComparisonObject2.isNull());
  bool result = false;
  const SpecialisedComparison* lastGoodEq = 0;
  bool simplifiedInInnerLoop;
  do
    {
      simplifiedInInnerLoop = false;
      for (const SpecialisedComparison* rel = &relations; rel; rel = rel->next())
	{
	  if (rel->assumedRelation() == AlwaysEqual)
	    {
	      if ((rel == lastGoodEq) && (!simplifiedInInnerLoop))
		return result;
	      bool eqWorks;
	      if (result)
		{
		  eqWorks = simplifyByEquality(*simplified,
					       *rel,
					       spareComparisonObject1,
					       spareComparisonObject2);
		  if (eqWorks)
		    {
		      simplified.swapWith(spareComparisonObject1);
		      if (strict(simplified->category()))
			return true;
		      lastGoodEq = rel;
		    };
		}
	      else 
		{
		  eqWorks = simplifyByEquality(comp,
					       *rel,
					       simplified,
					       spareComparisonObject1);
		  if (eqWorks) 
		    {
		      if (strict(simplified->category()))
			return true;
		      result = true;
		      lastGoodEq = rel;
		    };
		};
	      simplifiedInInnerLoop = simplifiedInInnerLoop || eqWorks;
	    };
	};
    }
  while (simplifiedInInnerLoop);
  return result;
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::isAlwaysGreaterUnderAssumptions(const SpecialisedComparison& comp, 
										 const SpecialisedComparison& assumedRelations,
										 Handle& spareComparisonObject1,
										 Handle& spareComparisonObject2,
										 Handle& spareComparisonObject3)
{
  CALL("isAlwaysGreaterUnderAssumptions(..)");
  ASSERT(!spareComparisonObject1.isNull());
  ASSERT(!spareComparisonObject2.isNull());
  ASSERT(!spareComparisonObject3.isNull());
  bool someRelationWorked = false;
  const SpecialisedComparison* lastGoodRel = 0;
  bool simplifiedInInnerLoop;
  do
    {
      simplifiedInInnerLoop = false;
      for (const SpecialisedComparison* rel = &assumedRelations; rel; rel = rel->next())
	{
	  if ((rel == lastGoodRel) && (!simplifiedInInnerLoop))
	    return false;
	  switch (rel->assumedRelation())
	    {
	    case AlwaysEqual:
	      {
		bool eqWorks;
		if (someRelationWorked)
		  {
		    eqWorks = simplifyByEquality(*spareComparisonObject1,
						 *rel,
						 spareComparisonObject2,
						 spareComparisonObject3);
		    if (eqWorks)
		      {
			spareComparisonObject1.swapWith(spareComparisonObject2);
			if (strict(spareComparisonObject1->category()))
			  {
			    return spareComparisonObject1->category() == AlwaysGreater;
			  };
			lastGoodRel = rel;
		      };
		  }
		else 
		  {
		    eqWorks = 
		      simplifyByEquality(comp,
					 *rel,
					 spareComparisonObject1,
					 spareComparisonObject2);
		    if (eqWorks) 
		      {
			if (strict(spareComparisonObject1->category()))
			  {
			    return spareComparisonObject1->category() == AlwaysGreater;
			  };
			someRelationWorked = true;
			lastGoodRel = rel;
		      };
		  };
		simplifiedInInnerLoop = simplifiedInInnerLoop || eqWorks;
	      };
	      break;
	    case CanBeGreaterOrEqual:
	      {
		bool ineqWorks;
		if (someRelationWorked)
		  {		    
		    ineqWorks = 
		      relaxAlwaysGreaterByInequality(*spareComparisonObject1,
						     *rel,
						     spareComparisonObject2,
						     spareComparisonObject3);
		    if (ineqWorks)
		      {
			spareComparisonObject1.swapWith(spareComparisonObject2);
			if (spareComparisonObject1->category() == BK::AlwaysGreater)
			  return true;
			ASSERT(!strict(spareComparisonObject1->category()));
			lastGoodRel = rel;
		      };
		  }
		else 
		  {		    
		    ineqWorks = relaxAlwaysGreaterByInequality(comp,
							       *rel,
							       spareComparisonObject1,
							       spareComparisonObject2);
		    if (ineqWorks) 
		      {			
			if (spareComparisonObject1->category() == BK::AlwaysGreater)
			  return true;
			ASSERT(!strict(spareComparisonObject1->category()));
			someRelationWorked = true;
			lastGoodRel = rel;
		      };
		  };
		simplifiedInInnerLoop = simplifiedInInnerLoop || ineqWorks;
	      };
	      break;
	    default: 
	      break;
	    }; 
	}; 
    }
  while (simplifiedInInnerLoop);
  return false;
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::isAlwaysLessUnderAssumptions(const SpecialisedComparison& comp, 
									      const SpecialisedComparison& assumedRelations,
									      Handle& spareComparisonObject1,
									      Handle& spareComparisonObject2,
									      Handle& spareComparisonObject3)
{
  CALL("isAlwaysLessUnderAssumptions(..)");
  ASSERT(!spareComparisonObject1.isNull());
  ASSERT(!spareComparisonObject2.isNull());
  ASSERT(!spareComparisonObject3.isNull());
  bool someRelationWorked = false;
  const SpecialisedComparison* lastGoodRel = 0;
  bool simplifiedInInnerLoop;
  do
    {
      simplifiedInInnerLoop = false;
      for (const SpecialisedComparison* rel = &assumedRelations; rel; rel = rel->next())
	{
	  if ((rel == lastGoodRel) && (!simplifiedInInnerLoop))
	    return false;
	  switch (rel->assumedRelation())
	    {
	    case AlwaysEqual:
	      {
		bool eqWorks;
		if (someRelationWorked)
		  {	    
		    eqWorks = 
		      simplifyByEquality(*spareComparisonObject1,
					 *rel,
					 spareComparisonObject2,
					 spareComparisonObject3);
		    if (eqWorks)
		      {
			spareComparisonObject1.swapWith(spareComparisonObject2);
			if (strict(spareComparisonObject1->category()))
			  {
			    return spareComparisonObject1->category() == AlwaysLess;
			  };
			lastGoodRel = rel;
		      };
		  }
		else 
		  {
		    eqWorks = simplifyByEquality(comp,
						 *rel,
						 spareComparisonObject1,
						 spareComparisonObject2);
		    if (eqWorks) 
		      {
			if (strict(spareComparisonObject1->category()))
			  {
			    return spareComparisonObject1->category() == AlwaysLess;
			  };
			someRelationWorked = true;
			lastGoodRel = rel;
		      };
		  };
		simplifiedInInnerLoop = simplifiedInInnerLoop || eqWorks;
	      };
	      break;
	    case CanBeGreaterOrEqual:
	      {
		bool ineqWorks;
		if (someRelationWorked)
		  {
		    ineqWorks = 
		      relaxAlwaysLessByInequality(*spareComparisonObject1,
						  *rel,
						  spareComparisonObject2,
						  spareComparisonObject3);
		    if (ineqWorks)
		      {
			spareComparisonObject1.swapWith(spareComparisonObject2);
			if (spareComparisonObject1->category() == BK::AlwaysLess)
			  return true;
			ASSERT(!strict(spareComparisonObject1->category()));
			lastGoodRel = rel;
		      };
		  }
		else 
		  {		    
		    ineqWorks = 
		      relaxAlwaysLessByInequality(comp,
						  *rel,
						  spareComparisonObject1,
						  spareComparisonObject2);
		    if (ineqWorks) 
		      {			
			if (spareComparisonObject1->category() == BK::AlwaysLess)
			  return true;
			ASSERT(!strict(spareComparisonObject1->category()));
			someRelationWorked = true;
			lastGoodRel = rel;
		      };
		  };
		simplifiedInInnerLoop = simplifiedInInnerLoop || ineqWorks;
	      };
	      break;
	    default: 
	      break;
	    }; 
	}; 
    }
  while (simplifiedInInnerLoop);
  return false;
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::simplifyByEquality(const SpecialisedComparison& comp, 
								    const SpecialisedComparison& equality, 
								    Handle& simplified,
								    Handle& spareComparisonObject)
{
  CALL("simplifyByEquality(..)");
  ASSERT(equality.assumedRelation() == AlwaysEqual);
  ASSERT(!simplified.isNull());
  ASSERT(!spareComparisonObject.isNull());
  long simplifyingCoeffForEquality;
  long simplifyingCoeffForComp;
  long bestSimplifyingCoeffForEquality;
  long bestSimplifyingCoeffForComp;
  ulong bestSimplificationStrength = 0UL;
  ulong origNumOfVariables = comp.numberOfVariables();
  static WeightPolynomial simplifiedPositive;
  static WeightPolynomial simplifiedNegative;
  const ulong* v;
  for (v = equality._positivePart.monomials().begin();
       v != equality._positivePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForEquality = -comp._positivePart.monomials().coefficient(*v);
      if (!simplifyingCoeffForEquality)
	{
	  simplifyingCoeffForEquality = comp._negativePart.monomials().coefficient(*v);
	};
      if (simplifyingCoeffForEquality) 
	{
	  simplifyingCoeffForComp = equality._positivePart.monomials().coefficient(*v);
	  applyRelation(comp,
			equality,
			simplifyingCoeffForEquality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {
	      return true;
	    };
	  ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	  if ((numOfVariablesInSimplified < origNumOfVariables) &&
	      (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
	    {
	      bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
	      bestSimplifyingCoeffForEquality = -simplifyingCoeffForEquality;
	      bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
	      simplified.swapWith(spareComparisonObject);
	    };
	};
    }; 
  for (v = equality._negativePart.monomials().begin();
       v != equality._negativePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForEquality = comp._positivePart.monomials().coefficient(*v);
      if (!simplifyingCoeffForEquality)
	{
	  simplifyingCoeffForEquality = -comp._negativePart.monomials().coefficient(*v);
	};
      if (simplifyingCoeffForEquality) 
	{
	  simplifyingCoeffForComp = equality._negativePart.monomials().coefficient(*v);
	  applyRelation(comp,
			equality,
			simplifyingCoeffForEquality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {
	      return true;
	    };
	  ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	  if ((numOfVariablesInSimplified < origNumOfVariables) &&
	      (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
	    {
	      bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
	      bestSimplifyingCoeffForEquality = -simplifyingCoeffForEquality;
	      bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
	      simplified.swapWith(spareComparisonObject);
	    };
	};
    }; 
  if (bestSimplificationStrength) 
    {      
      simplified.swapWith(spareComparisonObject);
      return true;
    }
  else
    return false;
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::relaxAlwaysGreaterByInequality(const SpecialisedComparison& comp, 
										const SpecialisedComparison& inequality,
										Handle& simplified,
										Handle& spareComparisonObject)
{
  CALL("relaxAlwaysGreaterByInequality(..)");
  ASSERT(inequality.assumedRelation() == CanBeGreaterOrEqual);
  ASSERT(!simplified.isNull());
  ASSERT(!spareComparisonObject.isNull());
  long simplifyingCoeffForInequality;
  long simplifyingCoeffForComp;
  long bestSimplifyingCoeffForInequality;
  long bestSimplifyingCoeffForComp;
  ulong bestSimplificationStrength = 0UL;
  ulong origNumOfVariables = comp.numberOfVariables();
  static WeightPolynomial simplifiedPositive;
  static WeightPolynomial simplifiedNegative;
  const ulong* v;
  for (v = inequality._positivePart.monomials().begin();
       v != inequality._positivePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForInequality = -comp._positivePart.monomials().coefficient(*v);
      if (simplifyingCoeffForInequality) 
	{
	  simplifyingCoeffForComp = inequality._positivePart.monomials().coefficient(*v);
	  applyRelation(comp,
			inequality,
			simplifyingCoeffForInequality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {
	      if (simplified->category() == BK::AlwaysGreater)
		{
		  return true; 
		}
	      else
		return false; 
	    }
	  else
	    {
	      ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	      if ((numOfVariablesInSimplified < origNumOfVariables) &&
		  (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
		{
		  bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
		  bestSimplifyingCoeffForInequality = -simplifyingCoeffForInequality;
		  bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
		  simplified.swapWith(spareComparisonObject);
		};
	    };
	};
    }; 
  for (v = inequality._negativePart.monomials().begin();
       v != inequality._negativePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForInequality = -comp._negativePart.monomials().coefficient(*v);
      if (simplifyingCoeffForInequality) 
	{
	  simplifyingCoeffForComp = inequality._negativePart.monomials().coefficient(*v);
	  ASSERT(simplifyingCoeffForComp);
	  applyRelation(comp,
			inequality,
			simplifyingCoeffForInequality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {	      
	      if (simplified->category() == BK::AlwaysGreater)
		{
		  return true; 
		}
	      else
		return false; 
	    }
	  else
	    {
	      ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	      if ((numOfVariablesInSimplified < origNumOfVariables) &&
		  (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
		{
		  bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
		  bestSimplifyingCoeffForInequality = -simplifyingCoeffForInequality;
		  bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
		  simplified.swapWith(spareComparisonObject);
		};
	    };
	};
    }; 
  if (bestSimplificationStrength) 
    {      
      simplified.swapWith(spareComparisonObject);
      return true;
    }
  else
    return false;
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::relaxAlwaysLessByInequality(const SpecialisedComparison& comp, 
									     const SpecialisedComparison& inequality,
									     Handle& simplified,
									     Handle& spareComparisonObject)
{
  CALL("relaxAlwaysLessByInequality(..)");
  ASSERT(inequality.assumedRelation() == CanBeGreaterOrEqual);
  ASSERT(!simplified.isNull());
  ASSERT(!spareComparisonObject.isNull());
  long simplifyingCoeffForInequality;
  long simplifyingCoeffForComp;
  long bestSimplifyingCoeffForInequality;
  long bestSimplifyingCoeffForComp;
  ulong bestSimplificationStrength = 0UL;
  ulong origNumOfVariables = comp.numberOfVariables();
  static WeightPolynomial simplifiedPositive;
  static WeightPolynomial simplifiedNegative;
  const ulong* v;
  for (v = inequality._positivePart.monomials().begin();
       v != inequality._positivePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForInequality = comp._negativePart.monomials().coefficient(*v);
      if (simplifyingCoeffForInequality) 
	{
	  simplifyingCoeffForComp = inequality._positivePart.monomials().coefficient(*v);
	  applyRelation(comp,
			inequality,
			simplifyingCoeffForInequality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {
	      if (simplified->category() == BK::AlwaysLess)
		{
		  return true; 
		}
	      else
		return false; 
	    }
	  else
	    {
	      ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	      if ((numOfVariablesInSimplified < origNumOfVariables) &&
		  (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
		{
		  bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
		  bestSimplifyingCoeffForInequality = simplifyingCoeffForInequality;
		  bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
		  simplified.swapWith(spareComparisonObject);
		};
	    };
	};
    }; 
  for (v = inequality._negativePart.monomials().begin();
       v != inequality._negativePart.monomials().end();
       ++v)
    {
      simplifyingCoeffForInequality = comp._positivePart.monomials().coefficient(*v);
      if (simplifyingCoeffForInequality) 
	{
	  simplifyingCoeffForComp = inequality._negativePart.monomials().coefficient(*v);
	  ASSERT(simplifyingCoeffForComp);
	  applyRelation(comp,
			inequality,
			simplifyingCoeffForInequality,
			simplifyingCoeffForComp,
			*simplified);
	  if (strict(simplified->category()))
	    {	      
	      if (simplified->category() == BK::AlwaysLess)
		{
		  return true; 
		}
	      else
		return false; 
	    }
	  else
	    {
	      ulong numOfVariablesInSimplified = simplified->numberOfVariables();
	      if ((numOfVariablesInSimplified < origNumOfVariables) &&
		  (origNumOfVariables - numOfVariablesInSimplified > bestSimplificationStrength))
		{
		  bestSimplificationStrength = origNumOfVariables - numOfVariablesInSimplified;
		  bestSimplifyingCoeffForInequality = simplifyingCoeffForInequality;
		  bestSimplifyingCoeffForComp = simplifyingCoeffForComp;
		  simplified.swapWith(spareComparisonObject);
		};
	    };
	};
    }; 
  if (bestSimplificationStrength) 
    {      
      simplified.swapWith(spareComparisonObject);
      return true;
    }
  else
    return false;
}; 
void 
WeightPolynomial::SpecialisedComparison::Extras::applyRelation(const SpecialisedComparison& comp, 
							       const SpecialisedComparison& relation,
							       const long& coeffForRelation,
							       const long& coeffForComp,
							       SpecialisedComparison& result)
{
  CALL("applyRelation(..)");
  ASSERT(coeffForRelation != 0);
  ASSERT(coeffForComp > 0);
  static WeightPolynomial resultPositive;
  static WeightPolynomial resultNegative;
  resultPositive.reset();
  resultNegative.reset();
  resultPositive.add(coeffForComp,comp._positivePart);
  resultNegative.add(coeffForComp,comp._negativePart);
  if (coeffForRelation > 0)
    {
      resultPositive.add(coeffForRelation,relation._positivePart);
      resultNegative.add(coeffForRelation,relation._negativePart);
    }
  else
    {
      resultPositive.add(-coeffForRelation,relation._negativePart);
      resultNegative.add(-coeffForRelation,relation._positivePart);
    };
  result.compare(resultPositive,resultNegative);
}; 
bool 
WeightPolynomial::SpecialisedComparison::Extras::compileGreaterOrEqual(const FunctionComparison& cat,
								       const WeightPolynomial& posPart,
								       const WeightPolynomial& negPart,
								       Command*& code,
								       ulong maxCodeSize)
{
  CALL("compileGreaterOrEqual(const FunctionComparison& cat,..)");
  ASSERT(maxCodeSize);
  Command* start = code;
  const ulong* v;
  DOP(v = 0);
  switch (cat)
    {
    case FailureToCompare:
      code->setTag(Command::RetFailureToCompare);
      ++code;
      return true;
    case AlwaysLess:
      code->setTag(Command::RetFailureToCompare);
      ++code;
      return true;
    case AlwaysEqual:
      code->setTag(Command::RetAlwaysEqual);
      ++code;
      return true;
    case AlwaysGreater:
      code->setTag(Command::RetAlwaysGreater);
      ++code;
      return true;
    case CanBeLessOrEqual:
      {
	ASSERT((!negPart.isConstant()) && (!negPart.freeMember()));
	ASSERT(posPart.isConstant() && posPart.freeMember());	
	v = negPart.monomials().begin();
	while (v != negPart.monomials().last())
	  {
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    code->setTag(Command::IfXGr1RetFail_IfNotGrndRetFail);
	    code->setVar(*v);
	    ++v;
	    ++code;
	  };
	ASSERT(v == negPart.monomials().last());
	if ((code - start) + 1 == (long)maxCodeSize)
	  {
	    code->setTag(Command::RetFailureToCompare);
	    ++code;
	    return false;
	  };
	code->setTag(Command::IfXGr1RetFail_IfGrndRetAlwEq_RetFail);
	code->setVar(*v);
	++code;
	ASSERT((code - start) <= (long)maxCodeSize);
	return true;
      }; 
    case CanBeGreaterOrEqual:
      {
	ASSERT((!posPart.isConstant()) && (!posPart.freeMember()));
	ASSERT(negPart.isConstant() && negPart.freeMember());
	v = posPart.monomials().begin();
	if (posPart.numberOfVariables() == 1UL)
	  {
	    code->setTag(Command::IfXGr1RetAlwGr_IfGrndRetAlwEq_RetCanBeGrOrEq);
	    code->setVar(*v);
	    ++code;
	    return true;	    
	  }
	else 
 	  {
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetCanBeGreaterOrEqual);
		++code;
		return false;
	      };
	    code->setTag(Command::IfXGr1RetAlwGr_GrndToGRND);
	    code->setVar(*v);
	    ++code;
	    ++v;
	    while (v != posPart.monomials().last())
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetCanBeGreaterOrEqual);
		    ++code;
		    return false;
		  };
		code->setTag(Command::IfXGr1RetAlwGr_UpdateGRND);
		code->setVar(*v);
		++v;
		++code;
	      };
	    ASSERT(v == posPart.monomials().last());
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetCanBeGreaterOrEqual);
		++code;
		return false;
	      };
	    code->setTag(Command::IfXGr1RetAlwGr_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq);
	    code->setVar(*v);
	    ++code;
	    ASSERT((code - start) <= (long)maxCodeSize);
	    return true;
	  };
      }; 
    case VolatileButNearlyAlwaysLess:
      {
	ASSERT((!negPart.isConstant()) && (!negPart.freeMember()));
	ASSERT(posPart.isConstant() && posPart.freeMember());	
	v = negPart.monomials().begin();
	if (negPart.numberOfVariables() == 1UL)
	  {
	    if (negPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq);
	      }
	    else
	      {
		code->setTag(Command::IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq);
		code->setCoefficient(negPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    code->setWeight(posPart.freeMember());
	    ++code;
	    return true;	
	  }
	else 
	  {
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    if (negPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::IfNotGrndRetFail_WMinusXToACC_IfACCLe0RetFail);
	      }
	    else
	      {
		code->setTag(Command::IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail);
		code->setCoefficient(negPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    code->setWeight(posPart.freeMember() - (negPart.minimalValue() - negPart.coefficient(*v)));
	    ++code;
	  vol_but_nearly_alw_less_all_monomials_except_first:
	    ++v;
	    while (v != negPart.monomials().last())
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail);
		  }
		else
		  {
		    code->setTag(Command::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++v;
		++code;
	      };
	    ASSERT(v == negPart.monomials().last());
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    if (negPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr);
	      }
	    else
	      {
		code->setTag(Command::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr);
		code->setCoefficient(negPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    ++code;
	    ASSERT((code - start) <= (long)maxCodeSize);
	    return true;
	  };
      }; 
    case VolatileButNearlyAlwaysGreater:
      {
	ASSERT((!posPart.isConstant()) && (!posPart.freeMember()));
	ASSERT(negPart.isConstant() && negPart.freeMember());
	v = posPart.monomials().begin();
	if (posPart.numberOfVariables() == 1UL)
	  {
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq);
	      }
	    else
	      {
		code->setTag(Command::IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*(posPart.monomials().begin()));
	    code->setWeight(negPart.freeMember());
	    ++code;
	    return true;
	  }
	else 
 	  {
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::WMinusXToACC_IfACCLe0RetAlwGr_GrndToGRND);
	      }
	    else
	      {
		code->setTag(Command::WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    code->setWeight(negPart.freeMember() - (posPart.minimalValue() - posPart.coefficient(*v)));
	    ++code;
	  vol_but_nearly_alw_greater_all_monomials_except_first:
	    ++v;
	    while (v != posPart.monomials().last())
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (posPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::ACCMinusXPlus1ToACC_IfACCLe0RetAlwGr_UpdateGRND);
		  }
		else
		  {
		    code->setTag(Command::ACCMinusCXPlusCToACC_IfACCLe0RetAlwGr_UpdateGRND);
		    code->setCoefficient(posPart.coefficient(*v));
		  };
		code->setVar(*v);
		++v;
		++code;
	      };
	    ASSERT(v == posPart.monomials().last());
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::IfXGrACCPlus1RetAlwGr_IfXLeACCPlus1RetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq);
	      }
	    else
	      {
		code->setTag(Command::IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    ++code;
	    ASSERT((code - start) <= (long)maxCodeSize);
	    return true;
	  };
      }; 
    case AbsolutelyVolatile: 
      {	
	ASSERT((!posPart.isConstant()) && (!negPart.isConstant()));
	ASSERT((!posPart.freeMember()) || (!negPart.freeMember()));
	Command* commandWithLabel;
	DOP(commandWithLabel = 0);
	if (posPart.numberOfVariables() <= negPart.numberOfVariables())
	  {
	    long constPart = posPart.freeMember() - negPart.minimalValue();
	    v = posPart.monomials().begin();
	    if (posPart.numberOfVariables() == 1UL)
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (posPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL);
		  }
		else
		  {
		    code->setTag(Command::CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL);
		    code->setCoefficient(posPart.coefficient(*v));
		  };
		code->setWeight(constPart);
		code->setVar(*v);
		code->setLabel(code + 1); 
		commandWithLabel = code; 
		++code;
	      }
	    else 
	      {		
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (posPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XPlusWToPOSINST);
		  }
		else
		  {
		    code->setTag(Command::CXPlusWToPOSINST);
		    code->setCoefficient(posPart.coefficient(*v));
		  };
		code->setVar(*v);
		code->setWeight(constPart);
		++code;
		++v;
		while (v != posPart.monomials().last())
		  { 		    
		    if ((code - start) + 1 == (long)maxCodeSize)
		      {
			code->setTag(Command::RetFailureToCompare);
			++code;
			return false;
		      };
		    if (posPart.coefficient(*v) == 1)
		      {
			code->setTag(Command::POSINSTPlusXToPOSINST);
		      }
		    else
		      {			
			code->setTag(Command::POSINSTPlusCXToPOSINST);
			code->setCoefficient(posPart.coefficient(*v));
		      };
		    code->setVar(*v);
		    ++v;
		    ++code;
		  };  
		ASSERT(v == posPart.monomials().last());
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (posPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::POSINSTPlusXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL);
		  }
		else
		  {			
		    code->setTag(Command::POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL);
		    code->setCoefficient(posPart.coefficient(*v));
		  };
		code->setVar(*v);
		code->setLabel(code + 1); 
		commandWithLabel = code; 
		++code;
	      }; 
	    ASSERT(code - start == (long)posPart.numberOfVariables());
	    v = negPart.monomials().begin();
	    if (negPart.numberOfVariables() == 1UL)
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XMinus1ToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST);
		  }
		else
		  {
		    code->setTag(Command::CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST);
		    code->setCoefficient(negPart.coefficient(*v));
		  }; 
		code->setVar(*v);
		++code;
	      }
	    else 
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail);
		  }
		else
		  {
		    code->setTag(Command::CXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++code;
		++v;
		while (v != negPart.monomials().last())
		  { 		    
		    if ((code - start) + 1 == (long)maxCodeSize)
		      {
			code->setTag(Command::RetFailureToCompare);
			++code;
			return false;
		      };
		    if (negPart.coefficient(*v) == 1)
		      {
			code->setTag(Command::NEGINSTPlusXMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail);
		      }
		    else
		      {			
			code->setTag(Command::NEGINSTPlusCXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail);
			code->setCoefficient(negPart.coefficient(*v));
		      };
		    code->setVar(*v);
		    ++v;
		    ++code;
		  }; 
		ASSERT(v == negPart.monomials().last());  
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::NEGINSTPlusXMinus1ToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST);
		  }
		else
		  {			
		    code->setTag(Command::NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++code;
	      }; 
	    ASSERT(code - start == (long)(posPart.numberOfVariables() + negPart.numberOfVariables()));
	    ASSERT((code - 1)->isTerminal());
	    commandWithLabel->setLabel(code);
	    v = negPart.monomials().begin();
	    if (negPart.numberOfVariables() == 1UL)
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::IfNotGrndRetFail_IfXMinus1GrPOSINSTRetFail_IfXMinus1LePOSINSTRetAlwGr_RetAlwEq);		    
		  }
		else
		  {
		    code->setTag(Command::IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++code;
		return true;
	      }
	    else 
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::IfNotGrndRetFail_POSINSTPlus1MinusXToACC_IfACCLe0RetFail);		    
		  }
		else
		  {
		    code->setTag(Command::IfNotGrndRetFail_POSINSTPlusCMinusCXToACC_IfACCLe0RetFail);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++code;
		goto vol_but_nearly_alw_less_all_monomials_except_first;		
	      };
	    ICP("ICP10");
	  }
	else 
	  {
	    long constPart = negPart.freeMember() - posPart.minimalValue();
	    v = negPart.monomials().begin();
	    if (negPart.numberOfVariables() == 1UL)
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL);
		    code->setWeight(constPart);		     
		  }
		else 
		  {		
		    code->setTag(Command::CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL);
		    code->setWeight(constPart);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		code->setLabel(code + 1); 
		commandWithLabel = code; 
		++code;
	      }
	    else
	      {
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::XPlusWToNEGINST);
		    code->setWeight(constPart);
		  }
		else 
		  {		   
		    code->setTag(Command::CXPlusWToNEGINST);
		    code->setWeight(constPart);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		++code;
		++v;
		while (v != negPart.monomials().last())
		  { 
		    if ((code - start) + 1 == (long)maxCodeSize)
		      {
			code->setTag(Command::RetFailureToCompare);
			++code;
			return false;
		      };
		    if (negPart.coefficient(*v) == 1)
		      {
			code->setTag(Command::NEGINSTPlusXToNEGINST);
		      }
		    else
		      {			
			code->setTag(Command::NEGINSTPlusCXToNEGINST);
			code->setCoefficient(negPart.coefficient(*v));
		      };
		    code->setVar(*v);
		    ++v;
		    ++code;
		  }; 
		ASSERT(v == negPart.monomials().last());  
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (negPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::NEGINSTPlusXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL);
		  }
		else
		  {			
		    code->setTag(Command::NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL);
		    code->setCoefficient(negPart.coefficient(*v));
		  };
		code->setVar(*v);
		code->setLabel(code + 1); 
		commandWithLabel = code; 
		++code;
	      };
	    ASSERT(code - start == (long)negPart.numberOfVariables());
	    ASSERT(posPart.numberOfVariables() > 1UL); 
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    v = posPart.monomials().begin();
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::XMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr);
	      }
	    else
	      {
		code->setTag(Command::CXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    ++code;
	    ++v;
	    while (v != posPart.monomials().last())
	      { 
		if ((code - start) + 1 == (long)maxCodeSize)
		  {
		    code->setTag(Command::RetFailureToCompare);
		    ++code;
		    return false;
		  };
		if (posPart.coefficient(*v) == 1)
		  {
		    code->setTag(Command::POSINSTPlusXMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr);
		  }
		else
		  {			
		    code->setTag(Command::POSINSTPlusCXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr);
		    code->setCoefficient(posPart.coefficient(*v));
		  };
		code->setVar(*v);
		++v;
		++code;
	      }; 
	    ASSERT(v == posPart.monomials().last()); 
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::POSINSTPlusXMinus1ToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST);
	      }
	    else
	      {			
		code->setTag(Command::POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    ++code; 
	    ASSERT(code - start == (long)(posPart.numberOfVariables() + negPart.numberOfVariables()));
	    ASSERT((code - 1)->isTerminal());
	    ASSERT(posPart.numberOfVariables() > 1UL); 
	    if ((code - start) + 1 == (long)maxCodeSize)
	      {
		code->setTag(Command::RetFailureToCompare);
		++code;
		return false;
	      };
	    commandWithLabel->setLabel(code);
	    v = posPart.monomials().begin();
	    if (posPart.coefficient(*v) == 1)
	      {
		code->setTag(Command::NEGINSTPlus1MinusXToACC_IfACCLe0RetAlwGr_GrndToGRND);
	      }
	    else
	      {			
		code->setTag(Command::NEGINSTPlusCMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND);
		code->setCoefficient(posPart.coefficient(*v));
	      };
	    code->setVar(*v);
	    ++code; 
	    goto vol_but_nearly_alw_greater_all_monomials_except_first;
	  };
	ICP("ICP20");
      }; 
    };
  ICP("END");
  return false;
}; 
