/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef WEIGHT_POLYNOMIAL_HPP
#define WEIGHT_POLYNOMIAL_HPP
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "LinearPolynomial.hpp"
#include "Comparison.hpp"
#include "SharedLinearPolynomial.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "WeightPolynomial"
#endif
#include "debugMacros.hpp"
namespace VK
{
class WeightPolynomial
{
 public:
  typedef BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices> Monomials;
  typedef BK::SharedLinearPolynomial<BK::GlobAlloc,long,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices> Stored;
  typedef BK::LinearPolynomialSharing<BK::GlobAlloc,long,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices> Sharing;
  class SpecialisedComparison;
 public:
  WeightPolynomial() {};
  ~WeightPolynomial() {};
  void reset() { _linearPolynomial.reset(); };
  static void setSharing(Sharing* s) 
    { 
      _sharing = s; 
    };
  Monomials& monomials() { return _linearPolynomial.monomials(); };
  const Monomials& monomials() const { return _linearPolynomial.monomials(); };
  const long& coefficient(ulong var) const 
    { 
      return monomials().coefficient(var); 
    };
  bool containsVariable(ulong var) const 
  { 
	  return coefficient(var) != 0; 
  };
  bool isConstant() const { return monomials().isZero(); };
  bool isZero() const { return isConstant() && (!freeMember()); };
  ulong numberOfVariables() const { return monomials().length(); };
  long& freeMember() { return _linearPolynomial.freeMember(); };
  const long& freeMember() const { return _linearPolynomial.freeMember(); };
  long minimalValue() const
    {
      return freeMember() + monomials().sumOfCoefficients();
    };
  bool equal(const WeightPolynomial& p) const;
  BK::FunctionComparison compare(const WeightPolynomial& p) const;
  BK::FunctionComparison compareForGreaterOrEqual(const WeightPolynomial& p) const;
  BK::FunctionComparison compareForLessOrEqual(const WeightPolynomial& p) const;
  bool isAlwaysGreater(const WeightPolynomial& p) const;
  bool isNeverLess(const WeightPolynomial& p) const;
  bool isAlwaysLess(const WeightPolynomial& p) const
    {
      CALL("isAlwaysLess(const WeightPolynomial& p) const");
      return p.isAlwaysGreater(*this);
    };
  bool isNeverGreater(const WeightPolynomial& p) const
    {
      CALL("isNeverGreater(const WeightPolynomial& p) const");
      return p.isNeverLess(*this);
    };
  bool containsVariableWhichIsNotIn(const WeightPolynomial& wp) const
    {
      return _linearPolynomial.containsVariableWhichIsNotIn(wp._linearPolynomial);
    };
  void add(const long& constant) { freeMember() += constant; };
  void add1() { ++freeMember(); };
  void add(const long& coeff,ulong var) { monomials().add(coeff,var); };
  void add1(ulong var) { monomials().add1(var); };
  void add(const long& coeff,const WeightPolynomial& p)
  {
    _linearPolynomial.add(coeff,p._linearPolynomial);
  };
  void divideBy(const long& c)
  {
    _linearPolynomial.divideBy(c);
  };
  long greatestCommonDivisorOfCoefficients() const
  {
    CALL("greatestCommonDivisorOfCoefficients() const");
    long gcd;
    if (_linearPolynomial.freeMember())
      {
	gcd = _linearPolynomial.freeMember();
	for (const ulong* v = _linearPolynomial.monomials().begin();
	     v != _linearPolynomial.monomials().end();
	     ++v)
	  {
	    if (gcd == 1) return 1;
	    gcd = BK::Math::greatestCommonDivisor(gcd,_linearPolynomial.monomials().coefficient(*v));	    
	  };
      }
    else
      {
	if (_linearPolynomial.monomials().length())
	  {
	    const ulong* v = _linearPolynomial.monomials().begin();
	    gcd = _linearPolynomial.monomials().coefficient(*v);
	    for (++v; v != _linearPolynomial.monomials().end(); ++v)
	      {
		if (gcd == 1) return 1;
		gcd = 
		  BK::Math::greatestCommonDivisor(gcd,
							      _linearPolynomial.monomials().coefficient(*v));	    
	      };
	  }
	else
	  return 1;
      };
    return gcd;
  }; 
  Stored* createStoredCopy() const;
  static void destroyStoredCopy(Stored* polynomial);
#ifndef NO_DEBUG_VIS
  ostream& output(ostream& str) const
    {
      return str << _linearPolynomial;
    };
#endif
 private:
  BK::LinearPolynomial<long,VampireKernelConst::MaxNumOfVariables * VampireKernelConst::MaxNumOfIndices> _linearPolynomial;
  friend class SpecialisedComparison;
  static Sharing* _sharing;
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
inline
ostream& operator<<(ostream& str,
		    const VK::WeightPolynomial& wp)
{
  return wp.output(str);
};
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
namespace VK
{
class WeightPolynomial::SpecialisedComparison
{
 public:
  class Stored;
  class Command;
  class Extras;
 public:
  SpecialisedComparison() {};
  ~SpecialisedComparison() {};
  void compare(const WeightPolynomial& weight1,const WeightPolynomial& weight2);
  const BK::FunctionComparison& category() const { return _category; };  
  const WeightPolynomial& positivePart() const { return _positivePart; };
  const WeightPolynomial& negativePart() const { return _negativePart; };
  ulong numberOfVariables() const 
  {
    return _positivePart.numberOfVariables() + _negativePart.numberOfVariables();
  };
  const BK::FunctionComparison& assumedRelation() const { return _assumedRelation; };
  void assumeRelation(const BK::FunctionComparison& rel) 
  {
    _assumedRelation = rel;
  };
  const SpecialisedComparison* next() const { return _next; };
  void setNext(SpecialisedComparison* sc) { _next = sc; };	       
 private:
  void factoriseCoefficients();
 private:
  BK::FunctionComparison _category;
  WeightPolynomial _positivePart;
  WeightPolynomial _negativePart;
  BK::FunctionComparison _assumedRelation;
  SpecialisedComparison* _next;
  friend class Extras;
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
inline
ostream& operator<<(ostream& str,
		    const VK::WeightPolynomial::SpecialisedComparison& sc)
{
  str << "{" << sc.positivePart() << "} vs {"
      << sc.negativePart() << "} ~~ " << sc.category();
  return str;
};
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
#  define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison::Extras"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class WeightPolynomial::SpecialisedComparison::Extras
  {
  public:
    typedef BK::ObjectPool<SpecialisedComparison> Pool;
    typedef BK::ObjectPool<SpecialisedComparison>::EntryHandle Handle;
  public:
    static
    bool 
    compileGreaterOrEqual(const SpecialisedComparison& comp,  
			  Command*& code,
			  ulong maxCodeSize)
    {
      return compileGreaterOrEqual(comp.category(),comp.positivePart(),
				   comp.negativePart(),code,maxCodeSize);
    };
    static
    bool 
    simplifyByEqualities(const SpecialisedComparison& comp,
			 const SpecialisedComparison& equalities,
			 Handle& simplified,
			 Handle& spareComparisonObject1,
			 Handle& spareComparisonObject2);
    static
    bool 
    isAlwaysGreaterUnderAssumptions(const SpecialisedComparison& comp,
				    const SpecialisedComparison& assumedRelations,
				    Handle& spareComparisonObject1,
				    Handle& spareComparisonObject2,
				    Handle& spareComparisonObject3);
    static
    bool 
    isAlwaysLessUnderAssumptions(const SpecialisedComparison& comp,
				 const SpecialisedComparison& assumedRelations,
				 Handle& spareComparisonObject1,
				 Handle& spareComparisonObject2,
				 Handle& spareComparisonObject3);
  private:
    static
    bool 
    simplifyByEquality(const SpecialisedComparison& comp,
		       const SpecialisedComparison& equality, 
		       Handle& simplified,
		       Handle& spareComparisonObject);
    static
    bool 
    relaxAlwaysGreaterByInequality(const SpecialisedComparison& comp,
				   const SpecialisedComparison& inequality,
				   Handle& simplified,
				   Handle& spareComparisonObject);
    static
    bool 
    relaxAlwaysLessByInequality(const SpecialisedComparison& comp,
				const SpecialisedComparison& inequality,
				Handle& simplified,
				Handle& spareComparisonObject);
    static
    void 
    applyRelation(const SpecialisedComparison& comp,
		  const SpecialisedComparison& relation,
		  const long& coeffForRelation,
		  const long& coeffForComp,
		  SpecialisedComparison& result);
    static 
    bool 
    compileGreaterOrEqual(const BK::FunctionComparison& cat,
			  const WeightPolynomial& posPart,
			  const WeightPolynomial& negPart,
			  Command*& code,ulong maxCodeSize);
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "WeightPolynomial::SpecialisedComparison::Stored"
#endif
#include "debugMacros.hpp"
namespace VK
{
class WeightPolynomial::SpecialisedComparison::Stored
{
 public:
  Stored() :
    _positivePart(0),
    _negativePart(0)
    { 
      CALL("constructor Stored()");
    };
  Stored(const WeightPolynomial::SpecialisedComparison& raw)
    { 
      CALL("constructor Stored(const WeightPolynomial::SpecialisedComparison& raw)");
      _category = raw.category();
      if (BK::strict(_category))
	{	  
          _positivePart = 0;
          _negativePart = 0;
	}
      else
	{
	  _positivePart = raw.positivePart().createStoredCopy();
	  ASSERT(_positivePart);
	  _negativePart = raw.negativePart().createStoredCopy();
	  ASSERT(_negativePart);
	};
    };
  Stored(const Stored& sc) 
    { 
      CALL("constructor Stored(const Stored& sc)");
      _category = sc.category();
      _positivePart = sc.positivePart();
      if (_positivePart) _positivePart->incReferenceCounter();
      _negativePart = sc.negativePart();
      if (_negativePart) _negativePart->incReferenceCounter();
    };
  ~Stored()
    {
      CALL("destructor ~Stored()");
      if (BK::DestructionMode::isThorough())
	{
	  WeightPolynomial::destroyStoredCopy(_positivePart);
	  WeightPolynomial::destroyStoredCopy(_negativePart);
	};
    };
  Stored& operator=(const Stored& sc)
  {
    CALL("operator=(const Stored& sc)");
    WeightPolynomial::destroyStoredCopy(_positivePart);
    WeightPolynomial::destroyStoredCopy(_negativePart);
    _category = sc.category();
    _positivePart = sc.positivePart();
    if (_positivePart) _positivePart->incReferenceCounter();
    _negativePart = sc.negativePart();
    if (_negativePart) _negativePart->incReferenceCounter();
    return *this;
  };
  const BK::FunctionComparison& category() const 
    { 
      return _category; 
    }; 
  WeightPolynomial::Stored* positivePart() const { return _positivePart; };
  WeightPolynomial::Stored* negativePart() const { return _negativePart; };
 private:
  BK::FunctionComparison _category;
  WeightPolynomial::Stored* _positivePart;
  WeightPolynomial::Stored* _negativePart;
}; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
inline
ostream& operator<<(ostream& str,
		    const VK::WeightPolynomial::SpecialisedComparison::Stored& comp)
{
  return
    str << "{" << comp.positivePart() << "} vs {" 
	<< comp.negativePart() << "} ~~ " << comp.category();
};
inline
ostream& operator<<(ostream& str,
		    const VK::WeightPolynomial::SpecialisedComparison::Stored* comp)
{	
  return str << *comp;
};
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_WEIGHT_POLYNOMIAL
 #define DEBUG_NAMESPACE "WeightPolynomial"
#endif
#include "debugMacros.hpp"
namespace VK
{
inline
bool WeightPolynomial::equal(const WeightPolynomial& p) const
{
  CALL("equal(const WeightPolynomial& p) const");
  if (minimalValue() != p.minimalValue()) return false;
  return monomials().isEqualTo(p.monomials());
}; 
inline
BK::FunctionComparison 
WeightPolynomial::compare(const WeightPolynomial& p) const
{
  CALL("compare(const WeightPolynomial& p) const");
  long minVal = minimalValue();
  long pMinVal = p.minimalValue();
  if (minVal < pMinVal)
    {
      if (p.monomials().isGreaterOrEqual(monomials()))
	{
	  return BK::AlwaysLess;
	}
      else 
	return BK::AbsolutelyVolatile;
    };
  if (minVal > pMinVal)
    {
      if (monomials().isGreaterOrEqual(p.monomials()))
	{
	  return BK::AlwaysGreater;
	}
      else 
	return BK::AbsolutelyVolatile;
    };
  switch (monomials().compare(p.monomials()))
    {
    case BK::Less: return BK::CanBeLessOrEqual;
    case BK::Equal: return BK::AlwaysEqual;
    case BK::Greater: return BK::CanBeGreaterOrEqual; 
    case BK::Incomparable: return BK::AbsolutelyVolatile;
    };
  ICP("ICP0");
  return BK::FailureToCompare;
}; 
inline
BK::FunctionComparison 
WeightPolynomial::compareForGreaterOrEqual(const WeightPolynomial& p) const
{
  CALL("compareForGreaterOrEqual(const WeightPolynomial& p) const");
  long minVal = minimalValue();
  long pMinVal = p.minimalValue();
  if (minVal < pMinVal)
    {
      return BK::FailureToCompare; 
    };
  if (minVal > pMinVal)
    {
      if (monomials().isGreaterOrEqual(p.monomials()))
	{
	  return BK::AlwaysGreater;
	}
      else 
	return BK::FailureToCompare;
    };
  switch (monomials().compare(p.monomials()))
    {
    case BK::Less: return BK::FailureToCompare;
    case BK::Equal: return BK::AlwaysEqual;
    case BK::Greater: return BK::CanBeGreaterOrEqual; 
    case BK::Incomparable: return BK::FailureToCompare;
    };
  ICP("ICP0");
  return BK::FailureToCompare;
}; 
inline
BK::FunctionComparison 
WeightPolynomial::compareForLessOrEqual(const WeightPolynomial& p) const
{
  CALL("compareForLessOrEqual(const WeightPolynomial& p) const");
  long minVal = minimalValue();
  long pMinVal = p.minimalValue();
  if (minVal < pMinVal)
    {
      if (p.monomials().isGreaterOrEqual(monomials()))
	{
	  return BK::AlwaysLess;
	}
      else 
	return BK::FailureToCompare;
    };
  if (minVal > pMinVal)
    {
      return BK::FailureToCompare; 
    };
  switch (monomials().compare(p.monomials()))
    {
    case BK::Less: return BK::CanBeLessOrEqual;
    case BK::Equal: return BK::AlwaysEqual;
    case BK::Greater: return BK::FailureToCompare;
    case BK::Incomparable: return BK::FailureToCompare;
    };
  ICP("ICP0");
  return BK::FailureToCompare;
}; 
inline
bool WeightPolynomial::isAlwaysGreater(const WeightPolynomial& p) const
{
  CALL("isAlwaysGreater(const WeightPolynomial& p)");
  if (minimalValue() <= p.minimalValue()) return false;
  return monomials().isGreaterOrEqual(p.monomials());
}; 
inline
bool WeightPolynomial::isNeverLess(const WeightPolynomial& p) const
{
  CALL("isNeverLess(const WeightPolynomial& p) const");
  if (minimalValue() < p.minimalValue()) return false;
  return monomials().isGreaterOrEqual(p.monomials());
}; 
inline
WeightPolynomial::Stored* WeightPolynomial::createStoredCopy() const 
{
  CALL("createStoredCopy() const");
  if (isZero()) return 0;
  _sharing->resetIntegration();
  if (freeMember()) _sharing->integrConst(freeMember());
  const ulong* endOfVars = monomials().end();
  for (const ulong* v = monomials().begin(); v != endOfVars; ++v)
    { 
      _sharing->integrVar(coefficient(*v),*v);
    };
  Stored* res = _sharing->integrate();
  ASSERT(res->referenceCounter() >= 1L);
  res->incReferenceCounter();
  return res;
}; 
inline
void WeightPolynomial::destroyStoredCopy(Stored* polynomial)
{
  CALL("destroyStoredCopy(Stored* polynomial)");
  if (polynomial) 
    {
      ASSERT(polynomial->referenceCounter() >= 1L);
      polynomial->decReferenceCounter();
      if (polynomial->referenceCounter() == 1L) 
	_sharing->remove(polynomial);
    };
};
}; 
#endif
