/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <fstream>
#include "Assertion.hpp"
#include "Tracer.hpp"
#include "XML.hpp"
#include "VampireException.hpp"
#include "Options.hpp"
#include "VampireException.hpp"
using namespace Shell;
XML::Attribute::Attribute (Type type, const string& name)
  : _type (type),
    _name (name),
    _next (0),
    _previous (0)
{
  TRACER("XML::Attribute::Attribute");
}
XML::IntegerAttribute::IntegerAttribute (const string& name, int value)
  : Attribute (INTEGER, name),
    _value (value)
{
  TRACER("XML::IntegerAttribute::IntegerAttribute");
}
XML::LongAttribute::LongAttribute (const string& name, long value)
  : Attribute (LONG, name),
    _value (value)
{
  TRACER("XML::LongAttribute::LongAttribute");
}
XML::DoubleAttribute::DoubleAttribute (const string& name, double value)
  : Attribute (DOUBLE, name),
    _value (value)
{
  TRACER("XML::DoubleAttribute::DoubleAttribute");
}
XML::TextAttribute::TextAttribute (const string& name, const string& text)
  : Attribute (TEXT, name),
    _value (text)
{
  TRACER("XML::TextAttribute::TextAttribute");
}
void XML::DeepElement::destroyContent ()
{
  TRACER("XML::DeepElement::destroyContent");
  Content* current = _firstChild;
  while (current) {
    Content* next = current->_next;
    current->destroy();
    current = next;
  }
} 
void XML::Element::write (ostream& str) const
{
  TRACER("XML::Element::write");
  write(str,0);
} 
XML::Element::Element (Type tp, const string& name)
  :
  Content (tp),
  _name (name),
  _firstAttribute (0),
  _lastAttribute (0)
{
  TRACER("XML::Element::Element");
}
XML::Content::Content (Type ct)
  : _type (ct),
    _next (0),
    _previous (0)
{
  TRACER("XML::Content::Content");
} 
void XML::Element::addAttribute(Attribute* att)
{
  TRACER("XML::Element::addAttribute");
  if (_lastAttribute) {
    _lastAttribute->setNext(att);
    att->setPrevious(_lastAttribute);
    _lastAttribute = att;
    return;
  }
  _firstAttribute = att;
  _lastAttribute = att;
} 
void XML::DeepElement::addChild (Content* sub)
{
  TRACER("XML::DeepElement::addChild");
  if (_lastChild) {
    _lastChild->_next = sub;
    sub->_previous = _lastChild;
    _lastChild = sub;
    return;
  }
  _firstChild = sub;
  _lastChild = sub;
} 
XML::TextElement::TextElement (const string& name, const string& text)
  : Element(TEXT,name),
    _text(text)
{
  TRACER("XML::TextElement::TextElement");
}
XML::DeepElement::DeepElement (const string& name)
  : Element(DEEP, name),
    _firstChild(0),
    _lastChild(0)
{
  TRACER("XML::DeepElement::DeepElement");
}
XML::FloatElement::FloatElement (const string& name, float value)
  : Element(FLOAT, name),
    _value(value)
{
  TRACER("XML::FloatElement::FloatElement");
}
XML::LongElement::LongElement (const string& name, long value)
  : Element(LONG, name),
    _value(value)
{
  TRACER("XML::LongElement::LongElement");
}
XML::IntegerElement::IntegerElement (const string& name, int value)
  : Element(INTEGER, name),
    _value(value)
{
  TRACER("XML::IntegerElement::IntegerElement");
}
XML::EmptyElement::EmptyElement (const string& name)
  : Element(EMPTY, name)
{
  TRACER("XML::EmptyElement::EmptyElement");
}
void XML::Attribute::writeList (ostream& str) const
{
  TRACER("XML::Attribute::writeList (ostream& str)");
  for (const Attribute* attr = this; attr; attr = attr->_next) {
    str << ' ';
    XML::Element::writeString(str,attr->_name);
    str << "=\"";
    attr->writeValue(str);
    str << '\"';
  }
} 
void XML::TextAttribute::writeValue (ostream& str) const
{
  XML::Element::writeString(str,_value);
} 
void XML::IntegerAttribute::writeValue (ostream& str) const
{
  str << _value;
} 
void XML::LongAttribute::writeValue (ostream& str) const
{
  str << _value;
} 
void XML::DoubleAttribute::writeValue (ostream& str) const
{
  str << _value;
} 
void XML::DeepElement::writeContent(ostream& str, int indent) const
{
  TRACER("XML::DeepElement::writeContent(ostream& str, int indent)");
  if (_firstChild) {
    str << '\n';
    for (Content* sub = _firstChild; sub; sub = sub->_next) {
      sub->write(str,indent+1);
    }
    writeSpaces(str,indent);
  }
} 
void XML::TextElement::writeContent(ostream& str, int indent) const
{
  writeString(str,_text);
} 
void XML::FloatElement::writeContent(ostream& str, int indent) const
{
  str << _value;
} 
void XML::LongElement::writeContent(ostream& str, int indent) const
{
  str << _value;
} 
void XML::IntegerElement::writeContent(ostream& str, int indent) const
{
  str << _value;
} 
void XML::EmptyElement::writeContent(ostream& str, int indent) const
{
} 
void XML::Element::writeFooter(ostream& str) const
{
  str << "</" << _name << '>';
} 
void XML::Element::writeSpaces(ostream& str,int n)
{
  ASS(n >= 0);
  while (n > 0) {
    n--;
    str << ' ';
  }
} 
void XML::Element::write (ostream& str, int indent) const
{
  TRACER("XML::Element::write(ostream& str, int indent)");
  writeSpaces(str,indent);
  str << '<' << _name;
  _firstAttribute->writeList(str);
  writeEndOfHeader(str);
  str << '>';
  writeContent(str,indent);
  writeFooter(str);
  str << '\n';
} 
void XML::EmptyElement::writeEndOfHeader (ostream& str) const
{
  str << " /";
}
void XML::Element::append (const string& file)
{
  TRACER("XML::Element::append");
  ofstream stream(file.c_str(),ios::app);
  CONTROL("X1");
  if (! stream) { 
    USER_ERROR("cannot open file XML file for appending");
  }
  CONTROL("X2");
  write(stream);
  CONTROL("X3");
} 
void XML::Content::destroy ()
{
  TRACER("XML::Content::destroy");
  destroyAttributes();
  destroyContent();
  switch (_type) {
  case PCDATA:
    delete static_cast<PCData*>(this);
    return;
  case EMPTY:
    delete static_cast<EmptyElement*>(this);
    return;
  case TEXT:
    delete static_cast<TextElement*>(this);
    return;
  case FLOAT:
    delete static_cast<FloatElement*>(this);
    return;
  case LONG:
    delete static_cast<LongElement*>(this);
    return;
  case INTEGER:
    delete static_cast<IntegerElement*>(this);
    return;
  case DEEP:
    delete static_cast<DeepElement*>(this);
    return;
  }
} 
void XML::Element::destroyAttributes ()
{
  TRACER("XML::Element::destroyAttributes");
  _firstAttribute->destroy();
}
void XML::Attribute::destroy ()
{
  TRACER("XML::Attribute::destroy");
  Attribute* current = this;
  while (current) {
    Attribute* next = current->_next;
    switch (current->_type) {
    case TEXT:
      delete static_cast<TextAttribute*>(current);
      break;
    case LONG:
      delete static_cast<LongAttribute*>(current);
      break;
    case DOUBLE:
      delete static_cast<DoubleAttribute*>(current);
      break;
    case INTEGER:
      delete static_cast<IntegerAttribute*>(current);
      break;
    }
    current = next;
  }
} 
void XML::Attribute::operator delete (void* obj)
{
  ASS(false);
}
void XML::Content::operator delete (void* obj)
{
  ASS(false);
}
void XML::Element::operator delete (void* obj)
{
  ASS(false);
}
void XML::Element::writeString (ostream& str,const string& s)
{
  size_t length = s.length();
  for (size_t i=0;i < length;i++) {
    char c = s[i];
    switch (c) {
    case '>':
      str << "&gt;";
      break;
    case '<':
      str << "&lt;";
      break;
    case '&':
      str << "&amp;";
      break;
    case '\'':
      str << "&apos;";
      break;
    case '\"':
      str << "&quot;";
      break;
    default:
      str << c;
      break;
    }
  }
} 
void XML::Element::setAttributes (Attribute* attr)
{
  TRACER("XML::Element::setAttributes");
  _firstAttribute = attr;
  if (! attr) {
    _lastAttribute = 0;
    return;
  }
  while (attr->next()) {
    attr = attr->next();
  }
  _lastAttribute = attr;
} 
const XML::Attribute* XML::Attribute::find (const string& name) const
{
  TRACER("XML::Attribute::find");
  for (const XML::Attribute* attr = this;
       attr;
       attr = attr->_next) {
    if (attr->name() == name) {
      return attr;
    }
  }
  return 0;
} 
