/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <fstream>
#include <sstream>
#include "Bag.hpp"
#include "Environment.hpp"
#include "Int.hpp"
#include "Options.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
namespace Shell {
class Bag::Exception
  : public VampireException
{
public:                                
  Exception(string errorMessage,const XML::Content* element);
  void cry (ostream&);
  XML::Element* toXML () const;
private:
  const XML::Content* _content;
}; 
void Bag::interpret (const string& inputFileName)
{
  TRACER("Bag::interpret(const string& inputFileName)");
  std::ifstream inFile(inputFileName.c_str());
  if(! inFile) {
    USER_ERROR(string("cannot open input file ") + inputFileName);
  };
  interpret(inFile);
} 
void Bag::interpret (istream& str)
{
} 
void Bag::interpret (const XML::Element* e)
{
  TRACER("Bag::interpret(XML::Element*)");
  string name = e->name();
  if (name == "assert") {
    ass(e);
    return;
  }
  if (name == "bag") {
    const XML::DeepElement* d = toDeep(e);
    for (const XML::Element* child = toElement(d->firstChild());
	 child;
	 child = toElement(child->next())) {
      interpret(child);
    }
    return;
  }
  if (name == "bind") {
    bind(e);
    return;
  }
  if (name == "bye") {
    bye(e);
    return;
  }
  if (name == "comment") { 
    return;
  }
  if (name == "kb_drop") {
    dropKB(e);
    return;
  }
  if (name == "kb_load") {
    loadKB(e);
    return;
  }
  if (name == "kb_status") {
    kbStatus(e);
    return;
  }
  if (name == "options") {
    readOptions(e);
    return;
  }
  if (name == "query") {
    query(e);
    return;
  }
  if (name == "saturate") {
    saturate(e);
    return;
  }
  if (name == "set") {
    setParameter(e);
    return;
  }
  throw Exception("unknown command",e);
} 
const XML::DeepElement* Bag::toDeep(const XML::Content* e)
{
  TRACER("Bag::toDeep");
  if (e->kind() != XML::Content::DEEP) {
    throw Exception("wrong element format",e);
  }
  return static_cast<const XML::DeepElement*>(e);
} 
const XML::EmptyElement* Bag::toEmpty(const XML::Content* e)
{
  TRACER("Bag::toEmpty");
  if (e->kind() != XML::Content::EMPTY) {
    throw Exception("wrong element format",e);
  }
  return static_cast<const XML::EmptyElement*>(e);
} 
const XML::TextElement* Bag::toText(const XML::Content* e)
{
  TRACER("Bag::toEmpty");
  if (e->kind() != XML::Content::TEXT) {
    throw Exception("wrong element format",e);
  }
  return static_cast<const XML::TextElement*>(e);
} 
const XML::Element* Bag::toElement(const XML::Content* e)
{
  TRACER("Bag::toElement");
  if (e->kind() == XML::Content::PCDATA) {
    throw Exception("PCDATA instead of an XML element",e);
  }
  return static_cast<const XML::Element*>(e);
} 
Bag::Exception::Exception (string errorMessage,const XML::Content* content)
  : VampireException(errorMessage,0),
    _content(content)
{
} 
void Bag::Exception::cry (ostream& str)
{
  TRACER("Bag::Exception::cry");
  if (! _content) { 
    str << "Bag File Error: " <<_file << "\n";
    return;
  }
  str << "Bag File Error: "
      << _file;
  if (_content->kind() == XML::Content::PCDATA) {
    str << " in '";
    _content->write(str);
    str << "'\n";
    return;
  }
  str << " in the following element:\n";
  _content->write(str);
} 
XML::Element* Bag::Exception::toXML () const
{
  TRACER("Bag::Exception::toXML");
  XML::DeepElement* result = new XML::DeepElement("bag_exception");
  result->addChild(new XML::TextElement("error",_file));
  if (_content) { 
    result->addChild(const_cast<XML::Content*>(_content));
  }
  return result;
} 
void Bag::readOptions (const XML::Element* o)
{
  TRACER("Bag::readOptions");
  const XML::DeepElement* opt = toDeep(o);
  const string* name = find(opt,"name");
  const string* origin = find(opt,"origin");
  Options* options = new Options(*Environment::options(origin));
  for (const XML::Content* e = opt->firstChild();
       e;
       e = e->next()) {
    const XML::EmptyElement* op = toEmpty(e);
    if (op->name() != "option") {
      throw Exception("option element expected",op);
    }
    const string* nm = find(op,"name");
    const string* val = find(op,"value");
    if (! (nm && val)) {
      throw Exception("either name or value of option undefined",op);
    }
    options->set(*nm,*val);
  }
  Environment::setOptions(name,options);
} 
void Bag::setParameter (const XML::Element* e)
{
  TRACER("Bag::setParameter");
  toEmpty(e);
  const string* parameter = find(e,"parameter");
  const string* value = find(e,"value");
  if (! parameter) {
    throw Exception("parameter omitted in command 'set'",e);
  }
  if (! value) {
    throw Exception("value omitted in command 'set'",e);
  }
  if (! Environment::set(*parameter,*value)) {
    throw Exception("unknown parameter",e);
  }
} 
const string* Bag::find (const XML::Element* element,const string& name)
{
  const XML::Attribute* attr = element->firstAttribute()->find(name);
  if (! attr) {
    return 0;
  }
  return &static_cast<const XML::TextAttribute*>(attr)->value();
} 
void Bag::loadKB(const XML::Element* e)
{
  TRACER("Bag::loadKB");
  toEmpty(e);
  const string* kb = find(e,"kb");
  if (! kb) {
    throw Exception("attribute 'kb' omitted in command 'kb_load'",e);
  }
  const string* file = find(e,"file");
  if (! file) {
    throw Exception("attribute 'file' omitted in command 'kb_load'",e);
  }
  const string* active = find(e,"enabled");
  bool isActive = true;
  if (active) {
    if (! yesNoToBoolean(*active,isActive)) {
      throw Exception("unknown value for attribute 'active'",e);
    }
  }
  const string* syntax = find(e,"syntax");
  int loaded = Environment::loadKB(*kb,*file,active ? &isActive : 0,syntax);
} 
void Bag::dropKB (const XML::Element* e)
{
  TRACER("Bag::dropKB");
  toEmpty(e);
  const string* kb = find(e,"kb");
  if (! kb) {
    throw Exception("attribute 'kb' omitted in command 'kb_drop'",e);
  }
  int dropped = Environment::dropKB(*kb);
} 
bool Bag::yesNoToBoolean (const string& yesNo,bool& result)
{
  TRACER("Bag::yesNoToBoolean");
  if (yesNo == "yes") {
    result = true;
    return true;
  }
  if (yesNo == "no") {
    result = false;
    return true;
  }
  return false;
} 
void Bag::kbStatus (const XML::Element* e)
{
  TRACER("Bag::kbStatus");
  toEmpty(e);
  const string* kb = find(e,"kb");
  if (! kb) {
    throw Exception("attribute 'kb' omitted in command 'kb_status'",e);
  }
  const string* active = find(e,"enabled");
  if (! active) {
    throw Exception("attribute 'enabled' omitted in command 'kb_status'",e);
  }
  bool isActive;
  if (! yesNoToBoolean(*active,isActive)) {
    throw Exception("unknown value for attribute 'enabled'",e);
  }
  Environment::setKBStatus(*kb,isActive);
} 
void Bag::saturate (const XML::Element* e)
{
} 
void Bag::bind(const XML::Element* e)
{
  TRACER("Bag::bind");
  toEmpty(e);
  const string* name = find(e,"name");
  if (! name) {
    throw Exception("attribute 'name' omitted in command 'bind'",e);
  }
  const string* function = find(e,"function");
  const string* relation = find(e,"relation");
  if (function) {
    if (relation) {
      throw Exception("both function and relation in command 'bind'",e);
    }
    Environment::bindFunction(*name,*function);
  }
  else if (relation) {
    Environment::bindRelation(*name,*relation);
  }
  else {
    throw Exception("no function or relation in command 'bind'",e);
  }
} 
void Bag::bye (const XML::Element* e)
{
} 
void Bag::ass (const XML::Element* e)
{
} 
void Bag::query (const XML::Element* e)
{
} 
}
