/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef ARRAY_HPP
#define ARRAY_HPP
#include <cstdlib>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#ifdef DEBUG_ALLOC
#  include "ObjDesc.hpp"
#endif
#include "DestructionMode.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ARRAY
#  define DEBUG_NAMESPACE "Array<class Alloc,class ElType,class ElRelocation,ulong MinIncrementSize>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
  template <class Alloc,class ElType,class ElRelocation,ulong MinIncrementSize>
  class Array
  {
  public:
    Array() : 
      _memory(0), 
      _currSize(0UL) 
#ifdef DEBUG_NAMESPACE
      , _name("?")
#endif
#ifndef NO_DEBUG
      , _debugFrozen(false)
#endif
    {  
    };
    Array(const char* nm)  : 
      _memory(0), 
      _currSize(0UL) 
#ifdef DEBUG_NAMESPACE
      , _name(nm)
#endif
#ifndef NO_DEBUG
      , _debugFrozen(false)
#endif
    { 
    };
    Array(const ElType& initElemVal,int ) :
      _memory(0), 
      _currSize(0UL),
      _initialElementValue(initElemVal)
#ifdef DEBUG_NAMESPACE
      , _name("?")
#endif
#ifndef NO_DEBUG
      , _debugFrozen(false)
#endif
    {
    };
    Array(const ElType& initElemVal,const char* nm)  :
      _memory(0), 
      _currSize(0UL), 
      _initialElementValue(initElemVal)
#ifdef DEBUG_NAMESPACE
      , _name(nm)
#endif
#ifndef NO_DEBUG
      , _debugFrozen(false)
#endif
    {
    };
    Array(const Array& a) :
      _currSize(a._currSize),
      _initialElementValue(a._initialElementValue)
#ifdef DEBUG_NAMESPACE
      , _name(a._name)
#endif
#ifndef NO_DEBUG
      , _debugFrozen(false)
#endif
    {
      if (_currSize)
	{
	  _memory = 
	    static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*_currSize));
	  for (ulong j = 0; j < _currSize; j++)
	    {
	      ::new (static_cast<void*>(_memory + j)) ElType(a._memory[j]);
	    };
	}
      else 
	_memory = 0;
    };
    Array& operator=(const Array& a)
    {
      CALLM(_name,"operator=(const Array& a)");
      if (_currSize)
	{
	  for (ulong j = 0; j < _currSize; j++)
	    _memory[j].~ElType();
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = a._currSize;   
      _initialElementValue = a._initialElementValue; 
#ifdef DEBUG_NAMESPACE
      _name = a._name;
#endif
      if (_currSize)
	{
	  _memory = 
	    static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*_currSize));
	  for (ulong j = 0; j < _currSize; j++)
	    _memory[j] = a._memory[j]; 
	}
      else 
	_memory = 0;
#ifndef NO_DEBUG
      _debugFrozen = a._debugFrozen;
#endif
      return *this;
    };
    ~Array() 
    { 
      CALLM(_name,"destructor ~Array()");
      if (DestructionMode::isThorough())
	if (_currSize)
	  {
	    for (ulong j = 0; j < _currSize; j++)
	      _memory[j].~ElType();
	    Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	  };
    };
    void relocateTo(void* target)
    {
      CALLM(_name,"relocateTo(void* target)");
      static_cast<Array*>(target)->_memory = _memory;
      static_cast<Array*>(target)->_currSize = _currSize;
      ElRelocation::move(_initialElementValue,
			 static_cast<void*>(&(static_cast<Array*>(target)->_initialElementValue)));
#ifdef DEBUG_NAMESPACE
      static_cast<Array*>(target)->_name = _name; 
#endif
#ifndef NO_DEBUG
      static_cast<Array*>(target)->_debugFrozen = _debugFrozen;
#endif
      BK_CORRUPT(*this);
    }; 
    void* operator new(size_t) 
    {
      CALL("operator new (size_t)");
      return Alloc::allocate(sizeof(Array));
    };
    void operator delete(void* obj) 
    {
      CALL("operator delete(void* obj)");
      Alloc::deallocate(obj,sizeof(Array)); 
    };
    const ElType* memory() const { return _memory; }; 
    ElType* memory() { return _memory; };
    ElType* endOfMemory() { return _memory + _currSize; };
    ElType& operator[](ulong i) 
    {
      CALLM(_name,"operator[](ulong i)");  
      ASSERT(i < _currSize);
      COP("COP10");
      return _memory[i];
    };
    ElType& operator[](ulong i) const 
    {
      CALLM(_name,"operator[](ulong i) const");  
      ASSERT(i < _currSize);
      COP("COP10");
      return _memory[i];
    }; 
    ElType& sub(ulong i) 
    {
      CALLM(_name,"ElType& sub(ulong i)");
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"sub(..) called on frozen array.");
#endif   
      if (i >= _currSize) expand(i);
      ASSERT(i < _currSize);
      return _memory[i];
    };
    void expand(ulong i)
    {
      CALLM(_name,"expand(ulong i)");
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array expanded.");
#endif   
      if (i < _currSize) return;
      ulong newSize = (i/MinIncrementSize)*MinIncrementSize;
      if (i >= newSize) newSize += MinIncrementSize; 
      ASSERT(i < newSize);
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*newSize));
      ASSERT(newMemory);
      if (_currSize)
	{
	  for (ulong j = 0; j < _currSize; j++) 
	    {
	      ElRelocation::move(_memory[j],static_cast<void*>(newMemory + j));
	    };
	};
      for (ulong k = _currSize; k < newSize; k++) 
	{
	  ::new (static_cast<void*>(newMemory + k)) ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
    }; 
    bool tryToExpand(ulong i)
    {
      CALLM(_name,"tryToExpand(ulong i)");
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array expanded.");
#endif   
      if (i < _currSize) return true;
      ulong newSize = (i/MinIncrementSize)*MinIncrementSize;
      if (i >= newSize) newSize += MinIncrementSize; 
      ASSERT(i < newSize);
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::tryToAllocate(sizeof(ElType)*newSize));
      if (!newMemory) return false;
      if (_currSize)
	{
	  for (ulong j = 0; j < _currSize; j++) 
	    ElRelocation::move(_memory[j],static_cast<void*>(newMemory + j));
	};
      for (ulong k = _currSize; k < newSize; k++) 
	{
	  ::new (static_cast<void*>(newMemory + k)) ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
      return true;
    }; 
    void expand()
    {
      CALLM(_name,"expand()");
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array expanded.");
#endif   
      ulong newSize = _currSize + MinIncrementSize;
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*newSize));
      if (_currSize)
	{
	  for (ulong j = 0; j < _currSize; j++) 
	    ElRelocation::move(_memory[j],static_cast<void*>(newMemory + j));
	};
      for (ulong k = _currSize; k < newSize; k++) 
	{
	  ::new (static_cast<void*>(newMemory + k)) ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
    }; 
    bool tryToExpand()
    {
      CALLM(_name,"tryToExpand()");
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array expanded.");
#endif   
      ulong newSize = _currSize + MinIncrementSize;
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::tryToAllocate(sizeof(ElType)*newSize));
      if (!newMemory) return false;
      ASSERT(newMemory);
      if (_currSize)
	{
	  for (ulong j = 0; j < _currSize; j++) 
	    ElRelocation::move(_memory[j],static_cast<void*>(newMemory + j));
	};
      for (ulong k = _currSize; k < newSize; k++) 
	{
	  ::new (static_cast<void*>(newMemory + k)) ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
      return true;
    }; 
    void stretch(ulong cut,
		 ulong requiredExtraCapacity,
		 ulong& createdExtraCapacity)
    {
      CALL("stretch(long cut,ulong requiredExtraCapacity,ulong& createdExtraCapacity)");
      ASSERT(cut < _currSize);
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array stretched.");
#endif   
      if (requiredExtraCapacity == 0UL) 
	{
	  createdExtraCapacity = 0UL;
	  return;
	};    
      createdExtraCapacity = 
	((requiredExtraCapacity/MinIncrementSize)*MinIncrementSize);
      if (requiredExtraCapacity % MinIncrementSize)
	createdExtraCapacity += MinIncrementSize;
      ulong newSize = _currSize + createdExtraCapacity;
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*newSize));
      ASSERT(newMemory);
      if (_currSize)
	{       
	  for (ulong j = 0UL; j < cut; j++) 
	    ElRelocation::move(_memory[j],static_cast<void*>(newMemory + j));
	  for (ulong j = cut; j < _currSize; j++) 
	    ElRelocation::move(_memory[j],
			       static_cast<void*>(newMemory + (j + createdExtraCapacity)));
	};
      for (ulong k = 0UL; k < createdExtraCapacity; k++) 
	{
	  ::new (static_cast<void*>(newMemory + cut + k)) 
	      ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
    }; 
    void stretch(ulong cut,ulong& createdExtraCapacity)
    {
      CALL("stretch(ulong cut,ulong& createdExtraCapacity)");
      ASSERT(cut < _currSize);
#ifndef NO_DEBUG
      if (_debugFrozen) Debug::error(DEBUG_OUT_STR,"Frozen array stretched.");
#endif   
      createdExtraCapacity = MinIncrementSize; 
      ulong newSize = _currSize + MinIncrementSize;    
      ElType* newMemory = 
	static_cast<ElType*>(Alloc::allocate(sizeof(ElType)*newSize));
      ASSERT(newMemory);
      if (_currSize)
	{       
	  for (ulong i = 0UL; i < cut; i++) 
	    ElRelocation::move(_memory[i],static_cast<void*>(newMemory + i));
	  for (ulong j = cut; j < _currSize; j++) 
	    ElRelocation::move(_memory[j],static_cast<void*>(newMemory + (j + MinIncrementSize)));
	};
      for (ulong k = 0UL; k < MinIncrementSize; k++) 
	{
	  ::new (static_cast<void*>(newMemory + cut + k)) 
	      ElType(_initialElementValue);
	};
      if (_currSize)
	{
	  Alloc::deallocate(static_cast<void*>(_memory),sizeof(ElType)*_currSize);
	};
      _currSize = newSize;  
      _memory = newMemory;
    }; 
    ulong size() const { return _currSize; };
#ifndef NO_DEBUG
    void freeze() { _debugFrozen = true; };
    void unfreeze() { _debugFrozen = false; };
    bool checkObj()
    {
#ifdef DEBUG_ALLOC 
      if (_currSize)
	{
	  ObjDesc* objDesc = Alloc::hostObj(static_cast<void*>(_memory));
	  if (!objDesc) return false;
#ifdef DEBUG_ALLOC_OBJ_SIZE
	  if (objDesc->size <= 0) return false; 
#endif
#ifdef DEBUG_ALLOC_OBJ_STATUS
	  if (objDesc->status != ObjDesc::Persistent) return false;
#endif
	};
#endif
      return true;
    };
#endif
  private:
    ElType* _memory;
    ulong _currSize; 
    ElType _initialElementValue; 
#ifdef DEBUG_NAMESPACE
    const char* _name;
#endif
#ifndef NO_DEBUG
    bool _debugFrozen;
#endif
  }; 
}; 
#endif
