/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef OBJECT_POOL_HPP
#define OBJECT_POOL_HPP
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL
#  define DEBUG_NAMESPACE "ObjectPool<class C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{
template <class C>
class ObjectPool 
{
protected:
  class Entry;
public:
  class EntryHandle
  {
  public:
    EntryHandle() : _content(0) {};
    EntryHandle(const EntryHandle& h) :
      _content(h._content)
    { 
      CALL("EntryHandle::constructor EntryHandle(const EntryHandle& h)");
      DOP(reportAquisition());
      if (_content)
	_content->incRefCounter();
    };
    EntryHandle(EntryHandle& h,int ) :
      _content(h._content)
    { 
      CALL("EntryHandle::constructor EntryHandle(EntryHandle& h,int)");
      DOP(reportAquisition());
      DOP(h.reportLoss());
      h._content = 0;
      DOP(h.reportAquisition());
    };
    explicit EntryHandle(ObjectPool& pool)
    {
      CALL("EntryHandle::constructor EntryHandle(ObjectPool& pool)");
      _content = pool.reserveEntry();
      ASSERT(!_content->_referenceCounter);
      DOP(reportAquisition());
      ASSERT(_content);
      _content->_pool = &pool;
      _content->incRefCounter();
    };  
    explicit EntryHandle(ObjectPool* pool)
    {
      CALL("EntryHandle::constructor EntryHandle(ObjectPool* pool)");
      _content = pool->reserveEntry(); 
      ASSERT(!_content->_referenceCounter);   
      DOP(reportAquisition());
      ASSERT(_content);
      _content->_pool = pool;
      _content->incRefCounter();
    };  
    ~EntryHandle()
    {
      CALL("EntryHandle::destructor ~EntryHandle()");
      DOP(reportLoss());
      if (_content)
	{
	  _content->decRefCounter();
	  if (!_content->_referenceCounter)
	    {
	      hibernate(_content->content());
	      _content->_pool->recycleEntry(_content);
	    };
	};
    };
    EntryHandle& operator=(const EntryHandle& h)
    {
      CALL("EntryHandle::operator=(const EntryHandle& h)");
      DOP(reportLoss());
      if (_content)
	{
	  _content->decRefCounter();
	  if (!_content->_referenceCounter)
	    {
	      hibernate(_content->content());
	      _content->_pool->recycleEntry(_content);
	    };
	};
      _content = h._content;
      DOP(reportAquisition());
      if (_content)
	_content->incRefCounter();
      return *this;
    };
    bool operator==(const EntryHandle& h) const 
    { 
      return _content == h._content; 
    };
    bool isNull() const { return !_content; };
    void clear()
    {
      CALL("EntryHandle::clear()");      
      DOP(reportLoss());
      if (_content)
	{
	  _content->decRefCounter();
	  if (!_content->_referenceCounter)
	    {
	      hibernate(_content->content());
	      _content->_pool->recycleEntry(_content);
	    };
	};
      _content = 0;      
      DOP(reportAquisition());
    };
    C& operator*() const
    {
      CALL("EntryHandle::operator*() const");
      ASSERT(_content);
      return _content->content();
    };
    C* operator->() const
    {
      CALL("EntryHandle::operator->() const");
      ASSERT(_content);
      return _content->contentAddr();
    };
    C* object() const
    {
      CALL("EntryHandle::object() const");
      ASSERT(_content);
      return _content->contentAddr();
    };
    void swapWith(EntryHandle& h)
    {
      CALL("EntryHandle::swapWith(EntryHandle& h)");
      DOP(reportLoss());
      DOP(h.reportLoss());
      Entry* tmp = _content;
      _content = h._content;
      h._content = tmp;
      DOP(reportAquisition());
      DOP(h.reportAquisition());
    };
    void collectFrom(EntryHandle& h)
    {
      CALL("EntryHandle::collectFrom(EntryHandle& h)");      
      DOP(reportLoss());
      if (_content)
	{
	  _content->decRefCounter();
	  if (!_content->_referenceCounter)
	    {
	      hibernate(_content->content());
	      _content->_pool->recycleEntry(_content);
	    };
	};
      _content = h._content;     
      DOP(reportAquisition());  
      DOP(h.reportLoss());
      h._content = 0;
      DOP(h.reportAquisition());  
    };
    void passTo(EntryHandle& h)
    {
      CALL("EntryHandle::passTo(EntryHandle& h)");   
      DOP(reportLoss());  
      DOP(h.reportLoss());
      if (h._content)
	{
	  h._content->decRefCounter();
	  if (!h._content->_referenceCounter)
	    {
	      hibernate(h._content->content());
	      h._content->_pool->recycleEntry(h._content);
	    };
	};
      h._content = _content;
      _content = 0; 
      DOP(reportAquisition());  
      DOP(h.reportAquisition());  
    };
    void relocateTo(void* target)
    {
      CALL("EntryHandle::relocateTo(void* target)");
      DOP(reportLoss());
      DOP(reportAquisition());
      (static_cast<EntryHandle*>(target))->_content = _content;      
      DOP((static_cast<EntryHandle*>(target))->reportAquisition());
      BK_CORRUPT(*this);
    };
    EntryHandle grab()
    {
      CALL("grab()");     
      DOP(reportLoss());  
      Entry* tmp = _content;
      _content = 0;
      return EntryHandle(tmp);
    };
    static EntryHandle grab(EntryHandle& h)
    {
      CALL("grab(EntryHandle& h)");
      DOP(h.reportLoss());  
      Entry* tmp = h._content;
      h._content = 0;
      return EntryHandle(tmp);
    };
  private:
    EntryHandle(Entry* e) :
      _content(e)
    {
      CALL("EntryHandle::constructor EntryHandle(Entry* e)");
    };
#ifdef DEBUG_NAMESPACE    
    void reportAquisition()
    {
      reportAquisition(_content);
    };
    void reportAquisition(Entry* e)
    {
    };
    void reportLoss()
    {
      reportLoss(_content);
    };
    void reportLoss(Entry* e)
    {
    };
#endif
  private:
    Entry* _content;
    friend class ObjectPool<C>;
  }; 
public:
  ObjectPool() {};
  virtual ~ObjectPool() {};
  void reserveObject(EntryHandle& h);
  bool tryToReserveObject(EntryHandle& h);
  C* lendObject();
  static void returnObject(C* obj);
  virtual void prepareNObjects(ulong n) = 0;
protected:
  class Entry : private C
  {
  public:
    Entry() : _referenceCounter(0L) {};
    ~Entry() {};
    Entry*& nextFree() { return _nextFree; };
  private:
    C& content() { return *(static_cast<C*>(this)); };
    const C& content() const { return *(static_cast<C*>(this)); };
    C* contentAddr() { return static_cast<C*>(this); };
    const C* contentAddr() const { return static_cast<C*>(this);  };
    static Entry* entryFor(C* cont) { return static_cast<Entry*>(cont); };
    void incRefCounter() 
    {
      CALL("Entry::incRefCounter()");
      ASSERT(_referenceCounter >= 0L);
      ++_referenceCounter;
    };
    void decRefCounter() 
    {
      CALL("Entry::decRefCounter()");
      ASSERT(_referenceCounter > 0L);
      --_referenceCounter;
    };
  private:
    union
    {
      ObjectPool* _pool;
      Entry* _nextFree;
    };
    long _referenceCounter;
    friend class EntryHandle;
    friend class ObjectPool<C>;
  }; 
protected:
  virtual Entry* reserveEntry() = 0;
  virtual Entry* tryToReserveEntry() = 0;
  virtual void recycleEntry(Entry* e) = 0;
private:
  friend class EntryHandle;
}; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL
#  define DEBUG_NAMESPACE "<global>"
#endif
#include "debugMacros.hpp"
namespace BK
{
  template <class C>
  inline
  void relocate(typename ObjectPool<C>::EntryHandle& source,void* target)
  {
    source.relocateTo(target);
  };
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OBJECT_POOL
#  define DEBUG_NAMESPACE "ObjectPool<class C>"
#endif
#include "debugMacros.hpp"
namespace BK 
{  
  template <class C>
  inline 
  void 
  ObjectPool<C>::reserveObject(EntryHandle& h)
  {
    CALL("reserveObject(EntryHandle& h)");
    Entry* e = reserveEntry();
    ASSERT(!e->_referenceCounter);
    e->_pool = this;
    e->incRefCounter();
    if (h._content)
      {
	h._content->decRefCounter();
	if (!h._content->_referenceCounter)
	  {
	    hibernate(h._content->content());
	    h._content->_pool->recycleEntry(h._content);
	  };
      };
    h._content = e;
  };
  template <class C>
  inline 
  bool 
  ObjectPool<C>::tryToReserveObject(EntryHandle& h)
  {
    CALL("tryToReserveObject(EntryHandle& h)");
    Entry* e = tryToReserveEntry();
    if (!e) return false;
    ASSERT(!e->_referenceCounter);
    e->_pool = this;
    e->incRefCounter();
    if (h._content)
      {
	h._content->decRefCounter();
	if (!h._content->_referenceCounter)
	  {
	    hibernate(h._content->content());
	    h._content->_pool->recycleEntry(h._content);
	  };
      };
    h._content = e;
    return true;
  };
  template <class C>
  inline 
  C* 
  ObjectPool<C>::lendObject()
  {
    CALL("lendObject()");
    Entry* e = reserveEntry();
    ASSERT(!e->_referenceCounter);
    e->_pool = this;
    return e->contentAddr();
  };
  template <class C>
  inline 
  void 
  ObjectPool<C>::returnObject(C* obj)
  {
    CALL("returnObject(C* obj)");
    hibernate(Entry::entryFor(obj)->content());
    ASSERT(!Entry::entryFor(obj)->_referenceCounter);
    Entry::entryFor(obj)->_pool->recycleEntry(Entry::entryFor(obj));
  };
}; 
#endif
