/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#define CORRECTION(expr) (((expr)*21)/20)
#define SPIDER_MODE 0
#include "CASCMode.hpp"
#include "Environment.hpp"
#include "Int.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "Refutation.hpp"
#include "Signature.hpp"
#include "Statistics.hpp"
#include "TheoryFinder.hpp"
#include "Tracer.hpp"
#include "XML.hpp"
namespace Shell {
class CASCMode::Schedule {
public:
  explicit Schedule(const Property&);
  Options* next ();
  bool hasNext () const;
private:
  int _length;
  const char** _options;
  const int* _timeLimits;
  int _nextSlice;
}; 
CASCMode::CASCMode (Options& options)
  : Run(options,VarList())
{
} 
void CASCMode::run ()
{
  TRACER("CASCMode::run");
  RunSaver saver(this);
  UnitList units;
  Environment::collectUnits(units);
  Problem problem(_options.problemName(),units);
  Property property(problem);
  property.scanProblem();
  TheoryFinder tf(problem,&property);
  tf.search();
//   cout << property.toSpider() << "\n";
//   exit(0);
  Schedule schedule(property);
  runSchedule(units,schedule,property,_options.timeLimitInDeciseconds());
}  
void CASCMode::reportAllocationFailure ()
{
  TRACER("CASCMode::reportAllocationFailure");
#if SPIDER_MODE
  cout << "? " 
       << _options.problemName() << ' '
       << "0.0 casc\n";
#else
  cout << "--------------------- CANNOT PROVE ----------------------\n"
          "----------------- allocation failure ----------------------\n";
#endif
} 
void CASCMode::runSchedule (UnitList& units,
			    Schedule& schedule,
			    const Property& property,
			    int timeLimit)
{
  TRACER("CASCMode::runSchedule");
  XML::DeepElement* proofAttempt = 0;
  while (schedule.hasNext()) {
    Options* options = schedule.next();
    Signature::sig.resetSkolemFunctionNumbers();
    int elapsedTime = _timer.elapsedDeciseconds();
    if (elapsedTime >= timeLimit) {
#if SPIDER_MODE
      cout << "? " 
 	   << _options.problemName() << ' '
 	   << _timer.elapsedSeconds() << ' ' 
 	   << _options.testId() << '\n';
#else
      cout << "--------------------- CANNOT PROVE ----------------------\n"
 	   << flush;
#endif
      if (_log) {
	_log->addChild(new XML::FloatElement("total_time",
					     (float)_timer.elapsedDeciseconds()/10));
	_log->addChild(new XML::TextElement("result",Run::toString(UNKNOWN)));
      }
      return;
    }
    int remainingTime = timeLimit - elapsedTime;
    int sliceTimeLimit = CORRECTION(options->timeLimitInDeciseconds());
    if (options->timeLimitInDeciseconds() == 0 || 
	sliceTimeLimit > remainingTime || 
        !schedule.hasNext()) {            
      sliceTimeLimit = remainingTime;
    }
    options->setMemoryLimit(_options.memoryLimit());
    options->setTimeLimitInDeciseconds(sliceTimeLimit);
    options->setSimulatedTimeLimit(CORRECTION(options->simulatedTimeLimit()));
#if SPIDER_MODE
#else
    cout << "Running slice "
	 << options->testId() 
	 << " for "
	 << sliceTimeLimit
	 << " deciseconds\n";
#endif
    Kernel kernel(*options,Signature::sig);
    if (_log) {
      proofAttempt = new XML::DeepElement("proof_attempt");
      _log->addChild(proofAttempt);
      proofAttempt->addChild(options->toXML());
    }
    UnitList copyOfUnits;
    Stack<Unit> stack(copyOfUnits);
    Iterator<Unit> us(units);
    while (us.hasNext()) {
      stack.push(us.next());
    }
    Problem problem(_options.problemName(),copyOfUnits);
    Preprocess prepro(problem,*options,property);
    prepro.preprocess();
    BK::DestructionMode::makeThorough();
    RunResult result = kernel.run(sliceTimeLimit,problem,_timer,_log);
    switch (result) {
    case REFUTATION:
#if SPIDER_MODE
      cout << "+ " 
 	   << _options.problemName() << ' '
 	   << _timer.elapsedSeconds() << ' ' 
 	   << _options.testId() << '\n';
#else
      cout << "--------------------- PROVED ----------------------\n"
 	   << flush;
#endif
      if (_options.proof() != Options::PROOF_OFF) {
 	Refutation ref(kernel,problem,_options);
#if SPIDER_MODE
#else
 	ref.prettyPrint(cout,1);
 	cout << flush;
#endif
	if (_log) {
	  proofAttempt->addChild(ref.toXML());
	}
      }
      break;
    case SATISFIABLE:
      if (property.category() == EPR) {
#if SPIDER_MODE
	cout << "- " 
	     << _options.problemName() << ' '
	     << _timer.elapsedSeconds() << ' ' 
	     << _options.testId() << '\n';
#else
	cout << "--------------------- UNPROVABLE ----------------------\n"
	     << flush;
#endif
	return;
      }
      break;
    case UNKNOWN:
#if SPIDER_MODE
#else
      cout << "------------------ SLICE UNSUCCESSFUL -------------------\n"
	   << flush;
#endif
      break;
    }
    if (_log) {
      Statistics statistics(kernel);
      XML::Element* xmlStat = statistics.toXML();
      if (xmlStat) {
	proofAttempt->addChild(xmlStat);
      }
      string resultString = Run::toString(result);
      proofAttempt->addChild(new XML::TextElement("slice_result",resultString));
      if (result == REFUTATION || ! schedule.hasNext()) {
	_log->addChild(new XML::FloatElement("total_time",
					     (float)_timer.elapsedDeciseconds()/10));
	_log->addChild(new XML::TextElement("result",resultString));
      }
      BK::DestructionMode::makeFast();
      kernel.closeSession();
      delete options;
    }
    if (result == REFUTATION) {
      return;
    }
  }
#if SPIDER_MODE
      cout << "? " 
 	   << _options.problemName() << ' '
 	   << _timer.elapsedSeconds() << ' ' 
 	   << _options.testId() << '\n';
#else
      cout << "--------------------- CANNOT PROVE ----------------------\n"
 	   << flush;
#endif
} 
CASCMode::Schedule::Schedule (const Property& property)
  : _nextSlice(0)
{
  TRACER("CASCMode::Schedule::Schedule");
  switch (property.category()) {
  case PEQ: 
    {
      _length = 10;
      static int timeLimits[] = {544,65,152,389,447,93,112,225,331,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs-11_16_bas=on_bd=preordered_bs=off_ep=exp1_nwc=1.3_600",
        "lrs+1011_6_bd=off_bs=off_ep=exp1_fsr=off_nwc=2.3_600",
        "lrs+17_2:1_bas=on_bs=off_lcm=kinky_600",
        "lrs-11_4:3_bas=on_ep=exp1_fde=none_600",
        "dis+1011_25_bas=on_ep=exp1_fde=none_nwc=1.1_stl=60_600",
        "dis+8_7_bas=on_bd=preordered_bs=off_ep=exp1_nwc=2.2_sos=on_600",
        "lrs-1012_32_bs=off_ep=exp1_nwc=4.0_600",
        "lrs+8_7_bas=on_bs=off_fde=none_nwc=1.7_600",
        "lrs-11_4_bas=on_bs=off_ep=exp1_fde=none_600",
        "lrs-1012_68_bs=off_fde=none_nwc=1.4_2400"
      };
      _options = options;
      // Total time: 4217
    }
    break;

  case HNE: 
    {
      _length = 10;
      static int timeLimits[] = {172,595,125,517,69,308,8,15,81,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+10_2:3_bs=off_fde=none_nwc=2.8_600",
        "lrs+11_500_bas=on_bs=off_ep=on_900",
        "lrs+1012_5:1_bd=off_bs=off_sos=on_600",
        "dis+10_13_ep=exp1_nwc=3.3_wi=on_600",
        "dis+1010_5:1_bs=off_lcm=predicate_nwc=4.5_sp=on_600",
        "lrs-1012_2_bs=off_ep=on_nwc=1.8_900",
        "lrs+10_200_bas=on_ep=on_fde=none_nwc=3.3_wi=on_600",
        "lrs-11_256_bs=off_ep=on_nwc=4.0_600",
        "lrs+8_7_bas=on_bs=off_fde=none_nwc=1.7_600",
        "lrs+17_4_bs=off_nwc=1.4_600"
      };
      _options = options;
      // Total time: 2384
    }
    break;
  case HEQ: 
    {
      _length = 14;
      static int timeLimits[] = {514,215,22,463,73,15,192,1,25,190,207,753,508,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "dis+13_11_bs=off_nwc=9.0_600",
        "lrs-1012_3_bs=off_fde=none_lcm=predicate_nwc=2.1_600",
        "lrs-1012_2_bas=on_bd=off_bs=off_lcm=kinky_sos=on_600",
        "ott+16_45_ep=exp1_nwc=1.1_600",
        "lrs+1010_4_bd=preordered_bs=off_lcm=kinky_nwc=3.0_600",
        "lrs+17_2:1_bas=on_bs=off_lcm=kinky_600",
        "dis+16_6_nwc=5.0_600",
        "lrs-11_8_bd=off_bs=off_ep=on_nwc=2.0_600",
        "dis+10_2:3_nwc=2.5_600",
        "dis+1011_2:1_bs=off_ep=on_nwc=5.0_600",
        "lrs+1012_2:3_bd=off_ep=on_nwc=1.6_600",
        "dis+16_26_bd=preordered_bs=off_fde=none_nwc=3.6_wi=on_1200",
        "lrs+16_24_bs=off_ep=on_fde=none_nwc=1.8_600",
        "lrs-1012_68_bs=off_fde=none_nwc=1.4_2400"
      };
      _options = options;
      // Total time: 4697
    }
    break;
  case NNE: 
    if (property.hasProp(PR_HAS_FLD2)) {
      _length = 7;
      static int timeLimits[] = {547,395,118,347,368,594,911};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+13_3:1_bs=off_ep=on_nwc=2.9_sos=on_600",
        "dis+10_15_bas=on_bs=off_ep=on_600",
        "lrs+16_5:4_bs=off_nwc=4.0_600",
        "lrs+17_13_bs=off_ep=on_nwc=2.4_sos=on_600",
        "lrs+1011_8_bs=off_ep=exp1_fde=none_lcm=kinky_nwc=3.5_600",
        "lrs+1011_2:3_bs=off_lcm=predicate_sos=on_900",
        "lrs+17_2_bs=off_ep=on_nwc=2.6_1200"
      };
      _options = options;
      // Total time: 3280
      break;
    }
    else { 
      _length = 7;
      static int timeLimits[] = {355,851,388,390,491,1027,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+13_3:1_bs=off_ep=on_nwc=2.9_sos=on_600",
        "lrs+1011_3:2_ep=on_lcm=kinky_sos=on_900",
        "dis+1011_12_bs=off_sos=on_600",
        "lrs+8_14_ep=exp1_fde=none_nwc=5.2_wi=on_600",
        "dis+17_8_bas=on_bs=off_ep=on_sos=on_600",
        "dis+1012_5_bas=on_bs=off_ep=on_fde=none_lcm=kinky_sos=on_1200",
        "lrs+11_500_bas=on_bs=off_ep=on_900"
      };
      _options = options;
      // Total time: 3503
      break;
    }
  case NEQ: 
    if (property.props() == 3) {
      _length = 21;
      static int timeLimits[] = {211,371,288,493,97,182,90,137,21,305,34,470,82,447,2,20,155,271,287,1545,824};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+1010_1_lcm=predicate_600",
        "lrs+1011_2:1_bs=off_lcm=kinky_sos=on_600",
        "lrs+10_2:1_bas=on_sos=on_sp=on_600",
        "lrs-1012_1_ep=on_lcm=kinky_nwc=5.0_sos=on_sp=on_600",
        "dis+1011_3_bd=off_ep=on_nwc=3.5_600",
        "lrs+1011_2:3_bs=off_ep=on_lcm=kinky_nwc=5.8_sp=on_600",
        "lrs-1012_2_bas=on_bd=off_bs=off_lcm=kinky_sos=on_600",
        "dis+11_256_bd=off_bs=off_fde=none_sos=on_600",
        "dis+13_11_bs=off_nwc=9.0_600",
        "lrs+1011_3_bs=off_600",
        "dis+1013_32_nwc=1.5_600",
        "dis+16_3:1_bs=off_ep=on_nwc=3.9_wi=on_600",
        "lrs+1011_5:1_bd=off_bs=off_ep=on_fde=none_lcm=predicate_sos=on_600",
        "lrs+1011_5:1_bs=off_ep=on_sos=on_600",
        "lrs+1012_10_bas=on_bs=off_ep=exp1_fde=none_nwc=3.5_600",
        "lrs-1012_11_bs=off_nwc=4.0_600",
        "lrs+1011_5:4_bas=on_bs=off_ep=on_lcm=kinky_nwc=3.8_600",
        "lrs+17_4:3_bas=on_bs=off_ep=on_nwc=2.3_600",
        "lrs+11_1_bas=on_bd=off_bs=off_ep=on_lcm=predicate_nwc=1.7_600",
        "dis+13_43_bas=on_bd=off_ep=on_nwc=4.8_1800",
        "lrs+16_5:4_ep=exp1_lcm=predicate_nwc=1.1_sos=on_sp=on_900"
      };
      _options = options;
      // Total time: 6332
      break;
    }
    else if (property.props() == 7) {
      _length = 21;
      static int timeLimits[] = {271,6,46,31,222,118,198,14,35,238,30,1,12,21,47,171,1,4,45,399,479};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+17_4_bas=on_bs=off_fde=none_nwc=2.7_600",
        "lrs-1012_22_bas=on_bs=off_ep=on_fde=none_nwc=3.9_600",
        "lrs-1012_1_bas=on_bd=off_bs=off_ep=on_lcm=predicate_nwc=3.0_600",
        "lrs+16_24_bs=off_ep=on_fde=none_nwc=1.8_600",
        "dis-1012_2_bas=on_bs=off_ep=on_fde=none_lcm=predicate_nwc=3.8_600",
        "lrs+1011_5:1_bd=off_bs=off_ep=on_fde=none_lcm=predicate_sos=on_600",
        "lrs+1012_10_bas=on_bs=off_ep=exp1_fde=none_nwc=3.5_600",
        "dis+16_13_bas=on_ep=exp1_fsr=off_nwc=3.6_600",
        "dis+1011_36_bs=off_ep=on_lcm=predicate_nwc=4.0_600",
        "lrs+16_5:4_bs=off_nwc=4.0_600",
        "lrs-1012_32_bs=off_ep=exp1_nwc=4.0_600",
        "dis+13_11_bs=off_nwc=9.0_600",
        "dis-1012_5:1_bas=on_ep=exp1_fde=none_nwc=2.9_600",
        "lrs+1012_5:1_bd=off_bs=off_sos=on_600",
        "dis+16_6_nwc=5.0_600",
        "lrs+17_4:3_bas=on_bs=off_ep=on_nwc=2.3_600",
        "dis+16_32_bs=off_ep=on_nwc=2.0_600",
        "lrs-1012_11_bs=off_nwc=4.0_600",
        "dis+10_15_bas=on_bs=off_ep=on_600",
        "lrs+1011_20_bas=on_bs=off_ep=exp1_fde=none_nwc=3.2_600",
        "lrs-11_4_bs=off_ep=on_lcm=kinky_nwc=1.8_600"
      };
      _options = options;
      // Total time: 2388
      break;
    }
    else {
      _length = 17;
      static int timeLimits[] = {454,336,146,403,125,32,53,185,170,117,262,1,119,198,289,396,563};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+17_4:3_bas=on_bs=off_ep=on_nwc=2.3_600",
        "lrs+1011_5:4_bas=on_bs=off_ep=on_lcm=kinky_nwc=3.8_600",
        "lrs+17_4_bas=on_bs=off_fde=none_nwc=2.7_600",
        "lrs+10_2:1_bas=on_sos=on_sp=on_600",
        "lrs+16_6_bas=on_bs=off_nwc=2.0_600",
        "lrs+1013_6:1_ep=on_fde=none_nwc=2.0_600",
        "dis+16_2_bas=on_bs=off_ep=exp1_fde=none_nwc=3.7_600",
        "dis+1011_36_bs=off_ep=on_lcm=predicate_nwc=4.0_600",
        "lrs+17_11_bas=on_bs=off_fde=none_nwc=2.3_600",
        "lrs+10_2:3_bs=off_fde=none_nwc=2.8_600",
        "lrs+1011_3:2_bas=on_bs=off_ep=on_lcm=predicate_sos=on_600",
        "lrs+1011_20_bas=on_bs=off_ep=exp1_fde=none_nwc=3.2_600",
        "lrs+8_24_bd=off_bs=off_fsr=off_lcm=predicate_nwc=4.2_600",
        "dis+11_4_lcm=predicate_600",
        "lrs+11_1_bas=on_bd=off_bs=off_ep=on_lcm=predicate_nwc=1.7_600",
        "lrs+17_10_bas=on_bd=off_bs=off_fde=none_nwc=2.6_600",
        "lrs+10_4_lcm=kinky_nwc=3.0_600"
      };
      _options = options;
      // Total time: 3848
      break;
    }
  case UEQ:
    if (property.hasProp(PR_LO_GROUP)) {
      _length = 5;
      static int timeLimits[] = {1108,685,0,0,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+8_86_bas=on_bs=off_wi=on_1200",
        "lrs+8_12_bs=off_1200",
	"lrs+8_10_1200",
	"lrs+8_20_bas=on_bs=off_fde=none_fsr=off_nwc=2.5_1200",
	"lrs+8_10:1_bs=off_nwc=1.9_1200"
      };
      _options = options;
    }
    else if (property.hasProp(PR_COMBINATOR_B)) {
      _length = 8;
      static int timeLimits[] = {47,424,1014,17,4663,0,0,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+8_10_bd=off_bs=off_fde=none_nwc=4.0_1200",
        "lrs+8_0_bas=on_nwc=4.5_wi=on_1200",
        "lrs+8_13_fde=none_nwc=1.1_1200",
        "lrs+8_8_bd=preordered_bs=off_fde=none_nwc=2.5_1200",
        "dis+8_2_bas=on_fde=none_nwc=3.3_wi=on_4800",
	"lrs+8_10_1200",
	"lrs+8_20_bas=on_bs=off_fde=none_fsr=off_nwc=2.5_1200",
	"lrs+8_10:1_bs=off_nwc=1.9_1200"
      };
      _options = options;
    }
    else if (property.hasProp(PR_COMBINATOR)) {
      _length = 8;
      static int timeLimits[] = {383,989,677,758,863,0,0,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+8_10:1_bs=off_nwc=1.9_1200",
        "lrs+8_10_bd=off_bs=off_fde=none_nwc=4.0_1200",
        "lrs+8_20_bas=on_bs=off_fde=none_fsr=off_nwc=2.5_1200",
        "lrs+8_5_bs=off_fde=none_nwc=2.3_1200",
        "lrs+8_8:1_bas=on_bd=preordered_bs=off_nwc=1.3_3000",
	"lrs+8_10_1200",
	"lrs+8_20_bas=on_bs=off_fde=none_fsr=off_nwc=2.5_1200",
	"lrs+8_10:1_bs=off_nwc=1.9_1200"
      };
      _options = options;
    }
    else {
      _length = 9;
      static int timeLimits[] = {41,621,611,65,379,1801,0,0,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+8_2:3_600",
        "lrs+8_6_bs=off_fde=none_1200",
        "lrs+8_5:4_1200",
        "lrs+8_0_bd=preordered_bs=off_nwc=4.3_1200",
        "lrs+8_8:1_bas=on_bd=preordered_bs=off_nwc=1.3_3000",
        "dis+8_15_bas=on_bs=off_nwc=1.4_wi=on_2100",
	"lrs+8_10_1200",
	"lrs+8_20_bas=on_bs=off_fde=none_fsr=off_nwc=2.5_1200",
	"lrs+8_10:1_bs=off_nwc=1.9_1200"
      };
      _options = options;
    }
    break;
  case EPR:
    {
      _length = 2;
      static int timeLimits[] = {3000,0};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+10_4_nwc=2.0_600",
        "lrs+16_5:4_bs=off_nwc=4.0_600"
      };
      _options = options;
    }
    break;
  case FNE: 
    {
      _length = 3;
      static int timeLimits[] = {26,1,362};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+13_8_bs=off_ep=exp1_nwc=2.5_sp=on_600",
        "lrs+10_4:3_bas=on_bs=off_ep=on_nwc=1.3_600",
        "lrs+13_4:1_bas=on_bs=off_ep=exp1_nwc=1.5_600"
      };
      _options = options;
      // Total time: 389
    }
    break;
  case FEQ: 
    if (! property.props()) { 
      _length = 13;
      static int timeLimits[] = {751,310,12,860,332,448,126,350,214,396,433,454,609};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "dis+1010_112_bas=on_bd=off_ep=exp1_fde=none_nwc=3.1_1200",
        "dis+1013_32_nwc=1.5_600",
        "lrs-11_4_bs=off_ep=on_lcm=kinky_nwc=1.8_600",
        "dis+11_19_bas=on_bd=off_ep=exp1_lcm=predicate_nwc=2.3_900",
        "dis+1012_128_lcm=predicate_nwc=3.3_600",
        "lrs+1013_10_ep=exp1_fde=none_nwc=1.6_600",
        "lrs+8_6_fde=none_nwc=1.5_600",
        "lrs+16_27_bas=on_bd=off_bs=off_ep=exp1_nwc=1.2_600",
        "dis+1013_6_ep=exp1_fde=none_nwc=1.1_wi=on_600",
        "lrs+1013_10_bd=preordered_ep=exp1_fde=none_nwc=1.5_wi=on_600",
        "lrs+1013_50_bd=off_ep=exp1_fde=none_600",
        "dis-11_34_bd=off_bs=off_ep=exp1_fsr=off_nwc=1.2_wi=on_900",
        "dis+1010_8_bas=on_bs=off_ep=exp1_lcm=kinky_900"
      };
      _options = options;
      // Total time: 5295
      break;
    }
    else if (property.props() == 1) {
      _length = 19;
      static int timeLimits[] = {176,563,582,28,521,598,568,149,160,35,132,71,310,19,264,23,101,201,585};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+1012_35_bd=off_bs=off_fde=none_600",
        "lrs+10_10:1_bs=off_fde=none_lcm=predicate_nwc=1.5_600",
        "lrs+10_4:3_bas=on_bs=off_ep=on_nwc=1.3_600",
        "dis+11_61_bas=on_fde=none_nwc=4.0_600",
        "lrs+1011_44_bas=on_bd=off_bs=off_fde=none_lcm=kinky_nwc=1.3_600",
        "lrs+17_10:1_bd=off_bs=off_ep=exp1_nwc=1.4_600",
        "lrs+11_0_bas=on_bd=off_bs=off_ep=exp1_fde=none_lcm=kinky_nwc=2.3_600",
        "lrs+17_2:1_bas=on_bs=off_ep=exp1_lcm=kinky_nwc=1.5_600",
        "lrs+13_10:1_bas=on_bs=off_ep=exp1_lcm=predicate_nwc=1.1_600",
        "lrs+17_8:1_bas=on_bd=preordered_bs=off_ep=exp1_fde=none_fsr=off_lcm=predicate_nwc=1.1_sos=on_sp=on_600",
        "lrs+16_27_bas=on_bd=off_bs=off_ep=exp1_nwc=1.2_600",
        "lrs+1011_5:1_ep=exp1_fde=none_nwc=1.1_600",
        "dis+10_1_bd=preordered_bs=off_fde=none_lcm=predicate_nwc=2.1_600",
        "dis+16_6_nwc=5.0_600",
        "dis+1010_8_bas=on_bs=off_ep=exp1_lcm=kinky_900",
        "lrs+11_1_bas=on_bd=off_bs=off_ep=on_lcm=predicate_nwc=1.7_600",
        "dis+10_3:1_bas=on_bd=off_bs=off_lcm=predicate_nwc=1.1_600",
        "lrs+10_2:1_bs=off_ep=exp1_fde=none_lcm=predicate_nwc=1.8_600",
        "lrs+1010_4_bd=preordered_bs=off_lcm=kinky_nwc=3.0_600"
      };
      _options = options;
      // Total time: 5086
      break;
    }
    else {
      _length = 29;
      static int timeLimits[] = {144,209,87,319,266,8,261,540,5,29,34,4,0,504,2,50,54,171,199,2,13,195,129,149,169,192,232,454,547};
      _timeLimits = timeLimits;
      static const char* options[] = {
        "lrs+13_4:1_bas=on_bs=off_ep=exp1_nwc=1.5_600",
        "lrs+1013_10:1_bd=off_bs=off_nwc=1.9_600",
        "lrs+10_10:1_bs=off_fde=none_lcm=predicate_nwc=1.5_600",
        "lrs+1011_1_nwc=1.5_600",
        "lrs+17_8:1_bas=on_bd=preordered_bs=off_ep=exp1_fde=none_fsr=off_lcm=predicate_nwc=1.1_sos=on_sp=on_600",
        "dis+1010_112_bas=on_bd=off_ep=exp1_fde=none_nwc=3.1_1200",
        "lrs+17_2:1_bas=on_bs=off_ep=exp1_lcm=kinky_nwc=1.5_600",
        "dis+1010_3:2_bd=off_ep=exp1_fde=none_lcm=predicate_nwc=4.0_900",
        "lrs+11_3_bs=off_ep=exp1_fde=none_nwc=1.1_wi=on_600",
        "lrs-1012_5:1_sos=on_sp=on_600",
        "dis+10_29_ep=on_fde=none_wi=on_600",
        "lrs+11_6_bs=off_ep=exp1_fde=none_lcm=predicate_nwc=1.2_600",
        "lrs+17_5:4_ep=exp1_fde=none_nwc=5.4_600",
        "lrs+11_0_bas=on_bd=off_bs=off_ep=exp1_fde=none_lcm=kinky_nwc=2.3_600",
        "lrs+1011_5:1_bd=off_bs=off_ep=on_fde=none_lcm=predicate_sos=on_600",
        "lrs+17_2:1_bas=on_bs=off_lcm=kinky_600",
        "lrs+1010_4_bd=preordered_bs=off_lcm=kinky_nwc=3.0_600",
        "lrs+10_4:3_bas=on_bs=off_ep=on_nwc=1.3_600",
        "lrs+13_10:1_bas=on_bs=off_ep=exp1_lcm=predicate_nwc=1.1_600",
        "lrs+10_60_bas=on_ep=exp1_fde=none_fsr=off_lcm=kinky_nwc=1.8_sp=on_600",
        "lrs+10_2:1_bs=off_ep=exp1_fde=none_lcm=predicate_nwc=1.8_600",
        "lrs+1013_11_bas=on_ep=on_lcm=kinky_nwc=1.2_600",
        "lrs+10_2:1_bas=on_sos=on_sp=on_600",
        "lrs-1012_4_ep=on_nwc=5.0_600",
        "dis+10_3:1_bas=on_bd=off_bs=off_lcm=predicate_nwc=1.1_600",
        "lrs+8_6_fde=none_nwc=1.5_600",
        "lrs+16_2_bd=off_ep=exp1_fde=none_nwc=2.6_600",
        "dis+1013_32_nwc=1.5_600",
        "lrs+1013_10_ep=exp1_fde=none_nwc=1.6_600"
      };
      _options = options;
      // Total time: 4968
      break;
    }
  default: 
    {
      cout << "Does not work!\n";
      exit(0);
    }
    break;
  }
} 
bool CASCMode::Schedule::hasNext () const
{
  return _nextSlice < _length ? true : false;
} 
Options* CASCMode::Schedule::next ()
{
  TRACER("CASCMode::Schedule::next");
  const char* opts = _options[_nextSlice];
  int timeLimit = _timeLimits[_nextSlice];
  _nextSlice++;
  Options* options = new Options();
  options->readFromTestId(opts);

  int delta = 0; // time limit correction
  if (timeLimit) {
    delta = timeLimit / 20;
    if (delta < 10) {
      delta = 10;
    }
  }

  options->setSimulatedTimeLimit(options->timeLimitInDeciseconds());
  options->setTimeLimitInDeciseconds(timeLimit + delta);
  return options;
} 
}
