/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Chain__
#  define __Chain__
#include "Allocator.hpp"
namespace Shell {
template <class C>
class Chain
{
 public:
  class Link
  {
   private:
    Link* _prev;
    Link* _next;
    C _elem;
   public:
    inline Link (const C& elem) 
      : _elem (elem) 
      {
      }
    inline Link ()  
      { 
        _prev = this;
        _next = this; 
      }
    inline Link* prev () const { return _prev; }
    inline void prev ( Link* p ) { _prev = p; }
    inline Link* next () const { return _next; }
    inline void next ( Link* n ) { _next = n; }
    void insertBefore (const C& elem)
      {
        Link* l = new Link(elem);
        l->_next = this;
        l->_prev = _prev;
        _prev->_next = l;
        _prev = l;
      }
    void insertAfter (const C& elem)
      {
        Link* l = new Link(elem);
        l->_prev = this;
        l->_next = _next;
        _next->_prev = l;
        _next = l;
      }
    void del ()
      {
        ASS(_next != this && _prev != this);
        _prev->_next = _next;
        _next->_prev = _prev;
        delete this;
      }
    const C& content () const { return _elem; }
    CLASS_NAME("Chain::Link");
    USE_ALLOCATOR(Link);
  }; 
  inline const Link* header () const { return &_header; }
  inline void addLast (const C& elem)   
    { _header.insertBefore (elem); }
  inline void addFirst (const C& elem)   
    { _header.insertAfter (elem); }
  inline void delLast ()    
    { _header.prev()->del(); }
  void delFirst ()    
    { _header.next()->del(); }
  inline bool isEmpty () const
    { return _header.prev() == &_header; }
  Link* first () const
    { ASS(! isEmpty());
      return _header.next(); }
  const C& last () const
    { ASS(! isEmpty());
      return _header.prev(); }
  int length () const
    { int len = 0;
      Iterator links (*this);
      while (links.hasNext()) {
        links.next ();
        len ++;
      }
      return len;
    } 
  CLASS_NAME("Chain");
  USE_ALLOCATOR(Chain);
  class Iterator {
   public: 
    inline explicit 
    Iterator (Chain& c) 
      : _header (c.header()),
        _current (_header->next())
        {}
    Iterator (const Chain& c) 
      : _header (c.header()),
        _current (_header->next())
        {}
    inline 
    bool hasNext () 
      { return _header != _current; }
    inline 
    const C& next () 
      { ASS(hasNext());
        const C& result = _current->content();
        _current = _current->next();
        return result;
      }
   private:
    const Link* _header;
    Link* _current;
  };
  class DelIterator {
   public:
    inline 
    DelIterator ( Chain& c ) : 
      _header ( c.header() ),
      _current ( _header->next() )
      {}  
    inline 
    bool hasNext () const
      { return _header != _current; }
    inline 
    const C& next () 
      { 
        ASS(_current != _header);
        const C& result = _current->content();
        _current = _current->next();
        return result;
      }
    inline
    void del () 
      {
        ASS( _current->prev() != _header ); 
        _current->prev()->del();
      }
    inline
    void replace (const C& elem) 
      {
        ASS( _current->prev() != _header ); 
        _current->prev()->del();
        _current->insertBefore (elem);
      }
    inline
    void insert (const C& elem) 
      {
        ASS( _current->prev() != _header ); 
        _current->prev()->insertBefore (elem);
      }
   private:
    const Link* _header;
    Link* _current;
  }; 
  ~Chain ()
  {
    DelIterator it (*this);
    while (it.hasNext()) {
      it.del();
    }
  } 
 private:
  Link _header;
 friend class Iterator;
};  
}
#endif
