/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string>
#include "Int.hpp"
#include "Inference.hpp"
#include "Tracer.hpp"
#include "Tautology.hpp"
#include "Problem.hpp"
#include "Signature.hpp"
#include "DoubleSubstitution.hpp"
#include "ClausalDefinition.hpp"
namespace Shell {
class ClausalDefinition::Counter
{
public:
  Counter();
  int posFlat;
  int negFlat;
  int posNonFlat;
  int negNonFlat;
  const PredicateSymbol* symbol;
  Unit lastPosClause;
  Unit lastNegClause;
  bool hasRecursive;
  bool hasTwo;
  int whatToDo;
}; 
ClausalDefinition::Counter::Counter ()
  :
  posFlat(0),
  negFlat(0),
  posNonFlat(0),
  negNonFlat(0),
  symbol(0),
  hasRecursive(false),
  hasTwo(false),
  whatToDo(0)
{
  TRACER("ClausalDefinition::Counter::Counter");
}
ClausalDefinition::ClausalDefinition (Problem& problem,Signature& signature)
  : _problem(problem),
    _signature(signature)
{
  TRACER("ClausalDefinition::ClausalDefinition");
} 
int ClausalDefinition::findAndReplace ()
{
  TRACER("ClausalDefinition::findAndReplace");
  int totalReplaced = 0;
  for (;;) {
    Counter* counter = new Counter[_signature.noOfPreds()];
    UnitChain::Iterator units (_problem.units());
    while (units.hasNext()) {
      Unit u(units.next());
      u.setMark(0);
      ASS (u.unitType() == CLAUSE);
      const Clause& cls = u.clause();
      Iterator<Literal> lits(cls.literals());
      while (lits.hasNext()) {
	const Literal& lit = lits.next();
	const Atom& atom = lit.atom();
	const PredicateSymbol& pred = atom.functor();
	int p = pred.number();
	Counter& c = counter[p];
	c.symbol = &pred; 
	if (lit.sign()) {
	  if (c.lastPosClause == u) {
	    c.hasTwo = true;
	  }
	  else {
	    c.lastPosClause = u;
	  }
	  if (c.lastNegClause == u) {
	    c.hasRecursive = true;
	  }
	}
	else {
	  if (c.lastNegClause == u) {
	    c.hasTwo = true;
	  }
	  else {
	    c.lastNegClause = u;
	  }
	  if (c.lastPosClause == u) {
	    c.hasRecursive = true;
	  }
	}
	if (atom.args().isLinearSequenceOfVariables()) {
	  if (lit.sign()) {
	    c.posFlat++;
	  }
	  else {
	    c.negFlat++;
	  }
	}
	else { 
	  if (lit.sign()) {
	    c.posNonFlat++;
	  }
	  else {
	    c.negNonFlat++;
	  }
	}
      }
    }
    bool toDo = false;
    for (int i = _signature.noOfPreds()-1;i >= 0;i--) {
      Counter& c = counter[i];
      const PredicateSymbol* p = c.symbol;
      if (! p || p->isEquality() || p->isEvaluable() || p->isAnswer()) {
	continue;
      }
      int pos = c.posNonFlat + c.posFlat;
      int neg = c.negNonFlat + c.negFlat;
      if ((pos == 0 && neg != 0) || 
	  (pos != 0 && neg == 0)) { 
	c.whatToDo = 2;
	toDo = true;
      }
      if (pos == 1 &&
          (c.posNonFlat == 0 || c.negNonFlat == 0) &&
	  c.lastPosClause.clause().literals().length() <= 2) {
	if (c.hasRecursive) {
	  c.whatToDo = 2;
	  toDo = true;
	}
	else if (c.hasTwo) { 
	  if (c.posNonFlat == 0) { 
	    c.whatToDo = 1;
	    toDo = true;
	  }
	}
	else {
	  c.whatToDo = 1;
	  toDo = true;
	}
      }
      else if (neg == 1 && 
	       (c.posNonFlat == 0 || c.negNonFlat == 0) &&
	       c.lastNegClause.clause().literals().length() <= 2) {
	if (c.hasRecursive) {
	  c.whatToDo = 2;
	  toDo = true;
	}
	else if (c.hasTwo) {
	  if (c.negNonFlat == 0) { 
	    c.lastPosClause = c.lastNegClause; 
	    c.whatToDo = 1;
	    toDo = true;
	  }
	}
	else {
	  c.lastPosClause = c.lastNegClause; 
	  c.whatToDo = 1;
	  toDo = true;
	}
      }
    }
    if (! toDo) {
      delete [] counter;
      return totalReplaced;
    }
    int replaced = 0;
    bool toRepeat = false; 
    for (int j = _signature.noOfPreds()-1;j >= 0;j--) {
      Counter& c = counter[j];
      switch (c.whatToDo) {
      case 0:
	break;
      case 1:
	toRepeat = toRepeat || resolveAll(c.symbol,c.lastPosClause,counter);
	replaced++;
	break;
      case 2:
	deleteAll(c.symbol);
	replaced++;
	break;
      }
    }
    delete [] counter;
    totalReplaced += replaced;
    if (! toRepeat) {
      return totalReplaced;
    }
  }
} 
void ClausalDefinition::deleteAll (const PredicateSymbol* pred)
{
  TRACER("ClausalDefinition::deleteAll");
  UnitChain::DelIterator del (_problem.giveUnits());
  while (del.hasNext()) {
    Unit u(del.next());
    Iterator<Literal> lits(u.clause().literals());
    while (lits.hasNext()) {
      const Literal& lit = lits.next();
      const PredicateSymbol& p = lit.atom().functor();
      if (pred == &p) {
	u.setMark(1); 
	del.del();
	break;
      }
    }
  }
} 
bool ClausalDefinition::resolveAll (const PredicateSymbol* p, 
				    Unit& u, 
				    Counter* counters)
{
  TRACER("ClausalDefinition::resolveAll");
  counters[p->number()].whatToDo = 0; 
  if (u.mark() == 1) { 
    deleteAll(p);
    return false;
  }
  ASS(u.mark() != 2);
  u.setMark(1); 
  bool result = false;
  UnitChain::DelIterator del (_problem.giveUnits());
  while (del.hasNext()) {
    Unit u1(del.next());
    if (u == u1) {
      del.del();
      continue;
    }
    if (u1.mark() == 2) { 
      continue;
    }
    Iterator<Literal> lits(u1.clause().literals());
    while (lits.hasNext()) {
      const Literal& lit = lits.next();
      const PredicateSymbol& q = lit.atom().functor();
      if (p == &q) {
	Unit resolvent;
	resolveAllUpon(p,u,u1,resolvent);
	u1.setMark(1);
	bool deleted = false;
	lits.reset(resolvent.clause().literals());
	while (lits.hasNext()) {
	  const PredicateSymbol& q = lits.next().atom().functor();
	  Counter& c = counters[q.number()];
	  if (c.whatToDo == 2) {
	    deleted = true;
	    break;
	  }
	  if (c.whatToDo == 1) {
	    c.whatToDo = 0;
	    result = true;
	  }
	}
	if (deleted) {
	  del.del();
	}
	else {
	  Unit res;
	  switch (Tautology::isTautology(resolvent,res)) {
	  case -1:
	    res.setMark(2);
	    del.replace(res);
	    break;
	  case 0:
	    resolvent.setMark(2);
	    del.replace(resolvent);
	    break;
	  case 1: 
	    del.del();
	    break;
	  }
	}
	break;
      }
    }
  }
  return result;
} 
void ClausalDefinition::resolveAllUpon (const PredicateSymbol* s,
					const Unit& u1,
					const Unit& u2,
					Unit& result)
{
  TRACER("ClausalDefinition::resolveAllUpon");
  Literal lit;
  LiteralList lits;
  Stack<Literal> stack1(lits);
  Iterator<Literal> it(u1.clause().literals());
  while (it.hasNext()) {
    const Literal& l = it.next();
    if (&l.atom().functor() == s) { 
      ASS(lit.isNull());
      lit = l;
    }
    else {
      stack1.push(l);
    }
  }
  ASS(! lit.isNull());
  result = u2;
  for (;;) {
    Literal lit1(lit);
    LiteralList lits1(lits);
    bool resolved = false;
    Literal lit2;
    LiteralList lits2;
    Stack<Literal> stack2(lits2);
    it.reset(result.clause().literals());
    while (it.hasNext()) {
      const Literal& l2 = it.next();
      if (&l2.atom().functor() == s) {
	resolved = true;
	lit2 = l2;
	while (it.hasNext()) {
	  stack2.push(it.next());
	}
      }
      else {
	stack2.push(l2);
      }
    }
    if (! resolved) {
      return;
    }
    DoubleSubstitution subst;
#if DEBUG_SHELL
     ASS(subst.resolve(lit1,lit2));
#else
    subst.resolve(lit1,lit2);
#endif
    subst.apply(lits1,0);
    subst.apply(lits2,1);
    lits1.append(lits2);
    Inference* inference = new Inference(IR_RESOLUTION,
					 UnitList(u1,UnitList(result)));
    result = Unit(Clause(lits1),inference);
  }
} 
}
