/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Definition.hpp"
#include "FunctionDefinition.hpp"
#include "Unit.hpp"
#include "Problem.hpp"
#include "Tracer.hpp"
#include "FormulaVarIterator.hpp"
#include "AtomIterator.hpp"
#include "VarCounter.hpp"
#include "SymCounter.hpp"
#include "Signature.hpp"
#include "VampireException.hpp"
using namespace Shell;
bool Definition::makeDefinition (const Unit& unit, Unit& def)
{
  TRACER ("Definition::makeDefinition");
  switch (unit.unitType()) {
  case CLAUSE: 
    {
      Clause c(unit.clause());
      Term lt;
      Term rt;
      if (! FunctionDefinition::isFunctionDefinition(c,lt,rt)) {
	return false;
      }
      Unit current(unit);
      ASS(! c.isEmpty());
      Atom a (c.literals().head().atom());
      ASS(a.isEquality());
      TermList ts(a.args());
      if (ts.head() == lt) {
	current.setDefType (DT_FUN);
	def = current;
	return true;
      }
      Literal lit (true, Atom(ts.second(),ts.head()));
      LiteralList lits(lit);
      Clause c1(lits);
      Unit result(IR_REORDER_EQ, c1, current);
      result.setDefType (DT_FUN);
      def = result;
      return true;
    }
  case FORMULA: 
    {
      Formula f (unit.formula());
      Atom lhs;
      Formula rhs;
      Term lt;
      Term rt;
      bool isPredDef = isPredicateDefinition(f,lhs,rhs);
      if (! isPredDef) {
	if (! FunctionDefinition::isFunctionDefinition(f,lt, rt)) {
	  return false;
	}
      }
      Unit current (unit);
      if ( f.connective() == FORALL &&
           f.qarg().connective() == FORALL ) {
        VarList vs;
        Formula g(f);
        while (g.connective() == FORALL) {
          vs.append (g.vars());
          g = g.qarg ();
        }
        f = Formula (FORALL, vs, g);
        current = Unit(IR_FLATTEN, f, current);
      }
      if (isPredDef) { 
	Formula g (f.connective() == IFF ? f : f.qarg());
	ASS(g.connective() == IFF);
	if (g.right() == rhs) {
	  current.setDefType (DT_PRED);
	  def = current;
	  return true;
	}
	Formula g1(IFF, g.right(), g.left());
	if (f.connective() == FORALL) {
	  g1 = Formula(FORALL, f.vars(), g1);
	}
	current = Unit(IR_PERMUT,g1,current);
	current.setDefType(DT_PRED);
	def = current;
	return true;
      }
      Formula g (f.connective() == FORALL ? f.qarg() : f);
      ASS (g.connective() == ATOM);
      Atom a (g.atom ());
      ASS (a.isEquality());
      TermList ts(a.args());
      if (ts.head() == lt) {
	current.setDefType (DT_FUN);
	def = current;
	return true;
      }
      Formula g1 (Atom(ts.second(), ts.head()));
      if (f.connective() == FORALL) {
	g1 = Formula (FORALL, f.vars(), g1);
      }
      current = Unit(IR_REORDER_EQ, g1, current);
      current.setDefType(DT_FUN);
      def = current;
      return true;
    }
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
bool Definition::isPredicateDefinition (const Formula& g,
					Atom& lhs, 
					Formula& rhs)
{
  TRACER ("Definition::isPredicateDefinition");
  Formula f (g);
  while (f.connective() == FORALL) {
    f = f.qarg ();
  }
  if (f.connective() != IFF) {
    return 0;
  }
  Formula lft (f.left());
  Formula rht (f.right());
  if (lft.connective() == ATOM && isPredicateDefinition(lft.atom(), rht)) {
    lhs = lft.atom();
    rhs = rht;
    return true;
  }
  if (rht.connective() == ATOM && isPredicateDefinition(rht.atom(), lft)) {
    lhs = rht.atom();
    rhs = lft;
    return true;
  }
  return false;
} 
bool Definition::occurs (const PredicateSymbol& p, const Formula& f)
{
  TRACER("Definition::occurs (const PredicateSymbol& p, const Formula& f)");
  AtomIterator at(f);
  while (at.hasNext()) {
    if (at.next().functor() == p) {
      return true;
    }
  }
  return false;
} 
bool Definition::isPredicateDefinition (const Atom& lhs, const Formula& rhs)
{
  TRACER("Definition::isPredicateDefinition (const Atom& lhs...)");
  if (lhs.isEquality()) {
    return false;
  }
  if (occurs(lhs.functor(),rhs)) {
    return false;
  }
  VarCounter counter;
  Iterator<Term> ls (lhs.args());
  while (ls.hasNext()) {
    Term v (ls.next());
    if (v.tag() != VAR) {
      return false;
    }
    Var w (v.var());
    if (counter.getCount(w) != 0) {
      return false;
    }
    counter.incCount(w);
  }
  int found = 0;
  FormulaVarIterator vs(rhs);
  while (vs.hasNext()) {
    Var v = vs.next();
    switch (counter.getCount(v)) {
    case 0:
      return false;
    case 1:
      found++;
      counter.incCount(v);
      break;
    default:
      break;
    }
  }
  return lhs.args().hasLength(found);
} 
int Definition::findPredicateDefinitions (Problem& problem)
{
  TRACER ("Definition::findDefinitions");
  int found = 0;
  UnitChain::DelIterator unit (problem.giveUnits()); 
  while (unit.hasNext()) {
    Unit u (unit.next ());
    Unit d;
    if (makeDefinition(u,d)) {
      found ++;
      if (u != d) { 
        unit.replace(d);
      }
    }
  }
  return found;
} 
int Definition::removeUnusedPredicateDefinitions (Problem& problem)
{
  TRACER ("Definition::removeUnusedPredicateDefinitions");
  if (findPredicateDefinitions(problem) == 0) {
    return 0;
  }
  SymCounter cnt (Signature::sig);
  cnt.count (problem,1);
  int removed = 0;
  bool changed = true;
  UnitChain& units = problem.giveUnits();
  while (changed) {
    changed = false;
    UnitChain::DelIterator us (units);
    while (us.hasNext()) {
      Unit u (us.next ());
      switch (u.defType()) {
      case  DT_NONE:
	break;
      case DT_PRED: 
	{
	  Formula f (u.formula ());
	  while ( f.connective() == FORALL ) {
	    f = f.qarg ();
	  }
	  ASS(f.connective() == IFF);
	  ASS(f.left().connective() == ATOM);
	  const PredicateSymbol& p = f.left().atom().functor();
	  SymCounter::Pred* pcount = cnt.get (p);
	  if (pcount->docc() > 1) { 
	    break;
	  }
	  ASS (pcount->docc() == 1); 
	  if (pcount->pocc() > 0) { 
	    if (pcount->nocc() > 0) { 
	      break;
	    }
	    Formula g (IMP, f.left(), f.right());
	    if (u.formula().connective() == FORALL) {
	      g = Formula (FORALL, u.formula().vars(), g);
	    }
	    Unit v (IR_HALF_EQUIV, g, u);
	    cnt.count (u,-1);
	    cnt.count (v,1);
	    us.replace (v);
	    removed++;
	    changed = true;
	    break;
	  }
	  if (pcount->nocc() > 0) { 
	    Formula g (IMP, f.right(), f.left());
	    if (u.formula().connective() == FORALL) {
	      g = Formula (FORALL, u.formula().vars(), g);
	    }
	    Unit v (IR_HALF_EQUIV, g, u);
	    cnt.count (u,-1);
	    cnt.count (v,1);
	    us.replace (v);
	    removed++;
	    changed = true;
	    break;
	  }
	  cnt.count (u,-1);
	  us.del ();
	  changed = true;
	  removed ++;
	  break;
	}
      case DT_FUN: 
	{
	  SymCounter::Fun* fcount;
	  if (u.unitType() == CLAUSE) {
	    Clause c (u.clause ());
	    fcount = cnt.get(c.literals().head().atom().args().head().functor());
	  }
	  else { 
	    Formula form (u.formula ());
	    while ( form.connective() == FORALL ) {
	      form = form.qarg ();
	    }
	    ASS (form.connective() == ATOM);
	    fcount = cnt.get(form.atom().args().head().functor());
	  }
	  if ( fcount->occ() > 1 ) { 
	    break;
	  }
	  ASS (fcount->occ() == 1); 
	  cnt.count (u,-1);
	  us.del ();
	  changed = true;
	  removed ++;
	  break;
	}
      }
    }
  }
  return removed;
} 
