/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string>
#include "Formula.hpp"
#include "FormulaVarIterator.hpp"
#include "Literal.hpp"
#include "Signature.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "VarCounter.hpp"
#include "XML.hpp"
using namespace Shell;
Formula::Formula (const Literal& l)
{
  TRACER("Formula::Formula (const Literal& l)");
  if (l.isPositive()) {
    _data = new AtomicData(l.atom());
    return;
  }
  Formula arg(l.atom());
  _data = new UnaryData(NOT,arg);
} 
void Formula::makeFromLiteralList (const LiteralList& from,Formula& to)
{
  TRACER("Formula::makeFromLiteralList");
  if (from.isEmpty()) {
    to = Formula(FALSE);
    return;
  }
  if (from.tail().isEmpty()) {
    to = Formula(from.head());
    return;
  }
  FormulaList fs;
  FormulaList::makeFromLiteralList(from,fs);
  to = Formula(OR,fs);
} 
Formula::Formula (Connective c, const VarList& vs, const Formula& a)
  :
  _data (new QuantifiedData(c,vs,a))
{
} 
Formula::Formula (const Atom& atom)
  :
  _data (new AtomicData(atom))
{
} 
Formula::Formula (bool value)
  :
  _data (new BooleanData(value))
{
} 
Formula::Formula (Connective con)
  :
  _data (new Data(con))
{
  ASS(con == TRUE || con == FALSE);
} 
Formula::Formula (Connective c, const Formula& arg)
  :
  _data (new UnaryData(c,arg))
{
} 
Formula::Formula (Connective c, const Formula& lhs, const Formula& rhs)
  :
  _data (new BinaryData(c,lhs,rhs))
{
} 
Formula::Formula (Connective c, const FormulaList& args)
  :
  _data (new JunctionData(c,args))
{
} 
Formula::~Formula () 
{
  if (_data) {
    _data->deref ();
  }
} 
void Formula::operator = (const Formula& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
void Formula::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;
  if (_counter == 0) {
    destroy ();
  }
} 
FormulaList::FormulaList (const Formula &hd)
  :
  Lst<Formula> (hd)
{
} 
void FormulaList::makeFromLiteralList (const LiteralList& ls,FormulaList& fs)
{
  TRACER("FormulaList::makeFromLiteralList");
  Stack<Formula> stack(fs);
  Iterator<Literal> lits(ls);
  while (lits.hasNext()) {
    Formula f(lits.next());
    stack.push(f);
  }
} 
FormulaList::FormulaList (const Formula &hd, const FormulaList& tl)
  :
  Lst<Formula> (hd,tl)
{
} 
FormulaList::FormulaList (LstData<Formula>* d)
  :
  Lst<Formula> (d)
{
} 
FormulaList::FormulaList (const FormulaList& ts)
  :
  Lst<Formula> (ts)
{
} 
Formula::QuantifiedData::QuantifiedData (Connective c, 
					 const VarList& vs, 
					 const Formula& arg)
  :
  Data (c),
  _vars (vs),
  _arg (arg)
{ 
  ASS ( c == FORALL || c == EXISTS );
} 
Formula::JunctionData::JunctionData (Connective c, const FormulaList& args) 
  :
  Data (c),
  _args (args)
{ 
  ASS ( c == AND || c == OR );
} 
Formula::BinaryData::BinaryData (Connective c, 
				 const Formula& lhs, 
				 const Formula& rhs)
  :
  Data (c),
  _lhs (lhs),
  _rhs (rhs)
{ 
  ASS ( c == IMP || c == IFF || c == XOR );
} 
Formula::AtomicData::AtomicData (const Atom& atom) 
  :
  Data (ATOM),
  _atom (atom)
{ 
} 
Formula::BooleanData::BooleanData (bool value) 
  :
  Data (value ? TRUE : FALSE)
{ 
} 
Formula::UnaryData::UnaryData (Connective c, 
			       const Formula& arg) 
  :
  Data (c),
  _arg (arg)
{ 
  ASS ( c == NOT );
} 
void Formula::makeNull ()
{
  if (_data) {
    _data->deref ();
    _data = 0;
  }
} 
inline
void Formula::Data::destroy () 
{ 
  TRACER ("Formula::Data::destroy");
  ASS (this);
  ASS (_counter == 0);
  switch ( connective() ) {
  case ATOM:
    delete static_cast<AtomicData*>(this);
    return;
  case AND:
  case OR:
    delete static_cast<JunctionData*>(this);
    return;
  case IMP:
  case IFF:
  case XOR:
    delete static_cast<BinaryData*>(this);
    return;
  case NOT:
    delete static_cast<UnaryData*>(this);
    return;
  case FORALL:
  case EXISTS:
    delete static_cast<QuantifiedData*>(this);
    return;
  case TRUE:
  case FALSE:
    delete static_cast<BooleanData*>(this);
    return;
  default:
    ASS( false );
    return;
  }
} 
void Formula::makeJunction (Connective c, Formula& lhs, Formula& rhs)
{
  if (lhs.connective() == c) {
    FormulaList largs (lhs.args());
    if (rhs.connective() == c) {
      largs.append (rhs.args());
    }
    else { 
      largs.append (FormulaList(rhs));
    }
    *this = Formula (c, largs);
    return;
  }
  if (rhs.connective() == c) {
    *this = Formula (c, FormulaList(lhs,rhs.args()));
    return;
  }
  *this = Formula (c, 
		   FormulaList (lhs,
		   FormulaList (rhs)));
  return;
} 
XML::Element* Formula::toXML() const
{
  XML::DeepElement* f = new XML::DeepElement("formula");
  f->addAttribute(new XML::TextAttribute("connective",
					 _connectiveNames[connective()]));
  switch ( connective() ) {
  case ATOM:
    f->addChild(atom().toXML());
    return f;
  case AND:
  case OR:
    {
      Iterator<Formula> fs (args());
      while (fs.hasNext()) {
	f->addChild(fs.next().toXML());
      }
    }
    return f;
  case IMP:
  case IFF:
  case XOR:
    f->addChild(left().toXML());
    f->addChild(right().toXML());
    return f;
  case NOT:
    f->addChild(uarg().toXML());
    return f;
  case FORALL:
  case EXISTS:
    {
      XML::DeepElement* vs = new XML::DeepElement("variables");
      Iterator<Var> variables (vars());
      while (variables.hasNext()) {
	vs->addChild(variables.next().toXML());
      }
      f->addChild(vs);
      f->addChild(qarg().toXML());
    }
    return f;
  case TRUE:
  case FALSE:
    return f;
  default:
    ASS( false );
    return 0;
  }
} 
string Formula::toString (Connective c)
{
  static string names [] =
    { "", "&", "\\/", "=>", "<=>", "<~>", "~", "!", "?", "false", "true"};
  return names[(int)c];
} 
string Formula::toString () const
{
  TRACER("Formula::toString");
  Connective c = connective();
  string con = toString(c);
  switch (c) {
  case ATOM:
    return atom().toString();
  case AND:
  case OR: 
    {
      ASS (args().length() >= 2);
      string result = args().head().toStringInScopeOf(c);
      Iterator<Formula> arg (args().tail());
      while (arg.hasNext()) {
	result += ' ' + con + ' ' + arg.next().toStringInScopeOf(c);
      }
      return result;
    }
  case IMP:
  case IFF:
  case XOR:
    return left().toStringInScopeOf(c) + ' ' + con + ' ' +
           right().toStringInScopeOf(c);
  case NOT:
    {
      const Formula& arg = uarg();
      if (arg.connective() == ATOM) {
	const Atom& atom = arg.atom();
	if (atom.isEquality()) {
	  const TermList& ts = atom.args();
	  return ts.head().toString() + "!=" + ts.second().toString();
	}
      }
      return con + arg.toStringInScopeOf(c);
    }
  case FORALL:
  case EXISTS:
    {
      string result = "(" + con;
      Iterator<Var> vs(vars());
      while (vs.hasNext()) {
	result += string(" ") + vs.next().toString();
      }
      result += ")";
      return result + qarg().toStringInScopeOf(c);
    }
  case TRUE:
  case FALSE:
    return con;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
bool Formula::parenthesesRequired (Connective outer) const
{
  TRACER("Formula::parenthesesRequired");
  switch (connective()) 
    {
    case ATOM:
    case NOT:
    case FORALL:
    case EXISTS:
    case TRUE:
    case FALSE:
      return false;
    case OR:
    case AND:
      return outer != IMP &&
	     outer != IFF &&
             outer != XOR;
    case IMP:
      return outer != IFF &&
	     outer != XOR;
    case IFF:
    case XOR:
      return true;
    default:
      ASS( false );
      return false;
    }
} 
string Formula::toStringInScopeOf (Connective outer) const
{
  return parenthesesRequired(outer) ?
         string("(") + toString() + ")" :
         toString();
} 
bool Formula::equals (const Formula& f) const
{
  if (*this == f) {
    return true;
  }
  if (connective() != f.connective()) {
    return false;
  }
  switch (connective()) 
    {
    case ATOM:
      return atom().equals(f.atom());
    case AND:
    case OR:
      return args().equals(f.args());
    case IMP:
    case IFF:
    case XOR:
      return left().equals(f.left()) &&
             right().equals(f.right());
    case NOT:
      return uarg().equals(f.uarg());
    case FORALL:
    case EXISTS:
      return vars().equals(f.vars()) &&
             qarg().equals(f.qarg());
    case TRUE:
    case FALSE:
      return true;
#if DEBUG_SHELL
    default:
      ASSERTION_VIOLATION;
#endif
    }
}
bool FormulaList::equals (const FormulaList& fs) const
{
  if (isEmpty()) {
    return fs.isEmpty();
  }
  return fs.isNonEmpty() &&
         head().equals(fs.head()) &&
         tail().equals(fs.tail());
} 
void Formula::freeVariables (VarList& vs) const
{
  VarCounter vc;
  FormulaVarIterator fvi(*this);
  while (fvi.hasNext()) {
    Var v(fvi.next());
    if (! vc.getCount(v)) {
      vc.incCount(v);
      vs.push(v);
    }
  }
} 
