/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Formula__
#define __Formula__
#include "Atom.hpp"
#include "Connective.hpp"
namespace Shell {
class FormulaList;
class Literal;
class LiteralList;
class VarList;
class Formula 
{
 private:
  class Data;
  class BooleanData;
  class AtomicData;
  class UnaryData;
  class BinaryData;
  class QuantifiedData;
  class JunctionData;
public:
  Formula() { _data = 0; }
  explicit Formula(Connective);
  Formula(const Formula& f);
  ~Formula();
  void operator=(const Formula& rhs);
  Formula(Connective c,const Formula& l,const Formula& r);
  Formula(Connective c, const Formula& a);
  Formula(Connective c, const FormulaList& fs);
  Formula(Connective c, const VarList& vs, const Formula& a);
  explicit Formula(const Literal&);
  explicit Formula(const Atom& a);
  explicit Formula(bool value);
  void makeJunction(Connective c, Formula& lhs, Formula& rhs);
  static void makeFromLiteralList(const LiteralList& from,Formula& to);
  Connective connective () const;
  const FormulaList& args() const;
  const Formula& left() const;
  const Formula& right() const;
  const Formula& qarg() const;
  const VarList& vars() const;
  const Formula& uarg() const;
  const Atom& atom() const;
  bool operator == (const Formula& rhs) const;
  bool operator != (const Formula& rhs) const;
  bool isNull() const { return _data == 0; }
  void makeNull(); 
  void* operator new(size_t);
  bool equals(const Formula&) const;
  void freeVariables(VarList&) const;
  XML::Element* toXML () const;
  string toString() const;
  string toStringInScopeOf(Connective con) const;
  static string toString(Connective con);
  bool parenthesesRequired(Connective outer) const;
private:
  Data* _data;
  static string _connectiveNames[];
}; 
class FormulaList 
  : public Lst <Formula>
{
 public:
  FormulaList () : Lst<Formula>() {}
  FormulaList (const FormulaList&);
  explicit FormulaList (const Formula& t); 
  FormulaList (const Formula& head, const FormulaList& tail);
  explicit FormulaList (LstData<Formula>*);
  static void makeFromLiteralList (const LiteralList& from,FormulaList& to);
  const FormulaList& tail () const
    { return static_cast<const FormulaList&>(Lst<Formula>::tail()); }
  bool equals (const FormulaList&) const;
}; 
class Formula::Data 
{
 public:
  Data (Connective c);
  ~Data () { ASS (_counter == 0); }
  void ref () { ASS (this); _counter++; }
  void deref ();
  Connective connective () const { return _connective; }
  CLASS_NAME("Formula::Data");
  USE_ALLOCATOR(Data);
 protected:
  unsigned int _counter : 28;
  Connective _connective : 4;
  void destroy ();
}; 
class Formula::BooleanData :
  public Formula::Data
{
 public:
  explicit BooleanData(bool);
  CLASS_NAME("Formula::BooleanData");
  USE_ALLOCATOR(BooleanData);
}; 
class Formula::AtomicData :
  public Formula::Data
{
 public:
  explicit AtomicData (const Atom& atom);
  const Atom& atom () const { return _atom; }
  CLASS_NAME("Formula::AtomicData");
  USE_ALLOCATOR(AtomicData);
 protected:
  Atom _atom;
}; 
class Formula::UnaryData :
  public Formula::Data
{
 public:
  UnaryData (Connective c, const Formula& arg);
  const Formula& arg () const { return _arg; }
  CLASS_NAME("Formula::UnaryData");
  USE_ALLOCATOR(UnaryData);
 protected:
  Formula _arg;
}; 
class Formula::BinaryData :
  public Formula::Data
{
 public:
  BinaryData (Connective c, const Formula& lhs, const Formula& rhs);
  const Formula& left () const { return _lhs; }
  const Formula& right () const { return _rhs; }
  CLASS_NAME("Formula::BinaryData");
  USE_ALLOCATOR(BinaryData);
 protected:
  Formula _lhs;
  Formula _rhs;
}; 
class Formula::JunctionData :
  public Formula::Data
{
 public:
  JunctionData (Connective c, const FormulaList& lhs);
  const FormulaList& args () const { return _args; }
  CLASS_NAME("Formula::JunctionData");
  USE_ALLOCATOR(JunctionData);
 protected:
  FormulaList _args;
}; 
class Formula::QuantifiedData :
  public Formula::Data
{
 public:
  QuantifiedData (Connective c, const VarList& vs, const Formula& arg);
  const Formula& arg () const { return _arg; }
  const VarList& vars () const { return _vars; }
  CLASS_NAME("Formula::QuantifiedData");
  USE_ALLOCATOR(QuantifiedData);
 protected:
  VarList _vars;
  Formula _arg;
}; 
inline
Formula::Formula (const Formula& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 
inline
bool Formula::operator == (const Formula& rhs) const
{ 
  ASS (_data && rhs._data);
  return _data == rhs._data; 
} 
inline
bool Formula::operator != (const Formula& rhs) const
{ 
  ASS (_data && rhs._data);
  return _data != rhs._data; 
} 
inline
Connective Formula::connective () const
{
  return _data->connective();
} 
inline
const Atom& Formula::atom () const
{
  ASS (connective() == ATOM);
  return (static_cast<AtomicData*>(_data))->atom();
} 
inline
const FormulaList& Formula::args () const
{
  ASS (connective() == AND || connective() == OR);
  return (static_cast<JunctionData*>(_data))->args();
} 
inline
const Formula& Formula::left () const
{
  ASS (connective() == IFF || 
       connective() == XOR || 
       connective() == IMP);
  return (static_cast<BinaryData*>(_data))->left();
} 
inline
const Formula& Formula::right () const
{
  ASS (connective() == IFF || 
       connective() == XOR || 
       connective() == IMP);
  return (static_cast<BinaryData*>(_data))->right();
} 
inline
const Formula& Formula::uarg () const
{
  ASS (connective() == NOT);
  return (static_cast<UnaryData*>(_data))->arg();
} 
inline
const Formula& Formula::qarg () const
{
  ASS (connective() == FORALL || connective() == EXISTS);
  return (static_cast<QuantifiedData*>(_data))->arg();
} 
inline
const VarList& Formula::vars () const
{
  ASS (connective() == FORALL || connective() == EXISTS);
  return (static_cast<QuantifiedData*>(_data))->vars();
} 
inline
Formula::Data::Data (Connective c)
  : 
  _counter (1),
  _connective (c)
{
} 
}
#endif 
