/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "Lexer.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
Lexer::Lexer (istream& in)
  :
  _charCursor(0),
  _stream(in),
  _eof(false),
  _lineNumber(1)
{
  readNextChar();
} 
bool Lexer::readNextChar ()
{
  TRACER("Lexer::readNextChar");
  if (_eof) {
    return false;
  }
  _lastCharacter = _stream.get();
  if (_lastCharacter == -1) {
    _eof = true;
    return false;
  }
  if (_lastCharacter == '\n') {
    _lineNumber++;
  }
  return true;
} 
void Lexer::readNumber (Token& token)
{
  TRACER("Lexer::readNumber");
  readUnsignedInteger();
  if (_lastCharacter == '.') {
    saveLastChar();
    if (readNextChar()) {
      if (isDigit(_lastCharacter)) {
	readUnsignedInteger();
	token.tag = TT_REAL;
	saveTokenText(token);
	return;
      }
    }
    saveTokenText(token);
    throw LexerException((string)"incorrect number format in " + token.text,
			 *this);
  }
  token.tag = TT_INTEGER;
  saveTokenText(token);
} 
void Lexer::saveLastChar ()
{
  TRACER("Lexer::saveLastChar");
  if (_charCursor >= CHARACTER_BUFFER_SIZE) {
    throw LexerException("token size exceeded",*this);
  }
  _charBuffer[_charCursor++]  = (char)_lastCharacter;
} 
void Lexer::saveChar (int character)
{
  TRACER("Lexer::saveChar");
  if (_charCursor >= CHARACTER_BUFFER_SIZE) {
    throw LexerException("token size exceeded",*this);
  }
  _charBuffer[_charCursor++]  = (char)character;
} 
void Lexer::saveTokenText (Token& token)
{
  TRACER("Lexer::saveTokenText");
  ASS (_charCursor <= CHARACTER_BUFFER_SIZE);
  _charBuffer[_charCursor] = 0;
  token.text = _charBuffer;
} 
void Lexer::readUnsignedInteger ()
{
  TRACER("TPTPLexer::readUnsignedInteger");
  saveLastChar();
  while (readNextChar() && isDigit(_lastCharacter)) {
      saveLastChar();
  }
} 
LexerException::LexerException (string message,const Lexer& lexer)
  : _message (message)
{
  if (lexer.isAtEndOfFile()) {
    _message += " at end of input";
    return;
  }
  int line = lexer.lineNumber();
  if (lexer.lastCharacter() == '\n') {
    line--;
  }
  _message += " in line ";
  _message += Int::toString(line);
} 
void LexerException::cry (ostream& out)
{
  out << "Lexer exception: " << _message << '\n';
} 
XML::Element* LexerException::toXML () const
{
  return new XML::TextElement("lexer_error",_message);
} 
void Lexer::readSequence (const char* cs)
{
  while (*cs) {
    readNextChar();
    if (lastCharacter() != *cs) {
      throw LexerException((string)cs + 
			   " expected",*this);
    }
    cs++;
  }
  readNextChar();
} 
