/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __list__
#  define __list__
#include "Allocator.hpp"
namespace Shell {
template <class C>
class List
{
 public:
  inline
  List ( C head )
    :
    _head ( head ),
    _tail ( 0 )
  {
  } 
  inline
  List ( C head, List* tail )
    :
    _head ( head ),
    _tail ( tail )
  {
  } 
  inline
  List () {}
  inline
  static List* empty ()  
  {
    return 0;
  }
  inline
  bool isEmpty () const
  {
    return this == 0;
  } 
  inline
  bool isNonEmpty () const
  {
    return this != 0;
  } 
  inline
  List* tail () const
  {
    ASS ( this != 0 );
    return _tail;
  } 
  inline
  C head () const
  {
    ASS ( this != 0 );
    return _head;
  } 
  inline
  C second () const
  {
    ASS ( this != 0 && _tail != 0 );
    return _tail->_head;
  } 
  inline void setHead (C hd)
  {
    ASS ( this != 0 );
    _head = hd;
  } 
  inline void setTail (List* tl)
  {
    ASS ( this != 0 );
    _tail = tl;
  } 
  void destroy ()
  {
    if (isEmpty ()) {
      return;
    }
    List* current = this;
    for (;;) {
      List* next = current->tail ();
      delete current;
      if (next->isEmpty ())
        return;
      current = next;
    }
  } 
  List* copy () const
  {
    if ( isEmpty () )
      return empty ();
    List* result = new List;
    result->head ( head () );
    List* previous = result;
    List* rest = tail ();
    while ( ! rest->isEmpty () ) {
      List* tmp = new List;
      tmp->head ( rest->head () );
      previous->tail ( tmp );
      previous = tmp;
      rest = rest->tail ();
    }
    previous->tail ( empty () );
    return result;
  }  
  List* append (List* snd) 
  {
    if ( isEmpty () )
      return snd;
    List* result = new List;
    result->head ( head () );
    List* previous = result;
    List* rest = tail ();
    while ( ! rest->isEmpty () ) {
      List* tmp = new List;
      tmp->head ( rest->head () );
      previous->tail ( tmp );
      previous = tmp;
      rest = rest->tail ();
    }
    previous->tail ( snd );
    return result;
  }  
  inline
  List* cons ( C elem )
  {
    return new List ( elem, this );
  } 
  inline
  static void push ( C elem, List* &lst )
  {
    lst = lst->cons ( elem );
  } 
  inline
  static C pop ( List* &lst )
  {
    ASS ( lst != 0 );
    List* tail = lst->tail ();
    C result = lst->head ();
    delete lst;
    lst = tail;
    return result;
  } 
  inline 
  static void move ( List*& from, List*& to )
  {
    List* tail = from->tail ();
    from->tail ( to );
    to = from;
    from = tail;
  } 
  static List* concat( List* first, List* second )
  {
    if ( first == 0 )
      return second;
    if ( second == 0 )
      return first;
    List* current = first;
    for (;;) {
      List* next = current->tail ();
      if ( ! next ) {
        current->tail ( second );
        return first;
      }
      current = next;
      next = next->tail ();
    }
  } 
  List* reverse ()
  {
    if ( isEmpty () )
      return empty ();
    List* result = empty ();
    List* lst = this;
    while ( ! lst->isEmpty () ) {
      List* tl = lst->tail ();
      lst->tail ( result );
      result = lst;
      lst = tl;
    }
    return result;
  } 
  int length () const
  {
    int len = 0;
    for ( const List* lst = this; ! lst->isEmpty () ; lst = lst->tail() )
      len ++;
    return len;
  } 
  bool member ( C elem )
  {
    for ( List* lst = this; ! lst->isEmpty () ; lst = lst->tail() ) {
      if ( lst->head() == elem )
        return true;
    }
    return false;
  } 
  List* remove ( C elem )         
  {
    if ( isEmpty() )
      return this;
    if ( head() == elem ) {
      List* result = tail ();
      delete this;
      return result;
    }
    if ( tail()->isEmpty() )
      return this;
    List* current = this;
    List* next = tail ();
    for (;;) {
      if ( next->head() == elem ) { 
        current->tail ( next->tail() );
        delete next;
        return this;
      }
      current = next;
      next = next->tail();
      if ( next->isEmpty() )
        return this;
    }
  } 
  C nth ( int N ) const 
  {     
    ASS ( N >= 0 );
    const List* l = this;
    while ( N != 0 ) {
      ASS ( l->isNonEmpty() );
      l = l->tail ();
      N-- ;
    }
    return l->head();
  } 
  static C deleteNth ( List*& lst, int N ) 
  {     
    ASS ( N >= 0 );
    C result;
    List* l = lst;
    ASS ( lst->isNonEmpty() );
    if ( N == 0 ) {
      result = l->head();
      lst = l->tail();
      delete l;
      return result;
    }
    List* next = l->tail();
    while ( --N != 0 ) {
      l = next;
      next = next->tail ();
      ASS ( next->isNonEmpty() );
    }
    result = next->head();
    l->tail (next->tail());
    delete next;
    return result;
  } 
  List* addLast ( C elem )
  {
    if ( ! this ) 
      return new List (elem);
    List* current;
    for ( current = this; current->_tail; current = current->_tail )
      ;
    current->tail ( new List (elem) );
    return this;
  } 
  List* split ( int n, List*& rest )
  {
    if ( ! this ) {
      ASS( n == 0 );
      rest = empty ();
      return empty ();
    }
    if ( n == 0 ) {
      rest = empty ();
      return this;
    }
    List* nth = this;
    while ( --n > 0 ) {
      ASS( nth );
      nth = nth->_tail;
    }
    ASS( nth );
    rest = nth->_tail;
    nth->_tail = empty ();
    return this;
  } 
  class Iterator {
   public:
    inline explicit 
    Iterator ( List* l ) : _lst (l) {}  
    inline explicit 
    Iterator ( const List* l ) : _lst (const_cast<List*>(l)) {}  
    inline 
    C next () 
      { 
        C result = _lst->head(); 
        _lst = _lst->tail();
        return result;
      }
    inline bool hasNext () const 
    { 
      return ! _lst->isEmpty(); 
    }
    inline
    void reset ( List* l ) { _lst = l; }
   private:
    List* _lst;
  };
  class DelIterator {
   public:
    inline 
    DelIterator ( List*& l ) : 
      _lst ( l ),
      #if DEBUG_SHELL
        _prev ( 0 ), 
      #endif
      _cur ( 0 )
      {}  
    inline 
    C next () 
      { 
        if ( _cur ) { 
          _prev = _cur;
          _cur = _cur->tail();
          ASS( _cur );
        }
        else { 
          _cur = _lst;
        }
        return _cur->head();
      }
    inline bool hasNext () 
      { 
        if ( _cur ) { 
          return _cur->tail() != 0;
        }
        return ! _lst->isEmpty(); 
      }
    void del () 
      {
        ASS( _cur ); 
        ASS( _cur != _prev ); 
        if ( _cur == _lst ) { 
          _lst = _lst->tail ();
          delete _cur;
          _cur = 0;
          return;
        }
        _prev->tail (_cur->tail());
        delete _cur;
        _cur = _prev;
      }
    void restore ( C elem ) 
      {
        if ( _cur ) { 
          _cur = new List (elem,_cur->tail());
          _prev->tail (_cur);
          return;
        }
        push (elem, _lst);
        _cur = _lst;
      }
   private:
    List*& _lst;
    List* _prev;
    List* _cur; 
  };
  CLASS_NAME("List");
  USE_ALLOCATOR(List);
 protected:  
  C _head;
  List* _tail;
};  
}
#endif
