/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Lst__
#define __Lst__
#include "Allocator.hpp"
namespace Shell {
template <class C> class LstData;
template <class C>
class Lst 
{
 public:
  Lst ();
  Lst (const Lst&);
  Lst (const C& head, const Lst& tail);
  explicit Lst (const C& t); 
  explicit Lst (LstData<C>*);
  ~Lst ();
  void operator = (const Lst& rhs);
  void* operator new (size_t);
  void append (const Lst&);
  bool isEmpty () const;
  bool isNonEmpty () const;
  const C& head () const;
  const C& second () const;
  const C& third () const;
  const Lst& tail () const;
  void setTail (Lst&);
  void pop ();
  bool operator == (const Lst& rhs) const;
  bool operator != (const Lst& rhs) const;
  void makeEmpty ();
  int length () const;
  bool hasLength (int length) const;
  bool member (const C& elem) const;
  void push (const C& elem);
  void reverse (Lst& result) const;
  void reverse (Lst& result, const Lst& appendTo) const;
  void copy (Lst& to) const;
  Lst& last() const;
 protected:
  LstData<C>* _data;
}; 
template <class C>
class LstData 
{
 public:
  LstData ();
  inline LstData (const C& t, const Lst<C>& ts)
    : _counter (1),
      _head (t),
      _tail (ts)
  {
  } 
  inline explicit LstData (const C& t)
    : _counter (1),
      _head (t),
      _tail ((LstData<C>*)0)
  {
  } 
  inline ~LstData ()
  {
    ASS (_counter == 0);
  } 
  inline void ref ()
    { 
      ASS (this);
      _counter++;
    } 
  inline void deref ()
  { 
    ASS (this);
    ASS (_counter > 0);
    _counter--;
    if (_counter == 0) {
      delete this;
    }
  } 
  inline const C& head () const
  { 
    ASS (this);
    ASS (_counter > 0);
    return _head; 
  } 
  inline const Lst<C>& tail () const
  { 
    ASS (this);
    ASS (_counter > 0);
    return _tail;
  } 
  inline void setTail (Lst<C>& tl)
  {
    _tail = tl;
  }
  CLASS_NAME("Lst");
  USE_ALLOCATOR(LstData);
 private:
  unsigned _counter;
  C _head;
  Lst<C> _tail;
}; 
template <class C>
class Iterator
{
 public: 
  inline explicit Iterator (const Lst<C>& lst)
    : 
    _list (lst) 
  {
  } 
  inline bool hasNext () const
  { 
    return _list.isNonEmpty(); 
  } 
  inline C& next ()
  {
    ASS( hasNext() );
    C& result = (const_cast<C&>(_list.head()));
    _list.pop ();
    return result;
  } 
  inline void reset (const Lst<C>& lst)
  {
    _list = lst;
  } 
 private:
  Lst<C> _list;
}; 
template <class C>
class Stack
{
 public: 
  inline explicit Stack (Lst<C>& lst)
    : 
    _list (&lst) 
  {
    ASS(lst.isEmpty());
  } 
  inline void push (const C& element)
  {
    _list->push(element);
    _list = const_cast<Lst<C>*>(&(_list->tail()));
  } 
 private:
  Lst<C>* _list;
}; 
template <class C>
inline
Lst<C>::Lst () 
  : 
  _data (0) 
{
} 
template <class C>
inline
Lst<C>::~Lst () 
{
  if (_data) {
    _data->deref ();
  }
} 
template <class C>
inline
Lst<C>::Lst (const Lst& ts)
  :
  _data (ts._data)
{
  if (_data) {
    _data->ref ();
  }
} 
template <class C>
inline
Lst<C>::Lst (LstData<C>* d)
  :
  _data (d)
{
  if (d) {
    d->ref ();
  }
} 
template <class C>
inline
Lst<C>::Lst (const C &hd, const Lst& tl)
  :
  _data (new LstData<C> (hd,tl))
{
} 
template <class C>
inline
Lst<C>::Lst (const C &hd)
  :
  _data (new LstData<C> (hd))
{
} 
template <class C>
inline
bool Lst<C>::operator == (const Lst& rhs) const
{ 
  return _data == rhs._data; 
} 
template <class C>
inline
bool Lst<C>::operator != (const Lst& rhs) const
{ 
  return _data != rhs._data; 
} 
template <class C>
inline
bool Lst<C>::isEmpty () const 
{ 
  return _data == 0; 
} 
template <class C>
inline
bool Lst<C>::isNonEmpty () const 
{ 
  return _data != 0; 
} 
template <class C>
inline
const C& Lst<C>::head () const 
{ 
  return _data->head (); 
} 
template <class C>
inline
const C& Lst<C>::second () const 
{ 
  return _data->tail().head(); 
}
template <class C>
inline
const C& Lst<C>::third () const 
{ 
  return _data->tail().tail().head(); 
}
template <class C>
inline
const Lst<C>& Lst<C>::tail () const 
{ 
  return _data->tail(); 
}
template <class C>
void Lst<C>::pop ()
{
  ASS (_data);
  LstData<C>* dtl = tail()._data;
  if (! dtl) {
    _data->deref();
    _data = 0;
    return;
  }
  dtl->ref ();
  _data->deref ();
  _data = dtl;
} 
template <class C>
bool Lst<C>::member (const C& elem) const
{
  Iterator<C> it (*this);
  while (it.hasNext()) {
    if (it.next() == elem) {
      return true;
    }
  }
  return false;
} 
template <class C>
inline
void Lst<C>::push (const C& elem)
{
  *this = Lst (elem,*this);
} 
template <class C>
inline
void Lst<C>::makeEmpty ()
{
  if (_data) {
    _data->deref ();
    _data = 0;
  }
} 
template <class C>
void Lst<C>::operator = (const Lst& t)
{
  if (t._data) {
    t._data->ref ();
  }
  if (_data) {
    _data->deref ();
  }
  _data = t._data;
} 
template <class C>
void Lst<C>::append (const Lst& lst)
{
  if (isEmpty()) {
    *this = lst;
    return;
  }
  Lst tl(tail());
  tl.append(lst);
  *this = Lst(head(),tl);
} 
template <class C>
Lst<C>& Lst<C>::last () const
{
  ASS(! isEmpty());
  if (tail().isEmpty()) {
    return const_cast<Lst&>(*this);
  }
  return tail().last();
} 
template <class C>
int Lst<C>::length () const
{
  int len = 0;
  for (LstData<C>* d = _data; d; d = d->tail()._data) {
    len ++;
  }
  return len;
} 
template <class C>
bool Lst<C>::hasLength (int l) const
{
  for (LstData<C>* d = _data; d; d = d->tail()._data) {
    l--;
    if (l < 0) {
      return false;
    }
  }
  return l == 0;
} 
template <class C>
void Lst<C>::reverse (Lst& result, const Lst& appendTo) const
{
  Lst rev (appendTo);
  Iterator<C> it (*this);
  while (it.hasNext()) {
    rev.push (it.next());
  }
  result = rev;
} 
template <class C>
inline
void Lst<C>::reverse (Lst& result) const
{
  Lst empty;
  reverse (result,empty);
} 
template <class C>
inline
void Lst<C>::copy (Lst& to) const
{
  if (isEmpty()) {
    return;
  }
  to = Lst (head());
  tail().copy (const_cast<Lst&>(to.tail()));
} 
template <class C>
inline
void Lst<C>::setTail (Lst& tl)
{
  ASS(_data);
  _data->setTail(tl);
} 
}
#endif 
