/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __Map__
#define __Map__

#include "Tracer.hpp"
#include "VampireException.hpp"

namespace Shell {

template <typename Key, typename Val,class Hash>
class Map 
{
protected:
  class Entry
  {
  public:
    inline Entry ()
      : code(0)
    {
    } // Map::Entry::Entry

    inline bool occupied () const
    {
      return code;
    } // Map::Entry::occupied

    void* operator new (size_t);

    unsigned code;
    Key key;
    Val value;
  }; // class Map::Entry

 public:
  Map ()
    : _capacity(0),
      _noOfEntries(0),
      _entries(0)
  {
    expand();
  } // Map::Map

  inline ~Map ()
  {
    delete [] _entries;
  } // Map::~Map

  bool find (Key key, Val& found) const
  {
    TRACER( "Map::find/2" );

    unsigned code = Hash::hash(key);
    if (code == 0) {
      code = 1;
    }
    Entry* entry;
    for (entry = firstEntryForCode(code);
	 entry->occupied();
	 entry = nextEntry(entry)) {
      if (entry->code == code &&
	  entry->key == key) {
	found = entry->value;
	return true;
      }
    }

    return false;
  } // Map::find
 
  inline Entry* firstEntryForCode(unsigned code) const
  {
    return _entries + (code % _capacity);
  } // Map::firstEntryForCode

  inline Entry* nextEntry(Entry* entry) const
  {
    entry ++;
    // check if the entry is a valid one
    return entry == _afterLast ? _entries : entry;
  } // nextEntry

  inline Val& insert(const Key& key,const Val& val)
  {
    TRACER("Map::insert");
    
    if (_noOfEntries >= _maxEntries) { // too many entries
      expand();
    }
    unsigned code = Hash::hash(key);
    if (code == 0) {
      code = 1;
    }
    return insert(key,val,code);
  } // Map::insert

  Val& insert(const Key& key, const Val& val,unsigned code)
  {
    TRACER("Map::insert/2");
    
    Entry* entry;
    for (entry = firstEntryForCode(code);
	 entry->occupied();
	 entry = nextEntry(entry)) {
      if (entry->code == code &&
	  entry->key == key) {
	return entry->value;
      }
    }
    // entry is not occupied
    _noOfEntries++;
    entry->key = key;
    entry->value = val;
    entry->code = code;
    return entry->value;
  } // Map::insert

  void replaceOrInsert(const Key& key,const Val& val)
  {
    TRACER("Map::insertOrReplace");
    
    if (_noOfEntries >= _maxEntries) { // too many entries
      expand();
    }
    unsigned code = Hash::hash(key);
    if (code == 0) {
      code = 1;
    }
    Entry* entry;
    for (entry = firstEntryForCode(code);
	 entry->occupied();
	 entry = nextEntry(entry)) {
      if (entry->code == code &&
	  entry->key == key) {
	entry->value = val;
	return;
      }
    }
    // entry is not occupied
    _noOfEntries++;
    entry->key = key;
    entry->value = val;
    entry->code = code;
  } // Map::replaceOrInsert


  void replace(const Key& key,const Val& val)
  {
    TRACER("Map::replace");
    
    if (_noOfEntries >= _maxEntries) { // too many entries
      expand();
    }
    unsigned code = Hash::hash(key);
    if (code == 0) {
      code = 1;
    }
    Entry* entry;
    for (entry = firstEntryForCode(code);
	 entry->occupied();
	 entry = nextEntry(entry)) {
      if (entry->code == code &&
	  entry->key == key) {
	entry->value = val;
	return;
      }
    }
#if VDEBUG
    ASSERTION_VIOLATION;
#endif
  } // Map::replace


  void deleteAll()
  {
    for (int i = _capacity-1;i >= 0;i--) {
      Entry& e = _entries[i];
      if (e.occupied()) {
	delete e.value;
      }
    }
  } // deleteAll

  inline int numberOfElements()
  {
    return _noOfEntries;
  }

  void destroyAll()
  {
    for (int i = _capacity-1;i >= 0;i--) {
      Entry& e = _entries[i];
      if (e.occupied()) {
	e.value->destroy();
      }
    }
  } // destroyAll

  int _capacity;
  int _noOfEntries;
  Entry* _entries;
  Entry* _afterLast; // entry after the last one
  int _maxEntries;

  void expand()
  {
    TRACER("Map::expand");

    _capacity = _capacity ? _capacity * 2 : 32;

    Entry* oldEntries = _entries;
    _entries = new Entry [_capacity];

    _afterLast = _entries + _capacity;
    _maxEntries = (int)(_capacity * 0.8);
    // experiments using (a) random numbers (b) consecutive numbers
    // and (1) int->int 20M allocations (2) string->int 10M allocations
    // and 30,000,000 allocations
    // 0.6 : 6.80 4.87 20.8 14.9 32.6 14
    // 0.7 : 6.58 5.61 23.1 15.2 35.2 16.6
    // 0.8 : 6.36 5.77 24.0 15.4 36.0 17.4
    // 0.9 : 7.54 6.04 25.4 15.2 37.0 18.4
    // copy old entries
    Entry* current = oldEntries;
    int remaining = _noOfEntries;
    _noOfEntries = 0;
    while (remaining != 0) {
      // find first occupied entry
      while (! current->occupied()) {
	current ++;
      }
      // now current is occupied
      insert(current->key,current->value,current->code);
      current ++;
      remaining --;
    }

    delete [] oldEntries;
  } // Map::expand

public:
  class Iterator {
  public:
    inline Iterator(const Map& map)
      : _next(map._entries),
	_last(map._afterLast)
    {
    } // Map::Iterator

    bool hasNext()
    {
      while (_next != _last) {
	if (_next->occupied()) {
	  return true;
	}
	_next++;
      }
      return false;
    }

    Val& next()
    {
      ASS(_next != _last);
      ASS(_next->occupied());
      Val& result = _next->value;
      _next++;
      return result;
    }
  private:
    /** iterator will look for the next occupied cell starting with this one */
    Entry* _next;
    /** iterator will stop looking for the next cell after reaching this one */
    Entry* _last;
  };
}; // class Map


}
#endif 
