/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Tracer.hpp"
#include "Normalisation.hpp"
#include "Unit.hpp"
#include "Sort.hpp"
#include "Problem.hpp"
#include "Signature.hpp"
#include "Compare.hpp"
using namespace Shell;
void Normalisation::normalise (Unit& u)
{
  switch (u.unitType()) {
  case FORMULA:
    {
      Formula f(u.formula());
      normalise(f);
      if (! f.equals(u.formula())) {
	u = Unit(IR_NORMALIZE, f, u);
      }
      return;
    }
  case CLAUSE:
    {
      Clause c(u.clause());
      normalise(c);
      if (! c.equals(u.clause())) {
	u = Unit(IR_NORMALIZE, c, u);
      }
      return;
    }
  }
} 
void Normalisation::normalise (Clause& c)
{
  TRACER ("Clause::normalize");
  int length = c.literals().length();
  Compare comparator;  
  Sort <Literal,Compare> srt (length,comparator);
  Iterator<Literal> lits (c.literals());
  while (lits.hasNext()) {
    Literal l (lits.next());
    normalise(l);
    srt.add (l);
  }
  srt.sort();
  LiteralList sorted;
  for (int i = length-1; i >=0; i--) {
    sorted.push (srt[i]);
  }
  c = Clause(sorted);
} 
void Normalisation::normalise (Literal& l)
{ 
  Atom a (l.atom());
  normalise(a);
  if(! (a == l.atom())) {
    l = Literal(l.sign(),a);
  }
} 
void Normalisation::normalise (Atom& a)
{
  if (a.isEquality()) {
    TermList as (a.args());
    Term l (as.head());
    Term r (as.second());
    if (Compare::compare(l,r) == GREATER) {
      a = Atom(a.functor(), 
	       TermList(r,TermList (l)));
    }
  }
} 
void Normalisation::normalise (Formula& f)
{
  switch (f.connective()) {
  case ATOM: 
    {
      Atom a (f.atom());
      normalise(a);
      if (! (a == f.atom())) {
	f = Formula (a);
      }
      return;
    }
  case AND:
  case OR: 
    {
      int len = f.args().length();
      Compare comparator;
      Sort<Formula,Compare> sort(len,comparator);
      Iterator<Formula> as (f.args());
      while (as.hasNext()) {
	Formula f (as.next());
	normalise(f);
	sort.add (f);
      }
      sort.sort();
      FormulaList fs;
      for (int i = len-1; i >= 0; i--) {
	fs.push (sort[i]);
      }
      f = Formula (f.connective(), fs);
      return;
    }
  case IMP: 
    {
      Formula lft (f.left());
      Formula rht (f.right());
      normalise(lft);
      normalise(rht);
      if (!(f.left() == lft && f.right() == rht)) {
	f = Formula(IMP,lft,rht);
      }
      return;
    }
  case IFF:
  case XOR: 
    {
      Formula lft (f.left());
      Formula rht (f.right());
      normalise(lft);
      normalise(rht);
      Compare cmp;
      if (cmp.lessThan(lft,rht)) {
	if (!(f.left() == lft && f.right() == rht)) {
	  f = Formula(f.connective(),lft,rht);
	}
	return;
      }
      f = Formula(f.connective(),rht,lft);
      return;
    }
  case NOT:
    {
      Formula arg (f.uarg());
      normalise(arg);
      if (!(f.uarg() == arg)) {
	f = Formula(NOT, arg);
      }
      return;
    }
  case FORALL:
  case EXISTS:
    {
      Formula arg (f.qarg());
      normalise(arg);
      if (!(f.qarg() == arg)) {
	f = Formula(f.connective(), f.vars(), arg);
      }
      return;
    }
  case FALSE:
  case TRUE:
    return;
#if DEBUG_SHELL  
  default:
    ASS(false);
#endif
  }
} 
