/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "CNF.hpp"
#include "ClausalDefinition.hpp"
#include "Compare.hpp"
#include "Definition.hpp"
#include "Environment.hpp"
// #include "EqualityAxioms.hpp"
#include "EqualityProxy.hpp"
#include "FunctionDefinition.hpp"
#include "Flattening.hpp"
#include "LiteralOrderingFinder.hpp"
#include "Miniscope.hpp"
#include "Naming.hpp"
#include "NNF.hpp"
#include "Normalisation.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "TheoryFinder.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "PureLiteral.hpp"
#include "Rectify.hpp"
#include "Signature.hpp"
#include "SimplifyFalseTrue.hpp"
#include "Skolem.hpp"
#include "Sort.hpp"
#include "SymCounter.hpp"
#include "Tautology.hpp"
#include "TPTP.hpp"
#include "Tracer.hpp"
#include "XML.hpp"
using namespace Shell;
Preprocess::Preprocess (Problem& problem, 
			const Options& options,
			const Property& property)
  : _problem(problem),
    _options(options),
    _property(property)
{
} 
void Preprocess::preprocess ()
{
  TRACER("Preprocess::preprocess");
  if (_options.normalize()) {
    Compare::createSymbolCounter(_problem);
  }
  UnitChain::DelIterator us (_problem.giveUnits());
  while (us.hasNext()) {
    Unit u (us.next());
    Unit v(u);
    preprocess1(v);
    if (v != u) {
      us.replace(v);
    }
  }
  if (_options.normalize()) {
    normalise();
  }
  switch (Environment::mode) 
    {
    case MODE_KIF: 
      break;
    default:
//       removeEqualityAxioms();
      if (_problem.hasFormulas() && 
 	  _options.unusedPredicateDefinitionRemoval()) {
	Definition::removeUnusedPredicateDefinitions(_problem);
      }
      break;
    }
  UnitChain::DelIterator us2 (_problem.giveUnits());
  while (us2.hasNext()) {
    Unit u (us2.next());
    Unit v (u);
    preprocess2(v);
    if (v != u) {
      us2.replace(v);
    }
  }
  if (_options.naming()) {
    Naming naming(_options.naming(),Signature::sig);
    UnitChain::DelIterator nit (_problem.giveUnits());
    while (nit.hasNext()) {
      Unit u(nit.next());
      Unit v(u);
      naming.apply(v);
      if (v != u) {
	nit.replace(v);
      }
    }
    UnitList defs(naming.getDefinitions());
    Iterator<Unit> dfs(defs);
    while (dfs.hasNext()) {
      _problem.addUnit(dfs.next());
    }
  }
  UnitChain::DelIterator us3 (_problem.giveUnits());
  while (us3.hasNext()) {
    Unit u(us3.next());
    Unit v(u);
    preprocess3(v);
    if (v != u) {
      us3.replace(v);
    }
  }
  LiteralOrderingFinder lof(_problem,_options,Signature::sig);
  lof.findOrdering();
  CNF::clausify (_problem);
   if (_options.functionDefinitionElimination() == Options::FDE_ALL &&
       _property.hasProp(PR_HAS_FUNCTION_DEFINITIONS)) {
    FunctionDefinition fd(_problem,Signature::sig);
    fd.removeAllDefinitions();
  }
  UnitChain::DelIterator units (_problem.giveUnits());
  while (units.hasNext()) {
    Unit newUnit;
    switch (Tautology::isTautology(units.next(),newUnit)) {
    case -1:
      units.replace(newUnit);
      break;
    case 0:
      break;
    case 1:
      units.del();
      break;
    }
  }
  PureLiteral pure(_problem,Signature::sig);
  pure.findAndRemove();
  ClausalDefinition cdef(_problem,Signature::sig);
  cdef.findAndReplace();
  if (_property.hasProp(PR_HAS_X_EQUALS_Y)) {
    switch (_options.equalityProxy()) {
    case Options::EP_OFF:
      break;
    case Options::EP_ON:
    case Options::EP_EXP1:
      {
	EqualityProxy proxy(Signature::sig);
	proxy.apply(_problem);
      }
    }
  }
} 
void Preprocess::preprocess1 (Unit& u)
{
  TRACER("Preprocess::preprocess1");
  switch (u.unitType()) {
  case CLAUSE:
    if (_options.normalize()) {
      Normalisation::normalise(u);
    }
    return;
  case FORMULA:
    {
      Rectify::rectify(u);
      SimplifyFalseTrue::simplify(u);
      Flattening::flatten(u);
      if (_options.normalize()) {
	Normalisation::normalise(u);
      }
      return;
    }
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
void Preprocess::preprocess2 (Unit& unit)
{
  TRACER("Preprocess::preprocess2");
  switch (unit.unitType()) {
  case CLAUSE:
    return;
  case FORMULA:
    NNF::ennf(unit);
    Flattening::flatten(unit);
    return;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
void Preprocess::preprocess3 (Unit& unit)
{
  TRACER("Preprocess::preprocess3");
  switch (unit.unitType()) {
  case CLAUSE:
    return;
  case FORMULA:
    NNF::nnf(unit);
    Flattening::flatten(unit);
    if (_options.miniscope()) {
      Miniscope::miniscope(unit);
    }
    Skolem::skolemise(unit);
    return;
#if DEBUG_SHELL
  default:
    ASS(false);
#endif
  }
} 
void Preprocess::normalise ()
{
  TRACER ( "Preprocess::normalise" );
  UnitChain& units = _problem.giveUnits();
  Compare cmp;
  Sort <Unit,Compare> sort (units.length(),cmp);
  UnitChain::DelIterator unit1 (units);
  while (unit1.hasNext()) {
    sort.add(unit1.next());
    unit1.del();
  }
  sort.sort ();
  for (int i = sort.length()-1;i >= 0;i--) {
    units.addFirst(sort[i]);
  }
} 
// int Preprocess::removeEqualityAxioms ()
// {
//   TRACER("Preprocess::removeEqualityAxioms");
//   int removed = 0;
//   UnitChain::DelIterator us (_problem.giveUnits());
//   while (us.hasNext()) {
//     if (EqualityAxioms::isEqualityAxiom(us.next())) {
//       us.del();
//       removed++;
//     }
//   }
//   return removed;
// } 
