/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Int.hpp"
#include "Hash.hpp"
#include "ProofChecker.hpp"
#include "Refutation.hpp"
#include "Inference.hpp"
#include "Map.hpp"
#include "Matching.hpp"
#include "XML.hpp"
#include "VarCounter.hpp"
#include "Substitution.hpp"
#include "TermVarIterator.hpp"
#include "AtomIterator.hpp"
#include "FormulaVarIterator.hpp"
#include "SubformulaIterator.hpp"
#include "SubtermIterator.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
using namespace Shell;
void ProofChecker::check (const Refutation& ref)
{
  TRACER("ProofChecker::check");
  Unit lastUnit;
  if (ref.refutation().isEmpty()) {
    PROOF_CHECKING_ERR("refutation is empty");
  }
  ProofChecker checker;
  Map<long,Unit,Hash> checked;
  Iterator<Unit> units (ref.refutation());
  while (units.hasNext()) {
    Unit conclusion(units.next());
    int number = conclusion.number();
    Iterator<Unit> premises (conclusion.inference()->premises());
    while (premises.hasNext()) {
      Unit premise(premises.next());
      Unit previous;
      if (! checked.find(premise.number(),previous)) {
	PROOF_CHECKING_ERR(string("premise ") + 
			   Int::toString(premise.number()) +
			   " of unit " +
			   Int::toString(conclusion.number()) +
			   " is not derived");
      }
      if (!(premise == previous)) {
	PROOF_CHECKING_ERR(string("premise ") + 
			   Int::toString(premise.number()) +
			   " of unit " +
			   Int::toString(conclusion.number()) +
			   " is not derived");
      }
    }
    checker.checkInference(conclusion);
    checked.insert(number,conclusion);
    lastUnit = conclusion;
  }
  if (! isGoal(lastUnit)) {
    PROOF_CHECKING_ERR(string("last unit ") + 
		       Int::toString(lastUnit.number()) +
		       " of refutation is not a goal");
  }
} 
void ProofChecker::checkInference (const Unit& conclusion)
{
  TRACER("ProofChecker::checkInference");
  Inference* inference = conclusion.inference();
  Formula premise;
  Formula premise2;
  Position position;
  switch (inference->rule()) {
  case IR_CLAUSIFY:
    if (! hasSinglePremise(inference)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    checkClausify(conclusion, inference->premises().head());
    return;
  case IR_KERNEL: 
    return;
  case IR_INPUT:
    checkInput(conclusion);
    return;
  case IR_RECTIFY:
    isFormulaInference(conclusion,premise,position);
    checkRectify(conclusion, premise);
    return;
  case IR_CLOSURE:
    isFormulaInference(conclusion,premise,position);
    checkClosure(conclusion, premise);
    return;
  case IR_FLATTEN:
    isFormulaInference(conclusion,premise,position);
    checkFlattening(conclusion,premise,position);
    return;
  case IR_NOT_AND:
    isFormulaInference(conclusion,premise,position);
    checkNotAnd(conclusion,premise,position,true);
    return;
  case IR_NOT_OR:
    isFormulaInference(conclusion,premise,position);
    checkNotAnd(conclusion,premise,position,false);
    return;
  case IR_NOT_IMP:
    isFormulaInference(conclusion,premise,position);
    checkNotImp(conclusion,premise,position);
    return;
  case IR_NOT_IFF:
    isFormulaInference(conclusion,premise,position);
    checkNotIff(conclusion,premise,position,true);
    return;
  case IR_NOT_XOR:
    isFormulaInference(conclusion,premise,position);
    checkNotIff(conclusion,premise,position,false);
    return;
  case IR_NOT_NOT:
    isFormulaInference(conclusion,premise,position);
    checkNotNot(conclusion,premise,position);
    return;
  case IR_NOT_FORALL:
    isFormulaInference(conclusion,premise,position);
    checkNotForall(conclusion,premise,position,true);
    return;
  case IR_NOT_EXISTS:
    isFormulaInference(conclusion,premise,position);
    checkNotForall(conclusion,premise,position,false);
    return;
  case IR_FORALL_AND:
    isFormulaInference(conclusion,premise,position);
    checkForallAnd(conclusion,premise,position,true);
    return;
  case IR_EXISTS_OR:
    isFormulaInference(conclusion,premise,position);
    checkForallAnd(conclusion,premise,position,false);
    return;
  case IR_FORALL_OR:
    isFormulaInference(conclusion,premise,position);
    checkForallOr(conclusion,premise,position,true);
    return;
  case IR_EXISTS_AND:
    isFormulaInference(conclusion,premise,position);
    checkForallOr(conclusion,premise,position,false);
    return;
  case IR_IMP_TO_OR:
    isFormulaInference(conclusion,premise,position);
    checkImpToOr(conclusion,premise,position);
    return;
  case IR_IFF_TO_AND:
    isFormulaInference(conclusion,premise,position);
    checkIffToAnd(conclusion,premise,position);
    return;
  case IR_XOR_TO_AND:
    isFormulaInference(conclusion,premise,position);
    checkXorToAnd(conclusion,premise,position);
    return;
  case IR_CHOICE_AXIOM:
    checkChoiceAxiom(conclusion);
    return;
  case IR_MONOTONE_REPLACEMENT:
    isFormulaInference(conclusion,premise,premise2,position);
    checkMonotoneReplacement(conclusion,premise,premise2,position);
    return;
  case IR_FORALL_ELIMINATION:
    isFormulaInference(conclusion,premise,position);
    checkForallElimination(conclusion,premise,position);
    return;
  case IR_HALF_EQUIV:
    isFormulaInference(conclusion,premise,position);
    checkEquivalenceToImplication(conclusion,premise,position);
    return;
  case IR_DUMMY_QUANTIFIER_REMOVAL:
    isFormulaInference(conclusion,premise,position);
    checkDummyQuantifierRemoval(conclusion,premise,position);
    return;
  case IR_QUANTIFIER_SWAP:
    isFormulaInference(conclusion,premise,position);
    checkQuantifierSwap(conclusion,premise,position);
    return;
  default:
    inference->toXML()->write(cout);
    PROOF_CHECKING_ERROR(conclusion,false,0);
    return;
  }
} 
bool ProofChecker::hasSinglePremise (const Inference* inference)
{
  TRACER("ProofChecker::hasSinglePremise");
  UnitList premises (inference->premises());
  return premises.isNonEmpty() &&
         premises.tail().isEmpty();
} 
ProofCheckingException::ProofCheckingException (const string& file,
						int line,
						const Unit& unit,
						bool isPremise,
						const InferenceInfo* info)
  : _hasUnit(true),
    _file(file),
    _line(line),
    _isPremise(isPremise),
    _unit(unit),
    _info(info)
{
  TRACER("ProofCheckingException::ProofCheckingException-1");
}
ProofCheckingException::ProofCheckingException (const string& file,
						int line,
						const string& msg)
  : _hasUnit(false),
    _file(file),
    _line(line),
    _message(msg)
{
  TRACER("ProofCheckingException::ProofCheckingException-2");
}
bool ProofChecker::isGoal (const Unit& unit)
{
  TRACER("ProofChecker::isGoal");
  switch (unit.unitType()) {
  case CLAUSE:
    {
      Iterator<Literal> lits(unit.clause().literals());
      while (lits.hasNext()) {
	if (! lits.next().atom().functor().isAnswer()) {
	  return false;
	}
      }
      return true;
    }
  default:
    return false;
  }
} 
void ProofCheckingException::cry (ostream& str)
{
  TRACER("ProofCheckingException::cry");
  str << "Proof checking error (file: "
      << _file
      << ", line: "
      << _line
      << "), ";
  if (_hasUnit) {
    if (_isPremise) {
      str << "cannot make the following inference:\n"
	  << _info->toString() << "\n"
	  << " from the following unit:\n"
	  << _unit.toString()
	  << "\n";
    }
    else { 
      str << " in the inference given below\n"
	  << _unit.toStringAsInference();
    }
  }
  else { 
    str << ": " << _message;
  }
  str << '\n';
} 
XML::Element* ProofCheckingException::toXML () const
{
  TRACER("ProofCheckingException::toXML");
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","proof_checking"));
  xml->addChild(new XML::TextElement("file",_file));
  xml->addChild(new XML::IntegerElement("line",_line));
  if (_hasUnit) {
    XML::Element* unit = _unit.toXML();
    xml->addChild(unit);
    unit->addAttribute(new XML::TextAttribute("kind",
					      _isPremise ? 
					      "premise" : 
					      "conclusion"));
    if (_isPremise) {
      xml->addChild(_info->toXML());
    }
  }
  else {
    xml->addChild(new XML::TextElement("message",_message));
  }
  return xml;
} 
void ProofChecker::checkRectify (const Unit& conclusion,
				 const Formula& premise)
{
  TRACER("ProofChecker::checkRectify");
  if (! areVariants(conclusion.formula(), premise)) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  if (! isRectified(conclusion.formula())) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
bool ProofChecker::areVariants (const Formula& f1, const Formula& f2)
{
  TRACER("ProofChecker::areVariants");
  Substitution sbst1;
  if (! Matching::match(f1,f2,sbst1)) {
    return false;
  }
  Substitution sbst2;
  return Matching::match(f2,f1,sbst2);
} 
bool ProofChecker::isRectified (const Formula& f)
{
  TRACER("ProofChecker::isRectified");
  VarCounter vc;
  SubformulaIterator sub(f);
  while (sub.hasNext()) {
    const Formula& g = sub.next();
    switch (g.connective()) {
    case FORALL:
    case EXISTS:
      {
	Iterator<Var> vs(g.vars());
	while (vs.hasNext()) {
	  Var v (vs.next());
	  if (vc.getCount(v) != 0) { 
	    return false;
	  }
	  vc.incCount(v);
	}
	break;
      }
    default:
      break;
    }
  }
  FormulaVarIterator fv(f);
  while (fv.hasNext()) {
    if (vc.getCount(fv.next()) != 0) {
      return false;
    }
  }
  return true;
} 
void ProofChecker::checkClosure (const Unit& conclusion,
				 const Formula& premise)
{
  TRACER("ProofChecker::checkClosure");
  Formula c (conclusion.formula());
  int freeVars = 0;
  VarCounter counter;
  FormulaVarIterator vs(premise);
  while (vs.hasNext()) {
    Var v (vs.next());
    if (counter.getCount(v) == 0) { 
      counter.incCount(v);
      freeVars++;
    }
  }
  if (freeVars == 0) {
    if (! c.equals(premise)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    return;
  }
  if (c.connective() != FORALL || !premise.equals(c.qarg())) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  Iterator<Var> ws(c.vars());
  while (ws.hasNext()) {
    Var w (ws.next());
    if (counter.getCount(w) != 1) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    freeVars--;
    counter.incCount(w);
  }
  if (freeVars != 0) { 
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkFlattening (const Unit& conclusion,
				    const Formula& premise,
				    const Position& position)
{
  TRACER("ProofChecker::checkFlattening");
  try {
    const Formula& F = position.subformula(premise);
    const Formula& G = position.subformula(conclusion.formula());
    if (F.connective() != G.connective()) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    switch (F.connective()) {
    case AND:
    case OR:
      {
	FormulaList remainder (G.args());
	if (! (scanFlattenedFormOf(F.connective(),F.args(),remainder) &&
	       remainder.isEmpty())) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
      }
      return;
    case FORALL:
    case EXISTS:
      {
	const Formula& f = F.qarg();
	if (f.connective() != F.connective()) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
	if (! (f.qarg().equals(G.qarg()))) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
	VarList fs (F.vars());
	fs.append(f.vars());
	if (! (fs.equals(G.vars()))) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
      }
      return;
    default:
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
bool ProofChecker::scanFlattenedFormOf (Connective c,
					const FormulaList& Fs,
					FormulaList& remainder)
{
  TRACER("ProofChecker::scanFlattenedFormOf");
  FormulaList fs(Fs);
  while (fs.isNonEmpty()) {
    if (remainder.isEmpty()) {
      return false;
    }
    Formula f(fs.head());
    if (f.connective() != c) {
      if (! f.equals(remainder.head())) {
	return false;
      }
      else {
	remainder.pop();
      }
    }
    else { 
      Iterator<Formula> gs(f.args());
      while (gs.hasNext()) {
	if (remainder.isEmpty()) {
	  return false;
	}
	if (! (gs.next().equals(remainder.head()))) {
	  return false;
	}
	remainder.pop();
      }
    }
    fs.pop();
  }
  return true;
} 
void ProofChecker::checkClausify (const Unit& conclusion,
				  const Unit& premise)
{
  TRACER("ProofChecker::checkClausify");
  if (conclusion.unitType() != CLAUSE ||
      premise.unitType() != FORMULA) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  InferenceInfo* info = conclusion.inference()->info();
  if (! info) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  LiteralList lits;
  conclusion.clause().literals().reverse(lits);
  Position pos;
  info->position().reverse(pos);
  if (! isDisjunctIn(lits,premise.formula(),pos) ||
      lits.isNonEmpty() ||  
      pos.isNonEmpty() ) {    
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
bool ProofChecker::isDisjunctIn (LiteralList& lits, 
				 const Formula& f,
				 Position& pos)
{
  TRACER("ProofChecker::isDisjunctIn");
  switch (f.connective()) {
  case ATOM:
    {
      if (lits.isEmpty()) {
	return false;
      }
      Literal lit(lits.head());
      if (lit.isNegative() || lit.atom() != f.atom()) {
	return false;
      }
      lits.pop();
      return true;
    }
  case AND:
    {
      if (pos.isEmpty()) {
	return false;
      }
      int p = pos.head();
      ASS(p >= 0);
      Iterator<Formula> args(f.args());
      while (p != 0) {
	p--;
	if (! args.hasNext()) {
	  return false;
	}
	args.next();
      }
      if (! args.hasNext()) {
	return false;
      }
      pos.pop();
      return isDisjunctIn (lits,args.next(),pos);
    }
  case OR:
    {
      Iterator<Formula> args(f.args());
      while (args.hasNext()) {
	if (! isDisjunctIn (lits,args.next(),pos)) {
	  return false;
	}
      }
      return true;
    }
  case NOT:
    {
      if (lits.isEmpty()) {
	return false;
      }
      Formula arg(f.uarg());
      Literal lit(lits.head());
      if (lit.isPositive() || 
	  arg.connective() != ATOM ||
	  lit.atom() != arg.atom()) {
	return false;
      }
      lits.pop();
      return true;
    }
  default:
    return false;
  }
} 
bool ProofChecker::checkReplace (const Formula& F, 
				 const Position& p,
				 const Formula& g, 
				 const Formula& G)
{
  TRACER("ProofChecker::checkReplace");
  if (p.isEmpty()) {
    return g.equals(G);
  }
  if (F.connective() != G.connective()) {
    return false;
  }
  switch (F.connective()) {
  case ATOM:
  case FALSE:
  case TRUE:
    return false;
  case AND:
  case OR:
    {
      FormulaList Fs(F.args());
      FormulaList Gs(G.args());
      int k = p.head();
      while (k > 0) {
	if (Fs.isEmpty() || 
	    Gs.isEmpty() || 
	    !(Fs.head().equals(Gs.head()))) {
	  return false;
	}
	k--;
	Fs.pop();
	Gs.pop();
      }
      return Fs.isNonEmpty()
	&& Gs.isNonEmpty()
	&& Fs.tail().equals(Gs.tail())
	&& checkReplace(Fs.head(),p.tail(),g,Gs.head());
    }
  case IMP:
  case IFF:
  case XOR:
    switch (p.head()) {
    case 0:
      return F.right().equals(G.right())
	&& checkReplace(F.left(),p.tail(),g,G.left());
    case 1:
      return F.left().equals(G.left())
	&& checkReplace(F.right(),p.tail(),g,G.right());
    default:
      return false;
    }
  case NOT:
    return p.head() == 0
      && checkReplace(F.uarg(),p.tail(),g,G.uarg());
  case FORALL:
  case EXISTS:
    return p.head() == 0
      && F.vars().equals(G.vars())
      && checkReplace(F.qarg(),p.tail(),g,G.qarg());
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
void ProofChecker::isFormulaInference (const Unit& conclusion, 
				       FormulaList& premises,
				       Position& position) 
{
  TRACER("ProofChecker::isFormulaInference");
  if (conclusion.unitType() != FORMULA) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  Inference* inf = conclusion.inference();
  if (! inf) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  InferenceInfo* info = inf->info();
  if (! info) {
    switch (inf->rule()) {
    case IR_RECTIFY:
    case IR_HALF_EQUIV:
      break;
    default:
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  else {
    position = info->position();
    if (inf->rule() != info->rule()) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  Stack<Formula> stack(premises);
  Iterator<Unit>prems(inf->premises());
  while (prems.hasNext()) {
    const Unit& parent = prems.next();
    if (parent.unitType() != FORMULA) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    stack.push(parent.formula());
  }
}  
void ProofChecker::isFormulaInference (const Unit& conclusion, 
				       Formula& premise,
				       Position& position) 
{
  TRACER("ProofChecker::isFormulaInference");
  FormulaList premises;
  isFormulaInference(conclusion,premises,position);
  if (! premises.hasLength(1)) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  premise = premises.head();
}  
void ProofChecker::isFormulaInference (const Unit& conclusion, 
				       Formula& premise1,
				       Formula& premise2,
				       Position& position) 
{
  TRACER("ProofChecker::isFormulaInference");
  FormulaList premises;
  isFormulaInference(conclusion,premises,position);
  if (! premises.hasLength(2)) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  premise1 = premises.head();
  premise2 = premises.second();
}  
void ProofChecker::checkNotAnd (const Unit& conclusion,
				const Formula& premise,
				const Position& position,
				bool isAnd)
{
  TRACER("ProofChecker::checkNotAnd");
  try {
    const Formula& G = position.subformula(premise);
    if (G.connective() != NOT) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& G1 = G.uarg();
    if (G1.connective() != (isAnd ? AND : OR)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& F = position.subformula(conclusion.formula());
    if (F.connective() != (isAnd ? OR : AND)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! isNegationOf(G1.args(),F.args())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
bool ProofChecker::isNegationOf (const FormulaList& Fs, const FormulaList Gs)
{
  TRACER("ProofChecker::isNegationOf");
  if (Fs.isEmpty()) {
    return Gs.isEmpty();
  }
  if (Gs.isEmpty()) {
    return false;
  }
  const Formula& F = Fs.head();
  const Formula& G = Gs.head();
  return G.connective() == NOT &&
         F.equals(G.uarg()) &&
         isNegationOf(Fs.tail(),Gs.tail());
} 
void ProofChecker::checkNotImp (const Unit& conclusion,
				const Formula& premise,
				const Position& position)
{
  TRACER("ProofChecker::checkNotImp");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != NOT) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.uarg(); 
    if (f.connective() != IMP) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! G.equals(Formula(AND,
			   FormulaList(f.left(),
			   FormulaList(Formula(NOT,f.right())))))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkNotIff (const Unit& conclusion,
				const Formula& premise,
				const Position& position,
				bool isIff)
{
  TRACER("ProofChecker::checkNotIff");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != NOT) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.uarg(); 
    if (f.connective() != (isIff ? IFF : XOR)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! G.equals(Formula((isIff ? XOR : IFF),f.left(),f.right()))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkNotNot (const Unit& conclusion,
				const Formula& premise,
				const Position& position)
{
  TRACER("ProofChecker::checkNotNot");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (! F.equals(Formula(NOT,
		   Formula(NOT,G)))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkNotForall (const Unit& conclusion,
				   const Formula& premise,
				   const Position& position,
				   bool isForall)
{
  TRACER("ProofChecker::checkNotForall");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != NOT) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.uarg(); 
    if (f.connective() != (isForall ? FORALL : EXISTS)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! G.equals(Formula(isForall ? EXISTS : FORALL,f.vars(),
		   Formula(NOT,f.qarg())))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkImpToOr (const Unit& conclusion,
				 const Formula& premise,
				 const Position& position)
{
  TRACER("ProofChecker::checkImpToOr");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != IMP) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! G.equals(Formula(OR,
			   FormulaList(Formula(NOT,F.left()),
			   FormulaList(F.right()))))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkChoiceAxiom (const Unit& conclusion)
{
  TRACER("ProofChecker::checkChoiceAxiom");
  if (conclusion.unitType() != FORMULA) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  Inference* inf = conclusion.inference();
  if (! inf) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  InferenceInfoTermList* info = static_cast<InferenceInfoTermList*>(inf->info());
  if (! info) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  const TermList& ts = info->terms();
  Formula F(conclusion.formula());
  FormulaVarIterator freevars(F);
  if (freevars.hasNext()) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  if (F.connective() == FORALL) {
    F = F.qarg();
  }
  if (F.connective() != IMP) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  const Formula& G = F.left();
  if (G.connective() != EXISTS) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  const VarList& xs = G.vars();
  if (xs.length() != ts.length()) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  Substitution subst;
  Iterator<Var> vs(xs);
  Iterator<Term> ss(ts);
  while (vs.hasNext()) {
    subst.bind(vs.next(),ss.next());
  }
  if (! obtainedBySubstitution(G.qarg(),F.right(),subst)) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  ss.reset(ts);
  while (ss.hasNext()) {
    const Term& t = ss.next();
    const FunctionSymbol& f = t.functor();
    if (_input.existsFun(f.name(),f.arity())) { 
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (_fresh.existsFun(f.name(),f.arity())) { 
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    _fresh.createFun(f.name(),f.arity());
  }
}  
bool ProofChecker::obtainedBySubstitution (const Formula& F, 
					   const Formula& G,
					   Substitution& s)
{
  VarCounter bound; 
  return obtainedBySubstitution(F,G,s,bound);
} 
bool ProofChecker::obtainedBySubstitution (const Formula& F, 
					   const Formula& G,
					   Substitution& s,
					   VarCounter& bound)
{
  if (F.connective() != G.connective()) {
    return false;
  }
  switch (F.connective()) {
  case ATOM:
    return obtainedBySubstitution(F.atom().args(),G.atom().args(),s,bound);
  case AND:
  case OR:
    return obtainedBySubstitution(F.args(),G.args(),s,bound);
  case IMP:
  case IFF:
  case XOR:
    return obtainedBySubstitution(F.left(),G.left(),s,bound) &&
           obtainedBySubstitution(F.right(),G.right(),s,bound);
  case NOT:
    return obtainedBySubstitution(F.uarg(),G.uarg(),s,bound);
  case FORALL:
  case EXISTS:
    {
      if (! F.vars().equals(G.vars())) {
	return false;
      }
      Iterator<Var> vs (F.vars());
      while (vs.hasNext()) {
	bound.incCount(vs.next());
      }
      if (! obtainedBySubstitution(F.qarg(),G.qarg(),s,bound)) {
	return false;
      }
      vs.reset(F.vars());
      while (vs.hasNext()) {
	bound.decCount(vs.next());
      }
      return true;
    }
  case FALSE:
  case TRUE:
    return true;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
bool ProofChecker::obtainedBySubstitution (const FormulaList& Fs, 
					   const FormulaList& Gs,
					   Substitution& s,
					   VarCounter& bound)
{
  if (Fs.length() != Gs.length()) {
    return false;
  }
  Iterator<Formula> fs(Fs);
  Iterator<Formula> gs(Gs);
  while (fs.hasNext()) {
    const Formula& F = fs.next();
    const Formula& G = gs.next();
    if (! obtainedBySubstitution(F,G,s,bound)) {
      return false;
    }
  }
  return true;
} 
bool ProofChecker::obtainedBySubstitution (const TermList& Ss, 
					   const TermList& Ts,
					   Substitution& subst,
					   VarCounter& bound)
{
  if (Ss.length() != Ts.length()) {
    return false;
  }
  Iterator<Term> ss(Ss);
  Iterator<Term> ts(Ts);
  while (ss.hasNext()) {
    const Term& s = ss.next();
    const Term& t = ts.next();
    if (! obtainedBySubstitution(s,t,subst,bound)) {
      return false;
    }
  }
  return true;
} 
bool ProofChecker::obtainedBySubstitution (const Term& s, 
					   const Term& t,
					   Substitution& subst,
					   VarCounter& bound)
{
  switch (s.tag()) {
  case VAR:
    {
      Var w(s.var());
      if (bound.getCount(w) > 0) { 
	return s.equals(t);
      }
      Term r;
      if (! subst.isBound(w,r)) { 
	return s.equals(t);
      }
      if (! r.equals(t)) {
	return false;
      }
      TermVarIterator vars(t);
      while (vars.hasNext()) {
	Var v(vars.next());
	if (bound.getCount(v) > 0) {
	  return false;
	}
      }
      return true;
    }
  case COMPOUND:
    return t.tag() == COMPOUND &&
           obtainedBySubstitution(s.args(),t.args(),subst,bound);
  case INTERPRETED:
    return s.equals(t);
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
void ProofChecker::checkInput (const Unit& conclusion)
{
  TRACER("ProofChecker::checkInput (const Unit&");
  switch (conclusion.unitType()) {
  case FORMULA:
    if (! checkInput(conclusion.formula())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    return;
  case CLAUSE:
    if (! checkInput(conclusion.formula())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    return;
  }
} 
bool ProofChecker::checkInput (const Formula& f)
{
  AtomIterator atm(f);
  while (atm.hasNext()) {
    if (! checkInput(atm.next())) {
      return false;
    }
  }
  return true;
} 
bool ProofChecker::checkInput (const Atom& a)
{
  if (! a.isEquality()) {
    const PredicateSymbol& p = a.functor();
    if (_fresh.existsPred(p.name(),p.arity())) { 
      return false;
    }
    _input.createPred(p.name(),p.arity());
  }
  SubtermIterator ts(a);
  while (ts.hasNext()) {
    const Term& t = ts.next();
    switch (t.tag()) {
    case VAR:
    case INTERPRETED:
      break;
    case COMPOUND:
      {
	const FunctionSymbol& f = t.functor();
	if (_fresh.existsFun(f.name(),f.arity())) { 
	  return false;
	}
	_input.createFun(f.name(),f.arity());
      }
      break;
    }
  }
  return true;
} 
void ProofChecker::checkMonotoneReplacement (const Unit& conclusion, 
					     const Formula& F,
					     const Formula& G1,
					     const Position& p)
{
  try {
    const Formula& H = F.connective() == FORALL ? F.qarg() : F;
    if (H.connective() != IMP) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& G2 = conclusion.formula();
    int polarity = p.polarity(G2);
    Formula F1;
    Formula F2;
    switch (polarity) {
    case 1:
      F1 = H.left();
      F2 = H.right();
      break;
    case -1:
      F2 = H.left();
      F1 = H.right();
      break;
    default: 
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& B1 = p.subformula(G1);
    const Formula& B2 = p.subformula(G2);
    Substitution s;
    if (! Matching::match(Formula(IMP,F1,F2),Formula(IMP,B1,B2),s)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    Position rev;
    p.reverse(rev);
    if (! checkReplace(G1,rev,B2,G2)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkForallElimination (const Unit& conclusion,
					   const Formula& G1,
					   const Position& p)
{
  TRACER("ProofChecker::checkForallElimination");
  try {
    if (p.polarity(G1) != 1) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    Formula F1 = p.subformula(G1);
    if (F1.connective() != FORALL) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    F1 = F1.qarg();
    const Formula& G2 = conclusion.formula();
    const Formula& F2 = p.subformula(G2);
    Substitution s;
    if (! Matching::match(F1,F2,s)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    Position rev;
    p.reverse(rev);
    if (! checkReplace(G1,rev,F2,G2)) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkEquivalenceToImplication (const Unit& conclusion, 
						  const Formula& f,
						  const Position& p)
{
  Formula F(f);
  Formula G(conclusion.formula());
  if (F.connective() == FORALL) {
    if (G.connective() != FORALL || ! F.vars().equals(G.vars())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    F = F.qarg();
    G = G.qarg();
  }
  if (F.connective() != IFF) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  if (G.connective() != IMP) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  if (F.left() == G.left() && F.right() == G.right()) {
    return;
  }
  if (F.left() == G.right() && F.right() == G.left()) {
    return;
  }
  PROOF_CHECKING_ERROR(conclusion,false,0);
} 
void ProofChecker::checkIffToAnd (const Unit& conclusion,
				  const Formula& premise,
				  const Position& position)
{
  TRACER("ProofChecker::checkIffToAnd");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != IFF) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& F1 = F.left();
    const Formula& F2 = F.right();
    if (! G.equals(Formula(AND,
			   FormulaList(Formula(IMP,F1,F2),
			   FormulaList(Formula(IMP,F2,F1)))))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkXorToAnd (const Unit& conclusion,
				  const Formula& premise,
				  const Position& position)
{
  TRACER("ProofChecker::checkXorToAnd");
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != XOR) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& F1 = F.left();
    const Formula& F2 = F.right();
    if (! G.equals(Formula(AND,
			   FormulaList(Formula(OR,
					       FormulaList(F1,
                                               FormulaList(F2))),
			   FormulaList(Formula(OR,
					       FormulaList(Formula(NOT,F1),
					       FormulaList(Formula(NOT,F2))))))))) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkForallAnd (const Unit& conclusion,
				   const Formula& premise,
				   const Position& position,
				   bool isAnd)
{
  TRACER("ProofChecker::checkForallAnd");
  Connective Q = isAnd ? FORALL : EXISTS;
  Connective c = isAnd ? AND : OR;
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != Q) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.qarg(); 
    if (f.connective() != c) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (G.connective() != c) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    FormulaList Fs(f.args());
    FormulaList Gs(G.args());
    if (Fs.length() != Gs.length()) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    while (Fs.isNonEmpty()) {
      if (! Gs.head().equals(Formula(Q,F.vars(),Fs.head()))) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      Fs.pop();
      Gs.pop();
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkForallOr (const Unit& conclusion,
				  const Formula& premise,
				  const Position& position,
				  bool isForall)
{
  TRACER("ProofChecker::checkForallOr");
  Connective Q = isForall ? FORALL : EXISTS;
  Connective c = isForall ? OR : AND;
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    if (F.connective() != Q) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.qarg(); 
    if (f.connective() != c) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    int nv = 0; 
    int ntv = 0; 
    VarCounter vars;
    Iterator<Var> ws(F.vars());
    while (ws.hasNext()) {
      Var w(ws.next());
      if (vars.getCount(w) == 0) {
	vars.incCount(w);
	nv++;
      }
    }
    Iterator<Term> ts(static_cast<InferenceInfoTermList*>(conclusion.inference()->info())->terms());
    while (ts.hasNext()) {
      const Term& t = ts.next();
      if (t.tag() == VAR) {
	Var v(t.var());
	switch (vars.getCount(v)) {
	case 0:
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	case 1:
	  vars.incCount(v);
	  ntv++;
	  break;
	case 2: 
	  break;
	}
      }
      else {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
    }
    FormulaList Gs; 
    if (G.connective() == c) {
      if (nv != ntv) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      Gs = G.args();
    }
    else if (G.connective() == Q) {
      const Formula& G1 = G.qarg();
      if (G1.connective() != c) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      Gs = G1.args();
      int remains = nv - ntv;
      Iterator<Var> vs(G.vars());
      while (vs.hasNext()) {
	Var v(vs.next());
	switch (vars.getCount(v)) {
	case 1:
	  vars.setCount(v,3);
	  remains--;
	  break;
	case 3:
	  break;
	default:
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
      }
      if (remains != 0) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      vs.reset(G.vars());
      while (vs.hasNext()) {
	vars.setCount(vs.next(),1);
      }
    }
    else {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    FormulaList Fs(f.args());
    FormulaList Hs; 
    bool found = false;
    while (Fs.isNonEmpty()) {
      Formula f(Fs.head());
      Fs.pop();
      bool toppledOccurs = false;
      FormulaVarIterator fvs(f);
      while (fvs.hasNext()) {
	Var v(fvs.next());
	if (vars.getCount(v) == 2) { 
	  toppledOccurs = true;
	  break;
	}
      }
      if (toppledOccurs) {
	if (Hs.isEmpty()) {
	  if (found) {
	    PROOF_CHECKING_ERROR(conclusion,false,0);
	  }
	  found = true;
	  if (Gs.isEmpty()) {
	    PROOF_CHECKING_ERROR(conclusion,false,0);
	  }
	  Formula g(Gs.head());
	  Gs.pop();
	  if (g.connective() != Q) {
	    PROOF_CHECKING_ERROR(conclusion,false,0);
	  }
	  int remains = ntv;
	  Iterator<Var> vs(g.vars());
	  while (vs.hasNext()) {
	    Var v(vs.next());
	    switch (vars.getCount(v)) {
	    case 2:
	      vars.incCount(v);
	      remains--;
	      break;
	    case 3:
	      break;
	    default:
	      PROOF_CHECKING_ERROR(conclusion,false,0);
	    }
	  }
	  if (remains != 0) {
	    PROOF_CHECKING_ERROR(conclusion,false,0);
	  }
	  vs.reset(g.vars());
	  while (vs.hasNext()) {
	    vars.setCount(vs.next(),2);
	  }
	  const Formula& g1 = g.qarg();
	  if (g1.connective() == c) {
	    Hs = g1.args();
	  }
	  else {
	    Hs = FormulaList(g1);
	  }
	}
	if (! f.equals(Hs.head())) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
	Hs.pop();
      }
      else if (Gs.isEmpty()) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      else {
	if (! f.equals(Gs.head())) {
	  PROOF_CHECKING_ERROR(conclusion,false,0);
	}
	Gs.pop();
      } 
    }
    if (Gs.isNonEmpty() || Hs.isNonEmpty()) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkDummyQuantifierRemoval (const Unit& conclusion,
						const Formula& premise,
						const Position& position)
{
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    Connective Q = F.connective();
    switch (Q) {
    case FORALL:
    case EXISTS:
      break;
    default:
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.qarg();
    VarCounter removed;
    Iterator<Var> vs(F.vars());
    while (vs.hasNext()) {
      removed.setCount(vs.next(),2);
    }
    if (! f.equals(G)) { 
      if (G.connective() != Q) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      if (! f.equals(G.qarg())) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
      vs.reset(G.vars());
      while (vs.hasNext()) {
	removed.setCount(vs.next(),1);
      }
    }
    FormulaVarIterator free(f);
    while (free.hasNext()) {
      if (removed.getCount(free.next()) > 1) {
	PROOF_CHECKING_ERROR(conclusion,false,0);
      }
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
void ProofChecker::checkQuantifierSwap (const Unit& conclusion,
					const Formula& premise,
					const Position& position)
{
  try {
    const Formula& G = position.subformula(conclusion.formula());
    const Formula& F = position.subformula(premise);
    Connective Q = F.connective();
    switch (Q) {
    case FORALL:
    case EXISTS:
      break;
    default:
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (G.connective() != Q) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& f = F.qarg();
    if (f.connective() != Q) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    const Formula& g = G.qarg();
    if (g.connective() != Q) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! F.vars().equals(g.vars())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! G.vars().equals(f.vars())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
    if (! g.qarg().equals(f.qarg())) {
      PROOF_CHECKING_ERROR(conclusion,false,0);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
} 
