/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <cstdio>
#include "Int.hpp"
#include "Formula.hpp"
#include "Reader.hpp"
#include "VampireException.hpp"
using namespace Vampire;
string Reader::toString (const Term* t) const
{
  switch (tag(t)) {
  case VAR:
    return varToString(var(t));
  case COMPOUND:
    return name(functor(t)) + toString(args(t),endOfArgs(t));
  case INTERPRETED:
    return value(t)->toString();
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
}
string Reader::toString (const TermList* ts, const TermList* end) const
{
  if (isEmpty(ts,end)) {
    return "";
  }
  string result = string("(") + toString(head(ts));
  ts = tail(ts,end);
  while (! isEmpty(ts,end)) {
    result += (string)"," + toString(head(ts));
    ts = tail(ts,end);
  }
  return result + ")";
} 
string Reader::varToString (int v) const
{
  static char tmp [20];
  tmp[0] = v >= 0 ? 'X' : '@';
  sprintf(tmp+1,"%d",abs(v));
  return string(tmp);
} 
string Reader::toString (const Atom* a) const
{
  const TermList* end = endOfArgs(a);
  if (isEquality(a)) {
    return toString(head(args(a))) + "=" +
      toString(head(tail(args(a),end)));
  }
  return name(functor(a)) + toString(args(a),end);
} 
string Reader::toString (const Literal* l) const
{
  const Atom* a = atom(l);
  const TermList* end = endOfArgs(a);
  if (isPositive(l)) {
    return toString(a);
  }
  if (isEquality(a)) {
    const TermList* ts = args(a);
    return toString(head(ts)) + "!=" +
      toString(head(tail(ts,end)));
  }
  return string("~") + toString(a);
} 
string Reader::toString (const LiteralList* ls,
			 const LiteralList* end) const
{
  if (isEmpty(ls,end)) {
    return "";
  }
  string result = toString(head(ls));
  ls = tail(ls,end);
  while (! isEmpty(ls,end)) {
    result += (string)" \\/ " + toString(head(ls));
    ls = tail(ls,end);
  }
  return result;
} 
string Reader::toString (const Clause* c) const
{
  const LiteralList* ls = literals(c);
  const LiteralList* end = endOfLiterals(c);
  if (isEmpty(ls,end)) {
    return "#";
  }
  return toString(ls,end);
} 
string Reader::toString (const Unit* u) const
{
  string prefix = Shell::Int::toString(number(u)) + ". ";
  Background b;
  saveBackground(u,b);
  if (b.isClause) {
    return prefix + toString(clause(u));
  }
  else {
    return prefix + toString(formula(u));
  }
} 
string Reader::toString (const Formula* f) const
{
  return reinterpret_cast<const Shell::Formula*>(f)->toString();
} 
const TermList* Reader::endOfArgs (const Term*) const
{
  return 0;
} 
const TermList* Reader::endOfArgs (const Atom*) const
{
  return 0;
} 
const LiteralList* Reader::endOfLiterals (const Clause*) const
{
  return 0;
} 
const UnitList* Reader::endOfPremises (const Unit*) const
{
  return 0;
} 
Background::Background ()
  : isClause(0),
    isDefinition(0),
    storageKind((StorageKind)0),
    derivedFromAxioms(0),
    isInput(0),
    isObtainedDuringPreprocessing(0),
    binaryResolutionUsed(0),
    forwardTransitiveChainingUsed(0),
    backwardTransitiveChainingUsed(0),
    forwardSuperpositionUsed(0),
    backwardSuperpositionUsed(0),
    equalityResolutionUsed(0),
    equalityFactoringUsed(0),
    forwardDemodulationUsed(0),
    backwardDemodulationUsed(0),
    simplificationByEqualityResolutionUsed(0),
    forwardSubsumptionResolutionUsed(0),
    backwardSubsumptionResolutionUsed(0),
    isReanimatedPassive(0),
    evaluationUsed(0),
    foldingUsed(0),
    clausificationUsed(0),
    choiceAxiom(0),
    monotoneReplacement(0),
    forallElimination(0),
    rectify(0),
    notAnd(0),
    notOr(0),
    notImp(0),
    notIff(0),
    notXor(0),
    notNot(0),
    notForall(0),
    notExists(0),
    impToOr(0),
    iffToAnd(0),
    xorToEnd(0),
    closure(0),
    flattening(0),
    normalisation(0),
    ennf(0),
    nnf(0),
    dummyQuantifierRemoval(0),
    forallAnd(0),
    existsOr(0),
    quantifierSwap(0),
    forallOr(0),
    existsAnd(0),
    permutation(0),
    reorderEquailities(0),
    halfEquivalence(0),
    miniscope(0),
    skolemise(0),
    clausify(0),
    removeDuplicateLiterals(0),
    kernelInference(0),
    equalityProxyReplacement(0),
    equalityProxyAxiom1(0),
    equalityProxyAxiom2(0),
    definitionUnfolding(0),
    rowVariableExpansion(0),
    predicateDefinition(0),
    reduceFalseTrue(0)
{
} 
Unit* Reader::origin (const Unit*) const
{
  return 0;
} 
