/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Environment.hpp"
#include "Inference.hpp"
#include "IntNameTable.hpp"
#include "Kernel.hpp"
#include "Map.hpp"
#include "Hash.hpp"
#include "Options.hpp"
#include "Problem.hpp"
#include "ProofChecker.hpp"
#include "Refutation.hpp"
#include "Permutation.hpp"
#include "TPTP.hpp"
#include "XML.hpp"
namespace Shell {
class Refutation::Map 
  : public Shell::Map<int,Unit,Hash>
{
}; 
Refutation::Refutation (const Kernel& kernel, 
			const Problem& problem,
			const Options& options)
  : _kernel (kernel),
    _problem(problem),
    _options(options),
    _proofCheck(options.proofChecking())
{
  TRACER ("Refutation::Refutation");
  Unit::setLastNumber(kernel.refutation()->number());
  Map map;
  collectKernelClause(kernel.refutation(), map);
  _goal = _refutation.head();
  UnitList refutation;
  int lastNumber = 0; 
  Iterator<Unit> units (_refutation);
  while (units.hasNext()) {
    Unit u (units.next());
    u.setNumberInRefutation(++ lastNumber);
    refutation.push(u);
  }
  _refutation = refutation;
  if (options.proofChecking()) {
    ProofChecker::check (*this);
  }
} 
Unit Refutation::collectKernelClause (const void* kernelClause, Map& map)
{
  TRACER ("Refutation::collectKernelClause (VampireKernel::Clause ...)");
  const VampireKernel::Clause* clause = reinterpret_cast<const VampireKernel::Clause*>(kernelClause);
  int number = clause->number();
  Unit unit;
  if (map.find(number, unit)) { 
    return unit;
  }
  if (clause->isInputClause()) { 
    Unit parent (clause->inputClauseOrigin());
    collect(parent,map);
    return parent;
  }
  if (clause->isNameIntroduction()) { 
    Formula ni;
    _kernel.getDefinition(clause,ni);
    Unit result(ni,clause->number());
    collect(result,map);
    return result;
  }
  UnitList parents = collectKernelClauseParents(clause->firstAncestor(),map);
  Clause cls;
  _kernel.getClause(clause,cls);
  if (parents.hasLength(1)) {
    Unit parent(parents.head());
    if (parent.unitType() == CLAUSE && 
	Permutation::isPermutation(parent.clause().literals(),cls.literals())) {
      return parent;
    }
  }
  VampireKernel::Clause::BackgroundFlagIterator flags(clause-> backgroundFlagIterator());
  Unit result(cls,parents,clause->number(),flags);
  collect(result,map);
  return result;
} 
void Refutation::collect (Unit& unit, Map& map)
{
  TRACER ("Refutation::collect (const Unit& ...)");
  int number = unit.number();
  Unit unit1;
  if (map.find(number,unit1)) { 
    return;
  }
  Iterator<Unit> parents (unit.inference()->premises());
  while (parents.hasNext()) {
    Unit par(parents.next());
    collect(par, map);
  }
  map.insert (number, unit);
  _refutation.push(unit);
} 
UnitList Refutation::collectKernelClauseParents (const void* ps,
						 Map& map)
{
  TRACER ("Refutation::collectKernelClauseParents");
  const VampireKernel::Clause::Ancestor* parent =
    reinterpret_cast<const VampireKernel::Clause::Ancestor*>(ps);
  if (! parent) {
    return UnitList ();
  }
  const VampireKernel::Clause* clause = parent->clause();
  const VampireKernel::Clause::Ancestor* rest = parent->next ();
  Unit u(collectKernelClause(clause, map));
  return UnitList(u, collectKernelClauseParents(rest, map));
} 
XML::Element* Refutation::toXML () const
{
  TRACER("Refutation::toXML");
  XML::DeepElement* ref = new XML::DeepElement("refutation");
  Iterator<Unit> units (_refutation);
  while (units.hasNext()) {
    ref->addChild(units.next().toXML());
  }
  return ref;
} 
void Refutation::prettyPrint (ostream& str, 
			      int answerNumber)
{
  TRACER("Refutation::prettyPrint");
  string xout = _options.xmlOutput();
  if (! _options.outputSuppressed()) {
    switch (Environment::outputSyntax) {
    case OS_VAMPIRE:
      {
	str << "=========== Refutation ==========\n";
	Iterator<Unit> units(refutation());
	while (units.hasNext()) {
	  str << units.next().toStringAsInference();
	}
	str << "======= End of refutation =======\n";
	break;
      }
    case OS_TPTP:
      str << TPTP::toString(*this);
      break;
#if DEBUG_SHELL
    default:
      ASS(false);
#endif
    }
  }
} 
}
