/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <string.h>
#include <stdio.h>
#include <cmath>
#include "Int.hpp"
#include "InterpretedConstant.hpp"
#include "Signature.hpp"
#include "Symbol.hpp"
#include "SymCounter.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
Signature::Signature () :
  _noOfPreds (0),
  _noOfFuns (0),
  _constants (0),
  _propositionalSymbols (0),
  _maxFunArity (0),
  _maxPredArity (0),
  _kernelSymbols (new Symbol*[1022]),
  _maxNumberOfKernelSymbols (1022),
  _lastSkolem (0),
  _lastName (0),
  _skolemPrefix("$s"),
  _namePrefix("$n")
{
  TRACER("Signature::Signature");
  if (! _kernelSymbols) {
    ALLOCATION_FAILURE;
  }
  for (int i = 0; i < 1022; i++) {
    _kernelSymbols[i] = 0;
  }
  for ( int j = NO_OF_BUCKETS - 1; j >= 0; j-- ) {
    _fbuckets [j] = SymbolList::empty ();
    _pbuckets [j] = SymbolList::empty ();
  }
  _equality = &createPred("equal", 2);
} 
Signature::~Signature () 
{
  TRACER("Signature::~Signature");
  for ( int i = NO_OF_BUCKETS - 1; i >= 0; i-- ) {
    SymbolList::Iterator fit (_fbuckets [i]);
    while (fit.hasNext()) {
      delete static_cast<FunctionSymbol*>(fit.next());
    }
    _fbuckets[i]->destroy();
    SymbolList::Iterator pit (_pbuckets [i]);
    while (pit.hasNext()) {
      delete static_cast<PredicateSymbol*>(pit.next());
    }
    _pbuckets[i]->destroy();
  }
} 
Symbol& Signature::create (const string& str, 
			   int arity, 
			   SymbolList** buckets, 
			   bool isPred )
{
  TRACER("Signature::create");
  int bucket = hash(str,arity);
  for (SymbolList* l = buckets[bucket]; ! l->isEmpty (); l = l->tail()) {
    Symbol* sym = l->head();
    if (arity == sym->arity() &&         
	str == sym->name()) {  
      return *sym;
    }
  }
  if (isPred) { 
    if (arity == 0) {
      _propositionalSymbols++;
    }
    else if (arity > _maxPredArity) {
      _maxPredArity = arity;
    }
  }
  else { 
    if (arity == 0) {
      _constants++;
    }
    else if (arity > _maxFunArity) {
      _maxFunArity = arity;
    }
  }
  Symbol* newf; 
  if (isPred) {
    newf = new PredicateSymbol (str,arity,_noOfPreds);
  }
  else {
    newf = new FunctionSymbol (str,arity,_noOfFuns);
  }
  buckets[bucket] = new SymbolList(newf, buckets[bucket]);
  return *newf;
} 
int Signature::hash (const string& str,int arity)
{
  TRACER("Signature::hash");
  int result = 0;
  int length = str.length();
  for (int i = 0; i < length; i++ ) {
    result = ( result * 10 + str[i] ) % NO_OF_BUCKETS;
  }
  return result;
} 
Signature::Iterator::Iterator (SymbolList** buckets)
  :
  _firstBucket (buckets),
  _currentBucket (buckets + NO_OF_BUCKETS - 1),
  _currentList (*_currentBucket)
{
  TRACER("Signature::Iterator::Iterator");
} 
bool Signature::Iterator::hasNext ()
{
  TRACER("Signature::Iterator::hasNext");
  for (;;) {
    if ( _currentList->isNonEmpty() ) {
      return true;
    }
    if (_currentBucket == _firstBucket) {
      return false;
    }
    _currentBucket--;
    _currentList = *_currentBucket;
  }
} 
Symbol& Signature::Iterator::next ()
{
  TRACER("Signature::Iterator::next");
  ASS( _currentList->isNonEmpty() );
  Symbol* result = _currentList->head ();
  _currentList = _currentList->tail ();
  return *result;
} 
Symbol* Signature::exists (const string& name, 
			   int arity, 
			   SymbolList** buckets)
{
  TRACER("Signature::exists");
  int bucket = hash(name,arity);
  for (SymbolList* l = buckets[bucket]; ! l->isEmpty (); l = l->tail()) {
    Symbol* sym = l->head();
    if (name == sym->name() && arity == sym->arity()) {  
      return sym;
    }
  }
  return 0;
} 
PredicateSymbol& Signature::createAnswerPredicate (int arity)
{
  TRACER("Signature::createAnswerPred");
  PredicateSymbol& answer = createPred("$answer",arity);
  answer.makeAnswer();
  return answer;
} 
string Signature::summary () 
{
  TRACER("Signature::summary");
  return string("Predicate symbols: ") +
    Int::toString(_noOfPreds) +
    " (propositional " +
    Int::toString(_propositionalSymbols) +
    ", maximal arity: " +
    Int::toString(_maxPredArity) +
    ")\n" +
    "Function symbols: " +
    Int::toString(_noOfFuns) +
    " (constants " +
    Int::toString(_constants) +
    ", skolem: " +
    Int::toString(_lastSkolem+1) +
    ", maximal arity: " +
    Int::toString(_maxFunArity) +
    ")\n";
} 
XML::Element* Signature::toXML () const
{
  TRACER("Signature::toXML");
  XML::DeepElement* signature = new XML::DeepElement("signature");
  XML::DeepElement* functions = new XML::DeepElement("functions");
  signature->addChild(functions);
  XML::DeepElement* predicates = new XML::DeepElement("predicates");
  signature->addChild(predicates);
  FunIterator fs(const_cast<Signature&>(*this));
  while (fs.hasNext()) {
    functions->addChild(fs.next().toXML());
  }
  PredIterator ps(const_cast<Signature&>(*this));
  while (ps.hasNext()) {
    predicates->addChild(ps.next().toXML());
  }
  return signature;
} 
void Signature::registerKernelNumber (Symbol& symbol, unsigned long n)
{
  TRACER("Signature::registerKernelNumber");
  ASS(&symbol != 0);
  expandKernelSymbolsToFit(n);
  _kernelSymbols[n] = &symbol;
  symbol.setKernelNumber(n);
} 
const FunctionSymbol& Signature::getFunctionByItsKernelNumber (unsigned long n,
							       int arity)
{
  TRACER("Signature::getFunctionByItsKernelNumber");
  expandKernelSymbolsToFit(n);
  Symbol* symbol = _kernelSymbols[n];
  if (symbol && symbol->arity() == arity && symbol->isFunction()) {
    return *(static_cast<FunctionSymbol*>(symbol));
  }
  static int lastFun = 0;
  static char name [12] = {'f', '_'}; 
  string nm;
  do {
    sprintf (name+2, "%d", ++lastFun);
    nm = name;
  }
  while (existsFun(nm,arity));
  FunctionSymbol& result = createFun(nm, arity);
  _kernelSymbols[n] = &result;
  return result;
} 
const FunctionSymbol& Signature::getFunctionByItsKernelNumber (unsigned long n)
{
  TRACER("Signature::getFunctionByItsKernelNumber");
  ASS (_maxNumberOfKernelSymbols > (long)n);
  Symbol* symbol = _kernelSymbols[n];
  ASS (symbol);
  ASS(symbol->isFunction());
  return *(static_cast<FunctionSymbol*>(symbol));
} 
const PredicateSymbol& Signature::getPredicateByItsKernelNumber (unsigned long n,
								 int arity)
{
  TRACER("Signature::getPredicateByItsKernelNumber");
  expandKernelSymbolsToFit(n);
  Symbol* symbol = _kernelSymbols[n];
  if (symbol && symbol->arity() == arity && ! symbol->isFunction()) {
    return *(static_cast<PredicateSymbol*>(symbol));
  }
  static int lastPred = 0;
  static char name [12] = {'p', '_'}; 
  string nm;
  do {
    sprintf (name+2, "%d", ++lastPred);
    nm = name;
  }
  while (existsPred(nm,arity));
  PredicateSymbol& result = createPred(nm, arity);
  result.makeName();
  _kernelSymbols[n] = &result;
  return result;
} 
const PredicateSymbol& Signature::getPredicateByItsKernelNumber (unsigned long n)
{
  TRACER("Signature::getPredicateByItsKernelNumber");
  ASS (_maxNumberOfKernelSymbols > (long)n);
  Symbol* symbol = _kernelSymbols[n];
  ASS (symbol);
  ASS(! symbol->isFunction());
  return *(static_cast<PredicateSymbol*>(symbol));
} 
void Signature::expandKernelSymbolsToFit (unsigned long n)
{
  TRACER("Signature::expandKernelSymbolsToFit");
  if (_maxNumberOfKernelSymbols > (int)n) {
    return;
  }
  int newLength = max(_maxNumberOfKernelSymbols + 1024,(int)n+1);
  Symbol** newArray = new Symbol* [newLength];
  if (! newArray) {
    ALLOCATION_FAILURE;
  }
  for (int i = 0;i < _maxNumberOfKernelSymbols;i++) {
    newArray[i] = _kernelSymbols[i];
  }
  for (int j = _maxNumberOfKernelSymbols;j < newLength;j++) {
    newArray[j] = 0;
  }
  delete [] _kernelSymbols;
  _kernelSymbols = newArray;
  _maxNumberOfKernelSymbols = newLength;
} 
PredicateSymbol& Signature::newName (int arity)
{
  TRACER("Signature::newSkolemName");
  for (;;) {
    string name(_namePrefix + Int::toString(_lastName++));
    PredicateSymbol* f = existsPred(name,arity);
    if (! f) {
      PredicateSymbol& result = createPred(name,arity);
      result.makeName();
      return result;
    }
  }
} 
FunctionSymbol& Signature::newSkolemFunction (int arity)
{
  TRACER("Signature::newSkolemFunction");
  for (;;) {
    string name(_skolemPrefix + Int::toString(_lastSkolem++));
    FunctionSymbol* f = existsFun (name,arity);
    if (! f) {
      FunctionSymbol& result = createFun(name, arity);
      result.makeSkolem();
      return result;
    }
  }
} 
void Signature::setSkolemPrefix (const string& prefix)
{
  _skolemPrefix = prefix;
} 
