/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Assertion.hpp"
#include "FormulaVarIterator.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Position.hpp"
#include "Rectify.hpp"
#include "Signature.hpp"
#include "Skolem.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
#include "VarCounter.hpp"
using namespace Shell;
void Skolem::skolemise (Unit& u)
{
  TRACER("Skolem::skolemise (Unit& u)");
  Rectify::rectify(u);
  ASS(u.unitType() == FORMULA);
  Formula f(u.formula());
  Skolem skol;
  skol.skolemiseNNF(f,Position(),Position());
  if (! f.equals(u.formula())) {
    Inference* inf = new Inference(IR_SKOLEMIZE,u);
    inf->setInfo(skol._info);
    u = Unit(f,inf);
  }
} 
void Skolem::skolemiseNNF (Formula& f, 
			   const Position& positionWithERemoved,
			   const Position& positionWithQRemoved)
{
  TRACER("Skolem::skolemiseNNF (Formula& f,...)");
  switch (f.connective()) {
  case ATOM: 
    {
      Atom a (f.atom());
      _subst.apply(a);
      if (a == f.atom()) {
	return;
      }
      f = Formula(a);
    }
    return;
  case AND:
  case OR: 
    {
      FormulaList fs (f.args());
      skolemiseNNF(fs, positionWithERemoved, positionWithQRemoved, 0);
      if (fs == f.args()) {
	return;
      }
      f = Formula(f.connective(), fs);
    }
    return;
  case NOT: 
    {
      Atom a(f.uarg().atom());
      _subst.apply(a);
      if (a == f.uarg().atom()) {
	return;
      }
      f = Formula (NOT, Formula(a));
    }
    return;
  case FORALL: 
    {
      VarList ws(f.vars());
      f = f.qarg();
      skolemiseNNF(f,Position(0,positionWithERemoved),positionWithQRemoved);
      _info = new InferenceInfoTermList(IR_SKOLEMIZE,
					positionWithQRemoved,
					_info,
					TermList(ws));
    }
    return;
  case EXISTS: 
    {
      TermList skolemTerms;
      VarList xs(f.vars());
      VarList ys;
      f.freeVariables(ys);
      makeSkolemTerms(xs, TermList(ys), ys.length(), skolemTerms);
      f = f.qarg();
      skolemiseNNF(f,positionWithERemoved,positionWithQRemoved);
      _info = new InferenceInfoTermList(IR_SKOLEMIZE,
					positionWithERemoved,
					_info,
					skolemTerms);
      Iterator<Var> ws (xs);
      while (ws.hasNext()) {
	_subst.undoBinding (ws.next());
      }
    }
    return;
  case TRUE:
  case FALSE:
    return;
#if DEBUG_SHELL
  case IMP:
  case IFF:
  case XOR:
  default:
    ASS( false );
#endif
  }
} 
void Skolem::skolemiseNNF (FormulaList& fs,
			   const Position& positionWithERemoved,
			   const Position& positionWithQRemoved,
			   int pos) 
{
  TRACER("skolemiseNNF (FormulaList& fs,...)");
  if (fs.isEmpty()) {
    return;
  }
  Formula h (fs.head());
  skolemiseNNF(h, Position(pos,positionWithERemoved), 
	       Position(pos,positionWithQRemoved));
  FormulaList t (fs.tail());
  skolemiseNNF(t, positionWithERemoved, 
	       positionWithQRemoved, pos+1);
  if (h == fs.head() && t == fs.tail()) {
    return;
  }
  fs = FormulaList(h,t);
} 
void Skolem::makeSkolemTerms (const VarList& vars, 
			      const TermList& skolemArgs,
			      int arity,
			      TermList& skolemTerms)
{
  if (vars.isEmpty()) {
    return;
  }
  Term skolemTerm(Signature::sig.newSkolemFunction(arity), skolemArgs);
  _subst.bind(vars.head(), skolemTerm);
  makeSkolemTerms(vars.tail(), skolemArgs, arity, skolemTerms);
  skolemTerms.push(skolemTerm);
} 
