/**
 * File SpiderMode.hpp
 * Defines several functions used in the Spider mode
 *
 * @since 14/03/2004 Manchester
 */

#include "Environment.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "Refutation.hpp"
#include "Signature.hpp"
#include "SpiderMode.hpp"
#include "Statistics.hpp"
#include "XML.hpp"
#include "VampireException.hpp"

using namespace Shell;

/**
 * Initialise the Spider mode.
 * @since 12/09/2004 Istanbul airport.
 */
SpiderMode::SpiderMode (Options& options)
  : Run(options,VarList())
{
} // SpiderMode::SpiderMode

/**
 * Running in the Spider mode.
 *
 * @since 14/03/2004 Manchester, copied from VampireMode
 * @since 23/10/2004 Manchester, exception handling added
 */
void SpiderMode::run ()
{
  bool resultPrinted = false;

  try {
    RunSaver saver(this);

    XML::DeepElement* proofAttempt = new XML::DeepElement("proof_attempt");
    if (_log) {
      _log->addChild(proofAttempt);
    }
    proofAttempt->addChild(_options.toXML());

    UnitList units;
    Environment::collectUnits(units);

    Problem problem(_options.problemName(),units);
    // preprocess the problem
    Property property(problem);
    property.scanProblem();

    Preprocess prepro(problem,_options,property);
    prepro.preprocess();

    // create a kernel
    Kernel kernel(_options,Signature::sig);

    int timeLimit = _options.timeLimitInDeciseconds() ? 
      _options.timeLimitInDeciseconds() : 
      INT_MAX;
    RunResult result = kernel.run(timeLimit,problem,_timer,0);
    Statistics stat(kernel);

    switch (result) {
    case REFUTATION: 
      if (_options.proof() != Options::PROOF_OFF) {
	Refutation ref(kernel,problem,_options);
	proofAttempt->addChild(ref.toXML());
      }
      cout << "+ ";
      break;

    case SATISFIABLE:
      cout << "- ";
      break;
      
    case UNKNOWN:
      cout << "? ";
      break;
    }

    cout << _options.problemName() << ' '
  	 << _timer.elapsedDeciseconds() << ' ' 
	 << _options.testId() << '\n';
    resultPrinted = true;

    // output statistics
    Statistics statistics(kernel);
    XML::Element* xmlStat = statistics.toXML();
    if (xmlStat) {
      proofAttempt->addChild(xmlStat);
    }
  
    string resultString = Run::toString(result);
    proofAttempt->addChild(new XML::TextElement("result",resultString));
    if (_log) {
      _log->addChild(new XML::TextElement("result",resultString));
    }

    BK::DestructionMode::makeFast();
    kernel.closeSession(); // not really necessary here
  }
  catch (VampireException& e) {
    if (! resultPrinted) {
      cout << "? " 
	   << _options.problemName() << ' '
 	   << _timer.elapsedDeciseconds() << ' ' 
	   << _options.testId() << '\n';
    }
  }
}  // SpiderMode::run


/**
 * Report allocation failure.
 * @since 21/03/2004 Manchester
 */
void SpiderMode::reportAllocationFailure ()
{
  cout << "? " 
       << _options.problemName() << ' '
       << "0 "
       << _options.testId() << '\n';
} // SpiderMode::reportAllocationFailure
