/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include <iostream>
#ifdef _MSC_VER
#  include <winsock2.h>
#  include <process.h>
#else 
#  include <unistd.h>
#endif
#include "Allocator.hpp"
#include "Int.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Statistics.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
Statistics::Statistics (const Kernel& kernel)
  : _kernel(kernel.kernel()),
    _statistics(const_cast<VampireKernel&>(_kernel).statistics()),
    _level(kernel.options().statistics())
{
} 
long Statistics::memoryInBytes () const
{
  return _statistics.memory + 
         GlobalAllocator::getUsedMemoryInBytes();
} 
void Statistics::output (ostream& str, const Options& opts) const
{
  TRACER("Statistics::output");
  Options::Statistics level = opts.statistics();
  if (level == Options::STATISTICS_NONE) {
    return;
  }
  str << toString();
} 
XML::Element* Statistics::toXML () const
{
  TRACER("Statistics::toXML");
  if (_level == Options::STATISTICS_NONE) {
    return 0;
  }
  bool detailed = (_level == Options::STATISTICS_FULL);
  XML::DeepElement* stat = new XML::DeepElement("statistics");
  stat->addChild(new XML::FloatElement("time", _statistics.time));
  stat->addChild(new XML::FloatElement("memory",
				       (float)(memoryInBytes()/100000)/10));
  stat->addChild(new XML::TextElement("termination_reason",
				      toString(_kernel.terminationReason())));
  if (detailed) {
    const VampireKernel::Statistics::DeductionInferences& gen = 
      _statistics.deductionInferences;
    const VampireKernel::Statistics::SimplifyingInferences& simp = 
      _statistics.simplifyingInferences;
    if (gen.totalAccomplished > 0) {
      XML::DeepElement* g = new XML::DeepElement("generating_inferences");
      stat->addChild(g);
      if (gen.resolutionAccomplished) {
	g->addChild(new XML::LongElement("resolution",
					 gen.resolutionAccomplished));
      }
      int superposition = gen.forwardSuperposition + gen.backwardSuperposition;
      if (superposition) {
	g->addChild(new XML::LongElement("superposition",superposition));
      }
      if (gen.equalityFactoring) {
	g->addChild(new XML::LongElement("equality_factoring",
					 gen.equalityFactoring));
      }
      long ers = gen.equalityResolution + simp.destructiveEqualityResolution;
      if (ers) {
	g->addChild(new XML::LongElement("equality_resolution",ers));
      }
      int chaining = gen.forwardChaining + gen.backwardChaining;
      if (chaining) {
	g->addChild(new XML::LongElement("transitivity_chaining",chaining));
      }
    }
    if (simp.totalForward || simp.totalBackward) {
      XML::DeepElement* s = new XML::DeepElement("simplifying_inferences");
      stat->addChild(s);
      if (simp.propositionalTautologyDeletion) {
	s->addChild(new XML::LongElement("propositional_tautology",
					 simp.propositionalTautologyDeletion));
      }
      if (simp.equationalTautologyDeletion) {
	s->addChild(new XML::LongElement("equational_tautology",
					 simp.equationalTautologyDeletion));
      }
      if (simp.forwardSubsumption) {
	s->addChild(new XML::LongElement("forward_subsumption",
					 simp.forwardSubsumption));
      }
      if (simp.forwardSubsumptionResolution) {
	s->addChild(new XML::LongElement("forward_subsumption_resolution",
					 simp.forwardSubsumptionResolution));
      }
      long fds = simp.forwardDemodulation + simp.onBranchForwardDemodulation;
      if (fds) {
	s->addChild(new XML::LongElement("forward_demodulation",fds));
      }
      if (simp.externalEvaluation) {
	s->addChild(new XML::LongElement("evaluation",
					 simp.externalEvaluation));
      }
      if (simp.backwardSubsumption) {
	s->addChild(new XML::LongElement("backward_subsumption",
					 simp.backwardSubsumption));
      }
      if (simp.backwardDemodulation) {
	s->addChild(new XML::LongElement("backward_demodulation",
					 simp.backwardDemodulation));
      }
      if (simp.onBranchBackwardDemodulation) {
	s->addChild(new XML::LongElement("branch_backward_demodulation",
					 simp.onBranchBackwardDemodulation));
      }
    }
    const VampireKernel::Statistics::PropositionalSplitting& prop = 
      _statistics.propositionalSplittingInferences;
    if (prop.accomplished) {
      XML::DeepElement* s = new XML::DeepElement("splitting");
      stat->addChild(s);
      s->addChild(new XML::LongElement("definition_folding",
				       prop.accomplished));
      s->addChild(new XML::FloatElement("average_number_of_components",
					prop.averageNumOfComponents));
      s->addChild(new XML::LongElement("distinct_components",
				       prop.introducedComponentNames));
    }
    const VampireKernel::Statistics::NegativeEqualitySplitting& spl = 
      _statistics.negativeEqualitySplittingInferences;
    if (spl.accomplished) {
      XML::DeepElement* s = new XML::DeepElement("inequality_splitting");
      stat->addChild(s);
      s->addChild(new XML::LongElement("definition_folding",
				       spl.accomplished));
      s->addChild(new XML::LongElement("new_symbols",
				       spl.introducedNames));
    }
    const VampireKernel::Statistics::GeneratedClauses& generated = 
      _statistics.generatedClauses;
    XML::DeepElement* g = new XML::DeepElement("generated_clauses");
    stat->addChild(g);
    g->addChild(new XML::LongElement("total",
				     generated.total));
    if (generated.totalRedundantDiscarded) {
      g->addChild(new XML::LongElement("discarded_as_redundant",
				       generated.totalRedundantDiscarded));
    }
    if (generated.totalNonredundantDiscarded) {
      g->addChild(new XML::LongElement("discarded_nonredundant",
				       generated.totalNonredundantDiscarded));
    }
    const VampireKernel::Statistics::RetainedClauses& retained = 
      _statistics.retainedClauses;
    XML::DeepElement* r = new XML::DeepElement("retained_clauses");
    stat->addChild(r);
    r->addChild(new XML::LongElement("total",
				     retained.total));
    r->addChild(new XML::LongElement("currently_active",
				     retained.currentlyActive));
    r->addChild(new XML::LongElement("currently_passive",
				     retained.currentlyPassive));
    r->addChild(new XML::LongElement("selected",
				     retained.selected));
    if (retained.totalOrphansMurdered) {
      r->addChild(new XML::LongElement("orphan_deletion",
				       retained.totalOrphansMurdered));
    }
    if (retained.totalRecycled) {
      r->addChild(new XML::LongElement("recycled",
				       retained.totalRecycled));
    }
  }
  return stat;
} 
string Statistics::toString (VampireKernel::TerminationReason reason)
{
  TRACER("Statistics::toString");
  switch (reason) 
    {
    case VampireKernel::UnknownTerminationReason:
      return "unknown";
    case VampireKernel::RefutationFound:
      return "refutation found";
    case VampireKernel::Satisfiable:
      return "satisfiability detected";
    case VampireKernel::EmptyPassive:
      return "no passive clauses left";
    case VampireKernel::MaxPassive:
      return "max_passive exceeded";
    case VampireKernel::MaxActive:
      return "max_active exceeded";
    case VampireKernel::TimeLimitExpired:
      return "time limit expired";
    case VampireKernel::AbortedOnAllocationFailure:
      return "allocation failure";
    case VampireKernel::AbortedOnExternalInterrupt:
      return "external interrupt";
    case VampireKernel::AbortedOnErrorSignal:
      return "error signal";
    case VampireKernel::AbortedOnError:
      return "error";
#if DEBUG_SHELL
    default:
      ASSERTION_VIOLATION;
#endif
    }
} 
string Statistics::toString () const
{
  TRACER("Statistics::toString");
  bool detailed = (_level == Options::STATISTICS_FULL);
  string str = "=========== Statistics ==========\n";
  if (detailed) {
    str += (string)"version: " + Int::toString(_kernel.versionNumber()) + " " +
	   _kernel.versionDescription() + "\n=== General:\n";
  }
  str += (string)"time: " + Int::toString(_statistics.time) + 
         "s\nmemory: " + Int::toString(((float)(memoryInBytes()/100000))/10) + 
         "Mb\ntermination reason: " + toString(_kernel.terminationReason()) + 
         '\n';
   if (detailed) {
    const VampireKernel::Statistics::DeductionInferences& gen = 
      _statistics.deductionInferences;
    if (gen.totalAccomplished > 0) {
      str += "=== Generating inferences:\n";
      if (gen.resolutionAccomplished) {
	str += (string)"resolution: " + 
	       Int::toString(gen.resolutionAccomplished)+ '\n';
      }
      int superposition = gen.forwardSuperposition + gen.backwardSuperposition;
      if (superposition) {
	str += (string)"superposition: " + 
               Int::toString(superposition)+ '\n';
      }
      if (gen.equalityFactoring) {
	str += (string)"equality_factoring: " + 
               Int::toString(gen.equalityFactoring)+ '\n';
      }
      if (gen.equalityResolution) {
	str += (string)"equality_resolution: " + 
               Int::toString(gen.equalityResolution)+ '\n';
      }
      int chaining = gen.forwardChaining + gen.backwardChaining;
      if (chaining) {
	str += (string)"transitivity_chaining: " + 
               Int::toString(chaining)+ '\n';
      }
    }
    const VampireKernel::Statistics::SimplifyingInferences& simp = 
      _statistics.simplifyingInferences;
    if (simp.totalForward || simp.totalBackward) {
      str += "=== Simplifying inferences:\n";
      if (simp.propositionalTautologyDeletion) {
	str += (string)"propositional_tautology: " + 
               Int::toString(simp.propositionalTautologyDeletion)+ '\n';
      }
      if (simp.equationalTautologyDeletion) {
	str += (string)"equational_tautology: " + 
               Int::toString(simp.equationalTautologyDeletion)+ '\n';
      }
      if (simp.forwardSubsumption) {
	str += (string)"forward_subsumption: " + 
               Int::toString(simp.forwardSubsumption)+ '\n';
      }
      if (simp.forwardSubsumptionResolution) {
	str += (string)"forward_subsumption_resolution: " + 
               Int::toString(simp.forwardSubsumptionResolution)+ '\n';
      }
      if (simp.forwardDemodulation) {
	str += (string)"forward_demodulation: " + 
               Int::toString(simp.forwardDemodulation)+ '\n';
      }
      if (simp.onBranchForwardDemodulation) {
	str += (string)"branch_forward_demodulation: " + 
               Int::toString(simp.onBranchForwardDemodulation)+ '\n';
      }
      if (simp.destructiveEqualityResolution) {
	str += (string)"equality_resolution: " + 
               Int::toString(simp.destructiveEqualityResolution)+ '\n';
      }
      if (simp.externalEvaluation) {
	str += (string)"evaluation: " + 
               Int::toString(simp.externalEvaluation)+ '\n';
      }
      if (simp.backwardSubsumption) {
	str += (string)"backward_subsumption: " + 
               Int::toString(simp.backwardSubsumption)+ '\n';
      }
      if (simp.backwardDemodulation) {
	str += (string)"backward_demodulation: " + 
               Int::toString(simp.backwardDemodulation)+ '\n';
      }
      if (simp.onBranchBackwardDemodulation) {
	str += (string)"branch_backward_demodulation: " + 
               Int::toString(simp.onBranchBackwardDemodulation) + '\n';
      }
    }
    const VampireKernel::Statistics::PropositionalSplitting& prop = 
      _statistics.propositionalSplittingInferences;
    if (prop.accomplished) {
      str += (string)"=== Splitting:\n";
      str += (string)"definition_folding: " + 
	Int::toString(prop.accomplished) + '\n';
      str += (string)"average_number_of_components: " + 
             Int::toString(prop.averageNumOfComponents)+ '\n';
      str += (string)"distinct_components: " + 
             Int::toString(prop.introducedComponentNames)+ '\n';
    }
    const VampireKernel::Statistics::NegativeEqualitySplitting& spl = 
      _statistics.negativeEqualitySplittingInferences;
    if (spl.accomplished) {
      str += "=== Inequality splitting:\n";
      str += (string)"definition_folding: " + Int::toString(spl.accomplished)+ '\n';
      str += (string)"new_symbols: " + Int::toString(spl.introducedNames)+ '\n';
    }
     const VampireKernel::Statistics::GeneratedClauses& generated = 
       _statistics.generatedClauses;
    str += "=== Generated clauses:\n";
     str += (string)"total: " + Int::toString(generated.total)+ '\n';
    if (generated.totalRedundantDiscarded) {
      str += (string)"discarded_as_redundant: " + 
             Int::toString(generated.totalRedundantDiscarded)+ '\n';
    }
    if (generated.totalNonredundantDiscarded) {
      str += (string)"discarded_nonredundant: " + 
             Int::toString(generated.totalNonredundantDiscarded)+ '\n';
    }
    const VampireKernel::Statistics::RetainedClauses& retained = 
      _statistics.retainedClauses;
    str += "=== Retained clauses:\n";
    str += (string)"total: " + Int::toString(retained.total)+ '\n';
    str += (string)"selected: " + Int::toString(retained.selected)+ '\n';
    str += (string)"currently_active: " + 
           Int::toString(retained.currentlyActive)+ '\n';
    str += (string)"currently_passive: " + 
           Int::toString(retained.currentlyPassive)+ '\n';
    if (retained.totalOrphansMurdered) {
      str += (string)"orphans_murdered: " + 
             Int::toString(retained.totalOrphansMurdered)+ '\n';
    }
    if (retained.totalRecycled) {
      str += (string)"recycled: " + Int::toString(retained.totalRecycled)+ '\n';
    }
  }
  str += "======= End of statistics =======\n";
  return str;
} 
