/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Environment.hpp"
#include "Int.hpp"
#include "TPTPLexer.hpp"
#include "TPTPParser.hpp"
#include "Signature.hpp"
#include "BlodKorv/Stack.hpp"
#include "Tracer.hpp"
#include "Unit.hpp"
#include "XML.hpp"
namespace Shell {
class TPTPParser::UnitStack
  : public Stack<Unit>
{
public:
  UnitStack(UnitList& units)
    : Stack<Unit>(units)
  {}
};
TPTPParser::TPTPParser (TPTPLexer& lexer)
  : Parser(lexer)
{
}
void TPTPParser::units (UnitList& us)
{
  TRACER("TPTPParser::units(UnitList&)");
  UnitStack stack(us);
  units(stack);
} 
void TPTPParser::units (UnitStack& stack)
{
  TRACER("TPTPParser::units(UnitStack&)");
  while (currentToken().tag != TT_EOF) {
    unit(stack);
  }
} 
void TPTPParser::unit (UnitStack& stack)
{
  TRACER("TPTPParser::unit");
  TokenType tag = currentToken().tag;
  switch (tag) {
  case TT_INPUT_FORMULA:
  case TT_INPUT_CLAUSE:
  case TT_CNF:
    break;
  case TT_INCLUDE:
    consumeToken();
    include(stack);
    return;
  default:
    throw ParserException("either input_formula or input_clause expected",
			  currentToken());
  }
  consumeToken();
  consumeToken(TT_LPAR);
  string nm = name();
  consumeToken(TT_COMMA);
  string tp = name();
  InputType it;
  if (tp == "axiom") {
    it = AXIOM;
  }
  else if (tp == "conjecture") {
    it = CONJECTURE;
  }
  else if (tp == "negated_conjecture") {
    it = CONJECTURE;
  }
  else if (tp == "hypothesis") {
    it = ASSUMPTION;
  }
  else if (tp == "lemma") {
    it = ASSUMPTION;
  }
  else {
    throw ParserException("axiom, conjecture or hypothesis expected", 
			  currentToken());
  }
  consumeToken(TT_COMMA);
  switch (tag) {
  case TT_INPUT_FORMULA:
    {
      Formula f;
      formula(f);
      if (it == CONJECTURE) {
	VarList vs;
	f.freeVariables(vs);
	if (vs.isEmpty()) {
	  f = Formula(NOT,f);
	}
	else {
	  f = Formula(NOT,Formula(FORALL,vs,f));
	}
      }
      stack.push(Unit(nm,it,f));
    }
    break;
  case TT_INPUT_CLAUSE:
    {
      Clause c;
      clause(c);
      stack.push(Unit(nm,it,c));
    }
    break;
  case TT_CNF:
    {
      Clause c;
      formulaClause(c);
      stack.push(Unit(nm,it,c));
    }
    break;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
  consumeToken(TT_RPAR);
  consumeToken(TT_DOT);
} 
void TPTPParser::formula (Formula& f)
{
  xorFormula(f);
  Token& t = currentToken();
  switch (t.tag) {
  case TT_IFF:
    {
      consumeToken();
      Formula rhs;
      formula (rhs);
      f = Formula(IFF, f, rhs);
    }
    return;
  default:
    return;
  }
} 
string TPTPParser::name ()
{
  Token& token = currentToken();
  if (token.tag != TT_NAME) {
    throw ParserException("name expected",
			  currentToken());
  }
  consumeToken();
  return token.text;
} 
void TPTPParser::clause (Clause& c)
{
  consumeToken(TT_LBRA);
  LiteralList ls;
  literals(ls);
  consumeToken(TT_RBRA);
  c = Clause(ls);
} 
void TPTPParser::formulaClause (Clause& c)
{
  bool par = false;
  if (currentToken().tag == TT_LPAR) { 
    consumeToken();
    par = true;
  }
  LiteralList ls;
  formulaLiterals(ls);
  c = Clause(ls);
  if (par) {
    consumeToken(TT_RPAR);
  }
} 
void TPTPParser::xorFormula (Formula& f)
{
  impFormula(f);
  switch (currentToken().tag) {
  case TT_XOR:
    {
      consumeToken();
      Formula rhs;
      xorFormula (rhs);
      f = Formula(XOR, f, rhs);
    }
    return;
  default:
    return;
  }
} 
void TPTPParser::impFormula (Formula& f)
{
  orFormula(f);
  Token& t = currentToken();
  switch (t.tag) {
  case TT_IMP:
    {
      consumeToken();
      Formula rhs;
      impFormula (rhs);
      f = Formula(IMP, f, rhs);
    }
    return;
  case TT_RIMP:
    {
      consumeToken();
      Formula lhs;
      impFormula(lhs);
      f = Formula(IMP, lhs, f);
    }
    return;
  default:
    return;
  }
} 
void TPTPParser::orFormula (Formula& f)
{
  andFormula(f);
  switch (currentToken().tag) {
  case TT_OR:
    {
      consumeToken();
      Formula rhs;
      orFormula (rhs);
      f.makeJunction(OR,f,rhs);
    }
    return;
  default:
    return;
  }
} 
void TPTPParser::andFormula (Formula& f)
{
  simpleFormula(f);
  switch (currentToken().tag) {
  case TT_AND:
    {
      consumeToken();
      Formula rhs;
      andFormula (rhs);
      f.makeJunction(AND,f,rhs);
    }
    return;
  default:
    return;
  }
} 
void TPTPParser::simpleFormula (Formula& f)
{
  TokenType tt = currentToken().tag;
  switch (tt) {
  case TT_NOT:
    consumeToken();
    simpleFormula(f);
    f = Formula(NOT,f);
    return;
  case TT_FORALL:
  case TT_EXISTS:
    {
      consumeToken();
      consumeToken(TT_LBRA);
      VarList vs;
      varList(vs);
      consumeToken(TT_RBRA);
      consumeToken(TT_COLON);
      simpleFormula(f);
      f = Formula(tt == TT_FORALL ? FORALL : EXISTS,vs,f);
    }
    return;
  case TT_LPAR:
    consumeToken();
    formula(f);
    consumeToken(TT_RPAR);
    return;
  case TT_TRUE:
    consumeToken();
    f = Formula(true);
    return;
  case TT_FALSE:
    consumeToken();
    f = Formula(false);
    return;
  default: 
    {
      bool sign;
      Atom a;
      atom(a,sign);
      if (sign) {
	f = Formula(a);
      }
      else {
	f = Formula(NOT,Formula(a));
      }
    }
    return;
  }
} 
void TPTPParser::varList (VarList& vs)
{
  Token& token = currentToken();
  if (token.tag != TT_VAR) {
    throw ParserException("variable expected",token);
  }
  Var v(var(token));
  consumeToken();
  vs = VarList(v);
  if (currentToken().tag != TT_COMMA) {
    return;
  }
  consumeToken();
  varList(const_cast<VarList&>(vs.tail()));
} 
void TPTPParser::atom (Atom& a,bool& positive)
{
  positive = true;
  Token& token = currentToken();
  switch (token.tag) {
  case TT_NAME: 
    {
      string functor = name();
      TermList ts;
      args(ts);
      switch (currentToken().tag) {
      case TT_NEQ:
	positive = false;
      case TT_EQUAL:
	{
	  consumeToken();
	  Term t2;
	  term(t2);
	  Term t1(Signature::sig.createFun(functor,ts.length()),ts);
	  a = Atom(t1,t2);
	}
	return;
      default:
	a = Atom(Signature::sig.createPred(functor,ts.length()),ts);
	return;
      }
    }
  case TT_TRUE: 
  case TT_FALSE: 
    throw ParserException("LOOK AT ME, THERE IS A PROBLEM, BUDDY!",token);
  case TT_REAL:
  case TT_INTEGER:
  case TT_VAR:
    {
      Term t1;
      term(t1);
      switch (currentToken().tag) {
      case TT_NEQ:
	positive = false;
      case TT_EQUAL:
	{
	  consumeToken();
	  Term t2;
	  term(t2);
	  a = Atom(t1,t2);
	}
	return;
      default:
	throw ParserException("= or != expected after term",token);
      }
    }
  default:
    throw ParserException("atom expected",token);
  }
} 
void TPTPParser::term (Term& t)
{
  Token& token = currentToken();
  switch (token.tag) {
  case TT_NAME:
    {
      string functor = token.text;
      consumeToken();
      TermList ts;
      args(ts);
      t = Term(Signature::sig.createFun(functor,ts.length()),ts);
    }
    return;
  case TT_INTEGER:
    t = Term(integer(token));
    consumeToken();
    return;
  case TT_REAL:
    t = Term(real(token));
    consumeToken();
    return;
  case TT_VAR:
    {
      Var v(var(token));
      consumeToken();
      t = Term(v);
    }
    return;
  default:
    throw ParserException("term expected",token);
  }
} 
void TPTPParser::terms (TermList& ts)
{
  Term t;
  term(t);
  if (currentToken().tag == TT_COMMA) {
    consumeToken();
    terms(ts);
  }
  ts = TermList(t,ts);
} 
void TPTPParser::args (TermList& ts)
{
  if (currentToken().tag != TT_LPAR) {
    return;
  }
  consumeToken();
  terms(ts);
  consumeToken(TT_RPAR);
} 
void TPTPParser::literals (LiteralList& ls)
{
  Literal l;
  literal(l);
  switch (currentToken().tag) {
  case TT_COMMA:
    consumeToken();
    literals(ls);
    break;
  default:
    break;
  }
  ls = LiteralList(l,ls);
} 
void TPTPParser::formulaLiterals (LiteralList& ls)
{
  switch (currentToken().tag) {
  case TT_FALSE:
    consumeToken();
    return;
  default:
    break;
  }
  Literal l;
  formulaLiteral(l);
  switch (currentToken().tag) {
  case TT_OR:
    consumeToken();
    formulaLiterals(ls);
    break;
  default:
    break;
  }
  ls = LiteralList(l,ls);
} 
void TPTPParser::literal (Literal& l)
{
  bool sign = false;
  Token& token = currentToken();
  switch (token.tag) {
  case TT_PP:
    sign = true;
    break;
  case TT_MM:
    break;
  default:
    throw ParserException("either ++ or -- expected",token);
  }
  consumeToken();
  bool atomSign;
  Atom a;
  atom(a,atomSign);
  if (! atomSign) {
    sign = ! sign;
  }
  l = Literal(sign,a);
} 
void TPTPParser::formulaLiteral (Literal& l)
{
  bool sign = true;
  Token& token = currentToken();
  switch (token.tag) {
  case TT_NOT:
    consumeToken();
    sign = false;
    break;
  default:
    break;
  }
  bool atomSign;
  Atom a;
  atom(a,atomSign);
  if (! atomSign) {
    sign = ! sign;
  }
  l = Literal(sign,a);
} 
XML::Element* ParserException::toXML () const
{
  return new XML::TextElement("parser_error",_message);
} 
void TPTPParser::include (UnitStack& stack)
{
  consumeToken(TT_LPAR);
  Token& token = currentToken();
  if (token.tag != TT_QUOTED_STRING) {
    throw ParserException("quoted string expected",token);
  }
  string fileName(Environment::include + token.text);
  consumeToken();
  consumeToken(TT_RPAR);
  consumeToken(TT_DOT);
  ifstream in(fileName.c_str());
  if (! in) {
    USER_ERROR((string)"cannot open file" + fileName);
  }
  TPTPLexer lexer(in);
  TPTPParser parser(lexer);
  parser.units(stack);
} 
}
