/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "CodeGenerator.hpp"
#include "Inference.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "PureLiteral.hpp"
#include "Timer.hpp"
#include "Tracer.hpp"
#include "TheoryFinder.hpp"
#include "Unit.hpp"
#include "VampireException.hpp"
#define DEBUG_THEORY_FINDER 0
namespace Shell {
class TheoryFinder::MatcherState 
{
public:
  MatcherState (const char* code,int* vars,const Symbol** sym);
  int match (int cp,const Term&);
  int match (int cp,const TermList&);
  bool addNewVar (int index,int var);
  bool addNewSymbol (int index,const Symbol& symb);
  const char* code;
  int* vars;
  const Symbol** sym;
};
class TheoryFinder::Matcher 
{
public:
  Matcher (MatcherState& state,int cp) 
    : _state(state), _cp(cp)
  {}
  virtual int nextMatch() = 0;
  virtual ~Matcher () {}
  static Matcher* createFormulaMatcher(MatcherState& state,int cp,const Formula& f);
  static Matcher* createQuantifiedFormulaMatcher(MatcherState& state,int cp,const Formula& f);
  static Matcher* createFormulaPairMatcher(MatcherState& state,int cp,
					   const Formula& formula1,const Formula& formula2);
protected:
  MatcherState& _state;
  int _cp;
};
class TheoryFinder::AtomMatcher 
  : public Matcher
{
public:
  AtomMatcher (MatcherState& state,int cp,const Atom& atom);
  int nextMatch();
  ~AtomMatcher() {}
private:
  const Atom& _atom;
  int _attempts;
};
class TheoryFinder::FormulaPairMatcher 
  : public Matcher
{
public:
  FormulaPairMatcher(MatcherState& state,
		     int cp,
		     Matcher* matcher1,
		     const Formula& formula2);
  int nextMatch();
  ~FormulaPairMatcher();
private:
  const Formula& _f2;
  Matcher* _fm1;
  Matcher* _fm2;
}; 
class TheoryFinder::LiteralListMatcher 
  : public Matcher
{
public:
  LiteralListMatcher (MatcherState& state,int cp,const LiteralList& lits);
  int nextMatch();
  ~LiteralListMatcher();
private:
  LiteralList _reserve;
  LiteralList _tried;
  AtomMatcher* _headMatcher;
  LiteralListMatcher* _tailMatcher;
};
class TheoryFinder::FormulaListMatcher 
  : public Matcher
{
public:
  FormulaListMatcher (MatcherState& state,int cp,const FormulaList& lits);
  int nextMatch();
  ~FormulaListMatcher();
private:
  FormulaList _reserve;
  FormulaList _tried;
  Matcher* _headMatcher;
  FormulaListMatcher* _tailMatcher;
}; 
class TheoryFinder::ClauseMatcher 
  : public Matcher
{
public:
  ClauseMatcher (MatcherState& state,const LiteralList& clause);
  int nextMatch();
  ~ClauseMatcher();
private:
  LiteralListMatcher* _delegate;
}; 
TheoryFinder::TheoryFinder (const Problem& problem,Property* property)
  : _problem(problem),
    _property(property)
{
  TRACER("TheoryFinder::TheoryFinder");
} 
TheoryFinder::~TheoryFinder ()
{
  TRACER("TheoryFinder::~TheoryFinder");
} 
int TheoryFinder::search ()
{
  TRACER("TheoryFinder::search");
  int found = 0;
  UnitChain::Iterator units(_problem.units());
  while (units.hasNext()) {
    if (matchAll(units.next())) {
      found++;
    }
  }
  if (found > 0) {
    PredicateSymbol& answer =_sig.createPred("answer",1);
    answer.makeAnswer();
    addTheory();
    Problem problem("",_metaTheory);
    PureLiteral pureLiteralEliminator(problem,_sig);
    pureLiteralEliminator.findAndRemove();
    if (problem.units().isEmpty()) {
#if DEBUG_THEORY_FINDER
      cout << "no clauses left\n";
#endif
    }
    else {
      Options options;
      options.setBackwardSubsumption(false);
      Timer timer;
      timer.start();
      Kernel kernel(options,_sig);
      RunResult result = kernel.run(INT_MAX,problem,timer,0);
      switch (result) {
      case REFUTATION: 
	{ 
	  Clause clause;
	  kernel.getClause(kernel.refutation(),clause);
	  analyse(clause);
	  while (kernel.rerun(INT_MAX) == REFUTATION) {
	    kernel.getClause(kernel.refutation(),clause);
	    analyse(clause);
	  }
#if DEBUG_THEORY_FINDER
	  cout << "Theory found. Thanks to Tanya!\n";
#endif
	}
	break;
      case SATISFIABLE:
#if DEBUG_THEORY_FINDER
	cout << "No theory found\n";
#endif
	break;
#if DEBUG_SHELL
      default:
	ASSERTION_VIOLATION;
#endif
      }
    }
    BK::DestructionMode::makeFast();
  }
  BK::DestructionMode::makeThorough();
  return found;
} 
bool TheoryFinder::matchAll (const Unit& u)
{
  TRACER("TheoryFinder::matchAll (const Unit& u)");
  Unit unit(u);
  _currentUnit = &unit;
  switch (unit.unitType()) {
  case CLAUSE:
    {
      const Clause& clause = u.clause();
      const LiteralList& ls = clause.literals();
      if (ls.hasLength(1) && ls.head().isPositive()) { 
	return matchAll(ls.head().atom());
      }
      return matchAll(unit.clause().literals());
    }
  case FORMULA:
    {
      Formula f(u.formula()); 
      stripForall(f);
      if (f.connective() == ATOM) {
	return matchAll(f.atom());
      }
      return matchAll(f);
    }
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
bool TheoryFinder::matchAll (const LiteralList& literals)
{
  TRACER("TheoryFinder::matchAll (const LiteralList&...)");
  return matchSubset(literals) ||
         matchCondensedDetachment1(literals) ||
         matchCondensedDetachment2(literals) ||
         matchExtensionality(literals) ||
         matchFLD1(literals) ||
         matchFLD2(literals);
} 
bool TheoryFinder::matchAll (const Formula& formula)
{
  TRACER("TheoryFinder::matchAll (const Formula&...)");
  Formula f(formula);
  return matchExtensionality(f) ||
         matchSubset(f);
} 
void TheoryFinder::stripForall (Formula& f)
{
  TRACER("TheoryFinder::stripForall");
  while (f.connective() == FORALL) {
    f = f.qarg();
  }
} 
bool TheoryFinder::matchC (const Atom& atom)
{
  TRACER("TheoryFinder::matchC");
  static const char code[] = 
  {EQL, 
    NEWSYM,0,2,NEWVAR,0,NEWVAR,1,  
    OLDSYM,0,OLDVAR,1,OLDVAR,0};   
  return matchCode(atom,code,"commutativity",1,0);
} 
bool TheoryFinder::matchA (const Atom& atom)
{
  TRACER("TheoryFinder::matchA");
  static const char code[] = 
  {EQL, 
    NEWSYM,0,2,OLDSYM,0,
    NEWVAR,0,NEWVAR,1,NEWVAR,2,  
    OLDSYM,0,OLDVAR,0,OLDSYM,0,
    OLDVAR,1,OLDVAR,2};   
  return matchCode(atom,code,"associativity",1,0);
} 
TheoryFinder::MatcherState::MatcherState (const char* c,
					  int* vs,
					  const Symbol** ss)
  : code(c),
    vars(vs),
    sym(ss)
{
  TRACER("TheoryFinder::MatcherState::MatcherState");
}
int TheoryFinder::AtomMatcher::nextMatch ()
{
  TRACER("TheoryFinder::AtomMatcher::nextMatch");
  const char* code = _state.code+_cp;
  for (;;) {
    switch (_attempts) {
    case 0:
      switch (code[0]) {
      case EQL:
	if (_atom.isEquality()) {
	  _attempts ++;
	  int newCp = _state.match(_cp+1,_atom.args());
	  if (newCp) {
	    return newCp;
	  }
	}
	else {
	  return 0;
	}
	break;
      case NEWSYM:
	if (_atom.isEquality()) {
	  return 0;
	}
	if (_atom.functor().arity() != code[2]) {
	  return 0;
	}
	_attempts++;
	if (!_state.addNewSymbol(code[1],_atom.functor())) {
	  return 0;
	}
	return _state.match(_cp+3,_atom.args());
      case OLDSYM:
	if (_state.sym[code[1]] != &_atom.functor()) {
	  return 0;
	}
	_attempts ++;
	return _state.match(_cp+2,_atom.args());
      default:
	return false;
      }
    case 1:
      if (code[0] == EQL) {
	int newCp = _state.match(_cp+1,_atom.args().second());
	if (! newCp) {
	  return 0;
	}
	_attempts++;
	return _state.match(newCp,_atom.args().head());
      }
      return 0;
    default:
      ASS(_attempts == 2);
      return 0;
    }
  }
} 
int TheoryFinder::MatcherState::match (int cp,const Term& t)
{
  TRACER("TheoryFinder::MatcherState::match");
  const char* cd = code + cp;
  switch (cd[0]) {
  case NEWVAR:
    if (t.tag() != VAR) {
      return 0;
    }
    if (! addNewVar(cd[1],t.var().toInt())) {
      return 0;
    }
    return cp+2;
  case OLDVAR:
    if (t.tag() != VAR) {
      return 0;
    }
    if (vars[cd[1]] != t.var().toInt()) {
      return 0;
    }
    return cp+2;
  case NEWSYM:
    if (t.tag() != COMPOUND) {
      return 0;
    }
    if (t.args().length() != cd[2]) {
      return 0;
    }
    if (!addNewSymbol(cd[1],t.functor())) {
      return 0;
    }
    return match(cp+3,t.args());
  case OLDSYM:
    if (t.tag() != COMPOUND) {
      return 0;
    }
    if (sym[cd[1]] != &t.functor()) {
      return 0;
    }
    return match(cp+2,t.args());
  default:
    return false;
  }
} 
int TheoryFinder::MatcherState::match (int cp,const TermList& ts)
{
  TRACER("TheoryFinder::MatcherState::match");
  TermList terms(ts);
  while (terms.isNonEmpty()) {
    cp = match(cp,terms.head());
    if (! cp) {
      return 0;
    }
    terms.pop();
  }
  return cp;
} 
TheoryFinder::AtomMatcher::AtomMatcher (MatcherState& state,
					int cp,
					const Atom& atom)
  : Matcher(state,cp),
    _atom(atom),
    _attempts(0)
{
  TRACER("TheoryFinder::AtomMatcher::AtomMatcher");
} 
bool TheoryFinder::MatcherState::addNewSymbol (int index,const Symbol& s)
{
  TRACER("TheoryFinder::MatcherState::addNewSymbol");
  for (int i = index-1;i >=0;i--) {
    if (sym[i] == &s) {
      return false;
    }
  }
  sym[index] = &s;
  return true;
} 
bool TheoryFinder::MatcherState::addNewVar (int index,int var)
{
  TRACER("TheoryFinder::MatcherState::addNewVar");
  for (int i = index-1;i >=0;i--) {
    if (vars[i] == var) {
      return false;
    }
  }
  vars[index] = var;
  return true;
} 
bool TheoryFinder::matchExtensionality (const LiteralList& c)
{
  TRACER("TheoryFinder::matchExtensionality (const LiteralList&...)");
  static char code[] =
  {3,                                      
    NEWSYM,0,2,                            
     NEWSYM,1,2,NEWVAR,0,NEWVAR,1,OLDVAR,0,
    OLDSYM,0,                              
     OLDSYM,1,OLDVAR,0,OLDVAR,1,OLDVAR,1,
    EQL,OLDVAR,0,OLDVAR,1}; 
  return matchCode(c,code,"extensionality",1,PR_HAS_EXTENSIONALITY);
} 
bool TheoryFinder::matchCondensedDetachment1 (const LiteralList& c)
{
  TRACER("TheoryFinder::CondensedDetachment1(const LiteralList&...)");
  static char code[] =
  {3,                                           
   NEWSYM,0,1,NEWVAR,0,                         
   CNOT,OLDSYM,0,NEWVAR,1,                      
   CNOT,OLDSYM,0,NEWSYM,1,2,OLDVAR,1,OLDVAR,0}; 
  return matchCode(c,code,"cd1",2,PR_HAS_CONDENSED_DETACHMENT1);
} 
bool TheoryFinder::matchCondensedDetachment2 (const LiteralList& c)
{
  TRACER("TheoryFinder::CondensedDetachment2(const LiteralList&...)");
  static char code[] =
  {3,                                                      
   NEWSYM,0,1,NEWVAR,0,                                    
   CNOT,OLDSYM,0,NEWVAR,1,                                 
   CNOT,OLDSYM,0,NEWSYM,1,2,NEWSYM,2,1,OLDVAR,1,OLDVAR,0}; 
  return matchCode(c,code,"cd2",3,PR_HAS_CONDENSED_DETACHMENT2);
} 
bool TheoryFinder::matchFLD1 (const LiteralList& c)
{
  TRACER("TheoryFinder::matchFLD1(const LiteralList&...)");
  static char code[] =
  {4,                                                      
   NEWSYM,0,2,NEWSYM,1,2,NEWSYM,2,2,NEWVAR,0,NEWVAR,1,     
   OLDSYM,2,NEWVAR,2,OLDVAR,1,                             
   OLDSYM,2,OLDSYM,1,OLDVAR,0,OLDVAR,2,OLDVAR,1,           
   CNOT,NEWSYM,3,1,OLDVAR,0,                               
   CNOT,OLDSYM,3,OLDVAR,2,                                 
   CNOT,OLDSYM,3,OLDVAR,1};                                
  return matchCode(c,code,"fld1",4,PR_HAS_FLD1);
} 
bool TheoryFinder::matchFLD2 (const LiteralList& c)
{
  TRACER("TheoryFinder::matchFLD2(const LiteralList&...)");
  static char code[] =
  {3,                                                      
   NEWSYM,0,3,NEWSYM,1,1,NEWVAR,0,OLDVAR,0,NEWSYM,2,0,     
   NEWSYM,3,3,NEWSYM,4,0,OLDVAR,0,OLDSYM,4,                
   CNOT,NEWSYM,5,1,OLDVAR,0};                              
  return matchCode(c,code,"fld2",4,PR_HAS_FLD2);
} 
bool TheoryFinder::matchSubset (const LiteralList& c)
{
  TRACER("TheoryFinder::matchSubset");
  static char code[] =
  {2,                                
    NEWSYM,0,2,                 
     NEWSYM,2,2,NEWVAR,0,NEWVAR,1,OLDVAR,0,
    NEWSYM,1,2,OLDVAR,0,OLDVAR,1}; 
  return matchCode(c,code,"subset",2,PR_HAS_SUBSET);
} 
TheoryFinder::ClauseMatcher::ClauseMatcher (MatcherState& state,
					    const LiteralList& clause)
  : Matcher(state,0)
{
  TRACER("TheoryFinder::ClauseMatcher::ClauseMatcher");
  _delegate = clause.hasLength(state.code[0]) ?
              new LiteralListMatcher(state,1,clause) : 
              0;
} 
int TheoryFinder::ClauseMatcher::nextMatch ()
{
  TRACER("TheoryFinder::ClauseMatcher::nextMatch");
  if (_delegate) {
    return _delegate->nextMatch();
  }
  return 0;
} 
TheoryFinder::ClauseMatcher::~ClauseMatcher ()
{
  TRACER("TheoryFinder::ClauseMatcher::~ClauseMatcher");
  delete _delegate;
} 
TheoryFinder::LiteralListMatcher::LiteralListMatcher (MatcherState& state,
						      int cp,
						      const LiteralList& lits)
  : Matcher(state,cp),
    _reserve(lits),
    _headMatcher(0),
    _tailMatcher(0)
{
  TRACER("TheoryFinder::LiteralListMatcher::LiteralListMatcher");
} 
TheoryFinder::FormulaListMatcher::FormulaListMatcher (MatcherState& state,
						      int cp,
						      const FormulaList& fs)
  : Matcher(state,cp),
    _reserve(fs),
    _headMatcher(0),
    _tailMatcher(0)
{
  TRACER("TheoryFinder::FormulaListMatcher::FormulaListMatcher");
} 
TheoryFinder::LiteralListMatcher::~LiteralListMatcher ()
{
  TRACER("TheoryFinder::LiteralListMatcher::~LiteralListMatcher");
  delete _headMatcher;
  delete _tailMatcher;
} 
TheoryFinder::FormulaListMatcher::~FormulaListMatcher ()
{
  TRACER("TheoryFinder::FormulaListMatcher::~FormulaListMatcher");
  delete _headMatcher;
  delete _tailMatcher;
} 
int TheoryFinder::LiteralListMatcher::nextMatch ()
{
  TRACER("TheoryFinder::LiteralListMatcher::nextMatch");
  for (;;) {
    if (_tailMatcher) {
      int cp = _tailMatcher->nextMatch();
      if (cp) {
	return cp;
      }
      delete _tailMatcher;
      _tailMatcher = 0;
    }
    else if (_headMatcher) {
      int cp = _headMatcher->nextMatch();
      if (cp) {
	if (_tried.isEmpty() && _reserve.tail().isEmpty()) {
	  return cp;
	}
	LiteralList remaining(_tried);
	remaining.append(_reserve.tail());
	_tailMatcher = new LiteralListMatcher(_state,cp,remaining);
      }
      else { 
	delete _headMatcher;
	_headMatcher = 0;
	_tried.push(_reserve.head());
	_reserve.pop();
      }
    }
    else if (_reserve.isEmpty()) {
      return 0;
    }
    else if (_state.code[_cp] == CNOT) { 
      const Literal& hd = _reserve.head();
      if (hd.isPositive()) { 
	_tried.push(_reserve.head());
	_reserve.pop();
      }
      else {
	_headMatcher = new AtomMatcher(_state,_cp+1,hd.atom());
      }
    }
    else { 
      const Literal& hd = _reserve.head();
      if (hd.isNegative()) { 
	_tried.push(_reserve.head());
	_reserve.pop();
      }
      else {
	_headMatcher = new AtomMatcher(_state,_cp,hd.atom());
      }
    }
  }
} 
int TheoryFinder::FormulaListMatcher::nextMatch ()
{
  TRACER("TheoryFinder::FormulaListMatcher::nextMatch");
  for (;;) {
    if (_tailMatcher) {
      int cp = _tailMatcher->nextMatch();
      if (cp) {
	return cp;
      }
      delete _tailMatcher;
      _tailMatcher = 0;
    }
    else if (_headMatcher) {
      int cp = _headMatcher->nextMatch();
      if (cp) {
	if (_tried.isEmpty() && _reserve.tail().isEmpty()) {
	  return cp;
	}
	FormulaList remaining(_tried);
	remaining.append(_reserve.tail());
	_tailMatcher = new FormulaListMatcher(_state,cp,remaining);
      }
      else { 
	delete _headMatcher;
	_headMatcher = 0;
	_tried.push(_reserve.head());
	_reserve.pop();
      }
    }
    else if (_reserve.isEmpty()) {
      return 0;
    }
    else { 
      _headMatcher = createFormulaMatcher(_state,_cp,_reserve.head());
      if (! _headMatcher) {
	_tried.push(_reserve.head());
	_reserve.pop();
      }
    }
  }
} 
bool TheoryFinder::matchSubset (const Formula& f)
{
  TRACER("TheoryFinder::matchSubset");
  static char code[] =
    {CIFF,                          
      NEWSYM,1,2,NEWVAR,0,NEWVAR,1, 
      CFORALL,1,2,CIMP,             
       NEWSYM,0,2,OLDVAR,2,OLDVAR,0,
       OLDSYM,0,OLDVAR,2,OLDVAR,1}; 
  return matchCode(f,code,"subset",2,PR_HAS_SUBSET);
} 
TheoryFinder::Matcher* TheoryFinder::Matcher::createFormulaMatcher (MatcherState& state,
								    int cp,
								    const Formula& formula)
{
  TRACER("TheoryFinder::Matcher::createFormulaMatcher");
  for (;;) {
    switch (state.code[cp]) {
    case CAND:
      if (formula.connective() == AND) {
	return new FormulaListMatcher(state,cp+1,formula.args());
      }
      return 0;
    case COR:
      if (formula.connective() == OR) {
	return new FormulaListMatcher(state,cp+1,formula.args());
      }
      return 0;
    case CIMP:
      if (formula.connective() == IMP) {
	return createFormulaPairMatcher(state,
					cp+1,
					formula.left(),formula.right());
      }
      return 0;
    case CNOT:
      if (formula.connective() == NOT) {
	return createFormulaMatcher(state,cp+1,formula.uarg());
      }
      return 0;
    case CIFF:
      if (formula.connective() == IFF) {
	FormulaList args(formula.right());
	args.push(formula.left());
	return new FormulaListMatcher(state,cp+1,args);
      }
      return 0;
    case CXOR:
      if (formula.connective() == XOR) {
	FormulaList args(formula.right());
	args.push(formula.left());
	return new FormulaListMatcher(state,cp+1,args);
      }
      return 0;
    case CFORALL:
      if (formula.connective() == FORALL) {
	return createQuantifiedFormulaMatcher(state,cp+1,formula);
      }
      return 0;
    case CEXISTS:
      if (formula.connective() == EXISTS) {
	return createQuantifiedFormulaMatcher(state,cp+1,formula);
      }
      return 0;
    default: 
      if (formula.connective() == ATOM) {
	return new AtomMatcher(state,cp,formula.atom());
      }
      return 0;
    }
  }
} 
TheoryFinder::Matcher*
TheoryFinder::Matcher::createQuantifiedFormulaMatcher (MatcherState& state,
						       int cp,
						       const Formula& formula)
{
  TRACER("TheoryFinder::Matcher::createQuantifiedFormulaMatcher");
  const VarList& vars = formula.vars();
  if (state.code[cp] != vars.length()) {
    return 0;
  }
  int v = state.code[cp+1]; 
  Iterator<Var> vs (vars);
  while (vs.hasNext()) {
    if (! state.addNewVar(v,vs.next().toInt())) {
      return 0;
    }
    v++;
  }
  return createFormulaMatcher(state,cp+2,formula.qarg());
} 
TheoryFinder::Matcher* TheoryFinder::Matcher::createFormulaPairMatcher (MatcherState& state,
									int cp,
									const Formula& formula1,
									const Formula& formula2)
{
  TRACER("TheoryFinder::Matcher::createFormulaPairMatcher");
  Matcher* fm1 = createFormulaMatcher(state,cp,formula1);
  if (! fm1) {
    return 0;
  }
  return new FormulaPairMatcher(state,cp,fm1,formula2);
} 
TheoryFinder::FormulaPairMatcher::FormulaPairMatcher (MatcherState& state,
						      int cp,
						      Matcher* fm1,
						      const Formula& f2)
  : Matcher(state,cp),
    _f2(f2),
    _fm1(fm1),
    _fm2(0)
{
  TRACER("TheoryFinder::FormulaPairMatcher::FormulaPairMatcher");
} 
TheoryFinder::FormulaPairMatcher::~FormulaPairMatcher ()
{
  TRACER("TheoryFinder::FormulaPairMatcher::~FormulaPairMatcher");
  delete _fm1;
  delete _fm2;
} 
int TheoryFinder::FormulaPairMatcher::nextMatch ()
{
  TRACER("TheoryFinder::FormulaPairMatcher::nextMatch");
  CONTROL("1");
  if (_fm2) {
    CONTROL("2");
    int cp = _fm2->nextMatch();
    if (cp) {
      return cp;
    }
    delete _fm2;
    _fm2 = 0;
  }
  CONTROL("3");
  for (;;) {
    CONTROL("4");
    int cp = _fm1->nextMatch();
    if (! cp) {
      return 0;
    }
    _fm2 = createFormulaMatcher(_state,cp,_f2);
    if (! _fm2) {
      continue;
    }
    int cp2 = _fm2->nextMatch();
    if (cp2) {
      return cp2;
    }
  }
} 
bool TheoryFinder::matchExtensionality (const Formula& f)
{
  TRACER("TheoryFinder::matchExtensionality (const Formula&...)");
  static char code1[] =
    {CIFF,                           
      CFORALL,1,0,CIFF,              
       NEWSYM,0,2,OLDVAR,0,NEWVAR,1, 
       OLDSYM,0,OLDVAR,0,NEWVAR,2,   
      EQL,OLDVAR,1,OLDVAR,2};         
  static char code2[] =
    {CIMP,                           
      CFORALL,1,0,CIFF,              
       NEWSYM,0,2,OLDVAR,0,NEWVAR,1, 
       OLDSYM,0,OLDVAR,0,NEWVAR,2,   
      EQL,OLDVAR,1,OLDVAR,2};         
  return matchCode(f,code1,"extensionality",1,PR_HAS_EXTENSIONALITY) ||
         matchCode(f,code2,"extensionality",1,PR_HAS_EXTENSIONALITY);
} 
void TheoryFinder::ass (const char* name,
			int arity,
			const Symbol** theorySymbols,
			Unit& result)
{
  TRACER("TheoryFinder::ass");
  TermList args;
  for (int i = arity-1; i >= 0;i--) {
    FunctionSymbol& constant =_sig.createFun(theorySymbols[i]->name(),0);
    args.push(constant);
  }
  Atom atom(_sig.createPred(name,arity),args);
  Literal lit(true,atom);
  LiteralList lits(lit);
  Clause cls(lits);
  result = Unit(cls,new Inference(IR_INPUT));
  _metaTheory.push(result);
#if DEBUG_THEORY_FINDER
  cout << result.toStringAsInference();
#endif
} 
bool TheoryFinder::matchLeftInverse (const Atom& atom)
{
  TRACER("TheoryFinder::matchLeftInverse");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWSYM,1,1,NEWVAR,0,OLDVAR,0,
    NEWSYM,2,0};                                           
  return matchCode(atom,code,"left_inverse",3,0);
} 
bool TheoryFinder::matchRightInverse (const Atom& atom)
{
  TRACER("TheoryFinder::matchRightInverse");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWVAR,0,NEWSYM,1,1,OLDVAR,0,
    NEWSYM,2,0};                                           
  return matchCode(atom,code,"right_inverse",3,0);
} 
bool TheoryFinder::matchLeftIdentity (const Atom& atom)
{
  TRACER("TheoryFinder::matchLeftIdentity");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWSYM,1,0,NEWVAR,0,  
    OLDVAR,0};                       
  return matchCode(atom,code,"left_identity",2,0);
} 
bool TheoryFinder::matchIdempotence (const Atom& atom)
{
  TRACER("TheoryFinder::matchIdempotence");
  static const char code[] = 
    {EQL,NEWSYM,0,2,NEWVAR,0,OLDVAR,0,OLDVAR,0}; 
  return matchCode(atom,code,"idempotence",1,0);
} 
bool TheoryFinder::matchRightIdentity (const Atom& atom)
{
  TRACER("TheoryFinder::matchRightIdentity");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWVAR,0,NEWSYM,1,0, 
    OLDVAR,0};                      
  return matchCode(atom,code,"right_identity",2,0);
} 
bool TheoryFinder::matchAssociator (const Atom& atom)
{
  TRACER("TheoryFinder::matchAssociator");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,3,NEWVAR,0,NEWVAR,1,NEWVAR,2,                     
    NEWSYM,1,2,NEWSYM,2,2,OLDSYM,2,OLDVAR,0,OLDVAR,1,OLDVAR,2, 
     NEWSYM,3,1,OLDSYM,2,OLDVAR,0,OLDSYM,2,OLDVAR,1,OLDVAR,2}; 
  return matchCode(atom,code,"associator",4,0);
} 
bool TheoryFinder::matchCommutator (const Atom& atom)
{
  TRACER("TheoryFinder::matchCommutator");
  static const char code[] = 
   {EQL,                                     
    NEWSYM,0,3,NEWVAR,0,NEWVAR,1,            
    NEWSYM,1,2,NEWSYM,2,2,OLDVAR,1,OLDVAR,0, 
     NEWSYM,3,1,OLDSYM,2,OLDVAR,0,OLDVAR,1}; 
  return matchCode(atom,code,"commutator",4,0);
} 
bool TheoryFinder::matchLeftDistributivity (const Atom& atom)
{
  TRACER("TheoryFinder::matchLeftDistributivity");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWSYM,1,2,NEWVAR,0,
    NEWVAR,1,NEWVAR,2,                                
    OLDSYM,1,OLDSYM,0,OLDVAR,0,OLDVAR,2,
    OLDSYM,0,OLDVAR,1,OLDVAR,2};                 
  return matchCode(atom,code,"left_distributivity",2,0);
} 
bool TheoryFinder::matchRightDistributivity (const Atom& atom)
{
  TRACER("TheoryFinder::matchRightDistributivity");
  static const char code[] = 
   {EQL, 
    NEWSYM,0,2,NEWVAR,0,NEWSYM,1,2,
    NEWVAR,1,NEWVAR,2,                     
    OLDSYM,1,OLDSYM,0,OLDVAR,0,OLDVAR,1,
    OLDSYM,0,OLDVAR,0,OLDVAR,2};           
  return matchCode(atom,code,"right_distributivity",2,0);
} 
bool TheoryFinder::matchCode (const Atom& atom,const char* code,
			      const char* name,int arity,int prop)
{
  TRACER("TheoryFinder::matchCode (const Atom&...)");
  int vars[8];
  const Symbol* sym[8];
  MatcherState state(code,vars,sym);
  AtomMatcher am(state,0,atom);
  if (am.nextMatch()) {
    Unit u;
    ass(name,arity,sym,u);
    if (_property && prop) {
      _property->addProp((Prop)prop);
    }
    return true;
  }
  return false;
} 
bool TheoryFinder::matchCode (const LiteralList& c,const char* code,
			      const char* name,int arity,int prop)
{
  TRACER("TheoryFinder::matchCode (const LiteralList&...)");
  int vars[8];
  const Symbol* sym[8];
  MatcherState state(code,vars,sym);
  ClauseMatcher matcher(state,c);
  if (matcher.nextMatch()) {
    Unit u;
    ass(name,arity,sym,u);
    if (_property && prop) {
      _property->addProp((Prop)prop);
    }
    return true;
  }
  return false;
} 
bool TheoryFinder::matchCode (const Formula& f,const char* code,
			      const char* name,int arity,int prop)
{
  TRACER("TheoryFinder::matchCode (const Formula&...)");
  int vars[8];
  const Symbol* sym[8];
  MatcherState state(code,vars,sym);
  Matcher* matcher = Matcher::createFormulaMatcher(state,0,f);
  if (! matcher) {
    return false;
  }
  if (matcher->nextMatch()) {
    Unit u;
    ass(name,arity,sym,u);
    if (_property && prop) {
      _property->addProp((Prop)prop);
    }
    delete matcher;
    return true;
  }
  delete matcher;
  return false;
} 
bool TheoryFinder::matchRobbins (const Atom& atom)
{
  TRACER("TheoryFinder::matchRobbins");
  static const char code1[] = 
   {EQL, 
     NEWSYM,0,1,NEWSYM,1,2,OLDSYM,0,OLDSYM,1,NEWVAR,0,NEWVAR,1,
      OLDSYM,0,OLDSYM,1,OLDVAR,0,OLDSYM,0,OLDVAR,1,            
     OLDVAR,0};                                                
  static const char code2[] = 
   {EQL, 
     NEWSYM,0,1,NEWSYM,1,2,OLDSYM,0,OLDSYM,1,NEWVAR,0,NEWVAR,1,
      OLDSYM,0,OLDSYM,1,OLDSYM,0,OLDVAR,0,OLDVAR,1,            
     OLDVAR,1};                                                
  static const char code3[] = 
   {EQL, 
     NEWSYM,0,1,NEWSYM,1,2,OLDSYM,0,OLDSYM,1,NEWVAR,0,OLDSYM,0,NEWVAR,1,
      OLDSYM,0,OLDSYM,1,OLDVAR,0,OLDVAR,1,                              
     OLDVAR,0};                                                         
  static const char code4[] = 
   {EQL, 
     NEWSYM,0,1,NEWSYM,1,2,OLDSYM,0,OLDSYM,1,OLDSYM,0,NEWVAR,0,NEWVAR,1,
      OLDSYM,0,OLDSYM,1,OLDVAR,0,OLDVAR,1,                              
     OLDVAR,1};                                                         
  return matchCode(atom,code1,"robbins",2,0) ||
         matchCode(atom,code2,"robbins",2,0) ||
         matchCode(atom,code3,"robbins",2,0) ||
         matchCode(atom,code4,"robbins",2,0);
} 
bool TheoryFinder::matchAlternative (const Atom& atom)
{
  TRACER("TheoryFinder::matchAlternative");
  static const char code1[] = 
   {EQL, 
     NEWSYM,0,2,OLDSYM,0,NEWVAR,0,OLDVAR,0,NEWVAR,1, 
     OLDSYM,0,OLDVAR,0,OLDSYM,0,OLDVAR,0,OLDVAR,1};  
  static const char code2[] = 
   {EQL, 
     NEWSYM,0,2,OLDSYM,0,NEWVAR,0,NEWVAR,1,OLDVAR,1, 
     OLDSYM,0,OLDVAR,0,OLDSYM,0,OLDVAR,1,OLDVAR,1};  
  return matchCode(atom,code1,"alternative",1,0) ||
         matchCode(atom,code2,"alternative",1,0);
} 
bool TheoryFinder::matchAbsorption (const Atom& atom)
{
  TRACER("TheoryFinder::matchAbsorption");
  static const char code[] = 
   {EQL,                                              
    NEWSYM,0,2,NEWVAR,0,NEWSYM,1,2,OLDVAR,0,NEWVAR,1, 
    OLDVAR,0};       
  return matchCode(atom,code,"absorption",2,0);
} 
bool TheoryFinder::matchCombinatorS (const Atom& atom)
{
  TRACER("TheoryFinder::matchCombinatorS");
  static const char code[] = 
   {EQL,                                     
    NEWSYM,0,2,OLDSYM,0,OLDSYM,0,NEWSYM,1,0,
     NEWVAR,0,NEWVAR,1,NEWVAR,2,             
    OLDSYM,0,OLDSYM,0,OLDVAR,0,OLDVAR,2,
     OLDSYM,0,OLDVAR,1,OLDVAR,2};            
  return matchCode(atom,code,"s",2,PR_COMBINATOR);
} 
bool TheoryFinder::matchCombinatorB (const Atom& atom)
{
  TRACER("TheoryFinder::matchCombinatorB");
  static const char code[] = 
   {EQL,                                           
    NEWSYM,0,2,OLDSYM,0,OLDSYM,0,NEWSYM,1,0,
     NEWVAR,0,NEWVAR,1,NEWVAR,2,                   
    OLDSYM,0,OLDVAR,0,OLDSYM,0,OLDVAR,1,OLDVAR,2}; 
  return matchCode(atom,code,"b",2,PR_COMBINATOR_B);
} 
bool TheoryFinder::matchCombinatorT (const Atom& atom)
{
  TRACER("TheoryFinder::matchCombinatorT");
  static const char code[] = 
   {EQL,                                              
    NEWSYM,0,2,OLDSYM,0,NEWSYM,1,0,NEWVAR,0,NEWVAR,1, 
    OLDSYM,0,OLDVAR,1,OLDVAR,0};                      
  return matchCode(atom,code,"t",2,PR_COMBINATOR);
} 
bool TheoryFinder::matchCombinatorO (const Atom& atom)
{
  TRACER("TheoryFinder::matchCombinatorO");
  static const char code[] = 
   {EQL,                                              
    NEWSYM,0,2,OLDSYM,0,NEWSYM,1,0,NEWVAR,0,NEWVAR,1, 
    OLDSYM,0,OLDVAR,1,OLDSYM,0,OLDVAR,0,OLDVAR,1};    
  return matchCode(atom,code,"o",2,PR_COMBINATOR);
} 
bool TheoryFinder::matchCombinatorQ (const Atom& atom)
{
  TRACER("TheoryFinder::matchCombinatorQ");
  static const char code[] = 
   {EQL,                                           
    NEWSYM,0,2,OLDSYM,0,OLDSYM,0,NEWSYM,1,0,
     NEWVAR,0,NEWVAR,1,NEWVAR,2,                   
    OLDSYM,0,OLDVAR,1,OLDSYM,0,OLDVAR,0,OLDVAR,2}; 
  return matchCode(atom,code,"q",2,PR_COMBINATOR);
} 
bool TheoryFinder::matchAll (const Atom& atom)
{
  TRACER("TheoryFinder::matchAll");
  return matchC(atom) ||
         matchA(atom) ||
         matchIdempotence(atom) ||
         matchLeftInverse(atom) ||
         matchLeftIdentity(atom) ||
         matchRightInverse(atom) ||
         matchRightIdentity(atom) ||
         matchLeftDistributivity(atom) ||
         matchRightDistributivity(atom) ||
         matchAssociator(atom) ||
         matchCommutator(atom) ||
         matchAlternative(atom) ||
         matchAbsorption(atom) ||
         matchRobbins(atom) || 
         matchCombinatorS(atom) ||
         matchCombinatorB(atom) ||
         matchCombinatorT(atom) ||
         matchCombinatorO(atom) ||
         matchCombinatorQ(atom);
} 
void TheoryFinder::analyse (const Clause& clause)
{
  TRACER("TheoryFinder::analyse");
  const Term& answer = clause.literals().head().atom().args().head();
  const string theory(answer.functor().name());
  if (theory == "group") {
    _property->addProp(PR_GROUP);
  }
  else if (theory == "ring") {
    _property->addProp(PR_RING);
  }
  else if (theory == "robbins_algebra") {
    _property->addProp(PR_ROBBINS_ALGEBRA);
  }
  else if (theory == "non_associative_ring") {
    _property->addProp(PR_NA_RING);
  }
  else if (theory == "boolean_algebra") {
    _property->addProp(PR_BOOLEAN_ALGEBRA);
  }
  else if (theory == "lattice") {
    _property->addProp(PR_LATTICE);
  }
  else if (theory == "lattice_ordered_group") {
    _property->addProp(PR_LO_GROUP);
  }
#if DEBUG_THEORY_FINDER
  cout << "THEORY FOUND: " << theory << "\n";
#endif
} 
}
