/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __unit__
#define __unit__
#include "Chain.hpp"
#include "Clause.hpp"
#include "Formula.hpp"
#include "InferenceRule.hpp"
#include "VampireKernel/VampireKernel.hpp"
namespace Shell {
class Options;
class Clause;
class Problem;
class UnitList;
class Inference;
enum InputType {
  AXIOM = 0u,
  ASSUMPTION = 1u,
  CONJECTURE = 2u
};
enum UnitType {
  CLAUSE = 0u,
  FORMULA = 1u
};
enum DefType {
  DT_NONE = 0u,
  DT_PRED = 1u,
  DT_FUN = 2u
};
class Unit 
{
 private:
  class Data;
  class ClauseData;
  class FormulaData;
 public:
  Unit ();
  Unit (const Unit& a);
  Unit (const string& name, InputType untype, const Formula&);
  Unit (const string& name, InputType untype, const Clause&);
  Unit (InferenceRule, const Formula&, const Unit& parent);
  Unit (const Formula&, Inference*);
  Unit (const Clause&, Inference*);
  Unit (InferenceRule, const Clause&, const Unit& parent);
  Unit (const Clause&, 
	const UnitList& parents, 
	long number,
	VampireKernel::Clause::BackgroundFlagIterator&);
  Unit (const Formula&, long number);
  Unit (const Formula&, const UnitList& parents);
  explicit Unit (void*); 
  ~Unit ();
  void operator = (const Unit& rhs);
  bool operator == (const Unit& rhs) const;
  bool operator != (const Unit& rhs) const;
  UnitType unitType () const;
  InputType inputType () const;
  void setInputType (InputType it);
  DefType defType () const;
  void setDefType (DefType);
  const Formula& formula () const;
  const Clause& clause () const;
  void setFormula (const Formula& f);
  const string& name () const;
  Inference* inference () const;
  void setInference (Inference* inf);
  long number () const;
  int numberInRefutation () const;
  void setNumberInRefutation (int newNumber);
  unsigned int mark () const;
  void setMark (unsigned int mark);
  void* giveAway () { return _data; }
  void* operator new (size_t);
  XML::Element* toXML () const;
  static string toString (InputType t);
  static string toString (UnitType t);
  static string toString (DefType dt);
  string toString () const;
  string toStringAsInference () const;
  static long nextNumber () { return ++_lastNumber; }
  static void setLastNumber (long n) { _lastNumber = n; }
  static InputType max (InputType it1,InputType it2);
 private:
  Data* _data;
  static long _lastNumber;
  friend class Data;
}; 
class UnitList 
: public Lst<Unit>
{
 public:
  UnitList ();
  UnitList (const UnitList&);
  explicit UnitList (const Unit &hd);
  UnitList (const Unit& head, const UnitList& tail);
  explicit UnitList (LstData<Unit>*);
  const UnitList& tail () const
    { return static_cast<const UnitList&>(Lst<Unit>::tail()); }
  UnitList& giveTail () const
    { return const_cast<UnitList&>(this->tail()); }
  UnitList& last ()
    { return static_cast<UnitList&>(Lst<Unit>::last()); }
  InputType maxInputType() const;
}; 
class Unit::Data
{
 public:
  ~Data ();
  void ref () { ASS (this); _counter++; }
  void deref ();
  UnitType unitType () const { return _untype; }
  InputType inputType () const { return _intype; }
  void setInputType (InputType it) { _intype = it; }
  DefType defType () const { return _defType; }
  void setDefType (DefType dt) { _defType = dt; }
  const string& name () const { return _name; }
  Inference* inference () const { return _inference; }
  void setInference (Inference* inf) { _inference = inf; }
  long number () const { return _number; }
  int numberInRefutation () const { return _numberInRefutation; }
  unsigned int mark () const { return _mark; }
  void setMark (unsigned int mark) { _mark = mark; }
  void setNumberInRefutation (int newNumber)
  {
    ASS(newNumber > 0);
    _numberInRefutation = newNumber;
  } 
  void* operator new (size_t);
 protected:
  Data (UnitType tp,const string& name, InputType intype);
  Data (UnitType tp,Inference* inf,long number);
  Data (UnitType tp,InferenceRule ir, const Unit& parent);
  unsigned int _counter : 24;
  UnitType _untype : 1;
  InputType _intype : 2;
  unsigned int _mark : 2;
  DefType _defType : 2;
  long _number;
  int _numberInRefutation;
  string _name;
  Inference* _inference;
}; 
class Unit::ClauseData
  : public Unit::Data
{
 public:
  ClauseData (const string& name, InputType untype, const Clause&);
  ClauseData (InferenceRule, const Clause&, const Unit& parent);
  ClauseData (const Clause&, 
	      const UnitList& parents, 
	      long number,
	      VampireKernel::Clause::BackgroundFlagIterator&);
  ClauseData (const Clause&, Inference*);
  const Clause& clause () const { return _clause; }
  CLASS_NAME("Unit::ClauseData");
  USE_ALLOCATOR(ClauseData);
 private:
  Clause _clause;
}; 
class Unit::FormulaData
  : public Unit::Data
{
 public:
  FormulaData (const string& name, InputType untype, const Formula&);
  FormulaData (InferenceRule, const Formula&, const Unit& parent);
  FormulaData (const Formula&, long number);
  FormulaData (const Formula&, const UnitList& parents);
  FormulaData (const Formula&, Inference*);
  const Formula& formula () const { return _formula; }
  void setFormula (const Formula& f) { _formula = f; }
  CLASS_NAME("Unit::FormulaData");
  USE_ALLOCATOR(FormulaData);
 protected:
  Formula _formula;
}; 
typedef Chain<Unit> UnitChain;
inline
Unit::Unit () 
  : 
  _data (0) 
{
} 
inline
Unit::~Unit () 
{
  if (_data) {
    _data->deref ();
  }
} 
inline
Unit::Unit (const Unit& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 
inline
Unit::Unit (void* d)
  :
  _data (reinterpret_cast<Data*>(d))
{
  _data->ref();
} 
inline
Unit::Unit (const Formula& f, Inference* inf)
  : 
  _data (new FormulaData(f, inf))
{
} 
inline
Unit::Unit (const Clause& c, Inference* inf)
  : 
  _data (new ClauseData(c, inf))
{
} 
inline
Unit::Unit (const string& name, InputType untype, const Formula& formula) 
  : 
  _data (new FormulaData(name, untype, formula))
{
} 
inline
Unit::Unit (const string& name, InputType untype, const Clause& clause) 
  : 
  _data (new ClauseData(name, untype, clause)) 
{
} 
inline
Unit::Unit (InferenceRule ir, const Formula& f, const Unit& parent) 
  : 
  _data (new FormulaData(ir, f, parent))
{
} 
inline
Unit::Unit (InferenceRule ir, const Clause& clause, const Unit& parent) 
  : 
  _data (new ClauseData(ir, clause, parent)) 
{
} 
inline
Unit::Unit (const Clause& clause,
	    const UnitList& parents,
	    long number,
	    VampireKernel::Clause::BackgroundFlagIterator& flags)
  : 
  _data (new ClauseData(clause, parents, number,flags)) 
{
} 
inline
Unit::Unit (const Formula& formula,long number)
  : 
  _data (new FormulaData(formula, number)) 
{
} 
inline
Unit::Unit (const Formula& formula, const UnitList& parents) 
  : 
  _data (new FormulaData(formula, parents)) 
{
} 
inline
bool Unit::operator == (const Unit& rhs) const
{ 
  return _data == rhs._data; 
} 
inline
bool Unit::operator != (const Unit& rhs) const
{ 
  return _data != rhs._data; 
} 
inline
UnitType Unit::unitType () const
{
  return _data->unitType ();
} 
inline
const Formula& Unit::formula () const
{
  ASS (unitType() == FORMULA);
  return static_cast<const FormulaData*>(_data)->formula();
} 
inline
void Unit::setFormula (const Formula& f)
{
  static_cast<FormulaData*>(_data)->setFormula(f);
} 
inline
DefType Unit::defType () const
{
  return _data->defType ();
} 
inline
void Unit::setDefType (DefType dt)
{
  _data->setDefType (dt);
} 
inline
const Clause& Unit::clause () const
{
  ASS (unitType() == CLAUSE);
  return static_cast<const ClauseData*>(_data)->clause();
} 
inline
InputType Unit::inputType () const
{
  return _data->inputType();
} 
inline
void Unit::setInputType (InputType it)
{
  _data->setInputType(it);
} 
inline
long Unit::number () const
{
  return _data->number ();
} 
inline
int Unit::numberInRefutation () const
{
  return _data->numberInRefutation ();
} 
inline
void Unit::setNumberInRefutation (int newNumber)
{
  _data->setNumberInRefutation (newNumber);
} 
inline
Inference* Unit::inference () const
{
  return _data->inference();
} 
inline
const string& Unit::name () const
{
  return _data->name ();
} 
inline
void Unit::setInference (Inference* inf)
{
  _data->setInference(inf);
} 
inline 
unsigned int Unit::mark () const 
{ 
  return _data->mark(); 
} 
inline
void Unit::setMark (unsigned int mark) 
{ 
  _data->setMark(mark);
} 
}
#endif 
