/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Run.hpp"
#include "Tracer.hpp"
#include "VampireException.hpp"
#include "XML.hpp"
using namespace Shell;
VampireException::VampireException (const string& file, int line)
  : _file(file),
    _line(line)
{
}
void VampireException::outputFileAndLine (ostream& str) const
{
  str << "(file: '" << _file
      << "', line: " << _line << ")";
} 
void VampireException::addFileAndLine (XML::DeepElement& xml) const
{
  xml.addChild(new XML::TextElement("file",_file));
  xml.addChild(new XML::IntegerElement("line",_line));
} 
UserErrorException::UserErrorException (const string& message)
  : _message(message)
{
}
void UserErrorException::cry (ostream& str)
{
  str << "Error: " <<_message << "\n";
} 
XML::Element* UserErrorException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","user error"));
  xml->addChild(new XML::TextElement("message",_message));
  addFileAndLine(*xml);
  return xml;
} 
AllocationFailureException::AllocationFailureException (const string& file, 
							int line)
  : VampireException(file,line)
{
}
void AllocationFailureException::cry (ostream& str)
{
  str << "Allocation failure ";
  outputFileAndLine(str);
  str << '\n';
} 
XML::Element* AllocationFailureException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","allocation failure"));
  addFileAndLine(*xml);
  return xml;
} 
AssertionViolationException::AssertionViolationException (const string& file, 
							  int line)
  : VampireException(file,line)
{
#if DEBUG_SHELL
  _stack = MyTracer::stackToXML();
#endif
  cry(cout);
}
void AssertionViolationException::cry (ostream& str)
{
  str << "Assertion violation ";
  outputFileAndLine(str);
  str << '\n';
#if DEBUG_SHELL
  MyTracer::printStack(cout);
#endif
} 
XML::Element* AssertionViolationException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","assertion violation"));
  addFileAndLine(*xml);
#if DEBUG_SHELL
  xml->addChild(_stack);
#endif
  return xml;
} 
VampireTerminationException::VampireTerminationException (const string& file, 
							  int line,
							  string reason,
							  RunResult result)
  : VampireException(file,line),
    _reason(reason),
    _result(result)
{
}
void VampireTerminationException::cry (ostream& str)
{
  str << "Vampire termination exception ";
  outputFileAndLine(str);
  str << ", reason: " << _reason 
      << ", result: " << Run::toString(_result) << '\n';
} 
XML::Element* VampireTerminationException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","vampire termination"));
  addFileAndLine(*xml);
  xml->addChild(new XML::TextElement("reason",_reason));
  xml->addChild(new XML::TextElement("result",Run::toString(_result)));
  return xml;
} 
