/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BackwardSuperposition.hpp"
#include "DestructionMode.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
#include "UnifTerm.hpp"
#include "Clause.inl"
using namespace BK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BackwardSuperposition"
#endif
#include "debugMacros.hpp"
VK::BackwardSuperposition::BackwardSuperposition(UnifIndex* unifIndex) : 
  OPEN_BR(1,"br 1") DCOMMA 
_statistics(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _simultaneousness(0L),
  _noInfBetweenRulesFlag(false),
  _paramodulateFromVariables(true),
  _mainPremiseSubsumption(false),
  _sidePremiseSubsumption(false),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _mainPremiseVarIndex(0UL),
  _sidePremiseVarIndex(1UL), 
  _dummyClauseCollector(DOP("BackwardSuperposition::_dummyClauseCollector"))
{
  _mainPremiseResidualSizeFormula.SetIndex(0UL);
  _mainPremiseSelectedLitResidualSizeFormula.SetIndex(0UL);
  _currentMainPosEqArg1Mem[0].symbol().MkFunc(Term((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity));
  _currentMainPosEqArg2Mem[0].symbol().MkFunc(Term((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity));
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _storedConstraintCheck = 
    SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; 
VK::BackwardSuperposition::~BackwardSuperposition() 
{ 
  CALL("destructor ~BackwardSuperposition()"); 
  if (DestructionMode::isThorough())
    {
      ASSERT(_storedConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
    };
};  
bool VK::BackwardSuperposition::SetQuery(Clause* cl) 
{ 
  CALL("SetQuery(Clause* cl)");
  ASSERT(!cl->isSupportClause());
  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_allMainPremiseLiterals->numOfSelectedLits());
  _mainSelectedLitIter.reset(_allMainPremiseLiterals);
  compileMainPremiseResidualSizeFormula();
 try_next_pos_eq:
  if (!findNextMainSelectedPosEq()) return false;
  ASSERT(_currentMainPosEqLit);
  if (!loadEq(_currentMainPosEqLit,_mainPremiseVarIndex))
    goto try_next_pos_eq;
  return true; 
}; 
bool VK::BackwardSuperposition::CollectResult()
{
  CALL("CollectResult()");
  if ((!_mainPremise->isPersistent()) 
      || (!_sidePremise->isPersistent()))
    return false;
  if (_mainPremiseResidualSize
      + _sidePremise->practicallyMinResidualSize()
      > _clauseSizeLimit)
    {
      ++(_statistics->inferences.deduction.backwardSuperposition.discardedBySize);
      return false;
    };            
  _clauseCollector->openNewClause(ClauseBackground::BackSup,
				  _mainPremise,
				  _sidePremise);
  _clauseCollector->announceNewClause();
  return
    collectSidePremiseLiterals() && 
    collectMainPremiseLiterals() && 
    _clauseCollector->finaliseClause();
}; 
bool VK::BackwardSuperposition::NextInference()
{
 process_literal:
  while (Paramodulate())
    { 
      if (CollectResult()) return true;
    };
 try_next_pos_eq:
  if (!findNextMainSelectedPosEq()) return false;
  ASSERT(_currentMainPosEqLit);
  if (!loadEq(_currentMainPosEqLit,_mainPremiseVarIndex))
    goto try_next_pos_eq;
  goto process_literal;
}; 
inline 
void 
VK::BackwardSuperposition::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();
  for (LiteralList::Iterator iter(_allMainPremiseLiterals,
				  _firstNonselectedMainLit);
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());
}; 
inline 
void 
VK::BackwardSuperposition::recompileMainPremiseSelectedLitResidualSizeFormula()
{
  CALL("recompileMainPremiseSelectedLitResidualSizeFormula()");
  _mainPremiseSelectedLitResidualSizeFormula.Reset();
  for (LiteralList::Iterator iter(_allMainPremiseLiterals); 
       iter.currentElement() != _firstNonselectedMainLit;
       iter.next())
    {
      ASSERT(!iter.currentLiteral()->isAnswerLiteral());
      if (iter.currentElement() == _currentMainPosEqLit)
	{              
	  ASSERT(iter.currentElement()->isEquality());
	  ASSERT(iter.currentElement()->positive());
	  if (iter.currentElement()->isOrientedEq())
	    _mainPremiseSelectedLitResidualSizeFormula.Compile(*(iter.currentElement()->equalityRHS()));
	}
      else 
	_mainPremiseSelectedLitResidualSizeFormula.Compile(iter.currentElement()->literal());
    }; 
}; 
inline 
bool
VK::BackwardSuperposition::findNextMainSelectedPosEq()
{
  CALL("findNextMainSelectedPosEq()");
  const Term* lit;
 check_lit:
  if (_mainSelectedLitIter.currentElement() == _firstNonselectedMainLit) 
    return false;  
  _currentMainPosEqLit = _mainSelectedLitIter.currentElement();
  lit = _mainSelectedLitIter.currentLiteral();
  if (lit->Positive() && 
      lit->IsEquality() &&
      _mainSelectedLitIter.currentElement()->properties().useEqForParamodulationFrom()) 
    {
      _mainSelectedLitIter.next();
      recompileMainPremiseSelectedLitResidualSizeFormula();
      return true;
    }; 
  _mainSelectedLitIter.next();
  goto check_lit;
}; 
inline 
bool 
VK::BackwardSuperposition::loadEq(const LiteralList::Element* eq_lit,
				  ulong eq_ind)
{
  CALL("loadEq(const LiteralList::Element*eq_lit,ulong eq_ind)");
  ASSERT(eq_lit->isEquality());
  ASSERT(eq_lit->positive());
  _mainPremiseVarIndex = eq_ind; 
  const Term* arg_ref = eq_lit->equalityLHS();
  UnifTerm* mem = _currentMainPosEqArg1Mem + 1;
  UnifTerm::StandardTermToUnifTerm(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg1Mem[0].SetBracket(mem);
  arg_ref = eq_lit->equalityRHS();
  mem = _currentMainPosEqArg2Mem + 1;
  UnifTerm::StandardTermToUnifTerm(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg2Mem[0].SetBracket(mem);
  _currentMainPosEqRHS = _currentMainPosEqArg2Mem + 1;
  _swapArguments = !eq_lit->isOrientedEq();
  _constraintRequired = eq_lit->isOrientableEq();
  if (!LoadLHS(_currentMainPosEqArg1Mem))
    {
      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedForOtherReasons);
      if (!_swapArguments) return false;
      _swapArguments = false;
      if (!LoadLHS(_currentMainPosEqArg2Mem))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedForOtherReasons);
	  return false;
	};
      _currentMainPosEqRHS = _currentMainPosEqArg1Mem + 1;
    };
  _currentCompiledConstraint = 0;
  if (_swapArguments && _constraintRequired)
    {
      _compiledConstraint->load(_currentMainPosEqArg2Mem,_currentMainPosEqArg1Mem);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _currentCompiledConstraint = _compiledConstraint;
	};
    };
  return true;
}; 
inline 
bool 
VK::BackwardSuperposition::Paramodulate()
{
  CALL("Paramodulate()");
 try_next_cand:
  if (NextCandidate(_currentCompiledConstraint,
		    _sidePremise,
		    _candLitRedexPair)) 
    {
      if (_noRulesAsSidePremises && _sidePremise->IsRule())
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.discardedForOtherReasons);
	  goto try_next_cand;
	};
      if (demodulateSelectedLits())
	goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;
      return true;
    };
  if (_swapArguments)
    {
      _swapArguments = false;
      if (!LoadLHS(_currentMainPosEqArg2Mem))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedForOtherReasons);
	  return false;
	};
      _currentMainPosEqRHS = _currentMainPosEqArg1Mem + 1;
      _currentCompiledConstraint = 0;
      if (_constraintRequired)
	{
	  _compiledConstraint->load(_currentMainPosEqArg1Mem,_currentMainPosEqArg2Mem);
	  if (_compiledConstraint->rhsVarInstMayOrder())
	    {
	      _compiledConstraint->compileGreaterOrEqual();
	      _currentCompiledConstraint = _compiledConstraint;
	    };
	};
      goto try_next_cand;          
    };
  return false; 
}; 
inline 
bool 
VK::BackwardSuperposition::collectMainPremiseLiterals() 
{
  CALL("collectMainPremiseLiterals()");
  LiteralProperties::SLModifier slmod = LiteralProperties::DefaultSLModifier;
  if (_currentMainPosEqLit->properties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_currentMainPosEqLit->properties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  slmod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_currentMainPosEqLit->properties().deploymentMode() == 
		 LiteralProperties::Subquery);
	  slmod = LiteralProperties::PostponeSelection;
	};
    };
  LiteralProperties prop;
  for (LiteralList::Iterator iter(_allMainPremiseLiterals); 
       iter.currentElement() != _firstNonselectedMainLit;
       iter.next())
    if (iter.currentElement() != _currentMainPosEqLit)
      {  
	prop.inheritFrom(iter.currentElement()->properties(),slmod);
	if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      prop)) 
	  return false;
      };      
  for (LiteralList::Iterator iter1(_allMainPremiseLiterals,_firstNonselectedMainLit);
       iter1.notEnd();
       iter1.next())
    {
      prop.inheritFrom(iter1.currentElement()->properties(),slmod);
      if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						    iter1.currentLiteral(),
						    prop)) 
	return false;   
    };  
  return true;
}; 
inline 
bool VK::BackwardSuperposition::collectSidePremiseLiterals() 
{
  CALL("collectSidePremiseLiterals()");
  const SimplificationOrdering::StoredConstraint* constr = 
    _candLitRedexPair->constraint();
  if (constr) 
    {
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantDiscarded);
	  return false;
	};
    };
  ulong rewritten_lit_num = _candLitRedexPair->litNum();
  LiteralList::Element* rewritten_lit = 
    _sidePremise->LitList()->nth(rewritten_lit_num);
  LiteralProperties::SLModifier rewLitSLMod =
    LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier slmod =
    LiteralProperties::DefaultSLModifier;
  if (rewritten_lit->properties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (rewritten_lit->properties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  slmod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(rewritten_lit->properties().deploymentMode() == 
		 LiteralProperties::Subquery);
	  slmod = LiteralProperties::PostponeSelection;
	};
    };
  LiteralProperties prop;
  for (LiteralList::Iterator iter(_sidePremise->LitList()); 
       iter.notEnd(); 
       iter.next())
    {
      if (iter.currentElement() == rewritten_lit)
	{        
	  prop.inheritFrom(iter.currentElement()->properties(),
			   rewLitSLMod);
	  if (rewritten_lit->isEquality())
	    { 
	      if (rewritten_lit->isPreorientedEq())
		{
		  if (!rewritten_lit->equalityLHS()->isVariable() &&
		      rewritten_lit->equalityLHS()->deref()->containsComplexSubterm(_candLitRedexPair->redex()))
		    prop.setDefaultEqualityOrientationMode();
		}
	      else
		if (!rewritten_lit->isOrientableEq())
		  prop.setDefaultEqualityOrientationMode();
	    };
	  if (!_clauseCollector->pushLiteralWithReplacement(_sidePremiseVarIndex,
							    iter.currentLiteral(),
							    _candLitRedexPair->redex(),
							    _currentMainPosEqRHS,
							    prop))
	    return false;
	}
      else
	{
	  if ((_simultaneousness > 0) && (iter.currentElement()->arity()))
	    {
	      prop.inheritFrom(iter.currentElement()->properties(),slmod);
	      if (!_clauseCollector->pushLiteralWithReplacement(_sidePremiseVarIndex,
								iter.currentLiteral(),
								_candLitRedexPair->redex(),
								_currentMainPosEqRHS,
								prop))
		return false;   
	    }
	  else
	    {
	      prop.inheritFrom(iter.currentElement()->properties(),slmod);
	      if (!_clauseCollector->pushLiteralWithUnifier(_sidePremiseVarIndex,
							    iter.currentElement()->literal(),
							    prop)) 
		return false;  
	    };
	};  
    };     
  return true;
};  
bool VK::BackwardSuperposition::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (collectSidePremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause)) 
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _sidePremise->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.backwardSuperposition.sidePremisesSubsumed);
 discarded:
  _sidePremise->unblockForSubsumption();
  return true;
}; 
bool VK::BackwardSuperposition::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");
  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties())) 
      return false; 
  return true;
}; 
bool 
VK::BackwardSuperposition::LoadLHS(UnifTerm* t)
{
  CALL("LoadLHS(UnifTerm* t)");
  if (!_retrievalOfUnifiable.CompileLiteral(t))
    return false;
  _newRetrievalAttempt = true;
  _isFromVariable = t->Arg1()->symbol().isVariable();
  return true;
}; 
inline 
bool 
VK::BackwardSuperposition::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
					 Clause*& candCl,
					 BackParaCandidates::LitRedexPair*& candLitRedexPair)
{
  CALL("NextCandidate(..)"); 
  if (_newRetrievalAttempt)
    {
      _newRetrievalAttempt = false;
      if ((!_paramodulateFromVariables) && _isFromVariable)
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedWithParamodulationsFromVariables);
	  return false;
	};
      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->BParaTree())))
	{	
	  _mainPremiseResidualSize =
	    _mainPremiseResidualSizeFormula.Compute() + 
	    _mainPremiseSelectedLitResidualSizeFormula.Compute();
	  if (_mainPremiseResidualSize > _clauseSizeLimit)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedBySize);
	      goto try_next_leaf;
	    };
	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.CurrentLeaf()->BPCandidates()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  return false;
	};
    }
  else 
    {
      if (_candidates.next(candCl,candLitRedexPair)) 
	{
	  return true;
	};
    try_next_leaf: 
      if (_retrievalOfUnifiable.FindNextLeaf())
	{	        
	  _mainPremiseResidualSize =
	    _mainPremiseResidualSizeFormula.Compute() + 
	    _mainPremiseSelectedLitResidualSizeFormula.Compute();
	  if (_mainPremiseResidualSize > _clauseSizeLimit)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedBySize);
	      goto try_next_leaf;
	    };
	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.CurrentLeaf()->BPCandidates()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  return false;  
	};
    };     
}; 
bool VK::BackwardSuperposition::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (collectMainPremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))   
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {      
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{	
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _mainPremise->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesSubsumed);
 discarded:
  _mainPremise->unblockForSubsumption();
  return true;
}; 
bool VK::BackwardSuperposition::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties())) 
      return false; 
  return true;
}; 
bool VK::BackwardSuperposition::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_sidePremise->paramodulationsIntoProtected() &&
      _mainPremise->Unit())
    return false; 
  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesDemodulated);
	      return true;
	    };
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
          _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() && 
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {		  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);     		      
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesDemodulated);
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	    }
	  else 
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (tautology)
		++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      return tautology;
	    };
	};
    };
  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
      if (collectSidePremiseSubtermsAvailableForParamodulationInto()&&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{		    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause); 
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++(_statistics->inferences.deduction.backwardSuperposition.sidePremisesDemodulated);
	      return true;
	    }
	  else
	    return false;
	}
      else 
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  bool tautology =
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (tautology)
	    ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  return tautology;
	};
    }
  else
    return false;
}; 
bool 
VK::BackwardSuperposition::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop; 
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      LiteralList::Element* elem = iter.currentElement();
      Term* lit = elem->literal();
      if (lit->IsEquality())
	{
	  if (elem->isOrientedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_mainPremiseVarIndex,
								      Term::dummyPositiveHeader1(),
								      *(elem->equalityLHS())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; 
bool 
VK::BackwardSuperposition::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      LiteralList::Element* elem = iter.currentElement();
      Term* lit = elem->literal();
      if (lit->IsEquality())
	{
	  if (elem->isOrientedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_sidePremiseVarIndex,
								      Term::dummyPositiveHeader1(),
								      *(elem->equalityLHS())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; 
