/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "BinaryResolution.hpp"
#include "Clause.inl"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
#  define DEBUG_NAMESPACE "BinaryResolution"
#endif
#include "debugMacros.hpp"
VK::BinaryResolution::BinaryResolution(UnifIndex* resIndex) : 
  OPEN_BR(1,"br 1") DCOMMA 
_statistics(0),
  _mainPremiseVarIndex(0),
  _sidePremiseVarIndex(1),
  _unificationIndex(resIndex),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _clauseCollector(0),
  _forwardProcessing(0),
  _dummyClauseCollector(DOP("BinaryResolution::_dummyClauseCollector")),
  _noInfBetweenRulesFlag(false)
{
  _residualSizeFormula.SetIndex(_mainPremiseVarIndex);
};
bool VK::BinaryResolution::Load(Clause* c)
{
  CALL("Load(Clause* c)");
  if (!eligible(c)) return false;
  ORMachine.SetIndex(_sidePremiseVarIndex);
  if (!_mainPremise.Load(_mainPremiseVarIndex,c))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);      return false;
    };
  compileResidualSize(c);
  _sideClauseCanNotBeRule = _noInfBetweenRulesFlag && c->IsRule();
  _currentMainLit = 0;
  _rotationPossible = true;
  _continuation = ContFirstLit;
  return true;
}; 
bool VK::BinaryResolution::NextInference()
{
  CALL("NextInference()");
  UnifIndex::IndexedObject* candidate;
  switch (_continuation)
    {
    case ContFirstLit: goto current_lit;
    case ContFirstFactor: goto first_factor;
    case ContNextFactor: goto next_factor; 
    };
  ICP("ICP1");
 current_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false; 
  ASSERT(_mainPremise.nthSelectedLit(_currentMainLit));
  if (!_mainPremise.nthSelectedLitEligibleForResolution(_currentMainLit))
    {
      ++_currentMainLit;
      goto current_lit;
    };
  if (!ORMachine.CompileLiteral(_mainPremise.nthSelectedLit(_currentMainLit)))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
      ++_currentMainLit;
      goto current_lit;
    };
  ORMachine.flipPolarity();
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  _rotationPossible = true;
 find_first_leaf:
  if (ORMachine.FindFirstLeaf(*(_unificationIndex->Tree(ORMachine.Code()->Symbol().Func()))))  
    {
    process_leaf: 
      _residualMainPremiseSize = _residualSizeFormula.Compute();
      if (_residualMainPremiseSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedBySize);
	  goto try_next_leaf;
	};
      ORMachine.ResetCandidatesFromLeaf();
      if (demodulateMainSelectedLits()) 
	{
	  while ((candidate  = ORMachine.NextCandidate()))
	    candidate->TheClause()->protectParamodulationsInto();
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;
    try_next_cand:
      candidate  = ORMachine.NextCandidate();
      if (!candidate) goto try_next_leaf;       
      if (_sideClauseCanNotBeRule && candidate->TheClause()->IsRule()) 
        {
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
        };
      if (_residualMainPremiseSize + 
	  candidate->TheClause()->practicallyMinResidualSize() >
	  _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedBySize);
	  goto try_next_cand;
        };   
      if (!_sidePremise.LoadOptimized(_sidePremiseVarIndex,candidate))
	{
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation))
	goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;
      if (!CollectResult()) 
        {
	  goto first_factor;
        };
      _continuation = ContFirstFactor;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (NextFactor())
	{
	  if (demodulateMainSelectedLits()) 
	    {
	      _sidePremise.TheClause()->protectParamodulationsInto();
	      goto next_factor;
	    }; 
	  if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation)) 
	    goto next_factor; 
	  if (subsumeMainPremise()) goto next_factor;  
	  if (subsumeSidePremise()) goto next_factor; 
	  if (!CollectResult()) goto next_factor;            
	  _continuation = ContNextFactor;
	  return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (ORMachine.FindNextLeaf()) { goto process_leaf; };
    };
  if (!_rotationPossible || !_mainPremise.nthSelectedLitRotation(_currentMainLit))
    {
      _mainPremise.UnmarkCut();
      ++_currentMainLit;
      goto current_lit;
    };
  if (!ORMachine.CompileLiteral(_mainPremise.nthSelectedLitRotation(_currentMainLit)))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
      ++_currentMainLit;
      goto current_lit;
    };
  ORMachine.flipPolarity();
  _rotationPossible = false;  
  goto find_first_leaf;
}; 
bool VK::BinaryResolution::NextFactor()
{
  CALL("NextFactor()");
  if (_sidePremise.WasFactored())
    {
      if (_sidePremise.GetNextFactor())
	{
	  return true;
	}
      else 
	{
	  _sidePremise.ResetFactoring();
	  if (_mainPremise.WasFactored())
	    {
	      return _mainPremise.GetNextFactor();
	    }
	  else
	    {
	      return _mainPremise.GetFirstFactor(_mainPremise.CutLit()+1);
	    };
	};
    }
  else 
    {
      if (_sidePremise.GetFirstFactor(0))
	{
	  return true;
	}
      else
	{
	  if (_mainPremise.WasFactored())
	    {
	      return _mainPremise.GetNextFactor();
	    }
	  else
	    {
	      return _mainPremise.GetFirstFactor(_mainPremise.CutLit()+1);
	    };
	};
    };
}; 
inline bool VK::BinaryResolution::CollectResult() 
{ 
  CALL("CollectResult()");
  if ((!_mainPremise.TheClause()->isPersistent())
      || (!_sidePremise.TheClause()->isPersistent()))
    return false;
  _clauseCollector->openNewClause(ClauseBackground::BinRes,
				  _mainPremise.TheClause(),
				  _sidePremise.TheClause());
  _clauseCollector->announceNewClause();
  LiteralProperties::SLModifier slmod1 = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier slmod2 = LiteralProperties::DefaultSLModifier;
  if (_mainPremise.resolvedLiteralProperties().deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (_mainPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::RuleHead)
	{
	  slmod1 = LiteralProperties::AccelerateSelection;
	}
      else 
	{
	  ASSERT(_mainPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);	    
	  slmod1 = LiteralProperties::PostponeSelection;
	};	
    };
  if (_sidePremise.resolvedLiteralProperties().deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (_sidePremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::RuleHead)
	{
	  slmod2 = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_sidePremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);	  
	  slmod2 = LiteralProperties::PostponeSelection;
	};	
    };
  return 
    _mainPremise.collectUnmarkedLiterals(*_clauseCollector,slmod1) && 
    _sidePremise.collectUnmarkedLiterals(*_clauseCollector,slmod2) &&
    _clauseCollector->finaliseClause();
}; 
inline void VK::BinaryResolution::compileResidualSize(const Clause* c)
{
  _residualSizeFormula.Reset();
  LiteralList* lits = c->LitList();
  for (LiteralList::Iterator iter(lits,lits->nth(lits->numOfSelectedLits()));
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _residualSizeFormula.Compile(iter.currentElement()->literal()); 
}; 
bool VK::BinaryResolution::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  ASSERT(_forwardProcessing);
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise.TheClause()->hasSubsumedPremiseInstance()) return false;
  _mainPremise.TheClause()->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (_mainPremise.collectAllLiterals(_dummyClauseCollector) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _mainPremise.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _mainPremise.TheClause()->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.resolution.mainPremisesSubsumed);
 discarded:
  _mainPremise.TheClause()->unblockForSubsumption();
  return true;
}; 
bool VK::BinaryResolution::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  ASSERT(_forwardProcessing);
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_sidePremise.TheClause()->hasSubsumedPremiseInstance()) return false;
  _sidePremise.TheClause()->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (_sidePremise.collectAllLiterals(_dummyClauseCollector) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _sidePremise.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _sidePremise.TheClause()->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.resolution.sidePremisesSubsumed);
 discarded:
  _sidePremise.TheClause()->unblockForSubsumption();
  return true;
}; 
bool VK::BinaryResolution::demodulateMainSelectedLits()
{
  CALL("demodulateMainSelectedLits()");
  ASSERT(_forwardProcessing);
  Clause* simplifyingClause;
  if (!_mainSelectedLitDemodulation) return false;
  _mainPremise.TheClause()->blockForDemodulationFrom();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (_mainPremise.collectSubtermsAvailableForParamodulationInto(_dummyClauseCollector,true) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      simplifyingClause = 
	_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
      ASSERT(simplifyingClause != _mainPremise.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _mainPremise.TheClause()->unblockForDemodulationFrom();
  return false;
 demodulated:
  _mainPremise.TheClause()->protectParamodulationsInto();
  ++(_statistics->inferences.deduction.resolution.mainPremisesDemodulated);
 discarded:
  _mainPremise.TheClause()->unblockForDemodulationFrom();
  return true;
}; 
bool VK::BinaryResolution::demodulateSideSelectedLits(bool includeMarked)
{
  CALL("demodulateSideSelectedLits(bool includeMarked)");
  ASSERT(_forwardProcessing);
  Clause* simplifyingClause;
  if (!_sideSelectedLitDemodulation) return false;
  _sidePremise.TheClause()->blockForDemodulationFrom();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (_sidePremise.collectSubtermsAvailableForParamodulationInto(_dummyClauseCollector,
							     includeMarked) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      simplifyingClause = 
	_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
      ASSERT(simplifyingClause != _sidePremise.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _sidePremise.TheClause()->unblockForDemodulationFrom();
  return false;
 demodulated:
  _mainPremise.TheClause()->protectParamodulationsInto();
  _sidePremise.TheClause()->protectParamodulationsInto();
  ++(_statistics->inferences.deduction.resolution.sidePremisesDemodulated);
 discarded:
  _sidePremise.TheClause()->unblockForDemodulationFrom();
  return true;
}; 
inline
bool VK::BinaryResolution::eligible(const Clause* cl)
{
  CALL("eligible(const Clause* cl)");
  const LiteralList* lits = cl->LitList();
  ulong numOfSelected = lits->numOfSelectedLits();
  for (const LiteralList::Element* el = cl->LitList()->firstElem();
       numOfSelected;
       (el = el->next(),--numOfSelected))
    {
      ASSERT(el);
      if (el->literal()->IsEquality())
	{
	  if (el->properties().useEqForResolution()) return true;
	}
      else 
	return true;
    };
  return false;
}; 
