/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Chaining.hpp"
#include "Clause.inl"
#include "RuntimeError.hpp"
#include "UnifTerm.hpp"
#include "Unifier.hpp"
#include "Term.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
Chaining::Chaining(UnifIndex* unifIndex) :
  OPEN_BR(1,"br 1") DCOMMA
  _statistics(0),
  _transitivity(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _mainVarIndex(0UL),
  _sideVarIndex(1UL),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _pseudoHeader((Term::Arity)1,(Term::Functor)0,Signature::PositivePolarity),
  _dummyClauseCollector(DOP("Chaining::_dummyClauseCollector")),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false)  
{
  _mainPremResidualSizeFormula.SetIndex(_mainVarIndex);
  _unificationMachine.SetIndex(_sideVarIndex);
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _sideConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; 
Chaining::~Chaining()
{  
  if (DestructionMode::isThorough())
    { 
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
      ASSERT(_sideConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_sideConstraintCheck);
    };
}; 
bool Chaining::loadQuery(Clause* cl)
{
  CALL("loadQuery(Clause* cl)");
  _query = cl;
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const Term* lit = iter.currentLiteral();
      const Clause* transLaw = 
	_transitivity->transitivityLaw(lit->functor());
      if (transLaw)
	{
	  _outerContinuation = LoadForwardPositive;
	  _mainPremResidualSizeFormula.Reset();
	  ulong remainingSelected = cl->LitList()->numOfSelectedLits();
	  for (LiteralList::Iterator iter1(cl->LitList()); 
	       iter1.notEnd(); 
	       iter1.next())
	    {
	      if (remainingSelected)
		{
		  if (!_transitivity->transitivityLaw(iter1.currentLiteral()->functor()))
		    {
		      _mainPremResidualSizeFormula.Compile(iter1.currentLiteral());
		    };
		  --remainingSelected;
		}
	      else 
		if (!iter1.currentLiteral()->isAnswerLiteral())
		  {
		    _mainPremResidualSizeFormula.Compile(iter1.currentLiteral());
		  };
	    };
	  if (_mainPremise.Load(_mainVarIndex,_query)) return true;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	}; 
      --numOfSelectedLits;
    };
  return false;
}; 
bool Chaining::nextInference()
{
  CALL("nextInference()");
  switch (_outerContinuation)
    {
    case LoadForwardPositive:      
      _innerContinuation = FirstLit;
    case NextForwardPositive:
      if (nextForwardPositive())
	{
	  _outerContinuation = NextForwardPositive;
	  return true;
	};         
      _innerContinuation = FirstLit;
    case NextBackwardPositive:
      if (nextBackwardPositive())
	{
	  _outerContinuation = NextBackwardPositive;
	  return true;
	};
      _innerContinuation = FirstLit;
    case NextForwardNegative1:
      if (nextForwardNegative1())
	{
	  _outerContinuation = NextForwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
    case NextBackwardNegative1:
      if (nextBackwardNegative1())
	{
	  _outerContinuation = NextBackwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
    case NextForwardNegative2:
      if (nextForwardNegative2()) 
	{
	  _outerContinuation = NextForwardNegative2;
	  return true;
	};         
      _innerContinuation = FirstLit;
    case NextBackwardNegative2:
      if (nextBackwardNegative2())
	{
	  _outerContinuation = NextBackwardNegative2;
	  return true;
	};
      return false;
#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
      return false;
#else
#  ifdef _SUPPRESS_WARNINGS_
      RuntimeError::report("Wrong _outerContinuation value in Chaining::nextInference()");
      return false;
#  endif
#endif
    };
  ICP("ICP1");
  RuntimeError::report("Illegal control point in Chaining::nextInference()");
  return false;
}; 
bool Chaining::nextForwardPositive()
{
  CALL("nextForwardPositive()");
  _forwardMode = true;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Negative())
    goto next_lit;
  _mainHead = lit->Arg2();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	  
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };
  _mainTail = lit->Arg1();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  ASSERT(_newLitHeader.Positive());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
      _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_mainTail,
			  _currentCandidate->chainingTail(),
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_mainTail,
			      _currentCandidate->chainingTail(),
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::nextBackwardPositive()
{
  CALL("nextBackwardPositive()");
  _forwardMode = false;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Negative())
    goto next_lit;
  _mainHead = lit->Arg1();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	 
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};  
    };
  _mainTail = lit->Arg2();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  ASSERT(_newLitHeader.Positive());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize > _clauseSizeLimit)
	{ 
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;        
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _mainTail,
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _mainTail,
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::nextForwardNegative1()
{
  CALL("nextForwardNegative1()");
  _forwardMode = true;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Negative())
    goto next_lit;
  _mainHead = lit->Arg1();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };
  _mainTail = lit->Arg2();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  _newLitHeader.flipPolarity();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg1ChainingTree(_newLitHeader.functor()))))
    {
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;  
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_mainTail,
			  _currentCandidate->chainingTail(),
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_mainTail,
			      _currentCandidate->chainingTail(),
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::nextBackwardNegative1()
{
  CALL("nextBackwardNegative1()");
  _forwardMode = false;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Positive())
    goto next_lit;
  _mainHead = lit->Arg1();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };
  _mainTail = lit->Arg2();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _mainTail,
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _mainTail,
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::nextForwardNegative2()
{
  CALL("nextForwardNegative2()");
  _forwardMode = true;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Negative())
    goto next_lit;
  _mainHead = lit->Arg2();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };
  _mainTail = lit->Arg1();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  _newLitHeader.flipPolarity();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg2ChainingTree(_newLitHeader.functor()))))
    {
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_currentCandidate->chainingTail(),
			  _mainTail,
			  false))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),
			      _mainTail,
			      false)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::nextBackwardNegative2()
{
  CALL("nextBackwardNegative2()");
  _forwardMode = false;
  UnifTerm* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentMainLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };
 next_lit:
  ++_currentMainLit;
 check_lit:
  if (_currentMainLit > _mainPremise.lastSelected()) return false;
  lit = _mainPremise.nthSelectedLit(_currentMainLit);
  ASSERT(lit);
  predNum = lit->symbol().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->symbol().Func().Positive())
    goto next_lit;
  _mainHead = lit->Arg2();
  varInHead = 0;
  if (_mainHead->symbol().isVariable())
    {
      varInHead = _mainHead->symbol().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  goto next_lit;
	};
    };
  _mainTail = lit->Arg1();
  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_mainHead,_mainTail);
  if (_headTailComparison == Less)
    {
      goto next_lit;
    };	
  _constraintOnMain = 0; 
  if ((_headTailComparison != Greater) &&
      (_headTailComparison != Equal))
    {
      ASSERT(_headTailComparison == Incomparable);
      _compiledConstraint->load(_mainTail,_mainHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreater();
	  _constraintOnMain = _compiledConstraint;
	};	
    };
  _newLitHeader = lit->symbol().Func();
  ASSERT(_newLitHeader.Negative());
  if (!_unificationMachine.CompilePseudoLiteral(_pseudoHeader,_mainHead))
    {
      ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
      goto next_lit;
    };
  _mainPremise.Cut(_currentMainLit);
  Unifier::current()->UnbindAll();
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
      _mainPremResidualSize = _mainPremResidualSizeFormula.Compute();
      if (_mainPremResidualSize >_clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_leaf;
	}; 
      if (_constraintOnMain &&
	  _constraintOnMain->holdsOnCurrentUnifier())
	{
	  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();
    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->practicallyMinResidualSize() 
	  + _query->practicallyMinResidualSize() 
	  > _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedBySize);
	  goto try_next_cand;
        }; 
      _sideConstraint = _currentCandidate->orderingConstraint();
      if (_sideConstraint)
	{	  
	  if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	      goto try_next_cand;
	    };
	};
      if (demodulateSelectedLits()) 
	{
	  if (_mainSelectedLitDemodulated) goto try_next_leaf; 
	  goto try_next_cand;
	};
      if (subsumeSidePremise()) goto try_next_cand;
      if (!_sidePremise.LoadOptimized(_sideVarIndex,_currentCandidate))
	{
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};
      if (!buildResolvent(_mainTail,
			  _currentCandidate->chainingTail(),
			  true))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;
    first_factor:
      _mainPremise.ResetFactoring();
      _sidePremise.ResetFactoring();
    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnMain &&
	      _constraintOnMain->holdsOnCurrentUnifier())
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
	      goto next_factor;
	    };
	  if (_sideConstraint)
	    {	  
	      if (_sideConstraintCheck->holds(_sideConstraint,_sideVarIndex))
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantDiscarded);
		  goto next_factor;
		};
	    };
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_mainTail,
			      _currentCandidate->chainingTail(),
			      true)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      goto try_next_cand;
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; 
  _mainPremise.UnmarkCut();
  goto next_lit;
}; 
bool Chaining::posChainingLitSelectedIn(Clause* cl)
{
  CALL("posChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const Term* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Positive()) return true;
	};
      --numOfSelectedLits;
    };
  return false;
}; 
bool Chaining::negChainingLitSelectedIn(Clause* cl)
{
  CALL("negChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const Term* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Negative()) return true;
	};
      --numOfSelectedLits;
    };
  return false;
}; 
bool Chaining::buildResolvent(UnifTerm* newLitArg1,
			      const Term& newLitArg2,
			      bool inheritFromMain)
{
  CALL("buildResolvent(UnifTerm* newLitArg1,const Term& newLitArg2,..)");
  ASSERT(newLitArg2.isVariable() || newLitArg2.IsReference()); 
  if ((!_query->isPersistent())
      || (!_sidePremise.TheClause()->isPersistent()))
    return false;
  if (_mainPremResidualSize + 
      _sidePremise.TheClause()->practicallyMinResidualSize() > 
      _clauseSizeLimit)
    {
      if (_forwardMode)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.discardedBySize);
	}
      else
	{
	  ++(_statistics->inferences.deduction.backwardChaining.discardedBySize);
	};
      return false;
    }; 
  _clauseCollector->openNewClause(((_forwardMode) ? 
				   ClauseBackground::ForwTransChain 
				   :
				   ClauseBackground::BackTransChain 
				   ),
				  _query,
				  _sidePremise.TheClause(),
				  const_cast<Clause*>(_currentTransitivityLaw));
  _clauseCollector->announceNewClause();
  LiteralProperties::SLModifier newLitSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier mainPremSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier sidePremSLMod = LiteralProperties::DefaultSLModifier;
  if (_mainPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_mainPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  mainPremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_mainPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  mainPremSLMod = LiteralProperties::PostponeSelection;	  
	};
    };
  if (_sidePremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_sidePremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  sidePremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_sidePremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  sidePremSLMod = LiteralProperties::PostponeSelection;
	};
    };
  if (!_mainPremise.collectUnmarkedLiterals(*_clauseCollector,mainPremSLMod))
    return false;
  if (!_sidePremise.collectUnmarkedLiterals(*_clauseCollector,sidePremSLMod))
    return false;
  LiteralProperties newLitProp;
  if (inheritFromMain)
    {
      newLitProp.inheritFrom(_mainPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    }
  else
    { 
      newLitProp.inheritFrom(_sidePremise.resolvedLiteralProperties(),
			     newLitSLMod);
    };
  newLitProp.setDeploymentMode(LiteralProperties::DefaultDeploymentMode);
  return
    _clauseCollector->openNewNonpropLiteral(_newLitHeader,newLitProp) &&
    _clauseCollector->pushTermWithUnifier(newLitArg1) &&
    _clauseCollector->pushTermWithUnifier(_sideVarIndex,newLitArg2) &&
    _clauseCollector->closeNonpropLiteral() &&
    _clauseCollector->finaliseClause();
}; 
bool Chaining::buildResolvent(const Term& newLitArg1,
			      UnifTerm* newLitArg2,
			      bool inheritFromMain)
{
  CALL("buildResolvent(const Term& newLitArg1,UnifTerm* newLitArg,..)");
  if ((!_query->isPersistent())
      || (!_sidePremise.TheClause()->isPersistent()))
    return false;
  if (_mainPremResidualSizeFormula.Compute() + 
      _sidePremise.TheClause()->practicallyMinResidualSize() > 
      _clauseSizeLimit)
    {
      if (_forwardMode)
	{
	  ++(_statistics->inferences.deduction.forwardChaining.discardedBySize);
	}
      else
	{
	  ++(_statistics->inferences.deduction.backwardChaining.discardedBySize);
	};
      return false;
    }; 
  _clauseCollector->openNewClause(((_forwardMode) ? 
				   ClauseBackground::ForwTransChain 
				   :
				   ClauseBackground::BackTransChain 
				   ),
				  _query,
				  _sidePremise.TheClause(),
				  const_cast<Clause*>(_currentTransitivityLaw));
  _clauseCollector->announceNewClause();
  LiteralProperties::SLModifier newLitSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier mainPremSLMod = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier sidePremSLMod = LiteralProperties::DefaultSLModifier;
  if (_mainPremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (inheritFromMain)
	newLitSLMod = LiteralProperties::AccelerateSelection; 
      if (_mainPremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  mainPremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_mainPremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  mainPremSLMod = LiteralProperties::PostponeSelection;	  
	};
    };
  if (_sidePremise.resolvedLiteralProperties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {      
      if (!inheritFromMain)
	newLitSLMod = LiteralProperties::AccelerateSelection; 
      if (_sidePremise.resolvedLiteralProperties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  sidePremSLMod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_sidePremise.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);
	  sidePremSLMod = LiteralProperties::PostponeSelection;
	};
    };
  if (!_mainPremise.collectUnmarkedLiterals(*_clauseCollector,mainPremSLMod))
    return false;
  if (!_sidePremise.collectUnmarkedLiterals(*_clauseCollector,sidePremSLMod))
    return false;
  LiteralProperties newLitProp;
  if (inheritFromMain)
    {
      newLitProp.inheritFrom(_mainPremise.resolvedLiteralProperties(),
			     newLitSLMod);
    }
  else
    { 
      newLitProp.inheritFrom(_sidePremise.resolvedLiteralProperties(),
			     newLitSLMod);
    };
  newLitProp.setDeploymentMode(LiteralProperties::DefaultDeploymentMode);
  return 
    _clauseCollector->openNewNonpropLiteral(_newLitHeader,newLitProp) &&
    _clauseCollector->pushTermWithUnifier(_sideVarIndex,newLitArg1) &&
    _clauseCollector->pushTermWithUnifier(newLitArg2) &&
    _clauseCollector->closeNonpropLiteral() &&
    _clauseCollector->finaliseClause();
}; 
bool Chaining::nextFactoring()
{
  CALL("nextFactoring()");
  if (_mainPremise.WasFactored())
   {
    if (_mainPremise.GetNextFactor())
     {
      return true;
     }
    else 
     {
      _mainPremise.ResetFactoring();
      if (_sidePremise.WasFactored())
       {
        return _sidePremise.GetNextFactor();
       }
      else
       {
        return _sidePremise.GetFirstFactor(_sidePremise.CutLit()+1);
       };
     };
   }
  else 
   {
    if (_mainPremise.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (_sidePremise.WasFactored())
       {
        return _sidePremise.GetNextFactor();
       }
      else
       {
        return _sidePremise.GetFirstFactor(_sidePremise.CutLit()+1);
       };
     };
   };
}; 
bool Chaining::subsumeMainPremise()
{
  CALL("subsumeMainPremise()"); 
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_query->hasSubsumedPremiseInstance()) return false;
  _query->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
  if (collectMainPremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause)) 
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _query);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _query->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  if (_forwardMode)
    {
      ++(_statistics->inferences.deduction.forwardChaining.mainPremisesSubsumed);
    }
  else
    {
      ++(_statistics->inferences.deduction.backwardChaining.mainPremisesSubsumed);
    };
 discarded:
  _query->unblockForSubsumption();
  return true;
}; 
bool Chaining::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_query->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; 
bool Chaining::subsumeSidePremise()
{
  CALL("subsumeSidePremise()"); 
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_currentCandidate->TheClause()->hasSubsumedPremiseInstance()) return false;
  _currentCandidate->TheClause()->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
  if (collectSidePremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause))  
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _currentCandidate->TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else 
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  goto discarded;
	};
      goto return_false;
    };
 return_false:
  _currentCandidate->TheClause()->unblockForSubsumption();
  return false;
 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  if (_forwardMode)
    {
      ++(_statistics->inferences.deduction.forwardChaining.sidePremisesSubsumed);
    }
  else
    {
      ++(_statistics->inferences.deduction.backwardChaining.sidePremisesSubsumed);
    }; 
 discarded:
  _currentCandidate->TheClause()->unblockForSubsumption();
  return true;
}; 
bool Chaining::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");
  for(LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sideVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties()))
      return false; 
  return true;
}; 
bool VK::Chaining::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  Clause* simplifyingClause;
  _mainSelectedLitDemodulated = false;
  if (_mainSelectedLitDemodulation)
    {   
      if (!_mainSelectedLitDemodulationAttempted)
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() &&
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {	  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);   
	      ASSERT(simplifyingClause != _query);
	      ASSERT(simplifyingClause != _currentCandidate->TheClause());
	      if (simplifyingClause)
		{	
		  _query->protectParamodulationsInto();
		  _currentCandidate->TheClause()->protectParamodulationsInto();
		  if (_forwardMode)
		    {
		      ++(_statistics->inferences.deduction.forwardChaining.mainPremisesDemodulated);
		    }
		  else
		    {
		      ++(_statistics->inferences.deduction.backwardChaining.mainPremisesDemodulated);
		    };
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	    }
	  else 
	    {
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (_forwardMode)
		{
		  ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
		}
	      else
		{
		  ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
		};
	      return tautology;
	    };
	};
    };
  if (_sideSelectedLitDemodulation)
    {
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 
      if (collectSidePremiseSubtermsAvailableForParamodulationInto() &&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{		    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
	  ASSERT(simplifyingClause != _query);
	  ASSERT(simplifyingClause != _currentCandidate->TheClause());
	  if (simplifyingClause)
	    {	
	      _query->protectParamodulationsInto();
	      _currentCandidate->TheClause()->protectParamodulationsInto();
	      if (_forwardMode)
		{
		  ++(_statistics->inferences.deduction.forwardChaining.sidePremisesDemodulated);
		}
	      else
		{
		  ++(_statistics->inferences.deduction.backwardChaining.sidePremisesDemodulated);
		};
	      return true;
	    }
	  else
	    return false;
	}
      else 
	{
	  bool tautology = 
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (_forwardMode)
	    {
	      ++(_statistics->inferences.deduction.forwardChaining.redundantBranchesDiscarded);
	    }
	  else
	    {
	      ++(_statistics->inferences.deduction.backwardChaining.redundantBranchesDiscarded);
	    };
	  return tautology; 
	};
    }
  else
    return false;
}; 
bool 
VK::Chaining::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");
  LiteralProperties prop;
  ulong numOfSelectedLits = _query->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_query->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      Term* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_mainVarIndex,
								      Term::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; 
bool 
VK::Chaining::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _currentCandidate->TheClause()->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      Term* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_sideVarIndex,
					     Term::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sideVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sideVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; 
