/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ClauseAssembler.hpp"
#include "LiteralList.hpp"
#include "BitWord.hpp"
#include "DiscTree.hpp"
#include "LiteralSelection.hpp"
#include "TemporaryLiteral.inl"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "TemporaryClause.hpp"
#include "Clause.inl"
#include "PagedPassiveClauseStorage.inl"
#include "LiteralProperties.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_ASSEMBLER
#define DEBUG_NAMESPACE "ClauseAssembler"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
ClauseAssembler::ClauseAssembler(LiteralSharing* litSharing,
				 const LiteralSelection* litSelection) 
  : 
  _litSharing(litSharing), 
  _litSelection(litSelection),
  _inheritNegativeSelection(false),
  _equalityIsPossible(true),
  _maxMaxWeight(VampireKernelConst::MaxClauseSize),
  _maxWeight(VampireKernelConst::MaxClauseSize)
{
};
void ClauseAssembler::reset()
{
  CALL("reset()");
  LiteralSharing* litSharing = _litSharing;
  const LiteralSelection* litSelection = _litSelection;
  BK::destroy(*this);
  BK::init(*this,litSharing,litSelection);
}; 
Clause* ClauseAssembler::assembleShared(TemporaryClause& cl)
{
  CALL("assembleShared(TemporaryClause& cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  ulong length = cl.literals()->length();
  Clause* res;
  if (length == 0UL)
    {
     res = assembleEmptyClause(cl.background());
     _litSharing->unlock();
     return res;
    };
  if (length == 1UL)
    {
      res = assembleUnitClause(*(cl.literals()->first()->literal()),
			       cl.background());
     _litSharing->unlock();
      return res;
    };
  res = assembleClause(*(cl.literals()),cl.background()); 
  _litSharing->unlock(); 
  return res;
}; 
Clause* ClauseAssembler::assembleSharedWithoutSelection(TemporaryClause& cl)
{
  CALL("assembleSharedWithoutSelection(TemporaryClause& cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  ulong length = cl.literals()->length();
  Clause* res;
  if (length == 0UL) 
    {
      res = assembleEmptyClause(cl.background());
     _litSharing->unlock();
     return res;
    };
  if (length == 1UL)
    {
      res =  assembleUnitClause(*(cl.literals()->first()->literal()),
				cl.background());
     _litSharing->unlock();
      return res;
    };
  res = assembleClauseWithoutSelection(*(cl.literals()),cl.background());
  _litSharing->unlock();
  return res;
}; 
bool 
ClauseAssembler::assembleInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						     TemporaryClause& cl,
						     long clauseNum,
						     ostream* showNewClause)
{
  CALL("assembleInPagedPassiveClauseStorage(..)");
  return
    assembleClauseInPagedPassiveClauseStorage(storage,
					      *(cl.literals()),
					      cl.background(),
					      clauseNum,
					      showNewClause);
}; 
Clause* ClauseAssembler::assembleNameIntro(const TemporaryLiteral& definedLit,
					   const BasicTempLitList& body)
{
  CALL("assembleNameIntro(..)");
  ASSERT(!body.isEmpty());  
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  LiteralList* lit_list = 
    new(body.length() + 1) 
    LiteralList(body.length() + 1,1UL);
  Clause* result = new Clause(lit_list);
  LiteralList::Iterator iter(lit_list);
  Term* lit;
  lit = shareTmpLiteral(definedLit);
  iter.currentElement()->setLiteral(lit);
  iter.currentElement()->properties() = definedLit.properties();
  iter.next();
  for (const BasicTempLitList::Element* el = body.first();
       el;
       el = el->next())
    {
      lit = shareTmpLiteral(*(el->literal()));
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      iter.currentElement()->properties() = el->literal()->properties();
      iter.next();
    };
  result->clearFlags();
  result->markAsNameIntro();
  ClauseBackground background;
  background.markAsNameIntro();
  background.copyTo(result->background());
  result->setSize(0);
  result->setWeightBase(0);
  result->setPracticallyMinResidualSize(0);
  result->setFlag(Clause::FlagLiteralsShared);
  _litSharing->unlock();
  return result; 
}; 
void ClauseAssembler::removeFromSharing(Clause* cl)
{  
  CALL("removeFromSharing(Clause* cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  for (LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
    removeFromSharing(iter.currentLiteral()); 
  _litSharing->unlock();
}; 
inline 
Term* 
ClauseAssembler::shareTmpLiteral(const TemporaryLiteral& lit) 
{
  CALL("shareTmpLiteral(const TemporaryLiteral& lit,const Clause* cl)");
  if (lit.isEquality())
    {
      return IntegrateEq(lit.header(),lit.arg1(),lit.arg2());
    }
  else 
    {
      return Integrate(lit.flatterm()); 
    };
}; 
inline Term* ClauseAssembler::Integrate(const Flatterm* lit)
{
  CALL("Integrate(const Flatterm* lit)");
  DiscTree::Node* leaf = _litSharing->Integrate(lit);
  leaf->entryMultiplicity()++;
  return leaf->term();
}; 
inline 
Term* 
ClauseAssembler::IntegrateEq(const Term& header,const Flatterm* arg1,const Flatterm* arg2)
{
  DiscTree::Node* leaf = _litSharing->IntegrateEq(header,arg1,arg2);
  leaf->entryMultiplicity()++;
  return leaf->term();
}; 
inline void ClauseAssembler::removeFromSharing(Term* lit)
{
  CALL("removeFromSharing(Term* lit,Clause* cl)");
  if (lit->arity())
    {
      DiscTree::Node** branch_to_destroy;
      DiscTree::Node* leaf;
      DiscTree::Node* last_fork;
      _litSharing->InfoForDeletion(lit,branch_to_destroy,leaf,last_fork);
      ASSERT(leaf->IsLeaf());
      leaf->entryMultiplicity()--;
      if (!(leaf->entryMultiplicity())) 
	{
	  Term* lit_copy = leaf->term();
	  DiscTree::DestroyBranchWithoutForks(*branch_to_destroy);
	  if (last_fork)
	    {
	      DiscTree::ReduceFork(last_fork);
	    }
	  else
	    {
	      *branch_to_destroy = 0;
	    }; 
	  _litSharing->destroyLiteralCopy(lit_copy);
	}; 
    }
  else 
    {
      ((*(_litSharing->Tree(*lit)))->entryMultiplicity())--;
      if (!(*(_litSharing->Tree(*lit)))->entryMultiplicity()) 
	{
	  delete *(_litSharing->Tree(*lit));
	  *(_litSharing->Tree(*lit)) = 0;
	};
    };
}; 
Clause* ClauseAssembler::assembleEmptyClause(OpenClauseBackground& ancestors) const
{
  CALL("assembleEmptyClause(OpenClauseBackground& ancestors) const");
  Clause* res = Clause::MakeEmptyClause();
  res->setSize(0);
  res->setPracticallyMinResidualSize(0);
  ancestors.writeNewBackgroundTo(res->background());
  return res;
}; 
Clause* ClauseAssembler::assembleUnitClause(TemporaryLiteral& lit,
					    OpenClauseBackground& ancestors) 
{
  CALL("assembleUnitClause(TemporaryLiteral& lit,OpenClauseBackground& ancestors)");
  lit.markAsSelected();
  lit.properties().setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
  lit.properties().setSelectionLayer(0UL);
  _intVarRenaming.Reset();
  lit.renameVariables(_intVarRenaming);
  LiteralList* lit_list;
  bool isAnswer = lit.isAnswerLiteral();
  long weightBase;
  bool posEqSelected;
  if (isAnswer)
    {
      weightBase = 0L;
      posEqSelected = false;
    }
  else
    {
      weightBase = lit.size();
      posEqSelected =
	lit.isEquality() &&
	lit.isPositive() && 
	lit.properties().useEqForParamodulationFrom();      
    };
  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqSelected);
  if (weight > _maxWeight) return 0;
  if (isAnswer)
    {
      lit_list = new (1) LiteralList(1,0);
    }
  else
    {
      lit_list = new (1) LiteralList(1,1);
    };
  Clause* result = new Clause(lit_list);
  lit_list->firstElem()->setLiteral(shareTmpLiteral(lit));
  LiteralProperties prop = lit.properties();
  if (_inheritNegativeSelection && 
      lit.isNegativeForSelection() &&
      !lit.isAnswerLiteral())
    {
      prop.setSelectabilityCategory(LiteralProperties::HighSelectability);
    };
  lit_list->firstElem()->properties() = prop;
  ancestors.writeNewBackgroundTo(result->background()); 
  result->setSize(lit.size());
  result->setWeightBase(weightBase);
  if (lit.isEquality() && lit.isPositive())
    { 
      result->setPracticallyMinResidualSize(lit.minPosEqLHSSize() + 1); 
    }
  else 
    result->setPracticallyMinResidualSize(0);
  result->setFlag(Clause::FlagLiteralsShared);
  if (isAnswer) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result;
}; 
Clause* ClauseAssembler::assembleClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)
{
  CALL("assembleClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(lits.checkIntegrity());
  if (lits.isEmpty()) return assembleEmptyClause(ancestors);
  long size;
  long weightBase;
  long num_of_pos_eq;
  if (_equalityIsPossible)
    {
      TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(lits,size,weightBase,num_of_pos_eq);
    }
  else
    { 
      TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
      num_of_pos_eq = 0;
    }; 
  BasicTempLitList max_lits;
  BasicTempLitList nonmax_lits; 
  BasicTempLitList tmp_lits;
  BasicTempLitList* ordered_max_lits;
  BasicTempLitList* ordered_nonmax_lits; 
  DOP(ulong debugLitsLength = lits.length());
  bool allMaximalSelected = false;
  if (ancestors.isSupport())
    {
      _litSelection->selectForSupport(lits,max_lits,nonmax_lits);
    }
  else 
    {
      _litSelection->select(lits,max_lits,nonmax_lits,allMaximalSelected);
    };
  ASSERT(lits.isEmpty());
  ASSERT((!max_lits.isEmpty()) || 
	 TempLitListExtras::containsOnlyAnswerLiterals(nonmax_lits));  
  bool posEqSelected = false;
  if (num_of_pos_eq)
    {
      for (const BasicTempLitList::Element* el = max_lits.first();
	   el && !posEqSelected;
	   el = el->next())
	{
	  posEqSelected = 
	    el->literal()->isEquality() &&
	    el->literal()->isPositive() &&
	    el->literal()->properties().useEqForParamodulationFrom();
	};
    };
  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqSelected);
  if (weight > _maxWeight) return 0;
  ulong num_of_max = max_lits.length();
  ulong num_of_nonmax = nonmax_lits.length();
  ASSERT(debugLitsLength == num_of_max + num_of_nonmax);
  if (num_of_max > 1)
    {
      _varWeighting.Reset();
      const BasicTempLitList::Element* el;
      for (el = max_lits.first(); el; el = el->next())
	_varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
      for (el = nonmax_lits.first(); el; el = el->next())
	_varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
      TempLitListExtras::passInOrder(max_lits,tmp_lits,_varWeighting);
      ordered_max_lits = &tmp_lits;
      if (num_of_nonmax > 1) 
	{
	  ASSERT(max_lits.isEmpty());
	  TempLitListExtras::passInOrder(nonmax_lits,max_lits,_varWeighting);
	  ordered_nonmax_lits = &max_lits; 
	}
      else 
	ordered_nonmax_lits = &nonmax_lits;  
    }
  else
    {
      ordered_max_lits = &max_lits;
      if (num_of_nonmax > 1)
	{
	  _varWeighting.Reset();
	  const BasicTempLitList::Element* el;
	  for (el = max_lits.first(); el; el = el->next())
	    _varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
	  for (el = nonmax_lits.first(); el; el = el->next())
	    _varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
	  TempLitListExtras::passInOrder(nonmax_lits,tmp_lits,_varWeighting);
	  ordered_nonmax_lits = &tmp_lits;
	}
      else 
	{
	  ordered_nonmax_lits = &nonmax_lits;
	};
    };  
  _intVarRenaming.Reset();
  TempLitListExtras::renameVariables(*ordered_max_lits,_intVarRenaming);
  TempLitListExtras::renameVariables(*ordered_nonmax_lits,_intVarRenaming);
  long weight_base_of_nonmax = 
    TempLitListExtras::computeWeightBase(*ordered_nonmax_lits);
  long min_size_of_max_pos_eq_lhs = 0;
  if (_equalityIsPossible)
    {
      min_size_of_max_pos_eq_lhs = 
	TempLitListExtras::minimalPosEqLHSSize(*ordered_max_lits);
      if (min_size_of_max_pos_eq_lhs) min_size_of_max_pos_eq_lhs++;
    };
  LiteralList* lit_list = 
    new(num_of_max + num_of_nonmax) 
    LiteralList(num_of_max + num_of_nonmax,num_of_max);
  Clause* result = new Clause(lit_list);
  LiteralList::Iterator iter(lit_list);
  Term* lit;
  LitHandle tmpLit;
  bool containsAnswerLiterals = false;
  while (num_of_max)
    {
      ordered_max_lits->popTo(tmpLit);
      lit = shareTmpLiteral(*tmpLit);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      ASSERT(!lit->isAnswerLiteral());
      LiteralProperties prop = tmpLit->properties();
      if (_inheritNegativeSelection &&
	  tmpLit->isNegativeForSelection())
	{	  
	  prop.setSelectabilityCategory(LiteralProperties::HighSelectability);
	};
      iter.currentElement()->properties() = prop;
      iter.next();
      num_of_max--;
    };
  while (num_of_nonmax)
    {
      ordered_nonmax_lits->popTo(tmpLit);
      lit = shareTmpLiteral(*tmpLit);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);   
      containsAnswerLiterals = containsAnswerLiterals || lit->isAnswerLiteral(); 
      iter.currentElement()->properties() = tmpLit->properties();
      iter.next();
      num_of_nonmax--;
    };
  ASSERT(iter.end());  
  ancestors.writeNewBackgroundTo(result->background());
  result->setSize(size);
  result->setWeightBase(weightBase);
  result->setPracticallyMinResidualSize(weight_base_of_nonmax + min_size_of_max_pos_eq_lhs);
  result->setFlag(Clause::FlagLiteralsShared);
  if (containsAnswerLiterals) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result; 
}; 
Clause* ClauseAssembler::assembleClauseWithoutSelection(BasicTempLitList& lits,
							OpenClauseBackground& ancestors)
{
  CALL("assembleClauseWithoutSelection(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(lits.checkIntegrity());
  if (lits.isEmpty()) return assembleEmptyClause(ancestors);
  long size;
  long weightBase;
  long num_of_pos_eq;
  if (_equalityIsPossible)
    {
      TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(lits,size,weightBase,num_of_pos_eq);
    }
  else
    { 
      TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
      num_of_pos_eq = 0;
    };    
  bool posEqSelected = false;
  if (num_of_pos_eq)
    {
      for (const BasicTempLitList::Element* el = lits.first();
	   el && !posEqSelected;
	   el = el->next())
	{
	  posEqSelected = 
	    el->literal()->isEquality() &&
	    el->literal()->isPositive() &&
	    el->literal()->properties().useEqForParamodulationFrom();
	};
    };
  _intVarRenaming.Reset();
  TempLitListExtras::renameVariables(lits,_intVarRenaming);
  long min_size_of_max_pos_eq_lhs = 0;
  if (_equalityIsPossible)
    {
      min_size_of_max_pos_eq_lhs = 
	TempLitListExtras::minimalPosEqLHSSize(lits);
      if (min_size_of_max_pos_eq_lhs) min_size_of_max_pos_eq_lhs++;
    };
  LiteralList* lit_list = 
    new(lits.length()) 
    LiteralList(lits.length(),lits.length());
  Clause* result = new Clause(lit_list);
  LiteralList::Iterator iter(lit_list);
  Term* lit;
  LitHandle tmpLit;
  bool containsAnswerLiterals = false;
  while (!lits.isEmpty())
    {
      lits.popTo(tmpLit);
      lit = shareTmpLiteral(*tmpLit);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      iter.currentElement()->properties() = tmpLit->properties();
      iter.next();
    };
  ASSERT(iter.end());  
  ancestors.writeNewBackgroundTo(result->background());
  result->setSize(size);
  result->setWeightBase(weightBase);
  result->setPracticallyMinResidualSize(min_size_of_max_pos_eq_lhs);
  result->setFlag(Clause::FlagLiteralsShared);
  if (containsAnswerLiterals) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result; 
}; 
bool 
ClauseAssembler::assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
							   BasicTempLitList& lits,
							   OpenClauseBackground& ancestors,
							   long clauseNum,
							   ostream* showNewClause)
{
  CALL("assembleClauseInPagedPassiveClauseStorage(..)");
  ASSERT(!lits.isEmpty());
  long size;
  long weightBase; 
  TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
  bool posEqHasToBeSelected = true;
  for (const BasicTempLitList::Element* el = lits.first();
       el && posEqHasToBeSelected;
       el = el->next())
    {
      posEqHasToBeSelected = 
	el->literal()->isEquality() &&
	el->literal()->isPositive() &&
	el->literal()->properties().useEqForParamodulationFrom();
    };
  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqHasToBeSelected);
  if (weight > _maxWeight) return false;
  BitWord rules = ancestors.Rules();
  ulong numOfLits = lits.length();
  ulong numOfAncestors = ancestors.numberOfAncestors();
  PagedPassiveClauseStorage::StoredClauseHandle newClauseHandle;
 try_to_open: 
  storage.openClause(newClauseHandle,clauseNum,size,weight,rules,numOfLits,numOfAncestors);
  if (newClauseHandle == 0)
    {
      if (storage.cancelWorstClause())
	goto try_to_open;
      return false;
    };
  writeIntoPagedPassiveClauseStorage(storage,lits);
  if (ancestors.numberOfAncestors())
    for (const Ancestor* anc = ancestors.last(); 
	 anc >= ancestors.begin(); 
	 anc--)
      {
	bool parent;
	Clause* cl = anc->strip(parent);
	storage.writeAncestor(cl,parent);
      };
  storage.endOfClause();
  if (showNewClause)
    {
      PagedPassiveClauseStorage::StoredClause newClause;
      PagedPassiveClauseStorage::loadClause(newClauseHandle,newClause);
      newClause.output(*showNewClause << "% ") << "\n";
    };
  return true;
}; 
void 
ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						    BasicTempLitList& lits)
{
  CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,BasicTempLitList& lits)"); 
  for (const BasicTempLitList::Element* el = lits.first(); 
       el;
       el = el->next())
    writeIntoPagedPassiveClauseStorage(storage,el->literal());   
}; 
void 
ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						    const LitHandle& lit)
{
  CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,const LitHandle& lit)");
  LiteralProperties prop = lit->properties();
  prop.makeSelected();
  storage.openLiteral(prop);
  Term header = lit->header();
  Flatterm::Iterator iter;
  if (header.IsEquality()) 
    {
      storage.writeSymbol(header);
      iter.Reset(lit->arg1());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym(); 
	}
      while (iter.NotEndOfTerm());
      iter.Reset(lit->arg2());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym();
	}
      while (iter.NotEndOfTerm());
    }
  else 
    {
      iter.Reset(lit->flatterm());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym();
	}
      while (iter.NotEndOfTerm());
    };
  storage.closeLiteral(); 
};  
