/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef FORWARD_MATCHING_SUBSTITUTION_CACHE_HPP
#define FORWARD_MATCHING_SUBSTITUTION_CACHE_HPP
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Flatterm.hpp"
#include "Term.hpp"
#ifndef NO_DEBUG
#  include "LinearCombination.hpp"
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
#  define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ForwardMatchingSubstitutionCache
  {
  public: 
    class Instance;
  public:
    ForwardMatchingSubstitutionCache() {};
    ~ForwardMatchingSubstitutionCache() {};
    void reset() 
    {
      DOP(_debugVariables.reset());
    };
    void associate(ulong var,const Flatterm* term)
    {
      CALL("associate(ulong var,const Flatterm* term)");
      ASSERT(!_debugVariables.coefficient(var));
      _subst[var] = term;
      DOP(_debugVariables.add1(var));
    };
    const Flatterm* map(ulong var) const 
    {
      CALL("map(ulong var) const");
      ASSERT(_debugVariables.coefficient(var));
      return _subst[var]; 
    };
#ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
      for (const ulong* v = _debugVariables.begin(); v < _debugVariables.end(); ++v)
	str << "X" << *v << " ---> " << map(*v) << "\n";
      return str;
    };
#endif
  private:
    const Flatterm* _subst[VampireKernelConst::MaxNumOfVariables];
#if (defined DEBUG_NAMESPACE) || (!(defined NO_DEBUG_VIS))
    BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _debugVariables;
#endif
  }; 
}; 
#ifndef NO_DEBUG_VIS
namespace std
{
  inline
  ostream& operator<<(ostream& str,const VK::ForwardMatchingSubstitutionCache* subst)
  {
    return subst->output(str);
  };
};
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
#define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class ForwardMatchingSubstitutionCache::Instance
  {
  public:
    Instance() : 
      _dummyTerm((Term::Arity)0,(Term::Functor)0,Signature::PositivePolarity)  
    {
    };
    ~Instance() {};
    void reset(const Term& t,const ForwardMatchingSubstitutionCache* subst);
    void reset(const Term* t,const ForwardMatchingSubstitutionCache* subst);
    const Term& symbol() const;
    bool next();
    bool after();
    bool substPosInSubstitution() const 
    {
      return !_onTop && _substIter.CurrPos()->isSubstitutionPosition();
    };
    static bool equal(const Term& t1,
		      const Term& t2,
		      const ForwardMatchingSubstitutionCache* subst);
  private:
    const ForwardMatchingSubstitutionCache* _subst;
    Term::Iterator _topLevel;
    Flatterm::ConstIterator _substIter;
    bool _onTop;
    Term _dummyTerm;
  }; 
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
#define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache::Instance"
#endif
#include "debugMacros.hpp"
namespace VK
{
  inline
  void 
  ForwardMatchingSubstitutionCache::Instance::reset(const Term& t,
						    const ForwardMatchingSubstitutionCache* subst)
  {
    CALL("reset(const Term& t,const ForwardMatchingSubstitutionCache* subst)");
    if (t.isVariable())
      {
	_subst = subst;
	_substIter.Reset(subst->map(t.var()));
	_topLevel.Reset(&_dummyTerm);
	NEVER(_topLevel.Next());
	_onTop = false;
      }
    else
      {
	ASSERT(t.IsReference());
	reset(t.deref(),subst);
      };
  }; 
  inline
  void 
  ForwardMatchingSubstitutionCache::Instance::reset(const Term* t,
						    const ForwardMatchingSubstitutionCache* subst)
  {
    CALL("reset(const Term* t,const ForwardMatchingSubstitutionCache* subst)");
    ASSERT(t->isComplex());
    _subst = subst;
    _topLevel.Reset(t);
    _onTop = true;
  }; 
  inline const Term& ForwardMatchingSubstitutionCache::Instance::symbol() const
  {
    CALL("symbol() const");
    if (_onTop) return _topLevel.symbolRef();
    ASSERT(!_substIter.EndOfTerm());
    return _substIter.CurrPos()->symbolRef();
  }; 
  inline bool ForwardMatchingSubstitutionCache::Instance::next()
  {
    CALL("next()");
    if (_onTop)
      {
      next_on_top:
	if (!_topLevel.Next()) return false;
	if (_topLevel.CurrentSymIsVar())
	  {
	    _onTop = false;
	    _substIter.Reset(_subst->map(_topLevel.symbolRef().var()));
	  };
	return true;
      }
    else 
      {
	_substIter.NextSym();
	if (!_substIter.EndOfTerm()) return true;
	_onTop = true;
	goto next_on_top;
      };
  }; 
  inline bool ForwardMatchingSubstitutionCache::Instance::after()
  {
    CALL("after()");
    if (_onTop)
      {
      next_on_top:
	if (!_topLevel.SkipTerm()) return false;
	if (_topLevel.CurrentSymIsVar())
	  {
	    _onTop = false;
	    _substIter.Reset(_subst->map(_topLevel.symbolRef().var()));
	  };
	return true;
      }
    else 
      {
	_substIter.after();
	if (!_substIter.EndOfTerm()) return true;
	_onTop = true;
	goto next_on_top;
      };
  }; 
  inline
  bool 
  ForwardMatchingSubstitutionCache::Instance::equal(const Term& t1,
						    const Term& t2,
						    const ForwardMatchingSubstitutionCache* subst)
  {
    CALL("equal(..)");
    Instance iter1;
    Instance iter2;
    iter1.reset(t1,subst);
    iter2.reset(t2,subst);
    do
      {
	if (iter1.symbol() != iter2.symbol()) return false;
	iter2.next();
      }
    while (iter1.next());
    return true;
  }; 
}; 
#endif
