/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef LITERAL_LIST_HPP
#define LITERAL_LIST_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
#include "Term.hpp"
#include "LiteralProperties.hpp"
using namespace std;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_LIST
#  define DEBUG_NAMESPACE "LiteralList"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class LiteralList
  {
  public:
    class Element
    {
    public: 
      bool isSelected() const { return _properties.isSelected(); };
      void markAsSelected() { _properties.makeSelected(); };
      void markAsUnselected() { _properties.makeUnselected(); };
      const LiteralProperties& properties() const { return _properties; };
      LiteralProperties& properties() { return _properties; };
      Term* literal() const
      {  
	CALL("Element::literal()");
	return _content;
      };
      Term literalRef() const 
      {
	CALL("literalRef() const");
	return Term(literal());
      }; 
      void setLiteral(Term* lit) 
      {
	CALL("setLiteral(Term* lit)");
	_content = lit;
      };
      Element& operator=(const Element& el) 
      {
	_content = el._content; 
	_properties = el._properties;
	return *this; 
      };
      Element* next() { return _next; };  
      const Element* next() const { return _next; };   
      Term::Arity arity() const { return literal()->arity(); };
      Term::Polarity polarity() const { return literal()->polarity(); };
      bool positive() const { return literal()->Positive(); };
      bool negative() const { return literal()->Negative(); };
      Term::Functor functor() const { return literal()->functor(); };
      ulong headerNum() const { return literal()->HeaderNum(); }; 
      bool isEquality() const { return literal()->IsEquality(); };
      bool isOrderedEq() const { return literal()->IsOrderedEq(); };
      bool isUnorderedEq() const { return literal()->IsUnorderedEq(); };
      bool isDefaultOrientableEq() const { 
	return 
	  properties().equalityOrientationMode() == 
	  LiteralProperties::DefaultEqualityOrientation;
      };
      bool isPreorientedEq() const {
	return 
	  properties().equalityOrientationMode() == 
	  LiteralProperties::GivenEqualityOrientation ||
	  properties().equalityOrientationMode() == 
	  LiteralProperties::ReverseGivenEqualityOrientation;
	  };
      bool isOrientableEq() const {
	return 
	  properties().equalityOrientationMode() != 
	  LiteralProperties::ProhibitEqualityOrientation;
      };
      bool isOrientedEq() const {
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::DefaultEqualityOrientation)
	  return isOrderedEq();
	return
	  properties().equalityOrientationMode() != 
	  LiteralProperties::ProhibitEqualityOrientation;
      };   
      bool orientationFollowsOrdering() const {
	CALL("orientationFollowsOrdering()");
	ASSERT(isOrientedEq());
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::DefaultEqualityOrientation)
	  return true;
	return 
	  isOrderedEq() && 
	  properties().equalityOrientationMode() == 
	  LiteralProperties::GivenEqualityOrientation;
      };
      Term* args() { return literal()->Args(); };
      const Term* args() const { return literal()->Args(); }; 
      Term* arg1() { return literal()->Arg1(); };
      const Term* arg1() const { return literal()->Arg1(); };
      Term* arg2() { return literal()->Arg2(); };
      const Term* arg2() const { return literal()->Arg2(); };
      Term* equalityLHS() {
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::ReverseGivenEqualityOrientation)
	  return literal()->Arg2();
	return literal()->Arg1();
      };
      const Term* equalityLHS() const {
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::ReverseGivenEqualityOrientation)
	  return literal()->Arg2();
	return literal()->Arg1();
      };
      Term* equalityRHS() {
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::ReverseGivenEqualityOrientation)
	  return literal()->Arg1();
	return literal()->Arg2();
      };
      const Term* equalityRHS() const {
	if (properties().equalityOrientationMode() == 
	    LiteralProperties::ReverseGivenEqualityOrientation)
	  return literal()->Arg1();
	return literal()->Arg2();
      };
    private:
      Element() {};
      Element(const Element&) {};
      Element(Element* nxt) { _next = nxt; };
      ~Element() {};
      void* operator new(size_t) 
      { 
	return BK::GlobAlloc::allocate(sizeof(Element)); 
      };
      void  operator delete(void* obj) 
      { 
	CALL("Element::operator delete(void* obj)");
	BK::GlobAlloc::deallocate(obj,sizeof(Element));
      };
    private:
      Term* _content;
      Element* _next;
      LiteralProperties _properties;
      friend class LiteralList; 
    }; 
    class Iterator;
    class ConstIterator;
  public:
    LiteralList(ulong numOfLits,ulong numOfSelectedLits)
    {
      CALL("constructor LiteralList(ulong numOfLits,ulong numOfSelectedLits)");
      ASSERT(numOfLits <= VampireKernelConst::MaxlNumOfLits);
      ASSERT(numOfSelectedLits <= numOfLits);
      ((_sizes = numOfLits) <<= 16) |= numOfSelectedLits;
      _elements = 0;
      for (ulong n = 0; n < numOfLits; n++)
	_elements = new Element(_elements);
    };
    ~LiteralList() 
    { 
      CALL("destructor ~LiteralList()");
      if (BK::DestructionMode::isThorough())
	{
	  while (_elements) 
	    { 
	      Element* tmp = _elements->next();
	      delete _elements;
	      _elements = tmp;  
	    };
	};
    };
    void* operator new(size_t,ulong numOfLits)
    {
      return BK::GlobAlloc::allocate(sizeof(LiteralList));
    };
    void operator delete(void* obj)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(LiteralList));
    };
    void operator delete(void* obj,ulong numOfLits)
    {
      BK::GlobAlloc::deallocate(obj,sizeof(LiteralList));
    };
    ulong length() const { return (_sizes >> 16); };
    ulong numOfSelectedLits() const { return ((_sizes << 16) >> 16); };
    Element* firstElem() 
    {
      return _elements; 
    };
    const Element* firstElem() const { return _elements; };
    Term* firstLit() const 
    {
      CALL("firstLit() const");
      ASSERT((length()) && (length() <= VampireKernelConst::MaxlNumOfLits));
      return _elements->literal(); 
    };
    Element* nth(ulong n)
    {
      CALL("nth(ulong n)");
      ASSERT(n <= length());
      Element* res = _elements;
      while (n) 
	{
	  res = res->next();
	  n--;
	};
      return res;
    };
    bool containsOnlyAnswerLiterals() const
    {
      for (const Element* el = firstElem(); el; el = el->next())
	if (!el->literal()->isAnswerLiteral()) return false;
      return true;
    };
    ostream& output(ostream& str,bool showLiteralProperties);
  private:
    LiteralList() { CALL("constructor LiteralList()"); ICP("0"); };
  private:
    ulong _sizes;  
    Element* _elements;
    friend class Iterator;
    friend class Element;
  }; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_LIST
#  define DEBUG_NAMESPACE "LiteralList::Iterator"
#endif 
#include "debugMacros.hpp"
  class LiteralList::Iterator
  {
  public:
    Iterator() : _current(0), _litNum(0UL)
    {
    };
    Iterator(LiteralList& lst) { reset(lst); };
    Iterator(LiteralList* lst) { reset(lst); };
    Iterator(LiteralList& lst,LiteralList::Element* start) { reset(lst,start); };
    Iterator(LiteralList* lst,LiteralList::Element* start) { reset(lst,start); };
    ~Iterator() {};
    void reset(LiteralList& lst) 
    {
      _current = lst.firstElem();
      _litNum = 0UL;
    };
    void reset(LiteralList* lst) 
    {
      _current = lst->firstElem();
      _litNum = 0UL;
    };
    void reset(LiteralList& lst,LiteralList::Element* start) 
    {
      CALL("reset(LiteralList& lst,LiteralList::Element* start)");
      _current = start;
      _litNum = 0UL;
    };
    void reset(LiteralList* lst,LiteralList::Element* start) 
    {
      CALL("reset(LiteralList* lst,LiteralList::Element* start)");
      _current = start;
      _litNum = 0UL;
    };
    operator bool() const { return !end(); };
    bool end() const { return !_current; };
    bool last() const 
    { 
      return (_current) && (!_current->next()); 
    }; 
    bool notEnd() const { return _current != 0; };
    void next() 
    {
      CALL("Iterator::next()");
      ASSERT(_current);
      _current = _current->next();
      _litNum++;
    };
    Element* currentElement() const
    {
      CALL("currentElement() const");
      return _current;
    };
    Term* currentLiteral() const
    {
      CALL("currentLiteral() const");
      ASSERT(_current);
      return _current->literal();
    };
    ulong litNum() const { return _litNum; };
  private:
    Element* _current;
    ulong _litNum; 
  }; 
}; 
#endif
