/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "LiteralSharing.hpp"
#include "Flatterm.hpp"
#include "DiscTree.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SHARING
 #define DEBUG_NAMESPACE "LiteralSharing"
#endif
#include "debugMacros.hpp"
using namespace BK;
using namespace VK;
LiteralSharing::LiteralSharing(TermSharing* si) : 
  _isLocked(false),
  _propositional(static_cast<Term*>(0),"LiteralSharing::_propositional"), 
  _termSharing(si),
  _discTree(static_cast<DiscTree::Node*>(0),"LiteralSharing::_discTree")
{
  CALL("constructor LiteralSharing(TermSharing* si)");
  DOP(_propositional.freeze());  
  DOP(_discTree.freeze()); 
}; 
LiteralSharing::~LiteralSharing() 
{ 
  CALL("destructor ~LiteralSharing()"); 
  if (DestructionMode::isThorough())
    {
      ulong n;
      for (n = 0UL; n < _discTree.size(); n++)
	{
	  DiscTree::Destroy(_discTree[n],DestroyLeafContent);
	  BK_CORRUPT(_discTree[n]);
	};
      for (n = 0UL; n < _propositional.size(); n++)
	if (_propositional[n]) 
	  {
	    delete (_propositional[n]);
	    BK_CORRUPT(_propositional[n]);
	  };
    };      
}; 
void LiteralSharing::reset()
{
  CALL("reset()");
  ulong n;
  for (n = 0UL; n < _discTree.size(); n++)
    {
      DiscTree::Destroy(_discTree[n],DestroyLeafContent);
      _discTree[n] = 0; 
    };
  for (n = 0UL; n < _propositional.size(); n++)
    if (_propositional[n])
      {
	ASSERT(!(_propositional[n]->arity()));
	delete (_propositional[n]);
	_propositional[n] = 0;
      };
  _isLocked = false;
}; 
DiscTree::Node* LiteralSharing::Integrate(Term* lit) 
{
  CALL("Integrate(Term* lit)");
  Term header = *lit;
  DiscTree::Node** tree = Tree(header);
  if (header.arity())
    {
      Term::Iterator iter;
      iter.Reset(lit);
      DiscTree::Surfer surf;
      surf.Reset(tree);
      if (*tree)
	{
	read_mode_next_sym:
	  if (iter.Next())
	    {
	      if (surf.Read(iter.symbolRef()))
		{
		  goto read_mode_next_sym;
		}
	      else
		{
		  surf.Write(iter.symbolRef());
		  goto write_mode_next_sym;
		}; 
	    }
	  else 
	    {
	      return surf.Leaf();
	    };
	}
      else
	{
	write_mode_next_sym:
	  if (iter.Next())
	    {
	      surf.Write(iter.symbolRef());
	      goto write_mode_next_sym;
	    }
	  else 
	    {
	      surf.CreateLeaf(_termSharing->CopyLiteral(lit));
	      return surf.Leaf();
	    };
	}; 
    }
  else 
    {
      if (!(*tree)) 
	{
	  *tree = new DiscTree::Node(ShareProp(header));
	};
      ASSERT((*tree)->IsLeaf());
      return *tree;
    };
}; 
DiscTree::Node* LiteralSharing::Integrate(const Flatterm* lit) 
{
  CALL("Integrate(const Flatterm* lit)");
  Term header = lit->symbolRef();
  DiscTree::Node** tree = Tree(header);
  if (header.arity())
    {
      Flatterm::ConstIterator iter;
      iter.Reset(lit);
      DiscTree::Surfer surf;
      surf.Reset(tree);
      if (*tree)
	{
	read_mode_next_sym:
	  iter.NextSym();
	  if (iter.EndOfTerm()) return surf.Leaf(); 
	  if (surf.Read(iter.CurrPos()->symbolRef())) { goto read_mode_next_sym; };
	  surf.Write(iter.CurrPos()->symbolRef()); 
	  goto write_mode_next_sym;
	}
      else
	{
	write_mode_next_sym:
	  iter.NextSym();
	  if (iter.EndOfTerm()) 
	    {
	      surf.CreateLeaf(_termSharing->CopyFlatLiteral(lit));
	      return surf.Leaf();
	    };       
	  surf.Write(iter.CurrPos()->symbolRef());
	  goto write_mode_next_sym;
	}; 
    }
  else 
    {
      if (!(*tree)) 
	{
	  *tree = new DiscTree::Node(ShareProp(header));
	};
      ASSERT((*tree)->IsLeaf());
      return *tree;
    };
}; 
DiscTree::Node* 
LiteralSharing::IntegrateEq(const Term& header,
			    const Flatterm* arg1,
			    const Flatterm* arg2) 
{
  CALL("IntegrateEq(const Term&,const Flatterm*,const Flatterm*)");
  DiscTree::Node** tree = Tree(header);
  Flatterm::ConstIterator iter;
  iter.Reset(arg1);
  bool first_arg = true;
  DiscTree::Surfer surf;
  surf.Reset(tree);
  if (*tree)
    {
    read_mode_next_sym:
      if (iter.EndOfTerm()) 
	{
	  if (first_arg)
	    {
	      iter.Reset(arg2);
	      first_arg = false;
	      goto read_mode_next_sym;
	    };
	  return surf.Leaf(); 
	};
      if (surf.Read(iter.CurrPos()->symbolRef())) 
	{
	  iter.NextSym();
	  goto read_mode_next_sym;
	};
      surf.Write(iter.CurrPos()->symbolRef()); 
      iter.NextSym();
      goto write_mode_next_sym;
    }
  else
    {
    write_mode_next_sym:
      if (iter.EndOfTerm()) 
	{
	  if (first_arg)
	    {
	      iter.Reset(arg2);
	      first_arg = false;
	      goto write_mode_next_sym;
	    };
	  surf.CreateLeaf(_termSharing->CopyFlatEq(header,arg1,arg2));
	  return surf.Leaf();
	};
      surf.Write(iter.CurrPos()->symbolRef());
      iter.NextSym();
      goto write_mode_next_sym;
    }; 
}; 
void LiteralSharing::InfoForDeletion(Term* lit,
                                     DiscTree::Node**& branch_to_destroy,
                                     DiscTree::Node*& leaf,
                                     DiscTree::Node*& last_fork)
{
  CALL("InfoForDeletion(Term*,DiscTree::Node**&,DiscTree::Node*&,DiscTree_NODE*&)");
  ASSERT(lit->arity());
  DiscTree::Node** tree = Tree(*lit); 
  ASSERT(*tree);
  ASSERT(!((*tree)->IsLeaf()));
  Term::Iterator iter;
  iter.Reset(lit);
  iter.Next(); 
  DiscTree::InfoForDeletion(iter,tree,branch_to_destroy,leaf,last_fork);
}; 
void LiteralSharing::DestroyLeafContent(DiscTree::Node* leaf)
{  
  CALL("DestroyLeafContent(DiscTree::Node* leaf)");
  if (leaf->term()->arity()) 
    {
      delete leaf->term();
    };
}; 
inline Term* LiteralSharing::ShareProp(const Term& header)
{
  CALL("ShareProp(const Term& header)");
  ASSERT(!header.arity());
  ulong hdNum = header.HeaderNum();
  DOP(_propositional.unfreeze());
  Term* res = _propositional.sub(hdNum);
  DOP(_propositional.freeze());
  if (!res) 
    {
      res = new (header.arity()) Term(header);
      _propositional[hdNum] = res;
    };
  return res;
};
DiscTree::Node** LiteralSharing::Tree(const Term& header)
{
  CALL("Tree(const Term& header)");
  DOP(_discTree.unfreeze());
  DiscTree::Node** res = &(_discTree.sub(header.HeaderNum()));
  DOP(_discTree.freeze());
  return res;
};
void LiteralSharing::destroyLiteralCopy(Term* lit)
{
  _termSharing->DestroyLiteralCopy(lit);
};
