/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PAGED_PASSIVE_CLAUSE_STORAGE_HPP
#define PAGED_PASSIVE_CLAUSE_STORAGE_HPP
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "GlobAlloc.hpp"
#include "BitWord.hpp"
#include "PageQueueDisciplineStorage.hpp"
#include "Term.hpp"
#include "ClauseBackground.hpp"
#include "Clause.hpp"
#include "LiteralProperties.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.hpp"
namespace VK
{
  class PagedPassiveClauseStorage
  {
  public:
    class ClauseProfile
    {
    public:
      ClauseProfile() {};
      ~ClauseProfile() {};
      long clauseNumber() const { return _clauseNumber; };
      long clauseSize() const { return _clauseSize; };
      long clauseWeight() const { return _clauseWeight; };
      const BK::BitWord& rules() const { return _rules; };
      ulong numOfLiterals() const { return _numOfLiterals; };
      ulong numOfAncestors() const { return _numOfAncestors; };
      ostream& output(ostream& str) const;
    private:
      long _clauseNumber;
      long _clauseSize;
      long _clauseWeight;
      BK::BitWord _rules;
      ulong _numOfLiterals;
      ulong _numOfAncestors;
      friend class PagedPassiveClauseStorage;
    }; 
    class ClauseLiterals
    {
    public:
      ClauseLiterals() {};
      ~ClauseLiterals() {};
      bool nextLiteral();
      bool nextSymbol(Term& sym);
      const LiteralProperties& currentLiteralProperties() const;
      void reset();
    private:
      const Term* _beginning;
      ulong _numberOfLiterals;
      const Term* _cursor;
      ulong _numOfRemainingLiterals;
      ulong _numOfRemainingSymbols;
      LiteralProperties _currentLiteralProperties;
      friend class PagedPassiveClauseStorage;
    }; 
    class ClauseAncestors
    {
    public:
      ClauseAncestors() {};
      ~ClauseAncestors() {};
      bool nextAncestor(Clause*& anc,bool& isParent);
      void reset();
    private:
      Clause** _beginning;
      ulong _numberOfAncestors;
      Clause** _cursor;
      ulong _numOfRemainingAncestors;
      friend class PagedPassiveClauseStorage;
    }; 
    class StoredClause
    {
    public:   
      StoredClause() {};
      ~StoredClause() {};
      ClauseProfile& profile() { return _profile; };
      ClauseLiterals& literals() { return _literals; };
      ClauseAncestors& ancestors() { return _ancestors; };
      long clauseNumber() const { return _profile.clauseNumber();  };
      long clauseSize() const { return _profile.clauseSize();  };
      const BK::BitWord& rules() const { return _profile.rules();  };
      ulong numOfLiterals() const { return  _profile.numOfLiterals(); };
      ulong numOfAncestors() const { return  _profile.numOfAncestors(); };   
      void resetLiterals() { _literals.reset(); };
      bool nextLiteral() { return _literals.nextLiteral(); };
      bool nextSymbol(Term& sym) { return _literals.nextSymbol(sym); };
      void resetAncestors() { _ancestors.reset(); };
      bool nextAncestor(Clause*& anc,bool& isParent) { 
	return _ancestors.nextAncestor(anc,isParent); 
      };
      bool isVIP() const { return rules().bit(ClauseBackground::VIP); };    
      bool isOrphan();
      ostream& output(ostream& str);    
    private: 
      ClauseProfile _profile;
      ClauseLiterals _literals;
      ClauseAncestors _ancestors;
      friend class PagedPassiveClauseStorage;
    }; 
    typedef void* StoredClauseHandle;
  public:
    PagedPassiveClauseStorage();
    ~PagedPassiveClauseStorage();
    void reset(); 
    static ulong pageSize() 
    { 
      return PageSize*sizeof(void*); 
    }; 
    void setAgeWeightRatio(long numerator,long denominator);
    void setBestGoodWeightRatio(long numerator,long denominator);
    void maximalPriorityOn() { _maximalPriorityFlag = true; };
    void maximalPriorityOff() { _maximalPriorityFlag = false; };
    bool empty() const { return _maxNonemptyCatNum < _minNonemptyCatNum; };
    bool nonempty() const { return _maxNonemptyCatNum >= _minNonemptyCatNum; };
    long totalNumberOfClauses() const { return _totalNumberOfClauses; };
    long numberOfSelectedClauses() const { return _numberOfSelectedClauses; };
    long numberOfCancelledClauses() const { return _numberOfCancelledClauses; };
    long numberOfCurrentlyKeptClauses() const 
    { 
      return _totalNumberOfClauses - _numberOfSelectedClauses - _numberOfCancelledClauses;
    };
    long maxPresentWeight() const { return _maxNonemptyCatNum; };
    void  openClause(StoredClauseHandle& handle,long clNum,long clSize,
		     long clWeight,const BK::BitWord rules,ulong numOfLits,
		     ulong numOfAncestors); 
    void openLiteral(const LiteralProperties& prop);
    void writeSymbol(const Term& sym);
    void closeLiteral();
    void writeAncestor(const Clause* ancestor,bool parent);
    void endOfClause();
    bool select(StoredClauseHandle& selectedClauseHandle);  
    bool select(StoredClause& selectedClause);
    void popSelected();
    bool cancelWorstClause();
    bool recyclePage()
    {
      CALL("recyclePage()");
      return _freePages.recyclePage();
    }; 
    static void loadClause(const StoredClauseHandle& handle,StoredClause& clause);
    ostream& outputProfile(ostream& str) const;
  private: 
    enum
    {
      PageSize = 16384UL 
    };
    typedef 
    BK::PageQueueDisciplineStorage<BK::GlobAlloc,PageSize> 
    Storage;
  private:
    long categoryNum(long clauseWeight);
    static ulong sizeInWords(ulong sizeInBytes);
    static void readFromMemory(const void* memory,
			       ClauseProfile& prof,
			       ClauseLiterals& lits,
			       ClauseAncestors& ancestors);
  private:
    enum 
    {
      MaximalCategoryNumber = VampireKernelConst::MaxClauseSize
    };
  private:
    long _ageWeightRatioNumerator;   
    long _ageWeightRatioDenominator;   
    long _bestGoodWeightNumerator;
    long _bestGoodWeightDenominator;
    bool _maximalPriorityFlag; 
    long _minNonemptyCatNum; 
    long _maxNonemptyCatNum;
    long _totalSelectedByAgeOrWeight;
    long _totalSelectedByWeight;
    long _totalNumberOfClauses;
    long _numberOfSelectedClauses;  
    long _numberOfCancelledClauses; 
    long _numberOfCurrentlyKept[MaximalCategoryNumber + 1];
    Storage::FreePages _freePages;
    Storage _category[MaximalCategoryNumber + 1];
    Storage* _currentCategory;
    long _selectedCategoryNumber;
    StoredClauseHandle _selectedClauseHandle;
    bool _lastSelectedByAge;
    bool _lastSelectedByWeight;
    void** _cursor;
    Term* _currentLiteral;
    ulong _currentLiteralSize;
#ifdef DEBUG_NAMESPACE
    bool _debugClauseOpenForWriting;
    void** _debugEndOfCurrentClause;
    ulong _debugNumberOfLiteralsToWrite;
    ulong _debugNumberOfAncestorsToWrite;
#endif  
    friend class Iterator;
    friend class ConstIterator;
  }; 
}; 
#endif
