/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef PATH_INDEXING_FOR_INSTANCE_RETRIEVAL_HPP
#define PATH_INDEXING_FOR_INSTANCE_RETRIEVAL_HPP 
#ifndef NO_DEBUG_VIS
#  include <iostream>
#endif  
#include "jargon.hpp"
#include "GlobalStopFlag.hpp"
#include "GlobAlloc.hpp"
#include "VampireKernelDebugFlags.hpp" 
#include "VampireKernelConst.hpp" 
#include "RuntimeError.hpp"
#include "Stack.hpp"
#include "PathIndex.hpp"
#include "CompactNodeSmallHeaderSkipList.hpp"
#include "GIncCountingSort.hpp"
#include "Multiset.hpp"
#include "LinearCombination.hpp"
#include "DestructionMode.hpp"
#include "Term.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEXING_FOR_INSTANCE_RETRIEVAL
#  define DEBUG_NAMESPACE "PathIndexingForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
namespace VK
{
  class PathIndexingForInstanceRetrieval
  { 
  public:
    PathIndexingForInstanceRetrieval();
    ~PathIndexingForInstanceRetrieval();
    bool integrate(const Term& integratedTermId);
    bool remove(const Term& removedTermId); 
    bool compileQuery(const Term& query);
    bool nextMatch(Term& matchId);
    const Term& subst(ulong var) const 
    { 
      CALL("subst(ulong var) const");
      ASSERT(_debugQueryVariables.coefficient(var));
      return _abstractMachine.substitution[var]; 
    };
    const Term* nativeVar(ulong varNum) const
    { 
      CALL("nativeVar(ulong varNum) const");
      ASSERT(_debugQueryVariables.coefficient(varNum));
      return _abstractMachine.substitution + varNum;
    };
#ifndef NO_DEBUG_VIS
    ostream& outputSubst(ostream& str) const;
#endif
  private:
    typedef BK::CNSHSkList<BK::GlobAlloc,Term::Arguments,Term> NonmaxPathRelation;
    typedef BK::CNSHVoidSkList<BK::GlobAlloc,Term> MaxPathRelation; 
    typedef PathIndex<NonmaxPathRelation,MaxPathRelation> PathIndexForInstanceRetrieval;
    class AbstractMachine
    {
    public: 
      enum InstructionTag
      {
	NonmaxEqualiseTermId,
	MaxEqualiseTermId,
	GetToTermId,
	LoadSubst,
	Compare,
	Success,
	GetToTermIdAndLoadSubst,
	GetToTermIdAndCompare  
      };
      class Instruction
      {
      public:
	Instruction() {}; 
	~Instruction() {}; 
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str,const AbstractMachine* machine,const Term* substitution) const;
#endif
      public:
	InstructionTag tag;
	union
	{
	  NonmaxPathRelation::Surfer* nonmaxPathRelationSurfer; 
	  MaxPathRelation::Surfer* maxPathRelationSurfer;
	}; 
	const NonmaxPathRelation::Node* const * nonmaxPathCurrNode; 
	ulong argNum;
	Term* subst;
      }; 
    public: 
      AbstractMachine();
      ~AbstractMachine();
      bool nextMatch(Term& matchId);
#ifndef NO_DEBUG_VIS
      ostream& outputCode(ostream& str) const; 
      ulong surferNum(const NonmaxPathRelation::Node* const * currNdAddr) const; 
      ulong surferNum(const NonmaxPathRelation::Surfer* surf) const;
      ulong surferNum(const MaxPathRelation::Surfer* surf) const;
#endif
    public:
      Instruction code[2*VampireKernelConst::MaxTermSize + 1];
#ifdef DEBUG_NAMESPACE
      Instruction* debugEndOfCode;
#endif
      Term substitution[VampireKernelConst::MaxTermSize];
      NonmaxPathRelation::Surfer nonmaxPathRelationSurfers[VampireKernelConst::MaxTermSize];
      MaxPathRelation::Surfer maxPathRelationSurfers[VampireKernelConst::MaxTermSize];
      bool firstRetrievalCall;
#ifdef INSTANCE_RETRIEVAL_INDEX_USE_WEIGHT_FILTERS
      ulong queryWeight;
#endif
#ifdef INSTANCE_RETRIEVAL_INDEX_USE_SIGNATURE_FILTERS
      SignatureFilter querySignatureFilter;
#endif
    }; 
    class IntermediateQueryRepresentation
    {
    public:
      IntermediateQueryRepresentation(AbstractMachine* abstractMachine); 
      ~IntermediateQueryRepresentation();
      void reset();
      void compound(ulong arity,const NonmaxPathRelation* pathRelation);
      void constant(const MaxPathRelation* pathRelation);
      void variable(ulong var);
      void endOfQuery();
      void generateCode();
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str);
#endif
    public:
      class Position
      {
      public:
	enum Tag { Start,Compound,Constant,Variable,End }; 
      public:
	Position() {};
#ifndef NO_DEBUG_VIS
	ostream& output(ostream& str,const Position* base) const;
#endif
      public: 
	Tag tag;
	union 
	{
	  const NonmaxPathRelation* nonmaxPathRelation;
	  const MaxPathRelation* maxPathRelation;
	};
	ulong argNum;      
	Position* nestPosition;
	ulong arity;
	ulong variable;
	ulong numOfVarArgs;
	Position* nextVarArg;
	bool hasNonvarArg;
	AbstractMachine::Instruction* mainInstruction;
      }; 
      class PositionPtr
      {
      public:
	PositionPtr() {};
	PositionPtr(Position* p) : ptr(p) {};
	~PositionPtr() {};
	ulong key() const 
	{
	  return (ptr->tag == Position::Constant) 
	    ? ptr->maxPathRelation->size() 
	    : ptr->nonmaxPathRelation->size();                     
	};
      public:
	Position* ptr;
      };   
    private: 
      void generatePrimaryCode();
      void generateSecondaryCode();
      void generateCodeForSingletons();
      void writeNonmaxEqualiseTermId(Position* primaryNest);
      void writeMaxEqualiseTermId(Position* primaryNest);
      void writeGetToTermId(Position* secondaryNest);
      void writeCompare(Position* varPos);
      void writeLoadSubst(Position* varPos);
      void superimposeCompare(AbstractMachine::Instruction* instr,
			      Position* varPos);
      void superimposeLoadSubst(AbstractMachine::Instruction* instr,
				Position* varPos);
    private: 
      Position _positions[VampireKernelConst::MaxTermSize + 2];
      AbstractMachine* _abstractMachine;
      Position* _nextPosition;
      ulong _currentHoles;
      Position* _currentNest;
      BK::Stack<ulong,VampireKernelConst::MaxTermDepth> _backtrackHoles;
      BK::Stack<Position*,VampireKernelConst::MaxTermDepth> _backtrackNonmaxPositions;
      BK::GIncCountingSort<ulong,PositionPtr,VampireKernelConst::MaxTermSize> _primaryNestSorting;
      BK::Stack<Position*,VampireKernelConst::MaxTermSize> _secondaryNests; 
      Position* _primaryPositions[VampireKernelConst::MaxTermSize]; 
      ulong _numOfPrimaryPositions;
      bool _substLoaded[VampireKernelConst::MaxNumOfVariables];  
      Position* _availableVarPosition[VampireKernelConst::MaxNumOfVariables];
      BK::Multiset<ulong,VampireKernelConst::MaxNumOfVariables> _internalVars;   
      BK::MultisetWithRemoval<ulong,VampireKernelConst::MaxNumOfVariables> _singletonVars;
      AbstractMachine::Instruction* _nextInstr;
      AbstractMachine::Instruction* _lastCompare;
    }; 
  private: 
    PathIndexForInstanceRetrieval::Trie _trie; 
    PathIndexForInstanceRetrieval::Integrator _integrator;
    Term::Traversal _integratedTermTraversal;
    PathIndexForInstanceRetrieval::Removal _removal;
    Term::Traversal _removedTermTraversal;
    Term::Traversal _queryTraversal;
    PathIndexForInstanceRetrieval::Retrieval _pathRelationRetrieval;
    IntermediateQueryRepresentation _intermediateQueryRepresentation; 
#ifndef NO_DEBUG
    BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _debugQueryVariables;
#endif
    AbstractMachine _abstractMachine;
  }; 
}; 
#endif
