/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "ProblemProfile.hpp"
#include "Clause.inl"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROBLEM_PROFILE
#  define DEBUG_NAMESPACE "ProblemProfile"
#endif
#include "debugMacros.hpp"
using namespace VK;
ProblemProfile::ProblemProfile() :
  _predInfoTable(PredInfo(),"ProblemProfile::_predInfoTable"),
  _funcInfoTable(FuncInfo(), "ProblemProfile::_funcInfoTable")
{
  CALL("constructor ProblemProfile()");
  DOP(_predInfoTable.freeze());
  DOP(_funcInfoTable.freeze());
  reset(); 
};
ProblemProfile::~ProblemProfile() { CALL("destructor ~ProblemProfile()"); };
void ProblemProfile::reset()
{
  CALL("reset()");
  _clauses.reset();
  _numOfEqLiterals = 0;
  _numOfPred = 0;
  _numOfFuncs = 0;
  for (ulong f = 0; f < _funcInfoTable.size(); f++) _funcInfoTable[f].reset();
  for (ulong p = 0; p < _predInfoTable.size(); p++) _predInfoTable[p].reset();
  _symbolInfoTable.reset();
}; 
void ProblemProfile::load(const Clause* cl) 
{
  CALL("load(const Clause* cl)"); 
  ASSERT(!cl->isNameIntro());
  ASSERT(!cl->isSupportClause());
  _clauses.enqueue(cl);
  _numOfEqLiterals += cl->NumOfEqLits();   
  ulong transPred;
  if (cl->isSymmetryLaw())
    {
      _symbolInfoTable.registerSymmetry(cl->LitList()->firstLit()->functor(),cl);
      predInfo(cl->LitList()->firstLit()->functor())._symmetric = true;
    }
  else
    if (cl->isTransitivityLaw(transPred) && 
	!VampireKernelConst::isEquality(transPred))
      {
	_symbolInfoTable.registerTransitivity(transPred,cl);
	predInfo(transPred)._transitive = true;
      };
  for(LiteralList::Iterator iter1(cl->LitList());iter1.notEnd();iter1.next())
    {
      const Term* lit1 = iter1.currentLiteral();
      ulong pred1 = lit1->functor(); 
      predInfo(pred1).regPred(*lit1);
    };               
  ulong commFunc;
  if (cl->isCommutativityLaw(commFunc))
    {
      _symbolInfoTable.registerCommutativity(commFunc,cl);
      funcInfo(commFunc)._commutative = true;
    };
  Term::Iterator iter;
  for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
    {
      if (litIter.currentLiteral()->arity())
	{
	  iter.Reset(litIter.currentLiteral());
	  ALWAYS(iter.Next());
	  do
	    {     
	      if (!iter.CurrentSymIsVar()) 
		{
		  funcInfo(iter.symbolRef().functor())._registered++; 
		  funcInfo(iter.symbolRef().functor())._symbol = iter.symbolRef();
		};
	    }
	  while (iter.Next());
	};
    };
}; 
void ProblemProfile::loadBackground(const Clause* cl) 
{
  CALL("loadBackground(const Clause* cl)"); 
  ASSERT(cl->isNameIntro() || cl->isSupportClause());
  _clauses.enqueue(cl);
  _numOfEqLiterals += cl->NumOfEqLits();   
  for(LiteralList::Iterator iter1(cl->LitList());iter1.notEnd();iter1.next())
    {
      const Term* lit1 = iter1.currentLiteral();
      ulong pred1 = lit1->functor(); 
      predInfo(pred1).regPred(*lit1);
    };               
  Term::Iterator iter;
  for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
    {
      if (litIter.currentLiteral()->arity())
	{
	  iter.Reset(litIter.currentLiteral());
	  ALWAYS(iter.Next());
	  do
	    {     
	      if (!iter.CurrentSymIsVar()) 
		{
		  funcInfo(iter.symbolRef().functor())._registered++; 
		  funcInfo(iter.symbolRef().functor())._symbol = iter.symbolRef();
		};
	    }
	  while (iter.Next());
	};
    };
}; 
void ProblemProfile::summarise()
{
  CALL("summarise()");
  for (ulong p = 0; p < _predInfoTable.size(); p++) 
    {                
      if ((predInfo(p)._registered) && (!predInfo(p)._symbol.IsEquality()))
	{
	  _numOfPred++;
	};
    };  
}; 
ostream& ProblemProfile::output(ostream& str) const
{
  str << "% Number of non-equality predicates: " << _numOfPred << '\n';
  if (_symbolInfoTable.numOfSymmetric()) 
    str << "% Number of symmetric predicates: " 
	<< _symbolInfoTable.numOfSymmetric()  << '\n';
  if (_symbolInfoTable.numOfTransitive()) 
    str << "% Number of transitive predicates: " 
	<< _symbolInfoTable.numOfTransitive()  << '\n';
  str << "% Predicates: {";           
  for (ulong p = 0; p < _predInfoTable.size(); p++)
    if ((_predInfoTable[p]._registered) && (!_predInfoTable[p]._symbol.IsEquality()))
      _predInfoTable[p].output(str << '[') << "] ";   
  str << "}\n"; 
  str << "% Number of functors: " << _numOfFuncs << '\n';
  if (_symbolInfoTable.numOfCommutative())
    str << "% Number of commutative functors: " 
	<< _symbolInfoTable.numOfCommutative() << '\n';
  str << "% Functors: {";
  for (ulong f = 0; f < _funcInfoTable.size(); f++)
    if (_funcInfoTable[f]._registered) _funcInfoTable[f].output(str << '[') << "] ";
  str << "}\n"; 
  return str;
}; 
