/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TemporaryClauseCollector.hpp"
#include "TemporaryLiteral.inl"
#include "PagedPassiveClauseStorage.inl"
#include "Clause.inl"
#include "InstanceWithUnifier.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_COLLECTOR
#  define DEBUG_NAMESPACE "TemporaryClauseCollector"
#endif
#include "debugMacros.hpp"
TemporaryClauseCollector::TemporaryClauseCollector() :  
  OPEN_BR(1,"br 1") DCOMMA
_markSubstitutionPositions(false),
  _showAnnouncedClauses(0),
  _showCompleteClauses(0),
  _preprocessingInProgress(false),
  _outputQueue(0),
  _rejectionReason(None),
  _statistics(0),
  _tautologyCheck(true),
  _literalFactoring(true),
  _sizeLimit(VampireKernelConst::MaxClauseSize),
  _literalSizeLimit(VampireKernelConst::MaxTermSize),
  _lengthLimit(VampireKernelConst::MaxlNumOfLits)
{    
#ifndef NO_DEBUG
  _name = "?";
#endif
  DOP(_debugTrash = true);
};
#ifndef NO_DEBUG
TemporaryClauseCollector::TemporaryClauseCollector(const char* name) :
  OPEN_BR(1,"br 1") DCOMMA 
_markSubstitutionPositions(false),
  _showAnnouncedClauses(0),
  _showCompleteClauses(0),
  _preprocessingInProgress(false),
  _outputQueue(0),
  _rejectionReason(None),
  _statistics(0),
  _tautologyCheck(true),
  _literalFactoring(true),
  _sizeLimit(VampireKernelConst::MaxClauseSize),
  _literalSizeLimit(VampireKernelConst::MaxTermSize),
  _lengthLimit(VampireKernelConst::MaxlNumOfLits)
{    
  _name = name;
  DOP(_debugTrash = true);
};
#endif
TemporaryClauseCollector::~TemporaryClauseCollector()
{
  CALLM(_name,"destructor ~TemporaryClauseCollector()");
}; 
void TemporaryClauseCollector::reset()
{
  CALLM(_name,"reset()");
  _markSubstitutionPositions = false; 
  _showAnnouncedClauses = 0;
  _showCompleteClauses = 0;
  _preprocessingInProgress = false;
  _outputQueue = 0;
  _rejectionReason = None;
  _newClause.clear();
  _newLiteral.clear();
  _statistics = 0;
  _tautologyCheck = true;
  _literalFactoring = true;
  _sizeLimit = VampireKernelConst::MaxClauseSize;
  _literalSizeLimit = VampireKernelConst::MaxTermSize;
  _lengthLimit = VampireKernelConst::MaxlNumOfLits;
  DOP(_debugTrash = true);
}; 
void 
TemporaryClauseCollector::openNewClause(ClauseBackground::Rule rule)
{
  CALLM(_name,"openNewClause(ClauseBackground::Rule rule)");
  registerGenerated(rule);
  _rejectionReason = None;
  if (_newClause.isNull())
    {
      TemporaryClause::Extras::freshClause(_newClause);
    }
  else
    _newClause->reset();
  _newClause->background().UsedRule(rule);
  if (_preprocessingInProgress)
    _newClause->background().UsedRule(ClauseBackground::Prepro);
  _unifierVarRenaming.Reset();
  _nonanswerSize = 0UL;
  _totalSize = 0UL;
  DOP(_debugTrash = false);
}; 
void TemporaryClauseCollector::announceNewClause()
{
  CALL("announceNewClause()");
  if (_showAnnouncedClauses)
    _newClause->outputAnnounced((*_showAnnouncedClauses) << "% ") << "\n";
}; 
bool TemporaryClauseCollector::finaliseClause()
{
  CALLM(_name,"finaliseClause()");
  ASSERT(!_debugTrash);
  TempLitListExtras::preorientEqualities(*(_newClause->literals()));
  _newClause->resetSimplificationFlags();
  _newClause->markAsWeeded();
  if (_showCompleteClauses)
    _newClause->outputComplete((*_showCompleteClauses) << "% ") << "\n";
  _outputQueue->collect(_newClause);
  DOP(_debugTrash = true);
  return true; 
}; 
bool TemporaryClauseCollector::finaliseClause(ClauseHandle& recipient)
{
  CALLM(_name,"finaliseClause(recipient)");  
  ASSERT(!_debugTrash);
  TempLitListExtras::preorientEqualities(*(_newClause->literals()));
  _newClause->resetSimplificationFlags();
  _newClause->markAsWeeded();
  if (_showCompleteClauses)
    _newClause->outputComplete((*_showCompleteClauses) << "% ") << "\n";
  recipient.collectFrom(_newClause);
  DOP(_debugTrash = true);
  return true; 
}; 
bool 
TemporaryClauseCollector::finalisePropNameIntroPart(const TemporaryClause& parentCl)
{
  CALLM(_name,"finalisePropNameIntroPart(const TemporaryClause& parentCl)");
  ASSERT(!_debugTrash);
  TempLitListExtras::preorientEqualities(*(_newClause->literals()));
  _newClause->resetSimplificationFlagsInPropIntroPart(parentCl);
  _newClause->markAsWeeded();
  if (_showCompleteClauses)
    _newClause->outputComplete((*_showCompleteClauses) << "% ") << "\n";
  _outputQueue->collect(_newClause);
  DOP(_debugTrash = true);
  return true; 
}; 
bool 
TemporaryClauseCollector::pushLiteral(const Term* lit,
				      const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteral(const Term* lit,const LiteralProperties& prop)");
  ASSERT(!_debugTrash);
  ASSERT(!_newClause.isNull());
  Term header = *lit;
  if (!(header.arity())) 
    {
      return pushPropLiteral(header,prop);
    };
  if (!openNewNonpropLiteral(header,prop)) 
    return false;
  Term::Iterator iter;
  iter.Reset(lit);
  while (iter.Next()) 
    {
      if (iter.CurrentSymIsVar())
	{
	  pushVar(iter.symbolRef(),false);
	}
      else
	{
	  if (iter.symbolRef().arity())
	    {
	      if (!pushNonconstFunc(iter.symbolRef(),false)) return false;
	    }
	  else 
	    pushConst(iter.symbolRef(),false); 
	}; 
    };
  return closeNonpropLiteral();
}; 
bool 
TemporaryClauseCollector::pushLiteral(const Term& header,
				      const Flatterm* arg1,
				      const WeightPolynomial::Monomials& varsForPadding,
				      const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteral(const Term& header,const Flatterm* arg1,..)");
  ASSERT(!_debugTrash);
  ASSERT(header.arity() == 1 + varsForPadding.length());  
  ASSERT(!header.IsEquality());
  if (_newLiteral.isNull())
    {
      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	{
	  reject(None);
	  return false;
	};
    };
  if (!_newLiteral->pushPseudoLit(header,arg1,varsForPadding,prop)) 
    {
      reject(LiteralTooDeep);
      return false;
    };
  _totalSize += _newLiteral->size();
  if (_totalSize > VampireKernelConst::MaxClauseSize)
    {
      reject(ClauseTooLarge);
      return false;
    };
  _collectingNonanswerLiteral = !header.isAnswerHeader();
  if (_collectingNonanswerLiteral)
    {
      _nonanswerSize += _newLiteral->size();
      ASSERT(_nonanswerSize <= _totalSize);
      if (_nonanswerSize > _sizeLimit)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  return closeNonpropLiteral();
}; 
bool 
TemporaryClauseCollector::loadClause(const Clause* c,
				     ClauseBackground::Rule rule,
				     Clause* parent)
{
  CALLM(_name,"loadClause(const Clause* c,..)");
  openNewClause(rule,parent);
  if (c->isPartOfNameIntroForPropSplitting()) 
    {
      addRule(ClauseBackground::PartOfNameIntroForPropSplitting);
    };
  announceNewClause();
  for(LiteralList::Iterator iter(c->LitList());
      iter.notEnd();
      iter.next())
    if (!(pushLiteral(iter.currentLiteral(),
		      iter.currentElement()->properties()))) 
      return false;       
  return true;
}; 
bool 
TemporaryClauseCollector::loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,
				     ClauseBackground::Rule rule)
{
  CALLM(_name,"loadClause(PagedPassiveClauseStorage::ClauseLiterals& literals,ClauseBackground::Rule rule)");
  literals.reset();
  openNewClause(rule);
  announceNewClause();
  while (literals.nextLiteral())
    {
      LiteralProperties prop = literals.currentLiteralProperties();
      Term sym;
      ALWAYS(literals.nextSymbol(sym)); 
      if (!sym.arity()) 
	{
	  if (!pushPropLiteral(sym,prop)) return false;
	}
      else
	{
	  if (!openNewNonpropLiteral(sym,prop)) return false; 
	  while (literals.nextSymbol(sym)) 
	    if (!pushSymbol(sym,false)) return false;
	  if (!closeNonpropLiteral()) return false;
	};
    };
  return true;
}; 
bool 
TemporaryClauseCollector::pushLiteralWithUnifier(ulong index,
						 const Term* lit,
						 const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithUnifier(ulong index,const Term* lit,const LiteralProperties& prop)");
  ASSERT(!_debugTrash);
  Term header = *lit;
  if (!header.arity()) return pushPropLiteral(header,prop);
  if (!openNewNonpropLiteral(header,prop)) return false;
  InstanceWithUnifier term;
  term.Reset(index,lit);
  term.SkipSym();
  UnifTerm::Symbol sym;
  if (_markSubstitutionPositions)
    {
      while (!term.Empty())
	{
	  sym = term.CurrSym();
	  if (sym.isVariable()) 
	    {
	      if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) return false; 
	    }
	  else
	    {
	      if (sym.Func().arity())
		{
		  if (!pushNonconstFunc(sym.Func(),
					term.currentPositionIsInSubstitution())) 
		    return false;
		}
	      else 
		{
		  pushConst(sym.Func(),
			    term.currentPositionIsInSubstitution());
		}; 
	    };
	  term.SkipSym();
	};
    }
  else 
    {
      while (!term.Empty())
	{
	  sym = term.CurrSym();
	  if (sym.isVariable()) 
	    {
	      if (!pushVar(_unifierVarRenaming.Rename(sym.var()))) return false; 
	    }
	  else
	    {
	      if (sym.Func().arity())
		{
		  if (!pushNonconstFunc(sym.Func())) return false;
		}
	      else 
		{
		  pushConst(sym.Func());
		}; 
	    };
	  term.SkipSym();
	};
    };
  return closeNonpropLiteral();  
}; 
bool 
TemporaryClauseCollector::pushLiteralWithUnifier(UnifTerm* lit,
						 const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithUnifier(UnifTerm* lit,..)");
  ASSERT(!_debugTrash);
  ASSERT(UnifTerm::CorrectComplexTerm(lit));
  Term header = lit->symbol().Func();
  if (!header.arity()) 
    return pushPropLiteral(header,prop);
  if (!openNewNonpropLiteral(header,prop)) return false;
  UnifTerm::Iter iter;
  iter.Reset(lit);
  Term sym;
  iter.SkipSym(); 
  if (_markSubstitutionPositions)
    {
      do
	{
	  if (iter.CurrVar()) 
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true))
		return false;   
	    }
	  else 
	    {
	      sym = iter.CurrSubterm()->symbol().Func();
	      if (sym.arity())
		{
		  if (!pushNonconstFunc(sym,
					iter.currentPositionIsInSubstitution())) 
		    return false;
		}
	      else 
		pushConst(sym,
			  iter.currentPositionIsInSubstitution());
	    };
	  iter.SkipSym();
	}
      while (!iter.Empty());
    }
  else 
    {
      do
	{
	  if (iter.CurrVar()) 
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar())))
		return false;   
	    }
	  else 
	    {
	      sym = iter.CurrSubterm()->symbol().Func();
	      if (sym.arity())
		{
		  if (!pushNonconstFunc(sym)) return false;
		}
	      else 
		pushConst(sym);
	    };
	  iter.SkipSym();
	}
      while (!iter.Empty());
    };
  return closeNonpropLiteral();
}; 
bool 
TemporaryClauseCollector::pushEqLiteralWithUnifier(const Term& header,
						   UnifTerm* arg1,
						   UnifTerm* arg2,
						   const LiteralProperties& prop)
{
  CALLM(_name,"pushEqLiteralWithUnifier(const Term& header,..)"); 
  ASSERT(!_debugTrash);
  ASSERT(header.IsEquality());   
  if (_markSubstitutionPositions)
    {
      return openNewNonpropLiteral(header,prop)
	&& pushTermWithUnifier(arg1,false)
	&& pushTermWithUnifier(arg2,false)
	&& closeNonpropLiteral();
    }
  else 
    {
      return openNewNonpropLiteral(header,prop)
	&& pushTermWithUnifier(arg1)
	&& pushTermWithUnifier(arg2)
	&& closeNonpropLiteral();
    };
}; 
bool 
TemporaryClauseCollector::pushLiteralWithReplacement(ulong index,
						     const Term* lit,
						     const Term* redex,
						     UnifTerm* replacingTerm,
						     const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithReplacement(ulong index, const Term* lit,...)");
  ASSERT(!_debugTrash);
  Term header = *lit;
  ASSERT(header.arity()); 
  if (header.IsEquality()) header.MakeUnordered();
  if (!openNewNonpropLiteral(header,prop)) 
    return false;
  InstanceWithUnifier iter;
  iter.Reset(index,lit);
  iter.SkipSym();
  UnifTerm::Symbol sym;
  if (_markSubstitutionPositions)
    {
      while (!iter.Empty())
	{
	  if (iter.CurrPosInBase() == redex)
	    {
	      if (!pushTermWithUnifier(replacingTerm,false)) return false;
	      iter.SkipTerm();
	    }
	  else
	    { 
	      sym = iter.CurrSym();
	      if (sym.isVariable()) 
		{
		  if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) 
		    return false;
		}
	      else
		{
		  if (sym.Func().arity())
		    {
		      if (!pushNonconstFunc(sym.Func(),
					    iter.currentPositionIsInSubstitution())) 
			return false;
		    }
		  else 
		    pushConst(sym.Func(),
			      iter.currentPositionIsInSubstitution()); 
		};
	      iter.SkipSym();
	    };  
	};
    }
  else 
    {      
      while (!iter.Empty())
	{
	  if (iter.CurrPosInBase() == redex)
	    {
	      if (!pushTermWithUnifier(replacingTerm)) return false;
	      iter.SkipTerm();
	    }
	  else
	    { 
	      sym = iter.CurrSym();
	      if (sym.isVariable()) 
		{
		  if (!pushVar(_unifierVarRenaming.Rename(sym.var()))) 
		    return false;
		}
	      else
		{
		  if (sym.Func().arity())
		    {
		      if (!pushNonconstFunc(sym.Func())) return false;
		    }
		  else 
		    pushConst(sym.Func()); 
		};
	      iter.SkipSym();
	    };  
	};
    };
  return closeNonpropLiteral();
}; 
bool 
TemporaryClauseCollector::pushLiteralWithReplacement(UnifTerm* lit,
						     UnifTerm* redex,
						     const Term& replacingTerm,
						     ulong index,
						     const LiteralProperties& prop)
{
  CALLM(_name,"pushLiteralWithReplacement(UnifTerm* lit,..)");
  ASSERT(!_debugTrash);
  Term header = lit->symbol().Func();
  ASSERT(header.arity());
  if (header.IsEquality()) header.MakeUnordered(); 
  if (!openNewNonpropLiteral(header,prop))
    return false;
  UnifTerm::Iter iter;
  iter.Reset(lit);
  Term sym;
  iter.SkipSym(); 
  if (_markSubstitutionPositions)
    {
      do
	{
	  if (iter.CurrVar()) 
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true))
		return false; 
	      iter.SkipSym(); 
	    }
	  else 
	    {
	      if (iter.CurrSubterm() == redex)
		{
		  if (!pushTermWithUnifier(index,replacingTerm,false)) return false;
		  iter.SkipTerm();
		}
	      else
		{
		  sym = iter.CurrSubterm()->symbol().Func();
		  if (sym.arity())
		    {
		      if (!pushNonconstFunc(sym,
					    iter.currentPositionIsInSubstitution())) 
			return false;
		    }
		  else 
		    pushConst(sym,
			      iter.currentPositionIsInSubstitution());
		  iter.SkipSym();
		}; 
	    };
	}
      while (!iter.Empty()); 
    }
  else 
    {
      do
	{
	  if (iter.CurrVar()) 
	    { 
	      if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar())))
		return false; 
	      iter.SkipSym(); 
	    }
	  else 
	    {
	      if (iter.CurrSubterm() == redex)
		{
		  if (!pushTermWithUnifier(index,replacingTerm)) return false;
		  iter.SkipTerm();
		}
	      else
		{
		  sym = iter.CurrSubterm()->symbol().Func();
		  if (sym.arity())
		    {
		      if (!pushNonconstFunc(sym)) return false;
		    }
		  else 
		    pushConst(sym);
		  iter.SkipSym();
		}; 
	    };
	}
      while (!iter.Empty()); 
    };
  return closeNonpropLiteral();
}; 
bool 
TemporaryClauseCollector::pushPseudoLiteralWithUnifier(ulong index,
						       const Term& header,
						       const Term& arg)
{
  CALLM(_name,"pushPseudoLiteralWithUnifier(..)");
  ASSERT(!_debugTrash);
  ASSERT(header.Positive());
  ASSERT(header.arity() == 1UL);
  LiteralProperties prop;
  if (!openNewNonpropLiteral(header,prop)) 
    return false;
  InstanceWithUnifier term;
  term.Reset(index,arg);
  UnifTerm::Symbol sym;
  do
    {
      sym = term.CurrSym();
      if (sym.isVariable()) 
	{
	  if (!pushVar(_unifierVarRenaming.Rename(sym.var()),false)) 
	    return false;
	}
      else
	{
	  if (sym.Func().arity())
	    {
	      if (!pushNonconstFunc(sym.Func(),false)) return false;
	    }
	  else 
	    {
	      pushConst(sym.Func(),false);
	    }; 
	};
      term.SkipSym();
    }
  while (!term.Empty());
  return closeNonpropLiteral();  
}; 
bool 
TemporaryClauseCollector::openNewNonpropLiteral(const Term& header,
						const LiteralProperties& prop)
{
  CALLM(_name,"openNewNonpropLiteral(..)");
  ASSERT(!_debugTrash);
  ASSERT(header.arity());
  (_totalSize += header.arity()) += 1;
  if (_totalSize > VampireKernelConst::MaxClauseSize)
    {
      reject(ClauseTooLarge);
      return false;
    };
  _collectingNonanswerLiteral = !header.isAnswerHeader();
  if (_collectingNonanswerLiteral)
    {
      (_nonanswerSize += header.arity()) += 1;
      ASSERT(_nonanswerSize <= _totalSize);
      if (_nonanswerSize > _sizeLimit)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  if (_newLiteral.isNull())
    {
      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	{
	  reject(None);
	  return false;
	};
    }
  _newLiteral->pushNonpropHeader(header,prop);
  _currentLiteral = _newLiteral;
  return true;
}; 
bool 
TemporaryClauseCollector::pushPropLiteral(const Term& header,
					  const LiteralProperties& prop)
{
  CALLM(_name,"pushPropLiteral(..)");    
  ASSERT(!_debugTrash);
  if (header == Term::builtInTrue() && _tautologyCheck)
    {
      reject(PropositionalTautology);
      return false;
    };
  if (header == Term::builtInFalse() && _literalFactoring)
    {
      return true;
    };    
  Term complementaryHeader = header;
  complementaryHeader.flipPolarity();
  for (const LitListElement* el = _newClause->literals()->first();
       el;
       el = el->next())
    {
      if (el->literal()->header() == header && _literalFactoring)
	{
	  el->literal()->properties().factoriseWith(prop,false);
	  return true;
	};
      if (el->literal()->header() == complementaryHeader && 
	  _tautologyCheck)
	{	    
	  reject(PropositionalTautology);
	  return false;
	};
    };
  ++_totalSize;
  if (_totalSize > VampireKernelConst::MaxClauseSize)
    {
      reject(ClauseTooLarge);
      return false;
    };
  _collectingNonanswerLiteral = !header.isAnswerHeader();
  if (_collectingNonanswerLiteral)
    {
      ++_nonanswerSize;
      ASSERT(_nonanswerSize <= _totalSize);
      if (_nonanswerSize > _sizeLimit)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  if (_newClause->literals()->length() >= _lengthLimit)
    {	
      reject(TooManyLiterals);
      return false;
    };
  if (_newLiteral.isNull())
    {
      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(_newLiteral))
	{
	  reject(None);
	  return false;
	};
    };
  _newLiteral->pushPropLit(header,prop);
  _currentLiteral = _newLiteral;
  _newClause->literals()->collect(_newLiteral);
  return true;
}; 
bool 
TemporaryClauseCollector::openNewLiteral(const Term& header,
					 const LiteralProperties& prop)
{
  CALLM(_name,"openNewLiteral(..)");
  if (header.arity())
    {
      _collectingNonpropLiteral = true;
      return openNewNonpropLiteral(header,prop);
    }
  else
    {
      _collectingNonpropLiteral = false;
      return pushPropLiteral(header,prop);
    };
}; 
bool TemporaryClauseCollector::collectLiteral(LitHandle& lit)
{
  CALLM(_name,"collectLiteral(LitHandle& lit)");
  ASSERT(!_debugTrash);
  Term header = lit->header();
  if (header.arity())
    {
      _totalSize += lit->size();
      if (_totalSize > VampireKernelConst::MaxClauseSize)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
      _collectingNonanswerLiteral = !header.isAnswerHeader();
      if (_collectingNonanswerLiteral)
	{
	  _nonanswerSize += lit->size();
	  ASSERT(_nonanswerSize <= _totalSize);
	  if (_nonanswerSize > _sizeLimit)
	    {
	      reject(ClauseTooLarge);
	      return false;
	    };
	};
      if (lit->isEquality())
	{	
	  ASSERT(!lit->isAnswerLiteral());
	  ASSERT(_collectingNonanswerLiteral);
	  if (lit->argsAreEqual())
	    {	
	      if (lit->isPositive() && _tautologyCheck)
		{
		  reject(EquationalTautology);
		  return false;
		}
	      else 
		if (_literalFactoring)
		  {
		    _totalSize -= lit->size();
		    _nonanswerSize -= lit->size();
		    return true;
		  }; 
	    };
	  for (LitListElement* el = _newClause->literals()->first();
	       el;
	       el = el->next())
	    {
	      if (el->literal()->isEquality())
		{
		  bool eqSymmetryUsed;
		  if (lit->equalEqLit(*(el->literal()),eqSymmetryUsed) &&
		      _literalFactoring)
		    {
		      el->literal()->properties().factoriseWith(lit->properties(),eqSymmetryUsed);
		      _totalSize -= lit->size();
		      _nonanswerSize -= lit->size();
		      return true;
		    };
		  if (lit->complementaryEqLit(*(el->literal())) &&
		      _tautologyCheck)
		    {
		      reject(PropositionalTautology);
		      return false;
		    };
		};
	    }; 
	  if (lit->size() > _literalSizeLimit)
	    {	      
	      reject(LiteralTooLarge);
	      return false;
	    };
	}
      else 
	{
	  for (LitListElement* el = _newClause->literals()->first();
	       el;
	       el = el->next())
	    {  
	      if (lit->equal(*(el->literal())) && _literalFactoring) 
		{
		  el->literal()->properties().factoriseWith(lit->properties(),false);
		  _totalSize -= lit->size();
		  if (_collectingNonanswerLiteral) 
		    {
		      ASSERT(!lit->isAnswerLiteral());
		      _nonanswerSize -= lit->size();
		    };
		  return true;
		}; 
	      if (lit->isComplementary(*(el->literal())) && _tautologyCheck)
		{
		  reject(PropositionalTautology);
		  return false;
		};
	    }; 
	  ASSERT(lit->size() <= VampireKernelConst::MaxTermSize);
	  if (_collectingNonanswerLiteral)
	    {
	      ASSERT(!lit->isAnswerLiteral());	    
	      if (lit->size() > _literalSizeLimit)
		{	      
		  reject(LiteralTooLarge);
		  return false;
		};
	    };
	};
      if (_newClause->literals()->length() >= _lengthLimit)
	{
	  reject(TooManyLiterals);
	  return false;
	};
      _newClause->literals()->collect(lit);
      return true;
    }
  else 
    {
      if (header == Term::builtInTrue() && _tautologyCheck)
	{
	  reject(PropositionalTautology);
	  return false;
	};
      if (header == Term::builtInFalse() && _literalFactoring)
	{
	  return true;
	};    
      Term complementaryHeader = header;
      complementaryHeader.flipPolarity();
      for (const LitListElement* el = _newClause->literals()->first();
	   el;
	   el = el->next())
	{
	  if (el->literal()->header() == header &&
	      _literalFactoring) 
	    {
	      el->literal()->properties().factoriseWith(lit->properties(),false);
	      return true;
	    };
	  if (el->literal()->header() == complementaryHeader &&
	      _tautologyCheck)
	    {	    
	      reject(PropositionalTautology);
	      return false;
	    };
	};
      ++_totalSize;
      if (_totalSize > VampireKernelConst::MaxClauseSize)
	{
	  reject(ClauseTooLarge);
	  return false;
	};
      _collectingNonanswerLiteral = !header.isAnswerHeader();
      if (_collectingNonanswerLiteral)
	{
	  ++_nonanswerSize;
	  ASSERT(_nonanswerSize <= _totalSize);
	  if (_nonanswerSize > _sizeLimit)
	    {
	      reject(ClauseTooLarge);
	      return false;
	    };
	};
      if (_newClause->literals()->length() >= _lengthLimit)
	{	
	  reject(TooManyLiterals);
	  return false;
	};
      _newClause->literals()->collect(lit);
      return true;
    }; 
}; 
bool TemporaryClauseCollector::collectLiteralList(BasicTempLitList& lst)
{
  CALLM(_name,"collectLiteralList(BasicTempLitList& lst)");  
  ASSERT(!_debugTrash);
  LitHandle lit;
  while (!lst.isEmpty())
    {
      lst.popTo(lit);
      if (!collectLiteral(lit))
	return false;
    };
  return true;
}; 
bool TemporaryClauseCollector::closeNonpropLiteral()
{
  CALLM(_name,"closeNonpropLiteral()");    
  ASSERT(!_debugTrash);
  ASSERT(!_newLiteral.isNull());
  _newLiteral->endOfLiteral();
  ASSERT(_newLiteral->arity());
  if (_newLiteral->isEquality())
    {	
      ASSERT(!_newLiteral->isAnswerLiteral());
      ASSERT(_collectingNonanswerLiteral);
      if (_newLiteral->argsAreEqual())
	{	
	  if (_newLiteral->isPositive() && _tautologyCheck)
	    {
	      reject(EquationalTautology);
	      return false;
	    }
	  else 
	    if (_literalFactoring)
	      {
		_totalSize -= _newLiteral->size();
		_nonanswerSize -= _newLiteral->size();
		return true;
	      }; 
	};
      for (LitListElement* el = _newClause->literals()->first();
	   el;
	   el = el->next())
	{	
	  if (el->literal()->isEquality())
	    {
	      bool eqSymmetryUsed;
	      if (_newLiteral->equalEqLit(*(el->literal()),eqSymmetryUsed) &&
		  _literalFactoring)
		{
		  el->literal()->properties().factoriseWith(_newLiteral->properties(),eqSymmetryUsed);
		  _totalSize -= _newLiteral->size();
		  _nonanswerSize -= _newLiteral->size();
		  return true;
		};
	      if (_newLiteral->complementaryEqLit(*(el->literal())) &&
		  _tautologyCheck)
		{
		  reject(PropositionalTautology);
		  return false;
		};
	    };
	}; 
      if (_newLiteral->size() > _literalSizeLimit)
	{	      
	  reject(LiteralTooLarge);
	  return false;
	};
      ASSERT(_nonanswerSize <= _sizeLimit);
    }
  else 
    {
      for (LitListElement* el = _newClause->literals()->first();
	   el;
	   el = el->next())
	{  
	  if (_newLiteral->equal(*(el->literal()))) 
	    {
	      _totalSize -= _newLiteral->size();
	      if (_collectingNonanswerLiteral) 
		{
		  ASSERT(!_newLiteral->isAnswerLiteral());
		  _nonanswerSize -= _newLiteral->size();
		};
	      return true;
	    }; 
	  if (_newLiteral->isComplementary(*(el->literal())) &&
	      _tautologyCheck)
	    {
	      reject(PropositionalTautology);
	      return false;
	    };
	}; 
      ASSERT(_newLiteral->size() <= VampireKernelConst::MaxTermSize);
      if (_collectingNonanswerLiteral)
	{
	  ASSERT(!_newLiteral->isAnswerLiteral());	    
	  if (_newLiteral->size() > _literalSizeLimit)
	    {	      
	      reject(LiteralTooLarge);
	      return false;
	    };
	  ASSERT(_nonanswerSize <= _sizeLimit);
	};
    };
  if (_newClause->literals()->length() >= _lengthLimit)
    {
      reject(TooManyLiterals);
      return false;
    };
  _newClause->literals()->collect(_newLiteral);
  return true;
}; 
bool TemporaryClauseCollector::closeLiteral()
{
  CALLM(_name,"closeLiteral()");
  _currentLiteral.clear();
  if (_collectingNonpropLiteral)
    return closeNonpropLiteral();
  return true;
};
bool TemporaryClauseCollector::pushNonconstFunc(const Term& f)
{
  CALLM(_name,"pushNonconstFunc(const Term& f)");
  ASSERT(!_debugTrash);
  _totalSize += f.arity();    
  if (_totalSize > VampireKernelConst::MaxClauseSize)  
    {
      reject(ClauseTooLarge);
      return false;
    };
  if (_collectingNonanswerLiteral)
    {
      _nonanswerSize += f.arity();
      if (_nonanswerSize > _sizeLimit) 
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  if (!_newLiteral->pushNonconstFun(f))
    {
      reject(LiteralTooDeep);
      return false;
    };
  if (_collectingNonanswerLiteral)
    {
      if (_newLiteral->size() > _literalSizeLimit)
	{
	  reject(LiteralTooLarge);
	  return false;
	}; 
    }
  else
    {
      if (_newLiteral->size() > VampireKernelConst::MaxTermSize)
	{
	  reject(LiteralTooLarge);
	  return false;
	}; 
    };
  return true;
}; 
bool TemporaryClauseCollector::pushNonconstFunc(const Term& f,bool substPos)
{
  CALLM(_name,"pushNonconstFunc(const Term& f,bool substPos)");
  ASSERT(!_debugTrash);
  _totalSize += f.arity();    
  if (_totalSize > VampireKernelConst::MaxClauseSize)  
    {
      reject(ClauseTooLarge);
      return false;
    };
  if (_collectingNonanswerLiteral)
    {
      _nonanswerSize += f.arity();
      if (_nonanswerSize > _sizeLimit) 
	{
	  reject(ClauseTooLarge);
	  return false;
	};
    };
  if (!_newLiteral->pushNonconstFun(f,substPos))
    {
      reject(LiteralTooDeep);
      return false;
    };
  if (_collectingNonanswerLiteral)
    {
      if (_newLiteral->size() > _literalSizeLimit)
	{
	  reject(LiteralTooLarge);
	  return false;
	}; 
    }
  else
    {
      if (_newLiteral->size() > VampireKernelConst::MaxTermSize)
	{
	  reject(LiteralTooLarge);
	  return false;
	}; 
    };
  return true;
}; 
bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const Term& t)
{
  CALLM(_name,"pushTermWithUnifier(ulong index,const Term& t)");
  ASSERT(!_debugTrash);
  if (t.isVariable())
    {
      bool freeVar;
      void* bind = Unifier::current()->variable(t.var(),index)->Unref(freeVar);
      if (freeVar) 
	{
	  return 
	    pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)));
	}
      else
	return pushComplexTermWithUnifier(static_cast<UnifTerm*>(bind));
    }
  else
    {
      ASSERT(t.IsReference());
      return pushComplexTermWithUnifier(index,t.deref());
    };
}; 
bool TemporaryClauseCollector::pushTermWithUnifier(ulong index,const Term& t,bool inSubst)
{
  CALLM(_name,"pushTermWithUnifier(ulong index,const Term& t,bool inSubst)");
  ASSERT(!_debugTrash);
  if (t.isVariable())
    {
      bool freeVar;
      void* bind = Unifier::current()->variable(t.var(),index)->Unref(freeVar);
      if (freeVar) 
	{
	  return 
	    pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)),true);
	}
      else
	return pushComplexTermWithUnifier(static_cast<UnifTerm*>(bind),true);
    }
  else
    {
      ASSERT(t.IsReference());
      return pushComplexTermWithUnifier(index,t.deref(),inSubst);
    };
}; 
bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const Term* t)  
{
  CALLM(_name,"pushComplexTermWithUnifier(ulong index,const Term* t)");
  ASSERT(!_debugTrash);
  InstanceWithUnifier term;
  term.Reset(index,t);
  UnifTerm::Symbol sym;
  do
    {
      sym = term.CurrSym();
      if (sym.isVariable()) 
	{
	  if (!(pushVar(_unifierVarRenaming.Rename(sym.var())))) return false; 
	}
      else
	{
	  if (sym.Func().arity())
	    {
	      if (!(pushNonconstFunc(sym.Func()))) return false;
	    }
	  else pushConst(sym.Func()); 
	};
      term.SkipSym();
    }
  while (!(term.Empty()));
  return true; 
}; 
bool TemporaryClauseCollector::pushComplexTermWithUnifier(ulong index,const Term* t,bool inSubst)  
{
  CALLM(_name,"pushComplexTermWithUnifier(ulong index,const Term* t,bool inSubst)");
  ASSERT(!_debugTrash);
  InstanceWithUnifier term;
  term.Reset(index,t);
  UnifTerm::Symbol sym;
  do
    {
      sym = term.CurrSym();
      if (sym.isVariable()) 
	{
	  if (!pushVar(_unifierVarRenaming.Rename(sym.var()),true)) return false; 
	}
      else
	{
	  if (sym.Func().arity())
	    {
	      if (!(pushNonconstFunc(sym.Func(),
				     inSubst || term.currentPositionIsInSubstitution()))) 
		return false;
	    }
	  else 
	    pushConst(sym.Func(),
		      inSubst || term.currentPositionIsInSubstitution()); 
	};
      term.SkipSym();
    }
  while (!(term.Empty()));
  return true; 
}; 
bool TemporaryClauseCollector::pushTermWithUnifier(UnifTerm* t)
{
  CALLM(_name,"pushTermWithUnifier(UnifTerm* t)");
  ASSERT(!_debugTrash);
  if (t->symbol().isVariable())
    {
      bool free_var;
      void* bind = t->symbol().var()->Unref(free_var);
      if (free_var) 
	return pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)));
      return pushComplexTermWithUnifier(static_cast<UnifTerm*>(bind)); 
    };
  return pushComplexTermWithUnifier(t);
}; 
bool TemporaryClauseCollector::pushTermWithUnifier(UnifTerm* t,bool inSubst)
{
  CALLM(_name,"pushTermWithUnifier(UnifTerm* t,bool inSubst)");
  ASSERT(!_debugTrash);
  if (t->symbol().isVariable())
    {
      bool free_var;
      void* bind = t->symbol().var()->Unref(free_var);
      if (free_var) 
	return pushVar(_unifierVarRenaming.Rename(static_cast<Unifier::Variable*>(bind)),true);
      return pushComplexTermWithUnifier(static_cast<UnifTerm*>(bind),true); 
    };
  return pushComplexTermWithUnifier(t,inSubst);
}; 
bool TemporaryClauseCollector::pushComplexTermWithUnifier(UnifTerm* t)
{
  CALLM(_name,"pushComplexTermWithUnifier(UnifTerm* t)");
  ASSERT(!_debugTrash);
  UnifTerm::Iter iter;
  iter.Reset(t);
  Term sym;
  do
    {
      if (iter.CurrVar()) 
	{ 
	  if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()))) return false; 
	}
      else 
	{
	  sym = iter.CurrSubterm()->symbol().Func();
	  if (sym.arity())
	    {
	      if (!(pushNonconstFunc(sym))) return false;
	    }
	  else 
	    pushConst(sym);
	};
      iter.SkipSym();
    }
  while (!(iter.Empty()));
  return true;
}; 
bool TemporaryClauseCollector::pushComplexTermWithUnifier(UnifTerm* t,bool inSubst)
{
  CALLM(_name,"pushComplexTermWithUnifier(UnifTerm* t,bool inSubst)");
  ASSERT(!_debugTrash);
  UnifTerm::Iter iter;
  iter.Reset(t);
  Term sym;
  do
    {
      if (iter.CurrVar()) 
	{ 
	  if (!pushVar(_unifierVarRenaming.Rename(iter.CurrVar()),true)) return false; 
	}
      else 
	{
	  sym = iter.CurrSubterm()->symbol().Func();
	  if (sym.arity())
	    {
	      if (!(pushNonconstFunc(sym,
				     inSubst || iter.currentPositionIsInSubstitution()))) 
		return false;
	    }
	  else 
	    pushConst(sym,
		      inSubst || iter.currentPositionIsInSubstitution());
	};
      iter.SkipSym();
    }
  while (!(iter.Empty()));
  return true;
}; 
void TemporaryClauseCollector::reject(const RejectionReason& reason)
{
  CALLM(_name,"reject(const RejectionReason& reason)");
  ASSERT(!_debugTrash);
  DOP(_debugTrash = true);
  _rejectionReason = reason;
  switch (reason)
    {	
    case None:
      ++(_statistics->clauses.generated.discardedForOtherReasons);
      break;
    case PropositionalTautology:
      ++(_statistics->clauses.generated.discardedPropositionalTautologies);
      ++(_statistics->inferences.simplifying.propositionalTautologyDeletion);
      break;
    case EquationalTautology:
      ++(_statistics->clauses.generated.discardedEquationalTautologies);
      ++(_statistics->inferences.simplifying.equationalTautologyDeletion);
      break;
    case ClauseTooLarge:
      ++(_statistics->clauses.generated.discardedTooLarge);
      break;
    case LiteralTooLarge:
      ++(_statistics->clauses.generated.discardedWithTooLargeLiterals);
      break;
    case TooManyLiterals:
      ++(_statistics->clauses.generated.discardedWithTooManyLiterals);
      break;
    case TooManyVariables: 
      ++(_statistics->clauses.generated.discardedWithTooManyVariables);
      break;
    case LiteralTooDeep:
      ++(_statistics->clauses.generated.discardedTooDeep);
      break;
    };
}; 
void 
TemporaryClauseCollector::registerGenerated(ClauseBackground::Rule rule)
{
  CALLM(_name,"registerGenerated(ClauseBackground::Rule rule)");
  ASSERT(_statistics);
  switch (rule)
    {
    case ClauseBackground::Input:
      ++(_statistics->clauses.generated.byOther);
      break;
    case ClauseBackground::Prepro:
      ++(_statistics->clauses.generated.byOther);
      break;
    case ClauseBackground::BinRes:
      ++(_statistics->clauses.generated.byResolution);
      ++(_statistics->inferences.deduction.resolution.accomplished);
      break;
    case ClauseBackground::ForwTransChain:
      ++(_statistics->clauses.generated.byForwardChaining);
      ++(_statistics->inferences.deduction.forwardChaining.accomplished);
      break;
    case ClauseBackground::BackTransChain:
      ++(_statistics->clauses.generated.byBackwardChaining);
      ++(_statistics->inferences.deduction.backwardChaining.accomplished);
      break;
    case ClauseBackground::ForwSup:
      ++(_statistics->clauses.generated.byForwardSuperposition);
      ++(_statistics->inferences.deduction.forwardSuperposition.accomplished);
      break;
    case ClauseBackground::BackSup:
      ++(_statistics->clauses.generated.byBackwardSuperposition);
      ++(_statistics->inferences.deduction.backwardSuperposition.accomplished);
      break;
    case ClauseBackground::EqRes:
      ++(_statistics->clauses.generated.byEqualityResolution);
      ++(_statistics->inferences.deduction.equalityResolution.accomplished);
      break;
    case ClauseBackground::EqFact:
      ++(_statistics->clauses.generated.byEqualityFactoring);
      ++(_statistics->inferences.deduction.equalityFactoring.accomplished);
      break;
    case ClauseBackground::BackDemod:
      ++(_statistics->clauses.generated.byBackwardDemodulation);
      break;
    case ClauseBackground::Reanimation:
      ++(_statistics->clauses.generated.byOther);
      break;
    case ClauseBackground::Folding:
      ++(_statistics->clauses.generated.byBackwardFolding);
      break;
    case ClauseBackground::Clausification:
      ++(_statistics->clauses.generated.byOther);
      break;
    case ClauseBackground::Unspecified:
      ++(_statistics->clauses.generated.byOther);
      break;
#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
#else
#  ifdef _SUPPRESS_WARNINGS_
    default:
      BK::RuntimeError::report("Wrong argument value in TemporaryClauseCollector::registerGenerated(ClauseBackground::Rule rule)");
#  endif
#endif
    };
}; 
