/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "UniformTermWeighting.hpp"
#include "UnifTerm.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "GlobAlloc.hpp"
#include "LocalUnification.hpp"
#include "InstanceWithUnifier.hpp"
using namespace BK;
using namespace VK;
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIFORM_TERM_WEIGHTING
#define DEBUG_NAMESPACE "UniformTermWeighting"
#endif
#include "debugMacros.hpp"
UniformTermWeighting uniformTermWeighting;
UniformTermWeighting::~UniformTermWeighting() {};
void* UniformTermWeighting::operator new(size_t)
{
  return GlobAlloc::allocate(sizeof(UniformTermWeighting));
}; 
void UniformTermWeighting::operator delete(void* obj)
{
  GlobAlloc::deallocate(obj,sizeof(UniformTermWeighting));
};
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,const Flatterm* term)
{
  CALL("collectWeight(WeightPolynomial& weight,const Flatterm* term)");
  Flatterm::ConstIterator iter(term);
  do
    {
      if (iter.CurrPos()->isVariable())
	{
	  weight.add1(iter.CurrPos()->Var());
	}
      else
	weight.add1();
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,long coefficient,const Flatterm* term)
{
  CALL("collectWeight(WeightPolynomial& weight,long coefficient,const Flatterm* term)");
  Flatterm::ConstIterator iter(term);
  do
    {
      if (iter.CurrPos()->isVariable())
	{
	  weight.add(coefficient,iter.CurrPos()->Var());
	}
      else
	weight.add(coefficient);
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
}; 
long UniformTermWeighting::computeMinimalInstanceWeight(const Flatterm* term,
								  bool& containsVariables)
{
  CALL("computeMinimalInstanceWeight(const Flatterm* term,bool& containsVariables)");
  long result = 0;
  containsVariables = false;
  Flatterm::ConstIterator iter(term);
  do
    {
      if (iter.CurrPos()->isVariable()) containsVariables = true;
      ++result;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());  
  return result;
}; 
long UniformTermWeighting::computeWeightIfGroundAndLessOrEqual(long coefficient,
									 const Flatterm* term,
									 long weightLimit)
{
  CALL("computeWeightIfGroundAndLessOrEqual(long coefficient,const Flatterm* term,long weightLimit)");
  long result = 0;
  Flatterm::ConstIterator iter(term);
  do
    {
      if (iter.CurrPos()->isVariable()) return 0L; 
      result += coefficient;
      if (result > weightLimit) return -1; 
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
  ASSERT(result > 0L);
  return result;
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,const Term& term)
{
  CALL("collectWeight(WeightPolynomial& weight,const Term& term");
  if (term.isVariable())
    {
      weight.add1(term.var());
    }
  else
    {
      ASSERT(term.IsReference());
      Term::Iterator iter(term.deref());
      do
	{
	  if (iter.CurrentSymIsVar())
	    {
	      weight.add1(iter.symbolRef().var());
	    }
	  else
	    weight.add1();
	}
      while (iter.Next());
    };
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,long coefficient,const Term& term)
{
  CALL("collectWeight(WeightPolynomial& weight,long coefficient,const Term& term");
  if (term.isVariable())
    {
      weight.add(coefficient,term.var());
    }
  else
    {
      ASSERT(term.IsReference());
      Term::Iterator iter(term.deref());
      do
	{
	  if (iter.CurrentSymIsVar())
	    {
	      weight.add(coefficient,iter.symbolRef().var());
	    }
	  else
	    weight.add(coefficient);
	}
      while (iter.Next());
    };
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,long coefficient,const UnifTerm* term)
{
  CALL("collectWeight(WeightPolynomial& weight,long coefficient,const UnifTerm* term)");
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  weight.add(coefficient,(static_cast<Unifier::Variable*>(instance))->absoluteVarNum());
	  return;
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar())
	{
	  weight.add(coefficient,iter.CurrVar()->absoluteVarNum());
	}
      else
	{
	  weight.add(coefficient);
	};
      iter.SkipSym();
    }
  while (iter.NotEmpty());
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,const UnifTerm* term)
{
  CALL("collectWeight(WeightPolynomial& weight,const UnifTerm* term)");
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  weight.add1((static_cast<Unifier::Variable*>(instance))->absoluteVarNum());
	  return;
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar())
	{
	  weight.add1(iter.CurrVar()->absoluteVarNum());
	}
      else
	{
	  weight.add1();
	};
      iter.SkipSym();
    }
  while (iter.NotEmpty());
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,const Term& term,ulong variableIndex)
{
  CALL("collectWeight(WeightPolynomial& weight,const Term& term,ulong variableIndex)");
  static InstanceWithUnifier iter;
  iter.Reset(variableIndex,(Term&)term);
  do
    {
      if (iter.CurrSym().isVariable())
	{
	  weight.add1(iter.CurrSym().var()->absoluteVarNum());
	}
      else
	{
	  weight.add1();
	};
      iter.SkipSym();
    }
  while (!iter.Empty());
}; 
void 
UniformTermWeighting::collectWeight(WeightPolynomial& weight,
				    long coefficient,
				    const Term* complexTerm,
				    const InstRet::Substitution* subst)
{
  CALL("collectWeight(WeightPolynomial& weight,long coefficient,const Term* complexTerm,const InstRet::Substitution* subst)");
  static InstRet::Substitution::Instance inst;
  inst.reset(complexTerm,subst);
  do
    {
      if (inst.symbol().isVariable())
	{
	  weight.add(coefficient,inst.symbol().var());
	}
      else
	weight.add(coefficient);
    }
  while (inst.next()); 
}; 
void 
UniformTermWeighting::collectWeight(WeightPolynomial& weight,
				    const Term* complexTerm,
				    const InstRet::Substitution* subst)
{
  CALL("collectWeight(WeightPolynomial& weight,const Term* complexTerm,const InstRet::Substitution* subst)");
  static InstRet::Substitution::Instance inst;
  inst.reset(complexTerm,subst);
  do
    {
      if (inst.symbol().isVariable())
	{
	  weight.add1(inst.symbol().var());
	}
      else
	weight.add1();
    }
  while (inst.next()); 
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,
					 const Term& term,
					 const ForwardMatchingSubstitutionCache* subst)
{
  CALL("collectWeight(WeightPolynomial& weight,const Term& term,const ForwardMatchingSubstitutionCache* subst)");
  static ForwardMatchingSubstitutionCache::Instance inst;
  inst.reset(term,subst);
  do
    {
      if (inst.symbol().isVariable())
	{
	  weight.add1(inst.symbol().var());
	}
      else
	weight.add1();
    }
  while (inst.next());
}; 
void UniformTermWeighting::collectWeight(WeightPolynomial& weight,
					 const Term& term,
					 const LocalUnification* unifier)
{
  CALL("collectWeight(WeightPolynomial& weight,const Term& term,const LocalUnification* unifier)");
  LocalUnification::TermTraversal inst(term,unifier);
  do
    {
      if (inst.term().isVariable())
	{
	  weight.add1(inst.term().var());
	}
      else
	weight.add1();
      inst.next();
    }
  while (inst);
}; 
long UniformTermWeighting::computeMinimalInstanceWeight(const UnifTerm* term,bool& containsVariables)
{
  CALL("computeMinimalInstanceWeight(const UnifTerm* term,bool& containsVariables)");
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  containsVariables = true;
	  return 1L; 
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());  
  long result = 0;
  containsVariables = false;
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar()) containsVariables = true;
      ++result;
      iter.SkipSym();
    }
  while (iter.NotEmpty());
  return result;
}; 
long UniformTermWeighting::computeWeightIfGroundAndLessOrEqual(long coefficient,const UnifTerm* term,long weightLimit)
{
  CALL("computeWeightIfGroundAndLessOrEqual(long coefficient,const UnifTerm* term,long weightLimit)");  
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  return 0L;  
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());  
  long result = 0;
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar()) return 0L; 
      result += coefficient;
      if (result > weightLimit) return -1; 
      iter.SkipSym();
    }
  while (iter.NotEmpty());
  ASSERT(result > 0L);
  return result;
}; 
long UniformTermWeighting::computeWeightIfGroundAndLessOrEqual(const UnifTerm* term,long weightLimit)
{
  CALL("computeWeightIfGroundAndLessOrEqual(const UnifTerm* term,long weightLimit)");  
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  return 0L;  
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());  
  long result = 0;
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar()) return 0L; 
      ++result;
      if (result > weightLimit) return -1; 
      iter.SkipSym();
    }
  while (iter.NotEmpty());
  ASSERT(result > 0L);
  return result;
}; 
long 
UniformTermWeighting::computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
								const UnifTerm* term,
								long weightLimit,
								bool& ground)
{
  CALL("computeMinimalInstanceWeightIfLessOrEqual(long coefficient,const UnifTerm* term,..)");
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  ground = false;
	  return (coefficient > weightLimit) ? -1 : coefficient;
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());  
  long result = 0;
  ground = true;
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar()) ground = false;
      result += coefficient;
      if (result > weightLimit) return -1; 
      iter.SkipSym();
    }
  while (iter.NotEmpty());
  ASSERT(result > 0L);
  return result;
}; 
long 
UniformTermWeighting::computeMinimalInstanceWeightIfLessOrEqual(const UnifTerm* term,
								long weightLimit,
								bool& ground)
{
  CALL("computeMinimalInstanceWeightIfLessOrEqual(const UnifTerm* term,..)");
  if (weightLimit <= 0L) return -1;
  if (term->symbol().isVariable())
    {
      Unifier::Variable* var = term->symbol().var();
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  ground = false;
	  return 1L; 
	}
      else
	{
	  term = static_cast<UnifTerm*>(instance);
	};
    };
  ASSERT(term->symbol().IsFunctor());  
  long result = 0;
  ground = true;
  UnifTerm::ConstIter iter(term);
  do
    {
      if (iter.CurrVar()) ground = false;
      ++result;
      if (result > weightLimit) return -1; 
      iter.SkipSym();
    }
  while (iter.NotEmpty());
  ASSERT(result > 0L);
  return result;
}; 
long 
UniformTermWeighting::computeMinimalInstanceWeight(const Term* term,bool& containsVariables)
{
  CALL("computeMinimalInstanceWeight(const Term* term,bool& containsVariables)");
  ASSERT(term->isComplex());
  _iter.Reset(term);
  long result = 0L;
  do
    {
      if (_iter.symbolRef().isVariable()) containsVariables = true;
      ++result;
    }
  while (_iter.Next());
  return result;
}; 
long 
UniformTermWeighting::computeWeightIfGroundAndLessOrEqual(long coefficient,
							  const Term* term,
							  long weightLimit)
{
  CALL("computeWeightIfGroundAndLessOrEqual(long coefficient,const Term* term,long weightLimit)");  
  ASSERT(term->isComplex());
  _iter.Reset(term);
  long result = 0L;
  do
    {
      if (_iter.symbolRef().isVariable()) return 0L; 
      result += coefficient;
      if (result > weightLimit) return -1; 
    }
  while (_iter.Next());
  return result;
}; 
long 
UniformTermWeighting::computeWeightIfGroundAndLessOrEqual(const Term* term,long weightLimit)
{
  CALL("computeWeightIfGroundAndLessOrEqual(const Term* term,long weightLimit)");  
  ASSERT(term->isComplex());
  _iter.Reset(term);
  long result = 0L;
  do
    {
      if (_iter.symbolRef().isVariable()) return 0L; 
      ++result;
      if (result > weightLimit) return -1; 
    }
  while (_iter.Next());
  return result;
}; 
long 
UniformTermWeighting::computeMinimalInstanceWeightIfLessOrEqual(long coefficient,
								const Term* term,
								long weightLimit,
								bool& ground)
{
  CALL("computeMinimalInstanceWeightIfLessOrEqual(long coefficient,const Term* term,..)");
  ASSERT(term->isComplex());
  _iter.Reset(term);
  long result = 0L;
  ground = true;
  do
    {
      if (_iter.symbolRef().isVariable()) ground = false; 
      result += coefficient;
      if (result > weightLimit) return -1; 
    }
  while (_iter.Next());
  return result;
}; 
long 
UniformTermWeighting::computeMinimalInstanceWeightIfLessOrEqual(const Term* term,
								long weightLimit,
								bool& ground)
{
  CALL("computeMinimalInstanceWeightIfLessOrEqual(const Term* term,..)");
  ASSERT(term->isComplex());
  _iter.Reset(term);
  long result = 0L;
  ground = true;
  do
    {
      if (_iter.symbolRef().isVariable()) ground = false; 
      ++result;
      if (result > weightLimit) return -1; 
    }
  while (_iter.Next());
  return result;
}; 
