/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#ifndef __XML__
#define __XML__
#include <string>
#include <iostream>
#include "Allocator.hpp"
using namespace std;
namespace Shell {
class XML
{
public:
  class Attribute
  {
  protected:
    enum Type {
      TEXT,
      LONG,
      INTEGER,
      DOUBLE
    };
  public:
    Attribute(Type type, const string& name);
    virtual ~Attribute() {}
    void writeList(ostream&) const;
    void* operator new(size_t);
    void operator delete(void* obj);
    void destroy();
    inline void setNext(Attribute* next) { _next = next; }
    inline void setPrevious(Attribute* previous) { _previous = previous; }
    inline Attribute* next() const { return _next; }
    inline string name() const { return _name; }
    const Attribute* find (const string& name) const;
  protected:
    Type _type;
    string _name;
    Attribute* _next;
    Attribute* _previous;
    virtual void writeValue(ostream&) const = 0;
  }; 
  class IntegerAttribute :
    public Attribute 
  {
  public:
    IntegerAttribute (const string& name, int value);
    ~IntegerAttribute() {}
    void writeValue (ostream&) const;
    CLASS_NAME("XML::IntegerAttribute");
    USE_ALLOCATOR(IntegerAttribute);
  private:
    int _value;
  }; 
  class LongAttribute :
    public Attribute 
  {
  public:
    LongAttribute (const string& name, long value);
    ~LongAttribute() {}
    void writeValue (ostream&) const;
    CLASS_NAME("XML::LongAttribute");
    USE_ALLOCATOR(LongAttribute);
  private:
    long _value;
  }; 
  class DoubleAttribute :
    public Attribute 
  {
  public:
    DoubleAttribute (const string& name, double value);
    ~DoubleAttribute() {}
    void writeValue (ostream&) const;
    CLASS_NAME("XML::DoubleAttribute");
    USE_ALLOCATOR(DoubleAttribute);
  private:
    double _value;
  }; 
  class TextAttribute :
    public Attribute 
  {
  public:
    TextAttribute(const string& name, const string& value);
    ~TextAttribute() {}
    void writeValue (ostream&) const;
    const string& value() const { return _value; }
    CLASS_NAME("XML::TextAttribute");
    USE_ALLOCATOR(TextAttribute);
  private:
    string _value;
  }; 
  class Content {
  public:
    enum Type {
      PCDATA,
      EMPTY,
      TEXT,
      FLOAT,
      LONG,
      INTEGER,
      DEEP
    };
    virtual void write (ostream&) const = 0;
    virtual void write (ostream&, int indent) const = 0;
    explicit Content (Type ct);
    virtual ~Content () {}
    void* operator new (size_t);
    void operator delete (void* obj);
    void destroy ();
    Type kind () const { return _type; }
    inline Content* next() const { return _next; }
  public:
    Type _type;
    Content* _next;
    Content* _previous;
  protected:
    virtual void destroyAttributes () {}
    virtual void destroyContent () {}
  }; 
  class PCData
    : public Content
  {
  public: 
    void write(ostream&) const;
    void write(ostream&,int indent) const;
    explicit PCData(const string& text);
    ~PCData() {}
    string text() const { return _text; }
    CLASS_NAME("XML::PCData");
    USE_ALLOCATOR(PCData);
  private:
    string _text;
  }; 
  class Element
    : public Content
  {
  public: 
    void append(const string& file); 
    virtual void write(ostream&) const; 
    virtual void write(ostream&, int indent) const; 
    Element(Type type, const string& name);
    virtual ~Element() {}
    void addAttribute(Attribute*);
    void* operator new(size_t);
    void operator delete(void* obj);
    static void writeString(ostream& str,const string& s);
    void setAttributes(Attribute*);
    string name () const { return _name; }
    Attribute* firstAttribute() const { return _firstAttribute; }
  protected:
    string _name;
    Attribute* _firstAttribute;
    Attribute* _lastAttribute;
    void addAfter(Element* next);
    static void writeSpaces(ostream& str,int n);
    void destroyAttributes();
  private:
    virtual void writeContent (ostream&,int indent) const = 0;
    virtual void writeFooter(ostream& str) const;
    virtual void writeEndOfHeader(ostream& str) const {}
  }; 
  class EmptyElement
    : public Element
  {
  public: 
    explicit EmptyElement(const string& name);
    ~EmptyElement() {}
    CLASS_NAME("XML::EmptyElement");
    USE_ALLOCATOR(EmptyElement);
  private:
    void writeContent (ostream&,int indent) const;
    void writeFooter(ostream& str) const {}
    void writeEndOfHeader(ostream& str) const;
  }; 
  class TextElement
    : public Element
  {
  public: 
    TextElement(const string& name, const string& text);
    ~TextElement() {}
    string text() const { return _text; }
    CLASS_NAME("XML::TextElement");
    USE_ALLOCATOR(TextElement);
  private:
    string _text;
    void writeContent (ostream&,int indent) const;
  }; 
  class FloatElement
    : public Element
  {
  public: 
    FloatElement(const string& name, float value);
    ~FloatElement() {}
    CLASS_NAME("XML::FloatElement");
    USE_ALLOCATOR(FloatElement);
  private:
    float _value;
    void writeContent (ostream&,int indent) const;
  }; 
  class LongElement
    : public Element
  {
  public: 
    LongElement(const string& name, long value);
    ~LongElement() {}
    CLASS_NAME("XML::LongElement");
    USE_ALLOCATOR(LongElement);
  private:
    long _value;
    void writeContent (ostream&,int indent) const;
  }; 
  class IntegerElement
    : public Element
  {
  public: 
    IntegerElement(const string& name, int value);
    ~IntegerElement() {}
    CLASS_NAME("XML::IntegerElement");
    USE_ALLOCATOR(IntegerElement);
  private:
    int _value;
    void writeContent (ostream&,int indent) const;
  }; 
  class DeepElement
    : public Element
  {
  public: 
    explicit DeepElement (const string& name);
    void addChild (Content* subelement);
    ~DeepElement () {}
    void destroyContent ();
    inline Content* firstChild() const { return _firstChild; }
    CLASS_NAME("XML::DeepElement");
    USE_ALLOCATOR(DeepElement);
  private:
    Content* _firstChild;
    Content* _lastChild;
    void writeContent (ostream&,int indent) const;
  }; 
}; 
}
#endif
