open Lib
open Printf

let ()= out_str "\n---------------- iProver v0.2 ---------------\n"

type options = 
    {mutable problem_path          : string; 
     mutable include_path          : string; 
     mutable problem_files         : string list;
     mutable inst_lit_sel          : string;  
     mutable solver_per_active     : int;
     mutable solver_per_clauses    : int;
     mutable inst_pass_queue1      : string; 
     mutable inst_pass_queue2      : string;
     mutable inst_pass_queue1_mult : int;
     mutable inst_pass_queue2_mult : int;
     mutable dismatching           : bool;
     mutable eager_unprocessed_to_passive : bool;
     mutable ground_splitting_input       : bool;
     mutable ground_splitting_full        : bool;
     mutable learning_loop_flag    : bool;
     mutable learning_start        : int;
     mutable learning_factor       : int;
     mutable fof                   : bool;
     mutable eprover_path          : string
}

(* --problem_path /home/TPTP-v3.0.1/Problems/ --input_file  Problems/ANA/ANA002-1.p*)      
(* --include_path /home/TPTP-v3.0.1/Problems/  "include('Axioms/SET003-0.ax')." *)

let default_options = 
  {problem_path  = "";
   include_path  = ""; 
   problem_files = [];
   solver_per_active = 750;
   solver_per_clauses = 7500;
(*   inst_lit_sel      = "[-num_var;+num_symb;-sign]";*)
(*   inst_lit_sel = "[+ground;+sign;+num_symb]";*)
   inst_lit_sel = "[+sign;+ground;-num_var;-num_symb]";
   inst_pass_queue1  = "[-num_var;-num_symb]";
   inst_pass_queue2  = "[+age;-num_symb]";
   inst_pass_queue1_mult = 15;
   inst_pass_queue2_mult = 5;
   dismatching           = true;
   eager_unprocessed_to_passive = false;
   ground_splitting_input       = false;
   ground_splitting_full        = false;
   learning_loop_flag           = true;
   learning_start               = 1000;
   learning_factor              = 2;
   fof                          = true;
   eprover_path                 = ""
}
 
let options  = default_options
    
(*files are added in reverse order *)

let add_file_options options file_name = 
  options.problem_files <- file_name::options.problem_files
					
let default_arg_fun = add_file_options options

let input_file_fun  = default_arg_fun     

let problem_path_fun path = 
  options.problem_path <- path 

let include_path_fun path = 
  options.include_path <- path 

let inst_lit_sel_opt       str   = options.inst_lit_sel <- str
let solver_per_active_opt  int   = options.solver_per_active <- int
let solver_per_clauses_opt  int  = options.solver_per_clauses <- int
let inst_pass_queue1_opt   str   = options.inst_pass_queue1 <-str
let inst_pass_queue2_opt   str   = options.inst_pass_queue2 <- str
let fof_opt                b     = options.fof <- b
let eprover_path_opt       str   = options.eprover_path <- str

let inst_pass_queue1_mult_opt int = 
  if int >= 0  then
    options.inst_pass_queue1_mult <- int
  else 
   (failwith "--inst_pass_queue1_mult: should be >= 0")

let inst_pass_queue2_mult_opt int = 
  if int >= 0  then
    options.inst_pass_queue2_mult <- int
  else 
    (failwith "--inst_pass_queue2_mult: should be >= 0")

let learning_start_opt int = 
 if int >= 0  then
   options.learning_start <- int
 else 
    (failwith "--learning_start: should be >=0")

let  learning_factor_opt int =
  if int >= 0  then
    options.learning_factor <- int
  else 
    (failwith "--learning_factor: should be >=0")

let dismatching_opt b = 
  options.dismatching  <- b
      
let eager_unprocessed_to_passive b = 
  options.eager_unprocessed_to_passive  <- b

let ground_splitting_input b = 
 options.ground_splitting_input <- b

let ground_splitting_full b = 
 options.ground_splitting_full <- b

let learning_loop_flag_opt b = 
  options.learning_loop_flag <- b


let info_str str = "\n    "^str

let spec_list = 
  [
   ("--problem_path", Arg.String(problem_path_fun), 
    (info_str "problem path prefix;") 
    ^(info_str "example: --problem_path /home/TPTP-v3.1.1/Problems/PUZ/  \n "));
   ("--include_path", Arg.String(include_path_fun), 
    (info_str "path prefix for includes;") 
    ^(info_str "example: --include_path /home/TPTP-v3.1.1/ \n"));
   ("--input_file", Arg.String(input_file_fun), 
    (info_str "input file name;") 
    ^(info_str " example: --input_file PUZ039-1.p \n"));
   ("--fof", Arg.Bool(fof_opt), 
    (info_str "to accept first order input; requires eprover for clausification;")
    ^(info_str "example: --fof true \n"));
   ("--eprover_path", Arg.String(eprover_path_opt), 
    (info_str "path to the eprover executable; eprover is used for clausification;") 
    ^(info_str "example: /home/EPROVER/ \n"));  
   ("--inst_lit_sel", Arg.String(inst_lit_sel_opt), 
    (info_str "literal selection function: max literal in  any lex combination of") 
    ^(info_str "sign,num_symb,num_var,ground,split;")
    ^(info_str	"example: --inst_lit_sel [-sign;+ground;+num_symb] \n"));
   ("--solver_per_active", Arg.Int(solver_per_active_opt), 
    (info_str "SAT solver calls per main loop iterations;")
     ^(info_str "example: --solver_per_active 750\n"));
   ("--solver_per_clauses", Arg.Int(solver_per_clauses_opt), 
    (info_str "SAT solver calls per new clauses;") 
    ^(info_str "example: --solver_per_clauses 7500 \n"));
   ("--inst_pass_queue1", Arg.String(inst_pass_queue1_opt), 
    (info_str "first passive priority queue, any lex combination of ")
    ^(info_str "num_symb, num_var, ground, age;") 
    ^(info_str " example: --inst_pass_queue1 [+age;-num_symb;-num_var]\n "));
   ("--inst_pass_queue2", Arg.String(inst_pass_queue2_opt), 
    (info_str "second passive priority queue, any lex combination of ")
    ^(info_str "num_symb, num_var, ground, age;")
    ^(info_str " example: --inst_pass_queue2 [-num_var;-num_symb]\n "));
   ("--inst_pass_queue1_mult", Arg.Int(inst_pass_queue1_mult_opt), 
    (info_str "first passive queue multiple;")
    ^(info_str " example: --inst_pass_queue1_mult 5 \n"));
   ("--inst_pass_queue2_mult", Arg.Int(inst_pass_queue2_mult_opt), 
    (info_str "second passive queue multiple;")
    ^(info_str " example: --inst_pass_queue2_mult 15 \n"));
   ("--dismatching", Arg.Bool(dismatching_opt), 
    (info_str  "dismatching constraints; ")
    ^(info_str " example:  --dismatching true \n"));
   ("--eager_unprocessed_to_passive", Arg.Bool(eager_unprocessed_to_passive), 
    (info_str "example: --eager_unprocessed_to_passive false \n"));
   ("--ground_splitting_input", Arg.Bool(ground_splitting_input), 
    (info_str "example: --ground_splitting_input  true\n"));
   ("--ground_splitting_full", Arg.Bool(ground_splitting_full), 
    (info_str "example: --ground_splitting_full false\n"));
   ("--learning_loop_flag", Arg.Bool(learning_loop_flag_opt), 
    (info_str "example: --learning_loop_flag true\n"));
   ("--learning_start", Arg.Int(learning_start_opt), 
    (info_str "num. of loop iter. when learning starts;") 
    ^(info_str "example: --learning_start 1000\n"));
   ("--learning_factor", Arg.Int(learning_factor_opt), 
    (info_str "learning bound is mult. by this factor after each round;") 
       ^(info_str "example: --learning_factor 2\n"))
]
        

(*let usage_msg = 
  "iproveropt [--problem_path path_prefix] [--include_path path_prefix] [--input_file file] input_file_1 ... input_file_n"*)

let usage_msg = "iproveropt [options] [filenames]" 
    
let read_args() = 
  Arg.parse spec_list default_arg_fun usage_msg
    
let () = read_args(); 
  if options.problem_files = [] 
  then
    (Arg.usage spec_list usage_msg;
     failwith "No problem files to solve")
  else ()

exception Parse_fail

let parse_list_opt parse_body_fn parse_fail str = 
  let str_ln = String.length str in
  try
    if (str.[0] = '[') & (str.[(str_ln-1)] = ']')
    then 
      let new_str = String.sub str 1 (str_ln-2) in
      let str_list = Str.split (Str.regexp "[|;|]") new_str in 
      parse_body_fn str_list 
    else 
      raise Parse_fail
  with 
    Parse_fail -> parse_fail ()

(**** Parse selection fun *)

let parse_sel_fail () =   
  Arg.usage spec_list usage_msg;
  failwith "--inst_lit_sel"
 

(* ["num_symb";"num_var"] makes fun list [num_symb;num_var]*)
let parse_sel_str_body str_list = 
  if str_list = [] then raise Parse_fail
  else 
    let str_to_fun str = 
      match str with 
      | "num_symb" | "+num_symb" -> Term.cmp_num_symb
      | "-num_symb"  -> (compose_12 (~-) Term.cmp_num_symb)
	    
      | "num_var" | "+num_var" -> Term.cmp_num_var
      | "-num_var"  -> (compose_12 (~-) Term.cmp_num_var)
	    
      | "ground" | "+ground" -> Term.cmp_ground
      | "-ground"  -> (compose_12 (~-) Term.cmp_ground)
	    
      | "sign" | "+sign" -> Term.cmp_sign
      | "-sign"  -> (compose_12 (~-) Term.cmp_sign)
      | "split" | "+split" -> Term.cmp_split
      | "-split"  -> (compose_12 (~-) Term.cmp_split)
      |_-> raise Parse_fail 
    in
    List.map str_to_fun str_list



let parse_sel_str () = 
  let sel_fun_list = 
    parse_list_opt parse_sel_str_body parse_sel_fail options.inst_lit_sel
  in add_param_str ("Inst Lit Sel: lex_comb "^options.inst_lit_sel^"\n");
  sel_fun_list

let sel_lex_compare = 
  lex_combination (parse_sel_str ()) 

let lit_sel_fun get_prop_truth clause = 
  let candidate_list = Clause.find_all get_prop_truth clause in
(*  out_str ("Sel_fun_cand: "^(Term.term_list_to_string candidate_list)^"\n");*)
	 (*  ^"Sel_truth_first"^get_prop_truth);*)
  list_find_max_element sel_lex_compare candidate_list

(* End Parse Selection *)

(* Parse two cmp functions for two Queues *)
let parse_queue1_fail () = 
  Arg.usage spec_list usage_msg;
  failwith "--inst_pass_queue1"

let parse_queue2_fail () = 
  Arg.usage spec_list usage_msg;
  failwith "--inst_pass_queue2"
 
let parse_queue_str_body str_list = 
  if str_list = [] then raise Parse_fail
  else 
    let str_to_fun str = 
      match str with 
      | "num_symb" | "+num_symb" -> Clause.cmp_num_symb
      | "-num_symb"  -> (compose_12 (~-) Clause.cmp_num_symb)
	    
      | "num_var" | "+num_var" -> Clause.cmp_num_var
      | "-num_var"  -> (compose_12 (~-) Clause.cmp_num_var)
	    
      | "ground" | "+ground" -> Clause.cmp_ground
      | "-ground"  -> (compose_12 (~-) Clause.cmp_ground)
	    
      | "age" | "+age" -> Clause.cmp_age
      | "-age"  -> (compose_12 (~-) Clause.cmp_age)
      |_-> raise Parse_fail 
    in
    List.map str_to_fun str_list


let parse_queue1_str () = 
  let queue1_fun_list = 
    parse_list_opt parse_queue_str_body parse_queue1_fail options.inst_pass_queue1
  in 
  add_param_str ("Passive Queue 1: lex_comb "
		 ^options.inst_pass_queue1
		 ^" Multiple: "
		 ^(string_of_int options.inst_pass_queue1_mult)^"\n");
  queue1_fun_list

let pass_queue1_cmp = 
  lex_combination ((parse_queue1_str ())@[(compose_12 (~-) Clause.compare)])


let parse_queue2_str () = 
  let queue2_fun_list = 
    parse_list_opt parse_queue_str_body parse_queue2_fail options.inst_pass_queue2
  in 
  add_param_str ("Passive Queue 2: lex_comb "^options.inst_pass_queue2
		 ^" Multiple: "
		 ^(string_of_int options.inst_pass_queue2_mult)^"\n");
  queue2_fun_list

let pass_queue2_cmp = 
  lex_combination ((parse_queue2_str ())@[(compose_12 (~-) Clause.compare)])


(* End Parse to Queues *)

(* Parse problems to DB *)

let () = 
  let add_path fn = options.problem_path^fn in
  options.problem_files <- (List.map add_path options.problem_files)

(*
let problem_files_out = 
  out_str_debug ("%Problem files: "^(list_of_str_to_str options.problem_files ",")^"\n")
*) 

let parsed_all = 
  if options.fof then 
    ParseFiles.eprover_clausify_parse options.eprover_path options.problem_files
  else
    ParseFiles.parse_files options.include_path options.problem_files

let _ =  Parsed_input_to_db.parsed_input_to_db parsed_all

(*let ()=out_str_debug ("debug iprover:\n "
		^(Clause.clause_list_to_string !Parsed_input_to_db.clause_list_ref))
*) 

(* End Parse problems to DB*)


let ()= add_param_str ("Dismatching constrains: "^(string_of_bool options.dismatching )^"\n")
 
let ()= add_param_str ("Eager unprocessed to passive: "^(string_of_bool options.eager_unprocessed_to_passive)^"\n")

let ()= add_param_str ("Ground_splitting_input: "^(string_of_bool options.ground_splitting_input)^"\n")
let ()= add_param_str ("Ground_splitting_full: "^(string_of_bool options.ground_splitting_full)^"\n")

module InstParam = 
  struct 
    let inst_lit_sel_fun      = lit_sel_fun
    let solver_per_active     = options.solver_per_active
    let solver_per_clauses    = options.solver_per_clauses
    let inst_pass_queue1_cmp  = pass_queue1_cmp
    let inst_pass_queue2_cmp  = pass_queue2_cmp
    let inst_pass_queue1_mult = options.inst_pass_queue1_mult
    let inst_pass_queue2_mult = options.inst_pass_queue2_mult
    let dismatching           = options.dismatching
    let eager_unprocessed_to_passive = options.eager_unprocessed_to_passive
    let ground_splitting_input       = options.ground_splitting_input
    let ground_splitting_full        = options.ground_splitting_full
    let learning_loop_flag           = options.learning_loop_flag
    let learning_start               = options.learning_start
    let learning_factor              = options.learning_factor
  end

module Instantiation = Instantiation.Make(InstParam)

(****************************************)

    let () = 
      try Instantiation.start_instantiation ()
      with 
	Termination_Signal -> out_str "\n Termination Signal\n"
      
(****************************************)
