open Lib
open Parser_types 

type var = Var.var
type clause = Clause.clause

type parsed_var = string
(*association*)
type var_map = (parsed_var*var) list 

(*to avoid passing max var as parameter to functions we assume global max_var*)


let max_var_ref   = ref (Var.get_first_var()) 
let symbol_db_ref = ref (SymbolDB.create_name "All_Symbols_DB")
let term_db_ref   = ref (TermDB.create_name "All_Terms_DB")
let clause_list_ref = ref []
(*
let clause_db_ref = ref (ClauseAssignDB.create_name "All_Clauses_DB")
*)

(*
(* theory symbols *)              
let neg_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "~" Symbol.Neg 1) symbol_db_ref
let () = Symbol.assign_is_input false neg_symb 

let true_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "true" Symbol.True 0) symbol_db_ref
let () = Symbol.assign_is_input false true_symb

let false_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "false" Symbol.False 0) symbol_db_ref
let () = Symbol.assign_is_input false false_symb

let equality_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "equal" Symbol.Equality 2) symbol_db_ref
let () = Symbol.assign_is_input false equality_symb

let plus_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "+" Symbol.Plus 2) symbol_db_ref
let () = Symbol.assign_is_input false plus_symb

let minus_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "-" Symbol.Minus 2) symbol_db_ref
let () = Symbol.assign_is_input false minus_symb

(* unary minus should be different since name is a key (it can be changed in future)*)
let unary_minus_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "-u" Symbol.UnaryMinus 1) symbol_db_ref
let () = Symbol.assign_is_input false unary_minus_symb

let bot_symb = 
  SymbolDB.add_ref 
    (Symbol.create_from_str_arity "bot" Symbol.Bot 0) symbol_db_ref
let () = Symbol.assign_is_input false bot_symb
*)

let bot_term = TermDB.add_ref (Term.create_fun_term Symbol.symb_bot []) term_db_ref

(* is_negative is not needed*)
let rec is_negative = function
  |UnaryFormula(Negation,formula)
    -> bool_plus true (is_negative formula)
  |_-> false 
      
let rec get_list_of_literals parsed_formula = 
  match parsed_formula with
  |BinaryFormula(binary_connective,formula1,formula2) ->
      if binary_connective = Or then 
	(get_list_of_literals formula1)@(get_list_of_literals formula2)
      else failwith 
	  "Some input formula is not in CNF: some binary connective is not |"
  | Atom(atom) -> 
      (match atom with  
      | TheoryTerm(NegEquality(t1,t2)) -> 
	  [UnaryFormula(Negation,Atom(TheoryTerm(Equality(t1,t2))))]    
(* remove false*)
      | TheoryTerm(False) -> []    
      |_-> [parsed_formula]
      )
  | UnaryFormula(Negation,formula) ->
     ( match formula with 
      |Atom(atom) -> 
          (match atom with 
	  |TheoryTerm(NegEquality(t1,t2)) -> 
	      [Atom(TheoryTerm(Equality(t1,t2)))] 
(* remove false *)
	  |TheoryTerm(True) -> []
	  |_-> [parsed_formula]
	  ) 
      |_-> failwith "Some input formula is not in CNF:" 
      ) 	    
  |QuantifiedFormula(_,_,_) ->
      failwith "CNF formula should be without quantifiers 
	(all variables are implicitly universally quantifiered)"
	

let rec parsed_term_to_term var_map_ref symbol_type = function 
  |TheoryTerm(theory_term) -> 
      theory_term_to_term var_map_ref symbol_type theory_term
  |UserTerm(user_term) ->  
      user_term_to_term var_map_ref symbol_type user_term
  |Var(parsed_var) -> 
      (try 
	Term.create_var_term (List.assoc parsed_var !var_map_ref)
      with
	Not_found -> 
          let new_var = !max_var_ref in 
          max_var_ref := Var.get_next_var !max_var_ref;
 	  var_map_ref:=(parsed_var,new_var)::!var_map_ref;
	  Term.create_var_term new_var
      )  

and theory_term_to_term var_map_ref symbol_type = function 
  |Equality(parsed_term1, parsed_term2) -> 
      Symbol.assign_is_input true Symbol.symb_equality;
      Term.create_fun_term Symbol.symb_equality 
	[(parsed_term_to_term var_map_ref Symbol.Fun parsed_term1); 
	 (parsed_term_to_term var_map_ref Symbol.Fun parsed_term2)]
	(* NegEquality is eliminated before*)
  |True ->   
      Symbol.assign_is_input true Symbol.symb_true;
      (Term.create_fun_term Symbol.symb_true [])
  |False ->  
      Symbol.assign_is_input true Symbol.symb_false;
      (Term.create_fun_term Symbol.symb_false [])
(* numbers as constants no theory treatment, can be added infuture*)
  |PositiveInteger (int)->  
      let int_symb = 
	(SymbolDB.add_ref 
	   (Symbol.create_from_str_arity 
	      (string_of_int int) Symbol.Fun 0) symbol_db_ref) in
      Symbol.assign_is_input true int_symb;
      Term.create_fun_term int_symb []
  |RealNumber(int1,int2)->
      let real_str = (string_of_int int1)^"."^(string_of_int int2) in 
      let real_symb = 
	SymbolDB.add_ref 
	  (Symbol.create_from_str_arity real_str  Symbol.Fun 0) 
	  symbol_db_ref in
      Symbol.assign_is_input true real_symb;
      Term.create_fun_term real_symb []
  |Plus(parsed_term1, parsed_term2)->
      Symbol.assign_is_input true Symbol.symb_plus;
      Term.create_fun_term Symbol.symb_plus 
	[(parsed_term_to_term var_map_ref Symbol.Fun parsed_term1); 
	 (parsed_term_to_term var_map_ref Symbol.Fun parsed_term2)]
	
  |Minus(parsed_term1, parsed_term2)->
      Symbol.assign_is_input true Symbol.symb_minus;
      Term.create_fun_term Symbol.symb_minus
	[(parsed_term_to_term var_map_ref Symbol.Fun parsed_term1); 
	 (parsed_term_to_term var_map_ref Symbol.Fun parsed_term2)]
  |UnaryMinus(parsed_term) ->
      Symbol.assign_is_input true Symbol.symb_unaryminus;
      Term.create_fun_term Symbol.symb_unaryminus
	[(parsed_term_to_term var_map_ref Symbol.Fun parsed_term)]

  |NegEquality(_,_)-> 
      failwith "strange error: NegEquality should of been eliminated" 	
          (* NegEquality is eliminated before*)	

and user_term_to_term var_map_ref symbol_type = function
  |Fun(parsed_symbol,parsed_term_list) ->
      let symb = 
	SymbolDB.add_ref 
	  (Symbol.create_from_str_arity 
	     parsed_symbol symbol_type (List.length parsed_term_list)) 
	  symbol_db_ref in
      Symbol.assign_is_input true symb;
      let args = 
	list_map_left (parsed_term_to_term var_map_ref Symbol.Fun) 
	  parsed_term_list in 
      Term.create_fun_term symb args
     
	
   
let rec parsed_literal_to_literal var_map_ref = function 
  |Atom(atom) -> 
      parsed_term_to_term var_map_ref Symbol.Pred atom
  |UnaryFormula(Negation,formula) -> 
(* here formula can be only Atom(...)*)  
      Symbol.assign_is_input true Symbol.symb_neg;
      (Term.create_fun_term Symbol.symb_neg
	 [(parsed_literal_to_literal var_map_ref formula)])
  |_ -> failwith 
	"parsed_literals_to_literals -- this shouldnot happen"

(* in list_of_literals_to_clauses
   clauses here are not normalised and 
   therefore terms are not added to db but symbols are
*)
	
let list_of_literals_to_clauses list_of_literals = 
  let var_map_ref = ref [] in 
  Clause.create 
    (list_map_left 
       (parsed_literal_to_literal var_map_ref) list_of_literals  
    )
    
let cnf_formula_to_db parsed_formula =  
  let clause_tmp = 
    list_of_literals_to_clauses (get_list_of_literals parsed_formula) in 
  let new_clause = (Clause.normalise term_db_ref clause_tmp) in
  Clause.assign_input_history new_clause;
  clause_list_ref := new_clause::!clause_list_ref;
  new_clause
    
(*
  ClauseAssignDB.add_ref 
    (Clause.normalise clause_tmp term_db_ref) clause_db_ref
*)

let rec parsed_input_to_db = function 
  |Formula(CNF,name,formula_type,parsed_formula,
	   formula_annotation_list)::tl -> 	    
             let clause = cnf_formula_to_db parsed_formula in
	     clause::(parsed_input_to_db tl)
  |Formula(_,_,_,_,_) :: tl -> 
      failwith "This language is not supported yet try to transfer to CNF" 
  |Include(_,_)::_ -> failwith "should be no includes here..."
  |Comment(_)::tl  -> (parsed_input_to_db tl)
  |Annotation(_)::tl  -> (parsed_input_to_db tl)
  |[] -> []

(*  List.iter top_element_to_db parsed_input *)
    
    
