open Lib  
(*type 'a param  = 'a Lib.param *)

type arity = int param
type key = string
(* type the symbol can have *)
type stype = 
  |Fun|Pred|Connective|Quantifier

type sproperty = Theory|Split


(* fast key can be used only when symb. are already put in symbolDB*)
type fast_key  = int param
type symbol = 
    {
     mutable fast_key: fast_key;
     name    : string; 
     arity   : arity; 
     stype   : stype;
     mutable sproperty : sproperty param;
     mutable group : int param;   
(* it is useful to partition all symbols to groups e.g. to represent 
   signature information in terms
   inv: group number should be less or equal to the Bit_vec.max_pos *)
     mutable is_input : bool param;
(* is_input -- the symbol is in the input of the problem *)
     mutable is_skolem : bool param
(* the symbol is a skolem function *)
   }

(****************** special symbols *)

(* Connectives *)
let symb_neg = 
  {
   fast_key  = Undef;
   name      = "~"; 
   arity     = Def(1);
   stype     = Connective; 
   sproperty = Undef;  
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
 }

let symb_and = 
  {
   fast_key  = Undef;
   name      = "&"; 
   arity     = Def(2); 
   stype     = Connective;
   sproperty = Undef;   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_or = 
  {
   fast_key  = Undef;
   name      = "\\/"; 
   arity     = Def(2); 
   stype     = Connective;
   sproperty = Undef;    
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_impl = 
  {
   fast_key  = Undef;
   name      = "->"; 
   arity     = Def(2); 
   stype     = Connective;
   sproperty = Undef;    
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

(* quantifiers *)
let symb_forall =
  {
   fast_key  = Undef;
   name      = "A"; 
   arity     = Def(1); 
   stype     = Quantifier;
   sproperty = Undef;   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_exists =
  {
   fast_key  = Undef;
   name      = "E"; 
   arity     = Def(1); 
   stype     = Quantifier;
   sproperty = Undef;   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}


(* theory symbols*)
let symb_true =
  {
   fast_key  = Undef;
   name      = "True"; 
   arity     = Def(0); 
   stype     = Pred;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_false =
  {
   fast_key  = Undef;
   name      = "False"; 
   arity     = Def(0); 
   stype     = Pred;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_equality =
  {
   fast_key  = Undef;
   name      = "="; 
   arity     = Def(2); 
   stype     = Pred;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_plus =
  {
   fast_key  = Undef;
   name      = "+"; 
   arity     = Def(2); 
   stype     = Fun;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_minus =
  {
   fast_key  = Undef;
   name      = "-"; 
   arity     = Def(2); 
   stype     = Fun;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}


let symb_unaryminus =
  {
   fast_key  = Undef;
   name      = "-"; 
   arity     = Def(1); 
   stype     = Fun;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Undef;
   is_skolem = Def(false);
}

let symb_bot =
  {
   fast_key  = Undef;
   name      = "iProver_bot"; 
   arity     = Def(0); 
   stype     = Fun;
   sproperty = Def(Theory);   
   group     = Undef;
   is_input  = Def(false);
   is_skolem = Def(false);
}


exception Symbol_fast_key_undef
exception Arity_undef
      
let create_from_str (name:string) (stype:stype) = 
  {name=name; stype=stype; 
   arity=Undef; fast_key=Undef; group=Undef; sproperty = Undef;
   is_input = Undef; is_skolem = Undef}

let create_from_str_arity (str:string) (stype:stype) (ar:int)  = 
  {name=str; stype=stype; 
   arity=Def(ar); fast_key=Undef; group=Undef; sproperty = Undef;
   is_input = Undef; is_skolem = Undef}



exception Symbol_fast_key_is_def

let assign_fast_key (s:symbol) (key:int) = 
 match s.fast_key with 
  |Undef -> s.fast_key <- Def(key)
  |_     -> raise Symbol_fast_key_is_def
    
let assign_group s group = 
 match s.group with 
 |Undef -> 
     if group <= Bit_vec.max_pos
     then s.group <- Def(group)
 |_-> failwith "group is already assigned"

let assign_property prop s = 
  s.sproperty <- Def(prop)

let get_property s = s.sproperty
  

let assign_is_input bval s =
  s.is_input <- Def(bval)


let assign_is_skolem bval s =
  s.is_skolem <- Def(bval)

exception Is_input_undef
let get_is_input s = 
  match s.is_input with 
  |Def(bv) -> bv
  |Undef -> false
(*raise Is_input_undef*)

exception Is_skolem_undef
let get_is_skolem s = 
  match s.is_skolem with 
  |Def(bv) -> bv
  |Undef -> raise Is_skolem_undef


let get_arity (s:symbol)    = 
  match s.arity with
  |Def(arity)   -> arity
  |Undef        -> raise  Arity_undef
	
let get_type (s:symbol) = s.stype
(*let get_key  (s:symbol) = s.name*)


exception Symbol_fast_key_undef

let  get_fast_key (s:symbol) =
  match s.fast_key with
  |Def(fkey)   -> fkey
  |Undef       -> raise  Symbol_fast_key_undef

exception Group_undef
let get_group s = 
  match s.group with
  |Def(group) -> group
  |Undef      -> raise Group_undef


(*let  compare_key (s1:symbol)(s2:symbol)  = *)
  (*(String.compare (get_key s1) (get_key s2))*)

let compare_name s1 s2 = (String.compare s1.name s2.name)
let compare_type s1 s2 = (compare s1.stype s2.stype)
let compare_arity s1 s2 = (compare s1.arity s2.arity)
let compare_property s1 s2 = (compare s1.sproperty s2.sproperty)

let  compare_key = 
  lex_combination [compare_name;compare_arity;compare_type;compare_property]


let  compare_fast_key (s1:symbol)(s2:symbol) = 
  (compare (get_fast_key s1) (get_fast_key s2)) 
     
let compare = compare_fast_key

let equal s1 s2 =  if (compare s1 s2)==cequal then true else false 

let hash  = get_fast_key 

(* unsafe compare    
let  compare (s1:t) (s2:t) = 
  try (compare_fast_key s1 s2) with 
    Fast_key_undef -> (compare_key s1 s2)
*)

let  to_string (s:symbol) = s.name
    
let to_prolog symb = 
  let symb_str = to_string symb in 
  match symb_str.[0] with 
    'a'..'z'  -> symb_str
  | 'A'..'Z'  -> symb_str.[0]<- (Char.lowercase symb_str.[0]); symb_str
  | '0'..'9'  -> "f"^symb_str
  |_ -> failwith "Symbol.to_prolog: unexpected char in symb_name"
	
