(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)
(*
let ppp t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i
*)

let is_eq equals term =
	(Convert.term_functor_symbol term) == equals

(* Exceptions *)
exception Redundant
exception Found
exception Failed

type atom =
	Term of Term.term
	| Eq of Equality.equation

let make_atom term st =
	if (Convert.term_functor_symbol term) == (Term.get_equality_symbol st) then (
		let l, r = Convert.term_pair term 
		in
			Eq (Equality.make_checked_equation term l r))
	else
		Term term    

let atom_to_term atom =
	match atom with
		Term t -> t
		| Eq eq -> Equality.get_term eq		
		
type sign =
    Pos
  | StratifiedNegAsFailure
  | StratifiedNegAsFailureNotSubsumed
  | Findall
  | Neg2

let lit_sign_and_atom_term st term =
  let f = Convert.term_functor_symbol term in
  if f == (Term.get_stratified_neg_as_failure_symbol st) ||
     f == (Term.get_stratified_neg_as_failure_alternate_symbol st) 
  then
    (StratifiedNegAsFailure, Convert.term_nth_arg term 0)
  else if f == (Term.get_stratified_neg_as_failure_not_subsumed_symbol st) then
    (StratifiedNegAsFailureNotSubsumed, Convert.term_nth_arg term 0)
  else if f == (Term.get_findall_symbol st) then
    (Findall, Convert.term_nth_arg term 1)
  else if f == (Term.get_neg2_symbol st) then
    (Neg2, Convert.term_nth_arg term 0)
  else
    (Pos, term)


    
(****************************************************************************)
(* Auxiliary *)
(****************************************************************************)

let renumber_variables term =
  let table = Array.make Term.max_vars (-1) in
  Term.renumber_variables_term table term;
  term

let remove_duplicates array =
	Array.mapi
		(fun i x ->
			if i==0 then x
			else
				(if x==array.(i-1) then ()
				 else x))
		array

let remove_term_duplicates array =
	let result = ref []
	in
		Array.iteri
			(fun i x ->
				if i>0 then
					(if Term.term_ident x array.(i-1) then ()
					else result := x :: !result)
				else result := x :: !result)
			array;
		Array.of_list !result

let rec remove_term_duplicates_from_list list =
	match list with
		[] -> []
		| head :: tail ->
			head :: (remove_term_duplicates_from_list_1 head tail)
				
and remove_term_duplicates_from_list_1 previous list =
	match list with
		[] -> []
		| head :: tail ->
			if Term.term_ident head previous then
				(remove_term_duplicates_from_list_1 head tail)
			else
				head :: (remove_term_duplicates_from_list_1 head tail)

let purge_literals_list equals lits =
	let oriented_lits =
		List.map
			(fun term ->
				if (Convert.term_functor_symbol term)==equals then
					(let l, r = Convert.term_pair term
					in
						if (LrpOrder.ground_like_lrpo_greater l r) then
							(term)
						else
							(Term.Compound
								(Term.make_compound
									equals
									[|r; l|])))
				else
					(term))
			lits
	in
	let s_lits = List.sort LrpOrder.ground_like_compare_lrpo_greater oriented_lits
	in
		remove_term_duplicates_from_list s_lits			

(****************************************************************************)
(* Body Literal *)
(****************************************************************************)

module BodyLiteral =
  struct


(****************************************************************************)
(* Optim Info:  Information about the optimization intended for humans. *)
(****************************************************************************)

    type optim_info =
	{ applicability_hint: Opt.applicability_value;
	  discrimination_hint: Opt.discrimination_value;
	  binding_pattern: Opt.binding_pattern;
	  extent_size: int }

    let make_optim_info
	applicability_hint
	discrimination_hint
	binding_pattern
	extent_size =
      { applicability_hint=applicability_hint;
	discrimination_hint=discrimination_hint;
	binding_pattern=binding_pattern;
	extent_size=extent_size }

    let make_default_optim_info () =
      make_optim_info 0.0 0.0 [||] 0 

(****************************************************************************)
(* BodyLiteral *)
(****************************************************************************)

    type literal =
	{mutable sign: sign;
	 mutable predicate: Predicate.predicate;
	 mutable atom: atom;
	 mutable term: Term.term;
	 mutable reader: Access.Reader.reader;
	 mutable foreign_context: Term.context;
	 mutable is_delta: bool;
	 mutable optim_info: optim_info;
	 mutable builtin: Term.p;
	 mutable subproof_pos: int;
	 mutable output_vars: int list}

    let get_sign lit = lit.sign

    let get_predicate lit = lit.predicate

    let get_atom lit = lit.atom
    
    let get_atom_term lit =     	
    	match lit.atom with
    		Term t -> t
    		|Eq eq -> Equality.get_term eq

	let get_is_oriented lit =
		match lit.atom with
			Term _ -> false
			|Eq eq -> Equality.get_is_oriented eq    		
    		
    let get_reader lit = lit.reader

    let get_term lit = lit.term
    
    let get_foreign_context lit = lit.foreign_context

    let get_is_delta lit = lit.is_delta

    let get_optim_info lit = lit.optim_info

    let get_builtin lit = lit.builtin

    let get_subproof_pos lit = lit.subproof_pos

    let set_subproof_pos lit n = lit.subproof_pos <- n
	

(****************************************************************************)
(* "Simple" Negation - i.e. Stratified Negation as Failure *)
(****************************************************************************)

(* Construct a builtin function, that interns the reader. *)

    exception Neg_failed
	
    let make_simpleneg_pdeterministic predicate equals =
      let reader = Predicate.get_all_bound_reader predicate in
      Term.PDeterministic
	((fun term context foreign_context trail ->
	  let saved_trail = Term.save_trail trail in
	  try
	  	(if (equals==(Predicate.get_symbol predicate)) then
	  		(let left, right = Convert.term_pair term
	  		in
	  		let trail1 = (Term.unify left context right context trail)
	  		in
	  			if Term.failed trail1 then ()
	  			else
	  				(raise Neg_failed));
	    Access.Reader.iter_unification_candidates    
	      (fun entry ->
		let trail1 =
		  Term.unify
		    term
		    context
		    (Fact.get_term entry)
		    foreign_context
		    trail in
		if Term.failed trail1 then
		  ()
		else
		  begin
		    ignore (Term.undo_bindings trail1 saved_trail);
		    raise Neg_failed
		  end)
	      reader
	      term
	      context;
	    trail)
	  with Neg_failed ->
	    Term.fail),
	 let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
	 (Array.make arity Opt.InputGround,
	  Array.make arity 1.0,
	  Opt.UnknownProof))

(****************************************************************************)
(* Stratified Negation as Failure - "Not Subsumed" Variation *)
(****************************************************************************)

    let make_subsneg_pdeterministic predicate =
      Term.PDeterministic
	((fun term context foreign_context trail ->
	  try
	    Access.Reader.iter_generalization_candidates 
	      (fun fact -> 
		if Term.subsumes_chk
		    (Fact.get_term fact)
		    (* ### This is not the best implementation, since
		       the term is copied here - to avoid this, we
		       would need a subsumes_chk that takes a context
		       produced by unify. Copying is done here
		       only for generalization candidates. *)
		    (Term.apply term context) 
		then
		  raise Neg_failed
		else
		  ())
	      (Predicate.get_all_bound_reader predicate)
	      term
	      context;
	    trail
	  with Neg_failed ->
	    Term.fail),
	 let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
	 (Array.make arity Opt.OutputFree,
	  Array.make arity 1.0,
	  Opt.UnknownProof))

(****************************************************************************)
(* Findall *)
(****************************************************************************)

(* Like Prolog's findall(Absterm, Query, List) but with sorted output
   list. Query argument must be a single positive literal. Variables
   in Absterm may only appear in Absterm and Query, and not anywhere else in
   the clause. All variables in Query that are not in term must appear
   in another subgoal (i.e. bound at call time). 

   Actually a findall subgoal is (### currently ?) handled with a
   mode that is InputGround for all Query argument positions that
   contain a variable not in Absterm.
*)


let rec term_list_rm_duplicates list =
  match list with
    e1 :: e2 :: list1 ->
      if (StdOrder.term_compare
	    e1 Term.null_context e2 Term.null_context) = 0 then
	term_list_rm_duplicates (e2 :: list1)
      else
	e1 :: term_list_rm_duplicates (e2 :: list1)
  | _ -> list
	
let make_findall_function symbol_table reader absterm list =
  let dot = Term.find_symbol symbol_table "." 2
  and nil = Term.find_symbol symbol_table "[]" 0 in
  (fun query context foreign_context trail ->
    let results = ref [] in
    Access.Reader.iter_unification_candidates
      (fun entry ->
	let saved_trail = Term.save_trail trail in
	let trail1 =
	  Term.unify
	    query
	    context
	    (Fact.get_term entry)
	    foreign_context
	    trail in
	if Term.failed trail1 then
	  ()
	else
	  begin
	    Listutils.push
	      (renumber_variables (Term.apply absterm context))
	      results;
	    ignore (Term.undo_bindings trail1 saved_trail)
	  end)
      reader
      query
      context;
    let output_list =
      Convert.list_of_term_to_term_1
	dot nil
	(term_list_rm_duplicates
	   (List.sort
	      (fun e1 e2 -> 
		StdOrder.term_compare 
		  e1 Term.null_context e2 Term.null_context)
	      !results)) in
    Term.unify
      list
      context
      output_list
      context
      trail)


let make_findall_pdeterministic symbol_table predicate query absterm list =
  let absvars = ref [] in
  Termutils.iter_vars
    (fun v -> Listutils.pushnew (Term.variable_number v) absvars)
    absterm;
  let arity = (Term.symbol_arity (Predicate.get_symbol predicate)) in
  let mode = Array.make arity Opt.OutputGround
  and bp = Array.make arity Opt.Free in
  if arity > 0 then
    match query with
      Term.Compound k ->
	let args = Term.compound_args k in
	for i = 0 to (arity - 1) do
	  try
	    Termutils.iter_vars
	      (fun v -> 
		if (not (List.mem (Term.variable_number v) !absvars)) then
		  raise Found
		else
		  ())
	      args.(i)
	  with Found ->
	    mode.(i) <- Opt.InputGround;
	    bp.(i) <- Opt.Bound
	done
    | _ ->
	raise (Failure "Bad findall arguments.")
  else
    ();
  let reader = Predicate.choose_best_reader predicate bp in
  Term.PDeterministic
    ((make_findall_function symbol_table reader absterm list),
       ( mode,
	 Array.make arity 1.0,
	 Opt.UnknownProof ))
    

(****************************************************************************)
(* BodyLiteral Constructors *)
(****************************************************************************)

    let aov_1 mode args vars =
      for i = 0 to (Array.length mode) - 1 do
	match mode.(i) with
	  Opt.OutputGround ->
	    Termutils.iter_vars
	      (fun v -> Listutils.pushnew 
		  (Term.variable_number v) vars)
	      args.(i)
	| Opt.InputGround | Opt.OutputFree ->
	    ()
      done

    let assign_output_vars lit term =
      let vars = ref [] in
      (match lit.sign with
	StratifiedNegAsFailure | StratifiedNegAsFailureNotSubsumed | Neg2 ->
	  ()
      | Findall ->
	  (* The 3rd arg is not contained in the atom. *)
	  Termutils.iter_vars
	    (fun v -> Listutils.pushnew (Term.variable_number v) vars)
	    (Convert.term_nth_arg lit.term 2)
      | Pos ->
	  match (get_atom_term lit) with
	    Term.Compound compound ->
	      (match lit.builtin with
		Term.PDeterministic (_, (mode, _, _)) ->
		  aov_1 mode (Term.compound_args compound) vars
	      | Term.PUnbound ->
		  try
		    match 
		      List.find
			(fun decl -> 
			  match decl with Opt.ModeDecl _ -> true | _ -> false)
			(Predicate.get_declarations lit.predicate)
		    with Opt.ModeDecl mode ->
		      aov_1 mode (Term.compound_args compound) vars
		    | _ ->
			()
		  with Not_found ->
		    Termutils.iter_vars
		      (fun v -> 
			Listutils.pushnew (Term.variable_number v) vars)
		      (get_atom_term lit))
	  | _ ->
	      ());
      lit.output_vars <- !vars

    let make_literal predicate_table symbol_table eq_convert term =
      let (sign, atom_term) = lit_sign_and_atom_term symbol_table term 
      and equals=Term.get_equality_symbol symbol_table
      in
      let atom_term =
      	if eq_convert && (not (Convert.term_functor_symbol atom_term==equals)) then
      		(Term.Compound
      			(Term.make_compound
      				equals
      				[|atom_term; Term.Unbound|]))
      	else
      		(atom_term)
      in
      let pred = (Predicate.find_predicate
		    predicate_table
		    (Convert.term_functor_symbol atom_term)) in
      Predicate.ensure_ready_to_run pred;
      let builtin =
	match sign with
	  StratifiedNegAsFailure ->
	    (make_simpleneg_pdeterministic pred equals)
	| StratifiedNegAsFailureNotSubsumed ->
	    (make_subsneg_pdeterministic pred)
	| Findall ->
	    (make_findall_pdeterministic
	       symbol_table
	       pred 
	       atom_term
	       (Convert.term_nth_arg term 0)
	       (Convert.term_nth_arg term 2))
	| _ ->
	    Term.symbol_predicate (Predicate.get_symbol pred) in
      let lit =
	{sign=sign;
	 predicate=pred;
	 term=term;
	 atom=(make_atom atom_term symbol_table);
	 is_delta=false;
	 builtin=builtin;
	 output_vars=[];
	 optim_info=make_default_optim_info (); (* Dummy init value. *)
	 reader=Access.Reader.null_reader (); (* Dummy init value. *)
	 foreign_context=Term.make_context 0; (* Dummy init value. *)
	 subproof_pos=(-1)} in (* Dummy init value. *)
      assign_output_vars lit term;
      lit
	  
    (** Returns a shallow copy of the given literal object, i.e.
	field values, including the atom term are shared. *)
    let make_literal_like literal =
      {sign=literal.sign;
       predicate=literal.predicate;
       atom=literal.atom;
       term=literal.term;
       is_delta=literal.is_delta;
       reader=literal.reader;
       foreign_context=literal.foreign_context;
       optim_info=literal.optim_info;
       builtin=literal.builtin;
       subproof_pos=literal.subproof_pos;
       output_vars=literal.output_vars}
	 
    let print_binding_pattern bp =
      for i = 0 to (Array.length bp) - 1 do
	match bp.(i) with
	  Opt.Bound ->
	    print_string "b"
	| Opt.Free ->
	    print_string "f"
      done

    let print_optim_info lit =
      let optim_info = lit.optim_info in
      Printf.printf "applic: %1.5f disc: %1.5f extent: %7d bpattern: "
	optim_info.applicability_hint
	optim_info.discrimination_hint
	optim_info.extent_size;
	print_binding_pattern optim_info.binding_pattern;
      print_endline ""
	
  end


(****************************************************************************)
(* HeadLiteral *)
(****************************************************************************)

	
module HeadLiteral =
  struct

    type literal =
	{mutable atom: atom;
	 mutable predicate: Predicate.predicate}

    let get_atom lit = lit.atom
    
    let get_atom_term lit =
    	match lit.atom with
    		Term t -> t
    		|Eq eq -> Equality.get_term eq

    let get_predicate lit = lit.predicate

    let make_literal predicate_table term = 
    	let atom = term 
    	in
	    	match atom with
    			Term t ->
					let pred = (Predicate.find_predicate
								    predicate_table
		    						(Convert.term_functor_symbol t)) 
		    		in 
						Predicate.ensure_ready_to_run pred;
							{predicate=pred;
							atom=atom}
	   			|Eq eq ->
	   				let pred = (Predicate.find_predicate
								    predicate_table
		    						(Convert.term_functor_symbol (Equality.get_term eq))) 
		    		in 
						Predicate.ensure_ready_to_run pred;
							{predicate=pred;
							atom=atom}
  end

  
    
(****************************************************************************)
(* Aux - Auxiliary datastructures passed to and kept by clauses. *)
(****************************************************************************)

let aux_id = ref 0

let gen_aux_id () =
  aux_id := !aux_id + 1;
  "aux_"^string_of_int(!aux_id)

let reset_aux_id () =
  aux_id := 0  


module Aux =

  struct 

    type aux = {
    aux_id: string;
	predicate_table: Predicate.predicate_table;
	symbol_table: Term.symbol_table;
	cfg: Cfg.cfg;
	negative_units_store: NegativeUnits.store;
	mutable proof_id: int}

    let make_aux cfg predicate_table symbol_table neg_store =
      {aux_id=gen_aux_id (); 
      cfg=cfg;
	predicate_table=predicate_table;
	symbol_table=symbol_table;
	negative_units_store=neg_store;
	proof_id=0}

  end

(****************************************************************************)
(* Disjunction *)
(****************************************************************************)

module DisjunctiveHead =

  struct

    type head =
	{literals:HeadLiteral.literal list}

    let get_literals disjunction =
      disjunction.literals

    let make_head literals =	
	      {literals=literals}

  end

(****************************************************************************)
(* Clause *)
(****************************************************************************)

let id = ref 0

let gen_id () =
  id := !id + 1;
  "clause_"^string_of_int(!id)

let reset_gen_id () =
  id := 0
  
type body =
    Body of BodyLiteral.literal list

type head =
    False
  | Unit of HeadLiteral.literal
  | Conjunction of HeadLiteral.literal list
  | Disjunction of DisjunctiveHead.head
	
type clause = {
    id: string;
    source: Term.term;
    mutable parent_source: Term.term;
    mutable body: body;
    head: head;
    mutable news: News.news;
    mutable subproofs: Fact.fact array;
    mutable parent_subproofs: Fact.fact array;
    cfg: Cfg.cfg;
    aux: Aux.aux;
    ground_head_terms: Term.term array;
    non_ground_head_terms: Term.term array;
    ground_body_terms: Term.term array;
    non_ground_body_terms: Term.term array;
    ground_head_eqs: Term.term array;
    non_ground_head_eqs: Term.term array;
    ground_body_eqs: Term.term array;
    non_ground_body_eqs: Term.term array;
    head_terms_functors_sum: int;
    head_eqs_functors_sum: int;
    body_terms_functors_sum: int;
    body_eqs_functors_sum: int;
    is_ground: bool;
    is_eq_unit: bool;
    is_oriented: bool;
    has_non_eqs: bool;
    sigterm: Term.term;
    fuzzy_sigterm: Term.term;
    impure_varnums: int list;
    is_redundant: bool ref;	
(* done as a ref and shared with delta clauses - switching off a clause will switch off 
all of its delta clauses as well *)
    mutable proof: Fact.proof;
    mutable is_in_current_stratum: bool;
    mutable dom_clause: bool
  }
    
  
let get_equation clause =
	if clause.is_eq_unit then
		(match clause.head with
			Unit lit ->
				(match (HeadLiteral.get_atom lit) with
					Eq eq -> eq
					|_ -> (raise (Err.Error ("Bad unit equation in Ic.get_equation."))))
			|_ -> (raise (Err.Error ("Bad unit equation in Ic.get_equation."))))
	else
		(raise (Err.Error ("Bad unit equation in Ic.get_equation.")))  
  
let get_ground_head_eqs clause = clause.ground_head_eqs		
		
let get_non_ground_head_eqs clause = clause.non_ground_head_eqs	

let get_body clause = clause.body

let get_head clause = clause.head

let get_news clause = clause.news

let get_cfg clause = clause.cfg

let set_news clause news =
  clause.news <- news

let get_id clause = clause.id

let is_eq_unit clause = clause.is_eq_unit

let is_oriented clause = clause.is_oriented

let is_redundant clause = !(clause.is_redundant)

let is_ground clause = clause.is_ground

let declare_redundant clause = clause.is_redundant := true

let undeclare_redundant clause = clause.is_redundant := false

let get_source clause = clause.source

let get_aux clause = clause.aux

let get_sigterm clause = clause.sigterm

let get_fuzzy_sigterm clause = clause.fuzzy_sigterm

let get_proof clause = clause.proof

let set_proof clause proof = clause.proof <- proof

let get_body_predicates clauses =
  let preds = ref [] 
  in
  	List.iter
  		(fun clause ->
  			(match clause.body with
					Body lits ->
  		  			List.iter
  						(fun lit ->
  							Listutils.pushnewq (BodyLiteral.get_predicate lit) preds)
  						lits)) 						
  		clauses;
	!preds
 
let get_impure_varnums clause = clause.impure_varnums

let is_in_current_stratum clause = clause.is_in_current_stratum

let set_is_in_current_stratum bval clause = clause.is_in_current_stratum <- bval

let rec rm_constant symbol terms =
  match terms with
    [] -> []
  | term :: terms1 ->
      if Convert.term_functor_symbol term == symbol then
	rm_constant symbol terms1
      else
	term :: rm_constant symbol terms1
		 
let make_head predicate_table symbol_table eq_convert term =
  let f = Convert.term_functor_symbol term
  and equals = Term.get_equality_symbol symbol_table
  in
  if f == Term.get_disjunction_symbol symbol_table then
    let lits = 
      List.map 
	(fun lit -> 
		if equals==(Convert.term_functor_symbol lit) then
			let l, r = Convert.term_pair lit
			in
				HeadLiteral.make_literal
					predicate_table
					(Eq (Equality.make_checked_equation lit l r))
		else		
			(if eq_convert then
				(let eq_term =
					Term.Compound
						(Term.make_compound
							equals
							[|lit; Term.Unbound|])
				in
					HeadLiteral.make_literal
						predicate_table
						(Eq (Equality.make_checked_equation eq_term lit Term.Unbound)))
			else
				(HeadLiteral.make_literal predicate_table (Term lit))))
	(purge_literals_list equals (		
	(rm_constant
	   (Term.get_false_symbol symbol_table)
	   (Convert.term_sequence_to_list
	      (Term.get_disjunction_symbol symbol_table)
	      term)))) in
	     
    match lits with
      [] -> False
    | [lit] -> Unit lit
    | _ -> Disjunction (DisjunctiveHead.make_head lits)
  else if f == Term.get_conjunction_symbol symbol_table then
    let lits = 
      List.map 
	(fun lit -> 
		if (Term.get_equality_symbol symbol_table)==(Convert.term_functor_symbol lit) then
			let l, r = Convert.term_pair lit
			in
				HeadLiteral.make_literal
					predicate_table
					(Eq (Equality.make_checked_equation lit l r))
		else		
			HeadLiteral.make_literal predicate_table (Term lit))
	(purge_literals_list equals (			
	(Convert.term_sequence_to_list
	   (Term.get_conjunction_symbol symbol_table)
	   term)))
	in
    Conjunction lits
  else if f == Term.get_false_symbol symbol_table then
    False
  else
  	if (Term.get_equality_symbol symbol_table)==(Convert.term_functor_symbol term) then
		let l, r = Convert.term_pair term
		in
			Unit 
				(HeadLiteral.make_literal
					predicate_table
					(Eq (Equality.make_checked_equation term l r)))
	else		
		Unit (HeadLiteral.make_literal predicate_table (Term term))
  
  
let make_body predicate_table symbol_table eq_convert term =
	let equals = Term.get_equality_symbol symbol_table
	in
  Body (List.map
	  (fun lit -> 
	    BodyLiteral.make_literal predicate_table symbol_table eq_convert lit)
	  (purge_literals_list equals (rm_constant
	     (Term.get_true_symbol symbol_table)
	     (Convert.term_sequence_to_list
		(Term.get_conjunction_symbol symbol_table)
		term))))

let get_head_lits_terms head =
	match head with
	    False -> []
  		| Unit lit -> (HeadLiteral.get_atom_term lit) :: []
  		| Conjunction conj -> List.map HeadLiteral.get_atom_term conj
  		| Disjunction disj -> List.map HeadLiteral.get_atom_term (DisjunctiveHead.get_literals disj)

let get_body_lits_terms body =		
	match body with
		Body lits -> List.map BodyLiteral.get_term lits
		
		
let rec make_sigterm symbol_table l1 l2 l3 l4 l5 l6 l7 l8 =		
	let base = Term.Constant (Term.get_meta_symbol symbol_table 0)
	in
	let sig1 = make_sigterm_1 symbol_table l8 1 base
	in
	let sig2 = make_sigterm_1 symbol_table l7 2 sig1
	in
	let sig3 = make_sigterm_1 symbol_table l6 3 sig2
	in
	let sig4 = make_sigterm_1 symbol_table l5 4 sig3
	in
	let sig5 = make_sigterm_1 symbol_table l4 5 sig4
	in
	let sig6 = make_sigterm_1 symbol_table l3 6 sig5
	in
	let sig7 = make_sigterm_1 symbol_table l2 7 sig6
	in
	let sig8 = make_sigterm_1 symbol_table l1 8 sig7
	in		
	sig8
	
and make_sigterm_1 symbol_table terms i sig_base =
	if (List.length terms)==0 then sig_base
	else (
		Term.Compound
			(Term.make_compound
				(Term.get_meta_symbol symbol_table i)
				[|make_sigterm_2 symbol_table terms sig_base|]))
		
and make_sigterm_2 symbol_table term_list sig_base =
	match term_list with
		head :: [] ->
			let sym = Convert.term_functor_symbol head
			in		
				Term.Compound
					(Term.make_compound 
						(Term.find_symbol 
							symbol_table 
							((Term.symbol_name sym)^(string_of_int (Term.symbol_arity sym)))
							1) 
						[|sig_base|])
		|head :: tail ->
			let sym = Convert.term_functor_symbol head
			in
				Term.Compound
					(Term.make_compound 
						(Term.find_symbol 
							symbol_table 
							((Term.symbol_name sym)^(string_of_int (Term.symbol_arity sym)))
							1) 
						[|make_sigterm_2 symbol_table tail sig_base|])
		|_ ->			
			(raise (Err.Error ("Empty clause in Ic.make_sigterm.")))
	
let rec make_fuzzy_sigterm symbol_table l1 l2 l3 l4 =					
	let base = Term.Constant (Term.get_meta_symbol symbol_table 0)
	in			
	let sig1 = make_sigterm_1 symbol_table l4 2 base
	in
	let sig2 = make_sigterm_1 symbol_table l3 4 sig1
	in
	let sig3 = make_sigterm_1 symbol_table l2 6 sig2
	in
	let sig4 = make_sigterm_1 symbol_table l1 8 sig3
	in
		sig4

		
let get_functors_sum term =
	let cnt = ref (-1)
	in
		Term.iter_term
			(fun t ->
				match t with
					Term.Variable _ -> ()
					|_ -> cnt := !cnt+1)
			term;
		!cnt		
		
let rec make_clause aux term = 
  let cfg = aux.Aux.cfg
  and false_ref = ref false
  and predicate_table = aux.Aux.predicate_table
  and symbol_table = aux.Aux.symbol_table
  in 
  let equals = (Term.get_equality_symbol symbol_table)
  and eq_convert = not (Cfg.get_hyperextension_flag cfg)
  in
  if (Convert.term_functor_symbol term) == equals then (
  	let head = make_head predicate_table symbol_table eq_convert term
  	and id = gen_id ()
  	in
  	if not (Cfg.get_ehyper cfg) then (Cfg.set_ehyper cfg true);
  	let eq = List.hd (get_head_lits_terms head)
  	and is_oriented = 
  		match head with
  			Unit eq_lit -> Equality.get_is_oriented 
  							(match (HeadLiteral.get_atom eq_lit) with
  								Eq equation -> equation
  								|_ -> (raise (Err.Error "Bad equation unit in Ic.make_clause.")))
  			|_ -> (raise (Err.Error "Bad equation unit in Ic.make_clause."))
  	in
  		if Termutils.is_ground term then 
			  {id=id;
			   source=eq;
			   parent_source=Term.Unbound;
			   body=Body [];
			   head=head;
			   news=News.null_news;
			   subproofs=[||];
			   parent_subproofs=[||];
			   cfg=cfg;
			   aux=aux;
			   ground_head_terms=[||];
		       non_ground_head_terms=[||];
		       ground_body_terms=[||];
		       non_ground_body_terms=[||];
			   ground_head_eqs=[|eq|];
		       non_ground_head_eqs=[||];
		       ground_body_eqs=[||];
		       non_ground_body_eqs=[||];
		       head_terms_functors_sum=0;
    		   head_eqs_functors_sum=(get_functors_sum eq);
    		   body_terms_functors_sum=0;
    		   body_eqs_functors_sum=0;
		       is_ground=true;
		       is_eq_unit=true;
		       is_oriented=is_oriented;
		       has_non_eqs=false;
		       sigterm=eq;
		       fuzzy_sigterm=eq;
		       impure_varnums=[];
		       is_redundant=false_ref;
		       proof=Fact.UnknownProof;
		       is_in_current_stratum=false;
		       dom_clause=false
			 }
  		else (
			  {id=id;
			   source=eq;
			   parent_source=Term.Unbound;
			   body=Body [];
			   head=head;
			   news=News.null_news;
			   subproofs=[||];
			   parent_subproofs=[||];
			   cfg=cfg;
			   aux=aux;
			   ground_head_terms=[||];
		       non_ground_head_terms=[||];
		       ground_body_terms=[||];
		       non_ground_body_terms=[||];
		       ground_head_eqs=[||];
		       non_ground_head_eqs=[|eq|];
		       ground_body_eqs=[||];
		       non_ground_body_eqs=[||];
		       head_terms_functors_sum=0;
    		   head_eqs_functors_sum=(get_functors_sum eq);
    		   body_terms_functors_sum=0;
    		   body_eqs_functors_sum=0;		       
		       is_ground=false;
		       is_eq_unit=true;
		       is_oriented=is_oriented;
		       has_non_eqs=false;
		       sigterm=eq;
		       fuzzy_sigterm=eq;
		       impure_varnums=[];
		       is_redundant=false_ref;
		       proof=Fact.UnknownProof;
		       is_in_current_stratum=false;
		       dom_clause=false		       
			 }))
	else (
  (* Assume term has :-/2 as main functor. *)
  let head = 
    make_head predicate_table symbol_table eq_convert (Convert.term_nth_arg term 0)
  and body = 
    make_body predicate_table symbol_table eq_convert (Convert.term_nth_arg term 1)
  and id = gen_id ()
  in
  
	let head_lits = get_head_lits_terms head
	and body_lits = get_body_lits_terms body
	in
	
	if ((List.length body_lits)=0)
	&& ((List.length head_lits)=1)
	&& ((Convert.term_functor_symbol (List.nth head_lits 0))==equals)
		then
			(make_clause aux (List.nth head_lits 0))	(* this was a unit equation in disguise - trivial body *)
		else (
	
  	let ground_head_lits, non_ground_head_lits =
  			List.partition Termutils.is_ground head_lits
  	and ground_body_lits, non_ground_body_lits =
  			List.partition Termutils.is_ground body_lits
  	and equals = Term.get_equality_symbol symbol_table
  	in
  	let ground_head_eqs_list, ground_head_terms_list =
  			List.partition
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				ground_head_lits
  	and ground_body_eqs_list, ground_body_terms_list =
  			List.partition 
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				ground_body_lits
  	and non_ground_head_eqs_list, non_ground_head_terms_list =
  			List.partition 
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				non_ground_head_lits
  	and non_ground_body_eqs_list, non_ground_body_terms_list =
  			List.partition 
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				non_ground_body_lits
  	and head_eqs_list, head_terms_list =
  			List.partition
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				head_lits
  	and body_eqs_list, body_terms_list =
  			List.partition
  				(fun lit -> 
  					equals == (Convert.term_functor_symbol lit))
  				body_lits
  	in
	if not (Cfg.get_ehyper cfg) &&
	((List.length ground_head_eqs_list)>0
	||(List.length non_ground_head_eqs_list)>0 	
	||(List.length ground_body_eqs_list)>0 	
	||(List.length non_ground_body_eqs_list)>0)
	then (Cfg.set_ehyper cfg true); 
	let is_ground = 
		not 
			((List.length non_ground_head_terms_list)>0
   			||(List.length non_ground_head_eqs_list)>0
   			||(List.length non_ground_body_terms_list)>0
   			||(List.length non_ground_body_eqs_list)>0)

   	and has_non_eqs =	
   			(List.length head_terms_list)>0
   			||(List.length body_terms_list)>0
   	and impure_varnums = (
   		match head with
   			Disjunction disj -> 
   				(let krh_dom = Term.get_krh_dom_symbol aux.Aux.symbol_table
   				and krh_dom_varnums = ref []
   				and shared_head_varnums =
   					Termutils.get_shared_varnums 
  						(List.map 
  							HeadLiteral.get_atom_term 
  								(DisjunctiveHead.get_literals disj))
   				in
	   				List.iter
	   					(fun ngbt ->
	   						if (Convert.term_functor_symbol ngbt) == krh_dom then 
	   							(Termutils.iter_vars
	   								(fun v -> 
	   									krh_dom_varnums := (Term.variable_number v) :: !krh_dom_varnums)
	   								ngbt))
	   					non_ground_body_terms_list;
	   				krh_dom_varnums := Listutils.sort_without_duplicates !krh_dom_varnums;
	   				let impure_varnums = ref []
	   				in
		   				List.iter
		   					(fun vnum ->
		   						if (List.mem vnum !krh_dom_varnums) then 
		   							(impure_varnums := vnum :: !impure_varnums))
		   					shared_head_varnums;
		   				!impure_varnums)
   			|_ -> [])
	and ground_head_terms = (Array.of_list ground_head_terms_list);
	and non_ground_head_terms = (Array.of_list non_ground_head_terms_list);
	and ground_body_terms = (Array.of_list ground_body_terms_list);
	and non_ground_body_terms = (Array.of_list non_ground_body_terms_list);
	and ground_head_eqs = (Array.of_list ground_head_eqs_list);
	and non_ground_head_eqs = (Array.of_list non_ground_head_eqs_list);
	and ground_body_eqs = (Array.of_list ground_body_eqs_list);
	and non_ground_body_eqs = (Array.of_list non_ground_body_eqs_list);
   	in	
   	let head_terms_functors_sum =
   		let cnt = ref 0
   		in
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				ground_head_terms;
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				non_ground_head_terms;
   				!cnt
   	and head_eqs_functors_sum =
   		let cnt = ref 0
   		in
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				ground_head_eqs;
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				non_ground_head_eqs;
   				!cnt   		
   	and body_terms_functors_sum =
   		let cnt = ref 0
   		in
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				ground_body_terms;
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				non_ground_body_terms;
   				!cnt   	
   	and body_eqs_functors_sum =
   		let cnt = ref 0
   		in
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				ground_body_eqs;
   			Array.iter
   				(fun t ->
   					cnt := !cnt + (get_functors_sum t))
   				non_ground_body_eqs;
   				!cnt     									   				
   	and source_term =
   		(let head_terms_eqs =
   			Array.concat
   				[ground_head_terms;
   				ground_head_eqs;
   				non_ground_head_terms;
   				non_ground_head_eqs]
   		and body_terms_eqs =
   			Array.concat
   				[ground_body_terms;
   				ground_body_eqs;
   				non_ground_body_terms;
   				non_ground_body_eqs]
   		in
   		let head_term =
   			(match head with
				Disjunction _ -> 
					Convert.array_of_disj_term_to_term 
						symbol_table
						head_terms_eqs
				|False -> (Term.Constant(Term.get_false_symbol symbol_table))
				|_ ->	
					Convert.array_of_conj_term_to_term 
						symbol_table
						head_terms_eqs)
		and body_term =
			Convert.array_of_conj_term_to_term 
				symbol_table
				body_terms_eqs
		in
			Term.Compound
				(Term.make_compound
					(Term.get_rule_symbol symbol_table)
					[|head_term; body_term|]))		
	in   		
  {id=id;
   source=source_term;
   parent_source=Term.Unbound;
   body=body;
   head=head;
   news=News.null_news;
   subproofs=[||];
   parent_subproofs=[||];
   cfg=cfg;
   aux=aux;
   ground_head_terms = ground_head_terms;
   non_ground_head_terms = non_ground_head_terms;
   ground_body_terms = ground_body_terms;
   non_ground_body_terms = non_ground_body_terms;
   ground_head_eqs = ground_head_eqs;
   non_ground_head_eqs = non_ground_head_eqs;
   ground_body_eqs = ground_body_eqs;
   non_ground_body_eqs = non_ground_body_eqs;
   head_terms_functors_sum=head_terms_functors_sum;
   head_eqs_functors_sum=head_eqs_functors_sum;
   body_terms_functors_sum=body_terms_functors_sum;
   body_eqs_functors_sum=body_eqs_functors_sum;   
   is_ground = is_ground;
   is_eq_unit=false;
   is_oriented=false;
   has_non_eqs=has_non_eqs;
   sigterm = make_sigterm 
   				symbol_table
   				ground_head_terms_list
   				non_ground_head_terms_list
   				ground_head_eqs_list
  				non_ground_head_eqs_list
  				ground_body_terms_list
  				non_ground_body_terms_list
  				ground_body_eqs_list
  				non_ground_body_eqs_list;
  	fuzzy_sigterm = make_fuzzy_sigterm
  	   			symbol_table
   				head_terms_list
   				head_eqs_list
  				body_terms_list
  				body_eqs_list;
	impure_varnums = impure_varnums;
	is_redundant=false_ref;
	proof=Fact.UnknownProof;
	is_in_current_stratum=false;
	dom_clause=false
 })
   				) (* end of else sorting out unit equations *)


 
let body_literals clause =
  match clause.body with
    Body lits -> lits
    
(** Returns a "shallow" copy of the given clause: the body has copies
   of the original literals, which however may share terms with
   the original ones, see [make_literal_like].
 *)
let make_clause_like clause =
  { id="delta_"^clause.id;
    source=clause.source;
    parent_source=clause.parent_source;
    news=clause.news;
    body=Body (List.map 
		 BodyLiteral.make_literal_like
		 (body_literals clause));
    head=clause.head;  (* ### may also make copies if needed. *)
    subproofs=Array.copy clause.subproofs;
    parent_subproofs=Array.copy clause.parent_subproofs;
    cfg=clause.cfg;
    aux=clause.aux;
    ground_head_terms=clause.ground_head_terms; 
    non_ground_head_terms=clause.non_ground_head_terms;
    ground_body_terms=clause.ground_body_terms; 
    non_ground_body_terms=clause.non_ground_body_terms;
    ground_head_eqs=clause.ground_head_eqs; 
    non_ground_head_eqs=clause.non_ground_head_eqs;
    ground_body_eqs=clause.ground_body_eqs; 
    non_ground_body_eqs=clause.non_ground_body_eqs;
    head_terms_functors_sum=clause.head_terms_functors_sum;
    head_eqs_functors_sum=clause.head_eqs_functors_sum;
    body_terms_functors_sum=clause.body_terms_functors_sum;
    body_eqs_functors_sum=clause.body_eqs_functors_sum;    
    is_ground=clause.is_ground;
    is_eq_unit=clause.is_eq_unit;
    is_oriented=clause.is_oriented;
    has_non_eqs=clause.has_non_eqs;
    sigterm=clause.sigterm;
    fuzzy_sigterm=clause.fuzzy_sigterm;
    impure_varnums=clause.impure_varnums;
    is_redundant=clause.is_redundant;
    proof=clause.proof;
	is_in_current_stratum=clause.is_in_current_stratum;
	dom_clause=clause.dom_clause
  }
    
let clause_delta_clauses delta_predicates clause =
  let cs = ref []
  and i = ref 0 in 
  if clause.is_eq_unit then
  	(cs := clause :: !cs)
  else (
  	List.iter 
	    (fun lit ->
	      let pred =lit.BodyLiteral.predicate in
	      if List.memq pred delta_predicates || ((Predicate.extent_size pred)>0) then
		begin
		  Predicate.ensure_delta_ready_to_run pred;
		  let clause1 = make_clause_like clause in
		  (List.nth (body_literals clause1) !i).BodyLiteral.is_delta <- 
		    true;
		  cs := clause1 :: !cs
		end
	      else ();
	      i := !i + 1)
	    (body_literals clause));
  !cs
  
let clause_delta_clauses_unchecked clause =
  let cs = ref []
  and i = ref 0 in 
  if clause.is_eq_unit then
  	(cs := clause :: !cs)
  else (	  
  	List.iter 
	    (fun lit ->
	      let pred = lit.BodyLiteral.predicate in
	
		  	Predicate.ensure_delta_ready_to_run pred;
		  	let clause1 = make_clause_like clause in
		  		(List.nth (body_literals clause1) !i).BodyLiteral.is_delta <- true;
		  		cs := clause1 :: !cs;
				i := !i + 1)
	    (body_literals clause)
  );
  !cs  
  
  
(*******************************************************************************)
(*	Storage of redundant clauses and facts, used for backtracking	*)
(*******************************************************************************)    
  

let rs_id = ref 0

let gen_rs_id () =
  rs_id := !rs_id + 1;
  "rs_"^string_of_int(!rs_id)

let reset_gen_rs_id () =
  rs_id := 0  
  
type redundant_store = 
	{rs_id : string;
	mutable clauses : clause list;
	mutable facts : Fact.fact list}  
	
let make_redundant_store () =
	let i = gen_rs_id ()
	in
		{rs_id = i;
		clauses = [];
		facts = []}
		
let declare_fact_redundant redundant_store fact =
	Fact.declare_redundant fact;
	redundant_store.facts <- fact :: redundant_store.facts
	
let declare_clause_redundant redundant_store clause =
	declare_redundant clause;
	(*if (clause.id="clause_109") then (pe "";pe "";pe "WAH!";pe "";pe "";);*)
	redundant_store.clauses <- clause :: redundant_store.clauses
	 	  
let undeclare_redundancies redundant_store = 
	List.iter
		undeclare_redundant
		redundant_store.clauses;
	List.iter
		Fact.undeclare_redundant
		redundant_store.facts
		

  
(*******************************************************************************)
(*	Clause indexing	*)
(*******************************************************************************)    
  
module SubtermEntry =
struct
	type entry = 
		Clause of (clause * int * bool * Equality.place)
		|Fact of (Fact.fact * Equality.place)
	let is_ground entry =
		match entry with
			Clause (clause, _, _, _) -> clause.is_ground
			|Fact (fact, _) -> Fact.is_ground fact
	let get_term entry =
		match entry with
			Clause (clause, _, _, _) -> clause.source
			|Fact (fact, _) -> Fact.get_term fact	
	let get_place_term entry =
		match entry with
			Clause (_, _, _, place) -> Equality.get_place place
			|Fact (_, place) -> Equality.get_place place
	let is_redundant entry =
		match entry with
			Clause (clause, _, _, _) -> (is_redundant clause)
			|Fact (fact, _) -> Fact.is_redundant fact
	let declare_redundant redundant_store entry =
		match entry with
			Clause (clause, _, _, _) -> declare_clause_redundant redundant_store clause
			|Fact (fact, _) -> declare_fact_redundant redundant_store fact		
end

module NewClausesTable =
	DtIndex.Make(SubtermEntry)(DtIndex.HashTableImplementation)

module SubtermStore =
  struct

    module HashDtIndex =
      LayeredIndex.Make(DtIndex.Make(SubtermEntry)(DtIndex.HashTableImplementation))
                      
    module AlistDtIndex =
      LayeredIndex.Make(DtIndex.Make(SubtermEntry)(DtIndex.AlistTableImplementation))

     
    type store =
	B1 of HashDtIndex.index
      | B2 of AlistDtIndex.index


    type state =
	S1 of HashDtIndex.state
      | S2 of AlistDtIndex.state

	    
    let make_store indextype =
      match indextype with
	Opt.HashDt -> B1 (HashDtIndex.make_index ())
      | Opt.AlistDt -> B2 (AlistDtIndex.make_index ())
	| _ -> raise (Err.Error ("Wrong indextype for clause subterm indexing."))

    let insert store =
      match store with
	B1 store -> HashDtIndex.insert store
      | B2 store -> AlistDtIndex.insert store

	    
    let remove_variant store =
      match store with
	B1 store -> HashDtIndex.remove_variant store
      | B2 store -> AlistDtIndex.remove_variant store


    let iter_all f store =
      match store with
	B1 store -> HashDtIndex.iter_all f store
      | B2 store -> AlistDtIndex.iter_all f store


    let iter_instance_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_instance_candidates f store
      | B2 store -> AlistDtIndex.iter_instance_candidates f store

    let iter_variant_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_variant_candidates f store
      | B2 store -> AlistDtIndex.iter_variant_candidates f store

      	    
    let iter_generalization_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_generalization_candidates f store
      | B2 store -> AlistDtIndex.iter_generalization_candidates f store


    let iter_unification_candidates f store =
      match store with
	B1 store -> HashDtIndex.iter_unification_candidates f store
      | B2 store -> AlistDtIndex.iter_unification_candidates f store

    let iter_instance_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_instance_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_instance_candidates_ml f store

    let iter_variant_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_variant_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_variant_candidates_ml f store

      	    
    let iter_generalization_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_generalization_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_generalization_candidates_ml f store


    let iter_unification_candidates_ml f store =
      match store with
	B1 store -> HashDtIndex.iter_unification_candidates_ml f store
      | B2 store -> AlistDtIndex.iter_unification_candidates_ml f store      
      

    let clear store =
      match store with
	B1 store -> HashDtIndex.clear store
      | B2 store -> AlistDtIndex.clear store


    let size store =
      match store with
	B1 store -> HashDtIndex.size store
      | B2 store -> AlistDtIndex.size store


    let new_layer store =
      match store with
	B1 store -> S1 (HashDtIndex.new_layer store)
      | B2 store -> S2 (AlistDtIndex.new_layer store)


    let add_layer state = 
      match state with
	S1 state -> S1 (HashDtIndex.add_layer state)
      | S2 state -> S2 (AlistDtIndex.add_layer state)


    let obtain_state store =
      match store with
	B1 store -> S1 (HashDtIndex.obtain_state store)
      | B2 store -> S2 (AlistDtIndex.obtain_state store)


    let install_state state =
      match state with
	S1 state -> HashDtIndex.install_state state
      | S2 state -> AlistDtIndex.install_state state


  end

(****************************************************************************)
(** Heuristics *)
(****************************************************************************)  
  
type applicability_pattern = Opt.applicability_value array

let applicability_relative_value_fun indexlength position =
  let n = float (indexlength - position) in
  n ** 2.0 (* ### Just some formula such that early args are preferred. *)

let applicability_value indexlength position =
  let sum = ref 0.0 in
  for i = 0 to indexlength - 1 do
    sum := !sum +. (applicability_relative_value_fun indexlength i)
  done;
  let factor = 1.0 /. !sum in
  (* ### Must we ensure that the sum is really 1.0 ? *)
  (applicability_relative_value_fun indexlength position) *. factor

let mk_applicability_pattern arity indexargs =
   let ap = Array.make arity 0.0
   and indexlength = (Array.length indexargs) in
   for i = 0 to indexlength - 1 do
     ap.(indexargs.(i)) <- applicability_value indexlength i
   done;
   ap

let mk_array_up_to n =
  let a = Array.make n 0 in
  for i = 0 to n - 1 do
    a.(i) <- i
  done;
  a
   
let indexargs_applicability_pattern store arity indexargs =
    let args =
	match indexargs with
	  Opt.WholeTerm -> mk_array_up_to arity
	| Opt.Argument n -> [|n|]
	| Opt.Arguments ns -> ns in
      mk_applicability_pattern arity args


(****************************************************************************)
(** Writer - write access layer above Store. *)
(****************************************************************************)

module Writer =
  struct

    type writer = {
	indexargs: Opt.indexargs;
	store: SubtermStore.store;
	predicate_arity: int
      }

    let get_store writer =
      writer.store

    let make_writer indextype indexargs predicate_arity =

      let store = SubtermStore.make_store indextype in
      {indexargs=indexargs;
       store=store;
       predicate_arity=predicate_arity}

    let indexargs writer =
      writer.indexargs

    let store writer =
      writer.store

    let predicate_arity writer =
      writer.predicate_arity

    let insert writer term entry =
      SubtermStore.insert 
	writer.store 
	(Access.key_term writer.indexargs term)
	entry

    let remove_variant writer term =
      SubtermStore.remove_variant 
	writer.store 
	(Access.key_term writer.indexargs term)
	
    let clear writer =
      SubtermStore.clear writer.store

    let size writer =
      SubtermStore.size writer.store
  end


(****************************************************************************)
(* Reader - read access layer above Store. The same store can be			*)
(* associated with several readers, e.g. for different binding patterns.	*)
(****************************************************************************)

module Reader =
  struct

    type reader = {
	applicability_pattern: applicability_pattern;
	indexargs: Opt.indexargs;
	store: SubtermStore.store;
	predicate_arity: int
      }

     (* DEBUG *)
    let print_info reader =
      print_string "READER: ";
      match reader.indexargs with
	Opt.WholeTerm ->
	  print_string "WholeTerm"
      | _ ->
	  print_string "Selected"

    let get_indexargs reader =
      reader.indexargs

    let make_readers writer =
      let store = Writer.store writer
      and indexargs = Writer.indexargs writer
      and arity = Writer.predicate_arity writer in
      let ap =
	indexargs_applicability_pattern store arity indexargs in
      [{applicability_pattern=ap;
	indexargs=indexargs;
	store=store;
	predicate_arity=arity}]

    let null_reader_obj = {
      applicability_pattern=[||];
      indexargs=Opt.Arguments [||];
      store=SubtermStore.make_store Opt.AlistDt;
      predicate_arity=0
    }

    let null_reader () =
      null_reader_obj

    let applicability_pattern reader =
      reader.applicability_pattern

    let applicability reader binding_pattern =
      if reader.predicate_arity = 0 then
	1.0
      else
	let sum = ref 0.0 in
	for i = 0 to reader.predicate_arity - 1 do
	  match binding_pattern.(i) with
	    Opt.Bound ->
	      sum := !sum +. reader.applicability_pattern.(i)
	  | Opt.Free ->
	      ()
	done;
	!sum

    let iter_unification_candidates f reader term context =
      SubtermStore.iter_unification_candidates
	f 
	reader.store
	(Access.key_term reader.indexargs term)
	context
	
    let iter_instance_candidates f reader term context =
      SubtermStore.iter_instance_candidates
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context
	
    let iter_variant_candidates f reader term context =
      SubtermStore.iter_variant_candidates
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context	

    let iter_generalization_candidates f reader term context =
      SubtermStore.iter_generalization_candidates
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context
	
	let iter_unification_candidates_ml f reader term context =
      SubtermStore.iter_unification_candidates_ml
	f 
	reader.store
	(Access.key_term reader.indexargs term)
	context
	
    let iter_instance_candidates_ml f reader term context =
      SubtermStore.iter_instance_candidates_ml
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context
	
    let iter_variant_candidates_ml f reader term context =
      SubtermStore.iter_variant_candidates_ml
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context	

    let iter_generalization_candidates_ml f reader term context =
      SubtermStore.iter_generalization_candidates_ml
	f
	reader.store
	(Access.key_term reader.indexargs term)
	context

    let iter_all f reader =
      SubtermStore.iter_all
	f
	reader.store

  end



type subterm = 
    {symbol: Term.symbol;
    mutable is_delta: bool;
    mutable delta_subterm: subterm;  
	mutable writer: Writer.writer;
	mutable reader: Reader.reader}   

module SubtermTable = Hashtbl.Make(
  struct
    type t = Term.symbol
    let equal = (==)
    let hash = Hashtbl.hash
  end)	

type subterm_table = subterm SubtermTable.t

let make_subterm_table () =
  SubtermTable.create 16  

let default_indextype symbol = 
    Opt.AlistDt  
  
    
let make_delta_subterm main_subterm =
	let symbol = main_subterm.symbol
	in
  		let writer =
    		Writer.make_writer
      			(default_indextype symbol)
      			Opt.WholeTerm
      			(Term.symbol_arity symbol) in
  		let reader =
  			match (Reader.make_readers writer) with
  				r::_ -> r
  				| _ -> raise (Failure "Bad reader construction during delta_subterm indexing.")
  	 	in
			{symbol=symbol;
			is_delta=true;
			delta_subterm=main_subterm;
			reader=reader;
			writer=writer}    
    
let make_subterm symbol =
  let writer =
    Writer.make_writer
      (default_indextype symbol)
      Opt.WholeTerm
      (Term.symbol_arity symbol) in
  let reader =
  	match (Reader.make_readers writer) with
  		r::_ -> r
  		| _ -> raise (Failure "Bad reader construction during subterm indexing.")
  	 in
  let rec subterm =
    {symbol=symbol;
     is_delta=false;
     delta_subterm=subterm;
     reader=reader;
     writer=writer}
  in
  subterm.delta_subterm <- make_delta_subterm subterm;
  subterm  
  
  
let find_subterm subterm_table symbol =
  try 
    SubtermTable.find subterm_table symbol 
  with Not_found ->
    let subterm = make_subterm symbol in
    SubtermTable.add subterm_table symbol subterm;
    subterm


let rec index_clause_subterms subterm_table clause =
	if clause.is_eq_unit then (		
(* Equation unit clauses get indexed by the head, in the manner of units. This also means that they must be indexed
by their predicate, the equation symbol. Thus, the equation symbol is the only predicate to be stored as a subterm in
the subterm table. Storing equation unit clauses in the normal predicate extent might have been nicer, but due to
the module dependencies the predicate index cannot store Ic.clauses.
In addition, each equation unit is stored twice, once in its original orientation and once with its arguments
switched. The reason for this is that during the computations a non-oriented, non-ground equation may search the
index for a unification/generalization/etc. candidate, and if there was no bi-directional indexing, then one would
have to do two search runs to cover all possible candidates. As index searches are more frequent than index writings,
it is more efficient in this case to write twice.
The same is not true for non-equation-unit clauses. Here the number of writings would double for each
equation literal, and further be multiplied by the number of possible equation literal orderings.
Therefore these clauses are only indexed once, with their literals sorted by predicate precedence,
and a more complex lookup method with backtracking takes care of the commutativity of equation literals.   *)
		if clause.is_ground then (	
			index_predicate_as_subterm subterm_table clause.ground_head_eqs.(0) 0 clause;
			index_predicate_as_subterm subterm_table (Term.binary_flip (clause.ground_head_eqs.(0))) 0 clause;
			index_literal_eq subterm_table clause.ground_head_eqs.(0) clause 0)
		else (
			index_predicate_as_subterm subterm_table clause.non_ground_head_eqs.(0) 0 clause;
			index_predicate_as_subterm subterm_table (Term.binary_flip (clause.non_ground_head_eqs.(0))) 0 clause;
			index_literal_eq subterm_table clause.non_ground_head_eqs.(0) clause 0)
	)
	else (
		let i = ref 0
		in
		Array.iter
			(fun term ->
				let place = Equality.make_array_place term
				in
					index_literal_term subterm_table term clause !i true place;
					i := !i+1)
			clause.ground_body_terms;
		Array.iter
			(fun term ->
				let place = Equality.make_array_place term
				in
					index_literal_term subterm_table term clause !i true place;
					i := !i+1)
			clause.non_ground_body_terms;
		Array.iter
			(fun eq ->
				index_literal_eq subterm_table eq clause !i;
				i := !i+1)
			clause.ground_body_eqs;
		Array.iter
			(fun eq ->
				index_literal_eq subterm_table eq clause !i;
				i := !i+1)
			clause.non_ground_body_eqs;
		let sig_key = find_subterm subterm_table (Convert.term_functor_symbol clause.sigterm)
		and fuzzy_sig_key = find_subterm subterm_table (Convert.term_functor_symbol clause.fuzzy_sigterm)
		in
		Writer.insert 
			sig_key.writer 
			clause.sigterm 
			(SubtermEntry.Clause (clause, (-1), false, Equality.null_place));
		Writer.insert 
			sig_key.delta_subterm.writer 
			clause.sigterm 
			(SubtermEntry.Clause (clause, (-1), false, Equality.null_place));	
		Writer.insert 
			fuzzy_sig_key.writer 
			clause.fuzzy_sigterm 
			(SubtermEntry.Clause (clause, (-1), false, Equality.null_place));
		Writer.insert 
			fuzzy_sig_key.delta_subterm.writer 
			clause.fuzzy_sigterm 
			(SubtermEntry.Clause (clause, (-1), false, Equality.null_place));
			
		Array.iter
			(fun term ->
				add_subterms subterm_table term)
			clause.ground_head_terms;
		Array.iter
			(fun term ->
				add_subterms subterm_table term)
			clause.non_ground_head_terms;
		Array.iter
			(fun eq ->
				add_subterms subterm_table eq)
			clause.ground_head_eqs;
		Array.iter
			(fun eq ->
				add_subterms subterm_table eq)
			clause.non_ground_head_eqs;
			
		match clause.head with
			Unit lit ->
				(let term = (HeadLiteral.get_atom_term lit)
				in
				if true (*(Termutils.is_ground term)*) then	
					(let key_term =
						Term.Compound
							(Term.make_compound
								(Term.get_meta_symbol clause.aux.Aux.symbol_table 1)
								[|term|])
					in
						index_predicate_as_subterm subterm_table key_term (-1) clause)
				else ())	
			|Disjunction disj ->
				(List.iter 
					(fun lit ->
						(let term = (HeadLiteral.get_atom_term lit)
							in
							if true (*(Termutils.is_ground term)*) then	
								(let key_term =
									Term.Compound
										(Term.make_compound
											(Term.get_meta_symbol clause.aux.Aux.symbol_table 1)
											[|term|])
								in
									index_predicate_as_subterm subterm_table key_term (-1) clause)
							else ()))	
					(DisjunctiveHead.get_literals disj))
			|_ -> ()


	)
	
and add_subterms subterm_table term =
	match term with
		Term.Variable v -> ()
		|Term.Compound k -> 
			Array.iter
				(fun arg ->
					add_subterms_1 subterm_table arg)
				(Term.compound_args k)
		|_ -> ()
		
and add_subterms_1 subterm_table term =
	match term with
		Term.Variable v -> ()
		|_ ->
			ignore (find_subterm subterm_table (Convert.term_functor_symbol term));
			match term with
				Term.Compound k ->
			 		Array.iter
						(fun arg ->
							add_subterms_1 subterm_table arg)
						(Term.compound_args k)
				|_ -> ()
		
and index_predicate_as_subterm subterm_table term indexed_at clause =
	let subterm = find_subterm subterm_table (Convert.term_functor_symbol term)
	and place = Equality.make_array_place term
	in
		Writer.insert subterm.writer term (SubtermEntry.Clause (clause, indexed_at, false, place));
		Writer.insert subterm.delta_subterm.writer term (SubtermEntry.Clause (clause, indexed_at, false, place))
		
and index_literal_term subterm_table key_t clause i left_is_indexed place = 
	match key_t with
		Term.Variable v -> ()
		|Term.Unbound -> ()
		|_ ->
			let key = find_subterm subterm_table (Convert.term_functor_symbol key_t)
			in
				Writer.insert key.writer key_t (SubtermEntry.Clause (clause, i, left_is_indexed, place));
				Writer.insert key.delta_subterm.writer key_t (SubtermEntry.Clause (clause, i, left_is_indexed, place));
				match key_t with
					Term.Compound k ->
						 Array.iteri
						 	(fun j arg_term ->
						 		let arg_place = Equality.ArrayPlace ((Term.compound_args k), j)
						 		in
						 			index_literal_term subterm_table arg_term clause i left_is_indexed arg_place)
						 	(Term.compound_args k)
					|_ -> ()
			
and index_literal_eq subterm_table key_eq clause i =
	let key = find_subterm subterm_table (Convert.term_functor_symbol key_eq)
	and l, r = Convert.term_pair key_eq
	in
	let place = Equality.make_array_place key_eq
	and l_place = Equality.ArrayPlace ((Term.get_compound_args key_eq), 0)
	and r_place = Equality.ArrayPlace ((Term.get_compound_args key_eq), 1)
	in
		Writer.insert key.writer key_eq (SubtermEntry.Clause (clause, i, true, place));
		Writer.insert key.delta_subterm.writer key_eq (SubtermEntry.Clause (clause, i, true, place));
		index_literal_term subterm_table l clause i true l_place;	
		index_literal_term subterm_table r clause i false r_place	
	
		
let rec index_unit_subterms_no_top subterm_table key_term value_term =
	match key_term with
		Term.Variable v -> () (* No indexing on variables. *)
		|Term.Unbound -> ()
		|_ ->
			match key_term with
				Term.Compound k ->
					Array.iteri
						(fun i arg_term ->
							let arg_place = Equality.ArrayPlace ((Term.compound_args k), i)
							in 
								index_unit_subterms subterm_table arg_term value_term arg_place)
						(Term.compound_args k)
				|_ -> ()
					
					
and index_unit_subterms subterm_table key_term value_term place =
	match key_term with
		Term.Variable v -> () (* No indexing on variables. *)
		|Term.Unbound -> ()
		|_ ->
			let key = find_subterm subterm_table (Convert.term_functor_symbol key_term)
			in 
				Writer.insert key.writer key_term (SubtermEntry.Fact(value_term, place));
				Writer.insert key.delta_subterm.writer key_term (SubtermEntry.Fact(value_term, place));
				match key_term with
					Term.Compound k ->
						Array.iteri
							(fun i arg_term ->
								let arg_place = Equality.ArrayPlace ((Term.compound_args k), i)
								in 
									index_unit_subterms subterm_table arg_term value_term arg_place)
							(Term.compound_args k)
					|_ -> ()		
											
				
					
let get_subterm_reader subterm = subterm.reader

let get_subterm_writer subterm = subterm.writer			
			
let get_delta_subterm subterm = subterm.delta_subterm

let get_main_subterm subterm = subterm.delta_subterm
	
let has_delta_subterm subterm = subterm.delta_subterm != subterm

let get_delta_subterm subterm = subterm.delta_subterm

let clear subterm = Writer.clear subterm.writer
      
let clear_delta_subterm subterm =
   	clear subterm.delta_subterm

			
let ensure_delta_ready_to_run subterm =
	if has_delta_subterm subterm then
    	()
  	else
    	let delta_subterm = make_delta_subterm subterm 
    	in
    	    subterm.delta_subterm <- delta_subterm		
    	    
    
(****************************************************************************)
(* Subgoal Sorting. *)
(****************************************************************************)

(* For now this is greedy, i.e. no backtracking ("query planning")
   involved. 
*)
    
let lit_arity lit =
  Term.symbol_arity (Predicate.get_symbol lit.BodyLiteral.predicate)
    
let lit_nth_arg lit n =
  Convert.term_nth_arg (BodyLiteral.get_atom_term lit) n
    
let mk_binding_pattern bound_vars lit =
  let arity = lit_arity lit in
  let bp = Array.make arity Opt.Free in
  for i = 0 to arity - 1 do
    let arg = (lit_nth_arg lit i) in
    if
      Termutils.is_nonvar arg ||
      List.mem (Termutils.varnum arg) bound_vars then
      bp.(i) <- Opt.Bound
    else
      ()
  done;
  bp
    
let permissible_by_modespec bound_vars modespec disc_hints lit =
  let arity = lit_arity lit in
  for i = 0 to arity - 1 do
    if modespec.(i) = Opt.InputGround then
      Termutils.iter_vars
	(fun v -> 
	  if not (List.mem (Term.variable_number v) bound_vars) then
	    raise Not_found
	  else
	    ())
	(lit_nth_arg lit i)
  done;
  let bp = (mk_binding_pattern bound_vars lit) in
  (1.0, 
   Opt.compute_discrimination_hint bp disc_hints,
   bp)
    
let permissible bound_vars lit =
  (* Returns (applicability_hint, discrimination_hint). Raises
     Not_found if the literal is not applicable at all with the
     given binding pattern. *)
  match lit.BodyLiteral.builtin with
    Term.PDeterministic (_, (mode, discs, _)) ->
      permissible_by_modespec bound_vars mode discs lit
  | Term.PUnbound ->
      let bp = mk_binding_pattern bound_vars lit
      and pred = lit.BodyLiteral.predicate in
      (Predicate.applicability pred bp,
       Opt.compute_discrimination_hint
	 bp (Predicate.get_discrimination_hints pred),
       bp)
	
let rec split_permissibles bound_vars lits y n =
  match lits with
    [] ->
      (y, n)
  | lit :: lits1 ->
      try
	let adb = permissible bound_vars lit in
	split_permissibles bound_vars lits1 ((lit, adb)::y) n
      with Not_found ->
	split_permissibles bound_vars lits1 y (lit::n)
	  

(****************************************************************************)

let compare_permissibles_by_to_end (l1, _) (l2, _) =
  if l1.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
    if l2.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
      0
    else
      1
  else if l2.BodyLiteral.sign = StratifiedNegAsFailureNotSubsumed then
    -1
  else
    0
      	  
let compare_permissibles_by_delta (l1, _) (l2, _) =
  if l1.BodyLiteral.is_delta then
    if l2.BodyLiteral.is_delta then
      0
    else
      -1
  else if l2.BodyLiteral.is_delta then
    1
  else
    0
      
let compare_permissibles_by_applicability_hint
    (_, (a1, _, _)) (_, (a2, _, _)) =
  compare a2 a1
    
let compare_permissibles_by_discrimination_hint 
    (_, (_, d1, _)) (_, (_, d2, _)) =
  compare d2 d1
    
let compare_permissibles_by_extent_size (l1, _) (l2, _) =
  (* ### 0 is preferred, but it effectively means "unknown", i.e.
     the effect is that IDB predicates are preferred. 
     Maybe change this. *)
  compare 
    (Predicate.extent_size l1.BodyLiteral.predicate)
    (Predicate.extent_size l2.BodyLiteral.predicate)
    
let compare_permissibles_somehow (l1, _) (l2, _) =
  (* Used to be independent on input ordering. *)
  compare 
    (l1.BodyLiteral.sign, BodyLiteral.get_atom_term l1)
    (l2.BodyLiteral.sign, BodyLiteral.get_atom_term l2)
    
let compare_permissibles lh1 lh2 =
  let r = compare_permissibles_by_to_end lh1 lh2 in
  if r <> 0 then r
  else
    let r = compare_permissibles_by_delta lh1 lh2 in
    if r <> 0 then r
    else
      let r = compare_permissibles_by_applicability_hint lh1 lh2 in
      if r <> 0 then r
      else
	let r = compare_permissibles_by_discrimination_hint lh1 lh2 in
	if r <> 0 then r
	else
	  let r = compare_permissibles_by_extent_size lh1 lh2 in
	  if r <> 0 then r
	  else
	    compare_permissibles_somehow lh1 lh2
	    
let choose_from_permissibles lhs =
  match List.sort compare_permissibles lhs with
    (l, (a, d, b)) :: rest ->
      (l, List.map fst rest, (a, d, b))
  | _ -> raise Not_found

let add_bound_vars bound_vars lit =
  List.append lit.BodyLiteral.output_vars bound_vars

let assign_optim_info lit (a, d, b) =
  lit.BodyLiteral.optim_info <-
    BodyLiteral.make_optim_info
      a d b (Predicate.extent_size lit.BodyLiteral.predicate)

(****************************************************************************)

(* Heuristics: if the greedy best next literal has application-hint and
   discrimation-hint of 0.0 - i.e. no index can be applied, the literal
   is chosen in a different way: that one which has the best successor
   literal if it would be the first is chosen.
*)

let is_poor_choice lit (applic_hint, discrim_hint, _) =
   (not lit.BodyLiteral.is_delta) &&
   applic_hint = 0.0 &&
   discrim_hint = 0.0

let rec map_split bef es  =
  match es with
    [] -> []
  | e::es1 ->
      (e, List.append bef es1) :: (map_split (e::bef) es1)

let choose_from_permissibles_by_second bound_vars lits =
  let lps = map_split [] lits in
  let lhs =
    List.map
      (fun (lit, lits1) ->
	let bound_vars_1 =
	  add_bound_vars bound_vars lit in
	let (next_permissibles, _) =
	  split_permissibles bound_vars_1 lits1 [] [] in
	match List.sort compare_permissibles next_permissibles with
	  (lit1, (a, d, b)) :: _ ->
	    (* Return this literal with values of the second one, obtained
	       by assuming that this literal is the first one. *)
	    (lit, (a, d, b))
	| _ ->
	    raise Not_found)
      lps in
  choose_from_permissibles lhs

(****************************************************************************)

let choose_permissible bound_vars lits =
  (* Throws Not_found. *)
  let (permissibles, rest1) =
    split_permissibles bound_vars lits [] [] in
  let (lit, rest2, h) =
    choose_from_permissibles permissibles in
  if is_poor_choice lit h && (List.length permissibles) > 1 then
    let (lit, rest2, h) =
      choose_from_permissibles_by_second
	bound_vars (List.map fst permissibles) in
    (lit, List.append rest1 rest2)
  else
    begin
      assign_optim_info lit h;
      (lit, List.append rest1 rest2)
    end


(****************************************************************************)

let rec select_delta deltaref lits =
  match lits with
    [] -> raise Not_found
  | lit :: lits1 ->
      if lit.BodyLiteral.is_delta then
	begin
	  deltaref := lit;
	  lits1
	end
      else lit :: select_delta deltaref lits1

let delta_to_front lits =
  match lits with
    [] -> []
  | lit :: _ ->
      let deltaref = ref lit in
	try
	  let rest = select_delta deltaref lits in
	  !deltaref :: rest
	with Not_found ->
	  lits


(****************************************************************************)
    
(** Raises Not_found if no permissible ordering can be found. *)

let rec sort_body_lits bound_vars lits =
  match lits with
    [] -> []
  | _ ->
      let (lit, lits1) = choose_permissible bound_vars lits in
      lit :: sort_body_lits 
	      (add_bound_vars bound_vars lit)
	      lits1
	  
(****************************************************************************)
	  
let rec assign_foreign_contexts n lits =
  match lits with
    [] ->
      ()
  | lit :: lits1 ->
      lit.BodyLiteral.foreign_context <- Term.make_context n;
      assign_foreign_contexts (n + 1) lits1
	
let rec assign_readers bound_vars lits =
  match lits with
    [] ->
      ()
  | lit :: lits1 ->
      begin
	match lit.BodyLiteral.builtin with
	  Term.PUnbound ->
	    let bp = mk_binding_pattern bound_vars lit
	    and pred = lit.BodyLiteral.predicate in
	    lit.BodyLiteral.reader <- 
	      Predicate.choose_best_reader
		(if lit.BodyLiteral.is_delta then
		  (Predicate.get_delta_predicate pred)
		else
		  pred)
		bp	  
	| _ ->
	    ()
      end;
      assign_readers
	(add_bound_vars bound_vars lit)
	lits1

let assign_subproof_stuff clause = 
  if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
    (match clause.body with
      Body lits ->
	clause.subproofs <- Fact.make_subproofs (List.length lits);
	let i = ref 0 in
	List.iter
	  (fun lit -> BodyLiteral.set_subproof_pos lit !i; i := !i+1)
	  lits)
  else
    ()
	    
let rec initialize_clause clause news = 
  clause.news <- news;		
  let bound_vars = [] in
  let lits = body_literals clause in
  let lits1 =
    match Cfg.get_body_sorting_method_param clause.cfg
    with
      0 ->
	(try
	  sort_body_lits bound_vars lits
	with Not_found ->
	  raise (Err.Error 
		   ("No permissible ordering for clause: "^
		    PrologPrint.print_to_string clause.source)))
    | 1 ->
	delta_to_front lits
    | 2 ->
	lits 
    | _ ->
	raise (Err.Error "Bad value for body_sorting_parameter.")
  in
  clause.body <- Body lits1;
  assign_readers bound_vars lits1;
  assign_foreign_contexts 1 lits1;
  assign_subproof_stuff clause


(** Restore the state of a clause, e.g. passed through a [BranchClosed]
   exception, such that it can be used for evaluation again. *)    
let refresh_clause clause =
  assign_foreign_contexts 1 (body_literals clause);
  assign_subproof_stuff clause


(****************************************************************************)
(* Runtime. *) 
(****************************************************************************)

exception BranchClosed of Fact.proof


				
(****************************************************************************)
(* Miscellaneous clause related functions required by the E-Hyper calculus *)
(****************************************************************************)

  	
let rec match_clauses clause_1 clause_2 = 
	match clause_1.head, clause_2.head with 
		False, False -> match_clauses_1 clause_1 clause_2
		| Unit _, Unit _ -> match_clauses_1 clause_1 clause_2
		| Conjunction _, Conjunction _ -> match_clauses_1 clause_1 clause_2
		| Disjunction _, Disjunction _ -> match_clauses_1 clause_1 clause_2
		| _, _ -> false
		
and match_clauses_1 clause_1 clause_2 = 
	let check = ref true
	and equals = Term.get_equality_symbol (clause_1.aux.Aux.symbol_table)
	in 
	if (Array.length clause_1.ground_head_terms)==(Array.length clause_2.ground_head_terms)
		&& (Array.length clause_1.ground_body_terms)==(Array.length clause_2.ground_body_terms)
		&& (Array.length clause_1.non_ground_head_terms)==(Array.length clause_2.non_ground_head_terms)
		&& (Array.length clause_1.non_ground_body_terms)==(Array.length clause_2.non_ground_body_terms)
		&& (Array.length clause_1.ground_head_eqs)==(Array.length clause_2.ground_head_eqs)
		&& (Array.length clause_1.ground_body_eqs)==(Array.length clause_2.ground_body_eqs)
		&& (Array.length clause_1.non_ground_head_eqs)==(Array.length clause_2.non_ground_head_eqs)
		&& (Array.length clause_1.non_ground_body_eqs)==(Array.length clause_2.non_ground_body_eqs)
		then (
			(try
				(Array.iteri
					(fun i x ->
						if (Term.term_ident x clause_2.ground_head_terms.(i)) then ()
						else raise Failed)
					clause_1.ground_head_terms;
				Array.iteri
					(fun i x ->
						if (Term.term_ident x clause_2.ground_body_terms.(i)) then ()
						else raise Failed)
					clause_1.ground_body_terms;
				Array.iteri
					(fun i x ->
						if (Term.term_ident x clause_2.ground_head_eqs.(i)) then ()
						else raise Failed)
					clause_1.ground_head_eqs;
				Array.iteri
					(fun i x ->
						if (Term.term_ident x clause_2.ground_body_eqs.(i)) then ()
						else raise Failed)
					clause_1.ground_body_eqs)
			with Failed -> check := false); 
			!check &&
			match_non_ground_literals
				equals
				(Array.append clause_1.non_ground_head_terms clause_1.non_ground_head_eqs)
				(Array.append clause_2.non_ground_head_terms clause_2.non_ground_head_eqs)
				(Array.append clause_1.non_ground_body_terms clause_1.non_ground_body_eqs)
				(Array.append clause_2.non_ground_body_terms clause_2.non_ground_body_eqs)
		)
	else false

(*	
and match_non_ground_literal_terms h1 h2 b1 b2 = 
	let var1_to_b = Array.make Term.max_vars (-1)
	and var2_to_b = Array.make Term.max_vars (-1)
	and var1_by_b = Array.make Term.max_vars (-1)
	and var2_by_b = Array.make Term.max_vars (-1)
	and offset = Array.length h1
	in
	let h1_b = Array.make offset (-1)
	and h2_b = Array.make offset (-1)
	and b1_b = Array.make (Array.length b1) (-1)
	and b2_b = Array.make (Array.length b2) (-1)
  	in
		match_seek_terms (Array.append h1 b1) 0 0 offset h2 b2 var1_to_b var2_to_b var1_by_b var2_by_b (Array.append h1_b b1_b) h2_b  b2_b

				
and match_non_ground_literal_eqs h1 h2 b1 b2 = 
	let var1_to_b = Array.make Term.max_vars (-1)
	and var2_to_b = Array.make Term.max_vars (-1)
	and var1_by_b = Array.make Term.max_vars (-1)
	and var2_by_b = Array.make Term.max_vars (-1)
	and offset = Array.length h1
	in 
	let h1_b = Array.make offset (-1)
	and h2_b = Array.make offset (-1)
	and b1_b = Array.make (Array.length b1) (-1)
	and b2_b = Array.make (Array.length b2) (-1)
	and not_flipped = Array.make (offset + (Array.length b1)) true
  	in 
		match_seek_eqs (Array.append h1 b1) 0 0 offset h2 b2 var1_to_b var2_to_b var1_by_b var2_by_b (Array.append h1_b b1_b) h2_b  b2_b not_flipped
*)

and match_non_ground_literals equals h1 h2 b1 b2 = 
	let var1_to_b = Array.make Term.max_vars (-1)
	and var2_to_b = Array.make Term.max_vars (-1)
	and var1_by_b = Array.make Term.max_vars (-1)
	and var2_by_b = Array.make Term.max_vars (-1)
	and offset = Array.length h1
	in 
	let h1_b = Array.make offset (-1)
	and h2_b = Array.make offset (-1)
	and b1_b = Array.make (Array.length b1) (-1)
	and b2_b = Array.make (Array.length b2) (-1)
	and not_flipped = Array.make (offset + (Array.length b1)) true
  	in 
		match_seek equals (Array.append h1 b1) 0 0 offset h2 b2 var1_to_b var2_to_b var1_by_b var2_by_b (Array.append h1_b b1_b) h2_b  b2_b not_flipped


and match_seek equals a1 index seek_index offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped =
	let seek_index = ref seek_index
	in 
	if index>=offset then (	(*Seeking in the body now, lots of offsets required. *) 
		if (Array.length a2b)>0 then (
			while !seek_index<(Array.length a1) && a1_b.(index)<0 do (* still seekspace in body && current a1 not matched yet *) (*MAINWHILE*)
			
				if a2b_b.(!seek_index-offset)<0 then (	(* observed bodylit unbound *)
					if (Term.variant_chk_bindings a1.(index) a2b.(!seek_index-offset) index var1_to_b var2_to_b var1_by_b var2_by_b)	(* observed bodylit matches*)
						then (
							a1_b.(index) <- !seek_index;
							a2b_b.(!seek_index-offset) <- index)
					else (
						if (is_eq equals a1.(index))
							&&(not_flipped.(index))
							&&(Term.variant_chk_bindings (Term.binary_flip a1.(index)) a2b.(!seek_index-offset) index var1_to_b var2_to_b var1_by_b var2_by_b)
							then (a1_b.(index) <- !seek_index;
							a2b_b.(!seek_index-offset) <- index;
							not_flipped.(index) <- true)
						else (	(*observed bodylit fails match *)
							seek_index := !seek_index+1
						)
					)
				)
				else (	(*observed bodylit is bound *)
					seek_index := !seek_index+1
				)
			done;
			if a1_b.(index)>(-1) then (	(*current has found match *)
				if index<((Array.length a1)-1) then (	(* still lits to match, go on in a1 *)
					match_seek equals a1 (index+1) offset offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped
				)
				else (true)		(* all done *)
			)
			else (	(* Begin heavy messing. *)
				if index=offset then false	(* should never happen, would require headless rule... *)
				else (
					let prev_index = index-1
					in
					let previous = a1.(prev_index)
					and prev_seek_index = a1_b.(prev_index)
					in
						Termutils.iter_vars
							(fun v ->
								let vnum = Term.variable_number v
								in
									if var1_by_b.(vnum) == prev_index then (
										var1_by_b.(vnum)<-(-1);
										var2_to_b.(var1_to_b.(vnum))<-(-1);
										var2_by_b.(var1_to_b.(vnum))<-(-1);
										var1_to_b.(vnum)<-(-1)
									)
									else ()
							)
							previous;
						if prev_index<offset then (		(* crossing the head/body border... *)
							a2h_b.(prev_seek_index) <- (-1);	
						)
						else (
							a2b_b.(prev_seek_index-offset) <- (-1)
						);

						a1_b.(prev_index) <- (-1);
						(* If the equation in the undone binding had not yet been flipped, there is a chance that its
						flipped version matches the same literal (when all four terms of both equations are variants
						of each other), but which will not cause the problem that initiated the backtracking. Thus,
						check whether a1.(prev_index) has been flipped, and if not, then try to match its flipped
						version to the same literal as before. If this works, continue with matchseek for a1.(index)
						from there. If not, seek a different matching literal for a1.(prev_index).*)
						if (is_eq equals previous)
							&&(not_flipped.(prev_index))
							&&(Term.variant_chk_bindings (Term.binary_flip previous) a2b.(prev_seek_index - offset) prev_index var1_to_b var2_to_b var1_by_b var2_by_b)
							then
								(a1_b.(prev_index) <- prev_seek_index; (* bind observed headlit to current*)
				 				a2b_b.(prev_seek_index-offset) <- prev_index; (* bind current to observed headlit*)
				 				not_flipped.(prev_index) <- false;
				 				match_seek equals a1 (prev_index + 1) (prev_seek_index + 1) offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			 			else 
							(not_flipped.(prev_index) <- true;
							match_seek equals a1 prev_index (prev_seek_index+1) offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
				)
			)
		)
		else true
	)
	else ( 
		while !seek_index<offset && a1_b.(index)<0 do (*still seekspace in head && current a1 not matched yet*)
			if a2h_b.(!seek_index)<0 then (	(* observed headlit unbound *)
				if (Term.variant_chk_bindings a1.(index) a2h.(!seek_index) index var1_to_b var2_to_b var1_by_b var2_by_b)
					then (	 (* observed headlit matches*)
				 		a1_b.(index) <- !seek_index; (* bind observed headlit to current*)
				 		a2h_b.(!seek_index) <- index (* bind current to observed headlit*)
				 	)
				 else (
				 	if (is_eq equals a1.(index))
				 		&&(not_flipped.(index)) 
				 		&&(Term.variant_chk_bindings (Term.binary_flip a1.(index)) a2h.(!seek_index) index var1_to_b var2_to_b var1_by_b var2_by_b)
				 		then (
				 			a1_b.(index) <- !seek_index; (* bind observed headlit to current*)
				 			a2h_b.(!seek_index) <- index; (* bind current to observed headlit*)
				 			not_flipped.(index) <- false
				 	)
				 	else (	(* observed headlit fails match *)
				 		seek_index := !seek_index+1	
					)
				 )
			)
			else (	(*observed headlit is bound *)
				seek_index := !seek_index+1	
			)
		done;	(* current has found match or there exists no match *)
		if a1_b.(index)>(-1) then ( (* current has found match *)
			if index<((Array.length a1)-1) then (	(* still literals waiting for match, proceed to next a1*)
				if index<(offset-1) then
					(match_seek equals a1 (index+1) 0 offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
				else
					(match_seek equals a1 (index+1) offset offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			)
			else true			(* All literals have matched, success! *)
		)
		else (	(* start backtracking *)
			if index=0 then false	(* easy: no match for first a1 -> no backtrack possible, all fails *)
			else (	(* So there was a previous a1, and something went wrong there. As current found no match,
					current has not modified any bindings. The last modification was by the previous a1.
					The plan: get index of previous a1, undo its bindings, and restart match_seek for previous a1
					but with increased seek_index from its last position *)
				let prev_index = index-1 
				in
				let previous = a1.(prev_index)
				and prev_seek_index = a1_b.(prev_index)
				in
					Termutils.iter_vars
						(fun v ->
							let vnum = Term.variable_number v
							in
								if var1_by_b.(vnum) == prev_index then (	(*this variable was bound by this literal*)
									var1_by_b.(vnum)<-(-1);		(*this variable is no longer bound by this literal *)
									var2_to_b.(var1_to_b.(vnum))<-(-1);	(*the variable bound to this variable is no longer bound to this variable *)
									var2_by_b.(var1_to_b.(vnum))<-(-1); (*the variable bound to this variable is no longer bound by this literal *)
									var1_to_b.(vnum)<-(-1)	(* this variable is no longer bound to another variable *)
								)
								else ()	(* this variable was bound by an earlier literal, it may not be set free by this literal *)
						)
						previous;	(* The variable bindings of previous have been undone. *)
					a2h_b.(prev_seek_index) <- (-1);
					a1_b.(prev_index) <- (-1); (* The literal bindings of previous have been undone. *)
					if (is_eq equals previous)
						&&(not_flipped.(prev_index)) 
						&&(Term.variant_chk_bindings (Term.binary_flip previous) a2h.(prev_seek_index) prev_index var1_to_b var2_to_b var1_by_b var2_by_b) 
						then (
			 			a1_b.(prev_index) <- prev_seek_index; (* bind observed headlit to current*)
			 			a2h_b.(prev_seek_index) <- prev_index; (* bind current to observed headlit*)
			 			not_flipped.(prev_index) <- false;
			 			match_seek equals a1 (prev_index + 1) (prev_seek_index + 1) offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped
			 		)
				 	else	(* flipped term did not match, unflip and test with the next literals *)
						(not_flipped.(prev_index) <- true;
						match_seek equals a1 prev_index (prev_seek_index + 1) offset a2h a2b var1_to_b var2_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			)
		)
	)	

	
let unchecked_match_clauses clause_1 clause_2 = 
	let check = ref true
	in
		(try
			(Array.iteri
				(fun i x ->
					if (Term.term_ident x clause_2.ground_head_terms.(i)) then ()
					else raise Failed)
				clause_1.ground_head_terms;
			Array.iteri
				(fun i x ->
					if (Term.term_ident x clause_2.ground_body_terms.(i)) then ()
					else raise Failed)
				clause_1.ground_body_terms;
			Array.iteri
				(fun i x ->
					if (Term.term_ident x clause_2.ground_head_eqs.(i)) then ()
					else raise Failed)
				clause_1.ground_head_eqs;
			Array.iteri
				(fun i x ->
					if (Term.term_ident x clause_2.ground_body_eqs.(i)) then ()
					else raise Failed)
				clause_1.ground_body_eqs)
		with Failed -> check := false);
		let equals =
			(Term.get_equality_symbol clause_1.aux.Aux.symbol_table)
		in
			!check &&
			match_non_ground_literals
				equals
				(Array.append clause_1.non_ground_head_terms clause_1.non_ground_head_eqs)
				(Array.append clause_2.non_ground_head_terms clause_2.non_ground_head_eqs)
				(Array.append clause_1.non_ground_body_terms clause_1.non_ground_body_eqs)
				(Array.append clause_2.non_ground_body_terms clause_2.non_ground_body_eqs)			
		
	
let rec compare_subclauses_lengths clause_1 clause_2 =
	match clause_1.head, clause_2.head with
		False, False -> compare_subclauses_lengths_1 clause_1 clause_2
		| False, _ -> (-1)
		| _, False -> 1
		| Unit _, Unit _ -> compare_subclauses_lengths_1 clause_1 clause_2
		| Unit _, _ -> (-1)
		| _, Unit _ -> 1
		| Conjunction _, Conjunction _ -> compare_subclauses_lengths_1 clause_1 clause_2
		| Conjunction _, _ -> (-1)
		| _, Conjunction _ -> 1
		| Disjunction _, Disjunction _ -> compare_subclauses_lengths_1 clause_1 clause_2
		(*| Disjunction _, _ -> (-1)*)
		(*| _, Disjunction _ -> 1*)

and compare_subclauses_lengths_1 clause_1 clause_2 =		
	match compare (Array.length clause_1.ground_head_terms) (Array.length clause_2.ground_head_terms) with
		0 ->
	(match compare (Array.length clause_1.ground_body_terms) (Array.length clause_2.ground_body_terms) with
		0 ->
	(match compare (Array.length clause_1.non_ground_head_terms) (Array.length clause_2.non_ground_head_terms) with
		0 ->
	(match compare (Array.length clause_1.non_ground_body_terms) (Array.length clause_2.non_ground_body_terms) with
		0 ->
	(match compare (Array.length clause_1.ground_head_eqs) (Array.length clause_2.ground_head_eqs) with
		0 ->
	(match compare (Array.length clause_1.ground_body_eqs) (Array.length clause_2.ground_body_eqs) with
		0 ->
	(match compare (Array.length clause_1.non_ground_head_eqs) (Array.length clause_2.non_ground_head_eqs) with
		0 -> compare (Array.length clause_1.non_ground_body_eqs) (Array.length clause_2.non_ground_body_eqs)
		|c -> c)
		|c -> c)
		|c -> c)
		|c -> c)
		|c -> c)
		|c -> c)
		|c -> c

		
		

		
		
			
let rec subsumed_by_clause_group clause clauses =  
	match clauses with
		[] -> false
		|head :: tail ->
			if ((compare_subclauses_lengths clause head)=0) then (
				if (unchecked_match_clauses clause head) then 
					(true)
				else
					(subsumed_by_clause_group clause tail)
				)
			else 
				false
	
			
					
let rec remove_clause_duplicates clauses =
	let sorted_clauses = List.sort compare_subclauses_lengths clauses
	in
		remove_clause_duplicates_1 sorted_clauses

and remove_clause_duplicates_1 clauses = 
	match clauses with
		[] -> []
		| head :: tail ->
			if subsumed_by_clause_group head tail then
				(remove_clause_duplicates_1 tail)
			else
				(head :: (remove_clause_duplicates_1 tail))


				
let rec subsume_seek equals a1 index seek_index offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped =
	let seek_index = ref seek_index
	in 
	if index>=offset then (	(*Seeking in the body now, lots of offsets required. *) 
		if (Array.length a2b)>0 then (
			while !seek_index<(Array.length a1) && a1_b.(index)<0 do (* still seekspace in body && current a1 not matched yet *) (*MAINWHILE*)
				if a2b_b.(!seek_index-offset)<0 then (	(* observed bodylit unbound *)
					if (Term.np_subsumes_chk_bindings a1.(index) a2b.(!seek_index-offset) index var1_to_b var1_by_b var2_by_b)	(* observed bodylit matches*)
						then (
							a1_b.(index) <- !seek_index;
							a2b_b.(!seek_index-offset) <- index)
					else (
						if (is_eq equals a1.(index))
							&&(not_flipped.(index))
							&&(Term.np_subsumes_chk_bindings (Term.binary_flip a1.(index)) a2b.(!seek_index-offset) index var1_to_b var1_by_b var2_by_b)
							then (a1_b.(index) <- !seek_index;
							a2b_b.(!seek_index-offset) <- index;
							not_flipped.(index) <- true)
						else (	(*observed bodylit fails match *)
							seek_index := !seek_index+1
						)
					)
				)
				else (	(*observed bodylit is bound *)
					seek_index := !seek_index+1
				)
			done;
			if a1_b.(index)>(-1) then (	(*current has found match *)
				if index<((Array.length a1)-1) then (	(* still lits to match, go on in a1 *)
					subsume_seek equals a1 (index+1) offset offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped
				)
				else true		(* all done *)
			)
			else (	(* Begin heavy messing. *)
				if index=offset then false	(* should never happen, would require headless rule... *)
				else (
					let prev_index = index-1
					in
					let previous = a1.(prev_index)
					and prev_seek_index = a1_b.(prev_index)
					in
						Termutils.iter_vars
							(fun v ->
								let vnum = Term.variable_number v
								in
									if var1_by_b.(vnum) == prev_index then (
										var1_by_b.(vnum)<-(-1);
										var1_to_b.(vnum)<-Term.Unbound
									)
									else ()
							)
							previous;
						if prev_index<offset then (		(* crossing the head/body border... *)
							a2h_b.(prev_seek_index) <- (-1);	
						)
						else (
							a2b_b.(prev_seek_index-offset) <- (-1)
						);

						a1_b.(prev_index) <- (-1);
						(* If the equation in the undone binding had not yet been flipped, there is a chance that its
						flipped version matches the same literal (when all four terms of both equations are variants
						of each other), but which will not cause the problem that initiated the backtracking. Thus,
						check whether a1.(prev_index) has been flipped, and if not, then try to match its flipped
						version to the same literal as before. If this works, continue with matchseek for a1.(index)
						from there. If not, seek a different matching literal for a1.(prev_index).*)
						if (is_eq equals previous)
						&&(not_flipped.(prev_index))
						&&(Term.np_subsumes_chk_bindings (Term.binary_flip previous) a2b.(prev_seek_index - offset) prev_index var1_to_b var1_by_b var2_by_b)
							then
								(a1_b.(prev_index) <- prev_seek_index; (* bind observed headlit to current*)
				 				a2b_b.(prev_seek_index-offset) <- prev_index; (* bind current to observed headlit*) (*escapes MAINWHILE*)
				 				not_flipped.(prev_index) <- false;
				 				subsume_seek equals a1 (prev_index + 1) (prev_seek_index + 1) offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			 			else 
							(not_flipped.(prev_index) <- true;
							subsume_seek equals a1 prev_index (prev_seek_index+1) offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
				)
			)
		)
		else true
	)
	else ( 
		while !seek_index<offset && a1_b.(index)<0 do (*still seekspace in head && current a1 not matched yet*) 
			if a2h_b.(!seek_index)<0 then (	(*observed headlit unbound*)
				if (Term.np_subsumes_chk_bindings a1.(index) a2h.(!seek_index) index var1_to_b var1_by_b var2_by_b)
					then (	 (* observed headlit matches*)
				 		a1_b.(index) <- !seek_index; (* bind observed headlit to current*)
				 		a2h_b.(!seek_index) <- index (* bind current to observed headlit*)
				 	)
				 else ( 
				 	if (is_eq equals a1.(index))
				 	&&(not_flipped.(index)) 
				 	&&(Term.np_subsumes_chk_bindings (Term.binary_flip a1.(index)) a2h.(!seek_index) index var1_to_b var1_by_b var2_by_b)
				 		then (
				 			a1_b.(index) <- !seek_index; (* bind observed headlit to current*)
				 			a2h_b.(!seek_index) <- index; (* bind current to observed headlit*)
				 			not_flipped.(index) <- false
				 	)
				 	else (	(*observed headlit fails match*)
				 		seek_index := !seek_index+1	
					)
				 )
			)
			else (	(*observed headlit is bound *)
				seek_index := !seek_index+1	
			)
		done;	(* main loop is over, this means that current has found match or that there exists no match*)
		if a1_b.(index)>(-1) then ( (* current has found match *)
			if index<((Array.length a1)-1) then (	(* still literals waiting for match, proceed to next a1 *)
				if index<(offset-1) then
					(subsume_seek equals a1 (index+1) 0 offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
				else
					(subsume_seek equals a1 (index+1) offset offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			)
			else true			(* all literals have matched, success! *)
		)
		else (	(* start backtracking *)
			if index=0 then (false)	(* easy: no match for first a1 -> no backtracking possible, all fails *)
			else (	(* So there was a previous a1, and something went wrong there. As current found no match,
					current has not modified any bindings. The last modification was by the previous a1.
					The plan: get index of previous a1, undo its bindings, and restart match_seek for previous a1
					but with increased seek_index from its last position *)
				let prev_index = index-1 
				in
				let previous = a1.(prev_index)
				and prev_seek_index = a1_b.(prev_index)
				in
					Termutils.iter_vars
						(fun v ->
							let vnum = Term.variable_number v
							in
								if var1_by_b.(vnum) == prev_index then (	(* this variable was bound by this literal *)
									var1_by_b.(vnum)<-(-1);		(* this variable is no longer bound by this literal *)
									var1_to_b.(vnum)<-Term.Unbound	(* this variable is no longer bound to another variable *)
								)
								else ()	(* this variable was bound by an earlier literal, it may not be set free by this literal *)
						)
						previous;	(* The variable bindings of previous have been undone. *)
					a2h_b.(prev_seek_index) <- (-1);
					a1_b.(prev_index) <- (-1); (* The literal bindings of previous have been undone. *)
					if (is_eq equals previous)
						&&(not_flipped.(prev_index)) 
						&&(Term.np_subsumes_chk_bindings (Term.binary_flip previous) a2h.(prev_seek_index) prev_index var1_to_b var1_by_b var2_by_b) 
						then (
			 			a1_b.(prev_index) <- prev_seek_index; (* bind observed headlit to current*)
			 			a2h_b.(prev_seek_index) <- prev_index; (* bind current to observed headlit*)
			 			not_flipped.(prev_index) <- false;
			 			subsume_seek equals a1 (prev_index + 1) (prev_seek_index + 1) offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped
			 		)
				 	else	(* flipped term did not match, unflip and test with the next literals *)
						(not_flipped.(prev_index) <- true;
						subsume_seek equals a1 prev_index (prev_seek_index + 1) offset a2h a2b var1_to_b var1_by_b var2_by_b a1_b a2h_b a2b_b not_flipped)
			)
		)
	)	
		

		
		
let rec np_subsumes_check_clauses clause_1 clause_2 =
(*	np_subsumes_check_clauses_st clause_1 clause_2		*)
	np_subsumes_check_clauses_nst clause_1 clause_2
	
and np_subsumes_check_clauses_st clause_1 clause_2 = 
	match clause_1.head, clause_2.head with
		False, False -> unchecked_np_subsumes_check_clauses_st clause_1 clause_2
		| Unit _, Unit _ -> unchecked_np_subsumes_check_clauses_st clause_1 clause_2
		| Conjunction _, Conjunction _ -> unchecked_np_subsumes_check_clauses_st clause_1 clause_2
		| Disjunction _, Disjunction _ -> unchecked_np_subsumes_check_clauses_st clause_1 clause_2
		| _, _ -> false
		
and np_subsumes_check_clauses_nst clause_1 clause_2 = 
	match clause_1.head, clause_2.head with
		False, False -> unchecked_np_subsumes_check_clauses_nst clause_1 clause_2
		| Unit _, Unit _ -> unchecked_np_subsumes_check_clauses_nst clause_1 clause_2
		| Conjunction _, Conjunction _ -> unchecked_np_subsumes_check_clauses_nst clause_1 clause_2
		| Disjunction _, Disjunction _ -> unchecked_np_subsumes_check_clauses_nst clause_1 clause_2
		| _, _ -> false		

and unchecked_np_subsumes_check_clauses clause_1 clause_2 =		
	unchecked_np_subsumes_check_clauses_st clause_1 clause_2
	
and unchecked_np_subsumes_check_clauses_st clause_1 clause_2 =
	if ((Array.length clause_1.ground_head_terms)+(Array.length clause_1.non_ground_head_terms))			
		== ((Array.length clause_2.ground_head_terms)+(Array.length clause_2.non_ground_head_terms))
		&& ((Array.length clause_1.ground_head_eqs)+(Array.length clause_1.non_ground_head_eqs))			
		== ((Array.length clause_2.ground_head_eqs)+(Array.length clause_2.non_ground_head_eqs))	
		&&((Array.length clause_1.ground_body_terms)+(Array.length clause_1.non_ground_body_terms))			
		== ((Array.length clause_2.ground_body_terms)+(Array.length clause_2.non_ground_body_terms))
		&&((Array.length clause_1.ground_body_eqs)+(Array.length clause_1.non_ground_body_eqs))			
		== ((Array.length clause_2.ground_body_eqs)+(Array.length clause_2.non_ground_body_eqs)) 
		then ( 
			(try
				(let (unmatched_head_terms_1, unmatched_head_terms_2) = 
					get_unmatched_lits clause_1.ground_head_terms clause_2.ground_head_terms
				in
				if (List.length unmatched_head_terms_1)>0 then raise Failed;
				let (unmatched_head_eqs_1, unmatched_head_eqs_2) = 
					get_unmatched_lits clause_1.ground_head_eqs clause_2.ground_head_eqs
				in
				if (List.length unmatched_head_eqs_1)>0 then raise Failed;				
				let (unmatched_body_terms_1, unmatched_body_terms_2) = 
					get_unmatched_lits clause_1.ground_body_terms clause_2.ground_body_terms
				in
				if (List.length unmatched_body_terms_1)>0 then raise Failed;
				let (unmatched_body_eqs_1, unmatched_body_eqs_2) = 
					get_unmatched_lits clause_1.ground_body_eqs clause_2.ground_body_eqs
				in
				if (List.length unmatched_body_eqs_1)>0 then raise Failed;				
				let clause_2_head_terms = 
					Array.append clause_2.non_ground_head_terms (Array.of_list unmatched_head_terms_2)
				and clause_2_head_eqs = 
					Array.append clause_2.non_ground_head_eqs (Array.of_list unmatched_head_eqs_2)
				and clause_2_body_terms = 
					Array.append clause_2.non_ground_body_terms (Array.of_list unmatched_body_terms_2)
				and clause_2_body_eqs = 
					Array.append clause_2.non_ground_body_eqs (Array.of_list unmatched_body_eqs_2)
				and equals = Term.get_equality_symbol (clause_1.aux.Aux.symbol_table)
				in 
					np_subsumes_chk
						equals
						(Array.append clause_1.non_ground_head_terms clause_1.non_ground_head_eqs)
						(Array.append clause_2_head_terms clause_2_head_eqs)
						(Array.append clause_1.non_ground_body_terms clause_1.non_ground_body_eqs)
						(Array.append clause_2_body_terms clause_2_body_eqs))
			with Failed -> false))
		else false
	
and unchecked_np_subsumes_check_clauses_nst clause_1 clause_2 = (*ps "Checking:";ppp clause_1.source;pe ".";ppp clause_2.source;pe ".";pe "";*)
	try
		(let (unmatched_head_terms_1, unmatched_head_terms_2) = 
			get_unmatched_lits clause_1.ground_head_terms clause_2.ground_head_terms
		in
		if (List.length unmatched_head_terms_1)>0 then raise Failed;
		let (unmatched_head_eqs_1, unmatched_head_eqs_2) = 
			get_unmatched_lits clause_1.ground_head_eqs clause_2.ground_head_eqs
		in
		if (List.length unmatched_head_eqs_1)>0 then raise Failed;				
		let (unmatched_body_terms_1, unmatched_body_terms_2) = 
			get_unmatched_lits clause_1.ground_body_terms clause_2.ground_body_terms
		in
		if (List.length unmatched_body_terms_1)>0 then raise Failed;
		let (unmatched_body_eqs_1, unmatched_body_eqs_2) = 
			get_unmatched_lits clause_1.ground_body_eqs clause_2.ground_body_eqs
		in
		if (List.length unmatched_body_eqs_1)>0 then raise Failed;				
		let clause_2_head_terms = 
			Array.append clause_2.non_ground_head_terms (Array.of_list unmatched_head_terms_2)
		and clause_2_head_eqs = 
			Array.append clause_2.non_ground_head_eqs (Array.of_list unmatched_head_eqs_2)
		and clause_2_body_terms = 
			Array.append clause_2.non_ground_body_terms (Array.of_list unmatched_body_terms_2)
		and clause_2_body_eqs = 
			Array.append clause_2.non_ground_body_eqs (Array.of_list unmatched_body_eqs_2)
		and equals = Term.get_equality_symbol (clause_1.aux.Aux.symbol_table)
		in 
			np_subsumes_chk
				equals
				(Array.append clause_1.non_ground_head_terms clause_1.non_ground_head_eqs)
				(Array.append clause_2_head_terms clause_2_head_eqs)
				(Array.append clause_1.non_ground_body_terms clause_1.non_ground_body_eqs)
				(Array.append clause_2_body_terms clause_2_body_eqs))
	with Failed -> (false)


and get_unmatched_lits lits_1 lits_2 =
	let i = ref 0
	and j = ref 0
	and unmatched_lits_1 = ref []
	and unmatched_lits_2 = ref []
	and len_1 = Array.length lits_1
	and len_2 = Array.length lits_2
	in 		
		if (len_1 > 0) && (len_2 > 0) then (		
			while (!i < len_1)&&(!j < len_2) do
				if (Term.term_ident lits_1.(!i) lits_2.(!j)) then (
					i := !i+1;
					j := !j+1
				)
				else (
					if (LrpOrder.lrpo_greater lits_1.(!i) lits_2.(!j)) then (
						unmatched_lits_2 := lits_2.(!j) :: !unmatched_lits_2;
						j := !j+1
					)
					else (unmatched_lits_1 := lits_1.(!i) :: !unmatched_lits_1;
						i := !i+1
					)
				)
			done;	
			while (!i < len_1) do
				unmatched_lits_1 := lits_1.(!i) :: !unmatched_lits_1;
				i := !i+1
			done;	
			while (!j < len_2) do
				unmatched_lits_2 := lits_2.(!j) :: !unmatched_lits_2;
				j := !j+1
			done;	
			(!unmatched_lits_1, !unmatched_lits_2)
		)
		else (
			((Array.to_list lits_1), (Array.to_list lits_2))
		)
		
					
 
and np_subsumes_chk equals h1 h2 b1 b2 = 
	let var1_to_b = Array.make Term.max_vars Term.Unbound
	and var1_by_b = Array.make Term.max_vars (-1)
	and var2_by_b = Array.make Term.max_vars (-1)
	and offset = Array.length h1
	in 
	let h1_b = Array.make offset (-1)
	and h2_b = Array.make offset (-1)
	and b1_b = Array.make (Array.length b1) (-1)
	and b2_b = Array.make (Array.length b2) (-1)
	and not_flipped = Array.make (offset + (Array.length b1)) true
  	in 
		subsume_seek equals (Array.append h1 b1) 0 0 offset h2 b2 var1_to_b var1_by_b var2_by_b (Array.append h1_b b1_b) h2_b  b2_b not_flipped
 		
(*		
and pa ar =
	Array.iter
		(fun t -> ppp t;ps " ")
		ar		
*)			
let rec ng_compare_subclauses_lengths clause_1 clause_2 =
	match clause_1.head, clause_2.head with
		False, False -> ng_compare_subclauses_lengths_1 clause_1 clause_2
		| False, _ -> (-1)
		| _, False -> 1
		| Unit _, Unit _ -> ng_compare_subclauses_lengths_1 clause_1 clause_2
		| Unit _, _ -> (-1)
		| _, Unit _ -> 1
		| Conjunction _, Conjunction _ -> ng_compare_subclauses_lengths_1 clause_1 clause_2
		| Conjunction _, _ -> (-1)
		| _, Conjunction _ -> 1
		| Disjunction _, Disjunction _ -> ng_compare_subclauses_lengths_1 clause_1 clause_2

and ng_compare_subclauses_lengths_1 clause_1 clause_2 =		
	match compare ((Array.length clause_1.ground_head_terms)+(Array.length clause_1.non_ground_head_terms)) 
					((Array.length clause_2.ground_head_terms)+(Array.length clause_2.non_ground_head_terms)) with
		0 ->
	(match compare ((Array.length clause_1.ground_body_terms)+(Array.length clause_1.non_ground_body_terms)) 
					((Array.length clause_2.ground_body_terms)+(Array.length clause_2.non_ground_body_terms)) with
		0 ->
	(match compare ((Array.length clause_1.ground_head_eqs)+(Array.length clause_1.non_ground_head_eqs)) 
					((Array.length clause_2.ground_head_eqs)+(Array.length clause_2.non_ground_head_eqs)) with
		0 -> compare ((Array.length clause_1.ground_head_eqs)+(Array.length clause_1.non_ground_head_eqs)) 
					((Array.length clause_2.ground_head_eqs)+(Array.length clause_2.non_ground_head_eqs))
		|c -> c)
		|c -> c)
		|c -> c	
		
let rec np_subsumed_by_clause_group clause clauses =  
	match clauses with
		[] -> false
		|head :: tail -> 
			if ((ng_compare_subclauses_lengths clause head)=0) then (
				if (unchecked_np_subsumes_check_clauses head clause) then 
					(true)
				else
					(np_subsumed_by_clause_group clause tail)
				)
			else 
				false	

						
let rec remove_subsumed_clauses clauses = 
	let sorted_clauses = List.sort ng_compare_subclauses_lengths clauses
	in
		remove_subsumed_clauses_1 sorted_clauses

and remove_subsumed_clauses_1 clauses = 
	match clauses with
		[] -> []
		| head :: tail -> 
			if np_subsumed_by_clause_group head tail then
				(remove_subsumed_clauses_1 tail)
			else
				(head :: (remove_subsumed_clauses_1 tail))		
	
				
				
let is_indexed subterm_table clause = 
	let sig_key = find_subterm subterm_table (Convert.term_functor_symbol clause.sigterm)
	and check = ref false
	in 
	(try (
		Reader.iter_instance_candidates 
			(fun entry -> 
				match entry with
					SubtermEntry.Clause (candidate, _, _, _) -> 
						if (match_clauses clause candidate) then 		
							(raise Found)
					|_ -> ())
			sig_key.reader	
			clause.sigterm
			Term.null_context)
	with Found -> (check := true));
	!check

let functors_sums_check clause_1 clause_2 =
	(clause_1.head_terms_functors_sum <= clause_2.head_terms_functors_sum)
	&& 	
	(clause_1.head_eqs_functors_sum <= clause_2.head_eqs_functors_sum)
	&&
	(clause_1.body_terms_functors_sum <= clause_2.body_terms_functors_sum)
	&& 	
	(clause_1.body_eqs_functors_sum <= clause_2.body_eqs_functors_sum)
	
let functors_sums_eq_check clause_1 clause_2 =
	(clause_1.head_terms_functors_sum = clause_2.head_terms_functors_sum)
	&& 	
	(clause_1.head_eqs_functors_sum = clause_2.head_eqs_functors_sum)
	&&
	(clause_1.body_terms_functors_sum = clause_2.body_terms_functors_sum)
	&& 	
	(clause_1.body_eqs_functors_sum = clause_2.body_eqs_functors_sum)	
	
let is_trivial_eq_unit clause =
	match clause.head with
		Unit lit ->
			(match (HeadLiteral.get_atom lit) with
				Eq eq -> 
					(if (Term.term_ident (Equality.get_left eq) (Equality.get_right eq)) then true
					else false)
				|_ -> (raise (Err.Error ("Bad clause in Ic.is_trivial_eq_unit."))))
		|_ -> (raise (Err.Error ("Bad clause in Ic.is_trivial_eq_unit.")))	
	
let np_subsumption_processing_clause subterm_table redundant_store clause = 
	let sig_key = find_subterm subterm_table (Convert.term_functor_symbol clause.fuzzy_sigterm)
	and iter_instance_candidates =
		if (Cfg.get_full_branch_redundancy_flag clause.aux.Aux.cfg) then
			Reader.iter_instance_candidates
		else
			Reader.iter_instance_candidates_ml
	in
	try ( 
		if clause.is_eq_unit then (
			if (is_trivial_eq_unit clause) then (raise Found);
			let flip_term = Term.binary_flip clause.source
			in
			Reader.iter_generalization_candidates
				(fun entry -> 
					match entry with
						SubtermEntry.Clause (candidate_clause, _, _, _) -> 
							if (is_redundant candidate_clause) then ()
							else (	
								if (functors_sums_check candidate_clause clause) then
									(if (Term.subsumes_chk candidate_clause.source clause.source) then
										(raise Found);
									if not (clause.is_oriented && candidate_clause.is_oriented) 
										&& (Term.subsumes_chk candidate_clause.source flip_term)
										then
											(raise Found)))	
						|_ ->())
				sig_key.reader
				clause.fuzzy_sigterm
				Term.null_context;
			iter_instance_candidates
				(fun entry -> 
					match entry with
						SubtermEntry.Clause (candidate_clause, _, _, _) -> 
							if (is_redundant candidate_clause) then ()
							else (
								if (functors_sums_check clause candidate_clause) then
									(if (Term.subsumes_chk clause.source candidate_clause.source) then
										(declare_clause_redundant redundant_store candidate_clause);
									if not (clause.is_oriented && candidate_clause.is_oriented) 
										&& (Term.subsumes_chk flip_term candidate_clause.source)
										then
											(declare_clause_redundant redundant_store candidate_clause)))	
						|_ ->())
				sig_key.reader
				clause.fuzzy_sigterm
				Term.null_context)
		else (
			Reader.iter_instance_candidates
				(fun entry -> 
					match entry with
						SubtermEntry.Clause (candidate_clause, _, _, _) -> 
							if (is_redundant candidate_clause) then ()
							else (	
								if (functors_sums_check candidate_clause clause) &&
									(np_subsumes_check_clauses candidate_clause clause) then
									(raise Found))
						|_ ->())
				sig_key.reader
				clause.fuzzy_sigterm
				Term.null_context;
			if clause.is_ground then
				(iter_instance_candidates
					(fun entry -> 
						match entry with
							SubtermEntry.Clause (candidate_clause, _, _, _) -> 
								if (is_redundant candidate_clause)||(not candidate_clause.is_ground) then ()
								else (
									if (functors_sums_eq_check candidate_clause clause) &&
										(match_clauses clause candidate_clause) then
										(declare_clause_redundant redundant_store candidate_clause))
							|_ ->())
					sig_key.reader
					clause.sigterm
					Term.null_context)
			else
				(iter_instance_candidates
					(fun entry -> 
						match entry with
							SubtermEntry.Clause (candidate_clause, _, _, _) -> 
								if (is_redundant candidate_clause) then ()
								else (
									if (functors_sums_check clause candidate_clause) &&
										(np_subsumes_check_clauses clause candidate_clause) then
										(declare_clause_redundant redundant_store candidate_clause))
							|_ ->())
					sig_key.reader
					clause.fuzzy_sigterm
					Term.null_context)))
	with Found -> 
			(raise Redundant)			
			
			
let clause_np_subsumed_by_clauses subterm_table clause = 
	let sig_key = find_subterm subterm_table (Convert.term_functor_symbol clause.fuzzy_sigterm)
	in
	try ( 
		if clause.is_eq_unit then (
			let flip_term = Term.binary_flip clause.source
			in
			Reader.iter_generalization_candidates
				(fun entry -> 
					match entry with
						SubtermEntry.Clause (candidate_clause, _, _, _) -> 
							if (is_redundant candidate_clause) then ()
							else (	
								if (functors_sums_check candidate_clause clause) then
									(if (Term.subsumes_chk candidate_clause.source clause.source) then
										(raise Found);
									if not (clause.is_oriented && candidate_clause.is_oriented) 
										&& (Term.subsumes_chk candidate_clause.source flip_term)
										then
											(raise Found)))
						|_ ->())
				sig_key.reader
				clause.fuzzy_sigterm
				Term.null_context)
		else (
			Reader.iter_instance_candidates
				(fun entry -> 
					match entry with
						SubtermEntry.Clause (candidate_clause, _, _, _) -> 
							if (is_redundant candidate_clause) then ()
							else (	
								if (functors_sums_check candidate_clause clause) &&
									(np_subsumes_check_clauses candidate_clause clause) then
									(raise Found))
						|_ ->())
				sig_key.reader
				clause.fuzzy_sigterm
				Term.null_context);
		false)
	with Found -> 
			(true)			
			
	
let non_ground_to_ground_clauses_lrpo_greater clause_1 clause_2 =
	if (not clause_1.is_ground) && clause_2.is_ground then (
		let (clause_1_unmatched_body_terms, clause_2_unmatched_body_terms) = 
			get_unmatched_lits 
				clause_1.ground_body_terms
				clause_2.ground_body_terms
		and (clause_1_unmatched_head_terms, clause_2_unmatched_head_terms) = 
			get_unmatched_lits 
				clause_1.ground_head_terms
				clause_2.ground_head_terms
		and (clause_1_unmatched_body_eqs, clause_2_unmatched_body_eqs) = 
			get_unmatched_lits 
				clause_1.ground_body_eqs
				clause_2.ground_body_eqs
		and (clause_1_unmatched_head_eqs, clause_2_unmatched_head_eqs) = 
			get_unmatched_lits 
				clause_1.ground_head_eqs
				clause_2.ground_head_eqs
		in
			if (List.length clause_1_unmatched_body_terms)=0
				&& (List.length clause_1_unmatched_head_terms)=0
				&& (List.length clause_1_unmatched_body_eqs)=0
				&& (List.length clause_1_unmatched_head_eqs)=0
				then true
			else false
	)
	else (raise (Err.Error ("Ground error in Ic.non_ground_to_ground_clauses_lrpo_greater.")))
		
let rec ground_clauses_lrpo_greater clause_1 clause_2 =
	if clause_1.is_ground && clause_2.is_ground then
		(not (match_clauses clause_1 clause_2)
		&& (ground_clauses_lrpo_greater_1 clause_1 clause_2))
	else (raise (Err.Error ("Non-ground clause in Ic.ground_clauses_lrpo_greater.")))
	
and ground_clauses_lrpo_greater_1 clause_1 clause_2 = 
	let (clause_1_unmatched_body_terms, clause_2_unmatched_body_terms) = 
		get_unmatched_lits 
			clause_1.ground_body_terms
			clause_2.ground_body_terms
	and (clause_1_unmatched_head_terms, clause_2_unmatched_head_terms) = 
		get_unmatched_lits 
			clause_1.ground_head_terms
			clause_2.ground_head_terms
	and (clause_1_unmatched_body_eqs, clause_2_unmatched_body_eqs) = 
		get_unmatched_lits 
			clause_1.ground_body_eqs
			clause_2.ground_body_eqs
	and (clause_1_unmatched_head_eqs, clause_2_unmatched_head_eqs) = 
		get_unmatched_lits 
			clause_1.ground_head_eqs
			clause_2.ground_head_eqs
	in 
	let c1bt = 
		if (List.length clause_1_unmatched_body_terms)>0 then
			(List.hd clause_1_unmatched_body_terms)
		else (Term.Unbound)
	and c2bt = 
		if (List.length clause_2_unmatched_body_terms)>0 then
			(List.hd clause_2_unmatched_body_terms)
		else (Term.Unbound)
	and c1ht = 
		if (List.length clause_1_unmatched_head_terms)>0 then
			(List.hd clause_1_unmatched_head_terms)
		else (Term.Unbound)
	and c2ht = 
		if (List.length clause_2_unmatched_head_terms)>0 then
			(List.hd clause_2_unmatched_head_terms)
		else (Term.Unbound)
	and c1be_l, c1be_r = 
		if (List.length clause_1_unmatched_body_eqs)>0 then
			Convert.term_pair (List.hd clause_1_unmatched_body_eqs)
		else (Term.Unbound, Term.Unbound)
	and c2be_l, c2be_r = 
		if (List.length clause_2_unmatched_body_eqs)>0 then
			Convert.term_pair (List.hd clause_2_unmatched_body_eqs)
		else (Term.Unbound, Term.Unbound)
	and c1he_l, c1he_r = 
		if (List.length clause_1_unmatched_head_eqs)>0 then
			Convert.term_pair (List.hd clause_1_unmatched_head_eqs)
		else (Term.Unbound, Term.Unbound)
	and c2he_l, c2he_r = 
		if (List.length clause_2_unmatched_head_eqs)>0 then
			Convert.term_pair (List.hd clause_2_unmatched_head_eqs)
		else (Term.Unbound, Term.Unbound)		
	and gr x y = LrpOrder.lrpo_greater x y
	in
	let greq x y = (Term.term_ident x y) || (gr x y)
	in
	let egr l1 r1 l2 r2 = (gr l1 l2) || ((Term.term_ident l1 l2) && (gr r1 r2))
	in
	let egreq l1 r1 l2 r2 = ((Term.term_ident l1 l2) && (Term.term_ident r1 r2)) || (egr l1 r1 l2 r2)
	in
		if ((gr c1bt c2bt) && (greq c1bt c2ht))	
			||
			((gr c1ht c2bt) && (gr c1ht c2ht)) then true
		else (
			if (c2bt==Term.Unbound) && (c2ht==Term.Unbound) then (
				if ((egr c1be_l c1be_r c2be_l c2be_r)&&(egreq c1be_l c1be_r c2he_l c2he_r))
					||
					((egr c1he_l c1he_r c2be_l c2be_r)&&(egr c1he_l c1he_r c2he_l c2he_r))
					then true
				else false)
			else false)
		
			
(****************************************************************************)
(* New clause candidate storage.                                            *) 
(****************************************************************************)
(* This is similar to a news object from News.ml, except for storing        *)
(* clauses instead of units.                                                *)

module NewClauseCandidates =
	struct

let ncc_id = ref 0

let gen_ncc_id () =
  ncc_id := !ncc_id + 1;	
  "ncc_"^string_of_int(!ncc_id)

let reset_gen_ncc_id () =
  ncc_id := 0  	
	
type new_clause_candidates =
	{ncc_id: string;
	current_stratum_table: NewClausesTable.index;
	other_stratum_table: NewClausesTable.index;	
	delayed_table: NewClausesTable.index;
	term_weighing_method: Weight.term_weighing_method;
	mutable current_stratum_candidates: clause list;
	mutable other_stratum_candidates: clause list;
	mutable delayed_candidates: clause list;
	mutable max_literal_weight: int;
	mutable weight_limit_exceeded: int}
	
let make_new_clause_candidates cfg =
	{ncc_id = gen_ncc_id ();
	current_stratum_table = NewClausesTable.make_index ();
	other_stratum_table = NewClausesTable.make_index ();
	delayed_table = NewClausesTable.make_index ();
	term_weighing_method=
		(match Cfg.get_weighing_method_param cfg with
			0 -> Weight.SymbolCount
			| 1 -> Weight.Depth
			| _ -> raise
			(Err.Error "Bad value for term_weighing_method parameter."));
	current_stratum_candidates = [];
	other_stratum_candidates = [];
	delayed_candidates = [];
	max_literal_weight = Cfg.get_max_weight_initial_param cfg;
	weight_limit_exceeded = 0}		
	
let term_array_max_weight term_array term_weighing_method max_w =
	Array.iter
		(fun term ->
			let w = Weight.term_weight term_weighing_method term
			in
				if (w > !max_w) then (max_w := w))
		term_array
	
let eq_array_max_weight_1 eq_array term_weighing_method max_w =		
	Array.iter
	(fun eq_term ->
		let l, r = Convert.term_pair eq_term
		in
			let w =
				max
					(Weight.term_weight term_weighing_method l)
					(Weight.term_weight term_weighing_method r)
			in
				if (w > !max_w) then (max_w := w))
	eq_array
				
let eq_array_max_weight eq_array term_weighing_method max_w =
	term_array_max_weight eq_array term_weighing_method max_w
(*	eq_array_max_weight_1 eq_array term_weighing_method max_w*)
	
		
	
let weight_check new_clause_candidates new_clause = (*pi new_clause_candidates.weight_limit_exceeded;	pe " <Before";*)
	if new_clause_candidates.max_literal_weight != -1 &&
	    (let w = 
	    	(let max_w = ref 0
	    	and term_weighing_method = new_clause_candidates.term_weighing_method
	    	in
	    		term_array_max_weight new_clause.ground_head_terms term_weighing_method max_w;
	    		term_array_max_weight new_clause.non_ground_head_terms term_weighing_method max_w;
	    		eq_array_max_weight new_clause.ground_head_eqs term_weighing_method max_w;
	    		eq_array_max_weight new_clause.non_ground_head_eqs term_weighing_method max_w;
	    		term_array_max_weight new_clause.ground_body_terms term_weighing_method max_w;
	    		term_array_max_weight new_clause.non_ground_body_terms term_weighing_method max_w;
	    		eq_array_max_weight new_clause.ground_body_eqs term_weighing_method max_w;
	    		eq_array_max_weight new_clause.non_ground_body_eqs term_weighing_method max_w;
	    		!max_w - new_clause_candidates.max_literal_weight) 
		 in
		    if w > 0 then (
				new_clause_candidates.weight_limit_exceeded <- 
			  		if new_clause_candidates.weight_limit_exceeded > 0 then
			    		min new_clause_candidates.weight_limit_exceeded w
			  		else w;	
					true
			)			
		    else false) 
		then (false)
	else (true)

let clause_is_subsumed_by_new_clauses table clause =
	try 
		(if clause.is_eq_unit then 
			(let flip_term = Term.binary_flip clause.source
			in
			NewClausesTable.iter_generalization_candidates
				(fun entry ->
					if (SubtermEntry.is_redundant entry) then ()
					else (				
						match entry with
							(SubtermEntry.Clause (candidate_clause, _, _, _)) ->
								if (Term.subsumes_chk candidate_clause.source clause.source) then
									(raise Found);
								if (not (candidate_clause.is_oriented && clause.is_oriented))
									&& (Term.subsumes_chk candidate_clause.source flip_term) then
									(raise Found)
							|_ -> ()))
				table
				clause.fuzzy_sigterm
				Term.null_context)
		else 
			(NewClausesTable.iter_instance_candidates
				(fun entry ->		
					if (SubtermEntry.is_redundant entry) then ()
					else (				
						match entry with
							(SubtermEntry.Clause (candidate_clause, _, _, _)) ->	
								if (np_subsumes_check_clauses candidate_clause clause) then
									(raise Found)
							|_ -> ()))
				table
				clause.fuzzy_sigterm
				Term.null_context);
		false)
	with Found -> true	
	

let remove_subsumed_from_new_clauses table clause =
	if clause.is_eq_unit then 
		(let flip_term = Term.binary_flip clause.source
		in
		NewClausesTable.iter_instance_candidates
			(fun entry ->
				if (SubtermEntry.is_redundant entry) then ()
				else (
					match entry with
						(SubtermEntry.Clause (candidate_clause, _, _, _)) ->
							if (Term.subsumes_chk clause.source candidate_clause.source) then
								(declare_redundant candidate_clause);
							if (not (candidate_clause.is_oriented && clause.is_oriented))
								&& (Term.subsumes_chk flip_term candidate_clause.source) then
								(declare_redundant candidate_clause)
						|_ -> ()))
			table
			clause.fuzzy_sigterm
			Term.null_context)
	else 
		(NewClausesTable.iter_instance_candidates
			(fun entry ->
				if (SubtermEntry.is_redundant entry) then ()
				else (			
					match entry with
						(SubtermEntry.Clause (candidate_clause, _, _, _)) ->
							if (np_subsumes_check_clauses clause candidate_clause) then
								(declare_redundant candidate_clause)
							|_ -> ()))
			table
			clause.fuzzy_sigterm
			Term.null_context)
		
let insert_delayed_candidate new_clause_candidates new_clause =
	if (Cfg.get_ehyper new_clause.aux.Aux.cfg) then
		(if not (clause_is_subsumed_by_new_clauses 
					new_clause_candidates.delayed_table 
					new_clause)
		then 
			(remove_subsumed_from_new_clauses
			new_clause_candidates.delayed_table
			new_clause;
			NewClausesTable.insert 
				new_clause_candidates.delayed_table 
					new_clause.fuzzy_sigterm 
					(SubtermEntry.Clause (new_clause, 0, false, Equality.null_place));
			if new_clause.is_eq_unit then
				(NewClausesTable.insert
					new_clause_candidates.delayed_table
					(Term.binary_flip new_clause.fuzzy_sigterm)
					(SubtermEntry.Clause (new_clause, 0, true, Equality.null_place)));
			new_clause_candidates.delayed_candidates <-
				new_clause :: new_clause_candidates.delayed_candidates))
	
	
let insert_current_stratum_candidate new_clause_candidates new_clause = 
	if (weight_check new_clause_candidates new_clause) then 
		(if not (clause_is_subsumed_by_new_clauses 
						new_clause_candidates.current_stratum_table 
						new_clause)
		then (
			remove_subsumed_from_new_clauses
				new_clause_candidates.current_stratum_table
				new_clause;	
			NewClausesTable.insert 
				new_clause_candidates.current_stratum_table 
					new_clause.fuzzy_sigterm 
					(SubtermEntry.Clause (new_clause, 0, false, Equality.null_place));
			if new_clause.is_eq_unit then
				(NewClausesTable.insert
					new_clause_candidates.current_stratum_table
					(Term.binary_flip new_clause.fuzzy_sigterm)
					(SubtermEntry.Clause (new_clause, 0, true, Equality.null_place)));
			new_clause_candidates.current_stratum_candidates <-
				new_clause :: new_clause_candidates.current_stratum_candidates)) 
		else
			(insert_delayed_candidate new_clause_candidates new_clause)
			
let insert_other_stratum_candidate new_clause_candidates new_clause =
	if (weight_check new_clause_candidates new_clause) then
		(if not (clause_is_subsumed_by_new_clauses 
					new_clause_candidates.other_stratum_table 
					new_clause)
			then (
				remove_subsumed_from_new_clauses
					new_clause_candidates.other_stratum_table
					new_clause;	
				NewClausesTable.insert 
					new_clause_candidates.other_stratum_table 
						new_clause.fuzzy_sigterm 
						(SubtermEntry.Clause (new_clause, 0, false, Equality.null_place));
				if new_clause.is_eq_unit then
					(NewClausesTable.insert
						new_clause_candidates.other_stratum_table
						(Term.binary_flip new_clause.fuzzy_sigterm)
						(SubtermEntry.Clause (new_clause, 0, true, Equality.null_place)));
				new_clause_candidates.other_stratum_candidates <-
					new_clause :: new_clause_candidates.other_stratum_candidates))
	else
		(insert_delayed_candidate new_clause_candidates new_clause)
		
let dump_table table =
	let clauses = ref []
	in
		NewClausesTable.iter_all
			(fun entry ->
				if (SubtermEntry.is_redundant entry) then ()
				else (
					match entry with
						SubtermEntry.Clause (clause, _, second_entry, _) ->
							if second_entry then ()
							else (clauses := clause :: !clauses)
						|_ -> ()))
			table;
		!clauses			
			
let get_current_stratum_candidates new_clause_candidates =
	let clauses = ref []
	in
		List.iter
			(fun clause ->
				if (is_redundant clause) then ()
				else
					(clauses := clause :: !clauses))
			new_clause_candidates.current_stratum_candidates;
		!clauses


let get_other_stratum_candidates new_clause_candidates =
	let clauses = ref []
	in
		List.iter
			(fun clause ->
				if (is_redundant clause) then ()
				else
					(clauses := clause :: !clauses))
			new_clause_candidates.other_stratum_candidates;
		!clauses
		
let get_delayed_candidates new_clause_candidates =
	let clauses = ref []
	in
		List.iter
			(fun clause ->
				if (is_redundant clause) then ()
				else
					(clauses := clause :: !clauses))
			new_clause_candidates.delayed_candidates;
		!clauses		
		
let clear_new_clause_candidates new_clause_candidates =
	NewClausesTable.clear new_clause_candidates.current_stratum_table;
	NewClausesTable.clear new_clause_candidates.other_stratum_table;
	new_clause_candidates.current_stratum_candidates <- [];
	new_clause_candidates.other_stratum_candidates <- []
	
let clear_delayed_clause_candidates new_clause_candidates =
	NewClausesTable.clear new_clause_candidates.delayed_table;
	new_clause_candidates.delayed_candidates <- []

let get_ncc_id new_clause_candidates =
	new_clause_candidates.ncc_id
	
let set_max_literal_weight new_clause_candidates weight =
	new_clause_candidates.max_literal_weight <- weight

let get_max_literal_weight new_clause_candidates =
	new_clause_candidates.max_literal_weight
	
let reset_weight_limit_exceeded new_clause_candidates =
	new_clause_candidates.weight_limit_exceeded <- 0

let weight_limit_exceeded new_clause_candidates =
	new_clause_candidates.weight_limit_exceeded	
	
end			
			
(****************************************************************************)
(* term minimizing:											*)
(* reductive rewriting of a term until a minimum is reached	*)

let rec minimize_term subterm_table equals term =	
	if Term.term_ident term Term.Unbound then
		(term)
	else
	(let term2 = Term.copy_term term
	in	
	let place = Equality.make_array_place term2
	and context_lr = Term.make_context 0
	and context_st = Term.make_context 0
	and trail = Term.make_trail ()
	in	
		try
			(			
			Equality.iter_nonvar_subterms_place
				(minimize_subterm
					subterm_table
					equals
					term2
					context_lr
					context_st
					trail)
				term2
				place;	
			term)
		with Found ->
			(let result = (Equality.get_place place)
			in		
				if (Term.term_ident result term) then (term)
				else (minimize_term subterm_table equals result)))
	
and minimize_subterm subterm_table equals mainterm context_lr context_st trail term place =
	let access_eq_term =
		Term.Compound
			(Term.make_compound
				equals
				[|term; Term.Variable (Term.make_variable 0)|])
	and eq_reader =
		get_subterm_reader
			(find_subterm subterm_table equals)
	and context = Term.make_context 0
	in	
		Reader.iter_unification_candidates
			(fun entry ->	
				match entry with 
					SubtermEntry.Clause (clause, _, left_is_indexed, _) ->
						(if clause.is_eq_unit 
							&& left_is_indexed 
							&& not (is_redundant clause)
							then (
								let eq =
									match clause.head with
										Unit lit -> 
											(match (HeadLiteral.get_atom lit) with
												Eq eq -> eq
												|_ -> (raise (Err.Error "Bad clause in Ic.minimize_subterm.")))
										|_ -> (raise (Err.Error "Bad clause in Ic.minimize_subterm."))
								in
								let l = Equality.get_left eq
								and r = Equality.get_right eq
								in
									if (Equality.get_is_oriented eq) then 
										(if (Term.failed (Term.unify term context_st l context_lr trail)) then ()
										else 
											(Equality.set_place place (Term.apply r context_lr);
											raise Found))
									else 
										(if (Term.failed (Term.unify term context_st l context_lr trail)) then ()
										else 
											(let r2 = Term.apply r context_lr
											in
												if (LrpOrder.lrpo_greater term r2) then
													(Equality.set_place place r2;
													raise Found));
										if (Term.failed (Term.unify term context_st r context_lr trail)) then ()
										else 
											(let l2 = Term.apply l context_lr
											in
												if (LrpOrder.lrpo_greater term l2) then
													(Equality.set_place place l2;
													raise Found)))
								)
							else ())
					|_ ->())
			eq_reader
			access_eq_term
			context
			
			
			
let minimize_literal_term equals subterm_table term =
	match term with
		Term.Compound k ->
			let minimized_args =
				Array.map 
					(fun arg ->
						minimize_term
							subterm_table
							equals
							arg)
					(Term.compound_args k)
			in
				Term.Compound
					(Term.make_compound
						(Term.compound_functor k)
						minimized_args)
	|_ -> term
	
			
	
(* Similar to minimize_subterm, but intended for redundancy test of delayed equation units. Input is
minimized only when the minimizing input is smaller than the input, thus rendering the input redundant. *)	
let minimize_eq_subterm subterm_table equals mainterm context_lr context_st trail term place =
	let access_eq_term =
		Term.Compound
			(Term.make_compound
				equals
				[|term; Term.Variable (Term.make_variable 0)|])
	and eq_reader =
		get_subterm_reader
			(find_subterm subterm_table equals)
	and context = Term.make_context 0
	in	
		Reader.iter_unification_candidates
			(fun entry ->	
				match entry with 
					SubtermEntry.Clause (clause, _, left_is_indexed, _) ->
						(if clause.is_eq_unit
							&& clause.is_ground 
							&& left_is_indexed 
							&& not (is_redundant clause)
							then (
								let eq =
									match clause.head with
										Unit lit -> 
											(match (HeadLiteral.get_atom lit) with
												Eq eq -> eq
												|_ -> (raise (Err.Error "Bad clause in Ic.minimize_eq_subterm.")))
										|_ -> (raise (Err.Error "Bad clause in Ic.minimize_eq_subterm."))
								in
								let l = Equality.get_left eq
								and r = Equality.get_right eq
								in
									(if (Term.failed (Term.unify term context_st l context_lr trail)) then ()
									else 
										(if (LrpOrder.lrpo_greater mainterm clause.ground_head_eqs.(0)) then
											(Equality.set_place place (Term.apply r context_lr);
											raise Found)
										else ())))
							else ())
					|_ ->())
			eq_reader
			access_eq_term
			context	
	
let rec minimize_eq_term equals subterm_table eq_term =
	let eq_term2 = Term.copy_term eq_term
	in
	let l, _ = Convert.term_pair eq_term2	
	in
	let place = Equality.make_array_place l
	and context_lr = Term.make_context 0
	and context_st = Term.make_context 0
	and trail = Term.make_trail ()
	in
		try
			(			
			Equality.iter_nonvar_subterms_place
				(minimize_eq_subterm
					subterm_table
					equals
					eq_term2
					context_lr
					context_st
					trail)
				eq_term2
				place;	
			eq_term)
		with Found ->
			(if (Term.term_ident eq_term eq_term2) then 
				(minimize_eq_term_right equals subterm_table eq_term)
			else (minimize_eq_term equals subterm_table eq_term2))			
			
and minimize_eq_term_right equals subterm_table eq_term =
	let eq_term2 = Term.copy_term eq_term
	in
	let _, r = Convert.term_pair eq_term2	
	in
	let place = Equality.make_array_place r
	and context_lr = Term.make_context 0
	and context_st = Term.make_context 0
	and trail = Term.make_trail ()
	in
		try
			(			
			Equality.iter_nonvar_subterms_place
				(minimize_eq_subterm
					subterm_table
					equals
					eq_term2
					context_lr
					context_st
					trail)
				eq_term2
				place;	
			eq_term)
		with Found ->
			(if (Term.term_ident eq_term eq_term2) then 
				(eq_term)
			else (minimize_eq_term_right equals subterm_table eq_term2))
						
(****************************************************************************)	      

let check_new_unit clause fact =
  let aux = clause.aux in
  if (Cfg.get_negative_units_flag aux.Aux.cfg) then
    try
      let neg_fact = NegativeUnits.lookup aux.Aux.negative_units_store fact in
      let proof =
	if (Cfg.get_proof_terms_flag aux.Aux.cfg) then
	  (Fact.NegProof (clause.source, 
			 (Array.append
	  			[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
	  			(Array.copy clause.subproofs)),
			 neg_fact))
	else
	  Fact.BuiltinProof in
      refresh_clause clause;	
      raise (BranchClosed proof)
    with Not_found ->
      ()


(* analogous to check_new_unit, but used for units derived by unit-sup-right inferences *)      
let check_new_unit_sup_right clause into_fact fact =
  let aux = clause.aux in
  if (Cfg.get_negative_units_flag aux.Aux.cfg) then
    try
      let neg_fact = NegativeUnits.lookup aux.Aux.negative_units_store fact in
      let proof =
	if (Cfg.get_proof_terms_flag aux.Aux.cfg) then
	  (Fact.NegUnitSupRightProof ((Fact.get_term into_fact),
	  		clause.source,
	  		into_fact,
			(Fact.make_derived_equation_fact_with_proof (get_equation clause) true clause.proof),
			neg_fact))
	else
	  Fact.BuiltinProof in
      refresh_clause clause;	
      raise (BranchClosed proof)
    with Not_found ->
      ()      

(****************************************************************************)

(* E-Hyper calculus: ref-inference *)

(* right now no longer needed in ref_eval, but could be used elsewhere *)
(* A smarter way would be to match clause.head with False and clause.body with [];
requires no symbol table. But as long as the whole function is unnecessary... *)
let is_empty_rule symbol_table term  = 
	if (Convert.term_functor_symbol term)==(Term.get_rule_symbol symbol_table) then
		(let head, body = Convert.term_pair term
		in 
			if (Convert.term_functor_symbol head)==(Term.get_false_symbol symbol_table)
			&& (Convert.term_functor_symbol body)==(Term.get_true_symbol symbol_table) then true
			else (false))
	else false		
		
	
let rec ref_eval_clause clause new_clause_candidates redundant_store = 
	let head_terms =
		Array.concat
			[clause.ground_head_terms;
			clause.non_ground_head_terms;
			clause.ground_head_eqs;
			clause.non_ground_head_eqs]
	in 
	let ref_head =
		(match clause.head with
			Disjunction _ -> 
				Convert.array_of_disj_term_to_term 
					clause.aux.Aux.symbol_table
					head_terms
			|False -> (Term.Constant(Term.get_false_symbol clause.aux.Aux.symbol_table))
			|_ ->	
				Convert.array_of_conj_term_to_term 
					clause.aux.Aux.symbol_table
					head_terms)
	in	
	Array.iteri
		(fun i term ->			
			let l, r = Convert.term_pair term
			and ref_context = Term.make_context 0
			and ref_trail = Term.make_trail ()
			in	
				if Term.failed (Term.unify l ref_context r ref_context ref_trail) then ()
				else (
					let make_proofs = (Cfg.get_proof_terms_flag clause.aux.Aux.cfg)
					in
					let proof = 
						if make_proofs then
							(Fact.EqRefProof
								(clause.source,
								(Fact.make_derived_fact_with_proof clause.source clause.proof)))
						else
							(Fact.UnknownProof)
					in				
					if (match clause.body with
						Body lits ->
							(List.length lits) = 1)
					&&
						(match clause.head with
							False -> raise (BranchClosed proof)
							|Unit lit -> 
								(match (HeadLiteral.get_atom lit) with
									Term term ->
										(let new_term = 
											renumber_variables
												(Term.apply term ref_context)
										in
										let fact =
											(if make_proofs then
												(Fact.make_derived_fact_with_proof new_term proof)
											else
												(Fact.make_derived_fact new_term))
										in
											check_new_unit clause fact;
		  									News.insert_unit
		    									clause.news 
		    									lit.HeadLiteral.predicate 
		    									fact;
											true)
									|_ -> false)
							|_ -> false)
						then ()				
						else (
					let ref_non_ground_body_eqs = Array.copy clause.non_ground_body_eqs
					in
						ref_non_ground_body_eqs.(i) <- Term.Constant (Term.get_true_symbol clause.aux.Aux.symbol_table);
						let ref_body =
							Convert.array_of_conj_term_to_term 
								clause.aux.Aux.symbol_table
								(Array.concat
									[ref_non_ground_body_eqs;
									clause.ground_body_eqs;
									clause.ground_body_terms;
									clause.non_ground_body_terms])
						in 
						let check_rule =
								(Term.Compound
									(Term.make_compound
										(Term.get_rule_symbol clause.aux.Aux.symbol_table)
										[|ref_head; ref_body|]))
						in 
						let ref_rule =
							renumber_variables
								(Term.apply
									check_rule
									ref_context)
						in 
							let ref_clause = make_clause clause.aux ref_rule
							in 		
								ref_clause.proof <- proof;
								if (is_empty_rule clause.aux.Aux.symbol_table ref_rule) then
									(if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
										(raise (BranchClosed (ref_clause.proof)))
									else
										(raise (BranchClosed (Fact.BuiltinProof))));
								NewClauseCandidates.insert_current_stratum_candidate
									new_clause_candidates
									ref_clause;
								ref_eval_clause ref_clause new_clause_candidates redundant_store;
								if 
								(Term.term_ident check_rule ref_rule) || 
								(ref_clause.is_ground && 
								non_ground_to_ground_clauses_lrpo_greater clause ref_clause) then
									(declare_clause_redundant redundant_store clause)
						)
				)
		)
		clause.non_ground_body_eqs;
	
	Array.iteri
		(fun i term -> 
			let l, r = Convert.term_pair term
			in	
				if Term.term_ident l r then (
					let make_proofs = (Cfg.get_proof_terms_flag clause.aux.Aux.cfg)
					in
					let proof = 
						if make_proofs then
							(Fact.EqRefProof
								(clause.source,
								(Fact.make_derived_fact_with_proof clause.source clause.proof)))
						else
							(Fact.UnknownProof)
					in
					
					if (match clause.body with
						Body lits ->
							(List.length lits) = 1)
					&&
						(match clause.head with
							False -> (raise (BranchClosed proof))
							|Unit lit -> 
								(match (HeadLiteral.get_atom lit) with
									Term term ->
										(let fact =
											(if make_proofs then
												(Fact.make_derived_fact_with_proof term proof)
											else
												(Fact.make_derived_fact term))
										in
											check_new_unit clause fact;
		  									News.insert_unit
		    									clause.news 
		    									lit.HeadLiteral.predicate 
		    									fact;
											true)
									|_ -> false)
							|_ -> false)
						then ()
						else (
					let ref_ground_body_eqs = Array.copy clause.ground_body_eqs
					in
						ref_ground_body_eqs.(i) <- Term.Constant (Term.get_true_symbol clause.aux.Aux.symbol_table);
						let ref_body =
							Convert.array_of_conj_term_to_term 
								clause.aux.Aux.symbol_table
								(Array.concat
									[ref_ground_body_eqs;
									clause.non_ground_body_eqs;
									clause.ground_body_terms;
									clause.non_ground_body_terms])
						in
						let ref_rule =
							renumber_variables
								(Term.Compound
									(Term.make_compound
										(Term.get_rule_symbol clause.aux.Aux.symbol_table)
										[|ref_head; ref_body|]))
						in 
							let ref_clause = make_clause clause.aux ref_rule
							in	
							ref_clause.proof <- proof;							
							if (is_empty_rule clause.aux.Aux.symbol_table ref_rule) then
								(if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
									(raise (BranchClosed (ref_clause.proof)))
								else
									(raise (BranchClosed (Fact.BuiltinProof))));								
							NewClauseCandidates.insert_current_stratum_candidate
								new_clause_candidates
								ref_clause;
							ref_eval_clause ref_clause new_clause_candidates redundant_store;
							declare_clause_redundant redundant_store clause
						)
				)
		)
		clause.ground_body_eqs

		
		
		
		

(****************************************************************************)
(* Unit Heads *)
(****************************************************************************)



let indexed_in_extent predicate term =
	try (
		Access.Reader.iter_instance_candidates
			(fun fact -> 
				if (Term.term_ident term (Fact.get_term fact)) then raise Found)
			(Predicate.default_reader predicate)
			term
			Term.null_context;
		false)
	with Found -> true	
	
let nonred_indexed_in_extent predicate term =
	try (
		Access.Reader.iter_instance_candidates
			(fun fact -> 
				if (not (Fact.is_redundant fact))
					&&(Term.term_ident term (Fact.get_term fact)) then raise Found)
			(Predicate.default_reader predicate)
			term
			Term.null_context;
		false)
	with Found -> true		
	
let red_indexed_in_extent predicate term =
	try (
		Access.Reader.iter_instance_candidates
			(fun fact -> 
				if (Fact.is_redundant fact)
					&&(Term.term_ident term (Fact.get_term fact)) then raise Found)
			(Predicate.default_reader predicate)
			term
			Term.null_context;
		false)
	with Found -> true			


(* A function that minimizes a fact based on the current domain units. *)	
exception Found_fact of Fact.fact						

let rec get_reductive_chain_end dom_fact =
	if (Fact.is_redundant dom_fact) then
		(get_reductive_chain_end
			(Fact.get_reducer dom_fact))
	else
		(match (Fact.get_term dom_fact) with
			Term.Compound dk ->
				(Term.compound_args dk).(0)
			|_ -> failwith "Bad domain unit in Ic.minimize_rewrite_fact")
					
and minimize_rewrite_fact clause fact =
	if (Fact.is_ground fact) then
		(let term = Fact.get_term fact
		in
			match term with
				Term.Compound k ->
					(let pred_sym = Convert.term_functor_symbol term
					and krh_dom_sym = Term.get_krh_dom_symbol clause.aux.Aux.symbol_table
					in
					let krh_dom_pred = Predicate.find_predicate clause.aux.Aux.predicate_table krh_dom_sym
					in
					let minimized_args =
						Array.map
							(fun arg ->
								let krh_dom_arg =
									Term.Compound
										(Term.make_compound
											krh_dom_sym
											[|arg|])
								in
									try
										(Access.Reader.iter_instance_candidates
											(fun dom_fact ->
												if (Fact.is_redundant dom_fact) then
													(raise (Found_fact dom_fact)))
											(Predicate.default_reader krh_dom_pred)
											krh_dom_arg
											Term.null_context;
										arg)
									with
										Found_fact dom_fact ->
											get_reductive_chain_end dom_fact)
							(Term.compound_args k)
					in
					let minimized_term =
						Term.Compound
							(Term.make_compound
								pred_sym
								minimized_args)
					in
					let minimized_fact =
						if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
							(Fact.make_derived_fact_with_proof
								minimized_term
								(Fact.get_proof fact))
						else
							(Fact.make_derived_fact minimized_term)
					in
						minimized_fact)	
					
				|_ -> fact)
	else
		(fact)	
		
	
				
exception Reduced_fact of Fact.fact	
exception Reduced_eq_unit of clause


	
let rec do_rewrite_term_unit_sup_right_place place l r term clause fact sup equals =	
	let (rewritten_term, _) = Equality.unit_sup_right_lr_into_term_place l r term place equals sup
	in	
		let rewritten_fact = 
			(if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then 
				(Fact.make_derived_fact_with_proof
					(renumber_variables rewritten_term)
					(Fact.UnitSupRightProof 
						((Fact.get_term fact),
						  clause.source,
						  fact,
						 (Fact.make_derived_equation_fact_with_proof (get_equation clause) true clause.proof))))
			else 
				(Fact.make_derived_fact 
					(renumber_variables rewritten_term)))
		and rewritten_predicate_symbol =
			Convert.term_functor_symbol rewritten_term
		in
(* There shall not be non-ground terms in the Herbrand universe. *)
			if (rewritten_predicate_symbol == (Term.get_krh_dom_symbol clause.aux.Aux.symbol_table))
				&& (not(Termutils.is_ground rewritten_term)) then (raise Failed)
			else 
				(let rewritten_unit_predicate = 
					Predicate.find_predicate
						clause.aux.Aux.predicate_table
						rewritten_predicate_symbol
				in 				
					check_new_unit_sup_right clause fact rewritten_fact;
					(rewritten_unit_predicate, rewritten_fact))	
				
and rewrite_term_unit_sup_right_place place l r term clause fact oriented redundant_store sup equals eq_reader =	
	try 
		(let rewritten_unit_predicate, rewritten_fact =
			(do_rewrite_term_unit_sup_right_place 
				place 
				l r 
				term 
				clause 
				fact 
				sup
				equals)
		in
			if ((Termutils.is_ground term) 
			&& (oriented || LrpOrder.lrpo_greater_unshared_vars term (Fact.get_term rewritten_fact))) then
				(declare_fact_redundant redundant_store fact;
				if (Predicate.get_symbol rewritten_unit_predicate)
				== (Term.get_krh_dom_symbol clause.aux.Aux.symbol_table)
				then
					(Fact.declare_redundant_by_fact fact rewritten_fact));

				let rewritten_fact = minimize_rewrite_fact clause rewritten_fact
				in
				check_new_unit_sup_right clause fact rewritten_fact;	
				News.insert_unit
					clause.news
					rewritten_unit_predicate
					rewritten_fact)
	with Equality.Failed | Failed -> ()
		
and iter_subterms_place f term place =
	f term place;
	match term with
		Term.Compound k ->
			Array.iteri
				(fun index arg ->
					let next_place = Equality.ArrayPlace ((Term.compound_args k), index)
					in 
						iter_subterms_place f arg next_place) 
				(Term.compound_args k)
		|_ -> ()
				
	
	
	
and fact_minimize_rewrite equals eq_reader data_clause sup fact = 
	let fact_reduce_rewrite term place =
		(let access_eq_term =
			Term.Compound
				(Term.make_compound
					equals
					[|term; Term.Variable (Term.make_variable 0)|])	
		and context = Term.make_context 0					
		in
			Reader.iter_unification_candidates
				(fun entry -> (
					match entry with 
						|SubtermEntry.Clause (lr_clause, _, left_is_indexed, _) ->
							(if (not (is_redundant lr_clause)) && lr_clause.is_eq_unit then
								(try
									(let l, r =
										(let lr_eq = get_equation lr_clause
										in
										if left_is_indexed then
											(Equality.get_left lr_eq,
											Equality.get_right lr_eq)
										else
											(Equality.get_right lr_eq,
											Equality.get_left lr_eq))
									in
									let _, reduced_fact =
										(do_rewrite_term_unit_sup_right_place
											place 
											l r 
											term
											lr_clause 
											fact 
											sup
											equals)
									in
										raise (Reduced_fact reduced_fact))
								with Equality.Failed | Failed -> ())
							else ())
						|_ -> ()))
				eq_reader
				access_eq_term
				context)
	in		
	let reduced_fact =
		(if (Fact.is_ground fact) then
			(try
				(let term = Fact.get_term fact
				in 
				(match term with
					Term.Compound k ->
						(Array.iteri
							(fun index arg ->
								let place = Equality.ArrayPlace ((Term.compound_args k), index)
								in
									iter_subterms_place fact_reduce_rewrite arg place)
							(Term.compound_args k))
					|Term.Constant c ->
						(let place = Equality.make_array_place term
						in
							iter_subterms_place fact_reduce_rewrite term place)		
					|_ -> ());
				fact)
			with Reduced_fact reduced_fact ->
				reduced_fact)
		else
			(fact))
	in
		if (fact == reduced_fact) then
			(fact)
		else
			(fact_minimize_rewrite equals eq_reader data_clause sup reduced_fact)
				
				
				
					
(* Similar to minimize_subterm, but intended for redundancy test of delayed equation units. Input is
minimized only when the minimizing input is smaller than the input, thus rendering the input redundant. *)	
let minimize_eq_subterm subterm_table equals mainterm context_lr context_st trail term place =
	let access_eq_term =
		Term.Compound
			(Term.make_compound
				equals
				[|term; Term.Variable (Term.make_variable 0)|])
	and eq_reader =
		get_subterm_reader
			(find_subterm subterm_table equals)
	and context = Term.make_context 0
	in	
		Reader.iter_unification_candidates
			(fun entry ->	
				match entry with 
					SubtermEntry.Clause (clause, _, left_is_indexed, _) ->
						(if clause.is_eq_unit
							&& clause.is_ground 
							&& left_is_indexed 
							&& not (is_redundant clause)
							then (
								let eq =
									match clause.head with
										Unit lit -> 
											(match (HeadLiteral.get_atom lit) with
												Eq eq -> eq
												|_ -> (raise (Err.Error "Bad clause in Ic.minimize_eq_subterm.")))
										|_ -> (raise (Err.Error "Bad clause in Ic.minimize_eq_subterm."))
								in
								let l = Equality.get_left eq
								and r = Equality.get_right eq
								in
									(if (Term.failed (Term.unify term context_st l context_lr trail)) then ()
									else 
										(if (LrpOrder.lrpo_greater mainterm clause.ground_head_eqs.(0)) then
											(Equality.set_place place (Term.apply r context_lr);
											raise Found)
										else ())))
							else ())
					|_ ->())
			eq_reader
			access_eq_term
			context	
		
		
		
		
		
let array_apply a c =
	Array.iteri
		(fun i x ->
			a.(i) <- Term.apply x c)
		a			

let rec do_rewrite_clause_unit_sup_right_place parent_clause place equals l r eq lr_clause left_is_indexed sup =
	let (rewritten_eq_term, _) =
		(Equality.unit_sup_right_lr_into_eq_place l r eq place equals left_is_indexed sup)		
	in
	let rewritten_eq_term = renumber_variables rewritten_eq_term
	in	
		let s2, t2 = Convert.term_pair rewritten_eq_term
		in 
			if (Term.term_ident s2 t2) then (raise Failed)	
			else
				(let rewritten_eq = 
					Equality.make_checked_equation 
						rewritten_eq_term 
							s2 
							t2
				in
				let eq_clause = make_clause parent_clause.aux (Equality.get_term rewritten_eq)
				in	
					if (Cfg.get_proof_terms_flag parent_clause.aux.Aux.cfg) then 
						(let proof =
							(Fact.UnitSupRightProof 
								(parent_clause.source,
								 lr_clause.source,
								 (Fact.make_derived_equation_fact_with_proof (get_equation parent_clause) true parent_clause.proof),
								 (Fact.make_derived_equation_fact_with_proof (get_equation lr_clause) true lr_clause.proof)))
						in
							eq_clause.proof <- proof);
					eq_clause)

	
and rewrite_clause_unit_sup_right_place
	parent_clause place l r lr_clause left_is_indexed new_clause_candidates ground_eq redundant_store sup equals eq_reader =
if (left_is_indexed && lr_clause.is_oriented)||(not(lr_clause.is_oriented)) then (	
(* Pruning: if this condition is not met, then condition 5 of unit-sup-right in calculus cannot be met. *)
	try
		(let eq = get_equation parent_clause
		in
		let eq_term = (Equality.get_term eq)	
		in
		let rewritten_clause =
			(do_rewrite_clause_unit_sup_right_place
				parent_clause 
				place 
				equals
				l r 
				eq_term
				lr_clause 
				left_is_indexed 
				sup)
		in
		let s, t = Convert.term_pair eq_term	(* Equality.get_left eq, Equality.get_right eq *)
		in
			if (parent_clause.is_ground && ground_eq && (LrpOrder.eq_lrpo_greater s t l r)) then
				(declare_clause_redundant redundant_store parent_clause);
(*			let insert_clause = 
					eq_unit_minimize_rewrite 
						equals 
						eq_reader 
						rewritten_clause 
						sup
			in		
*)			NewClauseCandidates.insert_current_stratum_candidate 
				new_clause_candidates 
				rewritten_clause)				
	with Equality.Failed | Failed -> ())
	
and eq_unit_minimize_rewrite equals eq_reader clause sup =			
	let eq_unit_reduce_rewrite term place =
		(let access_eq_term =
			Term.Compound
				(Term.make_compound
					equals
					[|term; Term.Variable (Term.make_variable 0)|])	
		and context = Term.make_context 0					
		in
			Reader.iter_unification_candidates
				(fun entry -> (
					match entry with 
						|SubtermEntry.Clause (lr_clause, _, left_is_indexed, _) ->
							(if (not (is_redundant lr_clause)) && lr_clause.is_eq_unit then
								(try
									(let l, r =
										(let lr_eq = get_equation lr_clause
										in
										if left_is_indexed then
											(Equality.get_left lr_eq,
											Equality.get_right lr_eq)
										else
											(Equality.get_right lr_eq,
											Equality.get_left lr_eq))
									in
									let reduced_eq_unit =
										(do_rewrite_clause_unit_sup_right_place
											clause 
											place 
											equals
											l r 
											clause.ground_head_eqs.(0)
											lr_clause 
											left_is_indexed 
											sup)
									in
										raise (Reduced_eq_unit reduced_eq_unit))
								with Equality.Failed | Failed -> ())
							else ())
						|_ -> ()))
				eq_reader
				access_eq_term
				context)
	in		
	let reduced_eq_unit =
		(if (clause.is_ground) then
			(try
				(let term = clause.ground_head_eqs.(0)
				in
				(match term with
					Term.Compound k ->
						(Array.iteri
							(fun index arg ->
								let place = Equality.ArrayPlace ((Term.compound_args k), index)
								in
									iter_subterms_place eq_unit_reduce_rewrite arg place)
							(Term.compound_args k))
					|Term.Constant c ->
						(let place = Equality.make_array_place term
						in
							iter_subterms_place eq_unit_reduce_rewrite term place)		
					|_ -> ());
				clause)
			with Reduced_eq_unit reduced_eq_unit ->
				reduced_eq_unit)
		else
			(clause))
	in
		if (clause == reduced_eq_unit) then
			(clause)
		else
			(eq_unit_minimize_rewrite equals eq_reader reduced_eq_unit sup)		




let rewrite_clause_unit_sup_right_place_OLD
	parent_clause place l r lr_clause left_is_indexed new_clause_candidates ground_eq redundant_store sup bla blah =
if (left_is_indexed && lr_clause.is_oriented)||(not(lr_clause.is_oriented)) then (	
(* Pruning: if this condition is not met, then condition 5 of unit-sup-right in calculus cannot be met. *)
	let eq = 
		(if parent_clause.is_ground then
			(parent_clause.ground_head_eqs.(0))
		else
			(parent_clause.non_ground_head_eqs.(0)))
	and equals = Term.get_equality_symbol parent_clause.aux.Aux.symbol_table
	in		
	try
		(let (rewritten_eq_term, _) =
			(Equality.unit_sup_right_lr_into_eq_place l r eq place equals left_is_indexed sup)
		in
			let rewritten_eq_term = renumber_variables rewritten_eq_term
			in	
				let s2, t2 = Convert.term_pair rewritten_eq_term
				in 
				if (Term.term_ident s2 t2) then ()	
				else
					(let rewritten_eq = 
							Equality.make_checked_equation 
								rewritten_eq_term 
									s2 
									t2
					and s, t = Convert.term_pair eq
					in	
						if (parent_clause.is_ground && ground_eq && (LrpOrder.eq_lrpo_greater s t l r)) then
							(declare_clause_redundant redundant_store parent_clause);
						if (Term.term_ident s t) then ()
						else						
							(let eq_clause = make_clause parent_clause.aux (Equality.get_term rewritten_eq)
							in	
							
							if (Cfg.get_proof_terms_flag parent_clause.aux.Aux.cfg) then 
								(let proof =
									(Fact.UnitSupRightProof 
										(parent_clause.source,
										 lr_clause.source,
										 (Fact.make_derived_equation_fact_with_proof (get_equation parent_clause) true parent_clause.proof),
										 (Fact.make_derived_equation_fact_with_proof (get_equation lr_clause) true lr_clause.proof)))
								in
									eq_clause.proof <- proof);
							NewClauseCandidates.insert_current_stratum_candidate 
								new_clause_candidates 
								eq_clause
							)))
	with Equality.Failed -> ()
)				
		
			
				
let rewrite_clause_sup_left_place 
		clause place l r lr_clause rewrite_pos b1 b2 b3 b4 left_is_indexed indexed_at same_stratum new_clause_candidates oriented redundant_store sup =
let literal_is_oriented =
	BodyLiteral.get_is_oriented 
		(List.nth
			(match clause.body with
				Body lits -> lits)
			indexed_at) 
in 
if (left_is_indexed && literal_is_oriented)||(not literal_is_oriented) then (
	let equals = (Term.get_equality_symbol clause.aux.Aux.symbol_table)
	in 	 	
		try (	
			let (rewritten_term, context) = 
				(if (Convert.term_functor_symbol b1.(rewrite_pos)) == equals then 
					(Equality.sup_left_lr_into_eq_place 
								l r 
								b1.(rewrite_pos)
								place 
								equals 
								clause.source 
								left_is_indexed
								sup)
				else (Equality.sup_left_lr_into_term_place 
								l r 
								b1.(rewrite_pos)
								place 
								equals 
								clause.source 
								sup))
			and rw_b1 = Array.copy b1
			and rw_b2 = Array.copy b2
			and rw_b3 = Array.copy b3
			and rw_b4 = Array.copy b4
			and h1 = Array.copy clause.ground_head_terms
			and h2 = Array.copy clause.non_ground_head_terms
			and h3 = Array.copy clause.ground_head_eqs
			and h4 = Array.copy clause.non_ground_head_eqs
			in 
				array_apply rw_b1 context;
				array_apply rw_b2 context;
				array_apply rw_b3 context;
				array_apply rw_b4 context;
				array_apply h1 context;	
				array_apply h2 context;
				array_apply h3 context;	
				array_apply h4 context;
				rw_b1.(rewrite_pos) <- rewritten_term;
				let rw_body =
					Convert.array_of_conj_term_to_term 
						clause.aux.Aux.symbol_table
						(Array.concat [rw_b1; rw_b2; rw_b3; rw_b4])
				and rw_head =
					match clause.head with
						False ->
							Term.Constant (Term.get_false_symbol clause.aux.Aux.symbol_table)
						|Disjunction _ ->
							Convert.array_of_disj_term_to_term 
								clause.aux.Aux.symbol_table
								(Array.concat [h1; h2; h3; h4])
						|_ ->
							Convert.array_of_conj_term_to_term 
								clause.aux.Aux.symbol_table
								(Array.concat [h1; h2; h3; h4])
				in 
				let rw_rule = 
					renumber_variables
						(Term.Compound
							(Term.make_compound
								(Term.get_rule_symbol clause.aux.Aux.symbol_table)
								[|rw_head; rw_body|]))
				in 
					let rw_clause = make_clause clause.aux rw_rule
					in 
					if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
						(let proof =
							Fact.SupLeftProof
								(clause.source,
								 lr_clause.source,
								 (Fact.make_derived_fact_with_proof clause.source clause.proof),
								 (Fact.make_derived_equation_fact_with_proof (get_equation lr_clause) true lr_clause.proof))
						in
							rw_clause.proof <- proof);
					if (oriented && clause.is_ground && 
						(clause.has_non_eqs || (lr_clause.is_ground && (ground_clauses_lrpo_greater clause lr_clause))))	
						(* ground clause and reductive superposition 
						leads to smaller clause, making parent redundant *)
						then (declare_clause_redundant redundant_store clause);	
					if same_stratum then
						(NewClauseCandidates.insert_current_stratum_candidate 
							new_clause_candidates 
							rw_clause)
					else
						(NewClauseCandidates.insert_other_stratum_candidate 
							new_clause_candidates 
							rw_clause);
					ref_eval_clause rw_clause new_clause_candidates redundant_store)			
		with Equality.Failed -> ()
)
					
let get_rewriting_data clause i = 
	let gbt_len = Array.length clause.ground_body_terms
	in	
	if i < gbt_len then
		(i,
		clause.ground_body_terms,
		clause.non_ground_body_terms,
		clause.ground_body_eqs,
		clause.non_ground_body_eqs)
	else (
		let ngbt_len = Array.length clause.non_ground_body_terms
		in 
		if i < (gbt_len + ngbt_len) then
			(i - gbt_len,
			clause.non_ground_body_terms,
			clause.ground_body_terms,
			clause.ground_body_eqs,
			clause.non_ground_body_eqs)
		else (
			let gbe_len = Array.length clause.ground_body_eqs
			and gngbt_len = gbt_len + ngbt_len
			in
			if i < (gngbt_len + gbe_len) then
				(i - gngbt_len,
				clause.ground_body_eqs,
				clause.ground_body_terms,
				clause.non_ground_body_terms,
				clause.non_ground_body_eqs)
			else (
				i - gngbt_len - gbe_len,
				clause.non_ground_body_eqs,
				clause.ground_body_terms,
				clause.non_ground_body_terms,
				clause.ground_body_eqs)))
  
				
let print_clause_list clauses =
	List.iter
		(fun clause ->
			PrologPrint.print clause.source;
			print_endline ".")
		clauses	
				    
let process_head_literal clause delta context lit subterm_table clauses new_clause_candidates redundant_store = 
	match (HeadLiteral.get_atom lit) with
		Term term -> 
			let lit_term = renumber_variables (Term.apply (HeadLiteral.get_atom_term lit) context)
			in

		  		let fact = 
		  			(if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
		  				(Fact.make_derived_fact_with_proof
      						lit_term
      						(Fact.Proof (clause.source,
      									Array.append
      										[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
		   									(Array.copy clause.subproofs))))
		   			else
		   				(Fact.make_derived_fact lit_term))
		  		in
		  			check_new_unit clause fact;
(* The following code in comments is an attempt to sequentially rewrite a newly derived fact
as long as possible until it reaches a minimum, given the current set of equations. 		  			
Say [fact_minimize_rewrite ... fact_1] results in [fact_n], then [fact_n] is smaller than
[fact_1],...,[fact_n-1], each of them being redundant. The proof of [fact_n] will include
the previous facts to allow for an unbroken proof chain, but those facts do not get inserted
into the index.
In general the idea seems good, but in practice it slows the operation down, sometimes by a factor
of 10. To reduce a term like this, each of its subterms must be considered, as it might be reducible
by some equation. This means there are a lot of indexing tests to be done in order to find such
equations, even when the term in question turns out to be irreducible. *)
(*		  			let equals = Term.get_equality_symbol clause.aux.Aux.symbol_table
		  			in
		  			let eq_reader = get_subterm_reader
		  								(find_subterm subterm_table equals)
		  			in
		  			let insert_fact = 
		  				fact_minimize_rewrite
		  					equals 
		  					eq_reader 
		  					clause 
		  					(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
		  					fact
		  			in	
*)			  			
						let fact = minimize_rewrite_fact clause fact
						in
						check_new_unit_sup_right clause fact fact;
						News.insert_unit
			    			clause.news 
			    			lit.HeadLiteral.predicate 
			    			fact
	    |Eq eq -> 
	    	let eq_term = renumber_variables (Term.apply (HeadLiteral.get_atom_term lit) context)
	    	in 		
	    	let l_tmp, r_tmp = Convert.term_pair eq_term
	    	in 
	    	let eq = Equality.make_checked_equation eq_term l_tmp r_tmp
	    	in
	    	let equals = Term.get_equality_symbol clause.aux.Aux.symbol_table
	    	in
	    	let eq_reader = get_subterm_reader
	    						(find_subterm subterm_table equals)
	    	in 
(* If the following condition is met, then the currently processed equation head literal has been reached by hyper 
resolution from clause, meaning that clause has one or more body literals and that there is not yet a specific 
clause record for the current equation. In this case the possible superposition inferences for the equation have to be
delayed until next round, because doing otherwise would throw the delta mechanism and strata out of synch, technically 
introducing a new clause into the reasoning in the middle of a round. Instead, the equation will be stored together
with the other new clauses generated this round, and wait for the next round until it is actually used. *)
	    		if (not(clause.is_eq_unit)) then (
	    			let new_eq_clause = make_clause clause.aux eq_term
	    			in
	    				if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
	    					(let proof =
	    						Fact.Proof
	    							(clause.source,
	    							(Array.append
	    								[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
	    								(Array.copy clause.subproofs)))
	    					in
	    						new_eq_clause.proof <- proof);
(*	    				let insert_clause =
	    						eq_unit_minimize_rewrite
	    							equals 
	    							eq_reader 
	    							new_eq_clause 
	    							(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
				  		in		*)
	    				NewClauseCandidates.insert_current_stratum_candidate 
	    					new_clause_candidates 
	    					new_eq_clause)
	    		else (		
(* if lit_term is an equation, try unit-sup-right on all facts in the predicate extent. *)

				let l = Equality.get_left eq
				and r = Equality.get_right eq
				and context = Term.make_context 0
				in			
				if (Termutils.is_nonvar l) then (
					
				let l_sym = Convert.term_functor_symbol l
				in
				let	left_subterm = 
					if delta then
						(get_delta_subterm 
							(find_subterm 
								subterm_table 
								l_sym))
					else
						(find_subterm
							subterm_table
							l_sym)
				in		
(*	superposition inferences, direction left to right	*)	
				Reader.iter_unification_candidates
					(fun entry ->	
						if (SubtermEntry.is_redundant entry) then ()
						else (	
							match entry with
								SubtermEntry.Fact (fact, place) ->
									let term = Fact.get_term fact
									in						
										(rewrite_term_unit_sup_right_place
											place
											l r 
											term 
											clause 
											fact
											(Equality.get_is_oriented eq)
											redundant_store
											(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
											equals
											eq_reader)

								|SubtermEntry.Clause (rule_clause, indexed_at, left_is_indexed, place) -> 
									if (indexed_at < 0) then ()
									else
										(if rule_clause.is_eq_unit then (
											(rewrite_clause_unit_sup_right_place
												rule_clause
												place
												l r
												clause
												left_is_indexed
												new_clause_candidates
												(Equality.get_is_ground eq)
												redundant_store
												(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
												equals
												eq_reader)
										)		
												
												
										else (
												let rewrite_pos, b1, b2, b3, b4 = get_rewriting_data rule_clause indexed_at
												(*and same_stratum = List.memq rule_clause clauses*)
												and same_stratum = rule_clause.is_in_current_stratum
												in 		
													(rewrite_clause_sup_left_place 
														rule_clause 
														place
														l r
														clause 
														rewrite_pos 
														b1 b2 b3 b4 
														left_is_indexed
														indexed_at 
														same_stratum 
														new_clause_candidates 
														(Equality.get_is_oriented eq)
														redundant_store
														(Cfg.get_sup_left_flag clause.aux.Aux.cfg)))											
											)))
					(left_subterm.reader)
					l
					context);
					
(*	superposition inferences, direction right to left	*)
					if (*not (Term.term_ident l r) &&*) not (Equality.get_is_oriented eq) then (
						match r with
							Term.Variable _ ->
								SubtermTable.iter
									(fun symbol subterm ->
										try	(
											Reader.iter_all
												(fun entry -> 
													if (SubtermEntry.is_redundant entry) then ()
													else (	
														match entry with
															SubtermEntry.Fact (fact, place) ->
																	let term = Fact.get_term fact
																	in
																		(rewrite_term_unit_sup_right_place
																			place
																			r l 
																			term 
																			clause 
																			fact
																			(Equality.get_is_oriented eq)
																			redundant_store
																			(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
																			equals
																			eq_reader)
																																									
															|SubtermEntry.Clause (rule_clause, indexed_at, left_is_indexed, place) -> (*PrologPrint.print rule_clause.source; print_string " -->>> ";print_clause_list clauses;*)
																if (indexed_at < 0) then ()
																else
																	(if rule_clause.is_eq_unit then
																		(rewrite_clause_unit_sup_right_place
																			rule_clause
																			place
																			r l
																			clause
																			left_is_indexed
																			new_clause_candidates
																			(Equality.get_is_ground eq)
																			redundant_store
																			(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
																			equals
																			eq_reader)
																
																	else (
																			let rewrite_pos, b1, b2, b3, b4 = get_rewriting_data rule_clause indexed_at
																			(*and same_stratum = List.memq rule_clause clauses*)
																			and same_stratum = rule_clause.is_in_current_stratum
																			in 
																				(rewrite_clause_sup_left_place 
																					rule_clause
																					place 
																					r l 
																					clause
																					rewrite_pos 
																					b1 b2 b3 b4 
																					left_is_indexed
																					indexed_at 
																					same_stratum 
																					new_clause_candidates 
																					(Equality.get_is_oriented eq)
																					redundant_store
																					(Cfg.get_sup_left_flag clause.aux.Aux.cfg))
																	)
																)
													)
												)	
												(subterm.reader))
											with Not_found -> ())
										subterm_table
									
							|_ -> 
								let r_sym = Convert.term_functor_symbol r
								in
								let right_subterm =
										find_subterm
											subterm_table r_sym
								in
									Reader.iter_unification_candidates
										(fun entry ->
											if (SubtermEntry.is_redundant entry) then ()
											else (
												match entry with
													SubtermEntry.Fact (fact, place) ->
														let term = Fact.get_term fact
														in
															(rewrite_term_unit_sup_right_place
																place
																r l 
																term 
																clause 
																fact
																(Equality.get_is_oriented eq)
																redundant_store
																(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
																equals
																eq_reader)
												
													|SubtermEntry.Clause (rule_clause, indexed_at, left_is_indexed, place) -> (*PrologPrint.print rule_clause.source; print_string " -->>> ";print_clause_list clauses;*)
														if (indexed_at < 0) then ()
														else
															(if rule_clause.is_eq_unit then
																(rewrite_clause_unit_sup_right_place
																	rule_clause
																	place
																	r l
																	clause
																	left_is_indexed
																	new_clause_candidates
																	(Equality.get_is_ground eq)
																	redundant_store
																	(Cfg.get_unit_sup_right_flag clause.aux.Aux.cfg)
																	equals
																	eq_reader)
																																	
															else (
																let rewrite_pos, b1, b2, b3, b4 = get_rewriting_data rule_clause indexed_at
																(*and same_stratum = List.memq rule_clause clauses*)
																and same_stratum = rule_clause.is_in_current_stratum
																in 
																	(rewrite_clause_sup_left_place 
																		rule_clause 
																		place
																		r l 
																		clause
																		rewrite_pos 
																		b1 b2 b3 b4 
																		left_is_indexed
																		indexed_at 
																		same_stratum 
																		new_clause_candidates 
																		(Equality.get_is_oriented eq)
																		redundant_store
																		(Cfg.get_sup_left_flag clause.aux.Aux.cfg))
															)
														)
											)
										)
										(right_subterm.reader)
										r
										context)				
	)
	
let is_equation term symbol_table =
	(Convert.term_functor_symbol term) == (Term.get_equality_symbol symbol_table) 
    
let process_head_literal_with_proof clause context lit =
  let fact =
    Fact.make_derived_fact_with_proof
      (renumber_variables (Term.apply (HeadLiteral.get_atom_term lit) context))
      (Fact.Proof (clause.source, 
      		(Array.append
      			[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
		   		(Array.copy clause.subproofs)))) in
  check_new_unit clause fact;
  News.insert_unit
    clause.news 
    lit.HeadLiteral.predicate
    fact


(****************************************************************************)
(* Disjunctive Heads *)

let rec rcu_1 negunits nfs lits =
  match lits with
    [] -> []
  | lit :: lits1 ->
      try
	match lit with
	  (_, fact, _, _, _, _) -> 
	    let nf = NegativeUnits.lookup negunits fact in

	   
	     (* print_endline "RCU"; *)
	     

	    nfs := nf :: !nfs;
	    rcu_1 negunits nfs lits1
      with Not_found ->
	lit :: rcu_1 negunits nfs lits1

let remove_complemented_units negunits lits =
  let nfs = ref [] in
  let lits1 = rcu_1 negunits nfs lits in
  (lits1, !nfs)
    
(** Assuming that sorting had put identical literals in succession. *)
let rec factorize_identical_literals lits =
  match lits with
    l1 :: l2 :: lits1 ->
      begin
	match l1, l2 with
	  (_, f1, _, _, _, _), (_, f2, _, _, _, _) ->
	    if Term.term_ident (Fact.get_term f1) (Fact.get_term f2) then
	      factorize_identical_literals (l2 :: lits1)
	    else
	      l1 :: factorize_identical_literals (l2 :: lits1)
      end
  | _ -> lits

let rec add_complement_splits lits =
  match lits with
    [] -> []
  | (p, f, w, it, sb, ub) :: lits1 ->
      (p, f, w, it, sb, ub, List.map (fun (_, f1, _, _, _, _) -> f1) lits1) ::
      add_complement_splits lits1

			


      
      
let process_disjunctive_head_1 clause context head subterm_table new_clause_candidates redundant_store = (*PrologPrint.print clause.source;print_endline ".";*)
  let aux = clause.aux 
  and cfg = clause.cfg in	
 
  let weighing_method =
    (* ### this dispatch is also in news.ml *)
    (match Cfg.get_weighing_method_param cfg with
      0 -> Weight.SymbolCount
    | 1 -> Weight.Depth
    | _ -> raise
	  (Err.Error "Bad value for term_weighing_method parameter.")) in
  let var_table = Array.make Term.max_vars (-1) 
  and ehyper = Cfg.get_ehyper clause.aux.Aux.cfg
(*  and equals = Term.get_equality_symbol clause.aux.Aux.symbol_table	*)
  and flat_bindings = Term.extract_flat_bindings context
  and impure_varnums = clause.impure_varnums 
  and clause_bindings_table = Array.make Term.max_vars (-1)
  in
	Term.init_renum_table 
		clause_bindings_table
		clause.source;
(*	Array.iter
		(fun t ->
			Term.renumber_variables_term
				clause_bindings_table
				t)
		flat_bindings; 	*)
(*	let tmp_renum_term =
		Term.Compound
			(Term.make_compound
				equals
				flat_bindings)
	in
		Term.renumber_variables_term_specific 
			clause_bindings_table 
			tmp_renum_term;*)
  let pure_bindings, impure_bindings =
  	if ehyper then
  		(let pure_bindings = Array.make Term.max_vars Term.Unbound
  		and impure_bindings = Array.make Term.max_vars Term.Unbound
  		in
  			Array.iteri
  				(fun i term ->
  					if not (List.mem i impure_varnums) then
  						(pure_bindings.(i) <- flat_bindings.(i))
  					else
  						(impure_bindings.(i) <- flat_bindings.(i)))
  				flat_bindings;
  				(pure_bindings, impure_bindings))
  	else ([||], [||])
  	
  	in 
  	let minimized_impure_bindings = impure_bindings
(*  		if ehyper then
	  		(Array.map
	  			(fun term -> 
					minimize_term
						subterm_table
						equals
						term)
				impure_bindings)
		else ([||])
*)
	in 
	let all_bindings =
		if ehyper then
			(let all_bindings = Array.make Term.max_vars Term.Unbound
			in
				Array.iteri
					(fun i term ->
						if (all_bindings.(i) == Term.Unbound) then
							(all_bindings.(i) <- term))
					pure_bindings;
				Array.iteri
					(fun i term ->
						if (all_bindings.(i) == Term.Unbound) then
							(all_bindings.(i) <- term))
					minimized_impure_bindings;
				all_bindings)
		else ([||])
	in
	let context = 
		(if ehyper then
			(Term.make_bound_context all_bindings 0)
		else context)
					
	in
      
  let lits1 =
    List.map
      (fun lit -> 
      	match (HeadLiteral.get_atom lit) with
      		Term impure_term -> 
      			let pure_term = Term.apply impure_term context
      			in 
      				Term.renumber_variables_term var_table pure_term;
      				(lit.HeadLiteral.predicate,
	 				Fact.make_derived_fact pure_term,
	 				(Weight.term_weight weighing_method pure_term),
	 				impure_term,
	 				impure_bindings,
	 				pure_bindings)
      		|Eq eq -> 
      			let impure_term = Equality.get_term eq
      			in
      			let pure_term = Term.apply impure_term context
      			in 
      			Term.renumber_variables_term var_table pure_term;
      			let l, r = Convert.term_pair pure_term
      			in
      			let eq = Equality.make_checked_equation pure_term l r
      			in
      			let eq_l_fact = 
      					(Fact.make_derived_equation_fact eq true)
      			in
      			let eq_term = Equality.get_term eq
      			in	
      				(lit.HeadLiteral.predicate,
      				eq_l_fact,
      				(Weight.term_weight weighing_method	eq_term),
      				impure_term,
      				impure_bindings,
      				pure_bindings))
      (DisjunctiveHead.get_literals head) in      
      
      
    (*  List.iter (fun (_,fk,_,_,_,_) -> ppp (Fact.get_term fk);pe "<DIS") lits1;*)
(*  if (not ehyper) && 
  	Cfg.get_check_disjunctions_are_ground_flag cfg &&
    List.exists
      (fun (_, fact, _, _, _, _) -> 
	not (Termutils.is_ground (Fact.get_term fact)))
      lits1 then
    raise (Err.Error 
	     ("Nonground disjunct derived with clause: "^
	      (PrologPrint.print_to_string clause.source)));
*)

  let (lits2, nfs) =
    if (Cfg.get_negative_units_flag aux.Aux.cfg) then
      remove_complemented_units aux.Aux.negative_units_store lits1
    else
      (lits1, []) in

  let proof =
    if (Cfg.get_proof_terms_flag aux.Aux.cfg) then
      begin 
	aux.Aux.proof_id <- aux.Aux.proof_id + 1;

		if (clause.dom_clause) then (
		Fact.DisjProof (aux.Aux.proof_id,
			clause.parent_source,
			(Array.append
				[|Fact.make_derived_fact_with_proof clause.parent_source clause.proof|]
				(Array.copy clause.parent_subproofs)), 
			nfs
		))
		else (
	Fact.DisjProof (aux.Aux.proof_id,
			clause.source,
			(Array.append
				[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
				(Array.copy clause.subproofs)), 
			nfs)
		)
      end
    else
      Fact.BuiltinProof in
  if lits2 = [] then
    begin
      refresh_clause clause;
      raise (BranchClosed proof)
    end
  else
    let lits2a =
      if (Cfg.get_proof_terms_flag aux.Aux.cfg) then
	List.map
	  (fun (p,f,w,it,ib,pb) -> 
	  	match f with
	  		Fact.DerivedFact _ -> (*ps ">>>>   ";pi aux.Aux.proof_id;ps ": ";ppp (Fact.get_term f);ps " <> ";ppp clause.source; ps "  P: ";Fact.print_proof clause.proof;pe " continued";
	  		Array.iter (fun sup -> Fact.print_proof (Fact.get_proof sup)) clause.subproofs;pe "";pe "";*)
			    (p,
			     Fact.make_derived_fact_with_proof (Fact.get_term f) proof,
			     w,
			     it,
			     ib,
			     pb)
			|Fact.DerivedEquationFact _ ->
				(p,
				Fact.make_derived_equation_fact_with_proof (Fact.get_equation f) true proof,
				w,
				it,
				ib,
				pb)
			|_ -> (raise (Err.Error ("Bad fact in Ic.process_disjunctive_head_1."))))
	  lits2
      else
	lits2
    in
    let lits3 =
      if Cfg.get_sort_heads_flag cfg then
	List.stable_sort
	  (fun (_, _, w1, _, _, _) (_, _, w2, _, _, _) -> compare w1 w2)
	  lits2a
      else
	lits2a in
    let lits4 =
      factorize_identical_literals lits3 in
    match lits4 with
      [(pred, fact, _, _, _, _)] ->	
		(if (Fact.is_equation fact) then
			(match fact with
				Fact.DerivedEquationFact _ ->
					let eq = Fact.get_equation fact
					in
					let new_eq_clause = make_clause aux (Equality.get_term eq)
	    			in	
						new_eq_clause.proof <- proof;
	    				NewClauseCandidates.insert_current_stratum_candidate 
	    					new_clause_candidates 
	    					new_eq_clause
	    		|Fact.DerivedEquationFactWithProof _ ->			
					let eq = Fact.get_equation fact
					and proof = Fact.get_proof fact
					in
					let new_eq_clause = make_clause aux (Equality.get_term eq)
	    			in	
						new_eq_clause.proof <- proof;
	    				NewClauseCandidates.insert_current_stratum_candidate 
	    					new_clause_candidates 
	    					new_eq_clause
	    		|_ -> (raise (Err.Error ("Bad equation fact in Ic.process_disjunctive_head_1."))))	    		
		else 
			(News.insert_unit clause.news pred fact))
    | _ -> 
	let lits5 =
	  if (Cfg.get_negative_units_flag aux.Aux.cfg) && 
	    not (Cfg.get_wp_semantics_flag cfg) then
	    add_complement_splits lits4
	  else
	    List.map (fun (p,t,w,it,ib,pb) -> (p,t,w,it,ib,pb,[])) lits4 in	
(*	    
	    pe "INSERT DISJUNCTION:";
	    List.iter
	    	(fun (_,f,_,_,_,_) ->
	    		ppp (Fact.get_term f);ps "; ";)
	    	lits4;
	    pe ".";
*)	    
	let rejected_eqs = News.insert_disjunction clause.news lits5
	in
		List.iter
			(fun eq_term ->
				let eq_clause = make_clause aux eq_term
				in
					NewClauseCandidates.insert_delayed_candidate new_clause_candidates eq_clause)
			rejected_eqs
	

	
let process_disjunctive_head clause context head subterm_table new_clause_candidates redundant_store =
	let var_table = Array.make Term.max_vars (-1) 
	in
	let head_lits_terms =
		List.map
			(fun lit ->
				let term = Term.apply (HeadLiteral.get_atom_term lit) context
				in
					Term.renumber_variables_term var_table term;
					term)
			(DisjunctiveHead.get_literals head)
	in 
	let shared_varnums =
		Termutils.get_shared_varnums head_lits_terms
	in 
		match shared_varnums with
			[] -> (process_disjunctive_head_1 
						clause 
						context 
						head 
						subterm_table 
						new_clause_candidates 
						redundant_store)
			|_ -> 
				(let krh_dom_symbol = Term.get_krh_dom_symbol clause.aux.Aux.symbol_table
				in 
				let krh_dom_terms =
					List.map
						(fun i ->
							Term.Compound
								(Term.make_compound
									krh_dom_symbol
									[|(Term.Variable (Term.make_variable i))|]))
						shared_varnums
				in
					let new_head_term =
						Convert.array_of_disj_term_to_term
							clause.aux.Aux.symbol_table
							(Array.of_list head_lits_terms)
					and new_body_term =
						Convert.array_of_conj_term_to_term
							clause.aux.Aux.symbol_table
							(Array.of_list krh_dom_terms)
					in
						let new_clause_term =
							Term.Compound
								(Term.make_compound
									(Term.get_rule_symbol clause.aux.Aux.symbol_table)
									[|new_head_term; new_body_term|])
						in
							let new_clause =
								make_clause clause.aux new_clause_term
							in
								if (Cfg.get_proof_terms_flag new_clause.aux.Aux.cfg) then
									(let proof =
										(Fact.HyperPureProof (clause.source,
	      									Array.append
	      										[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
			   									(Array.copy clause.subproofs)))
									in
										new_clause.proof <- proof);

								new_clause.proof <- clause.proof;
								new_clause.parent_subproofs <- Array.copy clause.subproofs;
								new_clause.dom_clause <- true;
								new_clause.parent_source <- clause.source;
								NewClauseCandidates.insert_current_stratum_candidate
									new_clause_candidates
									new_clause)
									
							
	

			
	
(****************************************************************************)
(* Head Processing *)
      
let has_ground_head clause =
	if (Array.length clause.non_ground_head_terms)>0
		||(Array.length clause.non_ground_head_eqs)>0
		then false
	else true

let process_head clause delta context subterm_table clauses new_clause_candidates redundant_store =
(* In ehyper-mode (that is, running with equations) the normal hyperextension processing of hyper tableaux
represents the series of sup-left and ref inferences that resolve away the non-equational body literals of a
clause. This means that at the current point in the program execution (we have "reached" the head), the clause
can be considered redundant (and get blocked) if the following conditions hold:
- we are running in ehyper-mode (no redundancy treatment in regular KRHyper mode),
- the clause is not an equation unit (equation units are special clauses; they have no body and will always
reach the current point, yet they have to remain in the clause processing unlike the similarly bodyless
disjunctive or conjunctive facts (which are split and turned into separate facts or clauses on a per literal
basis, and can therefore possibly be blocked due to redundancy)),
- the clause head is ground; this is the only case where the clauses resulting from the head processing
are smaller (according to the LRPO) than the clause in question. *)

	if (Cfg.get_ehyper clause.cfg) && (not clause.is_eq_unit) && (has_ground_head clause) then
		(declare_clause_redundant redundant_store clause)
	else ();


match clause.head with
    False ->
      let proof = 
	if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
	  Fact.Proof 
	  	(clause.source, 
	  		(Array.append
	  			[|Fact.make_derived_fact_with_proof clause.source clause.proof|]
	  			(Array.copy clause.subproofs)))
	else
	  Fact.BuiltinProof in
      refresh_clause clause;
      raise (BranchClosed proof)
  | Unit lit ->	
     (* if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
	process_head_literal_with_proof clause context lit
      else *) 
      	(process_head_literal clause delta context lit subterm_table clauses new_clause_candidates redundant_store)
  | Conjunction lits ->
      if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) then
	List.iter
	  (fun lit -> process_head_literal_with_proof clause context lit)
	  lits
      else
	List.iter
	  (fun lit -> process_head_literal clause delta context lit subterm_table clauses new_clause_candidates redundant_store)
	  lits
  | Disjunction head ->
      (* TODO: ### Variant with proofs *)
      process_disjunctive_head clause context head subterm_table new_clause_candidates redundant_store
	

      
(****************************************************************************)
(* Clause Evaluation *)



let set_subproof clause lit entry =
  clause.subproofs.(BodyLiteral.get_subproof_pos(lit)) <- entry
      
let rec eval_clause clause delta subterm_table clauses new_clause_candidates redundant_store = 
  	if (is_redundant clause) then ()
  	else ( 
		let context = Term.make_context 0
	  	and trail = Term.make_trail ()
	  	and literals = match clause.body with Body lits -> lits in
	  		if (not delta) && (Cfg.get_ehyper clause.aux.Aux.cfg) then
	  			(ref_eval_clause clause new_clause_candidates redundant_store);	(*ppp clause.source;pe ".";*)
	  		if (is_redundant clause) then ()
	  		else (eval_clause_1
	  				clause 
	  				delta 
	  				context 
	  				literals 
	  				trail 
	  				subterm_table 
	  				clauses 
	  				new_clause_candidates
	  				redundant_store)
	
	)
	
and eval_clause_1 clause delta context1 lits trail subterm_table clauses new_clause_candidates redundant_store =
  match lits with
    [] -> 	
      	process_head clause delta context1 subterm_table clauses new_clause_candidates redundant_store
  |	lit :: lits1 ->
      match lit.BodyLiteral.builtin with
	Term.PUnbound -> 
		let term = BodyLiteral.get_atom_term lit
		in
	  Access.Reader.iter_unification_candidates
	    (fun entry -> 
	    	if (Fact.is_redundant entry) then ()
	    	else ( 
	      let saved_trail=Term.save_trail trail in
	      let trail1 =
		Term.unify
		  term
		  context1 
		  (Fact.get_term entry)
		  lit.BodyLiteral.foreign_context
		  trail in 								
		  if Term.failed trail1 then
			()
	      else
		begin
		  if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg)&&(not(clause.dom_clause)) then
		    set_subproof clause lit entry
		  else
		    ();
		  eval_clause_1 clause delta context1 lits1 trail1 subterm_table clauses new_clause_candidates redundant_store;
		  ignore (Term.undo_bindings trail1 saved_trail)
		end))
	    lit.BodyLiteral.reader
	    term
	    context1;

	(match (BodyLiteral.get_atom lit) with
		Eq eq ->
			let flip_term = Term.binary_flip term
			in 
		Access.Reader.iter_unification_candidates
	    (fun entry -> 
	    	if (Fact.is_redundant entry) then ()
	    	else (
	      let saved_trail=Term.save_trail trail in
	      let trail1 =
		Term.unify
		  flip_term
		  context1 
		  (Fact.get_term entry)
		  lit.BodyLiteral.foreign_context
		  trail in 
	      if Term.failed trail1 then
			() 
	      else
		begin
		  if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg)&&(not(clause.dom_clause)) then
		    set_subproof clause lit entry
		  else
		    ();
		  eval_clause_1 clause delta context1 lits1 trail1 subterm_table clauses new_clause_candidates redundant_store;
		  ignore (Term.undo_bindings trail1 saved_trail)
		end))
	    lit.BodyLiteral.reader
	    flip_term
	    context1

		|_ -> ())
      | Term.PDeterministic (f, (_, _, opt_proof)) ->
	  let saved_trail=Term.save_trail trail in
(*	  ppp (BodyLiteral.get_atom_term lit);pe " <<< ";*)
	  let trail1 =
	    f
	      (BodyLiteral.get_atom_term lit)
	      context1
	      lit.BodyLiteral.foreign_context
	      trail
	  in
	  if Term.failed trail1 then
	    ()
	  else
	    begin
	      if (Cfg.get_proof_terms_flag clause.aux.Aux.cfg) &&
		opt_proof = Opt.UnknownProof then
		set_subproof clause lit Fact.fact_with_unknown_proof
	      else
		();
	      eval_clause_1 clause delta context1 lits1 trail1 subterm_table clauses new_clause_candidates redundant_store;
	      ignore (Term.undo_bindings trail1 saved_trail)
	    end

					
	    
	    
	    
	    
let np_subsumption_processing_unit full_branch_redundancy redundant_store predicate fact =
	let term = (Fact.get_term fact)
	in
		if (News.unit_is_subsumed_by_extent predicate term) then (raise Redundant)
		else 
			(let iter_instance_candidates =
				if full_branch_redundancy then
					Access.Reader.iter_instance_candidates
				else
					Access.Reader.iter_instance_candidates_ml		
			in	
			iter_instance_candidates
		    	(fun old_fact -> 
		      		if (not (Fact.is_redundant old_fact))
		      			&& (Term.subsumes_chk term (Fact.get_term old_fact)) then
		      			(declare_fact_redundant redundant_store old_fact))
		    	(Predicate.default_reader predicate)
		    	term
		    	Term.null_context)
		    	
let minimize_clause clause subterm_table =
	if clause.is_ground then
		(let equals =
			(Term.get_equality_symbol clause.aux.Aux.symbol_table)
		in
		if clause.is_eq_unit then
			(let eq_term = clause.ground_head_eqs.(0)
			in
			let minimized_eq_term = minimize_eq_term equals subterm_table eq_term
			in
				if (Term.term_ident eq_term minimized_eq_term) then
					(clause)
				else
					(make_clause clause.aux minimized_eq_term))
		else
			(let minimized_ground_body_terms =
				Array.map
					(fun term ->
						minimize_literal_term equals subterm_table term)
					clause.ground_body_terms
			and minimized_ground_body_eqs =
				Array.map
					(fun eq_term ->
						minimize_eq_term equals subterm_table eq_term)
					clause.ground_body_eqs
			in
			let non_trivial_minimized_ground_body_eqs =
				Array.of_list
					(let tmp = ref []
					in
						Array.iter
							(fun eq_term ->
								let l, r = Convert.term_pair eq_term
								in
									if (Term.term_ident l r) then ()
									else
										(tmp := eq_term :: !tmp))
							minimized_ground_body_eqs;
						!tmp)
			and ground_head_terms_eqs =
				Array.append
					clause.ground_head_terms
					clause.ground_head_eqs
			in
				let head_term =
					match clause.head with
						Disjunction _ ->
							Convert.array_of_disj_term_to_term 
								clause.aux.Aux.symbol_table
								ground_head_terms_eqs
						|_ ->
							Convert.array_of_conj_term_to_term 
								clause.aux.Aux.symbol_table
								ground_head_terms_eqs	
				and body_term =
					Convert.array_of_conj_term_to_term 
						clause.aux.Aux.symbol_table
						(Array.append
							minimized_ground_body_terms
							non_trivial_minimized_ground_body_eqs)
				in
					let new_clause_term =
						Term.Compound
							(Term.make_compound
								(Term.get_equality_symbol clause.aux.Aux.symbol_table)
								[|head_term; body_term|])
					in
						(make_clause clause.aux new_clause_term)))
	else
		(clause)
