/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "StandardKBOrdering.hpp"
#include "TermWeighting.hpp"
#include "Unifier.hpp"
#include "ForwardMatchingSubstitutionCache.hpp"
#include "DestructionMode.hpp"
#include "Unification.hpp"
#include "InstanceWithUnifier.hpp"
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering"
#endif
#include "debugMacros.hpp"
using namespace VK;
using namespace BK;
StandardKBOrdering::StandardKBOrdering() : 
  _specialisedComparisonPool(0),
  _constraintCheckOnUnifierPool(0),
  _constraintCheckOnForwardMatchingSubstPool(0),
  _compiledConstraintOnUnifiersPool(0),
  _compiledConstraintForInstanceRetrievalPool(0),
  _numberOfCompiledConstraintsForInstanceRetrieval(1UL),
  _numberOfConstraintChecksForInstanceRetrieval(1UL),
  _numberOfCompiledConstraintsOnUnifiers(1UL),
  _numberOfConstraintChecksOnUnifiers(1UL)
{    
  setType(SimplificationOrdering::StandardKB);
};
StandardKBOrdering::~StandardKBOrdering()
{
  CALL("destructor ~StandardKBOrdering()");
  if (DestructionMode::isThorough()) 
    {
      while (_specialisedComparisonPool)
	{
	  SpecialisedComparison* tmp = _specialisedComparisonPool;
	  _specialisedComparisonPool = _specialisedComparisonPool->getNextInPool();
	  delete tmp;
	};
      while (_constraintCheckOnUnifierPool)
	{
	  ConstraintCheckOnUnifier* tmp = _constraintCheckOnUnifierPool;
	  _constraintCheckOnUnifierPool = _constraintCheckOnUnifierPool->getNextInPool();
	  delete tmp;
	};
      while (_constraintCheckOnForwardMatchingSubstPool)
	{
	  ConstraintCheckOnForwardMatchingSubst* tmp = _constraintCheckOnForwardMatchingSubstPool;
	  _constraintCheckOnForwardMatchingSubstPool = _constraintCheckOnForwardMatchingSubstPool->getNextInPool();
	  delete tmp;
	};
      while (_compiledConstraintOnUnifiersPool)
	{
	  CompiledConstraintForUnifiers* tmp = _compiledConstraintOnUnifiersPool;
	  _compiledConstraintOnUnifiersPool = _compiledConstraintOnUnifiersPool->getNextInPool();      
	  delete tmp;
	};
      while (_compiledConstraintForInstanceRetrievalPool)
	{
	  CompiledConstraintForInstanceRetrieval* tmp = _compiledConstraintForInstanceRetrievalPool;
	  _compiledConstraintForInstanceRetrievalPool = _compiledConstraintForInstanceRetrievalPool->getNextInPool();  
	  delete tmp;
	};  
    };
}; 
void* StandardKBOrdering::operator new(size_t)
{
  return GlobAlloc::allocate(sizeof(StandardKBOrdering));
};
void StandardKBOrdering::operator delete(void* obj)
{
  GlobAlloc::deallocate(obj,sizeof(StandardKBOrdering));
}; 
Comparison StandardKBOrdering::compare(const Flatterm* term1,
				       const WeightPolynomial& weight1,
				       const Flatterm* term2,
				       const WeightPolynomial& weight2)
{
  CALL("compare(..)");
  Comparison lexCmp;
  switch (weight1.compare(weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexicographically(term1,term2);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexicographically(term1,term2);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexicographically(term1,term2);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison StandardKBOrdering::compareAtoms(const Flatterm* lit1,
					    const WeightPolynomial& weight1,
					    const Flatterm* lit2,
					    const WeightPolynomial& weight2)
{
  CALL("compareAtoms(..)");
  Comparison lexCmp;
  switch (weight1.compare(weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareAtomsLexicographically(lit1,lit2);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareAtomsLexicographically(lit1,lit2);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareAtomsLexicographically(lit1,lit2);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison StandardKBOrdering::compareOnCurrentUnifier(UnifTerm* term1,
						       UnifTerm* term2)
{
  CALL("compareOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,term1);
  TermWeighting::current()->collectWeight(_weight2,term2);
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexOnCurrentUnifier(term1,term2);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexOnCurrentUnifier(term1,term2);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexOnCurrentUnifier(term1,term2);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison 
StandardKBOrdering::compareOnCurrentUnifier(const Term& term1,
					    const Term& term2,
					    ulong variableIndex)
{
  CALL("compareOnCurrentUnifier(const Term& term1,const Term& term2,ulong variableIndex)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,term1,variableIndex);
  TermWeighting::current()->collectWeight(_weight2,term2,variableIndex);
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexOnCurrentUnifier(term1,term2,variableIndex);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexOnCurrentUnifier(term1,term2,variableIndex);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexOnCurrentUnifier(term1,term2,variableIndex);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison StandardKBOrdering::compare(const Term* complexTerm1,
				       const Term* complexTerm2,
				       const InstRet::Substitution* subst)
{
  CALL("compare(const Term* complexTerm1,const Term* complexTerm2,const InstRet::Substitution* subst)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,complexTerm1,subst);
  TermWeighting::current()->collectWeight(_weight2,complexTerm2,subst);
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexicographically(complexTerm1,complexTerm2,subst);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexicographically(complexTerm1,complexTerm2,subst);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexicographically(complexTerm1,complexTerm2,subst);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison StandardKBOrdering::compare(const Term& term1,
				       const Term& term2,
				       const ForwardMatchingSubstitutionCache* subst)
{
  CALL("compare(const Term& term1,const Term& term2,const ForwardMatchingSubstitutionCache* subst)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,term1,subst);
  TermWeighting::current()->collectWeight(_weight2,term2,subst);
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexicographically(term1,term2,subst);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexicographically(term1,term2,subst);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexicographically(term1,term2,subst);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
SimplificationOrdering::SpecialisedComparison* 
StandardKBOrdering::reserveSpecialisedComparison()
{
  CALL("reserveSpecialisedComparison()");
  SpecialisedComparison* result = _specialisedComparisonPool;
  if (result)
    {
      _specialisedComparisonPool = _specialisedComparisonPool->getNextInPool();
    }
  else
    {
      result = new SpecialisedComparison(this);
    };
  return 
    static_cast<SimplificationOrdering::SpecialisedComparison*>(result);
}; 
void StandardKBOrdering::releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("releaseSpecialisedComparison(SimplificationOrdering::SpecialisedComparison* sc)");
  (static_cast<StandardKBOrdering::SpecialisedComparison*>(sc))->setNextInPool(_specialisedComparisonPool);
  _specialisedComparisonPool = static_cast<StandardKBOrdering::SpecialisedComparison*>(sc);
}; 
SimplificationOrdering::StoredConstraint*
StandardKBOrdering::storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintGreater(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintGreater();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
StandardKBOrdering::storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintGreaterOrEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintGreaterOrEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
StandardKBOrdering::storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintLess(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintLess();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
StandardKBOrdering::storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintLessOrEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintLessOrEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint*
StandardKBOrdering::storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc)
{
  CALL("storedConstraintNotEqual(SimplificationOrdering::SpecialisedComparison* sc)");
  StoredConstraint* result = (static_cast<SpecialisedComparison*>(sc))->storedConstraintNotEqual();
  result->incReferenceCounter();
  ASSERT(result->referenceCounter() >= 2L);
  return 
    static_cast<SimplificationOrdering::StoredConstraint*>(static_cast<void*>(result));
}; 
SimplificationOrdering::StoredConstraint* StandardKBOrdering::copy(SimplificationOrdering::StoredConstraint* constr)
{
  CALL("copy(SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(constr);
  (static_cast<StoredConstraint*>(static_cast<void*>(constr)))->incReferenceCounter();
  return constr;
}; 
void StandardKBOrdering::releaseConstraint(SimplificationOrdering::StoredConstraint* constr)
{
  CALL("releaseConstraint(SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(constr);
  (static_cast<StoredConstraint*>(static_cast<void*>(constr)))->decReferenceCounter();
  if (!(static_cast<StoredConstraint*>(static_cast<void*>(constr)))->referenceCounter())
    delete (static_cast<StoredConstraint*>(static_cast<void*>(constr)));
}; 
SimplificationOrdering::ConstraintCheckOnUnifier* 
StandardKBOrdering::reserveConstraintCheckOnUnifier()
{
  CALL("reserveConstraintCheckOnUnifier()");  
  ConstraintCheckOnUnifier* result = _constraintCheckOnUnifierPool;
  if (result)
    {
      _constraintCheckOnUnifierPool = _constraintCheckOnUnifierPool->getNextInPool();
    }
  else
    {
      result = new ConstraintCheckOnUnifier(this);
    };
  return 
    static_cast<SimplificationOrdering::ConstraintCheckOnUnifier*>(result);
}; 
void StandardKBOrdering::releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu)
{
  CALL("releaseConstraintCheckOnUnifier(SimplificationOrdering::ConstraintCheckOnUnifier* ccu)");
  (static_cast<StandardKBOrdering::ConstraintCheckOnUnifier*>(ccu))->setNextInPool(_constraintCheckOnUnifierPool);
  _constraintCheckOnUnifierPool = static_cast<StandardKBOrdering::ConstraintCheckOnUnifier*>(ccu);
}; 
SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* 
StandardKBOrdering::reserveConstraintCheckOnForwardMatchingSubst()
{
  CALL("reserveConstraintCheckOnForwardMatchingSubst()");
  ConstraintCheckOnForwardMatchingSubst* result = _constraintCheckOnForwardMatchingSubstPool;
  if (result)
    {
      _constraintCheckOnForwardMatchingSubstPool = 
	_constraintCheckOnForwardMatchingSubstPool->getNextInPool();
    }
  else
    {
      result = new ConstraintCheckOnForwardMatchingSubst(this);
    };
  return 
    static_cast<SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst*>(result);
}; 
void 
StandardKBOrdering::releaseConstraintCheckOnForwardMatchingSubst(SimplificationOrdering::ConstraintCheckOnForwardMatchingSubst* ccfms)
{
  CALL("releaseConstraintCheckOnForwardMatchingSubst(..)");
  (static_cast<StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst*>(ccfms))->
    setNextInPool(_constraintCheckOnForwardMatchingSubstPool);
  _constraintCheckOnForwardMatchingSubstPool = 
    static_cast<StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst*>(ccfms);
}; 
SimplificationOrdering::CompiledConstraintForUnifiers* 
StandardKBOrdering::reserveCompiledConstraintForUnifiers()
{
  CALL("reserveCompiledConstraintForUnifiers()");
  StandardKBOrdering::CompiledConstraintForUnifiers* result = _compiledConstraintOnUnifiersPool;
  if (result)
    {
      _compiledConstraintOnUnifiersPool = result->getNextInPool();
    }
  else
    {
      result = new CompiledConstraintForUnifiers(this);
    };
  ASSERT(result);
  return 
    static_cast<SimplificationOrdering::CompiledConstraintForUnifiers*>(result);
}; 
void 
StandardKBOrdering::releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc)
{
  CALL("releaseCompiledConstraintForUnifiers(SimplificationOrdering::CompiledConstraintForUnifiers* cc)");
  (static_cast<StandardKBOrdering::CompiledConstraintForUnifiers*>(cc))->setNextInPool(_compiledConstraintOnUnifiersPool);
  _compiledConstraintOnUnifiersPool = static_cast<StandardKBOrdering::CompiledConstraintForUnifiers*>(cc);
}; 
SimplificationOrdering::CompiledConstraintForInstanceRetrieval* 
StandardKBOrdering::reserveCompiledConstraintForInstanceRetrieval()
{
  CALL("reserveCompiledConstraintForInstanceRetrieval()");
  StandardKBOrdering::CompiledConstraintForInstanceRetrieval* result = _compiledConstraintForInstanceRetrievalPool;
  if (result)
    {
      _compiledConstraintForInstanceRetrievalPool = result->getNextInPool();
    }
  else
    {
      result = new CompiledConstraintForInstanceRetrieval(this);
    };
  ASSERT(result);
  return
    static_cast<SimplificationOrdering::CompiledConstraintForInstanceRetrieval*>(result);
}; 
void 
StandardKBOrdering::releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc)
{
  CALL("releaseCompiledConstraintForInstanceRetrieval(SimplificationOrdering::CompiledConstraintForInstanceRetrieval* cc)");
  (static_cast<StandardKBOrdering::CompiledConstraintForInstanceRetrieval*>(cc))->setNextInPool(_compiledConstraintForInstanceRetrievalPool);
  _compiledConstraintForInstanceRetrievalPool = static_cast<StandardKBOrdering::CompiledConstraintForInstanceRetrieval*>(cc);
}; 
Comparison StandardKBOrdering::compareWithTransposition(const Flatterm* term)
{  
  CALL("compareWithTransposition(const Flatterm* term)");
 ASSERT(term->isComplex());
 ASSERT(term->arity() == 2);
 const Flatterm* arg1 = term->next();
 const Flatterm* arg2 = arg1->after();
 return compare(arg1,arg2);
}; 
inline
Comparison StandardKBOrdering::compare(const Flatterm* term1,
				       const Flatterm* term2)
{
  CALL("compare(const Flatterm* term1,const Flatterm* term2)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,term1);
  TermWeighting::current()->collectWeight(_weight2,term2);
  return compare(term1,_weight1,term2,_weight2);
}; 
inline
Comparison 
StandardKBOrdering::compareLexicographically(const Flatterm* term1,
					     const Flatterm* term2)
{
  CALL("compareLexicographically(const Flatterm* term1,..)");
  if (term1->symbolRef().isVariable())
    {
      if (term1->symbolRef() == term2->symbolRef()) return Equal;
      if (term2->containsSymbol(term1->symbolRef()))
	{
	  return Less;
	};
      return Incomparable;
    }
  else
    if (term2->symbolRef().isVariable())
      {
        if (term1->containsSymbol(term2->symbolRef()))
	  {
	    return Greater;
	  };
	return Incomparable;
      }
    else
      {
	if (term1->symbolRef() != term2->symbolRef())
	  return term1->symbolRef().comparePrecedence(term2->symbolRef());
	Flatterm::ConstIterator iter1(term1);
	Flatterm::ConstIterator iter2(term2);
	iter1.NextSym();
	iter2.NextSym();
        while (iter1.NotEndOfTerm())
	  {
	    Comparison cmp = compare(iter1.CurrPos(),iter2.CurrPos());
	    if (cmp != Equal) return cmp;
	    iter1.after();
	    iter2.after();
	  };
	ASSERT(iter2.EndOfTerm());
	return Equal;
      };
}; 
inline
Comparison 
StandardKBOrdering::compareAtomsLexicographically(const Flatterm* lit1,
						  const Flatterm* lit2)
{
  CALL("compareAtomsLexicographically(..)");
  if (lit1->functor() != lit2->functor()) 
    {
      return lit1->symbolRef().comparePrecedence(lit2->symbolRef());
    };
  Flatterm::ConstIterator iter1(lit1);
  Flatterm::ConstIterator iter2(lit2);
  iter1.NextSym();
  iter2.NextSym();
  while (iter1.NotEndOfTerm())
    {
      Comparison cmp = compare(iter1.CurrPos(),iter2.CurrPos());
      if (cmp != Equal) return cmp;
      iter1.after();
      iter2.after();
    };
  ASSERT(iter2.EndOfTerm());
  return Equal;
}; 
inline
Comparison StandardKBOrdering::compareLexicographically(const Term* complexTerm1,
							const Term* complexTerm2)
{
  CALL("compareLexicographically(const Term* complexTerm1,const Term* complexTerm2)");
  if ((*complexTerm1) != (*complexTerm2)) 
    return complexTerm1->comparePrecedence(*complexTerm2);
  const Term* arg1 = complexTerm1->Args();
  const Term* arg2 = complexTerm2->Args();
  while (!arg1->isNotArg())
    {
      Comparison cmp = 
	compare(*arg1,*arg2);
      if (cmp != Equal) return cmp;
      arg1 = arg1->Next();
      arg2 = arg2->Next();
    };
  return Equal;
}; 
inline
Comparison StandardKBOrdering::compareLexicographically(const Term& term1,
							const Term& term2)
{
  CALL("compareLexicographically(const Term& term1,..)");
  if (term1 == term2) return Equal;
  if (term1.isVariable())
    {
      if (term2.isVariable()) return Incomparable;
      if (term2.deref()->containsSymbol(term1)) return Less;
      return Incomparable;
    }
  else
    if (term2.isVariable())
      {
	if (term1.deref()->containsSymbol(term2)) return Greater;	
	return Incomparable;
      }
    else 
      {
	ASSERT(term1.IsReference() && term2.IsReference());
	return compareLexicographically(term1.deref(),term2.deref());
      };
}; 
inline
bool 
StandardKBOrdering::termWithUnifierContainsVariable(UnifTerm* complexTerm,
						    Unifier::Variable* var)
{ 
  CALL("termWithUnifierContainsVariable(UnifTerm* complexTerm,Unifier::Variable* var)");
  UnifTerm::CorrectBracket(complexTerm);
  ObjectPool<UnifTerm::Iter>::EntryHandle 
    iter(static_cast<ObjectPool<UnifTerm::Iter>&>(_prefixIterPool));
  iter->Reset(complexTerm);
  iter->SkipSym();
  while (iter->NotEmpty())
    {
      if (iter->CurrVar() == var) return true;
      iter->SkipSym();
    };
  return false;
}; 
inline
Comparison 
StandardKBOrdering::compareLexOnCurrentUnifier(UnifTerm* term1,
					       UnifTerm* term2)
{
  CALL("compareLexOnCurrentUnifier(UnifTerm* term1,UnifTerm* term2)");
  if (term1->symbol().isVariable())
    {
      bool freeVar1;
      void* instance1 = term1->symbol().var()->Unref(freeVar1);
      if (freeVar1)
	{
	  if (term2->symbol().isVariable())
	    {
	      bool freeVar2;
	      void* instance2 = term2->symbol().var()->Unref(freeVar2);
	      if (freeVar2)
		{
		  if (instance1 == instance2) return Equal;
		};
	    };
	  return Incomparable;
	};
      term1 = static_cast<UnifTerm*>(instance1);	
    };
  ASSERT(term1->symbol().IsFunctor());
  if (term2->symbol().isVariable())
    {
      bool freeVar2;
      void* instance2 = term2->symbol().var()->Unref(freeVar2);
      if (freeVar2) return Incomparable;
      term2 = static_cast<UnifTerm*>(instance2);
    };
  ASSERT(term2->symbol().IsFunctor());
  if (term1->symbol() != term2->symbol())
    return term1->symbol().Func().comparePrecedence(term2->symbol().Func());  
  if (!term1->symbol().Func().arity()) return Equal;
  ObjectPool<UnifTerm::Iter>::EntryHandle iter1(_prefixIterPool);
  ObjectPool<UnifTerm::Iter>::EntryHandle iter2(_prefixIterPool);
  UnifTerm::CorrectBracket(term1);
  UnifTerm::CorrectBracket(term2);
  iter1->Reset(term1);
  iter2->Reset(term2);
  iter1->SkipSym();
  iter2->SkipSym();
  do
    {
      Unifier::Variable* var1 = iter1->CurrVar();
      Unifier::Variable* var2 = iter2->CurrVar();
      if (var1)
	{
	  if (var2)
	    {
	      if (var1 != var2) return Incomparable;
	      iter1->SkipSym();
	      iter2->SkipSym();
	    }
	  else
	    {
	      if (termWithUnifierContainsVariable(iter2->CurrSubterm(),var1))
		return Less;
	      return Incomparable;
	    };
	}
      else
	if (var2)
	  {
	    if (termWithUnifierContainsVariable(iter1->CurrSubterm(),var2))
	      return Greater;
	    return Incomparable;
	  }
	else 
	  {
	    UnifTerm::CorrectBracket(iter1->CurrSubterm());
	    UnifTerm::CorrectBracket(iter2->CurrSubterm());
	    Comparison cmp = compareOnCurrentUnifier(iter1->CurrSubterm(),
						     iter2->CurrSubterm());
	    if (cmp != Equal) return cmp;
	    iter1->SkipTerm();
	    iter2->SkipTerm();
	  };
    }
  while (iter1->NotEmpty());
  ASSERT(iter2->Empty());
  return Equal;
}; 
inline
bool 
StandardKBOrdering::termWithUnifierContainsVariable(const Term* complexTerm,
						    ulong variableIndex,
						    Unifier::Variable* var)
{
  CALL("termWithUnifierContainsVariable(const Term* complexTerm,ulong variableIndex,Unifier::Variable* var)");
  ObjectPool<InstanceWithUnifier>::EntryHandle iter(_stTermWithSubstPool);
  iter->Reset(variableIndex,complexTerm);
  iter->SkipSym();
  while (!iter->Empty())
    {
      if (iter->CurrSym().isVariable() &&
	  (iter->CurrSym().var() == var))
	return true;
      iter->SkipSym();
    };
  return false;
}; 
inline
bool 
StandardKBOrdering::termWithUnifierContainsVariable(const Term& term,
						    ulong variableIndex,
						    Unifier::Variable* var)
{
  CALL("termWithUnifierContainsVariable(const Term& term,ulong variableIndex,Unifier::Variable* var)");
  if (term.isVariable())
    {
      Unifier::Variable* var = 
	Unifier::current()->variable(term.var(),variableIndex);
      bool freeVar;
      void* instance = var->Unref(freeVar);
      if (freeVar)
	{
	  return (static_cast<Unifier::Variable*>(instance)) == var;
	}
      else
	{
	  return termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance),var);
	};
    } 
  else
    {
      return termWithUnifierContainsVariable(term.deref(),
					     variableIndex,
					     var);
    };
}; 
inline
Comparison 
StandardKBOrdering::compareOnCurrentUnifier(UnifTerm* complexTerm1,
					    const Term* complexTerm2,
					    ulong variableIndex)
{
  CALL("compareOnCurrentUnifier(UnifTerm* complexTerm1,const Term* complexTerm2,ulong variableIndex)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,complexTerm1);
  TermWeighting::current()->collectWeight(_weight2,
					  Term(const_cast<Term*>(complexTerm2)),
					  variableIndex);
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexOnCurrentUnifier(complexTerm1,complexTerm2,variableIndex);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexOnCurrentUnifier(complexTerm1,complexTerm2,variableIndex);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexOnCurrentUnifier(complexTerm1,complexTerm2,variableIndex);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
inline
Comparison 
StandardKBOrdering::compareLexOnCurrentUnifier(UnifTerm* complexTerm1,
					       const Term* complexTerm2,
					       ulong variableIndex)
{
  CALL("compareLexOnCurrentUnifier(UnifTerm* complexTerm1,const Term* complexTerm2,ulong variableIndex)");
  if (complexTerm1->symbol().Func() != (*complexTerm2))
    return complexTerm1->symbol().Func().comparePrecedence(*complexTerm2);
  UnifTerm::CorrectBracket(complexTerm1);
  ObjectPool<UnifTerm::Iter>::EntryHandle iter1(_prefixIterPool);
  iter1->Reset(complexTerm1);
  iter1->SkipSym();
  const Term* arg2 = complexTerm2->Args();
  while (iter1->NotEmpty())
    {
      Unifier::Variable* var1 = iter1->CurrVar();
      if (var1)
	{
          if (arg2->isVariable())
	    {
	      Unifier::Variable* var2 = 
		Unifier::current()->variable(arg2->var(),variableIndex);
	      bool freeVar2;
	      void* instance2 = var2->Unref(freeVar2);
	      if (freeVar2)
		{
		  if ((static_cast<Unifier::Variable*>(instance2)) == var1)
		    return Less;
		}
	      else
		{
		  if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance2),
						      var1))
		    return Less;
		};	    
	      return Incomparable;  
	    }
	  else
	    {
	      if (termWithUnifierContainsVariable(arg2->deref(),
						  variableIndex,
						  var1))
		return Less;
	      return Incomparable;
	    };	    
	}
      else 
	{
	  UnifTerm::CorrectBracket(iter1->CurrSubterm());
	  if (arg2->isVariable())
	    {
	      Unifier::Variable* var2 = 
		Unifier::current()->variable(arg2->var(),variableIndex);
	      bool freeVar2;
	      void* instance2 = var2->Unref(freeVar2);
	      if (freeVar2)
		{
		  if (termWithUnifierContainsVariable(iter1->CurrSubterm(),
						      static_cast<Unifier::Variable*>(instance2)))
		    return Greater;
		  return Incomparable;
		}
	      else
		{
		  Comparison cmp = 
		    compareOnCurrentUnifier(iter1->CurrSubterm(),
					    static_cast<UnifTerm*>(instance2));
		  if (cmp != Equal) return cmp;
		};	
	    }
	  else 
	    {
	      Comparison cmp = 
		compareOnCurrentUnifier(iter1->CurrSubterm(),
					arg2->deref(),
					variableIndex); 
	      if (cmp != Equal) return cmp;
	    };
	};
      iter1->SkipTerm();
      arg2 = arg2->Next();
    };
  return Equal;
}; 
inline
Comparison 
StandardKBOrdering::compareLexOnCurrentUnifier(UnifTerm* complexTerm1,
					       const Term& term2,
					       ulong variableIndex)
{
  CALL("compareLexOnCurrentUnifier(UnifTerm* complexTerm1,const Term& term2,ulong variableIndex)");
  if (term2.isVariable())
    {
      Unifier::Variable* var2 = 
	Unifier::current()->variable(term2.var(),variableIndex);
      bool freeVar2;
      void* instance2 = var2->Unref(freeVar2);
      if (freeVar2)
	{
	  if (termWithUnifierContainsVariable(complexTerm1,
					      static_cast<Unifier::Variable*>(instance2)))
	    return Greater;
	  return Incomparable;
	}
      else
	{
	  return compareLexOnCurrentUnifier(complexTerm1,
					    static_cast<UnifTerm*>(instance2));
	};
    }
  else
    return compareLexOnCurrentUnifier(complexTerm1,
				      term2.deref(),
				      variableIndex);
}; 
inline
Comparison 
StandardKBOrdering::compareLexOnCurrentUnifier(const Term& term1,
					       const Term& term2,
					       ulong variableIndex)
{
  CALL("compareLexOnCurrentUnifier(const Term& term1,const Term& term2,ulong variableIndex)");
  if (term1 == term2) return Equal;
  if (term1.isVariable())
    {
      Unifier::Variable* var1 = 
	Unifier::current()->variable(term1.var(),variableIndex);
      bool freeVar1;
      void* instance1 = var1->Unref(freeVar1);
      if (freeVar1)
	{
	  if (term2.isVariable())
	    {
	      Unifier::Variable* var2 = 
		Unifier::current()->variable(term2.var(),variableIndex);
	      bool freeVar2;
	      void* instance2 = var2->Unref(freeVar2);
	      if (freeVar2)
		{
		  return (instance1 == instance2) ? Equal : Incomparable;
		}
	      else
		{
		  if (termWithUnifierContainsVariable(static_cast<UnifTerm*>(instance2),
						      static_cast<Unifier::Variable*>(instance1)))
		    return Less;
		  return Incomparable;
		};
	    }
	  else 
	    {
	      if (termWithUnifierContainsVariable(term2.deref(),
						  variableIndex,
						  static_cast<Unifier::Variable*>(instance1)))
		return Less;
	      return Incomparable;
	    };
	}
      else
	{
	  return 
	    compareLexOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
				       term2,
				       variableIndex);
	};
    }
  else 
    if (term2.isVariable())
      {
	Unifier::Variable* var2 = 
	  Unifier::current()->variable(term2.var(),variableIndex);
	bool freeVar2;
	void* instance2 = var2->Unref(freeVar2);
	if (freeVar2)
	  {
	    if (termWithUnifierContainsVariable(term1.deref(),
						variableIndex,
						static_cast<Unifier::Variable*>(instance2)))
	      return Greater;
	    return Incomparable;
	  }
	else
	  {
	    Comparison cmp =
	      compareLexOnCurrentUnifier(static_cast<UnifTerm*>(instance2),
					 term1.deref(),
					 variableIndex);
	    return BK::inverse(cmp);
	  };
      }
    else 
      {
	if ((*term1.deref()) != (*term2.deref()))
	  {
	    return term1.deref()->comparePrecedence(*term2.deref());
	  };
	const Term* arg1 = term1.deref()->Args();
	const Term* arg2 = term2.deref()->Args();
	while (!arg1->isNotArg())
	  {
	    Comparison cmp = 
	      compareOnCurrentUnifier(*arg1,*arg2,variableIndex);
	    if (cmp != Equal) return cmp;
	    arg1 = arg1->Next();
	    arg2 = arg2->Next();
	  };
	return Equal;
      };
}; 
inline
Comparison 
StandardKBOrdering::compareLexicographically(const Term* complexTerm1,
					     const InstRet::Substitution* subst,
					     const Term* complexTerm2)
{
  CALL("compareLexicographically(const Term* complexTerm1,const InstRet::Substitution* subst,const Term* complexTerm2)");
  if ((*complexTerm1) != (*complexTerm2))
    return complexTerm1->comparePrecedence(*complexTerm2);
  const Term* arg1 = complexTerm1->Args();
  const Term* arg2 = complexTerm2->Args();
  while (!arg1->isNotArg())
    {
      if (arg1->isVariable())
	{
	  Term instance = subst->map(arg1->var());
	  Comparison cmp = compare(instance,*arg2);
	  if (cmp != Equal) return cmp;
	}
      else
	{
	  Comparison cmp = compare(arg1->deref(),subst,*arg2);
	  if (cmp != Equal) return cmp;
	};
      arg1 = arg1->Next();      
      arg2 = arg2->Next();      
    };
  ASSERT(arg2->isNotArg());
  return Equal;
}; 
inline
Comparison 
StandardKBOrdering::compareLexicographically(const Term* complexTerm1,
					     const InstRet::Substitution* subst,
					     const Term& term2)
{
  CALL("compareLexicographically(const Term* complexTerm1,const InstRet::Substitution* subst,const Term& term2)");
  if (term2.isVariable())
    {
      InstRet::Substitution::Instance iter1;
      iter1.reset(complexTerm1,subst);
      while (iter1.next())
	{
	  if (iter1.symbol() == term2) return Greater;
	};
      return Incomparable;
    }
  else
    return compareLexicographically(complexTerm1,subst,term2.deref());
}; 
inline
Comparison 
StandardKBOrdering::compare(const Term* complexTerm1,
			    const InstRet::Substitution* subst,
			    const Term& term2)
{
  CALL("compare(const Term* complexTerm1,const InstRet::Substitution* subst,const Term& term2)");
  if (term2.isVariable())
    {
      InstRet::Substitution::Instance iter1;
      iter1.reset(complexTerm1,subst);
      while (iter1.next())
	{
	  if (iter1.symbol() == term2) return Greater;
	};
      return Incomparable;
    }
  else
    {
      _weight1.reset();
      _weight2.reset();
      TermWeighting::current()->collectWeight(_weight1,complexTerm1,subst);
      TermWeighting::current()->collectWeight(_weight2,term2);
      Comparison lexCmp;
      switch (_weight1.compare(_weight2))
	{
	case BK::FailureToCompare:
	  return Incomparable;
	case BK::AlwaysLess:
	  return Less;
	case BK::AlwaysEqual:
	  return compareLexicographically(complexTerm1,subst,term2.deref());
	case BK::AlwaysGreater:
	  return Greater;
	case BK::CanBeLessOrEqual:
	  lexCmp = compareLexicographically(complexTerm1,subst,term2.deref());
	  if (lexCmp == Less) return Less;
	  ASSERT(lexCmp != Equal);
	  return Incomparable;
	case BK::CanBeGreaterOrEqual:
	  lexCmp = compareLexicographically(complexTerm1,subst,term2.deref());
	  if (lexCmp == Greater) return Greater;
	  ASSERT(lexCmp != Equal);
	  return Incomparable;
	case BK::VolatileButNearlyAlwaysLess:
	case BK::VolatileButNearlyAlwaysGreater:
	case BK::AbsolutelyVolatile:
	  return Incomparable;
	};
    };
  ICP("ICP0");
  return Incomparable;
}; 
inline 
Comparison 
StandardKBOrdering::compare(const Term* complexTerm1,const Term* complexTerm2)
{
  CALL("compare(const Term* complexTerm1,const Term* complexTerm2)");
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,Term(const_cast<Term*>(complexTerm1)));
  TermWeighting::current()->collectWeight(_weight2,Term(const_cast<Term*>(complexTerm2)));
  Comparison lexCmp;
  switch (_weight1.compare(_weight2))
    {
    case BK::FailureToCompare:
      return Incomparable;
    case BK::AlwaysLess:
      return Less;
    case BK::AlwaysEqual:
      return compareLexicographically(complexTerm1,complexTerm2);
    case BK::AlwaysGreater:
      return Greater;
    case BK::CanBeLessOrEqual:
      lexCmp = compareLexicographically(complexTerm1,complexTerm2);
      if (lexCmp == Less) return Less;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::CanBeGreaterOrEqual:
      lexCmp = compareLexicographically(complexTerm1,complexTerm2);
      if (lexCmp == Greater) return Greater;
      ASSERT(lexCmp != Equal);
      return Incomparable;
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      return Incomparable;
    };
  ICP("ICP0");
  return Incomparable;
}; 
Comparison 
StandardKBOrdering::compare(const Term& term1,
			    const Term& term2)
{
  CALL("compare(const Term& term1,const Term& term2)");
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  if (term1 == term2) return Equal;
	}
      else
	{
	  if (term2.deref()->containsSymbol(term1)) return Less;
	};
      return Incomparable;
    }
  else
    if (term2.isVariable())
      {
	if (term1.deref()->containsSymbol(term2)) return Greater;
	return Incomparable;
      }
    else 
      return compare(term1.deref(),term2.deref());
}; 
inline 
Comparison 
StandardKBOrdering::compare(const Term& term1,
			    const Term& term2,
			    const InstRet::Substitution* subst)
{
  CALL("compare(const Term& term1,const Term& term2,const InstRet::Substitution* subst)");
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  return compare(subst->map(term1.var()),subst->map(term2.var()));
	}
      else
	{
	  Comparison cmp =
	    compare(term2.deref(),subst,subst->map(term1.var()));
	  return BK::inverse(cmp);
	};
    }
  else
    if (term2.isVariable())
      {
	return compare(term1.deref(),subst,subst->map(term2.var()));
      }
    else
      return compare(term1.deref(),term2.deref(),subst);
}; 
inline 
Comparison 
StandardKBOrdering::compareLexicographically(const Term* complexTerm1,
					     const Term* complexTerm2,
					     const InstRet::Substitution* subst)
{
  CALL("compareLexicographically(const Term* complexTerm1,const Term* complexTerm2,const InstRet::Substitution* subst)");
  if ((*complexTerm1) != (*complexTerm2))
    return complexTerm1->comparePrecedence(*complexTerm2);
  const Term* arg1 = complexTerm1->Args();      
  const Term* arg2 = complexTerm2->Args();
  while (!arg1->isNotArg())
    {
      Comparison cmp = compare(*arg1,*arg2,subst);
      if (cmp != Equal) return cmp;
      arg1 = arg1->Next();
      arg2 = arg2->Next();
    };
  ASSERT(arg2->isNotArg());
  return Equal;
}; 
inline
Comparison 
StandardKBOrdering::compare(const Term& term1,
			    const ForwardMatchingSubstitutionCache* subst,
			    const Flatterm* term2)
{
  CALL("compare(const Term& term1,const ForwardMatchingSubstitutionCache* subst,const Flatterm* term2)");
  if (term1.isVariable())
    {
      return compare(subst->map(term1.var()),term2);
    }
  else
    {
      if (term2->isVariable())
	{
	  ForwardMatchingSubstitutionCache::Instance iter1;
	  iter1.reset(term1.deref(),subst);
	  while (iter1.next())
	    {
	      if (iter1.symbol() == term2->symbolRef())
		return Greater;
	    };
	  return Incomparable;
	}
      else
	{
	  const Term* complexTerm1 = term1.deref();
	  _weight1.reset();
	  _weight2.reset();
	  TermWeighting::current()->collectWeight(_weight1,term1,subst);
	  TermWeighting::current()->collectWeight(_weight2,term2);
	  Comparison lexCmp;
	  switch (_weight1.compare(_weight2))
	    {
	    case BK::FailureToCompare:
	      return Incomparable;
	    case BK::AlwaysLess:
	      return Less;
	    case BK::AlwaysEqual:
	      return compareLexicographically(complexTerm1,subst,term2);
	    case BK::AlwaysGreater:
	      return Greater;
	    case BK::CanBeLessOrEqual:
	      lexCmp = compareLexicographically(complexTerm1,subst,term2);
	      if (lexCmp == Less) return Less;
	      ASSERT(lexCmp != Equal);
	      return Incomparable;
	    case BK::CanBeGreaterOrEqual:
	      lexCmp = compareLexicographically(complexTerm1,subst,term2);
	      if (lexCmp == Greater) return Greater;
	      ASSERT(lexCmp != Equal);
	      return Incomparable;
	    case BK::VolatileButNearlyAlwaysLess:
	    case BK::VolatileButNearlyAlwaysGreater:
	    case BK::AbsolutelyVolatile:
	      return Incomparable;
	    };  	  
	};
    };
  ICP("ICP0");
  return Incomparable;
}; 
inline
Comparison 
StandardKBOrdering::compareLexicographically(const Term* complexTerm1,
					     const ForwardMatchingSubstitutionCache* subst,
					     const Flatterm* term2)
{
  CALL("compareLexicographically(const Term* complexTerm1,const ForwardMatchingSubstitutionCache* subst,const Flatterm* term2)");
  if (term2->isVariable())
    {
      ForwardMatchingSubstitutionCache::Instance iter1;
      iter1.reset(complexTerm1,subst);
      while (iter1.next())
	{
	  if (iter1.symbol() == term2->symbolRef()) return Greater;
	};
      return Incomparable;
    }
  else
    {
      if ((*complexTerm1) != term2->symbolRef())
	return complexTerm1->comparePrecedence(term2->symbolRef());
      const Term* arg1 = complexTerm1->Args();
      Flatterm::ConstIterator iter2(term2);
      iter2.NextSym();
      while (!arg1->isNotArg())
	{
	  Comparison cmp = compare(*arg1,subst,iter2.CurrPos());
	  if (cmp != Equal) return cmp;
	  arg1 = arg1->Next();
	  iter2.after();
	};
      ASSERT(iter2.EndOfTerm());
      return Equal;
    };
}; 
inline
Comparison 
StandardKBOrdering::compareLexicographically(const Term& term1,
					     const Term& term2,
					     const ForwardMatchingSubstitutionCache* subst)
{
  CALL("compareLexicographically(const Term& term1,const Term& term2,const ForwardMatchingSubstitutionCache* subst)");
  if (term1.isVariable())
    {
      if (term2.isVariable())
	{
	  return compareLexicographically(subst->map(term1.var()),
					  subst->map(term2.var()));
	}
      else
	{
	  Comparison cmp = compareLexicographically(term2.deref(),
						    subst,
						    subst->map(term1.var()));
	  return BK::inverse(cmp);
	};
    }
  else
    if (term2.isVariable())
      {
	return compareLexicographically(term1.deref(),
					subst,
					subst->map(term2.var()));
      }
    else
      {
	const Term* complexTerm1 = term1.deref();
	const Term* complexTerm2 = term2.deref();
	if ((*complexTerm1) != (*complexTerm2))
	  return complexTerm1->comparePrecedence(*complexTerm2);
	const Term* arg1 = complexTerm1->Args();
	const Term* arg2 = complexTerm2->Args();
	while (!arg1->isNotArg())
	  {
	    Comparison cmp = compare(*arg1,*arg2,subst);
	    if (cmp != Equal) return cmp;
	    arg1 = arg1->Next();
	    arg2 = arg2->Next();
	  };
	ASSERT(arg2->isNotArg());
	return Equal;
      };     
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::SpecialisedComparison"
#endif
#include "debugMacros.hpp"
StandardKBOrdering::SpecialisedComparison::~SpecialisedComparison()
{ 
  CALL("destructor ~SpecialisedComparison()");
  if (DestructionMode::isThorough())
    {  
      if (_constraintGreater)
	{
	  _constraintGreater->decReferenceCounter();
	  if (!_constraintGreater->referenceCounter())
	    delete _constraintGreater;
	};
      if (_constraintGreaterOrEqual)
	{
	  _constraintGreaterOrEqual->decReferenceCounter();
	  if (!_constraintGreaterOrEqual->referenceCounter())
	    delete _constraintGreaterOrEqual;
	};
      if (_constraintLess)
	{
	  _constraintLess->decReferenceCounter();
	  if (!_constraintLess->referenceCounter())
	    delete _constraintLess;
	};  
      if (_constraintLessOrEqual)
	{
	  _constraintLessOrEqual->decReferenceCounter();
	  if (!_constraintLessOrEqual->referenceCounter())
	    delete _constraintLessOrEqual;
	};
      if (_constraintNotEqual)
	{
	  _constraintNotEqual->decReferenceCounter();
	  if (!_constraintNotEqual->referenceCounter())
	    delete _constraintNotEqual;
	};
    };
}; 
void StandardKBOrdering::SpecialisedComparison::load(const Term& lhs,const Term& rhs)
{
  CALL("load(const Term& lhs,const Term& rhs)");
  _lhs = lhs;
  _rhs = rhs;
  _lhsWeight.reset();
  _rhsWeight.reset();
  TermWeighting::current()->collectWeight(_lhsWeight,_lhs);
  TermWeighting::current()->collectWeight(_rhsWeight,_rhs);
  _specialised = false;
  if (_constraintGreater)
    {
      _constraintGreater->decReferenceCounter();
      if (!_constraintGreater->referenceCounter())
	delete _constraintGreater;
      _constraintGreater = 0;
    };
  if (_constraintGreaterOrEqual)
    {
      _constraintGreaterOrEqual->decReferenceCounter();
      if (!_constraintGreaterOrEqual->referenceCounter())
	delete _constraintGreaterOrEqual;
      _constraintGreaterOrEqual = 0;
    };
  if (_constraintLess)
    {
      _constraintLess->decReferenceCounter();
      if (!_constraintLess->referenceCounter())
	delete _constraintLess;
      _constraintLess = 0;
    };
  if (_constraintLessOrEqual)
    {
      _constraintLessOrEqual->decReferenceCounter();
      if (!_constraintLessOrEqual->referenceCounter())
	delete _constraintLessOrEqual;
      _constraintLessOrEqual = 0;
    };
  if (_constraintNotEqual)
    {
      _constraintNotEqual->decReferenceCounter();
      if (!_constraintNotEqual->referenceCounter())
	delete _constraintNotEqual;
      _constraintNotEqual = 0;
    };
}; 
bool StandardKBOrdering::SpecialisedComparison::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return !_rhsWeight.containsVariableWhichIsNotIn(_lhsWeight);
}; 
bool StandardKBOrdering::SpecialisedComparison::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return !_lhsWeight.containsVariableWhichIsNotIn(_rhsWeight);
}; 
void StandardKBOrdering::SpecialisedComparison::specialiseGreater()
{
  specialise();
};
void StandardKBOrdering::SpecialisedComparison::specialiseGreaterOrEqual()
{
  specialise();
};
void StandardKBOrdering::SpecialisedComparison::specialiseLess()
{
  specialise();
};
void StandardKBOrdering::SpecialisedComparison::specialiseLessOrEqual()
{
  specialise();
};
void StandardKBOrdering::SpecialisedComparison::specialiseNotEqual()
{
  specialise();
};
inline
StandardKBOrdering::StoredConstraint* 
StandardKBOrdering::SpecialisedComparison::storedConstraintGreater()
{
  CALL("storedConstraintGreater()");
  if (!_constraintGreater) 
    {
      specialise();
      _constraintGreater = 
	new StoredConstraint(StoredConstraint::Greater,
			     _weightComparison,
			     _lhs,
			     _rhs);
      _constraintGreater->incReferenceCounter();      
    };
  return _constraintGreater;
}; 
inline
StandardKBOrdering::StoredConstraint* 
StandardKBOrdering::SpecialisedComparison::storedConstraintGreaterOrEqual()
{
  CALL("storedConstraintGreaterOrEqual()");
  if (!_constraintGreaterOrEqual) 
    {
      specialise();
      _constraintGreaterOrEqual = 
	new StoredConstraint(StoredConstraint::GreaterOrEqual,
			     _weightComparison,
			     _lhs,
			     _rhs);
      _constraintGreaterOrEqual->incReferenceCounter();      
    };
  return _constraintGreaterOrEqual;
}; 
inline
StandardKBOrdering::StoredConstraint* 
StandardKBOrdering::SpecialisedComparison::storedConstraintLess()
{
  CALL("storedConstraintLess()");
  if (!_constraintLess) 
    {
      specialise();
      _constraintLess = 
	new StoredConstraint(StoredConstraint::Less,
			     _weightComparison,
			     _lhs,
			     _rhs);
      _constraintLess->incReferenceCounter();      
    };
  return _constraintLess;
}; 
inline
StandardKBOrdering::StoredConstraint* 
StandardKBOrdering::SpecialisedComparison::storedConstraintLessOrEqual()
{
  CALL("storedConstraintLessOrEqual()");  
  if (!_constraintLessOrEqual) 
    {
      specialise();
      _constraintLessOrEqual = 
	new StoredConstraint(StoredConstraint::LessOrEqual,
			     _weightComparison,
			     _lhs,
			     _rhs);
      _constraintLessOrEqual->incReferenceCounter();      
    };
  return _constraintLessOrEqual;
}; 
inline
StandardKBOrdering::StoredConstraint* 
StandardKBOrdering::SpecialisedComparison::storedConstraintNotEqual()
{
  CALL("storedConstraintNotEqual()");  
  if (!_constraintNotEqual) 
    {
      specialise();
      _constraintNotEqual = 
	new StoredConstraint(StoredConstraint::NotEqual,
			     _weightComparison,
			     _lhs,
			     _rhs);
      _constraintNotEqual->incReferenceCounter();      
    };
  return _constraintNotEqual;
}; 
#ifndef NO_DEBUG_VIS
ostream& StandardKBOrdering::SpecialisedComparison::output(ostream& str) const
{
  str << _lhs << " VS " << _rhs << "\n";
  if (_specialised)
    {
      str << "CATEGORY: " << category() << "\n";
      str << " WEIGHT: " << _weightComparison << "\n";
    }
  else
    {
      str << " NOT SPECIALISED\n";
      str << "LHS WEIGHT: " << _lhsWeight << "\n";
      str << "RHS WEIGHT: " << _rhsWeight;
    };
  return str;
};
#endif
inline
void StandardKBOrdering::SpecialisedComparison::specialise()
{
  CALL("specialise()");
  if (_specialised) return;
  _weightComparison.compare(_lhsWeight,_rhsWeight);
  switch (_weightComparison.category())
    {
    case BK::FailureToCompare:
      ICP("ICP10");
      break;
    case BK::AlwaysLess: 
      _category = SimplificationOrdering::SpecialisedComparison::AlwaysLess;
      break;
    case BK::AlwaysEqual:
      if (_lhs == _rhs) 
	{
	  _category = SimplificationOrdering::SpecialisedComparison::AlwaysEqual;
	}
      else
	_category = SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile;
      break;
    case BK::AlwaysGreater:
      _category = SimplificationOrdering::SpecialisedComparison::AlwaysGreater;
      break;
    case BK::CanBeLessOrEqual:
      _category = 
	SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess;
      break;
    case BK::CanBeGreaterOrEqual:
      _category = 
	SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater;
      break;
    case BK::VolatileButNearlyAlwaysLess:
      _category = 
	SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess;
      break;
    case BK::VolatileButNearlyAlwaysGreater:
      _category = 
	SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater;
      break;
    case BK::AbsolutelyVolatile:
      _category = 
	SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile;
      break;         
    };      
  _specialised = true;
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#  define DEBUG_NAMESPACE "StandardKBOrdering::ConstraintCheckOnUnifier"
#endif
#include "debugMacros.hpp"
StandardKBOrdering::ConstraintCheckOnUnifier::~ConstraintCheckOnUnifier()
{
};
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::holds(const SimplificationOrdering::StoredConstraint* constr,
						    ulong variableIndex)
{
  CALL("holds(..)");
  ASSERT(constr);
  const StoredConstraint* nativeConstr = 
    static_cast<const StoredConstraint*>(static_cast<const void*>(constr));
  _currentVariableIndex = variableIndex;
  bool result;
#if (defined DEBUG_NAMESPACE) || (defined _SUPPRESS_WARNINGS_)
  result = false;
#endif
  DOP(Comparison debugCmp = \
      _ordering->compareOnCurrentUnifier((nativeConstr)->lhs(),\
					 (nativeConstr)->rhs(),\
					 variableIndex));
  switch ((nativeConstr)->targetRelation())
    {
    case StoredConstraint::Greater:
      result = holdsGreater(nativeConstr);
      ASSERT(result == (debugCmp == Greater));
      break;
    case StoredConstraint::GreaterOrEqual:
      result = holdsGreaterOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Greater) || (debugCmp == Equal)));
      break;
    case StoredConstraint::Less:
      result = holdsLess(nativeConstr);
      ASSERT(result == (debugCmp == Less));
      break;
    case StoredConstraint::LessOrEqual:
      result = holdsLessOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Less) || (debugCmp == Equal)));
      break;
    case StoredConstraint::NotEqual:
      result = holdsNotEqual(nativeConstr);
      ASSERT(debugCmp != Equal || !result);
      break;
    };
  return result;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::holdsGreater(const StandardKBOrdering::StoredConstraint* constr)
{
  CALL("holdsGreater(..)");
  switch (checkGreaterOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      ICP("ICP10");
      return false; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyGreater(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:
      return true;
    case BK::CanBeLessOrEqual:
      ICP("ICP30");
      return false;
    case BK::CanBeGreaterOrEqual:
      return lexicorgaphicallyGreater(constr->lhs(),constr->rhs());
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      ICP("ICP70");
      return false; 
    };
  ICP("ICP80");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::holdsGreaterOrEqual(const StandardKBOrdering::StoredConstraint* constr)
{
  CALL("holdsGreaterOrEqual(..)");
  switch (checkGreaterOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      ICP("ICP10");
      return false; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyGreaterOrEqual(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:
      return true;
    case BK::CanBeLessOrEqual:	      	      
      ICP("ICP30");
      return false;
    case BK::CanBeGreaterOrEqual:
      return lexicorgaphicallyGreaterOrEqual(constr->lhs(),constr->rhs());
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:      	      
      ICP("ICP50");
      return false; 
    };
  ICP("ICP3");
  return false;
}; 
inline
bool StandardKBOrdering::ConstraintCheckOnUnifier::holdsLess(const StandardKBOrdering::StoredConstraint* constr)
{
  CALL("holdsLess(..)");
  switch (checkLessOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      return true; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyLess(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:      	      	      
      ICP("ICP20");
      return false;
    case BK::CanBeLessOrEqual:
      return lexicorgaphicallyLess(constr->lhs(),constr->rhs());
    case BK::CanBeGreaterOrEqual: 
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:    	      	      
      ICP("ICP30");
      return false; 
    };
  ICP("ICP40");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::holdsLessOrEqual(const StandardKBOrdering::StoredConstraint* constr)
{
  CALL("holdsLessOrEqual(..)");
  switch (checkLessOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      return true; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyLessOrEqual(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:         
      ICP("ICP20");
      return false;
    case BK::CanBeLessOrEqual:
      return lexicorgaphicallyLessOrEqual(constr->lhs(),constr->rhs());
    case BK::CanBeGreaterOrEqual:
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile: 	      
      ICP("ICP40");
      return false;  	
    };
  ICP("ICP50");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::holdsNotEqual(const StandardKBOrdering::StoredConstraint* constr)
{
  CALL("holdsNotEqual(..)");
  switch (checkNotEqual(constr->weightComparison()))
    {
    case BK::DefinitelyFalse:
      return !InstanceWithUnifier::equal(_currentVariableIndex,
				      constr->lhs(),
				      _currentVariableIndex,
				      constr->rhs());
    case BK::DefinitelyTrue:
      return true;
    case BK::DontKnow:
      return false;
    };
  ICP("ICP50");
  return false;
}; 
inline
FunctionComparison 
StandardKBOrdering::ConstraintCheckOnUnifier::checkGreaterOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkGreaterOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::FailureToCompare;
    case BK::AlwaysLess:
      return BK::FailureToCompare;
    case BK::AlwaysEqual:
    case BK::AlwaysGreater:
      return weightComparison.category(); 
    case BK::CanBeLessOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightPosInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = 
		  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
										static_cast<UnifTerm*>(varInstance),
										weightPosInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightPosInst);
		weightPosInst -= w;
	      };
	  };
	ASSERT(weightPosInst == 0);
	return BK::AlwaysEqual;
      };
    case BK::CanBeGreaterOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		weightNegInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightNegInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };	   
	    if (weightNegInst < 0) 
	      return BK::AlwaysGreater;
	  };
	ASSERT(weightNegInst == 0); 
	if (ground)
	  return BK::AlwaysEqual;
	return BK::CanBeGreaterOrEqual;
      };
    case BK::VolatileButNearlyAlwaysLess:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightPosInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = 
		  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
										static_cast<UnifTerm*>(varInstance),
										weightPosInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightPosInst);
		weightPosInst -= w;
	      };
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0) 
	  return BK::AlwaysEqual;
	return BK::AlwaysGreater;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = 
	      Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		weightNegInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightNegInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };
	    if (weightNegInst < 0) 
	      return BK::AlwaysGreater;
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0)
	  {
	    if (ground) 
	      return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  return BK::FailureToCompare;
      };
    case BK::AbsolutelyVolatile:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      Unifier::Variable* var = 
		Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	      bool freeVar;
	      void* varInstance = var->Unref(freeVar);
	      if (freeVar)
		{
		  ulong varNum;
		  ulong varIndex;
		  (static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		  _weightPositivePartInstance.add(coeff,
						  varNum + 
						  (varIndex*VampireKernelConst::MaxNumOfVariables));
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  static_cast<UnifTerm*>(varInstance));
		};
	    }; 
	};
      {
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = 
	      Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		ulong varNum;
		ulong varIndex;
		(static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		_weightNegativePartInstance.add(coeff,
						varNum + (varIndex*VampireKernelConst::MaxNumOfVariables));
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							static_cast<UnifTerm*>(varInstance));
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (!_weightPositivePartInstance.isAlwaysGreater(_weightNegativePartInstance))
	      return BK::FailureToCompare;
	    goto next_member;
	  }
	else
	  {
	    FunctionComparison cmp =
	      _weightPositivePartInstance.compareForGreaterOrEqual(_weightNegativePartInstance);
	    switch (cmp)
	      {
	      case BK::FailureToCompare: 
		return BK::FailureToCompare;
	      case BK::AlwaysLess:
		ICP("ICP10");
		return BK::FailureToCompare;
	      case BK::AlwaysEqual:
		return BK::AlwaysEqual;
	      case BK::AlwaysGreater:
		return BK::AlwaysGreater;
	      case BK::CanBeLessOrEqual:
		ICP("ICP20");
		return BK::FailureToCompare;
	      case BK::CanBeGreaterOrEqual:
		return BK::CanBeGreaterOrEqual;
	      case BK::VolatileButNearlyAlwaysLess:
	      case BK::VolatileButNearlyAlwaysGreater:
	      case BK::AbsolutelyVolatile:
		ICP("ICP30");
		return BK::FailureToCompare;
	      };
	  };	
      };
    }; 
  ICP("END");
  return BK::FailureToCompare; 
}; 
inline
FunctionComparison 
StandardKBOrdering::ConstraintCheckOnUnifier::checkLessOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkLessOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::FailureToCompare;
    case BK::AlwaysLess:
    case BK::AlwaysEqual:
      return weightComparison.category(); 
    case BK::AlwaysGreater:
      return BK::FailureToCompare;
    case BK::CanBeLessOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		weightPosInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };	   
	    if (weightPosInst < 0) 
	      return BK::AlwaysLess;
	  };
	ASSERT(weightPosInst == 0); 
	if (ground)
	  return BK::AlwaysEqual;
	return BK::CanBeLessOrEqual;
      };
    case BK::CanBeGreaterOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 static_cast<UnifTerm*>(varInstance),
												 weightNegInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst == 0);
	return BK::AlwaysEqual;
      };
    case BK::VolatileButNearlyAlwaysLess:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		weightPosInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };
	    if (weightPosInst < 0) 
	      return BK::AlwaysLess;
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0)
	  {
	    if (ground) 
	      return BK::AlwaysEqual;
	    return BK::CanBeLessOrEqual;
	  }
	else
	  return BK::FailureToCompare;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 static_cast<UnifTerm*>(varInstance),
												 weightNegInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0) 
	  return BK::AlwaysEqual;
	return BK::AlwaysLess;
      };
    case BK::AbsolutelyVolatile:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	      bool freeVar;
	      void* varInstance = var->Unref(freeVar);
	      if (freeVar)
		{
		  ulong varNum;
		  ulong varIndex;
		  (static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		  _weightPositivePartInstance.add(coeff,
						  varNum + (varIndex*VampireKernelConst::MaxNumOfVariables));
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  static_cast<UnifTerm*>(varInstance));
		};
	    }; 
	};
      {	
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		ulong varNum;
		ulong varIndex;
		(static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		_weightNegativePartInstance.add(coeff,
						varNum + (varIndex*VampireKernelConst::MaxNumOfVariables));
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							static_cast<UnifTerm*>(varInstance));
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (_weightPositivePartInstance.isNeverGreater(_weightNegativePartInstance))
	      {
		return BK::AlwaysLess;
	      };
	    goto next_member;
	  }
	else
	  {
	    FunctionComparison cmp =
	      _weightPositivePartInstance.compareForLessOrEqual(_weightNegativePartInstance);
	    switch (cmp)
	      {
	      case BK::FailureToCompare:		
		return BK::FailureToCompare;
	      case BK::AlwaysLess:
		return BK::AlwaysLess;
	      case BK::AlwaysEqual:
		return BK::AlwaysEqual;
	      case BK::AlwaysGreater:
		ICP("ICP10");
		return BK::FailureToCompare;
	      case BK::CanBeLessOrEqual:
		return BK::CanBeLessOrEqual;
	      case BK::CanBeGreaterOrEqual:
	      case BK::VolatileButNearlyAlwaysLess:
	      case BK::VolatileButNearlyAlwaysGreater:
	      case BK::AbsolutelyVolatile:
		ICP("ICP20");
		return BK::FailureToCompare;
	      };
	  };	
      };
    }; 
  ICP("END");
  return BK::FailureToCompare;
}; 
inline
BK::BoolPlus
StandardKBOrdering::ConstraintCheckOnUnifier::checkNotEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkNotEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::DontKnow;
    case BK::AlwaysLess:
      return BK::DefinitelyTrue;
    case BK::AlwaysEqual:
      return BK::DefinitelyFalse;
    case BK::AlwaysGreater:
      return BK::DefinitelyTrue;
    case BK::CanBeLessOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
                return BK::DefinitelyTrue;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) return BK::DefinitelyTrue;
	      };	   
	    if (weightPosInst < 0) 
	      return BK::DefinitelyTrue;
	  };
	ASSERT(weightPosInst == 0); 
        return BK::DefinitelyFalse;
      };
    case BK::CanBeGreaterOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::DefinitelyTrue;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::DefinitelyTrue;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 static_cast<UnifTerm*>(varInstance),
												 weightNegInst);
		if (w <= 0) return BK::DefinitelyTrue;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst == 0);
	return BK::DefinitelyFalse;
      };
    case BK::VolatileButNearlyAlwaysLess:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
	        return BK::DefinitelyTrue;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(static_cast<UnifTerm*>(varInstance),
									  instanceContainsVariables));
		if (instanceContainsVariables) return BK::DefinitelyTrue;
	      };
	    if (weightPosInst < 0) 
	      return BK::DefinitelyTrue;
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0)
	  {
	    return BK::DefinitelyFalse;
	  }
	else
	  return BK::DefinitelyTrue;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::DefinitelyTrue;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		return BK::DefinitelyTrue;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 static_cast<UnifTerm*>(varInstance),
												 weightNegInst);
		if (w <= 0) return BK::DefinitelyTrue;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0) 
	  return BK::DefinitelyFalse;
	return BK::DefinitelyTrue;
      };
    case BK::AbsolutelyVolatile:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	      bool freeVar;
	      void* varInstance = var->Unref(freeVar);
	      if (freeVar)
		{
		  ulong varNum;
		  ulong varIndex;
		  (static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		  _weightPositivePartInstance.add(coeff,
						  varNum + (varIndex*VampireKernelConst::MaxNumOfVariables));
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  static_cast<UnifTerm*>(varInstance));
		};
	    }; 
	};
      {	
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    Unifier::Variable* var = Unifier::current()->variable(mem->hd().var(),_currentVariableIndex);
	    bool freeVar;
	    void* varInstance = var->Unref(freeVar);
	    if (freeVar)
	      {
		ulong varNum;
		ulong varIndex;
		(static_cast<Unifier::Variable*>(varInstance))->NumAndIndex(varNum,varIndex);	    
		_weightNegativePartInstance.add(coeff,
						varNum + (varIndex*VampireKernelConst::MaxNumOfVariables));
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							static_cast<UnifTerm*>(varInstance));
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (_weightPositivePartInstance.isNeverGreater(_weightNegativePartInstance))
	      {
		return BK::DefinitelyTrue;
	      };
	    goto next_member;
	  }
	else
	  {
	    return
	      BK::boolPlus(!_weightPositivePartInstance.equal(_weightNegativePartInstance));
	  };	
      };
    }; 
  ICP("END");
  return BK::DontKnow;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::lexicorgaphicallyGreater(const Term& lhs,
								       const Term& rhs)
{
  CALL("lexicorgaphicallyGreater(const Term& lhs,const Term& rhs)");
  Comparison cmp = 
    _ordering->compareLexOnCurrentUnifier(lhs,rhs,_currentVariableIndex);
  return (cmp == Greater);
}; 
inline
bool StandardKBOrdering::ConstraintCheckOnUnifier::lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs)
{
  CALL("lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs)");
  Comparison cmp = 
    _ordering->compareLexOnCurrentUnifier(lhs,rhs,_currentVariableIndex);
  return (cmp == Greater) || (cmp == Equal);   
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::lexicorgaphicallyLess(const Term& lhs,const Term& rhs)
{
  return lexicorgaphicallyGreater(rhs,lhs);
};
inline
bool 
StandardKBOrdering::ConstraintCheckOnUnifier::lexicorgaphicallyLessOrEqual(const Term& lhs,const Term& rhs)
{
  return lexicorgaphicallyGreaterOrEqual(rhs,lhs);
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIMPLIFICATION_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst"
#endif
#include "debugMacros.hpp"
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::~ConstraintCheckOnForwardMatchingSubst()
{
};
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holds(const SimplificationOrdering::StoredConstraint* constr,
								 const ForwardMatchingSubstitutionCache* subst)
{
  CALL("holds(..)");
  ASSERT(constr);
  const StoredConstraint* nativeConstr = 
    static_cast<const StoredConstraint*>(static_cast<const void*>(constr));
  bool result;
  DOP(Comparison debugCmp = \
      _ordering->compare(nativeConstr->lhs(),\
			 nativeConstr->rhs(),\
			 subst));
  _subst = subst;
  switch (nativeConstr->targetRelation())
    {
    case StoredConstraint::Greater:
      result = holdsGreater(nativeConstr);
      ASSERT(result == (debugCmp == Greater));
      return result;
    case StoredConstraint::GreaterOrEqual:
      result = holdsGreaterOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Greater) || (debugCmp == Equal)));
      return result;
    case StoredConstraint::Less:
      result = holdsLess(nativeConstr);
      ASSERT(result == (debugCmp == Less));
      return result;
    case StoredConstraint::LessOrEqual:
      result = holdsLessOrEqual(nativeConstr);
      ASSERT(result == ((debugCmp == Less) || (debugCmp == Equal)));
      return result;
    case StoredConstraint::NotEqual:
      result = holdsNotEqual(nativeConstr);
      ASSERT(debugCmp != Equal || !result);
      return result;
    };
  ICP("ICP0");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holdsGreater(const StoredConstraint* constr)
{
  CALL("holdsGreater(..)");
  switch (checkGreaterOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      ICP("ICP10");
      return false; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyGreater(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:
      return true;
    case BK::CanBeLessOrEqual:
      ICP("ICP30");
      return false;
    case BK::CanBeGreaterOrEqual:
      return lexicorgaphicallyGreater(constr->lhs(),constr->rhs());
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:     	      
      ICP("ICP50");
      return false;
    };
  ICP("ICP60");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holdsGreaterOrEqual(const StoredConstraint* constr)
{
  CALL("holdsGreaterOrEqual(..)");
  switch (checkGreaterOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      ICP("ICP10");
      return false; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyGreaterOrEqual(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:
      return true;
    case BK::CanBeLessOrEqual:
      ICP("ICP30");
      return false;
    case BK::CanBeGreaterOrEqual:
      return lexicorgaphicallyGreaterOrEqual(constr->lhs(),constr->rhs());
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:
      ICP("ICP50");
      return false; 
    };
  ICP("ICP60");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holdsLess(const StoredConstraint* constr)
{
  CALL("holdsLess(..)");
  FunctionComparison weightComp = 
    checkLessOrEqual(constr->weightComparison());
  switch (weightComp)
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      return true; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyLess(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:    	      
      ICP("ICP20");
      return false;
    case BK::CanBeLessOrEqual:
      return lexicorgaphicallyLess(constr->lhs(),constr->rhs());
    case BK::CanBeGreaterOrEqual:
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:            	      
      ICP("ICP40");
      return false; 
    };
  ICP("ICP50");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holdsLessOrEqual(const StoredConstraint* constr)
{
  CALL("holdsLessOrEqual(..)");
  switch (checkLessOrEqual(constr->weightComparison()))
    {
    case BK::FailureToCompare:
      return false;
    case BK::AlwaysLess:
      return true; 
    case BK::AlwaysEqual:
      return lexicorgaphicallyLessOrEqual(constr->lhs(),constr->rhs());
    case BK::AlwaysGreater:   	      
      ICP("ICP20");
      return false;
    case BK::CanBeLessOrEqual:
      return lexicorgaphicallyLessOrEqual(constr->lhs(),constr->rhs());
    case BK::CanBeGreaterOrEqual:
    case BK::VolatileButNearlyAlwaysLess:
    case BK::VolatileButNearlyAlwaysGreater:
    case BK::AbsolutelyVolatile:    	      
      ICP("ICP40");
      return false; 
    };
  ICP("ICP50");
  return false;
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::holdsNotEqual(const StoredConstraint* constr)
{
  CALL("holdsNotEqual(..)");
  switch (checkNotEqual(constr->weightComparison()))
    {
    case BK::DefinitelyFalse:
      return 
	!ForwardMatchingSubstitutionCache::Instance::equal(constr->lhs(),
							   constr->rhs(),
							   _subst);
    case BK::DefinitelyTrue:
      return true;
    case BK::DontKnow:
      return false;
    };
  ICP("ICP50");
  return false;
}; 
inline
FunctionComparison 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::checkGreaterOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkGreaterOrEqual(..)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::FailureToCompare;
    case BK::AlwaysLess:
      return BK::FailureToCompare;
    case BK::AlwaysEqual:
    case BK::AlwaysGreater:
      return weightComparison.category(); 
    case BK::CanBeLessOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightPosInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {			
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightPosInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightPosInst);
		weightPosInst -= w;
	      };
	  };
	ASSERT(weightPosInst == 0);
	return BK::AlwaysEqual;
      };
    case BK::CanBeGreaterOrEqual:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		weightNegInst -= coeff;
		ground = false;
	      }
	    else
	      {
		bool instanceContainsVariables;
		weightNegInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };	
	    if (weightNegInst < 0) 
	      return BK::AlwaysGreater;
	  };
	ASSERT(weightNegInst == 0); 
	if (ground)
	  return BK::AlwaysEqual;
	return BK::CanBeGreaterOrEqual;
      };
    case BK::VolatileButNearlyAlwaysLess:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightPosInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightPosInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightPosInst);
		weightPosInst -= w;
	      };
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0) 
	  return BK::AlwaysEqual;
	return BK::AlwaysGreater;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		weightNegInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightNegInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };
	    if (weightNegInst < 0) 
	      return BK::AlwaysGreater;
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0)
	  {
	    if (ground) 
	      return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  return BK::FailureToCompare;
      };
    case BK::AbsolutelyVolatile:
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      const Flatterm* varInstance = _subst->map(mem->hd().var());
	      if (varInstance->isVariable())
		{ 
		  _weightPositivePartInstance.add(coeff,varInstance->Var());
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  varInstance);
		};
	    }; 
	};
      {
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {    
		_weightNegativePartInstance.add(coeff,varInstance->Var());
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							varInstance);
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (!_weightPositivePartInstance.isAlwaysGreater(_weightNegativePartInstance))
	      return BK::FailureToCompare;
	    goto next_member;
	  }
	else
	  {
	    FunctionComparison cmp =
	      _weightPositivePartInstance.compareForGreaterOrEqual(_weightNegativePartInstance);
	    switch (cmp)
	      {
	      case BK::FailureToCompare:
		return BK::FailureToCompare;
	      case BK::AlwaysLess:
		ICP("ICP10");
		return BK::FailureToCompare;
	      case BK::AlwaysEqual:
		return BK::AlwaysEqual;
	      case BK::AlwaysGreater:
		return BK::AlwaysGreater;
	      case BK::CanBeLessOrEqual:
		ICP("ICP20");
		return BK::FailureToCompare;
	      case BK::CanBeGreaterOrEqual:
		return BK::CanBeGreaterOrEqual;
	      case BK::VolatileButNearlyAlwaysLess:
	      case BK::VolatileButNearlyAlwaysGreater:
	      case BK::AbsolutelyVolatile:
		ICP("ICP30");
		return BK::FailureToCompare;
	      };
	  };	
      };
    }; 
  ICP("END");
  return BK::FailureToCompare; 
}; 
inline
FunctionComparison 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::checkLessOrEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkLessOrEqual(..)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::FailureToCompare;
    case BK::AlwaysLess:
    case BK::AlwaysEqual:
      return weightComparison.category();
    case BK::AlwaysGreater:
      return BK::FailureToCompare;
    case BK::CanBeLessOrEqual:
      COP("case BK::CanBeLessOrEqual");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		weightPosInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };	   
	    if (weightPosInst < 0) 
	      return BK::AlwaysLess;
	  };
	ASSERT(weightPosInst == 0); 
	if (ground)
	  return BK::AlwaysEqual;
	return BK::CanBeLessOrEqual;
      };
    case BK::CanBeGreaterOrEqual:
      COP("case BK::CanBeGreaterOrEqual");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightNegInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst == 0);
	return BK::AlwaysEqual;
      };
    case BK::VolatileButNearlyAlwaysLess:
      COP("case BK::VolatileButNearlyAlwaysLess");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	bool ground = true;
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		weightPosInst -= coeff;
		ground = false;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) ground = false;
	      };
	    if (weightPosInst < 0) 
	      return BK::AlwaysLess;
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0)
	  {
	    if (ground) 
	      return BK::AlwaysEqual;
	    return BK::CanBeLessOrEqual;
	  }
	else
	  return BK::FailureToCompare;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      COP("case BK::VolatileButNearlyAlwaysGreater");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::FailureToCompare;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::FailureToCompare;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightNegInst);
		if (w <= 0) return BK::FailureToCompare;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0) 
	  return BK::AlwaysEqual;
	return BK::AlwaysLess;
      };
    case BK::AbsolutelyVolatile:
      COP("case BK::AbsolutelyVolatile");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      const Flatterm* varInstance = _subst->map(mem->hd().var());
	      if (varInstance->isVariable())
		{    
		  _weightPositivePartInstance.add(coeff,varInstance->Var());
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  varInstance);
		};
	    }; 
	};
      {
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {    
		_weightNegativePartInstance.add(coeff,varInstance->Var());
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							varInstance);
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (_weightPositivePartInstance.isNeverGreater(_weightNegativePartInstance))
	      return BK::AlwaysLess;
	    goto next_member;
	  }
	else
	  {
	    FunctionComparison cmp =
	      _weightPositivePartInstance.compareForLessOrEqual(_weightNegativePartInstance);
	    switch (cmp)
	      {
	      case BK::FailureToCompare:
		return BK::FailureToCompare;		
	      case BK::AlwaysLess:
		return BK::AlwaysLess;
	      case BK::AlwaysEqual:
		return BK::AlwaysEqual;
	      case BK::AlwaysGreater:
		ICP("ICP10");
		return BK::FailureToCompare;
	      case BK::CanBeLessOrEqual:
		return BK::CanBeLessOrEqual;
	      case BK::CanBeGreaterOrEqual:
	      case BK::VolatileButNearlyAlwaysLess:
	      case BK::VolatileButNearlyAlwaysGreater:
	      case BK::AbsolutelyVolatile:
		ICP("ICP20");
		return BK::FailureToCompare;		
	      };
	  };	
      };
    }; 
  ICP("END");
  return BK::FailureToCompare; 
}; 
inline
BK::BoolPlus
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::checkNotEqual(const WeightPolynomial::SpecialisedComparison::Stored& weightComparison)
{
  CALL("checkNotEqual(..)");
  switch (weightComparison.category())
    {       
    case BK::FailureToCompare:
      return BK::DontKnow;
    case BK::AlwaysLess:
      return BK::DefinitelyTrue;
    case BK::AlwaysEqual:
      return BK::DefinitelyFalse;
    case BK::AlwaysGreater:
      return BK::DefinitelyFalse;
    case BK::CanBeLessOrEqual:
      COP("case BK::CanBeLessOrEqual");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
                return BK::DefinitelyTrue;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) return BK::DefinitelyTrue;
	      };	   
	    if (weightPosInst < 0) 
	      return BK::DefinitelyTrue;
	  };
	ASSERT(weightPosInst == 0); 
	return BK::DefinitelyFalse;
      };
    case BK::CanBeGreaterOrEqual:
      COP("case BK::CanBeGreaterOrEqual");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::DefinitelyTrue;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::DefinitelyTrue;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightNegInst);
		if (w <= 0) return BK::DefinitelyTrue;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst == 0);
	return BK::DefinitelyFalse;
      };
    case BK::VolatileButNearlyAlwaysLess:
      COP("case BK::VolatileButNearlyAlwaysLess");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.positivePart()->hd().isConstant());
      ASSERT(weightComparison.positivePart()->hd().constant() > 0);
      ASSERT(!weightComparison.positivePart()->tl());
      {
	long weightPosInst = weightComparison.positivePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.negativePart();
	     mem;
	     mem = mem->tl())
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
                return BK::DefinitelyTrue;
	      }
	    else
	      {		
		bool instanceContainsVariables;
		weightPosInst -= 
		  (coeff * 
		   TermWeighting::current()->computeMinimalInstanceWeight(varInstance,
									  instanceContainsVariables));
		if (instanceContainsVariables) return BK::DefinitelyTrue;
	      };
	    if (weightPosInst < 0) 
	      return BK::DefinitelyTrue;
	  };
	ASSERT(weightPosInst >= 0);
	if (weightPosInst == 0)
	  {
	    return BK::DefinitelyFalse;
	  }
	else
	  return BK::DefinitelyTrue;
      };
    case BK::VolatileButNearlyAlwaysGreater:
      COP("case BK::VolatileButNearlyAlwaysGreater");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      ASSERT(weightComparison.negativePart()->hd().isConstant());
      ASSERT(weightComparison.negativePart()->hd().constant() > 0);
      ASSERT(!weightComparison.negativePart()->tl());
      {
	long weightNegInst = weightComparison.negativePart()->hd().constant();
	for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	     mem;
	     mem = mem->tl())
	  {
	    if (weightNegInst == 0) 
	      return BK::DefinitelyTrue;
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {
		return BK::DefinitelyTrue;
	      }
	    else
	      {		
		long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(coeff,
												 varInstance,
												 weightNegInst);
		if (w <= 0) return BK::DefinitelyTrue;
		ASSERT(w <= weightNegInst);
		weightNegInst -= w;
	      };
	  };
	ASSERT(weightNegInst >= 0);
	if (weightNegInst == 0) 
	  return BK::DefinitelyFalse;
	return BK::DefinitelyTrue;
      };
    case BK::AbsolutelyVolatile:
      COP("case BK::AbsolutelyVolatile");
      ASSERT(weightComparison.positivePart());
      ASSERT(weightComparison.negativePart()); 
      _weightPositivePartInstance.reset();
      for (WeightPolynomial::Stored* mem = weightComparison.positivePart();
	   mem;
	   mem = mem->tl())
	{
	  if (mem->hd().isConstant())
	    {
	      _weightPositivePartInstance.add(mem->hd().constant());
	    }
	  else 
	    {
	      ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	      long coeff = mem->hd().coeff();
	      const Flatterm* varInstance = _subst->map(mem->hd().var());
	      if (varInstance->isVariable())
		{    
		  _weightPositivePartInstance.add(coeff,varInstance->Var());
		}
	      else
		{
		  TermWeighting::current()->collectWeight(_weightPositivePartInstance,
							  coeff,
							  varInstance);
		};
	    }; 
	};
      {
	_weightNegativePartInstance.reset();
	WeightPolynomial::Stored* mem = weightComparison.negativePart();
      next_member:
	ASSERT(mem);
	if (mem->hd().isConstant())
	  {
	    _weightNegativePartInstance.add(mem->hd().constant());
	  }
	else 
	  {
	    ASSERT(mem->hd().isVariable() && (mem->hd().coeff() > 0));
	    long coeff = mem->hd().coeff();
	    const Flatterm* varInstance = _subst->map(mem->hd().var());
	    if (varInstance->isVariable())
	      {    
		_weightNegativePartInstance.add(coeff,varInstance->Var());
	      }
	    else
	      {
		TermWeighting::current()->collectWeight(_weightNegativePartInstance,
							coeff,
							varInstance);
	      };
	  };
	mem = mem->tl();
	if (mem) 
	  {
	    if (_weightPositivePartInstance.isNeverGreater(_weightNegativePartInstance))
	      return BK::DefinitelyTrue;
	    goto next_member;
	  }
	else
	  {
	    return
	      BK::boolPlus(!_weightPositivePartInstance.equal(_weightNegativePartInstance));
	  };	
      };
    }; 
  ICP("END");
  return BK::DontKnow; 
}; 
inline
bool StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::lexicorgaphicallyGreater(const Term& lhs,const Term& rhs)
{
  CALL("lexicorgaphicallyGreater(const Term& lhs,const Term& rhs)");
  Comparison cmp =
    _ordering->compareLexicographically(lhs,rhs,_subst);
  return (cmp == Greater);
}; 
inline
bool StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs)
{
  CALL("lexicorgaphicallyGreaterOrEqual(const Term& lhs,const Term& rhs)");  
  Comparison cmp =
    _ordering->compareLexicographically(lhs,rhs,_subst);
  return (cmp == Greater) || (cmp == Equal);
}; 
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::lexicorgaphicallyLess(const Term& lhs,const Term& rhs)
{
  return lexicorgaphicallyGreater(rhs,lhs);
};
inline
bool 
StandardKBOrdering::ConstraintCheckOnForwardMatchingSubst::lexicorgaphicallyLessOrEqual(const Term& lhs,const Term& rhs)
{
  return lexicorgaphicallyGreaterOrEqual(rhs,lhs);
};
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForUnifiers"
#endif
#include "debugMacros.hpp"
StandardKBOrdering::CompiledConstraintForUnifiers::~CompiledConstraintForUnifiers() 
{
};  
void StandardKBOrdering::CompiledConstraintForUnifiers::load(UnifTerm* lhs,UnifTerm* rhs)
{
  CALL("load(UnifTerm* lhs,UnifTerm* rhs)");
  _lhs = lhs;
  _rhs = rhs;
  _lhsWeight.reset();
  _rhsWeight.reset();
  TermWeighting::current()->collectWeight(_lhsWeight,1,_lhs);
  TermWeighting::current()->collectWeight(_rhsWeight,1,_rhs);  
  _topLevelWeightComparison.compare(_lhsWeight,_rhsWeight);
}; 
bool StandardKBOrdering::CompiledConstraintForUnifiers::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return !_rhsWeight.containsVariableWhichIsNotIn(_lhsWeight);
}; 
bool StandardKBOrdering::CompiledConstraintForUnifiers::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return !_lhsWeight.containsVariableWhichIsNotIn(_rhsWeight);
}; 
void StandardKBOrdering::CompiledConstraintForUnifiers::compileGreaterOrEqual()
{
  CALL("compileGreaterOrEqual()");
  ++(_ordering->_numberOfCompiledConstraintsOnUnifiers);
#ifndef NO_DEBUG
  _targetRelation = GreaterOrEqual;
#endif
  DOP(const void* debugUnifState = Unifier::current()->state());
  Command* mainCode;
  WeightComparisonCommand* weightComparisonCode;
  bool complete;
  ulong specialisationStrength;
 try_to_compile:
  mainCode = _mainCode.memory();
  weightComparisonCode = _weightComparisonCode.memory();
  specialisationStrength = currentSpecialisationStrength();
  if (!compileGreaterOrEqual(_lhs,
			     _rhs,
			     0,
			     mainCode,
			     _mainCode.memory() + _mainCode.size() - 1,
			     weightComparisonCode,
			     _weightComparisonCode.memory() + _weightComparisonCode.size(),
			     complete,
			     specialisationStrength))  
  {
    DOP(_mainCode.unfreeze());
    DOP(_weightComparisonCode.unfreeze());
    if (_mainCode.tryToExpand() &&
	_weightComparisonCode.tryToExpand())
      {
	DOP(_mainCode.freeze());
	DOP(_weightComparisonCode.freeze());
	goto try_to_compile;  
      };  
    mainCode = _mainCode.memory();
    mainCode->setTag(Command::RetFailure);
    complete = true;
  };
  if (!complete)  mainCode->setTag(Command::RetSuccess); 
  ASSERT(Unifier::current()->state() == debugUnifState);
}; 
void StandardKBOrdering::CompiledConstraintForUnifiers::compileGreater()
{
  CALL("compileGreater()");
  ++(_ordering->_numberOfCompiledConstraintsOnUnifiers);
#ifndef NO_DEBUG
  _targetRelation = Greater;
#endif
  DOP(const void* debugUnifState = Unifier::current()->state());
  Command* mainCode;
  WeightComparisonCommand* weightComparisonCode;
  bool complete;
  ulong specialisationStrength;
 try_to_compile:
  mainCode = _mainCode.memory();
  weightComparisonCode = _weightComparisonCode.memory();
  specialisationStrength = currentSpecialisationStrength();
  if (!compileGreaterOrEqual(_lhs,
			     _rhs,
			     0,
			     mainCode,
			     _mainCode.memory() + _mainCode.size() - 1,
			     weightComparisonCode,
			     _weightComparisonCode.memory() + _weightComparisonCode.size(),
			     complete,
			     specialisationStrength))  
  {  
    DOP(_mainCode.unfreeze());
    DOP(_weightComparisonCode.unfreeze());
    if (_mainCode.tryToExpand() &&
	_weightComparisonCode.tryToExpand())
      {
	DOP(_mainCode.freeze());
	DOP(_weightComparisonCode.freeze());
	goto try_to_compile;  
      };  
    mainCode = _mainCode.memory();
    mainCode->setTag(Command::RetFailure);
    complete = true;
  };
  if (!complete)  mainCode->setTag(Command::RetFailure); 
  ASSERT(Unifier::current()->state() == debugUnifState);
}; 
bool
StandardKBOrdering::CompiledConstraintForUnifiers::compileGreaterOrEqual(UnifTerm* lhs,
									 UnifTerm* rhs,
									 WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
									 Command*& mainCode,
									 const Command* endOfMainCode,
									 WeightComparisonCommand*& weightComparisonCode,
									 const WeightComparisonCommand* endOfWeightComparisonCode,
									 bool& complete,
									 ulong& specialisationStrength)
{
  CALL("compileGreaterOrEqual(UnifTerm* lhs,...)");
  ASSERT(mainCode <= endOfMainCode);
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  if (mainCode == endOfMainCode) return false;
  WeightComparisonCommand* c;
  bool result = true;
  complete = false;
  Command* lastGoto;
  bool lexComplete;
  ulong specialisationStrength1;
  Unifier::Variable* lhsVar = 0;
  UnifTerm* lhsComplexTerm = 0;
  Unifier::Variable* rhsVar = 0;
  UnifTerm* rhsComplexTerm = 0;
  if (lhs->symbol().isVariable())
    {
      lhsVar = lhs->symbol().var();
      bool freeVar;
      void* instance = lhsVar->Unref(freeVar);
      if (freeVar)
	{
	  lhsVar = static_cast<Unifier::Variable*>(instance);
	}
      else
	{
	  lhsComplexTerm = static_cast<UnifTerm*>(instance);
	};
    }
  else
    lhsComplexTerm = lhs;
  if (rhs->symbol().isVariable())
    {
      rhsVar = rhs->symbol().var();
      bool freeVar;
      void* instance = rhsVar->Unref(freeVar);
      if (freeVar)
	{
	  rhsVar = static_cast<Unifier::Variable*>(instance);
	}
      else
	{
	  rhsComplexTerm = static_cast<UnifTerm*>(instance);
	};
    }
  else
    rhsComplexTerm = rhs;
  ASSERT(mainCode < endOfMainCode);
  if (lhsVar)
    {     
      if (rhsVar)
	{
	  if (lhsVar != rhsVar)
	    {
	      mainCode->setTag(Command::CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	      mainCode->setVarX(lhsVar);
	      mainCode->setVarY(rhsVar);  
	      ++mainCode;  
	    };
	}
      else 
	{
	  if (UnifTerm::Occurs(lhsVar,rhsComplexTerm))
	    {
	      mainCode->setTag(Command::RetFailure);
	      complete = true;
	    }
	  else
	    {
	      mainCode->setTag(Command::CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	      mainCode->setVarX(lhsVar);
	      mainCode->setTermT(rhsComplexTerm);	
	    };  
	  ++mainCode;  
	};
      return true;
    }
  else
    {
      if (rhsVar)
	{
	  if (UnifTerm::Occurs(rhsVar,lhsComplexTerm))
	    {
	      mainCode->setTag(Command::RetSuccess);
	      complete = true;
	    }
	  else
	    {
	      mainCode->setTag(Command::CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	      mainCode->setTermS(lhsComplexTerm);
	      mainCode->setVarY(rhsVar);	  
	    };
	  ++mainCode;  
	  return true;
	};
    };
  ASSERT(!lhsVar);
  ASSERT(!rhsVar);
  UnifTerm::CorrectBracket(lhsComplexTerm);
  UnifTerm::CorrectBracket(rhsComplexTerm);
  if (!specialisationStrength)
    {
      mainCode->setTag(Command::CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail);
      mainCode->setTermS(lhsComplexTerm);
      mainCode->setTermT(rhsComplexTerm);  
      ++mainCode;  
      return true;
    };
  --specialisationStrength;
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    weightComparison(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    simplifiedWeightComparison(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    spareWeightComparison1(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle
    spareWeightComparison2(_ordering->specialisedWeightComparisonPool());
  WeightComparisonCommand* initWCCodePosition;
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,lhsComplexTerm);
  TermWeighting::current()->collectWeight(_weight2,rhsComplexTerm);
  weightComparison->compare(_weight1,_weight2);
  switch (weightComparison->category())
    {
    case BK::FailureToCompare:
    case BK::AlwaysLess:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetFailure);
      complete = true;
      ++mainCode;    
      goto finish;
    case BK::AlwaysEqual:
      result = 
	compileGreaterOrEqualLex(lhsComplexTerm,
				 rhsComplexTerm,
				 assumedWeightRelations,
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 complete,
				 specialisationStrength);  	
      goto finish;
    case BK::AlwaysGreater:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetSuccess);
      complete = true;
      ++mainCode;  
      goto finish;
    case BK::CanBeLessOrEqual: 
      break; 
    case BK::CanBeGreaterOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile: 
      if (WPSCExtras::isAlwaysGreaterUnderAssumptions(*weightComparison,
						      *assumedWeightRelations,
						      simplifiedWeightComparison, 
						      spareWeightComparison1,
						      spareWeightComparison2))
	{	  
	  ASSERT(mainCode < endOfMainCode);
	  mainCode->setTag(Command::RetSuccess);
	  complete = true;
	  ++mainCode;  
	  goto finish;
	};
      break;
    }; 
  switch (weightComparison->category())
    {	      
    case BK::CanBeLessOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile:
      if (WPSCExtras::isAlwaysLessUnderAssumptions(*weightComparison,
						   *assumedWeightRelations,
						   simplifiedWeightComparison, 
						   spareWeightComparison1,
						   spareWeightComparison2))
	{	  
	  ASSERT(mainCode < endOfMainCode);
	  mainCode->setTag(Command::RetFailure);
	  complete = true;
	  ++mainCode;  
	  goto finish;
	};
      break;
    default: 
      break;
    }; 
  if (WPSCExtras::simplifyByEqualities(*weightComparison,
				       *assumedWeightRelations,
				       simplifiedWeightComparison,
				       spareWeightComparison1,
				       spareWeightComparison2))
    {
      weightComparison.swapWith(simplifiedWeightComparison);
    };
  switch (weightComparison->category())
    {
    case BK::FailureToCompare: 
    case BK::AlwaysLess:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetFailure);
      complete = true;
      ++mainCode;    
      goto finish;
    case BK::AlwaysEqual:
      COP("COP200");
      result = 
	compileGreaterOrEqualLex(lhsComplexTerm,
				 rhsComplexTerm,
				 assumedWeightRelations,
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 complete,
				 specialisationStrength);  	
      goto finish;
    case BK::AlwaysGreater:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetSuccess);
      complete = true;
      ++mainCode;  
      goto finish;
    case BK::CanBeLessOrEqual: 
    case BK::CanBeGreaterOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile: 
      goto compile_weight_comparison;
    };
 compile_weight_comparison:
  initWCCodePosition = weightComparisonCode;
  if (endOfWeightComparisonCode - weightComparisonCode < 2)
    {
      result = false;
      goto finish;
    };  
  ASSERT(weightComparisonCode < endOfWeightComparisonCode);
  if (!WPSCExtras::compileGreaterOrEqual(*weightComparison,
					 weightComparisonCode,
					 endOfWeightComparisonCode - weightComparisonCode))
    {
      result = false;
      goto finish;
    };
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  for (c = initWCCodePosition;
       c < weightComparisonCode;
       ++c)
    if (c->requiresVar())
      {
	c->setNativeVar(static_cast<const void*>(Unifier::current()->variable(c->var())));
      };
  ASSERT(mainCode < endOfMainCode);
  if (weightComparison->category() == BK::CanBeLessOrEqual)
    {
      mainCode->setTag(Command::IfWCFailRetFail);
    }
  else
    {
      mainCode->setTag(Command::IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL);
    };
  mainCode->setWeightComparisonSubroutine(initWCCodePosition);
  lastGoto = mainCode;
  ++mainCode;  
  weightComparison->assumeRelation(BK::AlwaysEqual);
  weightComparison->setNext(assumedWeightRelations);
  if (weightComparison->category() == BK::CanBeLessOrEqual)
    {
      specialisationStrength1 = specialisationStrength;
      specialisationStrength = 0UL; 
    }
  else
    {
      specialisationStrength1 = specialisationStrength/2;
      specialisationStrength -= specialisationStrength1;
    };
  result = 
    compileGreaterOrEqualLex(lhsComplexTerm,
			     rhsComplexTerm,
			     weightComparison.object(),
			     mainCode,
			     endOfMainCode,
			     weightComparisonCode,
			     endOfWeightComparisonCode,
			     lexComplete,
			     specialisationStrength1);
  specialisationStrength += specialisationStrength1;
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  if (!result) goto finish;
  if (weightComparison->category() != BK::CanBeLessOrEqual)
    {
      ASSERT(lastGoto->requiresLabel());
      lastGoto->setLabelL(mainCode);
      weightComparison->assumeRelation(BK::CanBeGreaterOrEqual);
      weightComparison->setNext(assumedWeightRelations);
      result = 
	compileGreaterOrEqualLex(lhsComplexTerm,
				 rhsComplexTerm,
				 weightComparison.object(),
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 lexComplete,
				 specialisationStrength); 
    };
 finish:
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  return result;
}; 
bool 
StandardKBOrdering::CompiledConstraintForUnifiers::compileGreaterOrEqualLex(UnifTerm* lhsComplexTerm,
									    UnifTerm* rhsComplexTerm,
									    WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
									    Command*& mainCode,
									    const Command* endOfMainCode,
									    WeightComparisonCommand*& weightComparisonCode,
									    const WeightComparisonCommand* endOfWeightComparisonCode,
									    bool& complete,
									    ulong& specialisationStrength)
{
  CALL("compileGreaterOrEqualLex(..)");
  ASSERT(mainCode <= endOfMainCode);
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  ASSERT(lhsComplexTerm->symbol().IsFunctor());
  ASSERT(rhsComplexTerm->symbol().IsFunctor());
  if (lhsComplexTerm->symbol().Func() == rhsComplexTerm->symbol().Func())
    {
      if (lhsComplexTerm->symbol().Func().arity())
	{
	  ObjectPool<UnifTerm::Iter>::EntryHandle lhsIter(_ordering->_prefixIterPool);
	  ObjectPool<UnifTerm::Iter>::EntryHandle rhsIter(_ordering->_prefixIterPool);
	  lhsIter->Reset(lhsComplexTerm);
	  rhsIter->Reset(rhsComplexTerm);
	  lhsIter->SkipSym();
	  rhsIter->SkipSym();
	  Unifier::Variable*** unifierState = Unifier::current()->MarkState();
	  do 
	    {
	      UnifTerm::CorrectBracket(lhsIter->CurrSubterm());
	      UnifTerm::CorrectBracket(rhsIter->CurrSubterm());
	      if (!compileGreaterOrEqual(lhsIter->CurrSubterm(),
					 rhsIter->CurrSubterm(),
					 assumedWeightRelations,
					 mainCode,
					 endOfMainCode,
					 weightComparisonCode,
					 endOfWeightComparisonCode,
					 complete,
					 specialisationStrength)) 
		{
		  Unifier::current()->UnbindDownto(unifierState);
		  return false;
		};
	      ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
	      if (complete)
		{
		  Unifier::current()->UnbindDownto(unifierState);
		  return true;
		};
	      if (!Unification::UnifyArbitrary(lhsIter->CurrSubterm(),
					       rhsIter->CurrSubterm()))
		{
		  if (mainCode == endOfMainCode) 
		    {
		      Unifier::current()->UnbindDownto(unifierState);
		      return false;
		    };
		  mainCode->setTag(Command::RetFailure);
		  complete = true;
		  ++mainCode; 
		  Unifier::current()->UnbindDownto(unifierState);   
		  return true;
		};
	      lhsIter->SkipTerm();
	      rhsIter->SkipTerm();
	    }
	  while (lhsIter->NotEmpty());
	  ASSERT(rhsIter->Empty());
	  Unifier::current()->UnbindDownto(unifierState);
	};
    }
  else 
    {
      if (mainCode == endOfMainCode) return false;
      if (lhsComplexTerm->symbol().Func().hasGreaterPrecedenceThan(rhsComplexTerm->symbol().Func()))
	{
	  mainCode->setTag(Command::RetSuccess);
	}
      else
	{
	  mainCode->setTag(Command::RetFailure);
	};
      complete = true;
      ++mainCode;	
    };
  return true;
}; 
bool StandardKBOrdering::CompiledConstraintForUnifiers::holdsOnCurrentUnifier()
{
  CALL("holdsOnCurrentUnifier()");
  ++(_ordering->_numberOfConstraintChecksOnUnifiers);
  Command* command = _mainCode.memory();
  FunctionComparison weightComp;
  Comparison cmp;
  bool freeVar1;
  void* instance1;
  bool freeVar2;
  void* instance2;
 next_command:
  switch (command->tag())
    {
    case Command::RetSuccess:
      goto return_true; 
    case Command::RetFailure:
      goto return_false;
    case Command::CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compareOnCurrentUnifier(command->termS(),
					   command->termT());
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      instance2 = command->varY()->Unref(freeVar2);
      if (freeVar2)
	{
	  if (UnifTerm::Occurs(static_cast<Unifier::Variable*>(instance2),command->termS()))
	    {
	      goto return_true;
	    }
	  else
	    goto return_false;
	}
      else
	{	  
	  cmp =
	    _ordering->compareOnCurrentUnifier(command->termS(),
					       static_cast<UnifTerm*>(instance2));
	  if (cmp == BK::Greater) goto return_true;
	  if (cmp != BK::Equal) goto return_false;  
	  ++command;
	  goto next_command; 
	};
      ICP("ICP10");
    case Command::CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      instance1 = command->varX()->Unref(freeVar1);
      if (freeVar1)
	{
	  goto return_false;
	}
      else
	{
	  cmp =
	    _ordering->compareOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
					       command->termT());
	  if (cmp == BK::Greater) goto return_true;
	  if (cmp != BK::Equal) goto return_false;  
	  ++command;
	  goto next_command; 
	};
      ICP("ICP20");
    case Command::CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      instance1 = command->varX()->Unref(freeVar1);
      instance2 = command->varY()->Unref(freeVar2);
      if (freeVar1)
	{
	  if (instance1 != instance2) goto return_false;
	}
      else
	if (freeVar2)
	  {
	    if (UnifTerm::Occurs(static_cast<Unifier::Variable*>(instance2),
				   static_cast<UnifTerm*>(instance1)))
	      {
		goto return_true;
	      }
	    else
	      goto return_false;
	  }
	else
	  {
	    cmp =
	    _ordering->compareOnCurrentUnifier(static_cast<UnifTerm*>(instance1),
					       static_cast<UnifTerm*>(instance2));
	    if (cmp == BK::Greater) goto return_true;
	    if (cmp != BK::Equal) goto return_false;  	    
	  };
      ++command;
      goto next_command; 
    case Command::IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL:      
      weightComp = 
	interpretWeightComparisonCode(command->weightComparisonSubroutine());
      if (weightComp == BK::FailureToCompare) goto return_false;
      if (weightComp == BK::AlwaysGreater) goto return_true;
      ASSERT((weightComp == BK::CanBeGreaterOrEqual) || 
	     (weightComp == BK::AlwaysEqual));
      if (weightComp == BK::CanBeGreaterOrEqual) 
	{
	  command = command->labelL();
	}
      else
	++command;
      goto next_command;
    case Command::IfWCFailRetFail:
      weightComp = 
	interpretWeightComparisonCode(command->weightComparisonSubroutine());
      if (weightComp == BK::FailureToCompare) goto return_false;
      ASSERT(weightComp == BK::AlwaysEqual);
      ++command;
      goto next_command;
#ifdef DEBUG_NAMESPACE
    default: ICP("ICP50"); goto return_false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default: goto return_false;
#endif
#endif
    };
 return_true:
  ASSERT((_targetRelation != GreaterOrEqual) || (SimplificationOrdering::current()->greaterOrEqualOnCurrentUnifier(_lhs,_rhs)));
  ASSERT((_targetRelation != Greater) || (SimplificationOrdering::current()->greaterOnCurrentUnifier(_lhs,_rhs)));
  return true;
 return_false:
  ASSERT((_targetRelation != GreaterOrEqual) || (!SimplificationOrdering::current()->greaterOrEqualOnCurrentUnifier(_lhs,_rhs)));  
  ASSERT((_targetRelation != Greater) || (!SimplificationOrdering::current()->greaterOnCurrentUnifier(_lhs,_rhs)));
  return false;
}; 
#ifndef NO_DEBUG_VIS
ostream& StandardKBOrdering::CompiledConstraintForUnifiers::output(ostream& str) const
{
  UnifTerm::outputTerm(str << "CONSTRAINT: ",_lhs);
  switch (_targetRelation)
    {
    case GreaterOrEqual: str << " >= "; break;
    default: str << "???";
    };
  UnifTerm::outputTerm(str,_rhs) << "\n";
  cout << "TOP LEVEL WEIGHT COMPARISON:\n" << _topLevelWeightComparison << "\n";
  str << "*** CODE: **********************\n";
  const Command* mainCodeCommand = _mainCode.memory();
  ulong gotosPassed = 0UL;
 another_branch:
  while (!mainCodeCommand->isTerminal())
    {
      mainCodeCommand->output(str,_mainCode.memory());
      if (mainCodeCommand->requiresLabel())
	++gotosPassed;
      ++mainCodeCommand;
    };
   mainCodeCommand->output(str,_mainCode.memory());
   if (gotosPassed)
     {
       --gotosPassed;
       ++mainCodeCommand;
       goto another_branch;
     };
  str << "*********** END OF CODE ********\n";
  return str;
};
#endif
inline 
Comparison 
StandardKBOrdering::CompiledConstraintForUnifiers::compareLexicographically(UnifTerm* term1,UnifTerm* term2)
{
  CALL("compareLexicographically(UnifTerm* term1,UnifTerm* term2)");
  ASSERT(term1->symbol().IsFunctor());
  ASSERT(term2->symbol().IsFunctor());
  ASSERT(term1->symbol() == term2->symbol()); 
  return _ordering->compareLexOnCurrentUnifier(term1,term2);
}; 
inline
FunctionComparison 
StandardKBOrdering::CompiledConstraintForUnifiers::interpretWeightComparisonCode(WeightComparisonCommand* subroutine)
{
  CALL("interpretWeightComparisonCode(WeightComparisonCommand* subroutine)");
  const WeightComparisonCommand* command = subroutine;
 next_command:
  switch (command->tag())
    {
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetFailureToCompare: return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetAlwaysLess: ICP("ICP10"); return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetAlwaysEqual: return BK::AlwaysEqual;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetAlwaysGreater: return BK::AlwaysGreater;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetCanBeLessOrEqual: ICP("ICP20"); return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetCanBeGreaterOrEqual: return BK::CanBeGreaterOrEqual;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetVolatileButNearlyAlwaysLess: ICP("ICP30");  return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetVolatileButNearlyAlwaysGreater: ICP("ICP40"); return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::RetAbsolutelyVolatile: ICP("ICP50"); return BK::FailureToCompare;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetFail_IfGrndRetAlwEq_RetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar) return BK::FailureToCompare; 
	long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),1);
	if (w <= 0) return BK::FailureToCompare;
	return BK::AlwaysEqual;
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetFail_IfNotGrndRetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar) return BK::FailureToCompare; 
	long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),1);
	if (w <= 0) return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetAlwGr_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar) return BK::CanBeGreaterOrEqual;
	bool ground;
	long w = TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
											       1,
											       ground);
	if (w < 0) return BK::AlwaysGreater;
	ASSERT(w == 1);
	if (ground) return BK::AlwaysEqual;
	return BK::CanBeGreaterOrEqual;
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetAlwGr_GrndToGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    long w = TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
												   1,
												   ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetAlwGr_UpdateGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    long w = TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
												   1,
												   ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGr1RetAlwGr_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    bool ground;
	    long w = TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
												   1,
												   ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _grnd = _grnd && ground;
	    if (_grnd) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	    
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
											     command->weight());
	    if (w <= 0) return BK::FailureToCompare;
	    if (w < command->weight()) 
	      {
		return BK::AlwaysGreater;
	      }
	    else
	      {
		ASSERT(w == command->weight());
		return BK::AlwaysEqual;
	      };
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	    
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
											     static_cast<const UnifTerm*>(instance),
											     command->weight());
	    if (w <= 0) return BK::FailureToCompare;
	    if (w < command->weight()) 
	      {
		return BK::AlwaysGreater;
	      }
	    else
	      {
		ASSERT(w == command->weight());
		return BK::AlwaysEqual;
	      };
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_WMinusXToACC_IfACCLe0RetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
											     command->weight());
	    if (w <= 0) return BK::FailureToCompare;
	    _acc = command->weight() - w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
											     static_cast<const UnifTerm*>(instance),
											     command->weight());
	    if (w <= 0) return BK::FailureToCompare;
	    _acc = command->weight() - w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    ++_acc;
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
											     _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    _acc += command->coefficient();
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
											     static_cast<const UnifTerm*>(instance),
											     _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    ++_acc;
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
											     _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    if (_acc == w) return BK::AlwaysEqual;
	    return BK::AlwaysGreater;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	
	    _acc += command->coefficient(); 
	    long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
											     static_cast<const UnifTerm*>(instance),
											     _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    if (_acc == w) return BK::AlwaysEqual;
	    return BK::AlwaysGreater;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	ASSERT(command->weight() > 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (1 < command->weight()) return BK::FailureToCompare;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    ASSERT((w > 0) && (w <= command->weight()));
	    if (w < command->weight()) return BK::FailureToCompare;
	    if (ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	ASSERT(command->weight() > 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (command->coefficient() < command->weight()) return BK::FailureToCompare;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  static_cast<const UnifTerm*>(instance),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    ASSERT((w > 0) && (w <= command->weight()));
	    if (w < command->weight()) return BK::FailureToCompare;
	    if (ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::WMinusXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _acc = command->weight() - 1;
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc = command->weight() - w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _acc = command->weight() - command->coefficient();
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  static_cast<const UnifTerm*>(instance),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc = command->weight() - w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::ACCMinusXPlus1ToACC_IfACCLe0RetAlwGr_UpdateGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    ++_acc;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::ACCMinusCXPlusCToACC_IfACCLe0RetAlwGr_UpdateGRND:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    _acc += command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfXGrACCPlus1RetAlwGr_IfXLeACCPlus1RetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    if (_acc > 0) return BK::FailureToCompare;  
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    ++_acc;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    if (w < _acc) return BK::FailureToCompare; 
	    ASSERT(w == _acc);
	    if (_grnd && ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    if (_acc > 0) return BK::FailureToCompare;  
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    _acc += command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    if (w < _acc) return BK::FailureToCompare; 
	    ASSERT(w == _acc);
	    if (_grnd && ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	    if (_posinst.isConstant())
	      {
		if (_posinst.freeMember() < 0) return BK::FailureToCompare;  
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	    if (_posinst.isConstant())
	      {
		if (_posinst.freeMember() < 0) return BK::FailureToCompare;  
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XPlusWToPOSINST:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXPlusWToPOSINST:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusXToPOSINST:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusCXToPOSINST:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.isConstant())
	  {
	    if (_posinst.freeMember() < 0) return BK::FailureToCompare; 
	    command = command->label();
	    goto next_command;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {	    
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.isConstant())
	  {
	    if (_posinst.freeMember() < 0) return BK::FailureToCompare; 
	    command = command->label();
	    goto next_command;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XMinus1ToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.reset();
	_neginst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.reset();
	_neginst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.reset();
	_neginst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.reset();
	_neginst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusXMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusCXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusXMinus1ToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_IfXMinus1GrPOSINSTRetFail_IfXMinus1LePOSINSTRetAlwGr_RetAlwEq:
      {
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {
	    _posinst.add1();
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
									    _posinst.freeMember());
	    if (w <= 0) return BK::FailureToCompare;
	    ASSERT(w <= _posinst.freeMember());
	    if (w < _posinst.freeMember()) return BK::AlwaysGreater;
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq:
      {
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {
	    _posinst.add(command->coefficient());
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									    static_cast<const UnifTerm*>(instance),
									    _posinst.freeMember());
	    if (w <= 0) return BK::FailureToCompare;
	    ASSERT(w <= _posinst.freeMember());
	    if (w < _posinst.freeMember()) return BK::AlwaysGreater;
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_POSINSTPlus1MinusXToACC_IfACCLe0RetFail:
      {	
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	    
	    _acc = _posinst.freeMember() + 1;
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(static_cast<const UnifTerm*>(instance),
									    _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::IfNotGrndRetFail_POSINSTPlusCMinusCXToACC_IfACCLe0RetFail:
      {	
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	    
	    _acc = _posinst.freeMember() + command->coefficient();
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									    static_cast<const UnifTerm*>(instance),
									    _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XPlusWToNEGINST:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXPlusWToNEGINST:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusXToNEGINST:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusCXToNEGINST:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,static_cast<const UnifTerm*>(instance));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_neginst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::XMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.reset();
	_posinst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::CXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.reset();
	_posinst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusXMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusCXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusXMinus1ToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_posinst.add(-1);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add1((static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_posinst.add(-(command->coefficient()));
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Unifier::Variable*>(instance))->absoluteVarNum());
	  }
	else
	  {
	    TermWeighting::current()->collectWeight(_posinst,command->coefficient(),static_cast<const UnifTerm*>(instance));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlus1MinusXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	ASSERT(_neginst.isConstant());
	ASSERT(_neginst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _acc = _neginst.freeMember();
	    _grnd = false;
	  }
	else
	  {
	    _acc = _neginst.freeMember() + 1;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForUnifiers::WeightComparisonCommand::NEGINSTPlusCMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	ASSERT(_neginst.isConstant());
	ASSERT(_neginst.freeMember() >= 0);
	bool freeVar;
	const void* instance = (static_cast<const Unifier::Variable*>(command->nativeVar()))->Unref(freeVar);
	if (freeVar)
	  {
	    _acc = _neginst.freeMember();
	    _grnd = false;
	  }
	else
	  {
	    _acc = _neginst.freeMember() + command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  static_cast<const UnifTerm*>(instance),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    _grnd = ground;
	  };
      };
      break;
#ifdef DEBUG_NAMESPACE
    default: ICP("Bad instruction tag");
#endif
    };
  ASSERT(!command->isTerminal());
  ++command;
  goto next_command;
  ICP("END");
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForUnifiers::Command"
#endif
#include "debugMacros.hpp"
inline
bool 
StandardKBOrdering::CompiledConstraintForUnifiers::Command::isTerminal() const
{
  CALL("isTerminal() const");
  switch (tag())
    {
    case RetSuccess:
    case RetFailure:
      return true;
    default: return false;
    };
}; 
inline
bool 
StandardKBOrdering::CompiledConstraintForUnifiers::Command::requiresLabel() const
{
  CALL("requiresLabel() const");
  return _tag == IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL;
}; 
#ifndef NO_DEBUG_VIS
ostream& StandardKBOrdering::CompiledConstraintForUnifiers::Command::output(ostream& str,const Command* base) const
{
  str << "[" << (this - base) << "] ";
  switch (tag())
    {
    case RetSuccess: return str << "RetSuccess;\n";
    case RetFailure: return str << "RetFailure;\n";      
    case CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      UnifTerm::outputTerm(str << " S = ",termS()) << "\n";                 
      UnifTerm::outputTerm(str << " T = ",termT()) << "\n";                 
      str << " {\n"
	  << "   cmp = compare(S@,T@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      UnifTerm::outputTerm(str << " S = ",termS()) << "\n";    
      str << " Y = " << "X" << _varYNum << "\n";         
      str << " {\n"
	  << "   cmp = compare(S@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";        
      UnifTerm::outputTerm(str << " T = ",termT()) << "\n";                 
      str << " {\n"
	  << "   cmp = compare(X@,T@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";  
      str << " Y = " << "X" << _varYNum << "\n";        
      str << " {\n"
	  << "   cmp = compare(X@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL:
      str << "IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL\n";
      str << " L = [" << (labelL() - base) << "]\n";
      str << " {\n"
	  << "   if WC(@) = FailureToCompare return Failure;\n"
	  << "   if WC(@) = AlwaysGreater return Success;\n"
          << "   ASSERT(WC(@) = AlwaysEqual || WC(@) = CanBeGreaterOrEqual);\n"
	  << "   if WC(@) = CanBeGreaterOrEqual goto L;\n"
	  << " }\n";
      str << "       |== WEIGHT COMPARISON SUBROUTINE: ========\n";
      WeightPolynomial::SpecialisedComparison::Command::outputCode(str,"       |  ",weightComparisonSubroutine());
      str << "       |== END OF WEIGHT COMPARISON SUBROUTINE ==\n";
      return str;
    case IfWCFailRetFail:
      str << "IfWCFailRetFail\n";
      str << " {\n"
	  << "   if WC(@) = FailureToCompare return Failure;\n"
	  << "   ASSERT(WC(@) = AlwaysEqual);\n"
	  << " }\n";
      return str;
    default:
      return str << "????????????;\n";
    };
  return str;
}; 
#endif
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForInstanceRetrieval"
#endif
#include "debugMacros.hpp"
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::~CompiledConstraintForInstanceRetrieval()
{
};
void 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::loadGreater(const Term* lhs,const Term* rhs)
{
  CALL("loadGreater(const Term* lhs,const Term* rhs)");
  ASSERT(lhs->isComplex());
  ASSERT(rhs->isComplex());
  _targetRelation = Greater;
  _lhs.makeReference(const_cast<Term*>(lhs));
  _rhs.makeReference(const_cast<Term*>(rhs));
  _lhsWeight.reset();
  _rhsWeight.reset();
  TermWeighting::current()->collectWeight(_lhsWeight,_lhs);
  TermWeighting::current()->collectWeight(_rhsWeight,_rhs);
  _topLevelWeightComparison.compare(_lhsWeight,_rhsWeight);
  switch (_topLevelWeightComparison.category())
    {
    case BK::FailureToCompare: 
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::FailureToCompare); 
      break;
    case BK::AlwaysLess:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Contradiction); 
      break;
    case BK::AlwaysEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    case BK::AlwaysGreater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology); 
      break;
    case BK::CanBeLessOrEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyContradiction); 
      break;
    case BK::CanBeGreaterOrEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology); 
      break;
    case BK::VolatileButNearlyAlwaysLess:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyContradiction); 
      break;
    case BK::VolatileButNearlyAlwaysGreater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology); 
      break;
    case BK::AbsolutelyVolatile:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    };     
}; 
void 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::loadNotEqual(const Term* lhs,const Term* rhs)
{
  CALL("loadNotEqual(const Term* lhs,const Term* rhs)");
  ASSERT(lhs->isComplex());
  ASSERT(rhs->isComplex());
  _targetRelation = NotEqual;
  _lhs.makeReference(const_cast<Term*>(lhs));
  _rhs.makeReference(const_cast<Term*>(rhs));
  _lhsWeight.reset();
  _rhsWeight.reset();
  TermWeighting::current()->collectWeight(_lhsWeight,_lhs);
  TermWeighting::current()->collectWeight(_rhsWeight,_rhs);
  _topLevelWeightComparison.compare(_lhsWeight,_rhsWeight);
  switch (_topLevelWeightComparison.category())
    {
    case BK::FailureToCompare: 
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::FailureToCompare); 
      break;
    case BK::AlwaysLess:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology); 
      break;
    case BK::AlwaysEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    case BK::AlwaysGreater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Tautology); 
      break;
    case BK::CanBeLessOrEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    case BK::CanBeGreaterOrEqual:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    case BK::VolatileButNearlyAlwaysLess:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology); 
      break;
    case BK::VolatileButNearlyAlwaysGreater:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology); 
      break;
    case BK::AbsolutelyVolatile:
      setCategory(SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile); 
      break;
    };     
}; 
bool StandardKBOrdering::CompiledConstraintForInstanceRetrieval::lhsVarInstMayOrder()
{
  CALL("lhsVarInstMayOrder()");
  return !_rhsWeight.containsVariableWhichIsNotIn(_lhsWeight);
}; 
bool StandardKBOrdering::CompiledConstraintForInstanceRetrieval::rhsVarInstMayOrder()
{
  CALL("rhsVarInstMayOrder()");
  return !_lhsWeight.containsVariableWhichIsNotIn(_rhsWeight);
}; 
void 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::compile(const InstRet::Substitution* subst)
{
  CALL("compile(const InstRet::Substitution* subst)");
  ASSERT(_targetRelation == Greater || _targetRelation == NotEqual);
  ++(_ordering->_numberOfCompiledConstraintsForInstanceRetrieval);
  _subst = subst;
  if (_targetRelation == NotEqual)
    {
      return;
    };
  _assumedUnifier.reset();
  DOP(const LocalUnification::UnifierState* debugUnifState = _assumedUnifier.getUnifierState());
  Command* mainCode;
  WeightComparisonCommand* weightComparisonCode;
  bool complete;
  ulong specialisationStrength;
 try_to_compile:
  mainCode = _mainCode.memory();
  weightComparisonCode = _weightComparisonCode.memory();
  specialisationStrength = currentSpecialisationStrength();
  if (!compileGreaterOrEqual(_lhs,
			     _rhs,
			     0,
			     mainCode,
			     _mainCode.memory() + _mainCode.size() - 1,
			     weightComparisonCode,
			     _weightComparisonCode.memory() + _weightComparisonCode.size(),
			     complete,
			     specialisationStrength))  
  { 
    DOP(_mainCode.unfreeze());
    DOP(_weightComparisonCode.unfreeze());
    if (_mainCode.tryToExpand() &&
	_weightComparisonCode.tryToExpand())
      {
	DOP(_mainCode.freeze());
	DOP(_weightComparisonCode.freeze());
	goto try_to_compile;  
      };  
    mainCode = _mainCode.memory();
    mainCode->setTag(Command::RetFailure);
    complete = true;
  };
  if (!complete)  mainCode->setTag(Command::RetFailure);
  ASSERT(_assumedUnifier.getUnifierState() == debugUnifState);
}; 
bool 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::compileGreaterOrEqual(const Term& lhs,
										  const Term& rhs,
										  WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
										  Command*& mainCode,
										  const Command* endOfMainCode,
										  WeightComparisonCommand*& weightComparisonCode,
										  const WeightComparisonCommand* endOfWeightComparisonCode,
										  bool& complete,
										  ulong& specialisationStrength)
{
  CALL("compileGreaterOrEqual(..)");
  ASSERT(mainCode <= endOfMainCode);
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  if (mainCode == endOfMainCode) return false;
  WeightComparisonCommand* c;
  bool result = true;
  complete = false;
  Command* lastGoto;
  bool lexComplete;
  ulong specialisationStrength1;
  if (!specialisationStrength)
    {
      ASSERT(mainCode < endOfMainCode);
      if (lhs.isVariable())
	{
	  if (rhs.isVariable())
	    {
	      if (lhs != rhs)
		{
		  mainCode->setTag(Command::CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
		  mainCode->setVarX(_subst->nativeVar(lhs.var()));
		  mainCode->setVarY(_subst->nativeVar(rhs.var()));  
		  ++mainCode; 
		};
	    }
	  else 
	    {
	      mainCode->setTag(Command::CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	      mainCode->setVarX(_subst->nativeVar(lhs.var()));
	      mainCode->setTermT(rhs);
	      ++mainCode; 
	    };
	}
      else
	if (rhs.isVariable())
	  {
	    mainCode->setTag(Command::CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	    mainCode->setTermS(lhs);
	    mainCode->setVarY(_subst->nativeVar(rhs.var()));
	    ++mainCode; 
	  }
	else 
	  {
	    mainCode->setTag(Command::CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	    mainCode->setTermS(lhs);
	    mainCode->setTermT(rhs);
	    ++mainCode; 
	  };	  
      return true;
    }; 
  --specialisationStrength;
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    weightComparison(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    simplifiedWeightComparison(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle 
    spareWeightComparison1(_ordering->specialisedWeightComparisonPool());
  ObjectPool<WeightPolynomial::SpecialisedComparison>::EntryHandle
    spareWeightComparison2(_ordering->specialisedWeightComparisonPool());
  WeightComparisonCommand* initWCCodePosition;
  _weight1.reset();
  _weight2.reset();
  TermWeighting::current()->collectWeight(_weight1,lhs,&_assumedUnifier);
  TermWeighting::current()->collectWeight(_weight2,rhs,&_assumedUnifier);
  weightComparison->compare(_weight1,_weight2);
  switch (weightComparison->category())
    {
    case BK::FailureToCompare:
    case BK::AlwaysLess:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetFailure);
      complete = true;
      ++mainCode;    
      goto finish;
    case BK::AlwaysEqual:
      result = 
	compileGreaterOrEqualLex(lhs,
				 rhs,
				 assumedWeightRelations,
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 complete,
				 specialisationStrength);  	
      goto finish;
    case BK::AlwaysGreater:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetSuccess);
      complete = true;
      ++mainCode;  
      goto finish;
    case BK::CanBeLessOrEqual: 
      break; 
    case BK::CanBeGreaterOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile: 
      if (WPSCExtras::isAlwaysGreaterUnderAssumptions(*weightComparison,
						      *assumedWeightRelations,
						      simplifiedWeightComparison, 
						      spareWeightComparison1,
						      spareWeightComparison2))
	{	  
	  ASSERT(mainCode < endOfMainCode);
	  mainCode->setTag(Command::RetSuccess);
	  complete = true;
	  ++mainCode;  
	  goto finish;
	};
      break;
    }; 
  switch (weightComparison->category())
    {	      
    case BK::CanBeLessOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile:
      if (WPSCExtras::isAlwaysLessUnderAssumptions(*weightComparison,
						   *assumedWeightRelations,
						   simplifiedWeightComparison, 
						   spareWeightComparison1,
						   spareWeightComparison2))
	{	  
	  ASSERT(mainCode < endOfMainCode);
	  mainCode->setTag(Command::RetFailure);
	  complete = true;
	  ++mainCode;  
	  goto finish;
	};
      break;
    default: 
      break;
    }; 
  if (WPSCExtras::simplifyByEqualities(*weightComparison,
				       *assumedWeightRelations,
				       simplifiedWeightComparison,
				       spareWeightComparison1,
				       spareWeightComparison2))
    {
      weightComparison.swapWith(simplifiedWeightComparison);
    };
  switch (weightComparison->category())
    {
    case BK::FailureToCompare: 
    case BK::AlwaysLess:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetFailure);
      complete = true;
      ++mainCode;    
      goto finish;
    case BK::AlwaysEqual:
      COP("COP200");
      result = 
	compileGreaterOrEqualLex(lhs,
				 rhs,
				 assumedWeightRelations,
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 complete,
				 specialisationStrength);  	
      goto finish;
    case BK::AlwaysGreater:
      ASSERT(mainCode < endOfMainCode);
      mainCode->setTag(Command::RetSuccess);
      complete = true;
      ++mainCode;  
      goto finish;
    case BK::CanBeLessOrEqual: 
    case BK::CanBeGreaterOrEqual: 
    case BK::VolatileButNearlyAlwaysLess: 
    case BK::VolatileButNearlyAlwaysGreater: 
    case BK::AbsolutelyVolatile: 
      goto compile_weight_comparison;
    };
 compile_weight_comparison:
  initWCCodePosition = weightComparisonCode;
  if (weightComparisonCode >= endOfWeightComparisonCode - 1)
    {
      result = false;
      goto finish;
    };  
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  if (!WPSCExtras::compileGreaterOrEqual(*weightComparison,
					 weightComparisonCode,
					 endOfWeightComparisonCode - weightComparisonCode))
    {
      result = false;
      goto finish;
    };
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  for (c = initWCCodePosition;
       c < weightComparisonCode;
       ++c)
    if (c->requiresVar())
      {
	c->setNativeVar(static_cast<const void*>(_subst->nativeVar(c->var())));
      };
  ASSERT(mainCode < endOfMainCode);
  if (weightComparison->category() == BK::CanBeLessOrEqual)
    {
      mainCode->setTag(Command::IfWCFailRetFail);
    }
  else
    {
      mainCode->setTag(Command::IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL);
    };
  mainCode->setWeightComparisonSubroutine(initWCCodePosition);
  lastGoto = mainCode;
  ++mainCode;  
  weightComparison->assumeRelation(BK::AlwaysEqual);
  weightComparison->setNext(assumedWeightRelations);
  if (weightComparison->category() == BK::CanBeLessOrEqual)
    {
      specialisationStrength1 = specialisationStrength;
      specialisationStrength = 0UL; 
    }
  else
    {
      specialisationStrength1 = specialisationStrength/2;
      specialisationStrength -= specialisationStrength1;
    };
  result = 
    compileGreaterOrEqualLex(lhs,
			     rhs,
			     weightComparison.object(),
			     mainCode,
			     endOfMainCode,
			     weightComparisonCode,
			     endOfWeightComparisonCode,
			     lexComplete,
			     specialisationStrength1);
  specialisationStrength += specialisationStrength1;
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  if (!result) goto finish;
  if (weightComparison->category() != BK::CanBeLessOrEqual)
    {
      ASSERT(lastGoto->requiresLabel());
      lastGoto->setLabelL(mainCode);
      weightComparison->assumeRelation(BK::CanBeGreaterOrEqual);
      weightComparison->setNext(assumedWeightRelations);
      result = 
	compileGreaterOrEqualLex(lhs,
				 rhs,
				 weightComparison.object(),
				 mainCode,
				 endOfMainCode,
				 weightComparisonCode,
				 endOfWeightComparisonCode,
				 lexComplete,
				 specialisationStrength); 
    };
 finish:
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  return result;
}; 
bool 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::compileGreaterOrEqualLex(const Term& lhs,
										     const Term& rhs,
										     WeightPolynomial::SpecialisedComparison* assumedWeightRelations,
										     Command*& mainCode,
										     const Command* endOfMainCode,
										     WeightComparisonCommand*& weightComparisonCode,
										     const WeightComparisonCommand* endOfWeightComparisonCode,
										     bool& complete,
										     ulong& specialisationStrength)
{
  CALL("compileGreaterOrEqualLex(..)");
  ASSERT(mainCode <= endOfMainCode);
  ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
  LocalUnification::TermTraversal lhsInst;
  LocalUnification::TermTraversal rhsInst;
  complete = false;
  lhsInst.reset(lhs,&_assumedUnifier);
  rhsInst.reset(rhs,&_assumedUnifier);
  if (lhsInst.term().isVariable())
    {
      if (rhsInst.term().isVariable())
	{
	  if (lhsInst.term() == rhsInst.term())
	    {
	    }
	  else 
	    {
	      if (mainCode == endOfMainCode) return false;
	      mainCode->setTag(Command::CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	      mainCode->setVarX(_subst->nativeVar(lhsInst.term().var()));
	      mainCode->setVarY(_subst->nativeVar(rhsInst.term().var()));
#ifndef NO_DEBUG
	      mainCode->setVarXNum(lhsInst.term().var());
	      mainCode->setVarYNum(rhsInst.term().var());
#endif
	      ++mainCode;		     
	    };
	}
      else 
	{	  
	  if (mainCode == endOfMainCode) return false;
	  mainCode->setTag(Command::CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	  mainCode->setVarX(_subst->nativeVar(lhsInst.term().var()));
#ifndef NO_DEBUG
	  mainCode->setVarXNum(lhsInst.term().var());
#endif
	  mainCode->setTermT(rhsInst.term());
	  ++mainCode;		      
	};
    }
  else 
    {
      if (rhsInst.term().isVariable())
	{
	  if (mainCode == endOfMainCode) return false;
	  mainCode->setTag(Command::CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail);
	  mainCode->setTermS(lhsInst.term());
	  mainCode->setVarY(_subst->nativeVar(rhsInst.term().var()));
#ifndef NO_DEBUG
	  mainCode->setVarYNum(rhsInst.term().var());
#endif
	  ++mainCode;	 	       
	}
      else 
	{
	  if (lhsInst.symbol() == rhsInst.symbol())
	    { 
	      if (lhsInst.symbol().arity())
		{
		  const LocalUnification::UnifierState* unifierState = 
		    _assumedUnifier.getUnifierState();
		  lhsInst.next();
		  rhsInst.next();
		  COP("COP100");
		  do
		    {
		      if (!compileGreaterOrEqual(lhsInst.term(),
						 rhsInst.term(),
						 assumedWeightRelations,
						 mainCode,
						 endOfMainCode,
						 weightComparisonCode,
						 endOfWeightComparisonCode,
						 complete,
						 specialisationStrength)) 
			{
			  _assumedUnifier.rollUnifierBackTo(unifierState);
			  return false;
			};
		      COP("AFTER compileGreaterOrEqual(..)");
		      ASSERT(weightComparisonCode <= endOfWeightComparisonCode);
		      if (complete)
			{
			  _assumedUnifier.rollUnifierBackTo(unifierState);
			  return true;
			};
		      if (!_assumedUnifier.unify(lhsInst.term(),rhsInst.term()))
			{
			  if (mainCode == endOfMainCode) 
			    {
			      _assumedUnifier.rollUnifierBackTo(unifierState);
			      return false;
			    };
			  mainCode->setTag(Command::RetFailure);
			  complete = true;
			  ++mainCode;    
			  _assumedUnifier.rollUnifierBackTo(unifierState);
			  return true;
			};
		      lhsInst.after();
		      rhsInst.after();
		    }
		  while (lhsInst);
		  _assumedUnifier.rollUnifierBackTo(unifierState);
		};
	    }
	  else 
	    {
	      if (mainCode == endOfMainCode) return false;
	      if (lhsInst.symbol().hasGreaterPrecedenceThan(rhsInst.symbol()))
		{
		  mainCode->setTag(Command::RetSuccess);
		}
	      else
		{
		  mainCode->setTag(Command::RetFailure);
		};
	      complete = true;
	      ++mainCode;	 
	    };
	};
    };
  return true;
}; 
bool StandardKBOrdering::CompiledConstraintForInstanceRetrieval::holds()
{
  CALL("holds()"); 
  ++(_ordering->_numberOfConstraintChecksForInstanceRetrieval);
  ASSERT(_targetRelation == Greater || 
	 _targetRelation == NotEqual);
  if (_targetRelation == NotEqual)
    {
      bool res = 
	!InstRet::Substitution::Instance::equal(_lhs,_rhs,_subst);
      return res;
    };
  ASSERT(_targetRelation == Greater);
  Command* command = _mainCode.memory();
  FunctionComparison weightComp;
  Comparison cmp;
  Term instanceForX;
  Term instanceForY;
 next_command:
  switch (command->tag())
    {
    case Command::RetSuccess:
      goto return_true; 
    case Command::RetFailure:
      goto return_false;
    case Command::CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compare(command->termS(),
			   command->termT(),
			   _subst);
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compare(command->termS().deref(),
			   _subst,
			   *(command->varY()));
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compare(command->termT().deref(),
			   _subst,
			   *(command->varX()));
      if (cmp == BK::Less) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compare(*(command->varX()),*(command->varY()));
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp =
	_ordering->compareLexicographically(*(command->varX()),
					    *(command->varY()));
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp = 
	_ordering->compareLexicographically(command->termT().deref(),
					    _subst,
					    *(command->varX()));
      if (cmp == BK::Less) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      cmp = 
	_ordering->compareLexicographically(command->termS().deref(),
					    _subst,
					    *(command->varY()));      
      if (cmp == BK::Greater) goto return_true;
      if (cmp != BK::Equal) goto return_false;      
      ++command;
      goto next_command;
    case Command::IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL:      
      weightComp = 
	interpretWeightComparisonCode(command->weightComparisonSubroutine());
      if (weightComp == BK::FailureToCompare) goto return_false;
      if (weightComp == BK::AlwaysGreater) goto return_true;
      ASSERT((weightComp == BK::CanBeGreaterOrEqual) || 
	     (weightComp == BK::AlwaysEqual));
      if (weightComp == BK::CanBeGreaterOrEqual) 
	{
	  command = command->labelL();
	}
      else
	++command;
      goto next_command;
    case Command::IfWCFailRetFail:
      weightComp = 
	interpretWeightComparisonCode(command->weightComparisonSubroutine());
      if (weightComp == BK::FailureToCompare) goto return_false;
      ASSERT(weightComp == BK::AlwaysEqual);
      ++command;
      goto next_command;
#ifdef DEBUG_NAMESPACE
    default: ICP("ICP10"); goto return_false;
#else
#ifdef _SUPPRESS_WARNINGS_
    default: goto return_false;
#endif
#endif
    };
 return_true:
  ASSERT(_targetRelation != Greater || 
	 SimplificationOrdering::current()->greater(_lhs.deref(),_rhs.deref(),_subst));
  return true;
 return_false:
  ASSERT(_targetRelation != Greater 
	 || !SimplificationOrdering::current()->greater(_lhs.deref(),_rhs.deref(),_subst));
  return false;
}; 
#ifndef NO_DEBUG_VIS
ostream& StandardKBOrdering::CompiledConstraintForInstanceRetrieval::output(ostream& str) const
{
  str << "CONSTRAINT: " << _lhs;
  switch (_targetRelation)
    {
    case Greater: str << " > "; break;
    default: str << "???";
    };
  str << _rhs << "\n";
  cout << "TOP LEVEL WEIGHT COMPARISON:\n" << _topLevelWeightComparison << "\n";
  str << "*** CODE: **********************\n";
  const Command* mainCodeCommand = _mainCode.memory();
  ulong gotosPassed = 0UL;
 another_branch:
  while (!mainCodeCommand->isTerminal())
    {
      mainCodeCommand->output(str,_mainCode.memory());
      if (mainCodeCommand->requiresLabel())
	++gotosPassed;
      ++mainCodeCommand;
    };
   mainCodeCommand->output(str,_mainCode.memory());
   if (gotosPassed)
     {
       --gotosPassed;
       ++mainCodeCommand;
       goto another_branch;
     };
  str << "*********** END OF CODE ********\n";
  return str;
};
#endif
inline
FunctionComparison 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::interpretWeightComparisonCode(WeightComparisonCommand* subroutine)
{
  CALL("interpretWeightComparisonCode(WeightComparisonCommand* subroutine)");
  const WeightComparisonCommand* command = subroutine;
 next_command:
  switch (command->tag())
    {
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetFailureToCompare: return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetAlwaysLess: ICP("ICP10"); return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetAlwaysEqual: return BK::AlwaysEqual;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetAlwaysGreater: return BK::AlwaysGreater;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetCanBeLessOrEqual: ICP("ICP20"); return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetCanBeGreaterOrEqual: return BK::CanBeGreaterOrEqual;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetVolatileButNearlyAlwaysLess: ICP("ICP30");  return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetVolatileButNearlyAlwaysGreater: ICP("ICP40"); return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::RetAbsolutelyVolatile: ICP("ICP50"); return BK::FailureToCompare;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetFail_IfGrndRetAlwEq_RetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare; 
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									1);
	if (w <= 0) return BK::FailureToCompare;
	return BK::AlwaysEqual;
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetFail_IfNotGrndRetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare; 
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									1);
	if (w <= 0) return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetAlwGr_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::CanBeGreaterOrEqual;
	bool ground;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									      1,
									      ground);
	if (w < 0) return BK::AlwaysGreater;
	ASSERT(w == 1);
	if (ground) return BK::AlwaysEqual;
	return BK::CanBeGreaterOrEqual;
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetAlwGr_GrndToGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  1,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetAlwGr_UpdateGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _grnd = false;
	  }
	else
	  {
	    bool ground;
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  1,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGr1RetAlwGr_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::CanBeGreaterOrEqual;
	bool ground;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									      1,
									      ground);
	if (w < 0) return BK::AlwaysGreater;
	_grnd = _grnd && ground;
	if (_grnd) return BK::AlwaysEqual;
	return BK::CanBeGreaterOrEqual;
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_IfXGrWRetFail_IfXLeWRetAlwGr_RetAlwEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									command->weight());
	if (w <= 0) return BK::FailureToCompare;
	if (w < command->weight()) 
	  {
	    return BK::AlwaysGreater;
	  }
	else
	  {
	    ASSERT(w == command->weight());
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_IfCXGrWRetFail_IfCXLeWRetAlwGr_RetAlwEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									(static_cast<const Term*>(command->nativeVar()))->deref(),
									command->weight());
	if (w <= 0) return BK::FailureToCompare;
	if (w < command->weight()) 
	  {
	    return BK::AlwaysGreater;
	  }
	else
	  {
	    ASSERT(w == command->weight());
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_WMinusXToACC_IfACCLe0RetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									command->weight());
	if (w <= 0) return BK::FailureToCompare;
	_acc = command->weight() - w;
	ASSERT(_acc >= 0);
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_WMinusCXToACC_IfACCLe0RetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									(static_cast<const Term*>(command->nativeVar()))->deref(),
									command->weight());
	if (w <= 0) return BK::FailureToCompare;
	_acc = command->weight() - w;
	ASSERT(_acc >= 0);
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	++_acc;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									_acc);
	if (w <= 0) return BK::FailureToCompare;
	_acc -= w;
	ASSERT(_acc >= 0);
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	_acc += command->coefficient();
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
											 (static_cast<const Term*>(command->nativeVar()))->deref(),
											 _acc);
	if (w <= 0) return BK::FailureToCompare;
	_acc -= w;
	ASSERT(_acc >= 0);
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusXPlus1ToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	++_acc;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									_acc);
	if (w <= 0) return BK::FailureToCompare;
	if (_acc == w) return BK::AlwaysEqual;
	return BK::AlwaysGreater;
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_ACCMinusCXPlusCToACC_IfACCLe0RetFail_IfACCEq0RetAlwEq_RetAlwGr:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  return BK::FailureToCompare;
	ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	_acc += command->coefficient(); 
	long w = 
	  TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									(static_cast<const Term*>(command->nativeVar()))->deref(),
									_acc);
	if (w <= 0) return BK::FailureToCompare;
	if (_acc == w) return BK::AlwaysEqual;
	return BK::AlwaysGreater;
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGrWRetAlwGr_IfXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	ASSERT(command->weight() > 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (1 < command->weight()) return BK::FailureToCompare;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    ASSERT((w > 0) && (w <= command->weight()));
	    if (w < command->weight()) return BK::FailureToCompare;
	    if (ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfCXGrWRetAlwGr_IfCXLeWRetFail_IfGrndRetAlwEq_RetCanBeGrOrEq:
      {
	ASSERT(command->weight() > 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (command->coefficient() < command->weight()) return BK::FailureToCompare;
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  (static_cast<const Term*>(command->nativeVar()))->deref(),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    ASSERT((w > 0) && (w <= command->weight()));
	    if (w < command->weight()) return BK::FailureToCompare;
	    if (ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::WMinusXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _acc = command->weight() - 1;
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc = command->weight() - w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::WMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _acc = command->weight() - command->coefficient();
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  (static_cast<const Term*>(command->nativeVar()))->deref(),
										  command->weight(),
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc = command->weight() - w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::ACCMinusXPlus1ToACC_IfACCLe0RetAlwGr_UpdateGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    ++_acc;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::ACCMinusCXPlusCToACC_IfACCLe0RetAlwGr_UpdateGRND:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    _grnd = false;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _acc += command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  (static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	    _grnd = _grnd && ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfXGrACCPlus1RetAlwGr_IfXLeACCPlus1RetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    if (_acc > 0) return BK::FailureToCompare;  
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    ++_acc;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    if (w < _acc) return BK::FailureToCompare; 
	    ASSERT(w == _acc);
	    if (_grnd && ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfCXGrACCPlusCRetAlwGr_IfCXLeACCPlusCRetFail_UpdateGRND_IfGRNDRetAlwEq_RetCanBeGrOrEq:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    if (_acc < 0) return BK::AlwaysGreater;
	    if (_acc > 0) return BK::FailureToCompare;  
	    return BK::CanBeGreaterOrEqual;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _acc += command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  (static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    if (w < _acc) return BK::FailureToCompare; 
	    ASSERT(w == _acc);
	    if (_grnd && ground) return BK::AlwaysEqual;
	    return BK::CanBeGreaterOrEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,*(static_cast<const Term*>(command->nativeVar())));
	    if (_posinst.isConstant())
	      {
		if (_posinst.freeMember() < 0) return BK::FailureToCompare;  
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXPlusWToPOSINST_IfConstPOSINSAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {	
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	    if (_posinst.isConstant())
	      {
		if (_posinst.freeMember() < 0) return BK::FailureToCompare;  
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XPlusWToPOSINST:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXPlusWToPOSINST:
      {
	_posinst.reset();
	_posinst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {			
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusXToPOSINST:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {				
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusCXToPOSINST:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {					
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());	    
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {						
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());    
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.isConstant())
	  {
	    if (_posinst.freeMember() < 0) return BK::FailureToCompare; 
	    command = command->label();
	    goto next_command;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusCXToPOSINST_IfConstPOSINSTAndPOSINSTLe0RetFail_IfConstPOSINSTGotoL:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());      
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.isConstant())
	  {
	    if (_posinst.freeMember() < 0) return BK::FailureToCompare; 
	    command = command->label();
	    goto next_command;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XMinus1ToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.reset();
	_neginst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference()); 
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXMinusCToNEGINSTRetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.reset();
	_neginst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference()); 
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.reset();
	_neginst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference()); 
	    TermWeighting::current()->collectWeight(_neginst,*(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.reset();
	_neginst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference()); 
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusXMinus1ToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference()); 
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusCXMinusCToNEGINST_IfCompForGrOrEqPOSINSTvNEGINSTEqFailRetFail:
      {
	_neginst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::FailureToCompare) 
	  return BK::FailureToCompare;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusXMinus1ToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusCXMinusCToNEGINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_neginst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_IfXMinus1GrPOSINSTRetFail_IfXMinus1LePOSINSTRetAlwGr_RetAlwEq:
      {
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {							
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _posinst.add1();
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									    _posinst.freeMember());
	    if (w <= 0) return BK::FailureToCompare;
	    ASSERT(w <= _posinst.freeMember());
	    if (w < _posinst.freeMember()) return BK::AlwaysGreater;
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_IfCXMinusCGrPOSINSTRetFail_IfCXMinusCLePOSINSTRetAlwGr_RetAlwEq:
      {
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _posinst.add(command->coefficient());
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									    (static_cast<const Term*>(command->nativeVar()))->deref(),
									    _posinst.freeMember());
	    if (w <= 0) return BK::FailureToCompare;
	    ASSERT(w <= _posinst.freeMember());
	    if (w < _posinst.freeMember()) return BK::AlwaysGreater;
	    return BK::AlwaysEqual;
	  };
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_POSINSTPlus1MinusXToACC_IfACCLe0RetFail:
      {	
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {	 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());   
	    _acc = _posinst.freeMember() + 1;
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
									    _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::IfNotGrndRetFail_POSINSTPlusCMinusCXToACC_IfACCLe0RetFail:
      {	
	ASSERT(_posinst.isConstant());
	ASSERT(_posinst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    return BK::FailureToCompare;
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());       
	    _acc = _posinst.freeMember() + command->coefficient();
	    long w = 
	      TermWeighting::current()->computeWeightIfGroundAndLessOrEqual(command->coefficient(),
									    (static_cast<const Term*>(command->nativeVar()))->deref(),
									    _acc);
	    if (w <= 0) return BK::FailureToCompare;
	    _acc -= w;
	    ASSERT(_acc >= 0);
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());   
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXPlusWToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());  
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XPlusWToNEGINST:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXPlusWToNEGINST:
      {
	_neginst.reset();
	_neginst.add(command->weight());
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusXToNEGINST:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusCXToNEGINST:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    *(static_cast<const Term*>(command->nativeVar())));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusCXToNEGINST_IfConstNEGINSTAndNEGINSTLe0RetAlwGr_IfConstNEGINSTGotoL:
      {
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _neginst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_neginst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	    if (_neginst.isConstant())
	      {
		if (_neginst.freeMember() < 0) return BK::AlwaysGreater;
		command = command->label();
		goto next_command;
	      };
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::XMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.reset();
	_posinst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::CXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.reset();
	_posinst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusXMinus1ToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusCXMinusCToPOSINST_IfCompForGrOrEqPOSINSTvNEGINSTEqAlwGrRetAlwGr:
      {
	_posinst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	if (_posinst.compareForGreaterOrEqual(_neginst) == BK::AlwaysGreater)
	  return BK::AlwaysGreater;
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusXMinus1ToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_posinst.add(-1);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add1((static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::POSINSTPlusCXMinusCToPOSINST_RetCompForGrOrEqPOSINSTvNEGINST:
      {
	_posinst.add(-(command->coefficient()));
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _posinst.add(command->coefficient(),(static_cast<const Term*>(command->nativeVar()))->var());
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    TermWeighting::current()->collectWeight(_posinst,
						    command->coefficient(),
						    *(static_cast<const Term*>(command->nativeVar())));
	  };
	return _posinst.compareForGreaterOrEqual(_neginst);
      };
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlus1MinusXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	ASSERT(_neginst.isConstant());
	ASSERT(_neginst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _acc = _neginst.freeMember();
	    _grnd = false;
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _acc = _neginst.freeMember() + 1;
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual((static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    _grnd = ground;
	  };
      };
      break;
    case CompiledConstraintForInstanceRetrieval::WeightComparisonCommand::NEGINSTPlusCMinusCXToACC_IfACCLe0RetAlwGr_GrndToGRND:
      {
	ASSERT(_neginst.isConstant());
	ASSERT(_neginst.freeMember() >= 0);
	if ((static_cast<const Term*>(command->nativeVar()))->isVariable())
	  {
	    _acc = _neginst.freeMember();
	    _grnd = false;
	  }
	else
	  {		 
	    ASSERT((static_cast<const Term*>(command->nativeVar()))->IsReference());
	    _acc = _neginst.freeMember() + command->coefficient();
	    bool ground;
	    long w = 
	      TermWeighting::current()->computeMinimalInstanceWeightIfLessOrEqual(command->coefficient(),
										  (static_cast<const Term*>(command->nativeVar()))->deref(),
										  _acc,
										  ground);
	    if (w < 0) return BK::AlwaysGreater;
	    _acc -= w;
	    _grnd = ground;
	  };
      };
      break;
#ifdef DEBUG_NAMESPACE
    default: ICP("Bad instruction tag");
#endif
    };
  ASSERT(!command->isTerminal());
  ++command;
  goto next_command;
  ICP("END");
}; 
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_KB_ORDERING
#define DEBUG_NAMESPACE "StandardKBOrdering::CompiledConstraintForInstanceRetrieval::Command"
#endif
#include "debugMacros.hpp"
inline
bool 
StandardKBOrdering::CompiledConstraintForInstanceRetrieval::Command::isTerminal() const
{
  CALL("isTerminal() const");
  switch (tag())
    {
    case RetSuccess:
    case RetFailure:
      return true;
    default: return false;
    };
}; 
#ifndef NO_DEBUG_VIS
ostream& StandardKBOrdering::CompiledConstraintForInstanceRetrieval::Command::output(ostream& str,const Command* base) const
{
  str << "[" << (this - base) << "] ";
  switch (tag())
    {
    case RetSuccess: return str << "RetSuccess;\n";
    case RetFailure: return str << "RetFailure;\n";      
    case CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " S = " << termS() << "\n";                 
      str << " T = " << termT() << "\n";                 
      str << " {\n"
	  << "   cmp = compare(S@,T@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " S = " << termS() << "\n";    
      str << " Y = " << "X" << _varYNum << "\n";         
      str << " {\n"
	  << "   cmp = compare(S@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";        
      str << " T = " << termT() << "\n";                 
      str << " {\n"
	  << "   cmp = compare(X@,T@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";  
      str << " Y = " << "X" << _varYNum << "\n";        
      str << " {\n"
	  << "   cmp = compare(X@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;     
    case CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";   
      str << " Y = " << "X" << _varYNum << "\n";         
      str << " {\n"
	  << "   cmp = compareLex(X@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;
    case CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";  
      str << " T = " << termT() << "\n";           
      str << " {\n"
	  << "   cmp = compareLex(X@,T@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;
    case CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " S = " << termS() << "\n";     
      str << " Y = " << "X" << _varYNum << "\n";         
      str << " {\n"
	  << "   cmp = compareLex(S@,Y@);\n"
          << "   if (cmp == Greater) return Success;\n"
	  << "   if (cmp != AlwaysEqual) return Failure;\n"
	  << " }\n";
      return str;
    case IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL:
      str << "IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL\n";
      str << " L = [" << (labelL() - base) << "]\n";
      str << " {\n"
	  << "   if WC(@) = FailureToCompare return Failure;\n"
	  << "   if WC(@) = AlwaysGreater return Success;\n"
          << "   ASSERT(WC(@) = AlwaysEqual || WC(@) = CanBeGreaterOrEqual);\n"
	  << "   if WC(@) = CanBeGreaterOrEqual goto L;\n"
	  << " }\n";
      str << "       |== WEIGHT COMPARISON SUBROUTINE: ========\n";
      WeightPolynomial::SpecialisedComparison::Command::outputCode(str,"       |  ",weightComparisonSubroutine());
      str << "       |== END OF WEIGHT COMPARISON SUBROUTINE ==\n";
      return str;
    case IfWCFailRetFail:
      str << "IfWCFailRetFail\n";
      str << " {\n"
	  << "   if WC(@) = FailureToCompare return Failure;\n"
	  << "   ASSERT(WC(@) = AlwaysEqual);\n"
	  << " }\n";
      return str;
    default:
      return str << "????????????;\n";
    };
  return str;
}; 
ostream& StandardKBOrdering::CompiledConstraintForInstanceRetrieval::Command::outputShort(ostream& str,const Command* base) const
{
  switch (tag())
    {
    case RetSuccess: return str << "RetSuccess\n";
    case RetFailure: return str << "RetFailure\n";
    case CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareST_IfCmpGrRetSucc_IfCmpNotEqRetFail\n"; 
      str << " S = " << termS() << "\n";   
      str << " T = " << termT() << "\n";
      return str;
    case CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareSY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n"; 
      str << " S = " << termS() << "\n"; 
      str << " Y = " << "X" << _varYNum << "\n"; 
      return str;
    case CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXT_IfCmpGrRetSucc_IfCmpNotEqRetFail\n"; 
      str << " X = " << "X" << _varXNum << "\n";   
      str << " T = " << termT() << "\n";
      return str;
    case CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareXY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n"; 
      str << " X = " << "X" << _varXNum << "\n";   
      str << " Y = " << "X" << _varYNum << "\n"; 
      return str;
    case CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexXY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";   
      str << " Y = " << "X" << _varYNum << "\n"; 
      return str;
    case CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexXT_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " X = " << "X" << _varXNum << "\n";  
      str << " T = " << termT() << "\n";
      return str;
    case CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail:
      str << "CompareLexSY_IfCmpGrRetSucc_IfCmpNotEqRetFail\n";
      str << " S = " << termS() << "\n";     
      str << " Y = " << "X" << _varYNum << "\n";
      return str;
    case IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL:
      str << "IfWCFailRetFail_IfWCAlwGrRetSucc_IfWCCanBeGrOrEqGotoL\n";
      str << " L = [" << (labelL() - base) << "]\n";
      return str;
    case IfWCFailRetFail:
      str << "IfWCFailRetFail\n";
      return str;
    default: return str << "????????????;\n";
    };
  return str;
}; 
#endif
